/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 3130
#define LARGE_STATE_COUNT 762
#define SYMBOL_COUNT 231
#define ALIAS_COUNT 0
#define TOKEN_COUNT 105
#define EXTERNAL_TOKEN_COUNT 7
#define FIELD_COUNT 30
#define MAX_ALIAS_SEQUENCE_LENGTH 8
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 79
#define SUPERTYPE_COUNT 4

enum ts_symbol_identifiers {
  sym_identifier = 1,
  anon_sym_SEMI = 2,
  anon_sym_EQ = 3,
  anon_sym_return = 4,
  anon_sym_async = 5,
  anon_sym_def = 6,
  anon_sym_DASH_GT = 7,
  anon_sym_COLON = 8,
  anon_sym_PLUS_EQ = 9,
  anon_sym_DASH_EQ = 10,
  anon_sym_STAR_EQ = 11,
  anon_sym_SLASH_EQ = 12,
  anon_sym_SLASH_SLASH_EQ = 13,
  anon_sym_PERCENT_EQ = 14,
  anon_sym_STAR_STAR_EQ = 15,
  anon_sym_GT_GT_EQ = 16,
  anon_sym_LT_LT_EQ = 17,
  anon_sym_GT_GT_GT_EQ = 18,
  anon_sym_AMP_EQ = 19,
  anon_sym_CARET_EQ = 20,
  anon_sym_PIPE_EQ = 21,
  anon_sym_COMMA = 22,
  anon_sym_LPAREN = 23,
  anon_sym_RPAREN = 24,
  anon_sym_LBRACK = 25,
  anon_sym_RBRACK = 26,
  anon_sym_del = 27,
  anon_sym_raise = 28,
  sym_pass_statement = 29,
  sym_break_statement = 30,
  sym_continue_statement = 31,
  sym_debugger_statement = 32,
  anon_sym_assert = 33,
  anon_sym_global = 34,
  anon_sym_nonlocal = 35,
  anon_sym_yield = 36,
  anon_sym_from = 37,
  anon_sym_import = 38,
  anon_sym_as = 39,
  anon_sym___python__ = 40,
  anon_sym_DOT = 41,
  anon_sym_if = 42,
  anon_sym_elif = 43,
  anon_sym_else = 44,
  anon_sym_for = 45,
  anon_sym_in = 46,
  anon_sym_while = 47,
  anon_sym_do = 48,
  anon_sym_try = 49,
  anon_sym_except = 50,
  anon_sym_finally = 51,
  anon_sym_with = 52,
  anon_sym_STAR = 53,
  anon_sym_STAR_STAR = 54,
  anon_sym_class = 55,
  anon_sym_AT = 56,
  anon_sym_not = 57,
  anon_sym_and = 58,
  anon_sym_or = 59,
  anon_sym_PLUS = 60,
  anon_sym_DASH = 61,
  anon_sym_SLASH = 62,
  anon_sym_SLASH_SLASH = 63,
  anon_sym_PERCENT = 64,
  anon_sym_PIPE = 65,
  anon_sym_CARET = 66,
  anon_sym_AMP = 67,
  anon_sym_LT_LT = 68,
  anon_sym_GT_GT = 69,
  anon_sym_GT_GT_GT = 70,
  anon_sym_TILDE = 71,
  anon_sym_typeof = 72,
  anon_sym_void = 73,
  anon_sym_LT = 74,
  anon_sym_LT_EQ = 75,
  anon_sym_EQ_EQ = 76,
  anon_sym_EQ_EQ_EQ = 77,
  anon_sym_BANG_EQ = 78,
  anon_sym_BANG_EQ_EQ = 79,
  anon_sym_GT_EQ = 80,
  anon_sym_GT = 81,
  anon_sym_is = 82,
  anon_sym_instanceof = 83,
  anon_sym_new = 84,
  anon_sym_QMARK = 85,
  anon_sym_await = 86,
  anon_sym_LBRACE = 87,
  anon_sym_RBRACE = 88,
  aux_sym_interpolation_token1 = 89,
  sym_this = 90,
  sym_true = 91,
  sym_false = 92,
  sym_none = 93,
  sym_string = 94,
  sym_verbatim = 95,
  sym_number = 96,
  sym_comment = 97,
  sym__newline = 98,
  sym__indent = 99,
  sym__dedent = 100,
  sym_regex = 101,
  sym_fstring_start = 102,
  sym_fstring_content = 103,
  sym_fstring_end = 104,
  sym_module = 105,
  sym__statement = 106,
  sym__simple_statements = 107,
  sym__simple_statement = 108,
  sym__compound_statement = 109,
  sym_block_tailed_statement = 110,
  sym__block_anonymous_function = 111,
  sym__expression_statement = 112,
  sym_assignment = 113,
  sym_augmented_assignment = 114,
  sym__assign_lhs = 115,
  sym__assign_rhs = 116,
  sym_expression_list = 117,
  sym_pattern_list = 118,
  sym_pattern = 119,
  sym_tuple_pattern = 120,
  sym_list_pattern = 121,
  sym__pattern_seq = 122,
  sym_return_statement = 123,
  sym_delete_statement = 124,
  sym_raise_statement = 125,
  sym_assert_statement = 126,
  sym_global_statement = 127,
  sym_nonlocal_statement = 128,
  sym_yield_statement = 129,
  sym_yield = 130,
  sym_import_statement = 131,
  sym_aliased_import = 132,
  sym_import_from_statement = 133,
  sym__import_list = 134,
  sym_aliased_import_name = 135,
  sym_scoped_flag_statement = 136,
  sym__flag_list = 137,
  sym_dotted_name = 138,
  sym_if_statement = 139,
  sym_elif_clause = 140,
  sym_else_clause = 141,
  sym_for_statement = 142,
  sym__for_target = 143,
  sym_for_js_statement = 144,
  sym_while_statement = 145,
  sym_do_statement = 146,
  sym_try_statement = 147,
  sym_except_clause = 148,
  sym_finally_clause = 149,
  sym_with_statement = 150,
  sym_with_clause = 151,
  sym_function_definition = 152,
  sym_async_function_definition = 153,
  sym_parameters = 154,
  sym__parameters = 155,
  sym__parameter = 156,
  sym_typed_parameter = 157,
  sym_default_parameter = 158,
  sym_typed_default_parameter = 159,
  sym_list_splat_parameter = 160,
  sym_dictionary_splat_parameter = 161,
  sym_class_definition = 162,
  sym_decorated_definition = 163,
  sym_decorator = 164,
  sym__suite = 165,
  sym_block = 166,
  sym__expression = 167,
  sym__primary_expression = 168,
  sym_not_operator = 169,
  sym_boolean_operator = 170,
  sym_binary_operator = 171,
  sym_unary_operator = 172,
  sym_comparison_operator = 173,
  sym_conditional_expression = 174,
  sym_attribute = 175,
  sym_subscript = 176,
  sym_slice = 177,
  sym_call = 178,
  sym_slice_call = 179,
  sym_argument_list = 180,
  sym__argument = 181,
  sym_keyword_argument = 182,
  sym_list_splat_argument = 183,
  sym_dictionary_splat_argument = 184,
  sym_new_expression = 185,
  sym_existential = 186,
  sym_await = 187,
  sym_anonymous_function = 188,
  sym__anon_suite = 189,
  sym__inline_body = 190,
  sym__inline_simple_statement = 191,
  sym_list = 192,
  sym__collection_element = 193,
  sym_set = 194,
  sym_tuple = 195,
  sym_parenthesized_expression = 196,
  sym_dictionary = 197,
  sym_pair = 198,
  sym_list_comprehension = 199,
  sym_set_comprehension = 200,
  sym_dictionary_comprehension = 201,
  sym_generator_expression = 202,
  sym__comprehension_clauses = 203,
  sym_for_in_clause = 204,
  sym_if_clause = 205,
  sym_concatenated_string = 206,
  sym_f_string = 207,
  sym_interpolation = 208,
  aux_sym_module_repeat1 = 209,
  aux_sym__simple_statements_repeat1 = 210,
  aux_sym_expression_list_repeat1 = 211,
  aux_sym_pattern_list_repeat1 = 212,
  aux_sym_global_statement_repeat1 = 213,
  aux_sym_import_statement_repeat1 = 214,
  aux_sym__import_list_repeat1 = 215,
  aux_sym__flag_list_repeat1 = 216,
  aux_sym_dotted_name_repeat1 = 217,
  aux_sym_if_statement_repeat1 = 218,
  aux_sym_try_statement_repeat1 = 219,
  aux_sym_with_statement_repeat1 = 220,
  aux_sym__parameters_repeat1 = 221,
  aux_sym_decorated_definition_repeat1 = 222,
  aux_sym_comparison_operator_repeat1 = 223,
  aux_sym_argument_list_repeat1 = 224,
  aux_sym__inline_body_repeat1 = 225,
  aux_sym_list_repeat1 = 226,
  aux_sym_dictionary_repeat1 = 227,
  aux_sym__comprehension_clauses_repeat1 = 228,
  aux_sym_concatenated_string_repeat1 = 229,
  aux_sym_f_string_repeat1 = 230,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [anon_sym_SEMI] = ";",
  [anon_sym_EQ] = "=",
  [anon_sym_return] = "return",
  [anon_sym_async] = "async",
  [anon_sym_def] = "def",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_COLON] = ":",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_SLASH_SLASH_EQ] = "//=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_STAR_STAR_EQ] = "**=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_GT_EQ] = ">>>=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_COMMA] = ",",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_del] = "del",
  [anon_sym_raise] = "raise",
  [sym_pass_statement] = "pass_statement",
  [sym_break_statement] = "break_statement",
  [sym_continue_statement] = "continue_statement",
  [sym_debugger_statement] = "debugger_statement",
  [anon_sym_assert] = "assert",
  [anon_sym_global] = "global",
  [anon_sym_nonlocal] = "nonlocal",
  [anon_sym_yield] = "yield",
  [anon_sym_from] = "from",
  [anon_sym_import] = "import",
  [anon_sym_as] = "as",
  [anon_sym___python__] = "__python__",
  [anon_sym_DOT] = ".",
  [anon_sym_if] = "if",
  [anon_sym_elif] = "elif",
  [anon_sym_else] = "else",
  [anon_sym_for] = "for",
  [anon_sym_in] = "in",
  [anon_sym_while] = "while",
  [anon_sym_do] = "do",
  [anon_sym_try] = "try",
  [anon_sym_except] = "except",
  [anon_sym_finally] = "finally",
  [anon_sym_with] = "with",
  [anon_sym_STAR] = "*",
  [anon_sym_STAR_STAR] = "**",
  [anon_sym_class] = "class",
  [anon_sym_AT] = "@",
  [anon_sym_not] = "not",
  [anon_sym_and] = "and",
  [anon_sym_or] = "or",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_SLASH] = "/",
  [anon_sym_SLASH_SLASH] = "//",
  [anon_sym_PERCENT] = "%",
  [anon_sym_PIPE] = "|",
  [anon_sym_CARET] = "^",
  [anon_sym_AMP] = "&",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_GT_GT_GT] = ">>>",
  [anon_sym_TILDE] = "~",
  [anon_sym_typeof] = "typeof",
  [anon_sym_void] = "void",
  [anon_sym_LT] = "<",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_EQ_EQ_EQ] = "===",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_BANG_EQ_EQ] = "!==",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_GT] = ">",
  [anon_sym_is] = "is",
  [anon_sym_instanceof] = "instanceof",
  [anon_sym_new] = "new",
  [anon_sym_QMARK] = "\?",
  [anon_sym_await] = "await",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [aux_sym_interpolation_token1] = "interpolation_token1",
  [sym_this] = "this",
  [sym_true] = "true",
  [sym_false] = "false",
  [sym_none] = "none",
  [sym_string] = "string",
  [sym_verbatim] = "verbatim",
  [sym_number] = "number",
  [sym_comment] = "comment",
  [sym__newline] = "_newline",
  [sym__indent] = "_indent",
  [sym__dedent] = "_dedent",
  [sym_regex] = "regex",
  [sym_fstring_start] = "fstring_start",
  [sym_fstring_content] = "fstring_content",
  [sym_fstring_end] = "fstring_end",
  [sym_module] = "module",
  [sym__statement] = "_statement",
  [sym__simple_statements] = "_simple_statements",
  [sym__simple_statement] = "_simple_statement",
  [sym__compound_statement] = "_compound_statement",
  [sym_block_tailed_statement] = "block_tailed_statement",
  [sym__block_anonymous_function] = "anonymous_function",
  [sym__expression_statement] = "_expression_statement",
  [sym_assignment] = "assignment",
  [sym_augmented_assignment] = "augmented_assignment",
  [sym__assign_lhs] = "_assign_lhs",
  [sym__assign_rhs] = "_assign_rhs",
  [sym_expression_list] = "expression_list",
  [sym_pattern_list] = "pattern_list",
  [sym_pattern] = "pattern",
  [sym_tuple_pattern] = "tuple_pattern",
  [sym_list_pattern] = "list_pattern",
  [sym__pattern_seq] = "_pattern_seq",
  [sym_return_statement] = "return_statement",
  [sym_delete_statement] = "delete_statement",
  [sym_raise_statement] = "raise_statement",
  [sym_assert_statement] = "assert_statement",
  [sym_global_statement] = "global_statement",
  [sym_nonlocal_statement] = "nonlocal_statement",
  [sym_yield_statement] = "yield_statement",
  [sym_yield] = "yield",
  [sym_import_statement] = "import_statement",
  [sym_aliased_import] = "aliased_import",
  [sym_import_from_statement] = "import_from_statement",
  [sym__import_list] = "_import_list",
  [sym_aliased_import_name] = "aliased_import_name",
  [sym_scoped_flag_statement] = "scoped_flag_statement",
  [sym__flag_list] = "_flag_list",
  [sym_dotted_name] = "dotted_name",
  [sym_if_statement] = "if_statement",
  [sym_elif_clause] = "elif_clause",
  [sym_else_clause] = "else_clause",
  [sym_for_statement] = "for_statement",
  [sym__for_target] = "_for_target",
  [sym_for_js_statement] = "for_js_statement",
  [sym_while_statement] = "while_statement",
  [sym_do_statement] = "do_statement",
  [sym_try_statement] = "try_statement",
  [sym_except_clause] = "except_clause",
  [sym_finally_clause] = "finally_clause",
  [sym_with_statement] = "with_statement",
  [sym_with_clause] = "with_clause",
  [sym_function_definition] = "function_definition",
  [sym_async_function_definition] = "async_function_definition",
  [sym_parameters] = "parameters",
  [sym__parameters] = "_parameters",
  [sym__parameter] = "_parameter",
  [sym_typed_parameter] = "typed_parameter",
  [sym_default_parameter] = "default_parameter",
  [sym_typed_default_parameter] = "typed_default_parameter",
  [sym_list_splat_parameter] = "list_splat_parameter",
  [sym_dictionary_splat_parameter] = "dictionary_splat_parameter",
  [sym_class_definition] = "class_definition",
  [sym_decorated_definition] = "decorated_definition",
  [sym_decorator] = "decorator",
  [sym__suite] = "_suite",
  [sym_block] = "block",
  [sym__expression] = "_expression",
  [sym__primary_expression] = "_primary_expression",
  [sym_not_operator] = "not_operator",
  [sym_boolean_operator] = "boolean_operator",
  [sym_binary_operator] = "binary_operator",
  [sym_unary_operator] = "unary_operator",
  [sym_comparison_operator] = "comparison_operator",
  [sym_conditional_expression] = "conditional_expression",
  [sym_attribute] = "attribute",
  [sym_subscript] = "subscript",
  [sym_slice] = "slice",
  [sym_call] = "call",
  [sym_slice_call] = "slice_call",
  [sym_argument_list] = "argument_list",
  [sym__argument] = "_argument",
  [sym_keyword_argument] = "keyword_argument",
  [sym_list_splat_argument] = "list_splat_argument",
  [sym_dictionary_splat_argument] = "dictionary_splat_argument",
  [sym_new_expression] = "new_expression",
  [sym_existential] = "existential",
  [sym_await] = "await",
  [sym_anonymous_function] = "anonymous_function",
  [sym__anon_suite] = "_anon_suite",
  [sym__inline_body] = "block",
  [sym__inline_simple_statement] = "_inline_simple_statement",
  [sym_list] = "list",
  [sym__collection_element] = "_collection_element",
  [sym_set] = "set",
  [sym_tuple] = "tuple",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_dictionary] = "dictionary",
  [sym_pair] = "pair",
  [sym_list_comprehension] = "list_comprehension",
  [sym_set_comprehension] = "set_comprehension",
  [sym_dictionary_comprehension] = "dictionary_comprehension",
  [sym_generator_expression] = "generator_expression",
  [sym__comprehension_clauses] = "_comprehension_clauses",
  [sym_for_in_clause] = "for_in_clause",
  [sym_if_clause] = "if_clause",
  [sym_concatenated_string] = "concatenated_string",
  [sym_f_string] = "f_string",
  [sym_interpolation] = "interpolation",
  [aux_sym_module_repeat1] = "module_repeat1",
  [aux_sym__simple_statements_repeat1] = "_simple_statements_repeat1",
  [aux_sym_expression_list_repeat1] = "expression_list_repeat1",
  [aux_sym_pattern_list_repeat1] = "pattern_list_repeat1",
  [aux_sym_global_statement_repeat1] = "global_statement_repeat1",
  [aux_sym_import_statement_repeat1] = "import_statement_repeat1",
  [aux_sym__import_list_repeat1] = "_import_list_repeat1",
  [aux_sym__flag_list_repeat1] = "_flag_list_repeat1",
  [aux_sym_dotted_name_repeat1] = "dotted_name_repeat1",
  [aux_sym_if_statement_repeat1] = "if_statement_repeat1",
  [aux_sym_try_statement_repeat1] = "try_statement_repeat1",
  [aux_sym_with_statement_repeat1] = "with_statement_repeat1",
  [aux_sym__parameters_repeat1] = "_parameters_repeat1",
  [aux_sym_decorated_definition_repeat1] = "decorated_definition_repeat1",
  [aux_sym_comparison_operator_repeat1] = "comparison_operator_repeat1",
  [aux_sym_argument_list_repeat1] = "argument_list_repeat1",
  [aux_sym__inline_body_repeat1] = "_inline_body_repeat1",
  [aux_sym_list_repeat1] = "list_repeat1",
  [aux_sym_dictionary_repeat1] = "dictionary_repeat1",
  [aux_sym__comprehension_clauses_repeat1] = "_comprehension_clauses_repeat1",
  [aux_sym_concatenated_string_repeat1] = "concatenated_string_repeat1",
  [aux_sym_f_string_repeat1] = "f_string_repeat1",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_return] = anon_sym_return,
  [anon_sym_async] = anon_sym_async,
  [anon_sym_def] = anon_sym_def,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_SLASH_SLASH_EQ] = anon_sym_SLASH_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_STAR_STAR_EQ] = anon_sym_STAR_STAR_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_GT_EQ] = anon_sym_GT_GT_GT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_del] = anon_sym_del,
  [anon_sym_raise] = anon_sym_raise,
  [sym_pass_statement] = sym_pass_statement,
  [sym_break_statement] = sym_break_statement,
  [sym_continue_statement] = sym_continue_statement,
  [sym_debugger_statement] = sym_debugger_statement,
  [anon_sym_assert] = anon_sym_assert,
  [anon_sym_global] = anon_sym_global,
  [anon_sym_nonlocal] = anon_sym_nonlocal,
  [anon_sym_yield] = anon_sym_yield,
  [anon_sym_from] = anon_sym_from,
  [anon_sym_import] = anon_sym_import,
  [anon_sym_as] = anon_sym_as,
  [anon_sym___python__] = anon_sym___python__,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_elif] = anon_sym_elif,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_try] = anon_sym_try,
  [anon_sym_except] = anon_sym_except,
  [anon_sym_finally] = anon_sym_finally,
  [anon_sym_with] = anon_sym_with,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_STAR_STAR] = anon_sym_STAR_STAR,
  [anon_sym_class] = anon_sym_class,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_not] = anon_sym_not,
  [anon_sym_and] = anon_sym_and,
  [anon_sym_or] = anon_sym_or,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_SLASH_SLASH] = anon_sym_SLASH_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_GT_GT_GT] = anon_sym_GT_GT_GT,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_typeof] = anon_sym_typeof,
  [anon_sym_void] = anon_sym_void,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_EQ_EQ_EQ] = anon_sym_EQ_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_BANG_EQ_EQ] = anon_sym_BANG_EQ_EQ,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_is] = anon_sym_is,
  [anon_sym_instanceof] = anon_sym_instanceof,
  [anon_sym_new] = anon_sym_new,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_await] = anon_sym_await,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [aux_sym_interpolation_token1] = aux_sym_interpolation_token1,
  [sym_this] = sym_this,
  [sym_true] = sym_true,
  [sym_false] = sym_false,
  [sym_none] = sym_none,
  [sym_string] = sym_string,
  [sym_verbatim] = sym_verbatim,
  [sym_number] = sym_number,
  [sym_comment] = sym_comment,
  [sym__newline] = sym__newline,
  [sym__indent] = sym__indent,
  [sym__dedent] = sym__dedent,
  [sym_regex] = sym_regex,
  [sym_fstring_start] = sym_fstring_start,
  [sym_fstring_content] = sym_fstring_content,
  [sym_fstring_end] = sym_fstring_end,
  [sym_module] = sym_module,
  [sym__statement] = sym__statement,
  [sym__simple_statements] = sym__simple_statements,
  [sym__simple_statement] = sym__simple_statement,
  [sym__compound_statement] = sym__compound_statement,
  [sym_block_tailed_statement] = sym_block_tailed_statement,
  [sym__block_anonymous_function] = sym_anonymous_function,
  [sym__expression_statement] = sym__expression_statement,
  [sym_assignment] = sym_assignment,
  [sym_augmented_assignment] = sym_augmented_assignment,
  [sym__assign_lhs] = sym__assign_lhs,
  [sym__assign_rhs] = sym__assign_rhs,
  [sym_expression_list] = sym_expression_list,
  [sym_pattern_list] = sym_pattern_list,
  [sym_pattern] = sym_pattern,
  [sym_tuple_pattern] = sym_tuple_pattern,
  [sym_list_pattern] = sym_list_pattern,
  [sym__pattern_seq] = sym__pattern_seq,
  [sym_return_statement] = sym_return_statement,
  [sym_delete_statement] = sym_delete_statement,
  [sym_raise_statement] = sym_raise_statement,
  [sym_assert_statement] = sym_assert_statement,
  [sym_global_statement] = sym_global_statement,
  [sym_nonlocal_statement] = sym_nonlocal_statement,
  [sym_yield_statement] = sym_yield_statement,
  [sym_yield] = sym_yield,
  [sym_import_statement] = sym_import_statement,
  [sym_aliased_import] = sym_aliased_import,
  [sym_import_from_statement] = sym_import_from_statement,
  [sym__import_list] = sym__import_list,
  [sym_aliased_import_name] = sym_aliased_import_name,
  [sym_scoped_flag_statement] = sym_scoped_flag_statement,
  [sym__flag_list] = sym__flag_list,
  [sym_dotted_name] = sym_dotted_name,
  [sym_if_statement] = sym_if_statement,
  [sym_elif_clause] = sym_elif_clause,
  [sym_else_clause] = sym_else_clause,
  [sym_for_statement] = sym_for_statement,
  [sym__for_target] = sym__for_target,
  [sym_for_js_statement] = sym_for_js_statement,
  [sym_while_statement] = sym_while_statement,
  [sym_do_statement] = sym_do_statement,
  [sym_try_statement] = sym_try_statement,
  [sym_except_clause] = sym_except_clause,
  [sym_finally_clause] = sym_finally_clause,
  [sym_with_statement] = sym_with_statement,
  [sym_with_clause] = sym_with_clause,
  [sym_function_definition] = sym_function_definition,
  [sym_async_function_definition] = sym_async_function_definition,
  [sym_parameters] = sym_parameters,
  [sym__parameters] = sym__parameters,
  [sym__parameter] = sym__parameter,
  [sym_typed_parameter] = sym_typed_parameter,
  [sym_default_parameter] = sym_default_parameter,
  [sym_typed_default_parameter] = sym_typed_default_parameter,
  [sym_list_splat_parameter] = sym_list_splat_parameter,
  [sym_dictionary_splat_parameter] = sym_dictionary_splat_parameter,
  [sym_class_definition] = sym_class_definition,
  [sym_decorated_definition] = sym_decorated_definition,
  [sym_decorator] = sym_decorator,
  [sym__suite] = sym__suite,
  [sym_block] = sym_block,
  [sym__expression] = sym__expression,
  [sym__primary_expression] = sym__primary_expression,
  [sym_not_operator] = sym_not_operator,
  [sym_boolean_operator] = sym_boolean_operator,
  [sym_binary_operator] = sym_binary_operator,
  [sym_unary_operator] = sym_unary_operator,
  [sym_comparison_operator] = sym_comparison_operator,
  [sym_conditional_expression] = sym_conditional_expression,
  [sym_attribute] = sym_attribute,
  [sym_subscript] = sym_subscript,
  [sym_slice] = sym_slice,
  [sym_call] = sym_call,
  [sym_slice_call] = sym_slice_call,
  [sym_argument_list] = sym_argument_list,
  [sym__argument] = sym__argument,
  [sym_keyword_argument] = sym_keyword_argument,
  [sym_list_splat_argument] = sym_list_splat_argument,
  [sym_dictionary_splat_argument] = sym_dictionary_splat_argument,
  [sym_new_expression] = sym_new_expression,
  [sym_existential] = sym_existential,
  [sym_await] = sym_await,
  [sym_anonymous_function] = sym_anonymous_function,
  [sym__anon_suite] = sym__anon_suite,
  [sym__inline_body] = sym_block,
  [sym__inline_simple_statement] = sym__inline_simple_statement,
  [sym_list] = sym_list,
  [sym__collection_element] = sym__collection_element,
  [sym_set] = sym_set,
  [sym_tuple] = sym_tuple,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_dictionary] = sym_dictionary,
  [sym_pair] = sym_pair,
  [sym_list_comprehension] = sym_list_comprehension,
  [sym_set_comprehension] = sym_set_comprehension,
  [sym_dictionary_comprehension] = sym_dictionary_comprehension,
  [sym_generator_expression] = sym_generator_expression,
  [sym__comprehension_clauses] = sym__comprehension_clauses,
  [sym_for_in_clause] = sym_for_in_clause,
  [sym_if_clause] = sym_if_clause,
  [sym_concatenated_string] = sym_concatenated_string,
  [sym_f_string] = sym_f_string,
  [sym_interpolation] = sym_interpolation,
  [aux_sym_module_repeat1] = aux_sym_module_repeat1,
  [aux_sym__simple_statements_repeat1] = aux_sym__simple_statements_repeat1,
  [aux_sym_expression_list_repeat1] = aux_sym_expression_list_repeat1,
  [aux_sym_pattern_list_repeat1] = aux_sym_pattern_list_repeat1,
  [aux_sym_global_statement_repeat1] = aux_sym_global_statement_repeat1,
  [aux_sym_import_statement_repeat1] = aux_sym_import_statement_repeat1,
  [aux_sym__import_list_repeat1] = aux_sym__import_list_repeat1,
  [aux_sym__flag_list_repeat1] = aux_sym__flag_list_repeat1,
  [aux_sym_dotted_name_repeat1] = aux_sym_dotted_name_repeat1,
  [aux_sym_if_statement_repeat1] = aux_sym_if_statement_repeat1,
  [aux_sym_try_statement_repeat1] = aux_sym_try_statement_repeat1,
  [aux_sym_with_statement_repeat1] = aux_sym_with_statement_repeat1,
  [aux_sym__parameters_repeat1] = aux_sym__parameters_repeat1,
  [aux_sym_decorated_definition_repeat1] = aux_sym_decorated_definition_repeat1,
  [aux_sym_comparison_operator_repeat1] = aux_sym_comparison_operator_repeat1,
  [aux_sym_argument_list_repeat1] = aux_sym_argument_list_repeat1,
  [aux_sym__inline_body_repeat1] = aux_sym__inline_body_repeat1,
  [aux_sym_list_repeat1] = aux_sym_list_repeat1,
  [aux_sym_dictionary_repeat1] = aux_sym_dictionary_repeat1,
  [aux_sym__comprehension_clauses_repeat1] = aux_sym__comprehension_clauses_repeat1,
  [aux_sym_concatenated_string_repeat1] = aux_sym_concatenated_string_repeat1,
  [aux_sym_f_string_repeat1] = aux_sym_f_string_repeat1,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_async] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_def] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_del] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_raise] = {
    .visible = true,
    .named = false,
  },
  [sym_pass_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_break_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_continue_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_debugger_statement] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_assert] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_global] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nonlocal] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_yield] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_from] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_import] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [anon_sym___python__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_elif] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_except] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_finally] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_with] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_not] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_and] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_or] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_typeof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_void] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_is] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_instanceof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_new] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_await] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_interpolation_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_this] = {
    .visible = true,
    .named = true,
  },
  [sym_true] = {
    .visible = true,
    .named = true,
  },
  [sym_false] = {
    .visible = true,
    .named = true,
  },
  [sym_none] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_verbatim] = {
    .visible = true,
    .named = true,
  },
  [sym_number] = {
    .visible = true,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__newline] = {
    .visible = false,
    .named = true,
  },
  [sym__indent] = {
    .visible = false,
    .named = true,
  },
  [sym__dedent] = {
    .visible = false,
    .named = true,
  },
  [sym_regex] = {
    .visible = true,
    .named = true,
  },
  [sym_fstring_start] = {
    .visible = true,
    .named = true,
  },
  [sym_fstring_content] = {
    .visible = true,
    .named = true,
  },
  [sym_fstring_end] = {
    .visible = true,
    .named = true,
  },
  [sym_module] = {
    .visible = true,
    .named = true,
  },
  [sym__statement] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_statements] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_statement] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__compound_statement] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_block_tailed_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__block_anonymous_function] = {
    .visible = true,
    .named = true,
  },
  [sym__expression_statement] = {
    .visible = false,
    .named = true,
  },
  [sym_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_augmented_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym__assign_lhs] = {
    .visible = false,
    .named = true,
  },
  [sym__assign_rhs] = {
    .visible = false,
    .named = true,
  },
  [sym_expression_list] = {
    .visible = true,
    .named = true,
  },
  [sym_pattern_list] = {
    .visible = true,
    .named = true,
  },
  [sym_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_list_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__pattern_seq] = {
    .visible = false,
    .named = true,
  },
  [sym_return_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_delete_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_raise_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_assert_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_global_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_nonlocal_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_yield_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_yield] = {
    .visible = true,
    .named = true,
  },
  [sym_import_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_aliased_import] = {
    .visible = true,
    .named = true,
  },
  [sym_import_from_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__import_list] = {
    .visible = false,
    .named = true,
  },
  [sym_aliased_import_name] = {
    .visible = true,
    .named = true,
  },
  [sym_scoped_flag_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__flag_list] = {
    .visible = false,
    .named = true,
  },
  [sym_dotted_name] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_elif_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__for_target] = {
    .visible = false,
    .named = true,
  },
  [sym_for_js_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_while_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_do_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_try_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_except_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_finally_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_with_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_with_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_async_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym__parameters] = {
    .visible = false,
    .named = true,
  },
  [sym__parameter] = {
    .visible = false,
    .named = true,
  },
  [sym_typed_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_default_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_default_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_list_splat_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_dictionary_splat_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_class_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_decorated_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_decorator] = {
    .visible = true,
    .named = true,
  },
  [sym__suite] = {
    .visible = false,
    .named = true,
  },
  [sym_block] = {
    .visible = true,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__primary_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_not_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_comparison_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_subscript] = {
    .visible = true,
    .named = true,
  },
  [sym_slice] = {
    .visible = true,
    .named = true,
  },
  [sym_call] = {
    .visible = true,
    .named = true,
  },
  [sym_slice_call] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym__argument] = {
    .visible = false,
    .named = true,
  },
  [sym_keyword_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_list_splat_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_dictionary_splat_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_new_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_existential] = {
    .visible = true,
    .named = true,
  },
  [sym_await] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_function] = {
    .visible = true,
    .named = true,
  },
  [sym__anon_suite] = {
    .visible = false,
    .named = true,
  },
  [sym__inline_body] = {
    .visible = true,
    .named = true,
  },
  [sym__inline_simple_statement] = {
    .visible = false,
    .named = true,
  },
  [sym_list] = {
    .visible = true,
    .named = true,
  },
  [sym__collection_element] = {
    .visible = false,
    .named = true,
  },
  [sym_set] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_dictionary] = {
    .visible = true,
    .named = true,
  },
  [sym_pair] = {
    .visible = true,
    .named = true,
  },
  [sym_list_comprehension] = {
    .visible = true,
    .named = true,
  },
  [sym_set_comprehension] = {
    .visible = true,
    .named = true,
  },
  [sym_dictionary_comprehension] = {
    .visible = true,
    .named = true,
  },
  [sym_generator_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__comprehension_clauses] = {
    .visible = false,
    .named = true,
  },
  [sym_for_in_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_if_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_concatenated_string] = {
    .visible = true,
    .named = true,
  },
  [sym_f_string] = {
    .visible = true,
    .named = true,
  },
  [sym_interpolation] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_module_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__simple_statements_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expression_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pattern_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_global_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_import_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__import_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__flag_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_dotted_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_try_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_with_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_decorated_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_comparison_operator_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__inline_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_dictionary_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__comprehension_clauses_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_concatenated_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_f_string_repeat1] = {
    .visible = false,
    .named = false,
  },
};

enum ts_field_identifiers {
  field_alias = 1,
  field_alternative = 2,
  field_argument = 3,
  field_arguments = 4,
  field_attribute = 5,
  field_body = 6,
  field_condition = 7,
  field_consequence = 8,
  field_constructor = 9,
  field_default = 10,
  field_definition = 11,
  field_expression = 12,
  field_flag = 13,
  field_format_spec = 14,
  field_from = 15,
  field_function = 16,
  field_key = 17,
  field_left = 18,
  field_module_name = 19,
  field_name = 20,
  field_object = 21,
  field_operator = 22,
  field_parameters = 23,
  field_return_type = 24,
  field_right = 25,
  field_subscript = 26,
  field_superclasses = 27,
  field_type = 28,
  field_type_conversion = 29,
  field_value = 30,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_arguments] = "arguments",
  [field_attribute] = "attribute",
  [field_body] = "body",
  [field_condition] = "condition",
  [field_consequence] = "consequence",
  [field_constructor] = "constructor",
  [field_default] = "default",
  [field_definition] = "definition",
  [field_expression] = "expression",
  [field_flag] = "flag",
  [field_format_spec] = "format_spec",
  [field_from] = "from",
  [field_function] = "function",
  [field_key] = "key",
  [field_left] = "left",
  [field_module_name] = "module_name",
  [field_name] = "name",
  [field_object] = "object",
  [field_operator] = "operator",
  [field_parameters] = "parameters",
  [field_return_type] = "return_type",
  [field_right] = "right",
  [field_subscript] = "subscript",
  [field_superclasses] = "superclasses",
  [field_type] = "type",
  [field_type_conversion] = "type_conversion",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 4},
  [2] = {.index = 4, .length = 4},
  [3] = {.index = 8, .length = 1},
  [4] = {.index = 9, .length = 1},
  [5] = {.index = 10, .length = 1},
  [6] = {.index = 11, .length = 2},
  [7] = {.index = 13, .length = 1},
  [8] = {.index = 14, .length = 1},
  [9] = {.index = 15, .length = 1},
  [10] = {.index = 16, .length = 2},
  [11] = {.index = 18, .length = 1},
  [12] = {.index = 19, .length = 1},
  [13] = {.index = 20, .length = 2},
  [15] = {.index = 22, .length = 5},
  [16] = {.index = 27, .length = 2},
  [17] = {.index = 29, .length = 3},
  [18] = {.index = 32, .length = 1},
  [19] = {.index = 33, .length = 2},
  [20] = {.index = 35, .length = 2},
  [21] = {.index = 37, .length = 2},
  [22] = {.index = 39, .length = 2},
  [23] = {.index = 41, .length = 1},
  [24] = {.index = 42, .length = 2},
  [25] = {.index = 44, .length = 1},
  [26] = {.index = 45, .length = 1},
  [27] = {.index = 46, .length = 1},
  [28] = {.index = 47, .length = 1},
  [29] = {.index = 48, .length = 2},
  [30] = {.index = 50, .length = 2},
  [31] = {.index = 52, .length = 2},
  [32] = {.index = 54, .length = 2},
  [33] = {.index = 56, .length = 2},
  [34] = {.index = 58, .length = 1},
  [35] = {.index = 59, .length = 1},
  [36] = {.index = 60, .length = 2},
  [37] = {.index = 62, .length = 2},
  [38] = {.index = 64, .length = 2},
  [39] = {.index = 66, .length = 2},
  [40] = {.index = 68, .length = 2},
  [41] = {.index = 70, .length = 1},
  [42] = {.index = 71, .length = 2},
  [43] = {.index = 73, .length = 2},
  [44] = {.index = 75, .length = 2},
  [45] = {.index = 77, .length = 2},
  [46] = {.index = 79, .length = 2},
  [47] = {.index = 81, .length = 3},
  [48] = {.index = 84, .length = 2},
  [49] = {.index = 86, .length = 2},
  [50] = {.index = 88, .length = 1},
  [51] = {.index = 89, .length = 3},
  [52] = {.index = 92, .length = 3},
  [53] = {.index = 95, .length = 3},
  [54] = {.index = 98, .length = 1},
  [55] = {.index = 99, .length = 3},
  [56] = {.index = 102, .length = 2},
  [57] = {.index = 104, .length = 2},
  [58] = {.index = 106, .length = 3},
  [59] = {.index = 109, .length = 2},
  [60] = {.index = 111, .length = 3},
  [61] = {.index = 114, .length = 2},
  [62] = {.index = 116, .length = 1},
  [63] = {.index = 117, .length = 1},
  [64] = {.index = 118, .length = 2},
  [65] = {.index = 120, .length = 2},
  [66] = {.index = 122, .length = 4},
  [67] = {.index = 126, .length = 2},
  [68] = {.index = 128, .length = 3},
  [69] = {.index = 131, .length = 2},
  [70] = {.index = 133, .length = 3},
  [71] = {.index = 136, .length = 3},
  [72] = {.index = 139, .length = 3},
  [73] = {.index = 142, .length = 4},
  [74] = {.index = 146, .length = 4},
  [75] = {.index = 150, .length = 4},
  [76] = {.index = 154, .length = 2},
  [77] = {.index = 156, .length = 2},
  [78] = {.index = 158, .length = 2},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_body, 0, .inherited = true},
    {field_parameters, 0, .inherited = true},
    {field_return_type, 0, .inherited = true},
    {field_value, 0},
  [4] =
    {field_body, 1, .inherited = true},
    {field_parameters, 1, .inherited = true},
    {field_return_type, 1, .inherited = true},
    {field_value, 1},
  [8] =
    {field_name, 1},
  [9] =
    {field_value, 0},
  [10] =
    {field_argument, 1},
  [11] =
    {field_argument, 1},
    {field_operator, 0},
  [13] =
    {field_constructor, 1},
  [14] =
    {field_operator, 1, .inherited = true},
  [15] =
    {field_object, 0},
  [16] =
    {field_arguments, 1},
    {field_function, 0},
  [18] =
    {field_definition, 1},
  [19] =
    {field_from, 2},
  [20] =
    {field_name, 1},
    {field_name, 2, .inherited = true},
  [22] =
    {field_body, 2, .inherited = true},
    {field_left, 0},
    {field_parameters, 2, .inherited = true},
    {field_return_type, 2, .inherited = true},
    {field_value, 2},
  [27] =
    {field_left, 0},
    {field_right, 2},
  [29] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [32] =
    {field_operator, 0},
  [33] =
    {field_operator, 0, .inherited = true},
    {field_operator, 1, .inherited = true},
  [35] =
    {field_attribute, 2},
    {field_object, 0},
  [37] =
    {field_default, 2},
    {field_object, 0},
  [39] =
    {field_arguments, 2},
    {field_object, 0},
  [41] =
    {field_expression, 1},
  [42] =
    {field_body, 3},
    {field_parameters, 1},
  [44] =
    {field_body, 1},
  [45] =
    {field_flag, 0},
  [46] =
    {field_flag, 3, .inherited = true},
  [47] =
    {field_name, 0},
  [48] =
    {field_module_name, 1},
    {field_name, 3, .inherited = true},
  [50] =
    {field_name, 0, .inherited = true},
    {field_name, 1, .inherited = true},
  [52] =
    {field_alias, 2},
    {field_name, 0},
  [54] =
    {field_condition, 1},
    {field_consequence, 3},
  [56] =
    {field_body, 3},
    {field_condition, 1},
  [58] =
    {field_body, 2},
  [59] =
    {field_body, 3},
  [60] =
    {field_alias, 2},
    {field_value, 0},
  [62] =
    {field_body, 3},
    {field_name, 1},
  [64] =
    {field_key, 0},
    {field_value, 2},
  [66] =
    {field_operator, 0},
    {field_operator, 1},
  [68] =
    {field_object, 0},
    {field_subscript, 2},
  [70] =
    {field_function, 0},
  [71] =
    {field_attribute, 3},
    {field_object, 0},
  [73] =
    {field_expression, 1},
    {field_type_conversion, 2},
  [75] =
    {field_body, 4},
    {field_parameters, 2},
  [77] =
    {field_name, 0},
    {field_value, 2},
  [79] =
    {field_name, 0},
    {field_type, 2},
  [81] =
    {field_body, 4},
    {field_name, 1},
    {field_parameters, 2},
  [84] =
    {field_flag, 0},
    {field_flag, 1, .inherited = true},
  [86] =
    {field_name, 0},
    {field_name, 1, .inherited = true},
  [88] =
    {field_alternative, 0},
  [89] =
    {field_alternative, 4},
    {field_condition, 1},
    {field_consequence, 3},
  [92] =
    {field_alternative, 4, .inherited = true},
    {field_condition, 1},
    {field_consequence, 3},
  [95] =
    {field_alternative, 4},
    {field_body, 3},
    {field_condition, 1},
  [98] =
    {field_body, 4},
  [99] =
    {field_body, 4},
    {field_name, 1},
    {field_superclasses, 2},
  [102] =
    {field_object, 0},
    {field_subscript, 3},
  [104] =
    {field_expression, 1},
    {field_format_spec, 3},
  [106] =
    {field_body, 5},
    {field_name, 2},
    {field_parameters, 3},
  [109] =
    {field_name, 1},
    {field_type, 3},
  [111] =
    {field_body, 5},
    {field_parameters, 1},
    {field_return_type, 3},
  [114] =
    {field_left, 1},
    {field_right, 3},
  [116] =
    {field_flag, 4, .inherited = true},
  [117] =
    {field_flag, 1},
  [118] =
    {field_flag, 0, .inherited = true},
    {field_flag, 1, .inherited = true},
  [120] =
    {field_module_name, 1},
    {field_name, 4, .inherited = true},
  [122] =
    {field_alternative, 4, .inherited = true},
    {field_alternative, 5},
    {field_condition, 1},
    {field_consequence, 3},
  [126] =
    {field_alternative, 0, .inherited = true},
    {field_alternative, 1, .inherited = true},
  [128] =
    {field_body, 5},
    {field_left, 1},
    {field_right, 3},
  [131] =
    {field_body, 2},
    {field_condition, 5},
  [133] =
    {field_expression, 1},
    {field_format_spec, 4},
    {field_type_conversion, 2},
  [136] =
    {field_body, 6},
    {field_parameters, 2},
    {field_return_type, 4},
  [139] =
    {field_name, 0},
    {field_type, 2},
    {field_value, 4},
  [142] =
    {field_body, 6},
    {field_name, 1},
    {field_parameters, 2},
    {field_return_type, 4},
  [146] =
    {field_alternative, 6},
    {field_body, 5},
    {field_left, 1},
    {field_right, 3},
  [150] =
    {field_body, 7},
    {field_name, 2},
    {field_parameters, 3},
    {field_return_type, 5},
  [154] =
    {field_alias, 2},
    {field_body, 4},
  [156] =
    {field_alias, 3},
    {field_body, 5},
  [158] =
    {field_alias, 4},
    {field_body, 6},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [14] = {
    [0] = sym_block,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__simple_statements, 2,
    sym__simple_statements,
    sym_block,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 2,
  [5] = 3,
  [6] = 3,
  [7] = 3,
  [8] = 2,
  [9] = 2,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 11,
  [14] = 14,
  [15] = 11,
  [16] = 10,
  [17] = 14,
  [18] = 11,
  [19] = 10,
  [20] = 11,
  [21] = 10,
  [22] = 11,
  [23] = 10,
  [24] = 11,
  [25] = 10,
  [26] = 11,
  [27] = 10,
  [28] = 11,
  [29] = 10,
  [30] = 11,
  [31] = 10,
  [32] = 11,
  [33] = 10,
  [34] = 11,
  [35] = 10,
  [36] = 11,
  [37] = 10,
  [38] = 11,
  [39] = 10,
  [40] = 10,
  [41] = 41,
  [42] = 41,
  [43] = 2,
  [44] = 3,
  [45] = 45,
  [46] = 45,
  [47] = 47,
  [48] = 47,
  [49] = 49,
  [50] = 2,
  [51] = 3,
  [52] = 2,
  [53] = 3,
  [54] = 54,
  [55] = 47,
  [56] = 56,
  [57] = 45,
  [58] = 54,
  [59] = 2,
  [60] = 56,
  [61] = 49,
  [62] = 45,
  [63] = 47,
  [64] = 3,
  [65] = 2,
  [66] = 3,
  [67] = 3,
  [68] = 3,
  [69] = 2,
  [70] = 2,
  [71] = 49,
  [72] = 49,
  [73] = 3,
  [74] = 2,
  [75] = 2,
  [76] = 3,
  [77] = 2,
  [78] = 3,
  [79] = 2,
  [80] = 3,
  [81] = 2,
  [82] = 3,
  [83] = 2,
  [84] = 3,
  [85] = 3,
  [86] = 3,
  [87] = 2,
  [88] = 3,
  [89] = 2,
  [90] = 2,
  [91] = 45,
  [92] = 45,
  [93] = 45,
  [94] = 47,
  [95] = 45,
  [96] = 41,
  [97] = 97,
  [98] = 47,
  [99] = 99,
  [100] = 45,
  [101] = 49,
  [102] = 54,
  [103] = 41,
  [104] = 97,
  [105] = 56,
  [106] = 106,
  [107] = 107,
  [108] = 99,
  [109] = 106,
  [110] = 45,
  [111] = 45,
  [112] = 47,
  [113] = 41,
  [114] = 107,
  [115] = 47,
  [116] = 41,
  [117] = 56,
  [118] = 47,
  [119] = 56,
  [120] = 47,
  [121] = 45,
  [122] = 54,
  [123] = 56,
  [124] = 47,
  [125] = 41,
  [126] = 45,
  [127] = 49,
  [128] = 49,
  [129] = 54,
  [130] = 45,
  [131] = 41,
  [132] = 49,
  [133] = 45,
  [134] = 45,
  [135] = 45,
  [136] = 54,
  [137] = 47,
  [138] = 54,
  [139] = 41,
  [140] = 49,
  [141] = 56,
  [142] = 45,
  [143] = 47,
  [144] = 49,
  [145] = 54,
  [146] = 41,
  [147] = 47,
  [148] = 45,
  [149] = 47,
  [150] = 45,
  [151] = 47,
  [152] = 49,
  [153] = 47,
  [154] = 56,
  [155] = 155,
  [156] = 156,
  [157] = 157,
  [158] = 158,
  [159] = 159,
  [160] = 160,
  [161] = 161,
  [162] = 162,
  [163] = 163,
  [164] = 164,
  [165] = 165,
  [166] = 166,
  [167] = 167,
  [168] = 168,
  [169] = 169,
  [170] = 170,
  [171] = 171,
  [172] = 49,
  [173] = 155,
  [174] = 174,
  [175] = 175,
  [176] = 160,
  [177] = 163,
  [178] = 168,
  [179] = 179,
  [180] = 49,
  [181] = 179,
  [182] = 49,
  [183] = 156,
  [184] = 184,
  [185] = 54,
  [186] = 47,
  [187] = 164,
  [188] = 188,
  [189] = 56,
  [190] = 169,
  [191] = 159,
  [192] = 49,
  [193] = 49,
  [194] = 169,
  [195] = 174,
  [196] = 157,
  [197] = 158,
  [198] = 166,
  [199] = 167,
  [200] = 175,
  [201] = 54,
  [202] = 47,
  [203] = 171,
  [204] = 56,
  [205] = 188,
  [206] = 169,
  [207] = 207,
  [208] = 170,
  [209] = 207,
  [210] = 161,
  [211] = 47,
  [212] = 162,
  [213] = 49,
  [214] = 184,
  [215] = 165,
  [216] = 216,
  [217] = 217,
  [218] = 218,
  [219] = 219,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 222,
  [224] = 218,
  [225] = 221,
  [226] = 217,
  [227] = 220,
  [228] = 222,
  [229] = 222,
  [230] = 219,
  [231] = 49,
  [232] = 49,
  [233] = 218,
  [234] = 216,
  [235] = 217,
  [236] = 216,
  [237] = 216,
  [238] = 217,
  [239] = 218,
  [240] = 49,
  [241] = 241,
  [242] = 242,
  [243] = 241,
  [244] = 242,
  [245] = 241,
  [246] = 241,
  [247] = 242,
  [248] = 241,
  [249] = 242,
  [250] = 242,
  [251] = 241,
  [252] = 242,
  [253] = 241,
  [254] = 242,
  [255] = 241,
  [256] = 242,
  [257] = 241,
  [258] = 242,
  [259] = 259,
  [260] = 260,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 265,
  [267] = 261,
  [268] = 262,
  [269] = 269,
  [270] = 263,
  [271] = 261,
  [272] = 272,
  [273] = 264,
  [274] = 262,
  [275] = 265,
  [276] = 261,
  [277] = 262,
  [278] = 261,
  [279] = 279,
  [280] = 280,
  [281] = 264,
  [282] = 260,
  [283] = 265,
  [284] = 272,
  [285] = 263,
  [286] = 261,
  [287] = 269,
  [288] = 272,
  [289] = 262,
  [290] = 260,
  [291] = 291,
  [292] = 280,
  [293] = 264,
  [294] = 279,
  [295] = 260,
  [296] = 265,
  [297] = 297,
  [298] = 263,
  [299] = 280,
  [300] = 261,
  [301] = 272,
  [302] = 262,
  [303] = 262,
  [304] = 264,
  [305] = 265,
  [306] = 261,
  [307] = 262,
  [308] = 264,
  [309] = 265,
  [310] = 261,
  [311] = 262,
  [312] = 264,
  [313] = 265,
  [314] = 261,
  [315] = 262,
  [316] = 264,
  [317] = 265,
  [318] = 261,
  [319] = 262,
  [320] = 291,
  [321] = 280,
  [322] = 260,
  [323] = 263,
  [324] = 272,
  [325] = 264,
  [326] = 265,
  [327] = 261,
  [328] = 262,
  [329] = 264,
  [330] = 265,
  [331] = 261,
  [332] = 262,
  [333] = 280,
  [334] = 260,
  [335] = 263,
  [336] = 265,
  [337] = 272,
  [338] = 264,
  [339] = 265,
  [340] = 261,
  [341] = 262,
  [342] = 264,
  [343] = 265,
  [344] = 261,
  [345] = 262,
  [346] = 264,
  [347] = 280,
  [348] = 280,
  [349] = 260,
  [350] = 263,
  [351] = 272,
  [352] = 264,
  [353] = 265,
  [354] = 261,
  [355] = 262,
  [356] = 264,
  [357] = 265,
  [358] = 261,
  [359] = 262,
  [360] = 264,
  [361] = 280,
  [362] = 260,
  [363] = 263,
  [364] = 272,
  [365] = 264,
  [366] = 265,
  [367] = 261,
  [368] = 262,
  [369] = 264,
  [370] = 265,
  [371] = 261,
  [372] = 262,
  [373] = 280,
  [374] = 260,
  [375] = 263,
  [376] = 272,
  [377] = 264,
  [378] = 265,
  [379] = 261,
  [380] = 262,
  [381] = 264,
  [382] = 264,
  [383] = 280,
  [384] = 260,
  [385] = 263,
  [386] = 272,
  [387] = 280,
  [388] = 260,
  [389] = 263,
  [390] = 272,
  [391] = 280,
  [392] = 260,
  [393] = 263,
  [394] = 272,
  [395] = 280,
  [396] = 260,
  [397] = 263,
  [398] = 272,
  [399] = 265,
  [400] = 280,
  [401] = 260,
  [402] = 263,
  [403] = 272,
  [404] = 280,
  [405] = 260,
  [406] = 263,
  [407] = 272,
  [408] = 280,
  [409] = 260,
  [410] = 263,
  [411] = 272,
  [412] = 280,
  [413] = 260,
  [414] = 263,
  [415] = 272,
  [416] = 280,
  [417] = 260,
  [418] = 263,
  [419] = 272,
  [420] = 280,
  [421] = 260,
  [422] = 263,
  [423] = 272,
  [424] = 280,
  [425] = 260,
  [426] = 263,
  [427] = 272,
  [428] = 280,
  [429] = 260,
  [430] = 263,
  [431] = 272,
  [432] = 261,
  [433] = 262,
  [434] = 265,
  [435] = 297,
  [436] = 2,
  [437] = 3,
  [438] = 438,
  [439] = 439,
  [440] = 440,
  [441] = 441,
  [442] = 442,
  [443] = 440,
  [444] = 444,
  [445] = 441,
  [446] = 442,
  [447] = 447,
  [448] = 439,
  [449] = 440,
  [450] = 450,
  [451] = 439,
  [452] = 447,
  [453] = 444,
  [454] = 447,
  [455] = 439,
  [456] = 440,
  [457] = 438,
  [458] = 450,
  [459] = 447,
  [460] = 460,
  [461] = 461,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 462,
  [466] = 463,
  [467] = 467,
  [468] = 468,
  [469] = 469,
  [470] = 470,
  [471] = 471,
  [472] = 472,
  [473] = 473,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 479,
  [480] = 480,
  [481] = 481,
  [482] = 467,
  [483] = 468,
  [484] = 484,
  [485] = 470,
  [486] = 471,
  [487] = 472,
  [488] = 473,
  [489] = 474,
  [490] = 475,
  [491] = 476,
  [492] = 477,
  [493] = 478,
  [494] = 479,
  [495] = 480,
  [496] = 481,
  [497] = 469,
  [498] = 498,
  [499] = 462,
  [500] = 500,
  [501] = 501,
  [502] = 502,
  [503] = 503,
  [504] = 462,
  [505] = 462,
  [506] = 501,
  [507] = 498,
  [508] = 502,
  [509] = 503,
  [510] = 510,
  [511] = 498,
  [512] = 462,
  [513] = 513,
  [514] = 462,
  [515] = 515,
  [516] = 516,
  [517] = 510,
  [518] = 460,
  [519] = 462,
  [520] = 462,
  [521] = 462,
  [522] = 462,
  [523] = 462,
  [524] = 462,
  [525] = 462,
  [526] = 515,
  [527] = 527,
  [528] = 462,
  [529] = 529,
  [530] = 462,
  [531] = 462,
  [532] = 529,
  [533] = 462,
  [534] = 534,
  [535] = 516,
  [536] = 462,
  [537] = 534,
  [538] = 464,
  [539] = 498,
  [540] = 500,
  [541] = 462,
  [542] = 542,
  [543] = 543,
  [544] = 544,
  [545] = 545,
  [546] = 546,
  [547] = 542,
  [548] = 543,
  [549] = 527,
  [550] = 544,
  [551] = 545,
  [552] = 546,
  [553] = 462,
  [554] = 513,
  [555] = 555,
  [556] = 461,
  [557] = 555,
  [558] = 484,
  [559] = 559,
  [560] = 560,
  [561] = 561,
  [562] = 562,
  [563] = 544,
  [564] = 564,
  [565] = 461,
  [566] = 566,
  [567] = 567,
  [568] = 568,
  [569] = 467,
  [570] = 570,
  [571] = 571,
  [572] = 572,
  [573] = 468,
  [574] = 477,
  [575] = 575,
  [576] = 576,
  [577] = 577,
  [578] = 578,
  [579] = 579,
  [580] = 484,
  [581] = 470,
  [582] = 582,
  [583] = 578,
  [584] = 501,
  [585] = 542,
  [586] = 529,
  [587] = 471,
  [588] = 460,
  [589] = 589,
  [590] = 555,
  [591] = 591,
  [592] = 478,
  [593] = 560,
  [594] = 479,
  [595] = 589,
  [596] = 472,
  [597] = 545,
  [598] = 502,
  [599] = 599,
  [600] = 600,
  [601] = 601,
  [602] = 480,
  [603] = 603,
  [604] = 604,
  [605] = 605,
  [606] = 561,
  [607] = 562,
  [608] = 608,
  [609] = 480,
  [610] = 591,
  [611] = 544,
  [612] = 612,
  [613] = 566,
  [614] = 529,
  [615] = 460,
  [616] = 616,
  [617] = 617,
  [618] = 618,
  [619] = 503,
  [620] = 568,
  [621] = 546,
  [622] = 474,
  [623] = 545,
  [624] = 624,
  [625] = 473,
  [626] = 599,
  [627] = 627,
  [628] = 628,
  [629] = 570,
  [630] = 475,
  [631] = 600,
  [632] = 481,
  [633] = 471,
  [634] = 601,
  [635] = 474,
  [636] = 475,
  [637] = 470,
  [638] = 603,
  [639] = 503,
  [640] = 582,
  [641] = 641,
  [642] = 476,
  [643] = 608,
  [644] = 472,
  [645] = 612,
  [646] = 641,
  [647] = 477,
  [648] = 648,
  [649] = 649,
  [650] = 616,
  [651] = 651,
  [652] = 463,
  [653] = 478,
  [654] = 617,
  [655] = 655,
  [656] = 618,
  [657] = 657,
  [658] = 571,
  [659] = 659,
  [660] = 567,
  [661] = 546,
  [662] = 542,
  [663] = 555,
  [664] = 479,
  [665] = 649,
  [666] = 666,
  [667] = 461,
  [668] = 651,
  [669] = 572,
  [670] = 501,
  [671] = 481,
  [672] = 463,
  [673] = 575,
  [674] = 674,
  [675] = 655,
  [676] = 543,
  [677] = 473,
  [678] = 564,
  [679] = 605,
  [680] = 502,
  [681] = 648,
  [682] = 666,
  [683] = 657,
  [684] = 674,
  [685] = 624,
  [686] = 576,
  [687] = 687,
  [688] = 688,
  [689] = 689,
  [690] = 543,
  [691] = 691,
  [692] = 484,
  [693] = 577,
  [694] = 694,
  [695] = 695,
  [696] = 695,
  [697] = 691,
  [698] = 627,
  [699] = 628,
  [700] = 687,
  [701] = 559,
  [702] = 659,
  [703] = 476,
  [704] = 688,
  [705] = 467,
  [706] = 468,
  [707] = 579,
  [708] = 689,
  [709] = 694,
  [710] = 604,
  [711] = 711,
  [712] = 711,
  [713] = 713,
  [714] = 713,
  [715] = 713,
  [716] = 713,
  [717] = 713,
  [718] = 713,
  [719] = 713,
  [720] = 713,
  [721] = 713,
  [722] = 713,
  [723] = 723,
  [724] = 724,
  [725] = 725,
  [726] = 726,
  [727] = 727,
  [728] = 724,
  [729] = 723,
  [730] = 726,
  [731] = 724,
  [732] = 726,
  [733] = 733,
  [734] = 726,
  [735] = 724,
  [736] = 724,
  [737] = 726,
  [738] = 724,
  [739] = 725,
  [740] = 727,
  [741] = 724,
  [742] = 724,
  [743] = 733,
  [744] = 724,
  [745] = 723,
  [746] = 733,
  [747] = 726,
  [748] = 724,
  [749] = 726,
  [750] = 726,
  [751] = 726,
  [752] = 727,
  [753] = 724,
  [754] = 725,
  [755] = 727,
  [756] = 725,
  [757] = 723,
  [758] = 733,
  [759] = 724,
  [760] = 726,
  [761] = 726,
  [762] = 762,
  [763] = 762,
  [764] = 762,
  [765] = 762,
  [766] = 766,
  [767] = 767,
  [768] = 768,
  [769] = 768,
  [770] = 768,
  [771] = 45,
  [772] = 768,
  [773] = 768,
  [774] = 768,
  [775] = 767,
  [776] = 768,
  [777] = 768,
  [778] = 768,
  [779] = 768,
  [780] = 768,
  [781] = 768,
  [782] = 782,
  [783] = 783,
  [784] = 782,
  [785] = 785,
  [786] = 786,
  [787] = 786,
  [788] = 782,
  [789] = 782,
  [790] = 790,
  [791] = 45,
  [792] = 792,
  [793] = 782,
  [794] = 41,
  [795] = 786,
  [796] = 796,
  [797] = 797,
  [798] = 792,
  [799] = 782,
  [800] = 790,
  [801] = 792,
  [802] = 802,
  [803] = 802,
  [804] = 802,
  [805] = 805,
  [806] = 805,
  [807] = 807,
  [808] = 808,
  [809] = 785,
  [810] = 796,
  [811] = 54,
  [812] = 807,
  [813] = 808,
  [814] = 805,
  [815] = 786,
  [816] = 786,
  [817] = 782,
  [818] = 786,
  [819] = 786,
  [820] = 782,
  [821] = 786,
  [822] = 786,
  [823] = 782,
  [824] = 782,
  [825] = 54,
  [826] = 785,
  [827] = 796,
  [828] = 783,
  [829] = 797,
  [830] = 786,
  [831] = 797,
  [832] = 790,
  [833] = 792,
  [834] = 786,
  [835] = 802,
  [836] = 805,
  [837] = 56,
  [838] = 786,
  [839] = 807,
  [840] = 808,
  [841] = 786,
  [842] = 785,
  [843] = 796,
  [844] = 782,
  [845] = 56,
  [846] = 807,
  [847] = 786,
  [848] = 786,
  [849] = 808,
  [850] = 782,
  [851] = 782,
  [852] = 786,
  [853] = 782,
  [854] = 786,
  [855] = 782,
  [856] = 786,
  [857] = 782,
  [858] = 797,
  [859] = 782,
  [860] = 790,
  [861] = 786,
  [862] = 782,
  [863] = 786,
  [864] = 782,
  [865] = 782,
  [866] = 866,
  [867] = 41,
  [868] = 47,
  [869] = 869,
  [870] = 54,
  [871] = 49,
  [872] = 872,
  [873] = 869,
  [874] = 866,
  [875] = 56,
  [876] = 869,
  [877] = 869,
  [878] = 869,
  [879] = 45,
  [880] = 56,
  [881] = 54,
  [882] = 56,
  [883] = 883,
  [884] = 884,
  [885] = 883,
  [886] = 54,
  [887] = 887,
  [888] = 49,
  [889] = 889,
  [890] = 889,
  [891] = 889,
  [892] = 892,
  [893] = 893,
  [894] = 894,
  [895] = 889,
  [896] = 889,
  [897] = 889,
  [898] = 889,
  [899] = 889,
  [900] = 889,
  [901] = 889,
  [902] = 889,
  [903] = 889,
  [904] = 889,
  [905] = 889,
  [906] = 889,
  [907] = 893,
  [908] = 908,
  [909] = 892,
  [910] = 889,
  [911] = 889,
  [912] = 908,
  [913] = 913,
  [914] = 889,
  [915] = 889,
  [916] = 916,
  [917] = 917,
  [918] = 889,
  [919] = 919,
  [920] = 920,
  [921] = 921,
  [922] = 922,
  [923] = 923,
  [924] = 924,
  [925] = 925,
  [926] = 926,
  [927] = 926,
  [928] = 925,
  [929] = 929,
  [930] = 929,
  [931] = 931,
  [932] = 932,
  [933] = 933,
  [934] = 931,
  [935] = 932,
  [936] = 216,
  [937] = 937,
  [938] = 938,
  [939] = 217,
  [940] = 218,
  [941] = 938,
  [942] = 942,
  [943] = 943,
  [944] = 921,
  [945] = 922,
  [946] = 924,
  [947] = 947,
  [948] = 948,
  [949] = 949,
  [950] = 950,
  [951] = 951,
  [952] = 925,
  [953] = 929,
  [954] = 931,
  [955] = 932,
  [956] = 938,
  [957] = 942,
  [958] = 943,
  [959] = 921,
  [960] = 922,
  [961] = 924,
  [962] = 947,
  [963] = 963,
  [964] = 948,
  [965] = 949,
  [966] = 950,
  [967] = 951,
  [968] = 968,
  [969] = 969,
  [970] = 970,
  [971] = 933,
  [972] = 947,
  [973] = 920,
  [974] = 963,
  [975] = 942,
  [976] = 943,
  [977] = 921,
  [978] = 929,
  [979] = 931,
  [980] = 932,
  [981] = 922,
  [982] = 924,
  [983] = 920,
  [984] = 933,
  [985] = 947,
  [986] = 963,
  [987] = 968,
  [988] = 969,
  [989] = 948,
  [990] = 949,
  [991] = 929,
  [992] = 931,
  [993] = 932,
  [994] = 950,
  [995] = 951,
  [996] = 938,
  [997] = 942,
  [998] = 943,
  [999] = 921,
  [1000] = 922,
  [1001] = 924,
  [1002] = 947,
  [1003] = 963,
  [1004] = 948,
  [1005] = 949,
  [1006] = 950,
  [1007] = 951,
  [1008] = 920,
  [1009] = 968,
  [1010] = 920,
  [1011] = 969,
  [1012] = 970,
  [1013] = 933,
  [1014] = 970,
  [1015] = 938,
  [1016] = 942,
  [1017] = 943,
  [1018] = 921,
  [1019] = 922,
  [1020] = 924,
  [1021] = 947,
  [1022] = 963,
  [1023] = 948,
  [1024] = 949,
  [1025] = 950,
  [1026] = 951,
  [1027] = 920,
  [1028] = 968,
  [1029] = 969,
  [1030] = 970,
  [1031] = 968,
  [1032] = 969,
  [1033] = 948,
  [1034] = 219,
  [1035] = 949,
  [1036] = 938,
  [1037] = 942,
  [1038] = 943,
  [1039] = 921,
  [1040] = 922,
  [1041] = 924,
  [1042] = 947,
  [1043] = 963,
  [1044] = 925,
  [1045] = 948,
  [1046] = 949,
  [1047] = 950,
  [1048] = 951,
  [1049] = 920,
  [1050] = 968,
  [1051] = 969,
  [1052] = 925,
  [1053] = 929,
  [1054] = 931,
  [1055] = 932,
  [1056] = 938,
  [1057] = 970,
  [1058] = 942,
  [1059] = 943,
  [1060] = 921,
  [1061] = 922,
  [1062] = 924,
  [1063] = 947,
  [1064] = 963,
  [1065] = 948,
  [1066] = 949,
  [1067] = 950,
  [1068] = 951,
  [1069] = 970,
  [1070] = 968,
  [1071] = 969,
  [1072] = 970,
  [1073] = 933,
  [1074] = 933,
  [1075] = 933,
  [1076] = 220,
  [1077] = 221,
  [1078] = 222,
  [1079] = 920,
  [1080] = 950,
  [1081] = 951,
  [1082] = 938,
  [1083] = 942,
  [1084] = 943,
  [1085] = 921,
  [1086] = 922,
  [1087] = 924,
  [1088] = 947,
  [1089] = 963,
  [1090] = 948,
  [1091] = 949,
  [1092] = 950,
  [1093] = 951,
  [1094] = 920,
  [1095] = 925,
  [1096] = 929,
  [1097] = 931,
  [1098] = 932,
  [1099] = 938,
  [1100] = 942,
  [1101] = 943,
  [1102] = 921,
  [1103] = 922,
  [1104] = 924,
  [1105] = 947,
  [1106] = 963,
  [1107] = 948,
  [1108] = 949,
  [1109] = 950,
  [1110] = 951,
  [1111] = 968,
  [1112] = 969,
  [1113] = 970,
  [1114] = 933,
  [1115] = 968,
  [1116] = 920,
  [1117] = 969,
  [1118] = 970,
  [1119] = 970,
  [1120] = 925,
  [1121] = 929,
  [1122] = 931,
  [1123] = 932,
  [1124] = 938,
  [1125] = 942,
  [1126] = 943,
  [1127] = 921,
  [1128] = 922,
  [1129] = 924,
  [1130] = 947,
  [1131] = 963,
  [1132] = 948,
  [1133] = 949,
  [1134] = 950,
  [1135] = 951,
  [1136] = 968,
  [1137] = 969,
  [1138] = 933,
  [1139] = 970,
  [1140] = 1140,
  [1141] = 920,
  [1142] = 925,
  [1143] = 929,
  [1144] = 931,
  [1145] = 932,
  [1146] = 938,
  [1147] = 942,
  [1148] = 943,
  [1149] = 921,
  [1150] = 922,
  [1151] = 924,
  [1152] = 947,
  [1153] = 963,
  [1154] = 948,
  [1155] = 949,
  [1156] = 950,
  [1157] = 951,
  [1158] = 968,
  [1159] = 969,
  [1160] = 1160,
  [1161] = 970,
  [1162] = 920,
  [1163] = 1163,
  [1164] = 1164,
  [1165] = 1165,
  [1166] = 925,
  [1167] = 929,
  [1168] = 925,
  [1169] = 929,
  [1170] = 931,
  [1171] = 932,
  [1172] = 938,
  [1173] = 942,
  [1174] = 943,
  [1175] = 921,
  [1176] = 922,
  [1177] = 924,
  [1178] = 947,
  [1179] = 963,
  [1180] = 948,
  [1181] = 949,
  [1182] = 950,
  [1183] = 951,
  [1184] = 968,
  [1185] = 969,
  [1186] = 931,
  [1187] = 970,
  [1188] = 932,
  [1189] = 1189,
  [1190] = 920,
  [1191] = 1191,
  [1192] = 1192,
  [1193] = 938,
  [1194] = 942,
  [1195] = 943,
  [1196] = 921,
  [1197] = 922,
  [1198] = 924,
  [1199] = 947,
  [1200] = 963,
  [1201] = 948,
  [1202] = 949,
  [1203] = 950,
  [1204] = 951,
  [1205] = 968,
  [1206] = 969,
  [1207] = 970,
  [1208] = 920,
  [1209] = 938,
  [1210] = 1210,
  [1211] = 920,
  [1212] = 938,
  [1213] = 938,
  [1214] = 920,
  [1215] = 1215,
  [1216] = 1216,
  [1217] = 942,
  [1218] = 943,
  [1219] = 938,
  [1220] = 921,
  [1221] = 922,
  [1222] = 924,
  [1223] = 947,
  [1224] = 963,
  [1225] = 948,
  [1226] = 949,
  [1227] = 950,
  [1228] = 951,
  [1229] = 937,
  [1230] = 925,
  [1231] = 1231,
  [1232] = 920,
  [1233] = 929,
  [1234] = 1234,
  [1235] = 1235,
  [1236] = 1236,
  [1237] = 1237,
  [1238] = 968,
  [1239] = 969,
  [1240] = 1240,
  [1241] = 1241,
  [1242] = 931,
  [1243] = 932,
  [1244] = 1244,
  [1245] = 919,
  [1246] = 1240,
  [1247] = 923,
  [1248] = 926,
  [1249] = 1163,
  [1250] = 1210,
  [1251] = 1237,
  [1252] = 925,
  [1253] = 929,
  [1254] = 931,
  [1255] = 932,
  [1256] = 938,
  [1257] = 925,
  [1258] = 933,
  [1259] = 970,
  [1260] = 933,
  [1261] = 925,
  [1262] = 938,
  [1263] = 942,
  [1264] = 943,
  [1265] = 921,
  [1266] = 922,
  [1267] = 924,
  [1268] = 947,
  [1269] = 963,
  [1270] = 948,
  [1271] = 949,
  [1272] = 950,
  [1273] = 951,
  [1274] = 968,
  [1275] = 969,
  [1276] = 920,
  [1277] = 968,
  [1278] = 969,
  [1279] = 1279,
  [1280] = 970,
  [1281] = 925,
  [1282] = 929,
  [1283] = 931,
  [1284] = 932,
  [1285] = 942,
  [1286] = 943,
  [1287] = 921,
  [1288] = 922,
  [1289] = 924,
  [1290] = 947,
  [1291] = 963,
  [1292] = 948,
  [1293] = 949,
  [1294] = 950,
  [1295] = 951,
  [1296] = 968,
  [1297] = 969,
  [1298] = 970,
  [1299] = 919,
  [1300] = 925,
  [1301] = 929,
  [1302] = 920,
  [1303] = 933,
  [1304] = 929,
  [1305] = 931,
  [1306] = 932,
  [1307] = 925,
  [1308] = 929,
  [1309] = 931,
  [1310] = 932,
  [1311] = 942,
  [1312] = 943,
  [1313] = 921,
  [1314] = 922,
  [1315] = 924,
  [1316] = 947,
  [1317] = 963,
  [1318] = 948,
  [1319] = 949,
  [1320] = 950,
  [1321] = 951,
  [1322] = 968,
  [1323] = 969,
  [1324] = 970,
  [1325] = 920,
  [1326] = 938,
  [1327] = 919,
  [1328] = 931,
  [1329] = 923,
  [1330] = 1163,
  [1331] = 1210,
  [1332] = 1237,
  [1333] = 925,
  [1334] = 932,
  [1335] = 938,
  [1336] = 942,
  [1337] = 943,
  [1338] = 921,
  [1339] = 922,
  [1340] = 924,
  [1341] = 947,
  [1342] = 963,
  [1343] = 948,
  [1344] = 949,
  [1345] = 950,
  [1346] = 951,
  [1347] = 920,
  [1348] = 968,
  [1349] = 969,
  [1350] = 970,
  [1351] = 925,
  [1352] = 929,
  [1353] = 931,
  [1354] = 932,
  [1355] = 933,
  [1356] = 929,
  [1357] = 931,
  [1358] = 932,
  [1359] = 919,
  [1360] = 1192,
  [1361] = 1163,
  [1362] = 1237,
  [1363] = 925,
  [1364] = 938,
  [1365] = 938,
  [1366] = 942,
  [1367] = 943,
  [1368] = 921,
  [1369] = 922,
  [1370] = 924,
  [1371] = 947,
  [1372] = 963,
  [1373] = 948,
  [1374] = 949,
  [1375] = 950,
  [1376] = 951,
  [1377] = 920,
  [1378] = 968,
  [1379] = 969,
  [1380] = 970,
  [1381] = 933,
  [1382] = 929,
  [1383] = 931,
  [1384] = 932,
  [1385] = 919,
  [1386] = 942,
  [1387] = 1163,
  [1388] = 1237,
  [1389] = 919,
  [1390] = 1163,
  [1391] = 1237,
  [1392] = 1392,
  [1393] = 1163,
  [1394] = 1237,
  [1395] = 919,
  [1396] = 1163,
  [1397] = 1237,
  [1398] = 919,
  [1399] = 1163,
  [1400] = 1237,
  [1401] = 919,
  [1402] = 1163,
  [1403] = 1237,
  [1404] = 919,
  [1405] = 1163,
  [1406] = 1237,
  [1407] = 919,
  [1408] = 1163,
  [1409] = 1237,
  [1410] = 919,
  [1411] = 1163,
  [1412] = 1237,
  [1413] = 919,
  [1414] = 1163,
  [1415] = 1237,
  [1416] = 919,
  [1417] = 1163,
  [1418] = 1237,
  [1419] = 919,
  [1420] = 1163,
  [1421] = 1237,
  [1422] = 919,
  [1423] = 1163,
  [1424] = 1237,
  [1425] = 919,
  [1426] = 1163,
  [1427] = 1237,
  [1428] = 919,
  [1429] = 1163,
  [1430] = 1237,
  [1431] = 919,
  [1432] = 1163,
  [1433] = 1237,
  [1434] = 1244,
  [1435] = 1216,
  [1436] = 1234,
  [1437] = 1164,
  [1438] = 1392,
  [1439] = 1439,
  [1440] = 923,
  [1441] = 1163,
  [1442] = 1165,
  [1443] = 1210,
  [1444] = 1215,
  [1445] = 1235,
  [1446] = 1237,
  [1447] = 1279,
  [1448] = 925,
  [1449] = 943,
  [1450] = 1450,
  [1451] = 1451,
  [1452] = 1439,
  [1453] = 923,
  [1454] = 1210,
  [1455] = 925,
  [1456] = 923,
  [1457] = 1210,
  [1458] = 923,
  [1459] = 1210,
  [1460] = 923,
  [1461] = 1210,
  [1462] = 923,
  [1463] = 1210,
  [1464] = 923,
  [1465] = 1210,
  [1466] = 923,
  [1467] = 1210,
  [1468] = 923,
  [1469] = 1210,
  [1470] = 923,
  [1471] = 1210,
  [1472] = 923,
  [1473] = 1210,
  [1474] = 923,
  [1475] = 1210,
  [1476] = 923,
  [1477] = 1210,
  [1478] = 923,
  [1479] = 1210,
  [1480] = 923,
  [1481] = 1210,
  [1482] = 923,
  [1483] = 1210,
  [1484] = 923,
  [1485] = 1210,
  [1486] = 923,
  [1487] = 1210,
  [1488] = 963,
  [1489] = 216,
  [1490] = 217,
  [1491] = 218,
  [1492] = 222,
  [1493] = 1493,
  [1494] = 216,
  [1495] = 217,
  [1496] = 218,
  [1497] = 1497,
  [1498] = 1497,
  [1499] = 1493,
  [1500] = 218,
  [1501] = 1501,
  [1502] = 1502,
  [1503] = 1503,
  [1504] = 1501,
  [1505] = 216,
  [1506] = 217,
  [1507] = 1507,
  [1508] = 1508,
  [1509] = 1508,
  [1510] = 217,
  [1511] = 218,
  [1512] = 220,
  [1513] = 216,
  [1514] = 1514,
  [1515] = 221,
  [1516] = 222,
  [1517] = 216,
  [1518] = 1518,
  [1519] = 217,
  [1520] = 218,
  [1521] = 222,
  [1522] = 222,
  [1523] = 217,
  [1524] = 218,
  [1525] = 1514,
  [1526] = 1507,
  [1527] = 216,
  [1528] = 219,
  [1529] = 1518,
  [1530] = 1530,
  [1531] = 1530,
  [1532] = 1532,
  [1533] = 1533,
  [1534] = 460,
  [1535] = 1535,
  [1536] = 450,
  [1537] = 438,
  [1538] = 1538,
  [1539] = 1539,
  [1540] = 1540,
  [1541] = 1541,
  [1542] = 1532,
  [1543] = 1543,
  [1544] = 1533,
  [1545] = 1540,
  [1546] = 1535,
  [1547] = 1538,
  [1548] = 1548,
  [1549] = 1539,
  [1550] = 1548,
  [1551] = 1551,
  [1552] = 1543,
  [1553] = 1551,
  [1554] = 222,
  [1555] = 1541,
  [1556] = 442,
  [1557] = 529,
  [1558] = 460,
  [1559] = 529,
  [1560] = 1551,
  [1561] = 1548,
  [1562] = 529,
  [1563] = 1548,
  [1564] = 529,
  [1565] = 1565,
  [1566] = 460,
  [1567] = 1567,
  [1568] = 1568,
  [1569] = 1551,
  [1570] = 1535,
  [1571] = 1571,
  [1572] = 1535,
  [1573] = 1565,
  [1574] = 460,
  [1575] = 1571,
  [1576] = 1539,
  [1577] = 1577,
  [1578] = 1568,
  [1579] = 1577,
  [1580] = 1567,
  [1581] = 1539,
  [1582] = 460,
  [1583] = 529,
  [1584] = 460,
  [1585] = 1535,
  [1586] = 605,
  [1587] = 1548,
  [1588] = 529,
  [1589] = 1539,
  [1590] = 1551,
  [1591] = 1535,
  [1592] = 655,
  [1593] = 1548,
  [1594] = 1539,
  [1595] = 1551,
  [1596] = 529,
  [1597] = 460,
  [1598] = 1598,
  [1599] = 1598,
  [1600] = 1600,
  [1601] = 1601,
  [1602] = 1602,
  [1603] = 1603,
  [1604] = 1604,
  [1605] = 1605,
  [1606] = 1606,
  [1607] = 1607,
  [1608] = 464,
  [1609] = 1609,
  [1610] = 1551,
  [1611] = 1611,
  [1612] = 1607,
  [1613] = 1539,
  [1614] = 1614,
  [1615] = 1609,
  [1616] = 1616,
  [1617] = 1617,
  [1618] = 1602,
  [1619] = 1539,
  [1620] = 529,
  [1621] = 1621,
  [1622] = 1616,
  [1623] = 460,
  [1624] = 1604,
  [1625] = 1611,
  [1626] = 1601,
  [1627] = 1598,
  [1628] = 1628,
  [1629] = 1605,
  [1630] = 1603,
  [1631] = 1631,
  [1632] = 1632,
  [1633] = 1633,
  [1634] = 1628,
  [1635] = 1635,
  [1636] = 1636,
  [1637] = 1637,
  [1638] = 529,
  [1639] = 1598,
  [1640] = 1640,
  [1641] = 1631,
  [1642] = 1548,
  [1643] = 1606,
  [1644] = 1535,
  [1645] = 1614,
  [1646] = 460,
  [1647] = 1535,
  [1648] = 1632,
  [1649] = 1635,
  [1650] = 1617,
  [1651] = 1621,
  [1652] = 1551,
  [1653] = 1633,
  [1654] = 1636,
  [1655] = 1548,
  [1656] = 1637,
  [1657] = 1600,
  [1658] = 1640,
  [1659] = 648,
  [1660] = 687,
  [1661] = 651,
  [1662] = 604,
  [1663] = 608,
  [1664] = 572,
  [1665] = 582,
  [1666] = 612,
  [1667] = 575,
  [1668] = 578,
  [1669] = 589,
  [1670] = 576,
  [1671] = 577,
  [1672] = 591,
  [1673] = 617,
  [1674] = 559,
  [1675] = 579,
  [1676] = 616,
  [1677] = 599,
  [1678] = 657,
  [1679] = 659,
  [1680] = 564,
  [1681] = 688,
  [1682] = 641,
  [1683] = 689,
  [1684] = 691,
  [1685] = 561,
  [1686] = 694,
  [1687] = 600,
  [1688] = 695,
  [1689] = 627,
  [1690] = 628,
  [1691] = 566,
  [1692] = 568,
  [1693] = 666,
  [1694] = 560,
  [1695] = 570,
  [1696] = 601,
  [1697] = 567,
  [1698] = 649,
  [1699] = 603,
  [1700] = 618,
  [1701] = 562,
  [1702] = 571,
  [1703] = 674,
  [1704] = 624,
  [1705] = 447,
  [1706] = 441,
  [1707] = 440,
  [1708] = 444,
  [1709] = 439,
  [1710] = 440,
  [1711] = 503,
  [1712] = 441,
  [1713] = 1713,
  [1714] = 447,
  [1715] = 447,
  [1716] = 510,
  [1717] = 447,
  [1718] = 481,
  [1719] = 444,
  [1720] = 544,
  [1721] = 439,
  [1722] = 478,
  [1723] = 439,
  [1724] = 475,
  [1725] = 527,
  [1726] = 439,
  [1727] = 463,
  [1728] = 500,
  [1729] = 543,
  [1730] = 480,
  [1731] = 546,
  [1732] = 555,
  [1733] = 440,
  [1734] = 440,
  [1735] = 502,
  [1736] = 473,
  [1737] = 442,
  [1738] = 479,
  [1739] = 444,
  [1740] = 441,
  [1741] = 447,
  [1742] = 440,
  [1743] = 1713,
  [1744] = 461,
  [1745] = 440,
  [1746] = 450,
  [1747] = 467,
  [1748] = 472,
  [1749] = 439,
  [1750] = 545,
  [1751] = 474,
  [1752] = 444,
  [1753] = 439,
  [1754] = 447,
  [1755] = 501,
  [1756] = 439,
  [1757] = 542,
  [1758] = 477,
  [1759] = 498,
  [1760] = 441,
  [1761] = 468,
  [1762] = 484,
  [1763] = 438,
  [1764] = 440,
  [1765] = 1765,
  [1766] = 1765,
  [1767] = 447,
  [1768] = 470,
  [1769] = 471,
  [1770] = 476,
  [1771] = 461,
  [1772] = 480,
  [1773] = 481,
  [1774] = 447,
  [1775] = 444,
  [1776] = 439,
  [1777] = 501,
  [1778] = 503,
  [1779] = 441,
  [1780] = 502,
  [1781] = 440,
  [1782] = 447,
  [1783] = 439,
  [1784] = 463,
  [1785] = 463,
  [1786] = 544,
  [1787] = 545,
  [1788] = 546,
  [1789] = 463,
  [1790] = 500,
  [1791] = 544,
  [1792] = 545,
  [1793] = 546,
  [1794] = 500,
  [1795] = 544,
  [1796] = 545,
  [1797] = 1503,
  [1798] = 510,
  [1799] = 546,
  [1800] = 498,
  [1801] = 467,
  [1802] = 468,
  [1803] = 484,
  [1804] = 470,
  [1805] = 471,
  [1806] = 472,
  [1807] = 473,
  [1808] = 474,
  [1809] = 475,
  [1810] = 476,
  [1811] = 477,
  [1812] = 478,
  [1813] = 479,
  [1814] = 480,
  [1815] = 481,
  [1816] = 501,
  [1817] = 502,
  [1818] = 503,
  [1819] = 498,
  [1820] = 542,
  [1821] = 444,
  [1822] = 1502,
  [1823] = 527,
  [1824] = 543,
  [1825] = 441,
  [1826] = 498,
  [1827] = 479,
  [1828] = 527,
  [1829] = 542,
  [1830] = 555,
  [1831] = 461,
  [1832] = 543,
  [1833] = 510,
  [1834] = 467,
  [1835] = 468,
  [1836] = 484,
  [1837] = 470,
  [1838] = 471,
  [1839] = 472,
  [1840] = 473,
  [1841] = 474,
  [1842] = 475,
  [1843] = 476,
  [1844] = 477,
  [1845] = 478,
  [1846] = 479,
  [1847] = 480,
  [1848] = 481,
  [1849] = 501,
  [1850] = 502,
  [1851] = 503,
  [1852] = 498,
  [1853] = 542,
  [1854] = 543,
  [1855] = 555,
  [1856] = 527,
  [1857] = 555,
  [1858] = 461,
  [1859] = 510,
  [1860] = 478,
  [1861] = 447,
  [1862] = 500,
  [1863] = 440,
  [1864] = 467,
  [1865] = 468,
  [1866] = 484,
  [1867] = 470,
  [1868] = 471,
  [1869] = 472,
  [1870] = 473,
  [1871] = 474,
  [1872] = 475,
  [1873] = 476,
  [1874] = 439,
  [1875] = 477,
  [1876] = 440,
  [1877] = 503,
  [1878] = 1878,
  [1879] = 463,
  [1880] = 447,
  [1881] = 440,
  [1882] = 444,
  [1883] = 500,
  [1884] = 544,
  [1885] = 545,
  [1886] = 439,
  [1887] = 467,
  [1888] = 468,
  [1889] = 484,
  [1890] = 470,
  [1891] = 471,
  [1892] = 472,
  [1893] = 473,
  [1894] = 474,
  [1895] = 475,
  [1896] = 476,
  [1897] = 463,
  [1898] = 542,
  [1899] = 477,
  [1900] = 478,
  [1901] = 440,
  [1902] = 543,
  [1903] = 467,
  [1904] = 468,
  [1905] = 484,
  [1906] = 470,
  [1907] = 471,
  [1908] = 472,
  [1909] = 473,
  [1910] = 474,
  [1911] = 475,
  [1912] = 476,
  [1913] = 444,
  [1914] = 477,
  [1915] = 478,
  [1916] = 544,
  [1917] = 542,
  [1918] = 555,
  [1919] = 461,
  [1920] = 543,
  [1921] = 479,
  [1922] = 441,
  [1923] = 480,
  [1924] = 498,
  [1925] = 479,
  [1926] = 480,
  [1927] = 545,
  [1928] = 481,
  [1929] = 546,
  [1930] = 501,
  [1931] = 440,
  [1932] = 500,
  [1933] = 546,
  [1934] = 498,
  [1935] = 447,
  [1936] = 527,
  [1937] = 439,
  [1938] = 1878,
  [1939] = 502,
  [1940] = 440,
  [1941] = 447,
  [1942] = 439,
  [1943] = 441,
  [1944] = 527,
  [1945] = 605,
  [1946] = 467,
  [1947] = 468,
  [1948] = 484,
  [1949] = 470,
  [1950] = 471,
  [1951] = 472,
  [1952] = 473,
  [1953] = 498,
  [1954] = 474,
  [1955] = 475,
  [1956] = 476,
  [1957] = 477,
  [1958] = 478,
  [1959] = 479,
  [1960] = 480,
  [1961] = 481,
  [1962] = 544,
  [1963] = 501,
  [1964] = 503,
  [1965] = 502,
  [1966] = 498,
  [1967] = 447,
  [1968] = 545,
  [1969] = 498,
  [1970] = 510,
  [1971] = 501,
  [1972] = 498,
  [1973] = 546,
  [1974] = 542,
  [1975] = 543,
  [1976] = 502,
  [1977] = 1878,
  [1978] = 555,
  [1979] = 461,
  [1980] = 510,
  [1981] = 503,
  [1982] = 655,
  [1983] = 463,
  [1984] = 555,
  [1985] = 461,
  [1986] = 439,
  [1987] = 481,
  [1988] = 501,
  [1989] = 476,
  [1990] = 477,
  [1991] = 478,
  [1992] = 474,
  [1993] = 498,
  [1994] = 473,
  [1995] = 479,
  [1996] = 480,
  [1997] = 546,
  [1998] = 474,
  [1999] = 475,
  [2000] = 481,
  [2001] = 545,
  [2002] = 501,
  [2003] = 475,
  [2004] = 476,
  [2005] = 503,
  [2006] = 477,
  [2007] = 502,
  [2008] = 476,
  [2009] = 478,
  [2010] = 498,
  [2011] = 444,
  [2012] = 501,
  [2013] = 510,
  [2014] = 542,
  [2015] = 544,
  [2016] = 555,
  [2017] = 467,
  [2018] = 468,
  [2019] = 484,
  [2020] = 464,
  [2021] = 470,
  [2022] = 471,
  [2023] = 472,
  [2024] = 473,
  [2025] = 474,
  [2026] = 475,
  [2027] = 476,
  [2028] = 477,
  [2029] = 478,
  [2030] = 479,
  [2031] = 480,
  [2032] = 481,
  [2033] = 468,
  [2034] = 503,
  [2035] = 502,
  [2036] = 461,
  [2037] = 498,
  [2038] = 470,
  [2039] = 463,
  [2040] = 479,
  [2041] = 480,
  [2042] = 544,
  [2043] = 545,
  [2044] = 481,
  [2045] = 544,
  [2046] = 545,
  [2047] = 546,
  [2048] = 503,
  [2049] = 501,
  [2050] = 463,
  [2051] = 503,
  [2052] = 502,
  [2053] = 544,
  [2054] = 543,
  [2055] = 545,
  [2056] = 471,
  [2057] = 481,
  [2058] = 544,
  [2059] = 545,
  [2060] = 546,
  [2061] = 463,
  [2062] = 441,
  [2063] = 500,
  [2064] = 544,
  [2065] = 545,
  [2066] = 477,
  [2067] = 467,
  [2068] = 468,
  [2069] = 484,
  [2070] = 470,
  [2071] = 467,
  [2072] = 463,
  [2073] = 471,
  [2074] = 468,
  [2075] = 546,
  [2076] = 447,
  [2077] = 439,
  [2078] = 472,
  [2079] = 467,
  [2080] = 484,
  [2081] = 484,
  [2082] = 470,
  [2083] = 471,
  [2084] = 472,
  [2085] = 473,
  [2086] = 474,
  [2087] = 475,
  [2088] = 476,
  [2089] = 477,
  [2090] = 478,
  [2091] = 479,
  [2092] = 480,
  [2093] = 481,
  [2094] = 501,
  [2095] = 502,
  [2096] = 503,
  [2097] = 498,
  [2098] = 542,
  [2099] = 543,
  [2100] = 484,
  [2101] = 555,
  [2102] = 461,
  [2103] = 527,
  [2104] = 542,
  [2105] = 543,
  [2106] = 546,
  [2107] = 478,
  [2108] = 470,
  [2109] = 546,
  [2110] = 542,
  [2111] = 543,
  [2112] = 542,
  [2113] = 543,
  [2114] = 471,
  [2115] = 542,
  [2116] = 543,
  [2117] = 527,
  [2118] = 467,
  [2119] = 463,
  [2120] = 479,
  [2121] = 480,
  [2122] = 472,
  [2123] = 468,
  [2124] = 472,
  [2125] = 484,
  [2126] = 470,
  [2127] = 473,
  [2128] = 474,
  [2129] = 527,
  [2130] = 475,
  [2131] = 471,
  [2132] = 472,
  [2133] = 476,
  [2134] = 510,
  [2135] = 555,
  [2136] = 461,
  [2137] = 473,
  [2138] = 474,
  [2139] = 555,
  [2140] = 461,
  [2141] = 555,
  [2142] = 461,
  [2143] = 555,
  [2144] = 461,
  [2145] = 477,
  [2146] = 478,
  [2147] = 502,
  [2148] = 479,
  [2149] = 480,
  [2150] = 473,
  [2151] = 475,
  [2152] = 481,
  [2153] = 498,
  [2154] = 501,
  [2155] = 503,
  [2156] = 502,
  [2157] = 463,
  [2158] = 500,
  [2159] = 467,
  [2160] = 468,
  [2161] = 555,
  [2162] = 544,
  [2163] = 481,
  [2164] = 545,
  [2165] = 463,
  [2166] = 546,
  [2167] = 501,
  [2168] = 542,
  [2169] = 627,
  [2170] = 604,
  [2171] = 628,
  [2172] = 503,
  [2173] = 543,
  [2174] = 463,
  [2175] = 560,
  [2176] = 641,
  [2177] = 612,
  [2178] = 2178,
  [2179] = 561,
  [2180] = 502,
  [2181] = 467,
  [2182] = 468,
  [2183] = 484,
  [2184] = 470,
  [2185] = 471,
  [2186] = 472,
  [2187] = 473,
  [2188] = 474,
  [2189] = 546,
  [2190] = 501,
  [2191] = 475,
  [2192] = 476,
  [2193] = 463,
  [2194] = 477,
  [2195] = 564,
  [2196] = 467,
  [2197] = 468,
  [2198] = 484,
  [2199] = 470,
  [2200] = 471,
  [2201] = 472,
  [2202] = 473,
  [2203] = 474,
  [2204] = 475,
  [2205] = 476,
  [2206] = 477,
  [2207] = 478,
  [2208] = 479,
  [2209] = 480,
  [2210] = 481,
  [2211] = 501,
  [2212] = 659,
  [2213] = 503,
  [2214] = 502,
  [2215] = 559,
  [2216] = 562,
  [2217] = 566,
  [2218] = 478,
  [2219] = 648,
  [2220] = 616,
  [2221] = 591,
  [2222] = 674,
  [2223] = 568,
  [2224] = 544,
  [2225] = 545,
  [2226] = 2226,
  [2227] = 502,
  [2228] = 570,
  [2229] = 571,
  [2230] = 599,
  [2231] = 687,
  [2232] = 600,
  [2233] = 608,
  [2234] = 688,
  [2235] = 546,
  [2236] = 2236,
  [2237] = 544,
  [2238] = 545,
  [2239] = 617,
  [2240] = 689,
  [2241] = 567,
  [2242] = 657,
  [2243] = 618,
  [2244] = 529,
  [2245] = 694,
  [2246] = 2226,
  [2247] = 572,
  [2248] = 2226,
  [2249] = 460,
  [2250] = 589,
  [2251] = 542,
  [2252] = 543,
  [2253] = 695,
  [2254] = 601,
  [2255] = 498,
  [2256] = 467,
  [2257] = 479,
  [2258] = 649,
  [2259] = 480,
  [2260] = 468,
  [2261] = 484,
  [2262] = 470,
  [2263] = 542,
  [2264] = 543,
  [2265] = 575,
  [2266] = 481,
  [2267] = 576,
  [2268] = 666,
  [2269] = 471,
  [2270] = 577,
  [2271] = 579,
  [2272] = 651,
  [2273] = 624,
  [2274] = 582,
  [2275] = 2236,
  [2276] = 472,
  [2277] = 473,
  [2278] = 474,
  [2279] = 475,
  [2280] = 691,
  [2281] = 461,
  [2282] = 555,
  [2283] = 461,
  [2284] = 476,
  [2285] = 603,
  [2286] = 503,
  [2287] = 2236,
  [2288] = 477,
  [2289] = 578,
  [2290] = 478,
  [2291] = 479,
  [2292] = 480,
  [2293] = 555,
  [2294] = 461,
  [2295] = 2226,
  [2296] = 2236,
  [2297] = 2297,
  [2298] = 2298,
  [2299] = 2299,
  [2300] = 2300,
  [2301] = 2298,
  [2302] = 2297,
  [2303] = 501,
  [2304] = 2297,
  [2305] = 2305,
  [2306] = 2298,
  [2307] = 2297,
  [2308] = 503,
  [2309] = 2297,
  [2310] = 2310,
  [2311] = 2297,
  [2312] = 502,
  [2313] = 2300,
  [2314] = 2297,
  [2315] = 2305,
  [2316] = 2297,
  [2317] = 2297,
  [2318] = 2310,
  [2319] = 2310,
  [2320] = 2297,
  [2321] = 2300,
  [2322] = 2297,
  [2323] = 2305,
  [2324] = 2297,
  [2325] = 2300,
  [2326] = 2297,
  [2327] = 2298,
  [2328] = 2297,
  [2329] = 2297,
  [2330] = 2297,
  [2331] = 2297,
  [2332] = 2297,
  [2333] = 2297,
  [2334] = 2305,
  [2335] = 2310,
  [2336] = 2297,
  [2337] = 2297,
  [2338] = 2300,
  [2339] = 441,
  [2340] = 447,
  [2341] = 2341,
  [2342] = 2341,
  [2343] = 439,
  [2344] = 2341,
  [2345] = 2341,
  [2346] = 444,
  [2347] = 2299,
  [2348] = 2348,
  [2349] = 2341,
  [2350] = 2348,
  [2351] = 2351,
  [2352] = 2352,
  [2353] = 2353,
  [2354] = 2354,
  [2355] = 527,
  [2356] = 527,
  [2357] = 2357,
  [2358] = 527,
  [2359] = 498,
  [2360] = 441,
  [2361] = 2351,
  [2362] = 2362,
  [2363] = 691,
  [2364] = 2364,
  [2365] = 2365,
  [2366] = 2366,
  [2367] = 2367,
  [2368] = 2352,
  [2369] = 501,
  [2370] = 2370,
  [2371] = 2371,
  [2372] = 2372,
  [2373] = 2373,
  [2374] = 2374,
  [2375] = 2367,
  [2376] = 502,
  [2377] = 2354,
  [2378] = 2367,
  [2379] = 2379,
  [2380] = 2370,
  [2381] = 2381,
  [2382] = 503,
  [2383] = 2357,
  [2384] = 2384,
  [2385] = 2385,
  [2386] = 2386,
  [2387] = 2387,
  [2388] = 2386,
  [2389] = 2385,
  [2390] = 2390,
  [2391] = 2387,
  [2392] = 2387,
  [2393] = 2393,
  [2394] = 2390,
  [2395] = 2393,
  [2396] = 2386,
  [2397] = 2386,
  [2398] = 2398,
  [2399] = 2386,
  [2400] = 2386,
  [2401] = 2401,
  [2402] = 2385,
  [2403] = 2387,
  [2404] = 2386,
  [2405] = 2401,
  [2406] = 2390,
  [2407] = 2407,
  [2408] = 2385,
  [2409] = 2401,
  [2410] = 2385,
  [2411] = 2411,
  [2412] = 2387,
  [2413] = 2411,
  [2414] = 2401,
  [2415] = 2415,
  [2416] = 2401,
  [2417] = 2385,
  [2418] = 2390,
  [2419] = 2386,
  [2420] = 2401,
  [2421] = 2385,
  [2422] = 2386,
  [2423] = 2401,
  [2424] = 2385,
  [2425] = 2385,
  [2426] = 2387,
  [2427] = 2401,
  [2428] = 2386,
  [2429] = 2385,
  [2430] = 2390,
  [2431] = 2386,
  [2432] = 2386,
  [2433] = 2401,
  [2434] = 2385,
  [2435] = 2401,
  [2436] = 2385,
  [2437] = 2437,
  [2438] = 2401,
  [2439] = 2385,
  [2440] = 2401,
  [2441] = 2386,
  [2442] = 2385,
  [2443] = 2401,
  [2444] = 2385,
  [2445] = 2390,
  [2446] = 2401,
  [2447] = 2401,
  [2448] = 2386,
  [2449] = 2401,
  [2450] = 2386,
  [2451] = 2437,
  [2452] = 2401,
  [2453] = 2387,
  [2454] = 2385,
  [2455] = 2386,
  [2456] = 2401,
  [2457] = 2385,
  [2458] = 2387,
  [2459] = 2459,
  [2460] = 2390,
  [2461] = 2461,
  [2462] = 2387,
  [2463] = 2401,
  [2464] = 502,
  [2465] = 2401,
  [2466] = 2466,
  [2467] = 2390,
  [2468] = 2468,
  [2469] = 2385,
  [2470] = 2387,
  [2471] = 2385,
  [2472] = 2387,
  [2473] = 2390,
  [2474] = 2474,
  [2475] = 2387,
  [2476] = 2476,
  [2477] = 2390,
  [2478] = 2386,
  [2479] = 2387,
  [2480] = 2387,
  [2481] = 2390,
  [2482] = 2398,
  [2483] = 2390,
  [2484] = 2390,
  [2485] = 2401,
  [2486] = 2387,
  [2487] = 2407,
  [2488] = 2488,
  [2489] = 2461,
  [2490] = 2387,
  [2491] = 2491,
  [2492] = 2390,
  [2493] = 2386,
  [2494] = 2390,
  [2495] = 2491,
  [2496] = 2387,
  [2497] = 2390,
  [2498] = 2390,
  [2499] = 2387,
  [2500] = 2386,
  [2501] = 2390,
  [2502] = 2387,
  [2503] = 2387,
  [2504] = 2390,
  [2505] = 2386,
  [2506] = 2387,
  [2507] = 2474,
  [2508] = 2466,
  [2509] = 2390,
  [2510] = 2385,
  [2511] = 2390,
  [2512] = 2385,
  [2513] = 217,
  [2514] = 216,
  [2515] = 218,
  [2516] = 2516,
  [2517] = 2517,
  [2518] = 2518,
  [2519] = 2519,
  [2520] = 2520,
  [2521] = 2520,
  [2522] = 2522,
  [2523] = 2523,
  [2524] = 2522,
  [2525] = 2523,
  [2526] = 2526,
  [2527] = 2527,
  [2528] = 2528,
  [2529] = 2528,
  [2530] = 2528,
  [2531] = 2528,
  [2532] = 2528,
  [2533] = 2528,
  [2534] = 2528,
  [2535] = 2528,
  [2536] = 2528,
  [2537] = 2537,
  [2538] = 2528,
  [2539] = 2528,
  [2540] = 2528,
  [2541] = 2528,
  [2542] = 2528,
  [2543] = 2528,
  [2544] = 2528,
  [2545] = 2528,
  [2546] = 2528,
  [2547] = 2528,
  [2548] = 2528,
  [2549] = 2537,
  [2550] = 2550,
  [2551] = 2551,
  [2552] = 2552,
  [2553] = 2553,
  [2554] = 2554,
  [2555] = 2553,
  [2556] = 500,
  [2557] = 510,
  [2558] = 545,
  [2559] = 500,
  [2560] = 2560,
  [2561] = 2561,
  [2562] = 544,
  [2563] = 219,
  [2564] = 2564,
  [2565] = 2565,
  [2566] = 2566,
  [2567] = 2561,
  [2568] = 2561,
  [2569] = 546,
  [2570] = 2566,
  [2571] = 2571,
  [2572] = 2560,
  [2573] = 2573,
  [2574] = 2561,
  [2575] = 220,
  [2576] = 221,
  [2577] = 2560,
  [2578] = 2566,
  [2579] = 2573,
  [2580] = 2580,
  [2581] = 2573,
  [2582] = 2582,
  [2583] = 500,
  [2584] = 2584,
  [2585] = 2582,
  [2586] = 2586,
  [2587] = 2587,
  [2588] = 2588,
  [2589] = 2589,
  [2590] = 2586,
  [2591] = 2591,
  [2592] = 2592,
  [2593] = 2588,
  [2594] = 2589,
  [2595] = 2595,
  [2596] = 2582,
  [2597] = 2592,
  [2598] = 2592,
  [2599] = 2592,
  [2600] = 2582,
  [2601] = 2601,
  [2602] = 2602,
  [2603] = 2603,
  [2604] = 2604,
  [2605] = 2605,
  [2606] = 2606,
  [2607] = 2607,
  [2608] = 2608,
  [2609] = 2609,
  [2610] = 2610,
  [2611] = 2611,
  [2612] = 510,
  [2613] = 2613,
  [2614] = 2614,
  [2615] = 2615,
  [2616] = 2616,
  [2617] = 2617,
  [2618] = 510,
  [2619] = 2619,
  [2620] = 2620,
  [2621] = 500,
  [2622] = 510,
  [2623] = 2623,
  [2624] = 2624,
  [2625] = 2625,
  [2626] = 2626,
  [2627] = 2627,
  [2628] = 2601,
  [2629] = 2629,
  [2630] = 2608,
  [2631] = 2631,
  [2632] = 500,
  [2633] = 2633,
  [2634] = 2634,
  [2635] = 2635,
  [2636] = 2636,
  [2637] = 2637,
  [2638] = 2520,
  [2639] = 2639,
  [2640] = 2640,
  [2641] = 2641,
  [2642] = 2642,
  [2643] = 2640,
  [2644] = 2527,
  [2645] = 2634,
  [2646] = 2580,
  [2647] = 2605,
  [2648] = 2648,
  [2649] = 2565,
  [2650] = 2564,
  [2651] = 2651,
  [2652] = 2652,
  [2653] = 2653,
  [2654] = 2654,
  [2655] = 1551,
  [2656] = 2656,
  [2657] = 2657,
  [2658] = 2658,
  [2659] = 1535,
  [2660] = 2660,
  [2661] = 2639,
  [2662] = 2641,
  [2663] = 2640,
  [2664] = 2639,
  [2665] = 2641,
  [2666] = 2666,
  [2667] = 2667,
  [2668] = 2640,
  [2669] = 2669,
  [2670] = 1548,
  [2671] = 2667,
  [2672] = 2652,
  [2673] = 2673,
  [2674] = 2652,
  [2675] = 2675,
  [2676] = 2636,
  [2677] = 2653,
  [2678] = 2635,
  [2679] = 2610,
  [2680] = 2624,
  [2681] = 2627,
  [2682] = 2637,
  [2683] = 2683,
  [2684] = 2684,
  [2685] = 2685,
  [2686] = 2683,
  [2687] = 2687,
  [2688] = 2666,
  [2689] = 2640,
  [2690] = 2615,
  [2691] = 2691,
  [2692] = 2666,
  [2693] = 2687,
  [2694] = 2694,
  [2695] = 2669,
  [2696] = 2639,
  [2697] = 2666,
  [2698] = 2641,
  [2699] = 2666,
  [2700] = 2700,
  [2701] = 2640,
  [2702] = 2694,
  [2703] = 2652,
  [2704] = 2653,
  [2705] = 2705,
  [2706] = 2683,
  [2707] = 2707,
  [2708] = 2685,
  [2709] = 2687,
  [2710] = 2710,
  [2711] = 2694,
  [2712] = 2666,
  [2713] = 2700,
  [2714] = 2714,
  [2715] = 2715,
  [2716] = 2716,
  [2717] = 2667,
  [2718] = 2718,
  [2719] = 2637,
  [2720] = 2640,
  [2721] = 2714,
  [2722] = 2722,
  [2723] = 2714,
  [2724] = 2724,
  [2725] = 2637,
  [2726] = 2726,
  [2727] = 2613,
  [2728] = 2669,
  [2729] = 2694,
  [2730] = 2614,
  [2731] = 2731,
  [2732] = 2667,
  [2733] = 2637,
  [2734] = 2616,
  [2735] = 2619,
  [2736] = 2620,
  [2737] = 2669,
  [2738] = 2687,
  [2739] = 2739,
  [2740] = 2667,
  [2741] = 2637,
  [2742] = 2653,
  [2743] = 2667,
  [2744] = 2637,
  [2745] = 2716,
  [2746] = 2666,
  [2747] = 2667,
  [2748] = 2637,
  [2749] = 2714,
  [2750] = 2640,
  [2751] = 2667,
  [2752] = 2637,
  [2753] = 2715,
  [2754] = 2667,
  [2755] = 2637,
  [2756] = 2756,
  [2757] = 2667,
  [2758] = 2637,
  [2759] = 2666,
  [2760] = 2667,
  [2761] = 2637,
  [2762] = 2667,
  [2763] = 2637,
  [2764] = 2675,
  [2765] = 2683,
  [2766] = 2667,
  [2767] = 2637,
  [2768] = 2768,
  [2769] = 2667,
  [2770] = 2637,
  [2771] = 2667,
  [2772] = 2637,
  [2773] = 1539,
  [2774] = 2666,
  [2775] = 2667,
  [2776] = 2637,
  [2777] = 2777,
  [2778] = 2718,
  [2779] = 2667,
  [2780] = 2637,
  [2781] = 2781,
  [2782] = 2782,
  [2783] = 2667,
  [2784] = 2637,
  [2785] = 2667,
  [2786] = 2786,
  [2787] = 2667,
  [2788] = 2637,
  [2789] = 2789,
  [2790] = 2667,
  [2791] = 2637,
  [2792] = 2792,
  [2793] = 2640,
  [2794] = 2794,
  [2795] = 2795,
  [2796] = 2796,
  [2797] = 2797,
  [2798] = 2798,
  [2799] = 2796,
  [2800] = 2800,
  [2801] = 2794,
  [2802] = 2795,
  [2803] = 2796,
  [2804] = 2804,
  [2805] = 2796,
  [2806] = 2800,
  [2807] = 2794,
  [2808] = 2795,
  [2809] = 2809,
  [2810] = 2797,
  [2811] = 2800,
  [2812] = 2796,
  [2813] = 2800,
  [2814] = 2794,
  [2815] = 2795,
  [2816] = 2809,
  [2817] = 2809,
  [2818] = 2796,
  [2819] = 2800,
  [2820] = 2809,
  [2821] = 2795,
  [2822] = 2795,
  [2823] = 2797,
  [2824] = 2796,
  [2825] = 2800,
  [2826] = 2794,
  [2827] = 2795,
  [2828] = 2809,
  [2829] = 2829,
  [2830] = 2796,
  [2831] = 2800,
  [2832] = 2794,
  [2833] = 2795,
  [2834] = 2797,
  [2835] = 2835,
  [2836] = 2797,
  [2837] = 2837,
  [2838] = 2796,
  [2839] = 2800,
  [2840] = 2794,
  [2841] = 2795,
  [2842] = 2842,
  [2843] = 2796,
  [2844] = 2800,
  [2845] = 2794,
  [2846] = 2795,
  [2847] = 2591,
  [2848] = 2848,
  [2849] = 2849,
  [2850] = 2850,
  [2851] = 2796,
  [2852] = 2794,
  [2853] = 2794,
  [2854] = 2809,
  [2855] = 2795,
  [2856] = 2809,
  [2857] = 2797,
  [2858] = 2809,
  [2859] = 2859,
  [2860] = 2797,
  [2861] = 2861,
  [2862] = 2796,
  [2863] = 2863,
  [2864] = 2864,
  [2865] = 2809,
  [2866] = 2797,
  [2867] = 2867,
  [2868] = 2794,
  [2869] = 2864,
  [2870] = 2800,
  [2871] = 2871,
  [2872] = 2707,
  [2873] = 2797,
  [2874] = 2809,
  [2875] = 2875,
  [2876] = 2797,
  [2877] = 2710,
  [2878] = 2797,
  [2879] = 2859,
  [2880] = 2792,
  [2881] = 2797,
  [2882] = 2809,
  [2883] = 2883,
  [2884] = 2884,
  [2885] = 2809,
  [2886] = 2795,
  [2887] = 2797,
  [2888] = 2722,
  [2889] = 2724,
  [2890] = 2796,
  [2891] = 2797,
  [2892] = 2892,
  [2893] = 2796,
  [2894] = 2800,
  [2895] = 2794,
  [2896] = 2795,
  [2897] = 2798,
  [2898] = 2898,
  [2899] = 2800,
  [2900] = 2800,
  [2901] = 2809,
  [2902] = 2809,
  [2903] = 2863,
  [2904] = 2809,
  [2905] = 2905,
  [2906] = 2797,
  [2907] = 2796,
  [2908] = 2800,
  [2909] = 2794,
  [2910] = 2795,
  [2911] = 2797,
  [2912] = 2794,
  [2913] = 2809,
  [2914] = 2809,
  [2915] = 2915,
  [2916] = 2916,
  [2917] = 2795,
  [2918] = 2797,
  [2919] = 2809,
  [2920] = 2795,
  [2921] = 2796,
  [2922] = 2800,
  [2923] = 2867,
  [2924] = 2794,
  [2925] = 2795,
  [2926] = 2871,
  [2927] = 2927,
  [2928] = 2797,
  [2929] = 2929,
  [2930] = 2892,
  [2931] = 2864,
  [2932] = 2796,
  [2933] = 2800,
  [2934] = 2794,
  [2935] = 2795,
  [2936] = 2800,
  [2937] = 2796,
  [2938] = 2938,
  [2939] = 2794,
  [2940] = 2927,
  [2941] = 2797,
  [2942] = 2809,
  [2943] = 2796,
  [2944] = 2800,
  [2945] = 2794,
  [2946] = 2795,
  [2947] = 2797,
  [2948] = 2800,
  [2949] = 2915,
  [2950] = 2809,
  [2951] = 2809,
  [2952] = 2796,
  [2953] = 2800,
  [2954] = 2794,
  [2955] = 2795,
  [2956] = 2794,
  [2957] = 2875,
  [2958] = 2797,
  [2959] = 2795,
  [2960] = 2864,
  [2961] = 2961,
  [2962] = 2796,
  [2963] = 2800,
  [2964] = 2961,
  [2965] = 2794,
  [2966] = 2861,
  [2967] = 2967,
  [2968] = 2968,
  [2969] = 2969,
  [2970] = 2970,
  [2971] = 2971,
  [2972] = 2972,
  [2973] = 2973,
  [2974] = 2974,
  [2975] = 2975,
  [2976] = 2976,
  [2977] = 2973,
  [2978] = 2978,
  [2979] = 2979,
  [2980] = 2980,
  [2981] = 2981,
  [2982] = 2982,
  [2983] = 2973,
  [2984] = 2984,
  [2985] = 2972,
  [2986] = 2980,
  [2987] = 2987,
  [2988] = 2988,
  [2989] = 2978,
  [2990] = 2972,
  [2991] = 2991,
  [2992] = 2970,
  [2993] = 2993,
  [2994] = 2978,
  [2995] = 2987,
  [2996] = 2969,
  [2997] = 2997,
  [2998] = 2998,
  [2999] = 2967,
  [3000] = 3000,
  [3001] = 3001,
  [3002] = 2997,
  [3003] = 3003,
  [3004] = 2987,
  [3005] = 3005,
  [3006] = 3001,
  [3007] = 2971,
  [3008] = 3008,
  [3009] = 2971,
  [3010] = 3010,
  [3011] = 2972,
  [3012] = 2973,
  [3013] = 3001,
  [3014] = 3014,
  [3015] = 3015,
  [3016] = 2975,
  [3017] = 2980,
  [3018] = 3018,
  [3019] = 3019,
  [3020] = 3020,
  [3021] = 3021,
  [3022] = 2968,
  [3023] = 2993,
  [3024] = 3024,
  [3025] = 3025,
  [3026] = 3001,
  [3027] = 3027,
  [3028] = 3028,
  [3029] = 2982,
  [3030] = 2984,
  [3031] = 2981,
  [3032] = 3032,
  [3033] = 3033,
  [3034] = 2991,
  [3035] = 2972,
  [3036] = 3036,
  [3037] = 2975,
  [3038] = 2979,
  [3039] = 3025,
  [3040] = 3025,
  [3041] = 2972,
  [3042] = 3000,
  [3043] = 3005,
  [3044] = 2973,
  [3045] = 3045,
  [3046] = 2967,
  [3047] = 2980,
  [3048] = 2975,
  [3049] = 2970,
  [3050] = 2970,
  [3051] = 3051,
  [3052] = 3025,
  [3053] = 2972,
  [3054] = 2997,
  [3055] = 3055,
  [3056] = 2987,
  [3057] = 2972,
  [3058] = 2972,
  [3059] = 3059,
  [3060] = 2993,
  [3061] = 3061,
  [3062] = 2972,
  [3063] = 2987,
  [3064] = 2976,
  [3065] = 3065,
  [3066] = 2993,
  [3067] = 3067,
  [3068] = 3068,
  [3069] = 2968,
  [3070] = 2972,
  [3071] = 2972,
  [3072] = 2972,
  [3073] = 2978,
  [3074] = 2997,
  [3075] = 3010,
  [3076] = 2991,
  [3077] = 2971,
  [3078] = 2997,
  [3079] = 2972,
  [3080] = 3080,
  [3081] = 3080,
  [3082] = 3065,
  [3083] = 3024,
  [3084] = 2972,
  [3085] = 3032,
  [3086] = 3086,
  [3087] = 2967,
  [3088] = 3088,
  [3089] = 2993,
  [3090] = 3090,
  [3091] = 3091,
  [3092] = 2972,
  [3093] = 3093,
  [3094] = 2968,
  [3095] = 3095,
  [3096] = 2991,
  [3097] = 3091,
  [3098] = 2972,
  [3099] = 2972,
  [3100] = 2988,
  [3101] = 3093,
  [3102] = 3028,
  [3103] = 2967,
  [3104] = 2998,
  [3105] = 3105,
  [3106] = 3018,
  [3107] = 3045,
  [3108] = 3003,
  [3109] = 2972,
  [3110] = 2967,
  [3111] = 2967,
  [3112] = 2972,
  [3113] = 2967,
  [3114] = 2967,
  [3115] = 2967,
  [3116] = 2967,
  [3117] = 2967,
  [3118] = 2967,
  [3119] = 2967,
  [3120] = 2967,
  [3121] = 2967,
  [3122] = 2967,
  [3123] = 2967,
  [3124] = 2967,
  [3125] = 2967,
  [3126] = 2967,
  [3127] = 3090,
  [3128] = 3055,
  [3129] = 3088,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym__compound_statement,
  sym__expression,
  sym__primary_expression,
  sym__simple_statement,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym__compound_statement] = {.index = 0, .length = 12},
  [sym__expression] = {.index = 12, .length = 10},
  [sym__primary_expression] = {.index = 22, .length = 27},
  [sym__simple_statement] = {.index = 49, .length = 16},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    sym_async_function_definition,
    sym_block_tailed_statement,
    sym_class_definition,
    sym_decorated_definition,
    sym_do_statement,
    sym_for_js_statement,
    sym_for_statement,
    sym_function_definition,
    sym_if_statement,
    sym_try_statement,
    sym_while_statement,
    sym_with_statement,
  [12] =
    sym__primary_expression,
    sym_assignment,
    sym_augmented_assignment,
    sym_await,
    sym_binary_operator,
    sym_boolean_operator,
    sym_comparison_operator,
    sym_conditional_expression,
    sym_not_operator,
    sym_unary_operator,
  [22] =
    sym_anonymous_function,
    sym_attribute,
    sym_call,
    sym_concatenated_string,
    sym_dictionary,
    sym_dictionary_comprehension,
    sym_existential,
    sym_f_string,
    sym_false,
    sym_generator_expression,
    sym_identifier,
    sym_list,
    sym_list_comprehension,
    sym_new_expression,
    sym_none,
    sym_number,
    sym_parenthesized_expression,
    sym_regex,
    sym_set,
    sym_set_comprehension,
    sym_slice_call,
    sym_string,
    sym_subscript,
    sym_this,
    sym_true,
    sym_tuple,
    sym_verbatim,
  [49] =
    sym__expression,
    sym_assert_statement,
    sym_break_statement,
    sym_continue_statement,
    sym_debugger_statement,
    sym_delete_statement,
    sym_expression_list,
    sym_global_statement,
    sym_import_from_statement,
    sym_import_statement,
    sym_nonlocal_statement,
    sym_pass_statement,
    sym_raise_statement,
    sym_return_statement,
    sym_scoped_flag_statement,
    sym_yield_statement,
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(157);
      ADVANCE_MAP(
        '!', 62,
        '"', 36,
        '#', 360,
        '%', 230,
        '&', 236,
        '\'', 48,
        '(', 181,
        ')', 182,
        '*', 207,
        '+', 221,
        ',', 180,
        '-', 224,
        '.', 190,
        '/', 225,
        '0', 268,
        ':', 166,
        ';', 158,
        '<', 244,
        '=', 160,
        '>', 252,
        '?', 258,
        '@', 213,
        'F', 354,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(138);
      if (lookahead == ']') ADVANCE(184);
      if (lookahead == '^') ADVANCE(234);
      if (lookahead == 'a') ADVANCE(327);
      if (lookahead == 'c') ADVANCE(320);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'e') ADVANCE(321);
      if (lookahead == 'f') ADVANCE(318);
      if (lookahead == 'i') ADVANCE(311);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'o') ADVANCE(339);
      if (lookahead == 'w') ADVANCE(317);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '|') ADVANCE(232);
      if (lookahead == '}') ADVANCE(260);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(283);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(0);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (('$' <= lookahead && lookahead <= '_') ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 1:
      if (lookahead == '\n') SKIP(26);
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(26);
      if (lookahead == '\r') SKIP(1);
      END_STATE();
    case 3:
      if (lookahead == '\n') SKIP(25);
      END_STATE();
    case 4:
      if (lookahead == '\n') SKIP(25);
      if (lookahead == '\r') SKIP(3);
      END_STATE();
    case 5:
      if (lookahead == '\n') SKIP(27);
      END_STATE();
    case 6:
      if (lookahead == '\n') SKIP(27);
      if (lookahead == '\r') SKIP(5);
      END_STATE();
    case 7:
      if (lookahead == '\n') SKIP(28);
      END_STATE();
    case 8:
      if (lookahead == '\n') SKIP(28);
      if (lookahead == '\r') SKIP(7);
      END_STATE();
    case 9:
      if (lookahead == '\n') SKIP(33);
      END_STATE();
    case 10:
      if (lookahead == '\n') SKIP(33);
      if (lookahead == '\r') SKIP(9);
      END_STATE();
    case 11:
      if (lookahead == '\n') SKIP(30);
      END_STATE();
    case 12:
      if (lookahead == '\n') SKIP(30);
      if (lookahead == '\r') SKIP(11);
      END_STATE();
    case 13:
      if (lookahead == '\n') SKIP(31);
      END_STATE();
    case 14:
      if (lookahead == '\n') SKIP(31);
      if (lookahead == '\r') SKIP(13);
      END_STATE();
    case 15:
      if (lookahead == '\n') SKIP(32);
      END_STATE();
    case 16:
      if (lookahead == '\n') SKIP(32);
      if (lookahead == '\r') SKIP(15);
      END_STATE();
    case 17:
      if (lookahead == '\n') SKIP(29);
      END_STATE();
    case 18:
      if (lookahead == '\n') SKIP(29);
      if (lookahead == '\r') SKIP(17);
      END_STATE();
    case 19:
      if (lookahead == '\n') SKIP(34);
      END_STATE();
    case 20:
      if (lookahead == '\n') SKIP(34);
      if (lookahead == '\r') SKIP(19);
      END_STATE();
    case 21:
      if (lookahead == '\n') SKIP(46);
      END_STATE();
    case 22:
      if (lookahead == '\n') SKIP(46);
      if (lookahead == '\r') SKIP(21);
      END_STATE();
    case 23:
      if (lookahead == '\n') SKIP(47);
      END_STATE();
    case 24:
      if (lookahead == '\n') SKIP(47);
      if (lookahead == '\r') SKIP(23);
      END_STATE();
    case 25:
      ADVANCE_MAP(
        '!', 62,
        '"', 36,
        '#', 360,
        '%', 229,
        '&', 235,
        '\'', 48,
        '(', 181,
        ')', 182,
        '*', 208,
        '+', 220,
        ',', 180,
        '-', 222,
        '.', 190,
        '/', 226,
        '0', 268,
        ':', 166,
        ';', 158,
        '<', 245,
        '=', 160,
        '>', 253,
        '?', 258,
        '@', 213,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(4);
      if (lookahead == ']') ADVANCE(184);
      if (lookahead == '^') ADVANCE(233);
      if (lookahead == 'a') ADVANCE(328);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'i') ADVANCE(313);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'o') ADVANCE(339);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '|') ADVANCE(231);
      if (lookahead == '}') ADVANCE(260);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(354);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(293);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(25);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (('$' <= lookahead && lookahead <= '_') ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 26:
      ADVANCE_MAP(
        '!', 61,
        '"', 36,
        '#', 360,
        '%', 229,
        '&', 235,
        '\'', 48,
        '(', 181,
        ')', 182,
        '*', 208,
        '+', 220,
        ',', 180,
        '-', 222,
        '.', 190,
        '/', 226,
        '0', 268,
        ':', 166,
        ';', 158,
        '<', 245,
        '=', 68,
        '>', 253,
        '?', 258,
        '@', 213,
        'F', 354,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(2);
      if (lookahead == ']') ADVANCE(184);
      if (lookahead == '^') ADVANCE(233);
      if (lookahead == 'a') ADVANCE(328);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'f') ADVANCE(334);
      if (lookahead == 'i') ADVANCE(313);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'o') ADVANCE(339);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '|') ADVANCE(231);
      if (lookahead == '}') ADVANCE(260);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(285);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(26);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (('$' <= lookahead && lookahead <= '_') ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 27:
      ADVANCE_MAP(
        '!', 61,
        '"', 36,
        '#', 360,
        '%', 229,
        '&', 235,
        '\'', 48,
        '(', 181,
        '*', 208,
        '+', 220,
        ',', 180,
        '-', 222,
        '.', 190,
        '/', 226,
        '0', 268,
        ':', 166,
        ';', 158,
        '<', 245,
        '=', 68,
        '>', 253,
        '?', 258,
        '@', 213,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(6);
      if (lookahead == '^') ADVANCE(233);
      if (lookahead == 'a') ADVANCE(327);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'i') ADVANCE(313);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'o') ADVANCE(339);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '|') ADVANCE(231);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(354);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(291);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(27);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 28:
      ADVANCE_MAP(
        '!', 61,
        '"', 36,
        '#', 360,
        '%', 229,
        '&', 235,
        '\'', 48,
        '(', 181,
        '*', 208,
        '+', 220,
        ',', 180,
        '-', 222,
        '.', 190,
        '/', 226,
        '0', 268,
        ';', 158,
        '<', 245,
        '=', 68,
        '>', 253,
        '?', 258,
        '@', 213,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(8);
      if (lookahead == '^') ADVANCE(233);
      if (lookahead == 'a') ADVANCE(328);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'e') ADVANCE(322);
      if (lookahead == 'i') ADVANCE(313);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'o') ADVANCE(339);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '|') ADVANCE(231);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(354);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(292);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(28);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 29:
      ADVANCE_MAP(
        '"', 36,
        '#', 360,
        '%', 63,
        '&', 64,
        '\'', 48,
        '(', 181,
        ')', 182,
        '*', 58,
        '+', 65,
        '-', 66,
        '.', 126,
        '/', 59,
        '0', 268,
        '<', 60,
        '=', 159,
        '>', 76,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(18);
      if (lookahead == ']') ADVANCE(184);
      if (lookahead == '^') ADVANCE(69);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '|') ADVANCE(70);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(354);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(297);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(29);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= '_') ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 30:
      ADVANCE_MAP(
        '"', 36,
        '#', 360,
        '\'', 48,
        '(', 181,
        ')', 182,
        '*', 208,
        '+', 220,
        '-', 222,
        '.', 190,
        '0', 268,
        ':', 166,
        ';', 158,
        '?', 258,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(12);
      if (lookahead == ']') ADVANCE(184);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '}') ADVANCE(260);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(354);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(296);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(30);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 31:
      ADVANCE_MAP(
        '"', 36,
        '#', 360,
        '\'', 48,
        '(', 181,
        ')', 182,
        '+', 220,
        '-', 222,
        '.', 126,
        '0', 268,
        'F', 354,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(14);
      if (lookahead == ']') ADVANCE(184);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'f') ADVANCE(334);
      if (lookahead == 'i') ADVANCE(314);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '}') ADVANCE(260);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(290);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(31);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 32:
      ADVANCE_MAP(
        '"', 36,
        '#', 360,
        '\'', 48,
        '(', 181,
        '+', 220,
        '-', 222,
        '.', 126,
        '0', 268,
        ':', 166,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(16);
      if (lookahead == 'a') ADVANCE(342);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(354);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(298);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(32);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 33:
      ADVANCE_MAP(
        '"', 36,
        '#', 360,
        '\'', 48,
        '(', 181,
        '+', 220,
        '-', 222,
        '.', 126,
        '0', 268,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(10);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'i') ADVANCE(326);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(354);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(294);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(33);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 34:
      if (lookahead == '"') ADVANCE(36);
      if (lookahead == '#') ADVANCE(360);
      if (lookahead == '\'') ADVANCE(48);
      if (lookahead == '(') ADVANCE(181);
      if (lookahead == '.') ADVANCE(126);
      if (lookahead == '0') ADVANCE(268);
      if (lookahead == '[') ADVANCE(183);
      if (lookahead == '\\') SKIP(20);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'i') ADVANCE(329);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(354);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(295);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(34);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 35:
      ADVANCE_MAP(
        '"', 36,
        '\'', 48,
        'B', 35,
        'R', 35,
        'U', 35,
        'b', 35,
        'r', 35,
        'u', 35,
      );
      END_STATE();
    case 36:
      if (lookahead == '"') ADVANCE(263);
      if (lookahead == '\\') ADVANCE(131);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(37);
      END_STATE();
    case 37:
      if (lookahead == '"') ADVANCE(262);
      if (lookahead == '\\') ADVANCE(131);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(37);
      END_STATE();
    case 38:
      if (lookahead == '"') ADVANCE(262);
      if (lookahead != 0) ADVANCE(40);
      END_STATE();
    case 39:
      if (lookahead == '"') ADVANCE(266);
      if (lookahead == '\\') ADVANCE(133);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(41);
      END_STATE();
    case 40:
      if (lookahead == '"') ADVANCE(43);
      if (lookahead == '\\') ADVANCE(129);
      if (lookahead != 0) ADVANCE(40);
      END_STATE();
    case 41:
      if (lookahead == '"') ADVANCE(265);
      if (lookahead == '\\') ADVANCE(133);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(41);
      END_STATE();
    case 42:
      if (lookahead == '"') ADVANCE(265);
      if (lookahead != 0) ADVANCE(45);
      END_STATE();
    case 43:
      if (lookahead == '"') ADVANCE(38);
      if (lookahead != 0) ADVANCE(40);
      END_STATE();
    case 44:
      if (lookahead == '"') ADVANCE(42);
      if (lookahead != 0) ADVANCE(45);
      END_STATE();
    case 45:
      if (lookahead == '"') ADVANCE(44);
      if (lookahead == '\\') ADVANCE(135);
      if (lookahead != 0) ADVANCE(45);
      END_STATE();
    case 46:
      ADVANCE_MAP(
        '#', 360,
        '%', 63,
        '&', 64,
        '(', 181,
        ')', 182,
        '*', 58,
        '+', 65,
        ',', 180,
        '-', 67,
        '.', 189,
        '/', 59,
        ':', 166,
        '<', 60,
        '=', 159,
        '>', 76,
        '?', 258,
        '@', 213,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(22);
      if (lookahead == ']') ADVANCE(184);
      if (lookahead == '^') ADVANCE(69);
      if (lookahead == 'a') ADVANCE(115);
      if (lookahead == 'c') ADVANCE(97);
      if (lookahead == 'd') ADVANCE(84);
      if (lookahead == 'i') ADVANCE(102);
      if (lookahead == '|') ADVANCE(70);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0xfeff) SKIP(46);
      END_STATE();
    case 47:
      if (lookahead == '#') ADVANCE(360);
      if (lookahead == '(') ADVANCE(181);
      if (lookahead == ')') ADVANCE(182);
      if (lookahead == '*') ADVANCE(208);
      if (lookahead == ';') ADVANCE(158);
      if (lookahead == '\\') SKIP(24);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(355);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(47);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 48:
      if (lookahead == '\'') ADVANCE(264);
      if (lookahead == '\\') ADVANCE(132);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(49);
      END_STATE();
    case 49:
      if (lookahead == '\'') ADVANCE(262);
      if (lookahead == '\\') ADVANCE(132);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(49);
      END_STATE();
    case 50:
      if (lookahead == '\'') ADVANCE(262);
      if (lookahead != 0) ADVANCE(52);
      END_STATE();
    case 51:
      if (lookahead == '\'') ADVANCE(267);
      if (lookahead == '\\') ADVANCE(134);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(53);
      END_STATE();
    case 52:
      if (lookahead == '\'') ADVANCE(55);
      if (lookahead == '\\') ADVANCE(130);
      if (lookahead != 0) ADVANCE(52);
      END_STATE();
    case 53:
      if (lookahead == '\'') ADVANCE(265);
      if (lookahead == '\\') ADVANCE(134);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(53);
      END_STATE();
    case 54:
      if (lookahead == '\'') ADVANCE(265);
      if (lookahead != 0) ADVANCE(57);
      END_STATE();
    case 55:
      if (lookahead == '\'') ADVANCE(50);
      if (lookahead != 0) ADVANCE(52);
      END_STATE();
    case 56:
      if (lookahead == '\'') ADVANCE(54);
      if (lookahead != 0) ADVANCE(57);
      END_STATE();
    case 57:
      if (lookahead == '\'') ADVANCE(56);
      if (lookahead == '\\') ADVANCE(136);
      if (lookahead != 0) ADVANCE(57);
      END_STATE();
    case 58:
      if (lookahead == '*') ADVANCE(71);
      if (lookahead == '=') ADVANCE(169);
      END_STATE();
    case 59:
      if (lookahead == '/') ADVANCE(72);
      if (lookahead == '=') ADVANCE(170);
      END_STATE();
    case 60:
      if (lookahead == '<') ADVANCE(73);
      END_STATE();
    case 61:
      if (lookahead == '=') ADVANCE(249);
      END_STATE();
    case 62:
      if (lookahead == '=') ADVANCE(249);
      if (lookahead == 'a' ||
          lookahead == 'r' ||
          lookahead == 's') ADVANCE(261);
      END_STATE();
    case 63:
      if (lookahead == '=') ADVANCE(172);
      END_STATE();
    case 64:
      if (lookahead == '=') ADVANCE(177);
      END_STATE();
    case 65:
      if (lookahead == '=') ADVANCE(167);
      END_STATE();
    case 66:
      if (lookahead == '=') ADVANCE(168);
      END_STATE();
    case 67:
      if (lookahead == '=') ADVANCE(168);
      if (lookahead == '>') ADVANCE(165);
      END_STATE();
    case 68:
      if (lookahead == '=') ADVANCE(247);
      END_STATE();
    case 69:
      if (lookahead == '=') ADVANCE(178);
      END_STATE();
    case 70:
      if (lookahead == '=') ADVANCE(179);
      END_STATE();
    case 71:
      if (lookahead == '=') ADVANCE(173);
      END_STATE();
    case 72:
      if (lookahead == '=') ADVANCE(171);
      END_STATE();
    case 73:
      if (lookahead == '=') ADVANCE(175);
      END_STATE();
    case 74:
      if (lookahead == '=') ADVANCE(174);
      if (lookahead == '>') ADVANCE(75);
      END_STATE();
    case 75:
      if (lookahead == '=') ADVANCE(176);
      END_STATE();
    case 76:
      if (lookahead == '>') ADVANCE(74);
      END_STATE();
    case 77:
      if (lookahead == 'a') ADVANCE(104);
      END_STATE();
    case 78:
      if (lookahead == 'a') ADVANCE(98);
      END_STATE();
    case 79:
      if (lookahead == 'a') ADVANCE(116);
      END_STATE();
    case 80:
      if (lookahead == 'c') ADVANCE(161);
      END_STATE();
    case 81:
      if (lookahead == 'c') ADVANCE(87);
      END_STATE();
    case 82:
      if (lookahead == 'c') ADVANCE(88);
      END_STATE();
    case 83:
      if (lookahead == 'd') ADVANCE(216);
      END_STATE();
    case 84:
      if (lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 85:
      if (lookahead == 'e') ADVANCE(193);
      END_STATE();
    case 86:
      if (lookahead == 'e') ADVANCE(201);
      END_STATE();
    case 87:
      if (lookahead == 'e') ADVANCE(109);
      END_STATE();
    case 88:
      if (lookahead == 'e') ADVANCE(106);
      END_STATE();
    case 89:
      if (lookahead == 'f') ADVANCE(191);
      if (lookahead == 'm') ADVANCE(108);
      if (lookahead == 'n') ADVANCE(199);
      if (lookahead == 's') ADVANCE(254);
      END_STATE();
    case 90:
      if (lookahead == 'f') ADVANCE(163);
      END_STATE();
    case 91:
      if (lookahead == 'f') ADVANCE(256);
      END_STATE();
    case 92:
      if (lookahead == 'h') ADVANCE(94);
      END_STATE();
    case 93:
      if (lookahead == 'i') ADVANCE(101);
      if (lookahead == 'o') ADVANCE(111);
      END_STATE();
    case 94:
      if (lookahead == 'i') ADVANCE(99);
      END_STATE();
    case 95:
      if (lookahead == 'l') ADVANCE(114);
      if (lookahead == 'x') ADVANCE(81);
      END_STATE();
    case 96:
      if (lookahead == 'l') ADVANCE(121);
      END_STATE();
    case 97:
      if (lookahead == 'l') ADVANCE(79);
      END_STATE();
    case 98:
      if (lookahead == 'l') ADVANCE(96);
      END_STATE();
    case 99:
      if (lookahead == 'l') ADVANCE(86);
      END_STATE();
    case 100:
      if (lookahead == 'n') ADVANCE(83);
      if (lookahead == 's') ADVANCE(187);
      END_STATE();
    case 101:
      if (lookahead == 'n') ADVANCE(78);
      END_STATE();
    case 102:
      if (lookahead == 'n') ADVANCE(197);
      END_STATE();
    case 103:
      if (lookahead == 'n') ADVANCE(80);
      END_STATE();
    case 104:
      if (lookahead == 'n') ADVANCE(82);
      END_STATE();
    case 105:
      if (lookahead == 'o') ADVANCE(117);
      END_STATE();
    case 106:
      if (lookahead == 'o') ADVANCE(91);
      END_STATE();
    case 107:
      if (lookahead == 'o') ADVANCE(112);
      END_STATE();
    case 108:
      if (lookahead == 'p') ADVANCE(107);
      END_STATE();
    case 109:
      if (lookahead == 'p') ADVANCE(118);
      END_STATE();
    case 110:
      if (lookahead == 'r') ADVANCE(218);
      END_STATE();
    case 111:
      if (lookahead == 'r') ADVANCE(195);
      END_STATE();
    case 112:
      if (lookahead == 'r') ADVANCE(119);
      END_STATE();
    case 113:
      if (lookahead == 's') ADVANCE(211);
      END_STATE();
    case 114:
      if (lookahead == 's') ADVANCE(85);
      END_STATE();
    case 115:
      if (lookahead == 's') ADVANCE(122);
      END_STATE();
    case 116:
      if (lookahead == 's') ADVANCE(113);
      END_STATE();
    case 117:
      if (lookahead == 't') ADVANCE(214);
      END_STATE();
    case 118:
      if (lookahead == 't') ADVANCE(203);
      END_STATE();
    case 119:
      if (lookahead == 't') ADVANCE(185);
      END_STATE();
    case 120:
      if (lookahead == 't') ADVANCE(77);
      END_STATE();
    case 121:
      if (lookahead == 'y') ADVANCE(205);
      END_STATE();
    case 122:
      if (lookahead == 'y') ADVANCE(103);
      END_STATE();
    case 123:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(127);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(277);
      END_STATE();
    case 124:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(273);
      END_STATE();
    case 125:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(275);
      END_STATE();
    case 126:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(272);
      END_STATE();
    case 127:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(277);
      END_STATE();
    case 128:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(279);
      END_STATE();
    case 129:
      if (lookahead != 0) ADVANCE(40);
      END_STATE();
    case 130:
      if (lookahead != 0) ADVANCE(52);
      END_STATE();
    case 131:
      if (lookahead != 0) ADVANCE(37);
      END_STATE();
    case 132:
      if (lookahead != 0) ADVANCE(49);
      END_STATE();
    case 133:
      if (lookahead != 0) ADVANCE(41);
      END_STATE();
    case 134:
      if (lookahead != 0) ADVANCE(53);
      END_STATE();
    case 135:
      if (lookahead != 0) ADVANCE(45);
      END_STATE();
    case 136:
      if (lookahead != 0) ADVANCE(57);
      END_STATE();
    case 137:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(0);
      END_STATE();
    case 138:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(0);
      if (lookahead == '\r') SKIP(137);
      END_STATE();
    case 139:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(153);
      END_STATE();
    case 140:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(153);
      if (lookahead == '\r') SKIP(139);
      END_STATE();
    case 141:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(152);
      END_STATE();
    case 142:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(152);
      if (lookahead == '\r') SKIP(141);
      END_STATE();
    case 143:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(151);
      END_STATE();
    case 144:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(151);
      if (lookahead == '\r') SKIP(143);
      END_STATE();
    case 145:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(154);
      END_STATE();
    case 146:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(154);
      if (lookahead == '\r') SKIP(145);
      END_STATE();
    case 147:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(156);
      END_STATE();
    case 148:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(156);
      if (lookahead == '\r') SKIP(147);
      END_STATE();
    case 149:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(155);
      END_STATE();
    case 150:
      if (eof) ADVANCE(157);
      if (lookahead == '\n') SKIP(155);
      if (lookahead == '\r') SKIP(149);
      END_STATE();
    case 151:
      if (eof) ADVANCE(157);
      ADVANCE_MAP(
        '!', 62,
        '"', 36,
        '#', 360,
        '%', 230,
        '&', 236,
        '\'', 48,
        '(', 181,
        ')', 182,
        '*', 207,
        '+', 221,
        ',', 180,
        '-', 223,
        '.', 189,
        '/', 225,
        ':', 166,
        ';', 158,
        '<', 244,
        '=', 160,
        '>', 252,
        '?', 258,
        '@', 213,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(144);
      if (lookahead == ']') ADVANCE(184);
      if (lookahead == '^') ADVANCE(234);
      if (lookahead == 'a') ADVANCE(100);
      if (lookahead == 'd') ADVANCE(84);
      if (lookahead == 'e') ADVANCE(95);
      if (lookahead == 'f') ADVANCE(93);
      if (lookahead == 'i') ADVANCE(89);
      if (lookahead == 'n') ADVANCE(105);
      if (lookahead == 'o') ADVANCE(110);
      if (lookahead == 'w') ADVANCE(92);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '|') ADVANCE(232);
      if (lookahead == '}') ADVANCE(260);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(35);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0xfeff) SKIP(151);
      END_STATE();
    case 152:
      if (eof) ADVANCE(157);
      ADVANCE_MAP(
        '!', 61,
        '"', 36,
        '#', 360,
        '%', 230,
        '&', 236,
        '\'', 48,
        '(', 181,
        '*', 207,
        '+', 221,
        ',', 180,
        '-', 223,
        '.', 190,
        '/', 225,
        '0', 268,
        ';', 158,
        '<', 244,
        '=', 160,
        '>', 252,
        '?', 258,
        '@', 213,
        'F', 354,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(142);
      if (lookahead == '^') ADVANCE(234);
      if (lookahead == 'a') ADVANCE(328);
      if (lookahead == 'c') ADVANCE(320);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'f') ADVANCE(334);
      if (lookahead == 'i') ADVANCE(311);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'o') ADVANCE(339);
      if (lookahead == 'w') ADVANCE(317);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '|') ADVANCE(232);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(284);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(152);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('e' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 153:
      if (eof) ADVANCE(157);
      ADVANCE_MAP(
        '"', 36,
        '#', 360,
        '\'', 48,
        '(', 181,
        '+', 220,
        '-', 222,
        '.', 126,
        '0', 268,
        '@', 213,
        'F', 354,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(140);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'c') ADVANCE(320);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'f') ADVANCE(334);
      if (lookahead == 'i') ADVANCE(312);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'w') ADVANCE(317);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(289);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(153);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('e' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 154:
      if (eof) ADVANCE(157);
      ADVANCE_MAP(
        '"', 36,
        '#', 360,
        '\'', 48,
        '(', 181,
        '+', 220,
        '-', 222,
        '.', 126,
        '0', 268,
        '@', 213,
        'F', 354,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(146);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'c') ADVANCE(320);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'e') ADVANCE(321);
      if (lookahead == 'f') ADVANCE(318);
      if (lookahead == 'i') ADVANCE(312);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'w') ADVANCE(317);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(286);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(154);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 155:
      if (eof) ADVANCE(157);
      ADVANCE_MAP(
        '"', 36,
        '#', 360,
        '\'', 48,
        '(', 181,
        '+', 220,
        '-', 222,
        '.', 126,
        '0', 268,
        '@', 213,
        'F', 354,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(150);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'c') ADVANCE(320);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'e') ADVANCE(322);
      if (lookahead == 'f') ADVANCE(334);
      if (lookahead == 'i') ADVANCE(312);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'w') ADVANCE(317);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(287);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(155);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 156:
      if (eof) ADVANCE(157);
      ADVANCE_MAP(
        '"', 36,
        '#', 360,
        '\'', 48,
        '(', 181,
        '+', 220,
        '-', 222,
        '.', 126,
        '0', 268,
        '@', 213,
        'F', 354,
        '[', 183,
      );
      if (lookahead == '\\') SKIP(148);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'c') ADVANCE(320);
      if (lookahead == 'd') ADVANCE(306);
      if (lookahead == 'f') ADVANCE(318);
      if (lookahead == 'i') ADVANCE(312);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'w') ADVANCE(317);
      if (lookahead == '{') ADVANCE(259);
      if (lookahead == '~') ADVANCE(243);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(282);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(288);
      if (lookahead == 'B' ||
          lookahead == 'R' ||
          lookahead == 'U' ||
          lookahead == 'b' ||
          lookahead == 'r' ||
          lookahead == 'u') ADVANCE(281);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0) SKIP(156);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(270);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('e' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(247);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_async);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_async);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_def);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_def);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH_EQ);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_STAR_STAR_EQ);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT_EQ);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_import);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_import);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_as);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == 'y') ADVANCE(331);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(272);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_if);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_else);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_for);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_in);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == 's') ADVANCE(350);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == 's') ADVANCE(120);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_while);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_except);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_except);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_finally);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_finally);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(210);
      if (lookahead == '=') ADVANCE(169);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(209);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      if (lookahead == '=') ADVANCE(173);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_class);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_class);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_not);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_not);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_and);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_and);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_or);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_or);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '=') ADVANCE(167);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '=') ADVANCE(168);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '=') ADVANCE(168);
      if (lookahead == '>') ADVANCE(165);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(228);
      if (lookahead == '=') ADVANCE(170);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(227);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_SLASH_SLASH);
      if (lookahead == '=') ADVANCE(171);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(172);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(179);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(178);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '=') ADVANCE(177);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(175);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(174);
      if (lookahead == '>') ADVANCE(242);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '>') ADVANCE(241);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT);
      if (lookahead == '=') ADVANCE(176);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(238);
      if (lookahead == '=') ADVANCE(246);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(237);
      if (lookahead == '=') ADVANCE(246);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      if (lookahead == '=') ADVANCE(248);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_EQ);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      if (lookahead == '=') ADVANCE(250);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(anon_sym_BANG_EQ_EQ);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(251);
      if (lookahead == '>') ADVANCE(239);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(251);
      if (lookahead == '>') ADVANCE(240);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(anon_sym_is);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(anon_sym_is);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(anon_sym_instanceof);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(anon_sym_instanceof);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(aux_sym_interpolation_token1);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(sym_string);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(sym_string);
      if (lookahead == '"') ADVANCE(40);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(sym_string);
      if (lookahead == '\'') ADVANCE(52);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(sym_verbatim);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(sym_verbatim);
      if (lookahead == '"') ADVANCE(45);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(sym_verbatim);
      if (lookahead == '\'') ADVANCE(57);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(sym_number);
      ADVANCE_MAP(
        '.', 272,
        'B', 124,
        'b', 124,
        'E', 123,
        'e', 123,
        'O', 125,
        'o', 125,
        'X', 128,
        'x', 128,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(270);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(sym_number);
      ADVANCE_MAP(
        '.', 272,
        'B', 356,
        'b', 356,
        'E', 353,
        'e', 353,
        'O', 357,
        'o', 357,
        'X', 358,
        'x', 358,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '.') ADVANCE(272);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(123);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(270);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '.') ADVANCE(272);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(353);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(123);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(272);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(273);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(274);
      if (lookahead == '$' ||
          ('2' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(sym_number);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(275);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(sym_number);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(276);
      if (lookahead == '$' ||
          lookahead == '8' ||
          lookahead == '9' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(sym_number);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(277);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(sym_number);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(278);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(sym_number);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(279);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(sym_number);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(280);
      if (lookahead == '$' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '"', 36,
        '\'', 48,
        'F', 354,
        'f', 354,
        'V', 282,
        'v', 282,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '"', 39,
        '\'', 51,
        'V', 282,
        'v', 282,
        'B', 282,
        'F', 282,
        'R', 282,
        'U', 282,
        'b', 282,
        'f', 282,
        'r', 282,
        'u', 282,
      );
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'F', 354,
        'a', 327,
        'c', 320,
        'd', 306,
        'e', 321,
        'f', 318,
        'i', 311,
        'n', 333,
        'o', 339,
        'w', 317,
        'V', 282,
        'v', 282,
        0x200b, 283,
        0xfeff, 283,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'F', 354,
        'a', 328,
        'c', 320,
        'd', 306,
        'f', 334,
        'i', 311,
        'n', 333,
        'o', 339,
        'w', 317,
        'V', 282,
        'v', 282,
        0x200b, 284,
        0xfeff, 284,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('e' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'F', 354,
        'a', 328,
        'd', 306,
        'f', 334,
        'i', 313,
        'n', 333,
        'o', 339,
        'V', 282,
        'v', 282,
        0x200b, 285,
        0xfeff, 285,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'F', 354,
        'a', 345,
        'c', 320,
        'd', 306,
        'e', 321,
        'f', 318,
        'i', 312,
        'n', 333,
        'w', 317,
        'V', 282,
        'v', 282,
        0x200b, 286,
        0xfeff, 286,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'F', 354,
        'a', 345,
        'c', 320,
        'd', 306,
        'e', 322,
        'f', 334,
        'i', 312,
        'n', 333,
        'w', 317,
        'V', 282,
        'v', 282,
        0x200b, 287,
        0xfeff, 287,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'F', 354,
        'a', 345,
        'c', 320,
        'd', 306,
        'f', 318,
        'i', 312,
        'n', 333,
        'w', 317,
        'V', 282,
        'v', 282,
        0x200b, 288,
        0xfeff, 288,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('e' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'F', 354,
        'a', 345,
        'c', 320,
        'd', 306,
        'f', 334,
        'i', 312,
        'n', 333,
        'w', 317,
        'V', 282,
        'v', 282,
        0x200b, 289,
        0xfeff, 289,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('e' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'F', 354,
        'a', 345,
        'd', 306,
        'f', 334,
        'i', 314,
        'n', 333,
        'V', 282,
        'v', 282,
        0x200b, 290,
        0xfeff, 290,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'a', 327,
        'd', 306,
        'i', 313,
        'n', 333,
        'o', 339,
        'F', 354,
        'f', 354,
        'V', 282,
        'v', 282,
        0x200b, 291,
        0xfeff, 291,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'a', 328,
        'd', 306,
        'e', 322,
        'i', 313,
        'n', 333,
        'o', 339,
        'F', 354,
        'f', 354,
        'V', 282,
        'v', 282,
        0x200b, 292,
        0xfeff, 292,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'a', 328,
        'd', 306,
        'i', 313,
        'n', 333,
        'o', 339,
        'F', 354,
        'f', 354,
        'V', 282,
        'v', 282,
        0x200b, 293,
        0xfeff, 293,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'a', 345,
        'd', 306,
        'i', 326,
        'n', 333,
        'F', 354,
        'f', 354,
        'V', 282,
        'v', 282,
        0x200b, 294,
        0xfeff, 294,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'a', 345,
        'd', 306,
        'i', 329,
        'F', 354,
        'f', 354,
        'V', 282,
        'v', 282,
        0x200b, 295,
        0xfeff, 295,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'a', 345,
        'd', 306,
        'n', 333,
        'F', 354,
        'f', 354,
        'V', 282,
        'v', 282,
        0x200b, 296,
        0xfeff, 296,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'a', 345,
        'd', 306,
        'F', 354,
        'f', 354,
        'V', 282,
        'v', 282,
        0x200b, 297,
        0xfeff, 297,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '0', 269,
        'a', 342,
        'd', 306,
        'n', 333,
        'F', 354,
        'f', 354,
        'V', 282,
        'v', 282,
        0x200b, 298,
        0xfeff, 298,
        'B', 281,
        'R', 281,
        'U', 281,
        'b', 281,
        'r', 281,
        'u', 281,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(271);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(346);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(324);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(332);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(162);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(309);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(310);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'd') ADVANCE(217);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(315);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(194);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(202);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(338);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(335);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'f') ADVANCE(192);
      if (lookahead == 'm') ADVANCE(337);
      if (lookahead == 'n') ADVANCE(198);
      if (lookahead == 's') ADVANCE(255);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'f') ADVANCE(192);
      if (lookahead == 'm') ADVANCE(337);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'f') ADVANCE(192);
      if (lookahead == 'n') ADVANCE(198);
      if (lookahead == 's') ADVANCE(255);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'f') ADVANCE(192);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'f') ADVANCE(164);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'f') ADVANCE(257);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'h') ADVANCE(319);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        'i', 330,
        'o', 340,
        'V', 282,
        'v', 282,
        'B', 354,
        'F', 354,
        'R', 354,
        'U', 354,
        'b', 354,
        'f', 354,
        'r', 354,
        'u', 354,
      );
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'i') ADVANCE(325);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(299);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(344);
      if (lookahead == 'x') ADVANCE(303);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(344);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(351);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(323);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(308);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'm') ADVANCE(337);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(305);
      if (lookahead == 's') ADVANCE(188);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(305);
      if (lookahead == 's') ADVANCE(352);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(200);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(300);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(302);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'n') ADVANCE(304);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(347);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        'o', 340,
        'V', 282,
        'v', 282,
        'B', 354,
        'F', 354,
        'R', 354,
        'U', 354,
        'b', 354,
        'f', 354,
        'r', 354,
        'u', 354,
      );
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(316);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(341);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'p') ADVANCE(336);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'p') ADVANCE(348);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(219);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(196);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(349);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(188);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(212);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(307);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(352);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(343);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(215);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(204);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(186);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 't') ADVANCE(301);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'y') ADVANCE(206);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'y') ADVANCE(331);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(127);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(278);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        'V', 282,
        'v', 282,
        'B', 354,
        'F', 354,
        'R', 354,
        'U', 354,
        'b', 354,
        'f', 354,
        'r', 354,
        'u', 354,
      );
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 0x200b ||
          lookahead == 0xfeff) ADVANCE(355);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(359);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 356:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(274);
      if (lookahead == '$' ||
          ('2' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 357:
      ACCEPT_TOKEN(sym_identifier);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(276);
      if (lookahead == '$' ||
          lookahead == '8' ||
          lookahead == '9' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 358:
      ACCEPT_TOKEN(sym_identifier);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(280);
      if (lookahead == '$' ||
          ('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z') ||
          (0xaa <= lookahead && lookahead <= 0xffff)) ADVANCE(359);
      END_STATE();
    case 360:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(360);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (lookahead == 'F') ADVANCE(1);
      if (lookahead == 'N') ADVANCE(2);
      if (lookahead == 'T') ADVANCE(3);
      if (lookahead == '\\') SKIP(4);
      if (lookahead == '_') ADVANCE(5);
      if (lookahead == 'a') ADVANCE(6);
      if (lookahead == 'b') ADVANCE(7);
      if (lookahead == 'c') ADVANCE(8);
      if (lookahead == 'd') ADVANCE(9);
      if (lookahead == 'e') ADVANCE(10);
      if (lookahead == 'f') ADVANCE(11);
      if (lookahead == 'g') ADVANCE(12);
      if (lookahead == 'n') ADVANCE(13);
      if (lookahead == 'p') ADVANCE(14);
      if (lookahead == 'r') ADVANCE(15);
      if (lookahead == 't') ADVANCE(16);
      if (lookahead == 'v') ADVANCE(17);
      if (lookahead == 'w') ADVANCE(18);
      if (lookahead == 'y') ADVANCE(19);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ' ||
          lookahead == 0xa0 ||
          lookahead == 0x200b ||
          lookahead == 0xfeff) SKIP(0);
      END_STATE();
    case 1:
      if (lookahead == 'a') ADVANCE(20);
      END_STATE();
    case 2:
      if (lookahead == 'o') ADVANCE(21);
      END_STATE();
    case 3:
      if (lookahead == 'r') ADVANCE(22);
      END_STATE();
    case 4:
      if (lookahead == '\n') SKIP(0);
      if (lookahead == '\r') SKIP(23);
      END_STATE();
    case 5:
      if (lookahead == '_') ADVANCE(24);
      END_STATE();
    case 6:
      if (lookahead == 's') ADVANCE(25);
      if (lookahead == 'w') ADVANCE(26);
      END_STATE();
    case 7:
      if (lookahead == 'r') ADVANCE(27);
      END_STATE();
    case 8:
      if (lookahead == 'o') ADVANCE(28);
      END_STATE();
    case 9:
      if (lookahead == 'e') ADVANCE(29);
      if (lookahead == 'o') ADVANCE(30);
      END_STATE();
    case 10:
      if (lookahead == 'l') ADVANCE(31);
      END_STATE();
    case 11:
      if (lookahead == 'r') ADVANCE(32);
      END_STATE();
    case 12:
      if (lookahead == 'l') ADVANCE(33);
      END_STATE();
    case 13:
      if (lookahead == 'e') ADVANCE(34);
      if (lookahead == 'o') ADVANCE(35);
      END_STATE();
    case 14:
      if (lookahead == 'a') ADVANCE(36);
      END_STATE();
    case 15:
      if (lookahead == 'a') ADVANCE(37);
      if (lookahead == 'e') ADVANCE(38);
      END_STATE();
    case 16:
      if (lookahead == 'h') ADVANCE(39);
      if (lookahead == 'r') ADVANCE(40);
      if (lookahead == 'y') ADVANCE(41);
      END_STATE();
    case 17:
      if (lookahead == 'o') ADVANCE(42);
      END_STATE();
    case 18:
      if (lookahead == 'i') ADVANCE(43);
      END_STATE();
    case 19:
      if (lookahead == 'i') ADVANCE(44);
      END_STATE();
    case 20:
      if (lookahead == 'l') ADVANCE(45);
      END_STATE();
    case 21:
      if (lookahead == 'n') ADVANCE(46);
      END_STATE();
    case 22:
      if (lookahead == 'u') ADVANCE(47);
      END_STATE();
    case 23:
      if (lookahead == '\n') SKIP(0);
      END_STATE();
    case 24:
      if (lookahead == 'p') ADVANCE(48);
      END_STATE();
    case 25:
      if (lookahead == 's') ADVANCE(49);
      END_STATE();
    case 26:
      if (lookahead == 'a') ADVANCE(50);
      END_STATE();
    case 27:
      if (lookahead == 'e') ADVANCE(51);
      END_STATE();
    case 28:
      if (lookahead == 'n') ADVANCE(52);
      END_STATE();
    case 29:
      if (lookahead == 'b') ADVANCE(53);
      if (lookahead == 'l') ADVANCE(54);
      END_STATE();
    case 30:
      ACCEPT_TOKEN(anon_sym_do);
      END_STATE();
    case 31:
      if (lookahead == 'i') ADVANCE(55);
      END_STATE();
    case 32:
      if (lookahead == 'o') ADVANCE(56);
      END_STATE();
    case 33:
      if (lookahead == 'o') ADVANCE(57);
      END_STATE();
    case 34:
      if (lookahead == 'w') ADVANCE(58);
      END_STATE();
    case 35:
      if (lookahead == 'n') ADVANCE(59);
      END_STATE();
    case 36:
      if (lookahead == 's') ADVANCE(60);
      END_STATE();
    case 37:
      if (lookahead == 'i') ADVANCE(61);
      END_STATE();
    case 38:
      if (lookahead == 't') ADVANCE(62);
      END_STATE();
    case 39:
      if (lookahead == 'i') ADVANCE(63);
      END_STATE();
    case 40:
      if (lookahead == 'y') ADVANCE(64);
      END_STATE();
    case 41:
      if (lookahead == 'p') ADVANCE(65);
      END_STATE();
    case 42:
      if (lookahead == 'i') ADVANCE(66);
      END_STATE();
    case 43:
      if (lookahead == 't') ADVANCE(67);
      END_STATE();
    case 44:
      if (lookahead == 'e') ADVANCE(68);
      END_STATE();
    case 45:
      if (lookahead == 's') ADVANCE(69);
      END_STATE();
    case 46:
      if (lookahead == 'e') ADVANCE(70);
      END_STATE();
    case 47:
      if (lookahead == 'e') ADVANCE(71);
      END_STATE();
    case 48:
      if (lookahead == 'y') ADVANCE(72);
      END_STATE();
    case 49:
      if (lookahead == 'e') ADVANCE(73);
      END_STATE();
    case 50:
      if (lookahead == 'i') ADVANCE(74);
      END_STATE();
    case 51:
      if (lookahead == 'a') ADVANCE(75);
      END_STATE();
    case 52:
      if (lookahead == 't') ADVANCE(76);
      END_STATE();
    case 53:
      if (lookahead == 'u') ADVANCE(77);
      END_STATE();
    case 54:
      ACCEPT_TOKEN(anon_sym_del);
      END_STATE();
    case 55:
      if (lookahead == 'f') ADVANCE(78);
      END_STATE();
    case 56:
      if (lookahead == 'm') ADVANCE(79);
      END_STATE();
    case 57:
      if (lookahead == 'b') ADVANCE(80);
      END_STATE();
    case 58:
      ACCEPT_TOKEN(anon_sym_new);
      END_STATE();
    case 59:
      if (lookahead == 'l') ADVANCE(81);
      END_STATE();
    case 60:
      if (lookahead == 's') ADVANCE(82);
      END_STATE();
    case 61:
      if (lookahead == 's') ADVANCE(83);
      END_STATE();
    case 62:
      if (lookahead == 'u') ADVANCE(84);
      END_STATE();
    case 63:
      if (lookahead == 's') ADVANCE(85);
      END_STATE();
    case 64:
      ACCEPT_TOKEN(anon_sym_try);
      END_STATE();
    case 65:
      if (lookahead == 'e') ADVANCE(86);
      END_STATE();
    case 66:
      if (lookahead == 'd') ADVANCE(87);
      END_STATE();
    case 67:
      if (lookahead == 'h') ADVANCE(88);
      END_STATE();
    case 68:
      if (lookahead == 'l') ADVANCE(89);
      END_STATE();
    case 69:
      if (lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 70:
      ACCEPT_TOKEN(sym_none);
      END_STATE();
    case 71:
      ACCEPT_TOKEN(sym_true);
      END_STATE();
    case 72:
      if (lookahead == 't') ADVANCE(91);
      END_STATE();
    case 73:
      if (lookahead == 'r') ADVANCE(92);
      END_STATE();
    case 74:
      if (lookahead == 't') ADVANCE(93);
      END_STATE();
    case 75:
      if (lookahead == 'k') ADVANCE(94);
      END_STATE();
    case 76:
      if (lookahead == 'i') ADVANCE(95);
      END_STATE();
    case 77:
      if (lookahead == 'g') ADVANCE(96);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(anon_sym_elif);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_from);
      END_STATE();
    case 80:
      if (lookahead == 'a') ADVANCE(97);
      END_STATE();
    case 81:
      if (lookahead == 'o') ADVANCE(98);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(sym_pass_statement);
      END_STATE();
    case 83:
      if (lookahead == 'e') ADVANCE(99);
      END_STATE();
    case 84:
      if (lookahead == 'r') ADVANCE(100);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(sym_this);
      END_STATE();
    case 86:
      if (lookahead == 'o') ADVANCE(101);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_void);
      END_STATE();
    case 88:
      ACCEPT_TOKEN(anon_sym_with);
      END_STATE();
    case 89:
      if (lookahead == 'd') ADVANCE(102);
      END_STATE();
    case 90:
      ACCEPT_TOKEN(sym_false);
      END_STATE();
    case 91:
      if (lookahead == 'h') ADVANCE(103);
      END_STATE();
    case 92:
      if (lookahead == 't') ADVANCE(104);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_await);
      END_STATE();
    case 94:
      ACCEPT_TOKEN(sym_break_statement);
      END_STATE();
    case 95:
      if (lookahead == 'n') ADVANCE(105);
      END_STATE();
    case 96:
      if (lookahead == 'g') ADVANCE(106);
      END_STATE();
    case 97:
      if (lookahead == 'l') ADVANCE(107);
      END_STATE();
    case 98:
      if (lookahead == 'c') ADVANCE(108);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(anon_sym_raise);
      END_STATE();
    case 100:
      if (lookahead == 'n') ADVANCE(109);
      END_STATE();
    case 101:
      if (lookahead == 'f') ADVANCE(110);
      END_STATE();
    case 102:
      ACCEPT_TOKEN(anon_sym_yield);
      END_STATE();
    case 103:
      if (lookahead == 'o') ADVANCE(111);
      END_STATE();
    case 104:
      ACCEPT_TOKEN(anon_sym_assert);
      END_STATE();
    case 105:
      if (lookahead == 'u') ADVANCE(112);
      END_STATE();
    case 106:
      if (lookahead == 'e') ADVANCE(113);
      END_STATE();
    case 107:
      ACCEPT_TOKEN(anon_sym_global);
      END_STATE();
    case 108:
      if (lookahead == 'a') ADVANCE(114);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(anon_sym_return);
      END_STATE();
    case 110:
      ACCEPT_TOKEN(anon_sym_typeof);
      END_STATE();
    case 111:
      if (lookahead == 'n') ADVANCE(115);
      END_STATE();
    case 112:
      if (lookahead == 'e') ADVANCE(116);
      END_STATE();
    case 113:
      if (lookahead == 'r') ADVANCE(117);
      END_STATE();
    case 114:
      if (lookahead == 'l') ADVANCE(118);
      END_STATE();
    case 115:
      if (lookahead == '_') ADVANCE(119);
      END_STATE();
    case 116:
      ACCEPT_TOKEN(sym_continue_statement);
      END_STATE();
    case 117:
      ACCEPT_TOKEN(sym_debugger_statement);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(anon_sym_nonlocal);
      END_STATE();
    case 119:
      if (lookahead == '_') ADVANCE(120);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(anon_sym___python__);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 153, .external_lex_state = 2},
  [2] = {.lex_state = 152, .external_lex_state = 3},
  [3] = {.lex_state = 152, .external_lex_state = 2},
  [4] = {.lex_state = 152, .external_lex_state = 2},
  [5] = {.lex_state = 152, .external_lex_state = 3},
  [6] = {.lex_state = 152, .external_lex_state = 2},
  [7] = {.lex_state = 152, .external_lex_state = 3},
  [8] = {.lex_state = 152, .external_lex_state = 2},
  [9] = {.lex_state = 152, .external_lex_state = 3},
  [10] = {.lex_state = 153, .external_lex_state = 3},
  [11] = {.lex_state = 153, .external_lex_state = 3},
  [12] = {.lex_state = 153, .external_lex_state = 2},
  [13] = {.lex_state = 153, .external_lex_state = 3},
  [14] = {.lex_state = 153, .external_lex_state = 2},
  [15] = {.lex_state = 153, .external_lex_state = 3},
  [16] = {.lex_state = 153, .external_lex_state = 3},
  [17] = {.lex_state = 153, .external_lex_state = 3},
  [18] = {.lex_state = 153, .external_lex_state = 3},
  [19] = {.lex_state = 153, .external_lex_state = 3},
  [20] = {.lex_state = 153, .external_lex_state = 3},
  [21] = {.lex_state = 153, .external_lex_state = 3},
  [22] = {.lex_state = 153, .external_lex_state = 3},
  [23] = {.lex_state = 153, .external_lex_state = 3},
  [24] = {.lex_state = 153, .external_lex_state = 3},
  [25] = {.lex_state = 153, .external_lex_state = 3},
  [26] = {.lex_state = 153, .external_lex_state = 3},
  [27] = {.lex_state = 153, .external_lex_state = 3},
  [28] = {.lex_state = 153, .external_lex_state = 3},
  [29] = {.lex_state = 153, .external_lex_state = 3},
  [30] = {.lex_state = 153, .external_lex_state = 3},
  [31] = {.lex_state = 153, .external_lex_state = 3},
  [32] = {.lex_state = 153, .external_lex_state = 3},
  [33] = {.lex_state = 153, .external_lex_state = 3},
  [34] = {.lex_state = 153, .external_lex_state = 3},
  [35] = {.lex_state = 153, .external_lex_state = 3},
  [36] = {.lex_state = 153, .external_lex_state = 3},
  [37] = {.lex_state = 153, .external_lex_state = 3},
  [38] = {.lex_state = 153, .external_lex_state = 3},
  [39] = {.lex_state = 153, .external_lex_state = 3},
  [40] = {.lex_state = 153, .external_lex_state = 3},
  [41] = {.lex_state = 152, .external_lex_state = 2},
  [42] = {.lex_state = 152, .external_lex_state = 3},
  [43] = {.lex_state = 26, .external_lex_state = 2},
  [44] = {.lex_state = 26, .external_lex_state = 2},
  [45] = {.lex_state = 152, .external_lex_state = 2},
  [46] = {.lex_state = 152, .external_lex_state = 3},
  [47] = {.lex_state = 152, .external_lex_state = 2},
  [48] = {.lex_state = 152, .external_lex_state = 3},
  [49] = {.lex_state = 152, .external_lex_state = 3},
  [50] = {.lex_state = 25, .external_lex_state = 2},
  [51] = {.lex_state = 25, .external_lex_state = 2},
  [52] = {.lex_state = 25, .external_lex_state = 2},
  [53] = {.lex_state = 25, .external_lex_state = 2},
  [54] = {.lex_state = 152, .external_lex_state = 2},
  [55] = {.lex_state = 152, .external_lex_state = 2},
  [56] = {.lex_state = 152, .external_lex_state = 2},
  [57] = {.lex_state = 152, .external_lex_state = 2},
  [58] = {.lex_state = 152, .external_lex_state = 3},
  [59] = {.lex_state = 26, .external_lex_state = 2},
  [60] = {.lex_state = 152, .external_lex_state = 3},
  [61] = {.lex_state = 152, .external_lex_state = 2},
  [62] = {.lex_state = 152, .external_lex_state = 3},
  [63] = {.lex_state = 152, .external_lex_state = 3},
  [64] = {.lex_state = 26, .external_lex_state = 2},
  [65] = {.lex_state = 25, .external_lex_state = 2},
  [66] = {.lex_state = 27, .external_lex_state = 2},
  [67] = {.lex_state = 25, .external_lex_state = 2},
  [68] = {.lex_state = 25, .external_lex_state = 2},
  [69] = {.lex_state = 27, .external_lex_state = 2},
  [70] = {.lex_state = 25, .external_lex_state = 2},
  [71] = {.lex_state = 152, .external_lex_state = 3},
  [72] = {.lex_state = 152, .external_lex_state = 2},
  [73] = {.lex_state = 26, .external_lex_state = 2},
  [74] = {.lex_state = 26, .external_lex_state = 2},
  [75] = {.lex_state = 25, .external_lex_state = 4},
  [76] = {.lex_state = 25, .external_lex_state = 4},
  [77] = {.lex_state = 26, .external_lex_state = 2},
  [78] = {.lex_state = 26, .external_lex_state = 2},
  [79] = {.lex_state = 25, .external_lex_state = 2},
  [80] = {.lex_state = 28, .external_lex_state = 2},
  [81] = {.lex_state = 28, .external_lex_state = 2},
  [82] = {.lex_state = 25, .external_lex_state = 2},
  [83] = {.lex_state = 26, .external_lex_state = 2},
  [84] = {.lex_state = 26, .external_lex_state = 2},
  [85] = {.lex_state = 28, .external_lex_state = 2},
  [86] = {.lex_state = 25, .external_lex_state = 2},
  [87] = {.lex_state = 25, .external_lex_state = 2},
  [88] = {.lex_state = 25, .external_lex_state = 4},
  [89] = {.lex_state = 25, .external_lex_state = 4},
  [90] = {.lex_state = 28, .external_lex_state = 2},
  [91] = {.lex_state = 26, .external_lex_state = 2},
  [92] = {.lex_state = 25, .external_lex_state = 2},
  [93] = {.lex_state = 25, .external_lex_state = 2},
  [94] = {.lex_state = 26, .external_lex_state = 2},
  [95] = {.lex_state = 26, .external_lex_state = 2},
  [96] = {.lex_state = 26, .external_lex_state = 2},
  [97] = {.lex_state = 33, .external_lex_state = 5},
  [98] = {.lex_state = 25, .external_lex_state = 2},
  [99] = {.lex_state = 33, .external_lex_state = 5},
  [100] = {.lex_state = 25, .external_lex_state = 2},
  [101] = {.lex_state = 26, .external_lex_state = 2},
  [102] = {.lex_state = 26, .external_lex_state = 2},
  [103] = {.lex_state = 25, .external_lex_state = 2},
  [104] = {.lex_state = 33, .external_lex_state = 5},
  [105] = {.lex_state = 26, .external_lex_state = 2},
  [106] = {.lex_state = 33, .external_lex_state = 5},
  [107] = {.lex_state = 33, .external_lex_state = 5},
  [108] = {.lex_state = 33, .external_lex_state = 5},
  [109] = {.lex_state = 33, .external_lex_state = 5},
  [110] = {.lex_state = 25, .external_lex_state = 2},
  [111] = {.lex_state = 27, .external_lex_state = 2},
  [112] = {.lex_state = 26, .external_lex_state = 2},
  [113] = {.lex_state = 26, .external_lex_state = 2},
  [114] = {.lex_state = 33, .external_lex_state = 5},
  [115] = {.lex_state = 25, .external_lex_state = 2},
  [116] = {.lex_state = 25, .external_lex_state = 2},
  [117] = {.lex_state = 26, .external_lex_state = 2},
  [118] = {.lex_state = 25, .external_lex_state = 2},
  [119] = {.lex_state = 25, .external_lex_state = 2},
  [120] = {.lex_state = 25, .external_lex_state = 2},
  [121] = {.lex_state = 26, .external_lex_state = 2},
  [122] = {.lex_state = 25, .external_lex_state = 2},
  [123] = {.lex_state = 25, .external_lex_state = 2},
  [124] = {.lex_state = 27, .external_lex_state = 2},
  [125] = {.lex_state = 27, .external_lex_state = 2},
  [126] = {.lex_state = 28, .external_lex_state = 2},
  [127] = {.lex_state = 26, .external_lex_state = 2},
  [128] = {.lex_state = 25, .external_lex_state = 2},
  [129] = {.lex_state = 26, .external_lex_state = 2},
  [130] = {.lex_state = 25, .external_lex_state = 4},
  [131] = {.lex_state = 25, .external_lex_state = 2},
  [132] = {.lex_state = 25, .external_lex_state = 2},
  [133] = {.lex_state = 26, .external_lex_state = 2},
  [134] = {.lex_state = 26, .external_lex_state = 2},
  [135] = {.lex_state = 25, .external_lex_state = 2},
  [136] = {.lex_state = 25, .external_lex_state = 2},
  [137] = {.lex_state = 25, .external_lex_state = 4},
  [138] = {.lex_state = 27, .external_lex_state = 2},
  [139] = {.lex_state = 25, .external_lex_state = 4},
  [140] = {.lex_state = 25, .external_lex_state = 2},
  [141] = {.lex_state = 27, .external_lex_state = 2},
  [142] = {.lex_state = 25, .external_lex_state = 4},
  [143] = {.lex_state = 25, .external_lex_state = 2},
  [144] = {.lex_state = 25, .external_lex_state = 2},
  [145] = {.lex_state = 25, .external_lex_state = 2},
  [146] = {.lex_state = 28, .external_lex_state = 2},
  [147] = {.lex_state = 28, .external_lex_state = 2},
  [148] = {.lex_state = 28, .external_lex_state = 2},
  [149] = {.lex_state = 26, .external_lex_state = 2},
  [150] = {.lex_state = 25, .external_lex_state = 2},
  [151] = {.lex_state = 26, .external_lex_state = 2},
  [152] = {.lex_state = 27, .external_lex_state = 2},
  [153] = {.lex_state = 26, .external_lex_state = 2},
  [154] = {.lex_state = 25, .external_lex_state = 2},
  [155] = {.lex_state = 33, .external_lex_state = 5},
  [156] = {.lex_state = 33, .external_lex_state = 5},
  [157] = {.lex_state = 33, .external_lex_state = 5},
  [158] = {.lex_state = 33, .external_lex_state = 5},
  [159] = {.lex_state = 33, .external_lex_state = 5},
  [160] = {.lex_state = 33, .external_lex_state = 5},
  [161] = {.lex_state = 33, .external_lex_state = 5},
  [162] = {.lex_state = 33, .external_lex_state = 5},
  [163] = {.lex_state = 33, .external_lex_state = 5},
  [164] = {.lex_state = 33, .external_lex_state = 5},
  [165] = {.lex_state = 33, .external_lex_state = 5},
  [166] = {.lex_state = 33, .external_lex_state = 5},
  [167] = {.lex_state = 33, .external_lex_state = 5},
  [168] = {.lex_state = 33, .external_lex_state = 5},
  [169] = {.lex_state = 33, .external_lex_state = 5},
  [170] = {.lex_state = 33, .external_lex_state = 5},
  [171] = {.lex_state = 33, .external_lex_state = 5},
  [172] = {.lex_state = 25, .external_lex_state = 2},
  [173] = {.lex_state = 33, .external_lex_state = 5},
  [174] = {.lex_state = 33, .external_lex_state = 5},
  [175] = {.lex_state = 33, .external_lex_state = 5},
  [176] = {.lex_state = 33, .external_lex_state = 5},
  [177] = {.lex_state = 33, .external_lex_state = 5},
  [178] = {.lex_state = 33, .external_lex_state = 5},
  [179] = {.lex_state = 33, .external_lex_state = 5},
  [180] = {.lex_state = 26, .external_lex_state = 2},
  [181] = {.lex_state = 33, .external_lex_state = 5},
  [182] = {.lex_state = 25, .external_lex_state = 4},
  [183] = {.lex_state = 33, .external_lex_state = 5},
  [184] = {.lex_state = 33, .external_lex_state = 5},
  [185] = {.lex_state = 28, .external_lex_state = 2},
  [186] = {.lex_state = 28, .external_lex_state = 2},
  [187] = {.lex_state = 33, .external_lex_state = 5},
  [188] = {.lex_state = 33, .external_lex_state = 5},
  [189] = {.lex_state = 28, .external_lex_state = 2},
  [190] = {.lex_state = 33, .external_lex_state = 5},
  [191] = {.lex_state = 33, .external_lex_state = 5},
  [192] = {.lex_state = 26, .external_lex_state = 2},
  [193] = {.lex_state = 28, .external_lex_state = 2},
  [194] = {.lex_state = 33, .external_lex_state = 5},
  [195] = {.lex_state = 33, .external_lex_state = 5},
  [196] = {.lex_state = 33, .external_lex_state = 5},
  [197] = {.lex_state = 33, .external_lex_state = 5},
  [198] = {.lex_state = 33, .external_lex_state = 5},
  [199] = {.lex_state = 33, .external_lex_state = 5},
  [200] = {.lex_state = 33, .external_lex_state = 5},
  [201] = {.lex_state = 25, .external_lex_state = 4},
  [202] = {.lex_state = 25, .external_lex_state = 4},
  [203] = {.lex_state = 33, .external_lex_state = 5},
  [204] = {.lex_state = 25, .external_lex_state = 4},
  [205] = {.lex_state = 33, .external_lex_state = 5},
  [206] = {.lex_state = 33, .external_lex_state = 5},
  [207] = {.lex_state = 33, .external_lex_state = 5},
  [208] = {.lex_state = 33, .external_lex_state = 5},
  [209] = {.lex_state = 33, .external_lex_state = 5},
  [210] = {.lex_state = 33, .external_lex_state = 5},
  [211] = {.lex_state = 25, .external_lex_state = 2},
  [212] = {.lex_state = 33, .external_lex_state = 5},
  [213] = {.lex_state = 26, .external_lex_state = 2},
  [214] = {.lex_state = 33, .external_lex_state = 5},
  [215] = {.lex_state = 33, .external_lex_state = 5},
  [216] = {.lex_state = 152, .external_lex_state = 2},
  [217] = {.lex_state = 152, .external_lex_state = 2},
  [218] = {.lex_state = 152, .external_lex_state = 2},
  [219] = {.lex_state = 152, .external_lex_state = 3},
  [220] = {.lex_state = 152, .external_lex_state = 3},
  [221] = {.lex_state = 152, .external_lex_state = 3},
  [222] = {.lex_state = 152, .external_lex_state = 3},
  [223] = {.lex_state = 152, .external_lex_state = 2},
  [224] = {.lex_state = 152, .external_lex_state = 3},
  [225] = {.lex_state = 152, .external_lex_state = 2},
  [226] = {.lex_state = 152, .external_lex_state = 3},
  [227] = {.lex_state = 152, .external_lex_state = 2},
  [228] = {.lex_state = 152, .external_lex_state = 3},
  [229] = {.lex_state = 152, .external_lex_state = 2},
  [230] = {.lex_state = 152, .external_lex_state = 2},
  [231] = {.lex_state = 25, .external_lex_state = 2},
  [232] = {.lex_state = 25, .external_lex_state = 4},
  [233] = {.lex_state = 152, .external_lex_state = 3},
  [234] = {.lex_state = 152, .external_lex_state = 2},
  [235] = {.lex_state = 152, .external_lex_state = 2},
  [236] = {.lex_state = 152, .external_lex_state = 3},
  [237] = {.lex_state = 152, .external_lex_state = 3},
  [238] = {.lex_state = 152, .external_lex_state = 3},
  [239] = {.lex_state = 152, .external_lex_state = 2},
  [240] = {.lex_state = 28, .external_lex_state = 2},
  [241] = {.lex_state = 33, .external_lex_state = 4},
  [242] = {.lex_state = 33, .external_lex_state = 4},
  [243] = {.lex_state = 33, .external_lex_state = 4},
  [244] = {.lex_state = 33, .external_lex_state = 4},
  [245] = {.lex_state = 33, .external_lex_state = 4},
  [246] = {.lex_state = 33, .external_lex_state = 4},
  [247] = {.lex_state = 33, .external_lex_state = 4},
  [248] = {.lex_state = 33, .external_lex_state = 4},
  [249] = {.lex_state = 33, .external_lex_state = 4},
  [250] = {.lex_state = 33, .external_lex_state = 4},
  [251] = {.lex_state = 33, .external_lex_state = 4},
  [252] = {.lex_state = 33, .external_lex_state = 4},
  [253] = {.lex_state = 33, .external_lex_state = 4},
  [254] = {.lex_state = 33, .external_lex_state = 4},
  [255] = {.lex_state = 33, .external_lex_state = 4},
  [256] = {.lex_state = 33, .external_lex_state = 4},
  [257] = {.lex_state = 33, .external_lex_state = 4},
  [258] = {.lex_state = 33, .external_lex_state = 4},
  [259] = {.lex_state = 33, .external_lex_state = 2},
  [260] = {.lex_state = 30, .external_lex_state = 5},
  [261] = {.lex_state = 30, .external_lex_state = 5},
  [262] = {.lex_state = 30, .external_lex_state = 5},
  [263] = {.lex_state = 30, .external_lex_state = 5},
  [264] = {.lex_state = 30, .external_lex_state = 5},
  [265] = {.lex_state = 30, .external_lex_state = 5},
  [266] = {.lex_state = 30, .external_lex_state = 5},
  [267] = {.lex_state = 30, .external_lex_state = 5},
  [268] = {.lex_state = 30, .external_lex_state = 5},
  [269] = {.lex_state = 30, .external_lex_state = 5},
  [270] = {.lex_state = 30, .external_lex_state = 5},
  [271] = {.lex_state = 30, .external_lex_state = 5},
  [272] = {.lex_state = 30, .external_lex_state = 5},
  [273] = {.lex_state = 30, .external_lex_state = 5},
  [274] = {.lex_state = 30, .external_lex_state = 5},
  [275] = {.lex_state = 30, .external_lex_state = 5},
  [276] = {.lex_state = 30, .external_lex_state = 5},
  [277] = {.lex_state = 30, .external_lex_state = 5},
  [278] = {.lex_state = 30, .external_lex_state = 5},
  [279] = {.lex_state = 30, .external_lex_state = 5},
  [280] = {.lex_state = 30, .external_lex_state = 5},
  [281] = {.lex_state = 30, .external_lex_state = 5},
  [282] = {.lex_state = 30, .external_lex_state = 5},
  [283] = {.lex_state = 30, .external_lex_state = 5},
  [284] = {.lex_state = 30, .external_lex_state = 5},
  [285] = {.lex_state = 30, .external_lex_state = 5},
  [286] = {.lex_state = 30, .external_lex_state = 5},
  [287] = {.lex_state = 30, .external_lex_state = 5},
  [288] = {.lex_state = 30, .external_lex_state = 5},
  [289] = {.lex_state = 30, .external_lex_state = 5},
  [290] = {.lex_state = 30, .external_lex_state = 5},
  [291] = {.lex_state = 30, .external_lex_state = 5},
  [292] = {.lex_state = 30, .external_lex_state = 5},
  [293] = {.lex_state = 30, .external_lex_state = 5},
  [294] = {.lex_state = 30, .external_lex_state = 5},
  [295] = {.lex_state = 30, .external_lex_state = 5},
  [296] = {.lex_state = 30, .external_lex_state = 5},
  [297] = {.lex_state = 30, .external_lex_state = 5},
  [298] = {.lex_state = 30, .external_lex_state = 5},
  [299] = {.lex_state = 30, .external_lex_state = 5},
  [300] = {.lex_state = 30, .external_lex_state = 5},
  [301] = {.lex_state = 30, .external_lex_state = 5},
  [302] = {.lex_state = 30, .external_lex_state = 5},
  [303] = {.lex_state = 30, .external_lex_state = 5},
  [304] = {.lex_state = 30, .external_lex_state = 5},
  [305] = {.lex_state = 30, .external_lex_state = 5},
  [306] = {.lex_state = 30, .external_lex_state = 5},
  [307] = {.lex_state = 30, .external_lex_state = 5},
  [308] = {.lex_state = 30, .external_lex_state = 5},
  [309] = {.lex_state = 30, .external_lex_state = 5},
  [310] = {.lex_state = 30, .external_lex_state = 5},
  [311] = {.lex_state = 30, .external_lex_state = 5},
  [312] = {.lex_state = 30, .external_lex_state = 5},
  [313] = {.lex_state = 30, .external_lex_state = 5},
  [314] = {.lex_state = 30, .external_lex_state = 5},
  [315] = {.lex_state = 30, .external_lex_state = 5},
  [316] = {.lex_state = 30, .external_lex_state = 5},
  [317] = {.lex_state = 30, .external_lex_state = 5},
  [318] = {.lex_state = 30, .external_lex_state = 5},
  [319] = {.lex_state = 30, .external_lex_state = 5},
  [320] = {.lex_state = 30, .external_lex_state = 5},
  [321] = {.lex_state = 30, .external_lex_state = 5},
  [322] = {.lex_state = 30, .external_lex_state = 5},
  [323] = {.lex_state = 30, .external_lex_state = 5},
  [324] = {.lex_state = 30, .external_lex_state = 5},
  [325] = {.lex_state = 30, .external_lex_state = 5},
  [326] = {.lex_state = 30, .external_lex_state = 5},
  [327] = {.lex_state = 30, .external_lex_state = 5},
  [328] = {.lex_state = 30, .external_lex_state = 5},
  [329] = {.lex_state = 30, .external_lex_state = 5},
  [330] = {.lex_state = 30, .external_lex_state = 5},
  [331] = {.lex_state = 30, .external_lex_state = 5},
  [332] = {.lex_state = 30, .external_lex_state = 5},
  [333] = {.lex_state = 30, .external_lex_state = 5},
  [334] = {.lex_state = 30, .external_lex_state = 5},
  [335] = {.lex_state = 30, .external_lex_state = 5},
  [336] = {.lex_state = 30, .external_lex_state = 5},
  [337] = {.lex_state = 30, .external_lex_state = 5},
  [338] = {.lex_state = 30, .external_lex_state = 5},
  [339] = {.lex_state = 30, .external_lex_state = 5},
  [340] = {.lex_state = 30, .external_lex_state = 5},
  [341] = {.lex_state = 30, .external_lex_state = 5},
  [342] = {.lex_state = 30, .external_lex_state = 5},
  [343] = {.lex_state = 30, .external_lex_state = 5},
  [344] = {.lex_state = 30, .external_lex_state = 5},
  [345] = {.lex_state = 30, .external_lex_state = 5},
  [346] = {.lex_state = 30, .external_lex_state = 5},
  [347] = {.lex_state = 30, .external_lex_state = 5},
  [348] = {.lex_state = 30, .external_lex_state = 5},
  [349] = {.lex_state = 30, .external_lex_state = 5},
  [350] = {.lex_state = 30, .external_lex_state = 5},
  [351] = {.lex_state = 30, .external_lex_state = 5},
  [352] = {.lex_state = 30, .external_lex_state = 5},
  [353] = {.lex_state = 30, .external_lex_state = 5},
  [354] = {.lex_state = 30, .external_lex_state = 5},
  [355] = {.lex_state = 30, .external_lex_state = 5},
  [356] = {.lex_state = 30, .external_lex_state = 5},
  [357] = {.lex_state = 30, .external_lex_state = 5},
  [358] = {.lex_state = 30, .external_lex_state = 5},
  [359] = {.lex_state = 30, .external_lex_state = 5},
  [360] = {.lex_state = 30, .external_lex_state = 5},
  [361] = {.lex_state = 30, .external_lex_state = 5},
  [362] = {.lex_state = 30, .external_lex_state = 5},
  [363] = {.lex_state = 30, .external_lex_state = 5},
  [364] = {.lex_state = 30, .external_lex_state = 5},
  [365] = {.lex_state = 30, .external_lex_state = 5},
  [366] = {.lex_state = 30, .external_lex_state = 5},
  [367] = {.lex_state = 30, .external_lex_state = 5},
  [368] = {.lex_state = 30, .external_lex_state = 5},
  [369] = {.lex_state = 30, .external_lex_state = 5},
  [370] = {.lex_state = 30, .external_lex_state = 5},
  [371] = {.lex_state = 30, .external_lex_state = 5},
  [372] = {.lex_state = 30, .external_lex_state = 5},
  [373] = {.lex_state = 30, .external_lex_state = 5},
  [374] = {.lex_state = 30, .external_lex_state = 5},
  [375] = {.lex_state = 30, .external_lex_state = 5},
  [376] = {.lex_state = 30, .external_lex_state = 5},
  [377] = {.lex_state = 30, .external_lex_state = 5},
  [378] = {.lex_state = 30, .external_lex_state = 5},
  [379] = {.lex_state = 30, .external_lex_state = 5},
  [380] = {.lex_state = 30, .external_lex_state = 5},
  [381] = {.lex_state = 30, .external_lex_state = 5},
  [382] = {.lex_state = 30, .external_lex_state = 5},
  [383] = {.lex_state = 30, .external_lex_state = 5},
  [384] = {.lex_state = 30, .external_lex_state = 5},
  [385] = {.lex_state = 30, .external_lex_state = 5},
  [386] = {.lex_state = 30, .external_lex_state = 5},
  [387] = {.lex_state = 30, .external_lex_state = 5},
  [388] = {.lex_state = 30, .external_lex_state = 5},
  [389] = {.lex_state = 30, .external_lex_state = 5},
  [390] = {.lex_state = 30, .external_lex_state = 5},
  [391] = {.lex_state = 30, .external_lex_state = 5},
  [392] = {.lex_state = 30, .external_lex_state = 5},
  [393] = {.lex_state = 30, .external_lex_state = 5},
  [394] = {.lex_state = 30, .external_lex_state = 5},
  [395] = {.lex_state = 30, .external_lex_state = 5},
  [396] = {.lex_state = 30, .external_lex_state = 5},
  [397] = {.lex_state = 30, .external_lex_state = 5},
  [398] = {.lex_state = 30, .external_lex_state = 5},
  [399] = {.lex_state = 30, .external_lex_state = 5},
  [400] = {.lex_state = 30, .external_lex_state = 5},
  [401] = {.lex_state = 30, .external_lex_state = 5},
  [402] = {.lex_state = 30, .external_lex_state = 5},
  [403] = {.lex_state = 30, .external_lex_state = 5},
  [404] = {.lex_state = 30, .external_lex_state = 5},
  [405] = {.lex_state = 30, .external_lex_state = 5},
  [406] = {.lex_state = 30, .external_lex_state = 5},
  [407] = {.lex_state = 30, .external_lex_state = 5},
  [408] = {.lex_state = 30, .external_lex_state = 5},
  [409] = {.lex_state = 30, .external_lex_state = 5},
  [410] = {.lex_state = 30, .external_lex_state = 5},
  [411] = {.lex_state = 30, .external_lex_state = 5},
  [412] = {.lex_state = 30, .external_lex_state = 5},
  [413] = {.lex_state = 30, .external_lex_state = 5},
  [414] = {.lex_state = 30, .external_lex_state = 5},
  [415] = {.lex_state = 30, .external_lex_state = 5},
  [416] = {.lex_state = 30, .external_lex_state = 5},
  [417] = {.lex_state = 30, .external_lex_state = 5},
  [418] = {.lex_state = 30, .external_lex_state = 5},
  [419] = {.lex_state = 30, .external_lex_state = 5},
  [420] = {.lex_state = 30, .external_lex_state = 5},
  [421] = {.lex_state = 30, .external_lex_state = 5},
  [422] = {.lex_state = 30, .external_lex_state = 5},
  [423] = {.lex_state = 30, .external_lex_state = 5},
  [424] = {.lex_state = 30, .external_lex_state = 5},
  [425] = {.lex_state = 30, .external_lex_state = 5},
  [426] = {.lex_state = 30, .external_lex_state = 5},
  [427] = {.lex_state = 30, .external_lex_state = 5},
  [428] = {.lex_state = 30, .external_lex_state = 5},
  [429] = {.lex_state = 30, .external_lex_state = 5},
  [430] = {.lex_state = 30, .external_lex_state = 5},
  [431] = {.lex_state = 30, .external_lex_state = 5},
  [432] = {.lex_state = 30, .external_lex_state = 5},
  [433] = {.lex_state = 30, .external_lex_state = 5},
  [434] = {.lex_state = 30, .external_lex_state = 5},
  [435] = {.lex_state = 30, .external_lex_state = 5},
  [436] = {.lex_state = 30, .external_lex_state = 2},
  [437] = {.lex_state = 30, .external_lex_state = 2},
  [438] = {.lex_state = 152, .external_lex_state = 2},
  [439] = {.lex_state = 152, .external_lex_state = 3},
  [440] = {.lex_state = 152, .external_lex_state = 2},
  [441] = {.lex_state = 152, .external_lex_state = 3},
  [442] = {.lex_state = 152, .external_lex_state = 2},
  [443] = {.lex_state = 152, .external_lex_state = 2},
  [444] = {.lex_state = 152, .external_lex_state = 2},
  [445] = {.lex_state = 152, .external_lex_state = 2},
  [446] = {.lex_state = 152, .external_lex_state = 3},
  [447] = {.lex_state = 152, .external_lex_state = 2},
  [448] = {.lex_state = 152, .external_lex_state = 2},
  [449] = {.lex_state = 152, .external_lex_state = 3},
  [450] = {.lex_state = 152, .external_lex_state = 3},
  [451] = {.lex_state = 152, .external_lex_state = 2},
  [452] = {.lex_state = 152, .external_lex_state = 2},
  [453] = {.lex_state = 152, .external_lex_state = 3},
  [454] = {.lex_state = 152, .external_lex_state = 3},
  [455] = {.lex_state = 152, .external_lex_state = 3},
  [456] = {.lex_state = 152, .external_lex_state = 3},
  [457] = {.lex_state = 152, .external_lex_state = 3},
  [458] = {.lex_state = 152, .external_lex_state = 2},
  [459] = {.lex_state = 152, .external_lex_state = 3},
  [460] = {.lex_state = 152, .external_lex_state = 6},
  [461] = {.lex_state = 152, .external_lex_state = 3},
  [462] = {.lex_state = 30, .external_lex_state = 2},
  [463] = {.lex_state = 152, .external_lex_state = 2},
  [464] = {.lex_state = 152, .external_lex_state = 2},
  [465] = {.lex_state = 30, .external_lex_state = 2},
  [466] = {.lex_state = 152, .external_lex_state = 3},
  [467] = {.lex_state = 152, .external_lex_state = 2},
  [468] = {.lex_state = 152, .external_lex_state = 2},
  [469] = {.lex_state = 152, .external_lex_state = 6},
  [470] = {.lex_state = 152, .external_lex_state = 2},
  [471] = {.lex_state = 152, .external_lex_state = 2},
  [472] = {.lex_state = 152, .external_lex_state = 2},
  [473] = {.lex_state = 152, .external_lex_state = 2},
  [474] = {.lex_state = 152, .external_lex_state = 2},
  [475] = {.lex_state = 152, .external_lex_state = 2},
  [476] = {.lex_state = 152, .external_lex_state = 2},
  [477] = {.lex_state = 152, .external_lex_state = 2},
  [478] = {.lex_state = 152, .external_lex_state = 2},
  [479] = {.lex_state = 152, .external_lex_state = 2},
  [480] = {.lex_state = 152, .external_lex_state = 2},
  [481] = {.lex_state = 152, .external_lex_state = 2},
  [482] = {.lex_state = 152, .external_lex_state = 3},
  [483] = {.lex_state = 152, .external_lex_state = 3},
  [484] = {.lex_state = 152, .external_lex_state = 3},
  [485] = {.lex_state = 152, .external_lex_state = 3},
  [486] = {.lex_state = 152, .external_lex_state = 3},
  [487] = {.lex_state = 152, .external_lex_state = 3},
  [488] = {.lex_state = 152, .external_lex_state = 3},
  [489] = {.lex_state = 152, .external_lex_state = 3},
  [490] = {.lex_state = 152, .external_lex_state = 3},
  [491] = {.lex_state = 152, .external_lex_state = 3},
  [492] = {.lex_state = 152, .external_lex_state = 3},
  [493] = {.lex_state = 152, .external_lex_state = 3},
  [494] = {.lex_state = 152, .external_lex_state = 3},
  [495] = {.lex_state = 152, .external_lex_state = 3},
  [496] = {.lex_state = 152, .external_lex_state = 3},
  [497] = {.lex_state = 152, .external_lex_state = 4},
  [498] = {.lex_state = 152, .external_lex_state = 3},
  [499] = {.lex_state = 30, .external_lex_state = 2},
  [500] = {.lex_state = 152, .external_lex_state = 2},
  [501] = {.lex_state = 152, .external_lex_state = 2},
  [502] = {.lex_state = 152, .external_lex_state = 2},
  [503] = {.lex_state = 152, .external_lex_state = 2},
  [504] = {.lex_state = 30, .external_lex_state = 2},
  [505] = {.lex_state = 30, .external_lex_state = 2},
  [506] = {.lex_state = 152, .external_lex_state = 3},
  [507] = {.lex_state = 152, .external_lex_state = 2},
  [508] = {.lex_state = 152, .external_lex_state = 3},
  [509] = {.lex_state = 152, .external_lex_state = 3},
  [510] = {.lex_state = 152, .external_lex_state = 3},
  [511] = {.lex_state = 152, .external_lex_state = 3},
  [512] = {.lex_state = 30, .external_lex_state = 2},
  [513] = {.lex_state = 152, .external_lex_state = 6},
  [514] = {.lex_state = 30, .external_lex_state = 2},
  [515] = {.lex_state = 152, .external_lex_state = 4},
  [516] = {.lex_state = 152, .external_lex_state = 6},
  [517] = {.lex_state = 152, .external_lex_state = 2},
  [518] = {.lex_state = 152, .external_lex_state = 4},
  [519] = {.lex_state = 30, .external_lex_state = 2},
  [520] = {.lex_state = 30, .external_lex_state = 2},
  [521] = {.lex_state = 30, .external_lex_state = 2},
  [522] = {.lex_state = 30, .external_lex_state = 2},
  [523] = {.lex_state = 30, .external_lex_state = 2},
  [524] = {.lex_state = 30, .external_lex_state = 2},
  [525] = {.lex_state = 30, .external_lex_state = 2},
  [526] = {.lex_state = 152, .external_lex_state = 6},
  [527] = {.lex_state = 152, .external_lex_state = 2},
  [528] = {.lex_state = 30, .external_lex_state = 2},
  [529] = {.lex_state = 152, .external_lex_state = 4},
  [530] = {.lex_state = 30, .external_lex_state = 2},
  [531] = {.lex_state = 30, .external_lex_state = 2},
  [532] = {.lex_state = 152, .external_lex_state = 6},
  [533] = {.lex_state = 30, .external_lex_state = 2},
  [534] = {.lex_state = 152, .external_lex_state = 6},
  [535] = {.lex_state = 152, .external_lex_state = 4},
  [536] = {.lex_state = 30, .external_lex_state = 2},
  [537] = {.lex_state = 152, .external_lex_state = 4},
  [538] = {.lex_state = 152, .external_lex_state = 3},
  [539] = {.lex_state = 152, .external_lex_state = 2},
  [540] = {.lex_state = 152, .external_lex_state = 3},
  [541] = {.lex_state = 30, .external_lex_state = 2},
  [542] = {.lex_state = 152, .external_lex_state = 2},
  [543] = {.lex_state = 152, .external_lex_state = 2},
  [544] = {.lex_state = 152, .external_lex_state = 2},
  [545] = {.lex_state = 152, .external_lex_state = 2},
  [546] = {.lex_state = 152, .external_lex_state = 2},
  [547] = {.lex_state = 152, .external_lex_state = 3},
  [548] = {.lex_state = 152, .external_lex_state = 3},
  [549] = {.lex_state = 152, .external_lex_state = 3},
  [550] = {.lex_state = 152, .external_lex_state = 3},
  [551] = {.lex_state = 152, .external_lex_state = 3},
  [552] = {.lex_state = 152, .external_lex_state = 3},
  [553] = {.lex_state = 30, .external_lex_state = 2},
  [554] = {.lex_state = 152, .external_lex_state = 4},
  [555] = {.lex_state = 152, .external_lex_state = 2},
  [556] = {.lex_state = 152, .external_lex_state = 2},
  [557] = {.lex_state = 152, .external_lex_state = 3},
  [558] = {.lex_state = 152, .external_lex_state = 2},
  [559] = {.lex_state = 152, .external_lex_state = 2},
  [560] = {.lex_state = 152, .external_lex_state = 2},
  [561] = {.lex_state = 152, .external_lex_state = 2},
  [562] = {.lex_state = 152, .external_lex_state = 2},
  [563] = {.lex_state = 152, .external_lex_state = 2},
  [564] = {.lex_state = 152, .external_lex_state = 2},
  [565] = {.lex_state = 152, .external_lex_state = 2},
  [566] = {.lex_state = 152, .external_lex_state = 2},
  [567] = {.lex_state = 152, .external_lex_state = 2},
  [568] = {.lex_state = 152, .external_lex_state = 2},
  [569] = {.lex_state = 152, .external_lex_state = 3},
  [570] = {.lex_state = 152, .external_lex_state = 2},
  [571] = {.lex_state = 152, .external_lex_state = 2},
  [572] = {.lex_state = 152, .external_lex_state = 2},
  [573] = {.lex_state = 152, .external_lex_state = 3},
  [574] = {.lex_state = 152, .external_lex_state = 2},
  [575] = {.lex_state = 152, .external_lex_state = 2},
  [576] = {.lex_state = 152, .external_lex_state = 2},
  [577] = {.lex_state = 152, .external_lex_state = 2},
  [578] = {.lex_state = 152, .external_lex_state = 3},
  [579] = {.lex_state = 152, .external_lex_state = 2},
  [580] = {.lex_state = 152, .external_lex_state = 3},
  [581] = {.lex_state = 152, .external_lex_state = 3},
  [582] = {.lex_state = 152, .external_lex_state = 2},
  [583] = {.lex_state = 152, .external_lex_state = 2},
  [584] = {.lex_state = 152, .external_lex_state = 2},
  [585] = {.lex_state = 152, .external_lex_state = 2},
  [586] = {.lex_state = 152, .external_lex_state = 2},
  [587] = {.lex_state = 152, .external_lex_state = 3},
  [588] = {.lex_state = 152, .external_lex_state = 2},
  [589] = {.lex_state = 152, .external_lex_state = 2},
  [590] = {.lex_state = 152, .external_lex_state = 2},
  [591] = {.lex_state = 152, .external_lex_state = 2},
  [592] = {.lex_state = 152, .external_lex_state = 2},
  [593] = {.lex_state = 152, .external_lex_state = 3},
  [594] = {.lex_state = 152, .external_lex_state = 2},
  [595] = {.lex_state = 152, .external_lex_state = 3},
  [596] = {.lex_state = 152, .external_lex_state = 3},
  [597] = {.lex_state = 152, .external_lex_state = 2},
  [598] = {.lex_state = 152, .external_lex_state = 2},
  [599] = {.lex_state = 152, .external_lex_state = 2},
  [600] = {.lex_state = 152, .external_lex_state = 2},
  [601] = {.lex_state = 152, .external_lex_state = 2},
  [602] = {.lex_state = 152, .external_lex_state = 3},
  [603] = {.lex_state = 152, .external_lex_state = 2},
  [604] = {.lex_state = 152, .external_lex_state = 2},
  [605] = {.lex_state = 152, .external_lex_state = 2},
  [606] = {.lex_state = 152, .external_lex_state = 3},
  [607] = {.lex_state = 152, .external_lex_state = 3},
  [608] = {.lex_state = 152, .external_lex_state = 2},
  [609] = {.lex_state = 152, .external_lex_state = 2},
  [610] = {.lex_state = 152, .external_lex_state = 3},
  [611] = {.lex_state = 152, .external_lex_state = 3},
  [612] = {.lex_state = 152, .external_lex_state = 2},
  [613] = {.lex_state = 152, .external_lex_state = 3},
  [614] = {.lex_state = 152, .external_lex_state = 3},
  [615] = {.lex_state = 152, .external_lex_state = 3},
  [616] = {.lex_state = 152, .external_lex_state = 2},
  [617] = {.lex_state = 152, .external_lex_state = 2},
  [618] = {.lex_state = 152, .external_lex_state = 2},
  [619] = {.lex_state = 152, .external_lex_state = 2},
  [620] = {.lex_state = 152, .external_lex_state = 3},
  [621] = {.lex_state = 152, .external_lex_state = 2},
  [622] = {.lex_state = 152, .external_lex_state = 2},
  [623] = {.lex_state = 152, .external_lex_state = 3},
  [624] = {.lex_state = 152, .external_lex_state = 2},
  [625] = {.lex_state = 152, .external_lex_state = 3},
  [626] = {.lex_state = 152, .external_lex_state = 3},
  [627] = {.lex_state = 152, .external_lex_state = 2},
  [628] = {.lex_state = 152, .external_lex_state = 2},
  [629] = {.lex_state = 152, .external_lex_state = 3},
  [630] = {.lex_state = 152, .external_lex_state = 2},
  [631] = {.lex_state = 152, .external_lex_state = 3},
  [632] = {.lex_state = 152, .external_lex_state = 2},
  [633] = {.lex_state = 152, .external_lex_state = 2},
  [634] = {.lex_state = 152, .external_lex_state = 3},
  [635] = {.lex_state = 152, .external_lex_state = 3},
  [636] = {.lex_state = 152, .external_lex_state = 3},
  [637] = {.lex_state = 152, .external_lex_state = 2},
  [638] = {.lex_state = 152, .external_lex_state = 3},
  [639] = {.lex_state = 152, .external_lex_state = 3},
  [640] = {.lex_state = 152, .external_lex_state = 3},
  [641] = {.lex_state = 152, .external_lex_state = 3},
  [642] = {.lex_state = 152, .external_lex_state = 3},
  [643] = {.lex_state = 152, .external_lex_state = 3},
  [644] = {.lex_state = 152, .external_lex_state = 2},
  [645] = {.lex_state = 152, .external_lex_state = 3},
  [646] = {.lex_state = 152, .external_lex_state = 2},
  [647] = {.lex_state = 152, .external_lex_state = 3},
  [648] = {.lex_state = 152, .external_lex_state = 2},
  [649] = {.lex_state = 152, .external_lex_state = 3},
  [650] = {.lex_state = 152, .external_lex_state = 3},
  [651] = {.lex_state = 152, .external_lex_state = 3},
  [652] = {.lex_state = 152, .external_lex_state = 3},
  [653] = {.lex_state = 152, .external_lex_state = 3},
  [654] = {.lex_state = 152, .external_lex_state = 3},
  [655] = {.lex_state = 152, .external_lex_state = 3},
  [656] = {.lex_state = 152, .external_lex_state = 3},
  [657] = {.lex_state = 152, .external_lex_state = 3},
  [658] = {.lex_state = 152, .external_lex_state = 3},
  [659] = {.lex_state = 152, .external_lex_state = 3},
  [660] = {.lex_state = 152, .external_lex_state = 3},
  [661] = {.lex_state = 152, .external_lex_state = 3},
  [662] = {.lex_state = 152, .external_lex_state = 3},
  [663] = {.lex_state = 152, .external_lex_state = 3},
  [664] = {.lex_state = 152, .external_lex_state = 3},
  [665] = {.lex_state = 152, .external_lex_state = 2},
  [666] = {.lex_state = 152, .external_lex_state = 2},
  [667] = {.lex_state = 152, .external_lex_state = 3},
  [668] = {.lex_state = 152, .external_lex_state = 2},
  [669] = {.lex_state = 152, .external_lex_state = 3},
  [670] = {.lex_state = 152, .external_lex_state = 3},
  [671] = {.lex_state = 152, .external_lex_state = 3},
  [672] = {.lex_state = 152, .external_lex_state = 2},
  [673] = {.lex_state = 152, .external_lex_state = 3},
  [674] = {.lex_state = 152, .external_lex_state = 2},
  [675] = {.lex_state = 152, .external_lex_state = 2},
  [676] = {.lex_state = 152, .external_lex_state = 3},
  [677] = {.lex_state = 152, .external_lex_state = 2},
  [678] = {.lex_state = 152, .external_lex_state = 3},
  [679] = {.lex_state = 152, .external_lex_state = 3},
  [680] = {.lex_state = 152, .external_lex_state = 3},
  [681] = {.lex_state = 152, .external_lex_state = 3},
  [682] = {.lex_state = 152, .external_lex_state = 3},
  [683] = {.lex_state = 152, .external_lex_state = 2},
  [684] = {.lex_state = 152, .external_lex_state = 3},
  [685] = {.lex_state = 152, .external_lex_state = 3},
  [686] = {.lex_state = 152, .external_lex_state = 3},
  [687] = {.lex_state = 152, .external_lex_state = 2},
  [688] = {.lex_state = 152, .external_lex_state = 2},
  [689] = {.lex_state = 152, .external_lex_state = 2},
  [690] = {.lex_state = 152, .external_lex_state = 2},
  [691] = {.lex_state = 152, .external_lex_state = 2},
  [692] = {.lex_state = 152, .external_lex_state = 2},
  [693] = {.lex_state = 152, .external_lex_state = 3},
  [694] = {.lex_state = 152, .external_lex_state = 2},
  [695] = {.lex_state = 152, .external_lex_state = 2},
  [696] = {.lex_state = 152, .external_lex_state = 3},
  [697] = {.lex_state = 152, .external_lex_state = 3},
  [698] = {.lex_state = 152, .external_lex_state = 3},
  [699] = {.lex_state = 152, .external_lex_state = 3},
  [700] = {.lex_state = 152, .external_lex_state = 3},
  [701] = {.lex_state = 152, .external_lex_state = 3},
  [702] = {.lex_state = 152, .external_lex_state = 2},
  [703] = {.lex_state = 152, .external_lex_state = 2},
  [704] = {.lex_state = 152, .external_lex_state = 3},
  [705] = {.lex_state = 152, .external_lex_state = 2},
  [706] = {.lex_state = 152, .external_lex_state = 2},
  [707] = {.lex_state = 152, .external_lex_state = 3},
  [708] = {.lex_state = 152, .external_lex_state = 3},
  [709] = {.lex_state = 152, .external_lex_state = 3},
  [710] = {.lex_state = 152, .external_lex_state = 3},
  [711] = {.lex_state = 152, .external_lex_state = 3},
  [712] = {.lex_state = 152, .external_lex_state = 2},
  [713] = {.lex_state = 30, .external_lex_state = 2},
  [714] = {.lex_state = 30, .external_lex_state = 2},
  [715] = {.lex_state = 30, .external_lex_state = 2},
  [716] = {.lex_state = 30, .external_lex_state = 2},
  [717] = {.lex_state = 30, .external_lex_state = 2},
  [718] = {.lex_state = 30, .external_lex_state = 2},
  [719] = {.lex_state = 30, .external_lex_state = 2},
  [720] = {.lex_state = 30, .external_lex_state = 2},
  [721] = {.lex_state = 30, .external_lex_state = 2},
  [722] = {.lex_state = 30, .external_lex_state = 2},
  [723] = {.lex_state = 30, .external_lex_state = 2},
  [724] = {.lex_state = 30, .external_lex_state = 2},
  [725] = {.lex_state = 30, .external_lex_state = 2},
  [726] = {.lex_state = 30, .external_lex_state = 2},
  [727] = {.lex_state = 30, .external_lex_state = 2},
  [728] = {.lex_state = 30, .external_lex_state = 2},
  [729] = {.lex_state = 30, .external_lex_state = 2},
  [730] = {.lex_state = 30, .external_lex_state = 2},
  [731] = {.lex_state = 30, .external_lex_state = 2},
  [732] = {.lex_state = 30, .external_lex_state = 2},
  [733] = {.lex_state = 30, .external_lex_state = 2},
  [734] = {.lex_state = 30, .external_lex_state = 2},
  [735] = {.lex_state = 30, .external_lex_state = 2},
  [736] = {.lex_state = 30, .external_lex_state = 2},
  [737] = {.lex_state = 30, .external_lex_state = 2},
  [738] = {.lex_state = 30, .external_lex_state = 2},
  [739] = {.lex_state = 30, .external_lex_state = 2},
  [740] = {.lex_state = 30, .external_lex_state = 2},
  [741] = {.lex_state = 30, .external_lex_state = 2},
  [742] = {.lex_state = 30, .external_lex_state = 2},
  [743] = {.lex_state = 30, .external_lex_state = 2},
  [744] = {.lex_state = 30, .external_lex_state = 2},
  [745] = {.lex_state = 30, .external_lex_state = 2},
  [746] = {.lex_state = 30, .external_lex_state = 2},
  [747] = {.lex_state = 30, .external_lex_state = 2},
  [748] = {.lex_state = 30, .external_lex_state = 2},
  [749] = {.lex_state = 30, .external_lex_state = 2},
  [750] = {.lex_state = 30, .external_lex_state = 2},
  [751] = {.lex_state = 30, .external_lex_state = 2},
  [752] = {.lex_state = 30, .external_lex_state = 2},
  [753] = {.lex_state = 30, .external_lex_state = 2},
  [754] = {.lex_state = 30, .external_lex_state = 2},
  [755] = {.lex_state = 30, .external_lex_state = 2},
  [756] = {.lex_state = 30, .external_lex_state = 2},
  [757] = {.lex_state = 30, .external_lex_state = 2},
  [758] = {.lex_state = 30, .external_lex_state = 2},
  [759] = {.lex_state = 30, .external_lex_state = 2},
  [760] = {.lex_state = 30, .external_lex_state = 2},
  [761] = {.lex_state = 30, .external_lex_state = 2},
  [762] = {.lex_state = 30, .external_lex_state = 2},
  [763] = {.lex_state = 30, .external_lex_state = 2},
  [764] = {.lex_state = 30, .external_lex_state = 2},
  [765] = {.lex_state = 30, .external_lex_state = 2},
  [766] = {.lex_state = 30, .external_lex_state = 2},
  [767] = {.lex_state = 30, .external_lex_state = 2},
  [768] = {.lex_state = 30, .external_lex_state = 2},
  [769] = {.lex_state = 30, .external_lex_state = 2},
  [770] = {.lex_state = 30, .external_lex_state = 2},
  [771] = {.lex_state = 30, .external_lex_state = 2},
  [772] = {.lex_state = 30, .external_lex_state = 2},
  [773] = {.lex_state = 30, .external_lex_state = 2},
  [774] = {.lex_state = 30, .external_lex_state = 2},
  [775] = {.lex_state = 30, .external_lex_state = 2},
  [776] = {.lex_state = 30, .external_lex_state = 2},
  [777] = {.lex_state = 30, .external_lex_state = 2},
  [778] = {.lex_state = 30, .external_lex_state = 2},
  [779] = {.lex_state = 30, .external_lex_state = 2},
  [780] = {.lex_state = 30, .external_lex_state = 2},
  [781] = {.lex_state = 30, .external_lex_state = 2},
  [782] = {.lex_state = 30, .external_lex_state = 2},
  [783] = {.lex_state = 30, .external_lex_state = 4},
  [784] = {.lex_state = 30, .external_lex_state = 2},
  [785] = {.lex_state = 30, .external_lex_state = 2},
  [786] = {.lex_state = 30, .external_lex_state = 2},
  [787] = {.lex_state = 30, .external_lex_state = 2},
  [788] = {.lex_state = 30, .external_lex_state = 2},
  [789] = {.lex_state = 30, .external_lex_state = 2},
  [790] = {.lex_state = 30, .external_lex_state = 2},
  [791] = {.lex_state = 30, .external_lex_state = 4},
  [792] = {.lex_state = 30, .external_lex_state = 2},
  [793] = {.lex_state = 30, .external_lex_state = 2},
  [794] = {.lex_state = 30, .external_lex_state = 2},
  [795] = {.lex_state = 30, .external_lex_state = 2},
  [796] = {.lex_state = 30, .external_lex_state = 2},
  [797] = {.lex_state = 30, .external_lex_state = 2},
  [798] = {.lex_state = 30, .external_lex_state = 2},
  [799] = {.lex_state = 30, .external_lex_state = 2},
  [800] = {.lex_state = 30, .external_lex_state = 2},
  [801] = {.lex_state = 30, .external_lex_state = 2},
  [802] = {.lex_state = 30, .external_lex_state = 2},
  [803] = {.lex_state = 30, .external_lex_state = 2},
  [804] = {.lex_state = 30, .external_lex_state = 2},
  [805] = {.lex_state = 30, .external_lex_state = 2},
  [806] = {.lex_state = 30, .external_lex_state = 2},
  [807] = {.lex_state = 30, .external_lex_state = 2},
  [808] = {.lex_state = 30, .external_lex_state = 2},
  [809] = {.lex_state = 30, .external_lex_state = 2},
  [810] = {.lex_state = 30, .external_lex_state = 2},
  [811] = {.lex_state = 30, .external_lex_state = 2},
  [812] = {.lex_state = 30, .external_lex_state = 2},
  [813] = {.lex_state = 30, .external_lex_state = 2},
  [814] = {.lex_state = 30, .external_lex_state = 2},
  [815] = {.lex_state = 30, .external_lex_state = 2},
  [816] = {.lex_state = 30, .external_lex_state = 2},
  [817] = {.lex_state = 30, .external_lex_state = 2},
  [818] = {.lex_state = 30, .external_lex_state = 2},
  [819] = {.lex_state = 30, .external_lex_state = 2},
  [820] = {.lex_state = 30, .external_lex_state = 2},
  [821] = {.lex_state = 30, .external_lex_state = 2},
  [822] = {.lex_state = 30, .external_lex_state = 2},
  [823] = {.lex_state = 30, .external_lex_state = 2},
  [824] = {.lex_state = 30, .external_lex_state = 2},
  [825] = {.lex_state = 31, .external_lex_state = 2},
  [826] = {.lex_state = 30, .external_lex_state = 2},
  [827] = {.lex_state = 30, .external_lex_state = 2},
  [828] = {.lex_state = 30, .external_lex_state = 4},
  [829] = {.lex_state = 30, .external_lex_state = 2},
  [830] = {.lex_state = 30, .external_lex_state = 2},
  [831] = {.lex_state = 30, .external_lex_state = 2},
  [832] = {.lex_state = 30, .external_lex_state = 2},
  [833] = {.lex_state = 30, .external_lex_state = 2},
  [834] = {.lex_state = 30, .external_lex_state = 2},
  [835] = {.lex_state = 30, .external_lex_state = 2},
  [836] = {.lex_state = 30, .external_lex_state = 2},
  [837] = {.lex_state = 30, .external_lex_state = 2},
  [838] = {.lex_state = 30, .external_lex_state = 2},
  [839] = {.lex_state = 30, .external_lex_state = 2},
  [840] = {.lex_state = 30, .external_lex_state = 2},
  [841] = {.lex_state = 30, .external_lex_state = 2},
  [842] = {.lex_state = 30, .external_lex_state = 2},
  [843] = {.lex_state = 30, .external_lex_state = 2},
  [844] = {.lex_state = 30, .external_lex_state = 2},
  [845] = {.lex_state = 31, .external_lex_state = 2},
  [846] = {.lex_state = 30, .external_lex_state = 2},
  [847] = {.lex_state = 30, .external_lex_state = 2},
  [848] = {.lex_state = 30, .external_lex_state = 2},
  [849] = {.lex_state = 30, .external_lex_state = 2},
  [850] = {.lex_state = 30, .external_lex_state = 2},
  [851] = {.lex_state = 30, .external_lex_state = 2},
  [852] = {.lex_state = 30, .external_lex_state = 2},
  [853] = {.lex_state = 30, .external_lex_state = 2},
  [854] = {.lex_state = 30, .external_lex_state = 2},
  [855] = {.lex_state = 30, .external_lex_state = 2},
  [856] = {.lex_state = 30, .external_lex_state = 2},
  [857] = {.lex_state = 30, .external_lex_state = 2},
  [858] = {.lex_state = 30, .external_lex_state = 2},
  [859] = {.lex_state = 30, .external_lex_state = 2},
  [860] = {.lex_state = 30, .external_lex_state = 2},
  [861] = {.lex_state = 30, .external_lex_state = 2},
  [862] = {.lex_state = 30, .external_lex_state = 2},
  [863] = {.lex_state = 30, .external_lex_state = 2},
  [864] = {.lex_state = 30, .external_lex_state = 2},
  [865] = {.lex_state = 30, .external_lex_state = 2},
  [866] = {.lex_state = 30, .external_lex_state = 2},
  [867] = {.lex_state = 30, .external_lex_state = 4},
  [868] = {.lex_state = 30, .external_lex_state = 2},
  [869] = {.lex_state = 30, .external_lex_state = 2},
  [870] = {.lex_state = 31, .external_lex_state = 2},
  [871] = {.lex_state = 30, .external_lex_state = 2},
  [872] = {.lex_state = 30, .external_lex_state = 2},
  [873] = {.lex_state = 30, .external_lex_state = 2},
  [874] = {.lex_state = 30, .external_lex_state = 2},
  [875] = {.lex_state = 31, .external_lex_state = 2},
  [876] = {.lex_state = 30, .external_lex_state = 2},
  [877] = {.lex_state = 30, .external_lex_state = 2},
  [878] = {.lex_state = 30, .external_lex_state = 2},
  [879] = {.lex_state = 30, .external_lex_state = 2},
  [880] = {.lex_state = 30, .external_lex_state = 4},
  [881] = {.lex_state = 30, .external_lex_state = 2},
  [882] = {.lex_state = 30, .external_lex_state = 2},
  [883] = {.lex_state = 32, .external_lex_state = 2},
  [884] = {.lex_state = 30, .external_lex_state = 2},
  [885] = {.lex_state = 32, .external_lex_state = 2},
  [886] = {.lex_state = 30, .external_lex_state = 4},
  [887] = {.lex_state = 30, .external_lex_state = 2},
  [888] = {.lex_state = 30, .external_lex_state = 4},
  [889] = {.lex_state = 30, .external_lex_state = 2},
  [890] = {.lex_state = 30, .external_lex_state = 2},
  [891] = {.lex_state = 30, .external_lex_state = 2},
  [892] = {.lex_state = 30, .external_lex_state = 2},
  [893] = {.lex_state = 30, .external_lex_state = 2},
  [894] = {.lex_state = 30, .external_lex_state = 2},
  [895] = {.lex_state = 30, .external_lex_state = 2},
  [896] = {.lex_state = 30, .external_lex_state = 2},
  [897] = {.lex_state = 30, .external_lex_state = 2},
  [898] = {.lex_state = 30, .external_lex_state = 2},
  [899] = {.lex_state = 30, .external_lex_state = 2},
  [900] = {.lex_state = 30, .external_lex_state = 2},
  [901] = {.lex_state = 30, .external_lex_state = 2},
  [902] = {.lex_state = 30, .external_lex_state = 2},
  [903] = {.lex_state = 30, .external_lex_state = 2},
  [904] = {.lex_state = 30, .external_lex_state = 2},
  [905] = {.lex_state = 30, .external_lex_state = 2},
  [906] = {.lex_state = 30, .external_lex_state = 2},
  [907] = {.lex_state = 30, .external_lex_state = 2},
  [908] = {.lex_state = 30, .external_lex_state = 2},
  [909] = {.lex_state = 30, .external_lex_state = 2},
  [910] = {.lex_state = 30, .external_lex_state = 2},
  [911] = {.lex_state = 30, .external_lex_state = 2},
  [912] = {.lex_state = 30, .external_lex_state = 2},
  [913] = {.lex_state = 30, .external_lex_state = 2},
  [914] = {.lex_state = 30, .external_lex_state = 2},
  [915] = {.lex_state = 30, .external_lex_state = 2},
  [916] = {.lex_state = 30, .external_lex_state = 2},
  [917] = {.lex_state = 30, .external_lex_state = 2},
  [918] = {.lex_state = 30, .external_lex_state = 2},
  [919] = {.lex_state = 30, .external_lex_state = 2},
  [920] = {.lex_state = 30, .external_lex_state = 2},
  [921] = {.lex_state = 30, .external_lex_state = 2},
  [922] = {.lex_state = 30, .external_lex_state = 2},
  [923] = {.lex_state = 30, .external_lex_state = 2},
  [924] = {.lex_state = 30, .external_lex_state = 2},
  [925] = {.lex_state = 30, .external_lex_state = 2},
  [926] = {.lex_state = 30, .external_lex_state = 2},
  [927] = {.lex_state = 30, .external_lex_state = 2},
  [928] = {.lex_state = 30, .external_lex_state = 2},
  [929] = {.lex_state = 30, .external_lex_state = 2},
  [930] = {.lex_state = 30, .external_lex_state = 2},
  [931] = {.lex_state = 30, .external_lex_state = 2},
  [932] = {.lex_state = 30, .external_lex_state = 2},
  [933] = {.lex_state = 30, .external_lex_state = 2},
  [934] = {.lex_state = 30, .external_lex_state = 2},
  [935] = {.lex_state = 30, .external_lex_state = 2},
  [936] = {.lex_state = 151},
  [937] = {.lex_state = 30, .external_lex_state = 2},
  [938] = {.lex_state = 30, .external_lex_state = 2},
  [939] = {.lex_state = 151},
  [940] = {.lex_state = 151},
  [941] = {.lex_state = 30, .external_lex_state = 2},
  [942] = {.lex_state = 30, .external_lex_state = 2},
  [943] = {.lex_state = 30, .external_lex_state = 2},
  [944] = {.lex_state = 30, .external_lex_state = 2},
  [945] = {.lex_state = 30, .external_lex_state = 2},
  [946] = {.lex_state = 30, .external_lex_state = 2},
  [947] = {.lex_state = 30, .external_lex_state = 2},
  [948] = {.lex_state = 30, .external_lex_state = 2},
  [949] = {.lex_state = 30, .external_lex_state = 2},
  [950] = {.lex_state = 30, .external_lex_state = 2},
  [951] = {.lex_state = 30, .external_lex_state = 2},
  [952] = {.lex_state = 30, .external_lex_state = 2},
  [953] = {.lex_state = 30, .external_lex_state = 2},
  [954] = {.lex_state = 30, .external_lex_state = 2},
  [955] = {.lex_state = 30, .external_lex_state = 2},
  [956] = {.lex_state = 30, .external_lex_state = 2},
  [957] = {.lex_state = 30, .external_lex_state = 2},
  [958] = {.lex_state = 30, .external_lex_state = 2},
  [959] = {.lex_state = 30, .external_lex_state = 2},
  [960] = {.lex_state = 30, .external_lex_state = 2},
  [961] = {.lex_state = 30, .external_lex_state = 2},
  [962] = {.lex_state = 30, .external_lex_state = 2},
  [963] = {.lex_state = 30, .external_lex_state = 2},
  [964] = {.lex_state = 30, .external_lex_state = 2},
  [965] = {.lex_state = 30, .external_lex_state = 2},
  [966] = {.lex_state = 30, .external_lex_state = 2},
  [967] = {.lex_state = 30, .external_lex_state = 2},
  [968] = {.lex_state = 30, .external_lex_state = 2},
  [969] = {.lex_state = 30, .external_lex_state = 2},
  [970] = {.lex_state = 30, .external_lex_state = 2},
  [971] = {.lex_state = 30, .external_lex_state = 2},
  [972] = {.lex_state = 30, .external_lex_state = 2},
  [973] = {.lex_state = 30, .external_lex_state = 2},
  [974] = {.lex_state = 30, .external_lex_state = 2},
  [975] = {.lex_state = 30, .external_lex_state = 2},
  [976] = {.lex_state = 30, .external_lex_state = 2},
  [977] = {.lex_state = 30, .external_lex_state = 2},
  [978] = {.lex_state = 30, .external_lex_state = 2},
  [979] = {.lex_state = 30, .external_lex_state = 2},
  [980] = {.lex_state = 30, .external_lex_state = 2},
  [981] = {.lex_state = 30, .external_lex_state = 2},
  [982] = {.lex_state = 30, .external_lex_state = 2},
  [983] = {.lex_state = 30, .external_lex_state = 2},
  [984] = {.lex_state = 30, .external_lex_state = 2},
  [985] = {.lex_state = 30, .external_lex_state = 2},
  [986] = {.lex_state = 30, .external_lex_state = 2},
  [987] = {.lex_state = 30, .external_lex_state = 2},
  [988] = {.lex_state = 30, .external_lex_state = 2},
  [989] = {.lex_state = 30, .external_lex_state = 2},
  [990] = {.lex_state = 30, .external_lex_state = 2},
  [991] = {.lex_state = 30, .external_lex_state = 2},
  [992] = {.lex_state = 30, .external_lex_state = 2},
  [993] = {.lex_state = 30, .external_lex_state = 2},
  [994] = {.lex_state = 30, .external_lex_state = 2},
  [995] = {.lex_state = 30, .external_lex_state = 2},
  [996] = {.lex_state = 30, .external_lex_state = 2},
  [997] = {.lex_state = 30, .external_lex_state = 2},
  [998] = {.lex_state = 30, .external_lex_state = 2},
  [999] = {.lex_state = 30, .external_lex_state = 2},
  [1000] = {.lex_state = 30, .external_lex_state = 2},
  [1001] = {.lex_state = 30, .external_lex_state = 2},
  [1002] = {.lex_state = 30, .external_lex_state = 2},
  [1003] = {.lex_state = 30, .external_lex_state = 2},
  [1004] = {.lex_state = 30, .external_lex_state = 2},
  [1005] = {.lex_state = 30, .external_lex_state = 2},
  [1006] = {.lex_state = 30, .external_lex_state = 2},
  [1007] = {.lex_state = 30, .external_lex_state = 2},
  [1008] = {.lex_state = 30, .external_lex_state = 2},
  [1009] = {.lex_state = 30, .external_lex_state = 2},
  [1010] = {.lex_state = 30, .external_lex_state = 2},
  [1011] = {.lex_state = 30, .external_lex_state = 2},
  [1012] = {.lex_state = 30, .external_lex_state = 2},
  [1013] = {.lex_state = 30, .external_lex_state = 2},
  [1014] = {.lex_state = 30, .external_lex_state = 2},
  [1015] = {.lex_state = 30, .external_lex_state = 2},
  [1016] = {.lex_state = 30, .external_lex_state = 2},
  [1017] = {.lex_state = 30, .external_lex_state = 2},
  [1018] = {.lex_state = 30, .external_lex_state = 2},
  [1019] = {.lex_state = 30, .external_lex_state = 2},
  [1020] = {.lex_state = 30, .external_lex_state = 2},
  [1021] = {.lex_state = 30, .external_lex_state = 2},
  [1022] = {.lex_state = 30, .external_lex_state = 2},
  [1023] = {.lex_state = 30, .external_lex_state = 2},
  [1024] = {.lex_state = 30, .external_lex_state = 2},
  [1025] = {.lex_state = 30, .external_lex_state = 2},
  [1026] = {.lex_state = 30, .external_lex_state = 2},
  [1027] = {.lex_state = 30, .external_lex_state = 2},
  [1028] = {.lex_state = 30, .external_lex_state = 2},
  [1029] = {.lex_state = 30, .external_lex_state = 2},
  [1030] = {.lex_state = 30, .external_lex_state = 2},
  [1031] = {.lex_state = 30, .external_lex_state = 2},
  [1032] = {.lex_state = 30, .external_lex_state = 2},
  [1033] = {.lex_state = 30, .external_lex_state = 2},
  [1034] = {.lex_state = 151},
  [1035] = {.lex_state = 30, .external_lex_state = 2},
  [1036] = {.lex_state = 30, .external_lex_state = 2},
  [1037] = {.lex_state = 30, .external_lex_state = 2},
  [1038] = {.lex_state = 30, .external_lex_state = 2},
  [1039] = {.lex_state = 30, .external_lex_state = 2},
  [1040] = {.lex_state = 30, .external_lex_state = 2},
  [1041] = {.lex_state = 30, .external_lex_state = 2},
  [1042] = {.lex_state = 30, .external_lex_state = 2},
  [1043] = {.lex_state = 30, .external_lex_state = 2},
  [1044] = {.lex_state = 30, .external_lex_state = 2},
  [1045] = {.lex_state = 30, .external_lex_state = 2},
  [1046] = {.lex_state = 30, .external_lex_state = 2},
  [1047] = {.lex_state = 30, .external_lex_state = 2},
  [1048] = {.lex_state = 30, .external_lex_state = 2},
  [1049] = {.lex_state = 30, .external_lex_state = 2},
  [1050] = {.lex_state = 30, .external_lex_state = 2},
  [1051] = {.lex_state = 30, .external_lex_state = 2},
  [1052] = {.lex_state = 30, .external_lex_state = 2},
  [1053] = {.lex_state = 30, .external_lex_state = 2},
  [1054] = {.lex_state = 30, .external_lex_state = 2},
  [1055] = {.lex_state = 30, .external_lex_state = 2},
  [1056] = {.lex_state = 30, .external_lex_state = 2},
  [1057] = {.lex_state = 30, .external_lex_state = 2},
  [1058] = {.lex_state = 30, .external_lex_state = 2},
  [1059] = {.lex_state = 30, .external_lex_state = 2},
  [1060] = {.lex_state = 30, .external_lex_state = 2},
  [1061] = {.lex_state = 30, .external_lex_state = 2},
  [1062] = {.lex_state = 30, .external_lex_state = 2},
  [1063] = {.lex_state = 30, .external_lex_state = 2},
  [1064] = {.lex_state = 30, .external_lex_state = 2},
  [1065] = {.lex_state = 30, .external_lex_state = 2},
  [1066] = {.lex_state = 30, .external_lex_state = 2},
  [1067] = {.lex_state = 30, .external_lex_state = 2},
  [1068] = {.lex_state = 30, .external_lex_state = 2},
  [1069] = {.lex_state = 30, .external_lex_state = 2},
  [1070] = {.lex_state = 30, .external_lex_state = 2},
  [1071] = {.lex_state = 30, .external_lex_state = 2},
  [1072] = {.lex_state = 30, .external_lex_state = 2},
  [1073] = {.lex_state = 30, .external_lex_state = 2},
  [1074] = {.lex_state = 30, .external_lex_state = 2},
  [1075] = {.lex_state = 30, .external_lex_state = 2},
  [1076] = {.lex_state = 151},
  [1077] = {.lex_state = 151},
  [1078] = {.lex_state = 151},
  [1079] = {.lex_state = 30, .external_lex_state = 2},
  [1080] = {.lex_state = 30, .external_lex_state = 2},
  [1081] = {.lex_state = 30, .external_lex_state = 2},
  [1082] = {.lex_state = 30, .external_lex_state = 2},
  [1083] = {.lex_state = 30, .external_lex_state = 2},
  [1084] = {.lex_state = 30, .external_lex_state = 2},
  [1085] = {.lex_state = 30, .external_lex_state = 2},
  [1086] = {.lex_state = 30, .external_lex_state = 2},
  [1087] = {.lex_state = 30, .external_lex_state = 2},
  [1088] = {.lex_state = 30, .external_lex_state = 2},
  [1089] = {.lex_state = 30, .external_lex_state = 2},
  [1090] = {.lex_state = 30, .external_lex_state = 2},
  [1091] = {.lex_state = 30, .external_lex_state = 2},
  [1092] = {.lex_state = 30, .external_lex_state = 2},
  [1093] = {.lex_state = 30, .external_lex_state = 2},
  [1094] = {.lex_state = 30, .external_lex_state = 2},
  [1095] = {.lex_state = 30, .external_lex_state = 2},
  [1096] = {.lex_state = 30, .external_lex_state = 2},
  [1097] = {.lex_state = 30, .external_lex_state = 2},
  [1098] = {.lex_state = 30, .external_lex_state = 2},
  [1099] = {.lex_state = 30, .external_lex_state = 2},
  [1100] = {.lex_state = 30, .external_lex_state = 2},
  [1101] = {.lex_state = 30, .external_lex_state = 2},
  [1102] = {.lex_state = 30, .external_lex_state = 2},
  [1103] = {.lex_state = 30, .external_lex_state = 2},
  [1104] = {.lex_state = 30, .external_lex_state = 2},
  [1105] = {.lex_state = 30, .external_lex_state = 2},
  [1106] = {.lex_state = 30, .external_lex_state = 2},
  [1107] = {.lex_state = 30, .external_lex_state = 2},
  [1108] = {.lex_state = 30, .external_lex_state = 2},
  [1109] = {.lex_state = 30, .external_lex_state = 2},
  [1110] = {.lex_state = 30, .external_lex_state = 2},
  [1111] = {.lex_state = 30, .external_lex_state = 2},
  [1112] = {.lex_state = 30, .external_lex_state = 2},
  [1113] = {.lex_state = 30, .external_lex_state = 2},
  [1114] = {.lex_state = 30, .external_lex_state = 2},
  [1115] = {.lex_state = 30, .external_lex_state = 2},
  [1116] = {.lex_state = 30, .external_lex_state = 2},
  [1117] = {.lex_state = 30, .external_lex_state = 2},
  [1118] = {.lex_state = 30, .external_lex_state = 2},
  [1119] = {.lex_state = 30, .external_lex_state = 2},
  [1120] = {.lex_state = 30, .external_lex_state = 2},
  [1121] = {.lex_state = 30, .external_lex_state = 2},
  [1122] = {.lex_state = 30, .external_lex_state = 2},
  [1123] = {.lex_state = 30, .external_lex_state = 2},
  [1124] = {.lex_state = 30, .external_lex_state = 2},
  [1125] = {.lex_state = 30, .external_lex_state = 2},
  [1126] = {.lex_state = 30, .external_lex_state = 2},
  [1127] = {.lex_state = 30, .external_lex_state = 2},
  [1128] = {.lex_state = 30, .external_lex_state = 2},
  [1129] = {.lex_state = 30, .external_lex_state = 2},
  [1130] = {.lex_state = 30, .external_lex_state = 2},
  [1131] = {.lex_state = 30, .external_lex_state = 2},
  [1132] = {.lex_state = 30, .external_lex_state = 2},
  [1133] = {.lex_state = 30, .external_lex_state = 2},
  [1134] = {.lex_state = 30, .external_lex_state = 2},
  [1135] = {.lex_state = 30, .external_lex_state = 2},
  [1136] = {.lex_state = 30, .external_lex_state = 2},
  [1137] = {.lex_state = 30, .external_lex_state = 2},
  [1138] = {.lex_state = 30, .external_lex_state = 2},
  [1139] = {.lex_state = 30, .external_lex_state = 2},
  [1140] = {.lex_state = 30, .external_lex_state = 2},
  [1141] = {.lex_state = 30, .external_lex_state = 2},
  [1142] = {.lex_state = 30, .external_lex_state = 2},
  [1143] = {.lex_state = 30, .external_lex_state = 2},
  [1144] = {.lex_state = 30, .external_lex_state = 2},
  [1145] = {.lex_state = 30, .external_lex_state = 2},
  [1146] = {.lex_state = 30, .external_lex_state = 2},
  [1147] = {.lex_state = 30, .external_lex_state = 2},
  [1148] = {.lex_state = 30, .external_lex_state = 2},
  [1149] = {.lex_state = 30, .external_lex_state = 2},
  [1150] = {.lex_state = 30, .external_lex_state = 2},
  [1151] = {.lex_state = 30, .external_lex_state = 2},
  [1152] = {.lex_state = 30, .external_lex_state = 2},
  [1153] = {.lex_state = 30, .external_lex_state = 2},
  [1154] = {.lex_state = 30, .external_lex_state = 2},
  [1155] = {.lex_state = 30, .external_lex_state = 2},
  [1156] = {.lex_state = 30, .external_lex_state = 2},
  [1157] = {.lex_state = 30, .external_lex_state = 2},
  [1158] = {.lex_state = 30, .external_lex_state = 2},
  [1159] = {.lex_state = 30, .external_lex_state = 2},
  [1160] = {.lex_state = 30, .external_lex_state = 2},
  [1161] = {.lex_state = 30, .external_lex_state = 2},
  [1162] = {.lex_state = 30, .external_lex_state = 2},
  [1163] = {.lex_state = 30, .external_lex_state = 2},
  [1164] = {.lex_state = 30, .external_lex_state = 2},
  [1165] = {.lex_state = 30, .external_lex_state = 2},
  [1166] = {.lex_state = 30, .external_lex_state = 2},
  [1167] = {.lex_state = 30, .external_lex_state = 2},
  [1168] = {.lex_state = 30, .external_lex_state = 2},
  [1169] = {.lex_state = 30, .external_lex_state = 2},
  [1170] = {.lex_state = 30, .external_lex_state = 2},
  [1171] = {.lex_state = 30, .external_lex_state = 2},
  [1172] = {.lex_state = 30, .external_lex_state = 2},
  [1173] = {.lex_state = 30, .external_lex_state = 2},
  [1174] = {.lex_state = 30, .external_lex_state = 2},
  [1175] = {.lex_state = 30, .external_lex_state = 2},
  [1176] = {.lex_state = 30, .external_lex_state = 2},
  [1177] = {.lex_state = 30, .external_lex_state = 2},
  [1178] = {.lex_state = 30, .external_lex_state = 2},
  [1179] = {.lex_state = 30, .external_lex_state = 2},
  [1180] = {.lex_state = 30, .external_lex_state = 2},
  [1181] = {.lex_state = 30, .external_lex_state = 2},
  [1182] = {.lex_state = 30, .external_lex_state = 2},
  [1183] = {.lex_state = 30, .external_lex_state = 2},
  [1184] = {.lex_state = 30, .external_lex_state = 2},
  [1185] = {.lex_state = 30, .external_lex_state = 2},
  [1186] = {.lex_state = 30, .external_lex_state = 2},
  [1187] = {.lex_state = 30, .external_lex_state = 2},
  [1188] = {.lex_state = 30, .external_lex_state = 2},
  [1189] = {.lex_state = 30, .external_lex_state = 2},
  [1190] = {.lex_state = 30, .external_lex_state = 2},
  [1191] = {.lex_state = 30, .external_lex_state = 2},
  [1192] = {.lex_state = 30, .external_lex_state = 2},
  [1193] = {.lex_state = 30, .external_lex_state = 2},
  [1194] = {.lex_state = 30, .external_lex_state = 2},
  [1195] = {.lex_state = 30, .external_lex_state = 2},
  [1196] = {.lex_state = 30, .external_lex_state = 2},
  [1197] = {.lex_state = 30, .external_lex_state = 2},
  [1198] = {.lex_state = 30, .external_lex_state = 2},
  [1199] = {.lex_state = 30, .external_lex_state = 2},
  [1200] = {.lex_state = 30, .external_lex_state = 2},
  [1201] = {.lex_state = 30, .external_lex_state = 2},
  [1202] = {.lex_state = 30, .external_lex_state = 2},
  [1203] = {.lex_state = 30, .external_lex_state = 2},
  [1204] = {.lex_state = 30, .external_lex_state = 2},
  [1205] = {.lex_state = 30, .external_lex_state = 2},
  [1206] = {.lex_state = 30, .external_lex_state = 2},
  [1207] = {.lex_state = 30, .external_lex_state = 2},
  [1208] = {.lex_state = 30, .external_lex_state = 2},
  [1209] = {.lex_state = 30, .external_lex_state = 2},
  [1210] = {.lex_state = 30, .external_lex_state = 2},
  [1211] = {.lex_state = 30, .external_lex_state = 2},
  [1212] = {.lex_state = 30, .external_lex_state = 2},
  [1213] = {.lex_state = 30, .external_lex_state = 2},
  [1214] = {.lex_state = 30, .external_lex_state = 2},
  [1215] = {.lex_state = 30, .external_lex_state = 2},
  [1216] = {.lex_state = 30, .external_lex_state = 2},
  [1217] = {.lex_state = 30, .external_lex_state = 2},
  [1218] = {.lex_state = 30, .external_lex_state = 2},
  [1219] = {.lex_state = 30, .external_lex_state = 2},
  [1220] = {.lex_state = 30, .external_lex_state = 2},
  [1221] = {.lex_state = 30, .external_lex_state = 2},
  [1222] = {.lex_state = 30, .external_lex_state = 2},
  [1223] = {.lex_state = 30, .external_lex_state = 2},
  [1224] = {.lex_state = 30, .external_lex_state = 2},
  [1225] = {.lex_state = 30, .external_lex_state = 2},
  [1226] = {.lex_state = 30, .external_lex_state = 2},
  [1227] = {.lex_state = 30, .external_lex_state = 2},
  [1228] = {.lex_state = 30, .external_lex_state = 2},
  [1229] = {.lex_state = 30, .external_lex_state = 2},
  [1230] = {.lex_state = 30, .external_lex_state = 2},
  [1231] = {.lex_state = 30, .external_lex_state = 2},
  [1232] = {.lex_state = 30, .external_lex_state = 2},
  [1233] = {.lex_state = 30, .external_lex_state = 2},
  [1234] = {.lex_state = 30, .external_lex_state = 2},
  [1235] = {.lex_state = 30, .external_lex_state = 2},
  [1236] = {.lex_state = 30, .external_lex_state = 2},
  [1237] = {.lex_state = 30, .external_lex_state = 2},
  [1238] = {.lex_state = 30, .external_lex_state = 2},
  [1239] = {.lex_state = 30, .external_lex_state = 2},
  [1240] = {.lex_state = 30, .external_lex_state = 2},
  [1241] = {.lex_state = 30, .external_lex_state = 2},
  [1242] = {.lex_state = 30, .external_lex_state = 2},
  [1243] = {.lex_state = 30, .external_lex_state = 2},
  [1244] = {.lex_state = 30, .external_lex_state = 2},
  [1245] = {.lex_state = 30, .external_lex_state = 2},
  [1246] = {.lex_state = 30, .external_lex_state = 2},
  [1247] = {.lex_state = 30, .external_lex_state = 2},
  [1248] = {.lex_state = 30, .external_lex_state = 2},
  [1249] = {.lex_state = 30, .external_lex_state = 2},
  [1250] = {.lex_state = 30, .external_lex_state = 2},
  [1251] = {.lex_state = 30, .external_lex_state = 2},
  [1252] = {.lex_state = 30, .external_lex_state = 2},
  [1253] = {.lex_state = 30, .external_lex_state = 2},
  [1254] = {.lex_state = 30, .external_lex_state = 2},
  [1255] = {.lex_state = 30, .external_lex_state = 2},
  [1256] = {.lex_state = 30, .external_lex_state = 2},
  [1257] = {.lex_state = 30, .external_lex_state = 2},
  [1258] = {.lex_state = 30, .external_lex_state = 2},
  [1259] = {.lex_state = 30, .external_lex_state = 2},
  [1260] = {.lex_state = 30, .external_lex_state = 2},
  [1261] = {.lex_state = 30, .external_lex_state = 2},
  [1262] = {.lex_state = 30, .external_lex_state = 2},
  [1263] = {.lex_state = 30, .external_lex_state = 2},
  [1264] = {.lex_state = 30, .external_lex_state = 2},
  [1265] = {.lex_state = 30, .external_lex_state = 2},
  [1266] = {.lex_state = 30, .external_lex_state = 2},
  [1267] = {.lex_state = 30, .external_lex_state = 2},
  [1268] = {.lex_state = 30, .external_lex_state = 2},
  [1269] = {.lex_state = 30, .external_lex_state = 2},
  [1270] = {.lex_state = 30, .external_lex_state = 2},
  [1271] = {.lex_state = 30, .external_lex_state = 2},
  [1272] = {.lex_state = 30, .external_lex_state = 2},
  [1273] = {.lex_state = 30, .external_lex_state = 2},
  [1274] = {.lex_state = 30, .external_lex_state = 2},
  [1275] = {.lex_state = 30, .external_lex_state = 2},
  [1276] = {.lex_state = 30, .external_lex_state = 2},
  [1277] = {.lex_state = 30, .external_lex_state = 2},
  [1278] = {.lex_state = 30, .external_lex_state = 2},
  [1279] = {.lex_state = 30, .external_lex_state = 2},
  [1280] = {.lex_state = 30, .external_lex_state = 2},
  [1281] = {.lex_state = 30, .external_lex_state = 2},
  [1282] = {.lex_state = 30, .external_lex_state = 2},
  [1283] = {.lex_state = 30, .external_lex_state = 2},
  [1284] = {.lex_state = 30, .external_lex_state = 2},
  [1285] = {.lex_state = 30, .external_lex_state = 2},
  [1286] = {.lex_state = 30, .external_lex_state = 2},
  [1287] = {.lex_state = 30, .external_lex_state = 2},
  [1288] = {.lex_state = 30, .external_lex_state = 2},
  [1289] = {.lex_state = 30, .external_lex_state = 2},
  [1290] = {.lex_state = 30, .external_lex_state = 2},
  [1291] = {.lex_state = 30, .external_lex_state = 2},
  [1292] = {.lex_state = 30, .external_lex_state = 2},
  [1293] = {.lex_state = 30, .external_lex_state = 2},
  [1294] = {.lex_state = 30, .external_lex_state = 2},
  [1295] = {.lex_state = 30, .external_lex_state = 2},
  [1296] = {.lex_state = 30, .external_lex_state = 2},
  [1297] = {.lex_state = 30, .external_lex_state = 2},
  [1298] = {.lex_state = 30, .external_lex_state = 2},
  [1299] = {.lex_state = 30, .external_lex_state = 2},
  [1300] = {.lex_state = 30, .external_lex_state = 2},
  [1301] = {.lex_state = 30, .external_lex_state = 2},
  [1302] = {.lex_state = 30, .external_lex_state = 2},
  [1303] = {.lex_state = 30, .external_lex_state = 2},
  [1304] = {.lex_state = 30, .external_lex_state = 2},
  [1305] = {.lex_state = 30, .external_lex_state = 2},
  [1306] = {.lex_state = 30, .external_lex_state = 2},
  [1307] = {.lex_state = 30, .external_lex_state = 2},
  [1308] = {.lex_state = 30, .external_lex_state = 2},
  [1309] = {.lex_state = 30, .external_lex_state = 2},
  [1310] = {.lex_state = 30, .external_lex_state = 2},
  [1311] = {.lex_state = 30, .external_lex_state = 2},
  [1312] = {.lex_state = 30, .external_lex_state = 2},
  [1313] = {.lex_state = 30, .external_lex_state = 2},
  [1314] = {.lex_state = 30, .external_lex_state = 2},
  [1315] = {.lex_state = 30, .external_lex_state = 2},
  [1316] = {.lex_state = 30, .external_lex_state = 2},
  [1317] = {.lex_state = 30, .external_lex_state = 2},
  [1318] = {.lex_state = 30, .external_lex_state = 2},
  [1319] = {.lex_state = 30, .external_lex_state = 2},
  [1320] = {.lex_state = 30, .external_lex_state = 2},
  [1321] = {.lex_state = 30, .external_lex_state = 2},
  [1322] = {.lex_state = 30, .external_lex_state = 2},
  [1323] = {.lex_state = 30, .external_lex_state = 2},
  [1324] = {.lex_state = 30, .external_lex_state = 2},
  [1325] = {.lex_state = 30, .external_lex_state = 2},
  [1326] = {.lex_state = 30, .external_lex_state = 2},
  [1327] = {.lex_state = 30, .external_lex_state = 2},
  [1328] = {.lex_state = 30, .external_lex_state = 2},
  [1329] = {.lex_state = 30, .external_lex_state = 2},
  [1330] = {.lex_state = 30, .external_lex_state = 2},
  [1331] = {.lex_state = 30, .external_lex_state = 2},
  [1332] = {.lex_state = 30, .external_lex_state = 2},
  [1333] = {.lex_state = 30, .external_lex_state = 2},
  [1334] = {.lex_state = 30, .external_lex_state = 2},
  [1335] = {.lex_state = 30, .external_lex_state = 2},
  [1336] = {.lex_state = 30, .external_lex_state = 2},
  [1337] = {.lex_state = 30, .external_lex_state = 2},
  [1338] = {.lex_state = 30, .external_lex_state = 2},
  [1339] = {.lex_state = 30, .external_lex_state = 2},
  [1340] = {.lex_state = 30, .external_lex_state = 2},
  [1341] = {.lex_state = 30, .external_lex_state = 2},
  [1342] = {.lex_state = 30, .external_lex_state = 2},
  [1343] = {.lex_state = 30, .external_lex_state = 2},
  [1344] = {.lex_state = 30, .external_lex_state = 2},
  [1345] = {.lex_state = 30, .external_lex_state = 2},
  [1346] = {.lex_state = 30, .external_lex_state = 2},
  [1347] = {.lex_state = 30, .external_lex_state = 2},
  [1348] = {.lex_state = 30, .external_lex_state = 2},
  [1349] = {.lex_state = 30, .external_lex_state = 2},
  [1350] = {.lex_state = 30, .external_lex_state = 2},
  [1351] = {.lex_state = 30, .external_lex_state = 2},
  [1352] = {.lex_state = 30, .external_lex_state = 2},
  [1353] = {.lex_state = 30, .external_lex_state = 2},
  [1354] = {.lex_state = 30, .external_lex_state = 2},
  [1355] = {.lex_state = 30, .external_lex_state = 2},
  [1356] = {.lex_state = 30, .external_lex_state = 2},
  [1357] = {.lex_state = 30, .external_lex_state = 2},
  [1358] = {.lex_state = 30, .external_lex_state = 2},
  [1359] = {.lex_state = 30, .external_lex_state = 2},
  [1360] = {.lex_state = 30, .external_lex_state = 2},
  [1361] = {.lex_state = 30, .external_lex_state = 2},
  [1362] = {.lex_state = 30, .external_lex_state = 2},
  [1363] = {.lex_state = 30, .external_lex_state = 2},
  [1364] = {.lex_state = 30, .external_lex_state = 2},
  [1365] = {.lex_state = 30, .external_lex_state = 2},
  [1366] = {.lex_state = 30, .external_lex_state = 2},
  [1367] = {.lex_state = 30, .external_lex_state = 2},
  [1368] = {.lex_state = 30, .external_lex_state = 2},
  [1369] = {.lex_state = 30, .external_lex_state = 2},
  [1370] = {.lex_state = 30, .external_lex_state = 2},
  [1371] = {.lex_state = 30, .external_lex_state = 2},
  [1372] = {.lex_state = 30, .external_lex_state = 2},
  [1373] = {.lex_state = 30, .external_lex_state = 2},
  [1374] = {.lex_state = 30, .external_lex_state = 2},
  [1375] = {.lex_state = 30, .external_lex_state = 2},
  [1376] = {.lex_state = 30, .external_lex_state = 2},
  [1377] = {.lex_state = 30, .external_lex_state = 2},
  [1378] = {.lex_state = 30, .external_lex_state = 2},
  [1379] = {.lex_state = 30, .external_lex_state = 2},
  [1380] = {.lex_state = 30, .external_lex_state = 2},
  [1381] = {.lex_state = 30, .external_lex_state = 2},
  [1382] = {.lex_state = 30, .external_lex_state = 2},
  [1383] = {.lex_state = 30, .external_lex_state = 2},
  [1384] = {.lex_state = 30, .external_lex_state = 2},
  [1385] = {.lex_state = 30, .external_lex_state = 2},
  [1386] = {.lex_state = 30, .external_lex_state = 2},
  [1387] = {.lex_state = 30, .external_lex_state = 2},
  [1388] = {.lex_state = 30, .external_lex_state = 2},
  [1389] = {.lex_state = 30, .external_lex_state = 2},
  [1390] = {.lex_state = 30, .external_lex_state = 2},
  [1391] = {.lex_state = 30, .external_lex_state = 2},
  [1392] = {.lex_state = 30, .external_lex_state = 2},
  [1393] = {.lex_state = 30, .external_lex_state = 2},
  [1394] = {.lex_state = 30, .external_lex_state = 2},
  [1395] = {.lex_state = 30, .external_lex_state = 2},
  [1396] = {.lex_state = 30, .external_lex_state = 2},
  [1397] = {.lex_state = 30, .external_lex_state = 2},
  [1398] = {.lex_state = 30, .external_lex_state = 2},
  [1399] = {.lex_state = 30, .external_lex_state = 2},
  [1400] = {.lex_state = 30, .external_lex_state = 2},
  [1401] = {.lex_state = 30, .external_lex_state = 2},
  [1402] = {.lex_state = 30, .external_lex_state = 2},
  [1403] = {.lex_state = 30, .external_lex_state = 2},
  [1404] = {.lex_state = 30, .external_lex_state = 2},
  [1405] = {.lex_state = 30, .external_lex_state = 2},
  [1406] = {.lex_state = 30, .external_lex_state = 2},
  [1407] = {.lex_state = 30, .external_lex_state = 2},
  [1408] = {.lex_state = 30, .external_lex_state = 2},
  [1409] = {.lex_state = 30, .external_lex_state = 2},
  [1410] = {.lex_state = 30, .external_lex_state = 2},
  [1411] = {.lex_state = 30, .external_lex_state = 2},
  [1412] = {.lex_state = 30, .external_lex_state = 2},
  [1413] = {.lex_state = 30, .external_lex_state = 2},
  [1414] = {.lex_state = 30, .external_lex_state = 2},
  [1415] = {.lex_state = 30, .external_lex_state = 2},
  [1416] = {.lex_state = 30, .external_lex_state = 2},
  [1417] = {.lex_state = 30, .external_lex_state = 2},
  [1418] = {.lex_state = 30, .external_lex_state = 2},
  [1419] = {.lex_state = 30, .external_lex_state = 2},
  [1420] = {.lex_state = 30, .external_lex_state = 2},
  [1421] = {.lex_state = 30, .external_lex_state = 2},
  [1422] = {.lex_state = 30, .external_lex_state = 2},
  [1423] = {.lex_state = 30, .external_lex_state = 2},
  [1424] = {.lex_state = 30, .external_lex_state = 2},
  [1425] = {.lex_state = 30, .external_lex_state = 2},
  [1426] = {.lex_state = 30, .external_lex_state = 2},
  [1427] = {.lex_state = 30, .external_lex_state = 2},
  [1428] = {.lex_state = 30, .external_lex_state = 2},
  [1429] = {.lex_state = 30, .external_lex_state = 2},
  [1430] = {.lex_state = 30, .external_lex_state = 2},
  [1431] = {.lex_state = 30, .external_lex_state = 2},
  [1432] = {.lex_state = 30, .external_lex_state = 2},
  [1433] = {.lex_state = 30, .external_lex_state = 2},
  [1434] = {.lex_state = 30, .external_lex_state = 2},
  [1435] = {.lex_state = 30, .external_lex_state = 2},
  [1436] = {.lex_state = 30, .external_lex_state = 2},
  [1437] = {.lex_state = 30, .external_lex_state = 2},
  [1438] = {.lex_state = 30, .external_lex_state = 2},
  [1439] = {.lex_state = 30, .external_lex_state = 2},
  [1440] = {.lex_state = 30, .external_lex_state = 2},
  [1441] = {.lex_state = 30, .external_lex_state = 2},
  [1442] = {.lex_state = 30, .external_lex_state = 2},
  [1443] = {.lex_state = 30, .external_lex_state = 2},
  [1444] = {.lex_state = 30, .external_lex_state = 2},
  [1445] = {.lex_state = 30, .external_lex_state = 2},
  [1446] = {.lex_state = 30, .external_lex_state = 2},
  [1447] = {.lex_state = 30, .external_lex_state = 2},
  [1448] = {.lex_state = 30, .external_lex_state = 2},
  [1449] = {.lex_state = 30, .external_lex_state = 2},
  [1450] = {.lex_state = 30, .external_lex_state = 2},
  [1451] = {.lex_state = 30, .external_lex_state = 2},
  [1452] = {.lex_state = 30, .external_lex_state = 2},
  [1453] = {.lex_state = 30, .external_lex_state = 2},
  [1454] = {.lex_state = 30, .external_lex_state = 2},
  [1455] = {.lex_state = 30, .external_lex_state = 2},
  [1456] = {.lex_state = 30, .external_lex_state = 2},
  [1457] = {.lex_state = 30, .external_lex_state = 2},
  [1458] = {.lex_state = 30, .external_lex_state = 2},
  [1459] = {.lex_state = 30, .external_lex_state = 2},
  [1460] = {.lex_state = 30, .external_lex_state = 2},
  [1461] = {.lex_state = 30, .external_lex_state = 2},
  [1462] = {.lex_state = 30, .external_lex_state = 2},
  [1463] = {.lex_state = 30, .external_lex_state = 2},
  [1464] = {.lex_state = 30, .external_lex_state = 2},
  [1465] = {.lex_state = 30, .external_lex_state = 2},
  [1466] = {.lex_state = 30, .external_lex_state = 2},
  [1467] = {.lex_state = 30, .external_lex_state = 2},
  [1468] = {.lex_state = 30, .external_lex_state = 2},
  [1469] = {.lex_state = 30, .external_lex_state = 2},
  [1470] = {.lex_state = 30, .external_lex_state = 2},
  [1471] = {.lex_state = 30, .external_lex_state = 2},
  [1472] = {.lex_state = 30, .external_lex_state = 2},
  [1473] = {.lex_state = 30, .external_lex_state = 2},
  [1474] = {.lex_state = 30, .external_lex_state = 2},
  [1475] = {.lex_state = 30, .external_lex_state = 2},
  [1476] = {.lex_state = 30, .external_lex_state = 2},
  [1477] = {.lex_state = 30, .external_lex_state = 2},
  [1478] = {.lex_state = 30, .external_lex_state = 2},
  [1479] = {.lex_state = 30, .external_lex_state = 2},
  [1480] = {.lex_state = 30, .external_lex_state = 2},
  [1481] = {.lex_state = 30, .external_lex_state = 2},
  [1482] = {.lex_state = 30, .external_lex_state = 2},
  [1483] = {.lex_state = 30, .external_lex_state = 2},
  [1484] = {.lex_state = 30, .external_lex_state = 2},
  [1485] = {.lex_state = 30, .external_lex_state = 2},
  [1486] = {.lex_state = 30, .external_lex_state = 2},
  [1487] = {.lex_state = 30, .external_lex_state = 2},
  [1488] = {.lex_state = 30, .external_lex_state = 2},
  [1489] = {.lex_state = 151},
  [1490] = {.lex_state = 151},
  [1491] = {.lex_state = 151},
  [1492] = {.lex_state = 151},
  [1493] = {.lex_state = 29, .external_lex_state = 2},
  [1494] = {.lex_state = 151},
  [1495] = {.lex_state = 151},
  [1496] = {.lex_state = 151},
  [1497] = {.lex_state = 29, .external_lex_state = 2},
  [1498] = {.lex_state = 29, .external_lex_state = 2},
  [1499] = {.lex_state = 29, .external_lex_state = 2},
  [1500] = {.lex_state = 151},
  [1501] = {.lex_state = 154, .external_lex_state = 3},
  [1502] = {.lex_state = 29, .external_lex_state = 2},
  [1503] = {.lex_state = 29, .external_lex_state = 2},
  [1504] = {.lex_state = 154, .external_lex_state = 2},
  [1505] = {.lex_state = 151},
  [1506] = {.lex_state = 151},
  [1507] = {.lex_state = 151},
  [1508] = {.lex_state = 155, .external_lex_state = 2},
  [1509] = {.lex_state = 155, .external_lex_state = 3},
  [1510] = {.lex_state = 151},
  [1511] = {.lex_state = 151},
  [1512] = {.lex_state = 151, .external_lex_state = 7},
  [1513] = {.lex_state = 151},
  [1514] = {.lex_state = 155, .external_lex_state = 3},
  [1515] = {.lex_state = 151, .external_lex_state = 7},
  [1516] = {.lex_state = 151, .external_lex_state = 7},
  [1517] = {.lex_state = 151, .external_lex_state = 7},
  [1518] = {.lex_state = 154, .external_lex_state = 3},
  [1519] = {.lex_state = 151, .external_lex_state = 7},
  [1520] = {.lex_state = 151, .external_lex_state = 7},
  [1521] = {.lex_state = 151, .external_lex_state = 7},
  [1522] = {.lex_state = 151},
  [1523] = {.lex_state = 151, .external_lex_state = 7},
  [1524] = {.lex_state = 151, .external_lex_state = 7},
  [1525] = {.lex_state = 155, .external_lex_state = 2},
  [1526] = {.lex_state = 151},
  [1527] = {.lex_state = 151, .external_lex_state = 7},
  [1528] = {.lex_state = 151, .external_lex_state = 7},
  [1529] = {.lex_state = 154, .external_lex_state = 2},
  [1530] = {.lex_state = 155, .external_lex_state = 3},
  [1531] = {.lex_state = 155, .external_lex_state = 2},
  [1532] = {.lex_state = 154, .external_lex_state = 2},
  [1533] = {.lex_state = 154, .external_lex_state = 2},
  [1534] = {.lex_state = 154, .external_lex_state = 2},
  [1535] = {.lex_state = 154, .external_lex_state = 2},
  [1536] = {.lex_state = 151, .external_lex_state = 8},
  [1537] = {.lex_state = 151, .external_lex_state = 8},
  [1538] = {.lex_state = 154, .external_lex_state = 3},
  [1539] = {.lex_state = 154, .external_lex_state = 2},
  [1540] = {.lex_state = 154, .external_lex_state = 2},
  [1541] = {.lex_state = 154, .external_lex_state = 3},
  [1542] = {.lex_state = 154, .external_lex_state = 3},
  [1543] = {.lex_state = 154, .external_lex_state = 3},
  [1544] = {.lex_state = 154, .external_lex_state = 3},
  [1545] = {.lex_state = 154, .external_lex_state = 3},
  [1546] = {.lex_state = 154, .external_lex_state = 3},
  [1547] = {.lex_state = 154, .external_lex_state = 2},
  [1548] = {.lex_state = 154, .external_lex_state = 3},
  [1549] = {.lex_state = 154, .external_lex_state = 3},
  [1550] = {.lex_state = 154, .external_lex_state = 2},
  [1551] = {.lex_state = 154, .external_lex_state = 2},
  [1552] = {.lex_state = 154, .external_lex_state = 2},
  [1553] = {.lex_state = 154, .external_lex_state = 3},
  [1554] = {.lex_state = 151},
  [1555] = {.lex_state = 154, .external_lex_state = 2},
  [1556] = {.lex_state = 151, .external_lex_state = 8},
  [1557] = {.lex_state = 154, .external_lex_state = 3},
  [1558] = {.lex_state = 154, .external_lex_state = 3},
  [1559] = {.lex_state = 154, .external_lex_state = 2},
  [1560] = {.lex_state = 155, .external_lex_state = 3},
  [1561] = {.lex_state = 155, .external_lex_state = 3},
  [1562] = {.lex_state = 155, .external_lex_state = 2},
  [1563] = {.lex_state = 155, .external_lex_state = 2},
  [1564] = {.lex_state = 155, .external_lex_state = 3},
  [1565] = {.lex_state = 155, .external_lex_state = 2},
  [1566] = {.lex_state = 155, .external_lex_state = 3},
  [1567] = {.lex_state = 155, .external_lex_state = 3},
  [1568] = {.lex_state = 156, .external_lex_state = 3},
  [1569] = {.lex_state = 155, .external_lex_state = 2},
  [1570] = {.lex_state = 155, .external_lex_state = 3},
  [1571] = {.lex_state = 155, .external_lex_state = 2},
  [1572] = {.lex_state = 155, .external_lex_state = 2},
  [1573] = {.lex_state = 155, .external_lex_state = 3},
  [1574] = {.lex_state = 155, .external_lex_state = 2},
  [1575] = {.lex_state = 155, .external_lex_state = 3},
  [1576] = {.lex_state = 155, .external_lex_state = 3},
  [1577] = {.lex_state = 155, .external_lex_state = 3},
  [1578] = {.lex_state = 156, .external_lex_state = 2},
  [1579] = {.lex_state = 155, .external_lex_state = 2},
  [1580] = {.lex_state = 155, .external_lex_state = 2},
  [1581] = {.lex_state = 155, .external_lex_state = 2},
  [1582] = {.lex_state = 151},
  [1583] = {.lex_state = 156, .external_lex_state = 3},
  [1584] = {.lex_state = 156, .external_lex_state = 3},
  [1585] = {.lex_state = 156, .external_lex_state = 2},
  [1586] = {.lex_state = 151, .external_lex_state = 8},
  [1587] = {.lex_state = 156, .external_lex_state = 2},
  [1588] = {.lex_state = 151},
  [1589] = {.lex_state = 156, .external_lex_state = 2},
  [1590] = {.lex_state = 156, .external_lex_state = 2},
  [1591] = {.lex_state = 156, .external_lex_state = 3},
  [1592] = {.lex_state = 151, .external_lex_state = 8},
  [1593] = {.lex_state = 156, .external_lex_state = 3},
  [1594] = {.lex_state = 156, .external_lex_state = 3},
  [1595] = {.lex_state = 156, .external_lex_state = 3},
  [1596] = {.lex_state = 156, .external_lex_state = 2},
  [1597] = {.lex_state = 156, .external_lex_state = 2},
  [1598] = {.lex_state = 156, .external_lex_state = 2},
  [1599] = {.lex_state = 156, .external_lex_state = 3},
  [1600] = {.lex_state = 153, .external_lex_state = 2},
  [1601] = {.lex_state = 153, .external_lex_state = 2},
  [1602] = {.lex_state = 153, .external_lex_state = 2},
  [1603] = {.lex_state = 153, .external_lex_state = 3},
  [1604] = {.lex_state = 153, .external_lex_state = 2},
  [1605] = {.lex_state = 153, .external_lex_state = 3},
  [1606] = {.lex_state = 153, .external_lex_state = 3},
  [1607] = {.lex_state = 153, .external_lex_state = 2},
  [1608] = {.lex_state = 151},
  [1609] = {.lex_state = 153, .external_lex_state = 2},
  [1610] = {.lex_state = 153, .external_lex_state = 2},
  [1611] = {.lex_state = 153, .external_lex_state = 2},
  [1612] = {.lex_state = 153, .external_lex_state = 3},
  [1613] = {.lex_state = 153, .external_lex_state = 3},
  [1614] = {.lex_state = 153, .external_lex_state = 2},
  [1615] = {.lex_state = 153, .external_lex_state = 3},
  [1616] = {.lex_state = 153, .external_lex_state = 3},
  [1617] = {.lex_state = 153, .external_lex_state = 3},
  [1618] = {.lex_state = 153, .external_lex_state = 3},
  [1619] = {.lex_state = 153, .external_lex_state = 2},
  [1620] = {.lex_state = 153, .external_lex_state = 3},
  [1621] = {.lex_state = 153, .external_lex_state = 3},
  [1622] = {.lex_state = 153, .external_lex_state = 2},
  [1623] = {.lex_state = 153, .external_lex_state = 3},
  [1624] = {.lex_state = 153, .external_lex_state = 3},
  [1625] = {.lex_state = 153, .external_lex_state = 3},
  [1626] = {.lex_state = 153, .external_lex_state = 3},
  [1627] = {.lex_state = 153, .external_lex_state = 2},
  [1628] = {.lex_state = 153, .external_lex_state = 2},
  [1629] = {.lex_state = 153, .external_lex_state = 2},
  [1630] = {.lex_state = 153, .external_lex_state = 2},
  [1631] = {.lex_state = 153, .external_lex_state = 2},
  [1632] = {.lex_state = 153, .external_lex_state = 2},
  [1633] = {.lex_state = 153, .external_lex_state = 3},
  [1634] = {.lex_state = 153, .external_lex_state = 3},
  [1635] = {.lex_state = 153, .external_lex_state = 2},
  [1636] = {.lex_state = 153, .external_lex_state = 2},
  [1637] = {.lex_state = 153, .external_lex_state = 3},
  [1638] = {.lex_state = 153, .external_lex_state = 2},
  [1639] = {.lex_state = 153, .external_lex_state = 3},
  [1640] = {.lex_state = 153, .external_lex_state = 3},
  [1641] = {.lex_state = 153, .external_lex_state = 3},
  [1642] = {.lex_state = 153, .external_lex_state = 3},
  [1643] = {.lex_state = 153, .external_lex_state = 2},
  [1644] = {.lex_state = 153, .external_lex_state = 3},
  [1645] = {.lex_state = 153, .external_lex_state = 3},
  [1646] = {.lex_state = 153, .external_lex_state = 2},
  [1647] = {.lex_state = 153, .external_lex_state = 2},
  [1648] = {.lex_state = 153, .external_lex_state = 3},
  [1649] = {.lex_state = 153, .external_lex_state = 3},
  [1650] = {.lex_state = 153, .external_lex_state = 2},
  [1651] = {.lex_state = 153, .external_lex_state = 2},
  [1652] = {.lex_state = 153, .external_lex_state = 3},
  [1653] = {.lex_state = 153, .external_lex_state = 2},
  [1654] = {.lex_state = 153, .external_lex_state = 3},
  [1655] = {.lex_state = 153, .external_lex_state = 2},
  [1656] = {.lex_state = 153, .external_lex_state = 2},
  [1657] = {.lex_state = 153, .external_lex_state = 3},
  [1658] = {.lex_state = 153, .external_lex_state = 2},
  [1659] = {.lex_state = 151},
  [1660] = {.lex_state = 151},
  [1661] = {.lex_state = 151},
  [1662] = {.lex_state = 151},
  [1663] = {.lex_state = 151},
  [1664] = {.lex_state = 151},
  [1665] = {.lex_state = 151},
  [1666] = {.lex_state = 151},
  [1667] = {.lex_state = 151},
  [1668] = {.lex_state = 151},
  [1669] = {.lex_state = 151},
  [1670] = {.lex_state = 151},
  [1671] = {.lex_state = 151},
  [1672] = {.lex_state = 151},
  [1673] = {.lex_state = 151},
  [1674] = {.lex_state = 151},
  [1675] = {.lex_state = 151},
  [1676] = {.lex_state = 151},
  [1677] = {.lex_state = 151},
  [1678] = {.lex_state = 151},
  [1679] = {.lex_state = 151},
  [1680] = {.lex_state = 151},
  [1681] = {.lex_state = 151},
  [1682] = {.lex_state = 151},
  [1683] = {.lex_state = 151},
  [1684] = {.lex_state = 151},
  [1685] = {.lex_state = 151},
  [1686] = {.lex_state = 151},
  [1687] = {.lex_state = 151},
  [1688] = {.lex_state = 151},
  [1689] = {.lex_state = 151},
  [1690] = {.lex_state = 151},
  [1691] = {.lex_state = 151},
  [1692] = {.lex_state = 151},
  [1693] = {.lex_state = 151},
  [1694] = {.lex_state = 151},
  [1695] = {.lex_state = 151},
  [1696] = {.lex_state = 151},
  [1697] = {.lex_state = 151},
  [1698] = {.lex_state = 151},
  [1699] = {.lex_state = 151},
  [1700] = {.lex_state = 151},
  [1701] = {.lex_state = 151},
  [1702] = {.lex_state = 151},
  [1703] = {.lex_state = 151},
  [1704] = {.lex_state = 151},
  [1705] = {.lex_state = 151},
  [1706] = {.lex_state = 151},
  [1707] = {.lex_state = 151},
  [1708] = {.lex_state = 151},
  [1709] = {.lex_state = 151},
  [1710] = {.lex_state = 151},
  [1711] = {.lex_state = 151},
  [1712] = {.lex_state = 151},
  [1713] = {.lex_state = 29, .external_lex_state = 2},
  [1714] = {.lex_state = 151},
  [1715] = {.lex_state = 151},
  [1716] = {.lex_state = 151},
  [1717] = {.lex_state = 151},
  [1718] = {.lex_state = 151},
  [1719] = {.lex_state = 151},
  [1720] = {.lex_state = 151},
  [1721] = {.lex_state = 151},
  [1722] = {.lex_state = 151},
  [1723] = {.lex_state = 151},
  [1724] = {.lex_state = 151},
  [1725] = {.lex_state = 151},
  [1726] = {.lex_state = 151},
  [1727] = {.lex_state = 151},
  [1728] = {.lex_state = 151},
  [1729] = {.lex_state = 151},
  [1730] = {.lex_state = 151},
  [1731] = {.lex_state = 151},
  [1732] = {.lex_state = 151},
  [1733] = {.lex_state = 151},
  [1734] = {.lex_state = 151},
  [1735] = {.lex_state = 151},
  [1736] = {.lex_state = 151},
  [1737] = {.lex_state = 151, .external_lex_state = 9},
  [1738] = {.lex_state = 151},
  [1739] = {.lex_state = 151},
  [1740] = {.lex_state = 151},
  [1741] = {.lex_state = 151},
  [1742] = {.lex_state = 151},
  [1743] = {.lex_state = 29, .external_lex_state = 2},
  [1744] = {.lex_state = 151},
  [1745] = {.lex_state = 151},
  [1746] = {.lex_state = 151, .external_lex_state = 9},
  [1747] = {.lex_state = 151},
  [1748] = {.lex_state = 151},
  [1749] = {.lex_state = 151},
  [1750] = {.lex_state = 151},
  [1751] = {.lex_state = 151},
  [1752] = {.lex_state = 151},
  [1753] = {.lex_state = 151},
  [1754] = {.lex_state = 151},
  [1755] = {.lex_state = 151},
  [1756] = {.lex_state = 151},
  [1757] = {.lex_state = 151},
  [1758] = {.lex_state = 151},
  [1759] = {.lex_state = 151},
  [1760] = {.lex_state = 151},
  [1761] = {.lex_state = 151},
  [1762] = {.lex_state = 151},
  [1763] = {.lex_state = 151, .external_lex_state = 9},
  [1764] = {.lex_state = 151},
  [1765] = {.lex_state = 29, .external_lex_state = 2},
  [1766] = {.lex_state = 29, .external_lex_state = 2},
  [1767] = {.lex_state = 151},
  [1768] = {.lex_state = 151},
  [1769] = {.lex_state = 151},
  [1770] = {.lex_state = 151},
  [1771] = {.lex_state = 151},
  [1772] = {.lex_state = 151},
  [1773] = {.lex_state = 151},
  [1774] = {.lex_state = 151},
  [1775] = {.lex_state = 151},
  [1776] = {.lex_state = 151},
  [1777] = {.lex_state = 151},
  [1778] = {.lex_state = 151},
  [1779] = {.lex_state = 151},
  [1780] = {.lex_state = 151},
  [1781] = {.lex_state = 151},
  [1782] = {.lex_state = 151},
  [1783] = {.lex_state = 151},
  [1784] = {.lex_state = 151},
  [1785] = {.lex_state = 151},
  [1786] = {.lex_state = 151},
  [1787] = {.lex_state = 151},
  [1788] = {.lex_state = 151},
  [1789] = {.lex_state = 151},
  [1790] = {.lex_state = 151},
  [1791] = {.lex_state = 151},
  [1792] = {.lex_state = 151},
  [1793] = {.lex_state = 151},
  [1794] = {.lex_state = 151},
  [1795] = {.lex_state = 151},
  [1796] = {.lex_state = 151},
  [1797] = {.lex_state = 34, .external_lex_state = 2},
  [1798] = {.lex_state = 151},
  [1799] = {.lex_state = 151},
  [1800] = {.lex_state = 151},
  [1801] = {.lex_state = 151},
  [1802] = {.lex_state = 151},
  [1803] = {.lex_state = 151},
  [1804] = {.lex_state = 151},
  [1805] = {.lex_state = 151},
  [1806] = {.lex_state = 151},
  [1807] = {.lex_state = 151},
  [1808] = {.lex_state = 151},
  [1809] = {.lex_state = 151},
  [1810] = {.lex_state = 151},
  [1811] = {.lex_state = 151},
  [1812] = {.lex_state = 151},
  [1813] = {.lex_state = 151},
  [1814] = {.lex_state = 151},
  [1815] = {.lex_state = 151},
  [1816] = {.lex_state = 151},
  [1817] = {.lex_state = 151},
  [1818] = {.lex_state = 151},
  [1819] = {.lex_state = 151},
  [1820] = {.lex_state = 151},
  [1821] = {.lex_state = 151},
  [1822] = {.lex_state = 34, .external_lex_state = 2},
  [1823] = {.lex_state = 151},
  [1824] = {.lex_state = 151},
  [1825] = {.lex_state = 151},
  [1826] = {.lex_state = 151},
  [1827] = {.lex_state = 151},
  [1828] = {.lex_state = 151},
  [1829] = {.lex_state = 151},
  [1830] = {.lex_state = 151},
  [1831] = {.lex_state = 151},
  [1832] = {.lex_state = 151},
  [1833] = {.lex_state = 151},
  [1834] = {.lex_state = 151},
  [1835] = {.lex_state = 151},
  [1836] = {.lex_state = 151},
  [1837] = {.lex_state = 151},
  [1838] = {.lex_state = 151},
  [1839] = {.lex_state = 151},
  [1840] = {.lex_state = 151},
  [1841] = {.lex_state = 151},
  [1842] = {.lex_state = 151},
  [1843] = {.lex_state = 151},
  [1844] = {.lex_state = 151},
  [1845] = {.lex_state = 151},
  [1846] = {.lex_state = 151},
  [1847] = {.lex_state = 151},
  [1848] = {.lex_state = 151},
  [1849] = {.lex_state = 151},
  [1850] = {.lex_state = 151},
  [1851] = {.lex_state = 151},
  [1852] = {.lex_state = 151},
  [1853] = {.lex_state = 151},
  [1854] = {.lex_state = 151},
  [1855] = {.lex_state = 151},
  [1856] = {.lex_state = 151},
  [1857] = {.lex_state = 151},
  [1858] = {.lex_state = 151},
  [1859] = {.lex_state = 151},
  [1860] = {.lex_state = 151},
  [1861] = {.lex_state = 151},
  [1862] = {.lex_state = 151},
  [1863] = {.lex_state = 151},
  [1864] = {.lex_state = 151},
  [1865] = {.lex_state = 151},
  [1866] = {.lex_state = 151},
  [1867] = {.lex_state = 151},
  [1868] = {.lex_state = 151},
  [1869] = {.lex_state = 151},
  [1870] = {.lex_state = 151},
  [1871] = {.lex_state = 151},
  [1872] = {.lex_state = 151},
  [1873] = {.lex_state = 151},
  [1874] = {.lex_state = 151},
  [1875] = {.lex_state = 151},
  [1876] = {.lex_state = 151},
  [1877] = {.lex_state = 151},
  [1878] = {.lex_state = 29, .external_lex_state = 2},
  [1879] = {.lex_state = 151},
  [1880] = {.lex_state = 151, .external_lex_state = 7},
  [1881] = {.lex_state = 151, .external_lex_state = 7},
  [1882] = {.lex_state = 151, .external_lex_state = 7},
  [1883] = {.lex_state = 151},
  [1884] = {.lex_state = 151},
  [1885] = {.lex_state = 151},
  [1886] = {.lex_state = 151, .external_lex_state = 7},
  [1887] = {.lex_state = 151},
  [1888] = {.lex_state = 151},
  [1889] = {.lex_state = 151},
  [1890] = {.lex_state = 151},
  [1891] = {.lex_state = 151},
  [1892] = {.lex_state = 151},
  [1893] = {.lex_state = 151},
  [1894] = {.lex_state = 151},
  [1895] = {.lex_state = 151},
  [1896] = {.lex_state = 151},
  [1897] = {.lex_state = 151},
  [1898] = {.lex_state = 151},
  [1899] = {.lex_state = 151},
  [1900] = {.lex_state = 151},
  [1901] = {.lex_state = 151, .external_lex_state = 7},
  [1902] = {.lex_state = 151},
  [1903] = {.lex_state = 151},
  [1904] = {.lex_state = 151},
  [1905] = {.lex_state = 151},
  [1906] = {.lex_state = 151},
  [1907] = {.lex_state = 151},
  [1908] = {.lex_state = 151},
  [1909] = {.lex_state = 151},
  [1910] = {.lex_state = 151},
  [1911] = {.lex_state = 151},
  [1912] = {.lex_state = 151},
  [1913] = {.lex_state = 151},
  [1914] = {.lex_state = 151},
  [1915] = {.lex_state = 151},
  [1916] = {.lex_state = 151},
  [1917] = {.lex_state = 151},
  [1918] = {.lex_state = 151},
  [1919] = {.lex_state = 151},
  [1920] = {.lex_state = 151},
  [1921] = {.lex_state = 151},
  [1922] = {.lex_state = 151},
  [1923] = {.lex_state = 151},
  [1924] = {.lex_state = 151},
  [1925] = {.lex_state = 151},
  [1926] = {.lex_state = 151},
  [1927] = {.lex_state = 151},
  [1928] = {.lex_state = 151},
  [1929] = {.lex_state = 151},
  [1930] = {.lex_state = 151},
  [1931] = {.lex_state = 151},
  [1932] = {.lex_state = 151},
  [1933] = {.lex_state = 151},
  [1934] = {.lex_state = 151},
  [1935] = {.lex_state = 151},
  [1936] = {.lex_state = 151},
  [1937] = {.lex_state = 151},
  [1938] = {.lex_state = 29, .external_lex_state = 2},
  [1939] = {.lex_state = 151},
  [1940] = {.lex_state = 151},
  [1941] = {.lex_state = 151},
  [1942] = {.lex_state = 151},
  [1943] = {.lex_state = 151, .external_lex_state = 7},
  [1944] = {.lex_state = 151},
  [1945] = {.lex_state = 151, .external_lex_state = 9},
  [1946] = {.lex_state = 151},
  [1947] = {.lex_state = 151},
  [1948] = {.lex_state = 151},
  [1949] = {.lex_state = 151},
  [1950] = {.lex_state = 151},
  [1951] = {.lex_state = 151},
  [1952] = {.lex_state = 151},
  [1953] = {.lex_state = 151},
  [1954] = {.lex_state = 151},
  [1955] = {.lex_state = 151},
  [1956] = {.lex_state = 151},
  [1957] = {.lex_state = 151},
  [1958] = {.lex_state = 151},
  [1959] = {.lex_state = 151},
  [1960] = {.lex_state = 151},
  [1961] = {.lex_state = 151},
  [1962] = {.lex_state = 151},
  [1963] = {.lex_state = 151},
  [1964] = {.lex_state = 151},
  [1965] = {.lex_state = 151},
  [1966] = {.lex_state = 151},
  [1967] = {.lex_state = 151, .external_lex_state = 7},
  [1968] = {.lex_state = 151},
  [1969] = {.lex_state = 151},
  [1970] = {.lex_state = 151},
  [1971] = {.lex_state = 151},
  [1972] = {.lex_state = 151},
  [1973] = {.lex_state = 151},
  [1974] = {.lex_state = 151},
  [1975] = {.lex_state = 151},
  [1976] = {.lex_state = 151},
  [1977] = {.lex_state = 29, .external_lex_state = 2},
  [1978] = {.lex_state = 151},
  [1979] = {.lex_state = 151},
  [1980] = {.lex_state = 151},
  [1981] = {.lex_state = 151},
  [1982] = {.lex_state = 151, .external_lex_state = 9},
  [1983] = {.lex_state = 151},
  [1984] = {.lex_state = 151},
  [1985] = {.lex_state = 151},
  [1986] = {.lex_state = 151, .external_lex_state = 7},
  [1987] = {.lex_state = 151},
  [1988] = {.lex_state = 151},
  [1989] = {.lex_state = 151},
  [1990] = {.lex_state = 151},
  [1991] = {.lex_state = 151},
  [1992] = {.lex_state = 151},
  [1993] = {.lex_state = 151},
  [1994] = {.lex_state = 151, .external_lex_state = 7},
  [1995] = {.lex_state = 151},
  [1996] = {.lex_state = 151},
  [1997] = {.lex_state = 151, .external_lex_state = 7},
  [1998] = {.lex_state = 151, .external_lex_state = 7},
  [1999] = {.lex_state = 151, .external_lex_state = 7},
  [2000] = {.lex_state = 151},
  [2001] = {.lex_state = 151, .external_lex_state = 7},
  [2002] = {.lex_state = 151},
  [2003] = {.lex_state = 151},
  [2004] = {.lex_state = 151},
  [2005] = {.lex_state = 151},
  [2006] = {.lex_state = 151},
  [2007] = {.lex_state = 151},
  [2008] = {.lex_state = 151, .external_lex_state = 7},
  [2009] = {.lex_state = 151},
  [2010] = {.lex_state = 151},
  [2011] = {.lex_state = 151},
  [2012] = {.lex_state = 151, .external_lex_state = 7},
  [2013] = {.lex_state = 151, .external_lex_state = 7},
  [2014] = {.lex_state = 151, .external_lex_state = 7},
  [2015] = {.lex_state = 151, .external_lex_state = 7},
  [2016] = {.lex_state = 151, .external_lex_state = 7},
  [2017] = {.lex_state = 151},
  [2018] = {.lex_state = 151},
  [2019] = {.lex_state = 151},
  [2020] = {.lex_state = 151, .external_lex_state = 7},
  [2021] = {.lex_state = 151},
  [2022] = {.lex_state = 151},
  [2023] = {.lex_state = 151},
  [2024] = {.lex_state = 151},
  [2025] = {.lex_state = 151},
  [2026] = {.lex_state = 151},
  [2027] = {.lex_state = 151},
  [2028] = {.lex_state = 151},
  [2029] = {.lex_state = 151},
  [2030] = {.lex_state = 151},
  [2031] = {.lex_state = 151},
  [2032] = {.lex_state = 151},
  [2033] = {.lex_state = 151},
  [2034] = {.lex_state = 151},
  [2035] = {.lex_state = 151},
  [2036] = {.lex_state = 151, .external_lex_state = 7},
  [2037] = {.lex_state = 151, .external_lex_state = 7},
  [2038] = {.lex_state = 151},
  [2039] = {.lex_state = 151},
  [2040] = {.lex_state = 151},
  [2041] = {.lex_state = 151},
  [2042] = {.lex_state = 151},
  [2043] = {.lex_state = 151},
  [2044] = {.lex_state = 151},
  [2045] = {.lex_state = 151},
  [2046] = {.lex_state = 151},
  [2047] = {.lex_state = 151},
  [2048] = {.lex_state = 151, .external_lex_state = 7},
  [2049] = {.lex_state = 151},
  [2050] = {.lex_state = 151},
  [2051] = {.lex_state = 151},
  [2052] = {.lex_state = 151},
  [2053] = {.lex_state = 151},
  [2054] = {.lex_state = 151, .external_lex_state = 7},
  [2055] = {.lex_state = 151},
  [2056] = {.lex_state = 151},
  [2057] = {.lex_state = 151, .external_lex_state = 7},
  [2058] = {.lex_state = 151},
  [2059] = {.lex_state = 151},
  [2060] = {.lex_state = 151},
  [2061] = {.lex_state = 151},
  [2062] = {.lex_state = 151},
  [2063] = {.lex_state = 151, .external_lex_state = 7},
  [2064] = {.lex_state = 151},
  [2065] = {.lex_state = 151},
  [2066] = {.lex_state = 151, .external_lex_state = 7},
  [2067] = {.lex_state = 151},
  [2068] = {.lex_state = 151},
  [2069] = {.lex_state = 151},
  [2070] = {.lex_state = 151},
  [2071] = {.lex_state = 151, .external_lex_state = 7},
  [2072] = {.lex_state = 151},
  [2073] = {.lex_state = 151},
  [2074] = {.lex_state = 151, .external_lex_state = 7},
  [2075] = {.lex_state = 151},
  [2076] = {.lex_state = 151},
  [2077] = {.lex_state = 151},
  [2078] = {.lex_state = 151},
  [2079] = {.lex_state = 151},
  [2080] = {.lex_state = 151},
  [2081] = {.lex_state = 151},
  [2082] = {.lex_state = 151},
  [2083] = {.lex_state = 151},
  [2084] = {.lex_state = 151},
  [2085] = {.lex_state = 151},
  [2086] = {.lex_state = 151},
  [2087] = {.lex_state = 151},
  [2088] = {.lex_state = 151},
  [2089] = {.lex_state = 151},
  [2090] = {.lex_state = 151},
  [2091] = {.lex_state = 151},
  [2092] = {.lex_state = 151},
  [2093] = {.lex_state = 151},
  [2094] = {.lex_state = 151},
  [2095] = {.lex_state = 151},
  [2096] = {.lex_state = 151},
  [2097] = {.lex_state = 151},
  [2098] = {.lex_state = 151},
  [2099] = {.lex_state = 151},
  [2100] = {.lex_state = 151, .external_lex_state = 7},
  [2101] = {.lex_state = 151},
  [2102] = {.lex_state = 151},
  [2103] = {.lex_state = 151, .external_lex_state = 7},
  [2104] = {.lex_state = 151},
  [2105] = {.lex_state = 151},
  [2106] = {.lex_state = 151},
  [2107] = {.lex_state = 151, .external_lex_state = 7},
  [2108] = {.lex_state = 151, .external_lex_state = 7},
  [2109] = {.lex_state = 151},
  [2110] = {.lex_state = 151},
  [2111] = {.lex_state = 151},
  [2112] = {.lex_state = 151},
  [2113] = {.lex_state = 151},
  [2114] = {.lex_state = 151, .external_lex_state = 7},
  [2115] = {.lex_state = 151},
  [2116] = {.lex_state = 151},
  [2117] = {.lex_state = 151},
  [2118] = {.lex_state = 151},
  [2119] = {.lex_state = 151, .external_lex_state = 7},
  [2120] = {.lex_state = 151, .external_lex_state = 7},
  [2121] = {.lex_state = 151, .external_lex_state = 7},
  [2122] = {.lex_state = 151, .external_lex_state = 7},
  [2123] = {.lex_state = 151},
  [2124] = {.lex_state = 151},
  [2125] = {.lex_state = 151},
  [2126] = {.lex_state = 151},
  [2127] = {.lex_state = 151},
  [2128] = {.lex_state = 151},
  [2129] = {.lex_state = 151},
  [2130] = {.lex_state = 151},
  [2131] = {.lex_state = 151},
  [2132] = {.lex_state = 151},
  [2133] = {.lex_state = 151},
  [2134] = {.lex_state = 151},
  [2135] = {.lex_state = 151},
  [2136] = {.lex_state = 151},
  [2137] = {.lex_state = 151},
  [2138] = {.lex_state = 151},
  [2139] = {.lex_state = 151},
  [2140] = {.lex_state = 151},
  [2141] = {.lex_state = 151},
  [2142] = {.lex_state = 151},
  [2143] = {.lex_state = 151},
  [2144] = {.lex_state = 151},
  [2145] = {.lex_state = 151},
  [2146] = {.lex_state = 151},
  [2147] = {.lex_state = 151, .external_lex_state = 7},
  [2148] = {.lex_state = 151},
  [2149] = {.lex_state = 151},
  [2150] = {.lex_state = 151},
  [2151] = {.lex_state = 151},
  [2152] = {.lex_state = 151},
  [2153] = {.lex_state = 151, .external_lex_state = 7},
  [2154] = {.lex_state = 151},
  [2155] = {.lex_state = 151},
  [2156] = {.lex_state = 151},
  [2157] = {.lex_state = 151},
  [2158] = {.lex_state = 151},
  [2159] = {.lex_state = 151},
  [2160] = {.lex_state = 151},
  [2161] = {.lex_state = 151},
  [2162] = {.lex_state = 151, .external_lex_state = 7},
  [2163] = {.lex_state = 151},
  [2164] = {.lex_state = 151, .external_lex_state = 7},
  [2165] = {.lex_state = 151, .external_lex_state = 7},
  [2166] = {.lex_state = 151, .external_lex_state = 7},
  [2167] = {.lex_state = 151},
  [2168] = {.lex_state = 151, .external_lex_state = 7},
  [2169] = {.lex_state = 151, .external_lex_state = 7},
  [2170] = {.lex_state = 151, .external_lex_state = 7},
  [2171] = {.lex_state = 151, .external_lex_state = 7},
  [2172] = {.lex_state = 151},
  [2173] = {.lex_state = 151, .external_lex_state = 7},
  [2174] = {.lex_state = 151},
  [2175] = {.lex_state = 151, .external_lex_state = 7},
  [2176] = {.lex_state = 151, .external_lex_state = 7},
  [2177] = {.lex_state = 151, .external_lex_state = 7},
  [2178] = {.lex_state = 151, .external_lex_state = 7},
  [2179] = {.lex_state = 151, .external_lex_state = 7},
  [2180] = {.lex_state = 151},
  [2181] = {.lex_state = 151, .external_lex_state = 7},
  [2182] = {.lex_state = 151, .external_lex_state = 7},
  [2183] = {.lex_state = 151, .external_lex_state = 7},
  [2184] = {.lex_state = 151, .external_lex_state = 7},
  [2185] = {.lex_state = 151, .external_lex_state = 7},
  [2186] = {.lex_state = 151, .external_lex_state = 7},
  [2187] = {.lex_state = 151, .external_lex_state = 7},
  [2188] = {.lex_state = 151, .external_lex_state = 7},
  [2189] = {.lex_state = 151},
  [2190] = {.lex_state = 151, .external_lex_state = 7},
  [2191] = {.lex_state = 151, .external_lex_state = 7},
  [2192] = {.lex_state = 151, .external_lex_state = 7},
  [2193] = {.lex_state = 151},
  [2194] = {.lex_state = 151, .external_lex_state = 7},
  [2195] = {.lex_state = 151, .external_lex_state = 7},
  [2196] = {.lex_state = 151},
  [2197] = {.lex_state = 151},
  [2198] = {.lex_state = 151},
  [2199] = {.lex_state = 151},
  [2200] = {.lex_state = 151},
  [2201] = {.lex_state = 151},
  [2202] = {.lex_state = 151},
  [2203] = {.lex_state = 151},
  [2204] = {.lex_state = 151},
  [2205] = {.lex_state = 151},
  [2206] = {.lex_state = 151},
  [2207] = {.lex_state = 151},
  [2208] = {.lex_state = 151},
  [2209] = {.lex_state = 151},
  [2210] = {.lex_state = 151},
  [2211] = {.lex_state = 151},
  [2212] = {.lex_state = 151, .external_lex_state = 7},
  [2213] = {.lex_state = 151},
  [2214] = {.lex_state = 151},
  [2215] = {.lex_state = 151, .external_lex_state = 7},
  [2216] = {.lex_state = 151, .external_lex_state = 7},
  [2217] = {.lex_state = 151, .external_lex_state = 7},
  [2218] = {.lex_state = 151, .external_lex_state = 7},
  [2219] = {.lex_state = 151, .external_lex_state = 7},
  [2220] = {.lex_state = 151, .external_lex_state = 7},
  [2221] = {.lex_state = 151, .external_lex_state = 7},
  [2222] = {.lex_state = 151, .external_lex_state = 7},
  [2223] = {.lex_state = 151, .external_lex_state = 7},
  [2224] = {.lex_state = 151},
  [2225] = {.lex_state = 151},
  [2226] = {.lex_state = 151},
  [2227] = {.lex_state = 151, .external_lex_state = 7},
  [2228] = {.lex_state = 151, .external_lex_state = 7},
  [2229] = {.lex_state = 151, .external_lex_state = 7},
  [2230] = {.lex_state = 151, .external_lex_state = 7},
  [2231] = {.lex_state = 151, .external_lex_state = 7},
  [2232] = {.lex_state = 151, .external_lex_state = 7},
  [2233] = {.lex_state = 151, .external_lex_state = 7},
  [2234] = {.lex_state = 151, .external_lex_state = 7},
  [2235] = {.lex_state = 151},
  [2236] = {.lex_state = 151},
  [2237] = {.lex_state = 151},
  [2238] = {.lex_state = 151},
  [2239] = {.lex_state = 151, .external_lex_state = 7},
  [2240] = {.lex_state = 151, .external_lex_state = 7},
  [2241] = {.lex_state = 151, .external_lex_state = 7},
  [2242] = {.lex_state = 151, .external_lex_state = 7},
  [2243] = {.lex_state = 151, .external_lex_state = 7},
  [2244] = {.lex_state = 151, .external_lex_state = 7},
  [2245] = {.lex_state = 151, .external_lex_state = 7},
  [2246] = {.lex_state = 151},
  [2247] = {.lex_state = 151, .external_lex_state = 7},
  [2248] = {.lex_state = 151},
  [2249] = {.lex_state = 151, .external_lex_state = 7},
  [2250] = {.lex_state = 151, .external_lex_state = 7},
  [2251] = {.lex_state = 151},
  [2252] = {.lex_state = 151},
  [2253] = {.lex_state = 151, .external_lex_state = 7},
  [2254] = {.lex_state = 151, .external_lex_state = 7},
  [2255] = {.lex_state = 151},
  [2256] = {.lex_state = 151},
  [2257] = {.lex_state = 151, .external_lex_state = 7},
  [2258] = {.lex_state = 151, .external_lex_state = 7},
  [2259] = {.lex_state = 151, .external_lex_state = 7},
  [2260] = {.lex_state = 151},
  [2261] = {.lex_state = 151},
  [2262] = {.lex_state = 151},
  [2263] = {.lex_state = 151},
  [2264] = {.lex_state = 151},
  [2265] = {.lex_state = 151, .external_lex_state = 7},
  [2266] = {.lex_state = 151, .external_lex_state = 7},
  [2267] = {.lex_state = 151, .external_lex_state = 7},
  [2268] = {.lex_state = 151, .external_lex_state = 7},
  [2269] = {.lex_state = 151},
  [2270] = {.lex_state = 151, .external_lex_state = 7},
  [2271] = {.lex_state = 151, .external_lex_state = 7},
  [2272] = {.lex_state = 151, .external_lex_state = 7},
  [2273] = {.lex_state = 151, .external_lex_state = 7},
  [2274] = {.lex_state = 151, .external_lex_state = 7},
  [2275] = {.lex_state = 151},
  [2276] = {.lex_state = 151},
  [2277] = {.lex_state = 151},
  [2278] = {.lex_state = 151},
  [2279] = {.lex_state = 151},
  [2280] = {.lex_state = 151, .external_lex_state = 7},
  [2281] = {.lex_state = 151},
  [2282] = {.lex_state = 151},
  [2283] = {.lex_state = 151},
  [2284] = {.lex_state = 151},
  [2285] = {.lex_state = 151, .external_lex_state = 7},
  [2286] = {.lex_state = 151, .external_lex_state = 7},
  [2287] = {.lex_state = 151},
  [2288] = {.lex_state = 151},
  [2289] = {.lex_state = 151, .external_lex_state = 7},
  [2290] = {.lex_state = 151},
  [2291] = {.lex_state = 151},
  [2292] = {.lex_state = 151},
  [2293] = {.lex_state = 151, .external_lex_state = 7},
  [2294] = {.lex_state = 151, .external_lex_state = 7},
  [2295] = {.lex_state = 151},
  [2296] = {.lex_state = 151},
  [2297] = {.lex_state = 29, .external_lex_state = 2},
  [2298] = {.lex_state = 151},
  [2299] = {.lex_state = 151},
  [2300] = {.lex_state = 151},
  [2301] = {.lex_state = 151},
  [2302] = {.lex_state = 29, .external_lex_state = 2},
  [2303] = {.lex_state = 151},
  [2304] = {.lex_state = 29, .external_lex_state = 2},
  [2305] = {.lex_state = 151},
  [2306] = {.lex_state = 151},
  [2307] = {.lex_state = 29, .external_lex_state = 2},
  [2308] = {.lex_state = 151},
  [2309] = {.lex_state = 29, .external_lex_state = 2},
  [2310] = {.lex_state = 151},
  [2311] = {.lex_state = 29, .external_lex_state = 2},
  [2312] = {.lex_state = 151},
  [2313] = {.lex_state = 151},
  [2314] = {.lex_state = 29, .external_lex_state = 2},
  [2315] = {.lex_state = 151},
  [2316] = {.lex_state = 29, .external_lex_state = 2},
  [2317] = {.lex_state = 29, .external_lex_state = 2},
  [2318] = {.lex_state = 151},
  [2319] = {.lex_state = 151},
  [2320] = {.lex_state = 29, .external_lex_state = 2},
  [2321] = {.lex_state = 151},
  [2322] = {.lex_state = 29, .external_lex_state = 2},
  [2323] = {.lex_state = 151},
  [2324] = {.lex_state = 29, .external_lex_state = 2},
  [2325] = {.lex_state = 151},
  [2326] = {.lex_state = 29, .external_lex_state = 2},
  [2327] = {.lex_state = 151},
  [2328] = {.lex_state = 29, .external_lex_state = 2},
  [2329] = {.lex_state = 29, .external_lex_state = 2},
  [2330] = {.lex_state = 29, .external_lex_state = 2},
  [2331] = {.lex_state = 29, .external_lex_state = 2},
  [2332] = {.lex_state = 29, .external_lex_state = 2},
  [2333] = {.lex_state = 29, .external_lex_state = 2},
  [2334] = {.lex_state = 151},
  [2335] = {.lex_state = 151},
  [2336] = {.lex_state = 29, .external_lex_state = 2},
  [2337] = {.lex_state = 29, .external_lex_state = 2},
  [2338] = {.lex_state = 151},
  [2339] = {.lex_state = 151, .external_lex_state = 7},
  [2340] = {.lex_state = 151, .external_lex_state = 7},
  [2341] = {.lex_state = 151},
  [2342] = {.lex_state = 151},
  [2343] = {.lex_state = 151, .external_lex_state = 7},
  [2344] = {.lex_state = 151},
  [2345] = {.lex_state = 151},
  [2346] = {.lex_state = 151, .external_lex_state = 7},
  [2347] = {.lex_state = 151},
  [2348] = {.lex_state = 151},
  [2349] = {.lex_state = 151},
  [2350] = {.lex_state = 151},
  [2351] = {.lex_state = 151},
  [2352] = {.lex_state = 151},
  [2353] = {.lex_state = 151},
  [2354] = {.lex_state = 151},
  [2355] = {.lex_state = 151},
  [2356] = {.lex_state = 151, .external_lex_state = 7},
  [2357] = {.lex_state = 151},
  [2358] = {.lex_state = 151},
  [2359] = {.lex_state = 151, .external_lex_state = 7},
  [2360] = {.lex_state = 151},
  [2361] = {.lex_state = 151},
  [2362] = {.lex_state = 151},
  [2363] = {.lex_state = 151},
  [2364] = {.lex_state = 151},
  [2365] = {.lex_state = 151},
  [2366] = {.lex_state = 151},
  [2367] = {.lex_state = 151},
  [2368] = {.lex_state = 151},
  [2369] = {.lex_state = 151, .external_lex_state = 7},
  [2370] = {.lex_state = 151},
  [2371] = {.lex_state = 151},
  [2372] = {.lex_state = 151},
  [2373] = {.lex_state = 151},
  [2374] = {.lex_state = 151},
  [2375] = {.lex_state = 151},
  [2376] = {.lex_state = 151, .external_lex_state = 7},
  [2377] = {.lex_state = 151},
  [2378] = {.lex_state = 151},
  [2379] = {.lex_state = 151},
  [2380] = {.lex_state = 151},
  [2381] = {.lex_state = 151},
  [2382] = {.lex_state = 151, .external_lex_state = 7},
  [2383] = {.lex_state = 151},
  [2384] = {.lex_state = 151},
  [2385] = {.lex_state = 151},
  [2386] = {.lex_state = 151},
  [2387] = {.lex_state = 151},
  [2388] = {.lex_state = 151},
  [2389] = {.lex_state = 151},
  [2390] = {.lex_state = 151},
  [2391] = {.lex_state = 151},
  [2392] = {.lex_state = 151},
  [2393] = {.lex_state = 151},
  [2394] = {.lex_state = 151},
  [2395] = {.lex_state = 151},
  [2396] = {.lex_state = 151},
  [2397] = {.lex_state = 151},
  [2398] = {.lex_state = 151},
  [2399] = {.lex_state = 151},
  [2400] = {.lex_state = 151},
  [2401] = {.lex_state = 151},
  [2402] = {.lex_state = 151},
  [2403] = {.lex_state = 151},
  [2404] = {.lex_state = 151},
  [2405] = {.lex_state = 151},
  [2406] = {.lex_state = 151},
  [2407] = {.lex_state = 151},
  [2408] = {.lex_state = 151},
  [2409] = {.lex_state = 151},
  [2410] = {.lex_state = 151},
  [2411] = {.lex_state = 151},
  [2412] = {.lex_state = 151},
  [2413] = {.lex_state = 151},
  [2414] = {.lex_state = 151},
  [2415] = {.lex_state = 151},
  [2416] = {.lex_state = 151},
  [2417] = {.lex_state = 151},
  [2418] = {.lex_state = 151},
  [2419] = {.lex_state = 151},
  [2420] = {.lex_state = 151},
  [2421] = {.lex_state = 151},
  [2422] = {.lex_state = 151},
  [2423] = {.lex_state = 151},
  [2424] = {.lex_state = 151},
  [2425] = {.lex_state = 151},
  [2426] = {.lex_state = 151},
  [2427] = {.lex_state = 151},
  [2428] = {.lex_state = 151},
  [2429] = {.lex_state = 151},
  [2430] = {.lex_state = 151},
  [2431] = {.lex_state = 151},
  [2432] = {.lex_state = 151},
  [2433] = {.lex_state = 151},
  [2434] = {.lex_state = 151},
  [2435] = {.lex_state = 151},
  [2436] = {.lex_state = 151},
  [2437] = {.lex_state = 151},
  [2438] = {.lex_state = 151},
  [2439] = {.lex_state = 151},
  [2440] = {.lex_state = 151},
  [2441] = {.lex_state = 151},
  [2442] = {.lex_state = 151},
  [2443] = {.lex_state = 151},
  [2444] = {.lex_state = 151},
  [2445] = {.lex_state = 151},
  [2446] = {.lex_state = 151},
  [2447] = {.lex_state = 151},
  [2448] = {.lex_state = 151},
  [2449] = {.lex_state = 151},
  [2450] = {.lex_state = 151},
  [2451] = {.lex_state = 151},
  [2452] = {.lex_state = 151},
  [2453] = {.lex_state = 151},
  [2454] = {.lex_state = 151},
  [2455] = {.lex_state = 151},
  [2456] = {.lex_state = 151},
  [2457] = {.lex_state = 151},
  [2458] = {.lex_state = 151},
  [2459] = {.lex_state = 151},
  [2460] = {.lex_state = 151},
  [2461] = {.lex_state = 151},
  [2462] = {.lex_state = 151},
  [2463] = {.lex_state = 151},
  [2464] = {.lex_state = 151},
  [2465] = {.lex_state = 151},
  [2466] = {.lex_state = 151},
  [2467] = {.lex_state = 151},
  [2468] = {.lex_state = 151},
  [2469] = {.lex_state = 151},
  [2470] = {.lex_state = 151},
  [2471] = {.lex_state = 151},
  [2472] = {.lex_state = 151},
  [2473] = {.lex_state = 151},
  [2474] = {.lex_state = 151},
  [2475] = {.lex_state = 151},
  [2476] = {.lex_state = 151, .external_lex_state = 7},
  [2477] = {.lex_state = 151},
  [2478] = {.lex_state = 151},
  [2479] = {.lex_state = 151},
  [2480] = {.lex_state = 151},
  [2481] = {.lex_state = 151},
  [2482] = {.lex_state = 151},
  [2483] = {.lex_state = 151},
  [2484] = {.lex_state = 151},
  [2485] = {.lex_state = 151},
  [2486] = {.lex_state = 151},
  [2487] = {.lex_state = 151},
  [2488] = {.lex_state = 151},
  [2489] = {.lex_state = 151},
  [2490] = {.lex_state = 151},
  [2491] = {.lex_state = 151},
  [2492] = {.lex_state = 151},
  [2493] = {.lex_state = 151},
  [2494] = {.lex_state = 151},
  [2495] = {.lex_state = 151},
  [2496] = {.lex_state = 151},
  [2497] = {.lex_state = 151},
  [2498] = {.lex_state = 151},
  [2499] = {.lex_state = 151},
  [2500] = {.lex_state = 151},
  [2501] = {.lex_state = 151},
  [2502] = {.lex_state = 151},
  [2503] = {.lex_state = 151},
  [2504] = {.lex_state = 151},
  [2505] = {.lex_state = 151},
  [2506] = {.lex_state = 151},
  [2507] = {.lex_state = 151},
  [2508] = {.lex_state = 151},
  [2509] = {.lex_state = 151},
  [2510] = {.lex_state = 151},
  [2511] = {.lex_state = 151},
  [2512] = {.lex_state = 151},
  [2513] = {.lex_state = 46},
  [2514] = {.lex_state = 46},
  [2515] = {.lex_state = 46},
  [2516] = {.lex_state = 46},
  [2517] = {.lex_state = 46},
  [2518] = {.lex_state = 46},
  [2519] = {.lex_state = 46},
  [2520] = {.lex_state = 151},
  [2521] = {.lex_state = 151},
  [2522] = {.lex_state = 151},
  [2523] = {.lex_state = 151},
  [2524] = {.lex_state = 151},
  [2525] = {.lex_state = 151},
  [2526] = {.lex_state = 151},
  [2527] = {.lex_state = 151},
  [2528] = {.lex_state = 151},
  [2529] = {.lex_state = 151},
  [2530] = {.lex_state = 151},
  [2531] = {.lex_state = 151},
  [2532] = {.lex_state = 151},
  [2533] = {.lex_state = 151},
  [2534] = {.lex_state = 151},
  [2535] = {.lex_state = 151},
  [2536] = {.lex_state = 151},
  [2537] = {.lex_state = 151},
  [2538] = {.lex_state = 151},
  [2539] = {.lex_state = 151},
  [2540] = {.lex_state = 151},
  [2541] = {.lex_state = 151},
  [2542] = {.lex_state = 151},
  [2543] = {.lex_state = 151},
  [2544] = {.lex_state = 151},
  [2545] = {.lex_state = 151},
  [2546] = {.lex_state = 151},
  [2547] = {.lex_state = 151},
  [2548] = {.lex_state = 151},
  [2549] = {.lex_state = 151},
  [2550] = {.lex_state = 47},
  [2551] = {.lex_state = 47},
  [2552] = {.lex_state = 47},
  [2553] = {.lex_state = 46},
  [2554] = {.lex_state = 47},
  [2555] = {.lex_state = 46},
  [2556] = {.lex_state = 151},
  [2557] = {.lex_state = 151},
  [2558] = {.lex_state = 151},
  [2559] = {.lex_state = 151},
  [2560] = {.lex_state = 151},
  [2561] = {.lex_state = 151},
  [2562] = {.lex_state = 151},
  [2563] = {.lex_state = 46},
  [2564] = {.lex_state = 151, .external_lex_state = 7},
  [2565] = {.lex_state = 151, .external_lex_state = 7},
  [2566] = {.lex_state = 151},
  [2567] = {.lex_state = 151},
  [2568] = {.lex_state = 151},
  [2569] = {.lex_state = 151},
  [2570] = {.lex_state = 151},
  [2571] = {.lex_state = 46},
  [2572] = {.lex_state = 151},
  [2573] = {.lex_state = 151},
  [2574] = {.lex_state = 151},
  [2575] = {.lex_state = 46},
  [2576] = {.lex_state = 46},
  [2577] = {.lex_state = 151},
  [2578] = {.lex_state = 151},
  [2579] = {.lex_state = 151},
  [2580] = {.lex_state = 151, .external_lex_state = 7},
  [2581] = {.lex_state = 151},
  [2582] = {.lex_state = 151, .external_lex_state = 10},
  [2583] = {.lex_state = 151},
  [2584] = {.lex_state = 151, .external_lex_state = 7},
  [2585] = {.lex_state = 151, .external_lex_state = 10},
  [2586] = {.lex_state = 151},
  [2587] = {.lex_state = 151},
  [2588] = {.lex_state = 151},
  [2589] = {.lex_state = 151},
  [2590] = {.lex_state = 151},
  [2591] = {.lex_state = 151, .external_lex_state = 7},
  [2592] = {.lex_state = 151, .external_lex_state = 10},
  [2593] = {.lex_state = 151},
  [2594] = {.lex_state = 151},
  [2595] = {.lex_state = 151, .external_lex_state = 10},
  [2596] = {.lex_state = 151, .external_lex_state = 10},
  [2597] = {.lex_state = 151, .external_lex_state = 10},
  [2598] = {.lex_state = 151, .external_lex_state = 10},
  [2599] = {.lex_state = 151, .external_lex_state = 10},
  [2600] = {.lex_state = 151, .external_lex_state = 10},
  [2601] = {.lex_state = 151, .external_lex_state = 7},
  [2602] = {.lex_state = 151, .external_lex_state = 7},
  [2603] = {.lex_state = 151, .external_lex_state = 7},
  [2604] = {.lex_state = 151, .external_lex_state = 7},
  [2605] = {.lex_state = 151, .external_lex_state = 7},
  [2606] = {.lex_state = 151, .external_lex_state = 7},
  [2607] = {.lex_state = 151, .external_lex_state = 7},
  [2608] = {.lex_state = 151},
  [2609] = {.lex_state = 151},
  [2610] = {.lex_state = 47, .external_lex_state = 7},
  [2611] = {.lex_state = 151},
  [2612] = {.lex_state = 151, .external_lex_state = 7},
  [2613] = {.lex_state = 151, .external_lex_state = 7},
  [2614] = {.lex_state = 151, .external_lex_state = 7},
  [2615] = {.lex_state = 151},
  [2616] = {.lex_state = 47, .external_lex_state = 7},
  [2617] = {.lex_state = 46},
  [2618] = {.lex_state = 151},
  [2619] = {.lex_state = 151, .external_lex_state = 7},
  [2620] = {.lex_state = 47, .external_lex_state = 7},
  [2621] = {.lex_state = 151},
  [2622] = {.lex_state = 151},
  [2623] = {.lex_state = 151, .external_lex_state = 7},
  [2624] = {.lex_state = 151, .external_lex_state = 7},
  [2625] = {.lex_state = 47},
  [2626] = {.lex_state = 151, .external_lex_state = 7},
  [2627] = {.lex_state = 151, .external_lex_state = 7},
  [2628] = {.lex_state = 151},
  [2629] = {.lex_state = 151, .external_lex_state = 7},
  [2630] = {.lex_state = 151},
  [2631] = {.lex_state = 151, .external_lex_state = 7},
  [2632] = {.lex_state = 151, .external_lex_state = 7},
  [2633] = {.lex_state = 151, .external_lex_state = 7},
  [2634] = {.lex_state = 151, .external_lex_state = 7},
  [2635] = {.lex_state = 151, .external_lex_state = 7},
  [2636] = {.lex_state = 151},
  [2637] = {.lex_state = 47},
  [2638] = {.lex_state = 46},
  [2639] = {.lex_state = 151},
  [2640] = {.lex_state = 151, .external_lex_state = 7},
  [2641] = {.lex_state = 151},
  [2642] = {.lex_state = 151, .external_lex_state = 7},
  [2643] = {.lex_state = 151, .external_lex_state = 7},
  [2644] = {.lex_state = 46},
  [2645] = {.lex_state = 151},
  [2646] = {.lex_state = 151},
  [2647] = {.lex_state = 151},
  [2648] = {.lex_state = 151, .external_lex_state = 7},
  [2649] = {.lex_state = 151},
  [2650] = {.lex_state = 151},
  [2651] = {.lex_state = 47},
  [2652] = {.lex_state = 151},
  [2653] = {.lex_state = 151},
  [2654] = {.lex_state = 151},
  [2655] = {.lex_state = 151},
  [2656] = {.lex_state = 151, .external_lex_state = 7},
  [2657] = {.lex_state = 151},
  [2658] = {.lex_state = 151, .external_lex_state = 7},
  [2659] = {.lex_state = 151},
  [2660] = {.lex_state = 151},
  [2661] = {.lex_state = 151},
  [2662] = {.lex_state = 151},
  [2663] = {.lex_state = 151, .external_lex_state = 7},
  [2664] = {.lex_state = 151},
  [2665] = {.lex_state = 151},
  [2666] = {.lex_state = 151, .external_lex_state = 7},
  [2667] = {.lex_state = 47},
  [2668] = {.lex_state = 151, .external_lex_state = 7},
  [2669] = {.lex_state = 151},
  [2670] = {.lex_state = 151},
  [2671] = {.lex_state = 47},
  [2672] = {.lex_state = 151},
  [2673] = {.lex_state = 151},
  [2674] = {.lex_state = 151},
  [2675] = {.lex_state = 151},
  [2676] = {.lex_state = 151},
  [2677] = {.lex_state = 151},
  [2678] = {.lex_state = 151},
  [2679] = {.lex_state = 47},
  [2680] = {.lex_state = 151},
  [2681] = {.lex_state = 151},
  [2682] = {.lex_state = 47},
  [2683] = {.lex_state = 151},
  [2684] = {.lex_state = 151, .external_lex_state = 10},
  [2685] = {.lex_state = 151},
  [2686] = {.lex_state = 151},
  [2687] = {.lex_state = 151},
  [2688] = {.lex_state = 151, .external_lex_state = 7},
  [2689] = {.lex_state = 151, .external_lex_state = 7},
  [2690] = {.lex_state = 151},
  [2691] = {.lex_state = 151},
  [2692] = {.lex_state = 151, .external_lex_state = 7},
  [2693] = {.lex_state = 151},
  [2694] = {.lex_state = 151},
  [2695] = {.lex_state = 151},
  [2696] = {.lex_state = 151},
  [2697] = {.lex_state = 151, .external_lex_state = 7},
  [2698] = {.lex_state = 151},
  [2699] = {.lex_state = 151, .external_lex_state = 7},
  [2700] = {.lex_state = 47},
  [2701] = {.lex_state = 151, .external_lex_state = 7},
  [2702] = {.lex_state = 151},
  [2703] = {.lex_state = 151},
  [2704] = {.lex_state = 151},
  [2705] = {.lex_state = 151},
  [2706] = {.lex_state = 151},
  [2707] = {.lex_state = 151, .external_lex_state = 7},
  [2708] = {.lex_state = 151},
  [2709] = {.lex_state = 151},
  [2710] = {.lex_state = 47, .external_lex_state = 7},
  [2711] = {.lex_state = 151},
  [2712] = {.lex_state = 151, .external_lex_state = 7},
  [2713] = {.lex_state = 47},
  [2714] = {.lex_state = 151},
  [2715] = {.lex_state = 47},
  [2716] = {.lex_state = 47},
  [2717] = {.lex_state = 47},
  [2718] = {.lex_state = 47},
  [2719] = {.lex_state = 47},
  [2720] = {.lex_state = 151, .external_lex_state = 7},
  [2721] = {.lex_state = 151},
  [2722] = {.lex_state = 151, .external_lex_state = 7},
  [2723] = {.lex_state = 151},
  [2724] = {.lex_state = 151, .external_lex_state = 7},
  [2725] = {.lex_state = 47},
  [2726] = {.lex_state = 47},
  [2727] = {.lex_state = 151},
  [2728] = {.lex_state = 151},
  [2729] = {.lex_state = 151},
  [2730] = {.lex_state = 151},
  [2731] = {.lex_state = 46},
  [2732] = {.lex_state = 47},
  [2733] = {.lex_state = 47},
  [2734] = {.lex_state = 47},
  [2735] = {.lex_state = 151},
  [2736] = {.lex_state = 47},
  [2737] = {.lex_state = 151},
  [2738] = {.lex_state = 151},
  [2739] = {.lex_state = 151},
  [2740] = {.lex_state = 47},
  [2741] = {.lex_state = 47},
  [2742] = {.lex_state = 151},
  [2743] = {.lex_state = 47},
  [2744] = {.lex_state = 47},
  [2745] = {.lex_state = 47},
  [2746] = {.lex_state = 151, .external_lex_state = 7},
  [2747] = {.lex_state = 47},
  [2748] = {.lex_state = 47},
  [2749] = {.lex_state = 151},
  [2750] = {.lex_state = 151, .external_lex_state = 7},
  [2751] = {.lex_state = 47},
  [2752] = {.lex_state = 47},
  [2753] = {.lex_state = 47},
  [2754] = {.lex_state = 47},
  [2755] = {.lex_state = 47},
  [2756] = {.lex_state = 151, .external_lex_state = 10},
  [2757] = {.lex_state = 47},
  [2758] = {.lex_state = 47},
  [2759] = {.lex_state = 151, .external_lex_state = 7},
  [2760] = {.lex_state = 47},
  [2761] = {.lex_state = 47},
  [2762] = {.lex_state = 47},
  [2763] = {.lex_state = 47},
  [2764] = {.lex_state = 151},
  [2765] = {.lex_state = 151},
  [2766] = {.lex_state = 47},
  [2767] = {.lex_state = 47},
  [2768] = {.lex_state = 151, .external_lex_state = 10},
  [2769] = {.lex_state = 47},
  [2770] = {.lex_state = 47},
  [2771] = {.lex_state = 47},
  [2772] = {.lex_state = 47},
  [2773] = {.lex_state = 151},
  [2774] = {.lex_state = 151, .external_lex_state = 7},
  [2775] = {.lex_state = 47},
  [2776] = {.lex_state = 47},
  [2777] = {.lex_state = 151, .external_lex_state = 10},
  [2778] = {.lex_state = 47},
  [2779] = {.lex_state = 47},
  [2780] = {.lex_state = 47},
  [2781] = {.lex_state = 47},
  [2782] = {.lex_state = 47},
  [2783] = {.lex_state = 47},
  [2784] = {.lex_state = 47},
  [2785] = {.lex_state = 47},
  [2786] = {.lex_state = 151},
  [2787] = {.lex_state = 47},
  [2788] = {.lex_state = 47},
  [2789] = {.lex_state = 47},
  [2790] = {.lex_state = 47},
  [2791] = {.lex_state = 47},
  [2792] = {.lex_state = 47, .external_lex_state = 7},
  [2793] = {.lex_state = 151, .external_lex_state = 7},
  [2794] = {.lex_state = 151},
  [2795] = {.lex_state = 46},
  [2796] = {.lex_state = 151},
  [2797] = {.lex_state = 46},
  [2798] = {.lex_state = 46},
  [2799] = {.lex_state = 151},
  [2800] = {.lex_state = 46},
  [2801] = {.lex_state = 151},
  [2802] = {.lex_state = 46},
  [2803] = {.lex_state = 151},
  [2804] = {.lex_state = 46},
  [2805] = {.lex_state = 151},
  [2806] = {.lex_state = 46},
  [2807] = {.lex_state = 151},
  [2808] = {.lex_state = 46},
  [2809] = {.lex_state = 46},
  [2810] = {.lex_state = 46},
  [2811] = {.lex_state = 46},
  [2812] = {.lex_state = 151},
  [2813] = {.lex_state = 46},
  [2814] = {.lex_state = 151},
  [2815] = {.lex_state = 46},
  [2816] = {.lex_state = 46},
  [2817] = {.lex_state = 46},
  [2818] = {.lex_state = 151},
  [2819] = {.lex_state = 46},
  [2820] = {.lex_state = 46},
  [2821] = {.lex_state = 46},
  [2822] = {.lex_state = 46},
  [2823] = {.lex_state = 46},
  [2824] = {.lex_state = 151},
  [2825] = {.lex_state = 46},
  [2826] = {.lex_state = 151},
  [2827] = {.lex_state = 46},
  [2828] = {.lex_state = 46},
  [2829] = {.lex_state = 151, .external_lex_state = 7},
  [2830] = {.lex_state = 151},
  [2831] = {.lex_state = 46},
  [2832] = {.lex_state = 151},
  [2833] = {.lex_state = 46},
  [2834] = {.lex_state = 46},
  [2835] = {.lex_state = 47},
  [2836] = {.lex_state = 46},
  [2837] = {.lex_state = 151, .external_lex_state = 7},
  [2838] = {.lex_state = 151},
  [2839] = {.lex_state = 46},
  [2840] = {.lex_state = 151},
  [2841] = {.lex_state = 46},
  [2842] = {.lex_state = 151, .external_lex_state = 7},
  [2843] = {.lex_state = 151},
  [2844] = {.lex_state = 46},
  [2845] = {.lex_state = 151},
  [2846] = {.lex_state = 46},
  [2847] = {.lex_state = 151},
  [2848] = {.lex_state = 151, .external_lex_state = 7},
  [2849] = {.lex_state = 151},
  [2850] = {.lex_state = 151},
  [2851] = {.lex_state = 151},
  [2852] = {.lex_state = 151},
  [2853] = {.lex_state = 151},
  [2854] = {.lex_state = 46},
  [2855] = {.lex_state = 46},
  [2856] = {.lex_state = 46},
  [2857] = {.lex_state = 46},
  [2858] = {.lex_state = 46},
  [2859] = {.lex_state = 151},
  [2860] = {.lex_state = 46},
  [2861] = {.lex_state = 46},
  [2862] = {.lex_state = 151},
  [2863] = {.lex_state = 151},
  [2864] = {.lex_state = 151},
  [2865] = {.lex_state = 46},
  [2866] = {.lex_state = 46},
  [2867] = {.lex_state = 151},
  [2868] = {.lex_state = 151},
  [2869] = {.lex_state = 151},
  [2870] = {.lex_state = 46},
  [2871] = {.lex_state = 151},
  [2872] = {.lex_state = 151},
  [2873] = {.lex_state = 46},
  [2874] = {.lex_state = 46},
  [2875] = {.lex_state = 47},
  [2876] = {.lex_state = 46},
  [2877] = {.lex_state = 47},
  [2878] = {.lex_state = 46},
  [2879] = {.lex_state = 151},
  [2880] = {.lex_state = 47},
  [2881] = {.lex_state = 46},
  [2882] = {.lex_state = 46},
  [2883] = {.lex_state = 46},
  [2884] = {.lex_state = 151, .external_lex_state = 7},
  [2885] = {.lex_state = 46},
  [2886] = {.lex_state = 46},
  [2887] = {.lex_state = 46},
  [2888] = {.lex_state = 151},
  [2889] = {.lex_state = 151},
  [2890] = {.lex_state = 151},
  [2891] = {.lex_state = 46},
  [2892] = {.lex_state = 46},
  [2893] = {.lex_state = 151},
  [2894] = {.lex_state = 46},
  [2895] = {.lex_state = 151},
  [2896] = {.lex_state = 46},
  [2897] = {.lex_state = 46},
  [2898] = {.lex_state = 151, .external_lex_state = 7},
  [2899] = {.lex_state = 46},
  [2900] = {.lex_state = 46},
  [2901] = {.lex_state = 46},
  [2902] = {.lex_state = 46},
  [2903] = {.lex_state = 151},
  [2904] = {.lex_state = 46},
  [2905] = {.lex_state = 151},
  [2906] = {.lex_state = 46},
  [2907] = {.lex_state = 151},
  [2908] = {.lex_state = 46},
  [2909] = {.lex_state = 151},
  [2910] = {.lex_state = 46},
  [2911] = {.lex_state = 46},
  [2912] = {.lex_state = 151},
  [2913] = {.lex_state = 46},
  [2914] = {.lex_state = 46},
  [2915] = {.lex_state = 46},
  [2916] = {.lex_state = 151},
  [2917] = {.lex_state = 46},
  [2918] = {.lex_state = 46},
  [2919] = {.lex_state = 46},
  [2920] = {.lex_state = 46},
  [2921] = {.lex_state = 151},
  [2922] = {.lex_state = 46},
  [2923] = {.lex_state = 151},
  [2924] = {.lex_state = 151},
  [2925] = {.lex_state = 46},
  [2926] = {.lex_state = 151},
  [2927] = {.lex_state = 46},
  [2928] = {.lex_state = 46},
  [2929] = {.lex_state = 151},
  [2930] = {.lex_state = 46},
  [2931] = {.lex_state = 151},
  [2932] = {.lex_state = 151},
  [2933] = {.lex_state = 46},
  [2934] = {.lex_state = 151},
  [2935] = {.lex_state = 46},
  [2936] = {.lex_state = 46},
  [2937] = {.lex_state = 151},
  [2938] = {.lex_state = 151},
  [2939] = {.lex_state = 151},
  [2940] = {.lex_state = 46},
  [2941] = {.lex_state = 46},
  [2942] = {.lex_state = 46},
  [2943] = {.lex_state = 151},
  [2944] = {.lex_state = 46},
  [2945] = {.lex_state = 151},
  [2946] = {.lex_state = 46},
  [2947] = {.lex_state = 46},
  [2948] = {.lex_state = 46},
  [2949] = {.lex_state = 46},
  [2950] = {.lex_state = 46},
  [2951] = {.lex_state = 46},
  [2952] = {.lex_state = 151},
  [2953] = {.lex_state = 46},
  [2954] = {.lex_state = 151},
  [2955] = {.lex_state = 46},
  [2956] = {.lex_state = 151},
  [2957] = {.lex_state = 47},
  [2958] = {.lex_state = 46},
  [2959] = {.lex_state = 46},
  [2960] = {.lex_state = 151},
  [2961] = {.lex_state = 46},
  [2962] = {.lex_state = 151},
  [2963] = {.lex_state = 46},
  [2964] = {.lex_state = 46},
  [2965] = {.lex_state = 151},
  [2966] = {.lex_state = 46},
  [2967] = {.lex_state = 151},
  [2968] = {.lex_state = 151},
  [2969] = {.lex_state = 46},
  [2970] = {.lex_state = 151},
  [2971] = {.lex_state = 47},
  [2972] = {.lex_state = 46},
  [2973] = {.lex_state = 47},
  [2974] = {.lex_state = 151},
  [2975] = {.lex_state = 151},
  [2976] = {.lex_state = 47},
  [2977] = {.lex_state = 47},
  [2978] = {.lex_state = 151},
  [2979] = {.lex_state = 47},
  [2980] = {.lex_state = 151},
  [2981] = {.lex_state = 47},
  [2982] = {.lex_state = 151},
  [2983] = {.lex_state = 47},
  [2984] = {.lex_state = 47},
  [2985] = {.lex_state = 46},
  [2986] = {.lex_state = 151},
  [2987] = {.lex_state = 151},
  [2988] = {.lex_state = 47},
  [2989] = {.lex_state = 151},
  [2990] = {.lex_state = 46},
  [2991] = {.lex_state = 151},
  [2992] = {.lex_state = 151},
  [2993] = {.lex_state = 151},
  [2994] = {.lex_state = 151},
  [2995] = {.lex_state = 151},
  [2996] = {.lex_state = 46},
  [2997] = {.lex_state = 151},
  [2998] = {.lex_state = 46},
  [2999] = {.lex_state = 151},
  [3000] = {.lex_state = 151},
  [3001] = {.lex_state = 151},
  [3002] = {.lex_state = 151},
  [3003] = {.lex_state = 47},
  [3004] = {.lex_state = 151},
  [3005] = {.lex_state = 151},
  [3006] = {.lex_state = 151},
  [3007] = {.lex_state = 47},
  [3008] = {.lex_state = 151},
  [3009] = {.lex_state = 47},
  [3010] = {.lex_state = 151},
  [3011] = {.lex_state = 46},
  [3012] = {.lex_state = 47},
  [3013] = {.lex_state = 151},
  [3014] = {.lex_state = 151},
  [3015] = {.lex_state = 151},
  [3016] = {.lex_state = 151},
  [3017] = {.lex_state = 151},
  [3018] = {.lex_state = 47},
  [3019] = {.lex_state = 47},
  [3020] = {.lex_state = 47},
  [3021] = {.lex_state = 151},
  [3022] = {.lex_state = 151},
  [3023] = {.lex_state = 151},
  [3024] = {.lex_state = 151},
  [3025] = {.lex_state = 151},
  [3026] = {.lex_state = 151},
  [3027] = {.lex_state = 47},
  [3028] = {.lex_state = 151},
  [3029] = {.lex_state = 151},
  [3030] = {.lex_state = 47},
  [3031] = {.lex_state = 47},
  [3032] = {.lex_state = 151},
  [3033] = {.lex_state = 47},
  [3034] = {.lex_state = 151},
  [3035] = {.lex_state = 46},
  [3036] = {.lex_state = 47},
  [3037] = {.lex_state = 151},
  [3038] = {.lex_state = 47},
  [3039] = {.lex_state = 151},
  [3040] = {.lex_state = 151},
  [3041] = {.lex_state = 46},
  [3042] = {.lex_state = 151},
  [3043] = {.lex_state = 151},
  [3044] = {.lex_state = 47},
  [3045] = {.lex_state = 47},
  [3046] = {.lex_state = 151},
  [3047] = {.lex_state = 151},
  [3048] = {.lex_state = 151},
  [3049] = {.lex_state = 151},
  [3050] = {.lex_state = 151},
  [3051] = {.lex_state = 151, .external_lex_state = 11},
  [3052] = {.lex_state = 151},
  [3053] = {.lex_state = 46},
  [3054] = {.lex_state = 151},
  [3055] = {.lex_state = 151},
  [3056] = {.lex_state = 151},
  [3057] = {.lex_state = 46},
  [3058] = {.lex_state = 46},
  [3059] = {.lex_state = 151},
  [3060] = {.lex_state = 151},
  [3061] = {.lex_state = 151},
  [3062] = {.lex_state = 46},
  [3063] = {.lex_state = 151},
  [3064] = {.lex_state = 47},
  [3065] = {.lex_state = 47},
  [3066] = {.lex_state = 151},
  [3067] = {.lex_state = 151},
  [3068] = {.lex_state = 151, .external_lex_state = 11},
  [3069] = {.lex_state = 151},
  [3070] = {.lex_state = 46},
  [3071] = {.lex_state = 46},
  [3072] = {.lex_state = 46},
  [3073] = {.lex_state = 151},
  [3074] = {.lex_state = 151},
  [3075] = {.lex_state = 151},
  [3076] = {.lex_state = 151},
  [3077] = {.lex_state = 47},
  [3078] = {.lex_state = 151},
  [3079] = {.lex_state = 46},
  [3080] = {.lex_state = 151},
  [3081] = {.lex_state = 151},
  [3082] = {.lex_state = 47},
  [3083] = {.lex_state = 151},
  [3084] = {.lex_state = 46},
  [3085] = {.lex_state = 151},
  [3086] = {.lex_state = 47},
  [3087] = {.lex_state = 151},
  [3088] = {.lex_state = 151},
  [3089] = {.lex_state = 151},
  [3090] = {.lex_state = 151},
  [3091] = {.lex_state = 151},
  [3092] = {.lex_state = 46},
  [3093] = {.lex_state = 151},
  [3094] = {.lex_state = 151},
  [3095] = {.lex_state = 151},
  [3096] = {.lex_state = 151},
  [3097] = {.lex_state = 151},
  [3098] = {.lex_state = 46},
  [3099] = {.lex_state = 46},
  [3100] = {.lex_state = 47},
  [3101] = {.lex_state = 151},
  [3102] = {.lex_state = 151},
  [3103] = {.lex_state = 151},
  [3104] = {.lex_state = 46},
  [3105] = {.lex_state = 151},
  [3106] = {.lex_state = 47},
  [3107] = {.lex_state = 47},
  [3108] = {.lex_state = 47},
  [3109] = {.lex_state = 46},
  [3110] = {.lex_state = 151},
  [3111] = {.lex_state = 151},
  [3112] = {.lex_state = 46},
  [3113] = {.lex_state = 151},
  [3114] = {.lex_state = 151},
  [3115] = {.lex_state = 151},
  [3116] = {.lex_state = 151},
  [3117] = {.lex_state = 151},
  [3118] = {.lex_state = 151},
  [3119] = {.lex_state = 151},
  [3120] = {.lex_state = 151},
  [3121] = {.lex_state = 151},
  [3122] = {.lex_state = 151},
  [3123] = {.lex_state = 151},
  [3124] = {.lex_state = 151},
  [3125] = {.lex_state = 151},
  [3126] = {.lex_state = 151},
  [3127] = {.lex_state = 151},
  [3128] = {.lex_state = 151},
  [3129] = {.lex_state = 151},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_return] = ACTIONS(1),
    [anon_sym_async] = ACTIONS(1),
    [anon_sym_def] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_del] = ACTIONS(1),
    [anon_sym_raise] = ACTIONS(1),
    [sym_pass_statement] = ACTIONS(1),
    [sym_break_statement] = ACTIONS(1),
    [sym_continue_statement] = ACTIONS(1),
    [sym_debugger_statement] = ACTIONS(1),
    [anon_sym_assert] = ACTIONS(1),
    [anon_sym_global] = ACTIONS(1),
    [anon_sym_nonlocal] = ACTIONS(1),
    [anon_sym_yield] = ACTIONS(1),
    [anon_sym_from] = ACTIONS(1),
    [anon_sym_import] = ACTIONS(1),
    [anon_sym_as] = ACTIONS(1),
    [anon_sym___python__] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_elif] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_try] = ACTIONS(1),
    [anon_sym_except] = ACTIONS(1),
    [anon_sym_finally] = ACTIONS(1),
    [anon_sym_with] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_STAR_STAR] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_not] = ACTIONS(1),
    [anon_sym_and] = ACTIONS(1),
    [anon_sym_or] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_SLASH_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_GT_GT_GT] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_typeof] = ACTIONS(1),
    [anon_sym_void] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_is] = ACTIONS(1),
    [anon_sym_instanceof] = ACTIONS(1),
    [anon_sym_new] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_await] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [aux_sym_interpolation_token1] = ACTIONS(1),
    [sym_this] = ACTIONS(1),
    [sym_true] = ACTIONS(1),
    [sym_false] = ACTIONS(1),
    [sym_none] = ACTIONS(1),
    [sym_string] = ACTIONS(1),
    [sym_verbatim] = ACTIONS(1),
    [sym_number] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1),
    [sym__indent] = ACTIONS(1),
    [sym__dedent] = ACTIONS(1),
    [sym_regex] = ACTIONS(1),
    [sym_fstring_start] = ACTIONS(1),
    [sym_fstring_content] = ACTIONS(1),
    [sym_fstring_end] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_module] = STATE(3095),
    [sym__statement] = STATE(12),
    [sym__simple_statements] = STATE(12),
    [sym__simple_statement] = STATE(2666),
    [sym__compound_statement] = STATE(12),
    [sym_block_tailed_statement] = STATE(1600),
    [sym__block_anonymous_function] = STATE(1611),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2549),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2549),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1600),
    [sym_for_statement] = STATE(1600),
    [sym_for_js_statement] = STATE(1600),
    [sym_while_statement] = STATE(1600),
    [sym_do_statement] = STATE(1600),
    [sym_try_statement] = STATE(1600),
    [sym_with_statement] = STATE(1600),
    [sym_function_definition] = STATE(1600),
    [sym_async_function_definition] = STATE(1600),
    [sym_class_definition] = STATE(1600),
    [sym_decorated_definition] = STATE(1600),
    [sym_decorator] = STATE(2553),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(12),
    [aux_sym_decorated_definition_repeat1] = STATE(2553),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(9),
    [anon_sym_async] = ACTIONS(11),
    [anon_sym_def] = ACTIONS(13),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_with] = ACTIONS(47),
    [anon_sym_class] = ACTIONS(49),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(2)] = {
    [sym__expression_statement] = STATE(656),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(656),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(656),
    [sym_delete_statement] = STATE(656),
    [sym_raise_statement] = STATE(656),
    [sym_assert_statement] = STATE(656),
    [sym_yield_statement] = STATE(656),
    [sym_yield] = STATE(651),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__inline_simple_statement] = STATE(656),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(89),
    [sym_break_statement] = ACTIONS(89),
    [sym_continue_statement] = ACTIONS(89),
    [sym_debugger_statement] = ACTIONS(73),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_global] = ACTIONS(73),
    [anon_sym_nonlocal] = ACTIONS(73),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_from] = ACTIONS(73),
    [anon_sym_import] = ACTIONS(73),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_while] = ACTIONS(73),
    [anon_sym_do] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(73),
    [anon_sym_with] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_class] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(109),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(3)] = {
    [sym__expression_statement] = STATE(618),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(618),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(618),
    [sym_delete_statement] = STATE(618),
    [sym_raise_statement] = STATE(618),
    [sym_assert_statement] = STATE(618),
    [sym_yield_statement] = STATE(618),
    [sym_yield] = STATE(668),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__inline_simple_statement] = STATE(618),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(115),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(135),
    [sym_break_statement] = ACTIONS(135),
    [sym_continue_statement] = ACTIONS(135),
    [sym_debugger_statement] = ACTIONS(119),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_global] = ACTIONS(119),
    [anon_sym_nonlocal] = ACTIONS(119),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_from] = ACTIONS(119),
    [anon_sym_import] = ACTIONS(119),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_while] = ACTIONS(119),
    [anon_sym_do] = ACTIONS(119),
    [anon_sym_try] = ACTIONS(119),
    [anon_sym_with] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_class] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(4)] = {
    [sym__expression_statement] = STATE(618),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(618),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(618),
    [sym_delete_statement] = STATE(618),
    [sym_raise_statement] = STATE(618),
    [sym_assert_statement] = STATE(618),
    [sym_yield_statement] = STATE(618),
    [sym_yield] = STATE(668),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__inline_simple_statement] = STATE(618),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(109),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(135),
    [sym_break_statement] = ACTIONS(135),
    [sym_continue_statement] = ACTIONS(135),
    [sym_debugger_statement] = ACTIONS(73),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_global] = ACTIONS(73),
    [anon_sym_nonlocal] = ACTIONS(73),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_from] = ACTIONS(73),
    [anon_sym_import] = ACTIONS(73),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_while] = ACTIONS(73),
    [anon_sym_do] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(73),
    [anon_sym_with] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_class] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(5)] = {
    [sym__expression_statement] = STATE(656),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(656),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(656),
    [sym_delete_statement] = STATE(656),
    [sym_raise_statement] = STATE(656),
    [sym_assert_statement] = STATE(656),
    [sym_yield_statement] = STATE(656),
    [sym_yield] = STATE(651),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__inline_simple_statement] = STATE(656),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(89),
    [sym_break_statement] = ACTIONS(89),
    [sym_continue_statement] = ACTIONS(89),
    [sym_debugger_statement] = ACTIONS(119),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_global] = ACTIONS(119),
    [anon_sym_nonlocal] = ACTIONS(119),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_from] = ACTIONS(119),
    [anon_sym_import] = ACTIONS(119),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_while] = ACTIONS(119),
    [anon_sym_do] = ACTIONS(119),
    [anon_sym_try] = ACTIONS(119),
    [anon_sym_with] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_class] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(115),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(6)] = {
    [sym__expression_statement] = STATE(618),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(618),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(618),
    [sym_delete_statement] = STATE(618),
    [sym_raise_statement] = STATE(618),
    [sym_assert_statement] = STATE(618),
    [sym_yield_statement] = STATE(618),
    [sym_yield] = STATE(668),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__inline_simple_statement] = STATE(618),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(115),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(135),
    [sym_break_statement] = ACTIONS(135),
    [sym_continue_statement] = ACTIONS(135),
    [sym_debugger_statement] = ACTIONS(119),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_global] = ACTIONS(119),
    [anon_sym_nonlocal] = ACTIONS(119),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_from] = ACTIONS(119),
    [anon_sym_import] = ACTIONS(119),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_while] = ACTIONS(119),
    [anon_sym_do] = ACTIONS(119),
    [anon_sym_try] = ACTIONS(119),
    [anon_sym_with] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_class] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(7)] = {
    [sym__expression_statement] = STATE(656),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(656),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(656),
    [sym_delete_statement] = STATE(656),
    [sym_raise_statement] = STATE(656),
    [sym_assert_statement] = STATE(656),
    [sym_yield_statement] = STATE(656),
    [sym_yield] = STATE(651),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__inline_simple_statement] = STATE(656),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(89),
    [sym_break_statement] = ACTIONS(89),
    [sym_continue_statement] = ACTIONS(89),
    [sym_debugger_statement] = ACTIONS(119),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_global] = ACTIONS(119),
    [anon_sym_nonlocal] = ACTIONS(119),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_from] = ACTIONS(119),
    [anon_sym_import] = ACTIONS(119),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_while] = ACTIONS(119),
    [anon_sym_do] = ACTIONS(119),
    [anon_sym_try] = ACTIONS(119),
    [anon_sym_with] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_class] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(115),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(8)] = {
    [sym__expression_statement] = STATE(618),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(618),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(618),
    [sym_delete_statement] = STATE(618),
    [sym_raise_statement] = STATE(618),
    [sym_assert_statement] = STATE(618),
    [sym_yield_statement] = STATE(618),
    [sym_yield] = STATE(668),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__inline_simple_statement] = STATE(618),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(109),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(135),
    [sym_break_statement] = ACTIONS(135),
    [sym_continue_statement] = ACTIONS(135),
    [sym_debugger_statement] = ACTIONS(73),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_global] = ACTIONS(73),
    [anon_sym_nonlocal] = ACTIONS(73),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_from] = ACTIONS(73),
    [anon_sym_import] = ACTIONS(73),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_while] = ACTIONS(73),
    [anon_sym_do] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(73),
    [anon_sym_with] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_class] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(9)] = {
    [sym__expression_statement] = STATE(656),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(656),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(656),
    [sym_delete_statement] = STATE(656),
    [sym_raise_statement] = STATE(656),
    [sym_assert_statement] = STATE(656),
    [sym_yield_statement] = STATE(656),
    [sym_yield] = STATE(651),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__inline_simple_statement] = STATE(656),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(89),
    [sym_break_statement] = ACTIONS(89),
    [sym_continue_statement] = ACTIONS(89),
    [sym_debugger_statement] = ACTIONS(73),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_global] = ACTIONS(73),
    [anon_sym_nonlocal] = ACTIONS(73),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_from] = ACTIONS(73),
    [anon_sym_import] = ACTIONS(73),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_while] = ACTIONS(73),
    [anon_sym_do] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(73),
    [anon_sym_with] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_class] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(109),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(10)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(195),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(11)] = {
    [sym__statement] = STATE(19),
    [sym__simple_statements] = STATE(19),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(19),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(19),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(197),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(12)] = {
    [sym__statement] = STATE(14),
    [sym__simple_statements] = STATE(14),
    [sym__simple_statement] = STATE(2666),
    [sym__compound_statement] = STATE(14),
    [sym_block_tailed_statement] = STATE(1600),
    [sym__block_anonymous_function] = STATE(1611),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2549),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2549),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1600),
    [sym_for_statement] = STATE(1600),
    [sym_for_js_statement] = STATE(1600),
    [sym_while_statement] = STATE(1600),
    [sym_do_statement] = STATE(1600),
    [sym_try_statement] = STATE(1600),
    [sym_with_statement] = STATE(1600),
    [sym_function_definition] = STATE(1600),
    [sym_async_function_definition] = STATE(1600),
    [sym_class_definition] = STATE(1600),
    [sym_decorated_definition] = STATE(1600),
    [sym_decorator] = STATE(2553),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(14),
    [aux_sym_decorated_definition_repeat1] = STATE(2553),
    [ts_builtin_sym_end] = ACTIONS(199),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(9),
    [anon_sym_async] = ACTIONS(11),
    [anon_sym_def] = ACTIONS(13),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_with] = ACTIONS(47),
    [anon_sym_class] = ACTIONS(49),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(13)] = {
    [sym__statement] = STATE(10),
    [sym__simple_statements] = STATE(10),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(10),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(10),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(201),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(14)] = {
    [sym__statement] = STATE(14),
    [sym__simple_statements] = STATE(14),
    [sym__simple_statement] = STATE(2666),
    [sym__compound_statement] = STATE(14),
    [sym_block_tailed_statement] = STATE(1600),
    [sym__block_anonymous_function] = STATE(1611),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2549),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2549),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1600),
    [sym_for_statement] = STATE(1600),
    [sym_for_js_statement] = STATE(1600),
    [sym_while_statement] = STATE(1600),
    [sym_do_statement] = STATE(1600),
    [sym_try_statement] = STATE(1600),
    [sym_with_statement] = STATE(1600),
    [sym_function_definition] = STATE(1600),
    [sym_async_function_definition] = STATE(1600),
    [sym_class_definition] = STATE(1600),
    [sym_decorated_definition] = STATE(1600),
    [sym_decorator] = STATE(2553),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(14),
    [aux_sym_decorated_definition_repeat1] = STATE(2553),
    [ts_builtin_sym_end] = ACTIONS(203),
    [sym_identifier] = ACTIONS(205),
    [anon_sym_return] = ACTIONS(208),
    [anon_sym_async] = ACTIONS(211),
    [anon_sym_def] = ACTIONS(214),
    [anon_sym_LPAREN] = ACTIONS(217),
    [anon_sym_LBRACK] = ACTIONS(220),
    [anon_sym_del] = ACTIONS(223),
    [anon_sym_raise] = ACTIONS(226),
    [sym_pass_statement] = ACTIONS(229),
    [sym_break_statement] = ACTIONS(229),
    [sym_continue_statement] = ACTIONS(229),
    [sym_debugger_statement] = ACTIONS(229),
    [anon_sym_assert] = ACTIONS(232),
    [anon_sym_global] = ACTIONS(235),
    [anon_sym_nonlocal] = ACTIONS(238),
    [anon_sym_yield] = ACTIONS(241),
    [anon_sym_from] = ACTIONS(244),
    [anon_sym_import] = ACTIONS(247),
    [anon_sym_if] = ACTIONS(250),
    [anon_sym_for] = ACTIONS(253),
    [anon_sym_while] = ACTIONS(256),
    [anon_sym_do] = ACTIONS(259),
    [anon_sym_try] = ACTIONS(262),
    [anon_sym_with] = ACTIONS(265),
    [anon_sym_class] = ACTIONS(268),
    [anon_sym_AT] = ACTIONS(271),
    [anon_sym_not] = ACTIONS(274),
    [anon_sym_PLUS] = ACTIONS(277),
    [anon_sym_DASH] = ACTIONS(277),
    [anon_sym_TILDE] = ACTIONS(277),
    [anon_sym_typeof] = ACTIONS(277),
    [anon_sym_void] = ACTIONS(277),
    [anon_sym_new] = ACTIONS(280),
    [anon_sym_await] = ACTIONS(283),
    [anon_sym_LBRACE] = ACTIONS(286),
    [sym_this] = ACTIONS(205),
    [sym_true] = ACTIONS(289),
    [sym_false] = ACTIONS(289),
    [sym_none] = ACTIONS(289),
    [sym_string] = ACTIONS(292),
    [sym_verbatim] = ACTIONS(289),
    [sym_number] = ACTIONS(289),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(295),
    [sym_fstring_start] = ACTIONS(298),
  },
  [STATE(15)] = {
    [sym__statement] = STATE(16),
    [sym__simple_statements] = STATE(16),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(16),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(16),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(301),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(16)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(303),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(17)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(205),
    [anon_sym_return] = ACTIONS(305),
    [anon_sym_async] = ACTIONS(308),
    [anon_sym_def] = ACTIONS(311),
    [anon_sym_LPAREN] = ACTIONS(217),
    [anon_sym_LBRACK] = ACTIONS(220),
    [anon_sym_del] = ACTIONS(223),
    [anon_sym_raise] = ACTIONS(226),
    [sym_pass_statement] = ACTIONS(229),
    [sym_break_statement] = ACTIONS(229),
    [sym_continue_statement] = ACTIONS(229),
    [sym_debugger_statement] = ACTIONS(229),
    [anon_sym_assert] = ACTIONS(232),
    [anon_sym_global] = ACTIONS(235),
    [anon_sym_nonlocal] = ACTIONS(238),
    [anon_sym_yield] = ACTIONS(241),
    [anon_sym_from] = ACTIONS(244),
    [anon_sym_import] = ACTIONS(247),
    [anon_sym_if] = ACTIONS(314),
    [anon_sym_for] = ACTIONS(317),
    [anon_sym_while] = ACTIONS(320),
    [anon_sym_do] = ACTIONS(323),
    [anon_sym_try] = ACTIONS(326),
    [anon_sym_with] = ACTIONS(329),
    [anon_sym_class] = ACTIONS(332),
    [anon_sym_AT] = ACTIONS(271),
    [anon_sym_not] = ACTIONS(274),
    [anon_sym_PLUS] = ACTIONS(277),
    [anon_sym_DASH] = ACTIONS(277),
    [anon_sym_TILDE] = ACTIONS(277),
    [anon_sym_typeof] = ACTIONS(277),
    [anon_sym_void] = ACTIONS(277),
    [anon_sym_new] = ACTIONS(280),
    [anon_sym_await] = ACTIONS(283),
    [anon_sym_LBRACE] = ACTIONS(286),
    [sym_this] = ACTIONS(205),
    [sym_true] = ACTIONS(289),
    [sym_false] = ACTIONS(289),
    [sym_none] = ACTIONS(289),
    [sym_string] = ACTIONS(292),
    [sym_verbatim] = ACTIONS(289),
    [sym_number] = ACTIONS(289),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(203),
    [sym_regex] = ACTIONS(295),
    [sym_fstring_start] = ACTIONS(298),
  },
  [STATE(18)] = {
    [sym__statement] = STATE(40),
    [sym__simple_statements] = STATE(40),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(40),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(40),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(335),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(19)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(337),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(20)] = {
    [sym__statement] = STATE(21),
    [sym__simple_statements] = STATE(21),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(21),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(21),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(339),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(21)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(341),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(22)] = {
    [sym__statement] = STATE(23),
    [sym__simple_statements] = STATE(23),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(23),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(23),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(343),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(23)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(345),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(24)] = {
    [sym__statement] = STATE(25),
    [sym__simple_statements] = STATE(25),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(25),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(25),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(347),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(25)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(349),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(26)] = {
    [sym__statement] = STATE(27),
    [sym__simple_statements] = STATE(27),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(27),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(27),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(351),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(27)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(353),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(28)] = {
    [sym__statement] = STATE(29),
    [sym__simple_statements] = STATE(29),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(29),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(29),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(355),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(29)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(357),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(30)] = {
    [sym__statement] = STATE(31),
    [sym__simple_statements] = STATE(31),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(31),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(31),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(359),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(31)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(361),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(32)] = {
    [sym__statement] = STATE(33),
    [sym__simple_statements] = STATE(33),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(33),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(33),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(363),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(33)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(365),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(34)] = {
    [sym__statement] = STATE(35),
    [sym__simple_statements] = STATE(35),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(35),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(35),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(367),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(35)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(369),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(36)] = {
    [sym__statement] = STATE(37),
    [sym__simple_statements] = STATE(37),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(37),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(37),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(371),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(37)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(373),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(38)] = {
    [sym__statement] = STATE(39),
    [sym__simple_statements] = STATE(39),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(39),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(39),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(375),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(39)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(377),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(40)] = {
    [sym__statement] = STATE(17),
    [sym__simple_statements] = STATE(17),
    [sym__simple_statement] = STATE(2774),
    [sym__compound_statement] = STATE(17),
    [sym_block_tailed_statement] = STATE(1657),
    [sym__block_anonymous_function] = STATE(1625),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2537),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2537),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym_if_statement] = STATE(1657),
    [sym_for_statement] = STATE(1657),
    [sym_for_js_statement] = STATE(1657),
    [sym_while_statement] = STATE(1657),
    [sym_do_statement] = STATE(1657),
    [sym_try_statement] = STATE(1657),
    [sym_with_statement] = STATE(1657),
    [sym_function_definition] = STATE(1657),
    [sym_async_function_definition] = STATE(1657),
    [sym_class_definition] = STATE(1657),
    [sym_decorated_definition] = STATE(1657),
    [sym_decorator] = STATE(2555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [aux_sym_module_repeat1] = STATE(17),
    [aux_sym_decorated_definition_repeat1] = STATE(2555),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(175),
    [anon_sym_async] = ACTIONS(177),
    [anon_sym_def] = ACTIONS(179),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_if] = ACTIONS(181),
    [anon_sym_for] = ACTIONS(183),
    [anon_sym_while] = ACTIONS(185),
    [anon_sym_do] = ACTIONS(187),
    [anon_sym_try] = ACTIONS(189),
    [anon_sym_with] = ACTIONS(191),
    [anon_sym_class] = ACTIONS(193),
    [anon_sym_AT] = ACTIONS(51),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(379),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(41)] = {
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(683),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(444),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(381),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_return] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(383),
    [anon_sym_raise] = ACTIONS(383),
    [sym_pass_statement] = ACTIONS(383),
    [sym_break_statement] = ACTIONS(383),
    [sym_continue_statement] = ACTIONS(383),
    [sym_debugger_statement] = ACTIONS(383),
    [anon_sym_assert] = ACTIONS(383),
    [anon_sym_global] = ACTIONS(383),
    [anon_sym_nonlocal] = ACTIONS(383),
    [anon_sym_yield] = ACTIONS(383),
    [anon_sym_from] = ACTIONS(383),
    [anon_sym_import] = ACTIONS(383),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_for] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_while] = ACTIONS(383),
    [anon_sym_do] = ACTIONS(383),
    [anon_sym_try] = ACTIONS(383),
    [anon_sym_with] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_class] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(42)] = {
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(657),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(453),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_return] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(383),
    [anon_sym_raise] = ACTIONS(383),
    [sym_pass_statement] = ACTIONS(383),
    [sym_break_statement] = ACTIONS(383),
    [sym_continue_statement] = ACTIONS(383),
    [sym_debugger_statement] = ACTIONS(383),
    [anon_sym_assert] = ACTIONS(383),
    [anon_sym_global] = ACTIONS(383),
    [anon_sym_nonlocal] = ACTIONS(383),
    [anon_sym_yield] = ACTIONS(383),
    [anon_sym_from] = ACTIONS(383),
    [anon_sym_import] = ACTIONS(383),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_for] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_while] = ACTIONS(383),
    [anon_sym_do] = ACTIONS(383),
    [anon_sym_try] = ACTIONS(383),
    [anon_sym_with] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_class] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(381),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(43)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(73),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(73),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(73),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(44)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(119),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(119),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(119),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(45)] = {
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(567),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(445),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(425),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_return] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(427),
    [anon_sym_raise] = ACTIONS(427),
    [sym_pass_statement] = ACTIONS(427),
    [sym_break_statement] = ACTIONS(427),
    [sym_continue_statement] = ACTIONS(427),
    [sym_debugger_statement] = ACTIONS(427),
    [anon_sym_assert] = ACTIONS(427),
    [anon_sym_global] = ACTIONS(427),
    [anon_sym_nonlocal] = ACTIONS(427),
    [anon_sym_yield] = ACTIONS(427),
    [anon_sym_from] = ACTIONS(429),
    [anon_sym_import] = ACTIONS(427),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_for] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_while] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(427),
    [anon_sym_try] = ACTIONS(427),
    [anon_sym_with] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_class] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(46)] = {
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(660),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(441),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_return] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(427),
    [anon_sym_raise] = ACTIONS(427),
    [sym_pass_statement] = ACTIONS(427),
    [sym_break_statement] = ACTIONS(427),
    [sym_continue_statement] = ACTIONS(427),
    [sym_debugger_statement] = ACTIONS(427),
    [anon_sym_assert] = ACTIONS(427),
    [anon_sym_global] = ACTIONS(427),
    [anon_sym_nonlocal] = ACTIONS(427),
    [anon_sym_yield] = ACTIONS(427),
    [anon_sym_from] = ACTIONS(431),
    [anon_sym_import] = ACTIONS(427),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_for] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_while] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(427),
    [anon_sym_try] = ACTIONS(427),
    [anon_sym_with] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_class] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(425),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(47)] = {
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(478),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_argument_list] = STATE(561),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(433),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_return] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(437),
    [anon_sym_LBRACK] = ACTIONS(439),
    [anon_sym_del] = ACTIONS(435),
    [anon_sym_raise] = ACTIONS(435),
    [sym_pass_statement] = ACTIONS(435),
    [sym_break_statement] = ACTIONS(435),
    [sym_continue_statement] = ACTIONS(435),
    [sym_debugger_statement] = ACTIONS(435),
    [anon_sym_assert] = ACTIONS(435),
    [anon_sym_global] = ACTIONS(435),
    [anon_sym_nonlocal] = ACTIONS(435),
    [anon_sym_yield] = ACTIONS(435),
    [anon_sym_from] = ACTIONS(435),
    [anon_sym_import] = ACTIONS(435),
    [anon_sym_DOT] = ACTIONS(441),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_for] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_while] = ACTIONS(435),
    [anon_sym_do] = ACTIONS(435),
    [anon_sym_try] = ACTIONS(435),
    [anon_sym_with] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_class] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(48)] = {
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(493),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_argument_list] = STATE(606),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_return] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(443),
    [anon_sym_LBRACK] = ACTIONS(445),
    [anon_sym_del] = ACTIONS(435),
    [anon_sym_raise] = ACTIONS(435),
    [sym_pass_statement] = ACTIONS(435),
    [sym_break_statement] = ACTIONS(435),
    [sym_continue_statement] = ACTIONS(435),
    [sym_debugger_statement] = ACTIONS(435),
    [anon_sym_assert] = ACTIONS(435),
    [anon_sym_global] = ACTIONS(435),
    [anon_sym_nonlocal] = ACTIONS(435),
    [anon_sym_yield] = ACTIONS(435),
    [anon_sym_from] = ACTIONS(435),
    [anon_sym_import] = ACTIONS(435),
    [anon_sym_DOT] = ACTIONS(447),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_for] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_while] = ACTIONS(435),
    [anon_sym_do] = ACTIONS(435),
    [anon_sym_try] = ACTIONS(435),
    [anon_sym_with] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_class] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(433),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(49)] = {
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(506),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_return] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(449),
    [anon_sym_raise] = ACTIONS(449),
    [sym_pass_statement] = ACTIONS(449),
    [sym_break_statement] = ACTIONS(449),
    [sym_continue_statement] = ACTIONS(449),
    [sym_debugger_statement] = ACTIONS(449),
    [anon_sym_assert] = ACTIONS(449),
    [anon_sym_global] = ACTIONS(449),
    [anon_sym_nonlocal] = ACTIONS(449),
    [anon_sym_yield] = ACTIONS(449),
    [anon_sym_from] = ACTIONS(449),
    [anon_sym_import] = ACTIONS(449),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_for] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_while] = ACTIONS(449),
    [anon_sym_do] = ACTIONS(449),
    [anon_sym_try] = ACTIONS(449),
    [anon_sym_with] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_class] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(451),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(50)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(73),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(73),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(73),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(51)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(119),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(119),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(119),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(52)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(73),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(73),
    [aux_sym_interpolation_token1] = ACTIONS(73),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(53)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(119),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(119),
    [aux_sym_interpolation_token1] = ACTIONS(119),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(54)] = {
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(527),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(503),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_return] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(505),
    [anon_sym_raise] = ACTIONS(505),
    [sym_pass_statement] = ACTIONS(505),
    [sym_break_statement] = ACTIONS(505),
    [sym_continue_statement] = ACTIONS(505),
    [sym_debugger_statement] = ACTIONS(505),
    [anon_sym_assert] = ACTIONS(505),
    [anon_sym_global] = ACTIONS(505),
    [anon_sym_nonlocal] = ACTIONS(505),
    [anon_sym_yield] = ACTIONS(505),
    [anon_sym_from] = ACTIONS(505),
    [anon_sym_import] = ACTIONS(505),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_do] = ACTIONS(505),
    [anon_sym_try] = ACTIONS(505),
    [anon_sym_with] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_class] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(55)] = {
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2534),
    [sym_pattern_list] = STATE(2534),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(592),
    [sym__primary_expression] = STATE(672),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(216),
    [sym_subscript] = STATE(216),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_argument_list] = STATE(561),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(433),
    [sym_identifier] = ACTIONS(507),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_return] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(509),
    [anon_sym_def] = ACTIONS(511),
    [anon_sym_LPAREN] = ACTIONS(513),
    [anon_sym_LBRACK] = ACTIONS(515),
    [anon_sym_del] = ACTIONS(435),
    [anon_sym_raise] = ACTIONS(435),
    [sym_pass_statement] = ACTIONS(435),
    [sym_break_statement] = ACTIONS(435),
    [sym_continue_statement] = ACTIONS(435),
    [sym_debugger_statement] = ACTIONS(435),
    [anon_sym_assert] = ACTIONS(435),
    [anon_sym_global] = ACTIONS(435),
    [anon_sym_nonlocal] = ACTIONS(435),
    [anon_sym_yield] = ACTIONS(435),
    [anon_sym_from] = ACTIONS(435),
    [anon_sym_import] = ACTIONS(435),
    [anon_sym_DOT] = ACTIONS(441),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_for] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_while] = ACTIONS(435),
    [anon_sym_do] = ACTIONS(435),
    [anon_sym_try] = ACTIONS(435),
    [anon_sym_with] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_class] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(517),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(519),
    [anon_sym_DASH] = ACTIONS(519),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(519),
    [anon_sym_typeof] = ACTIONS(519),
    [anon_sym_void] = ACTIONS(519),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(521),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(523),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(507),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(56)] = {
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(527),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(525),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_return] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(527),
    [anon_sym_raise] = ACTIONS(527),
    [sym_pass_statement] = ACTIONS(527),
    [sym_break_statement] = ACTIONS(527),
    [sym_continue_statement] = ACTIONS(527),
    [sym_debugger_statement] = ACTIONS(527),
    [anon_sym_assert] = ACTIONS(527),
    [anon_sym_global] = ACTIONS(527),
    [anon_sym_nonlocal] = ACTIONS(527),
    [anon_sym_yield] = ACTIONS(527),
    [anon_sym_from] = ACTIONS(527),
    [anon_sym_import] = ACTIONS(527),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_for] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_while] = ACTIONS(527),
    [anon_sym_do] = ACTIONS(527),
    [anon_sym_try] = ACTIONS(527),
    [anon_sym_with] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_class] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(57)] = {
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(567),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(445),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(425),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_return] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(427),
    [anon_sym_raise] = ACTIONS(427),
    [sym_pass_statement] = ACTIONS(427),
    [sym_break_statement] = ACTIONS(427),
    [sym_continue_statement] = ACTIONS(427),
    [sym_debugger_statement] = ACTIONS(427),
    [anon_sym_assert] = ACTIONS(427),
    [anon_sym_global] = ACTIONS(427),
    [anon_sym_nonlocal] = ACTIONS(427),
    [anon_sym_yield] = ACTIONS(427),
    [anon_sym_from] = ACTIONS(529),
    [anon_sym_import] = ACTIONS(427),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_for] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_while] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(427),
    [anon_sym_try] = ACTIONS(427),
    [anon_sym_with] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_class] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(58)] = {
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(549),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_return] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(505),
    [anon_sym_raise] = ACTIONS(505),
    [sym_pass_statement] = ACTIONS(505),
    [sym_break_statement] = ACTIONS(505),
    [sym_continue_statement] = ACTIONS(505),
    [sym_debugger_statement] = ACTIONS(505),
    [anon_sym_assert] = ACTIONS(505),
    [anon_sym_global] = ACTIONS(505),
    [anon_sym_nonlocal] = ACTIONS(505),
    [anon_sym_yield] = ACTIONS(505),
    [anon_sym_from] = ACTIONS(505),
    [anon_sym_import] = ACTIONS(505),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_do] = ACTIONS(505),
    [anon_sym_try] = ACTIONS(505),
    [anon_sym_with] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_class] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(503),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(59)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_COLON] = ACTIONS(73),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(73),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(60)] = {
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(549),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_return] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(527),
    [anon_sym_raise] = ACTIONS(527),
    [sym_pass_statement] = ACTIONS(527),
    [sym_break_statement] = ACTIONS(527),
    [sym_continue_statement] = ACTIONS(527),
    [sym_debugger_statement] = ACTIONS(527),
    [anon_sym_assert] = ACTIONS(527),
    [anon_sym_global] = ACTIONS(527),
    [anon_sym_nonlocal] = ACTIONS(527),
    [anon_sym_yield] = ACTIONS(527),
    [anon_sym_from] = ACTIONS(527),
    [anon_sym_import] = ACTIONS(527),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_for] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_while] = ACTIONS(527),
    [anon_sym_do] = ACTIONS(527),
    [anon_sym_try] = ACTIONS(527),
    [anon_sym_with] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_class] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(525),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(61)] = {
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(501),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(451),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_return] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(449),
    [anon_sym_raise] = ACTIONS(449),
    [sym_pass_statement] = ACTIONS(449),
    [sym_break_statement] = ACTIONS(449),
    [sym_continue_statement] = ACTIONS(449),
    [sym_debugger_statement] = ACTIONS(449),
    [anon_sym_assert] = ACTIONS(449),
    [anon_sym_global] = ACTIONS(449),
    [anon_sym_nonlocal] = ACTIONS(449),
    [anon_sym_yield] = ACTIONS(449),
    [anon_sym_from] = ACTIONS(449),
    [anon_sym_import] = ACTIONS(449),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_for] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_while] = ACTIONS(449),
    [anon_sym_do] = ACTIONS(449),
    [anon_sym_try] = ACTIONS(449),
    [anon_sym_with] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_class] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(62)] = {
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(660),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(441),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_return] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(427),
    [anon_sym_raise] = ACTIONS(427),
    [sym_pass_statement] = ACTIONS(427),
    [sym_break_statement] = ACTIONS(427),
    [sym_continue_statement] = ACTIONS(427),
    [sym_debugger_statement] = ACTIONS(427),
    [anon_sym_assert] = ACTIONS(427),
    [anon_sym_global] = ACTIONS(427),
    [anon_sym_nonlocal] = ACTIONS(427),
    [anon_sym_yield] = ACTIONS(427),
    [anon_sym_from] = ACTIONS(553),
    [anon_sym_import] = ACTIONS(427),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_for] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_while] = ACTIONS(427),
    [anon_sym_do] = ACTIONS(427),
    [anon_sym_try] = ACTIONS(427),
    [anon_sym_with] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_class] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(425),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(63)] = {
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2529),
    [sym_pattern_list] = STATE(2529),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(653),
    [sym__primary_expression] = STATE(652),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(236),
    [sym_subscript] = STATE(236),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_argument_list] = STATE(606),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(555),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_return] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(557),
    [anon_sym_def] = ACTIONS(559),
    [anon_sym_LPAREN] = ACTIONS(561),
    [anon_sym_LBRACK] = ACTIONS(563),
    [anon_sym_del] = ACTIONS(435),
    [anon_sym_raise] = ACTIONS(435),
    [sym_pass_statement] = ACTIONS(435),
    [sym_break_statement] = ACTIONS(435),
    [sym_continue_statement] = ACTIONS(435),
    [sym_debugger_statement] = ACTIONS(435),
    [anon_sym_assert] = ACTIONS(435),
    [anon_sym_global] = ACTIONS(435),
    [anon_sym_nonlocal] = ACTIONS(435),
    [anon_sym_yield] = ACTIONS(435),
    [anon_sym_from] = ACTIONS(435),
    [anon_sym_import] = ACTIONS(435),
    [anon_sym_DOT] = ACTIONS(447),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_for] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_while] = ACTIONS(435),
    [anon_sym_do] = ACTIONS(435),
    [anon_sym_try] = ACTIONS(435),
    [anon_sym_with] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_class] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(565),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(567),
    [anon_sym_DASH] = ACTIONS(567),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(567),
    [anon_sym_typeof] = ACTIONS(567),
    [anon_sym_void] = ACTIONS(567),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(569),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(571),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(555),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(433),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(64)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_COLON] = ACTIONS(119),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(119),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(65)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(73),
    [aux_sym_interpolation_token1] = ACTIONS(73),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(66)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_COLON] = ACTIONS(119),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_as] = ACTIONS(119),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(67)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_EQ] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_RPAREN] = ACTIONS(119),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(68)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(119),
    [aux_sym_interpolation_token1] = ACTIONS(119),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(69)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_COLON] = ACTIONS(73),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_as] = ACTIONS(73),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(70)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_EQ] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_RPAREN] = ACTIONS(73),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(71)] = {
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2529),
    [sym_pattern_list] = STATE(2529),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(670),
    [sym__primary_expression] = STATE(652),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(236),
    [sym_subscript] = STATE(236),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(555),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_return] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(557),
    [anon_sym_def] = ACTIONS(559),
    [anon_sym_LPAREN] = ACTIONS(631),
    [anon_sym_LBRACK] = ACTIONS(633),
    [anon_sym_del] = ACTIONS(449),
    [anon_sym_raise] = ACTIONS(449),
    [sym_pass_statement] = ACTIONS(449),
    [sym_break_statement] = ACTIONS(449),
    [sym_continue_statement] = ACTIONS(449),
    [sym_debugger_statement] = ACTIONS(449),
    [anon_sym_assert] = ACTIONS(449),
    [anon_sym_global] = ACTIONS(449),
    [anon_sym_nonlocal] = ACTIONS(449),
    [anon_sym_yield] = ACTIONS(449),
    [anon_sym_from] = ACTIONS(449),
    [anon_sym_import] = ACTIONS(449),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_for] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_while] = ACTIONS(449),
    [anon_sym_do] = ACTIONS(449),
    [anon_sym_try] = ACTIONS(449),
    [anon_sym_with] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_class] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(565),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(567),
    [anon_sym_DASH] = ACTIONS(567),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(567),
    [anon_sym_typeof] = ACTIONS(567),
    [anon_sym_void] = ACTIONS(567),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(569),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(571),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(555),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(451),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(72)] = {
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2534),
    [sym_pattern_list] = STATE(2534),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(584),
    [sym__primary_expression] = STATE(672),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(216),
    [sym_subscript] = STATE(216),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [ts_builtin_sym_end] = ACTIONS(451),
    [sym_identifier] = ACTIONS(507),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_return] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(509),
    [anon_sym_def] = ACTIONS(511),
    [anon_sym_LPAREN] = ACTIONS(635),
    [anon_sym_LBRACK] = ACTIONS(637),
    [anon_sym_del] = ACTIONS(449),
    [anon_sym_raise] = ACTIONS(449),
    [sym_pass_statement] = ACTIONS(449),
    [sym_break_statement] = ACTIONS(449),
    [sym_continue_statement] = ACTIONS(449),
    [sym_debugger_statement] = ACTIONS(449),
    [anon_sym_assert] = ACTIONS(449),
    [anon_sym_global] = ACTIONS(449),
    [anon_sym_nonlocal] = ACTIONS(449),
    [anon_sym_yield] = ACTIONS(449),
    [anon_sym_from] = ACTIONS(449),
    [anon_sym_import] = ACTIONS(449),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_for] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_while] = ACTIONS(449),
    [anon_sym_do] = ACTIONS(449),
    [anon_sym_try] = ACTIONS(449),
    [anon_sym_with] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_class] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(517),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(519),
    [anon_sym_DASH] = ACTIONS(519),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(519),
    [anon_sym_typeof] = ACTIONS(519),
    [anon_sym_void] = ACTIONS(519),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(521),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(523),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(507),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(73)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(119),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(74)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(73),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(75)] = {
    [sym__expression_statement] = STATE(2243),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2243),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2243),
    [sym_delete_statement] = STATE(2243),
    [sym_raise_statement] = STATE(2243),
    [sym_assert_statement] = STATE(2243),
    [sym_yield_statement] = STATE(2243),
    [sym_yield] = STATE(2272),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__inline_simple_statement] = STATE(2243),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(657),
    [sym_break_statement] = ACTIONS(657),
    [sym_continue_statement] = ACTIONS(657),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(109),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(76)] = {
    [sym__expression_statement] = STATE(2243),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2243),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2243),
    [sym_delete_statement] = STATE(2243),
    [sym_raise_statement] = STATE(2243),
    [sym_assert_statement] = STATE(2243),
    [sym_yield_statement] = STATE(2243),
    [sym_yield] = STATE(2272),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__inline_simple_statement] = STATE(2243),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(657),
    [sym_break_statement] = ACTIONS(657),
    [sym_continue_statement] = ACTIONS(657),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(115),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(77)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(73),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(78)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(119),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(79)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(73),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(80)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_COMMA] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_else] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(81)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_else] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(82)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(119),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(119),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(83)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(73),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(84)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(119),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_for] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(85)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_else] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(86)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(119),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(87)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(73),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(88)] = {
    [sym__expression_statement] = STATE(2243),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2243),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2243),
    [sym_delete_statement] = STATE(2243),
    [sym_raise_statement] = STATE(2243),
    [sym_assert_statement] = STATE(2243),
    [sym_yield_statement] = STATE(2243),
    [sym_yield] = STATE(2272),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__inline_simple_statement] = STATE(2243),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(119),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(657),
    [sym_break_statement] = ACTIONS(657),
    [sym_continue_statement] = ACTIONS(657),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_if] = ACTIONS(119),
    [anon_sym_in] = ACTIONS(119),
    [anon_sym_STAR] = ACTIONS(119),
    [anon_sym_STAR_STAR] = ACTIONS(119),
    [anon_sym_AT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(119),
    [anon_sym_or] = ACTIONS(119),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(119),
    [anon_sym_SLASH_SLASH] = ACTIONS(119),
    [anon_sym_PERCENT] = ACTIONS(119),
    [anon_sym_PIPE] = ACTIONS(119),
    [anon_sym_CARET] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(119),
    [anon_sym_LT_LT] = ACTIONS(119),
    [anon_sym_GT_GT] = ACTIONS(119),
    [anon_sym_GT_GT_GT] = ACTIONS(119),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(119),
    [anon_sym_LT_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ] = ACTIONS(119),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ] = ACTIONS(119),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(119),
    [anon_sym_GT_EQ] = ACTIONS(119),
    [anon_sym_GT] = ACTIONS(119),
    [anon_sym_is] = ACTIONS(119),
    [anon_sym_instanceof] = ACTIONS(119),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(115),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(89)] = {
    [sym__expression_statement] = STATE(2243),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2243),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2243),
    [sym_delete_statement] = STATE(2243),
    [sym_raise_statement] = STATE(2243),
    [sym_assert_statement] = STATE(2243),
    [sym_yield_statement] = STATE(2243),
    [sym_yield] = STATE(2272),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__inline_simple_statement] = STATE(2243),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(657),
    [sym_break_statement] = ACTIONS(657),
    [sym_continue_statement] = ACTIONS(657),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(109),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(90)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(73),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_else] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(91)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1706),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(427),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(427),
    [anon_sym_from] = ACTIONS(729),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_for] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(92)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1760),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(427),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(427),
    [anon_sym_from] = ACTIONS(731),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(93)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1712),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_from] = ACTIONS(733),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(427),
    [aux_sym_interpolation_token1] = ACTIONS(427),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(94)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1722),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(435),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(735),
    [anon_sym_RPAREN] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_RBRACK] = ACTIONS(435),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_for] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(95)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1740),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_from] = ACTIONS(741),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_for] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(427),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(96)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1678),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1708),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(383),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(383),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(383),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_for] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(97)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2178),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2178),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2178),
    [sym_delete_statement] = STATE(2178),
    [sym_raise_statement] = STATE(2178),
    [sym_assert_statement] = STATE(2178),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2178),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1609),
    [sym_block] = STATE(497),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2221),
    [sym__inline_body] = STATE(2221),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(743),
    [sym_break_statement] = ACTIONS(743),
    [sym_continue_statement] = ACTIONS(743),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(745),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(98)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1845),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(435),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(735),
    [anon_sym_LBRACK] = ACTIONS(747),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(435),
    [aux_sym_interpolation_token1] = ACTIONS(435),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(99)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2178),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2178),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2178),
    [sym_delete_statement] = STATE(2178),
    [sym_raise_statement] = STATE(2178),
    [sym_assert_statement] = STATE(2178),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2178),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1636),
    [sym_block] = STATE(497),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2215),
    [sym__inline_body] = STATE(2215),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(743),
    [sym_break_statement] = ACTIONS(743),
    [sym_continue_statement] = ACTIONS(743),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(745),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(100)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1779),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_EQ] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_RPAREN] = ACTIONS(427),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_from] = ACTIONS(749),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(101)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1755),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(449),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(449),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_for] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(102)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1725),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(505),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(505),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(505),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(103)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1678),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1752),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(383),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(383),
    [aux_sym_interpolation_token1] = ACTIONS(383),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(104)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2178),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2178),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2178),
    [sym_delete_statement] = STATE(2178),
    [sym_raise_statement] = STATE(2178),
    [sym_assert_statement] = STATE(2178),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2178),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1615),
    [sym_block] = STATE(469),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2221),
    [sym__inline_body] = STATE(2221),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(743),
    [sym_break_statement] = ACTIONS(743),
    [sym_continue_statement] = ACTIONS(743),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(751),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(105)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1725),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(527),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(527),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(527),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_for] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(106)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2178),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2178),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2178),
    [sym_delete_statement] = STATE(2178),
    [sym_raise_statement] = STATE(2178),
    [sym_assert_statement] = STATE(2178),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2178),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1612),
    [sym_block] = STATE(469),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2220),
    [sym__inline_body] = STATE(2220),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(743),
    [sym_break_statement] = ACTIONS(743),
    [sym_continue_statement] = ACTIONS(743),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(751),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(107)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2178),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2178),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2178),
    [sym_delete_statement] = STATE(2178),
    [sym_raise_statement] = STATE(2178),
    [sym_assert_statement] = STATE(2178),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2178),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1605),
    [sym_block] = STATE(469),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2171),
    [sym__inline_body] = STATE(2171),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(743),
    [sym_break_statement] = ACTIONS(743),
    [sym_continue_statement] = ACTIONS(743),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(751),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(108)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2178),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2178),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2178),
    [sym_delete_statement] = STATE(2178),
    [sym_raise_statement] = STATE(2178),
    [sym_assert_statement] = STATE(2178),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2178),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1654),
    [sym_block] = STATE(469),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2215),
    [sym__inline_body] = STATE(2215),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(743),
    [sym_break_statement] = ACTIONS(743),
    [sym_continue_statement] = ACTIONS(743),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(751),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(109)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2178),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2178),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2178),
    [sym_delete_statement] = STATE(2178),
    [sym_raise_statement] = STATE(2178),
    [sym_assert_statement] = STATE(2178),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2178),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1607),
    [sym_block] = STATE(497),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2220),
    [sym__inline_body] = STATE(2220),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(743),
    [sym_break_statement] = ACTIONS(743),
    [sym_continue_statement] = ACTIONS(743),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(745),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(110)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1712),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_from] = ACTIONS(753),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(427),
    [aux_sym_interpolation_token1] = ACTIONS(427),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(111)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1825),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_from] = ACTIONS(755),
    [anon_sym_as] = ACTIONS(427),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(112)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1812),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_COLON] = ACTIONS(435),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(735),
    [anon_sym_LBRACK] = ACTIONS(747),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_for] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(435),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(113)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1678),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1739),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_COLON] = ACTIONS(383),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_for] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(383),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(114)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2178),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2178),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2178),
    [sym_delete_statement] = STATE(2178),
    [sym_raise_statement] = STATE(2178),
    [sym_assert_statement] = STATE(2178),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2178),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1629),
    [sym_block] = STATE(497),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2171),
    [sym__inline_body] = STATE(2171),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(743),
    [sym_break_statement] = ACTIONS(743),
    [sym_continue_statement] = ACTIONS(743),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(745),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(115)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1860),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(435),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(735),
    [anon_sym_RPAREN] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(737),
    [anon_sym_RBRACK] = ACTIONS(435),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(116)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1678),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1719),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(383),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(383),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(383),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(117)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1856),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_COLON] = ACTIONS(527),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_for] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(527),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(118)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2545),
    [sym_pattern_list] = STATE(2545),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1915),
    [sym__primary_expression] = STATE(1897),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(759),
    [anon_sym_def] = ACTIONS(761),
    [anon_sym_COLON] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(763),
    [anon_sym_LBRACK] = ACTIONS(765),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(767),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(769),
    [anon_sym_DASH] = ACTIONS(769),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(769),
    [anon_sym_typeof] = ACTIONS(769),
    [anon_sym_void] = ACTIONS(769),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(771),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(773),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(435),
    [aux_sym_interpolation_token1] = ACTIONS(435),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(119)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1823),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(527),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(527),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(527),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(120)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1958),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_EQ] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(775),
    [anon_sym_RPAREN] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(777),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(121)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1740),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_from] = ACTIONS(779),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_for] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(427),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(122)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1828),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(505),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(505),
    [aux_sym_interpolation_token1] = ACTIONS(505),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(123)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1828),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(527),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(527),
    [aux_sym_interpolation_token1] = ACTIONS(527),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(124)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1900),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_COLON] = ACTIONS(435),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(735),
    [anon_sym_LBRACK] = ACTIONS(747),
    [anon_sym_as] = ACTIONS(435),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(125)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1678),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1821),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_COLON] = ACTIONS(383),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_as] = ACTIONS(383),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(126)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1922),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_from] = ACTIONS(781),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_else] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(127)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1816),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_COLON] = ACTIONS(449),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_for] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(449),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(128)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1849),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_COLON] = ACTIONS(449),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(449),
    [aux_sym_interpolation_token1] = ACTIONS(449),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(129)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1856),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_COLON] = ACTIONS(505),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(505),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(130)] = {
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2241),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1943),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_COMMA] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_from] = ACTIONS(783),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(425),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(131)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1678),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1775),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_EQ] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_RPAREN] = ACTIONS(383),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(132)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1777),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(449),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(449),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(133)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1706),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(427),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_from] = ACTIONS(785),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_for] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(134)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1706),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(427),
    [anon_sym_from] = ACTIONS(787),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_for] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(135)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1760),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(427),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(427),
    [anon_sym_from] = ACTIONS(789),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(136)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1823),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_COLON] = ACTIONS(505),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(505),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_RBRACK] = ACTIONS(505),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(137)] = {
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2107),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_argument_list] = STATE(2179),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(791),
    [anon_sym_LBRACK] = ACTIONS(793),
    [anon_sym_DOT] = ACTIONS(795),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(433),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(138)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1936),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_COLON] = ACTIONS(505),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_as] = ACTIONS(505),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(139)] = {
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2242),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1882),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(381),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(140)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2545),
    [sym_pattern_list] = STATE(2545),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1930),
    [sym__primary_expression] = STATE(1897),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(759),
    [anon_sym_def] = ACTIONS(761),
    [anon_sym_COLON] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(797),
    [anon_sym_LBRACK] = ACTIONS(799),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(767),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(769),
    [anon_sym_DASH] = ACTIONS(769),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(769),
    [anon_sym_typeof] = ACTIONS(769),
    [anon_sym_void] = ACTIONS(769),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(771),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(773),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(449),
    [aux_sym_interpolation_token1] = ACTIONS(449),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(141)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1936),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_COLON] = ACTIONS(527),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_as] = ACTIONS(527),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(142)] = {
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2241),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1943),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_from] = ACTIONS(801),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(425),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(143)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2533),
    [sym_pattern_list] = STATE(2533),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1991),
    [sym__primary_expression] = STATE(2157),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1505),
    [sym_subscript] = STATE(1505),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(803),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(805),
    [anon_sym_def] = ACTIONS(807),
    [anon_sym_COLON] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(763),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_RBRACK] = ACTIONS(435),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(811),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(813),
    [anon_sym_DASH] = ACTIONS(813),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(813),
    [anon_sym_typeof] = ACTIONS(813),
    [anon_sym_void] = ACTIONS(813),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(815),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(817),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(803),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(144)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1963),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_EQ] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_RPAREN] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(145)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1944),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_EQ] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_RPAREN] = ACTIONS(505),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(146)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(1678),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1913),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(383),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_COMMA] = ACTIONS(383),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(383),
    [anon_sym_if] = ACTIONS(383),
    [anon_sym_else] = ACTIONS(383),
    [anon_sym_in] = ACTIONS(383),
    [anon_sym_STAR] = ACTIONS(383),
    [anon_sym_STAR_STAR] = ACTIONS(383),
    [anon_sym_AT] = ACTIONS(383),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(383),
    [anon_sym_or] = ACTIONS(383),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(383),
    [anon_sym_SLASH_SLASH] = ACTIONS(383),
    [anon_sym_PERCENT] = ACTIONS(383),
    [anon_sym_PIPE] = ACTIONS(383),
    [anon_sym_CARET] = ACTIONS(383),
    [anon_sym_AMP] = ACTIONS(383),
    [anon_sym_LT_LT] = ACTIONS(383),
    [anon_sym_GT_GT] = ACTIONS(383),
    [anon_sym_GT_GT_GT] = ACTIONS(383),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(383),
    [anon_sym_LT_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ] = ACTIONS(383),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ] = ACTIONS(383),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(383),
    [anon_sym_GT_EQ] = ACTIONS(383),
    [anon_sym_GT] = ACTIONS(383),
    [anon_sym_is] = ACTIONS(383),
    [anon_sym_instanceof] = ACTIONS(383),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(383),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(147)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2090),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_COMMA] = ACTIONS(435),
    [anon_sym_LPAREN] = ACTIONS(735),
    [anon_sym_LBRACK] = ACTIONS(747),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_else] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(148)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1922),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_from] = ACTIONS(819),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_else] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(149)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2546),
    [sym_pattern_list] = STATE(2546),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2146),
    [sym__primary_expression] = STATE(2050),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(821),
    [anon_sym_def] = ACTIONS(823),
    [anon_sym_LPAREN] = ACTIONS(763),
    [anon_sym_RPAREN] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(765),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_for] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(825),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(827),
    [anon_sym_DASH] = ACTIONS(827),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(827),
    [anon_sym_typeof] = ACTIONS(827),
    [anon_sym_void] = ACTIONS(827),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(829),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(831),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(150)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1697),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1760),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_SEMI] = ACTIONS(427),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(427),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_from] = ACTIONS(833),
    [anon_sym_DOT] = ACTIONS(427),
    [anon_sym_if] = ACTIONS(427),
    [anon_sym_in] = ACTIONS(427),
    [anon_sym_STAR] = ACTIONS(427),
    [anon_sym_STAR_STAR] = ACTIONS(427),
    [anon_sym_AT] = ACTIONS(427),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_and] = ACTIONS(427),
    [anon_sym_or] = ACTIONS(427),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_SLASH] = ACTIONS(427),
    [anon_sym_SLASH_SLASH] = ACTIONS(427),
    [anon_sym_PERCENT] = ACTIONS(427),
    [anon_sym_PIPE] = ACTIONS(427),
    [anon_sym_CARET] = ACTIONS(427),
    [anon_sym_AMP] = ACTIONS(427),
    [anon_sym_LT_LT] = ACTIONS(427),
    [anon_sym_GT_GT] = ACTIONS(427),
    [anon_sym_GT_GT_GT] = ACTIONS(427),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_LT] = ACTIONS(427),
    [anon_sym_LT_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ] = ACTIONS(427),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ] = ACTIONS(427),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(427),
    [anon_sym_GT_EQ] = ACTIONS(427),
    [anon_sym_GT] = ACTIONS(427),
    [anon_sym_is] = ACTIONS(427),
    [anon_sym_instanceof] = ACTIONS(427),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_QMARK] = ACTIONS(427),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(151)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2538),
    [sym_pattern_list] = STATE(2538),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2029),
    [sym__primary_expression] = STATE(2072),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(835),
    [anon_sym_def] = ACTIONS(837),
    [anon_sym_LPAREN] = ACTIONS(763),
    [anon_sym_LBRACK] = ACTIONS(765),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_for] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(839),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(841),
    [anon_sym_DASH] = ACTIONS(841),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(841),
    [anon_sym_typeof] = ACTIONS(841),
    [anon_sym_void] = ACTIONS(841),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(843),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(845),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(435),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(152)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1971),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_COLON] = ACTIONS(449),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_as] = ACTIONS(449),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(153)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2531),
    [sym_pattern_list] = STATE(2531),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2009),
    [sym__primary_expression] = STATE(2039),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1505),
    [sym_subscript] = STATE(1505),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(803),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(847),
    [anon_sym_def] = ACTIONS(849),
    [anon_sym_LPAREN] = ACTIONS(763),
    [anon_sym_LBRACK] = ACTIONS(809),
    [anon_sym_RBRACK] = ACTIONS(435),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_for] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(851),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(853),
    [anon_sym_DASH] = ACTIONS(853),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(853),
    [anon_sym_typeof] = ACTIONS(853),
    [anon_sym_void] = ACTIONS(853),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(855),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(857),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(803),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(154)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1944),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_EQ] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_RPAREN] = ACTIONS(527),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(155)] = {
    [sym__simple_statements] = STATE(1563),
    [sym__simple_statement] = STATE(2692),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1579),
    [sym_block] = STATE(1579),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(861),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(156)] = {
    [sym__simple_statements] = STATE(1561),
    [sym__simple_statement] = STATE(2712),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1567),
    [sym_block] = STATE(1567),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(863),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(157)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1641),
    [sym_block] = STATE(1641),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(158)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1645),
    [sym_block] = STATE(1645),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(159)] = {
    [sym__simple_statements] = STATE(2670),
    [sym__simple_statement] = STATE(2699),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(3091),
    [sym_block] = STATE(3091),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(160)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1629),
    [sym_block] = STATE(1629),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(161)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1633),
    [sym_block] = STATE(1633),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(162)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1637),
    [sym_block] = STATE(1637),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(163)] = {
    [sym__simple_statements] = STATE(1550),
    [sym__simple_statement] = STATE(2697),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1533),
    [sym_block] = STATE(1533),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(871),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(164)] = {
    [sym__simple_statements] = STATE(1561),
    [sym__simple_statement] = STATE(2712),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1573),
    [sym_block] = STATE(1573),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(863),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(165)] = {
    [sym__simple_statements] = STATE(1548),
    [sym__simple_statement] = STATE(2688),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1538),
    [sym_block] = STATE(1538),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(873),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(166)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1616),
    [sym_block] = STATE(1616),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(167)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1615),
    [sym_block] = STATE(1615),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(168)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1636),
    [sym_block] = STATE(1636),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(169)] = {
    [sym__simple_statements] = STATE(1587),
    [sym__simple_statement] = STATE(2759),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1598),
    [sym_block] = STATE(1598),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(875),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(170)] = {
    [sym__simple_statements] = STATE(1548),
    [sym__simple_statement] = STATE(2688),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1541),
    [sym_block] = STATE(1541),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(873),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(171)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1612),
    [sym_block] = STATE(1612),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(172)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2533),
    [sym_pattern_list] = STATE(2533),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2002),
    [sym__primary_expression] = STATE(2157),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1505),
    [sym_subscript] = STATE(1505),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(803),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(805),
    [anon_sym_def] = ACTIONS(807),
    [anon_sym_COLON] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(877),
    [anon_sym_LBRACK] = ACTIONS(879),
    [anon_sym_RBRACK] = ACTIONS(449),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(811),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(813),
    [anon_sym_DASH] = ACTIONS(813),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(813),
    [anon_sym_typeof] = ACTIONS(813),
    [anon_sym_void] = ACTIONS(813),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(815),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(817),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(803),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(173)] = {
    [sym__simple_statements] = STATE(1561),
    [sym__simple_statement] = STATE(2712),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1577),
    [sym_block] = STATE(1577),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(863),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(174)] = {
    [sym__simple_statements] = STATE(1548),
    [sym__simple_statement] = STATE(2688),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1542),
    [sym_block] = STATE(1542),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(873),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(175)] = {
    [sym__simple_statements] = STATE(1548),
    [sym__simple_statement] = STATE(2688),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1543),
    [sym_block] = STATE(1543),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(873),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(176)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1605),
    [sym_block] = STATE(1605),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(177)] = {
    [sym__simple_statements] = STATE(1548),
    [sym__simple_statement] = STATE(2688),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1544),
    [sym_block] = STATE(1544),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(873),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(178)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1654),
    [sym_block] = STATE(1654),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(179)] = {
    [sym__simple_statements] = STATE(1548),
    [sym__simple_statement] = STATE(2688),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1545),
    [sym_block] = STATE(1545),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(873),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(180)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2546),
    [sym_pattern_list] = STATE(2546),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2154),
    [sym__primary_expression] = STATE(2050),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(821),
    [anon_sym_def] = ACTIONS(823),
    [anon_sym_LPAREN] = ACTIONS(797),
    [anon_sym_RPAREN] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(799),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_for] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(825),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(827),
    [anon_sym_DASH] = ACTIONS(827),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(827),
    [anon_sym_typeof] = ACTIONS(827),
    [anon_sym_void] = ACTIONS(827),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(829),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(831),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(181)] = {
    [sym__simple_statements] = STATE(1550),
    [sym__simple_statement] = STATE(2697),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1540),
    [sym_block] = STATE(1540),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(871),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(182)] = {
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2012),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(451),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(183)] = {
    [sym__simple_statements] = STATE(1563),
    [sym__simple_statement] = STATE(2692),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1580),
    [sym_block] = STATE(1580),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(861),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(184)] = {
    [sym__simple_statements] = STATE(1563),
    [sym__simple_statement] = STATE(2692),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1508),
    [sym_block] = STATE(1508),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(861),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(185)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2129),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_else] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(186)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2542),
    [sym_pattern_list] = STATE(2542),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2290),
    [sym__primary_expression] = STATE(2174),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_def] = ACTIONS(883),
    [anon_sym_LPAREN] = ACTIONS(763),
    [anon_sym_LBRACK] = ACTIONS(765),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_else] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(885),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(887),
    [anon_sym_DASH] = ACTIONS(887),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(887),
    [anon_sym_typeof] = ACTIONS(887),
    [anon_sym_void] = ACTIONS(887),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(889),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(891),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(187)] = {
    [sym__simple_statements] = STATE(1563),
    [sym__simple_statement] = STATE(2692),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1565),
    [sym_block] = STATE(1565),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(861),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(188)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1626),
    [sym_block] = STATE(1626),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(189)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2129),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_else] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(190)] = {
    [sym__simple_statements] = STATE(1642),
    [sym__simple_statement] = STATE(2774),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1639),
    [sym_block] = STATE(1639),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(865),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(191)] = {
    [sym__simple_statements] = STATE(2670),
    [sym__simple_statement] = STATE(2699),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(3097),
    [sym_block] = STATE(3097),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(192)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2531),
    [sym_pattern_list] = STATE(2531),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2049),
    [sym__primary_expression] = STATE(2039),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1505),
    [sym_subscript] = STATE(1505),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(803),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(847),
    [anon_sym_def] = ACTIONS(849),
    [anon_sym_LPAREN] = ACTIONS(877),
    [anon_sym_LBRACK] = ACTIONS(879),
    [anon_sym_RBRACK] = ACTIONS(449),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_for] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(851),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(853),
    [anon_sym_DASH] = ACTIONS(853),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(853),
    [anon_sym_typeof] = ACTIONS(853),
    [anon_sym_void] = ACTIONS(853),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(855),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(857),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(803),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(193)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2094),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_COMMA] = ACTIONS(449),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_else] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(194)] = {
    [sym__simple_statements] = STATE(1593),
    [sym__simple_statement] = STATE(2746),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1599),
    [sym_block] = STATE(1599),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(893),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(195)] = {
    [sym__simple_statements] = STATE(1550),
    [sym__simple_statement] = STATE(2697),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1532),
    [sym_block] = STATE(1532),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(871),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(196)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1631),
    [sym_block] = STATE(1631),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(197)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1614),
    [sym_block] = STATE(1614),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(198)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1622),
    [sym_block] = STATE(1622),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(199)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1609),
    [sym_block] = STATE(1609),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(200)] = {
    [sym__simple_statements] = STATE(1550),
    [sym__simple_statement] = STATE(2697),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1552),
    [sym_block] = STATE(1552),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(871),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(201)] = {
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2103),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_COMMA] = ACTIONS(505),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(503),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(202)] = {
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2539),
    [sym_pattern_list] = STATE(2539),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2218),
    [sym__primary_expression] = STATE(2165),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1517),
    [sym_subscript] = STATE(1517),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_argument_list] = STATE(2179),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(895),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(897),
    [anon_sym_def] = ACTIONS(899),
    [anon_sym_LPAREN] = ACTIONS(901),
    [anon_sym_LBRACK] = ACTIONS(903),
    [anon_sym_DOT] = ACTIONS(795),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(905),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(907),
    [anon_sym_DASH] = ACTIONS(907),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(907),
    [anon_sym_typeof] = ACTIONS(907),
    [anon_sym_void] = ACTIONS(907),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(909),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(911),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(895),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(433),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(203)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1607),
    [sym_block] = STATE(1607),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(204)] = {
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2103),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(527),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_COMMA] = ACTIONS(527),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_DOT] = ACTIONS(527),
    [anon_sym_if] = ACTIONS(527),
    [anon_sym_in] = ACTIONS(527),
    [anon_sym_STAR] = ACTIONS(527),
    [anon_sym_STAR_STAR] = ACTIONS(527),
    [anon_sym_AT] = ACTIONS(527),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_and] = ACTIONS(527),
    [anon_sym_or] = ACTIONS(527),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_SLASH] = ACTIONS(527),
    [anon_sym_SLASH_SLASH] = ACTIONS(527),
    [anon_sym_PERCENT] = ACTIONS(527),
    [anon_sym_PIPE] = ACTIONS(527),
    [anon_sym_CARET] = ACTIONS(527),
    [anon_sym_AMP] = ACTIONS(527),
    [anon_sym_LT_LT] = ACTIONS(527),
    [anon_sym_GT_GT] = ACTIONS(527),
    [anon_sym_GT_GT_GT] = ACTIONS(527),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_LT] = ACTIONS(527),
    [anon_sym_LT_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ] = ACTIONS(527),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ] = ACTIONS(527),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(527),
    [anon_sym_GT_EQ] = ACTIONS(527),
    [anon_sym_GT] = ACTIONS(527),
    [anon_sym_is] = ACTIONS(527),
    [anon_sym_instanceof] = ACTIONS(527),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_QMARK] = ACTIONS(527),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(525),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(205)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1601),
    [sym_block] = STATE(1601),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(206)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1627),
    [sym_block] = STATE(1627),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(207)] = {
    [sym__simple_statements] = STATE(2670),
    [sym__simple_statement] = STATE(2699),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(2590),
    [sym_block] = STATE(2590),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(208)] = {
    [sym__simple_statements] = STATE(1550),
    [sym__simple_statement] = STATE(2697),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1555),
    [sym_block] = STATE(1555),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(871),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(209)] = {
    [sym__simple_statements] = STATE(2670),
    [sym__simple_statement] = STATE(2699),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(2586),
    [sym_block] = STATE(2586),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(210)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1653),
    [sym_block] = STATE(1653),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(211)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2528),
    [sym_pattern_list] = STATE(2528),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2207),
    [sym__primary_expression] = STATE(2193),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_argument_list] = STATE(1685),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(435),
    [anon_sym_async] = ACTIONS(913),
    [anon_sym_def] = ACTIONS(915),
    [anon_sym_LPAREN] = ACTIONS(763),
    [anon_sym_RPAREN] = ACTIONS(435),
    [anon_sym_LBRACK] = ACTIONS(765),
    [anon_sym_DOT] = ACTIONS(739),
    [anon_sym_if] = ACTIONS(435),
    [anon_sym_in] = ACTIONS(435),
    [anon_sym_STAR] = ACTIONS(435),
    [anon_sym_STAR_STAR] = ACTIONS(435),
    [anon_sym_AT] = ACTIONS(435),
    [anon_sym_not] = ACTIONS(917),
    [anon_sym_and] = ACTIONS(435),
    [anon_sym_or] = ACTIONS(435),
    [anon_sym_PLUS] = ACTIONS(919),
    [anon_sym_DASH] = ACTIONS(919),
    [anon_sym_SLASH] = ACTIONS(435),
    [anon_sym_SLASH_SLASH] = ACTIONS(435),
    [anon_sym_PERCENT] = ACTIONS(435),
    [anon_sym_PIPE] = ACTIONS(435),
    [anon_sym_CARET] = ACTIONS(435),
    [anon_sym_AMP] = ACTIONS(435),
    [anon_sym_LT_LT] = ACTIONS(435),
    [anon_sym_GT_GT] = ACTIONS(435),
    [anon_sym_GT_GT_GT] = ACTIONS(435),
    [anon_sym_TILDE] = ACTIONS(919),
    [anon_sym_typeof] = ACTIONS(919),
    [anon_sym_void] = ACTIONS(919),
    [anon_sym_LT] = ACTIONS(435),
    [anon_sym_LT_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ] = ACTIONS(435),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ] = ACTIONS(435),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(435),
    [anon_sym_GT_EQ] = ACTIONS(435),
    [anon_sym_GT] = ACTIONS(435),
    [anon_sym_is] = ACTIONS(435),
    [anon_sym_instanceof] = ACTIONS(435),
    [anon_sym_new] = ACTIONS(921),
    [anon_sym_QMARK] = ACTIONS(435),
    [anon_sym_await] = ACTIONS(923),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(212)] = {
    [sym__simple_statements] = STATE(1655),
    [sym__simple_statement] = STATE(2666),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1656),
    [sym_block] = STATE(1656),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(869),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(213)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2538),
    [sym_pattern_list] = STATE(2538),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(1988),
    [sym__primary_expression] = STATE(2072),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(835),
    [anon_sym_def] = ACTIONS(837),
    [anon_sym_LPAREN] = ACTIONS(797),
    [anon_sym_LBRACK] = ACTIONS(799),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_for] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(839),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(841),
    [anon_sym_DASH] = ACTIONS(841),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(841),
    [anon_sym_typeof] = ACTIONS(841),
    [anon_sym_void] = ACTIONS(841),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(843),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(845),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(449),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(214)] = {
    [sym__simple_statements] = STATE(1561),
    [sym__simple_statement] = STATE(2712),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1509),
    [sym_block] = STATE(1509),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(863),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(215)] = {
    [sym__simple_statements] = STATE(1550),
    [sym__simple_statement] = STATE(2697),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__suite] = STATE(1547),
    [sym_block] = STATE(1547),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(871),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(216)] = {
    [ts_builtin_sym_end] = ACTIONS(925),
    [sym_identifier] = ACTIONS(927),
    [anon_sym_SEMI] = ACTIONS(927),
    [anon_sym_EQ] = ACTIONS(929),
    [anon_sym_return] = ACTIONS(927),
    [anon_sym_async] = ACTIONS(927),
    [anon_sym_def] = ACTIONS(927),
    [anon_sym_PLUS_EQ] = ACTIONS(929),
    [anon_sym_DASH_EQ] = ACTIONS(929),
    [anon_sym_STAR_EQ] = ACTIONS(929),
    [anon_sym_SLASH_EQ] = ACTIONS(929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(929),
    [anon_sym_PERCENT_EQ] = ACTIONS(929),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(929),
    [anon_sym_GT_GT_EQ] = ACTIONS(929),
    [anon_sym_LT_LT_EQ] = ACTIONS(929),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(929),
    [anon_sym_AMP_EQ] = ACTIONS(929),
    [anon_sym_CARET_EQ] = ACTIONS(929),
    [anon_sym_PIPE_EQ] = ACTIONS(929),
    [anon_sym_COMMA] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(927),
    [anon_sym_LBRACK] = ACTIONS(927),
    [anon_sym_del] = ACTIONS(927),
    [anon_sym_raise] = ACTIONS(927),
    [sym_pass_statement] = ACTIONS(927),
    [sym_break_statement] = ACTIONS(927),
    [sym_continue_statement] = ACTIONS(927),
    [sym_debugger_statement] = ACTIONS(927),
    [anon_sym_assert] = ACTIONS(927),
    [anon_sym_global] = ACTIONS(927),
    [anon_sym_nonlocal] = ACTIONS(927),
    [anon_sym_yield] = ACTIONS(927),
    [anon_sym_from] = ACTIONS(927),
    [anon_sym_import] = ACTIONS(927),
    [anon_sym_DOT] = ACTIONS(927),
    [anon_sym_if] = ACTIONS(927),
    [anon_sym_for] = ACTIONS(927),
    [anon_sym_in] = ACTIONS(927),
    [anon_sym_while] = ACTIONS(927),
    [anon_sym_do] = ACTIONS(927),
    [anon_sym_try] = ACTIONS(927),
    [anon_sym_with] = ACTIONS(927),
    [anon_sym_STAR] = ACTIONS(927),
    [anon_sym_STAR_STAR] = ACTIONS(927),
    [anon_sym_class] = ACTIONS(927),
    [anon_sym_AT] = ACTIONS(927),
    [anon_sym_not] = ACTIONS(927),
    [anon_sym_and] = ACTIONS(927),
    [anon_sym_or] = ACTIONS(927),
    [anon_sym_PLUS] = ACTIONS(927),
    [anon_sym_DASH] = ACTIONS(927),
    [anon_sym_SLASH] = ACTIONS(927),
    [anon_sym_SLASH_SLASH] = ACTIONS(927),
    [anon_sym_PERCENT] = ACTIONS(927),
    [anon_sym_PIPE] = ACTIONS(927),
    [anon_sym_CARET] = ACTIONS(927),
    [anon_sym_AMP] = ACTIONS(927),
    [anon_sym_LT_LT] = ACTIONS(927),
    [anon_sym_GT_GT] = ACTIONS(927),
    [anon_sym_GT_GT_GT] = ACTIONS(927),
    [anon_sym_TILDE] = ACTIONS(927),
    [anon_sym_typeof] = ACTIONS(927),
    [anon_sym_void] = ACTIONS(927),
    [anon_sym_LT] = ACTIONS(927),
    [anon_sym_LT_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(927),
    [anon_sym_GT_EQ] = ACTIONS(927),
    [anon_sym_GT] = ACTIONS(927),
    [anon_sym_is] = ACTIONS(927),
    [anon_sym_instanceof] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
    [anon_sym_QMARK] = ACTIONS(927),
    [anon_sym_await] = ACTIONS(927),
    [anon_sym_LBRACE] = ACTIONS(927),
    [sym_this] = ACTIONS(927),
    [sym_true] = ACTIONS(927),
    [sym_false] = ACTIONS(927),
    [sym_none] = ACTIONS(927),
    [sym_string] = ACTIONS(927),
    [sym_verbatim] = ACTIONS(927),
    [sym_number] = ACTIONS(927),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(925),
    [sym_fstring_start] = ACTIONS(925),
  },
  [STATE(217)] = {
    [ts_builtin_sym_end] = ACTIONS(931),
    [sym_identifier] = ACTIONS(933),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym_EQ] = ACTIONS(935),
    [anon_sym_return] = ACTIONS(933),
    [anon_sym_async] = ACTIONS(933),
    [anon_sym_def] = ACTIONS(933),
    [anon_sym_PLUS_EQ] = ACTIONS(935),
    [anon_sym_DASH_EQ] = ACTIONS(935),
    [anon_sym_STAR_EQ] = ACTIONS(935),
    [anon_sym_SLASH_EQ] = ACTIONS(935),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(935),
    [anon_sym_PERCENT_EQ] = ACTIONS(935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_LT_LT_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_AMP_EQ] = ACTIONS(935),
    [anon_sym_CARET_EQ] = ACTIONS(935),
    [anon_sym_PIPE_EQ] = ACTIONS(935),
    [anon_sym_COMMA] = ACTIONS(935),
    [anon_sym_LPAREN] = ACTIONS(933),
    [anon_sym_LBRACK] = ACTIONS(933),
    [anon_sym_del] = ACTIONS(933),
    [anon_sym_raise] = ACTIONS(933),
    [sym_pass_statement] = ACTIONS(933),
    [sym_break_statement] = ACTIONS(933),
    [sym_continue_statement] = ACTIONS(933),
    [sym_debugger_statement] = ACTIONS(933),
    [anon_sym_assert] = ACTIONS(933),
    [anon_sym_global] = ACTIONS(933),
    [anon_sym_nonlocal] = ACTIONS(933),
    [anon_sym_yield] = ACTIONS(933),
    [anon_sym_from] = ACTIONS(933),
    [anon_sym_import] = ACTIONS(933),
    [anon_sym_DOT] = ACTIONS(933),
    [anon_sym_if] = ACTIONS(933),
    [anon_sym_for] = ACTIONS(933),
    [anon_sym_in] = ACTIONS(933),
    [anon_sym_while] = ACTIONS(933),
    [anon_sym_do] = ACTIONS(933),
    [anon_sym_try] = ACTIONS(933),
    [anon_sym_with] = ACTIONS(933),
    [anon_sym_STAR] = ACTIONS(933),
    [anon_sym_STAR_STAR] = ACTIONS(933),
    [anon_sym_class] = ACTIONS(933),
    [anon_sym_AT] = ACTIONS(933),
    [anon_sym_not] = ACTIONS(933),
    [anon_sym_and] = ACTIONS(933),
    [anon_sym_or] = ACTIONS(933),
    [anon_sym_PLUS] = ACTIONS(933),
    [anon_sym_DASH] = ACTIONS(933),
    [anon_sym_SLASH] = ACTIONS(933),
    [anon_sym_SLASH_SLASH] = ACTIONS(933),
    [anon_sym_PERCENT] = ACTIONS(933),
    [anon_sym_PIPE] = ACTIONS(933),
    [anon_sym_CARET] = ACTIONS(933),
    [anon_sym_AMP] = ACTIONS(933),
    [anon_sym_LT_LT] = ACTIONS(933),
    [anon_sym_GT_GT] = ACTIONS(933),
    [anon_sym_GT_GT_GT] = ACTIONS(933),
    [anon_sym_TILDE] = ACTIONS(933),
    [anon_sym_typeof] = ACTIONS(933),
    [anon_sym_void] = ACTIONS(933),
    [anon_sym_LT] = ACTIONS(933),
    [anon_sym_LT_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(933),
    [anon_sym_GT_EQ] = ACTIONS(933),
    [anon_sym_GT] = ACTIONS(933),
    [anon_sym_is] = ACTIONS(933),
    [anon_sym_instanceof] = ACTIONS(933),
    [anon_sym_new] = ACTIONS(933),
    [anon_sym_QMARK] = ACTIONS(933),
    [anon_sym_await] = ACTIONS(933),
    [anon_sym_LBRACE] = ACTIONS(933),
    [sym_this] = ACTIONS(933),
    [sym_true] = ACTIONS(933),
    [sym_false] = ACTIONS(933),
    [sym_none] = ACTIONS(933),
    [sym_string] = ACTIONS(933),
    [sym_verbatim] = ACTIONS(933),
    [sym_number] = ACTIONS(933),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(931),
    [sym_fstring_start] = ACTIONS(931),
  },
  [STATE(218)] = {
    [ts_builtin_sym_end] = ACTIONS(937),
    [sym_identifier] = ACTIONS(939),
    [anon_sym_SEMI] = ACTIONS(939),
    [anon_sym_EQ] = ACTIONS(941),
    [anon_sym_return] = ACTIONS(939),
    [anon_sym_async] = ACTIONS(939),
    [anon_sym_def] = ACTIONS(939),
    [anon_sym_PLUS_EQ] = ACTIONS(941),
    [anon_sym_DASH_EQ] = ACTIONS(941),
    [anon_sym_STAR_EQ] = ACTIONS(941),
    [anon_sym_SLASH_EQ] = ACTIONS(941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(941),
    [anon_sym_PERCENT_EQ] = ACTIONS(941),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(941),
    [anon_sym_GT_GT_EQ] = ACTIONS(941),
    [anon_sym_LT_LT_EQ] = ACTIONS(941),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(941),
    [anon_sym_AMP_EQ] = ACTIONS(941),
    [anon_sym_CARET_EQ] = ACTIONS(941),
    [anon_sym_PIPE_EQ] = ACTIONS(941),
    [anon_sym_COMMA] = ACTIONS(941),
    [anon_sym_LPAREN] = ACTIONS(939),
    [anon_sym_LBRACK] = ACTIONS(939),
    [anon_sym_del] = ACTIONS(939),
    [anon_sym_raise] = ACTIONS(939),
    [sym_pass_statement] = ACTIONS(939),
    [sym_break_statement] = ACTIONS(939),
    [sym_continue_statement] = ACTIONS(939),
    [sym_debugger_statement] = ACTIONS(939),
    [anon_sym_assert] = ACTIONS(939),
    [anon_sym_global] = ACTIONS(939),
    [anon_sym_nonlocal] = ACTIONS(939),
    [anon_sym_yield] = ACTIONS(939),
    [anon_sym_from] = ACTIONS(939),
    [anon_sym_import] = ACTIONS(939),
    [anon_sym_DOT] = ACTIONS(939),
    [anon_sym_if] = ACTIONS(939),
    [anon_sym_for] = ACTIONS(939),
    [anon_sym_in] = ACTIONS(939),
    [anon_sym_while] = ACTIONS(939),
    [anon_sym_do] = ACTIONS(939),
    [anon_sym_try] = ACTIONS(939),
    [anon_sym_with] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(939),
    [anon_sym_STAR_STAR] = ACTIONS(939),
    [anon_sym_class] = ACTIONS(939),
    [anon_sym_AT] = ACTIONS(939),
    [anon_sym_not] = ACTIONS(939),
    [anon_sym_and] = ACTIONS(939),
    [anon_sym_or] = ACTIONS(939),
    [anon_sym_PLUS] = ACTIONS(939),
    [anon_sym_DASH] = ACTIONS(939),
    [anon_sym_SLASH] = ACTIONS(939),
    [anon_sym_SLASH_SLASH] = ACTIONS(939),
    [anon_sym_PERCENT] = ACTIONS(939),
    [anon_sym_PIPE] = ACTIONS(939),
    [anon_sym_CARET] = ACTIONS(939),
    [anon_sym_AMP] = ACTIONS(939),
    [anon_sym_LT_LT] = ACTIONS(939),
    [anon_sym_GT_GT] = ACTIONS(939),
    [anon_sym_GT_GT_GT] = ACTIONS(939),
    [anon_sym_TILDE] = ACTIONS(939),
    [anon_sym_typeof] = ACTIONS(939),
    [anon_sym_void] = ACTIONS(939),
    [anon_sym_LT] = ACTIONS(939),
    [anon_sym_LT_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(939),
    [anon_sym_GT_EQ] = ACTIONS(939),
    [anon_sym_GT] = ACTIONS(939),
    [anon_sym_is] = ACTIONS(939),
    [anon_sym_instanceof] = ACTIONS(939),
    [anon_sym_new] = ACTIONS(939),
    [anon_sym_QMARK] = ACTIONS(939),
    [anon_sym_await] = ACTIONS(939),
    [anon_sym_LBRACE] = ACTIONS(939),
    [sym_this] = ACTIONS(939),
    [sym_true] = ACTIONS(939),
    [sym_false] = ACTIONS(939),
    [sym_none] = ACTIONS(939),
    [sym_string] = ACTIONS(939),
    [sym_verbatim] = ACTIONS(939),
    [sym_number] = ACTIONS(939),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(937),
    [sym_fstring_start] = ACTIONS(937),
  },
  [STATE(219)] = {
    [sym_identifier] = ACTIONS(943),
    [anon_sym_SEMI] = ACTIONS(943),
    [anon_sym_EQ] = ACTIONS(943),
    [anon_sym_return] = ACTIONS(943),
    [anon_sym_async] = ACTIONS(943),
    [anon_sym_def] = ACTIONS(943),
    [anon_sym_PLUS_EQ] = ACTIONS(943),
    [anon_sym_DASH_EQ] = ACTIONS(943),
    [anon_sym_STAR_EQ] = ACTIONS(943),
    [anon_sym_SLASH_EQ] = ACTIONS(943),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(943),
    [anon_sym_PERCENT_EQ] = ACTIONS(943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(943),
    [anon_sym_GT_GT_EQ] = ACTIONS(943),
    [anon_sym_LT_LT_EQ] = ACTIONS(943),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(943),
    [anon_sym_AMP_EQ] = ACTIONS(943),
    [anon_sym_CARET_EQ] = ACTIONS(943),
    [anon_sym_PIPE_EQ] = ACTIONS(943),
    [anon_sym_COMMA] = ACTIONS(943),
    [anon_sym_LPAREN] = ACTIONS(943),
    [anon_sym_LBRACK] = ACTIONS(943),
    [anon_sym_del] = ACTIONS(943),
    [anon_sym_raise] = ACTIONS(943),
    [sym_pass_statement] = ACTIONS(943),
    [sym_break_statement] = ACTIONS(943),
    [sym_continue_statement] = ACTIONS(943),
    [sym_debugger_statement] = ACTIONS(943),
    [anon_sym_assert] = ACTIONS(943),
    [anon_sym_global] = ACTIONS(943),
    [anon_sym_nonlocal] = ACTIONS(943),
    [anon_sym_yield] = ACTIONS(943),
    [anon_sym_from] = ACTIONS(943),
    [anon_sym_import] = ACTIONS(943),
    [anon_sym_DOT] = ACTIONS(943),
    [anon_sym_if] = ACTIONS(943),
    [anon_sym_for] = ACTIONS(943),
    [anon_sym_in] = ACTIONS(943),
    [anon_sym_while] = ACTIONS(943),
    [anon_sym_do] = ACTIONS(943),
    [anon_sym_try] = ACTIONS(943),
    [anon_sym_with] = ACTIONS(943),
    [anon_sym_STAR] = ACTIONS(943),
    [anon_sym_STAR_STAR] = ACTIONS(943),
    [anon_sym_class] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(943),
    [anon_sym_not] = ACTIONS(943),
    [anon_sym_and] = ACTIONS(943),
    [anon_sym_or] = ACTIONS(943),
    [anon_sym_PLUS] = ACTIONS(943),
    [anon_sym_DASH] = ACTIONS(943),
    [anon_sym_SLASH] = ACTIONS(943),
    [anon_sym_SLASH_SLASH] = ACTIONS(943),
    [anon_sym_PERCENT] = ACTIONS(943),
    [anon_sym_PIPE] = ACTIONS(943),
    [anon_sym_CARET] = ACTIONS(943),
    [anon_sym_AMP] = ACTIONS(943),
    [anon_sym_LT_LT] = ACTIONS(943),
    [anon_sym_GT_GT] = ACTIONS(943),
    [anon_sym_GT_GT_GT] = ACTIONS(943),
    [anon_sym_TILDE] = ACTIONS(943),
    [anon_sym_typeof] = ACTIONS(943),
    [anon_sym_void] = ACTIONS(943),
    [anon_sym_LT] = ACTIONS(943),
    [anon_sym_LT_EQ] = ACTIONS(943),
    [anon_sym_EQ_EQ] = ACTIONS(943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(943),
    [anon_sym_BANG_EQ] = ACTIONS(943),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(943),
    [anon_sym_GT_EQ] = ACTIONS(943),
    [anon_sym_GT] = ACTIONS(943),
    [anon_sym_is] = ACTIONS(943),
    [anon_sym_instanceof] = ACTIONS(943),
    [anon_sym_new] = ACTIONS(943),
    [anon_sym_QMARK] = ACTIONS(943),
    [anon_sym_await] = ACTIONS(943),
    [anon_sym_LBRACE] = ACTIONS(943),
    [sym_this] = ACTIONS(943),
    [sym_true] = ACTIONS(943),
    [sym_false] = ACTIONS(943),
    [sym_none] = ACTIONS(943),
    [sym_string] = ACTIONS(943),
    [sym_verbatim] = ACTIONS(943),
    [sym_number] = ACTIONS(943),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(945),
    [sym_regex] = ACTIONS(945),
    [sym_fstring_start] = ACTIONS(945),
  },
  [STATE(220)] = {
    [sym_identifier] = ACTIONS(947),
    [anon_sym_SEMI] = ACTIONS(947),
    [anon_sym_EQ] = ACTIONS(947),
    [anon_sym_return] = ACTIONS(947),
    [anon_sym_async] = ACTIONS(947),
    [anon_sym_def] = ACTIONS(947),
    [anon_sym_PLUS_EQ] = ACTIONS(947),
    [anon_sym_DASH_EQ] = ACTIONS(947),
    [anon_sym_STAR_EQ] = ACTIONS(947),
    [anon_sym_SLASH_EQ] = ACTIONS(947),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(947),
    [anon_sym_PERCENT_EQ] = ACTIONS(947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(947),
    [anon_sym_GT_GT_EQ] = ACTIONS(947),
    [anon_sym_LT_LT_EQ] = ACTIONS(947),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(947),
    [anon_sym_AMP_EQ] = ACTIONS(947),
    [anon_sym_CARET_EQ] = ACTIONS(947),
    [anon_sym_PIPE_EQ] = ACTIONS(947),
    [anon_sym_COMMA] = ACTIONS(947),
    [anon_sym_LPAREN] = ACTIONS(947),
    [anon_sym_LBRACK] = ACTIONS(947),
    [anon_sym_del] = ACTIONS(947),
    [anon_sym_raise] = ACTIONS(947),
    [sym_pass_statement] = ACTIONS(947),
    [sym_break_statement] = ACTIONS(947),
    [sym_continue_statement] = ACTIONS(947),
    [sym_debugger_statement] = ACTIONS(947),
    [anon_sym_assert] = ACTIONS(947),
    [anon_sym_global] = ACTIONS(947),
    [anon_sym_nonlocal] = ACTIONS(947),
    [anon_sym_yield] = ACTIONS(947),
    [anon_sym_from] = ACTIONS(947),
    [anon_sym_import] = ACTIONS(947),
    [anon_sym_DOT] = ACTIONS(947),
    [anon_sym_if] = ACTIONS(947),
    [anon_sym_for] = ACTIONS(947),
    [anon_sym_in] = ACTIONS(947),
    [anon_sym_while] = ACTIONS(947),
    [anon_sym_do] = ACTIONS(947),
    [anon_sym_try] = ACTIONS(947),
    [anon_sym_with] = ACTIONS(947),
    [anon_sym_STAR] = ACTIONS(947),
    [anon_sym_STAR_STAR] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_AT] = ACTIONS(947),
    [anon_sym_not] = ACTIONS(947),
    [anon_sym_and] = ACTIONS(947),
    [anon_sym_or] = ACTIONS(947),
    [anon_sym_PLUS] = ACTIONS(947),
    [anon_sym_DASH] = ACTIONS(947),
    [anon_sym_SLASH] = ACTIONS(947),
    [anon_sym_SLASH_SLASH] = ACTIONS(947),
    [anon_sym_PERCENT] = ACTIONS(947),
    [anon_sym_PIPE] = ACTIONS(947),
    [anon_sym_CARET] = ACTIONS(947),
    [anon_sym_AMP] = ACTIONS(947),
    [anon_sym_LT_LT] = ACTIONS(947),
    [anon_sym_GT_GT] = ACTIONS(947),
    [anon_sym_GT_GT_GT] = ACTIONS(947),
    [anon_sym_TILDE] = ACTIONS(947),
    [anon_sym_typeof] = ACTIONS(947),
    [anon_sym_void] = ACTIONS(947),
    [anon_sym_LT] = ACTIONS(947),
    [anon_sym_LT_EQ] = ACTIONS(947),
    [anon_sym_EQ_EQ] = ACTIONS(947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(947),
    [anon_sym_BANG_EQ] = ACTIONS(947),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(947),
    [anon_sym_GT_EQ] = ACTIONS(947),
    [anon_sym_GT] = ACTIONS(947),
    [anon_sym_is] = ACTIONS(947),
    [anon_sym_instanceof] = ACTIONS(947),
    [anon_sym_new] = ACTIONS(947),
    [anon_sym_QMARK] = ACTIONS(947),
    [anon_sym_await] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(947),
    [sym_this] = ACTIONS(947),
    [sym_true] = ACTIONS(947),
    [sym_false] = ACTIONS(947),
    [sym_none] = ACTIONS(947),
    [sym_string] = ACTIONS(947),
    [sym_verbatim] = ACTIONS(947),
    [sym_number] = ACTIONS(947),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(949),
    [sym_regex] = ACTIONS(949),
    [sym_fstring_start] = ACTIONS(949),
  },
  [STATE(221)] = {
    [sym_identifier] = ACTIONS(947),
    [anon_sym_SEMI] = ACTIONS(947),
    [anon_sym_EQ] = ACTIONS(947),
    [anon_sym_return] = ACTIONS(947),
    [anon_sym_async] = ACTIONS(947),
    [anon_sym_def] = ACTIONS(947),
    [anon_sym_PLUS_EQ] = ACTIONS(947),
    [anon_sym_DASH_EQ] = ACTIONS(947),
    [anon_sym_STAR_EQ] = ACTIONS(947),
    [anon_sym_SLASH_EQ] = ACTIONS(947),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(947),
    [anon_sym_PERCENT_EQ] = ACTIONS(947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(947),
    [anon_sym_GT_GT_EQ] = ACTIONS(947),
    [anon_sym_LT_LT_EQ] = ACTIONS(947),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(947),
    [anon_sym_AMP_EQ] = ACTIONS(947),
    [anon_sym_CARET_EQ] = ACTIONS(947),
    [anon_sym_PIPE_EQ] = ACTIONS(947),
    [anon_sym_COMMA] = ACTIONS(947),
    [anon_sym_LPAREN] = ACTIONS(947),
    [anon_sym_LBRACK] = ACTIONS(947),
    [anon_sym_del] = ACTIONS(947),
    [anon_sym_raise] = ACTIONS(947),
    [sym_pass_statement] = ACTIONS(947),
    [sym_break_statement] = ACTIONS(947),
    [sym_continue_statement] = ACTIONS(947),
    [sym_debugger_statement] = ACTIONS(947),
    [anon_sym_assert] = ACTIONS(947),
    [anon_sym_global] = ACTIONS(947),
    [anon_sym_nonlocal] = ACTIONS(947),
    [anon_sym_yield] = ACTIONS(947),
    [anon_sym_from] = ACTIONS(947),
    [anon_sym_import] = ACTIONS(947),
    [anon_sym_DOT] = ACTIONS(947),
    [anon_sym_if] = ACTIONS(947),
    [anon_sym_for] = ACTIONS(947),
    [anon_sym_in] = ACTIONS(947),
    [anon_sym_while] = ACTIONS(947),
    [anon_sym_do] = ACTIONS(947),
    [anon_sym_try] = ACTIONS(947),
    [anon_sym_with] = ACTIONS(947),
    [anon_sym_STAR] = ACTIONS(947),
    [anon_sym_STAR_STAR] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_AT] = ACTIONS(947),
    [anon_sym_not] = ACTIONS(947),
    [anon_sym_and] = ACTIONS(947),
    [anon_sym_or] = ACTIONS(947),
    [anon_sym_PLUS] = ACTIONS(947),
    [anon_sym_DASH] = ACTIONS(947),
    [anon_sym_SLASH] = ACTIONS(947),
    [anon_sym_SLASH_SLASH] = ACTIONS(947),
    [anon_sym_PERCENT] = ACTIONS(947),
    [anon_sym_PIPE] = ACTIONS(947),
    [anon_sym_CARET] = ACTIONS(947),
    [anon_sym_AMP] = ACTIONS(947),
    [anon_sym_LT_LT] = ACTIONS(947),
    [anon_sym_GT_GT] = ACTIONS(947),
    [anon_sym_GT_GT_GT] = ACTIONS(947),
    [anon_sym_TILDE] = ACTIONS(947),
    [anon_sym_typeof] = ACTIONS(947),
    [anon_sym_void] = ACTIONS(947),
    [anon_sym_LT] = ACTIONS(947),
    [anon_sym_LT_EQ] = ACTIONS(947),
    [anon_sym_EQ_EQ] = ACTIONS(947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(947),
    [anon_sym_BANG_EQ] = ACTIONS(947),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(947),
    [anon_sym_GT_EQ] = ACTIONS(947),
    [anon_sym_GT] = ACTIONS(947),
    [anon_sym_is] = ACTIONS(947),
    [anon_sym_instanceof] = ACTIONS(947),
    [anon_sym_new] = ACTIONS(947),
    [anon_sym_QMARK] = ACTIONS(947),
    [anon_sym_await] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(947),
    [sym_this] = ACTIONS(947),
    [sym_true] = ACTIONS(947),
    [sym_false] = ACTIONS(947),
    [sym_none] = ACTIONS(947),
    [sym_string] = ACTIONS(947),
    [sym_verbatim] = ACTIONS(947),
    [sym_number] = ACTIONS(947),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(949),
    [sym_regex] = ACTIONS(949),
    [sym_fstring_start] = ACTIONS(949),
  },
  [STATE(222)] = {
    [sym_identifier] = ACTIONS(951),
    [anon_sym_SEMI] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(935),
    [anon_sym_return] = ACTIONS(951),
    [anon_sym_async] = ACTIONS(951),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_PLUS_EQ] = ACTIONS(935),
    [anon_sym_DASH_EQ] = ACTIONS(935),
    [anon_sym_STAR_EQ] = ACTIONS(935),
    [anon_sym_SLASH_EQ] = ACTIONS(935),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(935),
    [anon_sym_PERCENT_EQ] = ACTIONS(935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_LT_LT_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_AMP_EQ] = ACTIONS(935),
    [anon_sym_CARET_EQ] = ACTIONS(935),
    [anon_sym_PIPE_EQ] = ACTIONS(935),
    [anon_sym_COMMA] = ACTIONS(954),
    [anon_sym_LPAREN] = ACTIONS(951),
    [anon_sym_LBRACK] = ACTIONS(951),
    [anon_sym_del] = ACTIONS(951),
    [anon_sym_raise] = ACTIONS(951),
    [sym_pass_statement] = ACTIONS(951),
    [sym_break_statement] = ACTIONS(951),
    [sym_continue_statement] = ACTIONS(951),
    [sym_debugger_statement] = ACTIONS(951),
    [anon_sym_assert] = ACTIONS(951),
    [anon_sym_global] = ACTIONS(951),
    [anon_sym_nonlocal] = ACTIONS(951),
    [anon_sym_yield] = ACTIONS(951),
    [anon_sym_from] = ACTIONS(951),
    [anon_sym_import] = ACTIONS(951),
    [anon_sym_DOT] = ACTIONS(951),
    [anon_sym_if] = ACTIONS(951),
    [anon_sym_for] = ACTIONS(951),
    [anon_sym_in] = ACTIONS(951),
    [anon_sym_while] = ACTIONS(951),
    [anon_sym_do] = ACTIONS(951),
    [anon_sym_try] = ACTIONS(951),
    [anon_sym_with] = ACTIONS(951),
    [anon_sym_STAR] = ACTIONS(951),
    [anon_sym_STAR_STAR] = ACTIONS(951),
    [anon_sym_class] = ACTIONS(951),
    [anon_sym_AT] = ACTIONS(951),
    [anon_sym_not] = ACTIONS(951),
    [anon_sym_and] = ACTIONS(951),
    [anon_sym_or] = ACTIONS(951),
    [anon_sym_PLUS] = ACTIONS(951),
    [anon_sym_DASH] = ACTIONS(951),
    [anon_sym_SLASH] = ACTIONS(951),
    [anon_sym_SLASH_SLASH] = ACTIONS(951),
    [anon_sym_PERCENT] = ACTIONS(951),
    [anon_sym_PIPE] = ACTIONS(951),
    [anon_sym_CARET] = ACTIONS(951),
    [anon_sym_AMP] = ACTIONS(951),
    [anon_sym_LT_LT] = ACTIONS(951),
    [anon_sym_GT_GT] = ACTIONS(951),
    [anon_sym_GT_GT_GT] = ACTIONS(951),
    [anon_sym_TILDE] = ACTIONS(951),
    [anon_sym_typeof] = ACTIONS(951),
    [anon_sym_void] = ACTIONS(951),
    [anon_sym_LT] = ACTIONS(951),
    [anon_sym_LT_EQ] = ACTIONS(951),
    [anon_sym_EQ_EQ] = ACTIONS(951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(951),
    [anon_sym_BANG_EQ] = ACTIONS(951),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(951),
    [anon_sym_GT_EQ] = ACTIONS(951),
    [anon_sym_GT] = ACTIONS(951),
    [anon_sym_is] = ACTIONS(951),
    [anon_sym_instanceof] = ACTIONS(951),
    [anon_sym_new] = ACTIONS(951),
    [anon_sym_QMARK] = ACTIONS(951),
    [anon_sym_await] = ACTIONS(951),
    [anon_sym_LBRACE] = ACTIONS(951),
    [sym_this] = ACTIONS(951),
    [sym_true] = ACTIONS(951),
    [sym_false] = ACTIONS(951),
    [sym_none] = ACTIONS(951),
    [sym_string] = ACTIONS(951),
    [sym_verbatim] = ACTIONS(951),
    [sym_number] = ACTIONS(951),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(958),
    [sym_regex] = ACTIONS(958),
    [sym_fstring_start] = ACTIONS(958),
  },
  [STATE(223)] = {
    [ts_builtin_sym_end] = ACTIONS(958),
    [sym_identifier] = ACTIONS(951),
    [anon_sym_SEMI] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(935),
    [anon_sym_return] = ACTIONS(951),
    [anon_sym_async] = ACTIONS(951),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_PLUS_EQ] = ACTIONS(935),
    [anon_sym_DASH_EQ] = ACTIONS(935),
    [anon_sym_STAR_EQ] = ACTIONS(935),
    [anon_sym_SLASH_EQ] = ACTIONS(935),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(935),
    [anon_sym_PERCENT_EQ] = ACTIONS(935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_LT_LT_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_AMP_EQ] = ACTIONS(935),
    [anon_sym_CARET_EQ] = ACTIONS(935),
    [anon_sym_PIPE_EQ] = ACTIONS(935),
    [anon_sym_COMMA] = ACTIONS(954),
    [anon_sym_LPAREN] = ACTIONS(951),
    [anon_sym_LBRACK] = ACTIONS(951),
    [anon_sym_del] = ACTIONS(951),
    [anon_sym_raise] = ACTIONS(951),
    [sym_pass_statement] = ACTIONS(951),
    [sym_break_statement] = ACTIONS(951),
    [sym_continue_statement] = ACTIONS(951),
    [sym_debugger_statement] = ACTIONS(951),
    [anon_sym_assert] = ACTIONS(951),
    [anon_sym_global] = ACTIONS(951),
    [anon_sym_nonlocal] = ACTIONS(951),
    [anon_sym_yield] = ACTIONS(951),
    [anon_sym_from] = ACTIONS(951),
    [anon_sym_import] = ACTIONS(951),
    [anon_sym_DOT] = ACTIONS(951),
    [anon_sym_if] = ACTIONS(951),
    [anon_sym_for] = ACTIONS(951),
    [anon_sym_in] = ACTIONS(951),
    [anon_sym_while] = ACTIONS(951),
    [anon_sym_do] = ACTIONS(951),
    [anon_sym_try] = ACTIONS(951),
    [anon_sym_with] = ACTIONS(951),
    [anon_sym_STAR] = ACTIONS(951),
    [anon_sym_STAR_STAR] = ACTIONS(951),
    [anon_sym_class] = ACTIONS(951),
    [anon_sym_AT] = ACTIONS(951),
    [anon_sym_not] = ACTIONS(951),
    [anon_sym_and] = ACTIONS(951),
    [anon_sym_or] = ACTIONS(951),
    [anon_sym_PLUS] = ACTIONS(951),
    [anon_sym_DASH] = ACTIONS(951),
    [anon_sym_SLASH] = ACTIONS(951),
    [anon_sym_SLASH_SLASH] = ACTIONS(951),
    [anon_sym_PERCENT] = ACTIONS(951),
    [anon_sym_PIPE] = ACTIONS(951),
    [anon_sym_CARET] = ACTIONS(951),
    [anon_sym_AMP] = ACTIONS(951),
    [anon_sym_LT_LT] = ACTIONS(951),
    [anon_sym_GT_GT] = ACTIONS(951),
    [anon_sym_GT_GT_GT] = ACTIONS(951),
    [anon_sym_TILDE] = ACTIONS(951),
    [anon_sym_typeof] = ACTIONS(951),
    [anon_sym_void] = ACTIONS(951),
    [anon_sym_LT] = ACTIONS(951),
    [anon_sym_LT_EQ] = ACTIONS(951),
    [anon_sym_EQ_EQ] = ACTIONS(951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(951),
    [anon_sym_BANG_EQ] = ACTIONS(951),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(951),
    [anon_sym_GT_EQ] = ACTIONS(951),
    [anon_sym_GT] = ACTIONS(951),
    [anon_sym_is] = ACTIONS(951),
    [anon_sym_instanceof] = ACTIONS(951),
    [anon_sym_new] = ACTIONS(951),
    [anon_sym_QMARK] = ACTIONS(951),
    [anon_sym_await] = ACTIONS(951),
    [anon_sym_LBRACE] = ACTIONS(951),
    [sym_this] = ACTIONS(951),
    [sym_true] = ACTIONS(951),
    [sym_false] = ACTIONS(951),
    [sym_none] = ACTIONS(951),
    [sym_string] = ACTIONS(951),
    [sym_verbatim] = ACTIONS(951),
    [sym_number] = ACTIONS(951),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(958),
    [sym_fstring_start] = ACTIONS(958),
  },
  [STATE(224)] = {
    [sym_identifier] = ACTIONS(939),
    [anon_sym_SEMI] = ACTIONS(939),
    [anon_sym_EQ] = ACTIONS(941),
    [anon_sym_return] = ACTIONS(939),
    [anon_sym_async] = ACTIONS(939),
    [anon_sym_def] = ACTIONS(939),
    [anon_sym_PLUS_EQ] = ACTIONS(941),
    [anon_sym_DASH_EQ] = ACTIONS(941),
    [anon_sym_STAR_EQ] = ACTIONS(941),
    [anon_sym_SLASH_EQ] = ACTIONS(941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(941),
    [anon_sym_PERCENT_EQ] = ACTIONS(941),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(941),
    [anon_sym_GT_GT_EQ] = ACTIONS(941),
    [anon_sym_LT_LT_EQ] = ACTIONS(941),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(941),
    [anon_sym_AMP_EQ] = ACTIONS(941),
    [anon_sym_CARET_EQ] = ACTIONS(941),
    [anon_sym_PIPE_EQ] = ACTIONS(941),
    [anon_sym_COMMA] = ACTIONS(941),
    [anon_sym_LPAREN] = ACTIONS(939),
    [anon_sym_LBRACK] = ACTIONS(939),
    [anon_sym_del] = ACTIONS(939),
    [anon_sym_raise] = ACTIONS(939),
    [sym_pass_statement] = ACTIONS(939),
    [sym_break_statement] = ACTIONS(939),
    [sym_continue_statement] = ACTIONS(939),
    [sym_debugger_statement] = ACTIONS(939),
    [anon_sym_assert] = ACTIONS(939),
    [anon_sym_global] = ACTIONS(939),
    [anon_sym_nonlocal] = ACTIONS(939),
    [anon_sym_yield] = ACTIONS(939),
    [anon_sym_from] = ACTIONS(939),
    [anon_sym_import] = ACTIONS(939),
    [anon_sym_DOT] = ACTIONS(939),
    [anon_sym_if] = ACTIONS(939),
    [anon_sym_for] = ACTIONS(939),
    [anon_sym_in] = ACTIONS(939),
    [anon_sym_while] = ACTIONS(939),
    [anon_sym_do] = ACTIONS(939),
    [anon_sym_try] = ACTIONS(939),
    [anon_sym_with] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(939),
    [anon_sym_STAR_STAR] = ACTIONS(939),
    [anon_sym_class] = ACTIONS(939),
    [anon_sym_AT] = ACTIONS(939),
    [anon_sym_not] = ACTIONS(939),
    [anon_sym_and] = ACTIONS(939),
    [anon_sym_or] = ACTIONS(939),
    [anon_sym_PLUS] = ACTIONS(939),
    [anon_sym_DASH] = ACTIONS(939),
    [anon_sym_SLASH] = ACTIONS(939),
    [anon_sym_SLASH_SLASH] = ACTIONS(939),
    [anon_sym_PERCENT] = ACTIONS(939),
    [anon_sym_PIPE] = ACTIONS(939),
    [anon_sym_CARET] = ACTIONS(939),
    [anon_sym_AMP] = ACTIONS(939),
    [anon_sym_LT_LT] = ACTIONS(939),
    [anon_sym_GT_GT] = ACTIONS(939),
    [anon_sym_GT_GT_GT] = ACTIONS(939),
    [anon_sym_TILDE] = ACTIONS(939),
    [anon_sym_typeof] = ACTIONS(939),
    [anon_sym_void] = ACTIONS(939),
    [anon_sym_LT] = ACTIONS(939),
    [anon_sym_LT_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(939),
    [anon_sym_GT_EQ] = ACTIONS(939),
    [anon_sym_GT] = ACTIONS(939),
    [anon_sym_is] = ACTIONS(939),
    [anon_sym_instanceof] = ACTIONS(939),
    [anon_sym_new] = ACTIONS(939),
    [anon_sym_QMARK] = ACTIONS(939),
    [anon_sym_await] = ACTIONS(939),
    [anon_sym_LBRACE] = ACTIONS(939),
    [sym_this] = ACTIONS(939),
    [sym_true] = ACTIONS(939),
    [sym_false] = ACTIONS(939),
    [sym_none] = ACTIONS(939),
    [sym_string] = ACTIONS(939),
    [sym_verbatim] = ACTIONS(939),
    [sym_number] = ACTIONS(939),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(937),
    [sym_regex] = ACTIONS(937),
    [sym_fstring_start] = ACTIONS(937),
  },
  [STATE(225)] = {
    [ts_builtin_sym_end] = ACTIONS(949),
    [sym_identifier] = ACTIONS(947),
    [anon_sym_SEMI] = ACTIONS(947),
    [anon_sym_EQ] = ACTIONS(947),
    [anon_sym_return] = ACTIONS(947),
    [anon_sym_async] = ACTIONS(947),
    [anon_sym_def] = ACTIONS(947),
    [anon_sym_PLUS_EQ] = ACTIONS(947),
    [anon_sym_DASH_EQ] = ACTIONS(947),
    [anon_sym_STAR_EQ] = ACTIONS(947),
    [anon_sym_SLASH_EQ] = ACTIONS(947),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(947),
    [anon_sym_PERCENT_EQ] = ACTIONS(947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(947),
    [anon_sym_GT_GT_EQ] = ACTIONS(947),
    [anon_sym_LT_LT_EQ] = ACTIONS(947),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(947),
    [anon_sym_AMP_EQ] = ACTIONS(947),
    [anon_sym_CARET_EQ] = ACTIONS(947),
    [anon_sym_PIPE_EQ] = ACTIONS(947),
    [anon_sym_COMMA] = ACTIONS(947),
    [anon_sym_LPAREN] = ACTIONS(947),
    [anon_sym_LBRACK] = ACTIONS(947),
    [anon_sym_del] = ACTIONS(947),
    [anon_sym_raise] = ACTIONS(947),
    [sym_pass_statement] = ACTIONS(947),
    [sym_break_statement] = ACTIONS(947),
    [sym_continue_statement] = ACTIONS(947),
    [sym_debugger_statement] = ACTIONS(947),
    [anon_sym_assert] = ACTIONS(947),
    [anon_sym_global] = ACTIONS(947),
    [anon_sym_nonlocal] = ACTIONS(947),
    [anon_sym_yield] = ACTIONS(947),
    [anon_sym_from] = ACTIONS(947),
    [anon_sym_import] = ACTIONS(947),
    [anon_sym_DOT] = ACTIONS(947),
    [anon_sym_if] = ACTIONS(947),
    [anon_sym_for] = ACTIONS(947),
    [anon_sym_in] = ACTIONS(947),
    [anon_sym_while] = ACTIONS(947),
    [anon_sym_do] = ACTIONS(947),
    [anon_sym_try] = ACTIONS(947),
    [anon_sym_with] = ACTIONS(947),
    [anon_sym_STAR] = ACTIONS(947),
    [anon_sym_STAR_STAR] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_AT] = ACTIONS(947),
    [anon_sym_not] = ACTIONS(947),
    [anon_sym_and] = ACTIONS(947),
    [anon_sym_or] = ACTIONS(947),
    [anon_sym_PLUS] = ACTIONS(947),
    [anon_sym_DASH] = ACTIONS(947),
    [anon_sym_SLASH] = ACTIONS(947),
    [anon_sym_SLASH_SLASH] = ACTIONS(947),
    [anon_sym_PERCENT] = ACTIONS(947),
    [anon_sym_PIPE] = ACTIONS(947),
    [anon_sym_CARET] = ACTIONS(947),
    [anon_sym_AMP] = ACTIONS(947),
    [anon_sym_LT_LT] = ACTIONS(947),
    [anon_sym_GT_GT] = ACTIONS(947),
    [anon_sym_GT_GT_GT] = ACTIONS(947),
    [anon_sym_TILDE] = ACTIONS(947),
    [anon_sym_typeof] = ACTIONS(947),
    [anon_sym_void] = ACTIONS(947),
    [anon_sym_LT] = ACTIONS(947),
    [anon_sym_LT_EQ] = ACTIONS(947),
    [anon_sym_EQ_EQ] = ACTIONS(947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(947),
    [anon_sym_BANG_EQ] = ACTIONS(947),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(947),
    [anon_sym_GT_EQ] = ACTIONS(947),
    [anon_sym_GT] = ACTIONS(947),
    [anon_sym_is] = ACTIONS(947),
    [anon_sym_instanceof] = ACTIONS(947),
    [anon_sym_new] = ACTIONS(947),
    [anon_sym_QMARK] = ACTIONS(947),
    [anon_sym_await] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(947),
    [sym_this] = ACTIONS(947),
    [sym_true] = ACTIONS(947),
    [sym_false] = ACTIONS(947),
    [sym_none] = ACTIONS(947),
    [sym_string] = ACTIONS(947),
    [sym_verbatim] = ACTIONS(947),
    [sym_number] = ACTIONS(947),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(949),
    [sym_fstring_start] = ACTIONS(949),
  },
  [STATE(226)] = {
    [sym_identifier] = ACTIONS(933),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym_EQ] = ACTIONS(935),
    [anon_sym_return] = ACTIONS(933),
    [anon_sym_async] = ACTIONS(933),
    [anon_sym_def] = ACTIONS(933),
    [anon_sym_PLUS_EQ] = ACTIONS(935),
    [anon_sym_DASH_EQ] = ACTIONS(935),
    [anon_sym_STAR_EQ] = ACTIONS(935),
    [anon_sym_SLASH_EQ] = ACTIONS(935),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(935),
    [anon_sym_PERCENT_EQ] = ACTIONS(935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_LT_LT_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_AMP_EQ] = ACTIONS(935),
    [anon_sym_CARET_EQ] = ACTIONS(935),
    [anon_sym_PIPE_EQ] = ACTIONS(935),
    [anon_sym_COMMA] = ACTIONS(935),
    [anon_sym_LPAREN] = ACTIONS(933),
    [anon_sym_LBRACK] = ACTIONS(933),
    [anon_sym_del] = ACTIONS(933),
    [anon_sym_raise] = ACTIONS(933),
    [sym_pass_statement] = ACTIONS(933),
    [sym_break_statement] = ACTIONS(933),
    [sym_continue_statement] = ACTIONS(933),
    [sym_debugger_statement] = ACTIONS(933),
    [anon_sym_assert] = ACTIONS(933),
    [anon_sym_global] = ACTIONS(933),
    [anon_sym_nonlocal] = ACTIONS(933),
    [anon_sym_yield] = ACTIONS(933),
    [anon_sym_from] = ACTIONS(933),
    [anon_sym_import] = ACTIONS(933),
    [anon_sym_DOT] = ACTIONS(933),
    [anon_sym_if] = ACTIONS(933),
    [anon_sym_for] = ACTIONS(933),
    [anon_sym_in] = ACTIONS(933),
    [anon_sym_while] = ACTIONS(933),
    [anon_sym_do] = ACTIONS(933),
    [anon_sym_try] = ACTIONS(933),
    [anon_sym_with] = ACTIONS(933),
    [anon_sym_STAR] = ACTIONS(933),
    [anon_sym_STAR_STAR] = ACTIONS(933),
    [anon_sym_class] = ACTIONS(933),
    [anon_sym_AT] = ACTIONS(933),
    [anon_sym_not] = ACTIONS(933),
    [anon_sym_and] = ACTIONS(933),
    [anon_sym_or] = ACTIONS(933),
    [anon_sym_PLUS] = ACTIONS(933),
    [anon_sym_DASH] = ACTIONS(933),
    [anon_sym_SLASH] = ACTIONS(933),
    [anon_sym_SLASH_SLASH] = ACTIONS(933),
    [anon_sym_PERCENT] = ACTIONS(933),
    [anon_sym_PIPE] = ACTIONS(933),
    [anon_sym_CARET] = ACTIONS(933),
    [anon_sym_AMP] = ACTIONS(933),
    [anon_sym_LT_LT] = ACTIONS(933),
    [anon_sym_GT_GT] = ACTIONS(933),
    [anon_sym_GT_GT_GT] = ACTIONS(933),
    [anon_sym_TILDE] = ACTIONS(933),
    [anon_sym_typeof] = ACTIONS(933),
    [anon_sym_void] = ACTIONS(933),
    [anon_sym_LT] = ACTIONS(933),
    [anon_sym_LT_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(933),
    [anon_sym_GT_EQ] = ACTIONS(933),
    [anon_sym_GT] = ACTIONS(933),
    [anon_sym_is] = ACTIONS(933),
    [anon_sym_instanceof] = ACTIONS(933),
    [anon_sym_new] = ACTIONS(933),
    [anon_sym_QMARK] = ACTIONS(933),
    [anon_sym_await] = ACTIONS(933),
    [anon_sym_LBRACE] = ACTIONS(933),
    [sym_this] = ACTIONS(933),
    [sym_true] = ACTIONS(933),
    [sym_false] = ACTIONS(933),
    [sym_none] = ACTIONS(933),
    [sym_string] = ACTIONS(933),
    [sym_verbatim] = ACTIONS(933),
    [sym_number] = ACTIONS(933),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(931),
    [sym_regex] = ACTIONS(931),
    [sym_fstring_start] = ACTIONS(931),
  },
  [STATE(227)] = {
    [ts_builtin_sym_end] = ACTIONS(949),
    [sym_identifier] = ACTIONS(947),
    [anon_sym_SEMI] = ACTIONS(947),
    [anon_sym_EQ] = ACTIONS(947),
    [anon_sym_return] = ACTIONS(947),
    [anon_sym_async] = ACTIONS(947),
    [anon_sym_def] = ACTIONS(947),
    [anon_sym_PLUS_EQ] = ACTIONS(947),
    [anon_sym_DASH_EQ] = ACTIONS(947),
    [anon_sym_STAR_EQ] = ACTIONS(947),
    [anon_sym_SLASH_EQ] = ACTIONS(947),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(947),
    [anon_sym_PERCENT_EQ] = ACTIONS(947),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(947),
    [anon_sym_GT_GT_EQ] = ACTIONS(947),
    [anon_sym_LT_LT_EQ] = ACTIONS(947),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(947),
    [anon_sym_AMP_EQ] = ACTIONS(947),
    [anon_sym_CARET_EQ] = ACTIONS(947),
    [anon_sym_PIPE_EQ] = ACTIONS(947),
    [anon_sym_COMMA] = ACTIONS(947),
    [anon_sym_LPAREN] = ACTIONS(947),
    [anon_sym_LBRACK] = ACTIONS(947),
    [anon_sym_del] = ACTIONS(947),
    [anon_sym_raise] = ACTIONS(947),
    [sym_pass_statement] = ACTIONS(947),
    [sym_break_statement] = ACTIONS(947),
    [sym_continue_statement] = ACTIONS(947),
    [sym_debugger_statement] = ACTIONS(947),
    [anon_sym_assert] = ACTIONS(947),
    [anon_sym_global] = ACTIONS(947),
    [anon_sym_nonlocal] = ACTIONS(947),
    [anon_sym_yield] = ACTIONS(947),
    [anon_sym_from] = ACTIONS(947),
    [anon_sym_import] = ACTIONS(947),
    [anon_sym_DOT] = ACTIONS(947),
    [anon_sym_if] = ACTIONS(947),
    [anon_sym_for] = ACTIONS(947),
    [anon_sym_in] = ACTIONS(947),
    [anon_sym_while] = ACTIONS(947),
    [anon_sym_do] = ACTIONS(947),
    [anon_sym_try] = ACTIONS(947),
    [anon_sym_with] = ACTIONS(947),
    [anon_sym_STAR] = ACTIONS(947),
    [anon_sym_STAR_STAR] = ACTIONS(947),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_AT] = ACTIONS(947),
    [anon_sym_not] = ACTIONS(947),
    [anon_sym_and] = ACTIONS(947),
    [anon_sym_or] = ACTIONS(947),
    [anon_sym_PLUS] = ACTIONS(947),
    [anon_sym_DASH] = ACTIONS(947),
    [anon_sym_SLASH] = ACTIONS(947),
    [anon_sym_SLASH_SLASH] = ACTIONS(947),
    [anon_sym_PERCENT] = ACTIONS(947),
    [anon_sym_PIPE] = ACTIONS(947),
    [anon_sym_CARET] = ACTIONS(947),
    [anon_sym_AMP] = ACTIONS(947),
    [anon_sym_LT_LT] = ACTIONS(947),
    [anon_sym_GT_GT] = ACTIONS(947),
    [anon_sym_GT_GT_GT] = ACTIONS(947),
    [anon_sym_TILDE] = ACTIONS(947),
    [anon_sym_typeof] = ACTIONS(947),
    [anon_sym_void] = ACTIONS(947),
    [anon_sym_LT] = ACTIONS(947),
    [anon_sym_LT_EQ] = ACTIONS(947),
    [anon_sym_EQ_EQ] = ACTIONS(947),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(947),
    [anon_sym_BANG_EQ] = ACTIONS(947),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(947),
    [anon_sym_GT_EQ] = ACTIONS(947),
    [anon_sym_GT] = ACTIONS(947),
    [anon_sym_is] = ACTIONS(947),
    [anon_sym_instanceof] = ACTIONS(947),
    [anon_sym_new] = ACTIONS(947),
    [anon_sym_QMARK] = ACTIONS(947),
    [anon_sym_await] = ACTIONS(947),
    [anon_sym_LBRACE] = ACTIONS(947),
    [sym_this] = ACTIONS(947),
    [sym_true] = ACTIONS(947),
    [sym_false] = ACTIONS(947),
    [sym_none] = ACTIONS(947),
    [sym_string] = ACTIONS(947),
    [sym_verbatim] = ACTIONS(947),
    [sym_number] = ACTIONS(947),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(949),
    [sym_fstring_start] = ACTIONS(949),
  },
  [STATE(228)] = {
    [sym_identifier] = ACTIONS(951),
    [anon_sym_SEMI] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(935),
    [anon_sym_return] = ACTIONS(951),
    [anon_sym_async] = ACTIONS(951),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_PLUS_EQ] = ACTIONS(935),
    [anon_sym_DASH_EQ] = ACTIONS(935),
    [anon_sym_STAR_EQ] = ACTIONS(935),
    [anon_sym_SLASH_EQ] = ACTIONS(935),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(935),
    [anon_sym_PERCENT_EQ] = ACTIONS(935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_LT_LT_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_AMP_EQ] = ACTIONS(935),
    [anon_sym_CARET_EQ] = ACTIONS(935),
    [anon_sym_PIPE_EQ] = ACTIONS(935),
    [anon_sym_COMMA] = ACTIONS(935),
    [anon_sym_LPAREN] = ACTIONS(951),
    [anon_sym_LBRACK] = ACTIONS(951),
    [anon_sym_del] = ACTIONS(951),
    [anon_sym_raise] = ACTIONS(951),
    [sym_pass_statement] = ACTIONS(951),
    [sym_break_statement] = ACTIONS(951),
    [sym_continue_statement] = ACTIONS(951),
    [sym_debugger_statement] = ACTIONS(951),
    [anon_sym_assert] = ACTIONS(951),
    [anon_sym_global] = ACTIONS(951),
    [anon_sym_nonlocal] = ACTIONS(951),
    [anon_sym_yield] = ACTIONS(951),
    [anon_sym_from] = ACTIONS(951),
    [anon_sym_import] = ACTIONS(951),
    [anon_sym_DOT] = ACTIONS(951),
    [anon_sym_if] = ACTIONS(951),
    [anon_sym_for] = ACTIONS(951),
    [anon_sym_in] = ACTIONS(951),
    [anon_sym_while] = ACTIONS(951),
    [anon_sym_do] = ACTIONS(951),
    [anon_sym_try] = ACTIONS(951),
    [anon_sym_with] = ACTIONS(951),
    [anon_sym_STAR] = ACTIONS(951),
    [anon_sym_STAR_STAR] = ACTIONS(951),
    [anon_sym_class] = ACTIONS(951),
    [anon_sym_AT] = ACTIONS(951),
    [anon_sym_not] = ACTIONS(951),
    [anon_sym_and] = ACTIONS(951),
    [anon_sym_or] = ACTIONS(951),
    [anon_sym_PLUS] = ACTIONS(951),
    [anon_sym_DASH] = ACTIONS(951),
    [anon_sym_SLASH] = ACTIONS(951),
    [anon_sym_SLASH_SLASH] = ACTIONS(951),
    [anon_sym_PERCENT] = ACTIONS(951),
    [anon_sym_PIPE] = ACTIONS(951),
    [anon_sym_CARET] = ACTIONS(951),
    [anon_sym_AMP] = ACTIONS(951),
    [anon_sym_LT_LT] = ACTIONS(951),
    [anon_sym_GT_GT] = ACTIONS(951),
    [anon_sym_GT_GT_GT] = ACTIONS(951),
    [anon_sym_TILDE] = ACTIONS(951),
    [anon_sym_typeof] = ACTIONS(951),
    [anon_sym_void] = ACTIONS(951),
    [anon_sym_LT] = ACTIONS(951),
    [anon_sym_LT_EQ] = ACTIONS(951),
    [anon_sym_EQ_EQ] = ACTIONS(951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(951),
    [anon_sym_BANG_EQ] = ACTIONS(951),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(951),
    [anon_sym_GT_EQ] = ACTIONS(951),
    [anon_sym_GT] = ACTIONS(951),
    [anon_sym_is] = ACTIONS(951),
    [anon_sym_instanceof] = ACTIONS(951),
    [anon_sym_new] = ACTIONS(951),
    [anon_sym_QMARK] = ACTIONS(951),
    [anon_sym_await] = ACTIONS(951),
    [anon_sym_LBRACE] = ACTIONS(951),
    [sym_this] = ACTIONS(951),
    [sym_true] = ACTIONS(951),
    [sym_false] = ACTIONS(951),
    [sym_none] = ACTIONS(951),
    [sym_string] = ACTIONS(951),
    [sym_verbatim] = ACTIONS(951),
    [sym_number] = ACTIONS(951),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(958),
    [sym_regex] = ACTIONS(958),
    [sym_fstring_start] = ACTIONS(958),
  },
  [STATE(229)] = {
    [ts_builtin_sym_end] = ACTIONS(958),
    [sym_identifier] = ACTIONS(951),
    [anon_sym_SEMI] = ACTIONS(951),
    [anon_sym_EQ] = ACTIONS(935),
    [anon_sym_return] = ACTIONS(951),
    [anon_sym_async] = ACTIONS(951),
    [anon_sym_def] = ACTIONS(951),
    [anon_sym_PLUS_EQ] = ACTIONS(935),
    [anon_sym_DASH_EQ] = ACTIONS(935),
    [anon_sym_STAR_EQ] = ACTIONS(935),
    [anon_sym_SLASH_EQ] = ACTIONS(935),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(935),
    [anon_sym_PERCENT_EQ] = ACTIONS(935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_LT_LT_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_AMP_EQ] = ACTIONS(935),
    [anon_sym_CARET_EQ] = ACTIONS(935),
    [anon_sym_PIPE_EQ] = ACTIONS(935),
    [anon_sym_COMMA] = ACTIONS(935),
    [anon_sym_LPAREN] = ACTIONS(951),
    [anon_sym_LBRACK] = ACTIONS(951),
    [anon_sym_del] = ACTIONS(951),
    [anon_sym_raise] = ACTIONS(951),
    [sym_pass_statement] = ACTIONS(951),
    [sym_break_statement] = ACTIONS(951),
    [sym_continue_statement] = ACTIONS(951),
    [sym_debugger_statement] = ACTIONS(951),
    [anon_sym_assert] = ACTIONS(951),
    [anon_sym_global] = ACTIONS(951),
    [anon_sym_nonlocal] = ACTIONS(951),
    [anon_sym_yield] = ACTIONS(951),
    [anon_sym_from] = ACTIONS(951),
    [anon_sym_import] = ACTIONS(951),
    [anon_sym_DOT] = ACTIONS(951),
    [anon_sym_if] = ACTIONS(951),
    [anon_sym_for] = ACTIONS(951),
    [anon_sym_in] = ACTIONS(951),
    [anon_sym_while] = ACTIONS(951),
    [anon_sym_do] = ACTIONS(951),
    [anon_sym_try] = ACTIONS(951),
    [anon_sym_with] = ACTIONS(951),
    [anon_sym_STAR] = ACTIONS(951),
    [anon_sym_STAR_STAR] = ACTIONS(951),
    [anon_sym_class] = ACTIONS(951),
    [anon_sym_AT] = ACTIONS(951),
    [anon_sym_not] = ACTIONS(951),
    [anon_sym_and] = ACTIONS(951),
    [anon_sym_or] = ACTIONS(951),
    [anon_sym_PLUS] = ACTIONS(951),
    [anon_sym_DASH] = ACTIONS(951),
    [anon_sym_SLASH] = ACTIONS(951),
    [anon_sym_SLASH_SLASH] = ACTIONS(951),
    [anon_sym_PERCENT] = ACTIONS(951),
    [anon_sym_PIPE] = ACTIONS(951),
    [anon_sym_CARET] = ACTIONS(951),
    [anon_sym_AMP] = ACTIONS(951),
    [anon_sym_LT_LT] = ACTIONS(951),
    [anon_sym_GT_GT] = ACTIONS(951),
    [anon_sym_GT_GT_GT] = ACTIONS(951),
    [anon_sym_TILDE] = ACTIONS(951),
    [anon_sym_typeof] = ACTIONS(951),
    [anon_sym_void] = ACTIONS(951),
    [anon_sym_LT] = ACTIONS(951),
    [anon_sym_LT_EQ] = ACTIONS(951),
    [anon_sym_EQ_EQ] = ACTIONS(951),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(951),
    [anon_sym_BANG_EQ] = ACTIONS(951),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(951),
    [anon_sym_GT_EQ] = ACTIONS(951),
    [anon_sym_GT] = ACTIONS(951),
    [anon_sym_is] = ACTIONS(951),
    [anon_sym_instanceof] = ACTIONS(951),
    [anon_sym_new] = ACTIONS(951),
    [anon_sym_QMARK] = ACTIONS(951),
    [anon_sym_await] = ACTIONS(951),
    [anon_sym_LBRACE] = ACTIONS(951),
    [sym_this] = ACTIONS(951),
    [sym_true] = ACTIONS(951),
    [sym_false] = ACTIONS(951),
    [sym_none] = ACTIONS(951),
    [sym_string] = ACTIONS(951),
    [sym_verbatim] = ACTIONS(951),
    [sym_number] = ACTIONS(951),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(958),
    [sym_fstring_start] = ACTIONS(958),
  },
  [STATE(230)] = {
    [ts_builtin_sym_end] = ACTIONS(945),
    [sym_identifier] = ACTIONS(943),
    [anon_sym_SEMI] = ACTIONS(943),
    [anon_sym_EQ] = ACTIONS(943),
    [anon_sym_return] = ACTIONS(943),
    [anon_sym_async] = ACTIONS(943),
    [anon_sym_def] = ACTIONS(943),
    [anon_sym_PLUS_EQ] = ACTIONS(943),
    [anon_sym_DASH_EQ] = ACTIONS(943),
    [anon_sym_STAR_EQ] = ACTIONS(943),
    [anon_sym_SLASH_EQ] = ACTIONS(943),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(943),
    [anon_sym_PERCENT_EQ] = ACTIONS(943),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(943),
    [anon_sym_GT_GT_EQ] = ACTIONS(943),
    [anon_sym_LT_LT_EQ] = ACTIONS(943),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(943),
    [anon_sym_AMP_EQ] = ACTIONS(943),
    [anon_sym_CARET_EQ] = ACTIONS(943),
    [anon_sym_PIPE_EQ] = ACTIONS(943),
    [anon_sym_COMMA] = ACTIONS(943),
    [anon_sym_LPAREN] = ACTIONS(943),
    [anon_sym_LBRACK] = ACTIONS(943),
    [anon_sym_del] = ACTIONS(943),
    [anon_sym_raise] = ACTIONS(943),
    [sym_pass_statement] = ACTIONS(943),
    [sym_break_statement] = ACTIONS(943),
    [sym_continue_statement] = ACTIONS(943),
    [sym_debugger_statement] = ACTIONS(943),
    [anon_sym_assert] = ACTIONS(943),
    [anon_sym_global] = ACTIONS(943),
    [anon_sym_nonlocal] = ACTIONS(943),
    [anon_sym_yield] = ACTIONS(943),
    [anon_sym_from] = ACTIONS(943),
    [anon_sym_import] = ACTIONS(943),
    [anon_sym_DOT] = ACTIONS(943),
    [anon_sym_if] = ACTIONS(943),
    [anon_sym_for] = ACTIONS(943),
    [anon_sym_in] = ACTIONS(943),
    [anon_sym_while] = ACTIONS(943),
    [anon_sym_do] = ACTIONS(943),
    [anon_sym_try] = ACTIONS(943),
    [anon_sym_with] = ACTIONS(943),
    [anon_sym_STAR] = ACTIONS(943),
    [anon_sym_STAR_STAR] = ACTIONS(943),
    [anon_sym_class] = ACTIONS(943),
    [anon_sym_AT] = ACTIONS(943),
    [anon_sym_not] = ACTIONS(943),
    [anon_sym_and] = ACTIONS(943),
    [anon_sym_or] = ACTIONS(943),
    [anon_sym_PLUS] = ACTIONS(943),
    [anon_sym_DASH] = ACTIONS(943),
    [anon_sym_SLASH] = ACTIONS(943),
    [anon_sym_SLASH_SLASH] = ACTIONS(943),
    [anon_sym_PERCENT] = ACTIONS(943),
    [anon_sym_PIPE] = ACTIONS(943),
    [anon_sym_CARET] = ACTIONS(943),
    [anon_sym_AMP] = ACTIONS(943),
    [anon_sym_LT_LT] = ACTIONS(943),
    [anon_sym_GT_GT] = ACTIONS(943),
    [anon_sym_GT_GT_GT] = ACTIONS(943),
    [anon_sym_TILDE] = ACTIONS(943),
    [anon_sym_typeof] = ACTIONS(943),
    [anon_sym_void] = ACTIONS(943),
    [anon_sym_LT] = ACTIONS(943),
    [anon_sym_LT_EQ] = ACTIONS(943),
    [anon_sym_EQ_EQ] = ACTIONS(943),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(943),
    [anon_sym_BANG_EQ] = ACTIONS(943),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(943),
    [anon_sym_GT_EQ] = ACTIONS(943),
    [anon_sym_GT] = ACTIONS(943),
    [anon_sym_is] = ACTIONS(943),
    [anon_sym_instanceof] = ACTIONS(943),
    [anon_sym_new] = ACTIONS(943),
    [anon_sym_QMARK] = ACTIONS(943),
    [anon_sym_await] = ACTIONS(943),
    [anon_sym_LBRACE] = ACTIONS(943),
    [sym_this] = ACTIONS(943),
    [sym_true] = ACTIONS(943),
    [sym_false] = ACTIONS(943),
    [sym_none] = ACTIONS(943),
    [sym_string] = ACTIONS(943),
    [sym_verbatim] = ACTIONS(943),
    [sym_number] = ACTIONS(943),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(945),
    [sym_fstring_start] = ACTIONS(945),
  },
  [STATE(231)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2528),
    [sym_pattern_list] = STATE(2528),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2211),
    [sym__primary_expression] = STATE(2193),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(913),
    [anon_sym_def] = ACTIONS(915),
    [anon_sym_LPAREN] = ACTIONS(797),
    [anon_sym_RPAREN] = ACTIONS(449),
    [anon_sym_LBRACK] = ACTIONS(799),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(917),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(919),
    [anon_sym_DASH] = ACTIONS(919),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(919),
    [anon_sym_typeof] = ACTIONS(919),
    [anon_sym_void] = ACTIONS(919),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(921),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(923),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(232)] = {
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2539),
    [sym_pattern_list] = STATE(2539),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2190),
    [sym__primary_expression] = STATE(2165),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1517),
    [sym_subscript] = STATE(1517),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(895),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(897),
    [anon_sym_def] = ACTIONS(899),
    [anon_sym_LPAREN] = ACTIONS(961),
    [anon_sym_LBRACK] = ACTIONS(963),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(905),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(907),
    [anon_sym_DASH] = ACTIONS(907),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(907),
    [anon_sym_typeof] = ACTIONS(907),
    [anon_sym_void] = ACTIONS(907),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(909),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(911),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(895),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(451),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(233)] = {
    [sym_identifier] = ACTIONS(939),
    [anon_sym_SEMI] = ACTIONS(939),
    [anon_sym_EQ] = ACTIONS(941),
    [anon_sym_return] = ACTIONS(939),
    [anon_sym_async] = ACTIONS(939),
    [anon_sym_def] = ACTIONS(939),
    [anon_sym_PLUS_EQ] = ACTIONS(941),
    [anon_sym_DASH_EQ] = ACTIONS(941),
    [anon_sym_STAR_EQ] = ACTIONS(941),
    [anon_sym_SLASH_EQ] = ACTIONS(941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(941),
    [anon_sym_PERCENT_EQ] = ACTIONS(941),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(941),
    [anon_sym_GT_GT_EQ] = ACTIONS(941),
    [anon_sym_LT_LT_EQ] = ACTIONS(941),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(941),
    [anon_sym_AMP_EQ] = ACTIONS(941),
    [anon_sym_CARET_EQ] = ACTIONS(941),
    [anon_sym_PIPE_EQ] = ACTIONS(941),
    [anon_sym_COMMA] = ACTIONS(965),
    [anon_sym_LPAREN] = ACTIONS(939),
    [anon_sym_LBRACK] = ACTIONS(939),
    [anon_sym_del] = ACTIONS(939),
    [anon_sym_raise] = ACTIONS(939),
    [sym_pass_statement] = ACTIONS(939),
    [sym_break_statement] = ACTIONS(939),
    [sym_continue_statement] = ACTIONS(939),
    [sym_debugger_statement] = ACTIONS(939),
    [anon_sym_assert] = ACTIONS(939),
    [anon_sym_global] = ACTIONS(939),
    [anon_sym_nonlocal] = ACTIONS(939),
    [anon_sym_yield] = ACTIONS(939),
    [anon_sym_from] = ACTIONS(939),
    [anon_sym_import] = ACTIONS(939),
    [anon_sym_DOT] = ACTIONS(939),
    [anon_sym_if] = ACTIONS(939),
    [anon_sym_for] = ACTIONS(939),
    [anon_sym_in] = ACTIONS(939),
    [anon_sym_while] = ACTIONS(939),
    [anon_sym_do] = ACTIONS(939),
    [anon_sym_try] = ACTIONS(939),
    [anon_sym_with] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(939),
    [anon_sym_STAR_STAR] = ACTIONS(939),
    [anon_sym_class] = ACTIONS(939),
    [anon_sym_AT] = ACTIONS(939),
    [anon_sym_not] = ACTIONS(939),
    [anon_sym_and] = ACTIONS(939),
    [anon_sym_or] = ACTIONS(939),
    [anon_sym_PLUS] = ACTIONS(939),
    [anon_sym_DASH] = ACTIONS(939),
    [anon_sym_SLASH] = ACTIONS(939),
    [anon_sym_SLASH_SLASH] = ACTIONS(939),
    [anon_sym_PERCENT] = ACTIONS(939),
    [anon_sym_PIPE] = ACTIONS(939),
    [anon_sym_CARET] = ACTIONS(939),
    [anon_sym_AMP] = ACTIONS(939),
    [anon_sym_LT_LT] = ACTIONS(939),
    [anon_sym_GT_GT] = ACTIONS(939),
    [anon_sym_GT_GT_GT] = ACTIONS(939),
    [anon_sym_TILDE] = ACTIONS(939),
    [anon_sym_typeof] = ACTIONS(939),
    [anon_sym_void] = ACTIONS(939),
    [anon_sym_LT] = ACTIONS(939),
    [anon_sym_LT_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(939),
    [anon_sym_GT_EQ] = ACTIONS(939),
    [anon_sym_GT] = ACTIONS(939),
    [anon_sym_is] = ACTIONS(939),
    [anon_sym_instanceof] = ACTIONS(939),
    [anon_sym_new] = ACTIONS(939),
    [anon_sym_QMARK] = ACTIONS(939),
    [anon_sym_await] = ACTIONS(939),
    [anon_sym_LBRACE] = ACTIONS(939),
    [sym_this] = ACTIONS(939),
    [sym_true] = ACTIONS(939),
    [sym_false] = ACTIONS(939),
    [sym_none] = ACTIONS(939),
    [sym_string] = ACTIONS(939),
    [sym_verbatim] = ACTIONS(939),
    [sym_number] = ACTIONS(939),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(937),
    [sym_regex] = ACTIONS(937),
    [sym_fstring_start] = ACTIONS(937),
  },
  [STATE(234)] = {
    [ts_builtin_sym_end] = ACTIONS(925),
    [sym_identifier] = ACTIONS(927),
    [anon_sym_SEMI] = ACTIONS(927),
    [anon_sym_EQ] = ACTIONS(929),
    [anon_sym_return] = ACTIONS(927),
    [anon_sym_async] = ACTIONS(927),
    [anon_sym_def] = ACTIONS(927),
    [anon_sym_PLUS_EQ] = ACTIONS(929),
    [anon_sym_DASH_EQ] = ACTIONS(929),
    [anon_sym_STAR_EQ] = ACTIONS(929),
    [anon_sym_SLASH_EQ] = ACTIONS(929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(929),
    [anon_sym_PERCENT_EQ] = ACTIONS(929),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(929),
    [anon_sym_GT_GT_EQ] = ACTIONS(929),
    [anon_sym_LT_LT_EQ] = ACTIONS(929),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(929),
    [anon_sym_AMP_EQ] = ACTIONS(929),
    [anon_sym_CARET_EQ] = ACTIONS(929),
    [anon_sym_PIPE_EQ] = ACTIONS(929),
    [anon_sym_COMMA] = ACTIONS(968),
    [anon_sym_LPAREN] = ACTIONS(927),
    [anon_sym_LBRACK] = ACTIONS(927),
    [anon_sym_del] = ACTIONS(927),
    [anon_sym_raise] = ACTIONS(927),
    [sym_pass_statement] = ACTIONS(927),
    [sym_break_statement] = ACTIONS(927),
    [sym_continue_statement] = ACTIONS(927),
    [sym_debugger_statement] = ACTIONS(927),
    [anon_sym_assert] = ACTIONS(927),
    [anon_sym_global] = ACTIONS(927),
    [anon_sym_nonlocal] = ACTIONS(927),
    [anon_sym_yield] = ACTIONS(927),
    [anon_sym_from] = ACTIONS(927),
    [anon_sym_import] = ACTIONS(927),
    [anon_sym_DOT] = ACTIONS(927),
    [anon_sym_if] = ACTIONS(927),
    [anon_sym_for] = ACTIONS(927),
    [anon_sym_in] = ACTIONS(927),
    [anon_sym_while] = ACTIONS(927),
    [anon_sym_do] = ACTIONS(927),
    [anon_sym_try] = ACTIONS(927),
    [anon_sym_with] = ACTIONS(927),
    [anon_sym_STAR] = ACTIONS(927),
    [anon_sym_STAR_STAR] = ACTIONS(927),
    [anon_sym_class] = ACTIONS(927),
    [anon_sym_AT] = ACTIONS(927),
    [anon_sym_not] = ACTIONS(927),
    [anon_sym_and] = ACTIONS(927),
    [anon_sym_or] = ACTIONS(927),
    [anon_sym_PLUS] = ACTIONS(927),
    [anon_sym_DASH] = ACTIONS(927),
    [anon_sym_SLASH] = ACTIONS(927),
    [anon_sym_SLASH_SLASH] = ACTIONS(927),
    [anon_sym_PERCENT] = ACTIONS(927),
    [anon_sym_PIPE] = ACTIONS(927),
    [anon_sym_CARET] = ACTIONS(927),
    [anon_sym_AMP] = ACTIONS(927),
    [anon_sym_LT_LT] = ACTIONS(927),
    [anon_sym_GT_GT] = ACTIONS(927),
    [anon_sym_GT_GT_GT] = ACTIONS(927),
    [anon_sym_TILDE] = ACTIONS(927),
    [anon_sym_typeof] = ACTIONS(927),
    [anon_sym_void] = ACTIONS(927),
    [anon_sym_LT] = ACTIONS(927),
    [anon_sym_LT_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(927),
    [anon_sym_GT_EQ] = ACTIONS(927),
    [anon_sym_GT] = ACTIONS(927),
    [anon_sym_is] = ACTIONS(927),
    [anon_sym_instanceof] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
    [anon_sym_QMARK] = ACTIONS(927),
    [anon_sym_await] = ACTIONS(927),
    [anon_sym_LBRACE] = ACTIONS(927),
    [sym_this] = ACTIONS(927),
    [sym_true] = ACTIONS(927),
    [sym_false] = ACTIONS(927),
    [sym_none] = ACTIONS(927),
    [sym_string] = ACTIONS(927),
    [sym_verbatim] = ACTIONS(927),
    [sym_number] = ACTIONS(927),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(925),
    [sym_fstring_start] = ACTIONS(925),
  },
  [STATE(235)] = {
    [ts_builtin_sym_end] = ACTIONS(931),
    [sym_identifier] = ACTIONS(933),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym_EQ] = ACTIONS(935),
    [anon_sym_return] = ACTIONS(933),
    [anon_sym_async] = ACTIONS(933),
    [anon_sym_def] = ACTIONS(933),
    [anon_sym_PLUS_EQ] = ACTIONS(935),
    [anon_sym_DASH_EQ] = ACTIONS(935),
    [anon_sym_STAR_EQ] = ACTIONS(935),
    [anon_sym_SLASH_EQ] = ACTIONS(935),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(935),
    [anon_sym_PERCENT_EQ] = ACTIONS(935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_LT_LT_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_AMP_EQ] = ACTIONS(935),
    [anon_sym_CARET_EQ] = ACTIONS(935),
    [anon_sym_PIPE_EQ] = ACTIONS(935),
    [anon_sym_COMMA] = ACTIONS(971),
    [anon_sym_LPAREN] = ACTIONS(933),
    [anon_sym_LBRACK] = ACTIONS(933),
    [anon_sym_del] = ACTIONS(933),
    [anon_sym_raise] = ACTIONS(933),
    [sym_pass_statement] = ACTIONS(933),
    [sym_break_statement] = ACTIONS(933),
    [sym_continue_statement] = ACTIONS(933),
    [sym_debugger_statement] = ACTIONS(933),
    [anon_sym_assert] = ACTIONS(933),
    [anon_sym_global] = ACTIONS(933),
    [anon_sym_nonlocal] = ACTIONS(933),
    [anon_sym_yield] = ACTIONS(933),
    [anon_sym_from] = ACTIONS(933),
    [anon_sym_import] = ACTIONS(933),
    [anon_sym_DOT] = ACTIONS(933),
    [anon_sym_if] = ACTIONS(933),
    [anon_sym_for] = ACTIONS(933),
    [anon_sym_in] = ACTIONS(933),
    [anon_sym_while] = ACTIONS(933),
    [anon_sym_do] = ACTIONS(933),
    [anon_sym_try] = ACTIONS(933),
    [anon_sym_with] = ACTIONS(933),
    [anon_sym_STAR] = ACTIONS(933),
    [anon_sym_STAR_STAR] = ACTIONS(933),
    [anon_sym_class] = ACTIONS(933),
    [anon_sym_AT] = ACTIONS(933),
    [anon_sym_not] = ACTIONS(933),
    [anon_sym_and] = ACTIONS(933),
    [anon_sym_or] = ACTIONS(933),
    [anon_sym_PLUS] = ACTIONS(933),
    [anon_sym_DASH] = ACTIONS(933),
    [anon_sym_SLASH] = ACTIONS(933),
    [anon_sym_SLASH_SLASH] = ACTIONS(933),
    [anon_sym_PERCENT] = ACTIONS(933),
    [anon_sym_PIPE] = ACTIONS(933),
    [anon_sym_CARET] = ACTIONS(933),
    [anon_sym_AMP] = ACTIONS(933),
    [anon_sym_LT_LT] = ACTIONS(933),
    [anon_sym_GT_GT] = ACTIONS(933),
    [anon_sym_GT_GT_GT] = ACTIONS(933),
    [anon_sym_TILDE] = ACTIONS(933),
    [anon_sym_typeof] = ACTIONS(933),
    [anon_sym_void] = ACTIONS(933),
    [anon_sym_LT] = ACTIONS(933),
    [anon_sym_LT_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(933),
    [anon_sym_GT_EQ] = ACTIONS(933),
    [anon_sym_GT] = ACTIONS(933),
    [anon_sym_is] = ACTIONS(933),
    [anon_sym_instanceof] = ACTIONS(933),
    [anon_sym_new] = ACTIONS(933),
    [anon_sym_QMARK] = ACTIONS(933),
    [anon_sym_await] = ACTIONS(933),
    [anon_sym_LBRACE] = ACTIONS(933),
    [sym_this] = ACTIONS(933),
    [sym_true] = ACTIONS(933),
    [sym_false] = ACTIONS(933),
    [sym_none] = ACTIONS(933),
    [sym_string] = ACTIONS(933),
    [sym_verbatim] = ACTIONS(933),
    [sym_number] = ACTIONS(933),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(931),
    [sym_fstring_start] = ACTIONS(931),
  },
  [STATE(236)] = {
    [sym_identifier] = ACTIONS(927),
    [anon_sym_SEMI] = ACTIONS(927),
    [anon_sym_EQ] = ACTIONS(929),
    [anon_sym_return] = ACTIONS(927),
    [anon_sym_async] = ACTIONS(927),
    [anon_sym_def] = ACTIONS(927),
    [anon_sym_PLUS_EQ] = ACTIONS(929),
    [anon_sym_DASH_EQ] = ACTIONS(929),
    [anon_sym_STAR_EQ] = ACTIONS(929),
    [anon_sym_SLASH_EQ] = ACTIONS(929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(929),
    [anon_sym_PERCENT_EQ] = ACTIONS(929),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(929),
    [anon_sym_GT_GT_EQ] = ACTIONS(929),
    [anon_sym_LT_LT_EQ] = ACTIONS(929),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(929),
    [anon_sym_AMP_EQ] = ACTIONS(929),
    [anon_sym_CARET_EQ] = ACTIONS(929),
    [anon_sym_PIPE_EQ] = ACTIONS(929),
    [anon_sym_COMMA] = ACTIONS(929),
    [anon_sym_LPAREN] = ACTIONS(927),
    [anon_sym_LBRACK] = ACTIONS(927),
    [anon_sym_del] = ACTIONS(927),
    [anon_sym_raise] = ACTIONS(927),
    [sym_pass_statement] = ACTIONS(927),
    [sym_break_statement] = ACTIONS(927),
    [sym_continue_statement] = ACTIONS(927),
    [sym_debugger_statement] = ACTIONS(927),
    [anon_sym_assert] = ACTIONS(927),
    [anon_sym_global] = ACTIONS(927),
    [anon_sym_nonlocal] = ACTIONS(927),
    [anon_sym_yield] = ACTIONS(927),
    [anon_sym_from] = ACTIONS(927),
    [anon_sym_import] = ACTIONS(927),
    [anon_sym_DOT] = ACTIONS(927),
    [anon_sym_if] = ACTIONS(927),
    [anon_sym_for] = ACTIONS(927),
    [anon_sym_in] = ACTIONS(927),
    [anon_sym_while] = ACTIONS(927),
    [anon_sym_do] = ACTIONS(927),
    [anon_sym_try] = ACTIONS(927),
    [anon_sym_with] = ACTIONS(927),
    [anon_sym_STAR] = ACTIONS(927),
    [anon_sym_STAR_STAR] = ACTIONS(927),
    [anon_sym_class] = ACTIONS(927),
    [anon_sym_AT] = ACTIONS(927),
    [anon_sym_not] = ACTIONS(927),
    [anon_sym_and] = ACTIONS(927),
    [anon_sym_or] = ACTIONS(927),
    [anon_sym_PLUS] = ACTIONS(927),
    [anon_sym_DASH] = ACTIONS(927),
    [anon_sym_SLASH] = ACTIONS(927),
    [anon_sym_SLASH_SLASH] = ACTIONS(927),
    [anon_sym_PERCENT] = ACTIONS(927),
    [anon_sym_PIPE] = ACTIONS(927),
    [anon_sym_CARET] = ACTIONS(927),
    [anon_sym_AMP] = ACTIONS(927),
    [anon_sym_LT_LT] = ACTIONS(927),
    [anon_sym_GT_GT] = ACTIONS(927),
    [anon_sym_GT_GT_GT] = ACTIONS(927),
    [anon_sym_TILDE] = ACTIONS(927),
    [anon_sym_typeof] = ACTIONS(927),
    [anon_sym_void] = ACTIONS(927),
    [anon_sym_LT] = ACTIONS(927),
    [anon_sym_LT_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(927),
    [anon_sym_GT_EQ] = ACTIONS(927),
    [anon_sym_GT] = ACTIONS(927),
    [anon_sym_is] = ACTIONS(927),
    [anon_sym_instanceof] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
    [anon_sym_QMARK] = ACTIONS(927),
    [anon_sym_await] = ACTIONS(927),
    [anon_sym_LBRACE] = ACTIONS(927),
    [sym_this] = ACTIONS(927),
    [sym_true] = ACTIONS(927),
    [sym_false] = ACTIONS(927),
    [sym_none] = ACTIONS(927),
    [sym_string] = ACTIONS(927),
    [sym_verbatim] = ACTIONS(927),
    [sym_number] = ACTIONS(927),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(925),
    [sym_regex] = ACTIONS(925),
    [sym_fstring_start] = ACTIONS(925),
  },
  [STATE(237)] = {
    [sym_identifier] = ACTIONS(927),
    [anon_sym_SEMI] = ACTIONS(927),
    [anon_sym_EQ] = ACTIONS(929),
    [anon_sym_return] = ACTIONS(927),
    [anon_sym_async] = ACTIONS(927),
    [anon_sym_def] = ACTIONS(927),
    [anon_sym_PLUS_EQ] = ACTIONS(929),
    [anon_sym_DASH_EQ] = ACTIONS(929),
    [anon_sym_STAR_EQ] = ACTIONS(929),
    [anon_sym_SLASH_EQ] = ACTIONS(929),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(929),
    [anon_sym_PERCENT_EQ] = ACTIONS(929),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(929),
    [anon_sym_GT_GT_EQ] = ACTIONS(929),
    [anon_sym_LT_LT_EQ] = ACTIONS(929),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(929),
    [anon_sym_AMP_EQ] = ACTIONS(929),
    [anon_sym_CARET_EQ] = ACTIONS(929),
    [anon_sym_PIPE_EQ] = ACTIONS(929),
    [anon_sym_COMMA] = ACTIONS(968),
    [anon_sym_LPAREN] = ACTIONS(927),
    [anon_sym_LBRACK] = ACTIONS(927),
    [anon_sym_del] = ACTIONS(927),
    [anon_sym_raise] = ACTIONS(927),
    [sym_pass_statement] = ACTIONS(927),
    [sym_break_statement] = ACTIONS(927),
    [sym_continue_statement] = ACTIONS(927),
    [sym_debugger_statement] = ACTIONS(927),
    [anon_sym_assert] = ACTIONS(927),
    [anon_sym_global] = ACTIONS(927),
    [anon_sym_nonlocal] = ACTIONS(927),
    [anon_sym_yield] = ACTIONS(927),
    [anon_sym_from] = ACTIONS(927),
    [anon_sym_import] = ACTIONS(927),
    [anon_sym_DOT] = ACTIONS(927),
    [anon_sym_if] = ACTIONS(927),
    [anon_sym_for] = ACTIONS(927),
    [anon_sym_in] = ACTIONS(927),
    [anon_sym_while] = ACTIONS(927),
    [anon_sym_do] = ACTIONS(927),
    [anon_sym_try] = ACTIONS(927),
    [anon_sym_with] = ACTIONS(927),
    [anon_sym_STAR] = ACTIONS(927),
    [anon_sym_STAR_STAR] = ACTIONS(927),
    [anon_sym_class] = ACTIONS(927),
    [anon_sym_AT] = ACTIONS(927),
    [anon_sym_not] = ACTIONS(927),
    [anon_sym_and] = ACTIONS(927),
    [anon_sym_or] = ACTIONS(927),
    [anon_sym_PLUS] = ACTIONS(927),
    [anon_sym_DASH] = ACTIONS(927),
    [anon_sym_SLASH] = ACTIONS(927),
    [anon_sym_SLASH_SLASH] = ACTIONS(927),
    [anon_sym_PERCENT] = ACTIONS(927),
    [anon_sym_PIPE] = ACTIONS(927),
    [anon_sym_CARET] = ACTIONS(927),
    [anon_sym_AMP] = ACTIONS(927),
    [anon_sym_LT_LT] = ACTIONS(927),
    [anon_sym_GT_GT] = ACTIONS(927),
    [anon_sym_GT_GT_GT] = ACTIONS(927),
    [anon_sym_TILDE] = ACTIONS(927),
    [anon_sym_typeof] = ACTIONS(927),
    [anon_sym_void] = ACTIONS(927),
    [anon_sym_LT] = ACTIONS(927),
    [anon_sym_LT_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(927),
    [anon_sym_GT_EQ] = ACTIONS(927),
    [anon_sym_GT] = ACTIONS(927),
    [anon_sym_is] = ACTIONS(927),
    [anon_sym_instanceof] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
    [anon_sym_QMARK] = ACTIONS(927),
    [anon_sym_await] = ACTIONS(927),
    [anon_sym_LBRACE] = ACTIONS(927),
    [sym_this] = ACTIONS(927),
    [sym_true] = ACTIONS(927),
    [sym_false] = ACTIONS(927),
    [sym_none] = ACTIONS(927),
    [sym_string] = ACTIONS(927),
    [sym_verbatim] = ACTIONS(927),
    [sym_number] = ACTIONS(927),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(925),
    [sym_regex] = ACTIONS(925),
    [sym_fstring_start] = ACTIONS(925),
  },
  [STATE(238)] = {
    [sym_identifier] = ACTIONS(933),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym_EQ] = ACTIONS(935),
    [anon_sym_return] = ACTIONS(933),
    [anon_sym_async] = ACTIONS(933),
    [anon_sym_def] = ACTIONS(933),
    [anon_sym_PLUS_EQ] = ACTIONS(935),
    [anon_sym_DASH_EQ] = ACTIONS(935),
    [anon_sym_STAR_EQ] = ACTIONS(935),
    [anon_sym_SLASH_EQ] = ACTIONS(935),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(935),
    [anon_sym_PERCENT_EQ] = ACTIONS(935),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_LT_LT_EQ] = ACTIONS(935),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(935),
    [anon_sym_AMP_EQ] = ACTIONS(935),
    [anon_sym_CARET_EQ] = ACTIONS(935),
    [anon_sym_PIPE_EQ] = ACTIONS(935),
    [anon_sym_COMMA] = ACTIONS(971),
    [anon_sym_LPAREN] = ACTIONS(933),
    [anon_sym_LBRACK] = ACTIONS(933),
    [anon_sym_del] = ACTIONS(933),
    [anon_sym_raise] = ACTIONS(933),
    [sym_pass_statement] = ACTIONS(933),
    [sym_break_statement] = ACTIONS(933),
    [sym_continue_statement] = ACTIONS(933),
    [sym_debugger_statement] = ACTIONS(933),
    [anon_sym_assert] = ACTIONS(933),
    [anon_sym_global] = ACTIONS(933),
    [anon_sym_nonlocal] = ACTIONS(933),
    [anon_sym_yield] = ACTIONS(933),
    [anon_sym_from] = ACTIONS(933),
    [anon_sym_import] = ACTIONS(933),
    [anon_sym_DOT] = ACTIONS(933),
    [anon_sym_if] = ACTIONS(933),
    [anon_sym_for] = ACTIONS(933),
    [anon_sym_in] = ACTIONS(933),
    [anon_sym_while] = ACTIONS(933),
    [anon_sym_do] = ACTIONS(933),
    [anon_sym_try] = ACTIONS(933),
    [anon_sym_with] = ACTIONS(933),
    [anon_sym_STAR] = ACTIONS(933),
    [anon_sym_STAR_STAR] = ACTIONS(933),
    [anon_sym_class] = ACTIONS(933),
    [anon_sym_AT] = ACTIONS(933),
    [anon_sym_not] = ACTIONS(933),
    [anon_sym_and] = ACTIONS(933),
    [anon_sym_or] = ACTIONS(933),
    [anon_sym_PLUS] = ACTIONS(933),
    [anon_sym_DASH] = ACTIONS(933),
    [anon_sym_SLASH] = ACTIONS(933),
    [anon_sym_SLASH_SLASH] = ACTIONS(933),
    [anon_sym_PERCENT] = ACTIONS(933),
    [anon_sym_PIPE] = ACTIONS(933),
    [anon_sym_CARET] = ACTIONS(933),
    [anon_sym_AMP] = ACTIONS(933),
    [anon_sym_LT_LT] = ACTIONS(933),
    [anon_sym_GT_GT] = ACTIONS(933),
    [anon_sym_GT_GT_GT] = ACTIONS(933),
    [anon_sym_TILDE] = ACTIONS(933),
    [anon_sym_typeof] = ACTIONS(933),
    [anon_sym_void] = ACTIONS(933),
    [anon_sym_LT] = ACTIONS(933),
    [anon_sym_LT_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(933),
    [anon_sym_GT_EQ] = ACTIONS(933),
    [anon_sym_GT] = ACTIONS(933),
    [anon_sym_is] = ACTIONS(933),
    [anon_sym_instanceof] = ACTIONS(933),
    [anon_sym_new] = ACTIONS(933),
    [anon_sym_QMARK] = ACTIONS(933),
    [anon_sym_await] = ACTIONS(933),
    [anon_sym_LBRACE] = ACTIONS(933),
    [sym_this] = ACTIONS(933),
    [sym_true] = ACTIONS(933),
    [sym_false] = ACTIONS(933),
    [sym_none] = ACTIONS(933),
    [sym_string] = ACTIONS(933),
    [sym_verbatim] = ACTIONS(933),
    [sym_number] = ACTIONS(933),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(931),
    [sym_regex] = ACTIONS(931),
    [sym_fstring_start] = ACTIONS(931),
  },
  [STATE(239)] = {
    [ts_builtin_sym_end] = ACTIONS(937),
    [sym_identifier] = ACTIONS(939),
    [anon_sym_SEMI] = ACTIONS(939),
    [anon_sym_EQ] = ACTIONS(941),
    [anon_sym_return] = ACTIONS(939),
    [anon_sym_async] = ACTIONS(939),
    [anon_sym_def] = ACTIONS(939),
    [anon_sym_PLUS_EQ] = ACTIONS(941),
    [anon_sym_DASH_EQ] = ACTIONS(941),
    [anon_sym_STAR_EQ] = ACTIONS(941),
    [anon_sym_SLASH_EQ] = ACTIONS(941),
    [anon_sym_SLASH_SLASH_EQ] = ACTIONS(941),
    [anon_sym_PERCENT_EQ] = ACTIONS(941),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(941),
    [anon_sym_GT_GT_EQ] = ACTIONS(941),
    [anon_sym_LT_LT_EQ] = ACTIONS(941),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(941),
    [anon_sym_AMP_EQ] = ACTIONS(941),
    [anon_sym_CARET_EQ] = ACTIONS(941),
    [anon_sym_PIPE_EQ] = ACTIONS(941),
    [anon_sym_COMMA] = ACTIONS(965),
    [anon_sym_LPAREN] = ACTIONS(939),
    [anon_sym_LBRACK] = ACTIONS(939),
    [anon_sym_del] = ACTIONS(939),
    [anon_sym_raise] = ACTIONS(939),
    [sym_pass_statement] = ACTIONS(939),
    [sym_break_statement] = ACTIONS(939),
    [sym_continue_statement] = ACTIONS(939),
    [sym_debugger_statement] = ACTIONS(939),
    [anon_sym_assert] = ACTIONS(939),
    [anon_sym_global] = ACTIONS(939),
    [anon_sym_nonlocal] = ACTIONS(939),
    [anon_sym_yield] = ACTIONS(939),
    [anon_sym_from] = ACTIONS(939),
    [anon_sym_import] = ACTIONS(939),
    [anon_sym_DOT] = ACTIONS(939),
    [anon_sym_if] = ACTIONS(939),
    [anon_sym_for] = ACTIONS(939),
    [anon_sym_in] = ACTIONS(939),
    [anon_sym_while] = ACTIONS(939),
    [anon_sym_do] = ACTIONS(939),
    [anon_sym_try] = ACTIONS(939),
    [anon_sym_with] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(939),
    [anon_sym_STAR_STAR] = ACTIONS(939),
    [anon_sym_class] = ACTIONS(939),
    [anon_sym_AT] = ACTIONS(939),
    [anon_sym_not] = ACTIONS(939),
    [anon_sym_and] = ACTIONS(939),
    [anon_sym_or] = ACTIONS(939),
    [anon_sym_PLUS] = ACTIONS(939),
    [anon_sym_DASH] = ACTIONS(939),
    [anon_sym_SLASH] = ACTIONS(939),
    [anon_sym_SLASH_SLASH] = ACTIONS(939),
    [anon_sym_PERCENT] = ACTIONS(939),
    [anon_sym_PIPE] = ACTIONS(939),
    [anon_sym_CARET] = ACTIONS(939),
    [anon_sym_AMP] = ACTIONS(939),
    [anon_sym_LT_LT] = ACTIONS(939),
    [anon_sym_GT_GT] = ACTIONS(939),
    [anon_sym_GT_GT_GT] = ACTIONS(939),
    [anon_sym_TILDE] = ACTIONS(939),
    [anon_sym_typeof] = ACTIONS(939),
    [anon_sym_void] = ACTIONS(939),
    [anon_sym_LT] = ACTIONS(939),
    [anon_sym_LT_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(939),
    [anon_sym_GT_EQ] = ACTIONS(939),
    [anon_sym_GT] = ACTIONS(939),
    [anon_sym_is] = ACTIONS(939),
    [anon_sym_instanceof] = ACTIONS(939),
    [anon_sym_new] = ACTIONS(939),
    [anon_sym_QMARK] = ACTIONS(939),
    [anon_sym_await] = ACTIONS(939),
    [anon_sym_LBRACE] = ACTIONS(939),
    [sym_this] = ACTIONS(939),
    [sym_true] = ACTIONS(939),
    [sym_false] = ACTIONS(939),
    [sym_none] = ACTIONS(939),
    [sym_string] = ACTIONS(939),
    [sym_verbatim] = ACTIONS(939),
    [sym_number] = ACTIONS(939),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(937),
    [sym_fstring_start] = ACTIONS(937),
  },
  [STATE(240)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2542),
    [sym_pattern_list] = STATE(2542),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2167),
    [sym__primary_expression] = STATE(2174),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1489),
    [sym_subscript] = STATE(1489),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(757),
    [anon_sym_SEMI] = ACTIONS(449),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_def] = ACTIONS(883),
    [anon_sym_LPAREN] = ACTIONS(797),
    [anon_sym_LBRACK] = ACTIONS(799),
    [anon_sym_DOT] = ACTIONS(449),
    [anon_sym_if] = ACTIONS(449),
    [anon_sym_else] = ACTIONS(449),
    [anon_sym_in] = ACTIONS(449),
    [anon_sym_STAR] = ACTIONS(449),
    [anon_sym_STAR_STAR] = ACTIONS(449),
    [anon_sym_AT] = ACTIONS(449),
    [anon_sym_not] = ACTIONS(885),
    [anon_sym_and] = ACTIONS(449),
    [anon_sym_or] = ACTIONS(449),
    [anon_sym_PLUS] = ACTIONS(887),
    [anon_sym_DASH] = ACTIONS(887),
    [anon_sym_SLASH] = ACTIONS(449),
    [anon_sym_SLASH_SLASH] = ACTIONS(449),
    [anon_sym_PERCENT] = ACTIONS(449),
    [anon_sym_PIPE] = ACTIONS(449),
    [anon_sym_CARET] = ACTIONS(449),
    [anon_sym_AMP] = ACTIONS(449),
    [anon_sym_LT_LT] = ACTIONS(449),
    [anon_sym_GT_GT] = ACTIONS(449),
    [anon_sym_GT_GT_GT] = ACTIONS(449),
    [anon_sym_TILDE] = ACTIONS(887),
    [anon_sym_typeof] = ACTIONS(887),
    [anon_sym_void] = ACTIONS(887),
    [anon_sym_LT] = ACTIONS(449),
    [anon_sym_LT_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ] = ACTIONS(449),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ] = ACTIONS(449),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(449),
    [anon_sym_GT_EQ] = ACTIONS(449),
    [anon_sym_GT] = ACTIONS(449),
    [anon_sym_is] = ACTIONS(449),
    [anon_sym_instanceof] = ACTIONS(449),
    [anon_sym_new] = ACTIONS(889),
    [anon_sym_QMARK] = ACTIONS(449),
    [anon_sym_await] = ACTIONS(891),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(757),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(241)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(974),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(242)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(976),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(243)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(978),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(244)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(980),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(245)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(982),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(246)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(984),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(247)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(986),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(248)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(988),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(249)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(990),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(250)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(992),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(251)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(994),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(252)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(996),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(253)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(998),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(254)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1000),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(255)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1002),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(256)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1004),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(257)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1006),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(258)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1008),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(259)] = {
    [sym__simple_statement] = STATE(2884),
    [sym__expression_statement] = STATE(2837),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2837),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2837),
    [sym_delete_statement] = STATE(2837),
    [sym_raise_statement] = STATE(2837),
    [sym_assert_statement] = STATE(2837),
    [sym_global_statement] = STATE(2837),
    [sym_nonlocal_statement] = STATE(2837),
    [sym_yield_statement] = STATE(2837),
    [sym_yield] = STATE(2272),
    [sym_import_statement] = STATE(2837),
    [sym_import_from_statement] = STATE(2837),
    [sym_scoped_flag_statement] = STATE(2837),
    [sym__expression] = STATE(2340),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(859),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(19),
    [anon_sym_raise] = ACTIONS(21),
    [sym_pass_statement] = ACTIONS(23),
    [sym_break_statement] = ACTIONS(23),
    [sym_continue_statement] = ACTIONS(23),
    [sym_debugger_statement] = ACTIONS(23),
    [anon_sym_assert] = ACTIONS(25),
    [anon_sym_global] = ACTIONS(27),
    [anon_sym_nonlocal] = ACTIONS(29),
    [anon_sym_yield] = ACTIONS(31),
    [anon_sym_from] = ACTIONS(33),
    [anon_sym_import] = ACTIONS(35),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(260)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2250),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2250),
    [sym__inline_body] = STATE(2250),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(261)] = {
    [sym__expression_statement] = STATE(2042),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(2042),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2042),
    [sym_delete_statement] = STATE(2042),
    [sym_raise_statement] = STATE(2042),
    [sym_assert_statement] = STATE(2042),
    [sym_yield_statement] = STATE(2042),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(2042),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(1014),
    [sym_break_statement] = ACTIONS(1014),
    [sym_continue_statement] = ACTIONS(1014),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(262)] = {
    [sym__expression_statement] = STATE(2042),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(2042),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2042),
    [sym_delete_statement] = STATE(2042),
    [sym_raise_statement] = STATE(2042),
    [sym_assert_statement] = STATE(2042),
    [sym_yield_statement] = STATE(2042),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(2042),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(1014),
    [sym_break_statement] = ACTIONS(1014),
    [sym_continue_statement] = ACTIONS(1014),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(263)] = {
    [sym__expression_statement] = STATE(1720),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1720),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1720),
    [sym_delete_statement] = STATE(1720),
    [sym_raise_statement] = STATE(1720),
    [sym_assert_statement] = STATE(1720),
    [sym_yield_statement] = STATE(1720),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(1720),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(1016),
    [sym_break_statement] = ACTIONS(1016),
    [sym_continue_statement] = ACTIONS(1016),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(264)] = {
    [sym__expression_statement] = STATE(563),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(563),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(563),
    [sym_delete_statement] = STATE(563),
    [sym_raise_statement] = STATE(563),
    [sym_assert_statement] = STATE(563),
    [sym_yield_statement] = STATE(563),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(591),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(591),
    [sym__inline_body] = STATE(591),
    [sym__inline_simple_statement] = STATE(563),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(1018),
    [sym_break_statement] = ACTIONS(1018),
    [sym_continue_statement] = ACTIONS(1018),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(265)] = {
    [sym__expression_statement] = STATE(1720),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1720),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1720),
    [sym_delete_statement] = STATE(1720),
    [sym_raise_statement] = STATE(1720),
    [sym_assert_statement] = STATE(1720),
    [sym_yield_statement] = STATE(1720),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(1720),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(1016),
    [sym_break_statement] = ACTIONS(1016),
    [sym_continue_statement] = ACTIONS(1016),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(266)] = {
    [sym__expression_statement] = STATE(563),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(563),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(563),
    [sym_delete_statement] = STATE(563),
    [sym_raise_statement] = STATE(563),
    [sym_assert_statement] = STATE(563),
    [sym_yield_statement] = STATE(563),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(616),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(616),
    [sym__inline_body] = STATE(616),
    [sym__inline_simple_statement] = STATE(563),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(1018),
    [sym_break_statement] = ACTIONS(1018),
    [sym_continue_statement] = ACTIONS(1018),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(267)] = {
    [sym__expression_statement] = STATE(563),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(563),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(563),
    [sym_delete_statement] = STATE(563),
    [sym_raise_statement] = STATE(563),
    [sym_assert_statement] = STATE(563),
    [sym_yield_statement] = STATE(563),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(628),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(628),
    [sym__inline_body] = STATE(628),
    [sym__inline_simple_statement] = STATE(563),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(1018),
    [sym_break_statement] = ACTIONS(1018),
    [sym_continue_statement] = ACTIONS(1018),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(268)] = {
    [sym__expression_statement] = STATE(563),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(563),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(563),
    [sym_delete_statement] = STATE(563),
    [sym_raise_statement] = STATE(563),
    [sym_assert_statement] = STATE(563),
    [sym_yield_statement] = STATE(563),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(559),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(559),
    [sym__inline_body] = STATE(559),
    [sym__inline_simple_statement] = STATE(563),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(1018),
    [sym_break_statement] = ACTIONS(1018),
    [sym_continue_statement] = ACTIONS(1018),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(269)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(535),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2216),
    [sym__inline_body] = STATE(2216),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(745),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(270)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2239),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2239),
    [sym__inline_body] = STATE(2239),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(271)] = {
    [sym__expression_statement] = STATE(1720),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1720),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1720),
    [sym_delete_statement] = STATE(1720),
    [sym_raise_statement] = STATE(1720),
    [sym_assert_statement] = STATE(1720),
    [sym_yield_statement] = STATE(1720),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(1720),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(1016),
    [sym_break_statement] = ACTIONS(1016),
    [sym_continue_statement] = ACTIONS(1016),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(272)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2169),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2169),
    [sym__inline_body] = STATE(2169),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(273)] = {
    [sym__expression_statement] = STATE(611),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(611),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(611),
    [sym_delete_statement] = STATE(611),
    [sym_raise_statement] = STATE(611),
    [sym_assert_statement] = STATE(611),
    [sym_yield_statement] = STATE(611),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(610),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(610),
    [sym__inline_body] = STATE(610),
    [sym__inline_simple_statement] = STATE(611),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(1022),
    [sym_break_statement] = ACTIONS(1022),
    [sym_continue_statement] = ACTIONS(1022),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(274)] = {
    [sym__expression_statement] = STATE(1720),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1720),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1720),
    [sym_delete_statement] = STATE(1720),
    [sym_raise_statement] = STATE(1720),
    [sym_assert_statement] = STATE(1720),
    [sym_yield_statement] = STATE(1720),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(1720),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(1016),
    [sym_break_statement] = ACTIONS(1016),
    [sym_continue_statement] = ACTIONS(1016),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(275)] = {
    [sym__expression_statement] = STATE(611),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(611),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(611),
    [sym_delete_statement] = STATE(611),
    [sym_raise_statement] = STATE(611),
    [sym_assert_statement] = STATE(611),
    [sym_yield_statement] = STATE(611),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(650),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(650),
    [sym__inline_body] = STATE(650),
    [sym__inline_simple_statement] = STATE(611),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(1022),
    [sym_break_statement] = ACTIONS(1022),
    [sym_continue_statement] = ACTIONS(1022),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(276)] = {
    [sym__expression_statement] = STATE(611),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(611),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(611),
    [sym_delete_statement] = STATE(611),
    [sym_raise_statement] = STATE(611),
    [sym_assert_statement] = STATE(611),
    [sym_yield_statement] = STATE(611),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(699),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(699),
    [sym__inline_body] = STATE(699),
    [sym__inline_simple_statement] = STATE(611),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(1022),
    [sym_break_statement] = ACTIONS(1022),
    [sym_continue_statement] = ACTIONS(1022),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(277)] = {
    [sym__expression_statement] = STATE(611),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(611),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(611),
    [sym_delete_statement] = STATE(611),
    [sym_raise_statement] = STATE(611),
    [sym_assert_statement] = STATE(611),
    [sym_yield_statement] = STATE(611),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(701),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(701),
    [sym__inline_body] = STATE(701),
    [sym__inline_simple_statement] = STATE(611),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(1022),
    [sym_break_statement] = ACTIONS(1022),
    [sym_continue_statement] = ACTIONS(1022),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(278)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2171),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2171),
    [sym__inline_body] = STATE(2171),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(279)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(554),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2239),
    [sym__inline_body] = STATE(2239),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(745),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(280)] = {
    [sym__expression_statement] = STATE(2162),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2162),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2162),
    [sym_delete_statement] = STATE(2162),
    [sym_raise_statement] = STATE(2162),
    [sym_assert_statement] = STATE(2162),
    [sym_yield_statement] = STATE(2162),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2216),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2216),
    [sym__inline_body] = STATE(2216),
    [sym__inline_simple_statement] = STATE(2162),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(1026),
    [sym_break_statement] = ACTIONS(1026),
    [sym_continue_statement] = ACTIONS(1026),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(281)] = {
    [sym__expression_statement] = STATE(1916),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1916),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1916),
    [sym_delete_statement] = STATE(1916),
    [sym_raise_statement] = STATE(1916),
    [sym_assert_statement] = STATE(1916),
    [sym_yield_statement] = STATE(1916),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(1916),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(1028),
    [sym_break_statement] = ACTIONS(1028),
    [sym_continue_statement] = ACTIONS(1028),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(282)] = {
    [sym__expression_statement] = STATE(2162),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2162),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2162),
    [sym_delete_statement] = STATE(2162),
    [sym_raise_statement] = STATE(2162),
    [sym_assert_statement] = STATE(2162),
    [sym_yield_statement] = STATE(2162),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2250),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2250),
    [sym__inline_body] = STATE(2250),
    [sym__inline_simple_statement] = STATE(2162),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(1026),
    [sym_break_statement] = ACTIONS(1026),
    [sym_continue_statement] = ACTIONS(1026),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(283)] = {
    [sym__expression_statement] = STATE(1916),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1916),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1916),
    [sym_delete_statement] = STATE(1916),
    [sym_raise_statement] = STATE(1916),
    [sym_assert_statement] = STATE(1916),
    [sym_yield_statement] = STATE(1916),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(1916),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(1028),
    [sym_break_statement] = ACTIONS(1028),
    [sym_continue_statement] = ACTIONS(1028),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(284)] = {
    [sym__expression_statement] = STATE(1720),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1720),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1720),
    [sym_delete_statement] = STATE(1720),
    [sym_raise_statement] = STATE(1720),
    [sym_assert_statement] = STATE(1720),
    [sym_yield_statement] = STATE(1720),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(1720),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(1016),
    [sym_break_statement] = ACTIONS(1016),
    [sym_continue_statement] = ACTIONS(1016),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(285)] = {
    [sym__expression_statement] = STATE(2162),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2162),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2162),
    [sym_delete_statement] = STATE(2162),
    [sym_raise_statement] = STATE(2162),
    [sym_assert_statement] = STATE(2162),
    [sym_yield_statement] = STATE(2162),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2239),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2239),
    [sym__inline_body] = STATE(2239),
    [sym__inline_simple_statement] = STATE(2162),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(1026),
    [sym_break_statement] = ACTIONS(1026),
    [sym_continue_statement] = ACTIONS(1026),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(286)] = {
    [sym__expression_statement] = STATE(1916),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1916),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1916),
    [sym_delete_statement] = STATE(1916),
    [sym_raise_statement] = STATE(1916),
    [sym_assert_statement] = STATE(1916),
    [sym_yield_statement] = STATE(1916),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(1916),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(1028),
    [sym_break_statement] = ACTIONS(1028),
    [sym_continue_statement] = ACTIONS(1028),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(287)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(516),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2216),
    [sym__inline_body] = STATE(2216),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(751),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(288)] = {
    [sym__expression_statement] = STATE(2162),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2162),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2162),
    [sym_delete_statement] = STATE(2162),
    [sym_raise_statement] = STATE(2162),
    [sym_assert_statement] = STATE(2162),
    [sym_yield_statement] = STATE(2162),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2169),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2169),
    [sym__inline_body] = STATE(2169),
    [sym__inline_simple_statement] = STATE(2162),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(1026),
    [sym_break_statement] = ACTIONS(1026),
    [sym_continue_statement] = ACTIONS(1026),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(289)] = {
    [sym__expression_statement] = STATE(1916),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1916),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1916),
    [sym_delete_statement] = STATE(1916),
    [sym_raise_statement] = STATE(1916),
    [sym_assert_statement] = STATE(1916),
    [sym_yield_statement] = STATE(1916),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(1916),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(1028),
    [sym_break_statement] = ACTIONS(1028),
    [sym_continue_statement] = ACTIONS(1028),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(290)] = {
    [sym__expression_statement] = STATE(1720),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1720),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1720),
    [sym_delete_statement] = STATE(1720),
    [sym_raise_statement] = STATE(1720),
    [sym_assert_statement] = STATE(1720),
    [sym_yield_statement] = STATE(1720),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(1720),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(1016),
    [sym_break_statement] = ACTIONS(1016),
    [sym_continue_statement] = ACTIONS(1016),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(291)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(534),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2250),
    [sym__inline_body] = STATE(2250),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(751),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(292)] = {
    [sym__expression_statement] = STATE(1916),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1916),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1916),
    [sym_delete_statement] = STATE(1916),
    [sym_raise_statement] = STATE(1916),
    [sym_assert_statement] = STATE(1916),
    [sym_yield_statement] = STATE(1916),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(1916),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(1028),
    [sym_break_statement] = ACTIONS(1028),
    [sym_continue_statement] = ACTIONS(1028),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(293)] = {
    [sym__expression_statement] = STATE(2162),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2162),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2162),
    [sym_delete_statement] = STATE(2162),
    [sym_raise_statement] = STATE(2162),
    [sym_assert_statement] = STATE(2162),
    [sym_yield_statement] = STATE(2162),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2221),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2221),
    [sym__inline_body] = STATE(2221),
    [sym__inline_simple_statement] = STATE(2162),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(1026),
    [sym_break_statement] = ACTIONS(1026),
    [sym_continue_statement] = ACTIONS(1026),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(294)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(513),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2239),
    [sym__inline_body] = STATE(2239),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(751),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(295)] = {
    [sym__expression_statement] = STATE(1916),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1916),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1916),
    [sym_delete_statement] = STATE(1916),
    [sym_raise_statement] = STATE(1916),
    [sym_assert_statement] = STATE(1916),
    [sym_yield_statement] = STATE(1916),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(1916),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(1028),
    [sym_break_statement] = ACTIONS(1028),
    [sym_continue_statement] = ACTIONS(1028),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(296)] = {
    [sym__expression_statement] = STATE(2162),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2162),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2162),
    [sym_delete_statement] = STATE(2162),
    [sym_raise_statement] = STATE(2162),
    [sym_assert_statement] = STATE(2162),
    [sym_yield_statement] = STATE(2162),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2220),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2220),
    [sym__inline_body] = STATE(2220),
    [sym__inline_simple_statement] = STATE(2162),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(1026),
    [sym_break_statement] = ACTIONS(1026),
    [sym_continue_statement] = ACTIONS(1026),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(297)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(526),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2169),
    [sym__inline_body] = STATE(2169),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(751),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(298)] = {
    [sym__expression_statement] = STATE(1916),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1916),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1916),
    [sym_delete_statement] = STATE(1916),
    [sym_raise_statement] = STATE(1916),
    [sym_assert_statement] = STATE(1916),
    [sym_yield_statement] = STATE(1916),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(1916),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(1028),
    [sym_break_statement] = ACTIONS(1028),
    [sym_continue_statement] = ACTIONS(1028),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(299)] = {
    [sym__expression_statement] = STATE(1720),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1720),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1720),
    [sym_delete_statement] = STATE(1720),
    [sym_raise_statement] = STATE(1720),
    [sym_assert_statement] = STATE(1720),
    [sym_yield_statement] = STATE(1720),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(1720),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(1016),
    [sym_break_statement] = ACTIONS(1016),
    [sym_continue_statement] = ACTIONS(1016),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(300)] = {
    [sym__expression_statement] = STATE(2162),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2162),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2162),
    [sym_delete_statement] = STATE(2162),
    [sym_raise_statement] = STATE(2162),
    [sym_assert_statement] = STATE(2162),
    [sym_yield_statement] = STATE(2162),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2171),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2171),
    [sym__inline_body] = STATE(2171),
    [sym__inline_simple_statement] = STATE(2162),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(1026),
    [sym_break_statement] = ACTIONS(1026),
    [sym_continue_statement] = ACTIONS(1026),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(301)] = {
    [sym__expression_statement] = STATE(1916),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1916),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1916),
    [sym_delete_statement] = STATE(1916),
    [sym_raise_statement] = STATE(1916),
    [sym_assert_statement] = STATE(1916),
    [sym_yield_statement] = STATE(1916),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(1916),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(1028),
    [sym_break_statement] = ACTIONS(1028),
    [sym_continue_statement] = ACTIONS(1028),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(302)] = {
    [sym__expression_statement] = STATE(2162),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2162),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2162),
    [sym_delete_statement] = STATE(2162),
    [sym_raise_statement] = STATE(2162),
    [sym_assert_statement] = STATE(2162),
    [sym_yield_statement] = STATE(2162),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2215),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2215),
    [sym__inline_body] = STATE(2215),
    [sym__inline_simple_statement] = STATE(2162),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(1026),
    [sym_break_statement] = ACTIONS(1026),
    [sym_continue_statement] = ACTIONS(1026),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(303)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2215),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2215),
    [sym__inline_body] = STATE(2215),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(304)] = {
    [sym__expression_statement] = STATE(1791),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1791),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1791),
    [sym_delete_statement] = STATE(1791),
    [sym_raise_statement] = STATE(1791),
    [sym_assert_statement] = STATE(1791),
    [sym_yield_statement] = STATE(1791),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(1791),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(1030),
    [sym_break_statement] = ACTIONS(1030),
    [sym_continue_statement] = ACTIONS(1030),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(305)] = {
    [sym__expression_statement] = STATE(1791),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1791),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1791),
    [sym_delete_statement] = STATE(1791),
    [sym_raise_statement] = STATE(1791),
    [sym_assert_statement] = STATE(1791),
    [sym_yield_statement] = STATE(1791),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(1791),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(1030),
    [sym_break_statement] = ACTIONS(1030),
    [sym_continue_statement] = ACTIONS(1030),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(306)] = {
    [sym__expression_statement] = STATE(1791),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1791),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1791),
    [sym_delete_statement] = STATE(1791),
    [sym_raise_statement] = STATE(1791),
    [sym_assert_statement] = STATE(1791),
    [sym_yield_statement] = STATE(1791),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(1791),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(1030),
    [sym_break_statement] = ACTIONS(1030),
    [sym_continue_statement] = ACTIONS(1030),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(307)] = {
    [sym__expression_statement] = STATE(1791),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1791),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1791),
    [sym_delete_statement] = STATE(1791),
    [sym_raise_statement] = STATE(1791),
    [sym_assert_statement] = STATE(1791),
    [sym_yield_statement] = STATE(1791),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(1791),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(1030),
    [sym_break_statement] = ACTIONS(1030),
    [sym_continue_statement] = ACTIONS(1030),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(308)] = {
    [sym__expression_statement] = STATE(1962),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1962),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1962),
    [sym_delete_statement] = STATE(1962),
    [sym_raise_statement] = STATE(1962),
    [sym_assert_statement] = STATE(1962),
    [sym_yield_statement] = STATE(1962),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(1962),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(1032),
    [sym_break_statement] = ACTIONS(1032),
    [sym_continue_statement] = ACTIONS(1032),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(309)] = {
    [sym__expression_statement] = STATE(1962),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1962),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1962),
    [sym_delete_statement] = STATE(1962),
    [sym_raise_statement] = STATE(1962),
    [sym_assert_statement] = STATE(1962),
    [sym_yield_statement] = STATE(1962),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(1962),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(1032),
    [sym_break_statement] = ACTIONS(1032),
    [sym_continue_statement] = ACTIONS(1032),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(310)] = {
    [sym__expression_statement] = STATE(1962),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1962),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1962),
    [sym_delete_statement] = STATE(1962),
    [sym_raise_statement] = STATE(1962),
    [sym_assert_statement] = STATE(1962),
    [sym_yield_statement] = STATE(1962),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(1962),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(1032),
    [sym_break_statement] = ACTIONS(1032),
    [sym_continue_statement] = ACTIONS(1032),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(311)] = {
    [sym__expression_statement] = STATE(1962),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1962),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1962),
    [sym_delete_statement] = STATE(1962),
    [sym_raise_statement] = STATE(1962),
    [sym_assert_statement] = STATE(1962),
    [sym_yield_statement] = STATE(1962),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(1962),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(1032),
    [sym_break_statement] = ACTIONS(1032),
    [sym_continue_statement] = ACTIONS(1032),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(312)] = {
    [sym__expression_statement] = STATE(544),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(544),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(544),
    [sym_delete_statement] = STATE(544),
    [sym_raise_statement] = STATE(544),
    [sym_assert_statement] = STATE(544),
    [sym_yield_statement] = STATE(544),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(591),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(591),
    [sym__inline_body] = STATE(591),
    [sym__inline_simple_statement] = STATE(544),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(1034),
    [sym_break_statement] = ACTIONS(1034),
    [sym_continue_statement] = ACTIONS(1034),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(313)] = {
    [sym__expression_statement] = STATE(544),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(544),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(544),
    [sym_delete_statement] = STATE(544),
    [sym_raise_statement] = STATE(544),
    [sym_assert_statement] = STATE(544),
    [sym_yield_statement] = STATE(544),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(616),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(616),
    [sym__inline_body] = STATE(616),
    [sym__inline_simple_statement] = STATE(544),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(1034),
    [sym_break_statement] = ACTIONS(1034),
    [sym_continue_statement] = ACTIONS(1034),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(314)] = {
    [sym__expression_statement] = STATE(544),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(544),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(544),
    [sym_delete_statement] = STATE(544),
    [sym_raise_statement] = STATE(544),
    [sym_assert_statement] = STATE(544),
    [sym_yield_statement] = STATE(544),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(628),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(628),
    [sym__inline_body] = STATE(628),
    [sym__inline_simple_statement] = STATE(544),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(1034),
    [sym_break_statement] = ACTIONS(1034),
    [sym_continue_statement] = ACTIONS(1034),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(315)] = {
    [sym__expression_statement] = STATE(544),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(544),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(544),
    [sym_delete_statement] = STATE(544),
    [sym_raise_statement] = STATE(544),
    [sym_assert_statement] = STATE(544),
    [sym_yield_statement] = STATE(544),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(559),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(559),
    [sym__inline_body] = STATE(559),
    [sym__inline_simple_statement] = STATE(544),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(1034),
    [sym_break_statement] = ACTIONS(1034),
    [sym_continue_statement] = ACTIONS(1034),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(316)] = {
    [sym__expression_statement] = STATE(550),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(550),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(550),
    [sym_delete_statement] = STATE(550),
    [sym_raise_statement] = STATE(550),
    [sym_assert_statement] = STATE(550),
    [sym_yield_statement] = STATE(550),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(610),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(610),
    [sym__inline_body] = STATE(610),
    [sym__inline_simple_statement] = STATE(550),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(1036),
    [sym_break_statement] = ACTIONS(1036),
    [sym_continue_statement] = ACTIONS(1036),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(317)] = {
    [sym__expression_statement] = STATE(550),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(550),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(550),
    [sym_delete_statement] = STATE(550),
    [sym_raise_statement] = STATE(550),
    [sym_assert_statement] = STATE(550),
    [sym_yield_statement] = STATE(550),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(650),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(650),
    [sym__inline_body] = STATE(650),
    [sym__inline_simple_statement] = STATE(550),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(1036),
    [sym_break_statement] = ACTIONS(1036),
    [sym_continue_statement] = ACTIONS(1036),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(318)] = {
    [sym__expression_statement] = STATE(550),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(550),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(550),
    [sym_delete_statement] = STATE(550),
    [sym_raise_statement] = STATE(550),
    [sym_assert_statement] = STATE(550),
    [sym_yield_statement] = STATE(550),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(699),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(699),
    [sym__inline_body] = STATE(699),
    [sym__inline_simple_statement] = STATE(550),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(1036),
    [sym_break_statement] = ACTIONS(1036),
    [sym_continue_statement] = ACTIONS(1036),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(319)] = {
    [sym__expression_statement] = STATE(550),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(550),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(550),
    [sym_delete_statement] = STATE(550),
    [sym_raise_statement] = STATE(550),
    [sym_assert_statement] = STATE(550),
    [sym_yield_statement] = STATE(550),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(701),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(701),
    [sym__inline_body] = STATE(701),
    [sym__inline_simple_statement] = STATE(550),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(1036),
    [sym_break_statement] = ACTIONS(1036),
    [sym_continue_statement] = ACTIONS(1036),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(320)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(537),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2250),
    [sym__inline_body] = STATE(2250),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(745),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(321)] = {
    [sym__expression_statement] = STATE(2562),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(2562),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2562),
    [sym_delete_statement] = STATE(2562),
    [sym_raise_statement] = STATE(2562),
    [sym_assert_statement] = STATE(2562),
    [sym_yield_statement] = STATE(2562),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(2562),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(1044),
    [sym_break_statement] = ACTIONS(1044),
    [sym_continue_statement] = ACTIONS(1044),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(322)] = {
    [sym__expression_statement] = STATE(2562),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(2562),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2562),
    [sym_delete_statement] = STATE(2562),
    [sym_raise_statement] = STATE(2562),
    [sym_assert_statement] = STATE(2562),
    [sym_yield_statement] = STATE(2562),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(2562),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(1044),
    [sym_break_statement] = ACTIONS(1044),
    [sym_continue_statement] = ACTIONS(1044),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(323)] = {
    [sym__expression_statement] = STATE(2562),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(2562),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2562),
    [sym_delete_statement] = STATE(2562),
    [sym_raise_statement] = STATE(2562),
    [sym_assert_statement] = STATE(2562),
    [sym_yield_statement] = STATE(2562),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(2562),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(1044),
    [sym_break_statement] = ACTIONS(1044),
    [sym_continue_statement] = ACTIONS(1044),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(324)] = {
    [sym__expression_statement] = STATE(2562),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(2562),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2562),
    [sym_delete_statement] = STATE(2562),
    [sym_raise_statement] = STATE(2562),
    [sym_assert_statement] = STATE(2562),
    [sym_yield_statement] = STATE(2562),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(2562),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(1044),
    [sym_break_statement] = ACTIONS(1044),
    [sym_continue_statement] = ACTIONS(1044),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(325)] = {
    [sym__expression_statement] = STATE(2237),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2237),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2237),
    [sym_delete_statement] = STATE(2237),
    [sym_raise_statement] = STATE(2237),
    [sym_assert_statement] = STATE(2237),
    [sym_yield_statement] = STATE(2237),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(2237),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(1050),
    [sym_break_statement] = ACTIONS(1050),
    [sym_continue_statement] = ACTIONS(1050),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(326)] = {
    [sym__expression_statement] = STATE(2237),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2237),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2237),
    [sym_delete_statement] = STATE(2237),
    [sym_raise_statement] = STATE(2237),
    [sym_assert_statement] = STATE(2237),
    [sym_yield_statement] = STATE(2237),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(2237),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(1050),
    [sym_break_statement] = ACTIONS(1050),
    [sym_continue_statement] = ACTIONS(1050),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(327)] = {
    [sym__expression_statement] = STATE(2237),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2237),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2237),
    [sym_delete_statement] = STATE(2237),
    [sym_raise_statement] = STATE(2237),
    [sym_assert_statement] = STATE(2237),
    [sym_yield_statement] = STATE(2237),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(2237),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(1050),
    [sym_break_statement] = ACTIONS(1050),
    [sym_continue_statement] = ACTIONS(1050),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(328)] = {
    [sym__expression_statement] = STATE(2237),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2237),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2237),
    [sym_delete_statement] = STATE(2237),
    [sym_raise_statement] = STATE(2237),
    [sym_assert_statement] = STATE(2237),
    [sym_yield_statement] = STATE(2237),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(2237),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(1050),
    [sym_break_statement] = ACTIONS(1050),
    [sym_continue_statement] = ACTIONS(1050),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(329)] = {
    [sym__expression_statement] = STATE(1786),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1786),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1786),
    [sym_delete_statement] = STATE(1786),
    [sym_raise_statement] = STATE(1786),
    [sym_assert_statement] = STATE(1786),
    [sym_yield_statement] = STATE(1786),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(1786),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(1052),
    [sym_break_statement] = ACTIONS(1052),
    [sym_continue_statement] = ACTIONS(1052),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(330)] = {
    [sym__expression_statement] = STATE(1786),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1786),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1786),
    [sym_delete_statement] = STATE(1786),
    [sym_raise_statement] = STATE(1786),
    [sym_assert_statement] = STATE(1786),
    [sym_yield_statement] = STATE(1786),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(1786),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(1052),
    [sym_break_statement] = ACTIONS(1052),
    [sym_continue_statement] = ACTIONS(1052),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(331)] = {
    [sym__expression_statement] = STATE(1786),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1786),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1786),
    [sym_delete_statement] = STATE(1786),
    [sym_raise_statement] = STATE(1786),
    [sym_assert_statement] = STATE(1786),
    [sym_yield_statement] = STATE(1786),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(1786),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(1052),
    [sym_break_statement] = ACTIONS(1052),
    [sym_continue_statement] = ACTIONS(1052),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(332)] = {
    [sym__expression_statement] = STATE(1786),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1786),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1786),
    [sym_delete_statement] = STATE(1786),
    [sym_raise_statement] = STATE(1786),
    [sym_assert_statement] = STATE(1786),
    [sym_yield_statement] = STATE(1786),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(1786),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(1052),
    [sym_break_statement] = ACTIONS(1052),
    [sym_continue_statement] = ACTIONS(1052),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(333)] = {
    [sym__expression_statement] = STATE(1962),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1962),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1962),
    [sym_delete_statement] = STATE(1962),
    [sym_raise_statement] = STATE(1962),
    [sym_assert_statement] = STATE(1962),
    [sym_yield_statement] = STATE(1962),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(1962),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(1032),
    [sym_break_statement] = ACTIONS(1032),
    [sym_continue_statement] = ACTIONS(1032),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(334)] = {
    [sym__expression_statement] = STATE(1962),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1962),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1962),
    [sym_delete_statement] = STATE(1962),
    [sym_raise_statement] = STATE(1962),
    [sym_assert_statement] = STATE(1962),
    [sym_yield_statement] = STATE(1962),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(1962),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(1032),
    [sym_break_statement] = ACTIONS(1032),
    [sym_continue_statement] = ACTIONS(1032),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(335)] = {
    [sym__expression_statement] = STATE(1962),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1962),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1962),
    [sym_delete_statement] = STATE(1962),
    [sym_raise_statement] = STATE(1962),
    [sym_assert_statement] = STATE(1962),
    [sym_yield_statement] = STATE(1962),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(1962),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(1032),
    [sym_break_statement] = ACTIONS(1032),
    [sym_continue_statement] = ACTIONS(1032),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(336)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2220),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2220),
    [sym__inline_body] = STATE(2220),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(337)] = {
    [sym__expression_statement] = STATE(1962),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1962),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1962),
    [sym_delete_statement] = STATE(1962),
    [sym_raise_statement] = STATE(1962),
    [sym_assert_statement] = STATE(1962),
    [sym_yield_statement] = STATE(1962),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(1962),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(1032),
    [sym_break_statement] = ACTIONS(1032),
    [sym_continue_statement] = ACTIONS(1032),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(338)] = {
    [sym__expression_statement] = STATE(2064),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2064),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2064),
    [sym_delete_statement] = STATE(2064),
    [sym_raise_statement] = STATE(2064),
    [sym_assert_statement] = STATE(2064),
    [sym_yield_statement] = STATE(2064),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(2064),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(1054),
    [sym_break_statement] = ACTIONS(1054),
    [sym_continue_statement] = ACTIONS(1054),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(339)] = {
    [sym__expression_statement] = STATE(2064),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2064),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2064),
    [sym_delete_statement] = STATE(2064),
    [sym_raise_statement] = STATE(2064),
    [sym_assert_statement] = STATE(2064),
    [sym_yield_statement] = STATE(2064),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(2064),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(1054),
    [sym_break_statement] = ACTIONS(1054),
    [sym_continue_statement] = ACTIONS(1054),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(340)] = {
    [sym__expression_statement] = STATE(2064),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2064),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2064),
    [sym_delete_statement] = STATE(2064),
    [sym_raise_statement] = STATE(2064),
    [sym_assert_statement] = STATE(2064),
    [sym_yield_statement] = STATE(2064),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(2064),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(1054),
    [sym_break_statement] = ACTIONS(1054),
    [sym_continue_statement] = ACTIONS(1054),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(341)] = {
    [sym__expression_statement] = STATE(2064),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2064),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2064),
    [sym_delete_statement] = STATE(2064),
    [sym_raise_statement] = STATE(2064),
    [sym_assert_statement] = STATE(2064),
    [sym_yield_statement] = STATE(2064),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(2064),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(1054),
    [sym_break_statement] = ACTIONS(1054),
    [sym_continue_statement] = ACTIONS(1054),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(342)] = {
    [sym__expression_statement] = STATE(1795),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1795),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1795),
    [sym_delete_statement] = STATE(1795),
    [sym_raise_statement] = STATE(1795),
    [sym_assert_statement] = STATE(1795),
    [sym_yield_statement] = STATE(1795),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(1795),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(1056),
    [sym_break_statement] = ACTIONS(1056),
    [sym_continue_statement] = ACTIONS(1056),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(343)] = {
    [sym__expression_statement] = STATE(1795),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1795),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1795),
    [sym_delete_statement] = STATE(1795),
    [sym_raise_statement] = STATE(1795),
    [sym_assert_statement] = STATE(1795),
    [sym_yield_statement] = STATE(1795),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(1795),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(1056),
    [sym_break_statement] = ACTIONS(1056),
    [sym_continue_statement] = ACTIONS(1056),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(344)] = {
    [sym__expression_statement] = STATE(1795),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1795),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1795),
    [sym_delete_statement] = STATE(1795),
    [sym_raise_statement] = STATE(1795),
    [sym_assert_statement] = STATE(1795),
    [sym_yield_statement] = STATE(1795),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(1795),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(1056),
    [sym_break_statement] = ACTIONS(1056),
    [sym_continue_statement] = ACTIONS(1056),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(345)] = {
    [sym__expression_statement] = STATE(1795),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1795),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1795),
    [sym_delete_statement] = STATE(1795),
    [sym_raise_statement] = STATE(1795),
    [sym_assert_statement] = STATE(1795),
    [sym_yield_statement] = STATE(1795),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(1795),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(1056),
    [sym_break_statement] = ACTIONS(1056),
    [sym_continue_statement] = ACTIONS(1056),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(346)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2221),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2221),
    [sym__inline_body] = STATE(2221),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(347)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(2216),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2216),
    [sym__inline_body] = STATE(2216),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1012),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(348)] = {
    [sym__expression_statement] = STATE(1786),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1786),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1786),
    [sym_delete_statement] = STATE(1786),
    [sym_raise_statement] = STATE(1786),
    [sym_assert_statement] = STATE(1786),
    [sym_yield_statement] = STATE(1786),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(1786),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(1052),
    [sym_break_statement] = ACTIONS(1052),
    [sym_continue_statement] = ACTIONS(1052),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(349)] = {
    [sym__expression_statement] = STATE(1786),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1786),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1786),
    [sym_delete_statement] = STATE(1786),
    [sym_raise_statement] = STATE(1786),
    [sym_assert_statement] = STATE(1786),
    [sym_yield_statement] = STATE(1786),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(1786),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(1052),
    [sym_break_statement] = ACTIONS(1052),
    [sym_continue_statement] = ACTIONS(1052),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(350)] = {
    [sym__expression_statement] = STATE(1786),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1786),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1786),
    [sym_delete_statement] = STATE(1786),
    [sym_raise_statement] = STATE(1786),
    [sym_assert_statement] = STATE(1786),
    [sym_yield_statement] = STATE(1786),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(1786),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(1052),
    [sym_break_statement] = ACTIONS(1052),
    [sym_continue_statement] = ACTIONS(1052),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(351)] = {
    [sym__expression_statement] = STATE(1786),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1786),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1786),
    [sym_delete_statement] = STATE(1786),
    [sym_raise_statement] = STATE(1786),
    [sym_assert_statement] = STATE(1786),
    [sym_yield_statement] = STATE(1786),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(1786),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(1052),
    [sym_break_statement] = ACTIONS(1052),
    [sym_continue_statement] = ACTIONS(1052),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(352)] = {
    [sym__expression_statement] = STATE(1884),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1884),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1884),
    [sym_delete_statement] = STATE(1884),
    [sym_raise_statement] = STATE(1884),
    [sym_assert_statement] = STATE(1884),
    [sym_yield_statement] = STATE(1884),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(1884),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(1058),
    [sym_break_statement] = ACTIONS(1058),
    [sym_continue_statement] = ACTIONS(1058),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(353)] = {
    [sym__expression_statement] = STATE(1884),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1884),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1884),
    [sym_delete_statement] = STATE(1884),
    [sym_raise_statement] = STATE(1884),
    [sym_assert_statement] = STATE(1884),
    [sym_yield_statement] = STATE(1884),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(1884),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(1058),
    [sym_break_statement] = ACTIONS(1058),
    [sym_continue_statement] = ACTIONS(1058),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(354)] = {
    [sym__expression_statement] = STATE(1884),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1884),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1884),
    [sym_delete_statement] = STATE(1884),
    [sym_raise_statement] = STATE(1884),
    [sym_assert_statement] = STATE(1884),
    [sym_yield_statement] = STATE(1884),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(1884),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(1058),
    [sym_break_statement] = ACTIONS(1058),
    [sym_continue_statement] = ACTIONS(1058),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(355)] = {
    [sym__expression_statement] = STATE(1884),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1884),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1884),
    [sym_delete_statement] = STATE(1884),
    [sym_raise_statement] = STATE(1884),
    [sym_assert_statement] = STATE(1884),
    [sym_yield_statement] = STATE(1884),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(1884),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(1058),
    [sym_break_statement] = ACTIONS(1058),
    [sym_continue_statement] = ACTIONS(1058),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(356)] = {
    [sym__expression_statement] = STATE(2224),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2224),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2224),
    [sym_delete_statement] = STATE(2224),
    [sym_raise_statement] = STATE(2224),
    [sym_assert_statement] = STATE(2224),
    [sym_yield_statement] = STATE(2224),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(2224),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(1060),
    [sym_break_statement] = ACTIONS(1060),
    [sym_continue_statement] = ACTIONS(1060),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(357)] = {
    [sym__expression_statement] = STATE(2224),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2224),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2224),
    [sym_delete_statement] = STATE(2224),
    [sym_raise_statement] = STATE(2224),
    [sym_assert_statement] = STATE(2224),
    [sym_yield_statement] = STATE(2224),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(2224),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(1060),
    [sym_break_statement] = ACTIONS(1060),
    [sym_continue_statement] = ACTIONS(1060),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(358)] = {
    [sym__expression_statement] = STATE(2224),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2224),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2224),
    [sym_delete_statement] = STATE(2224),
    [sym_raise_statement] = STATE(2224),
    [sym_assert_statement] = STATE(2224),
    [sym_yield_statement] = STATE(2224),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(2224),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(1060),
    [sym_break_statement] = ACTIONS(1060),
    [sym_continue_statement] = ACTIONS(1060),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(359)] = {
    [sym__expression_statement] = STATE(2224),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2224),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2224),
    [sym_delete_statement] = STATE(2224),
    [sym_raise_statement] = STATE(2224),
    [sym_assert_statement] = STATE(2224),
    [sym_yield_statement] = STATE(2224),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(2224),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(1060),
    [sym_break_statement] = ACTIONS(1060),
    [sym_continue_statement] = ACTIONS(1060),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(360)] = {
    [sym__expression_statement] = STATE(2562),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(2562),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2562),
    [sym_delete_statement] = STATE(2562),
    [sym_raise_statement] = STATE(2562),
    [sym_assert_statement] = STATE(2562),
    [sym_yield_statement] = STATE(2562),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(2562),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(1044),
    [sym_break_statement] = ACTIONS(1044),
    [sym_continue_statement] = ACTIONS(1044),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(361)] = {
    [sym__expression_statement] = STATE(1791),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1791),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1791),
    [sym_delete_statement] = STATE(1791),
    [sym_raise_statement] = STATE(1791),
    [sym_assert_statement] = STATE(1791),
    [sym_yield_statement] = STATE(1791),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(1791),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(1030),
    [sym_break_statement] = ACTIONS(1030),
    [sym_continue_statement] = ACTIONS(1030),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(362)] = {
    [sym__expression_statement] = STATE(1791),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1791),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1791),
    [sym_delete_statement] = STATE(1791),
    [sym_raise_statement] = STATE(1791),
    [sym_assert_statement] = STATE(1791),
    [sym_yield_statement] = STATE(1791),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(1791),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(1030),
    [sym_break_statement] = ACTIONS(1030),
    [sym_continue_statement] = ACTIONS(1030),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(363)] = {
    [sym__expression_statement] = STATE(1791),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1791),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1791),
    [sym_delete_statement] = STATE(1791),
    [sym_raise_statement] = STATE(1791),
    [sym_assert_statement] = STATE(1791),
    [sym_yield_statement] = STATE(1791),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(1791),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(1030),
    [sym_break_statement] = ACTIONS(1030),
    [sym_continue_statement] = ACTIONS(1030),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(364)] = {
    [sym__expression_statement] = STATE(1791),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1791),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1791),
    [sym_delete_statement] = STATE(1791),
    [sym_raise_statement] = STATE(1791),
    [sym_assert_statement] = STATE(1791),
    [sym_yield_statement] = STATE(1791),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(1791),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(1030),
    [sym_break_statement] = ACTIONS(1030),
    [sym_continue_statement] = ACTIONS(1030),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(365)] = {
    [sym__expression_statement] = STATE(2058),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2058),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2058),
    [sym_delete_statement] = STATE(2058),
    [sym_raise_statement] = STATE(2058),
    [sym_assert_statement] = STATE(2058),
    [sym_yield_statement] = STATE(2058),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(2058),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(1062),
    [sym_break_statement] = ACTIONS(1062),
    [sym_continue_statement] = ACTIONS(1062),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(366)] = {
    [sym__expression_statement] = STATE(2058),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2058),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2058),
    [sym_delete_statement] = STATE(2058),
    [sym_raise_statement] = STATE(2058),
    [sym_assert_statement] = STATE(2058),
    [sym_yield_statement] = STATE(2058),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(2058),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(1062),
    [sym_break_statement] = ACTIONS(1062),
    [sym_continue_statement] = ACTIONS(1062),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(367)] = {
    [sym__expression_statement] = STATE(2058),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2058),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2058),
    [sym_delete_statement] = STATE(2058),
    [sym_raise_statement] = STATE(2058),
    [sym_assert_statement] = STATE(2058),
    [sym_yield_statement] = STATE(2058),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(2058),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(1062),
    [sym_break_statement] = ACTIONS(1062),
    [sym_continue_statement] = ACTIONS(1062),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(368)] = {
    [sym__expression_statement] = STATE(2058),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2058),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2058),
    [sym_delete_statement] = STATE(2058),
    [sym_raise_statement] = STATE(2058),
    [sym_assert_statement] = STATE(2058),
    [sym_yield_statement] = STATE(2058),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(2058),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(1062),
    [sym_break_statement] = ACTIONS(1062),
    [sym_continue_statement] = ACTIONS(1062),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(369)] = {
    [sym__expression_statement] = STATE(2053),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2053),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2053),
    [sym_delete_statement] = STATE(2053),
    [sym_raise_statement] = STATE(2053),
    [sym_assert_statement] = STATE(2053),
    [sym_yield_statement] = STATE(2053),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(2053),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(1064),
    [sym_break_statement] = ACTIONS(1064),
    [sym_continue_statement] = ACTIONS(1064),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(370)] = {
    [sym__expression_statement] = STATE(2053),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2053),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2053),
    [sym_delete_statement] = STATE(2053),
    [sym_raise_statement] = STATE(2053),
    [sym_assert_statement] = STATE(2053),
    [sym_yield_statement] = STATE(2053),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(2053),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(1064),
    [sym_break_statement] = ACTIONS(1064),
    [sym_continue_statement] = ACTIONS(1064),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(371)] = {
    [sym__expression_statement] = STATE(2053),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2053),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2053),
    [sym_delete_statement] = STATE(2053),
    [sym_raise_statement] = STATE(2053),
    [sym_assert_statement] = STATE(2053),
    [sym_yield_statement] = STATE(2053),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(2053),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(1064),
    [sym_break_statement] = ACTIONS(1064),
    [sym_continue_statement] = ACTIONS(1064),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(372)] = {
    [sym__expression_statement] = STATE(2053),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2053),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2053),
    [sym_delete_statement] = STATE(2053),
    [sym_raise_statement] = STATE(2053),
    [sym_assert_statement] = STATE(2053),
    [sym_yield_statement] = STATE(2053),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(2053),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(1064),
    [sym_break_statement] = ACTIONS(1064),
    [sym_continue_statement] = ACTIONS(1064),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(373)] = {
    [sym__expression_statement] = STATE(1795),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1795),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1795),
    [sym_delete_statement] = STATE(1795),
    [sym_raise_statement] = STATE(1795),
    [sym_assert_statement] = STATE(1795),
    [sym_yield_statement] = STATE(1795),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(1795),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(1056),
    [sym_break_statement] = ACTIONS(1056),
    [sym_continue_statement] = ACTIONS(1056),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(374)] = {
    [sym__expression_statement] = STATE(1795),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1795),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1795),
    [sym_delete_statement] = STATE(1795),
    [sym_raise_statement] = STATE(1795),
    [sym_assert_statement] = STATE(1795),
    [sym_yield_statement] = STATE(1795),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(1795),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(1056),
    [sym_break_statement] = ACTIONS(1056),
    [sym_continue_statement] = ACTIONS(1056),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(375)] = {
    [sym__expression_statement] = STATE(1795),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1795),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1795),
    [sym_delete_statement] = STATE(1795),
    [sym_raise_statement] = STATE(1795),
    [sym_assert_statement] = STATE(1795),
    [sym_yield_statement] = STATE(1795),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(1795),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(1056),
    [sym_break_statement] = ACTIONS(1056),
    [sym_continue_statement] = ACTIONS(1056),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(376)] = {
    [sym__expression_statement] = STATE(1795),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1795),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1795),
    [sym_delete_statement] = STATE(1795),
    [sym_raise_statement] = STATE(1795),
    [sym_assert_statement] = STATE(1795),
    [sym_yield_statement] = STATE(1795),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(1795),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(1056),
    [sym_break_statement] = ACTIONS(1056),
    [sym_continue_statement] = ACTIONS(1056),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(377)] = {
    [sym__expression_statement] = STATE(2045),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2045),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2045),
    [sym_delete_statement] = STATE(2045),
    [sym_raise_statement] = STATE(2045),
    [sym_assert_statement] = STATE(2045),
    [sym_yield_statement] = STATE(2045),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(2045),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(1066),
    [sym_break_statement] = ACTIONS(1066),
    [sym_continue_statement] = ACTIONS(1066),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(378)] = {
    [sym__expression_statement] = STATE(2045),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2045),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2045),
    [sym_delete_statement] = STATE(2045),
    [sym_raise_statement] = STATE(2045),
    [sym_assert_statement] = STATE(2045),
    [sym_yield_statement] = STATE(2045),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(2045),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(1066),
    [sym_break_statement] = ACTIONS(1066),
    [sym_continue_statement] = ACTIONS(1066),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(379)] = {
    [sym__expression_statement] = STATE(2045),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2045),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2045),
    [sym_delete_statement] = STATE(2045),
    [sym_raise_statement] = STATE(2045),
    [sym_assert_statement] = STATE(2045),
    [sym_yield_statement] = STATE(2045),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(2045),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(1066),
    [sym_break_statement] = ACTIONS(1066),
    [sym_continue_statement] = ACTIONS(1066),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(380)] = {
    [sym__expression_statement] = STATE(2045),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2045),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2045),
    [sym_delete_statement] = STATE(2045),
    [sym_raise_statement] = STATE(2045),
    [sym_assert_statement] = STATE(2045),
    [sym_yield_statement] = STATE(2045),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(2045),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(1066),
    [sym_break_statement] = ACTIONS(1066),
    [sym_continue_statement] = ACTIONS(1066),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(381)] = {
    [sym__expression_statement] = STATE(1720),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1720),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1720),
    [sym_delete_statement] = STATE(1720),
    [sym_raise_statement] = STATE(1720),
    [sym_assert_statement] = STATE(1720),
    [sym_yield_statement] = STATE(1720),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(1720),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(1016),
    [sym_break_statement] = ACTIONS(1016),
    [sym_continue_statement] = ACTIONS(1016),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(382)] = {
    [sym__expression_statement] = STATE(2042),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(2042),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2042),
    [sym_delete_statement] = STATE(2042),
    [sym_raise_statement] = STATE(2042),
    [sym_assert_statement] = STATE(2042),
    [sym_yield_statement] = STATE(2042),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1672),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1672),
    [sym__inline_body] = STATE(1672),
    [sym__inline_simple_statement] = STATE(2042),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(1014),
    [sym_break_statement] = ACTIONS(1014),
    [sym_continue_statement] = ACTIONS(1014),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(383)] = {
    [sym__expression_statement] = STATE(2224),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2224),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2224),
    [sym_delete_statement] = STATE(2224),
    [sym_raise_statement] = STATE(2224),
    [sym_assert_statement] = STATE(2224),
    [sym_yield_statement] = STATE(2224),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(2224),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(1060),
    [sym_break_statement] = ACTIONS(1060),
    [sym_continue_statement] = ACTIONS(1060),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(384)] = {
    [sym__expression_statement] = STATE(2224),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2224),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2224),
    [sym_delete_statement] = STATE(2224),
    [sym_raise_statement] = STATE(2224),
    [sym_assert_statement] = STATE(2224),
    [sym_yield_statement] = STATE(2224),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(2224),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(1060),
    [sym_break_statement] = ACTIONS(1060),
    [sym_continue_statement] = ACTIONS(1060),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(385)] = {
    [sym__expression_statement] = STATE(2224),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2224),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2224),
    [sym_delete_statement] = STATE(2224),
    [sym_raise_statement] = STATE(2224),
    [sym_assert_statement] = STATE(2224),
    [sym_yield_statement] = STATE(2224),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(2224),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(1060),
    [sym_break_statement] = ACTIONS(1060),
    [sym_continue_statement] = ACTIONS(1060),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(386)] = {
    [sym__expression_statement] = STATE(2224),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2224),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2224),
    [sym_delete_statement] = STATE(2224),
    [sym_raise_statement] = STATE(2224),
    [sym_assert_statement] = STATE(2224),
    [sym_yield_statement] = STATE(2224),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(2224),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(1060),
    [sym_break_statement] = ACTIONS(1060),
    [sym_continue_statement] = ACTIONS(1060),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(387)] = {
    [sym__expression_statement] = STATE(2237),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2237),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2237),
    [sym_delete_statement] = STATE(2237),
    [sym_raise_statement] = STATE(2237),
    [sym_assert_statement] = STATE(2237),
    [sym_yield_statement] = STATE(2237),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(2237),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(1050),
    [sym_break_statement] = ACTIONS(1050),
    [sym_continue_statement] = ACTIONS(1050),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(388)] = {
    [sym__expression_statement] = STATE(2237),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2237),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2237),
    [sym_delete_statement] = STATE(2237),
    [sym_raise_statement] = STATE(2237),
    [sym_assert_statement] = STATE(2237),
    [sym_yield_statement] = STATE(2237),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(2237),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(1050),
    [sym_break_statement] = ACTIONS(1050),
    [sym_continue_statement] = ACTIONS(1050),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(389)] = {
    [sym__expression_statement] = STATE(2237),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2237),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2237),
    [sym_delete_statement] = STATE(2237),
    [sym_raise_statement] = STATE(2237),
    [sym_assert_statement] = STATE(2237),
    [sym_yield_statement] = STATE(2237),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(2237),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(1050),
    [sym_break_statement] = ACTIONS(1050),
    [sym_continue_statement] = ACTIONS(1050),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(390)] = {
    [sym__expression_statement] = STATE(2237),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2237),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2237),
    [sym_delete_statement] = STATE(2237),
    [sym_raise_statement] = STATE(2237),
    [sym_assert_statement] = STATE(2237),
    [sym_yield_statement] = STATE(2237),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(2237),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(1050),
    [sym_break_statement] = ACTIONS(1050),
    [sym_continue_statement] = ACTIONS(1050),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(391)] = {
    [sym__expression_statement] = STATE(2064),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2064),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2064),
    [sym_delete_statement] = STATE(2064),
    [sym_raise_statement] = STATE(2064),
    [sym_assert_statement] = STATE(2064),
    [sym_yield_statement] = STATE(2064),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(2064),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(1054),
    [sym_break_statement] = ACTIONS(1054),
    [sym_continue_statement] = ACTIONS(1054),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(392)] = {
    [sym__expression_statement] = STATE(2064),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2064),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2064),
    [sym_delete_statement] = STATE(2064),
    [sym_raise_statement] = STATE(2064),
    [sym_assert_statement] = STATE(2064),
    [sym_yield_statement] = STATE(2064),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(2064),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(1054),
    [sym_break_statement] = ACTIONS(1054),
    [sym_continue_statement] = ACTIONS(1054),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(393)] = {
    [sym__expression_statement] = STATE(2064),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2064),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2064),
    [sym_delete_statement] = STATE(2064),
    [sym_raise_statement] = STATE(2064),
    [sym_assert_statement] = STATE(2064),
    [sym_yield_statement] = STATE(2064),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(2064),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(1054),
    [sym_break_statement] = ACTIONS(1054),
    [sym_continue_statement] = ACTIONS(1054),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(394)] = {
    [sym__expression_statement] = STATE(2064),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(2064),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2064),
    [sym_delete_statement] = STATE(2064),
    [sym_raise_statement] = STATE(2064),
    [sym_assert_statement] = STATE(2064),
    [sym_yield_statement] = STATE(2064),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(2064),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(1054),
    [sym_break_statement] = ACTIONS(1054),
    [sym_continue_statement] = ACTIONS(1054),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(395)] = {
    [sym__expression_statement] = STATE(1884),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1884),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1884),
    [sym_delete_statement] = STATE(1884),
    [sym_raise_statement] = STATE(1884),
    [sym_assert_statement] = STATE(1884),
    [sym_yield_statement] = STATE(1884),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(1884),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(1058),
    [sym_break_statement] = ACTIONS(1058),
    [sym_continue_statement] = ACTIONS(1058),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(396)] = {
    [sym__expression_statement] = STATE(1884),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1884),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1884),
    [sym_delete_statement] = STATE(1884),
    [sym_raise_statement] = STATE(1884),
    [sym_assert_statement] = STATE(1884),
    [sym_yield_statement] = STATE(1884),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(1884),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(1058),
    [sym_break_statement] = ACTIONS(1058),
    [sym_continue_statement] = ACTIONS(1058),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(397)] = {
    [sym__expression_statement] = STATE(1884),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1884),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1884),
    [sym_delete_statement] = STATE(1884),
    [sym_raise_statement] = STATE(1884),
    [sym_assert_statement] = STATE(1884),
    [sym_yield_statement] = STATE(1884),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(1884),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(1058),
    [sym_break_statement] = ACTIONS(1058),
    [sym_continue_statement] = ACTIONS(1058),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(398)] = {
    [sym__expression_statement] = STATE(1884),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1884),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1884),
    [sym_delete_statement] = STATE(1884),
    [sym_raise_statement] = STATE(1884),
    [sym_assert_statement] = STATE(1884),
    [sym_yield_statement] = STATE(1884),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(1884),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(1058),
    [sym_break_statement] = ACTIONS(1058),
    [sym_continue_statement] = ACTIONS(1058),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(399)] = {
    [sym__expression_statement] = STATE(2562),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(2562),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2562),
    [sym_delete_statement] = STATE(2562),
    [sym_raise_statement] = STATE(2562),
    [sym_assert_statement] = STATE(2562),
    [sym_yield_statement] = STATE(2562),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(2562),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(1044),
    [sym_break_statement] = ACTIONS(1044),
    [sym_continue_statement] = ACTIONS(1044),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(400)] = {
    [sym__expression_statement] = STATE(2058),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2058),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2058),
    [sym_delete_statement] = STATE(2058),
    [sym_raise_statement] = STATE(2058),
    [sym_assert_statement] = STATE(2058),
    [sym_yield_statement] = STATE(2058),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(2058),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(1062),
    [sym_break_statement] = ACTIONS(1062),
    [sym_continue_statement] = ACTIONS(1062),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(401)] = {
    [sym__expression_statement] = STATE(2058),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2058),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2058),
    [sym_delete_statement] = STATE(2058),
    [sym_raise_statement] = STATE(2058),
    [sym_assert_statement] = STATE(2058),
    [sym_yield_statement] = STATE(2058),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(2058),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(1062),
    [sym_break_statement] = ACTIONS(1062),
    [sym_continue_statement] = ACTIONS(1062),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(402)] = {
    [sym__expression_statement] = STATE(2058),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2058),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2058),
    [sym_delete_statement] = STATE(2058),
    [sym_raise_statement] = STATE(2058),
    [sym_assert_statement] = STATE(2058),
    [sym_yield_statement] = STATE(2058),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(2058),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(1062),
    [sym_break_statement] = ACTIONS(1062),
    [sym_continue_statement] = ACTIONS(1062),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(403)] = {
    [sym__expression_statement] = STATE(2058),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2058),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2058),
    [sym_delete_statement] = STATE(2058),
    [sym_raise_statement] = STATE(2058),
    [sym_assert_statement] = STATE(2058),
    [sym_yield_statement] = STATE(2058),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(2058),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(1062),
    [sym_break_statement] = ACTIONS(1062),
    [sym_continue_statement] = ACTIONS(1062),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(404)] = {
    [sym__expression_statement] = STATE(2045),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2045),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2045),
    [sym_delete_statement] = STATE(2045),
    [sym_raise_statement] = STATE(2045),
    [sym_assert_statement] = STATE(2045),
    [sym_yield_statement] = STATE(2045),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(2045),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(1066),
    [sym_break_statement] = ACTIONS(1066),
    [sym_continue_statement] = ACTIONS(1066),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(405)] = {
    [sym__expression_statement] = STATE(2045),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2045),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2045),
    [sym_delete_statement] = STATE(2045),
    [sym_raise_statement] = STATE(2045),
    [sym_assert_statement] = STATE(2045),
    [sym_yield_statement] = STATE(2045),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(2045),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(1066),
    [sym_break_statement] = ACTIONS(1066),
    [sym_continue_statement] = ACTIONS(1066),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(406)] = {
    [sym__expression_statement] = STATE(2045),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2045),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2045),
    [sym_delete_statement] = STATE(2045),
    [sym_raise_statement] = STATE(2045),
    [sym_assert_statement] = STATE(2045),
    [sym_yield_statement] = STATE(2045),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(2045),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(1066),
    [sym_break_statement] = ACTIONS(1066),
    [sym_continue_statement] = ACTIONS(1066),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(407)] = {
    [sym__expression_statement] = STATE(2045),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(2045),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2045),
    [sym_delete_statement] = STATE(2045),
    [sym_raise_statement] = STATE(2045),
    [sym_assert_statement] = STATE(2045),
    [sym_yield_statement] = STATE(2045),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(2045),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(1066),
    [sym_break_statement] = ACTIONS(1066),
    [sym_continue_statement] = ACTIONS(1066),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(408)] = {
    [sym__expression_statement] = STATE(2042),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(2042),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2042),
    [sym_delete_statement] = STATE(2042),
    [sym_raise_statement] = STATE(2042),
    [sym_assert_statement] = STATE(2042),
    [sym_yield_statement] = STATE(2042),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(2042),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(1014),
    [sym_break_statement] = ACTIONS(1014),
    [sym_continue_statement] = ACTIONS(1014),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(409)] = {
    [sym__expression_statement] = STATE(2042),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(2042),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2042),
    [sym_delete_statement] = STATE(2042),
    [sym_raise_statement] = STATE(2042),
    [sym_assert_statement] = STATE(2042),
    [sym_yield_statement] = STATE(2042),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(2042),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(1014),
    [sym_break_statement] = ACTIONS(1014),
    [sym_continue_statement] = ACTIONS(1014),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(410)] = {
    [sym__expression_statement] = STATE(2042),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(2042),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2042),
    [sym_delete_statement] = STATE(2042),
    [sym_raise_statement] = STATE(2042),
    [sym_assert_statement] = STATE(2042),
    [sym_yield_statement] = STATE(2042),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(2042),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(1014),
    [sym_break_statement] = ACTIONS(1014),
    [sym_continue_statement] = ACTIONS(1014),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(411)] = {
    [sym__expression_statement] = STATE(2042),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(2042),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2042),
    [sym_delete_statement] = STATE(2042),
    [sym_raise_statement] = STATE(2042),
    [sym_assert_statement] = STATE(2042),
    [sym_yield_statement] = STATE(2042),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(2042),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(1014),
    [sym_break_statement] = ACTIONS(1014),
    [sym_continue_statement] = ACTIONS(1014),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(412)] = {
    [sym__expression_statement] = STATE(2053),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2053),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2053),
    [sym_delete_statement] = STATE(2053),
    [sym_raise_statement] = STATE(2053),
    [sym_assert_statement] = STATE(2053),
    [sym_yield_statement] = STATE(2053),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1701),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1701),
    [sym__inline_body] = STATE(1701),
    [sym__inline_simple_statement] = STATE(2053),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(1064),
    [sym_break_statement] = ACTIONS(1064),
    [sym_continue_statement] = ACTIONS(1064),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(413)] = {
    [sym__expression_statement] = STATE(2053),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2053),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2053),
    [sym_delete_statement] = STATE(2053),
    [sym_raise_statement] = STATE(2053),
    [sym_assert_statement] = STATE(2053),
    [sym_yield_statement] = STATE(2053),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1669),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1669),
    [sym__inline_body] = STATE(1669),
    [sym__inline_simple_statement] = STATE(2053),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(1064),
    [sym_break_statement] = ACTIONS(1064),
    [sym_continue_statement] = ACTIONS(1064),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(414)] = {
    [sym__expression_statement] = STATE(2053),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2053),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2053),
    [sym_delete_statement] = STATE(2053),
    [sym_raise_statement] = STATE(2053),
    [sym_assert_statement] = STATE(2053),
    [sym_yield_statement] = STATE(2053),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1673),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1673),
    [sym__inline_body] = STATE(1673),
    [sym__inline_simple_statement] = STATE(2053),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(1064),
    [sym_break_statement] = ACTIONS(1064),
    [sym_continue_statement] = ACTIONS(1064),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(415)] = {
    [sym__expression_statement] = STATE(2053),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(2053),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2053),
    [sym_delete_statement] = STATE(2053),
    [sym_raise_statement] = STATE(2053),
    [sym_assert_statement] = STATE(2053),
    [sym_yield_statement] = STATE(2053),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1689),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1689),
    [sym__inline_body] = STATE(1689),
    [sym__inline_simple_statement] = STATE(2053),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(1064),
    [sym_break_statement] = ACTIONS(1064),
    [sym_continue_statement] = ACTIONS(1064),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(416)] = {
    [sym__expression_statement] = STATE(563),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(563),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(563),
    [sym_delete_statement] = STATE(563),
    [sym_raise_statement] = STATE(563),
    [sym_assert_statement] = STATE(563),
    [sym_yield_statement] = STATE(563),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(562),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(562),
    [sym__inline_body] = STATE(562),
    [sym__inline_simple_statement] = STATE(563),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(1018),
    [sym_break_statement] = ACTIONS(1018),
    [sym_continue_statement] = ACTIONS(1018),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(417)] = {
    [sym__expression_statement] = STATE(563),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(563),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(563),
    [sym_delete_statement] = STATE(563),
    [sym_raise_statement] = STATE(563),
    [sym_assert_statement] = STATE(563),
    [sym_yield_statement] = STATE(563),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(589),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(589),
    [sym__inline_body] = STATE(589),
    [sym__inline_simple_statement] = STATE(563),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(1018),
    [sym_break_statement] = ACTIONS(1018),
    [sym_continue_statement] = ACTIONS(1018),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(418)] = {
    [sym__expression_statement] = STATE(563),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(563),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(563),
    [sym_delete_statement] = STATE(563),
    [sym_raise_statement] = STATE(563),
    [sym_assert_statement] = STATE(563),
    [sym_yield_statement] = STATE(563),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(617),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(617),
    [sym__inline_body] = STATE(617),
    [sym__inline_simple_statement] = STATE(563),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(1018),
    [sym_break_statement] = ACTIONS(1018),
    [sym_continue_statement] = ACTIONS(1018),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(419)] = {
    [sym__expression_statement] = STATE(563),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(563),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(563),
    [sym_delete_statement] = STATE(563),
    [sym_raise_statement] = STATE(563),
    [sym_assert_statement] = STATE(563),
    [sym_yield_statement] = STATE(563),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(627),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(627),
    [sym__inline_body] = STATE(627),
    [sym__inline_simple_statement] = STATE(563),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(1018),
    [sym_break_statement] = ACTIONS(1018),
    [sym_continue_statement] = ACTIONS(1018),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(420)] = {
    [sym__expression_statement] = STATE(544),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(544),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(544),
    [sym_delete_statement] = STATE(544),
    [sym_raise_statement] = STATE(544),
    [sym_assert_statement] = STATE(544),
    [sym_yield_statement] = STATE(544),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(562),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(562),
    [sym__inline_body] = STATE(562),
    [sym__inline_simple_statement] = STATE(544),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(1034),
    [sym_break_statement] = ACTIONS(1034),
    [sym_continue_statement] = ACTIONS(1034),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(421)] = {
    [sym__expression_statement] = STATE(544),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(544),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(544),
    [sym_delete_statement] = STATE(544),
    [sym_raise_statement] = STATE(544),
    [sym_assert_statement] = STATE(544),
    [sym_yield_statement] = STATE(544),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(589),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(589),
    [sym__inline_body] = STATE(589),
    [sym__inline_simple_statement] = STATE(544),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(1034),
    [sym_break_statement] = ACTIONS(1034),
    [sym_continue_statement] = ACTIONS(1034),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(422)] = {
    [sym__expression_statement] = STATE(544),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(544),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(544),
    [sym_delete_statement] = STATE(544),
    [sym_raise_statement] = STATE(544),
    [sym_assert_statement] = STATE(544),
    [sym_yield_statement] = STATE(544),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(617),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(617),
    [sym__inline_body] = STATE(617),
    [sym__inline_simple_statement] = STATE(544),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(1034),
    [sym_break_statement] = ACTIONS(1034),
    [sym_continue_statement] = ACTIONS(1034),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(423)] = {
    [sym__expression_statement] = STATE(544),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(544),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(544),
    [sym_delete_statement] = STATE(544),
    [sym_raise_statement] = STATE(544),
    [sym_assert_statement] = STATE(544),
    [sym_yield_statement] = STATE(544),
    [sym_yield] = STATE(668),
    [sym_block] = STATE(627),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__anon_suite] = STATE(627),
    [sym__inline_body] = STATE(627),
    [sym__inline_simple_statement] = STATE(544),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(1034),
    [sym_break_statement] = ACTIONS(1034),
    [sym_continue_statement] = ACTIONS(1034),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1020),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(424)] = {
    [sym__expression_statement] = STATE(611),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(611),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(611),
    [sym_delete_statement] = STATE(611),
    [sym_raise_statement] = STATE(611),
    [sym_assert_statement] = STATE(611),
    [sym_yield_statement] = STATE(611),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(607),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(607),
    [sym__inline_body] = STATE(607),
    [sym__inline_simple_statement] = STATE(611),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(1022),
    [sym_break_statement] = ACTIONS(1022),
    [sym_continue_statement] = ACTIONS(1022),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(425)] = {
    [sym__expression_statement] = STATE(611),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(611),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(611),
    [sym_delete_statement] = STATE(611),
    [sym_raise_statement] = STATE(611),
    [sym_assert_statement] = STATE(611),
    [sym_yield_statement] = STATE(611),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(595),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(595),
    [sym__inline_body] = STATE(595),
    [sym__inline_simple_statement] = STATE(611),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(1022),
    [sym_break_statement] = ACTIONS(1022),
    [sym_continue_statement] = ACTIONS(1022),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(426)] = {
    [sym__expression_statement] = STATE(611),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(611),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(611),
    [sym_delete_statement] = STATE(611),
    [sym_raise_statement] = STATE(611),
    [sym_assert_statement] = STATE(611),
    [sym_yield_statement] = STATE(611),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(654),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(654),
    [sym__inline_body] = STATE(654),
    [sym__inline_simple_statement] = STATE(611),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(1022),
    [sym_break_statement] = ACTIONS(1022),
    [sym_continue_statement] = ACTIONS(1022),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(427)] = {
    [sym__expression_statement] = STATE(611),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(611),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(611),
    [sym_delete_statement] = STATE(611),
    [sym_raise_statement] = STATE(611),
    [sym_assert_statement] = STATE(611),
    [sym_yield_statement] = STATE(611),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(698),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(698),
    [sym__inline_body] = STATE(698),
    [sym__inline_simple_statement] = STATE(611),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(1022),
    [sym_break_statement] = ACTIONS(1022),
    [sym_continue_statement] = ACTIONS(1022),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(428)] = {
    [sym__expression_statement] = STATE(550),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(550),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(550),
    [sym_delete_statement] = STATE(550),
    [sym_raise_statement] = STATE(550),
    [sym_assert_statement] = STATE(550),
    [sym_yield_statement] = STATE(550),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(607),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(607),
    [sym__inline_body] = STATE(607),
    [sym__inline_simple_statement] = STATE(550),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(1036),
    [sym_break_statement] = ACTIONS(1036),
    [sym_continue_statement] = ACTIONS(1036),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(429)] = {
    [sym__expression_statement] = STATE(550),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(550),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(550),
    [sym_delete_statement] = STATE(550),
    [sym_raise_statement] = STATE(550),
    [sym_assert_statement] = STATE(550),
    [sym_yield_statement] = STATE(550),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(595),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(595),
    [sym__inline_body] = STATE(595),
    [sym__inline_simple_statement] = STATE(550),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(1036),
    [sym_break_statement] = ACTIONS(1036),
    [sym_continue_statement] = ACTIONS(1036),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(430)] = {
    [sym__expression_statement] = STATE(550),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(550),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(550),
    [sym_delete_statement] = STATE(550),
    [sym_raise_statement] = STATE(550),
    [sym_assert_statement] = STATE(550),
    [sym_yield_statement] = STATE(550),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(654),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(654),
    [sym__inline_body] = STATE(654),
    [sym__inline_simple_statement] = STATE(550),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(1036),
    [sym_break_statement] = ACTIONS(1036),
    [sym_continue_statement] = ACTIONS(1036),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(431)] = {
    [sym__expression_statement] = STATE(550),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(550),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(550),
    [sym_delete_statement] = STATE(550),
    [sym_raise_statement] = STATE(550),
    [sym_assert_statement] = STATE(550),
    [sym_yield_statement] = STATE(550),
    [sym_yield] = STATE(651),
    [sym_block] = STATE(698),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__anon_suite] = STATE(698),
    [sym__inline_body] = STATE(698),
    [sym__inline_simple_statement] = STATE(550),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(1036),
    [sym_break_statement] = ACTIONS(1036),
    [sym_continue_statement] = ACTIONS(1036),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1024),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(432)] = {
    [sym__expression_statement] = STATE(2562),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(2562),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2562),
    [sym_delete_statement] = STATE(2562),
    [sym_raise_statement] = STATE(2562),
    [sym_assert_statement] = STATE(2562),
    [sym_yield_statement] = STATE(2562),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1690),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1690),
    [sym__inline_body] = STATE(1690),
    [sym__inline_simple_statement] = STATE(2562),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(1044),
    [sym_break_statement] = ACTIONS(1044),
    [sym_continue_statement] = ACTIONS(1044),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(433)] = {
    [sym__expression_statement] = STATE(2562),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(2562),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2562),
    [sym_delete_statement] = STATE(2562),
    [sym_raise_statement] = STATE(2562),
    [sym_assert_statement] = STATE(2562),
    [sym_yield_statement] = STATE(2562),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1674),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1674),
    [sym__inline_body] = STATE(1674),
    [sym__inline_simple_statement] = STATE(2562),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(1044),
    [sym_break_statement] = ACTIONS(1044),
    [sym_continue_statement] = ACTIONS(1044),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(434)] = {
    [sym__expression_statement] = STATE(2042),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(2042),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2042),
    [sym_delete_statement] = STATE(2042),
    [sym_raise_statement] = STATE(2042),
    [sym_assert_statement] = STATE(2042),
    [sym_yield_statement] = STATE(2042),
    [sym_yield] = STATE(1661),
    [sym_block] = STATE(1676),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__anon_suite] = STATE(1676),
    [sym__inline_body] = STATE(1676),
    [sym__inline_simple_statement] = STATE(2042),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(1014),
    [sym_break_statement] = ACTIONS(1014),
    [sym_continue_statement] = ACTIONS(1014),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(867),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(435)] = {
    [sym__expression_statement] = STATE(2015),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2015),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2015),
    [sym_delete_statement] = STATE(2015),
    [sym_raise_statement] = STATE(2015),
    [sym_assert_statement] = STATE(2015),
    [sym_yield_statement] = STATE(2015),
    [sym_yield] = STATE(2272),
    [sym_block] = STATE(515),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__anon_suite] = STATE(2169),
    [sym__inline_body] = STATE(2169),
    [sym__inline_simple_statement] = STATE(2015),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(1010),
    [sym_break_statement] = ACTIONS(1010),
    [sym_continue_statement] = ACTIONS(1010),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(745),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(436)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(437)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_DOT] = ACTIONS(119),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_QMARK] = ACTIONS(119),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(438)] = {
    [sym_f_string] = STATE(442),
    [aux_sym_concatenated_string_repeat1] = STATE(442),
    [ts_builtin_sym_end] = ACTIONS(1068),
    [sym_identifier] = ACTIONS(1070),
    [anon_sym_SEMI] = ACTIONS(1070),
    [anon_sym_return] = ACTIONS(1070),
    [anon_sym_async] = ACTIONS(1070),
    [anon_sym_def] = ACTIONS(1070),
    [anon_sym_COMMA] = ACTIONS(1070),
    [anon_sym_LPAREN] = ACTIONS(1070),
    [anon_sym_LBRACK] = ACTIONS(1070),
    [anon_sym_del] = ACTIONS(1070),
    [anon_sym_raise] = ACTIONS(1070),
    [sym_pass_statement] = ACTIONS(1070),
    [sym_break_statement] = ACTIONS(1070),
    [sym_continue_statement] = ACTIONS(1070),
    [sym_debugger_statement] = ACTIONS(1070),
    [anon_sym_assert] = ACTIONS(1070),
    [anon_sym_global] = ACTIONS(1070),
    [anon_sym_nonlocal] = ACTIONS(1070),
    [anon_sym_yield] = ACTIONS(1070),
    [anon_sym_from] = ACTIONS(1070),
    [anon_sym_import] = ACTIONS(1070),
    [anon_sym_DOT] = ACTIONS(1070),
    [anon_sym_if] = ACTIONS(1070),
    [anon_sym_for] = ACTIONS(1070),
    [anon_sym_in] = ACTIONS(1070),
    [anon_sym_while] = ACTIONS(1070),
    [anon_sym_do] = ACTIONS(1070),
    [anon_sym_try] = ACTIONS(1070),
    [anon_sym_with] = ACTIONS(1070),
    [anon_sym_STAR] = ACTIONS(1070),
    [anon_sym_STAR_STAR] = ACTIONS(1070),
    [anon_sym_class] = ACTIONS(1070),
    [anon_sym_AT] = ACTIONS(1070),
    [anon_sym_not] = ACTIONS(1070),
    [anon_sym_and] = ACTIONS(1070),
    [anon_sym_or] = ACTIONS(1070),
    [anon_sym_PLUS] = ACTIONS(1070),
    [anon_sym_DASH] = ACTIONS(1070),
    [anon_sym_SLASH] = ACTIONS(1070),
    [anon_sym_SLASH_SLASH] = ACTIONS(1070),
    [anon_sym_PERCENT] = ACTIONS(1070),
    [anon_sym_PIPE] = ACTIONS(1070),
    [anon_sym_CARET] = ACTIONS(1070),
    [anon_sym_AMP] = ACTIONS(1070),
    [anon_sym_LT_LT] = ACTIONS(1070),
    [anon_sym_GT_GT] = ACTIONS(1070),
    [anon_sym_GT_GT_GT] = ACTIONS(1070),
    [anon_sym_TILDE] = ACTIONS(1070),
    [anon_sym_typeof] = ACTIONS(1070),
    [anon_sym_void] = ACTIONS(1070),
    [anon_sym_LT] = ACTIONS(1070),
    [anon_sym_LT_EQ] = ACTIONS(1070),
    [anon_sym_EQ_EQ] = ACTIONS(1070),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1070),
    [anon_sym_BANG_EQ] = ACTIONS(1070),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1070),
    [anon_sym_GT_EQ] = ACTIONS(1070),
    [anon_sym_GT] = ACTIONS(1070),
    [anon_sym_is] = ACTIONS(1070),
    [anon_sym_instanceof] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1070),
    [anon_sym_QMARK] = ACTIONS(1070),
    [anon_sym_await] = ACTIONS(1070),
    [anon_sym_LBRACE] = ACTIONS(1070),
    [sym_this] = ACTIONS(1070),
    [sym_true] = ACTIONS(1070),
    [sym_false] = ACTIONS(1070),
    [sym_none] = ACTIONS(1070),
    [sym_string] = ACTIONS(1070),
    [sym_verbatim] = ACTIONS(1070),
    [sym_number] = ACTIONS(1070),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1068),
    [sym_fstring_start] = ACTIONS(1068),
  },
  [STATE(439)] = {
    [aux_sym_expression_list_repeat1] = STATE(540),
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1072),
    [anon_sym_SEMI] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_async] = ACTIONS(1072),
    [anon_sym_def] = ACTIONS(1072),
    [anon_sym_COMMA] = ACTIONS(1074),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_del] = ACTIONS(1072),
    [anon_sym_raise] = ACTIONS(1072),
    [sym_pass_statement] = ACTIONS(1072),
    [sym_break_statement] = ACTIONS(1072),
    [sym_continue_statement] = ACTIONS(1072),
    [sym_debugger_statement] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_global] = ACTIONS(1072),
    [anon_sym_nonlocal] = ACTIONS(1072),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_from] = ACTIONS(1072),
    [anon_sym_import] = ACTIONS(1072),
    [anon_sym_DOT] = ACTIONS(1072),
    [anon_sym_if] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1072),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1072),
    [anon_sym_do] = ACTIONS(1072),
    [anon_sym_try] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(1072),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1072),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1072),
    [anon_sym_typeof] = ACTIONS(1072),
    [anon_sym_void] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_QMARK] = ACTIONS(1072),
    [anon_sym_await] = ACTIONS(1072),
    [anon_sym_LBRACE] = ACTIONS(1072),
    [sym_this] = ACTIONS(1072),
    [sym_true] = ACTIONS(1072),
    [sym_false] = ACTIONS(1072),
    [sym_none] = ACTIONS(1072),
    [sym_string] = ACTIONS(1072),
    [sym_verbatim] = ACTIONS(1072),
    [sym_number] = ACTIONS(1072),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1101),
    [sym_regex] = ACTIONS(1101),
    [sym_fstring_start] = ACTIONS(1101),
  },
  [STATE(440)] = {
    [aux_sym_expression_list_repeat1] = STATE(500),
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1103),
    [sym_identifier] = ACTIONS(1105),
    [anon_sym_SEMI] = ACTIONS(1105),
    [anon_sym_return] = ACTIONS(1105),
    [anon_sym_async] = ACTIONS(1105),
    [anon_sym_def] = ACTIONS(1105),
    [anon_sym_COMMA] = ACTIONS(1107),
    [anon_sym_LPAREN] = ACTIONS(1105),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_del] = ACTIONS(1105),
    [anon_sym_raise] = ACTIONS(1105),
    [sym_pass_statement] = ACTIONS(1105),
    [sym_break_statement] = ACTIONS(1105),
    [sym_continue_statement] = ACTIONS(1105),
    [sym_debugger_statement] = ACTIONS(1105),
    [anon_sym_assert] = ACTIONS(1105),
    [anon_sym_global] = ACTIONS(1105),
    [anon_sym_nonlocal] = ACTIONS(1105),
    [anon_sym_yield] = ACTIONS(1105),
    [anon_sym_from] = ACTIONS(1105),
    [anon_sym_import] = ACTIONS(1105),
    [anon_sym_DOT] = ACTIONS(1105),
    [anon_sym_if] = ACTIONS(1105),
    [anon_sym_for] = ACTIONS(1105),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1105),
    [anon_sym_do] = ACTIONS(1105),
    [anon_sym_try] = ACTIONS(1105),
    [anon_sym_with] = ACTIONS(1105),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1105),
    [anon_sym_typeof] = ACTIONS(1105),
    [anon_sym_void] = ACTIONS(1105),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1105),
    [anon_sym_QMARK] = ACTIONS(1105),
    [anon_sym_await] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1105),
    [sym_this] = ACTIONS(1105),
    [sym_true] = ACTIONS(1105),
    [sym_false] = ACTIONS(1105),
    [sym_none] = ACTIONS(1105),
    [sym_string] = ACTIONS(1105),
    [sym_verbatim] = ACTIONS(1105),
    [sym_number] = ACTIONS(1105),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1103),
    [sym_fstring_start] = ACTIONS(1103),
  },
  [STATE(441)] = {
    [aux_sym_expression_list_repeat1] = STATE(540),
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1133),
    [anon_sym_SEMI] = ACTIONS(1133),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_async] = ACTIONS(1133),
    [anon_sym_def] = ACTIONS(1133),
    [anon_sym_COMMA] = ACTIONS(1135),
    [anon_sym_LPAREN] = ACTIONS(1133),
    [anon_sym_LBRACK] = ACTIONS(1133),
    [anon_sym_del] = ACTIONS(1133),
    [anon_sym_raise] = ACTIONS(1133),
    [sym_pass_statement] = ACTIONS(1133),
    [sym_break_statement] = ACTIONS(1133),
    [sym_continue_statement] = ACTIONS(1133),
    [sym_debugger_statement] = ACTIONS(1133),
    [anon_sym_assert] = ACTIONS(1133),
    [anon_sym_global] = ACTIONS(1133),
    [anon_sym_nonlocal] = ACTIONS(1133),
    [anon_sym_yield] = ACTIONS(1133),
    [anon_sym_from] = ACTIONS(1133),
    [anon_sym_import] = ACTIONS(1133),
    [anon_sym_DOT] = ACTIONS(1133),
    [anon_sym_if] = ACTIONS(1133),
    [anon_sym_for] = ACTIONS(1133),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1133),
    [anon_sym_do] = ACTIONS(1133),
    [anon_sym_try] = ACTIONS(1133),
    [anon_sym_with] = ACTIONS(1133),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1133),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1133),
    [anon_sym_typeof] = ACTIONS(1133),
    [anon_sym_void] = ACTIONS(1133),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1133),
    [anon_sym_QMARK] = ACTIONS(1133),
    [anon_sym_await] = ACTIONS(1133),
    [anon_sym_LBRACE] = ACTIONS(1133),
    [sym_this] = ACTIONS(1133),
    [sym_true] = ACTIONS(1133),
    [sym_false] = ACTIONS(1133),
    [sym_none] = ACTIONS(1133),
    [sym_string] = ACTIONS(1133),
    [sym_verbatim] = ACTIONS(1133),
    [sym_number] = ACTIONS(1133),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1137),
    [sym_regex] = ACTIONS(1137),
    [sym_fstring_start] = ACTIONS(1137),
  },
  [STATE(442)] = {
    [sym_f_string] = STATE(442),
    [aux_sym_concatenated_string_repeat1] = STATE(442),
    [ts_builtin_sym_end] = ACTIONS(1139),
    [sym_identifier] = ACTIONS(1141),
    [anon_sym_SEMI] = ACTIONS(1141),
    [anon_sym_return] = ACTIONS(1141),
    [anon_sym_async] = ACTIONS(1141),
    [anon_sym_def] = ACTIONS(1141),
    [anon_sym_COMMA] = ACTIONS(1141),
    [anon_sym_LPAREN] = ACTIONS(1141),
    [anon_sym_LBRACK] = ACTIONS(1141),
    [anon_sym_del] = ACTIONS(1141),
    [anon_sym_raise] = ACTIONS(1141),
    [sym_pass_statement] = ACTIONS(1141),
    [sym_break_statement] = ACTIONS(1141),
    [sym_continue_statement] = ACTIONS(1141),
    [sym_debugger_statement] = ACTIONS(1141),
    [anon_sym_assert] = ACTIONS(1141),
    [anon_sym_global] = ACTIONS(1141),
    [anon_sym_nonlocal] = ACTIONS(1141),
    [anon_sym_yield] = ACTIONS(1141),
    [anon_sym_from] = ACTIONS(1141),
    [anon_sym_import] = ACTIONS(1141),
    [anon_sym_DOT] = ACTIONS(1141),
    [anon_sym_if] = ACTIONS(1141),
    [anon_sym_for] = ACTIONS(1141),
    [anon_sym_in] = ACTIONS(1141),
    [anon_sym_while] = ACTIONS(1141),
    [anon_sym_do] = ACTIONS(1141),
    [anon_sym_try] = ACTIONS(1141),
    [anon_sym_with] = ACTIONS(1141),
    [anon_sym_STAR] = ACTIONS(1141),
    [anon_sym_STAR_STAR] = ACTIONS(1141),
    [anon_sym_class] = ACTIONS(1141),
    [anon_sym_AT] = ACTIONS(1141),
    [anon_sym_not] = ACTIONS(1141),
    [anon_sym_and] = ACTIONS(1141),
    [anon_sym_or] = ACTIONS(1141),
    [anon_sym_PLUS] = ACTIONS(1141),
    [anon_sym_DASH] = ACTIONS(1141),
    [anon_sym_SLASH] = ACTIONS(1141),
    [anon_sym_SLASH_SLASH] = ACTIONS(1141),
    [anon_sym_PERCENT] = ACTIONS(1141),
    [anon_sym_PIPE] = ACTIONS(1141),
    [anon_sym_CARET] = ACTIONS(1141),
    [anon_sym_AMP] = ACTIONS(1141),
    [anon_sym_LT_LT] = ACTIONS(1141),
    [anon_sym_GT_GT] = ACTIONS(1141),
    [anon_sym_GT_GT_GT] = ACTIONS(1141),
    [anon_sym_TILDE] = ACTIONS(1141),
    [anon_sym_typeof] = ACTIONS(1141),
    [anon_sym_void] = ACTIONS(1141),
    [anon_sym_LT] = ACTIONS(1141),
    [anon_sym_LT_EQ] = ACTIONS(1141),
    [anon_sym_EQ_EQ] = ACTIONS(1141),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1141),
    [anon_sym_BANG_EQ] = ACTIONS(1141),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1141),
    [anon_sym_GT_EQ] = ACTIONS(1141),
    [anon_sym_GT] = ACTIONS(1141),
    [anon_sym_is] = ACTIONS(1141),
    [anon_sym_instanceof] = ACTIONS(1141),
    [anon_sym_new] = ACTIONS(1141),
    [anon_sym_QMARK] = ACTIONS(1141),
    [anon_sym_await] = ACTIONS(1141),
    [anon_sym_LBRACE] = ACTIONS(1141),
    [sym_this] = ACTIONS(1141),
    [sym_true] = ACTIONS(1141),
    [sym_false] = ACTIONS(1141),
    [sym_none] = ACTIONS(1141),
    [sym_string] = ACTIONS(1143),
    [sym_verbatim] = ACTIONS(1141),
    [sym_number] = ACTIONS(1141),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1139),
    [sym_fstring_start] = ACTIONS(1146),
  },
  [STATE(443)] = {
    [aux_sym_expression_list_repeat1] = STATE(500),
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1103),
    [sym_identifier] = ACTIONS(1105),
    [anon_sym_SEMI] = ACTIONS(1105),
    [anon_sym_return] = ACTIONS(1105),
    [anon_sym_async] = ACTIONS(1105),
    [anon_sym_def] = ACTIONS(1105),
    [anon_sym_COMMA] = ACTIONS(1149),
    [anon_sym_LPAREN] = ACTIONS(1105),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_del] = ACTIONS(1105),
    [anon_sym_raise] = ACTIONS(1105),
    [sym_pass_statement] = ACTIONS(1105),
    [sym_break_statement] = ACTIONS(1105),
    [sym_continue_statement] = ACTIONS(1105),
    [sym_debugger_statement] = ACTIONS(1105),
    [anon_sym_assert] = ACTIONS(1105),
    [anon_sym_global] = ACTIONS(1105),
    [anon_sym_nonlocal] = ACTIONS(1105),
    [anon_sym_yield] = ACTIONS(1105),
    [anon_sym_from] = ACTIONS(1105),
    [anon_sym_import] = ACTIONS(1105),
    [anon_sym_DOT] = ACTIONS(1105),
    [anon_sym_if] = ACTIONS(1105),
    [anon_sym_for] = ACTIONS(1105),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1105),
    [anon_sym_do] = ACTIONS(1105),
    [anon_sym_try] = ACTIONS(1105),
    [anon_sym_with] = ACTIONS(1105),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1105),
    [anon_sym_typeof] = ACTIONS(1105),
    [anon_sym_void] = ACTIONS(1105),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1105),
    [anon_sym_QMARK] = ACTIONS(1105),
    [anon_sym_await] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1105),
    [sym_this] = ACTIONS(1105),
    [sym_true] = ACTIONS(1105),
    [sym_false] = ACTIONS(1105),
    [sym_none] = ACTIONS(1105),
    [sym_string] = ACTIONS(1105),
    [sym_verbatim] = ACTIONS(1105),
    [sym_number] = ACTIONS(1105),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1103),
    [sym_fstring_start] = ACTIONS(1103),
  },
  [STATE(444)] = {
    [aux_sym_expression_list_repeat1] = STATE(500),
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1152),
    [sym_identifier] = ACTIONS(1154),
    [anon_sym_SEMI] = ACTIONS(1154),
    [anon_sym_return] = ACTIONS(1154),
    [anon_sym_async] = ACTIONS(1154),
    [anon_sym_def] = ACTIONS(1154),
    [anon_sym_COMMA] = ACTIONS(1107),
    [anon_sym_LPAREN] = ACTIONS(1154),
    [anon_sym_LBRACK] = ACTIONS(1154),
    [anon_sym_del] = ACTIONS(1154),
    [anon_sym_raise] = ACTIONS(1154),
    [sym_pass_statement] = ACTIONS(1154),
    [sym_break_statement] = ACTIONS(1154),
    [sym_continue_statement] = ACTIONS(1154),
    [sym_debugger_statement] = ACTIONS(1154),
    [anon_sym_assert] = ACTIONS(1154),
    [anon_sym_global] = ACTIONS(1154),
    [anon_sym_nonlocal] = ACTIONS(1154),
    [anon_sym_yield] = ACTIONS(1154),
    [anon_sym_from] = ACTIONS(1154),
    [anon_sym_import] = ACTIONS(1154),
    [anon_sym_DOT] = ACTIONS(1154),
    [anon_sym_if] = ACTIONS(1154),
    [anon_sym_for] = ACTIONS(1154),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1154),
    [anon_sym_do] = ACTIONS(1154),
    [anon_sym_try] = ACTIONS(1154),
    [anon_sym_with] = ACTIONS(1154),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1154),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1154),
    [anon_sym_typeof] = ACTIONS(1154),
    [anon_sym_void] = ACTIONS(1154),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1154),
    [anon_sym_QMARK] = ACTIONS(1154),
    [anon_sym_await] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1154),
    [sym_this] = ACTIONS(1154),
    [sym_true] = ACTIONS(1154),
    [sym_false] = ACTIONS(1154),
    [sym_none] = ACTIONS(1154),
    [sym_string] = ACTIONS(1154),
    [sym_verbatim] = ACTIONS(1154),
    [sym_number] = ACTIONS(1154),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1152),
    [sym_fstring_start] = ACTIONS(1152),
  },
  [STATE(445)] = {
    [aux_sym_expression_list_repeat1] = STATE(500),
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1137),
    [sym_identifier] = ACTIONS(1133),
    [anon_sym_SEMI] = ACTIONS(1133),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_async] = ACTIONS(1133),
    [anon_sym_def] = ACTIONS(1133),
    [anon_sym_COMMA] = ACTIONS(1107),
    [anon_sym_LPAREN] = ACTIONS(1133),
    [anon_sym_LBRACK] = ACTIONS(1133),
    [anon_sym_del] = ACTIONS(1133),
    [anon_sym_raise] = ACTIONS(1133),
    [sym_pass_statement] = ACTIONS(1133),
    [sym_break_statement] = ACTIONS(1133),
    [sym_continue_statement] = ACTIONS(1133),
    [sym_debugger_statement] = ACTIONS(1133),
    [anon_sym_assert] = ACTIONS(1133),
    [anon_sym_global] = ACTIONS(1133),
    [anon_sym_nonlocal] = ACTIONS(1133),
    [anon_sym_yield] = ACTIONS(1133),
    [anon_sym_from] = ACTIONS(1133),
    [anon_sym_import] = ACTIONS(1133),
    [anon_sym_DOT] = ACTIONS(1133),
    [anon_sym_if] = ACTIONS(1133),
    [anon_sym_for] = ACTIONS(1133),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1133),
    [anon_sym_do] = ACTIONS(1133),
    [anon_sym_try] = ACTIONS(1133),
    [anon_sym_with] = ACTIONS(1133),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1133),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1133),
    [anon_sym_typeof] = ACTIONS(1133),
    [anon_sym_void] = ACTIONS(1133),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1133),
    [anon_sym_QMARK] = ACTIONS(1133),
    [anon_sym_await] = ACTIONS(1133),
    [anon_sym_LBRACE] = ACTIONS(1133),
    [sym_this] = ACTIONS(1133),
    [sym_true] = ACTIONS(1133),
    [sym_false] = ACTIONS(1133),
    [sym_none] = ACTIONS(1133),
    [sym_string] = ACTIONS(1133),
    [sym_verbatim] = ACTIONS(1133),
    [sym_number] = ACTIONS(1133),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1137),
    [sym_fstring_start] = ACTIONS(1137),
  },
  [STATE(446)] = {
    [sym_f_string] = STATE(446),
    [aux_sym_concatenated_string_repeat1] = STATE(446),
    [sym_identifier] = ACTIONS(1141),
    [anon_sym_SEMI] = ACTIONS(1141),
    [anon_sym_return] = ACTIONS(1141),
    [anon_sym_async] = ACTIONS(1141),
    [anon_sym_def] = ACTIONS(1141),
    [anon_sym_COMMA] = ACTIONS(1141),
    [anon_sym_LPAREN] = ACTIONS(1141),
    [anon_sym_LBRACK] = ACTIONS(1141),
    [anon_sym_del] = ACTIONS(1141),
    [anon_sym_raise] = ACTIONS(1141),
    [sym_pass_statement] = ACTIONS(1141),
    [sym_break_statement] = ACTIONS(1141),
    [sym_continue_statement] = ACTIONS(1141),
    [sym_debugger_statement] = ACTIONS(1141),
    [anon_sym_assert] = ACTIONS(1141),
    [anon_sym_global] = ACTIONS(1141),
    [anon_sym_nonlocal] = ACTIONS(1141),
    [anon_sym_yield] = ACTIONS(1141),
    [anon_sym_from] = ACTIONS(1141),
    [anon_sym_import] = ACTIONS(1141),
    [anon_sym_DOT] = ACTIONS(1141),
    [anon_sym_if] = ACTIONS(1141),
    [anon_sym_for] = ACTIONS(1141),
    [anon_sym_in] = ACTIONS(1141),
    [anon_sym_while] = ACTIONS(1141),
    [anon_sym_do] = ACTIONS(1141),
    [anon_sym_try] = ACTIONS(1141),
    [anon_sym_with] = ACTIONS(1141),
    [anon_sym_STAR] = ACTIONS(1141),
    [anon_sym_STAR_STAR] = ACTIONS(1141),
    [anon_sym_class] = ACTIONS(1141),
    [anon_sym_AT] = ACTIONS(1141),
    [anon_sym_not] = ACTIONS(1141),
    [anon_sym_and] = ACTIONS(1141),
    [anon_sym_or] = ACTIONS(1141),
    [anon_sym_PLUS] = ACTIONS(1141),
    [anon_sym_DASH] = ACTIONS(1141),
    [anon_sym_SLASH] = ACTIONS(1141),
    [anon_sym_SLASH_SLASH] = ACTIONS(1141),
    [anon_sym_PERCENT] = ACTIONS(1141),
    [anon_sym_PIPE] = ACTIONS(1141),
    [anon_sym_CARET] = ACTIONS(1141),
    [anon_sym_AMP] = ACTIONS(1141),
    [anon_sym_LT_LT] = ACTIONS(1141),
    [anon_sym_GT_GT] = ACTIONS(1141),
    [anon_sym_GT_GT_GT] = ACTIONS(1141),
    [anon_sym_TILDE] = ACTIONS(1141),
    [anon_sym_typeof] = ACTIONS(1141),
    [anon_sym_void] = ACTIONS(1141),
    [anon_sym_LT] = ACTIONS(1141),
    [anon_sym_LT_EQ] = ACTIONS(1141),
    [anon_sym_EQ_EQ] = ACTIONS(1141),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1141),
    [anon_sym_BANG_EQ] = ACTIONS(1141),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1141),
    [anon_sym_GT_EQ] = ACTIONS(1141),
    [anon_sym_GT] = ACTIONS(1141),
    [anon_sym_is] = ACTIONS(1141),
    [anon_sym_instanceof] = ACTIONS(1141),
    [anon_sym_new] = ACTIONS(1141),
    [anon_sym_QMARK] = ACTIONS(1141),
    [anon_sym_await] = ACTIONS(1141),
    [anon_sym_LBRACE] = ACTIONS(1141),
    [sym_this] = ACTIONS(1141),
    [sym_true] = ACTIONS(1141),
    [sym_false] = ACTIONS(1141),
    [sym_none] = ACTIONS(1141),
    [sym_string] = ACTIONS(1156),
    [sym_verbatim] = ACTIONS(1141),
    [sym_number] = ACTIONS(1141),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1139),
    [sym_regex] = ACTIONS(1139),
    [sym_fstring_start] = ACTIONS(1159),
  },
  [STATE(447)] = {
    [aux_sym_expression_list_repeat1] = STATE(500),
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1162),
    [sym_identifier] = ACTIONS(1164),
    [anon_sym_SEMI] = ACTIONS(1164),
    [anon_sym_return] = ACTIONS(1164),
    [anon_sym_async] = ACTIONS(1164),
    [anon_sym_def] = ACTIONS(1164),
    [anon_sym_COMMA] = ACTIONS(1107),
    [anon_sym_LPAREN] = ACTIONS(1164),
    [anon_sym_LBRACK] = ACTIONS(1164),
    [anon_sym_del] = ACTIONS(1164),
    [anon_sym_raise] = ACTIONS(1164),
    [sym_pass_statement] = ACTIONS(1164),
    [sym_break_statement] = ACTIONS(1164),
    [sym_continue_statement] = ACTIONS(1164),
    [sym_debugger_statement] = ACTIONS(1164),
    [anon_sym_assert] = ACTIONS(1164),
    [anon_sym_global] = ACTIONS(1164),
    [anon_sym_nonlocal] = ACTIONS(1164),
    [anon_sym_yield] = ACTIONS(1164),
    [anon_sym_from] = ACTIONS(1164),
    [anon_sym_import] = ACTIONS(1164),
    [anon_sym_DOT] = ACTIONS(1164),
    [anon_sym_if] = ACTIONS(1164),
    [anon_sym_for] = ACTIONS(1164),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1164),
    [anon_sym_do] = ACTIONS(1164),
    [anon_sym_try] = ACTIONS(1164),
    [anon_sym_with] = ACTIONS(1164),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1164),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1164),
    [anon_sym_typeof] = ACTIONS(1164),
    [anon_sym_void] = ACTIONS(1164),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1164),
    [anon_sym_QMARK] = ACTIONS(1164),
    [anon_sym_await] = ACTIONS(1164),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [sym_this] = ACTIONS(1164),
    [sym_true] = ACTIONS(1164),
    [sym_false] = ACTIONS(1164),
    [sym_none] = ACTIONS(1164),
    [sym_string] = ACTIONS(1164),
    [sym_verbatim] = ACTIONS(1164),
    [sym_number] = ACTIONS(1164),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1162),
    [sym_fstring_start] = ACTIONS(1162),
  },
  [STATE(448)] = {
    [aux_sym_expression_list_repeat1] = STATE(500),
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1101),
    [sym_identifier] = ACTIONS(1072),
    [anon_sym_SEMI] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_async] = ACTIONS(1072),
    [anon_sym_def] = ACTIONS(1072),
    [anon_sym_COMMA] = ACTIONS(1107),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_del] = ACTIONS(1072),
    [anon_sym_raise] = ACTIONS(1072),
    [sym_pass_statement] = ACTIONS(1072),
    [sym_break_statement] = ACTIONS(1072),
    [sym_continue_statement] = ACTIONS(1072),
    [sym_debugger_statement] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_global] = ACTIONS(1072),
    [anon_sym_nonlocal] = ACTIONS(1072),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_from] = ACTIONS(1072),
    [anon_sym_import] = ACTIONS(1072),
    [anon_sym_DOT] = ACTIONS(1072),
    [anon_sym_if] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1072),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1072),
    [anon_sym_do] = ACTIONS(1072),
    [anon_sym_try] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(1072),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1072),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1072),
    [anon_sym_typeof] = ACTIONS(1072),
    [anon_sym_void] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_QMARK] = ACTIONS(1072),
    [anon_sym_await] = ACTIONS(1072),
    [anon_sym_LBRACE] = ACTIONS(1072),
    [sym_this] = ACTIONS(1072),
    [sym_true] = ACTIONS(1072),
    [sym_false] = ACTIONS(1072),
    [sym_none] = ACTIONS(1072),
    [sym_string] = ACTIONS(1072),
    [sym_verbatim] = ACTIONS(1072),
    [sym_number] = ACTIONS(1072),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1101),
    [sym_fstring_start] = ACTIONS(1101),
  },
  [STATE(449)] = {
    [aux_sym_expression_list_repeat1] = STATE(540),
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1105),
    [anon_sym_SEMI] = ACTIONS(1105),
    [anon_sym_return] = ACTIONS(1105),
    [anon_sym_async] = ACTIONS(1105),
    [anon_sym_def] = ACTIONS(1105),
    [anon_sym_COMMA] = ACTIONS(1135),
    [anon_sym_LPAREN] = ACTIONS(1105),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_del] = ACTIONS(1105),
    [anon_sym_raise] = ACTIONS(1105),
    [sym_pass_statement] = ACTIONS(1105),
    [sym_break_statement] = ACTIONS(1105),
    [sym_continue_statement] = ACTIONS(1105),
    [sym_debugger_statement] = ACTIONS(1105),
    [anon_sym_assert] = ACTIONS(1105),
    [anon_sym_global] = ACTIONS(1105),
    [anon_sym_nonlocal] = ACTIONS(1105),
    [anon_sym_yield] = ACTIONS(1105),
    [anon_sym_from] = ACTIONS(1105),
    [anon_sym_import] = ACTIONS(1105),
    [anon_sym_DOT] = ACTIONS(1105),
    [anon_sym_if] = ACTIONS(1105),
    [anon_sym_for] = ACTIONS(1105),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1105),
    [anon_sym_do] = ACTIONS(1105),
    [anon_sym_try] = ACTIONS(1105),
    [anon_sym_with] = ACTIONS(1105),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1105),
    [anon_sym_typeof] = ACTIONS(1105),
    [anon_sym_void] = ACTIONS(1105),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1105),
    [anon_sym_QMARK] = ACTIONS(1105),
    [anon_sym_await] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1105),
    [sym_this] = ACTIONS(1105),
    [sym_true] = ACTIONS(1105),
    [sym_false] = ACTIONS(1105),
    [sym_none] = ACTIONS(1105),
    [sym_string] = ACTIONS(1105),
    [sym_verbatim] = ACTIONS(1105),
    [sym_number] = ACTIONS(1105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1103),
    [sym_regex] = ACTIONS(1103),
    [sym_fstring_start] = ACTIONS(1103),
  },
  [STATE(450)] = {
    [sym_f_string] = STATE(457),
    [aux_sym_concatenated_string_repeat1] = STATE(457),
    [sym_identifier] = ACTIONS(927),
    [anon_sym_SEMI] = ACTIONS(927),
    [anon_sym_return] = ACTIONS(927),
    [anon_sym_async] = ACTIONS(927),
    [anon_sym_def] = ACTIONS(927),
    [anon_sym_COMMA] = ACTIONS(927),
    [anon_sym_LPAREN] = ACTIONS(927),
    [anon_sym_LBRACK] = ACTIONS(927),
    [anon_sym_del] = ACTIONS(927),
    [anon_sym_raise] = ACTIONS(927),
    [sym_pass_statement] = ACTIONS(927),
    [sym_break_statement] = ACTIONS(927),
    [sym_continue_statement] = ACTIONS(927),
    [sym_debugger_statement] = ACTIONS(927),
    [anon_sym_assert] = ACTIONS(927),
    [anon_sym_global] = ACTIONS(927),
    [anon_sym_nonlocal] = ACTIONS(927),
    [anon_sym_yield] = ACTIONS(927),
    [anon_sym_from] = ACTIONS(927),
    [anon_sym_import] = ACTIONS(927),
    [anon_sym_DOT] = ACTIONS(927),
    [anon_sym_if] = ACTIONS(927),
    [anon_sym_for] = ACTIONS(927),
    [anon_sym_in] = ACTIONS(927),
    [anon_sym_while] = ACTIONS(927),
    [anon_sym_do] = ACTIONS(927),
    [anon_sym_try] = ACTIONS(927),
    [anon_sym_with] = ACTIONS(927),
    [anon_sym_STAR] = ACTIONS(927),
    [anon_sym_STAR_STAR] = ACTIONS(927),
    [anon_sym_class] = ACTIONS(927),
    [anon_sym_AT] = ACTIONS(927),
    [anon_sym_not] = ACTIONS(927),
    [anon_sym_and] = ACTIONS(927),
    [anon_sym_or] = ACTIONS(927),
    [anon_sym_PLUS] = ACTIONS(927),
    [anon_sym_DASH] = ACTIONS(927),
    [anon_sym_SLASH] = ACTIONS(927),
    [anon_sym_SLASH_SLASH] = ACTIONS(927),
    [anon_sym_PERCENT] = ACTIONS(927),
    [anon_sym_PIPE] = ACTIONS(927),
    [anon_sym_CARET] = ACTIONS(927),
    [anon_sym_AMP] = ACTIONS(927),
    [anon_sym_LT_LT] = ACTIONS(927),
    [anon_sym_GT_GT] = ACTIONS(927),
    [anon_sym_GT_GT_GT] = ACTIONS(927),
    [anon_sym_TILDE] = ACTIONS(927),
    [anon_sym_typeof] = ACTIONS(927),
    [anon_sym_void] = ACTIONS(927),
    [anon_sym_LT] = ACTIONS(927),
    [anon_sym_LT_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(927),
    [anon_sym_GT_EQ] = ACTIONS(927),
    [anon_sym_GT] = ACTIONS(927),
    [anon_sym_is] = ACTIONS(927),
    [anon_sym_instanceof] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
    [anon_sym_QMARK] = ACTIONS(927),
    [anon_sym_await] = ACTIONS(927),
    [anon_sym_LBRACE] = ACTIONS(927),
    [sym_this] = ACTIONS(927),
    [sym_true] = ACTIONS(927),
    [sym_false] = ACTIONS(927),
    [sym_none] = ACTIONS(927),
    [sym_string] = ACTIONS(1166),
    [sym_verbatim] = ACTIONS(927),
    [sym_number] = ACTIONS(927),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(925),
    [sym_regex] = ACTIONS(925),
    [sym_fstring_start] = ACTIONS(1169),
  },
  [STATE(451)] = {
    [aux_sym_expression_list_repeat1] = STATE(500),
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1101),
    [sym_identifier] = ACTIONS(1072),
    [anon_sym_SEMI] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_async] = ACTIONS(1072),
    [anon_sym_def] = ACTIONS(1072),
    [anon_sym_COMMA] = ACTIONS(1172),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_del] = ACTIONS(1072),
    [anon_sym_raise] = ACTIONS(1072),
    [sym_pass_statement] = ACTIONS(1072),
    [sym_break_statement] = ACTIONS(1072),
    [sym_continue_statement] = ACTIONS(1072),
    [sym_debugger_statement] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_global] = ACTIONS(1072),
    [anon_sym_nonlocal] = ACTIONS(1072),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_from] = ACTIONS(1072),
    [anon_sym_import] = ACTIONS(1072),
    [anon_sym_DOT] = ACTIONS(1072),
    [anon_sym_if] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1072),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1072),
    [anon_sym_do] = ACTIONS(1072),
    [anon_sym_try] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(1072),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1072),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1072),
    [anon_sym_typeof] = ACTIONS(1072),
    [anon_sym_void] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_QMARK] = ACTIONS(1072),
    [anon_sym_await] = ACTIONS(1072),
    [anon_sym_LBRACE] = ACTIONS(1072),
    [sym_this] = ACTIONS(1072),
    [sym_true] = ACTIONS(1072),
    [sym_false] = ACTIONS(1072),
    [sym_none] = ACTIONS(1072),
    [sym_string] = ACTIONS(1072),
    [sym_verbatim] = ACTIONS(1072),
    [sym_number] = ACTIONS(1072),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1101),
    [sym_fstring_start] = ACTIONS(1101),
  },
  [STATE(452)] = {
    [aux_sym_expression_list_repeat1] = STATE(500),
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1162),
    [sym_identifier] = ACTIONS(1164),
    [anon_sym_SEMI] = ACTIONS(1164),
    [anon_sym_return] = ACTIONS(1164),
    [anon_sym_async] = ACTIONS(1164),
    [anon_sym_def] = ACTIONS(1164),
    [anon_sym_COMMA] = ACTIONS(1175),
    [anon_sym_LPAREN] = ACTIONS(1164),
    [anon_sym_LBRACK] = ACTIONS(1164),
    [anon_sym_del] = ACTIONS(1164),
    [anon_sym_raise] = ACTIONS(1164),
    [sym_pass_statement] = ACTIONS(1164),
    [sym_break_statement] = ACTIONS(1164),
    [sym_continue_statement] = ACTIONS(1164),
    [sym_debugger_statement] = ACTIONS(1164),
    [anon_sym_assert] = ACTIONS(1164),
    [anon_sym_global] = ACTIONS(1164),
    [anon_sym_nonlocal] = ACTIONS(1164),
    [anon_sym_yield] = ACTIONS(1164),
    [anon_sym_from] = ACTIONS(1164),
    [anon_sym_import] = ACTIONS(1164),
    [anon_sym_DOT] = ACTIONS(1164),
    [anon_sym_if] = ACTIONS(1164),
    [anon_sym_for] = ACTIONS(1164),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1164),
    [anon_sym_do] = ACTIONS(1164),
    [anon_sym_try] = ACTIONS(1164),
    [anon_sym_with] = ACTIONS(1164),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1164),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1164),
    [anon_sym_typeof] = ACTIONS(1164),
    [anon_sym_void] = ACTIONS(1164),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1164),
    [anon_sym_QMARK] = ACTIONS(1164),
    [anon_sym_await] = ACTIONS(1164),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [sym_this] = ACTIONS(1164),
    [sym_true] = ACTIONS(1164),
    [sym_false] = ACTIONS(1164),
    [sym_none] = ACTIONS(1164),
    [sym_string] = ACTIONS(1164),
    [sym_verbatim] = ACTIONS(1164),
    [sym_number] = ACTIONS(1164),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1162),
    [sym_fstring_start] = ACTIONS(1162),
  },
  [STATE(453)] = {
    [aux_sym_expression_list_repeat1] = STATE(540),
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1154),
    [anon_sym_SEMI] = ACTIONS(1154),
    [anon_sym_return] = ACTIONS(1154),
    [anon_sym_async] = ACTIONS(1154),
    [anon_sym_def] = ACTIONS(1154),
    [anon_sym_COMMA] = ACTIONS(1135),
    [anon_sym_LPAREN] = ACTIONS(1154),
    [anon_sym_LBRACK] = ACTIONS(1154),
    [anon_sym_del] = ACTIONS(1154),
    [anon_sym_raise] = ACTIONS(1154),
    [sym_pass_statement] = ACTIONS(1154),
    [sym_break_statement] = ACTIONS(1154),
    [sym_continue_statement] = ACTIONS(1154),
    [sym_debugger_statement] = ACTIONS(1154),
    [anon_sym_assert] = ACTIONS(1154),
    [anon_sym_global] = ACTIONS(1154),
    [anon_sym_nonlocal] = ACTIONS(1154),
    [anon_sym_yield] = ACTIONS(1154),
    [anon_sym_from] = ACTIONS(1154),
    [anon_sym_import] = ACTIONS(1154),
    [anon_sym_DOT] = ACTIONS(1154),
    [anon_sym_if] = ACTIONS(1154),
    [anon_sym_for] = ACTIONS(1154),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1154),
    [anon_sym_do] = ACTIONS(1154),
    [anon_sym_try] = ACTIONS(1154),
    [anon_sym_with] = ACTIONS(1154),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1154),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1154),
    [anon_sym_typeof] = ACTIONS(1154),
    [anon_sym_void] = ACTIONS(1154),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1154),
    [anon_sym_QMARK] = ACTIONS(1154),
    [anon_sym_await] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1154),
    [sym_this] = ACTIONS(1154),
    [sym_true] = ACTIONS(1154),
    [sym_false] = ACTIONS(1154),
    [sym_none] = ACTIONS(1154),
    [sym_string] = ACTIONS(1154),
    [sym_verbatim] = ACTIONS(1154),
    [sym_number] = ACTIONS(1154),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1152),
    [sym_regex] = ACTIONS(1152),
    [sym_fstring_start] = ACTIONS(1152),
  },
  [STATE(454)] = {
    [aux_sym_expression_list_repeat1] = STATE(540),
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1164),
    [anon_sym_SEMI] = ACTIONS(1164),
    [anon_sym_return] = ACTIONS(1164),
    [anon_sym_async] = ACTIONS(1164),
    [anon_sym_def] = ACTIONS(1164),
    [anon_sym_COMMA] = ACTIONS(1135),
    [anon_sym_LPAREN] = ACTIONS(1164),
    [anon_sym_LBRACK] = ACTIONS(1164),
    [anon_sym_del] = ACTIONS(1164),
    [anon_sym_raise] = ACTIONS(1164),
    [sym_pass_statement] = ACTIONS(1164),
    [sym_break_statement] = ACTIONS(1164),
    [sym_continue_statement] = ACTIONS(1164),
    [sym_debugger_statement] = ACTIONS(1164),
    [anon_sym_assert] = ACTIONS(1164),
    [anon_sym_global] = ACTIONS(1164),
    [anon_sym_nonlocal] = ACTIONS(1164),
    [anon_sym_yield] = ACTIONS(1164),
    [anon_sym_from] = ACTIONS(1164),
    [anon_sym_import] = ACTIONS(1164),
    [anon_sym_DOT] = ACTIONS(1164),
    [anon_sym_if] = ACTIONS(1164),
    [anon_sym_for] = ACTIONS(1164),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1164),
    [anon_sym_do] = ACTIONS(1164),
    [anon_sym_try] = ACTIONS(1164),
    [anon_sym_with] = ACTIONS(1164),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1164),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1164),
    [anon_sym_typeof] = ACTIONS(1164),
    [anon_sym_void] = ACTIONS(1164),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1164),
    [anon_sym_QMARK] = ACTIONS(1164),
    [anon_sym_await] = ACTIONS(1164),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [sym_this] = ACTIONS(1164),
    [sym_true] = ACTIONS(1164),
    [sym_false] = ACTIONS(1164),
    [sym_none] = ACTIONS(1164),
    [sym_string] = ACTIONS(1164),
    [sym_verbatim] = ACTIONS(1164),
    [sym_number] = ACTIONS(1164),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1162),
    [sym_regex] = ACTIONS(1162),
    [sym_fstring_start] = ACTIONS(1162),
  },
  [STATE(455)] = {
    [aux_sym_expression_list_repeat1] = STATE(540),
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1072),
    [anon_sym_SEMI] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_async] = ACTIONS(1072),
    [anon_sym_def] = ACTIONS(1072),
    [anon_sym_COMMA] = ACTIONS(1135),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_del] = ACTIONS(1072),
    [anon_sym_raise] = ACTIONS(1072),
    [sym_pass_statement] = ACTIONS(1072),
    [sym_break_statement] = ACTIONS(1072),
    [sym_continue_statement] = ACTIONS(1072),
    [sym_debugger_statement] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_global] = ACTIONS(1072),
    [anon_sym_nonlocal] = ACTIONS(1072),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_from] = ACTIONS(1072),
    [anon_sym_import] = ACTIONS(1072),
    [anon_sym_DOT] = ACTIONS(1072),
    [anon_sym_if] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1072),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1072),
    [anon_sym_do] = ACTIONS(1072),
    [anon_sym_try] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(1072),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1072),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1072),
    [anon_sym_typeof] = ACTIONS(1072),
    [anon_sym_void] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_QMARK] = ACTIONS(1072),
    [anon_sym_await] = ACTIONS(1072),
    [anon_sym_LBRACE] = ACTIONS(1072),
    [sym_this] = ACTIONS(1072),
    [sym_true] = ACTIONS(1072),
    [sym_false] = ACTIONS(1072),
    [sym_none] = ACTIONS(1072),
    [sym_string] = ACTIONS(1072),
    [sym_verbatim] = ACTIONS(1072),
    [sym_number] = ACTIONS(1072),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1101),
    [sym_regex] = ACTIONS(1101),
    [sym_fstring_start] = ACTIONS(1101),
  },
  [STATE(456)] = {
    [aux_sym_expression_list_repeat1] = STATE(540),
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1105),
    [anon_sym_SEMI] = ACTIONS(1105),
    [anon_sym_return] = ACTIONS(1105),
    [anon_sym_async] = ACTIONS(1105),
    [anon_sym_def] = ACTIONS(1105),
    [anon_sym_COMMA] = ACTIONS(1178),
    [anon_sym_LPAREN] = ACTIONS(1105),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_del] = ACTIONS(1105),
    [anon_sym_raise] = ACTIONS(1105),
    [sym_pass_statement] = ACTIONS(1105),
    [sym_break_statement] = ACTIONS(1105),
    [sym_continue_statement] = ACTIONS(1105),
    [sym_debugger_statement] = ACTIONS(1105),
    [anon_sym_assert] = ACTIONS(1105),
    [anon_sym_global] = ACTIONS(1105),
    [anon_sym_nonlocal] = ACTIONS(1105),
    [anon_sym_yield] = ACTIONS(1105),
    [anon_sym_from] = ACTIONS(1105),
    [anon_sym_import] = ACTIONS(1105),
    [anon_sym_DOT] = ACTIONS(1105),
    [anon_sym_if] = ACTIONS(1105),
    [anon_sym_for] = ACTIONS(1105),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1105),
    [anon_sym_do] = ACTIONS(1105),
    [anon_sym_try] = ACTIONS(1105),
    [anon_sym_with] = ACTIONS(1105),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1105),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1105),
    [anon_sym_typeof] = ACTIONS(1105),
    [anon_sym_void] = ACTIONS(1105),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1105),
    [anon_sym_QMARK] = ACTIONS(1105),
    [anon_sym_await] = ACTIONS(1105),
    [anon_sym_LBRACE] = ACTIONS(1105),
    [sym_this] = ACTIONS(1105),
    [sym_true] = ACTIONS(1105),
    [sym_false] = ACTIONS(1105),
    [sym_none] = ACTIONS(1105),
    [sym_string] = ACTIONS(1105),
    [sym_verbatim] = ACTIONS(1105),
    [sym_number] = ACTIONS(1105),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1103),
    [sym_regex] = ACTIONS(1103),
    [sym_fstring_start] = ACTIONS(1103),
  },
  [STATE(457)] = {
    [sym_f_string] = STATE(446),
    [aux_sym_concatenated_string_repeat1] = STATE(446),
    [sym_identifier] = ACTIONS(1070),
    [anon_sym_SEMI] = ACTIONS(1070),
    [anon_sym_return] = ACTIONS(1070),
    [anon_sym_async] = ACTIONS(1070),
    [anon_sym_def] = ACTIONS(1070),
    [anon_sym_COMMA] = ACTIONS(1070),
    [anon_sym_LPAREN] = ACTIONS(1070),
    [anon_sym_LBRACK] = ACTIONS(1070),
    [anon_sym_del] = ACTIONS(1070),
    [anon_sym_raise] = ACTIONS(1070),
    [sym_pass_statement] = ACTIONS(1070),
    [sym_break_statement] = ACTIONS(1070),
    [sym_continue_statement] = ACTIONS(1070),
    [sym_debugger_statement] = ACTIONS(1070),
    [anon_sym_assert] = ACTIONS(1070),
    [anon_sym_global] = ACTIONS(1070),
    [anon_sym_nonlocal] = ACTIONS(1070),
    [anon_sym_yield] = ACTIONS(1070),
    [anon_sym_from] = ACTIONS(1070),
    [anon_sym_import] = ACTIONS(1070),
    [anon_sym_DOT] = ACTIONS(1070),
    [anon_sym_if] = ACTIONS(1070),
    [anon_sym_for] = ACTIONS(1070),
    [anon_sym_in] = ACTIONS(1070),
    [anon_sym_while] = ACTIONS(1070),
    [anon_sym_do] = ACTIONS(1070),
    [anon_sym_try] = ACTIONS(1070),
    [anon_sym_with] = ACTIONS(1070),
    [anon_sym_STAR] = ACTIONS(1070),
    [anon_sym_STAR_STAR] = ACTIONS(1070),
    [anon_sym_class] = ACTIONS(1070),
    [anon_sym_AT] = ACTIONS(1070),
    [anon_sym_not] = ACTIONS(1070),
    [anon_sym_and] = ACTIONS(1070),
    [anon_sym_or] = ACTIONS(1070),
    [anon_sym_PLUS] = ACTIONS(1070),
    [anon_sym_DASH] = ACTIONS(1070),
    [anon_sym_SLASH] = ACTIONS(1070),
    [anon_sym_SLASH_SLASH] = ACTIONS(1070),
    [anon_sym_PERCENT] = ACTIONS(1070),
    [anon_sym_PIPE] = ACTIONS(1070),
    [anon_sym_CARET] = ACTIONS(1070),
    [anon_sym_AMP] = ACTIONS(1070),
    [anon_sym_LT_LT] = ACTIONS(1070),
    [anon_sym_GT_GT] = ACTIONS(1070),
    [anon_sym_GT_GT_GT] = ACTIONS(1070),
    [anon_sym_TILDE] = ACTIONS(1070),
    [anon_sym_typeof] = ACTIONS(1070),
    [anon_sym_void] = ACTIONS(1070),
    [anon_sym_LT] = ACTIONS(1070),
    [anon_sym_LT_EQ] = ACTIONS(1070),
    [anon_sym_EQ_EQ] = ACTIONS(1070),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1070),
    [anon_sym_BANG_EQ] = ACTIONS(1070),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1070),
    [anon_sym_GT_EQ] = ACTIONS(1070),
    [anon_sym_GT] = ACTIONS(1070),
    [anon_sym_is] = ACTIONS(1070),
    [anon_sym_instanceof] = ACTIONS(1070),
    [anon_sym_new] = ACTIONS(1070),
    [anon_sym_QMARK] = ACTIONS(1070),
    [anon_sym_await] = ACTIONS(1070),
    [anon_sym_LBRACE] = ACTIONS(1070),
    [sym_this] = ACTIONS(1070),
    [sym_true] = ACTIONS(1070),
    [sym_false] = ACTIONS(1070),
    [sym_none] = ACTIONS(1070),
    [sym_string] = ACTIONS(1070),
    [sym_verbatim] = ACTIONS(1070),
    [sym_number] = ACTIONS(1070),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1068),
    [sym_regex] = ACTIONS(1068),
    [sym_fstring_start] = ACTIONS(1068),
  },
  [STATE(458)] = {
    [sym_f_string] = STATE(438),
    [aux_sym_concatenated_string_repeat1] = STATE(438),
    [ts_builtin_sym_end] = ACTIONS(925),
    [sym_identifier] = ACTIONS(927),
    [anon_sym_SEMI] = ACTIONS(927),
    [anon_sym_return] = ACTIONS(927),
    [anon_sym_async] = ACTIONS(927),
    [anon_sym_def] = ACTIONS(927),
    [anon_sym_COMMA] = ACTIONS(927),
    [anon_sym_LPAREN] = ACTIONS(927),
    [anon_sym_LBRACK] = ACTIONS(927),
    [anon_sym_del] = ACTIONS(927),
    [anon_sym_raise] = ACTIONS(927),
    [sym_pass_statement] = ACTIONS(927),
    [sym_break_statement] = ACTIONS(927),
    [sym_continue_statement] = ACTIONS(927),
    [sym_debugger_statement] = ACTIONS(927),
    [anon_sym_assert] = ACTIONS(927),
    [anon_sym_global] = ACTIONS(927),
    [anon_sym_nonlocal] = ACTIONS(927),
    [anon_sym_yield] = ACTIONS(927),
    [anon_sym_from] = ACTIONS(927),
    [anon_sym_import] = ACTIONS(927),
    [anon_sym_DOT] = ACTIONS(927),
    [anon_sym_if] = ACTIONS(927),
    [anon_sym_for] = ACTIONS(927),
    [anon_sym_in] = ACTIONS(927),
    [anon_sym_while] = ACTIONS(927),
    [anon_sym_do] = ACTIONS(927),
    [anon_sym_try] = ACTIONS(927),
    [anon_sym_with] = ACTIONS(927),
    [anon_sym_STAR] = ACTIONS(927),
    [anon_sym_STAR_STAR] = ACTIONS(927),
    [anon_sym_class] = ACTIONS(927),
    [anon_sym_AT] = ACTIONS(927),
    [anon_sym_not] = ACTIONS(927),
    [anon_sym_and] = ACTIONS(927),
    [anon_sym_or] = ACTIONS(927),
    [anon_sym_PLUS] = ACTIONS(927),
    [anon_sym_DASH] = ACTIONS(927),
    [anon_sym_SLASH] = ACTIONS(927),
    [anon_sym_SLASH_SLASH] = ACTIONS(927),
    [anon_sym_PERCENT] = ACTIONS(927),
    [anon_sym_PIPE] = ACTIONS(927),
    [anon_sym_CARET] = ACTIONS(927),
    [anon_sym_AMP] = ACTIONS(927),
    [anon_sym_LT_LT] = ACTIONS(927),
    [anon_sym_GT_GT] = ACTIONS(927),
    [anon_sym_GT_GT_GT] = ACTIONS(927),
    [anon_sym_TILDE] = ACTIONS(927),
    [anon_sym_typeof] = ACTIONS(927),
    [anon_sym_void] = ACTIONS(927),
    [anon_sym_LT] = ACTIONS(927),
    [anon_sym_LT_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(927),
    [anon_sym_GT_EQ] = ACTIONS(927),
    [anon_sym_GT] = ACTIONS(927),
    [anon_sym_is] = ACTIONS(927),
    [anon_sym_instanceof] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
    [anon_sym_QMARK] = ACTIONS(927),
    [anon_sym_await] = ACTIONS(927),
    [anon_sym_LBRACE] = ACTIONS(927),
    [sym_this] = ACTIONS(927),
    [sym_true] = ACTIONS(927),
    [sym_false] = ACTIONS(927),
    [sym_none] = ACTIONS(927),
    [sym_string] = ACTIONS(1181),
    [sym_verbatim] = ACTIONS(927),
    [sym_number] = ACTIONS(927),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(925),
    [sym_fstring_start] = ACTIONS(1184),
  },
  [STATE(459)] = {
    [aux_sym_expression_list_repeat1] = STATE(540),
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1164),
    [anon_sym_SEMI] = ACTIONS(1164),
    [anon_sym_return] = ACTIONS(1164),
    [anon_sym_async] = ACTIONS(1164),
    [anon_sym_def] = ACTIONS(1164),
    [anon_sym_COMMA] = ACTIONS(1187),
    [anon_sym_LPAREN] = ACTIONS(1164),
    [anon_sym_LBRACK] = ACTIONS(1164),
    [anon_sym_del] = ACTIONS(1164),
    [anon_sym_raise] = ACTIONS(1164),
    [sym_pass_statement] = ACTIONS(1164),
    [sym_break_statement] = ACTIONS(1164),
    [sym_continue_statement] = ACTIONS(1164),
    [sym_debugger_statement] = ACTIONS(1164),
    [anon_sym_assert] = ACTIONS(1164),
    [anon_sym_global] = ACTIONS(1164),
    [anon_sym_nonlocal] = ACTIONS(1164),
    [anon_sym_yield] = ACTIONS(1164),
    [anon_sym_from] = ACTIONS(1164),
    [anon_sym_import] = ACTIONS(1164),
    [anon_sym_DOT] = ACTIONS(1164),
    [anon_sym_if] = ACTIONS(1164),
    [anon_sym_for] = ACTIONS(1164),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1164),
    [anon_sym_do] = ACTIONS(1164),
    [anon_sym_try] = ACTIONS(1164),
    [anon_sym_with] = ACTIONS(1164),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1164),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1164),
    [anon_sym_typeof] = ACTIONS(1164),
    [anon_sym_void] = ACTIONS(1164),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1164),
    [anon_sym_QMARK] = ACTIONS(1164),
    [anon_sym_await] = ACTIONS(1164),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [sym_this] = ACTIONS(1164),
    [sym_true] = ACTIONS(1164),
    [sym_false] = ACTIONS(1164),
    [sym_none] = ACTIONS(1164),
    [sym_string] = ACTIONS(1164),
    [sym_verbatim] = ACTIONS(1164),
    [sym_number] = ACTIONS(1164),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1162),
    [sym_regex] = ACTIONS(1162),
    [sym_fstring_start] = ACTIONS(1162),
  },
  [STATE(460)] = {
    [sym_identifier] = ACTIONS(1190),
    [anon_sym_SEMI] = ACTIONS(1190),
    [anon_sym_return] = ACTIONS(1190),
    [anon_sym_async] = ACTIONS(1190),
    [anon_sym_def] = ACTIONS(1190),
    [anon_sym_COMMA] = ACTIONS(1190),
    [anon_sym_LPAREN] = ACTIONS(1190),
    [anon_sym_LBRACK] = ACTIONS(1190),
    [anon_sym_del] = ACTIONS(1190),
    [anon_sym_raise] = ACTIONS(1190),
    [sym_pass_statement] = ACTIONS(1190),
    [sym_break_statement] = ACTIONS(1190),
    [sym_continue_statement] = ACTIONS(1190),
    [sym_debugger_statement] = ACTIONS(1190),
    [anon_sym_assert] = ACTIONS(1190),
    [anon_sym_global] = ACTIONS(1190),
    [anon_sym_nonlocal] = ACTIONS(1190),
    [anon_sym_yield] = ACTIONS(1190),
    [anon_sym_from] = ACTIONS(1190),
    [anon_sym_import] = ACTIONS(1190),
    [anon_sym_DOT] = ACTIONS(1190),
    [anon_sym_if] = ACTIONS(1190),
    [anon_sym_for] = ACTIONS(1190),
    [anon_sym_in] = ACTIONS(1190),
    [anon_sym_while] = ACTIONS(1190),
    [anon_sym_do] = ACTIONS(1190),
    [anon_sym_try] = ACTIONS(1190),
    [anon_sym_with] = ACTIONS(1190),
    [anon_sym_STAR] = ACTIONS(1190),
    [anon_sym_STAR_STAR] = ACTIONS(1190),
    [anon_sym_class] = ACTIONS(1190),
    [anon_sym_AT] = ACTIONS(1190),
    [anon_sym_not] = ACTIONS(1190),
    [anon_sym_and] = ACTIONS(1190),
    [anon_sym_or] = ACTIONS(1190),
    [anon_sym_PLUS] = ACTIONS(1190),
    [anon_sym_DASH] = ACTIONS(1190),
    [anon_sym_SLASH] = ACTIONS(1190),
    [anon_sym_SLASH_SLASH] = ACTIONS(1190),
    [anon_sym_PERCENT] = ACTIONS(1190),
    [anon_sym_PIPE] = ACTIONS(1190),
    [anon_sym_CARET] = ACTIONS(1190),
    [anon_sym_AMP] = ACTIONS(1190),
    [anon_sym_LT_LT] = ACTIONS(1190),
    [anon_sym_GT_GT] = ACTIONS(1190),
    [anon_sym_GT_GT_GT] = ACTIONS(1190),
    [anon_sym_TILDE] = ACTIONS(1190),
    [anon_sym_typeof] = ACTIONS(1190),
    [anon_sym_void] = ACTIONS(1190),
    [anon_sym_LT] = ACTIONS(1190),
    [anon_sym_LT_EQ] = ACTIONS(1190),
    [anon_sym_EQ_EQ] = ACTIONS(1190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1190),
    [anon_sym_BANG_EQ] = ACTIONS(1190),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1190),
    [anon_sym_GT_EQ] = ACTIONS(1190),
    [anon_sym_GT] = ACTIONS(1190),
    [anon_sym_is] = ACTIONS(1190),
    [anon_sym_instanceof] = ACTIONS(1190),
    [anon_sym_new] = ACTIONS(1190),
    [anon_sym_QMARK] = ACTIONS(1190),
    [anon_sym_await] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1190),
    [sym_this] = ACTIONS(1190),
    [sym_true] = ACTIONS(1190),
    [sym_false] = ACTIONS(1190),
    [sym_none] = ACTIONS(1190),
    [sym_string] = ACTIONS(1190),
    [sym_verbatim] = ACTIONS(1190),
    [sym_number] = ACTIONS(1190),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1192),
    [sym__dedent] = ACTIONS(1192),
    [sym_regex] = ACTIONS(1192),
    [sym_fstring_start] = ACTIONS(1192),
  },
  [STATE(461)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1194),
    [anon_sym_SEMI] = ACTIONS(1194),
    [anon_sym_return] = ACTIONS(1194),
    [anon_sym_async] = ACTIONS(1194),
    [anon_sym_def] = ACTIONS(1194),
    [anon_sym_COMMA] = ACTIONS(1194),
    [anon_sym_LPAREN] = ACTIONS(1194),
    [anon_sym_LBRACK] = ACTIONS(1194),
    [anon_sym_del] = ACTIONS(1194),
    [anon_sym_raise] = ACTIONS(1194),
    [sym_pass_statement] = ACTIONS(1194),
    [sym_break_statement] = ACTIONS(1194),
    [sym_continue_statement] = ACTIONS(1194),
    [sym_debugger_statement] = ACTIONS(1194),
    [anon_sym_assert] = ACTIONS(1194),
    [anon_sym_global] = ACTIONS(1194),
    [anon_sym_nonlocal] = ACTIONS(1194),
    [anon_sym_yield] = ACTIONS(1194),
    [anon_sym_from] = ACTIONS(1194),
    [anon_sym_import] = ACTIONS(1194),
    [anon_sym_DOT] = ACTIONS(1194),
    [anon_sym_if] = ACTIONS(1194),
    [anon_sym_for] = ACTIONS(1194),
    [anon_sym_in] = ACTIONS(1194),
    [anon_sym_while] = ACTIONS(1194),
    [anon_sym_do] = ACTIONS(1194),
    [anon_sym_try] = ACTIONS(1194),
    [anon_sym_with] = ACTIONS(1194),
    [anon_sym_STAR] = ACTIONS(1194),
    [anon_sym_STAR_STAR] = ACTIONS(1194),
    [anon_sym_class] = ACTIONS(1194),
    [anon_sym_AT] = ACTIONS(1194),
    [anon_sym_not] = ACTIONS(1194),
    [anon_sym_and] = ACTIONS(1194),
    [anon_sym_or] = ACTIONS(1194),
    [anon_sym_PLUS] = ACTIONS(1194),
    [anon_sym_DASH] = ACTIONS(1194),
    [anon_sym_SLASH] = ACTIONS(1194),
    [anon_sym_SLASH_SLASH] = ACTIONS(1194),
    [anon_sym_PERCENT] = ACTIONS(1194),
    [anon_sym_PIPE] = ACTIONS(1194),
    [anon_sym_CARET] = ACTIONS(1194),
    [anon_sym_AMP] = ACTIONS(1194),
    [anon_sym_LT_LT] = ACTIONS(1194),
    [anon_sym_GT_GT] = ACTIONS(1194),
    [anon_sym_GT_GT_GT] = ACTIONS(1194),
    [anon_sym_TILDE] = ACTIONS(1194),
    [anon_sym_typeof] = ACTIONS(1194),
    [anon_sym_void] = ACTIONS(1194),
    [anon_sym_LT] = ACTIONS(1194),
    [anon_sym_LT_EQ] = ACTIONS(1194),
    [anon_sym_EQ_EQ] = ACTIONS(1194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1194),
    [anon_sym_BANG_EQ] = ACTIONS(1194),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1194),
    [anon_sym_GT_EQ] = ACTIONS(1194),
    [anon_sym_GT] = ACTIONS(1194),
    [anon_sym_is] = ACTIONS(1194),
    [anon_sym_instanceof] = ACTIONS(1194),
    [anon_sym_new] = ACTIONS(1194),
    [anon_sym_QMARK] = ACTIONS(1194),
    [anon_sym_await] = ACTIONS(1194),
    [anon_sym_LBRACE] = ACTIONS(1194),
    [sym_this] = ACTIONS(1194),
    [sym_true] = ACTIONS(1194),
    [sym_false] = ACTIONS(1194),
    [sym_none] = ACTIONS(1194),
    [sym_string] = ACTIONS(1194),
    [sym_verbatim] = ACTIONS(1194),
    [sym_number] = ACTIONS(1194),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1196),
    [sym_regex] = ACTIONS(1196),
    [sym_fstring_start] = ACTIONS(1196),
  },
  [STATE(462)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1767),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(459),
    [anon_sym_raise] = ACTIONS(461),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(463),
    [anon_sym_yield] = ACTIONS(465),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(463)] = {
    [sym_argument_list] = STATE(564),
    [ts_builtin_sym_end] = ACTIONS(1198),
    [sym_identifier] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym_return] = ACTIONS(1200),
    [anon_sym_async] = ACTIONS(1200),
    [anon_sym_def] = ACTIONS(1200),
    [anon_sym_COMMA] = ACTIONS(1200),
    [anon_sym_LPAREN] = ACTIONS(1202),
    [anon_sym_LBRACK] = ACTIONS(1204),
    [anon_sym_del] = ACTIONS(1200),
    [anon_sym_raise] = ACTIONS(1200),
    [sym_pass_statement] = ACTIONS(1200),
    [sym_break_statement] = ACTIONS(1200),
    [sym_continue_statement] = ACTIONS(1200),
    [sym_debugger_statement] = ACTIONS(1200),
    [anon_sym_assert] = ACTIONS(1200),
    [anon_sym_global] = ACTIONS(1200),
    [anon_sym_nonlocal] = ACTIONS(1200),
    [anon_sym_yield] = ACTIONS(1200),
    [anon_sym_from] = ACTIONS(1200),
    [anon_sym_import] = ACTIONS(1200),
    [anon_sym_DOT] = ACTIONS(1206),
    [anon_sym_if] = ACTIONS(1200),
    [anon_sym_for] = ACTIONS(1200),
    [anon_sym_in] = ACTIONS(1200),
    [anon_sym_while] = ACTIONS(1200),
    [anon_sym_do] = ACTIONS(1200),
    [anon_sym_try] = ACTIONS(1200),
    [anon_sym_with] = ACTIONS(1200),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_STAR_STAR] = ACTIONS(1200),
    [anon_sym_class] = ACTIONS(1200),
    [anon_sym_AT] = ACTIONS(1200),
    [anon_sym_not] = ACTIONS(1200),
    [anon_sym_and] = ACTIONS(1200),
    [anon_sym_or] = ACTIONS(1200),
    [anon_sym_PLUS] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [anon_sym_SLASH] = ACTIONS(1200),
    [anon_sym_SLASH_SLASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(1200),
    [anon_sym_PIPE] = ACTIONS(1200),
    [anon_sym_CARET] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_LT_LT] = ACTIONS(1200),
    [anon_sym_GT_GT] = ACTIONS(1200),
    [anon_sym_GT_GT_GT] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_typeof] = ACTIONS(1200),
    [anon_sym_void] = ACTIONS(1200),
    [anon_sym_LT] = ACTIONS(1200),
    [anon_sym_LT_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1200),
    [anon_sym_GT_EQ] = ACTIONS(1200),
    [anon_sym_GT] = ACTIONS(1200),
    [anon_sym_is] = ACTIONS(1200),
    [anon_sym_instanceof] = ACTIONS(1200),
    [anon_sym_new] = ACTIONS(1200),
    [anon_sym_QMARK] = ACTIONS(1208),
    [anon_sym_await] = ACTIONS(1200),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [sym_this] = ACTIONS(1200),
    [sym_true] = ACTIONS(1200),
    [sym_false] = ACTIONS(1200),
    [sym_none] = ACTIONS(1200),
    [sym_string] = ACTIONS(1200),
    [sym_verbatim] = ACTIONS(1200),
    [sym_number] = ACTIONS(1200),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1198),
    [sym_fstring_start] = ACTIONS(1198),
  },
  [STATE(464)] = {
    [sym_argument_list] = STATE(564),
    [ts_builtin_sym_end] = ACTIONS(1210),
    [sym_identifier] = ACTIONS(1212),
    [anon_sym_SEMI] = ACTIONS(1212),
    [anon_sym_return] = ACTIONS(1212),
    [anon_sym_async] = ACTIONS(1212),
    [anon_sym_def] = ACTIONS(1212),
    [anon_sym_COMMA] = ACTIONS(1212),
    [anon_sym_LPAREN] = ACTIONS(1202),
    [anon_sym_LBRACK] = ACTIONS(1204),
    [anon_sym_del] = ACTIONS(1212),
    [anon_sym_raise] = ACTIONS(1212),
    [sym_pass_statement] = ACTIONS(1212),
    [sym_break_statement] = ACTIONS(1212),
    [sym_continue_statement] = ACTIONS(1212),
    [sym_debugger_statement] = ACTIONS(1212),
    [anon_sym_assert] = ACTIONS(1212),
    [anon_sym_global] = ACTIONS(1212),
    [anon_sym_nonlocal] = ACTIONS(1212),
    [anon_sym_yield] = ACTIONS(1212),
    [anon_sym_from] = ACTIONS(1212),
    [anon_sym_import] = ACTIONS(1212),
    [anon_sym_DOT] = ACTIONS(1206),
    [anon_sym_if] = ACTIONS(1212),
    [anon_sym_for] = ACTIONS(1212),
    [anon_sym_in] = ACTIONS(1212),
    [anon_sym_while] = ACTIONS(1212),
    [anon_sym_do] = ACTIONS(1212),
    [anon_sym_try] = ACTIONS(1212),
    [anon_sym_with] = ACTIONS(1212),
    [anon_sym_STAR] = ACTIONS(1212),
    [anon_sym_STAR_STAR] = ACTIONS(1212),
    [anon_sym_class] = ACTIONS(1212),
    [anon_sym_AT] = ACTIONS(1212),
    [anon_sym_not] = ACTIONS(1212),
    [anon_sym_and] = ACTIONS(1212),
    [anon_sym_or] = ACTIONS(1212),
    [anon_sym_PLUS] = ACTIONS(1212),
    [anon_sym_DASH] = ACTIONS(1212),
    [anon_sym_SLASH] = ACTIONS(1212),
    [anon_sym_SLASH_SLASH] = ACTIONS(1212),
    [anon_sym_PERCENT] = ACTIONS(1212),
    [anon_sym_PIPE] = ACTIONS(1212),
    [anon_sym_CARET] = ACTIONS(1212),
    [anon_sym_AMP] = ACTIONS(1212),
    [anon_sym_LT_LT] = ACTIONS(1212),
    [anon_sym_GT_GT] = ACTIONS(1212),
    [anon_sym_GT_GT_GT] = ACTIONS(1212),
    [anon_sym_TILDE] = ACTIONS(1212),
    [anon_sym_typeof] = ACTIONS(1212),
    [anon_sym_void] = ACTIONS(1212),
    [anon_sym_LT] = ACTIONS(1212),
    [anon_sym_LT_EQ] = ACTIONS(1212),
    [anon_sym_EQ_EQ] = ACTIONS(1212),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1212),
    [anon_sym_BANG_EQ] = ACTIONS(1212),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1212),
    [anon_sym_GT_EQ] = ACTIONS(1212),
    [anon_sym_GT] = ACTIONS(1212),
    [anon_sym_is] = ACTIONS(1212),
    [anon_sym_instanceof] = ACTIONS(1212),
    [anon_sym_new] = ACTIONS(1212),
    [anon_sym_QMARK] = ACTIONS(1212),
    [anon_sym_await] = ACTIONS(1212),
    [anon_sym_LBRACE] = ACTIONS(1212),
    [sym_this] = ACTIONS(1212),
    [sym_true] = ACTIONS(1212),
    [sym_false] = ACTIONS(1212),
    [sym_none] = ACTIONS(1212),
    [sym_string] = ACTIONS(1212),
    [sym_verbatim] = ACTIONS(1212),
    [sym_number] = ACTIONS(1212),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1210),
    [sym_fstring_start] = ACTIONS(1210),
  },
  [STATE(465)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2532),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2532),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1861),
    [sym__primary_expression] = STATE(1983),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(581),
    [anon_sym_async] = ACTIONS(583),
    [anon_sym_def] = ACTIONS(585),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(587),
    [anon_sym_raise] = ACTIONS(589),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(591),
    [anon_sym_yield] = ACTIONS(593),
    [anon_sym_not] = ACTIONS(595),
    [anon_sym_PLUS] = ACTIONS(597),
    [anon_sym_DASH] = ACTIONS(597),
    [anon_sym_TILDE] = ACTIONS(597),
    [anon_sym_typeof] = ACTIONS(597),
    [anon_sym_void] = ACTIONS(597),
    [anon_sym_new] = ACTIONS(599),
    [anon_sym_await] = ACTIONS(601),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(466)] = {
    [sym_argument_list] = STATE(678),
    [sym_identifier] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym_return] = ACTIONS(1200),
    [anon_sym_async] = ACTIONS(1200),
    [anon_sym_def] = ACTIONS(1200),
    [anon_sym_COMMA] = ACTIONS(1200),
    [anon_sym_LPAREN] = ACTIONS(1214),
    [anon_sym_LBRACK] = ACTIONS(1216),
    [anon_sym_del] = ACTIONS(1200),
    [anon_sym_raise] = ACTIONS(1200),
    [sym_pass_statement] = ACTIONS(1200),
    [sym_break_statement] = ACTIONS(1200),
    [sym_continue_statement] = ACTIONS(1200),
    [sym_debugger_statement] = ACTIONS(1200),
    [anon_sym_assert] = ACTIONS(1200),
    [anon_sym_global] = ACTIONS(1200),
    [anon_sym_nonlocal] = ACTIONS(1200),
    [anon_sym_yield] = ACTIONS(1200),
    [anon_sym_from] = ACTIONS(1200),
    [anon_sym_import] = ACTIONS(1200),
    [anon_sym_DOT] = ACTIONS(1218),
    [anon_sym_if] = ACTIONS(1200),
    [anon_sym_for] = ACTIONS(1200),
    [anon_sym_in] = ACTIONS(1200),
    [anon_sym_while] = ACTIONS(1200),
    [anon_sym_do] = ACTIONS(1200),
    [anon_sym_try] = ACTIONS(1200),
    [anon_sym_with] = ACTIONS(1200),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_STAR_STAR] = ACTIONS(1200),
    [anon_sym_class] = ACTIONS(1200),
    [anon_sym_AT] = ACTIONS(1200),
    [anon_sym_not] = ACTIONS(1200),
    [anon_sym_and] = ACTIONS(1200),
    [anon_sym_or] = ACTIONS(1200),
    [anon_sym_PLUS] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [anon_sym_SLASH] = ACTIONS(1200),
    [anon_sym_SLASH_SLASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(1200),
    [anon_sym_PIPE] = ACTIONS(1200),
    [anon_sym_CARET] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_LT_LT] = ACTIONS(1200),
    [anon_sym_GT_GT] = ACTIONS(1200),
    [anon_sym_GT_GT_GT] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_typeof] = ACTIONS(1200),
    [anon_sym_void] = ACTIONS(1200),
    [anon_sym_LT] = ACTIONS(1200),
    [anon_sym_LT_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1200),
    [anon_sym_GT_EQ] = ACTIONS(1200),
    [anon_sym_GT] = ACTIONS(1200),
    [anon_sym_is] = ACTIONS(1200),
    [anon_sym_instanceof] = ACTIONS(1200),
    [anon_sym_new] = ACTIONS(1200),
    [anon_sym_QMARK] = ACTIONS(1220),
    [anon_sym_await] = ACTIONS(1200),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [sym_this] = ACTIONS(1200),
    [sym_true] = ACTIONS(1200),
    [sym_false] = ACTIONS(1200),
    [sym_none] = ACTIONS(1200),
    [sym_string] = ACTIONS(1200),
    [sym_verbatim] = ACTIONS(1200),
    [sym_number] = ACTIONS(1200),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1198),
    [sym_regex] = ACTIONS(1198),
    [sym_fstring_start] = ACTIONS(1198),
  },
  [STATE(467)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1222),
    [sym_identifier] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym_return] = ACTIONS(1224),
    [anon_sym_async] = ACTIONS(1224),
    [anon_sym_def] = ACTIONS(1224),
    [anon_sym_COMMA] = ACTIONS(1224),
    [anon_sym_LPAREN] = ACTIONS(1224),
    [anon_sym_LBRACK] = ACTIONS(1224),
    [anon_sym_del] = ACTIONS(1224),
    [anon_sym_raise] = ACTIONS(1224),
    [sym_pass_statement] = ACTIONS(1224),
    [sym_break_statement] = ACTIONS(1224),
    [sym_continue_statement] = ACTIONS(1224),
    [sym_debugger_statement] = ACTIONS(1224),
    [anon_sym_assert] = ACTIONS(1224),
    [anon_sym_global] = ACTIONS(1224),
    [anon_sym_nonlocal] = ACTIONS(1224),
    [anon_sym_yield] = ACTIONS(1224),
    [anon_sym_from] = ACTIONS(1224),
    [anon_sym_import] = ACTIONS(1224),
    [anon_sym_DOT] = ACTIONS(1224),
    [anon_sym_if] = ACTIONS(1224),
    [anon_sym_for] = ACTIONS(1224),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1224),
    [anon_sym_do] = ACTIONS(1224),
    [anon_sym_try] = ACTIONS(1224),
    [anon_sym_with] = ACTIONS(1224),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1224),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_typeof] = ACTIONS(1224),
    [anon_sym_void] = ACTIONS(1224),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1224),
    [anon_sym_QMARK] = ACTIONS(1224),
    [anon_sym_await] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [sym_this] = ACTIONS(1224),
    [sym_true] = ACTIONS(1224),
    [sym_false] = ACTIONS(1224),
    [sym_none] = ACTIONS(1224),
    [sym_string] = ACTIONS(1224),
    [sym_verbatim] = ACTIONS(1224),
    [sym_number] = ACTIONS(1224),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1222),
    [sym_fstring_start] = ACTIONS(1222),
  },
  [STATE(468)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1228),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1228),
    [anon_sym_SLASH_SLASH] = ACTIONS(1228),
    [anon_sym_PERCENT] = ACTIONS(1228),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(469)] = {
    [sym_identifier] = ACTIONS(1230),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1230),
    [anon_sym_async] = ACTIONS(1230),
    [anon_sym_def] = ACTIONS(1230),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1234),
    [anon_sym_LBRACK] = ACTIONS(1234),
    [anon_sym_del] = ACTIONS(1230),
    [anon_sym_raise] = ACTIONS(1230),
    [sym_pass_statement] = ACTIONS(1230),
    [sym_break_statement] = ACTIONS(1230),
    [sym_continue_statement] = ACTIONS(1230),
    [sym_debugger_statement] = ACTIONS(1230),
    [anon_sym_assert] = ACTIONS(1230),
    [anon_sym_global] = ACTIONS(1230),
    [anon_sym_nonlocal] = ACTIONS(1230),
    [anon_sym_yield] = ACTIONS(1230),
    [anon_sym_from] = ACTIONS(1230),
    [anon_sym_import] = ACTIONS(1230),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1230),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1230),
    [anon_sym_do] = ACTIONS(1230),
    [anon_sym_try] = ACTIONS(1230),
    [anon_sym_with] = ACTIONS(1230),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1230),
    [anon_sym_AT] = ACTIONS(1234),
    [anon_sym_not] = ACTIONS(1234),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1230),
    [anon_sym_typeof] = ACTIONS(1230),
    [anon_sym_void] = ACTIONS(1230),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1230),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1230),
    [anon_sym_LBRACE] = ACTIONS(1230),
    [sym_this] = ACTIONS(1230),
    [sym_true] = ACTIONS(1230),
    [sym_false] = ACTIONS(1230),
    [sym_none] = ACTIONS(1230),
    [sym_string] = ACTIONS(1230),
    [sym_verbatim] = ACTIONS(1230),
    [sym_number] = ACTIONS(1230),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym__dedent] = ACTIONS(1239),
    [sym_regex] = ACTIONS(1239),
    [sym_fstring_start] = ACTIONS(1239),
  },
  [STATE(470)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1241),
    [sym_identifier] = ACTIONS(1243),
    [anon_sym_SEMI] = ACTIONS(1243),
    [anon_sym_return] = ACTIONS(1243),
    [anon_sym_async] = ACTIONS(1243),
    [anon_sym_def] = ACTIONS(1243),
    [anon_sym_COMMA] = ACTIONS(1243),
    [anon_sym_LPAREN] = ACTIONS(1243),
    [anon_sym_LBRACK] = ACTIONS(1243),
    [anon_sym_del] = ACTIONS(1243),
    [anon_sym_raise] = ACTIONS(1243),
    [sym_pass_statement] = ACTIONS(1243),
    [sym_break_statement] = ACTIONS(1243),
    [sym_continue_statement] = ACTIONS(1243),
    [sym_debugger_statement] = ACTIONS(1243),
    [anon_sym_assert] = ACTIONS(1243),
    [anon_sym_global] = ACTIONS(1243),
    [anon_sym_nonlocal] = ACTIONS(1243),
    [anon_sym_yield] = ACTIONS(1243),
    [anon_sym_from] = ACTIONS(1243),
    [anon_sym_import] = ACTIONS(1243),
    [anon_sym_DOT] = ACTIONS(1243),
    [anon_sym_if] = ACTIONS(1243),
    [anon_sym_for] = ACTIONS(1243),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1243),
    [anon_sym_do] = ACTIONS(1243),
    [anon_sym_try] = ACTIONS(1243),
    [anon_sym_with] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1243),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1243),
    [anon_sym_or] = ACTIONS(1243),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1243),
    [anon_sym_typeof] = ACTIONS(1243),
    [anon_sym_void] = ACTIONS(1243),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1243),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_await] = ACTIONS(1243),
    [anon_sym_LBRACE] = ACTIONS(1243),
    [sym_this] = ACTIONS(1243),
    [sym_true] = ACTIONS(1243),
    [sym_false] = ACTIONS(1243),
    [sym_none] = ACTIONS(1243),
    [sym_string] = ACTIONS(1243),
    [sym_verbatim] = ACTIONS(1243),
    [sym_number] = ACTIONS(1243),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1241),
    [sym_fstring_start] = ACTIONS(1241),
  },
  [STATE(471)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1241),
    [sym_identifier] = ACTIONS(1243),
    [anon_sym_SEMI] = ACTIONS(1243),
    [anon_sym_return] = ACTIONS(1243),
    [anon_sym_async] = ACTIONS(1243),
    [anon_sym_def] = ACTIONS(1243),
    [anon_sym_COMMA] = ACTIONS(1243),
    [anon_sym_LPAREN] = ACTIONS(1243),
    [anon_sym_LBRACK] = ACTIONS(1243),
    [anon_sym_del] = ACTIONS(1243),
    [anon_sym_raise] = ACTIONS(1243),
    [sym_pass_statement] = ACTIONS(1243),
    [sym_break_statement] = ACTIONS(1243),
    [sym_continue_statement] = ACTIONS(1243),
    [sym_debugger_statement] = ACTIONS(1243),
    [anon_sym_assert] = ACTIONS(1243),
    [anon_sym_global] = ACTIONS(1243),
    [anon_sym_nonlocal] = ACTIONS(1243),
    [anon_sym_yield] = ACTIONS(1243),
    [anon_sym_from] = ACTIONS(1243),
    [anon_sym_import] = ACTIONS(1243),
    [anon_sym_DOT] = ACTIONS(1243),
    [anon_sym_if] = ACTIONS(1243),
    [anon_sym_for] = ACTIONS(1243),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1243),
    [anon_sym_do] = ACTIONS(1243),
    [anon_sym_try] = ACTIONS(1243),
    [anon_sym_with] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1243),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1243),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1243),
    [anon_sym_typeof] = ACTIONS(1243),
    [anon_sym_void] = ACTIONS(1243),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1243),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_await] = ACTIONS(1243),
    [anon_sym_LBRACE] = ACTIONS(1243),
    [sym_this] = ACTIONS(1243),
    [sym_true] = ACTIONS(1243),
    [sym_false] = ACTIONS(1243),
    [sym_none] = ACTIONS(1243),
    [sym_string] = ACTIONS(1243),
    [sym_verbatim] = ACTIONS(1243),
    [sym_number] = ACTIONS(1243),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1241),
    [sym_fstring_start] = ACTIONS(1241),
  },
  [STATE(472)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(473)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(474)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(475)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(476)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(477)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(477),
    [ts_builtin_sym_end] = ACTIONS(1245),
    [sym_identifier] = ACTIONS(1247),
    [anon_sym_SEMI] = ACTIONS(1247),
    [anon_sym_return] = ACTIONS(1247),
    [anon_sym_async] = ACTIONS(1247),
    [anon_sym_def] = ACTIONS(1247),
    [anon_sym_COMMA] = ACTIONS(1247),
    [anon_sym_LPAREN] = ACTIONS(1247),
    [anon_sym_LBRACK] = ACTIONS(1247),
    [anon_sym_del] = ACTIONS(1247),
    [anon_sym_raise] = ACTIONS(1247),
    [sym_pass_statement] = ACTIONS(1247),
    [sym_break_statement] = ACTIONS(1247),
    [sym_continue_statement] = ACTIONS(1247),
    [sym_debugger_statement] = ACTIONS(1247),
    [anon_sym_assert] = ACTIONS(1247),
    [anon_sym_global] = ACTIONS(1247),
    [anon_sym_nonlocal] = ACTIONS(1247),
    [anon_sym_yield] = ACTIONS(1247),
    [anon_sym_from] = ACTIONS(1247),
    [anon_sym_import] = ACTIONS(1247),
    [anon_sym_DOT] = ACTIONS(1247),
    [anon_sym_if] = ACTIONS(1247),
    [anon_sym_for] = ACTIONS(1247),
    [anon_sym_in] = ACTIONS(1249),
    [anon_sym_while] = ACTIONS(1247),
    [anon_sym_do] = ACTIONS(1247),
    [anon_sym_try] = ACTIONS(1247),
    [anon_sym_with] = ACTIONS(1247),
    [anon_sym_STAR] = ACTIONS(1247),
    [anon_sym_STAR_STAR] = ACTIONS(1247),
    [anon_sym_class] = ACTIONS(1247),
    [anon_sym_AT] = ACTIONS(1247),
    [anon_sym_not] = ACTIONS(1252),
    [anon_sym_and] = ACTIONS(1247),
    [anon_sym_or] = ACTIONS(1247),
    [anon_sym_PLUS] = ACTIONS(1247),
    [anon_sym_DASH] = ACTIONS(1247),
    [anon_sym_SLASH] = ACTIONS(1247),
    [anon_sym_SLASH_SLASH] = ACTIONS(1247),
    [anon_sym_PERCENT] = ACTIONS(1247),
    [anon_sym_PIPE] = ACTIONS(1247),
    [anon_sym_CARET] = ACTIONS(1247),
    [anon_sym_AMP] = ACTIONS(1247),
    [anon_sym_LT_LT] = ACTIONS(1247),
    [anon_sym_GT_GT] = ACTIONS(1247),
    [anon_sym_GT_GT_GT] = ACTIONS(1247),
    [anon_sym_TILDE] = ACTIONS(1247),
    [anon_sym_typeof] = ACTIONS(1247),
    [anon_sym_void] = ACTIONS(1247),
    [anon_sym_LT] = ACTIONS(1249),
    [anon_sym_LT_EQ] = ACTIONS(1249),
    [anon_sym_EQ_EQ] = ACTIONS(1249),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1249),
    [anon_sym_BANG_EQ] = ACTIONS(1249),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1249),
    [anon_sym_GT_EQ] = ACTIONS(1249),
    [anon_sym_GT] = ACTIONS(1249),
    [anon_sym_is] = ACTIONS(1255),
    [anon_sym_instanceof] = ACTIONS(1249),
    [anon_sym_new] = ACTIONS(1247),
    [anon_sym_QMARK] = ACTIONS(1247),
    [anon_sym_await] = ACTIONS(1247),
    [anon_sym_LBRACE] = ACTIONS(1247),
    [sym_this] = ACTIONS(1247),
    [sym_true] = ACTIONS(1247),
    [sym_false] = ACTIONS(1247),
    [sym_none] = ACTIONS(1247),
    [sym_string] = ACTIONS(1247),
    [sym_verbatim] = ACTIONS(1247),
    [sym_number] = ACTIONS(1247),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1245),
    [sym_fstring_start] = ACTIONS(1245),
  },
  [STATE(478)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1258),
    [sym_identifier] = ACTIONS(1260),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_return] = ACTIONS(1260),
    [anon_sym_async] = ACTIONS(1260),
    [anon_sym_def] = ACTIONS(1260),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LPAREN] = ACTIONS(1260),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [anon_sym_del] = ACTIONS(1260),
    [anon_sym_raise] = ACTIONS(1260),
    [sym_pass_statement] = ACTIONS(1260),
    [sym_break_statement] = ACTIONS(1260),
    [sym_continue_statement] = ACTIONS(1260),
    [sym_debugger_statement] = ACTIONS(1260),
    [anon_sym_assert] = ACTIONS(1260),
    [anon_sym_global] = ACTIONS(1260),
    [anon_sym_nonlocal] = ACTIONS(1260),
    [anon_sym_yield] = ACTIONS(1260),
    [anon_sym_from] = ACTIONS(1260),
    [anon_sym_import] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1260),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1260),
    [anon_sym_do] = ACTIONS(1260),
    [anon_sym_try] = ACTIONS(1260),
    [anon_sym_with] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1260),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_typeof] = ACTIONS(1260),
    [anon_sym_void] = ACTIONS(1260),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1260),
    [anon_sym_QMARK] = ACTIONS(1260),
    [anon_sym_await] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [sym_this] = ACTIONS(1260),
    [sym_true] = ACTIONS(1260),
    [sym_false] = ACTIONS(1260),
    [sym_none] = ACTIONS(1260),
    [sym_string] = ACTIONS(1260),
    [sym_verbatim] = ACTIONS(1260),
    [sym_number] = ACTIONS(1260),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1258),
    [sym_fstring_start] = ACTIONS(1258),
  },
  [STATE(479)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1264),
    [sym_identifier] = ACTIONS(1266),
    [anon_sym_SEMI] = ACTIONS(1266),
    [anon_sym_return] = ACTIONS(1266),
    [anon_sym_async] = ACTIONS(1266),
    [anon_sym_def] = ACTIONS(1266),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LPAREN] = ACTIONS(1266),
    [anon_sym_LBRACK] = ACTIONS(1266),
    [anon_sym_del] = ACTIONS(1266),
    [anon_sym_raise] = ACTIONS(1266),
    [sym_pass_statement] = ACTIONS(1266),
    [sym_break_statement] = ACTIONS(1266),
    [sym_continue_statement] = ACTIONS(1266),
    [sym_debugger_statement] = ACTIONS(1266),
    [anon_sym_assert] = ACTIONS(1266),
    [anon_sym_global] = ACTIONS(1266),
    [anon_sym_nonlocal] = ACTIONS(1266),
    [anon_sym_yield] = ACTIONS(1266),
    [anon_sym_from] = ACTIONS(1266),
    [anon_sym_import] = ACTIONS(1266),
    [anon_sym_DOT] = ACTIONS(1266),
    [anon_sym_if] = ACTIONS(1266),
    [anon_sym_for] = ACTIONS(1266),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1266),
    [anon_sym_do] = ACTIONS(1266),
    [anon_sym_try] = ACTIONS(1266),
    [anon_sym_with] = ACTIONS(1266),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1266),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1266),
    [anon_sym_typeof] = ACTIONS(1266),
    [anon_sym_void] = ACTIONS(1266),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(1266),
    [anon_sym_await] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1266),
    [sym_this] = ACTIONS(1266),
    [sym_true] = ACTIONS(1266),
    [sym_false] = ACTIONS(1266),
    [sym_none] = ACTIONS(1266),
    [sym_string] = ACTIONS(1266),
    [sym_verbatim] = ACTIONS(1266),
    [sym_number] = ACTIONS(1266),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1264),
    [sym_fstring_start] = ACTIONS(1264),
  },
  [STATE(480)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_async] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1270),
    [anon_sym_del] = ACTIONS(1270),
    [anon_sym_raise] = ACTIONS(1270),
    [sym_pass_statement] = ACTIONS(1270),
    [sym_break_statement] = ACTIONS(1270),
    [sym_continue_statement] = ACTIONS(1270),
    [sym_debugger_statement] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_global] = ACTIONS(1270),
    [anon_sym_nonlocal] = ACTIONS(1270),
    [anon_sym_yield] = ACTIONS(1270),
    [anon_sym_from] = ACTIONS(1270),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_DOT] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_with] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1270),
    [anon_sym_or] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1270),
    [anon_sym_typeof] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1270),
    [anon_sym_await] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1270),
    [sym_this] = ACTIONS(1270),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [sym_none] = ACTIONS(1270),
    [sym_string] = ACTIONS(1270),
    [sym_verbatim] = ACTIONS(1270),
    [sym_number] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1268),
    [sym_fstring_start] = ACTIONS(1268),
  },
  [STATE(481)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1272),
    [sym_identifier] = ACTIONS(1274),
    [anon_sym_SEMI] = ACTIONS(1274),
    [anon_sym_return] = ACTIONS(1274),
    [anon_sym_async] = ACTIONS(1274),
    [anon_sym_def] = ACTIONS(1274),
    [anon_sym_COMMA] = ACTIONS(1274),
    [anon_sym_LPAREN] = ACTIONS(1274),
    [anon_sym_LBRACK] = ACTIONS(1274),
    [anon_sym_del] = ACTIONS(1274),
    [anon_sym_raise] = ACTIONS(1274),
    [sym_pass_statement] = ACTIONS(1274),
    [sym_break_statement] = ACTIONS(1274),
    [sym_continue_statement] = ACTIONS(1274),
    [sym_debugger_statement] = ACTIONS(1274),
    [anon_sym_assert] = ACTIONS(1274),
    [anon_sym_global] = ACTIONS(1274),
    [anon_sym_nonlocal] = ACTIONS(1274),
    [anon_sym_yield] = ACTIONS(1274),
    [anon_sym_from] = ACTIONS(1274),
    [anon_sym_import] = ACTIONS(1274),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1274),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1274),
    [anon_sym_do] = ACTIONS(1274),
    [anon_sym_try] = ACTIONS(1274),
    [anon_sym_with] = ACTIONS(1274),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1274),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1274),
    [anon_sym_typeof] = ACTIONS(1274),
    [anon_sym_void] = ACTIONS(1274),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1274),
    [anon_sym_QMARK] = ACTIONS(1274),
    [anon_sym_await] = ACTIONS(1274),
    [anon_sym_LBRACE] = ACTIONS(1274),
    [sym_this] = ACTIONS(1274),
    [sym_true] = ACTIONS(1274),
    [sym_false] = ACTIONS(1274),
    [sym_none] = ACTIONS(1274),
    [sym_string] = ACTIONS(1274),
    [sym_verbatim] = ACTIONS(1274),
    [sym_number] = ACTIONS(1274),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1272),
    [sym_fstring_start] = ACTIONS(1272),
  },
  [STATE(482)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym_return] = ACTIONS(1224),
    [anon_sym_async] = ACTIONS(1224),
    [anon_sym_def] = ACTIONS(1224),
    [anon_sym_COMMA] = ACTIONS(1224),
    [anon_sym_LPAREN] = ACTIONS(1224),
    [anon_sym_LBRACK] = ACTIONS(1224),
    [anon_sym_del] = ACTIONS(1224),
    [anon_sym_raise] = ACTIONS(1224),
    [sym_pass_statement] = ACTIONS(1224),
    [sym_break_statement] = ACTIONS(1224),
    [sym_continue_statement] = ACTIONS(1224),
    [sym_debugger_statement] = ACTIONS(1224),
    [anon_sym_assert] = ACTIONS(1224),
    [anon_sym_global] = ACTIONS(1224),
    [anon_sym_nonlocal] = ACTIONS(1224),
    [anon_sym_yield] = ACTIONS(1224),
    [anon_sym_from] = ACTIONS(1224),
    [anon_sym_import] = ACTIONS(1224),
    [anon_sym_DOT] = ACTIONS(1224),
    [anon_sym_if] = ACTIONS(1224),
    [anon_sym_for] = ACTIONS(1224),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1224),
    [anon_sym_do] = ACTIONS(1224),
    [anon_sym_try] = ACTIONS(1224),
    [anon_sym_with] = ACTIONS(1224),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1224),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_typeof] = ACTIONS(1224),
    [anon_sym_void] = ACTIONS(1224),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1224),
    [anon_sym_QMARK] = ACTIONS(1224),
    [anon_sym_await] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [sym_this] = ACTIONS(1224),
    [sym_true] = ACTIONS(1224),
    [sym_false] = ACTIONS(1224),
    [sym_none] = ACTIONS(1224),
    [sym_string] = ACTIONS(1224),
    [sym_verbatim] = ACTIONS(1224),
    [sym_number] = ACTIONS(1224),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1222),
    [sym_regex] = ACTIONS(1222),
    [sym_fstring_start] = ACTIONS(1222),
  },
  [STATE(483)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1228),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1228),
    [anon_sym_SLASH_SLASH] = ACTIONS(1228),
    [anon_sym_PERCENT] = ACTIONS(1228),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(484)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1228),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1228),
    [anon_sym_SLASH_SLASH] = ACTIONS(1228),
    [anon_sym_PERCENT] = ACTIONS(1228),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(485)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1243),
    [anon_sym_SEMI] = ACTIONS(1243),
    [anon_sym_return] = ACTIONS(1243),
    [anon_sym_async] = ACTIONS(1243),
    [anon_sym_def] = ACTIONS(1243),
    [anon_sym_COMMA] = ACTIONS(1243),
    [anon_sym_LPAREN] = ACTIONS(1243),
    [anon_sym_LBRACK] = ACTIONS(1243),
    [anon_sym_del] = ACTIONS(1243),
    [anon_sym_raise] = ACTIONS(1243),
    [sym_pass_statement] = ACTIONS(1243),
    [sym_break_statement] = ACTIONS(1243),
    [sym_continue_statement] = ACTIONS(1243),
    [sym_debugger_statement] = ACTIONS(1243),
    [anon_sym_assert] = ACTIONS(1243),
    [anon_sym_global] = ACTIONS(1243),
    [anon_sym_nonlocal] = ACTIONS(1243),
    [anon_sym_yield] = ACTIONS(1243),
    [anon_sym_from] = ACTIONS(1243),
    [anon_sym_import] = ACTIONS(1243),
    [anon_sym_DOT] = ACTIONS(1243),
    [anon_sym_if] = ACTIONS(1243),
    [anon_sym_for] = ACTIONS(1243),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1243),
    [anon_sym_do] = ACTIONS(1243),
    [anon_sym_try] = ACTIONS(1243),
    [anon_sym_with] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1243),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1243),
    [anon_sym_or] = ACTIONS(1243),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1243),
    [anon_sym_typeof] = ACTIONS(1243),
    [anon_sym_void] = ACTIONS(1243),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1243),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_await] = ACTIONS(1243),
    [anon_sym_LBRACE] = ACTIONS(1243),
    [sym_this] = ACTIONS(1243),
    [sym_true] = ACTIONS(1243),
    [sym_false] = ACTIONS(1243),
    [sym_none] = ACTIONS(1243),
    [sym_string] = ACTIONS(1243),
    [sym_verbatim] = ACTIONS(1243),
    [sym_number] = ACTIONS(1243),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1241),
    [sym_regex] = ACTIONS(1241),
    [sym_fstring_start] = ACTIONS(1241),
  },
  [STATE(486)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1243),
    [anon_sym_SEMI] = ACTIONS(1243),
    [anon_sym_return] = ACTIONS(1243),
    [anon_sym_async] = ACTIONS(1243),
    [anon_sym_def] = ACTIONS(1243),
    [anon_sym_COMMA] = ACTIONS(1243),
    [anon_sym_LPAREN] = ACTIONS(1243),
    [anon_sym_LBRACK] = ACTIONS(1243),
    [anon_sym_del] = ACTIONS(1243),
    [anon_sym_raise] = ACTIONS(1243),
    [sym_pass_statement] = ACTIONS(1243),
    [sym_break_statement] = ACTIONS(1243),
    [sym_continue_statement] = ACTIONS(1243),
    [sym_debugger_statement] = ACTIONS(1243),
    [anon_sym_assert] = ACTIONS(1243),
    [anon_sym_global] = ACTIONS(1243),
    [anon_sym_nonlocal] = ACTIONS(1243),
    [anon_sym_yield] = ACTIONS(1243),
    [anon_sym_from] = ACTIONS(1243),
    [anon_sym_import] = ACTIONS(1243),
    [anon_sym_DOT] = ACTIONS(1243),
    [anon_sym_if] = ACTIONS(1243),
    [anon_sym_for] = ACTIONS(1243),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1243),
    [anon_sym_do] = ACTIONS(1243),
    [anon_sym_try] = ACTIONS(1243),
    [anon_sym_with] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1243),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1243),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1243),
    [anon_sym_typeof] = ACTIONS(1243),
    [anon_sym_void] = ACTIONS(1243),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1243),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_await] = ACTIONS(1243),
    [anon_sym_LBRACE] = ACTIONS(1243),
    [sym_this] = ACTIONS(1243),
    [sym_true] = ACTIONS(1243),
    [sym_false] = ACTIONS(1243),
    [sym_none] = ACTIONS(1243),
    [sym_string] = ACTIONS(1243),
    [sym_verbatim] = ACTIONS(1243),
    [sym_number] = ACTIONS(1243),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1241),
    [sym_regex] = ACTIONS(1241),
    [sym_fstring_start] = ACTIONS(1241),
  },
  [STATE(487)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(488)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(489)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(490)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(491)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(492)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(492),
    [sym_identifier] = ACTIONS(1247),
    [anon_sym_SEMI] = ACTIONS(1247),
    [anon_sym_return] = ACTIONS(1247),
    [anon_sym_async] = ACTIONS(1247),
    [anon_sym_def] = ACTIONS(1247),
    [anon_sym_COMMA] = ACTIONS(1247),
    [anon_sym_LPAREN] = ACTIONS(1247),
    [anon_sym_LBRACK] = ACTIONS(1247),
    [anon_sym_del] = ACTIONS(1247),
    [anon_sym_raise] = ACTIONS(1247),
    [sym_pass_statement] = ACTIONS(1247),
    [sym_break_statement] = ACTIONS(1247),
    [sym_continue_statement] = ACTIONS(1247),
    [sym_debugger_statement] = ACTIONS(1247),
    [anon_sym_assert] = ACTIONS(1247),
    [anon_sym_global] = ACTIONS(1247),
    [anon_sym_nonlocal] = ACTIONS(1247),
    [anon_sym_yield] = ACTIONS(1247),
    [anon_sym_from] = ACTIONS(1247),
    [anon_sym_import] = ACTIONS(1247),
    [anon_sym_DOT] = ACTIONS(1247),
    [anon_sym_if] = ACTIONS(1247),
    [anon_sym_for] = ACTIONS(1247),
    [anon_sym_in] = ACTIONS(1276),
    [anon_sym_while] = ACTIONS(1247),
    [anon_sym_do] = ACTIONS(1247),
    [anon_sym_try] = ACTIONS(1247),
    [anon_sym_with] = ACTIONS(1247),
    [anon_sym_STAR] = ACTIONS(1247),
    [anon_sym_STAR_STAR] = ACTIONS(1247),
    [anon_sym_class] = ACTIONS(1247),
    [anon_sym_AT] = ACTIONS(1247),
    [anon_sym_not] = ACTIONS(1279),
    [anon_sym_and] = ACTIONS(1247),
    [anon_sym_or] = ACTIONS(1247),
    [anon_sym_PLUS] = ACTIONS(1247),
    [anon_sym_DASH] = ACTIONS(1247),
    [anon_sym_SLASH] = ACTIONS(1247),
    [anon_sym_SLASH_SLASH] = ACTIONS(1247),
    [anon_sym_PERCENT] = ACTIONS(1247),
    [anon_sym_PIPE] = ACTIONS(1247),
    [anon_sym_CARET] = ACTIONS(1247),
    [anon_sym_AMP] = ACTIONS(1247),
    [anon_sym_LT_LT] = ACTIONS(1247),
    [anon_sym_GT_GT] = ACTIONS(1247),
    [anon_sym_GT_GT_GT] = ACTIONS(1247),
    [anon_sym_TILDE] = ACTIONS(1247),
    [anon_sym_typeof] = ACTIONS(1247),
    [anon_sym_void] = ACTIONS(1247),
    [anon_sym_LT] = ACTIONS(1276),
    [anon_sym_LT_EQ] = ACTIONS(1276),
    [anon_sym_EQ_EQ] = ACTIONS(1276),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1276),
    [anon_sym_BANG_EQ] = ACTIONS(1276),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1276),
    [anon_sym_GT_EQ] = ACTIONS(1276),
    [anon_sym_GT] = ACTIONS(1276),
    [anon_sym_is] = ACTIONS(1282),
    [anon_sym_instanceof] = ACTIONS(1276),
    [anon_sym_new] = ACTIONS(1247),
    [anon_sym_QMARK] = ACTIONS(1247),
    [anon_sym_await] = ACTIONS(1247),
    [anon_sym_LBRACE] = ACTIONS(1247),
    [sym_this] = ACTIONS(1247),
    [sym_true] = ACTIONS(1247),
    [sym_false] = ACTIONS(1247),
    [sym_none] = ACTIONS(1247),
    [sym_string] = ACTIONS(1247),
    [sym_verbatim] = ACTIONS(1247),
    [sym_number] = ACTIONS(1247),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1245),
    [sym_regex] = ACTIONS(1245),
    [sym_fstring_start] = ACTIONS(1245),
  },
  [STATE(493)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1260),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_return] = ACTIONS(1260),
    [anon_sym_async] = ACTIONS(1260),
    [anon_sym_def] = ACTIONS(1260),
    [anon_sym_COMMA] = ACTIONS(1260),
    [anon_sym_LPAREN] = ACTIONS(1260),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [anon_sym_del] = ACTIONS(1260),
    [anon_sym_raise] = ACTIONS(1260),
    [sym_pass_statement] = ACTIONS(1260),
    [sym_break_statement] = ACTIONS(1260),
    [sym_continue_statement] = ACTIONS(1260),
    [sym_debugger_statement] = ACTIONS(1260),
    [anon_sym_assert] = ACTIONS(1260),
    [anon_sym_global] = ACTIONS(1260),
    [anon_sym_nonlocal] = ACTIONS(1260),
    [anon_sym_yield] = ACTIONS(1260),
    [anon_sym_from] = ACTIONS(1260),
    [anon_sym_import] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1260),
    [anon_sym_if] = ACTIONS(1285),
    [anon_sym_for] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1260),
    [anon_sym_do] = ACTIONS(1260),
    [anon_sym_try] = ACTIONS(1260),
    [anon_sym_with] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1260),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_typeof] = ACTIONS(1260),
    [anon_sym_void] = ACTIONS(1260),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1260),
    [anon_sym_QMARK] = ACTIONS(1260),
    [anon_sym_await] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [sym_this] = ACTIONS(1260),
    [sym_true] = ACTIONS(1260),
    [sym_false] = ACTIONS(1260),
    [sym_none] = ACTIONS(1260),
    [sym_string] = ACTIONS(1260),
    [sym_verbatim] = ACTIONS(1260),
    [sym_number] = ACTIONS(1260),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1258),
    [sym_regex] = ACTIONS(1258),
    [sym_fstring_start] = ACTIONS(1258),
  },
  [STATE(494)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1266),
    [anon_sym_SEMI] = ACTIONS(1266),
    [anon_sym_return] = ACTIONS(1266),
    [anon_sym_async] = ACTIONS(1266),
    [anon_sym_def] = ACTIONS(1266),
    [anon_sym_COMMA] = ACTIONS(1266),
    [anon_sym_LPAREN] = ACTIONS(1266),
    [anon_sym_LBRACK] = ACTIONS(1266),
    [anon_sym_del] = ACTIONS(1266),
    [anon_sym_raise] = ACTIONS(1266),
    [sym_pass_statement] = ACTIONS(1266),
    [sym_break_statement] = ACTIONS(1266),
    [sym_continue_statement] = ACTIONS(1266),
    [sym_debugger_statement] = ACTIONS(1266),
    [anon_sym_assert] = ACTIONS(1266),
    [anon_sym_global] = ACTIONS(1266),
    [anon_sym_nonlocal] = ACTIONS(1266),
    [anon_sym_yield] = ACTIONS(1266),
    [anon_sym_from] = ACTIONS(1266),
    [anon_sym_import] = ACTIONS(1266),
    [anon_sym_DOT] = ACTIONS(1266),
    [anon_sym_if] = ACTIONS(1266),
    [anon_sym_for] = ACTIONS(1266),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1266),
    [anon_sym_do] = ACTIONS(1266),
    [anon_sym_try] = ACTIONS(1266),
    [anon_sym_with] = ACTIONS(1266),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1266),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1266),
    [anon_sym_typeof] = ACTIONS(1266),
    [anon_sym_void] = ACTIONS(1266),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(1266),
    [anon_sym_await] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1266),
    [sym_this] = ACTIONS(1266),
    [sym_true] = ACTIONS(1266),
    [sym_false] = ACTIONS(1266),
    [sym_none] = ACTIONS(1266),
    [sym_string] = ACTIONS(1266),
    [sym_verbatim] = ACTIONS(1266),
    [sym_number] = ACTIONS(1266),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1264),
    [sym_regex] = ACTIONS(1264),
    [sym_fstring_start] = ACTIONS(1264),
  },
  [STATE(495)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_async] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1270),
    [anon_sym_del] = ACTIONS(1270),
    [anon_sym_raise] = ACTIONS(1270),
    [sym_pass_statement] = ACTIONS(1270),
    [sym_break_statement] = ACTIONS(1270),
    [sym_continue_statement] = ACTIONS(1270),
    [sym_debugger_statement] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_global] = ACTIONS(1270),
    [anon_sym_nonlocal] = ACTIONS(1270),
    [anon_sym_yield] = ACTIONS(1270),
    [anon_sym_from] = ACTIONS(1270),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_DOT] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_with] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1270),
    [anon_sym_or] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1270),
    [anon_sym_typeof] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1270),
    [anon_sym_await] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1270),
    [sym_this] = ACTIONS(1270),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [sym_none] = ACTIONS(1270),
    [sym_string] = ACTIONS(1270),
    [sym_verbatim] = ACTIONS(1270),
    [sym_number] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1268),
    [sym_regex] = ACTIONS(1268),
    [sym_fstring_start] = ACTIONS(1268),
  },
  [STATE(496)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1274),
    [anon_sym_SEMI] = ACTIONS(1274),
    [anon_sym_return] = ACTIONS(1274),
    [anon_sym_async] = ACTIONS(1274),
    [anon_sym_def] = ACTIONS(1274),
    [anon_sym_COMMA] = ACTIONS(1274),
    [anon_sym_LPAREN] = ACTIONS(1274),
    [anon_sym_LBRACK] = ACTIONS(1274),
    [anon_sym_del] = ACTIONS(1274),
    [anon_sym_raise] = ACTIONS(1274),
    [sym_pass_statement] = ACTIONS(1274),
    [sym_break_statement] = ACTIONS(1274),
    [sym_continue_statement] = ACTIONS(1274),
    [sym_debugger_statement] = ACTIONS(1274),
    [anon_sym_assert] = ACTIONS(1274),
    [anon_sym_global] = ACTIONS(1274),
    [anon_sym_nonlocal] = ACTIONS(1274),
    [anon_sym_yield] = ACTIONS(1274),
    [anon_sym_from] = ACTIONS(1274),
    [anon_sym_import] = ACTIONS(1274),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(1285),
    [anon_sym_for] = ACTIONS(1274),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1274),
    [anon_sym_do] = ACTIONS(1274),
    [anon_sym_try] = ACTIONS(1274),
    [anon_sym_with] = ACTIONS(1274),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1274),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1274),
    [anon_sym_typeof] = ACTIONS(1274),
    [anon_sym_void] = ACTIONS(1274),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1274),
    [anon_sym_QMARK] = ACTIONS(1274),
    [anon_sym_await] = ACTIONS(1274),
    [anon_sym_LBRACE] = ACTIONS(1274),
    [sym_this] = ACTIONS(1274),
    [sym_true] = ACTIONS(1274),
    [sym_false] = ACTIONS(1274),
    [sym_none] = ACTIONS(1274),
    [sym_string] = ACTIONS(1274),
    [sym_verbatim] = ACTIONS(1274),
    [sym_number] = ACTIONS(1274),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1272),
    [sym_regex] = ACTIONS(1272),
    [sym_fstring_start] = ACTIONS(1272),
  },
  [STATE(497)] = {
    [ts_builtin_sym_end] = ACTIONS(1239),
    [sym_identifier] = ACTIONS(1230),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1230),
    [anon_sym_async] = ACTIONS(1230),
    [anon_sym_def] = ACTIONS(1230),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1234),
    [anon_sym_LBRACK] = ACTIONS(1234),
    [anon_sym_del] = ACTIONS(1230),
    [anon_sym_raise] = ACTIONS(1230),
    [sym_pass_statement] = ACTIONS(1230),
    [sym_break_statement] = ACTIONS(1230),
    [sym_continue_statement] = ACTIONS(1230),
    [sym_debugger_statement] = ACTIONS(1230),
    [anon_sym_assert] = ACTIONS(1230),
    [anon_sym_global] = ACTIONS(1230),
    [anon_sym_nonlocal] = ACTIONS(1230),
    [anon_sym_yield] = ACTIONS(1230),
    [anon_sym_from] = ACTIONS(1230),
    [anon_sym_import] = ACTIONS(1230),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1230),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1230),
    [anon_sym_do] = ACTIONS(1230),
    [anon_sym_try] = ACTIONS(1230),
    [anon_sym_with] = ACTIONS(1230),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1230),
    [anon_sym_AT] = ACTIONS(1234),
    [anon_sym_not] = ACTIONS(1234),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1234),
    [anon_sym_DASH] = ACTIONS(1234),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1230),
    [anon_sym_typeof] = ACTIONS(1230),
    [anon_sym_void] = ACTIONS(1230),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1230),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1230),
    [anon_sym_LBRACE] = ACTIONS(1230),
    [sym_this] = ACTIONS(1230),
    [sym_true] = ACTIONS(1230),
    [sym_false] = ACTIONS(1230),
    [sym_none] = ACTIONS(1230),
    [sym_string] = ACTIONS(1230),
    [sym_verbatim] = ACTIONS(1230),
    [sym_number] = ACTIONS(1230),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym_regex] = ACTIONS(1239),
    [sym_fstring_start] = ACTIONS(1239),
  },
  [STATE(498)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1287),
    [anon_sym_SEMI] = ACTIONS(1287),
    [anon_sym_return] = ACTIONS(1287),
    [anon_sym_async] = ACTIONS(1287),
    [anon_sym_def] = ACTIONS(1287),
    [anon_sym_COMMA] = ACTIONS(1289),
    [anon_sym_LPAREN] = ACTIONS(1287),
    [anon_sym_LBRACK] = ACTIONS(1287),
    [anon_sym_del] = ACTIONS(1287),
    [anon_sym_raise] = ACTIONS(1287),
    [sym_pass_statement] = ACTIONS(1287),
    [sym_break_statement] = ACTIONS(1287),
    [sym_continue_statement] = ACTIONS(1287),
    [sym_debugger_statement] = ACTIONS(1287),
    [anon_sym_assert] = ACTIONS(1287),
    [anon_sym_global] = ACTIONS(1287),
    [anon_sym_nonlocal] = ACTIONS(1287),
    [anon_sym_yield] = ACTIONS(1287),
    [anon_sym_from] = ACTIONS(1287),
    [anon_sym_import] = ACTIONS(1287),
    [anon_sym_DOT] = ACTIONS(1287),
    [anon_sym_if] = ACTIONS(1287),
    [anon_sym_for] = ACTIONS(1287),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1287),
    [anon_sym_do] = ACTIONS(1287),
    [anon_sym_try] = ACTIONS(1287),
    [anon_sym_with] = ACTIONS(1287),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1287),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1287),
    [anon_sym_typeof] = ACTIONS(1287),
    [anon_sym_void] = ACTIONS(1287),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1287),
    [anon_sym_QMARK] = ACTIONS(1287),
    [anon_sym_await] = ACTIONS(1287),
    [anon_sym_LBRACE] = ACTIONS(1287),
    [sym_this] = ACTIONS(1287),
    [sym_true] = ACTIONS(1287),
    [sym_false] = ACTIONS(1287),
    [sym_none] = ACTIONS(1287),
    [sym_string] = ACTIONS(1287),
    [sym_verbatim] = ACTIONS(1287),
    [sym_number] = ACTIONS(1287),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1291),
    [sym_regex] = ACTIONS(1291),
    [sym_fstring_start] = ACTIONS(1291),
  },
  [STATE(499)] = {
    [sym__expression_statement] = STATE(618),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(618),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(618),
    [sym_delete_statement] = STATE(618),
    [sym_raise_statement] = STATE(618),
    [sym_assert_statement] = STATE(618),
    [sym_yield_statement] = STATE(618),
    [sym_yield] = STATE(668),
    [sym__expression] = STATE(452),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__inline_simple_statement] = STATE(618),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(131),
    [anon_sym_raise] = ACTIONS(133),
    [sym_pass_statement] = ACTIONS(135),
    [sym_break_statement] = ACTIONS(135),
    [sym_continue_statement] = ACTIONS(135),
    [anon_sym_assert] = ACTIONS(137),
    [anon_sym_yield] = ACTIONS(139),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(500)] = {
    [aux_sym_expression_list_repeat1] = STATE(517),
    [ts_builtin_sym_end] = ACTIONS(503),
    [sym_identifier] = ACTIONS(505),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_return] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(505),
    [anon_sym_def] = ACTIONS(505),
    [anon_sym_COMMA] = ACTIONS(1293),
    [anon_sym_LPAREN] = ACTIONS(505),
    [anon_sym_LBRACK] = ACTIONS(505),
    [anon_sym_del] = ACTIONS(505),
    [anon_sym_raise] = ACTIONS(505),
    [sym_pass_statement] = ACTIONS(505),
    [sym_break_statement] = ACTIONS(505),
    [sym_continue_statement] = ACTIONS(505),
    [sym_debugger_statement] = ACTIONS(505),
    [anon_sym_assert] = ACTIONS(505),
    [anon_sym_global] = ACTIONS(505),
    [anon_sym_nonlocal] = ACTIONS(505),
    [anon_sym_yield] = ACTIONS(505),
    [anon_sym_from] = ACTIONS(505),
    [anon_sym_import] = ACTIONS(505),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_do] = ACTIONS(505),
    [anon_sym_try] = ACTIONS(505),
    [anon_sym_with] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_class] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(505),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(505),
    [anon_sym_DASH] = ACTIONS(505),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(505),
    [anon_sym_typeof] = ACTIONS(505),
    [anon_sym_void] = ACTIONS(505),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(505),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(505),
    [anon_sym_LBRACE] = ACTIONS(505),
    [sym_this] = ACTIONS(505),
    [sym_true] = ACTIONS(505),
    [sym_false] = ACTIONS(505),
    [sym_none] = ACTIONS(505),
    [sym_string] = ACTIONS(505),
    [sym_verbatim] = ACTIONS(505),
    [sym_number] = ACTIONS(505),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(503),
    [sym_fstring_start] = ACTIONS(503),
  },
  [STATE(501)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1295),
    [sym_identifier] = ACTIONS(1297),
    [anon_sym_SEMI] = ACTIONS(1297),
    [anon_sym_return] = ACTIONS(1297),
    [anon_sym_async] = ACTIONS(1297),
    [anon_sym_def] = ACTIONS(1297),
    [anon_sym_COMMA] = ACTIONS(1297),
    [anon_sym_LPAREN] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1297),
    [anon_sym_del] = ACTIONS(1297),
    [anon_sym_raise] = ACTIONS(1297),
    [sym_pass_statement] = ACTIONS(1297),
    [sym_break_statement] = ACTIONS(1297),
    [sym_continue_statement] = ACTIONS(1297),
    [sym_debugger_statement] = ACTIONS(1297),
    [anon_sym_assert] = ACTIONS(1297),
    [anon_sym_global] = ACTIONS(1297),
    [anon_sym_nonlocal] = ACTIONS(1297),
    [anon_sym_yield] = ACTIONS(1297),
    [anon_sym_from] = ACTIONS(1297),
    [anon_sym_import] = ACTIONS(1297),
    [anon_sym_DOT] = ACTIONS(1297),
    [anon_sym_if] = ACTIONS(1297),
    [anon_sym_for] = ACTIONS(1297),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1297),
    [anon_sym_do] = ACTIONS(1297),
    [anon_sym_try] = ACTIONS(1297),
    [anon_sym_with] = ACTIONS(1297),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1297),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1297),
    [anon_sym_typeof] = ACTIONS(1297),
    [anon_sym_void] = ACTIONS(1297),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1297),
    [anon_sym_QMARK] = ACTIONS(1297),
    [anon_sym_await] = ACTIONS(1297),
    [anon_sym_LBRACE] = ACTIONS(1297),
    [sym_this] = ACTIONS(1297),
    [sym_true] = ACTIONS(1297),
    [sym_false] = ACTIONS(1297),
    [sym_none] = ACTIONS(1297),
    [sym_string] = ACTIONS(1297),
    [sym_verbatim] = ACTIONS(1297),
    [sym_number] = ACTIONS(1297),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1295),
    [sym_fstring_start] = ACTIONS(1295),
  },
  [STATE(502)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1299),
    [sym_identifier] = ACTIONS(1301),
    [anon_sym_SEMI] = ACTIONS(1301),
    [anon_sym_return] = ACTIONS(1301),
    [anon_sym_async] = ACTIONS(1301),
    [anon_sym_def] = ACTIONS(1301),
    [anon_sym_COMMA] = ACTIONS(1301),
    [anon_sym_LPAREN] = ACTIONS(1301),
    [anon_sym_LBRACK] = ACTIONS(1301),
    [anon_sym_del] = ACTIONS(1301),
    [anon_sym_raise] = ACTIONS(1301),
    [sym_pass_statement] = ACTIONS(1301),
    [sym_break_statement] = ACTIONS(1301),
    [sym_continue_statement] = ACTIONS(1301),
    [sym_debugger_statement] = ACTIONS(1301),
    [anon_sym_assert] = ACTIONS(1301),
    [anon_sym_global] = ACTIONS(1301),
    [anon_sym_nonlocal] = ACTIONS(1301),
    [anon_sym_yield] = ACTIONS(1301),
    [anon_sym_from] = ACTIONS(1301),
    [anon_sym_import] = ACTIONS(1301),
    [anon_sym_DOT] = ACTIONS(1301),
    [anon_sym_if] = ACTIONS(1301),
    [anon_sym_for] = ACTIONS(1301),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1301),
    [anon_sym_do] = ACTIONS(1301),
    [anon_sym_try] = ACTIONS(1301),
    [anon_sym_with] = ACTIONS(1301),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1301),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1301),
    [anon_sym_typeof] = ACTIONS(1301),
    [anon_sym_void] = ACTIONS(1301),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1301),
    [anon_sym_QMARK] = ACTIONS(1301),
    [anon_sym_await] = ACTIONS(1301),
    [anon_sym_LBRACE] = ACTIONS(1301),
    [sym_this] = ACTIONS(1301),
    [sym_true] = ACTIONS(1301),
    [sym_false] = ACTIONS(1301),
    [sym_none] = ACTIONS(1301),
    [sym_string] = ACTIONS(1301),
    [sym_verbatim] = ACTIONS(1301),
    [sym_number] = ACTIONS(1301),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1299),
    [sym_fstring_start] = ACTIONS(1299),
  },
  [STATE(503)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1303),
    [sym_identifier] = ACTIONS(1305),
    [anon_sym_SEMI] = ACTIONS(1305),
    [anon_sym_return] = ACTIONS(1305),
    [anon_sym_async] = ACTIONS(1305),
    [anon_sym_def] = ACTIONS(1305),
    [anon_sym_COMMA] = ACTIONS(1305),
    [anon_sym_LPAREN] = ACTIONS(1305),
    [anon_sym_LBRACK] = ACTIONS(1305),
    [anon_sym_del] = ACTIONS(1305),
    [anon_sym_raise] = ACTIONS(1305),
    [sym_pass_statement] = ACTIONS(1305),
    [sym_break_statement] = ACTIONS(1305),
    [sym_continue_statement] = ACTIONS(1305),
    [sym_debugger_statement] = ACTIONS(1305),
    [anon_sym_assert] = ACTIONS(1305),
    [anon_sym_global] = ACTIONS(1305),
    [anon_sym_nonlocal] = ACTIONS(1305),
    [anon_sym_yield] = ACTIONS(1305),
    [anon_sym_from] = ACTIONS(1305),
    [anon_sym_import] = ACTIONS(1305),
    [anon_sym_DOT] = ACTIONS(1305),
    [anon_sym_if] = ACTIONS(1305),
    [anon_sym_for] = ACTIONS(1305),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1305),
    [anon_sym_do] = ACTIONS(1305),
    [anon_sym_try] = ACTIONS(1305),
    [anon_sym_with] = ACTIONS(1305),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1305),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1305),
    [anon_sym_typeof] = ACTIONS(1305),
    [anon_sym_void] = ACTIONS(1305),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1305),
    [anon_sym_QMARK] = ACTIONS(1305),
    [anon_sym_await] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1305),
    [sym_this] = ACTIONS(1305),
    [sym_true] = ACTIONS(1305),
    [sym_false] = ACTIONS(1305),
    [sym_none] = ACTIONS(1305),
    [sym_string] = ACTIONS(1305),
    [sym_verbatim] = ACTIONS(1305),
    [sym_number] = ACTIONS(1305),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1303),
    [sym_fstring_start] = ACTIONS(1303),
  },
  [STATE(504)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1782),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(639),
    [anon_sym_raise] = ACTIONS(641),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(643),
    [anon_sym_yield] = ACTIONS(645),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(505)] = {
    [sym__expression_statement] = STATE(656),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(656),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(656),
    [sym_delete_statement] = STATE(656),
    [sym_raise_statement] = STATE(656),
    [sym_assert_statement] = STATE(656),
    [sym_yield_statement] = STATE(656),
    [sym_yield] = STATE(651),
    [sym__expression] = STATE(459),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__inline_simple_statement] = STATE(656),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(85),
    [anon_sym_raise] = ACTIONS(87),
    [sym_pass_statement] = ACTIONS(89),
    [sym_break_statement] = ACTIONS(89),
    [sym_continue_statement] = ACTIONS(89),
    [anon_sym_assert] = ACTIONS(91),
    [anon_sym_yield] = ACTIONS(93),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(506)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1297),
    [anon_sym_SEMI] = ACTIONS(1297),
    [anon_sym_return] = ACTIONS(1297),
    [anon_sym_async] = ACTIONS(1297),
    [anon_sym_def] = ACTIONS(1297),
    [anon_sym_COMMA] = ACTIONS(1297),
    [anon_sym_LPAREN] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1297),
    [anon_sym_del] = ACTIONS(1297),
    [anon_sym_raise] = ACTIONS(1297),
    [sym_pass_statement] = ACTIONS(1297),
    [sym_break_statement] = ACTIONS(1297),
    [sym_continue_statement] = ACTIONS(1297),
    [sym_debugger_statement] = ACTIONS(1297),
    [anon_sym_assert] = ACTIONS(1297),
    [anon_sym_global] = ACTIONS(1297),
    [anon_sym_nonlocal] = ACTIONS(1297),
    [anon_sym_yield] = ACTIONS(1297),
    [anon_sym_from] = ACTIONS(1297),
    [anon_sym_import] = ACTIONS(1297),
    [anon_sym_DOT] = ACTIONS(1297),
    [anon_sym_if] = ACTIONS(1297),
    [anon_sym_for] = ACTIONS(1297),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1297),
    [anon_sym_do] = ACTIONS(1297),
    [anon_sym_try] = ACTIONS(1297),
    [anon_sym_with] = ACTIONS(1297),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1297),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1297),
    [anon_sym_typeof] = ACTIONS(1297),
    [anon_sym_void] = ACTIONS(1297),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1297),
    [anon_sym_QMARK] = ACTIONS(1297),
    [anon_sym_await] = ACTIONS(1297),
    [anon_sym_LBRACE] = ACTIONS(1297),
    [sym_this] = ACTIONS(1297),
    [sym_true] = ACTIONS(1297),
    [sym_false] = ACTIONS(1297),
    [sym_none] = ACTIONS(1297),
    [sym_string] = ACTIONS(1297),
    [sym_verbatim] = ACTIONS(1297),
    [sym_number] = ACTIONS(1297),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1295),
    [sym_regex] = ACTIONS(1295),
    [sym_fstring_start] = ACTIONS(1295),
  },
  [STATE(507)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1291),
    [sym_identifier] = ACTIONS(1287),
    [anon_sym_SEMI] = ACTIONS(1287),
    [anon_sym_return] = ACTIONS(1287),
    [anon_sym_async] = ACTIONS(1287),
    [anon_sym_def] = ACTIONS(1287),
    [anon_sym_COMMA] = ACTIONS(1307),
    [anon_sym_LPAREN] = ACTIONS(1287),
    [anon_sym_LBRACK] = ACTIONS(1287),
    [anon_sym_del] = ACTIONS(1287),
    [anon_sym_raise] = ACTIONS(1287),
    [sym_pass_statement] = ACTIONS(1287),
    [sym_break_statement] = ACTIONS(1287),
    [sym_continue_statement] = ACTIONS(1287),
    [sym_debugger_statement] = ACTIONS(1287),
    [anon_sym_assert] = ACTIONS(1287),
    [anon_sym_global] = ACTIONS(1287),
    [anon_sym_nonlocal] = ACTIONS(1287),
    [anon_sym_yield] = ACTIONS(1287),
    [anon_sym_from] = ACTIONS(1287),
    [anon_sym_import] = ACTIONS(1287),
    [anon_sym_DOT] = ACTIONS(1287),
    [anon_sym_if] = ACTIONS(1287),
    [anon_sym_for] = ACTIONS(1287),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1287),
    [anon_sym_do] = ACTIONS(1287),
    [anon_sym_try] = ACTIONS(1287),
    [anon_sym_with] = ACTIONS(1287),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1287),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1287),
    [anon_sym_typeof] = ACTIONS(1287),
    [anon_sym_void] = ACTIONS(1287),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1287),
    [anon_sym_QMARK] = ACTIONS(1287),
    [anon_sym_await] = ACTIONS(1287),
    [anon_sym_LBRACE] = ACTIONS(1287),
    [sym_this] = ACTIONS(1287),
    [sym_true] = ACTIONS(1287),
    [sym_false] = ACTIONS(1287),
    [sym_none] = ACTIONS(1287),
    [sym_string] = ACTIONS(1287),
    [sym_verbatim] = ACTIONS(1287),
    [sym_number] = ACTIONS(1287),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1291),
    [sym_fstring_start] = ACTIONS(1291),
  },
  [STATE(508)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1301),
    [anon_sym_SEMI] = ACTIONS(1301),
    [anon_sym_return] = ACTIONS(1301),
    [anon_sym_async] = ACTIONS(1301),
    [anon_sym_def] = ACTIONS(1301),
    [anon_sym_COMMA] = ACTIONS(1301),
    [anon_sym_LPAREN] = ACTIONS(1301),
    [anon_sym_LBRACK] = ACTIONS(1301),
    [anon_sym_del] = ACTIONS(1301),
    [anon_sym_raise] = ACTIONS(1301),
    [sym_pass_statement] = ACTIONS(1301),
    [sym_break_statement] = ACTIONS(1301),
    [sym_continue_statement] = ACTIONS(1301),
    [sym_debugger_statement] = ACTIONS(1301),
    [anon_sym_assert] = ACTIONS(1301),
    [anon_sym_global] = ACTIONS(1301),
    [anon_sym_nonlocal] = ACTIONS(1301),
    [anon_sym_yield] = ACTIONS(1301),
    [anon_sym_from] = ACTIONS(1301),
    [anon_sym_import] = ACTIONS(1301),
    [anon_sym_DOT] = ACTIONS(1301),
    [anon_sym_if] = ACTIONS(1301),
    [anon_sym_for] = ACTIONS(1301),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1301),
    [anon_sym_do] = ACTIONS(1301),
    [anon_sym_try] = ACTIONS(1301),
    [anon_sym_with] = ACTIONS(1301),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1301),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1301),
    [anon_sym_typeof] = ACTIONS(1301),
    [anon_sym_void] = ACTIONS(1301),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1301),
    [anon_sym_QMARK] = ACTIONS(1301),
    [anon_sym_await] = ACTIONS(1301),
    [anon_sym_LBRACE] = ACTIONS(1301),
    [sym_this] = ACTIONS(1301),
    [sym_true] = ACTIONS(1301),
    [sym_false] = ACTIONS(1301),
    [sym_none] = ACTIONS(1301),
    [sym_string] = ACTIONS(1301),
    [sym_verbatim] = ACTIONS(1301),
    [sym_number] = ACTIONS(1301),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1299),
    [sym_regex] = ACTIONS(1299),
    [sym_fstring_start] = ACTIONS(1299),
  },
  [STATE(509)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1305),
    [anon_sym_SEMI] = ACTIONS(1305),
    [anon_sym_return] = ACTIONS(1305),
    [anon_sym_async] = ACTIONS(1305),
    [anon_sym_def] = ACTIONS(1305),
    [anon_sym_COMMA] = ACTIONS(1305),
    [anon_sym_LPAREN] = ACTIONS(1305),
    [anon_sym_LBRACK] = ACTIONS(1305),
    [anon_sym_del] = ACTIONS(1305),
    [anon_sym_raise] = ACTIONS(1305),
    [sym_pass_statement] = ACTIONS(1305),
    [sym_break_statement] = ACTIONS(1305),
    [sym_continue_statement] = ACTIONS(1305),
    [sym_debugger_statement] = ACTIONS(1305),
    [anon_sym_assert] = ACTIONS(1305),
    [anon_sym_global] = ACTIONS(1305),
    [anon_sym_nonlocal] = ACTIONS(1305),
    [anon_sym_yield] = ACTIONS(1305),
    [anon_sym_from] = ACTIONS(1305),
    [anon_sym_import] = ACTIONS(1305),
    [anon_sym_DOT] = ACTIONS(1305),
    [anon_sym_if] = ACTIONS(1305),
    [anon_sym_for] = ACTIONS(1305),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1305),
    [anon_sym_do] = ACTIONS(1305),
    [anon_sym_try] = ACTIONS(1305),
    [anon_sym_with] = ACTIONS(1305),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1305),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1305),
    [anon_sym_typeof] = ACTIONS(1305),
    [anon_sym_void] = ACTIONS(1305),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1305),
    [anon_sym_QMARK] = ACTIONS(1305),
    [anon_sym_await] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1305),
    [sym_this] = ACTIONS(1305),
    [sym_true] = ACTIONS(1305),
    [sym_false] = ACTIONS(1305),
    [sym_none] = ACTIONS(1305),
    [sym_string] = ACTIONS(1305),
    [sym_verbatim] = ACTIONS(1305),
    [sym_number] = ACTIONS(1305),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1303),
    [sym_regex] = ACTIONS(1303),
    [sym_fstring_start] = ACTIONS(1303),
  },
  [STATE(510)] = {
    [aux_sym_expression_list_repeat1] = STATE(510),
    [sym_identifier] = ACTIONS(1309),
    [anon_sym_SEMI] = ACTIONS(1309),
    [anon_sym_return] = ACTIONS(1309),
    [anon_sym_async] = ACTIONS(1309),
    [anon_sym_def] = ACTIONS(1309),
    [anon_sym_COMMA] = ACTIONS(1311),
    [anon_sym_LPAREN] = ACTIONS(1309),
    [anon_sym_LBRACK] = ACTIONS(1309),
    [anon_sym_del] = ACTIONS(1309),
    [anon_sym_raise] = ACTIONS(1309),
    [sym_pass_statement] = ACTIONS(1309),
    [sym_break_statement] = ACTIONS(1309),
    [sym_continue_statement] = ACTIONS(1309),
    [sym_debugger_statement] = ACTIONS(1309),
    [anon_sym_assert] = ACTIONS(1309),
    [anon_sym_global] = ACTIONS(1309),
    [anon_sym_nonlocal] = ACTIONS(1309),
    [anon_sym_yield] = ACTIONS(1309),
    [anon_sym_from] = ACTIONS(1309),
    [anon_sym_import] = ACTIONS(1309),
    [anon_sym_DOT] = ACTIONS(1309),
    [anon_sym_if] = ACTIONS(1309),
    [anon_sym_for] = ACTIONS(1309),
    [anon_sym_in] = ACTIONS(1309),
    [anon_sym_while] = ACTIONS(1309),
    [anon_sym_do] = ACTIONS(1309),
    [anon_sym_try] = ACTIONS(1309),
    [anon_sym_with] = ACTIONS(1309),
    [anon_sym_STAR] = ACTIONS(1309),
    [anon_sym_STAR_STAR] = ACTIONS(1309),
    [anon_sym_class] = ACTIONS(1309),
    [anon_sym_AT] = ACTIONS(1309),
    [anon_sym_not] = ACTIONS(1309),
    [anon_sym_and] = ACTIONS(1309),
    [anon_sym_or] = ACTIONS(1309),
    [anon_sym_PLUS] = ACTIONS(1309),
    [anon_sym_DASH] = ACTIONS(1309),
    [anon_sym_SLASH] = ACTIONS(1309),
    [anon_sym_SLASH_SLASH] = ACTIONS(1309),
    [anon_sym_PERCENT] = ACTIONS(1309),
    [anon_sym_PIPE] = ACTIONS(1309),
    [anon_sym_CARET] = ACTIONS(1309),
    [anon_sym_AMP] = ACTIONS(1309),
    [anon_sym_LT_LT] = ACTIONS(1309),
    [anon_sym_GT_GT] = ACTIONS(1309),
    [anon_sym_GT_GT_GT] = ACTIONS(1309),
    [anon_sym_TILDE] = ACTIONS(1309),
    [anon_sym_typeof] = ACTIONS(1309),
    [anon_sym_void] = ACTIONS(1309),
    [anon_sym_LT] = ACTIONS(1309),
    [anon_sym_LT_EQ] = ACTIONS(1309),
    [anon_sym_EQ_EQ] = ACTIONS(1309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1309),
    [anon_sym_BANG_EQ] = ACTIONS(1309),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1309),
    [anon_sym_GT_EQ] = ACTIONS(1309),
    [anon_sym_GT] = ACTIONS(1309),
    [anon_sym_is] = ACTIONS(1309),
    [anon_sym_instanceof] = ACTIONS(1309),
    [anon_sym_new] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(1309),
    [anon_sym_await] = ACTIONS(1309),
    [anon_sym_LBRACE] = ACTIONS(1309),
    [sym_this] = ACTIONS(1309),
    [sym_true] = ACTIONS(1309),
    [sym_false] = ACTIONS(1309),
    [sym_none] = ACTIONS(1309),
    [sym_string] = ACTIONS(1309),
    [sym_verbatim] = ACTIONS(1309),
    [sym_number] = ACTIONS(1309),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1314),
    [sym_regex] = ACTIONS(1314),
    [sym_fstring_start] = ACTIONS(1314),
  },
  [STATE(511)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1287),
    [anon_sym_SEMI] = ACTIONS(1287),
    [anon_sym_return] = ACTIONS(1287),
    [anon_sym_async] = ACTIONS(1287),
    [anon_sym_def] = ACTIONS(1287),
    [anon_sym_COMMA] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1287),
    [anon_sym_LBRACK] = ACTIONS(1287),
    [anon_sym_del] = ACTIONS(1287),
    [anon_sym_raise] = ACTIONS(1287),
    [sym_pass_statement] = ACTIONS(1287),
    [sym_break_statement] = ACTIONS(1287),
    [sym_continue_statement] = ACTIONS(1287),
    [sym_debugger_statement] = ACTIONS(1287),
    [anon_sym_assert] = ACTIONS(1287),
    [anon_sym_global] = ACTIONS(1287),
    [anon_sym_nonlocal] = ACTIONS(1287),
    [anon_sym_yield] = ACTIONS(1287),
    [anon_sym_from] = ACTIONS(1287),
    [anon_sym_import] = ACTIONS(1287),
    [anon_sym_DOT] = ACTIONS(1287),
    [anon_sym_if] = ACTIONS(1287),
    [anon_sym_for] = ACTIONS(1287),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1287),
    [anon_sym_do] = ACTIONS(1287),
    [anon_sym_try] = ACTIONS(1287),
    [anon_sym_with] = ACTIONS(1287),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1287),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1287),
    [anon_sym_typeof] = ACTIONS(1287),
    [anon_sym_void] = ACTIONS(1287),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1287),
    [anon_sym_QMARK] = ACTIONS(1287),
    [anon_sym_await] = ACTIONS(1287),
    [anon_sym_LBRACE] = ACTIONS(1287),
    [sym_this] = ACTIONS(1287),
    [sym_true] = ACTIONS(1287),
    [sym_false] = ACTIONS(1287),
    [sym_none] = ACTIONS(1287),
    [sym_string] = ACTIONS(1287),
    [sym_verbatim] = ACTIONS(1287),
    [sym_number] = ACTIONS(1287),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1291),
    [sym_regex] = ACTIONS(1291),
    [sym_fstring_start] = ACTIONS(1291),
  },
  [STATE(512)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(715),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(717),
    [anon_sym_yield] = ACTIONS(719),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(513)] = {
    [sym_identifier] = ACTIONS(1318),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1318),
    [anon_sym_async] = ACTIONS(1318),
    [anon_sym_def] = ACTIONS(1318),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1320),
    [anon_sym_LBRACK] = ACTIONS(1320),
    [anon_sym_del] = ACTIONS(1318),
    [anon_sym_raise] = ACTIONS(1318),
    [sym_pass_statement] = ACTIONS(1318),
    [sym_break_statement] = ACTIONS(1318),
    [sym_continue_statement] = ACTIONS(1318),
    [sym_debugger_statement] = ACTIONS(1318),
    [anon_sym_assert] = ACTIONS(1318),
    [anon_sym_global] = ACTIONS(1318),
    [anon_sym_nonlocal] = ACTIONS(1318),
    [anon_sym_yield] = ACTIONS(1318),
    [anon_sym_from] = ACTIONS(1318),
    [anon_sym_import] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1320),
    [anon_sym_for] = ACTIONS(1318),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1318),
    [anon_sym_do] = ACTIONS(1318),
    [anon_sym_try] = ACTIONS(1318),
    [anon_sym_with] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1318),
    [anon_sym_AT] = ACTIONS(1320),
    [anon_sym_not] = ACTIONS(1320),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_typeof] = ACTIONS(1318),
    [anon_sym_void] = ACTIONS(1318),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1318),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [sym_this] = ACTIONS(1318),
    [sym_true] = ACTIONS(1318),
    [sym_false] = ACTIONS(1318),
    [sym_none] = ACTIONS(1318),
    [sym_string] = ACTIONS(1318),
    [sym_verbatim] = ACTIONS(1318),
    [sym_number] = ACTIONS(1318),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym__dedent] = ACTIONS(1323),
    [sym_regex] = ACTIONS(1323),
    [sym_fstring_start] = ACTIONS(1323),
  },
  [STATE(514)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1714),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(531),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(537),
    [anon_sym_raise] = ACTIONS(539),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(541),
    [anon_sym_yield] = ACTIONS(543),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(515)] = {
    [ts_builtin_sym_end] = ACTIONS(1325),
    [sym_identifier] = ACTIONS(1327),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1327),
    [anon_sym_async] = ACTIONS(1327),
    [anon_sym_def] = ACTIONS(1327),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1329),
    [anon_sym_LBRACK] = ACTIONS(1329),
    [anon_sym_del] = ACTIONS(1327),
    [anon_sym_raise] = ACTIONS(1327),
    [sym_pass_statement] = ACTIONS(1327),
    [sym_break_statement] = ACTIONS(1327),
    [sym_continue_statement] = ACTIONS(1327),
    [sym_debugger_statement] = ACTIONS(1327),
    [anon_sym_assert] = ACTIONS(1327),
    [anon_sym_global] = ACTIONS(1327),
    [anon_sym_nonlocal] = ACTIONS(1327),
    [anon_sym_yield] = ACTIONS(1327),
    [anon_sym_from] = ACTIONS(1327),
    [anon_sym_import] = ACTIONS(1327),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1329),
    [anon_sym_for] = ACTIONS(1327),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1327),
    [anon_sym_do] = ACTIONS(1327),
    [anon_sym_try] = ACTIONS(1327),
    [anon_sym_with] = ACTIONS(1327),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1327),
    [anon_sym_AT] = ACTIONS(1329),
    [anon_sym_not] = ACTIONS(1329),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1329),
    [anon_sym_DASH] = ACTIONS(1329),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1327),
    [anon_sym_typeof] = ACTIONS(1327),
    [anon_sym_void] = ACTIONS(1327),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1327),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1327),
    [anon_sym_LBRACE] = ACTIONS(1327),
    [sym_this] = ACTIONS(1327),
    [sym_true] = ACTIONS(1327),
    [sym_false] = ACTIONS(1327),
    [sym_none] = ACTIONS(1327),
    [sym_string] = ACTIONS(1327),
    [sym_verbatim] = ACTIONS(1327),
    [sym_number] = ACTIONS(1327),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym_regex] = ACTIONS(1325),
    [sym_fstring_start] = ACTIONS(1325),
  },
  [STATE(516)] = {
    [sym_identifier] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1332),
    [anon_sym_async] = ACTIONS(1332),
    [anon_sym_def] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1334),
    [anon_sym_del] = ACTIONS(1332),
    [anon_sym_raise] = ACTIONS(1332),
    [sym_pass_statement] = ACTIONS(1332),
    [sym_break_statement] = ACTIONS(1332),
    [sym_continue_statement] = ACTIONS(1332),
    [sym_debugger_statement] = ACTIONS(1332),
    [anon_sym_assert] = ACTIONS(1332),
    [anon_sym_global] = ACTIONS(1332),
    [anon_sym_nonlocal] = ACTIONS(1332),
    [anon_sym_yield] = ACTIONS(1332),
    [anon_sym_from] = ACTIONS(1332),
    [anon_sym_import] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1332),
    [anon_sym_do] = ACTIONS(1332),
    [anon_sym_try] = ACTIONS(1332),
    [anon_sym_with] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1332),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_not] = ACTIONS(1334),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_typeof] = ACTIONS(1332),
    [anon_sym_void] = ACTIONS(1332),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1332),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [sym_this] = ACTIONS(1332),
    [sym_true] = ACTIONS(1332),
    [sym_false] = ACTIONS(1332),
    [sym_none] = ACTIONS(1332),
    [sym_string] = ACTIONS(1332),
    [sym_verbatim] = ACTIONS(1332),
    [sym_number] = ACTIONS(1332),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym__dedent] = ACTIONS(1337),
    [sym_regex] = ACTIONS(1337),
    [sym_fstring_start] = ACTIONS(1337),
  },
  [STATE(517)] = {
    [aux_sym_expression_list_repeat1] = STATE(517),
    [ts_builtin_sym_end] = ACTIONS(1314),
    [sym_identifier] = ACTIONS(1309),
    [anon_sym_SEMI] = ACTIONS(1309),
    [anon_sym_return] = ACTIONS(1309),
    [anon_sym_async] = ACTIONS(1309),
    [anon_sym_def] = ACTIONS(1309),
    [anon_sym_COMMA] = ACTIONS(1339),
    [anon_sym_LPAREN] = ACTIONS(1309),
    [anon_sym_LBRACK] = ACTIONS(1309),
    [anon_sym_del] = ACTIONS(1309),
    [anon_sym_raise] = ACTIONS(1309),
    [sym_pass_statement] = ACTIONS(1309),
    [sym_break_statement] = ACTIONS(1309),
    [sym_continue_statement] = ACTIONS(1309),
    [sym_debugger_statement] = ACTIONS(1309),
    [anon_sym_assert] = ACTIONS(1309),
    [anon_sym_global] = ACTIONS(1309),
    [anon_sym_nonlocal] = ACTIONS(1309),
    [anon_sym_yield] = ACTIONS(1309),
    [anon_sym_from] = ACTIONS(1309),
    [anon_sym_import] = ACTIONS(1309),
    [anon_sym_DOT] = ACTIONS(1309),
    [anon_sym_if] = ACTIONS(1309),
    [anon_sym_for] = ACTIONS(1309),
    [anon_sym_in] = ACTIONS(1309),
    [anon_sym_while] = ACTIONS(1309),
    [anon_sym_do] = ACTIONS(1309),
    [anon_sym_try] = ACTIONS(1309),
    [anon_sym_with] = ACTIONS(1309),
    [anon_sym_STAR] = ACTIONS(1309),
    [anon_sym_STAR_STAR] = ACTIONS(1309),
    [anon_sym_class] = ACTIONS(1309),
    [anon_sym_AT] = ACTIONS(1309),
    [anon_sym_not] = ACTIONS(1309),
    [anon_sym_and] = ACTIONS(1309),
    [anon_sym_or] = ACTIONS(1309),
    [anon_sym_PLUS] = ACTIONS(1309),
    [anon_sym_DASH] = ACTIONS(1309),
    [anon_sym_SLASH] = ACTIONS(1309),
    [anon_sym_SLASH_SLASH] = ACTIONS(1309),
    [anon_sym_PERCENT] = ACTIONS(1309),
    [anon_sym_PIPE] = ACTIONS(1309),
    [anon_sym_CARET] = ACTIONS(1309),
    [anon_sym_AMP] = ACTIONS(1309),
    [anon_sym_LT_LT] = ACTIONS(1309),
    [anon_sym_GT_GT] = ACTIONS(1309),
    [anon_sym_GT_GT_GT] = ACTIONS(1309),
    [anon_sym_TILDE] = ACTIONS(1309),
    [anon_sym_typeof] = ACTIONS(1309),
    [anon_sym_void] = ACTIONS(1309),
    [anon_sym_LT] = ACTIONS(1309),
    [anon_sym_LT_EQ] = ACTIONS(1309),
    [anon_sym_EQ_EQ] = ACTIONS(1309),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1309),
    [anon_sym_BANG_EQ] = ACTIONS(1309),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1309),
    [anon_sym_GT_EQ] = ACTIONS(1309),
    [anon_sym_GT] = ACTIONS(1309),
    [anon_sym_is] = ACTIONS(1309),
    [anon_sym_instanceof] = ACTIONS(1309),
    [anon_sym_new] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(1309),
    [anon_sym_await] = ACTIONS(1309),
    [anon_sym_LBRACE] = ACTIONS(1309),
    [sym_this] = ACTIONS(1309),
    [sym_true] = ACTIONS(1309),
    [sym_false] = ACTIONS(1309),
    [sym_none] = ACTIONS(1309),
    [sym_string] = ACTIONS(1309),
    [sym_verbatim] = ACTIONS(1309),
    [sym_number] = ACTIONS(1309),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1314),
    [sym_fstring_start] = ACTIONS(1314),
  },
  [STATE(518)] = {
    [ts_builtin_sym_end] = ACTIONS(1192),
    [sym_identifier] = ACTIONS(1190),
    [anon_sym_SEMI] = ACTIONS(1190),
    [anon_sym_return] = ACTIONS(1190),
    [anon_sym_async] = ACTIONS(1190),
    [anon_sym_def] = ACTIONS(1190),
    [anon_sym_COMMA] = ACTIONS(1190),
    [anon_sym_LPAREN] = ACTIONS(1190),
    [anon_sym_LBRACK] = ACTIONS(1190),
    [anon_sym_del] = ACTIONS(1190),
    [anon_sym_raise] = ACTIONS(1190),
    [sym_pass_statement] = ACTIONS(1190),
    [sym_break_statement] = ACTIONS(1190),
    [sym_continue_statement] = ACTIONS(1190),
    [sym_debugger_statement] = ACTIONS(1190),
    [anon_sym_assert] = ACTIONS(1190),
    [anon_sym_global] = ACTIONS(1190),
    [anon_sym_nonlocal] = ACTIONS(1190),
    [anon_sym_yield] = ACTIONS(1190),
    [anon_sym_from] = ACTIONS(1190),
    [anon_sym_import] = ACTIONS(1190),
    [anon_sym_DOT] = ACTIONS(1190),
    [anon_sym_if] = ACTIONS(1190),
    [anon_sym_for] = ACTIONS(1190),
    [anon_sym_in] = ACTIONS(1190),
    [anon_sym_while] = ACTIONS(1190),
    [anon_sym_do] = ACTIONS(1190),
    [anon_sym_try] = ACTIONS(1190),
    [anon_sym_with] = ACTIONS(1190),
    [anon_sym_STAR] = ACTIONS(1190),
    [anon_sym_STAR_STAR] = ACTIONS(1190),
    [anon_sym_class] = ACTIONS(1190),
    [anon_sym_AT] = ACTIONS(1190),
    [anon_sym_not] = ACTIONS(1190),
    [anon_sym_and] = ACTIONS(1190),
    [anon_sym_or] = ACTIONS(1190),
    [anon_sym_PLUS] = ACTIONS(1190),
    [anon_sym_DASH] = ACTIONS(1190),
    [anon_sym_SLASH] = ACTIONS(1190),
    [anon_sym_SLASH_SLASH] = ACTIONS(1190),
    [anon_sym_PERCENT] = ACTIONS(1190),
    [anon_sym_PIPE] = ACTIONS(1190),
    [anon_sym_CARET] = ACTIONS(1190),
    [anon_sym_AMP] = ACTIONS(1190),
    [anon_sym_LT_LT] = ACTIONS(1190),
    [anon_sym_GT_GT] = ACTIONS(1190),
    [anon_sym_GT_GT_GT] = ACTIONS(1190),
    [anon_sym_TILDE] = ACTIONS(1190),
    [anon_sym_typeof] = ACTIONS(1190),
    [anon_sym_void] = ACTIONS(1190),
    [anon_sym_LT] = ACTIONS(1190),
    [anon_sym_LT_EQ] = ACTIONS(1190),
    [anon_sym_EQ_EQ] = ACTIONS(1190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1190),
    [anon_sym_BANG_EQ] = ACTIONS(1190),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1190),
    [anon_sym_GT_EQ] = ACTIONS(1190),
    [anon_sym_GT] = ACTIONS(1190),
    [anon_sym_is] = ACTIONS(1190),
    [anon_sym_instanceof] = ACTIONS(1190),
    [anon_sym_new] = ACTIONS(1190),
    [anon_sym_QMARK] = ACTIONS(1190),
    [anon_sym_await] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1190),
    [sym_this] = ACTIONS(1190),
    [sym_true] = ACTIONS(1190),
    [sym_false] = ACTIONS(1190),
    [sym_none] = ACTIONS(1190),
    [sym_string] = ACTIONS(1190),
    [sym_verbatim] = ACTIONS(1190),
    [sym_number] = ACTIONS(1190),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1192),
    [sym_regex] = ACTIONS(1192),
    [sym_fstring_start] = ACTIONS(1192),
  },
  [STATE(519)] = {
    [sym__expression_statement] = STATE(618),
    [sym_assignment] = STATE(665),
    [sym_augmented_assignment] = STATE(665),
    [sym__assign_lhs] = STATE(2535),
    [sym_expression_list] = STATE(618),
    [sym_pattern_list] = STATE(2535),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(618),
    [sym_delete_statement] = STATE(618),
    [sym_raise_statement] = STATE(618),
    [sym_assert_statement] = STATE(618),
    [sym_yield_statement] = STATE(618),
    [sym_yield] = STATE(668),
    [sym__expression] = STATE(447),
    [sym__primary_expression] = STATE(463),
    [sym_not_operator] = STATE(665),
    [sym_boolean_operator] = STATE(665),
    [sym_binary_operator] = STATE(665),
    [sym_unary_operator] = STATE(665),
    [sym_comparison_operator] = STATE(665),
    [sym_conditional_expression] = STATE(665),
    [sym_attribute] = STATE(234),
    [sym_subscript] = STATE(234),
    [sym_call] = STATE(646),
    [sym_slice_call] = STATE(646),
    [sym_new_expression] = STATE(646),
    [sym_existential] = STATE(646),
    [sym_await] = STATE(665),
    [sym_anonymous_function] = STATE(646),
    [sym__inline_simple_statement] = STATE(618),
    [sym_list] = STATE(646),
    [sym_set] = STATE(646),
    [sym_tuple] = STATE(646),
    [sym_parenthesized_expression] = STATE(646),
    [sym_dictionary] = STATE(646),
    [sym_list_comprehension] = STATE(646),
    [sym_set_comprehension] = STATE(646),
    [sym_dictionary_comprehension] = STATE(646),
    [sym_generator_expression] = STATE(646),
    [sym_concatenated_string] = STATE(646),
    [sym_f_string] = STATE(458),
    [sym_identifier] = ACTIONS(117),
    [anon_sym_return] = ACTIONS(121),
    [anon_sym_async] = ACTIONS(123),
    [anon_sym_def] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(127),
    [anon_sym_LBRACK] = ACTIONS(129),
    [anon_sym_del] = ACTIONS(159),
    [anon_sym_raise] = ACTIONS(161),
    [sym_pass_statement] = ACTIONS(135),
    [sym_break_statement] = ACTIONS(135),
    [sym_continue_statement] = ACTIONS(135),
    [anon_sym_assert] = ACTIONS(163),
    [anon_sym_yield] = ACTIONS(165),
    [anon_sym_not] = ACTIONS(141),
    [anon_sym_PLUS] = ACTIONS(143),
    [anon_sym_DASH] = ACTIONS(143),
    [anon_sym_TILDE] = ACTIONS(143),
    [anon_sym_typeof] = ACTIONS(143),
    [anon_sym_void] = ACTIONS(143),
    [anon_sym_new] = ACTIONS(145),
    [anon_sym_await] = ACTIONS(147),
    [anon_sym_LBRACE] = ACTIONS(149),
    [sym_this] = ACTIONS(117),
    [sym_true] = ACTIONS(151),
    [sym_false] = ACTIONS(151),
    [sym_none] = ACTIONS(151),
    [sym_string] = ACTIONS(153),
    [sym_verbatim] = ACTIONS(151),
    [sym_number] = ACTIONS(151),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(155),
    [sym_fstring_start] = ACTIONS(157),
  },
  [STATE(520)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1741),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(453),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(671),
    [anon_sym_raise] = ACTIONS(673),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(675),
    [anon_sym_yield] = ACTIONS(677),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(521)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1715),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(487),
    [anon_sym_raise] = ACTIONS(489),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(491),
    [anon_sym_yield] = ACTIONS(493),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(522)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2548),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2548),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1774),
    [sym__primary_expression] = STATE(1879),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1513),
    [sym_subscript] = STATE(1513),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(603),
    [anon_sym_return] = ACTIONS(605),
    [anon_sym_async] = ACTIONS(607),
    [anon_sym_def] = ACTIONS(609),
    [anon_sym_LPAREN] = ACTIONS(611),
    [anon_sym_LBRACK] = ACTIONS(613),
    [anon_sym_del] = ACTIONS(615),
    [anon_sym_raise] = ACTIONS(617),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(619),
    [anon_sym_yield] = ACTIONS(621),
    [anon_sym_not] = ACTIONS(623),
    [anon_sym_PLUS] = ACTIONS(625),
    [anon_sym_DASH] = ACTIONS(625),
    [anon_sym_TILDE] = ACTIONS(625),
    [anon_sym_typeof] = ACTIONS(625),
    [anon_sym_void] = ACTIONS(625),
    [anon_sym_new] = ACTIONS(627),
    [anon_sym_await] = ACTIONS(629),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(603),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(523)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1941),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(707),
    [anon_sym_raise] = ACTIONS(709),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(711),
    [anon_sym_yield] = ACTIONS(713),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(524)] = {
    [sym__expression_statement] = STATE(2243),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2243),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2243),
    [sym_delete_statement] = STATE(2243),
    [sym_raise_statement] = STATE(2243),
    [sym_assert_statement] = STATE(2243),
    [sym_yield_statement] = STATE(2243),
    [sym_yield] = STATE(2272),
    [sym__expression] = STATE(1880),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__inline_simple_statement] = STATE(2243),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(653),
    [anon_sym_raise] = ACTIONS(655),
    [sym_pass_statement] = ACTIONS(657),
    [sym_break_statement] = ACTIONS(657),
    [sym_continue_statement] = ACTIONS(657),
    [anon_sym_assert] = ACTIONS(659),
    [anon_sym_yield] = ACTIONS(661),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(525)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(665),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(667),
    [anon_sym_yield] = ACTIONS(669),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(526)] = {
    [sym_identifier] = ACTIONS(1327),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1327),
    [anon_sym_async] = ACTIONS(1327),
    [anon_sym_def] = ACTIONS(1327),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1329),
    [anon_sym_LBRACK] = ACTIONS(1329),
    [anon_sym_del] = ACTIONS(1327),
    [anon_sym_raise] = ACTIONS(1327),
    [sym_pass_statement] = ACTIONS(1327),
    [sym_break_statement] = ACTIONS(1327),
    [sym_continue_statement] = ACTIONS(1327),
    [sym_debugger_statement] = ACTIONS(1327),
    [anon_sym_assert] = ACTIONS(1327),
    [anon_sym_global] = ACTIONS(1327),
    [anon_sym_nonlocal] = ACTIONS(1327),
    [anon_sym_yield] = ACTIONS(1327),
    [anon_sym_from] = ACTIONS(1327),
    [anon_sym_import] = ACTIONS(1327),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1329),
    [anon_sym_for] = ACTIONS(1327),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1327),
    [anon_sym_do] = ACTIONS(1327),
    [anon_sym_try] = ACTIONS(1327),
    [anon_sym_with] = ACTIONS(1327),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1327),
    [anon_sym_AT] = ACTIONS(1329),
    [anon_sym_not] = ACTIONS(1329),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1329),
    [anon_sym_DASH] = ACTIONS(1329),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1327),
    [anon_sym_typeof] = ACTIONS(1327),
    [anon_sym_void] = ACTIONS(1327),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1327),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1327),
    [anon_sym_LBRACE] = ACTIONS(1327),
    [sym_this] = ACTIONS(1327),
    [sym_true] = ACTIONS(1327),
    [sym_false] = ACTIONS(1327),
    [sym_none] = ACTIONS(1327),
    [sym_string] = ACTIONS(1327),
    [sym_verbatim] = ACTIONS(1327),
    [sym_number] = ACTIONS(1327),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym__dedent] = ACTIONS(1325),
    [sym_regex] = ACTIONS(1325),
    [sym_fstring_start] = ACTIONS(1325),
  },
  [STATE(527)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1314),
    [sym_identifier] = ACTIONS(1309),
    [anon_sym_SEMI] = ACTIONS(1309),
    [anon_sym_return] = ACTIONS(1309),
    [anon_sym_async] = ACTIONS(1309),
    [anon_sym_def] = ACTIONS(1309),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_LPAREN] = ACTIONS(1309),
    [anon_sym_LBRACK] = ACTIONS(1309),
    [anon_sym_del] = ACTIONS(1309),
    [anon_sym_raise] = ACTIONS(1309),
    [sym_pass_statement] = ACTIONS(1309),
    [sym_break_statement] = ACTIONS(1309),
    [sym_continue_statement] = ACTIONS(1309),
    [sym_debugger_statement] = ACTIONS(1309),
    [anon_sym_assert] = ACTIONS(1309),
    [anon_sym_global] = ACTIONS(1309),
    [anon_sym_nonlocal] = ACTIONS(1309),
    [anon_sym_yield] = ACTIONS(1309),
    [anon_sym_from] = ACTIONS(1309),
    [anon_sym_import] = ACTIONS(1309),
    [anon_sym_DOT] = ACTIONS(1309),
    [anon_sym_if] = ACTIONS(1309),
    [anon_sym_for] = ACTIONS(1309),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1309),
    [anon_sym_do] = ACTIONS(1309),
    [anon_sym_try] = ACTIONS(1309),
    [anon_sym_with] = ACTIONS(1309),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1309),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1309),
    [anon_sym_typeof] = ACTIONS(1309),
    [anon_sym_void] = ACTIONS(1309),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(1309),
    [anon_sym_await] = ACTIONS(1309),
    [anon_sym_LBRACE] = ACTIONS(1309),
    [sym_this] = ACTIONS(1309),
    [sym_true] = ACTIONS(1309),
    [sym_false] = ACTIONS(1309),
    [sym_none] = ACTIONS(1309),
    [sym_string] = ACTIONS(1309),
    [sym_verbatim] = ACTIONS(1309),
    [sym_number] = ACTIONS(1309),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1314),
    [sym_fstring_start] = ACTIONS(1314),
  },
  [STATE(528)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2547),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2547),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1935),
    [sym__primary_expression] = STATE(2061),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(679),
    [anon_sym_async] = ACTIONS(681),
    [anon_sym_def] = ACTIONS(683),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(685),
    [anon_sym_raise] = ACTIONS(687),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(689),
    [anon_sym_yield] = ACTIONS(691),
    [anon_sym_not] = ACTIONS(693),
    [anon_sym_PLUS] = ACTIONS(695),
    [anon_sym_DASH] = ACTIONS(695),
    [anon_sym_TILDE] = ACTIONS(695),
    [anon_sym_typeof] = ACTIONS(695),
    [anon_sym_void] = ACTIONS(695),
    [anon_sym_new] = ACTIONS(697),
    [anon_sym_await] = ACTIONS(699),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(529)] = {
    [ts_builtin_sym_end] = ACTIONS(1342),
    [sym_identifier] = ACTIONS(1344),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_return] = ACTIONS(1344),
    [anon_sym_async] = ACTIONS(1344),
    [anon_sym_def] = ACTIONS(1344),
    [anon_sym_COMMA] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_LBRACK] = ACTIONS(1344),
    [anon_sym_del] = ACTIONS(1344),
    [anon_sym_raise] = ACTIONS(1344),
    [sym_pass_statement] = ACTIONS(1344),
    [sym_break_statement] = ACTIONS(1344),
    [sym_continue_statement] = ACTIONS(1344),
    [sym_debugger_statement] = ACTIONS(1344),
    [anon_sym_assert] = ACTIONS(1344),
    [anon_sym_global] = ACTIONS(1344),
    [anon_sym_nonlocal] = ACTIONS(1344),
    [anon_sym_yield] = ACTIONS(1344),
    [anon_sym_from] = ACTIONS(1344),
    [anon_sym_import] = ACTIONS(1344),
    [anon_sym_DOT] = ACTIONS(1344),
    [anon_sym_if] = ACTIONS(1344),
    [anon_sym_for] = ACTIONS(1344),
    [anon_sym_in] = ACTIONS(1344),
    [anon_sym_while] = ACTIONS(1344),
    [anon_sym_do] = ACTIONS(1344),
    [anon_sym_try] = ACTIONS(1344),
    [anon_sym_with] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_STAR_STAR] = ACTIONS(1344),
    [anon_sym_class] = ACTIONS(1344),
    [anon_sym_AT] = ACTIONS(1344),
    [anon_sym_not] = ACTIONS(1344),
    [anon_sym_and] = ACTIONS(1344),
    [anon_sym_or] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_SLASH_SLASH] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_LT_LT] = ACTIONS(1344),
    [anon_sym_GT_GT] = ACTIONS(1344),
    [anon_sym_GT_GT_GT] = ACTIONS(1344),
    [anon_sym_TILDE] = ACTIONS(1344),
    [anon_sym_typeof] = ACTIONS(1344),
    [anon_sym_void] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_LT_EQ] = ACTIONS(1344),
    [anon_sym_EQ_EQ] = ACTIONS(1344),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1344),
    [anon_sym_GT_EQ] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_is] = ACTIONS(1344),
    [anon_sym_instanceof] = ACTIONS(1344),
    [anon_sym_new] = ACTIONS(1344),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_await] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [sym_this] = ACTIONS(1344),
    [sym_true] = ACTIONS(1344),
    [sym_false] = ACTIONS(1344),
    [sym_none] = ACTIONS(1344),
    [sym_string] = ACTIONS(1344),
    [sym_verbatim] = ACTIONS(1344),
    [sym_number] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1342),
    [sym_regex] = ACTIONS(1342),
    [sym_fstring_start] = ACTIONS(1342),
  },
  [STATE(530)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1705),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(397),
    [anon_sym_raise] = ACTIONS(399),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(403),
    [anon_sym_yield] = ACTIONS(405),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(531)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1717),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(385),
    [anon_sym_return] = ACTIONS(387),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_del] = ACTIONS(663),
    [anon_sym_raise] = ACTIONS(701),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(703),
    [anon_sym_yield] = ACTIONS(705),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(532)] = {
    [sym_identifier] = ACTIONS(1344),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_return] = ACTIONS(1344),
    [anon_sym_async] = ACTIONS(1344),
    [anon_sym_def] = ACTIONS(1344),
    [anon_sym_COMMA] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_LBRACK] = ACTIONS(1344),
    [anon_sym_del] = ACTIONS(1344),
    [anon_sym_raise] = ACTIONS(1344),
    [sym_pass_statement] = ACTIONS(1344),
    [sym_break_statement] = ACTIONS(1344),
    [sym_continue_statement] = ACTIONS(1344),
    [sym_debugger_statement] = ACTIONS(1344),
    [anon_sym_assert] = ACTIONS(1344),
    [anon_sym_global] = ACTIONS(1344),
    [anon_sym_nonlocal] = ACTIONS(1344),
    [anon_sym_yield] = ACTIONS(1344),
    [anon_sym_from] = ACTIONS(1344),
    [anon_sym_import] = ACTIONS(1344),
    [anon_sym_DOT] = ACTIONS(1344),
    [anon_sym_if] = ACTIONS(1344),
    [anon_sym_for] = ACTIONS(1344),
    [anon_sym_in] = ACTIONS(1344),
    [anon_sym_while] = ACTIONS(1344),
    [anon_sym_do] = ACTIONS(1344),
    [anon_sym_try] = ACTIONS(1344),
    [anon_sym_with] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_STAR_STAR] = ACTIONS(1344),
    [anon_sym_class] = ACTIONS(1344),
    [anon_sym_AT] = ACTIONS(1344),
    [anon_sym_not] = ACTIONS(1344),
    [anon_sym_and] = ACTIONS(1344),
    [anon_sym_or] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_SLASH_SLASH] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_LT_LT] = ACTIONS(1344),
    [anon_sym_GT_GT] = ACTIONS(1344),
    [anon_sym_GT_GT_GT] = ACTIONS(1344),
    [anon_sym_TILDE] = ACTIONS(1344),
    [anon_sym_typeof] = ACTIONS(1344),
    [anon_sym_void] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_LT_EQ] = ACTIONS(1344),
    [anon_sym_EQ_EQ] = ACTIONS(1344),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1344),
    [anon_sym_GT_EQ] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_is] = ACTIONS(1344),
    [anon_sym_instanceof] = ACTIONS(1344),
    [anon_sym_new] = ACTIONS(1344),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_await] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [sym_this] = ACTIONS(1344),
    [sym_true] = ACTIONS(1344),
    [sym_false] = ACTIONS(1344),
    [sym_none] = ACTIONS(1344),
    [sym_string] = ACTIONS(1344),
    [sym_verbatim] = ACTIONS(1344),
    [sym_number] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1342),
    [sym__dedent] = ACTIONS(1342),
    [sym_regex] = ACTIONS(1342),
    [sym_fstring_start] = ACTIONS(1342),
  },
  [STATE(533)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(1754),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(477),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(573),
    [anon_sym_raise] = ACTIONS(575),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(577),
    [anon_sym_yield] = ACTIONS(579),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(534)] = {
    [sym_identifier] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1346),
    [anon_sym_async] = ACTIONS(1346),
    [anon_sym_def] = ACTIONS(1346),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1348),
    [anon_sym_LBRACK] = ACTIONS(1348),
    [anon_sym_del] = ACTIONS(1346),
    [anon_sym_raise] = ACTIONS(1346),
    [sym_pass_statement] = ACTIONS(1346),
    [sym_break_statement] = ACTIONS(1346),
    [sym_continue_statement] = ACTIONS(1346),
    [sym_debugger_statement] = ACTIONS(1346),
    [anon_sym_assert] = ACTIONS(1346),
    [anon_sym_global] = ACTIONS(1346),
    [anon_sym_nonlocal] = ACTIONS(1346),
    [anon_sym_yield] = ACTIONS(1346),
    [anon_sym_from] = ACTIONS(1346),
    [anon_sym_import] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1346),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1346),
    [anon_sym_do] = ACTIONS(1346),
    [anon_sym_try] = ACTIONS(1346),
    [anon_sym_with] = ACTIONS(1346),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1346),
    [anon_sym_AT] = ACTIONS(1348),
    [anon_sym_not] = ACTIONS(1348),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1346),
    [anon_sym_typeof] = ACTIONS(1346),
    [anon_sym_void] = ACTIONS(1346),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1346),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1346),
    [anon_sym_LBRACE] = ACTIONS(1346),
    [sym_this] = ACTIONS(1346),
    [sym_true] = ACTIONS(1346),
    [sym_false] = ACTIONS(1346),
    [sym_none] = ACTIONS(1346),
    [sym_string] = ACTIONS(1346),
    [sym_verbatim] = ACTIONS(1346),
    [sym_number] = ACTIONS(1346),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym__dedent] = ACTIONS(1351),
    [sym_regex] = ACTIONS(1351),
    [sym_fstring_start] = ACTIONS(1351),
  },
  [STATE(535)] = {
    [ts_builtin_sym_end] = ACTIONS(1337),
    [sym_identifier] = ACTIONS(1332),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1332),
    [anon_sym_async] = ACTIONS(1332),
    [anon_sym_def] = ACTIONS(1332),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1334),
    [anon_sym_LBRACK] = ACTIONS(1334),
    [anon_sym_del] = ACTIONS(1332),
    [anon_sym_raise] = ACTIONS(1332),
    [sym_pass_statement] = ACTIONS(1332),
    [sym_break_statement] = ACTIONS(1332),
    [sym_continue_statement] = ACTIONS(1332),
    [sym_debugger_statement] = ACTIONS(1332),
    [anon_sym_assert] = ACTIONS(1332),
    [anon_sym_global] = ACTIONS(1332),
    [anon_sym_nonlocal] = ACTIONS(1332),
    [anon_sym_yield] = ACTIONS(1332),
    [anon_sym_from] = ACTIONS(1332),
    [anon_sym_import] = ACTIONS(1332),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1332),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1332),
    [anon_sym_do] = ACTIONS(1332),
    [anon_sym_try] = ACTIONS(1332),
    [anon_sym_with] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1332),
    [anon_sym_AT] = ACTIONS(1334),
    [anon_sym_not] = ACTIONS(1334),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1334),
    [anon_sym_DASH] = ACTIONS(1334),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1332),
    [anon_sym_typeof] = ACTIONS(1332),
    [anon_sym_void] = ACTIONS(1332),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1332),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [sym_this] = ACTIONS(1332),
    [sym_true] = ACTIONS(1332),
    [sym_false] = ACTIONS(1332),
    [sym_none] = ACTIONS(1332),
    [sym_string] = ACTIONS(1332),
    [sym_verbatim] = ACTIONS(1332),
    [sym_number] = ACTIONS(1332),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym_regex] = ACTIONS(1337),
    [sym_fstring_start] = ACTIONS(1337),
  },
  [STATE(536)] = {
    [sym__expression_statement] = STATE(656),
    [sym_assignment] = STATE(649),
    [sym_augmented_assignment] = STATE(649),
    [sym__assign_lhs] = STATE(2540),
    [sym_expression_list] = STATE(656),
    [sym_pattern_list] = STATE(2540),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(656),
    [sym_delete_statement] = STATE(656),
    [sym_raise_statement] = STATE(656),
    [sym_assert_statement] = STATE(656),
    [sym_yield_statement] = STATE(656),
    [sym_yield] = STATE(651),
    [sym__expression] = STATE(454),
    [sym__primary_expression] = STATE(466),
    [sym_not_operator] = STATE(649),
    [sym_boolean_operator] = STATE(649),
    [sym_binary_operator] = STATE(649),
    [sym_unary_operator] = STATE(649),
    [sym_comparison_operator] = STATE(649),
    [sym_conditional_expression] = STATE(649),
    [sym_attribute] = STATE(237),
    [sym_subscript] = STATE(237),
    [sym_call] = STATE(641),
    [sym_slice_call] = STATE(641),
    [sym_new_expression] = STATE(641),
    [sym_existential] = STATE(641),
    [sym_await] = STATE(649),
    [sym_anonymous_function] = STATE(641),
    [sym__inline_simple_statement] = STATE(656),
    [sym_list] = STATE(641),
    [sym_set] = STATE(641),
    [sym_tuple] = STATE(641),
    [sym_parenthesized_expression] = STATE(641),
    [sym_dictionary] = STATE(641),
    [sym_list_comprehension] = STATE(641),
    [sym_set_comprehension] = STATE(641),
    [sym_dictionary_comprehension] = STATE(641),
    [sym_generator_expression] = STATE(641),
    [sym_concatenated_string] = STATE(641),
    [sym_f_string] = STATE(450),
    [sym_identifier] = ACTIONS(71),
    [anon_sym_return] = ACTIONS(75),
    [anon_sym_async] = ACTIONS(77),
    [anon_sym_def] = ACTIONS(79),
    [anon_sym_LPAREN] = ACTIONS(81),
    [anon_sym_LBRACK] = ACTIONS(83),
    [anon_sym_del] = ACTIONS(167),
    [anon_sym_raise] = ACTIONS(169),
    [sym_pass_statement] = ACTIONS(89),
    [sym_break_statement] = ACTIONS(89),
    [sym_continue_statement] = ACTIONS(89),
    [anon_sym_assert] = ACTIONS(171),
    [anon_sym_yield] = ACTIONS(173),
    [anon_sym_not] = ACTIONS(95),
    [anon_sym_PLUS] = ACTIONS(97),
    [anon_sym_DASH] = ACTIONS(97),
    [anon_sym_TILDE] = ACTIONS(97),
    [anon_sym_typeof] = ACTIONS(97),
    [anon_sym_void] = ACTIONS(97),
    [anon_sym_new] = ACTIONS(99),
    [anon_sym_await] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(103),
    [sym_this] = ACTIONS(71),
    [sym_true] = ACTIONS(105),
    [sym_false] = ACTIONS(105),
    [sym_none] = ACTIONS(105),
    [sym_string] = ACTIONS(107),
    [sym_verbatim] = ACTIONS(105),
    [sym_number] = ACTIONS(105),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(111),
    [sym_fstring_start] = ACTIONS(113),
  },
  [STATE(537)] = {
    [ts_builtin_sym_end] = ACTIONS(1351),
    [sym_identifier] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1346),
    [anon_sym_async] = ACTIONS(1346),
    [anon_sym_def] = ACTIONS(1346),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1348),
    [anon_sym_LBRACK] = ACTIONS(1348),
    [anon_sym_del] = ACTIONS(1346),
    [anon_sym_raise] = ACTIONS(1346),
    [sym_pass_statement] = ACTIONS(1346),
    [sym_break_statement] = ACTIONS(1346),
    [sym_continue_statement] = ACTIONS(1346),
    [sym_debugger_statement] = ACTIONS(1346),
    [anon_sym_assert] = ACTIONS(1346),
    [anon_sym_global] = ACTIONS(1346),
    [anon_sym_nonlocal] = ACTIONS(1346),
    [anon_sym_yield] = ACTIONS(1346),
    [anon_sym_from] = ACTIONS(1346),
    [anon_sym_import] = ACTIONS(1346),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1348),
    [anon_sym_for] = ACTIONS(1346),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1346),
    [anon_sym_do] = ACTIONS(1346),
    [anon_sym_try] = ACTIONS(1346),
    [anon_sym_with] = ACTIONS(1346),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1346),
    [anon_sym_AT] = ACTIONS(1348),
    [anon_sym_not] = ACTIONS(1348),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1346),
    [anon_sym_typeof] = ACTIONS(1346),
    [anon_sym_void] = ACTIONS(1346),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1346),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1346),
    [anon_sym_LBRACE] = ACTIONS(1346),
    [sym_this] = ACTIONS(1346),
    [sym_true] = ACTIONS(1346),
    [sym_false] = ACTIONS(1346),
    [sym_none] = ACTIONS(1346),
    [sym_string] = ACTIONS(1346),
    [sym_verbatim] = ACTIONS(1346),
    [sym_number] = ACTIONS(1346),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym_regex] = ACTIONS(1351),
    [sym_fstring_start] = ACTIONS(1351),
  },
  [STATE(538)] = {
    [sym_argument_list] = STATE(678),
    [sym_identifier] = ACTIONS(1212),
    [anon_sym_SEMI] = ACTIONS(1212),
    [anon_sym_return] = ACTIONS(1212),
    [anon_sym_async] = ACTIONS(1212),
    [anon_sym_def] = ACTIONS(1212),
    [anon_sym_COMMA] = ACTIONS(1212),
    [anon_sym_LPAREN] = ACTIONS(1214),
    [anon_sym_LBRACK] = ACTIONS(1216),
    [anon_sym_del] = ACTIONS(1212),
    [anon_sym_raise] = ACTIONS(1212),
    [sym_pass_statement] = ACTIONS(1212),
    [sym_break_statement] = ACTIONS(1212),
    [sym_continue_statement] = ACTIONS(1212),
    [sym_debugger_statement] = ACTIONS(1212),
    [anon_sym_assert] = ACTIONS(1212),
    [anon_sym_global] = ACTIONS(1212),
    [anon_sym_nonlocal] = ACTIONS(1212),
    [anon_sym_yield] = ACTIONS(1212),
    [anon_sym_from] = ACTIONS(1212),
    [anon_sym_import] = ACTIONS(1212),
    [anon_sym_DOT] = ACTIONS(1218),
    [anon_sym_if] = ACTIONS(1212),
    [anon_sym_for] = ACTIONS(1212),
    [anon_sym_in] = ACTIONS(1212),
    [anon_sym_while] = ACTIONS(1212),
    [anon_sym_do] = ACTIONS(1212),
    [anon_sym_try] = ACTIONS(1212),
    [anon_sym_with] = ACTIONS(1212),
    [anon_sym_STAR] = ACTIONS(1212),
    [anon_sym_STAR_STAR] = ACTIONS(1212),
    [anon_sym_class] = ACTIONS(1212),
    [anon_sym_AT] = ACTIONS(1212),
    [anon_sym_not] = ACTIONS(1212),
    [anon_sym_and] = ACTIONS(1212),
    [anon_sym_or] = ACTIONS(1212),
    [anon_sym_PLUS] = ACTIONS(1212),
    [anon_sym_DASH] = ACTIONS(1212),
    [anon_sym_SLASH] = ACTIONS(1212),
    [anon_sym_SLASH_SLASH] = ACTIONS(1212),
    [anon_sym_PERCENT] = ACTIONS(1212),
    [anon_sym_PIPE] = ACTIONS(1212),
    [anon_sym_CARET] = ACTIONS(1212),
    [anon_sym_AMP] = ACTIONS(1212),
    [anon_sym_LT_LT] = ACTIONS(1212),
    [anon_sym_GT_GT] = ACTIONS(1212),
    [anon_sym_GT_GT_GT] = ACTIONS(1212),
    [anon_sym_TILDE] = ACTIONS(1212),
    [anon_sym_typeof] = ACTIONS(1212),
    [anon_sym_void] = ACTIONS(1212),
    [anon_sym_LT] = ACTIONS(1212),
    [anon_sym_LT_EQ] = ACTIONS(1212),
    [anon_sym_EQ_EQ] = ACTIONS(1212),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1212),
    [anon_sym_BANG_EQ] = ACTIONS(1212),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1212),
    [anon_sym_GT_EQ] = ACTIONS(1212),
    [anon_sym_GT] = ACTIONS(1212),
    [anon_sym_is] = ACTIONS(1212),
    [anon_sym_instanceof] = ACTIONS(1212),
    [anon_sym_new] = ACTIONS(1212),
    [anon_sym_QMARK] = ACTIONS(1212),
    [anon_sym_await] = ACTIONS(1212),
    [anon_sym_LBRACE] = ACTIONS(1212),
    [sym_this] = ACTIONS(1212),
    [sym_true] = ACTIONS(1212),
    [sym_false] = ACTIONS(1212),
    [sym_none] = ACTIONS(1212),
    [sym_string] = ACTIONS(1212),
    [sym_verbatim] = ACTIONS(1212),
    [sym_number] = ACTIONS(1212),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1210),
    [sym_regex] = ACTIONS(1210),
    [sym_fstring_start] = ACTIONS(1210),
  },
  [STATE(539)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1291),
    [sym_identifier] = ACTIONS(1287),
    [anon_sym_SEMI] = ACTIONS(1287),
    [anon_sym_return] = ACTIONS(1287),
    [anon_sym_async] = ACTIONS(1287),
    [anon_sym_def] = ACTIONS(1287),
    [anon_sym_COMMA] = ACTIONS(1353),
    [anon_sym_LPAREN] = ACTIONS(1287),
    [anon_sym_LBRACK] = ACTIONS(1287),
    [anon_sym_del] = ACTIONS(1287),
    [anon_sym_raise] = ACTIONS(1287),
    [sym_pass_statement] = ACTIONS(1287),
    [sym_break_statement] = ACTIONS(1287),
    [sym_continue_statement] = ACTIONS(1287),
    [sym_debugger_statement] = ACTIONS(1287),
    [anon_sym_assert] = ACTIONS(1287),
    [anon_sym_global] = ACTIONS(1287),
    [anon_sym_nonlocal] = ACTIONS(1287),
    [anon_sym_yield] = ACTIONS(1287),
    [anon_sym_from] = ACTIONS(1287),
    [anon_sym_import] = ACTIONS(1287),
    [anon_sym_DOT] = ACTIONS(1287),
    [anon_sym_if] = ACTIONS(1287),
    [anon_sym_for] = ACTIONS(1287),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1287),
    [anon_sym_do] = ACTIONS(1287),
    [anon_sym_try] = ACTIONS(1287),
    [anon_sym_with] = ACTIONS(1287),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1287),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1117),
    [anon_sym_or] = ACTIONS(1119),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1287),
    [anon_sym_typeof] = ACTIONS(1287),
    [anon_sym_void] = ACTIONS(1287),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1287),
    [anon_sym_QMARK] = ACTIONS(1287),
    [anon_sym_await] = ACTIONS(1287),
    [anon_sym_LBRACE] = ACTIONS(1287),
    [sym_this] = ACTIONS(1287),
    [sym_true] = ACTIONS(1287),
    [sym_false] = ACTIONS(1287),
    [sym_none] = ACTIONS(1287),
    [sym_string] = ACTIONS(1287),
    [sym_verbatim] = ACTIONS(1287),
    [sym_number] = ACTIONS(1287),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1291),
    [sym_fstring_start] = ACTIONS(1291),
  },
  [STATE(540)] = {
    [aux_sym_expression_list_repeat1] = STATE(510),
    [sym_identifier] = ACTIONS(505),
    [anon_sym_SEMI] = ACTIONS(505),
    [anon_sym_return] = ACTIONS(505),
    [anon_sym_async] = ACTIONS(505),
    [anon_sym_def] = ACTIONS(505),
    [anon_sym_COMMA] = ACTIONS(1355),
    [anon_sym_LPAREN] = ACTIONS(505),
    [anon_sym_LBRACK] = ACTIONS(505),
    [anon_sym_del] = ACTIONS(505),
    [anon_sym_raise] = ACTIONS(505),
    [sym_pass_statement] = ACTIONS(505),
    [sym_break_statement] = ACTIONS(505),
    [sym_continue_statement] = ACTIONS(505),
    [sym_debugger_statement] = ACTIONS(505),
    [anon_sym_assert] = ACTIONS(505),
    [anon_sym_global] = ACTIONS(505),
    [anon_sym_nonlocal] = ACTIONS(505),
    [anon_sym_yield] = ACTIONS(505),
    [anon_sym_from] = ACTIONS(505),
    [anon_sym_import] = ACTIONS(505),
    [anon_sym_DOT] = ACTIONS(505),
    [anon_sym_if] = ACTIONS(505),
    [anon_sym_for] = ACTIONS(505),
    [anon_sym_in] = ACTIONS(505),
    [anon_sym_while] = ACTIONS(505),
    [anon_sym_do] = ACTIONS(505),
    [anon_sym_try] = ACTIONS(505),
    [anon_sym_with] = ACTIONS(505),
    [anon_sym_STAR] = ACTIONS(505),
    [anon_sym_STAR_STAR] = ACTIONS(505),
    [anon_sym_class] = ACTIONS(505),
    [anon_sym_AT] = ACTIONS(505),
    [anon_sym_not] = ACTIONS(505),
    [anon_sym_and] = ACTIONS(505),
    [anon_sym_or] = ACTIONS(505),
    [anon_sym_PLUS] = ACTIONS(505),
    [anon_sym_DASH] = ACTIONS(505),
    [anon_sym_SLASH] = ACTIONS(505),
    [anon_sym_SLASH_SLASH] = ACTIONS(505),
    [anon_sym_PERCENT] = ACTIONS(505),
    [anon_sym_PIPE] = ACTIONS(505),
    [anon_sym_CARET] = ACTIONS(505),
    [anon_sym_AMP] = ACTIONS(505),
    [anon_sym_LT_LT] = ACTIONS(505),
    [anon_sym_GT_GT] = ACTIONS(505),
    [anon_sym_GT_GT_GT] = ACTIONS(505),
    [anon_sym_TILDE] = ACTIONS(505),
    [anon_sym_typeof] = ACTIONS(505),
    [anon_sym_void] = ACTIONS(505),
    [anon_sym_LT] = ACTIONS(505),
    [anon_sym_LT_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ] = ACTIONS(505),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ] = ACTIONS(505),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(505),
    [anon_sym_GT_EQ] = ACTIONS(505),
    [anon_sym_GT] = ACTIONS(505),
    [anon_sym_is] = ACTIONS(505),
    [anon_sym_instanceof] = ACTIONS(505),
    [anon_sym_new] = ACTIONS(505),
    [anon_sym_QMARK] = ACTIONS(505),
    [anon_sym_await] = ACTIONS(505),
    [anon_sym_LBRACE] = ACTIONS(505),
    [sym_this] = ACTIONS(505),
    [sym_true] = ACTIONS(505),
    [sym_false] = ACTIONS(505),
    [sym_none] = ACTIONS(505),
    [sym_string] = ACTIONS(505),
    [sym_verbatim] = ACTIONS(505),
    [sym_number] = ACTIONS(505),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(503),
    [sym_regex] = ACTIONS(503),
    [sym_fstring_start] = ACTIONS(503),
  },
  [STATE(541)] = {
    [sym__expression_statement] = STATE(2243),
    [sym_assignment] = STATE(2258),
    [sym_augmented_assignment] = STATE(2258),
    [sym__assign_lhs] = STATE(2530),
    [sym_expression_list] = STATE(2243),
    [sym_pattern_list] = STATE(2530),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(2243),
    [sym_delete_statement] = STATE(2243),
    [sym_raise_statement] = STATE(2243),
    [sym_assert_statement] = STATE(2243),
    [sym_yield_statement] = STATE(2243),
    [sym_yield] = STATE(2272),
    [sym__expression] = STATE(1967),
    [sym__primary_expression] = STATE(2119),
    [sym_not_operator] = STATE(2258),
    [sym_boolean_operator] = STATE(2258),
    [sym_binary_operator] = STATE(2258),
    [sym_unary_operator] = STATE(2258),
    [sym_comparison_operator] = STATE(2258),
    [sym_conditional_expression] = STATE(2258),
    [sym_attribute] = STATE(1527),
    [sym_subscript] = STATE(1527),
    [sym_call] = STATE(2176),
    [sym_slice_call] = STATE(2176),
    [sym_new_expression] = STATE(2176),
    [sym_existential] = STATE(2176),
    [sym_await] = STATE(2258),
    [sym_anonymous_function] = STATE(2176),
    [sym__inline_simple_statement] = STATE(2243),
    [sym_list] = STATE(2176),
    [sym_set] = STATE(2176),
    [sym_tuple] = STATE(2176),
    [sym_parenthesized_expression] = STATE(2176),
    [sym_dictionary] = STATE(2176),
    [sym_list_comprehension] = STATE(2176),
    [sym_set_comprehension] = STATE(2176),
    [sym_dictionary_comprehension] = STATE(2176),
    [sym_generator_expression] = STATE(2176),
    [sym_concatenated_string] = STATE(2176),
    [sym_f_string] = STATE(1746),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_return] = ACTIONS(647),
    [anon_sym_async] = ACTIONS(649),
    [anon_sym_def] = ACTIONS(651),
    [anon_sym_LPAREN] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_del] = ACTIONS(721),
    [anon_sym_raise] = ACTIONS(723),
    [sym_pass_statement] = ACTIONS(657),
    [sym_break_statement] = ACTIONS(657),
    [sym_continue_statement] = ACTIONS(657),
    [anon_sym_assert] = ACTIONS(725),
    [anon_sym_yield] = ACTIONS(727),
    [anon_sym_not] = ACTIONS(53),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(55),
    [anon_sym_typeof] = ACTIONS(55),
    [anon_sym_void] = ACTIONS(55),
    [anon_sym_new] = ACTIONS(57),
    [anon_sym_await] = ACTIONS(59),
    [anon_sym_LBRACE] = ACTIONS(61),
    [sym_this] = ACTIONS(7),
    [sym_true] = ACTIONS(63),
    [sym_false] = ACTIONS(63),
    [sym_none] = ACTIONS(63),
    [sym_string] = ACTIONS(65),
    [sym_verbatim] = ACTIONS(63),
    [sym_number] = ACTIONS(63),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(67),
    [sym_fstring_start] = ACTIONS(69),
  },
  [STATE(542)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_async] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1270),
    [anon_sym_del] = ACTIONS(1270),
    [anon_sym_raise] = ACTIONS(1270),
    [sym_pass_statement] = ACTIONS(1270),
    [sym_break_statement] = ACTIONS(1270),
    [sym_continue_statement] = ACTIONS(1270),
    [sym_debugger_statement] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_global] = ACTIONS(1270),
    [anon_sym_nonlocal] = ACTIONS(1270),
    [anon_sym_yield] = ACTIONS(1270),
    [anon_sym_from] = ACTIONS(1270),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_DOT] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_in] = ACTIONS(1109),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_with] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1111),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1111),
    [anon_sym_not] = ACTIONS(1115),
    [anon_sym_and] = ACTIONS(1270),
    [anon_sym_or] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1121),
    [anon_sym_DASH] = ACTIONS(1121),
    [anon_sym_SLASH] = ACTIONS(1111),
    [anon_sym_SLASH_SLASH] = ACTIONS(1111),
    [anon_sym_PERCENT] = ACTIONS(1111),
    [anon_sym_PIPE] = ACTIONS(1123),
    [anon_sym_CARET] = ACTIONS(1125),
    [anon_sym_AMP] = ACTIONS(1127),
    [anon_sym_LT_LT] = ACTIONS(1129),
    [anon_sym_GT_GT] = ACTIONS(1129),
    [anon_sym_GT_GT_GT] = ACTIONS(1129),
    [anon_sym_TILDE] = ACTIONS(1270),
    [anon_sym_typeof] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_LT] = ACTIONS(1109),
    [anon_sym_LT_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ] = ACTIONS(1109),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ] = ACTIONS(1109),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1109),
    [anon_sym_GT_EQ] = ACTIONS(1109),
    [anon_sym_GT] = ACTIONS(1109),
    [anon_sym_is] = ACTIONS(1131),
    [anon_sym_instanceof] = ACTIONS(1109),
    [anon_sym_new] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1270),
    [anon_sym_await] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1270),
    [sym_this] = ACTIONS(1270),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [sym_none] = ACTIONS(1270),
    [sym_string] = ACTIONS(1270),
    [sym_verbatim] = ACTIONS(1270),
    [sym_number] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1268),
    [sym_fstring_start] = ACTIONS(1268),
  },
  [STATE(543)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(477),
    [ts_builtin_sym_end] = ACTIONS(1357),
    [sym_identifier] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1359),
    [anon_sym_return] = ACTIONS(1359),
    [anon_sym_async] = ACTIONS(1359),
    [anon_sym_def] = ACTIONS(1359),
    [anon_sym_COMMA] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1359),
    [anon_sym_LBRACK] = ACTIONS(1359),
    [anon_sym_del] = ACTIONS(1359),
    [anon_sym_raise] = ACTIONS(1359),
    [sym_pass_statement] = ACTIONS(1359),
    [sym_break_statement] = ACTIONS(1359),
    [sym_continue_statement] = ACTIONS(1359),
    [sym_debugger_statement] = ACTIONS(1359),
    [anon_sym_assert] = ACTIONS(1359),
    [anon_sym_global] = ACTIONS(1359),
    [anon_sym_nonlocal] = ACTIONS(1359),
    [anon_sym_yield] = ACTIONS(1359),
    [anon_sym_from] = ACTIONS(1359),
    [anon_sym_import] = ACTIONS(1359),
    [anon_sym_DOT] = ACTIONS(1359),
    [anon_sym_if] = ACTIONS(1359),
    [anon_sym_for] = ACTIONS(1359),
    [anon_sym_in] = ACTIONS(1359),
    [anon_sym_while] = ACTIONS(1359),
    [anon_sym_do] = ACTIONS(1359),
    [anon_sym_try] = ACTIONS(1359),
    [anon_sym_with] = ACTIONS(1359),
    [anon_sym_STAR] = ACTIONS(1359),
    [anon_sym_STAR_STAR] = ACTIONS(1359),
    [anon_sym_class] = ACTIONS(1359),
    [anon_sym_AT] = ACTIONS(1359),
    [anon_sym_not] = ACTIONS(1359),
    [anon_sym_and] = ACTIONS(1359),
    [anon_sym_or] = ACTIONS(1359),
    [anon_sym_PLUS] = ACTIONS(1359),
    [anon_sym_DASH] = ACTIONS(1359),
    [anon_sym_SLASH] = ACTIONS(1359),
    [anon_sym_SLASH_SLASH] = ACTIONS(1359),
    [anon_sym_PERCENT] = ACTIONS(1359),
    [anon_sym_PIPE] = ACTIONS(1359),
    [anon_sym_CARET] = ACTIONS(1359),
    [anon_sym_AMP] = ACTIONS(1359),
    [anon_sym_LT_LT] = ACTIONS(1359),
    [anon_sym_GT_GT] = ACTIONS(1359),
    [anon_sym_GT_GT_GT] = ACTIONS(1359),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_typeof] = ACTIONS(1359),
    [anon_sym_void] = ACTIONS(1359),
    [anon_sym_LT] = ACTIONS(1359),
    [anon_sym_LT_EQ] = ACTIONS(1359),
    [anon_sym_EQ_EQ] = ACTIONS(1359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1359),
    [anon_sym_BANG_EQ] = ACTIONS(1359),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1359),
    [anon_sym_GT_EQ] = ACTIONS(1359),
    [anon_sym_GT] = ACTIONS(1359),
    [anon_sym_is] = ACTIONS(1359),
    [anon_sym_instanceof] = ACTIONS(1359),
    [anon_sym_new] = ACTIONS(1359),
    [anon_sym_QMARK] = ACTIONS(1359),
    [anon_sym_await] = ACTIONS(1359),
    [anon_sym_LBRACE] = ACTIONS(1359),
    [sym_this] = ACTIONS(1359),
    [sym_true] = ACTIONS(1359),
    [sym_false] = ACTIONS(1359),
    [sym_none] = ACTIONS(1359),
    [sym_string] = ACTIONS(1359),
    [sym_verbatim] = ACTIONS(1359),
    [sym_number] = ACTIONS(1359),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1357),
    [sym_fstring_start] = ACTIONS(1357),
  },
  [STATE(544)] = {
    [aux_sym__inline_body_repeat1] = STATE(545),
    [ts_builtin_sym_end] = ACTIONS(1361),
    [sym_identifier] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1365),
    [anon_sym_return] = ACTIONS(1363),
    [anon_sym_async] = ACTIONS(1363),
    [anon_sym_def] = ACTIONS(1363),
    [anon_sym_COMMA] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [anon_sym_del] = ACTIONS(1363),
    [anon_sym_raise] = ACTIONS(1363),
    [sym_pass_statement] = ACTIONS(1363),
    [sym_break_statement] = ACTIONS(1363),
    [sym_continue_statement] = ACTIONS(1363),
    [sym_debugger_statement] = ACTIONS(1363),
    [anon_sym_assert] = ACTIONS(1363),
    [anon_sym_global] = ACTIONS(1363),
    [anon_sym_nonlocal] = ACTIONS(1363),
    [anon_sym_yield] = ACTIONS(1363),
    [anon_sym_from] = ACTIONS(1363),
    [anon_sym_import] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_if] = ACTIONS(1363),
    [anon_sym_for] = ACTIONS(1363),
    [anon_sym_in] = ACTIONS(1363),
    [anon_sym_while] = ACTIONS(1363),
    [anon_sym_do] = ACTIONS(1363),
    [anon_sym_try] = ACTIONS(1363),
    [anon_sym_with] = ACTIONS(1363),
    [anon_sym_STAR] = ACTIONS(1363),
    [anon_sym_STAR_STAR] = ACTIONS(1363),
    [anon_sym_class] = ACTIONS(1363),
    [anon_sym_AT] = ACTIONS(1363),
    [anon_sym_not] = ACTIONS(1363),
    [anon_sym_and] = ACTIONS(1363),
    [anon_sym_or] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [anon_sym_SLASH] = ACTIONS(1363),
    [anon_sym_SLASH_SLASH] = ACTIONS(1363),
    [anon_sym_PERCENT] = ACTIONS(1363),
    [anon_sym_PIPE] = ACTIONS(1363),
    [anon_sym_CARET] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [anon_sym_LT_LT] = ACTIONS(1363),
    [anon_sym_GT_GT] = ACTIONS(1363),
    [anon_sym_GT_GT_GT] = ACTIONS(1363),
    [anon_sym_TILDE] = ACTIONS(1363),
    [anon_sym_typeof] = ACTIONS(1363),
    [anon_sym_void] = ACTIONS(1363),
    [anon_sym_LT] = ACTIONS(1363),
    [anon_sym_LT_EQ] = ACTIONS(1363),
    [anon_sym_EQ_EQ] = ACTIONS(1363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1363),
    [anon_sym_BANG_EQ] = ACTIONS(1363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1363),
    [anon_sym_GT_EQ] = ACTIONS(1363),
    [anon_sym_GT] = ACTIONS(1363),
    [anon_sym_is] = ACTIONS(1363),
    [anon_sym_instanceof] = ACTIONS(1363),
    [anon_sym_new] = ACTIONS(1363),
    [anon_sym_QMARK] = ACTIONS(1363),
    [anon_sym_await] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [sym_this] = ACTIONS(1363),
    [sym_true] = ACTIONS(1363),
    [sym_false] = ACTIONS(1363),
    [sym_none] = ACTIONS(1363),
    [sym_string] = ACTIONS(1363),
    [sym_verbatim] = ACTIONS(1363),
    [sym_number] = ACTIONS(1363),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1361),
    [sym_fstring_start] = ACTIONS(1361),
  },
  [STATE(545)] = {
    [aux_sym__inline_body_repeat1] = STATE(546),
    [ts_builtin_sym_end] = ACTIONS(109),
    [sym_identifier] = ACTIONS(73),
    [anon_sym_SEMI] = ACTIONS(1367),
    [anon_sym_return] = ACTIONS(73),
    [anon_sym_async] = ACTIONS(73),
    [anon_sym_def] = ACTIONS(73),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(73),
    [anon_sym_LBRACK] = ACTIONS(73),
    [anon_sym_del] = ACTIONS(73),
    [anon_sym_raise] = ACTIONS(73),
    [sym_pass_statement] = ACTIONS(73),
    [sym_break_statement] = ACTIONS(73),
    [sym_continue_statement] = ACTIONS(73),
    [sym_debugger_statement] = ACTIONS(73),
    [anon_sym_assert] = ACTIONS(73),
    [anon_sym_global] = ACTIONS(73),
    [anon_sym_nonlocal] = ACTIONS(73),
    [anon_sym_yield] = ACTIONS(73),
    [anon_sym_from] = ACTIONS(73),
    [anon_sym_import] = ACTIONS(73),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_while] = ACTIONS(73),
    [anon_sym_do] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(73),
    [anon_sym_with] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_class] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(73),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(73),
    [anon_sym_DASH] = ACTIONS(73),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(73),
    [anon_sym_typeof] = ACTIONS(73),
    [anon_sym_void] = ACTIONS(73),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(73),
    [anon_sym_LBRACE] = ACTIONS(73),
    [sym_this] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_none] = ACTIONS(73),
    [sym_string] = ACTIONS(73),
    [sym_verbatim] = ACTIONS(73),
    [sym_number] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(109),
    [sym_fstring_start] = ACTIONS(109),
  },
  [STATE(546)] = {
    [aux_sym__inline_body_repeat1] = STATE(546),
    [ts_builtin_sym_end] = ACTIONS(1369),
    [sym_identifier] = ACTIONS(1371),
    [anon_sym_SEMI] = ACTIONS(1373),
    [anon_sym_return] = ACTIONS(1371),
    [anon_sym_async] = ACTIONS(1371),
    [anon_sym_def] = ACTIONS(1371),
    [anon_sym_COMMA] = ACTIONS(1371),
    [anon_sym_LPAREN] = ACTIONS(1371),
    [anon_sym_LBRACK] = ACTIONS(1371),
    [anon_sym_del] = ACTIONS(1371),
    [anon_sym_raise] = ACTIONS(1371),
    [sym_pass_statement] = ACTIONS(1371),
    [sym_break_statement] = ACTIONS(1371),
    [sym_continue_statement] = ACTIONS(1371),
    [sym_debugger_statement] = ACTIONS(1371),
    [anon_sym_assert] = ACTIONS(1371),
    [anon_sym_global] = ACTIONS(1371),
    [anon_sym_nonlocal] = ACTIONS(1371),
    [anon_sym_yield] = ACTIONS(1371),
    [anon_sym_from] = ACTIONS(1371),
    [anon_sym_import] = ACTIONS(1371),
    [anon_sym_DOT] = ACTIONS(1371),
    [anon_sym_if] = ACTIONS(1371),
    [anon_sym_for] = ACTIONS(1371),
    [anon_sym_in] = ACTIONS(1371),
    [anon_sym_while] = ACTIONS(1371),
    [anon_sym_do] = ACTIONS(1371),
    [anon_sym_try] = ACTIONS(1371),
    [anon_sym_with] = ACTIONS(1371),
    [anon_sym_STAR] = ACTIONS(1371),
    [anon_sym_STAR_STAR] = ACTIONS(1371),
    [anon_sym_class] = ACTIONS(1371),
    [anon_sym_AT] = ACTIONS(1371),
    [anon_sym_not] = ACTIONS(1371),
    [anon_sym_and] = ACTIONS(1371),
    [anon_sym_or] = ACTIONS(1371),
    [anon_sym_PLUS] = ACTIONS(1371),
    [anon_sym_DASH] = ACTIONS(1371),
    [anon_sym_SLASH] = ACTIONS(1371),
    [anon_sym_SLASH_SLASH] = ACTIONS(1371),
    [anon_sym_PERCENT] = ACTIONS(1371),
    [anon_sym_PIPE] = ACTIONS(1371),
    [anon_sym_CARET] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(1371),
    [anon_sym_LT_LT] = ACTIONS(1371),
    [anon_sym_GT_GT] = ACTIONS(1371),
    [anon_sym_GT_GT_GT] = ACTIONS(1371),
    [anon_sym_TILDE] = ACTIONS(1371),
    [anon_sym_typeof] = ACTIONS(1371),
    [anon_sym_void] = ACTIONS(1371),
    [anon_sym_LT] = ACTIONS(1371),
    [anon_sym_LT_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1371),
    [anon_sym_GT_EQ] = ACTIONS(1371),
    [anon_sym_GT] = ACTIONS(1371),
    [anon_sym_is] = ACTIONS(1371),
    [anon_sym_instanceof] = ACTIONS(1371),
    [anon_sym_new] = ACTIONS(1371),
    [anon_sym_QMARK] = ACTIONS(1371),
    [anon_sym_await] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1371),
    [sym_this] = ACTIONS(1371),
    [sym_true] = ACTIONS(1371),
    [sym_false] = ACTIONS(1371),
    [sym_none] = ACTIONS(1371),
    [sym_string] = ACTIONS(1371),
    [sym_verbatim] = ACTIONS(1371),
    [sym_number] = ACTIONS(1371),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1369),
    [sym_fstring_start] = ACTIONS(1369),
  },
  [STATE(547)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_async] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_COMMA] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1270),
    [anon_sym_del] = ACTIONS(1270),
    [anon_sym_raise] = ACTIONS(1270),
    [sym_pass_statement] = ACTIONS(1270),
    [sym_break_statement] = ACTIONS(1270),
    [sym_continue_statement] = ACTIONS(1270),
    [sym_debugger_statement] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_global] = ACTIONS(1270),
    [anon_sym_nonlocal] = ACTIONS(1270),
    [anon_sym_yield] = ACTIONS(1270),
    [anon_sym_from] = ACTIONS(1270),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_DOT] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_with] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1270),
    [anon_sym_or] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1270),
    [anon_sym_typeof] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1270),
    [anon_sym_await] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1270),
    [sym_this] = ACTIONS(1270),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [sym_none] = ACTIONS(1270),
    [sym_string] = ACTIONS(1270),
    [sym_verbatim] = ACTIONS(1270),
    [sym_number] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1268),
    [sym_regex] = ACTIONS(1268),
    [sym_fstring_start] = ACTIONS(1268),
  },
  [STATE(548)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(492),
    [sym_identifier] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1359),
    [anon_sym_return] = ACTIONS(1359),
    [anon_sym_async] = ACTIONS(1359),
    [anon_sym_def] = ACTIONS(1359),
    [anon_sym_COMMA] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1359),
    [anon_sym_LBRACK] = ACTIONS(1359),
    [anon_sym_del] = ACTIONS(1359),
    [anon_sym_raise] = ACTIONS(1359),
    [sym_pass_statement] = ACTIONS(1359),
    [sym_break_statement] = ACTIONS(1359),
    [sym_continue_statement] = ACTIONS(1359),
    [sym_debugger_statement] = ACTIONS(1359),
    [anon_sym_assert] = ACTIONS(1359),
    [anon_sym_global] = ACTIONS(1359),
    [anon_sym_nonlocal] = ACTIONS(1359),
    [anon_sym_yield] = ACTIONS(1359),
    [anon_sym_from] = ACTIONS(1359),
    [anon_sym_import] = ACTIONS(1359),
    [anon_sym_DOT] = ACTIONS(1359),
    [anon_sym_if] = ACTIONS(1359),
    [anon_sym_for] = ACTIONS(1359),
    [anon_sym_in] = ACTIONS(1359),
    [anon_sym_while] = ACTIONS(1359),
    [anon_sym_do] = ACTIONS(1359),
    [anon_sym_try] = ACTIONS(1359),
    [anon_sym_with] = ACTIONS(1359),
    [anon_sym_STAR] = ACTIONS(1359),
    [anon_sym_STAR_STAR] = ACTIONS(1359),
    [anon_sym_class] = ACTIONS(1359),
    [anon_sym_AT] = ACTIONS(1359),
    [anon_sym_not] = ACTIONS(1359),
    [anon_sym_and] = ACTIONS(1359),
    [anon_sym_or] = ACTIONS(1359),
    [anon_sym_PLUS] = ACTIONS(1359),
    [anon_sym_DASH] = ACTIONS(1359),
    [anon_sym_SLASH] = ACTIONS(1359),
    [anon_sym_SLASH_SLASH] = ACTIONS(1359),
    [anon_sym_PERCENT] = ACTIONS(1359),
    [anon_sym_PIPE] = ACTIONS(1359),
    [anon_sym_CARET] = ACTIONS(1359),
    [anon_sym_AMP] = ACTIONS(1359),
    [anon_sym_LT_LT] = ACTIONS(1359),
    [anon_sym_GT_GT] = ACTIONS(1359),
    [anon_sym_GT_GT_GT] = ACTIONS(1359),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_typeof] = ACTIONS(1359),
    [anon_sym_void] = ACTIONS(1359),
    [anon_sym_LT] = ACTIONS(1359),
    [anon_sym_LT_EQ] = ACTIONS(1359),
    [anon_sym_EQ_EQ] = ACTIONS(1359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1359),
    [anon_sym_BANG_EQ] = ACTIONS(1359),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1359),
    [anon_sym_GT_EQ] = ACTIONS(1359),
    [anon_sym_GT] = ACTIONS(1359),
    [anon_sym_is] = ACTIONS(1359),
    [anon_sym_instanceof] = ACTIONS(1359),
    [anon_sym_new] = ACTIONS(1359),
    [anon_sym_QMARK] = ACTIONS(1359),
    [anon_sym_await] = ACTIONS(1359),
    [anon_sym_LBRACE] = ACTIONS(1359),
    [sym_this] = ACTIONS(1359),
    [sym_true] = ACTIONS(1359),
    [sym_false] = ACTIONS(1359),
    [sym_none] = ACTIONS(1359),
    [sym_string] = ACTIONS(1359),
    [sym_verbatim] = ACTIONS(1359),
    [sym_number] = ACTIONS(1359),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1357),
    [sym_regex] = ACTIONS(1357),
    [sym_fstring_start] = ACTIONS(1357),
  },
  [STATE(549)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1309),
    [anon_sym_SEMI] = ACTIONS(1309),
    [anon_sym_return] = ACTIONS(1309),
    [anon_sym_async] = ACTIONS(1309),
    [anon_sym_def] = ACTIONS(1309),
    [anon_sym_COMMA] = ACTIONS(1309),
    [anon_sym_LPAREN] = ACTIONS(1309),
    [anon_sym_LBRACK] = ACTIONS(1309),
    [anon_sym_del] = ACTIONS(1309),
    [anon_sym_raise] = ACTIONS(1309),
    [sym_pass_statement] = ACTIONS(1309),
    [sym_break_statement] = ACTIONS(1309),
    [sym_continue_statement] = ACTIONS(1309),
    [sym_debugger_statement] = ACTIONS(1309),
    [anon_sym_assert] = ACTIONS(1309),
    [anon_sym_global] = ACTIONS(1309),
    [anon_sym_nonlocal] = ACTIONS(1309),
    [anon_sym_yield] = ACTIONS(1309),
    [anon_sym_from] = ACTIONS(1309),
    [anon_sym_import] = ACTIONS(1309),
    [anon_sym_DOT] = ACTIONS(1309),
    [anon_sym_if] = ACTIONS(1309),
    [anon_sym_for] = ACTIONS(1309),
    [anon_sym_in] = ACTIONS(1077),
    [anon_sym_while] = ACTIONS(1309),
    [anon_sym_do] = ACTIONS(1309),
    [anon_sym_try] = ACTIONS(1309),
    [anon_sym_with] = ACTIONS(1309),
    [anon_sym_STAR] = ACTIONS(1079),
    [anon_sym_STAR_STAR] = ACTIONS(1081),
    [anon_sym_class] = ACTIONS(1309),
    [anon_sym_AT] = ACTIONS(1079),
    [anon_sym_not] = ACTIONS(1083),
    [anon_sym_and] = ACTIONS(1085),
    [anon_sym_or] = ACTIONS(1087),
    [anon_sym_PLUS] = ACTIONS(1089),
    [anon_sym_DASH] = ACTIONS(1089),
    [anon_sym_SLASH] = ACTIONS(1079),
    [anon_sym_SLASH_SLASH] = ACTIONS(1079),
    [anon_sym_PERCENT] = ACTIONS(1079),
    [anon_sym_PIPE] = ACTIONS(1091),
    [anon_sym_CARET] = ACTIONS(1093),
    [anon_sym_AMP] = ACTIONS(1095),
    [anon_sym_LT_LT] = ACTIONS(1097),
    [anon_sym_GT_GT] = ACTIONS(1097),
    [anon_sym_GT_GT_GT] = ACTIONS(1097),
    [anon_sym_TILDE] = ACTIONS(1309),
    [anon_sym_typeof] = ACTIONS(1309),
    [anon_sym_void] = ACTIONS(1309),
    [anon_sym_LT] = ACTIONS(1077),
    [anon_sym_LT_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ] = ACTIONS(1077),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ] = ACTIONS(1077),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1077),
    [anon_sym_GT_EQ] = ACTIONS(1077),
    [anon_sym_GT] = ACTIONS(1077),
    [anon_sym_is] = ACTIONS(1099),
    [anon_sym_instanceof] = ACTIONS(1077),
    [anon_sym_new] = ACTIONS(1309),
    [anon_sym_QMARK] = ACTIONS(1309),
    [anon_sym_await] = ACTIONS(1309),
    [anon_sym_LBRACE] = ACTIONS(1309),
    [sym_this] = ACTIONS(1309),
    [sym_true] = ACTIONS(1309),
    [sym_false] = ACTIONS(1309),
    [sym_none] = ACTIONS(1309),
    [sym_string] = ACTIONS(1309),
    [sym_verbatim] = ACTIONS(1309),
    [sym_number] = ACTIONS(1309),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1314),
    [sym_regex] = ACTIONS(1314),
    [sym_fstring_start] = ACTIONS(1314),
  },
  [STATE(550)] = {
    [aux_sym__inline_body_repeat1] = STATE(551),
    [sym_identifier] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1376),
    [anon_sym_return] = ACTIONS(1363),
    [anon_sym_async] = ACTIONS(1363),
    [anon_sym_def] = ACTIONS(1363),
    [anon_sym_COMMA] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [anon_sym_del] = ACTIONS(1363),
    [anon_sym_raise] = ACTIONS(1363),
    [sym_pass_statement] = ACTIONS(1363),
    [sym_break_statement] = ACTIONS(1363),
    [sym_continue_statement] = ACTIONS(1363),
    [sym_debugger_statement] = ACTIONS(1363),
    [anon_sym_assert] = ACTIONS(1363),
    [anon_sym_global] = ACTIONS(1363),
    [anon_sym_nonlocal] = ACTIONS(1363),
    [anon_sym_yield] = ACTIONS(1363),
    [anon_sym_from] = ACTIONS(1363),
    [anon_sym_import] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_if] = ACTIONS(1363),
    [anon_sym_for] = ACTIONS(1363),
    [anon_sym_in] = ACTIONS(1363),
    [anon_sym_while] = ACTIONS(1363),
    [anon_sym_do] = ACTIONS(1363),
    [anon_sym_try] = ACTIONS(1363),
    [anon_sym_with] = ACTIONS(1363),
    [anon_sym_STAR] = ACTIONS(1363),
    [anon_sym_STAR_STAR] = ACTIONS(1363),
    [anon_sym_class] = ACTIONS(1363),
    [anon_sym_AT] = ACTIONS(1363),
    [anon_sym_not] = ACTIONS(1363),
    [anon_sym_and] = ACTIONS(1363),
    [anon_sym_or] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [anon_sym_SLASH] = ACTIONS(1363),
    [anon_sym_SLASH_SLASH] = ACTIONS(1363),
    [anon_sym_PERCENT] = ACTIONS(1363),
    [anon_sym_PIPE] = ACTIONS(1363),
    [anon_sym_CARET] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [anon_sym_LT_LT] = ACTIONS(1363),
    [anon_sym_GT_GT] = ACTIONS(1363),
    [anon_sym_GT_GT_GT] = ACTIONS(1363),
    [anon_sym_TILDE] = ACTIONS(1363),
    [anon_sym_typeof] = ACTIONS(1363),
    [anon_sym_void] = ACTIONS(1363),
    [anon_sym_LT] = ACTIONS(1363),
    [anon_sym_LT_EQ] = ACTIONS(1363),
    [anon_sym_EQ_EQ] = ACTIONS(1363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1363),
    [anon_sym_BANG_EQ] = ACTIONS(1363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1363),
    [anon_sym_GT_EQ] = ACTIONS(1363),
    [anon_sym_GT] = ACTIONS(1363),
    [anon_sym_is] = ACTIONS(1363),
    [anon_sym_instanceof] = ACTIONS(1363),
    [anon_sym_new] = ACTIONS(1363),
    [anon_sym_QMARK] = ACTIONS(1363),
    [anon_sym_await] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [sym_this] = ACTIONS(1363),
    [sym_true] = ACTIONS(1363),
    [sym_false] = ACTIONS(1363),
    [sym_none] = ACTIONS(1363),
    [sym_string] = ACTIONS(1363),
    [sym_verbatim] = ACTIONS(1363),
    [sym_number] = ACTIONS(1363),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1361),
    [sym_regex] = ACTIONS(1361),
    [sym_fstring_start] = ACTIONS(1361),
  },
  [STATE(551)] = {
    [aux_sym__inline_body_repeat1] = STATE(552),
    [sym_identifier] = ACTIONS(73),
    [anon_sym_SEMI] = ACTIONS(1378),
    [anon_sym_return] = ACTIONS(73),
    [anon_sym_async] = ACTIONS(73),
    [anon_sym_def] = ACTIONS(73),
    [anon_sym_COMMA] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(73),
    [anon_sym_LBRACK] = ACTIONS(73),
    [anon_sym_del] = ACTIONS(73),
    [anon_sym_raise] = ACTIONS(73),
    [sym_pass_statement] = ACTIONS(73),
    [sym_break_statement] = ACTIONS(73),
    [sym_continue_statement] = ACTIONS(73),
    [sym_debugger_statement] = ACTIONS(73),
    [anon_sym_assert] = ACTIONS(73),
    [anon_sym_global] = ACTIONS(73),
    [anon_sym_nonlocal] = ACTIONS(73),
    [anon_sym_yield] = ACTIONS(73),
    [anon_sym_from] = ACTIONS(73),
    [anon_sym_import] = ACTIONS(73),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_while] = ACTIONS(73),
    [anon_sym_do] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(73),
    [anon_sym_with] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_class] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(73),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(73),
    [anon_sym_DASH] = ACTIONS(73),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(73),
    [anon_sym_typeof] = ACTIONS(73),
    [anon_sym_void] = ACTIONS(73),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(73),
    [anon_sym_LBRACE] = ACTIONS(73),
    [sym_this] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_none] = ACTIONS(73),
    [sym_string] = ACTIONS(73),
    [sym_verbatim] = ACTIONS(73),
    [sym_number] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(109),
    [sym_regex] = ACTIONS(109),
    [sym_fstring_start] = ACTIONS(109),
  },
  [STATE(552)] = {
    [aux_sym__inline_body_repeat1] = STATE(552),
    [sym_identifier] = ACTIONS(1371),
    [anon_sym_SEMI] = ACTIONS(1380),
    [anon_sym_return] = ACTIONS(1371),
    [anon_sym_async] = ACTIONS(1371),
    [anon_sym_def] = ACTIONS(1371),
    [anon_sym_COMMA] = ACTIONS(1371),
    [anon_sym_LPAREN] = ACTIONS(1371),
    [anon_sym_LBRACK] = ACTIONS(1371),
    [anon_sym_del] = ACTIONS(1371),
    [anon_sym_raise] = ACTIONS(1371),
    [sym_pass_statement] = ACTIONS(1371),
    [sym_break_statement] = ACTIONS(1371),
    [sym_continue_statement] = ACTIONS(1371),
    [sym_debugger_statement] = ACTIONS(1371),
    [anon_sym_assert] = ACTIONS(1371),
    [anon_sym_global] = ACTIONS(1371),
    [anon_sym_nonlocal] = ACTIONS(1371),
    [anon_sym_yield] = ACTIONS(1371),
    [anon_sym_from] = ACTIONS(1371),
    [anon_sym_import] = ACTIONS(1371),
    [anon_sym_DOT] = ACTIONS(1371),
    [anon_sym_if] = ACTIONS(1371),
    [anon_sym_for] = ACTIONS(1371),
    [anon_sym_in] = ACTIONS(1371),
    [anon_sym_while] = ACTIONS(1371),
    [anon_sym_do] = ACTIONS(1371),
    [anon_sym_try] = ACTIONS(1371),
    [anon_sym_with] = ACTIONS(1371),
    [anon_sym_STAR] = ACTIONS(1371),
    [anon_sym_STAR_STAR] = ACTIONS(1371),
    [anon_sym_class] = ACTIONS(1371),
    [anon_sym_AT] = ACTIONS(1371),
    [anon_sym_not] = ACTIONS(1371),
    [anon_sym_and] = ACTIONS(1371),
    [anon_sym_or] = ACTIONS(1371),
    [anon_sym_PLUS] = ACTIONS(1371),
    [anon_sym_DASH] = ACTIONS(1371),
    [anon_sym_SLASH] = ACTIONS(1371),
    [anon_sym_SLASH_SLASH] = ACTIONS(1371),
    [anon_sym_PERCENT] = ACTIONS(1371),
    [anon_sym_PIPE] = ACTIONS(1371),
    [anon_sym_CARET] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(1371),
    [anon_sym_LT_LT] = ACTIONS(1371),
    [anon_sym_GT_GT] = ACTIONS(1371),
    [anon_sym_GT_GT_GT] = ACTIONS(1371),
    [anon_sym_TILDE] = ACTIONS(1371),
    [anon_sym_typeof] = ACTIONS(1371),
    [anon_sym_void] = ACTIONS(1371),
    [anon_sym_LT] = ACTIONS(1371),
    [anon_sym_LT_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1371),
    [anon_sym_GT_EQ] = ACTIONS(1371),
    [anon_sym_GT] = ACTIONS(1371),
    [anon_sym_is] = ACTIONS(1371),
    [anon_sym_instanceof] = ACTIONS(1371),
    [anon_sym_new] = ACTIONS(1371),
    [anon_sym_QMARK] = ACTIONS(1371),
    [anon_sym_await] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1371),
    [sym_this] = ACTIONS(1371),
    [sym_true] = ACTIONS(1371),
    [sym_false] = ACTIONS(1371),
    [sym_none] = ACTIONS(1371),
    [sym_string] = ACTIONS(1371),
    [sym_verbatim] = ACTIONS(1371),
    [sym_number] = ACTIONS(1371),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1369),
    [sym_regex] = ACTIONS(1369),
    [sym_fstring_start] = ACTIONS(1369),
  },
  [STATE(553)] = {
    [sym__expression_statement] = STATE(1700),
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2544),
    [sym_expression_list] = STATE(1700),
    [sym_pattern_list] = STATE(2544),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym_return_statement] = STATE(1700),
    [sym_delete_statement] = STATE(1700),
    [sym_raise_statement] = STATE(1700),
    [sym_assert_statement] = STATE(1700),
    [sym_yield_statement] = STATE(1700),
    [sym_yield] = STATE(1661),
    [sym__expression] = STATE(2076),
    [sym__primary_expression] = STATE(1789),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym__inline_simple_statement] = STATE(1700),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_return] = ACTIONS(1038),
    [anon_sym_async] = ACTIONS(479),
    [anon_sym_def] = ACTIONS(481),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_del] = ACTIONS(1040),
    [anon_sym_raise] = ACTIONS(1042),
    [sym_pass_statement] = ACTIONS(401),
    [sym_break_statement] = ACTIONS(401),
    [sym_continue_statement] = ACTIONS(401),
    [anon_sym_assert] = ACTIONS(1046),
    [anon_sym_yield] = ACTIONS(1048),
    [anon_sym_not] = ACTIONS(495),
    [anon_sym_PLUS] = ACTIONS(497),
    [anon_sym_DASH] = ACTIONS(497),
    [anon_sym_TILDE] = ACTIONS(497),
    [anon_sym_typeof] = ACTIONS(497),
    [anon_sym_void] = ACTIONS(497),
    [anon_sym_new] = ACTIONS(499),
    [anon_sym_await] = ACTIONS(501),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(554)] = {
    [ts_builtin_sym_end] = ACTIONS(1323),
    [sym_identifier] = ACTIONS(1318),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_return] = ACTIONS(1318),
    [anon_sym_async] = ACTIONS(1318),
    [anon_sym_def] = ACTIONS(1318),
    [anon_sym_COMMA] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1320),
    [anon_sym_LBRACK] = ACTIONS(1320),
    [anon_sym_del] = ACTIONS(1318),
    [anon_sym_raise] = ACTIONS(1318),
    [sym_pass_statement] = ACTIONS(1318),
    [sym_break_statement] = ACTIONS(1318),
    [sym_continue_statement] = ACTIONS(1318),
    [sym_debugger_statement] = ACTIONS(1318),
    [anon_sym_assert] = ACTIONS(1318),
    [anon_sym_global] = ACTIONS(1318),
    [anon_sym_nonlocal] = ACTIONS(1318),
    [anon_sym_yield] = ACTIONS(1318),
    [anon_sym_from] = ACTIONS(1318),
    [anon_sym_import] = ACTIONS(1318),
    [anon_sym_DOT] = ACTIONS(1232),
    [anon_sym_if] = ACTIONS(1320),
    [anon_sym_for] = ACTIONS(1318),
    [anon_sym_in] = ACTIONS(1232),
    [anon_sym_while] = ACTIONS(1318),
    [anon_sym_do] = ACTIONS(1318),
    [anon_sym_try] = ACTIONS(1318),
    [anon_sym_with] = ACTIONS(1318),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_STAR_STAR] = ACTIONS(1232),
    [anon_sym_class] = ACTIONS(1318),
    [anon_sym_AT] = ACTIONS(1320),
    [anon_sym_not] = ACTIONS(1320),
    [anon_sym_and] = ACTIONS(1232),
    [anon_sym_or] = ACTIONS(1232),
    [anon_sym_PLUS] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_SLASH] = ACTIONS(1232),
    [anon_sym_SLASH_SLASH] = ACTIONS(1232),
    [anon_sym_PERCENT] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_CARET] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_LT_LT] = ACTIONS(1232),
    [anon_sym_GT_GT] = ACTIONS(1232),
    [anon_sym_GT_GT_GT] = ACTIONS(1232),
    [anon_sym_TILDE] = ACTIONS(1318),
    [anon_sym_typeof] = ACTIONS(1318),
    [anon_sym_void] = ACTIONS(1318),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_LT_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ] = ACTIONS(1232),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ] = ACTIONS(1232),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1232),
    [anon_sym_GT_EQ] = ACTIONS(1232),
    [anon_sym_GT] = ACTIONS(1232),
    [anon_sym_is] = ACTIONS(1232),
    [anon_sym_instanceof] = ACTIONS(1232),
    [anon_sym_new] = ACTIONS(1318),
    [anon_sym_QMARK] = ACTIONS(1232),
    [anon_sym_await] = ACTIONS(1318),
    [anon_sym_LBRACE] = ACTIONS(1318),
    [sym_this] = ACTIONS(1318),
    [sym_true] = ACTIONS(1318),
    [sym_false] = ACTIONS(1318),
    [sym_none] = ACTIONS(1318),
    [sym_string] = ACTIONS(1318),
    [sym_verbatim] = ACTIONS(1318),
    [sym_number] = ACTIONS(1318),
    [sym_comment] = ACTIONS(3),
    [sym__newline] = ACTIONS(1237),
    [sym_regex] = ACTIONS(1323),
    [sym_fstring_start] = ACTIONS(1323),
  },
  [STATE(555)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1383),
    [sym_identifier] = ACTIONS(1385),
    [anon_sym_SEMI] = ACTIONS(1385),
    [anon_sym_return] = ACTIONS(1385),
    [anon_sym_async] = ACTIONS(1385),
    [anon_sym_def] = ACTIONS(1385),
    [anon_sym_COMMA] = ACTIONS(1385),
    [anon_sym_LPAREN] = ACTIONS(1385),
    [anon_sym_LBRACK] = ACTIONS(1385),
    [anon_sym_del] = ACTIONS(1385),
    [anon_sym_raise] = ACTIONS(1385),
    [sym_pass_statement] = ACTIONS(1385),
    [sym_break_statement] = ACTIONS(1385),
    [sym_continue_statement] = ACTIONS(1385),
    [sym_debugger_statement] = ACTIONS(1385),
    [anon_sym_assert] = ACTIONS(1385),
    [anon_sym_global] = ACTIONS(1385),
    [anon_sym_nonlocal] = ACTIONS(1385),
    [anon_sym_yield] = ACTIONS(1385),
    [anon_sym_from] = ACTIONS(1385),
    [anon_sym_import] = ACTIONS(1385),
    [anon_sym_DOT] = ACTIONS(1385),
    [anon_sym_if] = ACTIONS(1385),
    [anon_sym_for] = ACTIONS(1385),
    [anon_sym_in] = ACTIONS(1385),
    [anon_sym_while] = ACTIONS(1385),
    [anon_sym_do] = ACTIONS(1385),
    [anon_sym_try] = ACTIONS(1385),
    [anon_sym_with] = ACTIONS(1385),
    [anon_sym_STAR] = ACTIONS(1385),
    [anon_sym_STAR_STAR] = ACTIONS(1385),
    [anon_sym_class] = ACTIONS(1385),
    [anon_sym_AT] = ACTIONS(1385),
    [anon_sym_not] = ACTIONS(1385),
    [anon_sym_and] = ACTIONS(1385),
    [anon_sym_or] = ACTIONS(1385),
    [anon_sym_PLUS] = ACTIONS(1385),
    [anon_sym_DASH] = ACTIONS(1385),
    [anon_sym_SLASH] = ACTIONS(1385),
    [anon_sym_SLASH_SLASH] = ACTIONS(1385),
    [anon_sym_PERCENT] = ACTIONS(1385),
    [anon_sym_PIPE] = ACTIONS(1385),
    [anon_sym_CARET] = ACTIONS(1385),
    [anon_sym_AMP] = ACTIONS(1385),
    [anon_sym_LT_LT] = ACTIONS(1385),
    [anon_sym_GT_GT] = ACTIONS(1385),
    [anon_sym_GT_GT_GT] = ACTIONS(1385),
    [anon_sym_TILDE] = ACTIONS(1385),
    [anon_sym_typeof] = ACTIONS(1385),
    [anon_sym_void] = ACTIONS(1385),
    [anon_sym_LT] = ACTIONS(1385),
    [anon_sym_LT_EQ] = ACTIONS(1385),
    [anon_sym_EQ_EQ] = ACTIONS(1385),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1385),
    [anon_sym_BANG_EQ] = ACTIONS(1385),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1385),
    [anon_sym_GT_EQ] = ACTIONS(1385),
    [anon_sym_GT] = ACTIONS(1385),
    [anon_sym_is] = ACTIONS(1385),
    [anon_sym_instanceof] = ACTIONS(1385),
    [anon_sym_new] = ACTIONS(1385),
    [anon_sym_QMARK] = ACTIONS(1385),
    [anon_sym_await] = ACTIONS(1385),
    [anon_sym_LBRACE] = ACTIONS(1385),
    [sym_this] = ACTIONS(1385),
    [sym_true] = ACTIONS(1385),
    [sym_false] = ACTIONS(1385),
    [sym_none] = ACTIONS(1385),
    [sym_string] = ACTIONS(1385),
    [sym_verbatim] = ACTIONS(1385),
    [sym_number] = ACTIONS(1385),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1383),
    [sym_fstring_start] = ACTIONS(1383),
  },
  [STATE(556)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1196),
    [sym_identifier] = ACTIONS(1194),
    [anon_sym_SEMI] = ACTIONS(1194),
    [anon_sym_return] = ACTIONS(1194),
    [anon_sym_async] = ACTIONS(1194),
    [anon_sym_def] = ACTIONS(1194),
    [anon_sym_COMMA] = ACTIONS(1194),
    [anon_sym_LPAREN] = ACTIONS(1194),
    [anon_sym_LBRACK] = ACTIONS(1194),
    [anon_sym_del] = ACTIONS(1194),
    [anon_sym_raise] = ACTIONS(1194),
    [sym_pass_statement] = ACTIONS(1194),
    [sym_break_statement] = ACTIONS(1194),
    [sym_continue_statement] = ACTIONS(1194),
    [sym_debugger_statement] = ACTIONS(1194),
    [anon_sym_assert] = ACTIONS(1194),
    [anon_sym_global] = ACTIONS(1194),
    [anon_sym_nonlocal] = ACTIONS(1194),
    [anon_sym_yield] = ACTIONS(1194),
    [anon_sym_from] = ACTIONS(1194),
    [anon_sym_import] = ACTIONS(1194),
    [anon_sym_DOT] = ACTIONS(1194),
    [anon_sym_if] = ACTIONS(1194),
    [anon_sym_for] = ACTIONS(1194),
    [anon_sym_in] = ACTIONS(1194),
    [anon_sym_while] = ACTIONS(1194),
    [anon_sym_do] = ACTIONS(1194),
    [anon_sym_try] = ACTIONS(1194),
    [anon_sym_with] = ACTIONS(1194),
    [anon_sym_STAR] = ACTIONS(1194),
    [anon_sym_STAR_STAR] = ACTIONS(1194),
    [anon_sym_class] = ACTIONS(1194),
    [anon_sym_AT] = ACTIONS(1194),
    [anon_sym_not] = ACTIONS(1194),
    [anon_sym_and] = ACTIONS(1194),
    [anon_sym_or] = ACTIONS(1194),
    [anon_sym_PLUS] = ACTIONS(1194),
    [anon_sym_DASH] = ACTIONS(1194),
    [anon_sym_SLASH] = ACTIONS(1194),
    [anon_sym_SLASH_SLASH] = ACTIONS(1194),
    [anon_sym_PERCENT] = ACTIONS(1194),
    [anon_sym_PIPE] = ACTIONS(1194),
    [anon_sym_CARET] = ACTIONS(1194),
    [anon_sym_AMP] = ACTIONS(1194),
    [anon_sym_LT_LT] = ACTIONS(1194),
    [anon_sym_GT_GT] = ACTIONS(1194),
    [anon_sym_GT_GT_GT] = ACTIONS(1194),
    [anon_sym_TILDE] = ACTIONS(1194),
    [anon_sym_typeof] = ACTIONS(1194),
    [anon_sym_void] = ACTIONS(1194),
    [anon_sym_LT] = ACTIONS(1194),
    [anon_sym_LT_EQ] = ACTIONS(1194),
    [anon_sym_EQ_EQ] = ACTIONS(1194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1194),
    [anon_sym_BANG_EQ] = ACTIONS(1194),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1194),
    [anon_sym_GT_EQ] = ACTIONS(1194),
    [anon_sym_GT] = ACTIONS(1194),
    [anon_sym_is] = ACTIONS(1194),
    [anon_sym_instanceof] = ACTIONS(1194),
    [anon_sym_new] = ACTIONS(1194),
    [anon_sym_QMARK] = ACTIONS(1194),
    [anon_sym_await] = ACTIONS(1194),
    [anon_sym_LBRACE] = ACTIONS(1194),
    [sym_this] = ACTIONS(1194),
    [sym_true] = ACTIONS(1194),
    [sym_false] = ACTIONS(1194),
    [sym_none] = ACTIONS(1194),
    [sym_string] = ACTIONS(1194),
    [sym_verbatim] = ACTIONS(1194),
    [sym_number] = ACTIONS(1194),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1196),
    [sym_fstring_start] = ACTIONS(1196),
  },
  [STATE(557)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(548),
    [sym_identifier] = ACTIONS(1385),
    [anon_sym_SEMI] = ACTIONS(1385),
    [anon_sym_return] = ACTIONS(1385),
    [anon_sym_async] = ACTIONS(1385),
    [anon_sym_def] = ACTIONS(1385),
    [anon_sym_COMMA] = ACTIONS(1385),
    [anon_sym_LPAREN] = ACTIONS(1385),
    [anon_sym_LBRACK] = ACTIONS(1385),
    [anon_sym_del] = ACTIONS(1385),
    [anon_sym_raise] = ACTIONS(1385),
    [sym_pass_statement] = ACTIONS(1385),
    [sym_break_statement] = ACTIONS(1385),
    [sym_continue_statement] = ACTIONS(1385),
    [sym_debugger_statement] = ACTIONS(1385),
    [anon_sym_assert] = ACTIONS(1385),
    [anon_sym_global] = ACTIONS(1385),
    [anon_sym_nonlocal] = ACTIONS(1385),
    [anon_sym_yield] = ACTIONS(1385),
    [anon_sym_from] = ACTIONS(1385),
    [anon_sym_import] = ACTIONS(1385),
    [anon_sym_DOT] = ACTIONS(1385),
    [anon_sym_if] = ACTIONS(1385),
    [anon_sym_for] = ACTIONS(1385),
    [anon_sym_in] = ACTIONS(1385),
    [anon_sym_while] = ACTIONS(1385),
    [anon_sym_do] = ACTIONS(1385),
    [anon_sym_try] = ACTIONS(1385),
    [anon_sym_with] = ACTIONS(1385),
    [anon_sym_STAR] = ACTIONS(1385),
    [anon_sym_STAR_STAR] = ACTIONS(1385),
    [anon_sym_class] = ACTIONS(1385),
    [anon_sym_AT] = ACTIONS(1385),
    [anon_sym_not] = ACTIONS(1385),
    [anon_sym_and] = ACTIONS(1385),
    [anon_sym_or] = ACTIONS(1385),
    [anon_sym_PLUS] = ACTIONS(1385),
    [anon_sym_DASH] = ACTIONS(1385),
    [anon_sym_SLASH] = ACTIONS(1385),
    [anon_sym_SLASH_SLASH] = ACTIONS(1385),
    [anon_sym_PERCENT] = ACTIONS(1385),
    [anon_sym_PIPE] = ACTIONS(1385),
    [anon_sym_CARET] = ACTIONS(1385),
    [anon_sym_AMP] = ACTIONS(1385),
    [anon_sym_LT_LT] = ACTIONS(1385),
    [anon_sym_GT_GT] = ACTIONS(1385),
    [anon_sym_GT_GT_GT] = ACTIONS(1385),
    [anon_sym_TILDE] = ACTIONS(1385),
    [anon_sym_typeof] = ACTIONS(1385),
    [anon_sym_void] = ACTIONS(1385),
    [anon_sym_LT] = ACTIONS(1385),
    [anon_sym_LT_EQ] = ACTIONS(1385),
    [anon_sym_EQ_EQ] = ACTIONS(1385),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1385),
    [anon_sym_BANG_EQ] = ACTIONS(1385),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1385),
    [anon_sym_GT_EQ] = ACTIONS(1385),
    [anon_sym_GT] = ACTIONS(1385),
    [anon_sym_is] = ACTIONS(1385),
    [anon_sym_instanceof] = ACTIONS(1385),
    [anon_sym_new] = ACTIONS(1385),
    [anon_sym_QMARK] = ACTIONS(1385),
    [anon_sym_await] = ACTIONS(1385),
    [anon_sym_LBRACE] = ACTIONS(1385),
    [sym_this] = ACTIONS(1385),
    [sym_true] = ACTIONS(1385),
    [sym_false] = ACTIONS(1385),
    [sym_none] = ACTIONS(1385),
    [sym_string] = ACTIONS(1385),
    [sym_verbatim] = ACTIONS(1385),
    [sym_number] = ACTIONS(1385),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1383),
    [sym_regex] = ACTIONS(1383),
    [sym_fstring_start] = ACTIONS(1383),
  },
  [STATE(558)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(543),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_COMMA] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_STAR_STAR] = ACTIONS(1113),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1228),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1228),
    [anon_sym_SLASH_SLASH] = ACTIONS(1228),
    [anon_sym_PERCENT] = ACTIONS(1228),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(559)] = {
    [ts_builtin_sym_end] = ACTIONS(1387),
    [sym_identifier] = ACTIONS(1389),
    [anon_sym_SEMI] = ACTIONS(1389),
    [anon_sym_return] = ACTIONS(1389),
    [anon_sym_async] = ACTIONS(1389),
    [anon_sym_def] = ACTIONS(1389),
    [anon_sym_COMMA] = ACTIONS(1389),
    [anon_sym_LPAREN] = ACTIONS(1389),
    [anon_sym_LBRACK] = ACTIONS(1389),
    [anon_sym_del] = ACTIONS(1389),
    [anon_sym_raise] = ACTIONS(1389),
    [sym_pass_statement] = ACTIONS(1389),
    [sym_break_statement] = ACTIONS(1389),
    [sym_continue_statement] = ACTIONS(1389),
    [sym_debugger_statement] = ACTIONS(1389),
    [anon_sym_assert] = ACTIONS(1389),
    [anon_sym_global] = ACTIONS(1389),
    [anon_sym_nonlocal] = ACTIONS(1389),
    [anon_sym_yield] = ACTIONS(1389),
    [anon_sym_from] = ACTIONS(1389),
    [anon_sym_import] = ACTIONS(1389),
    [anon_sym_DOT] = ACTIONS(1389),
    [anon_sym_if] = ACTIONS(1389),
    [anon_sym_for] = ACTIONS(1389),
    [anon_sym_in] = ACTIONS(1389),
    [anon_sym_while] = ACTIONS(1389),
    [anon_sym_do] = ACTIONS(1389),
    [anon_sym_try] = ACTIONS(1389),
    [anon_sym_with] = ACTIONS(1389),
    [anon_sym_STAR] = ACTIONS(1389),
    [anon_sym_STAR_STAR] = ACTIONS(1389),
    [anon_sym_class] = ACTIONS(1389),
    [anon_sym_AT] = ACTIONS(1389),
    [anon_sym_not] = ACTIONS(1389),
    [anon_sym_and] = ACTIONS(1389),
    [anon_sym_or] = ACTIONS(1389),
    [anon_sym_PLUS] = ACTIONS(1389),
    [anon_sym_DASH] = ACTIONS(1389),
    [anon_sym_SLASH] = ACTIONS(1389),
    [anon_sym_SLASH_SLASH] = ACTIONS(1389),
    [anon_sym_PERCENT] = ACTIONS(1389),
    [anon_sym_PIPE] = ACTIONS(1389),
    [anon_sym_CARET] = ACTIONS(1389),
    [anon_sym_AMP] = ACTIONS(1389),
    [anon_sym_LT_LT] = ACTIONS(1389),
    [anon_sym_GT_GT] = ACTIONS(1389),
    [anon_sym_GT_GT_GT] = ACTIONS(1389),
    [anon_sym_TILDE] = ACTIONS(1389),
    [anon_sym_typeof] = ACTIONS(1389),
    [anon_sym_void] = ACTIONS(1389),
    [anon_sym_LT] = ACTIONS(1389),
    [anon_sym_LT_EQ] = ACTIONS(1389),
    [anon_sym_EQ_EQ] = ACTIONS(1389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1389),
    [anon_sym_BANG_EQ] = ACTIONS(1389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1389),
    [anon_sym_GT_EQ] = ACTIONS(1389),
    [anon_sym_GT] = ACTIONS(1389),
    [anon_sym_is] = ACTIONS(1389),
    [anon_sym_instanceof] = ACTIONS(1389),
    [anon_sym_new] = ACTIONS(1389),
    [anon_sym_QMARK] = ACTIONS(1389),
    [anon_sym_await] = ACTIONS(1389),
    [anon_sym_LBRACE] = ACTIONS(1389),
    [sym_this] = ACTIONS(1389),
    [sym_true] = ACTIONS(1389),
    [sym_false] = ACTIONS(1389),
    [sym_none] = ACTIONS(1389),
    [sym_string] = ACTIONS(1389),
    [sym_verbatim] = ACTIONS(1389),
    [sym_number] = ACTIONS(1389),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1387),
    [sym_fstring_start] = ACTIONS(1387),
  },
  [STATE(560)] = {
    [ts_builtin_sym_end] = ACTIONS(1391),
    [sym_identifier] = ACTIONS(1393),
    [anon_sym_SEMI] = ACTIONS(1393),
    [anon_sym_return] = ACTIONS(1393),
    [anon_sym_async] = ACTIONS(1393),
    [anon_sym_def] = ACTIONS(1393),
    [anon_sym_COMMA] = ACTIONS(1393),
    [anon_sym_LPAREN] = ACTIONS(1393),
    [anon_sym_LBRACK] = ACTIONS(1393),
    [anon_sym_del] = ACTIONS(1393),
    [anon_sym_raise] = ACTIONS(1393),
    [sym_pass_statement] = ACTIONS(1393),
    [sym_break_statement] = ACTIONS(1393),
    [sym_continue_statement] = ACTIONS(1393),
    [sym_debugger_statement] = ACTIONS(1393),
    [anon_sym_assert] = ACTIONS(1393),
    [anon_sym_global] = ACTIONS(1393),
    [anon_sym_nonlocal] = ACTIONS(1393),
    [anon_sym_yield] = ACTIONS(1393),
    [anon_sym_from] = ACTIONS(1393),
    [anon_sym_import] = ACTIONS(1393),
    [anon_sym_DOT] = ACTIONS(1393),
    [anon_sym_if] = ACTIONS(1393),
    [anon_sym_for] = ACTIONS(1393),
    [anon_sym_in] = ACTIONS(1393),
    [anon_sym_while] = ACTIONS(1393),
    [anon_sym_do] = ACTIONS(1393),
    [anon_sym_try] = ACTIONS(1393),
    [anon_sym_with] = ACTIONS(1393),
    [anon_sym_STAR] = ACTIONS(1393),
    [anon_sym_STAR_STAR] = ACTIONS(1393),
    [anon_sym_class] = ACTIONS(1393),
    [anon_sym_AT] = ACTIONS(1393),
    [anon_sym_not] = ACTIONS(1393),
    [anon_sym_and] = ACTIONS(1393),
    [anon_sym_or] = ACTIONS(1393),
    [anon_sym_PLUS] = ACTIONS(1393),
    [anon_sym_DASH] = ACTIONS(1393),
    [anon_sym_SLASH] = ACTIONS(1393),
    [anon_sym_SLASH_SLASH] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1393),
    [anon_sym_PIPE] = ACTIONS(1393),
    [anon_sym_CARET] = ACTIONS(1393),
    [anon_sym_AMP] = ACTIONS(1393),
    [anon_sym_LT_LT] = ACTIONS(1393),
    [anon_sym_GT_GT] = ACTIONS(1393),
    [anon_sym_GT_GT_GT] = ACTIONS(1393),
    [anon_sym_TILDE] = ACTIONS(1393),
    [anon_sym_typeof] = ACTIONS(1393),
    [anon_sym_void] = ACTIONS(1393),
    [anon_sym_LT] = ACTIONS(1393),
    [anon_sym_LT_EQ] = ACTIONS(1393),
    [anon_sym_EQ_EQ] = ACTIONS(1393),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1393),
    [anon_sym_BANG_EQ] = ACTIONS(1393),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1393),
    [anon_sym_GT_EQ] = ACTIONS(1393),
    [anon_sym_GT] = ACTIONS(1393),
    [anon_sym_is] = ACTIONS(1393),
    [anon_sym_instanceof] = ACTIONS(1393),
    [anon_sym_new] = ACTIONS(1393),
    [anon_sym_QMARK] = ACTIONS(1393),
    [anon_sym_await] = ACTIONS(1393),
    [anon_sym_LBRACE] = ACTIONS(1393),
    [sym_this] = ACTIONS(1393),
    [sym_true] = ACTIONS(1393),
    [sym_false] = ACTIONS(1393),
    [sym_none] = ACTIONS(1393),
    [sym_string] = ACTIONS(1393),
    [sym_verbatim] = ACTIONS(1393),
    [sym_number] = ACTIONS(1393),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1391),
    [sym_fstring_start] = ACTIONS(1391),
  },
  [STATE(561)] = {
    [ts_builtin_sym_end] = ACTIONS(1395),
    [sym_identifier] = ACTIONS(1397),
    [anon_sym_SEMI] = ACTIONS(1397),
    [anon_sym_return] = ACTIONS(1397),
    [anon_sym_async] = ACTIONS(1397),
    [anon_sym_def] = ACTIONS(1397),
    [anon_sym_COMMA] = ACTIONS(1397),
    [anon_sym_LPAREN] = ACTIONS(1397),
    [anon_sym_LBRACK] = ACTIONS(1397),
    [anon_sym_del] = ACTIONS(1397),
    [anon_sym_raise] = ACTIONS(1397),
    [sym_pass_statement] = ACTIONS(1397),
    [sym_break_statement] = ACTIONS(1397),
    [sym_continue_statement] = ACTIONS(1397),
    [sym_debugger_statement] = ACTIONS(1397),
    [anon_sym_assert] = ACTIONS(1397),
    [anon_sym_global] = ACTIONS(1397),
    [anon_sym_nonlocal] = ACTIONS(1397),
    [anon_sym_yield] = ACTIONS(1397),
    [anon_sym_from] = ACTIONS(1397),
    [anon_sym_import] = ACTIONS(1397),
    [anon_sym_DOT] = ACTIONS(1397),
    [anon_sym_if] = ACTIONS(1397),
    [anon_sym_for] = ACTIONS(1397),
    [anon_sym_in] = ACTIONS(1397),
    [anon_sym_while] = ACTIONS(1397),
    [anon_sym_do] = ACTIONS(1397),
    [anon_sym_try] = ACTIONS(1397),
    [anon_sym_with] = ACTIONS(1397),
    [anon_sym_STAR] = ACTIONS(1397),
    [anon_sym_STAR_STAR] = ACTIONS(1397),
    [anon_sym_class] = ACTIONS(1397),
    [anon_sym_AT] = ACTIONS(1397),
    [anon_sym_not] = ACTIONS(1397),
    [anon_sym_and] = ACTIONS(1397),
    [anon_sym_or] = ACTIONS(1397),
    [anon_sym_PLUS] = ACTIONS(1397),
    [anon_sym_DASH] = ACTIONS(1397),
    [anon_sym_SLASH] = ACTIONS(1397),
    [anon_sym_SLASH_SLASH] = ACTIONS(1397),
    [anon_sym_PERCENT] = ACTIONS(1397),
    [anon_sym_PIPE] = ACTIONS(1397),
    [anon_sym_CARET] = ACTIONS(1397),
    [anon_sym_AMP] = ACTIONS(1397),
    [anon_sym_LT_LT] = ACTIONS(1397),
    [anon_sym_GT_GT] = ACTIONS(1397),
    [anon_sym_GT_GT_GT] = ACTIONS(1397),
    [anon_sym_TILDE] = ACTIONS(1397),
    [anon_sym_typeof] = ACTIONS(1397),
    [anon_sym_void] = ACTIONS(1397),
    [anon_sym_LT] = ACTIONS(1397),
    [anon_sym_LT_EQ] = ACTIONS(1397),
    [anon_sym_EQ_EQ] = ACTIONS(1397),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1397),
    [anon_sym_BANG_EQ] = ACTIONS(1397),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1397),
    [anon_sym_GT_EQ] = ACTIONS(1397),
    [anon_sym_GT] = ACTIONS(1397),
    [anon_sym_is] = ACTIONS(1397),
    [anon_sym_instanceof] = ACTIONS(1397),
    [anon_sym_new] = ACTIONS(1397),
    [anon_sym_QMARK] = ACTIONS(1397),
    [anon_sym_await] = ACTIONS(1397),
    [anon_sym_LBRACE] = ACTIONS(1397),
    [sym_this] = ACTIONS(1397),
    [sym_true] = ACTIONS(1397),
    [sym_false] = ACTIONS(1397),
    [sym_none] = ACTIONS(1397),
    [sym_string] = ACTIONS(1397),
    [sym_verbatim] = ACTIONS(1397),
    [sym_number] = ACTIONS(1397),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1395),
    [sym_fstring_start] = ACTIONS(1395),
  },
  [STATE(562)] = {
    [ts_builtin_sym_end] = ACTIONS(1399),
    [sym_identifier] = ACTIONS(1401),
    [anon_sym_SEMI] = ACTIONS(1401),
    [anon_sym_return] = ACTIONS(1401),
    [anon_sym_async] = ACTIONS(1401),
    [anon_sym_def] = ACTIONS(1401),
    [anon_sym_COMMA] = ACTIONS(1401),
    [anon_sym_LPAREN] = ACTIONS(1401),
    [anon_sym_LBRACK] = ACTIONS(1401),
    [anon_sym_del] = ACTIONS(1401),
    [anon_sym_raise] = ACTIONS(1401),
    [sym_pass_statement] = ACTIONS(1401),
    [sym_break_statement] = ACTIONS(1401),
    [sym_continue_statement] = ACTIONS(1401),
    [sym_debugger_statement] = ACTIONS(1401),
    [anon_sym_assert] = ACTIONS(1401),
    [anon_sym_global] = ACTIONS(1401),
    [anon_sym_nonlocal] = ACTIONS(1401),
    [anon_sym_yield] = ACTIONS(1401),
    [anon_sym_from] = ACTIONS(1401),
    [anon_sym_import] = ACTIONS(1401),
    [anon_sym_DOT] = ACTIONS(1401),
    [anon_sym_if] = ACTIONS(1401),
    [anon_sym_for] = ACTIONS(1401),
    [anon_sym_in] = ACTIONS(1401),
    [anon_sym_while] = ACTIONS(1401),
    [anon_sym_do] = ACTIONS(1401),
    [anon_sym_try] = ACTIONS(1401),
    [anon_sym_with] = ACTIONS(1401),
    [anon_sym_STAR] = ACTIONS(1401),
    [anon_sym_STAR_STAR] = ACTIONS(1401),
    [anon_sym_class] = ACTIONS(1401),
    [anon_sym_AT] = ACTIONS(1401),
    [anon_sym_not] = ACTIONS(1401),
    [anon_sym_and] = ACTIONS(1401),
    [anon_sym_or] = ACTIONS(1401),
    [anon_sym_PLUS] = ACTIONS(1401),
    [anon_sym_DASH] = ACTIONS(1401),
    [anon_sym_SLASH] = ACTIONS(1401),
    [anon_sym_SLASH_SLASH] = ACTIONS(1401),
    [anon_sym_PERCENT] = ACTIONS(1401),
    [anon_sym_PIPE] = ACTIONS(1401),
    [anon_sym_CARET] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(1401),
    [anon_sym_LT_LT] = ACTIONS(1401),
    [anon_sym_GT_GT] = ACTIONS(1401),
    [anon_sym_GT_GT_GT] = ACTIONS(1401),
    [anon_sym_TILDE] = ACTIONS(1401),
    [anon_sym_typeof] = ACTIONS(1401),
    [anon_sym_void] = ACTIONS(1401),
    [anon_sym_LT] = ACTIONS(1401),
    [anon_sym_LT_EQ] = ACTIONS(1401),
    [anon_sym_EQ_EQ] = ACTIONS(1401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1401),
    [anon_sym_BANG_EQ] = ACTIONS(1401),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1401),
    [anon_sym_GT_EQ] = ACTIONS(1401),
    [anon_sym_GT] = ACTIONS(1401),
    [anon_sym_is] = ACTIONS(1401),
    [anon_sym_instanceof] = ACTIONS(1401),
    [anon_sym_new] = ACTIONS(1401),
    [anon_sym_QMARK] = ACTIONS(1401),
    [anon_sym_await] = ACTIONS(1401),
    [anon_sym_LBRACE] = ACTIONS(1401),
    [sym_this] = ACTIONS(1401),
    [sym_true] = ACTIONS(1401),
    [sym_false] = ACTIONS(1401),
    [sym_none] = ACTIONS(1401),
    [sym_string] = ACTIONS(1401),
    [sym_verbatim] = ACTIONS(1401),
    [sym_number] = ACTIONS(1401),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1399),
    [sym_fstring_start] = ACTIONS(1399),
  },
  [STATE(563)] = {
    [aux_sym__inline_body_repeat1] = STATE(597),
    [ts_builtin_sym_end] = ACTIONS(1361),
    [sym_identifier] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1403),
    [anon_sym_return] = ACTIONS(1363),
    [anon_sym_async] = ACTIONS(1363),
    [anon_sym_def] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [anon_sym_del] = ACTIONS(1363),
    [anon_sym_raise] = ACTIONS(1363),
    [sym_pass_statement] = ACTIONS(1363),
    [sym_break_statement] = ACTIONS(1363),
    [sym_continue_statement] = ACTIONS(1363),
    [sym_debugger_statement] = ACTIONS(1363),
    [anon_sym_assert] = ACTIONS(1363),
    [anon_sym_global] = ACTIONS(1363),
    [anon_sym_nonlocal] = ACTIONS(1363),
    [anon_sym_yield] = ACTIONS(1363),
    [anon_sym_from] = ACTIONS(1363),
    [anon_sym_import] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_if] = ACTIONS(1363),
    [anon_sym_for] = ACTIONS(1363),
    [anon_sym_in] = ACTIONS(1363),
    [anon_sym_while] = ACTIONS(1363),
    [anon_sym_do] = ACTIONS(1363),
    [anon_sym_try] = ACTIONS(1363),
    [anon_sym_with] = ACTIONS(1363),
    [anon_sym_STAR] = ACTIONS(1363),
    [anon_sym_STAR_STAR] = ACTIONS(1363),
    [anon_sym_class] = ACTIONS(1363),
    [anon_sym_AT] = ACTIONS(1363),
    [anon_sym_not] = ACTIONS(1363),
    [anon_sym_and] = ACTIONS(1363),
    [anon_sym_or] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [anon_sym_SLASH] = ACTIONS(1363),
    [anon_sym_SLASH_SLASH] = ACTIONS(1363),
    [anon_sym_PERCENT] = ACTIONS(1363),
    [anon_sym_PIPE] = ACTIONS(1363),
    [anon_sym_CARET] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [anon_sym_LT_LT] = ACTIONS(1363),
    [anon_sym_GT_GT] = ACTIONS(1363),
    [anon_sym_GT_GT_GT] = ACTIONS(1363),
    [anon_sym_TILDE] = ACTIONS(1363),
    [anon_sym_typeof] = ACTIONS(1363),
    [anon_sym_void] = ACTIONS(1363),
    [anon_sym_LT] = ACTIONS(1363),
    [anon_sym_LT_EQ] = ACTIONS(1363),
    [anon_sym_EQ_EQ] = ACTIONS(1363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1363),
    [anon_sym_BANG_EQ] = ACTIONS(1363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1363),
    [anon_sym_GT_EQ] = ACTIONS(1363),
    [anon_sym_GT] = ACTIONS(1363),
    [anon_sym_is] = ACTIONS(1363),
    [anon_sym_instanceof] = ACTIONS(1363),
    [anon_sym_new] = ACTIONS(1363),
    [anon_sym_QMARK] = ACTIONS(1363),
    [anon_sym_await] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [sym_this] = ACTIONS(1363),
    [sym_true] = ACTIONS(1363),
    [sym_false] = ACTIONS(1363),
    [sym_none] = ACTIONS(1363),
    [sym_string] = ACTIONS(1363),
    [sym_verbatim] = ACTIONS(1363),
    [sym_number] = ACTIONS(1363),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1361),
    [sym_fstring_start] = ACTIONS(1361),
  },
  [STATE(564)] = {
    [ts_builtin_sym_end] = ACTIONS(1405),
    [sym_identifier] = ACTIONS(1407),
    [anon_sym_SEMI] = ACTIONS(1407),
    [anon_sym_return] = ACTIONS(1407),
    [anon_sym_async] = ACTIONS(1407),
    [anon_sym_def] = ACTIONS(1407),
    [anon_sym_COMMA] = ACTIONS(1407),
    [anon_sym_LPAREN] = ACTIONS(1407),
    [anon_sym_LBRACK] = ACTIONS(1407),
    [anon_sym_del] = ACTIONS(1407),
    [anon_sym_raise] = ACTIONS(1407),
    [sym_pass_statement] = ACTIONS(1407),
    [sym_break_statement] = ACTIONS(1407),
    [sym_continue_statement] = ACTIONS(1407),
    [sym_debugger_statement] = ACTIONS(1407),
    [anon_sym_assert] = ACTIONS(1407),
    [anon_sym_global] = ACTIONS(1407),
    [anon_sym_nonlocal] = ACTIONS(1407),
    [anon_sym_yield] = ACTIONS(1407),
    [anon_sym_from] = ACTIONS(1407),
    [anon_sym_import] = ACTIONS(1407),
    [anon_sym_DOT] = ACTIONS(1407),
    [anon_sym_if] = ACTIONS(1407),
    [anon_sym_for] = ACTIONS(1407),
    [anon_sym_in] = ACTIONS(1407),
    [anon_sym_while] = ACTIONS(1407),
    [anon_sym_do] = ACTIONS(1407),
    [anon_sym_try] = ACTIONS(1407),
    [anon_sym_with] = ACTIONS(1407),
    [anon_sym_STAR] = ACTIONS(1407),
    [anon_sym_STAR_STAR] = ACTIONS(1407),
    [anon_sym_class] = ACTIONS(1407),
    [anon_sym_AT] = ACTIONS(1407),
    [anon_sym_not] = ACTIONS(1407),
    [anon_sym_and] = ACTIONS(1407),
    [anon_sym_or] = ACTIONS(1407),
    [anon_sym_PLUS] = ACTIONS(1407),
    [anon_sym_DASH] = ACTIONS(1407),
    [anon_sym_SLASH] = ACTIONS(1407),
    [anon_sym_SLASH_SLASH] = ACTIONS(1407),
    [anon_sym_PERCENT] = ACTIONS(1407),
    [anon_sym_PIPE] = ACTIONS(1407),
    [anon_sym_CARET] = ACTIONS(1407),
    [anon_sym_AMP] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(1407),
    [anon_sym_GT_GT] = ACTIONS(1407),
    [anon_sym_GT_GT_GT] = ACTIONS(1407),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_typeof] = ACTIONS(1407),
    [anon_sym_void] = ACTIONS(1407),
    [anon_sym_LT] = ACTIONS(1407),
    [anon_sym_LT_EQ] = ACTIONS(1407),
    [anon_sym_EQ_EQ] = ACTIONS(1407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1407),
    [anon_sym_BANG_EQ] = ACTIONS(1407),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1407),
    [anon_sym_GT_EQ] = ACTIONS(1407),
    [anon_sym_GT] = ACTIONS(1407),
    [anon_sym_is] = ACTIONS(1407),
    [anon_sym_instanceof] = ACTIONS(1407),
    [anon_sym_new] = ACTIONS(1407),
    [anon_sym_QMARK] = ACTIONS(1407),
    [anon_sym_await] = ACTIONS(1407),
    [anon_sym_LBRACE] = ACTIONS(1407),
    [sym_this] = ACTIONS(1407),
    [sym_true] = ACTIONS(1407),
    [sym_false] = ACTIONS(1407),
    [sym_none] = ACTIONS(1407),
    [sym_string] = ACTIONS(1407),
    [sym_verbatim] = ACTIONS(1407),
    [sym_number] = ACTIONS(1407),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1405),
    [sym_fstring_start] = ACTIONS(1405),
  },
  [STATE(565)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1196),
    [sym_identifier] = ACTIONS(1194),
    [anon_sym_SEMI] = ACTIONS(1194),
    [anon_sym_return] = ACTIONS(1194),
    [anon_sym_async] = ACTIONS(1194),
    [anon_sym_def] = ACTIONS(1194),
    [anon_sym_LPAREN] = ACTIONS(1194),
    [anon_sym_LBRACK] = ACTIONS(1194),
    [anon_sym_del] = ACTIONS(1194),
    [anon_sym_raise] = ACTIONS(1194),
    [sym_pass_statement] = ACTIONS(1194),
    [sym_break_statement] = ACTIONS(1194),
    [sym_continue_statement] = ACTIONS(1194),
    [sym_debugger_statement] = ACTIONS(1194),
    [anon_sym_assert] = ACTIONS(1194),
    [anon_sym_global] = ACTIONS(1194),
    [anon_sym_nonlocal] = ACTIONS(1194),
    [anon_sym_yield] = ACTIONS(1194),
    [anon_sym_from] = ACTIONS(1194),
    [anon_sym_import] = ACTIONS(1194),
    [anon_sym_DOT] = ACTIONS(1194),
    [anon_sym_if] = ACTIONS(1194),
    [anon_sym_for] = ACTIONS(1194),
    [anon_sym_in] = ACTIONS(1194),
    [anon_sym_while] = ACTIONS(1194),
    [anon_sym_do] = ACTIONS(1194),
    [anon_sym_try] = ACTIONS(1194),
    [anon_sym_with] = ACTIONS(1194),
    [anon_sym_STAR] = ACTIONS(1194),
    [anon_sym_STAR_STAR] = ACTIONS(1194),
    [anon_sym_class] = ACTIONS(1194),
    [anon_sym_AT] = ACTIONS(1194),
    [anon_sym_not] = ACTIONS(1194),
    [anon_sym_and] = ACTIONS(1194),
    [anon_sym_or] = ACTIONS(1194),
    [anon_sym_PLUS] = ACTIONS(1194),
    [anon_sym_DASH] = ACTIONS(1194),
    [anon_sym_SLASH] = ACTIONS(1194),
    [anon_sym_SLASH_SLASH] = ACTIONS(1194),
    [anon_sym_PERCENT] = ACTIONS(1194),
    [anon_sym_PIPE] = ACTIONS(1194),
    [anon_sym_CARET] = ACTIONS(1194),
    [anon_sym_AMP] = ACTIONS(1194),
    [anon_sym_LT_LT] = ACTIONS(1194),
    [anon_sym_GT_GT] = ACTIONS(1194),
    [anon_sym_GT_GT_GT] = ACTIONS(1194),
    [anon_sym_TILDE] = ACTIONS(1194),
    [anon_sym_typeof] = ACTIONS(1194),
    [anon_sym_void] = ACTIONS(1194),
    [anon_sym_LT] = ACTIONS(1194),
    [anon_sym_LT_EQ] = ACTIONS(1194),
    [anon_sym_EQ_EQ] = ACTIONS(1194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1194),
    [anon_sym_BANG_EQ] = ACTIONS(1194),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1194),
    [anon_sym_GT_EQ] = ACTIONS(1194),
    [anon_sym_GT] = ACTIONS(1194),
    [anon_sym_is] = ACTIONS(1194),
    [anon_sym_instanceof] = ACTIONS(1194),
    [anon_sym_new] = ACTIONS(1194),
    [anon_sym_QMARK] = ACTIONS(1194),
    [anon_sym_await] = ACTIONS(1194),
    [anon_sym_LBRACE] = ACTIONS(1194),
    [sym_this] = ACTIONS(1194),
    [sym_true] = ACTIONS(1194),
    [sym_false] = ACTIONS(1194),
    [sym_none] = ACTIONS(1194),
    [sym_string] = ACTIONS(1194),
    [sym_verbatim] = ACTIONS(1194),
    [sym_number] = ACTIONS(1194),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1196),
    [sym_fstring_start] = ACTIONS(1196),
  },
  [STATE(566)] = {
    [ts_builtin_sym_end] = ACTIONS(1409),
    [sym_identifier] = ACTIONS(1411),
    [anon_sym_SEMI] = ACTIONS(1411),
    [anon_sym_return] = ACTIONS(1411),
    [anon_sym_async] = ACTIONS(1411),
    [anon_sym_def] = ACTIONS(1411),
    [anon_sym_COMMA] = ACTIONS(1411),
    [anon_sym_LPAREN] = ACTIONS(1411),
    [anon_sym_LBRACK] = ACTIONS(1411),
    [anon_sym_del] = ACTIONS(1411),
    [anon_sym_raise] = ACTIONS(1411),
    [sym_pass_statement] = ACTIONS(1411),
    [sym_break_statement] = ACTIONS(1411),
    [sym_continue_statement] = ACTIONS(1411),
    [sym_debugger_statement] = ACTIONS(1411),
    [anon_sym_assert] = ACTIONS(1411),
    [anon_sym_global] = ACTIONS(1411),
    [anon_sym_nonlocal] = ACTIONS(1411),
    [anon_sym_yield] = ACTIONS(1411),
    [anon_sym_from] = ACTIONS(1411),
    [anon_sym_import] = ACTIONS(1411),
    [anon_sym_DOT] = ACTIONS(1411),
    [anon_sym_if] = ACTIONS(1411),
    [anon_sym_for] = ACTIONS(1411),
    [anon_sym_in] = ACTIONS(1411),
    [anon_sym_while] = ACTIONS(1411),
    [anon_sym_do] = ACTIONS(1411),
    [anon_sym_try] = ACTIONS(1411),
    [anon_sym_with] = ACTIONS(1411),
    [anon_sym_STAR] = ACTIONS(1411),
    [anon_sym_STAR_STAR] = ACTIONS(1411),
    [anon_sym_class] = ACTIONS(1411),
    [anon_sym_AT] = ACTIONS(1411),
    [anon_sym_not] = ACTIONS(1411),
    [anon_sym_and] = ACTIONS(1411),
    [anon_sym_or] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1411),
    [anon_sym_DASH] = ACTIONS(1411),
    [anon_sym_SLASH] = ACTIONS(1411),
    [anon_sym_SLASH_SLASH] = ACTIONS(1411),
    [anon_sym_PERCENT] = ACTIONS(1411),
    [anon_sym_PIPE] = ACTIONS(1411),
    [anon_sym_CARET] = ACTIONS(1411),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_LT_LT] = ACTIONS(1411),
    [anon_sym_GT_GT] = ACTIONS(1411),
    [anon_sym_GT_GT_GT] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1411),
    [anon_sym_typeof] = ACTIONS(1411),
    [anon_sym_void] = ACTIONS(1411),
    [anon_sym_LT] = ACTIONS(1411),
    [anon_sym_LT_EQ] = ACTIONS(1411),
    [anon_sym_EQ_EQ] = ACTIONS(1411),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1411),
    [anon_sym_BANG_EQ] = ACTIONS(1411),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1411),
    [anon_sym_GT_EQ] = ACTIONS(1411),
    [anon_sym_GT] = ACTIONS(1411),
    [anon_sym_is] = ACTIONS(1411),
    [anon_sym_instanceof] = ACTIONS(1411),
    [anon_sym_new] = ACTIONS(1411),
    [anon_sym_QMARK] = ACTIONS(1411),
    [anon_sym_await] = ACTIONS(1411),
    [anon_sym_LBRACE] = ACTIONS(1411),
    [sym_this] = ACTIONS(1411),
    [sym_true] = ACTIONS(1411),
    [sym_false] = ACTIONS(1411),
    [sym_none] = ACTIONS(1411),
    [sym_string] = ACTIONS(1411),
    [sym_verbatim] = ACTIONS(1411),
    [sym_number] = ACTIONS(1411),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1409),
    [sym_fstring_start] = ACTIONS(1409),
  },
  [STATE(567)] = {
    [ts_builtin_sym_end] = ACTIONS(1137),
    [sym_identifier] = ACTIONS(1133),
    [anon_sym_SEMI] = ACTIONS(1133),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_async] = ACTIONS(1133),
    [anon_sym_def] = ACTIONS(1133),
    [anon_sym_COMMA] = ACTIONS(1133),
    [anon_sym_LPAREN] = ACTIONS(1133),
    [anon_sym_LBRACK] = ACTIONS(1133),
    [anon_sym_del] = ACTIONS(1133),
    [anon_sym_raise] = ACTIONS(1133),
    [sym_pass_statement] = ACTIONS(1133),
    [sym_break_statement] = ACTIONS(1133),
    [sym_continue_statement] = ACTIONS(1133),
    [sym_debugger_statement] = ACTIONS(1133),
    [anon_sym_assert] = ACTIONS(1133),
    [anon_sym_global] = ACTIONS(1133),
    [anon_sym_nonlocal] = ACTIONS(1133),
    [anon_sym_yield] = ACTIONS(1133),
    [anon_sym_from] = ACTIONS(1133),
    [anon_sym_import] = ACTIONS(1133),
    [anon_sym_DOT] = ACTIONS(1133),
    [anon_sym_if] = ACTIONS(1133),
    [anon_sym_for] = ACTIONS(1133),
    [anon_sym_in] = ACTIONS(1133),
    [anon_sym_while] = ACTIONS(1133),
    [anon_sym_do] = ACTIONS(1133),
    [anon_sym_try] = ACTIONS(1133),
    [anon_sym_with] = ACTIONS(1133),
    [anon_sym_STAR] = ACTIONS(1133),
    [anon_sym_STAR_STAR] = ACTIONS(1133),
    [anon_sym_class] = ACTIONS(1133),
    [anon_sym_AT] = ACTIONS(1133),
    [anon_sym_not] = ACTIONS(1133),
    [anon_sym_and] = ACTIONS(1133),
    [anon_sym_or] = ACTIONS(1133),
    [anon_sym_PLUS] = ACTIONS(1133),
    [anon_sym_DASH] = ACTIONS(1133),
    [anon_sym_SLASH] = ACTIONS(1133),
    [anon_sym_SLASH_SLASH] = ACTIONS(1133),
    [anon_sym_PERCENT] = ACTIONS(1133),
    [anon_sym_PIPE] = ACTIONS(1133),
    [anon_sym_CARET] = ACTIONS(1133),
    [anon_sym_AMP] = ACTIONS(1133),
    [anon_sym_LT_LT] = ACTIONS(1133),
    [anon_sym_GT_GT] = ACTIONS(1133),
    [anon_sym_GT_GT_GT] = ACTIONS(1133),
    [anon_sym_TILDE] = ACTIONS(1133),
    [anon_sym_typeof] = ACTIONS(1133),
    [anon_sym_void] = ACTIONS(1133),
    [anon_sym_LT] = ACTIONS(1133),
    [anon_sym_LT_EQ] = ACTIONS(1133),
    [anon_sym_EQ_EQ] = ACTIONS(1133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1133),
    [anon_sym_BANG_EQ] = ACTIONS(1133),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1133),
    [anon_sym_GT_EQ] = ACTIONS(1133),
    [anon_sym_GT] = ACTIONS(1133),
    [anon_sym_is] = ACTIONS(1133),
    [anon_sym_instanceof] = ACTIONS(1133),
    [anon_sym_new] = ACTIONS(1133),
    [anon_sym_QMARK] = ACTIONS(1133),
    [anon_sym_await] = ACTIONS(1133),
    [anon_sym_LBRACE] = ACTIONS(1133),
    [sym_this] = ACTIONS(1133),
    [sym_true] = ACTIONS(1133),
    [sym_false] = ACTIONS(1133),
    [sym_none] = ACTIONS(1133),
    [sym_string] = ACTIONS(1133),
    [sym_verbatim] = ACTIONS(1133),
    [sym_number] = ACTIONS(1133),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1137),
    [sym_fstring_start] = ACTIONS(1137),
  },
  [STATE(568)] = {
    [ts_builtin_sym_end] = ACTIONS(1413),
    [sym_identifier] = ACTIONS(1415),
    [anon_sym_SEMI] = ACTIONS(1415),
    [anon_sym_return] = ACTIONS(1415),
    [anon_sym_async] = ACTIONS(1415),
    [anon_sym_def] = ACTIONS(1415),
    [anon_sym_COMMA] = ACTIONS(1415),
    [anon_sym_LPAREN] = ACTIONS(1415),
    [anon_sym_LBRACK] = ACTIONS(1415),
    [anon_sym_del] = ACTIONS(1415),
    [anon_sym_raise] = ACTIONS(1415),
    [sym_pass_statement] = ACTIONS(1415),
    [sym_break_statement] = ACTIONS(1415),
    [sym_continue_statement] = ACTIONS(1415),
    [sym_debugger_statement] = ACTIONS(1415),
    [anon_sym_assert] = ACTIONS(1415),
    [anon_sym_global] = ACTIONS(1415),
    [anon_sym_nonlocal] = ACTIONS(1415),
    [anon_sym_yield] = ACTIONS(1415),
    [anon_sym_from] = ACTIONS(1415),
    [anon_sym_import] = ACTIONS(1415),
    [anon_sym_DOT] = ACTIONS(1415),
    [anon_sym_if] = ACTIONS(1415),
    [anon_sym_for] = ACTIONS(1415),
    [anon_sym_in] = ACTIONS(1415),
    [anon_sym_while] = ACTIONS(1415),
    [anon_sym_do] = ACTIONS(1415),
    [anon_sym_try] = ACTIONS(1415),
    [anon_sym_with] = ACTIONS(1415),
    [anon_sym_STAR] = ACTIONS(1415),
    [anon_sym_STAR_STAR] = ACTIONS(1415),
    [anon_sym_class] = ACTIONS(1415),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_not] = ACTIONS(1415),
    [anon_sym_and] = ACTIONS(1415),
    [anon_sym_or] = ACTIONS(1415),
    [anon_sym_PLUS] = ACTIONS(1415),
    [anon_sym_DASH] = ACTIONS(1415),
    [anon_sym_SLASH] = ACTIONS(1415),
    [anon_sym_SLASH_SLASH] = ACTIONS(1415),
    [anon_sym_PERCENT] = ACTIONS(1415),
    [anon_sym_PIPE] = ACTIONS(1415),
    [anon_sym_CARET] = ACTIONS(1415),
    [anon_sym_AMP] = ACTIONS(1415),
    [anon_sym_LT_LT] = ACTIONS(1415),
    [anon_sym_GT_GT] = ACTIONS(1415),
    [anon_sym_GT_GT_GT] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1415),
    [anon_sym_typeof] = ACTIONS(1415),
    [anon_sym_void] = ACTIONS(1415),
    [anon_sym_LT] = ACTIONS(1415),
    [anon_sym_LT_EQ] = ACTIONS(1415),
    [anon_sym_EQ_EQ] = ACTIONS(1415),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1415),
    [anon_sym_BANG_EQ] = ACTIONS(1415),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1415),
    [anon_sym_GT_EQ] = ACTIONS(1415),
    [anon_sym_GT] = ACTIONS(1415),
    [anon_sym_is] = ACTIONS(1415),
    [anon_sym_instanceof] = ACTIONS(1415),
    [anon_sym_new] = ACTIONS(1415),
    [anon_sym_QMARK] = ACTIONS(1415),
    [anon_sym_await] = ACTIONS(1415),
    [anon_sym_LBRACE] = ACTIONS(1415),
    [sym_this] = ACTIONS(1415),
    [sym_true] = ACTIONS(1415),
    [sym_false] = ACTIONS(1415),
    [sym_none] = ACTIONS(1415),
    [sym_string] = ACTIONS(1415),
    [sym_verbatim] = ACTIONS(1415),
    [sym_number] = ACTIONS(1415),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1413),
    [sym_fstring_start] = ACTIONS(1413),
  },
  [STATE(569)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym_return] = ACTIONS(1224),
    [anon_sym_async] = ACTIONS(1224),
    [anon_sym_def] = ACTIONS(1224),
    [anon_sym_LPAREN] = ACTIONS(1224),
    [anon_sym_LBRACK] = ACTIONS(1224),
    [anon_sym_del] = ACTIONS(1224),
    [anon_sym_raise] = ACTIONS(1224),
    [sym_pass_statement] = ACTIONS(1224),
    [sym_break_statement] = ACTIONS(1224),
    [sym_continue_statement] = ACTIONS(1224),
    [sym_debugger_statement] = ACTIONS(1224),
    [anon_sym_assert] = ACTIONS(1224),
    [anon_sym_global] = ACTIONS(1224),
    [anon_sym_nonlocal] = ACTIONS(1224),
    [anon_sym_yield] = ACTIONS(1224),
    [anon_sym_from] = ACTIONS(1224),
    [anon_sym_import] = ACTIONS(1224),
    [anon_sym_DOT] = ACTIONS(1224),
    [anon_sym_if] = ACTIONS(1224),
    [anon_sym_for] = ACTIONS(1224),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1224),
    [anon_sym_do] = ACTIONS(1224),
    [anon_sym_try] = ACTIONS(1224),
    [anon_sym_with] = ACTIONS(1224),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1224),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1425),
    [anon_sym_or] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_typeof] = ACTIONS(1224),
    [anon_sym_void] = ACTIONS(1224),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1224),
    [anon_sym_QMARK] = ACTIONS(1224),
    [anon_sym_await] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [sym_this] = ACTIONS(1224),
    [sym_true] = ACTIONS(1224),
    [sym_false] = ACTIONS(1224),
    [sym_none] = ACTIONS(1224),
    [sym_string] = ACTIONS(1224),
    [sym_verbatim] = ACTIONS(1224),
    [sym_number] = ACTIONS(1224),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1222),
    [sym_regex] = ACTIONS(1222),
    [sym_fstring_start] = ACTIONS(1222),
  },
  [STATE(570)] = {
    [ts_builtin_sym_end] = ACTIONS(1441),
    [sym_identifier] = ACTIONS(1443),
    [anon_sym_SEMI] = ACTIONS(1443),
    [anon_sym_return] = ACTIONS(1443),
    [anon_sym_async] = ACTIONS(1443),
    [anon_sym_def] = ACTIONS(1443),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LPAREN] = ACTIONS(1443),
    [anon_sym_LBRACK] = ACTIONS(1443),
    [anon_sym_del] = ACTIONS(1443),
    [anon_sym_raise] = ACTIONS(1443),
    [sym_pass_statement] = ACTIONS(1443),
    [sym_break_statement] = ACTIONS(1443),
    [sym_continue_statement] = ACTIONS(1443),
    [sym_debugger_statement] = ACTIONS(1443),
    [anon_sym_assert] = ACTIONS(1443),
    [anon_sym_global] = ACTIONS(1443),
    [anon_sym_nonlocal] = ACTIONS(1443),
    [anon_sym_yield] = ACTIONS(1443),
    [anon_sym_from] = ACTIONS(1443),
    [anon_sym_import] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1443),
    [anon_sym_if] = ACTIONS(1443),
    [anon_sym_for] = ACTIONS(1443),
    [anon_sym_in] = ACTIONS(1443),
    [anon_sym_while] = ACTIONS(1443),
    [anon_sym_do] = ACTIONS(1443),
    [anon_sym_try] = ACTIONS(1443),
    [anon_sym_with] = ACTIONS(1443),
    [anon_sym_STAR] = ACTIONS(1443),
    [anon_sym_STAR_STAR] = ACTIONS(1443),
    [anon_sym_class] = ACTIONS(1443),
    [anon_sym_AT] = ACTIONS(1443),
    [anon_sym_not] = ACTIONS(1443),
    [anon_sym_and] = ACTIONS(1443),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_PLUS] = ACTIONS(1443),
    [anon_sym_DASH] = ACTIONS(1443),
    [anon_sym_SLASH] = ACTIONS(1443),
    [anon_sym_SLASH_SLASH] = ACTIONS(1443),
    [anon_sym_PERCENT] = ACTIONS(1443),
    [anon_sym_PIPE] = ACTIONS(1443),
    [anon_sym_CARET] = ACTIONS(1443),
    [anon_sym_AMP] = ACTIONS(1443),
    [anon_sym_LT_LT] = ACTIONS(1443),
    [anon_sym_GT_GT] = ACTIONS(1443),
    [anon_sym_GT_GT_GT] = ACTIONS(1443),
    [anon_sym_TILDE] = ACTIONS(1443),
    [anon_sym_typeof] = ACTIONS(1443),
    [anon_sym_void] = ACTIONS(1443),
    [anon_sym_LT] = ACTIONS(1443),
    [anon_sym_LT_EQ] = ACTIONS(1443),
    [anon_sym_EQ_EQ] = ACTIONS(1443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1443),
    [anon_sym_BANG_EQ] = ACTIONS(1443),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1443),
    [anon_sym_GT_EQ] = ACTIONS(1443),
    [anon_sym_GT] = ACTIONS(1443),
    [anon_sym_is] = ACTIONS(1443),
    [anon_sym_instanceof] = ACTIONS(1443),
    [anon_sym_new] = ACTIONS(1443),
    [anon_sym_QMARK] = ACTIONS(1443),
    [anon_sym_await] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1443),
    [sym_this] = ACTIONS(1443),
    [sym_true] = ACTIONS(1443),
    [sym_false] = ACTIONS(1443),
    [sym_none] = ACTIONS(1443),
    [sym_string] = ACTIONS(1443),
    [sym_verbatim] = ACTIONS(1443),
    [sym_number] = ACTIONS(1443),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1441),
    [sym_fstring_start] = ACTIONS(1441),
  },
  [STATE(571)] = {
    [ts_builtin_sym_end] = ACTIONS(1445),
    [sym_identifier] = ACTIONS(1447),
    [anon_sym_SEMI] = ACTIONS(1447),
    [anon_sym_return] = ACTIONS(1447),
    [anon_sym_async] = ACTIONS(1447),
    [anon_sym_def] = ACTIONS(1447),
    [anon_sym_COMMA] = ACTIONS(1447),
    [anon_sym_LPAREN] = ACTIONS(1447),
    [anon_sym_LBRACK] = ACTIONS(1447),
    [anon_sym_del] = ACTIONS(1447),
    [anon_sym_raise] = ACTIONS(1447),
    [sym_pass_statement] = ACTIONS(1447),
    [sym_break_statement] = ACTIONS(1447),
    [sym_continue_statement] = ACTIONS(1447),
    [sym_debugger_statement] = ACTIONS(1447),
    [anon_sym_assert] = ACTIONS(1447),
    [anon_sym_global] = ACTIONS(1447),
    [anon_sym_nonlocal] = ACTIONS(1447),
    [anon_sym_yield] = ACTIONS(1447),
    [anon_sym_from] = ACTIONS(1447),
    [anon_sym_import] = ACTIONS(1447),
    [anon_sym_DOT] = ACTIONS(1447),
    [anon_sym_if] = ACTIONS(1447),
    [anon_sym_for] = ACTIONS(1447),
    [anon_sym_in] = ACTIONS(1447),
    [anon_sym_while] = ACTIONS(1447),
    [anon_sym_do] = ACTIONS(1447),
    [anon_sym_try] = ACTIONS(1447),
    [anon_sym_with] = ACTIONS(1447),
    [anon_sym_STAR] = ACTIONS(1447),
    [anon_sym_STAR_STAR] = ACTIONS(1447),
    [anon_sym_class] = ACTIONS(1447),
    [anon_sym_AT] = ACTIONS(1447),
    [anon_sym_not] = ACTIONS(1447),
    [anon_sym_and] = ACTIONS(1447),
    [anon_sym_or] = ACTIONS(1447),
    [anon_sym_PLUS] = ACTIONS(1447),
    [anon_sym_DASH] = ACTIONS(1447),
    [anon_sym_SLASH] = ACTIONS(1447),
    [anon_sym_SLASH_SLASH] = ACTIONS(1447),
    [anon_sym_PERCENT] = ACTIONS(1447),
    [anon_sym_PIPE] = ACTIONS(1447),
    [anon_sym_CARET] = ACTIONS(1447),
    [anon_sym_AMP] = ACTIONS(1447),
    [anon_sym_LT_LT] = ACTIONS(1447),
    [anon_sym_GT_GT] = ACTIONS(1447),
    [anon_sym_GT_GT_GT] = ACTIONS(1447),
    [anon_sym_TILDE] = ACTIONS(1447),
    [anon_sym_typeof] = ACTIONS(1447),
    [anon_sym_void] = ACTIONS(1447),
    [anon_sym_LT] = ACTIONS(1447),
    [anon_sym_LT_EQ] = ACTIONS(1447),
    [anon_sym_EQ_EQ] = ACTIONS(1447),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1447),
    [anon_sym_BANG_EQ] = ACTIONS(1447),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1447),
    [anon_sym_GT_EQ] = ACTIONS(1447),
    [anon_sym_GT] = ACTIONS(1447),
    [anon_sym_is] = ACTIONS(1447),
    [anon_sym_instanceof] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1447),
    [anon_sym_QMARK] = ACTIONS(1447),
    [anon_sym_await] = ACTIONS(1447),
    [anon_sym_LBRACE] = ACTIONS(1447),
    [sym_this] = ACTIONS(1447),
    [sym_true] = ACTIONS(1447),
    [sym_false] = ACTIONS(1447),
    [sym_none] = ACTIONS(1447),
    [sym_string] = ACTIONS(1447),
    [sym_verbatim] = ACTIONS(1447),
    [sym_number] = ACTIONS(1447),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1445),
    [sym_fstring_start] = ACTIONS(1445),
  },
  [STATE(572)] = {
    [ts_builtin_sym_end] = ACTIONS(1449),
    [sym_identifier] = ACTIONS(1451),
    [anon_sym_SEMI] = ACTIONS(1451),
    [anon_sym_return] = ACTIONS(1451),
    [anon_sym_async] = ACTIONS(1451),
    [anon_sym_def] = ACTIONS(1451),
    [anon_sym_COMMA] = ACTIONS(1451),
    [anon_sym_LPAREN] = ACTIONS(1451),
    [anon_sym_LBRACK] = ACTIONS(1451),
    [anon_sym_del] = ACTIONS(1451),
    [anon_sym_raise] = ACTIONS(1451),
    [sym_pass_statement] = ACTIONS(1451),
    [sym_break_statement] = ACTIONS(1451),
    [sym_continue_statement] = ACTIONS(1451),
    [sym_debugger_statement] = ACTIONS(1451),
    [anon_sym_assert] = ACTIONS(1451),
    [anon_sym_global] = ACTIONS(1451),
    [anon_sym_nonlocal] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1451),
    [anon_sym_from] = ACTIONS(1451),
    [anon_sym_import] = ACTIONS(1451),
    [anon_sym_DOT] = ACTIONS(1451),
    [anon_sym_if] = ACTIONS(1451),
    [anon_sym_for] = ACTIONS(1451),
    [anon_sym_in] = ACTIONS(1451),
    [anon_sym_while] = ACTIONS(1451),
    [anon_sym_do] = ACTIONS(1451),
    [anon_sym_try] = ACTIONS(1451),
    [anon_sym_with] = ACTIONS(1451),
    [anon_sym_STAR] = ACTIONS(1451),
    [anon_sym_STAR_STAR] = ACTIONS(1451),
    [anon_sym_class] = ACTIONS(1451),
    [anon_sym_AT] = ACTIONS(1451),
    [anon_sym_not] = ACTIONS(1451),
    [anon_sym_and] = ACTIONS(1451),
    [anon_sym_or] = ACTIONS(1451),
    [anon_sym_PLUS] = ACTIONS(1451),
    [anon_sym_DASH] = ACTIONS(1451),
    [anon_sym_SLASH] = ACTIONS(1451),
    [anon_sym_SLASH_SLASH] = ACTIONS(1451),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [anon_sym_PIPE] = ACTIONS(1451),
    [anon_sym_CARET] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(1451),
    [anon_sym_LT_LT] = ACTIONS(1451),
    [anon_sym_GT_GT] = ACTIONS(1451),
    [anon_sym_GT_GT_GT] = ACTIONS(1451),
    [anon_sym_TILDE] = ACTIONS(1451),
    [anon_sym_typeof] = ACTIONS(1451),
    [anon_sym_void] = ACTIONS(1451),
    [anon_sym_LT] = ACTIONS(1451),
    [anon_sym_LT_EQ] = ACTIONS(1451),
    [anon_sym_EQ_EQ] = ACTIONS(1451),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1451),
    [anon_sym_BANG_EQ] = ACTIONS(1451),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1451),
    [anon_sym_GT_EQ] = ACTIONS(1451),
    [anon_sym_GT] = ACTIONS(1451),
    [anon_sym_is] = ACTIONS(1451),
    [anon_sym_instanceof] = ACTIONS(1451),
    [anon_sym_new] = ACTIONS(1451),
    [anon_sym_QMARK] = ACTIONS(1451),
    [anon_sym_await] = ACTIONS(1451),
    [anon_sym_LBRACE] = ACTIONS(1451),
    [sym_this] = ACTIONS(1451),
    [sym_true] = ACTIONS(1451),
    [sym_false] = ACTIONS(1451),
    [sym_none] = ACTIONS(1451),
    [sym_string] = ACTIONS(1451),
    [sym_verbatim] = ACTIONS(1451),
    [sym_number] = ACTIONS(1451),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1449),
    [sym_fstring_start] = ACTIONS(1449),
  },
  [STATE(573)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1228),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1228),
    [anon_sym_SLASH_SLASH] = ACTIONS(1228),
    [anon_sym_PERCENT] = ACTIONS(1228),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(574)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(574),
    [ts_builtin_sym_end] = ACTIONS(1245),
    [sym_identifier] = ACTIONS(1247),
    [anon_sym_SEMI] = ACTIONS(1247),
    [anon_sym_return] = ACTIONS(1247),
    [anon_sym_async] = ACTIONS(1247),
    [anon_sym_def] = ACTIONS(1247),
    [anon_sym_LPAREN] = ACTIONS(1247),
    [anon_sym_LBRACK] = ACTIONS(1247),
    [anon_sym_del] = ACTIONS(1247),
    [anon_sym_raise] = ACTIONS(1247),
    [sym_pass_statement] = ACTIONS(1247),
    [sym_break_statement] = ACTIONS(1247),
    [sym_continue_statement] = ACTIONS(1247),
    [sym_debugger_statement] = ACTIONS(1247),
    [anon_sym_assert] = ACTIONS(1247),
    [anon_sym_global] = ACTIONS(1247),
    [anon_sym_nonlocal] = ACTIONS(1247),
    [anon_sym_yield] = ACTIONS(1247),
    [anon_sym_from] = ACTIONS(1247),
    [anon_sym_import] = ACTIONS(1247),
    [anon_sym_DOT] = ACTIONS(1247),
    [anon_sym_if] = ACTIONS(1247),
    [anon_sym_for] = ACTIONS(1247),
    [anon_sym_in] = ACTIONS(1453),
    [anon_sym_while] = ACTIONS(1247),
    [anon_sym_do] = ACTIONS(1247),
    [anon_sym_try] = ACTIONS(1247),
    [anon_sym_with] = ACTIONS(1247),
    [anon_sym_STAR] = ACTIONS(1247),
    [anon_sym_STAR_STAR] = ACTIONS(1247),
    [anon_sym_class] = ACTIONS(1247),
    [anon_sym_AT] = ACTIONS(1247),
    [anon_sym_not] = ACTIONS(1456),
    [anon_sym_and] = ACTIONS(1247),
    [anon_sym_or] = ACTIONS(1247),
    [anon_sym_PLUS] = ACTIONS(1247),
    [anon_sym_DASH] = ACTIONS(1247),
    [anon_sym_SLASH] = ACTIONS(1247),
    [anon_sym_SLASH_SLASH] = ACTIONS(1247),
    [anon_sym_PERCENT] = ACTIONS(1247),
    [anon_sym_PIPE] = ACTIONS(1247),
    [anon_sym_CARET] = ACTIONS(1247),
    [anon_sym_AMP] = ACTIONS(1247),
    [anon_sym_LT_LT] = ACTIONS(1247),
    [anon_sym_GT_GT] = ACTIONS(1247),
    [anon_sym_GT_GT_GT] = ACTIONS(1247),
    [anon_sym_TILDE] = ACTIONS(1247),
    [anon_sym_typeof] = ACTIONS(1247),
    [anon_sym_void] = ACTIONS(1247),
    [anon_sym_LT] = ACTIONS(1453),
    [anon_sym_LT_EQ] = ACTIONS(1453),
    [anon_sym_EQ_EQ] = ACTIONS(1453),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1453),
    [anon_sym_BANG_EQ] = ACTIONS(1453),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1453),
    [anon_sym_GT_EQ] = ACTIONS(1453),
    [anon_sym_GT] = ACTIONS(1453),
    [anon_sym_is] = ACTIONS(1459),
    [anon_sym_instanceof] = ACTIONS(1453),
    [anon_sym_new] = ACTIONS(1247),
    [anon_sym_QMARK] = ACTIONS(1247),
    [anon_sym_await] = ACTIONS(1247),
    [anon_sym_LBRACE] = ACTIONS(1247),
    [sym_this] = ACTIONS(1247),
    [sym_true] = ACTIONS(1247),
    [sym_false] = ACTIONS(1247),
    [sym_none] = ACTIONS(1247),
    [sym_string] = ACTIONS(1247),
    [sym_verbatim] = ACTIONS(1247),
    [sym_number] = ACTIONS(1247),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1245),
    [sym_fstring_start] = ACTIONS(1245),
  },
  [STATE(575)] = {
    [ts_builtin_sym_end] = ACTIONS(1462),
    [sym_identifier] = ACTIONS(1464),
    [anon_sym_SEMI] = ACTIONS(1464),
    [anon_sym_return] = ACTIONS(1464),
    [anon_sym_async] = ACTIONS(1464),
    [anon_sym_def] = ACTIONS(1464),
    [anon_sym_COMMA] = ACTIONS(1464),
    [anon_sym_LPAREN] = ACTIONS(1464),
    [anon_sym_LBRACK] = ACTIONS(1464),
    [anon_sym_del] = ACTIONS(1464),
    [anon_sym_raise] = ACTIONS(1464),
    [sym_pass_statement] = ACTIONS(1464),
    [sym_break_statement] = ACTIONS(1464),
    [sym_continue_statement] = ACTIONS(1464),
    [sym_debugger_statement] = ACTIONS(1464),
    [anon_sym_assert] = ACTIONS(1464),
    [anon_sym_global] = ACTIONS(1464),
    [anon_sym_nonlocal] = ACTIONS(1464),
    [anon_sym_yield] = ACTIONS(1464),
    [anon_sym_from] = ACTIONS(1464),
    [anon_sym_import] = ACTIONS(1464),
    [anon_sym_DOT] = ACTIONS(1464),
    [anon_sym_if] = ACTIONS(1464),
    [anon_sym_for] = ACTIONS(1464),
    [anon_sym_in] = ACTIONS(1464),
    [anon_sym_while] = ACTIONS(1464),
    [anon_sym_do] = ACTIONS(1464),
    [anon_sym_try] = ACTIONS(1464),
    [anon_sym_with] = ACTIONS(1464),
    [anon_sym_STAR] = ACTIONS(1464),
    [anon_sym_STAR_STAR] = ACTIONS(1464),
    [anon_sym_class] = ACTIONS(1464),
    [anon_sym_AT] = ACTIONS(1464),
    [anon_sym_not] = ACTIONS(1464),
    [anon_sym_and] = ACTIONS(1464),
    [anon_sym_or] = ACTIONS(1464),
    [anon_sym_PLUS] = ACTIONS(1464),
    [anon_sym_DASH] = ACTIONS(1464),
    [anon_sym_SLASH] = ACTIONS(1464),
    [anon_sym_SLASH_SLASH] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1464),
    [anon_sym_PIPE] = ACTIONS(1464),
    [anon_sym_CARET] = ACTIONS(1464),
    [anon_sym_AMP] = ACTIONS(1464),
    [anon_sym_LT_LT] = ACTIONS(1464),
    [anon_sym_GT_GT] = ACTIONS(1464),
    [anon_sym_GT_GT_GT] = ACTIONS(1464),
    [anon_sym_TILDE] = ACTIONS(1464),
    [anon_sym_typeof] = ACTIONS(1464),
    [anon_sym_void] = ACTIONS(1464),
    [anon_sym_LT] = ACTIONS(1464),
    [anon_sym_LT_EQ] = ACTIONS(1464),
    [anon_sym_EQ_EQ] = ACTIONS(1464),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1464),
    [anon_sym_BANG_EQ] = ACTIONS(1464),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1464),
    [anon_sym_GT_EQ] = ACTIONS(1464),
    [anon_sym_GT] = ACTIONS(1464),
    [anon_sym_is] = ACTIONS(1464),
    [anon_sym_instanceof] = ACTIONS(1464),
    [anon_sym_new] = ACTIONS(1464),
    [anon_sym_QMARK] = ACTIONS(1464),
    [anon_sym_await] = ACTIONS(1464),
    [anon_sym_LBRACE] = ACTIONS(1464),
    [sym_this] = ACTIONS(1464),
    [sym_true] = ACTIONS(1464),
    [sym_false] = ACTIONS(1464),
    [sym_none] = ACTIONS(1464),
    [sym_string] = ACTIONS(1464),
    [sym_verbatim] = ACTIONS(1464),
    [sym_number] = ACTIONS(1464),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1462),
    [sym_fstring_start] = ACTIONS(1462),
  },
  [STATE(576)] = {
    [ts_builtin_sym_end] = ACTIONS(1466),
    [sym_identifier] = ACTIONS(1468),
    [anon_sym_SEMI] = ACTIONS(1468),
    [anon_sym_return] = ACTIONS(1468),
    [anon_sym_async] = ACTIONS(1468),
    [anon_sym_def] = ACTIONS(1468),
    [anon_sym_COMMA] = ACTIONS(1468),
    [anon_sym_LPAREN] = ACTIONS(1468),
    [anon_sym_LBRACK] = ACTIONS(1468),
    [anon_sym_del] = ACTIONS(1468),
    [anon_sym_raise] = ACTIONS(1468),
    [sym_pass_statement] = ACTIONS(1468),
    [sym_break_statement] = ACTIONS(1468),
    [sym_continue_statement] = ACTIONS(1468),
    [sym_debugger_statement] = ACTIONS(1468),
    [anon_sym_assert] = ACTIONS(1468),
    [anon_sym_global] = ACTIONS(1468),
    [anon_sym_nonlocal] = ACTIONS(1468),
    [anon_sym_yield] = ACTIONS(1468),
    [anon_sym_from] = ACTIONS(1468),
    [anon_sym_import] = ACTIONS(1468),
    [anon_sym_DOT] = ACTIONS(1468),
    [anon_sym_if] = ACTIONS(1468),
    [anon_sym_for] = ACTIONS(1468),
    [anon_sym_in] = ACTIONS(1468),
    [anon_sym_while] = ACTIONS(1468),
    [anon_sym_do] = ACTIONS(1468),
    [anon_sym_try] = ACTIONS(1468),
    [anon_sym_with] = ACTIONS(1468),
    [anon_sym_STAR] = ACTIONS(1468),
    [anon_sym_STAR_STAR] = ACTIONS(1468),
    [anon_sym_class] = ACTIONS(1468),
    [anon_sym_AT] = ACTIONS(1468),
    [anon_sym_not] = ACTIONS(1468),
    [anon_sym_and] = ACTIONS(1468),
    [anon_sym_or] = ACTIONS(1468),
    [anon_sym_PLUS] = ACTIONS(1468),
    [anon_sym_DASH] = ACTIONS(1468),
    [anon_sym_SLASH] = ACTIONS(1468),
    [anon_sym_SLASH_SLASH] = ACTIONS(1468),
    [anon_sym_PERCENT] = ACTIONS(1468),
    [anon_sym_PIPE] = ACTIONS(1468),
    [anon_sym_CARET] = ACTIONS(1468),
    [anon_sym_AMP] = ACTIONS(1468),
    [anon_sym_LT_LT] = ACTIONS(1468),
    [anon_sym_GT_GT] = ACTIONS(1468),
    [anon_sym_GT_GT_GT] = ACTIONS(1468),
    [anon_sym_TILDE] = ACTIONS(1468),
    [anon_sym_typeof] = ACTIONS(1468),
    [anon_sym_void] = ACTIONS(1468),
    [anon_sym_LT] = ACTIONS(1468),
    [anon_sym_LT_EQ] = ACTIONS(1468),
    [anon_sym_EQ_EQ] = ACTIONS(1468),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1468),
    [anon_sym_BANG_EQ] = ACTIONS(1468),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1468),
    [anon_sym_GT_EQ] = ACTIONS(1468),
    [anon_sym_GT] = ACTIONS(1468),
    [anon_sym_is] = ACTIONS(1468),
    [anon_sym_instanceof] = ACTIONS(1468),
    [anon_sym_new] = ACTIONS(1468),
    [anon_sym_QMARK] = ACTIONS(1468),
    [anon_sym_await] = ACTIONS(1468),
    [anon_sym_LBRACE] = ACTIONS(1468),
    [sym_this] = ACTIONS(1468),
    [sym_true] = ACTIONS(1468),
    [sym_false] = ACTIONS(1468),
    [sym_none] = ACTIONS(1468),
    [sym_string] = ACTIONS(1468),
    [sym_verbatim] = ACTIONS(1468),
    [sym_number] = ACTIONS(1468),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1466),
    [sym_fstring_start] = ACTIONS(1466),
  },
  [STATE(577)] = {
    [ts_builtin_sym_end] = ACTIONS(1470),
    [sym_identifier] = ACTIONS(1472),
    [anon_sym_SEMI] = ACTIONS(1472),
    [anon_sym_return] = ACTIONS(1472),
    [anon_sym_async] = ACTIONS(1472),
    [anon_sym_def] = ACTIONS(1472),
    [anon_sym_COMMA] = ACTIONS(1472),
    [anon_sym_LPAREN] = ACTIONS(1472),
    [anon_sym_LBRACK] = ACTIONS(1472),
    [anon_sym_del] = ACTIONS(1472),
    [anon_sym_raise] = ACTIONS(1472),
    [sym_pass_statement] = ACTIONS(1472),
    [sym_break_statement] = ACTIONS(1472),
    [sym_continue_statement] = ACTIONS(1472),
    [sym_debugger_statement] = ACTIONS(1472),
    [anon_sym_assert] = ACTIONS(1472),
    [anon_sym_global] = ACTIONS(1472),
    [anon_sym_nonlocal] = ACTIONS(1472),
    [anon_sym_yield] = ACTIONS(1472),
    [anon_sym_from] = ACTIONS(1472),
    [anon_sym_import] = ACTIONS(1472),
    [anon_sym_DOT] = ACTIONS(1472),
    [anon_sym_if] = ACTIONS(1472),
    [anon_sym_for] = ACTIONS(1472),
    [anon_sym_in] = ACTIONS(1472),
    [anon_sym_while] = ACTIONS(1472),
    [anon_sym_do] = ACTIONS(1472),
    [anon_sym_try] = ACTIONS(1472),
    [anon_sym_with] = ACTIONS(1472),
    [anon_sym_STAR] = ACTIONS(1472),
    [anon_sym_STAR_STAR] = ACTIONS(1472),
    [anon_sym_class] = ACTIONS(1472),
    [anon_sym_AT] = ACTIONS(1472),
    [anon_sym_not] = ACTIONS(1472),
    [anon_sym_and] = ACTIONS(1472),
    [anon_sym_or] = ACTIONS(1472),
    [anon_sym_PLUS] = ACTIONS(1472),
    [anon_sym_DASH] = ACTIONS(1472),
    [anon_sym_SLASH] = ACTIONS(1472),
    [anon_sym_SLASH_SLASH] = ACTIONS(1472),
    [anon_sym_PERCENT] = ACTIONS(1472),
    [anon_sym_PIPE] = ACTIONS(1472),
    [anon_sym_CARET] = ACTIONS(1472),
    [anon_sym_AMP] = ACTIONS(1472),
    [anon_sym_LT_LT] = ACTIONS(1472),
    [anon_sym_GT_GT] = ACTIONS(1472),
    [anon_sym_GT_GT_GT] = ACTIONS(1472),
    [anon_sym_TILDE] = ACTIONS(1472),
    [anon_sym_typeof] = ACTIONS(1472),
    [anon_sym_void] = ACTIONS(1472),
    [anon_sym_LT] = ACTIONS(1472),
    [anon_sym_LT_EQ] = ACTIONS(1472),
    [anon_sym_EQ_EQ] = ACTIONS(1472),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1472),
    [anon_sym_BANG_EQ] = ACTIONS(1472),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1472),
    [anon_sym_GT_EQ] = ACTIONS(1472),
    [anon_sym_GT] = ACTIONS(1472),
    [anon_sym_is] = ACTIONS(1472),
    [anon_sym_instanceof] = ACTIONS(1472),
    [anon_sym_new] = ACTIONS(1472),
    [anon_sym_QMARK] = ACTIONS(1472),
    [anon_sym_await] = ACTIONS(1472),
    [anon_sym_LBRACE] = ACTIONS(1472),
    [sym_this] = ACTIONS(1472),
    [sym_true] = ACTIONS(1472),
    [sym_false] = ACTIONS(1472),
    [sym_none] = ACTIONS(1472),
    [sym_string] = ACTIONS(1472),
    [sym_verbatim] = ACTIONS(1472),
    [sym_number] = ACTIONS(1472),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1470),
    [sym_fstring_start] = ACTIONS(1470),
  },
  [STATE(578)] = {
    [sym_identifier] = ACTIONS(1474),
    [anon_sym_SEMI] = ACTIONS(1474),
    [anon_sym_return] = ACTIONS(1474),
    [anon_sym_async] = ACTIONS(1474),
    [anon_sym_def] = ACTIONS(1474),
    [anon_sym_COMMA] = ACTIONS(1474),
    [anon_sym_LPAREN] = ACTIONS(1474),
    [anon_sym_LBRACK] = ACTIONS(1474),
    [anon_sym_del] = ACTIONS(1474),
    [anon_sym_raise] = ACTIONS(1474),
    [sym_pass_statement] = ACTIONS(1474),
    [sym_break_statement] = ACTIONS(1474),
    [sym_continue_statement] = ACTIONS(1474),
    [sym_debugger_statement] = ACTIONS(1474),
    [anon_sym_assert] = ACTIONS(1474),
    [anon_sym_global] = ACTIONS(1474),
    [anon_sym_nonlocal] = ACTIONS(1474),
    [anon_sym_yield] = ACTIONS(1474),
    [anon_sym_from] = ACTIONS(1474),
    [anon_sym_import] = ACTIONS(1474),
    [anon_sym_DOT] = ACTIONS(1474),
    [anon_sym_if] = ACTIONS(1474),
    [anon_sym_for] = ACTIONS(1474),
    [anon_sym_in] = ACTIONS(1474),
    [anon_sym_while] = ACTIONS(1474),
    [anon_sym_do] = ACTIONS(1474),
    [anon_sym_try] = ACTIONS(1474),
    [anon_sym_with] = ACTIONS(1474),
    [anon_sym_STAR] = ACTIONS(1474),
    [anon_sym_STAR_STAR] = ACTIONS(1474),
    [anon_sym_class] = ACTIONS(1474),
    [anon_sym_AT] = ACTIONS(1474),
    [anon_sym_not] = ACTIONS(1474),
    [anon_sym_and] = ACTIONS(1474),
    [anon_sym_or] = ACTIONS(1474),
    [anon_sym_PLUS] = ACTIONS(1474),
    [anon_sym_DASH] = ACTIONS(1474),
    [anon_sym_SLASH] = ACTIONS(1474),
    [anon_sym_SLASH_SLASH] = ACTIONS(1474),
    [anon_sym_PERCENT] = ACTIONS(1474),
    [anon_sym_PIPE] = ACTIONS(1474),
    [anon_sym_CARET] = ACTIONS(1474),
    [anon_sym_AMP] = ACTIONS(1474),
    [anon_sym_LT_LT] = ACTIONS(1474),
    [anon_sym_GT_GT] = ACTIONS(1474),
    [anon_sym_GT_GT_GT] = ACTIONS(1474),
    [anon_sym_TILDE] = ACTIONS(1474),
    [anon_sym_typeof] = ACTIONS(1474),
    [anon_sym_void] = ACTIONS(1474),
    [anon_sym_LT] = ACTIONS(1474),
    [anon_sym_LT_EQ] = ACTIONS(1474),
    [anon_sym_EQ_EQ] = ACTIONS(1474),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1474),
    [anon_sym_BANG_EQ] = ACTIONS(1474),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1474),
    [anon_sym_GT_EQ] = ACTIONS(1474),
    [anon_sym_GT] = ACTIONS(1474),
    [anon_sym_is] = ACTIONS(1474),
    [anon_sym_instanceof] = ACTIONS(1474),
    [anon_sym_new] = ACTIONS(1474),
    [anon_sym_QMARK] = ACTIONS(1474),
    [anon_sym_await] = ACTIONS(1474),
    [anon_sym_LBRACE] = ACTIONS(1474),
    [sym_this] = ACTIONS(1474),
    [sym_true] = ACTIONS(1474),
    [sym_false] = ACTIONS(1474),
    [sym_none] = ACTIONS(1474),
    [sym_string] = ACTIONS(1474),
    [sym_verbatim] = ACTIONS(1474),
    [sym_number] = ACTIONS(1474),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1476),
    [sym_regex] = ACTIONS(1476),
    [sym_fstring_start] = ACTIONS(1476),
  },
  [STATE(579)] = {
    [ts_builtin_sym_end] = ACTIONS(1478),
    [sym_identifier] = ACTIONS(1480),
    [anon_sym_SEMI] = ACTIONS(1480),
    [anon_sym_return] = ACTIONS(1480),
    [anon_sym_async] = ACTIONS(1480),
    [anon_sym_def] = ACTIONS(1480),
    [anon_sym_COMMA] = ACTIONS(1480),
    [anon_sym_LPAREN] = ACTIONS(1480),
    [anon_sym_LBRACK] = ACTIONS(1480),
    [anon_sym_del] = ACTIONS(1480),
    [anon_sym_raise] = ACTIONS(1480),
    [sym_pass_statement] = ACTIONS(1480),
    [sym_break_statement] = ACTIONS(1480),
    [sym_continue_statement] = ACTIONS(1480),
    [sym_debugger_statement] = ACTIONS(1480),
    [anon_sym_assert] = ACTIONS(1480),
    [anon_sym_global] = ACTIONS(1480),
    [anon_sym_nonlocal] = ACTIONS(1480),
    [anon_sym_yield] = ACTIONS(1480),
    [anon_sym_from] = ACTIONS(1480),
    [anon_sym_import] = ACTIONS(1480),
    [anon_sym_DOT] = ACTIONS(1480),
    [anon_sym_if] = ACTIONS(1480),
    [anon_sym_for] = ACTIONS(1480),
    [anon_sym_in] = ACTIONS(1480),
    [anon_sym_while] = ACTIONS(1480),
    [anon_sym_do] = ACTIONS(1480),
    [anon_sym_try] = ACTIONS(1480),
    [anon_sym_with] = ACTIONS(1480),
    [anon_sym_STAR] = ACTIONS(1480),
    [anon_sym_STAR_STAR] = ACTIONS(1480),
    [anon_sym_class] = ACTIONS(1480),
    [anon_sym_AT] = ACTIONS(1480),
    [anon_sym_not] = ACTIONS(1480),
    [anon_sym_and] = ACTIONS(1480),
    [anon_sym_or] = ACTIONS(1480),
    [anon_sym_PLUS] = ACTIONS(1480),
    [anon_sym_DASH] = ACTIONS(1480),
    [anon_sym_SLASH] = ACTIONS(1480),
    [anon_sym_SLASH_SLASH] = ACTIONS(1480),
    [anon_sym_PERCENT] = ACTIONS(1480),
    [anon_sym_PIPE] = ACTIONS(1480),
    [anon_sym_CARET] = ACTIONS(1480),
    [anon_sym_AMP] = ACTIONS(1480),
    [anon_sym_LT_LT] = ACTIONS(1480),
    [anon_sym_GT_GT] = ACTIONS(1480),
    [anon_sym_GT_GT_GT] = ACTIONS(1480),
    [anon_sym_TILDE] = ACTIONS(1480),
    [anon_sym_typeof] = ACTIONS(1480),
    [anon_sym_void] = ACTIONS(1480),
    [anon_sym_LT] = ACTIONS(1480),
    [anon_sym_LT_EQ] = ACTIONS(1480),
    [anon_sym_EQ_EQ] = ACTIONS(1480),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1480),
    [anon_sym_BANG_EQ] = ACTIONS(1480),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1480),
    [anon_sym_GT_EQ] = ACTIONS(1480),
    [anon_sym_GT] = ACTIONS(1480),
    [anon_sym_is] = ACTIONS(1480),
    [anon_sym_instanceof] = ACTIONS(1480),
    [anon_sym_new] = ACTIONS(1480),
    [anon_sym_QMARK] = ACTIONS(1480),
    [anon_sym_await] = ACTIONS(1480),
    [anon_sym_LBRACE] = ACTIONS(1480),
    [sym_this] = ACTIONS(1480),
    [sym_true] = ACTIONS(1480),
    [sym_false] = ACTIONS(1480),
    [sym_none] = ACTIONS(1480),
    [sym_string] = ACTIONS(1480),
    [sym_verbatim] = ACTIONS(1480),
    [sym_number] = ACTIONS(1480),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1478),
    [sym_fstring_start] = ACTIONS(1478),
  },
  [STATE(580)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1228),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1228),
    [anon_sym_SLASH_SLASH] = ACTIONS(1228),
    [anon_sym_PERCENT] = ACTIONS(1228),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(581)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1243),
    [anon_sym_SEMI] = ACTIONS(1243),
    [anon_sym_return] = ACTIONS(1243),
    [anon_sym_async] = ACTIONS(1243),
    [anon_sym_def] = ACTIONS(1243),
    [anon_sym_LPAREN] = ACTIONS(1243),
    [anon_sym_LBRACK] = ACTIONS(1243),
    [anon_sym_del] = ACTIONS(1243),
    [anon_sym_raise] = ACTIONS(1243),
    [sym_pass_statement] = ACTIONS(1243),
    [sym_break_statement] = ACTIONS(1243),
    [sym_continue_statement] = ACTIONS(1243),
    [sym_debugger_statement] = ACTIONS(1243),
    [anon_sym_assert] = ACTIONS(1243),
    [anon_sym_global] = ACTIONS(1243),
    [anon_sym_nonlocal] = ACTIONS(1243),
    [anon_sym_yield] = ACTIONS(1243),
    [anon_sym_from] = ACTIONS(1243),
    [anon_sym_import] = ACTIONS(1243),
    [anon_sym_DOT] = ACTIONS(1243),
    [anon_sym_if] = ACTIONS(1243),
    [anon_sym_for] = ACTIONS(1243),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1243),
    [anon_sym_do] = ACTIONS(1243),
    [anon_sym_try] = ACTIONS(1243),
    [anon_sym_with] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1243),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1243),
    [anon_sym_or] = ACTIONS(1243),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1243),
    [anon_sym_typeof] = ACTIONS(1243),
    [anon_sym_void] = ACTIONS(1243),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1243),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_await] = ACTIONS(1243),
    [anon_sym_LBRACE] = ACTIONS(1243),
    [sym_this] = ACTIONS(1243),
    [sym_true] = ACTIONS(1243),
    [sym_false] = ACTIONS(1243),
    [sym_none] = ACTIONS(1243),
    [sym_string] = ACTIONS(1243),
    [sym_verbatim] = ACTIONS(1243),
    [sym_number] = ACTIONS(1243),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1241),
    [sym_regex] = ACTIONS(1241),
    [sym_fstring_start] = ACTIONS(1241),
  },
  [STATE(582)] = {
    [ts_builtin_sym_end] = ACTIONS(1482),
    [sym_identifier] = ACTIONS(1484),
    [anon_sym_SEMI] = ACTIONS(1484),
    [anon_sym_return] = ACTIONS(1484),
    [anon_sym_async] = ACTIONS(1484),
    [anon_sym_def] = ACTIONS(1484),
    [anon_sym_COMMA] = ACTIONS(1484),
    [anon_sym_LPAREN] = ACTIONS(1484),
    [anon_sym_LBRACK] = ACTIONS(1484),
    [anon_sym_del] = ACTIONS(1484),
    [anon_sym_raise] = ACTIONS(1484),
    [sym_pass_statement] = ACTIONS(1484),
    [sym_break_statement] = ACTIONS(1484),
    [sym_continue_statement] = ACTIONS(1484),
    [sym_debugger_statement] = ACTIONS(1484),
    [anon_sym_assert] = ACTIONS(1484),
    [anon_sym_global] = ACTIONS(1484),
    [anon_sym_nonlocal] = ACTIONS(1484),
    [anon_sym_yield] = ACTIONS(1484),
    [anon_sym_from] = ACTIONS(1484),
    [anon_sym_import] = ACTIONS(1484),
    [anon_sym_DOT] = ACTIONS(1484),
    [anon_sym_if] = ACTIONS(1484),
    [anon_sym_for] = ACTIONS(1484),
    [anon_sym_in] = ACTIONS(1484),
    [anon_sym_while] = ACTIONS(1484),
    [anon_sym_do] = ACTIONS(1484),
    [anon_sym_try] = ACTIONS(1484),
    [anon_sym_with] = ACTIONS(1484),
    [anon_sym_STAR] = ACTIONS(1484),
    [anon_sym_STAR_STAR] = ACTIONS(1484),
    [anon_sym_class] = ACTIONS(1484),
    [anon_sym_AT] = ACTIONS(1484),
    [anon_sym_not] = ACTIONS(1484),
    [anon_sym_and] = ACTIONS(1484),
    [anon_sym_or] = ACTIONS(1484),
    [anon_sym_PLUS] = ACTIONS(1484),
    [anon_sym_DASH] = ACTIONS(1484),
    [anon_sym_SLASH] = ACTIONS(1484),
    [anon_sym_SLASH_SLASH] = ACTIONS(1484),
    [anon_sym_PERCENT] = ACTIONS(1484),
    [anon_sym_PIPE] = ACTIONS(1484),
    [anon_sym_CARET] = ACTIONS(1484),
    [anon_sym_AMP] = ACTIONS(1484),
    [anon_sym_LT_LT] = ACTIONS(1484),
    [anon_sym_GT_GT] = ACTIONS(1484),
    [anon_sym_GT_GT_GT] = ACTIONS(1484),
    [anon_sym_TILDE] = ACTIONS(1484),
    [anon_sym_typeof] = ACTIONS(1484),
    [anon_sym_void] = ACTIONS(1484),
    [anon_sym_LT] = ACTIONS(1484),
    [anon_sym_LT_EQ] = ACTIONS(1484),
    [anon_sym_EQ_EQ] = ACTIONS(1484),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1484),
    [anon_sym_BANG_EQ] = ACTIONS(1484),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1484),
    [anon_sym_GT_EQ] = ACTIONS(1484),
    [anon_sym_GT] = ACTIONS(1484),
    [anon_sym_is] = ACTIONS(1484),
    [anon_sym_instanceof] = ACTIONS(1484),
    [anon_sym_new] = ACTIONS(1484),
    [anon_sym_QMARK] = ACTIONS(1484),
    [anon_sym_await] = ACTIONS(1484),
    [anon_sym_LBRACE] = ACTIONS(1484),
    [sym_this] = ACTIONS(1484),
    [sym_true] = ACTIONS(1484),
    [sym_false] = ACTIONS(1484),
    [sym_none] = ACTIONS(1484),
    [sym_string] = ACTIONS(1484),
    [sym_verbatim] = ACTIONS(1484),
    [sym_number] = ACTIONS(1484),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1482),
    [sym_fstring_start] = ACTIONS(1482),
  },
  [STATE(583)] = {
    [ts_builtin_sym_end] = ACTIONS(1476),
    [sym_identifier] = ACTIONS(1474),
    [anon_sym_SEMI] = ACTIONS(1474),
    [anon_sym_return] = ACTIONS(1474),
    [anon_sym_async] = ACTIONS(1474),
    [anon_sym_def] = ACTIONS(1474),
    [anon_sym_COMMA] = ACTIONS(1474),
    [anon_sym_LPAREN] = ACTIONS(1474),
    [anon_sym_LBRACK] = ACTIONS(1474),
    [anon_sym_del] = ACTIONS(1474),
    [anon_sym_raise] = ACTIONS(1474),
    [sym_pass_statement] = ACTIONS(1474),
    [sym_break_statement] = ACTIONS(1474),
    [sym_continue_statement] = ACTIONS(1474),
    [sym_debugger_statement] = ACTIONS(1474),
    [anon_sym_assert] = ACTIONS(1474),
    [anon_sym_global] = ACTIONS(1474),
    [anon_sym_nonlocal] = ACTIONS(1474),
    [anon_sym_yield] = ACTIONS(1474),
    [anon_sym_from] = ACTIONS(1474),
    [anon_sym_import] = ACTIONS(1474),
    [anon_sym_DOT] = ACTIONS(1474),
    [anon_sym_if] = ACTIONS(1474),
    [anon_sym_for] = ACTIONS(1474),
    [anon_sym_in] = ACTIONS(1474),
    [anon_sym_while] = ACTIONS(1474),
    [anon_sym_do] = ACTIONS(1474),
    [anon_sym_try] = ACTIONS(1474),
    [anon_sym_with] = ACTIONS(1474),
    [anon_sym_STAR] = ACTIONS(1474),
    [anon_sym_STAR_STAR] = ACTIONS(1474),
    [anon_sym_class] = ACTIONS(1474),
    [anon_sym_AT] = ACTIONS(1474),
    [anon_sym_not] = ACTIONS(1474),
    [anon_sym_and] = ACTIONS(1474),
    [anon_sym_or] = ACTIONS(1474),
    [anon_sym_PLUS] = ACTIONS(1474),
    [anon_sym_DASH] = ACTIONS(1474),
    [anon_sym_SLASH] = ACTIONS(1474),
    [anon_sym_SLASH_SLASH] = ACTIONS(1474),
    [anon_sym_PERCENT] = ACTIONS(1474),
    [anon_sym_PIPE] = ACTIONS(1474),
    [anon_sym_CARET] = ACTIONS(1474),
    [anon_sym_AMP] = ACTIONS(1474),
    [anon_sym_LT_LT] = ACTIONS(1474),
    [anon_sym_GT_GT] = ACTIONS(1474),
    [anon_sym_GT_GT_GT] = ACTIONS(1474),
    [anon_sym_TILDE] = ACTIONS(1474),
    [anon_sym_typeof] = ACTIONS(1474),
    [anon_sym_void] = ACTIONS(1474),
    [anon_sym_LT] = ACTIONS(1474),
    [anon_sym_LT_EQ] = ACTIONS(1474),
    [anon_sym_EQ_EQ] = ACTIONS(1474),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1474),
    [anon_sym_BANG_EQ] = ACTIONS(1474),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1474),
    [anon_sym_GT_EQ] = ACTIONS(1474),
    [anon_sym_GT] = ACTIONS(1474),
    [anon_sym_is] = ACTIONS(1474),
    [anon_sym_instanceof] = ACTIONS(1474),
    [anon_sym_new] = ACTIONS(1474),
    [anon_sym_QMARK] = ACTIONS(1474),
    [anon_sym_await] = ACTIONS(1474),
    [anon_sym_LBRACE] = ACTIONS(1474),
    [sym_this] = ACTIONS(1474),
    [sym_true] = ACTIONS(1474),
    [sym_false] = ACTIONS(1474),
    [sym_none] = ACTIONS(1474),
    [sym_string] = ACTIONS(1474),
    [sym_verbatim] = ACTIONS(1474),
    [sym_number] = ACTIONS(1474),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1476),
    [sym_fstring_start] = ACTIONS(1476),
  },
  [STATE(584)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1295),
    [sym_identifier] = ACTIONS(1297),
    [anon_sym_SEMI] = ACTIONS(1297),
    [anon_sym_return] = ACTIONS(1297),
    [anon_sym_async] = ACTIONS(1297),
    [anon_sym_def] = ACTIONS(1297),
    [anon_sym_LPAREN] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1297),
    [anon_sym_del] = ACTIONS(1297),
    [anon_sym_raise] = ACTIONS(1297),
    [sym_pass_statement] = ACTIONS(1297),
    [sym_break_statement] = ACTIONS(1297),
    [sym_continue_statement] = ACTIONS(1297),
    [sym_debugger_statement] = ACTIONS(1297),
    [anon_sym_assert] = ACTIONS(1297),
    [anon_sym_global] = ACTIONS(1297),
    [anon_sym_nonlocal] = ACTIONS(1297),
    [anon_sym_yield] = ACTIONS(1297),
    [anon_sym_from] = ACTIONS(1297),
    [anon_sym_import] = ACTIONS(1297),
    [anon_sym_DOT] = ACTIONS(1297),
    [anon_sym_if] = ACTIONS(1297),
    [anon_sym_for] = ACTIONS(1297),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1297),
    [anon_sym_do] = ACTIONS(1297),
    [anon_sym_try] = ACTIONS(1297),
    [anon_sym_with] = ACTIONS(1297),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1297),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1494),
    [anon_sym_or] = ACTIONS(1496),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1297),
    [anon_sym_typeof] = ACTIONS(1297),
    [anon_sym_void] = ACTIONS(1297),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1297),
    [anon_sym_QMARK] = ACTIONS(1297),
    [anon_sym_await] = ACTIONS(1297),
    [anon_sym_LBRACE] = ACTIONS(1297),
    [sym_this] = ACTIONS(1297),
    [sym_true] = ACTIONS(1297),
    [sym_false] = ACTIONS(1297),
    [sym_none] = ACTIONS(1297),
    [sym_string] = ACTIONS(1297),
    [sym_verbatim] = ACTIONS(1297),
    [sym_number] = ACTIONS(1297),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1295),
    [sym_fstring_start] = ACTIONS(1295),
  },
  [STATE(585)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_async] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1270),
    [anon_sym_del] = ACTIONS(1270),
    [anon_sym_raise] = ACTIONS(1270),
    [sym_pass_statement] = ACTIONS(1270),
    [sym_break_statement] = ACTIONS(1270),
    [sym_continue_statement] = ACTIONS(1270),
    [sym_debugger_statement] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_global] = ACTIONS(1270),
    [anon_sym_nonlocal] = ACTIONS(1270),
    [anon_sym_yield] = ACTIONS(1270),
    [anon_sym_from] = ACTIONS(1270),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_DOT] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_with] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1270),
    [anon_sym_or] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1270),
    [anon_sym_typeof] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1270),
    [anon_sym_await] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1270),
    [sym_this] = ACTIONS(1270),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [sym_none] = ACTIONS(1270),
    [sym_string] = ACTIONS(1270),
    [sym_verbatim] = ACTIONS(1270),
    [sym_number] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1268),
    [sym_fstring_start] = ACTIONS(1268),
  },
  [STATE(586)] = {
    [ts_builtin_sym_end] = ACTIONS(1342),
    [sym_identifier] = ACTIONS(1344),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_return] = ACTIONS(1344),
    [anon_sym_async] = ACTIONS(1344),
    [anon_sym_def] = ACTIONS(1344),
    [anon_sym_COMMA] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_LBRACK] = ACTIONS(1344),
    [anon_sym_del] = ACTIONS(1344),
    [anon_sym_raise] = ACTIONS(1344),
    [sym_pass_statement] = ACTIONS(1344),
    [sym_break_statement] = ACTIONS(1344),
    [sym_continue_statement] = ACTIONS(1344),
    [sym_debugger_statement] = ACTIONS(1344),
    [anon_sym_assert] = ACTIONS(1344),
    [anon_sym_global] = ACTIONS(1344),
    [anon_sym_nonlocal] = ACTIONS(1344),
    [anon_sym_yield] = ACTIONS(1344),
    [anon_sym_from] = ACTIONS(1344),
    [anon_sym_import] = ACTIONS(1344),
    [anon_sym_DOT] = ACTIONS(1344),
    [anon_sym_if] = ACTIONS(1344),
    [anon_sym_for] = ACTIONS(1344),
    [anon_sym_in] = ACTIONS(1344),
    [anon_sym_while] = ACTIONS(1344),
    [anon_sym_do] = ACTIONS(1344),
    [anon_sym_try] = ACTIONS(1344),
    [anon_sym_with] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_STAR_STAR] = ACTIONS(1344),
    [anon_sym_class] = ACTIONS(1344),
    [anon_sym_AT] = ACTIONS(1344),
    [anon_sym_not] = ACTIONS(1344),
    [anon_sym_and] = ACTIONS(1344),
    [anon_sym_or] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_SLASH_SLASH] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_LT_LT] = ACTIONS(1344),
    [anon_sym_GT_GT] = ACTIONS(1344),
    [anon_sym_GT_GT_GT] = ACTIONS(1344),
    [anon_sym_TILDE] = ACTIONS(1344),
    [anon_sym_typeof] = ACTIONS(1344),
    [anon_sym_void] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_LT_EQ] = ACTIONS(1344),
    [anon_sym_EQ_EQ] = ACTIONS(1344),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1344),
    [anon_sym_GT_EQ] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_is] = ACTIONS(1344),
    [anon_sym_instanceof] = ACTIONS(1344),
    [anon_sym_new] = ACTIONS(1344),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_await] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [sym_this] = ACTIONS(1344),
    [sym_true] = ACTIONS(1344),
    [sym_false] = ACTIONS(1344),
    [sym_none] = ACTIONS(1344),
    [sym_string] = ACTIONS(1344),
    [sym_verbatim] = ACTIONS(1344),
    [sym_number] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1342),
    [sym_fstring_start] = ACTIONS(1342),
  },
  [STATE(587)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1243),
    [anon_sym_SEMI] = ACTIONS(1243),
    [anon_sym_return] = ACTIONS(1243),
    [anon_sym_async] = ACTIONS(1243),
    [anon_sym_def] = ACTIONS(1243),
    [anon_sym_LPAREN] = ACTIONS(1243),
    [anon_sym_LBRACK] = ACTIONS(1243),
    [anon_sym_del] = ACTIONS(1243),
    [anon_sym_raise] = ACTIONS(1243),
    [sym_pass_statement] = ACTIONS(1243),
    [sym_break_statement] = ACTIONS(1243),
    [sym_continue_statement] = ACTIONS(1243),
    [sym_debugger_statement] = ACTIONS(1243),
    [anon_sym_assert] = ACTIONS(1243),
    [anon_sym_global] = ACTIONS(1243),
    [anon_sym_nonlocal] = ACTIONS(1243),
    [anon_sym_yield] = ACTIONS(1243),
    [anon_sym_from] = ACTIONS(1243),
    [anon_sym_import] = ACTIONS(1243),
    [anon_sym_DOT] = ACTIONS(1243),
    [anon_sym_if] = ACTIONS(1243),
    [anon_sym_for] = ACTIONS(1243),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1243),
    [anon_sym_do] = ACTIONS(1243),
    [anon_sym_try] = ACTIONS(1243),
    [anon_sym_with] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1243),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1425),
    [anon_sym_or] = ACTIONS(1243),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1243),
    [anon_sym_typeof] = ACTIONS(1243),
    [anon_sym_void] = ACTIONS(1243),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1243),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_await] = ACTIONS(1243),
    [anon_sym_LBRACE] = ACTIONS(1243),
    [sym_this] = ACTIONS(1243),
    [sym_true] = ACTIONS(1243),
    [sym_false] = ACTIONS(1243),
    [sym_none] = ACTIONS(1243),
    [sym_string] = ACTIONS(1243),
    [sym_verbatim] = ACTIONS(1243),
    [sym_number] = ACTIONS(1243),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1241),
    [sym_regex] = ACTIONS(1241),
    [sym_fstring_start] = ACTIONS(1241),
  },
  [STATE(588)] = {
    [ts_builtin_sym_end] = ACTIONS(1192),
    [sym_identifier] = ACTIONS(1190),
    [anon_sym_SEMI] = ACTIONS(1190),
    [anon_sym_return] = ACTIONS(1190),
    [anon_sym_async] = ACTIONS(1190),
    [anon_sym_def] = ACTIONS(1190),
    [anon_sym_COMMA] = ACTIONS(1190),
    [anon_sym_LPAREN] = ACTIONS(1190),
    [anon_sym_LBRACK] = ACTIONS(1190),
    [anon_sym_del] = ACTIONS(1190),
    [anon_sym_raise] = ACTIONS(1190),
    [sym_pass_statement] = ACTIONS(1190),
    [sym_break_statement] = ACTIONS(1190),
    [sym_continue_statement] = ACTIONS(1190),
    [sym_debugger_statement] = ACTIONS(1190),
    [anon_sym_assert] = ACTIONS(1190),
    [anon_sym_global] = ACTIONS(1190),
    [anon_sym_nonlocal] = ACTIONS(1190),
    [anon_sym_yield] = ACTIONS(1190),
    [anon_sym_from] = ACTIONS(1190),
    [anon_sym_import] = ACTIONS(1190),
    [anon_sym_DOT] = ACTIONS(1190),
    [anon_sym_if] = ACTIONS(1190),
    [anon_sym_for] = ACTIONS(1190),
    [anon_sym_in] = ACTIONS(1190),
    [anon_sym_while] = ACTIONS(1190),
    [anon_sym_do] = ACTIONS(1190),
    [anon_sym_try] = ACTIONS(1190),
    [anon_sym_with] = ACTIONS(1190),
    [anon_sym_STAR] = ACTIONS(1190),
    [anon_sym_STAR_STAR] = ACTIONS(1190),
    [anon_sym_class] = ACTIONS(1190),
    [anon_sym_AT] = ACTIONS(1190),
    [anon_sym_not] = ACTIONS(1190),
    [anon_sym_and] = ACTIONS(1190),
    [anon_sym_or] = ACTIONS(1190),
    [anon_sym_PLUS] = ACTIONS(1190),
    [anon_sym_DASH] = ACTIONS(1190),
    [anon_sym_SLASH] = ACTIONS(1190),
    [anon_sym_SLASH_SLASH] = ACTIONS(1190),
    [anon_sym_PERCENT] = ACTIONS(1190),
    [anon_sym_PIPE] = ACTIONS(1190),
    [anon_sym_CARET] = ACTIONS(1190),
    [anon_sym_AMP] = ACTIONS(1190),
    [anon_sym_LT_LT] = ACTIONS(1190),
    [anon_sym_GT_GT] = ACTIONS(1190),
    [anon_sym_GT_GT_GT] = ACTIONS(1190),
    [anon_sym_TILDE] = ACTIONS(1190),
    [anon_sym_typeof] = ACTIONS(1190),
    [anon_sym_void] = ACTIONS(1190),
    [anon_sym_LT] = ACTIONS(1190),
    [anon_sym_LT_EQ] = ACTIONS(1190),
    [anon_sym_EQ_EQ] = ACTIONS(1190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1190),
    [anon_sym_BANG_EQ] = ACTIONS(1190),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1190),
    [anon_sym_GT_EQ] = ACTIONS(1190),
    [anon_sym_GT] = ACTIONS(1190),
    [anon_sym_is] = ACTIONS(1190),
    [anon_sym_instanceof] = ACTIONS(1190),
    [anon_sym_new] = ACTIONS(1190),
    [anon_sym_QMARK] = ACTIONS(1190),
    [anon_sym_await] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1190),
    [sym_this] = ACTIONS(1190),
    [sym_true] = ACTIONS(1190),
    [sym_false] = ACTIONS(1190),
    [sym_none] = ACTIONS(1190),
    [sym_string] = ACTIONS(1190),
    [sym_verbatim] = ACTIONS(1190),
    [sym_number] = ACTIONS(1190),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1192),
    [sym_fstring_start] = ACTIONS(1192),
  },
  [STATE(589)] = {
    [ts_builtin_sym_end] = ACTIONS(1510),
    [sym_identifier] = ACTIONS(1512),
    [anon_sym_SEMI] = ACTIONS(1512),
    [anon_sym_return] = ACTIONS(1512),
    [anon_sym_async] = ACTIONS(1512),
    [anon_sym_def] = ACTIONS(1512),
    [anon_sym_COMMA] = ACTIONS(1512),
    [anon_sym_LPAREN] = ACTIONS(1512),
    [anon_sym_LBRACK] = ACTIONS(1512),
    [anon_sym_del] = ACTIONS(1512),
    [anon_sym_raise] = ACTIONS(1512),
    [sym_pass_statement] = ACTIONS(1512),
    [sym_break_statement] = ACTIONS(1512),
    [sym_continue_statement] = ACTIONS(1512),
    [sym_debugger_statement] = ACTIONS(1512),
    [anon_sym_assert] = ACTIONS(1512),
    [anon_sym_global] = ACTIONS(1512),
    [anon_sym_nonlocal] = ACTIONS(1512),
    [anon_sym_yield] = ACTIONS(1512),
    [anon_sym_from] = ACTIONS(1512),
    [anon_sym_import] = ACTIONS(1512),
    [anon_sym_DOT] = ACTIONS(1512),
    [anon_sym_if] = ACTIONS(1512),
    [anon_sym_for] = ACTIONS(1512),
    [anon_sym_in] = ACTIONS(1512),
    [anon_sym_while] = ACTIONS(1512),
    [anon_sym_do] = ACTIONS(1512),
    [anon_sym_try] = ACTIONS(1512),
    [anon_sym_with] = ACTIONS(1512),
    [anon_sym_STAR] = ACTIONS(1512),
    [anon_sym_STAR_STAR] = ACTIONS(1512),
    [anon_sym_class] = ACTIONS(1512),
    [anon_sym_AT] = ACTIONS(1512),
    [anon_sym_not] = ACTIONS(1512),
    [anon_sym_and] = ACTIONS(1512),
    [anon_sym_or] = ACTIONS(1512),
    [anon_sym_PLUS] = ACTIONS(1512),
    [anon_sym_DASH] = ACTIONS(1512),
    [anon_sym_SLASH] = ACTIONS(1512),
    [anon_sym_SLASH_SLASH] = ACTIONS(1512),
    [anon_sym_PERCENT] = ACTIONS(1512),
    [anon_sym_PIPE] = ACTIONS(1512),
    [anon_sym_CARET] = ACTIONS(1512),
    [anon_sym_AMP] = ACTIONS(1512),
    [anon_sym_LT_LT] = ACTIONS(1512),
    [anon_sym_GT_GT] = ACTIONS(1512),
    [anon_sym_GT_GT_GT] = ACTIONS(1512),
    [anon_sym_TILDE] = ACTIONS(1512),
    [anon_sym_typeof] = ACTIONS(1512),
    [anon_sym_void] = ACTIONS(1512),
    [anon_sym_LT] = ACTIONS(1512),
    [anon_sym_LT_EQ] = ACTIONS(1512),
    [anon_sym_EQ_EQ] = ACTIONS(1512),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1512),
    [anon_sym_BANG_EQ] = ACTIONS(1512),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1512),
    [anon_sym_GT_EQ] = ACTIONS(1512),
    [anon_sym_GT] = ACTIONS(1512),
    [anon_sym_is] = ACTIONS(1512),
    [anon_sym_instanceof] = ACTIONS(1512),
    [anon_sym_new] = ACTIONS(1512),
    [anon_sym_QMARK] = ACTIONS(1512),
    [anon_sym_await] = ACTIONS(1512),
    [anon_sym_LBRACE] = ACTIONS(1512),
    [sym_this] = ACTIONS(1512),
    [sym_true] = ACTIONS(1512),
    [sym_false] = ACTIONS(1512),
    [sym_none] = ACTIONS(1512),
    [sym_string] = ACTIONS(1512),
    [sym_verbatim] = ACTIONS(1512),
    [sym_number] = ACTIONS(1512),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1510),
    [sym_fstring_start] = ACTIONS(1510),
  },
  [STATE(590)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1383),
    [sym_identifier] = ACTIONS(1385),
    [anon_sym_SEMI] = ACTIONS(1385),
    [anon_sym_return] = ACTIONS(1385),
    [anon_sym_async] = ACTIONS(1385),
    [anon_sym_def] = ACTIONS(1385),
    [anon_sym_LPAREN] = ACTIONS(1385),
    [anon_sym_LBRACK] = ACTIONS(1385),
    [anon_sym_del] = ACTIONS(1385),
    [anon_sym_raise] = ACTIONS(1385),
    [sym_pass_statement] = ACTIONS(1385),
    [sym_break_statement] = ACTIONS(1385),
    [sym_continue_statement] = ACTIONS(1385),
    [sym_debugger_statement] = ACTIONS(1385),
    [anon_sym_assert] = ACTIONS(1385),
    [anon_sym_global] = ACTIONS(1385),
    [anon_sym_nonlocal] = ACTIONS(1385),
    [anon_sym_yield] = ACTIONS(1385),
    [anon_sym_from] = ACTIONS(1385),
    [anon_sym_import] = ACTIONS(1385),
    [anon_sym_DOT] = ACTIONS(1385),
    [anon_sym_if] = ACTIONS(1385),
    [anon_sym_for] = ACTIONS(1385),
    [anon_sym_in] = ACTIONS(1385),
    [anon_sym_while] = ACTIONS(1385),
    [anon_sym_do] = ACTIONS(1385),
    [anon_sym_try] = ACTIONS(1385),
    [anon_sym_with] = ACTIONS(1385),
    [anon_sym_STAR] = ACTIONS(1385),
    [anon_sym_STAR_STAR] = ACTIONS(1385),
    [anon_sym_class] = ACTIONS(1385),
    [anon_sym_AT] = ACTIONS(1385),
    [anon_sym_not] = ACTIONS(1385),
    [anon_sym_and] = ACTIONS(1385),
    [anon_sym_or] = ACTIONS(1385),
    [anon_sym_PLUS] = ACTIONS(1385),
    [anon_sym_DASH] = ACTIONS(1385),
    [anon_sym_SLASH] = ACTIONS(1385),
    [anon_sym_SLASH_SLASH] = ACTIONS(1385),
    [anon_sym_PERCENT] = ACTIONS(1385),
    [anon_sym_PIPE] = ACTIONS(1385),
    [anon_sym_CARET] = ACTIONS(1385),
    [anon_sym_AMP] = ACTIONS(1385),
    [anon_sym_LT_LT] = ACTIONS(1385),
    [anon_sym_GT_GT] = ACTIONS(1385),
    [anon_sym_GT_GT_GT] = ACTIONS(1385),
    [anon_sym_TILDE] = ACTIONS(1385),
    [anon_sym_typeof] = ACTIONS(1385),
    [anon_sym_void] = ACTIONS(1385),
    [anon_sym_LT] = ACTIONS(1385),
    [anon_sym_LT_EQ] = ACTIONS(1385),
    [anon_sym_EQ_EQ] = ACTIONS(1385),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1385),
    [anon_sym_BANG_EQ] = ACTIONS(1385),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1385),
    [anon_sym_GT_EQ] = ACTIONS(1385),
    [anon_sym_GT] = ACTIONS(1385),
    [anon_sym_is] = ACTIONS(1385),
    [anon_sym_instanceof] = ACTIONS(1385),
    [anon_sym_new] = ACTIONS(1385),
    [anon_sym_QMARK] = ACTIONS(1385),
    [anon_sym_await] = ACTIONS(1385),
    [anon_sym_LBRACE] = ACTIONS(1385),
    [sym_this] = ACTIONS(1385),
    [sym_true] = ACTIONS(1385),
    [sym_false] = ACTIONS(1385),
    [sym_none] = ACTIONS(1385),
    [sym_string] = ACTIONS(1385),
    [sym_verbatim] = ACTIONS(1385),
    [sym_number] = ACTIONS(1385),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1383),
    [sym_fstring_start] = ACTIONS(1383),
  },
  [STATE(591)] = {
    [ts_builtin_sym_end] = ACTIONS(1514),
    [sym_identifier] = ACTIONS(1516),
    [anon_sym_SEMI] = ACTIONS(1516),
    [anon_sym_return] = ACTIONS(1516),
    [anon_sym_async] = ACTIONS(1516),
    [anon_sym_def] = ACTIONS(1516),
    [anon_sym_COMMA] = ACTIONS(1516),
    [anon_sym_LPAREN] = ACTIONS(1516),
    [anon_sym_LBRACK] = ACTIONS(1516),
    [anon_sym_del] = ACTIONS(1516),
    [anon_sym_raise] = ACTIONS(1516),
    [sym_pass_statement] = ACTIONS(1516),
    [sym_break_statement] = ACTIONS(1516),
    [sym_continue_statement] = ACTIONS(1516),
    [sym_debugger_statement] = ACTIONS(1516),
    [anon_sym_assert] = ACTIONS(1516),
    [anon_sym_global] = ACTIONS(1516),
    [anon_sym_nonlocal] = ACTIONS(1516),
    [anon_sym_yield] = ACTIONS(1516),
    [anon_sym_from] = ACTIONS(1516),
    [anon_sym_import] = ACTIONS(1516),
    [anon_sym_DOT] = ACTIONS(1516),
    [anon_sym_if] = ACTIONS(1516),
    [anon_sym_for] = ACTIONS(1516),
    [anon_sym_in] = ACTIONS(1516),
    [anon_sym_while] = ACTIONS(1516),
    [anon_sym_do] = ACTIONS(1516),
    [anon_sym_try] = ACTIONS(1516),
    [anon_sym_with] = ACTIONS(1516),
    [anon_sym_STAR] = ACTIONS(1516),
    [anon_sym_STAR_STAR] = ACTIONS(1516),
    [anon_sym_class] = ACTIONS(1516),
    [anon_sym_AT] = ACTIONS(1516),
    [anon_sym_not] = ACTIONS(1516),
    [anon_sym_and] = ACTIONS(1516),
    [anon_sym_or] = ACTIONS(1516),
    [anon_sym_PLUS] = ACTIONS(1516),
    [anon_sym_DASH] = ACTIONS(1516),
    [anon_sym_SLASH] = ACTIONS(1516),
    [anon_sym_SLASH_SLASH] = ACTIONS(1516),
    [anon_sym_PERCENT] = ACTIONS(1516),
    [anon_sym_PIPE] = ACTIONS(1516),
    [anon_sym_CARET] = ACTIONS(1516),
    [anon_sym_AMP] = ACTIONS(1516),
    [anon_sym_LT_LT] = ACTIONS(1516),
    [anon_sym_GT_GT] = ACTIONS(1516),
    [anon_sym_GT_GT_GT] = ACTIONS(1516),
    [anon_sym_TILDE] = ACTIONS(1516),
    [anon_sym_typeof] = ACTIONS(1516),
    [anon_sym_void] = ACTIONS(1516),
    [anon_sym_LT] = ACTIONS(1516),
    [anon_sym_LT_EQ] = ACTIONS(1516),
    [anon_sym_EQ_EQ] = ACTIONS(1516),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1516),
    [anon_sym_BANG_EQ] = ACTIONS(1516),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1516),
    [anon_sym_GT_EQ] = ACTIONS(1516),
    [anon_sym_GT] = ACTIONS(1516),
    [anon_sym_is] = ACTIONS(1516),
    [anon_sym_instanceof] = ACTIONS(1516),
    [anon_sym_new] = ACTIONS(1516),
    [anon_sym_QMARK] = ACTIONS(1516),
    [anon_sym_await] = ACTIONS(1516),
    [anon_sym_LBRACE] = ACTIONS(1516),
    [sym_this] = ACTIONS(1516),
    [sym_true] = ACTIONS(1516),
    [sym_false] = ACTIONS(1516),
    [sym_none] = ACTIONS(1516),
    [sym_string] = ACTIONS(1516),
    [sym_verbatim] = ACTIONS(1516),
    [sym_number] = ACTIONS(1516),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1514),
    [sym_fstring_start] = ACTIONS(1514),
  },
  [STATE(592)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1258),
    [sym_identifier] = ACTIONS(1260),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_return] = ACTIONS(1260),
    [anon_sym_async] = ACTIONS(1260),
    [anon_sym_def] = ACTIONS(1260),
    [anon_sym_LPAREN] = ACTIONS(1260),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [anon_sym_del] = ACTIONS(1260),
    [anon_sym_raise] = ACTIONS(1260),
    [sym_pass_statement] = ACTIONS(1260),
    [sym_break_statement] = ACTIONS(1260),
    [sym_continue_statement] = ACTIONS(1260),
    [sym_debugger_statement] = ACTIONS(1260),
    [anon_sym_assert] = ACTIONS(1260),
    [anon_sym_global] = ACTIONS(1260),
    [anon_sym_nonlocal] = ACTIONS(1260),
    [anon_sym_yield] = ACTIONS(1260),
    [anon_sym_from] = ACTIONS(1260),
    [anon_sym_import] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1260),
    [anon_sym_if] = ACTIONS(1518),
    [anon_sym_for] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1260),
    [anon_sym_do] = ACTIONS(1260),
    [anon_sym_try] = ACTIONS(1260),
    [anon_sym_with] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1260),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1494),
    [anon_sym_or] = ACTIONS(1496),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_typeof] = ACTIONS(1260),
    [anon_sym_void] = ACTIONS(1260),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1260),
    [anon_sym_QMARK] = ACTIONS(1260),
    [anon_sym_await] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [sym_this] = ACTIONS(1260),
    [sym_true] = ACTIONS(1260),
    [sym_false] = ACTIONS(1260),
    [sym_none] = ACTIONS(1260),
    [sym_string] = ACTIONS(1260),
    [sym_verbatim] = ACTIONS(1260),
    [sym_number] = ACTIONS(1260),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1258),
    [sym_fstring_start] = ACTIONS(1258),
  },
  [STATE(593)] = {
    [sym_identifier] = ACTIONS(1393),
    [anon_sym_SEMI] = ACTIONS(1393),
    [anon_sym_return] = ACTIONS(1393),
    [anon_sym_async] = ACTIONS(1393),
    [anon_sym_def] = ACTIONS(1393),
    [anon_sym_COMMA] = ACTIONS(1393),
    [anon_sym_LPAREN] = ACTIONS(1393),
    [anon_sym_LBRACK] = ACTIONS(1393),
    [anon_sym_del] = ACTIONS(1393),
    [anon_sym_raise] = ACTIONS(1393),
    [sym_pass_statement] = ACTIONS(1393),
    [sym_break_statement] = ACTIONS(1393),
    [sym_continue_statement] = ACTIONS(1393),
    [sym_debugger_statement] = ACTIONS(1393),
    [anon_sym_assert] = ACTIONS(1393),
    [anon_sym_global] = ACTIONS(1393),
    [anon_sym_nonlocal] = ACTIONS(1393),
    [anon_sym_yield] = ACTIONS(1393),
    [anon_sym_from] = ACTIONS(1393),
    [anon_sym_import] = ACTIONS(1393),
    [anon_sym_DOT] = ACTIONS(1393),
    [anon_sym_if] = ACTIONS(1393),
    [anon_sym_for] = ACTIONS(1393),
    [anon_sym_in] = ACTIONS(1393),
    [anon_sym_while] = ACTIONS(1393),
    [anon_sym_do] = ACTIONS(1393),
    [anon_sym_try] = ACTIONS(1393),
    [anon_sym_with] = ACTIONS(1393),
    [anon_sym_STAR] = ACTIONS(1393),
    [anon_sym_STAR_STAR] = ACTIONS(1393),
    [anon_sym_class] = ACTIONS(1393),
    [anon_sym_AT] = ACTIONS(1393),
    [anon_sym_not] = ACTIONS(1393),
    [anon_sym_and] = ACTIONS(1393),
    [anon_sym_or] = ACTIONS(1393),
    [anon_sym_PLUS] = ACTIONS(1393),
    [anon_sym_DASH] = ACTIONS(1393),
    [anon_sym_SLASH] = ACTIONS(1393),
    [anon_sym_SLASH_SLASH] = ACTIONS(1393),
    [anon_sym_PERCENT] = ACTIONS(1393),
    [anon_sym_PIPE] = ACTIONS(1393),
    [anon_sym_CARET] = ACTIONS(1393),
    [anon_sym_AMP] = ACTIONS(1393),
    [anon_sym_LT_LT] = ACTIONS(1393),
    [anon_sym_GT_GT] = ACTIONS(1393),
    [anon_sym_GT_GT_GT] = ACTIONS(1393),
    [anon_sym_TILDE] = ACTIONS(1393),
    [anon_sym_typeof] = ACTIONS(1393),
    [anon_sym_void] = ACTIONS(1393),
    [anon_sym_LT] = ACTIONS(1393),
    [anon_sym_LT_EQ] = ACTIONS(1393),
    [anon_sym_EQ_EQ] = ACTIONS(1393),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1393),
    [anon_sym_BANG_EQ] = ACTIONS(1393),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1393),
    [anon_sym_GT_EQ] = ACTIONS(1393),
    [anon_sym_GT] = ACTIONS(1393),
    [anon_sym_is] = ACTIONS(1393),
    [anon_sym_instanceof] = ACTIONS(1393),
    [anon_sym_new] = ACTIONS(1393),
    [anon_sym_QMARK] = ACTIONS(1393),
    [anon_sym_await] = ACTIONS(1393),
    [anon_sym_LBRACE] = ACTIONS(1393),
    [sym_this] = ACTIONS(1393),
    [sym_true] = ACTIONS(1393),
    [sym_false] = ACTIONS(1393),
    [sym_none] = ACTIONS(1393),
    [sym_string] = ACTIONS(1393),
    [sym_verbatim] = ACTIONS(1393),
    [sym_number] = ACTIONS(1393),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1391),
    [sym_regex] = ACTIONS(1391),
    [sym_fstring_start] = ACTIONS(1391),
  },
  [STATE(594)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1264),
    [sym_identifier] = ACTIONS(1266),
    [anon_sym_SEMI] = ACTIONS(1266),
    [anon_sym_return] = ACTIONS(1266),
    [anon_sym_async] = ACTIONS(1266),
    [anon_sym_def] = ACTIONS(1266),
    [anon_sym_LPAREN] = ACTIONS(1266),
    [anon_sym_LBRACK] = ACTIONS(1266),
    [anon_sym_del] = ACTIONS(1266),
    [anon_sym_raise] = ACTIONS(1266),
    [sym_pass_statement] = ACTIONS(1266),
    [sym_break_statement] = ACTIONS(1266),
    [sym_continue_statement] = ACTIONS(1266),
    [sym_debugger_statement] = ACTIONS(1266),
    [anon_sym_assert] = ACTIONS(1266),
    [anon_sym_global] = ACTIONS(1266),
    [anon_sym_nonlocal] = ACTIONS(1266),
    [anon_sym_yield] = ACTIONS(1266),
    [anon_sym_from] = ACTIONS(1266),
    [anon_sym_import] = ACTIONS(1266),
    [anon_sym_DOT] = ACTIONS(1266),
    [anon_sym_if] = ACTIONS(1266),
    [anon_sym_for] = ACTIONS(1266),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1266),
    [anon_sym_do] = ACTIONS(1266),
    [anon_sym_try] = ACTIONS(1266),
    [anon_sym_with] = ACTIONS(1266),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1266),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1494),
    [anon_sym_or] = ACTIONS(1496),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1266),
    [anon_sym_typeof] = ACTIONS(1266),
    [anon_sym_void] = ACTIONS(1266),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(1266),
    [anon_sym_await] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1266),
    [sym_this] = ACTIONS(1266),
    [sym_true] = ACTIONS(1266),
    [sym_false] = ACTIONS(1266),
    [sym_none] = ACTIONS(1266),
    [sym_string] = ACTIONS(1266),
    [sym_verbatim] = ACTIONS(1266),
    [sym_number] = ACTIONS(1266),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1264),
    [sym_fstring_start] = ACTIONS(1264),
  },
  [STATE(595)] = {
    [sym_identifier] = ACTIONS(1512),
    [anon_sym_SEMI] = ACTIONS(1512),
    [anon_sym_return] = ACTIONS(1512),
    [anon_sym_async] = ACTIONS(1512),
    [anon_sym_def] = ACTIONS(1512),
    [anon_sym_COMMA] = ACTIONS(1512),
    [anon_sym_LPAREN] = ACTIONS(1512),
    [anon_sym_LBRACK] = ACTIONS(1512),
    [anon_sym_del] = ACTIONS(1512),
    [anon_sym_raise] = ACTIONS(1512),
    [sym_pass_statement] = ACTIONS(1512),
    [sym_break_statement] = ACTIONS(1512),
    [sym_continue_statement] = ACTIONS(1512),
    [sym_debugger_statement] = ACTIONS(1512),
    [anon_sym_assert] = ACTIONS(1512),
    [anon_sym_global] = ACTIONS(1512),
    [anon_sym_nonlocal] = ACTIONS(1512),
    [anon_sym_yield] = ACTIONS(1512),
    [anon_sym_from] = ACTIONS(1512),
    [anon_sym_import] = ACTIONS(1512),
    [anon_sym_DOT] = ACTIONS(1512),
    [anon_sym_if] = ACTIONS(1512),
    [anon_sym_for] = ACTIONS(1512),
    [anon_sym_in] = ACTIONS(1512),
    [anon_sym_while] = ACTIONS(1512),
    [anon_sym_do] = ACTIONS(1512),
    [anon_sym_try] = ACTIONS(1512),
    [anon_sym_with] = ACTIONS(1512),
    [anon_sym_STAR] = ACTIONS(1512),
    [anon_sym_STAR_STAR] = ACTIONS(1512),
    [anon_sym_class] = ACTIONS(1512),
    [anon_sym_AT] = ACTIONS(1512),
    [anon_sym_not] = ACTIONS(1512),
    [anon_sym_and] = ACTIONS(1512),
    [anon_sym_or] = ACTIONS(1512),
    [anon_sym_PLUS] = ACTIONS(1512),
    [anon_sym_DASH] = ACTIONS(1512),
    [anon_sym_SLASH] = ACTIONS(1512),
    [anon_sym_SLASH_SLASH] = ACTIONS(1512),
    [anon_sym_PERCENT] = ACTIONS(1512),
    [anon_sym_PIPE] = ACTIONS(1512),
    [anon_sym_CARET] = ACTIONS(1512),
    [anon_sym_AMP] = ACTIONS(1512),
    [anon_sym_LT_LT] = ACTIONS(1512),
    [anon_sym_GT_GT] = ACTIONS(1512),
    [anon_sym_GT_GT_GT] = ACTIONS(1512),
    [anon_sym_TILDE] = ACTIONS(1512),
    [anon_sym_typeof] = ACTIONS(1512),
    [anon_sym_void] = ACTIONS(1512),
    [anon_sym_LT] = ACTIONS(1512),
    [anon_sym_LT_EQ] = ACTIONS(1512),
    [anon_sym_EQ_EQ] = ACTIONS(1512),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1512),
    [anon_sym_BANG_EQ] = ACTIONS(1512),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1512),
    [anon_sym_GT_EQ] = ACTIONS(1512),
    [anon_sym_GT] = ACTIONS(1512),
    [anon_sym_is] = ACTIONS(1512),
    [anon_sym_instanceof] = ACTIONS(1512),
    [anon_sym_new] = ACTIONS(1512),
    [anon_sym_QMARK] = ACTIONS(1512),
    [anon_sym_await] = ACTIONS(1512),
    [anon_sym_LBRACE] = ACTIONS(1512),
    [sym_this] = ACTIONS(1512),
    [sym_true] = ACTIONS(1512),
    [sym_false] = ACTIONS(1512),
    [sym_none] = ACTIONS(1512),
    [sym_string] = ACTIONS(1512),
    [sym_verbatim] = ACTIONS(1512),
    [sym_number] = ACTIONS(1512),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1510),
    [sym_regex] = ACTIONS(1510),
    [sym_fstring_start] = ACTIONS(1510),
  },
  [STATE(596)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(597)] = {
    [aux_sym__inline_body_repeat1] = STATE(621),
    [ts_builtin_sym_end] = ACTIONS(109),
    [sym_identifier] = ACTIONS(73),
    [anon_sym_SEMI] = ACTIONS(1520),
    [anon_sym_return] = ACTIONS(73),
    [anon_sym_async] = ACTIONS(73),
    [anon_sym_def] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(73),
    [anon_sym_LBRACK] = ACTIONS(73),
    [anon_sym_del] = ACTIONS(73),
    [anon_sym_raise] = ACTIONS(73),
    [sym_pass_statement] = ACTIONS(73),
    [sym_break_statement] = ACTIONS(73),
    [sym_continue_statement] = ACTIONS(73),
    [sym_debugger_statement] = ACTIONS(73),
    [anon_sym_assert] = ACTIONS(73),
    [anon_sym_global] = ACTIONS(73),
    [anon_sym_nonlocal] = ACTIONS(73),
    [anon_sym_yield] = ACTIONS(73),
    [anon_sym_from] = ACTIONS(73),
    [anon_sym_import] = ACTIONS(73),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_while] = ACTIONS(73),
    [anon_sym_do] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(73),
    [anon_sym_with] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_class] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(73),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(73),
    [anon_sym_DASH] = ACTIONS(73),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(73),
    [anon_sym_typeof] = ACTIONS(73),
    [anon_sym_void] = ACTIONS(73),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(73),
    [anon_sym_LBRACE] = ACTIONS(73),
    [sym_this] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_none] = ACTIONS(73),
    [sym_string] = ACTIONS(73),
    [sym_verbatim] = ACTIONS(73),
    [sym_number] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(109),
    [sym_fstring_start] = ACTIONS(109),
  },
  [STATE(598)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1299),
    [sym_identifier] = ACTIONS(1301),
    [anon_sym_SEMI] = ACTIONS(1301),
    [anon_sym_return] = ACTIONS(1301),
    [anon_sym_async] = ACTIONS(1301),
    [anon_sym_def] = ACTIONS(1301),
    [anon_sym_LPAREN] = ACTIONS(1301),
    [anon_sym_LBRACK] = ACTIONS(1301),
    [anon_sym_del] = ACTIONS(1301),
    [anon_sym_raise] = ACTIONS(1301),
    [sym_pass_statement] = ACTIONS(1301),
    [sym_break_statement] = ACTIONS(1301),
    [sym_continue_statement] = ACTIONS(1301),
    [sym_debugger_statement] = ACTIONS(1301),
    [anon_sym_assert] = ACTIONS(1301),
    [anon_sym_global] = ACTIONS(1301),
    [anon_sym_nonlocal] = ACTIONS(1301),
    [anon_sym_yield] = ACTIONS(1301),
    [anon_sym_from] = ACTIONS(1301),
    [anon_sym_import] = ACTIONS(1301),
    [anon_sym_DOT] = ACTIONS(1301),
    [anon_sym_if] = ACTIONS(1301),
    [anon_sym_for] = ACTIONS(1301),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1301),
    [anon_sym_do] = ACTIONS(1301),
    [anon_sym_try] = ACTIONS(1301),
    [anon_sym_with] = ACTIONS(1301),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1301),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1494),
    [anon_sym_or] = ACTIONS(1496),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1301),
    [anon_sym_typeof] = ACTIONS(1301),
    [anon_sym_void] = ACTIONS(1301),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1301),
    [anon_sym_QMARK] = ACTIONS(1301),
    [anon_sym_await] = ACTIONS(1301),
    [anon_sym_LBRACE] = ACTIONS(1301),
    [sym_this] = ACTIONS(1301),
    [sym_true] = ACTIONS(1301),
    [sym_false] = ACTIONS(1301),
    [sym_none] = ACTIONS(1301),
    [sym_string] = ACTIONS(1301),
    [sym_verbatim] = ACTIONS(1301),
    [sym_number] = ACTIONS(1301),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1299),
    [sym_fstring_start] = ACTIONS(1299),
  },
  [STATE(599)] = {
    [ts_builtin_sym_end] = ACTIONS(1522),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_SEMI] = ACTIONS(1524),
    [anon_sym_return] = ACTIONS(1524),
    [anon_sym_async] = ACTIONS(1524),
    [anon_sym_def] = ACTIONS(1524),
    [anon_sym_COMMA] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1524),
    [anon_sym_LBRACK] = ACTIONS(1524),
    [anon_sym_del] = ACTIONS(1524),
    [anon_sym_raise] = ACTIONS(1524),
    [sym_pass_statement] = ACTIONS(1524),
    [sym_break_statement] = ACTIONS(1524),
    [sym_continue_statement] = ACTIONS(1524),
    [sym_debugger_statement] = ACTIONS(1524),
    [anon_sym_assert] = ACTIONS(1524),
    [anon_sym_global] = ACTIONS(1524),
    [anon_sym_nonlocal] = ACTIONS(1524),
    [anon_sym_yield] = ACTIONS(1524),
    [anon_sym_from] = ACTIONS(1524),
    [anon_sym_import] = ACTIONS(1524),
    [anon_sym_DOT] = ACTIONS(1524),
    [anon_sym_if] = ACTIONS(1524),
    [anon_sym_for] = ACTIONS(1524),
    [anon_sym_in] = ACTIONS(1524),
    [anon_sym_while] = ACTIONS(1524),
    [anon_sym_do] = ACTIONS(1524),
    [anon_sym_try] = ACTIONS(1524),
    [anon_sym_with] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1524),
    [anon_sym_STAR_STAR] = ACTIONS(1524),
    [anon_sym_class] = ACTIONS(1524),
    [anon_sym_AT] = ACTIONS(1524),
    [anon_sym_not] = ACTIONS(1524),
    [anon_sym_and] = ACTIONS(1524),
    [anon_sym_or] = ACTIONS(1524),
    [anon_sym_PLUS] = ACTIONS(1524),
    [anon_sym_DASH] = ACTIONS(1524),
    [anon_sym_SLASH] = ACTIONS(1524),
    [anon_sym_SLASH_SLASH] = ACTIONS(1524),
    [anon_sym_PERCENT] = ACTIONS(1524),
    [anon_sym_PIPE] = ACTIONS(1524),
    [anon_sym_CARET] = ACTIONS(1524),
    [anon_sym_AMP] = ACTIONS(1524),
    [anon_sym_LT_LT] = ACTIONS(1524),
    [anon_sym_GT_GT] = ACTIONS(1524),
    [anon_sym_GT_GT_GT] = ACTIONS(1524),
    [anon_sym_TILDE] = ACTIONS(1524),
    [anon_sym_typeof] = ACTIONS(1524),
    [anon_sym_void] = ACTIONS(1524),
    [anon_sym_LT] = ACTIONS(1524),
    [anon_sym_LT_EQ] = ACTIONS(1524),
    [anon_sym_EQ_EQ] = ACTIONS(1524),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1524),
    [anon_sym_BANG_EQ] = ACTIONS(1524),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1524),
    [anon_sym_GT_EQ] = ACTIONS(1524),
    [anon_sym_GT] = ACTIONS(1524),
    [anon_sym_is] = ACTIONS(1524),
    [anon_sym_instanceof] = ACTIONS(1524),
    [anon_sym_new] = ACTIONS(1524),
    [anon_sym_QMARK] = ACTIONS(1524),
    [anon_sym_await] = ACTIONS(1524),
    [anon_sym_LBRACE] = ACTIONS(1524),
    [sym_this] = ACTIONS(1524),
    [sym_true] = ACTIONS(1524),
    [sym_false] = ACTIONS(1524),
    [sym_none] = ACTIONS(1524),
    [sym_string] = ACTIONS(1524),
    [sym_verbatim] = ACTIONS(1524),
    [sym_number] = ACTIONS(1524),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1522),
    [sym_fstring_start] = ACTIONS(1522),
  },
  [STATE(600)] = {
    [ts_builtin_sym_end] = ACTIONS(1526),
    [sym_identifier] = ACTIONS(1528),
    [anon_sym_SEMI] = ACTIONS(1528),
    [anon_sym_return] = ACTIONS(1528),
    [anon_sym_async] = ACTIONS(1528),
    [anon_sym_def] = ACTIONS(1528),
    [anon_sym_COMMA] = ACTIONS(1528),
    [anon_sym_LPAREN] = ACTIONS(1528),
    [anon_sym_LBRACK] = ACTIONS(1528),
    [anon_sym_del] = ACTIONS(1528),
    [anon_sym_raise] = ACTIONS(1528),
    [sym_pass_statement] = ACTIONS(1528),
    [sym_break_statement] = ACTIONS(1528),
    [sym_continue_statement] = ACTIONS(1528),
    [sym_debugger_statement] = ACTIONS(1528),
    [anon_sym_assert] = ACTIONS(1528),
    [anon_sym_global] = ACTIONS(1528),
    [anon_sym_nonlocal] = ACTIONS(1528),
    [anon_sym_yield] = ACTIONS(1528),
    [anon_sym_from] = ACTIONS(1528),
    [anon_sym_import] = ACTIONS(1528),
    [anon_sym_DOT] = ACTIONS(1528),
    [anon_sym_if] = ACTIONS(1528),
    [anon_sym_for] = ACTIONS(1528),
    [anon_sym_in] = ACTIONS(1528),
    [anon_sym_while] = ACTIONS(1528),
    [anon_sym_do] = ACTIONS(1528),
    [anon_sym_try] = ACTIONS(1528),
    [anon_sym_with] = ACTIONS(1528),
    [anon_sym_STAR] = ACTIONS(1528),
    [anon_sym_STAR_STAR] = ACTIONS(1528),
    [anon_sym_class] = ACTIONS(1528),
    [anon_sym_AT] = ACTIONS(1528),
    [anon_sym_not] = ACTIONS(1528),
    [anon_sym_and] = ACTIONS(1528),
    [anon_sym_or] = ACTIONS(1528),
    [anon_sym_PLUS] = ACTIONS(1528),
    [anon_sym_DASH] = ACTIONS(1528),
    [anon_sym_SLASH] = ACTIONS(1528),
    [anon_sym_SLASH_SLASH] = ACTIONS(1528),
    [anon_sym_PERCENT] = ACTIONS(1528),
    [anon_sym_PIPE] = ACTIONS(1528),
    [anon_sym_CARET] = ACTIONS(1528),
    [anon_sym_AMP] = ACTIONS(1528),
    [anon_sym_LT_LT] = ACTIONS(1528),
    [anon_sym_GT_GT] = ACTIONS(1528),
    [anon_sym_GT_GT_GT] = ACTIONS(1528),
    [anon_sym_TILDE] = ACTIONS(1528),
    [anon_sym_typeof] = ACTIONS(1528),
    [anon_sym_void] = ACTIONS(1528),
    [anon_sym_LT] = ACTIONS(1528),
    [anon_sym_LT_EQ] = ACTIONS(1528),
    [anon_sym_EQ_EQ] = ACTIONS(1528),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1528),
    [anon_sym_BANG_EQ] = ACTIONS(1528),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1528),
    [anon_sym_GT_EQ] = ACTIONS(1528),
    [anon_sym_GT] = ACTIONS(1528),
    [anon_sym_is] = ACTIONS(1528),
    [anon_sym_instanceof] = ACTIONS(1528),
    [anon_sym_new] = ACTIONS(1528),
    [anon_sym_QMARK] = ACTIONS(1528),
    [anon_sym_await] = ACTIONS(1528),
    [anon_sym_LBRACE] = ACTIONS(1528),
    [sym_this] = ACTIONS(1528),
    [sym_true] = ACTIONS(1528),
    [sym_false] = ACTIONS(1528),
    [sym_none] = ACTIONS(1528),
    [sym_string] = ACTIONS(1528),
    [sym_verbatim] = ACTIONS(1528),
    [sym_number] = ACTIONS(1528),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1526),
    [sym_fstring_start] = ACTIONS(1526),
  },
  [STATE(601)] = {
    [ts_builtin_sym_end] = ACTIONS(1530),
    [sym_identifier] = ACTIONS(1532),
    [anon_sym_SEMI] = ACTIONS(1532),
    [anon_sym_return] = ACTIONS(1532),
    [anon_sym_async] = ACTIONS(1532),
    [anon_sym_def] = ACTIONS(1532),
    [anon_sym_COMMA] = ACTIONS(1532),
    [anon_sym_LPAREN] = ACTIONS(1532),
    [anon_sym_LBRACK] = ACTIONS(1532),
    [anon_sym_del] = ACTIONS(1532),
    [anon_sym_raise] = ACTIONS(1532),
    [sym_pass_statement] = ACTIONS(1532),
    [sym_break_statement] = ACTIONS(1532),
    [sym_continue_statement] = ACTIONS(1532),
    [sym_debugger_statement] = ACTIONS(1532),
    [anon_sym_assert] = ACTIONS(1532),
    [anon_sym_global] = ACTIONS(1532),
    [anon_sym_nonlocal] = ACTIONS(1532),
    [anon_sym_yield] = ACTIONS(1532),
    [anon_sym_from] = ACTIONS(1532),
    [anon_sym_import] = ACTIONS(1532),
    [anon_sym_DOT] = ACTIONS(1532),
    [anon_sym_if] = ACTIONS(1532),
    [anon_sym_for] = ACTIONS(1532),
    [anon_sym_in] = ACTIONS(1532),
    [anon_sym_while] = ACTIONS(1532),
    [anon_sym_do] = ACTIONS(1532),
    [anon_sym_try] = ACTIONS(1532),
    [anon_sym_with] = ACTIONS(1532),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_STAR] = ACTIONS(1532),
    [anon_sym_class] = ACTIONS(1532),
    [anon_sym_AT] = ACTIONS(1532),
    [anon_sym_not] = ACTIONS(1532),
    [anon_sym_and] = ACTIONS(1532),
    [anon_sym_or] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1532),
    [anon_sym_DASH] = ACTIONS(1532),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_SLASH_SLASH] = ACTIONS(1532),
    [anon_sym_PERCENT] = ACTIONS(1532),
    [anon_sym_PIPE] = ACTIONS(1532),
    [anon_sym_CARET] = ACTIONS(1532),
    [anon_sym_AMP] = ACTIONS(1532),
    [anon_sym_LT_LT] = ACTIONS(1532),
    [anon_sym_GT_GT] = ACTIONS(1532),
    [anon_sym_GT_GT_GT] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1532),
    [anon_sym_typeof] = ACTIONS(1532),
    [anon_sym_void] = ACTIONS(1532),
    [anon_sym_LT] = ACTIONS(1532),
    [anon_sym_LT_EQ] = ACTIONS(1532),
    [anon_sym_EQ_EQ] = ACTIONS(1532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1532),
    [anon_sym_BANG_EQ] = ACTIONS(1532),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1532),
    [anon_sym_GT_EQ] = ACTIONS(1532),
    [anon_sym_GT] = ACTIONS(1532),
    [anon_sym_is] = ACTIONS(1532),
    [anon_sym_instanceof] = ACTIONS(1532),
    [anon_sym_new] = ACTIONS(1532),
    [anon_sym_QMARK] = ACTIONS(1532),
    [anon_sym_await] = ACTIONS(1532),
    [anon_sym_LBRACE] = ACTIONS(1532),
    [sym_this] = ACTIONS(1532),
    [sym_true] = ACTIONS(1532),
    [sym_false] = ACTIONS(1532),
    [sym_none] = ACTIONS(1532),
    [sym_string] = ACTIONS(1532),
    [sym_verbatim] = ACTIONS(1532),
    [sym_number] = ACTIONS(1532),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1530),
    [sym_fstring_start] = ACTIONS(1530),
  },
  [STATE(602)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_async] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1270),
    [anon_sym_del] = ACTIONS(1270),
    [anon_sym_raise] = ACTIONS(1270),
    [sym_pass_statement] = ACTIONS(1270),
    [sym_break_statement] = ACTIONS(1270),
    [sym_continue_statement] = ACTIONS(1270),
    [sym_debugger_statement] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_global] = ACTIONS(1270),
    [anon_sym_nonlocal] = ACTIONS(1270),
    [anon_sym_yield] = ACTIONS(1270),
    [anon_sym_from] = ACTIONS(1270),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_DOT] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_with] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1270),
    [anon_sym_or] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1270),
    [anon_sym_typeof] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1270),
    [anon_sym_await] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1270),
    [sym_this] = ACTIONS(1270),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [sym_none] = ACTIONS(1270),
    [sym_string] = ACTIONS(1270),
    [sym_verbatim] = ACTIONS(1270),
    [sym_number] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1268),
    [sym_regex] = ACTIONS(1268),
    [sym_fstring_start] = ACTIONS(1268),
  },
  [STATE(603)] = {
    [ts_builtin_sym_end] = ACTIONS(1534),
    [sym_identifier] = ACTIONS(1536),
    [anon_sym_SEMI] = ACTIONS(1536),
    [anon_sym_return] = ACTIONS(1536),
    [anon_sym_async] = ACTIONS(1536),
    [anon_sym_def] = ACTIONS(1536),
    [anon_sym_COMMA] = ACTIONS(1536),
    [anon_sym_LPAREN] = ACTIONS(1536),
    [anon_sym_LBRACK] = ACTIONS(1536),
    [anon_sym_del] = ACTIONS(1536),
    [anon_sym_raise] = ACTIONS(1536),
    [sym_pass_statement] = ACTIONS(1536),
    [sym_break_statement] = ACTIONS(1536),
    [sym_continue_statement] = ACTIONS(1536),
    [sym_debugger_statement] = ACTIONS(1536),
    [anon_sym_assert] = ACTIONS(1536),
    [anon_sym_global] = ACTIONS(1536),
    [anon_sym_nonlocal] = ACTIONS(1536),
    [anon_sym_yield] = ACTIONS(1536),
    [anon_sym_from] = ACTIONS(1536),
    [anon_sym_import] = ACTIONS(1536),
    [anon_sym_DOT] = ACTIONS(1536),
    [anon_sym_if] = ACTIONS(1536),
    [anon_sym_for] = ACTIONS(1536),
    [anon_sym_in] = ACTIONS(1536),
    [anon_sym_while] = ACTIONS(1536),
    [anon_sym_do] = ACTIONS(1536),
    [anon_sym_try] = ACTIONS(1536),
    [anon_sym_with] = ACTIONS(1536),
    [anon_sym_STAR] = ACTIONS(1536),
    [anon_sym_STAR_STAR] = ACTIONS(1536),
    [anon_sym_class] = ACTIONS(1536),
    [anon_sym_AT] = ACTIONS(1536),
    [anon_sym_not] = ACTIONS(1536),
    [anon_sym_and] = ACTIONS(1536),
    [anon_sym_or] = ACTIONS(1536),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_SLASH] = ACTIONS(1536),
    [anon_sym_SLASH_SLASH] = ACTIONS(1536),
    [anon_sym_PERCENT] = ACTIONS(1536),
    [anon_sym_PIPE] = ACTIONS(1536),
    [anon_sym_CARET] = ACTIONS(1536),
    [anon_sym_AMP] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1536),
    [anon_sym_GT_GT] = ACTIONS(1536),
    [anon_sym_GT_GT_GT] = ACTIONS(1536),
    [anon_sym_TILDE] = ACTIONS(1536),
    [anon_sym_typeof] = ACTIONS(1536),
    [anon_sym_void] = ACTIONS(1536),
    [anon_sym_LT] = ACTIONS(1536),
    [anon_sym_LT_EQ] = ACTIONS(1536),
    [anon_sym_EQ_EQ] = ACTIONS(1536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1536),
    [anon_sym_BANG_EQ] = ACTIONS(1536),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1536),
    [anon_sym_GT_EQ] = ACTIONS(1536),
    [anon_sym_GT] = ACTIONS(1536),
    [anon_sym_is] = ACTIONS(1536),
    [anon_sym_instanceof] = ACTIONS(1536),
    [anon_sym_new] = ACTIONS(1536),
    [anon_sym_QMARK] = ACTIONS(1536),
    [anon_sym_await] = ACTIONS(1536),
    [anon_sym_LBRACE] = ACTIONS(1536),
    [sym_this] = ACTIONS(1536),
    [sym_true] = ACTIONS(1536),
    [sym_false] = ACTIONS(1536),
    [sym_none] = ACTIONS(1536),
    [sym_string] = ACTIONS(1536),
    [sym_verbatim] = ACTIONS(1536),
    [sym_number] = ACTIONS(1536),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1534),
    [sym_fstring_start] = ACTIONS(1534),
  },
  [STATE(604)] = {
    [ts_builtin_sym_end] = ACTIONS(1538),
    [sym_identifier] = ACTIONS(1540),
    [anon_sym_SEMI] = ACTIONS(1540),
    [anon_sym_return] = ACTIONS(1540),
    [anon_sym_async] = ACTIONS(1540),
    [anon_sym_def] = ACTIONS(1540),
    [anon_sym_COMMA] = ACTIONS(1540),
    [anon_sym_LPAREN] = ACTIONS(1540),
    [anon_sym_LBRACK] = ACTIONS(1540),
    [anon_sym_del] = ACTIONS(1540),
    [anon_sym_raise] = ACTIONS(1540),
    [sym_pass_statement] = ACTIONS(1540),
    [sym_break_statement] = ACTIONS(1540),
    [sym_continue_statement] = ACTIONS(1540),
    [sym_debugger_statement] = ACTIONS(1540),
    [anon_sym_assert] = ACTIONS(1540),
    [anon_sym_global] = ACTIONS(1540),
    [anon_sym_nonlocal] = ACTIONS(1540),
    [anon_sym_yield] = ACTIONS(1540),
    [anon_sym_from] = ACTIONS(1540),
    [anon_sym_import] = ACTIONS(1540),
    [anon_sym_DOT] = ACTIONS(1540),
    [anon_sym_if] = ACTIONS(1540),
    [anon_sym_for] = ACTIONS(1540),
    [anon_sym_in] = ACTIONS(1540),
    [anon_sym_while] = ACTIONS(1540),
    [anon_sym_do] = ACTIONS(1540),
    [anon_sym_try] = ACTIONS(1540),
    [anon_sym_with] = ACTIONS(1540),
    [anon_sym_STAR] = ACTIONS(1540),
    [anon_sym_STAR_STAR] = ACTIONS(1540),
    [anon_sym_class] = ACTIONS(1540),
    [anon_sym_AT] = ACTIONS(1540),
    [anon_sym_not] = ACTIONS(1540),
    [anon_sym_and] = ACTIONS(1540),
    [anon_sym_or] = ACTIONS(1540),
    [anon_sym_PLUS] = ACTIONS(1540),
    [anon_sym_DASH] = ACTIONS(1540),
    [anon_sym_SLASH] = ACTIONS(1540),
    [anon_sym_SLASH_SLASH] = ACTIONS(1540),
    [anon_sym_PERCENT] = ACTIONS(1540),
    [anon_sym_PIPE] = ACTIONS(1540),
    [anon_sym_CARET] = ACTIONS(1540),
    [anon_sym_AMP] = ACTIONS(1540),
    [anon_sym_LT_LT] = ACTIONS(1540),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1540),
    [anon_sym_TILDE] = ACTIONS(1540),
    [anon_sym_typeof] = ACTIONS(1540),
    [anon_sym_void] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1540),
    [anon_sym_LT_EQ] = ACTIONS(1540),
    [anon_sym_EQ_EQ] = ACTIONS(1540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1540),
    [anon_sym_BANG_EQ] = ACTIONS(1540),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1540),
    [anon_sym_GT_EQ] = ACTIONS(1540),
    [anon_sym_GT] = ACTIONS(1540),
    [anon_sym_is] = ACTIONS(1540),
    [anon_sym_instanceof] = ACTIONS(1540),
    [anon_sym_new] = ACTIONS(1540),
    [anon_sym_QMARK] = ACTIONS(1540),
    [anon_sym_await] = ACTIONS(1540),
    [anon_sym_LBRACE] = ACTIONS(1540),
    [sym_this] = ACTIONS(1540),
    [sym_true] = ACTIONS(1540),
    [sym_false] = ACTIONS(1540),
    [sym_none] = ACTIONS(1540),
    [sym_string] = ACTIONS(1540),
    [sym_verbatim] = ACTIONS(1540),
    [sym_number] = ACTIONS(1540),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1538),
    [sym_fstring_start] = ACTIONS(1538),
  },
  [STATE(605)] = {
    [ts_builtin_sym_end] = ACTIONS(1542),
    [sym_identifier] = ACTIONS(1544),
    [anon_sym_SEMI] = ACTIONS(1544),
    [anon_sym_return] = ACTIONS(1544),
    [anon_sym_async] = ACTIONS(1544),
    [anon_sym_def] = ACTIONS(1544),
    [anon_sym_COMMA] = ACTIONS(1544),
    [anon_sym_LPAREN] = ACTIONS(1544),
    [anon_sym_LBRACK] = ACTIONS(1544),
    [anon_sym_del] = ACTIONS(1544),
    [anon_sym_raise] = ACTIONS(1544),
    [sym_pass_statement] = ACTIONS(1544),
    [sym_break_statement] = ACTIONS(1544),
    [sym_continue_statement] = ACTIONS(1544),
    [sym_debugger_statement] = ACTIONS(1544),
    [anon_sym_assert] = ACTIONS(1544),
    [anon_sym_global] = ACTIONS(1544),
    [anon_sym_nonlocal] = ACTIONS(1544),
    [anon_sym_yield] = ACTIONS(1544),
    [anon_sym_from] = ACTIONS(1544),
    [anon_sym_import] = ACTIONS(1544),
    [anon_sym_DOT] = ACTIONS(1544),
    [anon_sym_if] = ACTIONS(1544),
    [anon_sym_for] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1544),
    [anon_sym_while] = ACTIONS(1544),
    [anon_sym_do] = ACTIONS(1544),
    [anon_sym_try] = ACTIONS(1544),
    [anon_sym_with] = ACTIONS(1544),
    [anon_sym_STAR] = ACTIONS(1544),
    [anon_sym_STAR_STAR] = ACTIONS(1544),
    [anon_sym_class] = ACTIONS(1544),
    [anon_sym_AT] = ACTIONS(1544),
    [anon_sym_not] = ACTIONS(1544),
    [anon_sym_and] = ACTIONS(1544),
    [anon_sym_or] = ACTIONS(1544),
    [anon_sym_PLUS] = ACTIONS(1544),
    [anon_sym_DASH] = ACTIONS(1544),
    [anon_sym_SLASH] = ACTIONS(1544),
    [anon_sym_SLASH_SLASH] = ACTIONS(1544),
    [anon_sym_PERCENT] = ACTIONS(1544),
    [anon_sym_PIPE] = ACTIONS(1544),
    [anon_sym_CARET] = ACTIONS(1544),
    [anon_sym_AMP] = ACTIONS(1544),
    [anon_sym_LT_LT] = ACTIONS(1544),
    [anon_sym_GT_GT] = ACTIONS(1544),
    [anon_sym_GT_GT_GT] = ACTIONS(1544),
    [anon_sym_TILDE] = ACTIONS(1544),
    [anon_sym_typeof] = ACTIONS(1544),
    [anon_sym_void] = ACTIONS(1544),
    [anon_sym_LT] = ACTIONS(1544),
    [anon_sym_LT_EQ] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1544),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1544),
    [anon_sym_BANG_EQ] = ACTIONS(1544),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1544),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_GT] = ACTIONS(1544),
    [anon_sym_is] = ACTIONS(1544),
    [anon_sym_instanceof] = ACTIONS(1544),
    [anon_sym_new] = ACTIONS(1544),
    [anon_sym_QMARK] = ACTIONS(1544),
    [anon_sym_await] = ACTIONS(1544),
    [anon_sym_LBRACE] = ACTIONS(1544),
    [sym_this] = ACTIONS(1544),
    [sym_true] = ACTIONS(1544),
    [sym_false] = ACTIONS(1544),
    [sym_none] = ACTIONS(1544),
    [sym_string] = ACTIONS(1544),
    [sym_verbatim] = ACTIONS(1544),
    [sym_number] = ACTIONS(1544),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1542),
    [sym_fstring_start] = ACTIONS(1542),
  },
  [STATE(606)] = {
    [sym_identifier] = ACTIONS(1397),
    [anon_sym_SEMI] = ACTIONS(1397),
    [anon_sym_return] = ACTIONS(1397),
    [anon_sym_async] = ACTIONS(1397),
    [anon_sym_def] = ACTIONS(1397),
    [anon_sym_COMMA] = ACTIONS(1397),
    [anon_sym_LPAREN] = ACTIONS(1397),
    [anon_sym_LBRACK] = ACTIONS(1397),
    [anon_sym_del] = ACTIONS(1397),
    [anon_sym_raise] = ACTIONS(1397),
    [sym_pass_statement] = ACTIONS(1397),
    [sym_break_statement] = ACTIONS(1397),
    [sym_continue_statement] = ACTIONS(1397),
    [sym_debugger_statement] = ACTIONS(1397),
    [anon_sym_assert] = ACTIONS(1397),
    [anon_sym_global] = ACTIONS(1397),
    [anon_sym_nonlocal] = ACTIONS(1397),
    [anon_sym_yield] = ACTIONS(1397),
    [anon_sym_from] = ACTIONS(1397),
    [anon_sym_import] = ACTIONS(1397),
    [anon_sym_DOT] = ACTIONS(1397),
    [anon_sym_if] = ACTIONS(1397),
    [anon_sym_for] = ACTIONS(1397),
    [anon_sym_in] = ACTIONS(1397),
    [anon_sym_while] = ACTIONS(1397),
    [anon_sym_do] = ACTIONS(1397),
    [anon_sym_try] = ACTIONS(1397),
    [anon_sym_with] = ACTIONS(1397),
    [anon_sym_STAR] = ACTIONS(1397),
    [anon_sym_STAR_STAR] = ACTIONS(1397),
    [anon_sym_class] = ACTIONS(1397),
    [anon_sym_AT] = ACTIONS(1397),
    [anon_sym_not] = ACTIONS(1397),
    [anon_sym_and] = ACTIONS(1397),
    [anon_sym_or] = ACTIONS(1397),
    [anon_sym_PLUS] = ACTIONS(1397),
    [anon_sym_DASH] = ACTIONS(1397),
    [anon_sym_SLASH] = ACTIONS(1397),
    [anon_sym_SLASH_SLASH] = ACTIONS(1397),
    [anon_sym_PERCENT] = ACTIONS(1397),
    [anon_sym_PIPE] = ACTIONS(1397),
    [anon_sym_CARET] = ACTIONS(1397),
    [anon_sym_AMP] = ACTIONS(1397),
    [anon_sym_LT_LT] = ACTIONS(1397),
    [anon_sym_GT_GT] = ACTIONS(1397),
    [anon_sym_GT_GT_GT] = ACTIONS(1397),
    [anon_sym_TILDE] = ACTIONS(1397),
    [anon_sym_typeof] = ACTIONS(1397),
    [anon_sym_void] = ACTIONS(1397),
    [anon_sym_LT] = ACTIONS(1397),
    [anon_sym_LT_EQ] = ACTIONS(1397),
    [anon_sym_EQ_EQ] = ACTIONS(1397),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1397),
    [anon_sym_BANG_EQ] = ACTIONS(1397),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1397),
    [anon_sym_GT_EQ] = ACTIONS(1397),
    [anon_sym_GT] = ACTIONS(1397),
    [anon_sym_is] = ACTIONS(1397),
    [anon_sym_instanceof] = ACTIONS(1397),
    [anon_sym_new] = ACTIONS(1397),
    [anon_sym_QMARK] = ACTIONS(1397),
    [anon_sym_await] = ACTIONS(1397),
    [anon_sym_LBRACE] = ACTIONS(1397),
    [sym_this] = ACTIONS(1397),
    [sym_true] = ACTIONS(1397),
    [sym_false] = ACTIONS(1397),
    [sym_none] = ACTIONS(1397),
    [sym_string] = ACTIONS(1397),
    [sym_verbatim] = ACTIONS(1397),
    [sym_number] = ACTIONS(1397),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1395),
    [sym_regex] = ACTIONS(1395),
    [sym_fstring_start] = ACTIONS(1395),
  },
  [STATE(607)] = {
    [sym_identifier] = ACTIONS(1401),
    [anon_sym_SEMI] = ACTIONS(1401),
    [anon_sym_return] = ACTIONS(1401),
    [anon_sym_async] = ACTIONS(1401),
    [anon_sym_def] = ACTIONS(1401),
    [anon_sym_COMMA] = ACTIONS(1401),
    [anon_sym_LPAREN] = ACTIONS(1401),
    [anon_sym_LBRACK] = ACTIONS(1401),
    [anon_sym_del] = ACTIONS(1401),
    [anon_sym_raise] = ACTIONS(1401),
    [sym_pass_statement] = ACTIONS(1401),
    [sym_break_statement] = ACTIONS(1401),
    [sym_continue_statement] = ACTIONS(1401),
    [sym_debugger_statement] = ACTIONS(1401),
    [anon_sym_assert] = ACTIONS(1401),
    [anon_sym_global] = ACTIONS(1401),
    [anon_sym_nonlocal] = ACTIONS(1401),
    [anon_sym_yield] = ACTIONS(1401),
    [anon_sym_from] = ACTIONS(1401),
    [anon_sym_import] = ACTIONS(1401),
    [anon_sym_DOT] = ACTIONS(1401),
    [anon_sym_if] = ACTIONS(1401),
    [anon_sym_for] = ACTIONS(1401),
    [anon_sym_in] = ACTIONS(1401),
    [anon_sym_while] = ACTIONS(1401),
    [anon_sym_do] = ACTIONS(1401),
    [anon_sym_try] = ACTIONS(1401),
    [anon_sym_with] = ACTIONS(1401),
    [anon_sym_STAR] = ACTIONS(1401),
    [anon_sym_STAR_STAR] = ACTIONS(1401),
    [anon_sym_class] = ACTIONS(1401),
    [anon_sym_AT] = ACTIONS(1401),
    [anon_sym_not] = ACTIONS(1401),
    [anon_sym_and] = ACTIONS(1401),
    [anon_sym_or] = ACTIONS(1401),
    [anon_sym_PLUS] = ACTIONS(1401),
    [anon_sym_DASH] = ACTIONS(1401),
    [anon_sym_SLASH] = ACTIONS(1401),
    [anon_sym_SLASH_SLASH] = ACTIONS(1401),
    [anon_sym_PERCENT] = ACTIONS(1401),
    [anon_sym_PIPE] = ACTIONS(1401),
    [anon_sym_CARET] = ACTIONS(1401),
    [anon_sym_AMP] = ACTIONS(1401),
    [anon_sym_LT_LT] = ACTIONS(1401),
    [anon_sym_GT_GT] = ACTIONS(1401),
    [anon_sym_GT_GT_GT] = ACTIONS(1401),
    [anon_sym_TILDE] = ACTIONS(1401),
    [anon_sym_typeof] = ACTIONS(1401),
    [anon_sym_void] = ACTIONS(1401),
    [anon_sym_LT] = ACTIONS(1401),
    [anon_sym_LT_EQ] = ACTIONS(1401),
    [anon_sym_EQ_EQ] = ACTIONS(1401),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1401),
    [anon_sym_BANG_EQ] = ACTIONS(1401),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1401),
    [anon_sym_GT_EQ] = ACTIONS(1401),
    [anon_sym_GT] = ACTIONS(1401),
    [anon_sym_is] = ACTIONS(1401),
    [anon_sym_instanceof] = ACTIONS(1401),
    [anon_sym_new] = ACTIONS(1401),
    [anon_sym_QMARK] = ACTIONS(1401),
    [anon_sym_await] = ACTIONS(1401),
    [anon_sym_LBRACE] = ACTIONS(1401),
    [sym_this] = ACTIONS(1401),
    [sym_true] = ACTIONS(1401),
    [sym_false] = ACTIONS(1401),
    [sym_none] = ACTIONS(1401),
    [sym_string] = ACTIONS(1401),
    [sym_verbatim] = ACTIONS(1401),
    [sym_number] = ACTIONS(1401),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1399),
    [sym_regex] = ACTIONS(1399),
    [sym_fstring_start] = ACTIONS(1399),
  },
  [STATE(608)] = {
    [ts_builtin_sym_end] = ACTIONS(1546),
    [sym_identifier] = ACTIONS(1548),
    [anon_sym_SEMI] = ACTIONS(1548),
    [anon_sym_return] = ACTIONS(1548),
    [anon_sym_async] = ACTIONS(1548),
    [anon_sym_def] = ACTIONS(1548),
    [anon_sym_COMMA] = ACTIONS(1548),
    [anon_sym_LPAREN] = ACTIONS(1548),
    [anon_sym_LBRACK] = ACTIONS(1548),
    [anon_sym_del] = ACTIONS(1548),
    [anon_sym_raise] = ACTIONS(1548),
    [sym_pass_statement] = ACTIONS(1548),
    [sym_break_statement] = ACTIONS(1548),
    [sym_continue_statement] = ACTIONS(1548),
    [sym_debugger_statement] = ACTIONS(1548),
    [anon_sym_assert] = ACTIONS(1548),
    [anon_sym_global] = ACTIONS(1548),
    [anon_sym_nonlocal] = ACTIONS(1548),
    [anon_sym_yield] = ACTIONS(1548),
    [anon_sym_from] = ACTIONS(1548),
    [anon_sym_import] = ACTIONS(1548),
    [anon_sym_DOT] = ACTIONS(1548),
    [anon_sym_if] = ACTIONS(1548),
    [anon_sym_for] = ACTIONS(1548),
    [anon_sym_in] = ACTIONS(1548),
    [anon_sym_while] = ACTIONS(1548),
    [anon_sym_do] = ACTIONS(1548),
    [anon_sym_try] = ACTIONS(1548),
    [anon_sym_with] = ACTIONS(1548),
    [anon_sym_STAR] = ACTIONS(1548),
    [anon_sym_STAR_STAR] = ACTIONS(1548),
    [anon_sym_class] = ACTIONS(1548),
    [anon_sym_AT] = ACTIONS(1548),
    [anon_sym_not] = ACTIONS(1548),
    [anon_sym_and] = ACTIONS(1548),
    [anon_sym_or] = ACTIONS(1548),
    [anon_sym_PLUS] = ACTIONS(1548),
    [anon_sym_DASH] = ACTIONS(1548),
    [anon_sym_SLASH] = ACTIONS(1548),
    [anon_sym_SLASH_SLASH] = ACTIONS(1548),
    [anon_sym_PERCENT] = ACTIONS(1548),
    [anon_sym_PIPE] = ACTIONS(1548),
    [anon_sym_CARET] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1548),
    [anon_sym_LT_LT] = ACTIONS(1548),
    [anon_sym_GT_GT] = ACTIONS(1548),
    [anon_sym_GT_GT_GT] = ACTIONS(1548),
    [anon_sym_TILDE] = ACTIONS(1548),
    [anon_sym_typeof] = ACTIONS(1548),
    [anon_sym_void] = ACTIONS(1548),
    [anon_sym_LT] = ACTIONS(1548),
    [anon_sym_LT_EQ] = ACTIONS(1548),
    [anon_sym_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_GT_EQ] = ACTIONS(1548),
    [anon_sym_GT] = ACTIONS(1548),
    [anon_sym_is] = ACTIONS(1548),
    [anon_sym_instanceof] = ACTIONS(1548),
    [anon_sym_new] = ACTIONS(1548),
    [anon_sym_QMARK] = ACTIONS(1548),
    [anon_sym_await] = ACTIONS(1548),
    [anon_sym_LBRACE] = ACTIONS(1548),
    [sym_this] = ACTIONS(1548),
    [sym_true] = ACTIONS(1548),
    [sym_false] = ACTIONS(1548),
    [sym_none] = ACTIONS(1548),
    [sym_string] = ACTIONS(1548),
    [sym_verbatim] = ACTIONS(1548),
    [sym_number] = ACTIONS(1548),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1546),
    [sym_fstring_start] = ACTIONS(1546),
  },
  [STATE(609)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_async] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1270),
    [anon_sym_del] = ACTIONS(1270),
    [anon_sym_raise] = ACTIONS(1270),
    [sym_pass_statement] = ACTIONS(1270),
    [sym_break_statement] = ACTIONS(1270),
    [sym_continue_statement] = ACTIONS(1270),
    [sym_debugger_statement] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_global] = ACTIONS(1270),
    [anon_sym_nonlocal] = ACTIONS(1270),
    [anon_sym_yield] = ACTIONS(1270),
    [anon_sym_from] = ACTIONS(1270),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_DOT] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_with] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1270),
    [anon_sym_or] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1270),
    [anon_sym_typeof] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1270),
    [anon_sym_await] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1270),
    [sym_this] = ACTIONS(1270),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [sym_none] = ACTIONS(1270),
    [sym_string] = ACTIONS(1270),
    [sym_verbatim] = ACTIONS(1270),
    [sym_number] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1268),
    [sym_fstring_start] = ACTIONS(1268),
  },
  [STATE(610)] = {
    [sym_identifier] = ACTIONS(1516),
    [anon_sym_SEMI] = ACTIONS(1516),
    [anon_sym_return] = ACTIONS(1516),
    [anon_sym_async] = ACTIONS(1516),
    [anon_sym_def] = ACTIONS(1516),
    [anon_sym_COMMA] = ACTIONS(1516),
    [anon_sym_LPAREN] = ACTIONS(1516),
    [anon_sym_LBRACK] = ACTIONS(1516),
    [anon_sym_del] = ACTIONS(1516),
    [anon_sym_raise] = ACTIONS(1516),
    [sym_pass_statement] = ACTIONS(1516),
    [sym_break_statement] = ACTIONS(1516),
    [sym_continue_statement] = ACTIONS(1516),
    [sym_debugger_statement] = ACTIONS(1516),
    [anon_sym_assert] = ACTIONS(1516),
    [anon_sym_global] = ACTIONS(1516),
    [anon_sym_nonlocal] = ACTIONS(1516),
    [anon_sym_yield] = ACTIONS(1516),
    [anon_sym_from] = ACTIONS(1516),
    [anon_sym_import] = ACTIONS(1516),
    [anon_sym_DOT] = ACTIONS(1516),
    [anon_sym_if] = ACTIONS(1516),
    [anon_sym_for] = ACTIONS(1516),
    [anon_sym_in] = ACTIONS(1516),
    [anon_sym_while] = ACTIONS(1516),
    [anon_sym_do] = ACTIONS(1516),
    [anon_sym_try] = ACTIONS(1516),
    [anon_sym_with] = ACTIONS(1516),
    [anon_sym_STAR] = ACTIONS(1516),
    [anon_sym_STAR_STAR] = ACTIONS(1516),
    [anon_sym_class] = ACTIONS(1516),
    [anon_sym_AT] = ACTIONS(1516),
    [anon_sym_not] = ACTIONS(1516),
    [anon_sym_and] = ACTIONS(1516),
    [anon_sym_or] = ACTIONS(1516),
    [anon_sym_PLUS] = ACTIONS(1516),
    [anon_sym_DASH] = ACTIONS(1516),
    [anon_sym_SLASH] = ACTIONS(1516),
    [anon_sym_SLASH_SLASH] = ACTIONS(1516),
    [anon_sym_PERCENT] = ACTIONS(1516),
    [anon_sym_PIPE] = ACTIONS(1516),
    [anon_sym_CARET] = ACTIONS(1516),
    [anon_sym_AMP] = ACTIONS(1516),
    [anon_sym_LT_LT] = ACTIONS(1516),
    [anon_sym_GT_GT] = ACTIONS(1516),
    [anon_sym_GT_GT_GT] = ACTIONS(1516),
    [anon_sym_TILDE] = ACTIONS(1516),
    [anon_sym_typeof] = ACTIONS(1516),
    [anon_sym_void] = ACTIONS(1516),
    [anon_sym_LT] = ACTIONS(1516),
    [anon_sym_LT_EQ] = ACTIONS(1516),
    [anon_sym_EQ_EQ] = ACTIONS(1516),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1516),
    [anon_sym_BANG_EQ] = ACTIONS(1516),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1516),
    [anon_sym_GT_EQ] = ACTIONS(1516),
    [anon_sym_GT] = ACTIONS(1516),
    [anon_sym_is] = ACTIONS(1516),
    [anon_sym_instanceof] = ACTIONS(1516),
    [anon_sym_new] = ACTIONS(1516),
    [anon_sym_QMARK] = ACTIONS(1516),
    [anon_sym_await] = ACTIONS(1516),
    [anon_sym_LBRACE] = ACTIONS(1516),
    [sym_this] = ACTIONS(1516),
    [sym_true] = ACTIONS(1516),
    [sym_false] = ACTIONS(1516),
    [sym_none] = ACTIONS(1516),
    [sym_string] = ACTIONS(1516),
    [sym_verbatim] = ACTIONS(1516),
    [sym_number] = ACTIONS(1516),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1514),
    [sym_regex] = ACTIONS(1514),
    [sym_fstring_start] = ACTIONS(1514),
  },
  [STATE(611)] = {
    [aux_sym__inline_body_repeat1] = STATE(623),
    [sym_identifier] = ACTIONS(1363),
    [anon_sym_SEMI] = ACTIONS(1550),
    [anon_sym_return] = ACTIONS(1363),
    [anon_sym_async] = ACTIONS(1363),
    [anon_sym_def] = ACTIONS(1363),
    [anon_sym_LPAREN] = ACTIONS(1363),
    [anon_sym_LBRACK] = ACTIONS(1363),
    [anon_sym_del] = ACTIONS(1363),
    [anon_sym_raise] = ACTIONS(1363),
    [sym_pass_statement] = ACTIONS(1363),
    [sym_break_statement] = ACTIONS(1363),
    [sym_continue_statement] = ACTIONS(1363),
    [sym_debugger_statement] = ACTIONS(1363),
    [anon_sym_assert] = ACTIONS(1363),
    [anon_sym_global] = ACTIONS(1363),
    [anon_sym_nonlocal] = ACTIONS(1363),
    [anon_sym_yield] = ACTIONS(1363),
    [anon_sym_from] = ACTIONS(1363),
    [anon_sym_import] = ACTIONS(1363),
    [anon_sym_DOT] = ACTIONS(1363),
    [anon_sym_if] = ACTIONS(1363),
    [anon_sym_for] = ACTIONS(1363),
    [anon_sym_in] = ACTIONS(1363),
    [anon_sym_while] = ACTIONS(1363),
    [anon_sym_do] = ACTIONS(1363),
    [anon_sym_try] = ACTIONS(1363),
    [anon_sym_with] = ACTIONS(1363),
    [anon_sym_STAR] = ACTIONS(1363),
    [anon_sym_STAR_STAR] = ACTIONS(1363),
    [anon_sym_class] = ACTIONS(1363),
    [anon_sym_AT] = ACTIONS(1363),
    [anon_sym_not] = ACTIONS(1363),
    [anon_sym_and] = ACTIONS(1363),
    [anon_sym_or] = ACTIONS(1363),
    [anon_sym_PLUS] = ACTIONS(1363),
    [anon_sym_DASH] = ACTIONS(1363),
    [anon_sym_SLASH] = ACTIONS(1363),
    [anon_sym_SLASH_SLASH] = ACTIONS(1363),
    [anon_sym_PERCENT] = ACTIONS(1363),
    [anon_sym_PIPE] = ACTIONS(1363),
    [anon_sym_CARET] = ACTIONS(1363),
    [anon_sym_AMP] = ACTIONS(1363),
    [anon_sym_LT_LT] = ACTIONS(1363),
    [anon_sym_GT_GT] = ACTIONS(1363),
    [anon_sym_GT_GT_GT] = ACTIONS(1363),
    [anon_sym_TILDE] = ACTIONS(1363),
    [anon_sym_typeof] = ACTIONS(1363),
    [anon_sym_void] = ACTIONS(1363),
    [anon_sym_LT] = ACTIONS(1363),
    [anon_sym_LT_EQ] = ACTIONS(1363),
    [anon_sym_EQ_EQ] = ACTIONS(1363),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1363),
    [anon_sym_BANG_EQ] = ACTIONS(1363),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1363),
    [anon_sym_GT_EQ] = ACTIONS(1363),
    [anon_sym_GT] = ACTIONS(1363),
    [anon_sym_is] = ACTIONS(1363),
    [anon_sym_instanceof] = ACTIONS(1363),
    [anon_sym_new] = ACTIONS(1363),
    [anon_sym_QMARK] = ACTIONS(1363),
    [anon_sym_await] = ACTIONS(1363),
    [anon_sym_LBRACE] = ACTIONS(1363),
    [sym_this] = ACTIONS(1363),
    [sym_true] = ACTIONS(1363),
    [sym_false] = ACTIONS(1363),
    [sym_none] = ACTIONS(1363),
    [sym_string] = ACTIONS(1363),
    [sym_verbatim] = ACTIONS(1363),
    [sym_number] = ACTIONS(1363),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1361),
    [sym_regex] = ACTIONS(1361),
    [sym_fstring_start] = ACTIONS(1361),
  },
  [STATE(612)] = {
    [ts_builtin_sym_end] = ACTIONS(1552),
    [sym_identifier] = ACTIONS(1554),
    [anon_sym_SEMI] = ACTIONS(1554),
    [anon_sym_return] = ACTIONS(1554),
    [anon_sym_async] = ACTIONS(1554),
    [anon_sym_def] = ACTIONS(1554),
    [anon_sym_COMMA] = ACTIONS(1554),
    [anon_sym_LPAREN] = ACTIONS(1554),
    [anon_sym_LBRACK] = ACTIONS(1554),
    [anon_sym_del] = ACTIONS(1554),
    [anon_sym_raise] = ACTIONS(1554),
    [sym_pass_statement] = ACTIONS(1554),
    [sym_break_statement] = ACTIONS(1554),
    [sym_continue_statement] = ACTIONS(1554),
    [sym_debugger_statement] = ACTIONS(1554),
    [anon_sym_assert] = ACTIONS(1554),
    [anon_sym_global] = ACTIONS(1554),
    [anon_sym_nonlocal] = ACTIONS(1554),
    [anon_sym_yield] = ACTIONS(1554),
    [anon_sym_from] = ACTIONS(1554),
    [anon_sym_import] = ACTIONS(1554),
    [anon_sym_DOT] = ACTIONS(1554),
    [anon_sym_if] = ACTIONS(1554),
    [anon_sym_for] = ACTIONS(1554),
    [anon_sym_in] = ACTIONS(1554),
    [anon_sym_while] = ACTIONS(1554),
    [anon_sym_do] = ACTIONS(1554),
    [anon_sym_try] = ACTIONS(1554),
    [anon_sym_with] = ACTIONS(1554),
    [anon_sym_STAR] = ACTIONS(1554),
    [anon_sym_STAR_STAR] = ACTIONS(1554),
    [anon_sym_class] = ACTIONS(1554),
    [anon_sym_AT] = ACTIONS(1554),
    [anon_sym_not] = ACTIONS(1554),
    [anon_sym_and] = ACTIONS(1554),
    [anon_sym_or] = ACTIONS(1554),
    [anon_sym_PLUS] = ACTIONS(1554),
    [anon_sym_DASH] = ACTIONS(1554),
    [anon_sym_SLASH] = ACTIONS(1554),
    [anon_sym_SLASH_SLASH] = ACTIONS(1554),
    [anon_sym_PERCENT] = ACTIONS(1554),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_CARET] = ACTIONS(1554),
    [anon_sym_AMP] = ACTIONS(1554),
    [anon_sym_LT_LT] = ACTIONS(1554),
    [anon_sym_GT_GT] = ACTIONS(1554),
    [anon_sym_GT_GT_GT] = ACTIONS(1554),
    [anon_sym_TILDE] = ACTIONS(1554),
    [anon_sym_typeof] = ACTIONS(1554),
    [anon_sym_void] = ACTIONS(1554),
    [anon_sym_LT] = ACTIONS(1554),
    [anon_sym_LT_EQ] = ACTIONS(1554),
    [anon_sym_EQ_EQ] = ACTIONS(1554),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1554),
    [anon_sym_BANG_EQ] = ACTIONS(1554),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1554),
    [anon_sym_GT_EQ] = ACTIONS(1554),
    [anon_sym_GT] = ACTIONS(1554),
    [anon_sym_is] = ACTIONS(1554),
    [anon_sym_instanceof] = ACTIONS(1554),
    [anon_sym_new] = ACTIONS(1554),
    [anon_sym_QMARK] = ACTIONS(1554),
    [anon_sym_await] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [sym_this] = ACTIONS(1554),
    [sym_true] = ACTIONS(1554),
    [sym_false] = ACTIONS(1554),
    [sym_none] = ACTIONS(1554),
    [sym_string] = ACTIONS(1554),
    [sym_verbatim] = ACTIONS(1554),
    [sym_number] = ACTIONS(1554),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1552),
    [sym_fstring_start] = ACTIONS(1552),
  },
  [STATE(613)] = {
    [sym_identifier] = ACTIONS(1411),
    [anon_sym_SEMI] = ACTIONS(1411),
    [anon_sym_return] = ACTIONS(1411),
    [anon_sym_async] = ACTIONS(1411),
    [anon_sym_def] = ACTIONS(1411),
    [anon_sym_COMMA] = ACTIONS(1411),
    [anon_sym_LPAREN] = ACTIONS(1411),
    [anon_sym_LBRACK] = ACTIONS(1411),
    [anon_sym_del] = ACTIONS(1411),
    [anon_sym_raise] = ACTIONS(1411),
    [sym_pass_statement] = ACTIONS(1411),
    [sym_break_statement] = ACTIONS(1411),
    [sym_continue_statement] = ACTIONS(1411),
    [sym_debugger_statement] = ACTIONS(1411),
    [anon_sym_assert] = ACTIONS(1411),
    [anon_sym_global] = ACTIONS(1411),
    [anon_sym_nonlocal] = ACTIONS(1411),
    [anon_sym_yield] = ACTIONS(1411),
    [anon_sym_from] = ACTIONS(1411),
    [anon_sym_import] = ACTIONS(1411),
    [anon_sym_DOT] = ACTIONS(1411),
    [anon_sym_if] = ACTIONS(1411),
    [anon_sym_for] = ACTIONS(1411),
    [anon_sym_in] = ACTIONS(1411),
    [anon_sym_while] = ACTIONS(1411),
    [anon_sym_do] = ACTIONS(1411),
    [anon_sym_try] = ACTIONS(1411),
    [anon_sym_with] = ACTIONS(1411),
    [anon_sym_STAR] = ACTIONS(1411),
    [anon_sym_STAR_STAR] = ACTIONS(1411),
    [anon_sym_class] = ACTIONS(1411),
    [anon_sym_AT] = ACTIONS(1411),
    [anon_sym_not] = ACTIONS(1411),
    [anon_sym_and] = ACTIONS(1411),
    [anon_sym_or] = ACTIONS(1411),
    [anon_sym_PLUS] = ACTIONS(1411),
    [anon_sym_DASH] = ACTIONS(1411),
    [anon_sym_SLASH] = ACTIONS(1411),
    [anon_sym_SLASH_SLASH] = ACTIONS(1411),
    [anon_sym_PERCENT] = ACTIONS(1411),
    [anon_sym_PIPE] = ACTIONS(1411),
    [anon_sym_CARET] = ACTIONS(1411),
    [anon_sym_AMP] = ACTIONS(1411),
    [anon_sym_LT_LT] = ACTIONS(1411),
    [anon_sym_GT_GT] = ACTIONS(1411),
    [anon_sym_GT_GT_GT] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1411),
    [anon_sym_typeof] = ACTIONS(1411),
    [anon_sym_void] = ACTIONS(1411),
    [anon_sym_LT] = ACTIONS(1411),
    [anon_sym_LT_EQ] = ACTIONS(1411),
    [anon_sym_EQ_EQ] = ACTIONS(1411),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1411),
    [anon_sym_BANG_EQ] = ACTIONS(1411),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1411),
    [anon_sym_GT_EQ] = ACTIONS(1411),
    [anon_sym_GT] = ACTIONS(1411),
    [anon_sym_is] = ACTIONS(1411),
    [anon_sym_instanceof] = ACTIONS(1411),
    [anon_sym_new] = ACTIONS(1411),
    [anon_sym_QMARK] = ACTIONS(1411),
    [anon_sym_await] = ACTIONS(1411),
    [anon_sym_LBRACE] = ACTIONS(1411),
    [sym_this] = ACTIONS(1411),
    [sym_true] = ACTIONS(1411),
    [sym_false] = ACTIONS(1411),
    [sym_none] = ACTIONS(1411),
    [sym_string] = ACTIONS(1411),
    [sym_verbatim] = ACTIONS(1411),
    [sym_number] = ACTIONS(1411),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1409),
    [sym_regex] = ACTIONS(1409),
    [sym_fstring_start] = ACTIONS(1409),
  },
  [STATE(614)] = {
    [sym_identifier] = ACTIONS(1344),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_return] = ACTIONS(1344),
    [anon_sym_async] = ACTIONS(1344),
    [anon_sym_def] = ACTIONS(1344),
    [anon_sym_COMMA] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_LBRACK] = ACTIONS(1344),
    [anon_sym_del] = ACTIONS(1344),
    [anon_sym_raise] = ACTIONS(1344),
    [sym_pass_statement] = ACTIONS(1344),
    [sym_break_statement] = ACTIONS(1344),
    [sym_continue_statement] = ACTIONS(1344),
    [sym_debugger_statement] = ACTIONS(1344),
    [anon_sym_assert] = ACTIONS(1344),
    [anon_sym_global] = ACTIONS(1344),
    [anon_sym_nonlocal] = ACTIONS(1344),
    [anon_sym_yield] = ACTIONS(1344),
    [anon_sym_from] = ACTIONS(1344),
    [anon_sym_import] = ACTIONS(1344),
    [anon_sym_DOT] = ACTIONS(1344),
    [anon_sym_if] = ACTIONS(1344),
    [anon_sym_for] = ACTIONS(1344),
    [anon_sym_in] = ACTIONS(1344),
    [anon_sym_while] = ACTIONS(1344),
    [anon_sym_do] = ACTIONS(1344),
    [anon_sym_try] = ACTIONS(1344),
    [anon_sym_with] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_STAR_STAR] = ACTIONS(1344),
    [anon_sym_class] = ACTIONS(1344),
    [anon_sym_AT] = ACTIONS(1344),
    [anon_sym_not] = ACTIONS(1344),
    [anon_sym_and] = ACTIONS(1344),
    [anon_sym_or] = ACTIONS(1344),
    [anon_sym_PLUS] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_SLASH] = ACTIONS(1344),
    [anon_sym_SLASH_SLASH] = ACTIONS(1344),
    [anon_sym_PERCENT] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_CARET] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_LT_LT] = ACTIONS(1344),
    [anon_sym_GT_GT] = ACTIONS(1344),
    [anon_sym_GT_GT_GT] = ACTIONS(1344),
    [anon_sym_TILDE] = ACTIONS(1344),
    [anon_sym_typeof] = ACTIONS(1344),
    [anon_sym_void] = ACTIONS(1344),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_LT_EQ] = ACTIONS(1344),
    [anon_sym_EQ_EQ] = ACTIONS(1344),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ] = ACTIONS(1344),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1344),
    [anon_sym_GT_EQ] = ACTIONS(1344),
    [anon_sym_GT] = ACTIONS(1344),
    [anon_sym_is] = ACTIONS(1344),
    [anon_sym_instanceof] = ACTIONS(1344),
    [anon_sym_new] = ACTIONS(1344),
    [anon_sym_QMARK] = ACTIONS(1344),
    [anon_sym_await] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [sym_this] = ACTIONS(1344),
    [sym_true] = ACTIONS(1344),
    [sym_false] = ACTIONS(1344),
    [sym_none] = ACTIONS(1344),
    [sym_string] = ACTIONS(1344),
    [sym_verbatim] = ACTIONS(1344),
    [sym_number] = ACTIONS(1344),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1342),
    [sym_regex] = ACTIONS(1342),
    [sym_fstring_start] = ACTIONS(1342),
  },
  [STATE(615)] = {
    [sym_identifier] = ACTIONS(1190),
    [anon_sym_SEMI] = ACTIONS(1190),
    [anon_sym_return] = ACTIONS(1190),
    [anon_sym_async] = ACTIONS(1190),
    [anon_sym_def] = ACTIONS(1190),
    [anon_sym_COMMA] = ACTIONS(1190),
    [anon_sym_LPAREN] = ACTIONS(1190),
    [anon_sym_LBRACK] = ACTIONS(1190),
    [anon_sym_del] = ACTIONS(1190),
    [anon_sym_raise] = ACTIONS(1190),
    [sym_pass_statement] = ACTIONS(1190),
    [sym_break_statement] = ACTIONS(1190),
    [sym_continue_statement] = ACTIONS(1190),
    [sym_debugger_statement] = ACTIONS(1190),
    [anon_sym_assert] = ACTIONS(1190),
    [anon_sym_global] = ACTIONS(1190),
    [anon_sym_nonlocal] = ACTIONS(1190),
    [anon_sym_yield] = ACTIONS(1190),
    [anon_sym_from] = ACTIONS(1190),
    [anon_sym_import] = ACTIONS(1190),
    [anon_sym_DOT] = ACTIONS(1190),
    [anon_sym_if] = ACTIONS(1190),
    [anon_sym_for] = ACTIONS(1190),
    [anon_sym_in] = ACTIONS(1190),
    [anon_sym_while] = ACTIONS(1190),
    [anon_sym_do] = ACTIONS(1190),
    [anon_sym_try] = ACTIONS(1190),
    [anon_sym_with] = ACTIONS(1190),
    [anon_sym_STAR] = ACTIONS(1190),
    [anon_sym_STAR_STAR] = ACTIONS(1190),
    [anon_sym_class] = ACTIONS(1190),
    [anon_sym_AT] = ACTIONS(1190),
    [anon_sym_not] = ACTIONS(1190),
    [anon_sym_and] = ACTIONS(1190),
    [anon_sym_or] = ACTIONS(1190),
    [anon_sym_PLUS] = ACTIONS(1190),
    [anon_sym_DASH] = ACTIONS(1190),
    [anon_sym_SLASH] = ACTIONS(1190),
    [anon_sym_SLASH_SLASH] = ACTIONS(1190),
    [anon_sym_PERCENT] = ACTIONS(1190),
    [anon_sym_PIPE] = ACTIONS(1190),
    [anon_sym_CARET] = ACTIONS(1190),
    [anon_sym_AMP] = ACTIONS(1190),
    [anon_sym_LT_LT] = ACTIONS(1190),
    [anon_sym_GT_GT] = ACTIONS(1190),
    [anon_sym_GT_GT_GT] = ACTIONS(1190),
    [anon_sym_TILDE] = ACTIONS(1190),
    [anon_sym_typeof] = ACTIONS(1190),
    [anon_sym_void] = ACTIONS(1190),
    [anon_sym_LT] = ACTIONS(1190),
    [anon_sym_LT_EQ] = ACTIONS(1190),
    [anon_sym_EQ_EQ] = ACTIONS(1190),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1190),
    [anon_sym_BANG_EQ] = ACTIONS(1190),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1190),
    [anon_sym_GT_EQ] = ACTIONS(1190),
    [anon_sym_GT] = ACTIONS(1190),
    [anon_sym_is] = ACTIONS(1190),
    [anon_sym_instanceof] = ACTIONS(1190),
    [anon_sym_new] = ACTIONS(1190),
    [anon_sym_QMARK] = ACTIONS(1190),
    [anon_sym_await] = ACTIONS(1190),
    [anon_sym_LBRACE] = ACTIONS(1190),
    [sym_this] = ACTIONS(1190),
    [sym_true] = ACTIONS(1190),
    [sym_false] = ACTIONS(1190),
    [sym_none] = ACTIONS(1190),
    [sym_string] = ACTIONS(1190),
    [sym_verbatim] = ACTIONS(1190),
    [sym_number] = ACTIONS(1190),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1192),
    [sym_regex] = ACTIONS(1192),
    [sym_fstring_start] = ACTIONS(1192),
  },
  [STATE(616)] = {
    [ts_builtin_sym_end] = ACTIONS(1556),
    [sym_identifier] = ACTIONS(1558),
    [anon_sym_SEMI] = ACTIONS(1558),
    [anon_sym_return] = ACTIONS(1558),
    [anon_sym_async] = ACTIONS(1558),
    [anon_sym_def] = ACTIONS(1558),
    [anon_sym_COMMA] = ACTIONS(1558),
    [anon_sym_LPAREN] = ACTIONS(1558),
    [anon_sym_LBRACK] = ACTIONS(1558),
    [anon_sym_del] = ACTIONS(1558),
    [anon_sym_raise] = ACTIONS(1558),
    [sym_pass_statement] = ACTIONS(1558),
    [sym_break_statement] = ACTIONS(1558),
    [sym_continue_statement] = ACTIONS(1558),
    [sym_debugger_statement] = ACTIONS(1558),
    [anon_sym_assert] = ACTIONS(1558),
    [anon_sym_global] = ACTIONS(1558),
    [anon_sym_nonlocal] = ACTIONS(1558),
    [anon_sym_yield] = ACTIONS(1558),
    [anon_sym_from] = ACTIONS(1558),
    [anon_sym_import] = ACTIONS(1558),
    [anon_sym_DOT] = ACTIONS(1558),
    [anon_sym_if] = ACTIONS(1558),
    [anon_sym_for] = ACTIONS(1558),
    [anon_sym_in] = ACTIONS(1558),
    [anon_sym_while] = ACTIONS(1558),
    [anon_sym_do] = ACTIONS(1558),
    [anon_sym_try] = ACTIONS(1558),
    [anon_sym_with] = ACTIONS(1558),
    [anon_sym_STAR] = ACTIONS(1558),
    [anon_sym_STAR_STAR] = ACTIONS(1558),
    [anon_sym_class] = ACTIONS(1558),
    [anon_sym_AT] = ACTIONS(1558),
    [anon_sym_not] = ACTIONS(1558),
    [anon_sym_and] = ACTIONS(1558),
    [anon_sym_or] = ACTIONS(1558),
    [anon_sym_PLUS] = ACTIONS(1558),
    [anon_sym_DASH] = ACTIONS(1558),
    [anon_sym_SLASH] = ACTIONS(1558),
    [anon_sym_SLASH_SLASH] = ACTIONS(1558),
    [anon_sym_PERCENT] = ACTIONS(1558),
    [anon_sym_PIPE] = ACTIONS(1558),
    [anon_sym_CARET] = ACTIONS(1558),
    [anon_sym_AMP] = ACTIONS(1558),
    [anon_sym_LT_LT] = ACTIONS(1558),
    [anon_sym_GT_GT] = ACTIONS(1558),
    [anon_sym_GT_GT_GT] = ACTIONS(1558),
    [anon_sym_TILDE] = ACTIONS(1558),
    [anon_sym_typeof] = ACTIONS(1558),
    [anon_sym_void] = ACTIONS(1558),
    [anon_sym_LT] = ACTIONS(1558),
    [anon_sym_LT_EQ] = ACTIONS(1558),
    [anon_sym_EQ_EQ] = ACTIONS(1558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1558),
    [anon_sym_BANG_EQ] = ACTIONS(1558),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1558),
    [anon_sym_GT_EQ] = ACTIONS(1558),
    [anon_sym_GT] = ACTIONS(1558),
    [anon_sym_is] = ACTIONS(1558),
    [anon_sym_instanceof] = ACTIONS(1558),
    [anon_sym_new] = ACTIONS(1558),
    [anon_sym_QMARK] = ACTIONS(1558),
    [anon_sym_await] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [sym_this] = ACTIONS(1558),
    [sym_true] = ACTIONS(1558),
    [sym_false] = ACTIONS(1558),
    [sym_none] = ACTIONS(1558),
    [sym_string] = ACTIONS(1558),
    [sym_verbatim] = ACTIONS(1558),
    [sym_number] = ACTIONS(1558),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1556),
    [sym_fstring_start] = ACTIONS(1556),
  },
  [STATE(617)] = {
    [ts_builtin_sym_end] = ACTIONS(1560),
    [sym_identifier] = ACTIONS(1562),
    [anon_sym_SEMI] = ACTIONS(1562),
    [anon_sym_return] = ACTIONS(1562),
    [anon_sym_async] = ACTIONS(1562),
    [anon_sym_def] = ACTIONS(1562),
    [anon_sym_COMMA] = ACTIONS(1562),
    [anon_sym_LPAREN] = ACTIONS(1562),
    [anon_sym_LBRACK] = ACTIONS(1562),
    [anon_sym_del] = ACTIONS(1562),
    [anon_sym_raise] = ACTIONS(1562),
    [sym_pass_statement] = ACTIONS(1562),
    [sym_break_statement] = ACTIONS(1562),
    [sym_continue_statement] = ACTIONS(1562),
    [sym_debugger_statement] = ACTIONS(1562),
    [anon_sym_assert] = ACTIONS(1562),
    [anon_sym_global] = ACTIONS(1562),
    [anon_sym_nonlocal] = ACTIONS(1562),
    [anon_sym_yield] = ACTIONS(1562),
    [anon_sym_from] = ACTIONS(1562),
    [anon_sym_import] = ACTIONS(1562),
    [anon_sym_DOT] = ACTIONS(1562),
    [anon_sym_if] = ACTIONS(1562),
    [anon_sym_for] = ACTIONS(1562),
    [anon_sym_in] = ACTIONS(1562),
    [anon_sym_while] = ACTIONS(1562),
    [anon_sym_do] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1562),
    [anon_sym_with] = ACTIONS(1562),
    [anon_sym_STAR] = ACTIONS(1562),
    [anon_sym_STAR_STAR] = ACTIONS(1562),
    [anon_sym_class] = ACTIONS(1562),
    [anon_sym_AT] = ACTIONS(1562),
    [anon_sym_not] = ACTIONS(1562),
    [anon_sym_and] = ACTIONS(1562),
    [anon_sym_or] = ACTIONS(1562),
    [anon_sym_PLUS] = ACTIONS(1562),
    [anon_sym_DASH] = ACTIONS(1562),
    [anon_sym_SLASH] = ACTIONS(1562),
    [anon_sym_SLASH_SLASH] = ACTIONS(1562),
    [anon_sym_PERCENT] = ACTIONS(1562),
    [anon_sym_PIPE] = ACTIONS(1562),
    [anon_sym_CARET] = ACTIONS(1562),
    [anon_sym_AMP] = ACTIONS(1562),
    [anon_sym_LT_LT] = ACTIONS(1562),
    [anon_sym_GT_GT] = ACTIONS(1562),
    [anon_sym_GT_GT_GT] = ACTIONS(1562),
    [anon_sym_TILDE] = ACTIONS(1562),
    [anon_sym_typeof] = ACTIONS(1562),
    [anon_sym_void] = ACTIONS(1562),
    [anon_sym_LT] = ACTIONS(1562),
    [anon_sym_LT_EQ] = ACTIONS(1562),
    [anon_sym_EQ_EQ] = ACTIONS(1562),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1562),
    [anon_sym_BANG_EQ] = ACTIONS(1562),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1562),
    [anon_sym_GT_EQ] = ACTIONS(1562),
    [anon_sym_GT] = ACTIONS(1562),
    [anon_sym_is] = ACTIONS(1562),
    [anon_sym_instanceof] = ACTIONS(1562),
    [anon_sym_new] = ACTIONS(1562),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_await] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [sym_this] = ACTIONS(1562),
    [sym_true] = ACTIONS(1562),
    [sym_false] = ACTIONS(1562),
    [sym_none] = ACTIONS(1562),
    [sym_string] = ACTIONS(1562),
    [sym_verbatim] = ACTIONS(1562),
    [sym_number] = ACTIONS(1562),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1560),
    [sym_fstring_start] = ACTIONS(1560),
  },
  [STATE(618)] = {
    [ts_builtin_sym_end] = ACTIONS(1369),
    [sym_identifier] = ACTIONS(1371),
    [anon_sym_SEMI] = ACTIONS(1371),
    [anon_sym_return] = ACTIONS(1371),
    [anon_sym_async] = ACTIONS(1371),
    [anon_sym_def] = ACTIONS(1371),
    [anon_sym_COMMA] = ACTIONS(1371),
    [anon_sym_LPAREN] = ACTIONS(1371),
    [anon_sym_LBRACK] = ACTIONS(1371),
    [anon_sym_del] = ACTIONS(1371),
    [anon_sym_raise] = ACTIONS(1371),
    [sym_pass_statement] = ACTIONS(1371),
    [sym_break_statement] = ACTIONS(1371),
    [sym_continue_statement] = ACTIONS(1371),
    [sym_debugger_statement] = ACTIONS(1371),
    [anon_sym_assert] = ACTIONS(1371),
    [anon_sym_global] = ACTIONS(1371),
    [anon_sym_nonlocal] = ACTIONS(1371),
    [anon_sym_yield] = ACTIONS(1371),
    [anon_sym_from] = ACTIONS(1371),
    [anon_sym_import] = ACTIONS(1371),
    [anon_sym_DOT] = ACTIONS(1371),
    [anon_sym_if] = ACTIONS(1371),
    [anon_sym_for] = ACTIONS(1371),
    [anon_sym_in] = ACTIONS(1371),
    [anon_sym_while] = ACTIONS(1371),
    [anon_sym_do] = ACTIONS(1371),
    [anon_sym_try] = ACTIONS(1371),
    [anon_sym_with] = ACTIONS(1371),
    [anon_sym_STAR] = ACTIONS(1371),
    [anon_sym_STAR_STAR] = ACTIONS(1371),
    [anon_sym_class] = ACTIONS(1371),
    [anon_sym_AT] = ACTIONS(1371),
    [anon_sym_not] = ACTIONS(1371),
    [anon_sym_and] = ACTIONS(1371),
    [anon_sym_or] = ACTIONS(1371),
    [anon_sym_PLUS] = ACTIONS(1371),
    [anon_sym_DASH] = ACTIONS(1371),
    [anon_sym_SLASH] = ACTIONS(1371),
    [anon_sym_SLASH_SLASH] = ACTIONS(1371),
    [anon_sym_PERCENT] = ACTIONS(1371),
    [anon_sym_PIPE] = ACTIONS(1371),
    [anon_sym_CARET] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(1371),
    [anon_sym_LT_LT] = ACTIONS(1371),
    [anon_sym_GT_GT] = ACTIONS(1371),
    [anon_sym_GT_GT_GT] = ACTIONS(1371),
    [anon_sym_TILDE] = ACTIONS(1371),
    [anon_sym_typeof] = ACTIONS(1371),
    [anon_sym_void] = ACTIONS(1371),
    [anon_sym_LT] = ACTIONS(1371),
    [anon_sym_LT_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1371),
    [anon_sym_GT_EQ] = ACTIONS(1371),
    [anon_sym_GT] = ACTIONS(1371),
    [anon_sym_is] = ACTIONS(1371),
    [anon_sym_instanceof] = ACTIONS(1371),
    [anon_sym_new] = ACTIONS(1371),
    [anon_sym_QMARK] = ACTIONS(1371),
    [anon_sym_await] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1371),
    [sym_this] = ACTIONS(1371),
    [sym_true] = ACTIONS(1371),
    [sym_false] = ACTIONS(1371),
    [sym_none] = ACTIONS(1371),
    [sym_string] = ACTIONS(1371),
    [sym_verbatim] = ACTIONS(1371),
    [sym_number] = ACTIONS(1371),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1369),
    [sym_fstring_start] = ACTIONS(1369),
  },
  [STATE(619)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1303),
    [sym_identifier] = ACTIONS(1305),
    [anon_sym_SEMI] = ACTIONS(1305),
    [anon_sym_return] = ACTIONS(1305),
    [anon_sym_async] = ACTIONS(1305),
    [anon_sym_def] = ACTIONS(1305),
    [anon_sym_LPAREN] = ACTIONS(1305),
    [anon_sym_LBRACK] = ACTIONS(1305),
    [anon_sym_del] = ACTIONS(1305),
    [anon_sym_raise] = ACTIONS(1305),
    [sym_pass_statement] = ACTIONS(1305),
    [sym_break_statement] = ACTIONS(1305),
    [sym_continue_statement] = ACTIONS(1305),
    [sym_debugger_statement] = ACTIONS(1305),
    [anon_sym_assert] = ACTIONS(1305),
    [anon_sym_global] = ACTIONS(1305),
    [anon_sym_nonlocal] = ACTIONS(1305),
    [anon_sym_yield] = ACTIONS(1305),
    [anon_sym_from] = ACTIONS(1305),
    [anon_sym_import] = ACTIONS(1305),
    [anon_sym_DOT] = ACTIONS(1305),
    [anon_sym_if] = ACTIONS(1305),
    [anon_sym_for] = ACTIONS(1305),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1305),
    [anon_sym_do] = ACTIONS(1305),
    [anon_sym_try] = ACTIONS(1305),
    [anon_sym_with] = ACTIONS(1305),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1305),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1494),
    [anon_sym_or] = ACTIONS(1496),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1305),
    [anon_sym_typeof] = ACTIONS(1305),
    [anon_sym_void] = ACTIONS(1305),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1305),
    [anon_sym_QMARK] = ACTIONS(1305),
    [anon_sym_await] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1305),
    [sym_this] = ACTIONS(1305),
    [sym_true] = ACTIONS(1305),
    [sym_false] = ACTIONS(1305),
    [sym_none] = ACTIONS(1305),
    [sym_string] = ACTIONS(1305),
    [sym_verbatim] = ACTIONS(1305),
    [sym_number] = ACTIONS(1305),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1303),
    [sym_fstring_start] = ACTIONS(1303),
  },
  [STATE(620)] = {
    [sym_identifier] = ACTIONS(1415),
    [anon_sym_SEMI] = ACTIONS(1415),
    [anon_sym_return] = ACTIONS(1415),
    [anon_sym_async] = ACTIONS(1415),
    [anon_sym_def] = ACTIONS(1415),
    [anon_sym_COMMA] = ACTIONS(1415),
    [anon_sym_LPAREN] = ACTIONS(1415),
    [anon_sym_LBRACK] = ACTIONS(1415),
    [anon_sym_del] = ACTIONS(1415),
    [anon_sym_raise] = ACTIONS(1415),
    [sym_pass_statement] = ACTIONS(1415),
    [sym_break_statement] = ACTIONS(1415),
    [sym_continue_statement] = ACTIONS(1415),
    [sym_debugger_statement] = ACTIONS(1415),
    [anon_sym_assert] = ACTIONS(1415),
    [anon_sym_global] = ACTIONS(1415),
    [anon_sym_nonlocal] = ACTIONS(1415),
    [anon_sym_yield] = ACTIONS(1415),
    [anon_sym_from] = ACTIONS(1415),
    [anon_sym_import] = ACTIONS(1415),
    [anon_sym_DOT] = ACTIONS(1415),
    [anon_sym_if] = ACTIONS(1415),
    [anon_sym_for] = ACTIONS(1415),
    [anon_sym_in] = ACTIONS(1415),
    [anon_sym_while] = ACTIONS(1415),
    [anon_sym_do] = ACTIONS(1415),
    [anon_sym_try] = ACTIONS(1415),
    [anon_sym_with] = ACTIONS(1415),
    [anon_sym_STAR] = ACTIONS(1415),
    [anon_sym_STAR_STAR] = ACTIONS(1415),
    [anon_sym_class] = ACTIONS(1415),
    [anon_sym_AT] = ACTIONS(1415),
    [anon_sym_not] = ACTIONS(1415),
    [anon_sym_and] = ACTIONS(1415),
    [anon_sym_or] = ACTIONS(1415),
    [anon_sym_PLUS] = ACTIONS(1415),
    [anon_sym_DASH] = ACTIONS(1415),
    [anon_sym_SLASH] = ACTIONS(1415),
    [anon_sym_SLASH_SLASH] = ACTIONS(1415),
    [anon_sym_PERCENT] = ACTIONS(1415),
    [anon_sym_PIPE] = ACTIONS(1415),
    [anon_sym_CARET] = ACTIONS(1415),
    [anon_sym_AMP] = ACTIONS(1415),
    [anon_sym_LT_LT] = ACTIONS(1415),
    [anon_sym_GT_GT] = ACTIONS(1415),
    [anon_sym_GT_GT_GT] = ACTIONS(1415),
    [anon_sym_TILDE] = ACTIONS(1415),
    [anon_sym_typeof] = ACTIONS(1415),
    [anon_sym_void] = ACTIONS(1415),
    [anon_sym_LT] = ACTIONS(1415),
    [anon_sym_LT_EQ] = ACTIONS(1415),
    [anon_sym_EQ_EQ] = ACTIONS(1415),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1415),
    [anon_sym_BANG_EQ] = ACTIONS(1415),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1415),
    [anon_sym_GT_EQ] = ACTIONS(1415),
    [anon_sym_GT] = ACTIONS(1415),
    [anon_sym_is] = ACTIONS(1415),
    [anon_sym_instanceof] = ACTIONS(1415),
    [anon_sym_new] = ACTIONS(1415),
    [anon_sym_QMARK] = ACTIONS(1415),
    [anon_sym_await] = ACTIONS(1415),
    [anon_sym_LBRACE] = ACTIONS(1415),
    [sym_this] = ACTIONS(1415),
    [sym_true] = ACTIONS(1415),
    [sym_false] = ACTIONS(1415),
    [sym_none] = ACTIONS(1415),
    [sym_string] = ACTIONS(1415),
    [sym_verbatim] = ACTIONS(1415),
    [sym_number] = ACTIONS(1415),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1413),
    [sym_regex] = ACTIONS(1413),
    [sym_fstring_start] = ACTIONS(1413),
  },
  [STATE(621)] = {
    [aux_sym__inline_body_repeat1] = STATE(621),
    [ts_builtin_sym_end] = ACTIONS(1369),
    [sym_identifier] = ACTIONS(1371),
    [anon_sym_SEMI] = ACTIONS(1564),
    [anon_sym_return] = ACTIONS(1371),
    [anon_sym_async] = ACTIONS(1371),
    [anon_sym_def] = ACTIONS(1371),
    [anon_sym_LPAREN] = ACTIONS(1371),
    [anon_sym_LBRACK] = ACTIONS(1371),
    [anon_sym_del] = ACTIONS(1371),
    [anon_sym_raise] = ACTIONS(1371),
    [sym_pass_statement] = ACTIONS(1371),
    [sym_break_statement] = ACTIONS(1371),
    [sym_continue_statement] = ACTIONS(1371),
    [sym_debugger_statement] = ACTIONS(1371),
    [anon_sym_assert] = ACTIONS(1371),
    [anon_sym_global] = ACTIONS(1371),
    [anon_sym_nonlocal] = ACTIONS(1371),
    [anon_sym_yield] = ACTIONS(1371),
    [anon_sym_from] = ACTIONS(1371),
    [anon_sym_import] = ACTIONS(1371),
    [anon_sym_DOT] = ACTIONS(1371),
    [anon_sym_if] = ACTIONS(1371),
    [anon_sym_for] = ACTIONS(1371),
    [anon_sym_in] = ACTIONS(1371),
    [anon_sym_while] = ACTIONS(1371),
    [anon_sym_do] = ACTIONS(1371),
    [anon_sym_try] = ACTIONS(1371),
    [anon_sym_with] = ACTIONS(1371),
    [anon_sym_STAR] = ACTIONS(1371),
    [anon_sym_STAR_STAR] = ACTIONS(1371),
    [anon_sym_class] = ACTIONS(1371),
    [anon_sym_AT] = ACTIONS(1371),
    [anon_sym_not] = ACTIONS(1371),
    [anon_sym_and] = ACTIONS(1371),
    [anon_sym_or] = ACTIONS(1371),
    [anon_sym_PLUS] = ACTIONS(1371),
    [anon_sym_DASH] = ACTIONS(1371),
    [anon_sym_SLASH] = ACTIONS(1371),
    [anon_sym_SLASH_SLASH] = ACTIONS(1371),
    [anon_sym_PERCENT] = ACTIONS(1371),
    [anon_sym_PIPE] = ACTIONS(1371),
    [anon_sym_CARET] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(1371),
    [anon_sym_LT_LT] = ACTIONS(1371),
    [anon_sym_GT_GT] = ACTIONS(1371),
    [anon_sym_GT_GT_GT] = ACTIONS(1371),
    [anon_sym_TILDE] = ACTIONS(1371),
    [anon_sym_typeof] = ACTIONS(1371),
    [anon_sym_void] = ACTIONS(1371),
    [anon_sym_LT] = ACTIONS(1371),
    [anon_sym_LT_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1371),
    [anon_sym_GT_EQ] = ACTIONS(1371),
    [anon_sym_GT] = ACTIONS(1371),
    [anon_sym_is] = ACTIONS(1371),
    [anon_sym_instanceof] = ACTIONS(1371),
    [anon_sym_new] = ACTIONS(1371),
    [anon_sym_QMARK] = ACTIONS(1371),
    [anon_sym_await] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1371),
    [sym_this] = ACTIONS(1371),
    [sym_true] = ACTIONS(1371),
    [sym_false] = ACTIONS(1371),
    [sym_none] = ACTIONS(1371),
    [sym_string] = ACTIONS(1371),
    [sym_verbatim] = ACTIONS(1371),
    [sym_number] = ACTIONS(1371),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1369),
    [sym_fstring_start] = ACTIONS(1369),
  },
  [STATE(622)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(623)] = {
    [aux_sym__inline_body_repeat1] = STATE(661),
    [sym_identifier] = ACTIONS(73),
    [anon_sym_SEMI] = ACTIONS(1567),
    [anon_sym_return] = ACTIONS(73),
    [anon_sym_async] = ACTIONS(73),
    [anon_sym_def] = ACTIONS(73),
    [anon_sym_LPAREN] = ACTIONS(73),
    [anon_sym_LBRACK] = ACTIONS(73),
    [anon_sym_del] = ACTIONS(73),
    [anon_sym_raise] = ACTIONS(73),
    [sym_pass_statement] = ACTIONS(73),
    [sym_break_statement] = ACTIONS(73),
    [sym_continue_statement] = ACTIONS(73),
    [sym_debugger_statement] = ACTIONS(73),
    [anon_sym_assert] = ACTIONS(73),
    [anon_sym_global] = ACTIONS(73),
    [anon_sym_nonlocal] = ACTIONS(73),
    [anon_sym_yield] = ACTIONS(73),
    [anon_sym_from] = ACTIONS(73),
    [anon_sym_import] = ACTIONS(73),
    [anon_sym_DOT] = ACTIONS(73),
    [anon_sym_if] = ACTIONS(73),
    [anon_sym_for] = ACTIONS(73),
    [anon_sym_in] = ACTIONS(73),
    [anon_sym_while] = ACTIONS(73),
    [anon_sym_do] = ACTIONS(73),
    [anon_sym_try] = ACTIONS(73),
    [anon_sym_with] = ACTIONS(73),
    [anon_sym_STAR] = ACTIONS(73),
    [anon_sym_STAR_STAR] = ACTIONS(73),
    [anon_sym_class] = ACTIONS(73),
    [anon_sym_AT] = ACTIONS(73),
    [anon_sym_not] = ACTIONS(73),
    [anon_sym_and] = ACTIONS(73),
    [anon_sym_or] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(73),
    [anon_sym_DASH] = ACTIONS(73),
    [anon_sym_SLASH] = ACTIONS(73),
    [anon_sym_SLASH_SLASH] = ACTIONS(73),
    [anon_sym_PERCENT] = ACTIONS(73),
    [anon_sym_PIPE] = ACTIONS(73),
    [anon_sym_CARET] = ACTIONS(73),
    [anon_sym_AMP] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(73),
    [anon_sym_GT_GT] = ACTIONS(73),
    [anon_sym_GT_GT_GT] = ACTIONS(73),
    [anon_sym_TILDE] = ACTIONS(73),
    [anon_sym_typeof] = ACTIONS(73),
    [anon_sym_void] = ACTIONS(73),
    [anon_sym_LT] = ACTIONS(73),
    [anon_sym_LT_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ] = ACTIONS(73),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ] = ACTIONS(73),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(73),
    [anon_sym_GT_EQ] = ACTIONS(73),
    [anon_sym_GT] = ACTIONS(73),
    [anon_sym_is] = ACTIONS(73),
    [anon_sym_instanceof] = ACTIONS(73),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_QMARK] = ACTIONS(73),
    [anon_sym_await] = ACTIONS(73),
    [anon_sym_LBRACE] = ACTIONS(73),
    [sym_this] = ACTIONS(73),
    [sym_true] = ACTIONS(73),
    [sym_false] = ACTIONS(73),
    [sym_none] = ACTIONS(73),
    [sym_string] = ACTIONS(73),
    [sym_verbatim] = ACTIONS(73),
    [sym_number] = ACTIONS(73),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(109),
    [sym_regex] = ACTIONS(109),
    [sym_fstring_start] = ACTIONS(109),
  },
  [STATE(624)] = {
    [ts_builtin_sym_end] = ACTIONS(1569),
    [sym_identifier] = ACTIONS(1571),
    [anon_sym_SEMI] = ACTIONS(1571),
    [anon_sym_return] = ACTIONS(1571),
    [anon_sym_async] = ACTIONS(1571),
    [anon_sym_def] = ACTIONS(1571),
    [anon_sym_COMMA] = ACTIONS(1571),
    [anon_sym_LPAREN] = ACTIONS(1571),
    [anon_sym_LBRACK] = ACTIONS(1571),
    [anon_sym_del] = ACTIONS(1571),
    [anon_sym_raise] = ACTIONS(1571),
    [sym_pass_statement] = ACTIONS(1571),
    [sym_break_statement] = ACTIONS(1571),
    [sym_continue_statement] = ACTIONS(1571),
    [sym_debugger_statement] = ACTIONS(1571),
    [anon_sym_assert] = ACTIONS(1571),
    [anon_sym_global] = ACTIONS(1571),
    [anon_sym_nonlocal] = ACTIONS(1571),
    [anon_sym_yield] = ACTIONS(1571),
    [anon_sym_from] = ACTIONS(1571),
    [anon_sym_import] = ACTIONS(1571),
    [anon_sym_DOT] = ACTIONS(1571),
    [anon_sym_if] = ACTIONS(1571),
    [anon_sym_for] = ACTIONS(1571),
    [anon_sym_in] = ACTIONS(1571),
    [anon_sym_while] = ACTIONS(1571),
    [anon_sym_do] = ACTIONS(1571),
    [anon_sym_try] = ACTIONS(1571),
    [anon_sym_with] = ACTIONS(1571),
    [anon_sym_STAR] = ACTIONS(1571),
    [anon_sym_STAR_STAR] = ACTIONS(1571),
    [anon_sym_class] = ACTIONS(1571),
    [anon_sym_AT] = ACTIONS(1571),
    [anon_sym_not] = ACTIONS(1571),
    [anon_sym_and] = ACTIONS(1571),
    [anon_sym_or] = ACTIONS(1571),
    [anon_sym_PLUS] = ACTIONS(1571),
    [anon_sym_DASH] = ACTIONS(1571),
    [anon_sym_SLASH] = ACTIONS(1571),
    [anon_sym_SLASH_SLASH] = ACTIONS(1571),
    [anon_sym_PERCENT] = ACTIONS(1571),
    [anon_sym_PIPE] = ACTIONS(1571),
    [anon_sym_CARET] = ACTIONS(1571),
    [anon_sym_AMP] = ACTIONS(1571),
    [anon_sym_LT_LT] = ACTIONS(1571),
    [anon_sym_GT_GT] = ACTIONS(1571),
    [anon_sym_GT_GT_GT] = ACTIONS(1571),
    [anon_sym_TILDE] = ACTIONS(1571),
    [anon_sym_typeof] = ACTIONS(1571),
    [anon_sym_void] = ACTIONS(1571),
    [anon_sym_LT] = ACTIONS(1571),
    [anon_sym_LT_EQ] = ACTIONS(1571),
    [anon_sym_EQ_EQ] = ACTIONS(1571),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1571),
    [anon_sym_BANG_EQ] = ACTIONS(1571),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1571),
    [anon_sym_GT_EQ] = ACTIONS(1571),
    [anon_sym_GT] = ACTIONS(1571),
    [anon_sym_is] = ACTIONS(1571),
    [anon_sym_instanceof] = ACTIONS(1571),
    [anon_sym_new] = ACTIONS(1571),
    [anon_sym_QMARK] = ACTIONS(1571),
    [anon_sym_await] = ACTIONS(1571),
    [anon_sym_LBRACE] = ACTIONS(1571),
    [sym_this] = ACTIONS(1571),
    [sym_true] = ACTIONS(1571),
    [sym_false] = ACTIONS(1571),
    [sym_none] = ACTIONS(1571),
    [sym_string] = ACTIONS(1571),
    [sym_verbatim] = ACTIONS(1571),
    [sym_number] = ACTIONS(1571),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1569),
    [sym_fstring_start] = ACTIONS(1569),
  },
  [STATE(625)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(626)] = {
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_SEMI] = ACTIONS(1524),
    [anon_sym_return] = ACTIONS(1524),
    [anon_sym_async] = ACTIONS(1524),
    [anon_sym_def] = ACTIONS(1524),
    [anon_sym_COMMA] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1524),
    [anon_sym_LBRACK] = ACTIONS(1524),
    [anon_sym_del] = ACTIONS(1524),
    [anon_sym_raise] = ACTIONS(1524),
    [sym_pass_statement] = ACTIONS(1524),
    [sym_break_statement] = ACTIONS(1524),
    [sym_continue_statement] = ACTIONS(1524),
    [sym_debugger_statement] = ACTIONS(1524),
    [anon_sym_assert] = ACTIONS(1524),
    [anon_sym_global] = ACTIONS(1524),
    [anon_sym_nonlocal] = ACTIONS(1524),
    [anon_sym_yield] = ACTIONS(1524),
    [anon_sym_from] = ACTIONS(1524),
    [anon_sym_import] = ACTIONS(1524),
    [anon_sym_DOT] = ACTIONS(1524),
    [anon_sym_if] = ACTIONS(1524),
    [anon_sym_for] = ACTIONS(1524),
    [anon_sym_in] = ACTIONS(1524),
    [anon_sym_while] = ACTIONS(1524),
    [anon_sym_do] = ACTIONS(1524),
    [anon_sym_try] = ACTIONS(1524),
    [anon_sym_with] = ACTIONS(1524),
    [anon_sym_STAR] = ACTIONS(1524),
    [anon_sym_STAR_STAR] = ACTIONS(1524),
    [anon_sym_class] = ACTIONS(1524),
    [anon_sym_AT] = ACTIONS(1524),
    [anon_sym_not] = ACTIONS(1524),
    [anon_sym_and] = ACTIONS(1524),
    [anon_sym_or] = ACTIONS(1524),
    [anon_sym_PLUS] = ACTIONS(1524),
    [anon_sym_DASH] = ACTIONS(1524),
    [anon_sym_SLASH] = ACTIONS(1524),
    [anon_sym_SLASH_SLASH] = ACTIONS(1524),
    [anon_sym_PERCENT] = ACTIONS(1524),
    [anon_sym_PIPE] = ACTIONS(1524),
    [anon_sym_CARET] = ACTIONS(1524),
    [anon_sym_AMP] = ACTIONS(1524),
    [anon_sym_LT_LT] = ACTIONS(1524),
    [anon_sym_GT_GT] = ACTIONS(1524),
    [anon_sym_GT_GT_GT] = ACTIONS(1524),
    [anon_sym_TILDE] = ACTIONS(1524),
    [anon_sym_typeof] = ACTIONS(1524),
    [anon_sym_void] = ACTIONS(1524),
    [anon_sym_LT] = ACTIONS(1524),
    [anon_sym_LT_EQ] = ACTIONS(1524),
    [anon_sym_EQ_EQ] = ACTIONS(1524),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1524),
    [anon_sym_BANG_EQ] = ACTIONS(1524),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1524),
    [anon_sym_GT_EQ] = ACTIONS(1524),
    [anon_sym_GT] = ACTIONS(1524),
    [anon_sym_is] = ACTIONS(1524),
    [anon_sym_instanceof] = ACTIONS(1524),
    [anon_sym_new] = ACTIONS(1524),
    [anon_sym_QMARK] = ACTIONS(1524),
    [anon_sym_await] = ACTIONS(1524),
    [anon_sym_LBRACE] = ACTIONS(1524),
    [sym_this] = ACTIONS(1524),
    [sym_true] = ACTIONS(1524),
    [sym_false] = ACTIONS(1524),
    [sym_none] = ACTIONS(1524),
    [sym_string] = ACTIONS(1524),
    [sym_verbatim] = ACTIONS(1524),
    [sym_number] = ACTIONS(1524),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1522),
    [sym_regex] = ACTIONS(1522),
    [sym_fstring_start] = ACTIONS(1522),
  },
  [STATE(627)] = {
    [ts_builtin_sym_end] = ACTIONS(1573),
    [sym_identifier] = ACTIONS(1575),
    [anon_sym_SEMI] = ACTIONS(1575),
    [anon_sym_return] = ACTIONS(1575),
    [anon_sym_async] = ACTIONS(1575),
    [anon_sym_def] = ACTIONS(1575),
    [anon_sym_COMMA] = ACTIONS(1575),
    [anon_sym_LPAREN] = ACTIONS(1575),
    [anon_sym_LBRACK] = ACTIONS(1575),
    [anon_sym_del] = ACTIONS(1575),
    [anon_sym_raise] = ACTIONS(1575),
    [sym_pass_statement] = ACTIONS(1575),
    [sym_break_statement] = ACTIONS(1575),
    [sym_continue_statement] = ACTIONS(1575),
    [sym_debugger_statement] = ACTIONS(1575),
    [anon_sym_assert] = ACTIONS(1575),
    [anon_sym_global] = ACTIONS(1575),
    [anon_sym_nonlocal] = ACTIONS(1575),
    [anon_sym_yield] = ACTIONS(1575),
    [anon_sym_from] = ACTIONS(1575),
    [anon_sym_import] = ACTIONS(1575),
    [anon_sym_DOT] = ACTIONS(1575),
    [anon_sym_if] = ACTIONS(1575),
    [anon_sym_for] = ACTIONS(1575),
    [anon_sym_in] = ACTIONS(1575),
    [anon_sym_while] = ACTIONS(1575),
    [anon_sym_do] = ACTIONS(1575),
    [anon_sym_try] = ACTIONS(1575),
    [anon_sym_with] = ACTIONS(1575),
    [anon_sym_STAR] = ACTIONS(1575),
    [anon_sym_STAR_STAR] = ACTIONS(1575),
    [anon_sym_class] = ACTIONS(1575),
    [anon_sym_AT] = ACTIONS(1575),
    [anon_sym_not] = ACTIONS(1575),
    [anon_sym_and] = ACTIONS(1575),
    [anon_sym_or] = ACTIONS(1575),
    [anon_sym_PLUS] = ACTIONS(1575),
    [anon_sym_DASH] = ACTIONS(1575),
    [anon_sym_SLASH] = ACTIONS(1575),
    [anon_sym_SLASH_SLASH] = ACTIONS(1575),
    [anon_sym_PERCENT] = ACTIONS(1575),
    [anon_sym_PIPE] = ACTIONS(1575),
    [anon_sym_CARET] = ACTIONS(1575),
    [anon_sym_AMP] = ACTIONS(1575),
    [anon_sym_LT_LT] = ACTIONS(1575),
    [anon_sym_GT_GT] = ACTIONS(1575),
    [anon_sym_GT_GT_GT] = ACTIONS(1575),
    [anon_sym_TILDE] = ACTIONS(1575),
    [anon_sym_typeof] = ACTIONS(1575),
    [anon_sym_void] = ACTIONS(1575),
    [anon_sym_LT] = ACTIONS(1575),
    [anon_sym_LT_EQ] = ACTIONS(1575),
    [anon_sym_EQ_EQ] = ACTIONS(1575),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1575),
    [anon_sym_BANG_EQ] = ACTIONS(1575),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1575),
    [anon_sym_GT_EQ] = ACTIONS(1575),
    [anon_sym_GT] = ACTIONS(1575),
    [anon_sym_is] = ACTIONS(1575),
    [anon_sym_instanceof] = ACTIONS(1575),
    [anon_sym_new] = ACTIONS(1575),
    [anon_sym_QMARK] = ACTIONS(1575),
    [anon_sym_await] = ACTIONS(1575),
    [anon_sym_LBRACE] = ACTIONS(1575),
    [sym_this] = ACTIONS(1575),
    [sym_true] = ACTIONS(1575),
    [sym_false] = ACTIONS(1575),
    [sym_none] = ACTIONS(1575),
    [sym_string] = ACTIONS(1575),
    [sym_verbatim] = ACTIONS(1575),
    [sym_number] = ACTIONS(1575),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1573),
    [sym_fstring_start] = ACTIONS(1573),
  },
  [STATE(628)] = {
    [ts_builtin_sym_end] = ACTIONS(1577),
    [sym_identifier] = ACTIONS(1579),
    [anon_sym_SEMI] = ACTIONS(1579),
    [anon_sym_return] = ACTIONS(1579),
    [anon_sym_async] = ACTIONS(1579),
    [anon_sym_def] = ACTIONS(1579),
    [anon_sym_COMMA] = ACTIONS(1579),
    [anon_sym_LPAREN] = ACTIONS(1579),
    [anon_sym_LBRACK] = ACTIONS(1579),
    [anon_sym_del] = ACTIONS(1579),
    [anon_sym_raise] = ACTIONS(1579),
    [sym_pass_statement] = ACTIONS(1579),
    [sym_break_statement] = ACTIONS(1579),
    [sym_continue_statement] = ACTIONS(1579),
    [sym_debugger_statement] = ACTIONS(1579),
    [anon_sym_assert] = ACTIONS(1579),
    [anon_sym_global] = ACTIONS(1579),
    [anon_sym_nonlocal] = ACTIONS(1579),
    [anon_sym_yield] = ACTIONS(1579),
    [anon_sym_from] = ACTIONS(1579),
    [anon_sym_import] = ACTIONS(1579),
    [anon_sym_DOT] = ACTIONS(1579),
    [anon_sym_if] = ACTIONS(1579),
    [anon_sym_for] = ACTIONS(1579),
    [anon_sym_in] = ACTIONS(1579),
    [anon_sym_while] = ACTIONS(1579),
    [anon_sym_do] = ACTIONS(1579),
    [anon_sym_try] = ACTIONS(1579),
    [anon_sym_with] = ACTIONS(1579),
    [anon_sym_STAR] = ACTIONS(1579),
    [anon_sym_STAR_STAR] = ACTIONS(1579),
    [anon_sym_class] = ACTIONS(1579),
    [anon_sym_AT] = ACTIONS(1579),
    [anon_sym_not] = ACTIONS(1579),
    [anon_sym_and] = ACTIONS(1579),
    [anon_sym_or] = ACTIONS(1579),
    [anon_sym_PLUS] = ACTIONS(1579),
    [anon_sym_DASH] = ACTIONS(1579),
    [anon_sym_SLASH] = ACTIONS(1579),
    [anon_sym_SLASH_SLASH] = ACTIONS(1579),
    [anon_sym_PERCENT] = ACTIONS(1579),
    [anon_sym_PIPE] = ACTIONS(1579),
    [anon_sym_CARET] = ACTIONS(1579),
    [anon_sym_AMP] = ACTIONS(1579),
    [anon_sym_LT_LT] = ACTIONS(1579),
    [anon_sym_GT_GT] = ACTIONS(1579),
    [anon_sym_GT_GT_GT] = ACTIONS(1579),
    [anon_sym_TILDE] = ACTIONS(1579),
    [anon_sym_typeof] = ACTIONS(1579),
    [anon_sym_void] = ACTIONS(1579),
    [anon_sym_LT] = ACTIONS(1579),
    [anon_sym_LT_EQ] = ACTIONS(1579),
    [anon_sym_EQ_EQ] = ACTIONS(1579),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1579),
    [anon_sym_BANG_EQ] = ACTIONS(1579),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1579),
    [anon_sym_GT_EQ] = ACTIONS(1579),
    [anon_sym_GT] = ACTIONS(1579),
    [anon_sym_is] = ACTIONS(1579),
    [anon_sym_instanceof] = ACTIONS(1579),
    [anon_sym_new] = ACTIONS(1579),
    [anon_sym_QMARK] = ACTIONS(1579),
    [anon_sym_await] = ACTIONS(1579),
    [anon_sym_LBRACE] = ACTIONS(1579),
    [sym_this] = ACTIONS(1579),
    [sym_true] = ACTIONS(1579),
    [sym_false] = ACTIONS(1579),
    [sym_none] = ACTIONS(1579),
    [sym_string] = ACTIONS(1579),
    [sym_verbatim] = ACTIONS(1579),
    [sym_number] = ACTIONS(1579),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1577),
    [sym_fstring_start] = ACTIONS(1577),
  },
  [STATE(629)] = {
    [sym_identifier] = ACTIONS(1443),
    [anon_sym_SEMI] = ACTIONS(1443),
    [anon_sym_return] = ACTIONS(1443),
    [anon_sym_async] = ACTIONS(1443),
    [anon_sym_def] = ACTIONS(1443),
    [anon_sym_COMMA] = ACTIONS(1443),
    [anon_sym_LPAREN] = ACTIONS(1443),
    [anon_sym_LBRACK] = ACTIONS(1443),
    [anon_sym_del] = ACTIONS(1443),
    [anon_sym_raise] = ACTIONS(1443),
    [sym_pass_statement] = ACTIONS(1443),
    [sym_break_statement] = ACTIONS(1443),
    [sym_continue_statement] = ACTIONS(1443),
    [sym_debugger_statement] = ACTIONS(1443),
    [anon_sym_assert] = ACTIONS(1443),
    [anon_sym_global] = ACTIONS(1443),
    [anon_sym_nonlocal] = ACTIONS(1443),
    [anon_sym_yield] = ACTIONS(1443),
    [anon_sym_from] = ACTIONS(1443),
    [anon_sym_import] = ACTIONS(1443),
    [anon_sym_DOT] = ACTIONS(1443),
    [anon_sym_if] = ACTIONS(1443),
    [anon_sym_for] = ACTIONS(1443),
    [anon_sym_in] = ACTIONS(1443),
    [anon_sym_while] = ACTIONS(1443),
    [anon_sym_do] = ACTIONS(1443),
    [anon_sym_try] = ACTIONS(1443),
    [anon_sym_with] = ACTIONS(1443),
    [anon_sym_STAR] = ACTIONS(1443),
    [anon_sym_STAR_STAR] = ACTIONS(1443),
    [anon_sym_class] = ACTIONS(1443),
    [anon_sym_AT] = ACTIONS(1443),
    [anon_sym_not] = ACTIONS(1443),
    [anon_sym_and] = ACTIONS(1443),
    [anon_sym_or] = ACTIONS(1443),
    [anon_sym_PLUS] = ACTIONS(1443),
    [anon_sym_DASH] = ACTIONS(1443),
    [anon_sym_SLASH] = ACTIONS(1443),
    [anon_sym_SLASH_SLASH] = ACTIONS(1443),
    [anon_sym_PERCENT] = ACTIONS(1443),
    [anon_sym_PIPE] = ACTIONS(1443),
    [anon_sym_CARET] = ACTIONS(1443),
    [anon_sym_AMP] = ACTIONS(1443),
    [anon_sym_LT_LT] = ACTIONS(1443),
    [anon_sym_GT_GT] = ACTIONS(1443),
    [anon_sym_GT_GT_GT] = ACTIONS(1443),
    [anon_sym_TILDE] = ACTIONS(1443),
    [anon_sym_typeof] = ACTIONS(1443),
    [anon_sym_void] = ACTIONS(1443),
    [anon_sym_LT] = ACTIONS(1443),
    [anon_sym_LT_EQ] = ACTIONS(1443),
    [anon_sym_EQ_EQ] = ACTIONS(1443),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1443),
    [anon_sym_BANG_EQ] = ACTIONS(1443),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1443),
    [anon_sym_GT_EQ] = ACTIONS(1443),
    [anon_sym_GT] = ACTIONS(1443),
    [anon_sym_is] = ACTIONS(1443),
    [anon_sym_instanceof] = ACTIONS(1443),
    [anon_sym_new] = ACTIONS(1443),
    [anon_sym_QMARK] = ACTIONS(1443),
    [anon_sym_await] = ACTIONS(1443),
    [anon_sym_LBRACE] = ACTIONS(1443),
    [sym_this] = ACTIONS(1443),
    [sym_true] = ACTIONS(1443),
    [sym_false] = ACTIONS(1443),
    [sym_none] = ACTIONS(1443),
    [sym_string] = ACTIONS(1443),
    [sym_verbatim] = ACTIONS(1443),
    [sym_number] = ACTIONS(1443),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1441),
    [sym_regex] = ACTIONS(1441),
    [sym_fstring_start] = ACTIONS(1441),
  },
  [STATE(630)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(631)] = {
    [sym_identifier] = ACTIONS(1528),
    [anon_sym_SEMI] = ACTIONS(1528),
    [anon_sym_return] = ACTIONS(1528),
    [anon_sym_async] = ACTIONS(1528),
    [anon_sym_def] = ACTIONS(1528),
    [anon_sym_COMMA] = ACTIONS(1528),
    [anon_sym_LPAREN] = ACTIONS(1528),
    [anon_sym_LBRACK] = ACTIONS(1528),
    [anon_sym_del] = ACTIONS(1528),
    [anon_sym_raise] = ACTIONS(1528),
    [sym_pass_statement] = ACTIONS(1528),
    [sym_break_statement] = ACTIONS(1528),
    [sym_continue_statement] = ACTIONS(1528),
    [sym_debugger_statement] = ACTIONS(1528),
    [anon_sym_assert] = ACTIONS(1528),
    [anon_sym_global] = ACTIONS(1528),
    [anon_sym_nonlocal] = ACTIONS(1528),
    [anon_sym_yield] = ACTIONS(1528),
    [anon_sym_from] = ACTIONS(1528),
    [anon_sym_import] = ACTIONS(1528),
    [anon_sym_DOT] = ACTIONS(1528),
    [anon_sym_if] = ACTIONS(1528),
    [anon_sym_for] = ACTIONS(1528),
    [anon_sym_in] = ACTIONS(1528),
    [anon_sym_while] = ACTIONS(1528),
    [anon_sym_do] = ACTIONS(1528),
    [anon_sym_try] = ACTIONS(1528),
    [anon_sym_with] = ACTIONS(1528),
    [anon_sym_STAR] = ACTIONS(1528),
    [anon_sym_STAR_STAR] = ACTIONS(1528),
    [anon_sym_class] = ACTIONS(1528),
    [anon_sym_AT] = ACTIONS(1528),
    [anon_sym_not] = ACTIONS(1528),
    [anon_sym_and] = ACTIONS(1528),
    [anon_sym_or] = ACTIONS(1528),
    [anon_sym_PLUS] = ACTIONS(1528),
    [anon_sym_DASH] = ACTIONS(1528),
    [anon_sym_SLASH] = ACTIONS(1528),
    [anon_sym_SLASH_SLASH] = ACTIONS(1528),
    [anon_sym_PERCENT] = ACTIONS(1528),
    [anon_sym_PIPE] = ACTIONS(1528),
    [anon_sym_CARET] = ACTIONS(1528),
    [anon_sym_AMP] = ACTIONS(1528),
    [anon_sym_LT_LT] = ACTIONS(1528),
    [anon_sym_GT_GT] = ACTIONS(1528),
    [anon_sym_GT_GT_GT] = ACTIONS(1528),
    [anon_sym_TILDE] = ACTIONS(1528),
    [anon_sym_typeof] = ACTIONS(1528),
    [anon_sym_void] = ACTIONS(1528),
    [anon_sym_LT] = ACTIONS(1528),
    [anon_sym_LT_EQ] = ACTIONS(1528),
    [anon_sym_EQ_EQ] = ACTIONS(1528),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1528),
    [anon_sym_BANG_EQ] = ACTIONS(1528),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1528),
    [anon_sym_GT_EQ] = ACTIONS(1528),
    [anon_sym_GT] = ACTIONS(1528),
    [anon_sym_is] = ACTIONS(1528),
    [anon_sym_instanceof] = ACTIONS(1528),
    [anon_sym_new] = ACTIONS(1528),
    [anon_sym_QMARK] = ACTIONS(1528),
    [anon_sym_await] = ACTIONS(1528),
    [anon_sym_LBRACE] = ACTIONS(1528),
    [sym_this] = ACTIONS(1528),
    [sym_true] = ACTIONS(1528),
    [sym_false] = ACTIONS(1528),
    [sym_none] = ACTIONS(1528),
    [sym_string] = ACTIONS(1528),
    [sym_verbatim] = ACTIONS(1528),
    [sym_number] = ACTIONS(1528),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1526),
    [sym_regex] = ACTIONS(1526),
    [sym_fstring_start] = ACTIONS(1526),
  },
  [STATE(632)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1272),
    [sym_identifier] = ACTIONS(1274),
    [anon_sym_SEMI] = ACTIONS(1274),
    [anon_sym_return] = ACTIONS(1274),
    [anon_sym_async] = ACTIONS(1274),
    [anon_sym_def] = ACTIONS(1274),
    [anon_sym_LPAREN] = ACTIONS(1274),
    [anon_sym_LBRACK] = ACTIONS(1274),
    [anon_sym_del] = ACTIONS(1274),
    [anon_sym_raise] = ACTIONS(1274),
    [sym_pass_statement] = ACTIONS(1274),
    [sym_break_statement] = ACTIONS(1274),
    [sym_continue_statement] = ACTIONS(1274),
    [sym_debugger_statement] = ACTIONS(1274),
    [anon_sym_assert] = ACTIONS(1274),
    [anon_sym_global] = ACTIONS(1274),
    [anon_sym_nonlocal] = ACTIONS(1274),
    [anon_sym_yield] = ACTIONS(1274),
    [anon_sym_from] = ACTIONS(1274),
    [anon_sym_import] = ACTIONS(1274),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(1518),
    [anon_sym_for] = ACTIONS(1274),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1274),
    [anon_sym_do] = ACTIONS(1274),
    [anon_sym_try] = ACTIONS(1274),
    [anon_sym_with] = ACTIONS(1274),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1274),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1494),
    [anon_sym_or] = ACTIONS(1496),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1274),
    [anon_sym_typeof] = ACTIONS(1274),
    [anon_sym_void] = ACTIONS(1274),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1274),
    [anon_sym_QMARK] = ACTIONS(1274),
    [anon_sym_await] = ACTIONS(1274),
    [anon_sym_LBRACE] = ACTIONS(1274),
    [sym_this] = ACTIONS(1274),
    [sym_true] = ACTIONS(1274),
    [sym_false] = ACTIONS(1274),
    [sym_none] = ACTIONS(1274),
    [sym_string] = ACTIONS(1274),
    [sym_verbatim] = ACTIONS(1274),
    [sym_number] = ACTIONS(1274),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1272),
    [sym_fstring_start] = ACTIONS(1272),
  },
  [STATE(633)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1241),
    [sym_identifier] = ACTIONS(1243),
    [anon_sym_SEMI] = ACTIONS(1243),
    [anon_sym_return] = ACTIONS(1243),
    [anon_sym_async] = ACTIONS(1243),
    [anon_sym_def] = ACTIONS(1243),
    [anon_sym_LPAREN] = ACTIONS(1243),
    [anon_sym_LBRACK] = ACTIONS(1243),
    [anon_sym_del] = ACTIONS(1243),
    [anon_sym_raise] = ACTIONS(1243),
    [sym_pass_statement] = ACTIONS(1243),
    [sym_break_statement] = ACTIONS(1243),
    [sym_continue_statement] = ACTIONS(1243),
    [sym_debugger_statement] = ACTIONS(1243),
    [anon_sym_assert] = ACTIONS(1243),
    [anon_sym_global] = ACTIONS(1243),
    [anon_sym_nonlocal] = ACTIONS(1243),
    [anon_sym_yield] = ACTIONS(1243),
    [anon_sym_from] = ACTIONS(1243),
    [anon_sym_import] = ACTIONS(1243),
    [anon_sym_DOT] = ACTIONS(1243),
    [anon_sym_if] = ACTIONS(1243),
    [anon_sym_for] = ACTIONS(1243),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1243),
    [anon_sym_do] = ACTIONS(1243),
    [anon_sym_try] = ACTIONS(1243),
    [anon_sym_with] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1243),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1494),
    [anon_sym_or] = ACTIONS(1243),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1243),
    [anon_sym_typeof] = ACTIONS(1243),
    [anon_sym_void] = ACTIONS(1243),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1243),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_await] = ACTIONS(1243),
    [anon_sym_LBRACE] = ACTIONS(1243),
    [sym_this] = ACTIONS(1243),
    [sym_true] = ACTIONS(1243),
    [sym_false] = ACTIONS(1243),
    [sym_none] = ACTIONS(1243),
    [sym_string] = ACTIONS(1243),
    [sym_verbatim] = ACTIONS(1243),
    [sym_number] = ACTIONS(1243),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1241),
    [sym_fstring_start] = ACTIONS(1241),
  },
  [STATE(634)] = {
    [sym_identifier] = ACTIONS(1532),
    [anon_sym_SEMI] = ACTIONS(1532),
    [anon_sym_return] = ACTIONS(1532),
    [anon_sym_async] = ACTIONS(1532),
    [anon_sym_def] = ACTIONS(1532),
    [anon_sym_COMMA] = ACTIONS(1532),
    [anon_sym_LPAREN] = ACTIONS(1532),
    [anon_sym_LBRACK] = ACTIONS(1532),
    [anon_sym_del] = ACTIONS(1532),
    [anon_sym_raise] = ACTIONS(1532),
    [sym_pass_statement] = ACTIONS(1532),
    [sym_break_statement] = ACTIONS(1532),
    [sym_continue_statement] = ACTIONS(1532),
    [sym_debugger_statement] = ACTIONS(1532),
    [anon_sym_assert] = ACTIONS(1532),
    [anon_sym_global] = ACTIONS(1532),
    [anon_sym_nonlocal] = ACTIONS(1532),
    [anon_sym_yield] = ACTIONS(1532),
    [anon_sym_from] = ACTIONS(1532),
    [anon_sym_import] = ACTIONS(1532),
    [anon_sym_DOT] = ACTIONS(1532),
    [anon_sym_if] = ACTIONS(1532),
    [anon_sym_for] = ACTIONS(1532),
    [anon_sym_in] = ACTIONS(1532),
    [anon_sym_while] = ACTIONS(1532),
    [anon_sym_do] = ACTIONS(1532),
    [anon_sym_try] = ACTIONS(1532),
    [anon_sym_with] = ACTIONS(1532),
    [anon_sym_STAR] = ACTIONS(1532),
    [anon_sym_STAR_STAR] = ACTIONS(1532),
    [anon_sym_class] = ACTIONS(1532),
    [anon_sym_AT] = ACTIONS(1532),
    [anon_sym_not] = ACTIONS(1532),
    [anon_sym_and] = ACTIONS(1532),
    [anon_sym_or] = ACTIONS(1532),
    [anon_sym_PLUS] = ACTIONS(1532),
    [anon_sym_DASH] = ACTIONS(1532),
    [anon_sym_SLASH] = ACTIONS(1532),
    [anon_sym_SLASH_SLASH] = ACTIONS(1532),
    [anon_sym_PERCENT] = ACTIONS(1532),
    [anon_sym_PIPE] = ACTIONS(1532),
    [anon_sym_CARET] = ACTIONS(1532),
    [anon_sym_AMP] = ACTIONS(1532),
    [anon_sym_LT_LT] = ACTIONS(1532),
    [anon_sym_GT_GT] = ACTIONS(1532),
    [anon_sym_GT_GT_GT] = ACTIONS(1532),
    [anon_sym_TILDE] = ACTIONS(1532),
    [anon_sym_typeof] = ACTIONS(1532),
    [anon_sym_void] = ACTIONS(1532),
    [anon_sym_LT] = ACTIONS(1532),
    [anon_sym_LT_EQ] = ACTIONS(1532),
    [anon_sym_EQ_EQ] = ACTIONS(1532),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1532),
    [anon_sym_BANG_EQ] = ACTIONS(1532),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1532),
    [anon_sym_GT_EQ] = ACTIONS(1532),
    [anon_sym_GT] = ACTIONS(1532),
    [anon_sym_is] = ACTIONS(1532),
    [anon_sym_instanceof] = ACTIONS(1532),
    [anon_sym_new] = ACTIONS(1532),
    [anon_sym_QMARK] = ACTIONS(1532),
    [anon_sym_await] = ACTIONS(1532),
    [anon_sym_LBRACE] = ACTIONS(1532),
    [sym_this] = ACTIONS(1532),
    [sym_true] = ACTIONS(1532),
    [sym_false] = ACTIONS(1532),
    [sym_none] = ACTIONS(1532),
    [sym_string] = ACTIONS(1532),
    [sym_verbatim] = ACTIONS(1532),
    [sym_number] = ACTIONS(1532),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1530),
    [sym_regex] = ACTIONS(1530),
    [sym_fstring_start] = ACTIONS(1530),
  },
  [STATE(635)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(636)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(637)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1241),
    [sym_identifier] = ACTIONS(1243),
    [anon_sym_SEMI] = ACTIONS(1243),
    [anon_sym_return] = ACTIONS(1243),
    [anon_sym_async] = ACTIONS(1243),
    [anon_sym_def] = ACTIONS(1243),
    [anon_sym_LPAREN] = ACTIONS(1243),
    [anon_sym_LBRACK] = ACTIONS(1243),
    [anon_sym_del] = ACTIONS(1243),
    [anon_sym_raise] = ACTIONS(1243),
    [sym_pass_statement] = ACTIONS(1243),
    [sym_break_statement] = ACTIONS(1243),
    [sym_continue_statement] = ACTIONS(1243),
    [sym_debugger_statement] = ACTIONS(1243),
    [anon_sym_assert] = ACTIONS(1243),
    [anon_sym_global] = ACTIONS(1243),
    [anon_sym_nonlocal] = ACTIONS(1243),
    [anon_sym_yield] = ACTIONS(1243),
    [anon_sym_from] = ACTIONS(1243),
    [anon_sym_import] = ACTIONS(1243),
    [anon_sym_DOT] = ACTIONS(1243),
    [anon_sym_if] = ACTIONS(1243),
    [anon_sym_for] = ACTIONS(1243),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1243),
    [anon_sym_do] = ACTIONS(1243),
    [anon_sym_try] = ACTIONS(1243),
    [anon_sym_with] = ACTIONS(1243),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1243),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1243),
    [anon_sym_or] = ACTIONS(1243),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1243),
    [anon_sym_typeof] = ACTIONS(1243),
    [anon_sym_void] = ACTIONS(1243),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1243),
    [anon_sym_QMARK] = ACTIONS(1243),
    [anon_sym_await] = ACTIONS(1243),
    [anon_sym_LBRACE] = ACTIONS(1243),
    [sym_this] = ACTIONS(1243),
    [sym_true] = ACTIONS(1243),
    [sym_false] = ACTIONS(1243),
    [sym_none] = ACTIONS(1243),
    [sym_string] = ACTIONS(1243),
    [sym_verbatim] = ACTIONS(1243),
    [sym_number] = ACTIONS(1243),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1241),
    [sym_fstring_start] = ACTIONS(1241),
  },
  [STATE(638)] = {
    [sym_identifier] = ACTIONS(1536),
    [anon_sym_SEMI] = ACTIONS(1536),
    [anon_sym_return] = ACTIONS(1536),
    [anon_sym_async] = ACTIONS(1536),
    [anon_sym_def] = ACTIONS(1536),
    [anon_sym_COMMA] = ACTIONS(1536),
    [anon_sym_LPAREN] = ACTIONS(1536),
    [anon_sym_LBRACK] = ACTIONS(1536),
    [anon_sym_del] = ACTIONS(1536),
    [anon_sym_raise] = ACTIONS(1536),
    [sym_pass_statement] = ACTIONS(1536),
    [sym_break_statement] = ACTIONS(1536),
    [sym_continue_statement] = ACTIONS(1536),
    [sym_debugger_statement] = ACTIONS(1536),
    [anon_sym_assert] = ACTIONS(1536),
    [anon_sym_global] = ACTIONS(1536),
    [anon_sym_nonlocal] = ACTIONS(1536),
    [anon_sym_yield] = ACTIONS(1536),
    [anon_sym_from] = ACTIONS(1536),
    [anon_sym_import] = ACTIONS(1536),
    [anon_sym_DOT] = ACTIONS(1536),
    [anon_sym_if] = ACTIONS(1536),
    [anon_sym_for] = ACTIONS(1536),
    [anon_sym_in] = ACTIONS(1536),
    [anon_sym_while] = ACTIONS(1536),
    [anon_sym_do] = ACTIONS(1536),
    [anon_sym_try] = ACTIONS(1536),
    [anon_sym_with] = ACTIONS(1536),
    [anon_sym_STAR] = ACTIONS(1536),
    [anon_sym_STAR_STAR] = ACTIONS(1536),
    [anon_sym_class] = ACTIONS(1536),
    [anon_sym_AT] = ACTIONS(1536),
    [anon_sym_not] = ACTIONS(1536),
    [anon_sym_and] = ACTIONS(1536),
    [anon_sym_or] = ACTIONS(1536),
    [anon_sym_PLUS] = ACTIONS(1536),
    [anon_sym_DASH] = ACTIONS(1536),
    [anon_sym_SLASH] = ACTIONS(1536),
    [anon_sym_SLASH_SLASH] = ACTIONS(1536),
    [anon_sym_PERCENT] = ACTIONS(1536),
    [anon_sym_PIPE] = ACTIONS(1536),
    [anon_sym_CARET] = ACTIONS(1536),
    [anon_sym_AMP] = ACTIONS(1536),
    [anon_sym_LT_LT] = ACTIONS(1536),
    [anon_sym_GT_GT] = ACTIONS(1536),
    [anon_sym_GT_GT_GT] = ACTIONS(1536),
    [anon_sym_TILDE] = ACTIONS(1536),
    [anon_sym_typeof] = ACTIONS(1536),
    [anon_sym_void] = ACTIONS(1536),
    [anon_sym_LT] = ACTIONS(1536),
    [anon_sym_LT_EQ] = ACTIONS(1536),
    [anon_sym_EQ_EQ] = ACTIONS(1536),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1536),
    [anon_sym_BANG_EQ] = ACTIONS(1536),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1536),
    [anon_sym_GT_EQ] = ACTIONS(1536),
    [anon_sym_GT] = ACTIONS(1536),
    [anon_sym_is] = ACTIONS(1536),
    [anon_sym_instanceof] = ACTIONS(1536),
    [anon_sym_new] = ACTIONS(1536),
    [anon_sym_QMARK] = ACTIONS(1536),
    [anon_sym_await] = ACTIONS(1536),
    [anon_sym_LBRACE] = ACTIONS(1536),
    [sym_this] = ACTIONS(1536),
    [sym_true] = ACTIONS(1536),
    [sym_false] = ACTIONS(1536),
    [sym_none] = ACTIONS(1536),
    [sym_string] = ACTIONS(1536),
    [sym_verbatim] = ACTIONS(1536),
    [sym_number] = ACTIONS(1536),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1534),
    [sym_regex] = ACTIONS(1534),
    [sym_fstring_start] = ACTIONS(1534),
  },
  [STATE(639)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1305),
    [anon_sym_SEMI] = ACTIONS(1305),
    [anon_sym_return] = ACTIONS(1305),
    [anon_sym_async] = ACTIONS(1305),
    [anon_sym_def] = ACTIONS(1305),
    [anon_sym_LPAREN] = ACTIONS(1305),
    [anon_sym_LBRACK] = ACTIONS(1305),
    [anon_sym_del] = ACTIONS(1305),
    [anon_sym_raise] = ACTIONS(1305),
    [sym_pass_statement] = ACTIONS(1305),
    [sym_break_statement] = ACTIONS(1305),
    [sym_continue_statement] = ACTIONS(1305),
    [sym_debugger_statement] = ACTIONS(1305),
    [anon_sym_assert] = ACTIONS(1305),
    [anon_sym_global] = ACTIONS(1305),
    [anon_sym_nonlocal] = ACTIONS(1305),
    [anon_sym_yield] = ACTIONS(1305),
    [anon_sym_from] = ACTIONS(1305),
    [anon_sym_import] = ACTIONS(1305),
    [anon_sym_DOT] = ACTIONS(1305),
    [anon_sym_if] = ACTIONS(1305),
    [anon_sym_for] = ACTIONS(1305),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1305),
    [anon_sym_do] = ACTIONS(1305),
    [anon_sym_try] = ACTIONS(1305),
    [anon_sym_with] = ACTIONS(1305),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1305),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1425),
    [anon_sym_or] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1305),
    [anon_sym_typeof] = ACTIONS(1305),
    [anon_sym_void] = ACTIONS(1305),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1305),
    [anon_sym_QMARK] = ACTIONS(1305),
    [anon_sym_await] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1305),
    [sym_this] = ACTIONS(1305),
    [sym_true] = ACTIONS(1305),
    [sym_false] = ACTIONS(1305),
    [sym_none] = ACTIONS(1305),
    [sym_string] = ACTIONS(1305),
    [sym_verbatim] = ACTIONS(1305),
    [sym_number] = ACTIONS(1305),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1303),
    [sym_regex] = ACTIONS(1303),
    [sym_fstring_start] = ACTIONS(1303),
  },
  [STATE(640)] = {
    [sym_identifier] = ACTIONS(1484),
    [anon_sym_SEMI] = ACTIONS(1484),
    [anon_sym_return] = ACTIONS(1484),
    [anon_sym_async] = ACTIONS(1484),
    [anon_sym_def] = ACTIONS(1484),
    [anon_sym_COMMA] = ACTIONS(1484),
    [anon_sym_LPAREN] = ACTIONS(1484),
    [anon_sym_LBRACK] = ACTIONS(1484),
    [anon_sym_del] = ACTIONS(1484),
    [anon_sym_raise] = ACTIONS(1484),
    [sym_pass_statement] = ACTIONS(1484),
    [sym_break_statement] = ACTIONS(1484),
    [sym_continue_statement] = ACTIONS(1484),
    [sym_debugger_statement] = ACTIONS(1484),
    [anon_sym_assert] = ACTIONS(1484),
    [anon_sym_global] = ACTIONS(1484),
    [anon_sym_nonlocal] = ACTIONS(1484),
    [anon_sym_yield] = ACTIONS(1484),
    [anon_sym_from] = ACTIONS(1484),
    [anon_sym_import] = ACTIONS(1484),
    [anon_sym_DOT] = ACTIONS(1484),
    [anon_sym_if] = ACTIONS(1484),
    [anon_sym_for] = ACTIONS(1484),
    [anon_sym_in] = ACTIONS(1484),
    [anon_sym_while] = ACTIONS(1484),
    [anon_sym_do] = ACTIONS(1484),
    [anon_sym_try] = ACTIONS(1484),
    [anon_sym_with] = ACTIONS(1484),
    [anon_sym_STAR] = ACTIONS(1484),
    [anon_sym_STAR_STAR] = ACTIONS(1484),
    [anon_sym_class] = ACTIONS(1484),
    [anon_sym_AT] = ACTIONS(1484),
    [anon_sym_not] = ACTIONS(1484),
    [anon_sym_and] = ACTIONS(1484),
    [anon_sym_or] = ACTIONS(1484),
    [anon_sym_PLUS] = ACTIONS(1484),
    [anon_sym_DASH] = ACTIONS(1484),
    [anon_sym_SLASH] = ACTIONS(1484),
    [anon_sym_SLASH_SLASH] = ACTIONS(1484),
    [anon_sym_PERCENT] = ACTIONS(1484),
    [anon_sym_PIPE] = ACTIONS(1484),
    [anon_sym_CARET] = ACTIONS(1484),
    [anon_sym_AMP] = ACTIONS(1484),
    [anon_sym_LT_LT] = ACTIONS(1484),
    [anon_sym_GT_GT] = ACTIONS(1484),
    [anon_sym_GT_GT_GT] = ACTIONS(1484),
    [anon_sym_TILDE] = ACTIONS(1484),
    [anon_sym_typeof] = ACTIONS(1484),
    [anon_sym_void] = ACTIONS(1484),
    [anon_sym_LT] = ACTIONS(1484),
    [anon_sym_LT_EQ] = ACTIONS(1484),
    [anon_sym_EQ_EQ] = ACTIONS(1484),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1484),
    [anon_sym_BANG_EQ] = ACTIONS(1484),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1484),
    [anon_sym_GT_EQ] = ACTIONS(1484),
    [anon_sym_GT] = ACTIONS(1484),
    [anon_sym_is] = ACTIONS(1484),
    [anon_sym_instanceof] = ACTIONS(1484),
    [anon_sym_new] = ACTIONS(1484),
    [anon_sym_QMARK] = ACTIONS(1484),
    [anon_sym_await] = ACTIONS(1484),
    [anon_sym_LBRACE] = ACTIONS(1484),
    [sym_this] = ACTIONS(1484),
    [sym_true] = ACTIONS(1484),
    [sym_false] = ACTIONS(1484),
    [sym_none] = ACTIONS(1484),
    [sym_string] = ACTIONS(1484),
    [sym_verbatim] = ACTIONS(1484),
    [sym_number] = ACTIONS(1484),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1482),
    [sym_regex] = ACTIONS(1482),
    [sym_fstring_start] = ACTIONS(1482),
  },
  [STATE(641)] = {
    [sym_identifier] = ACTIONS(927),
    [anon_sym_SEMI] = ACTIONS(927),
    [anon_sym_return] = ACTIONS(927),
    [anon_sym_async] = ACTIONS(927),
    [anon_sym_def] = ACTIONS(927),
    [anon_sym_COMMA] = ACTIONS(927),
    [anon_sym_LPAREN] = ACTIONS(927),
    [anon_sym_LBRACK] = ACTIONS(927),
    [anon_sym_del] = ACTIONS(927),
    [anon_sym_raise] = ACTIONS(927),
    [sym_pass_statement] = ACTIONS(927),
    [sym_break_statement] = ACTIONS(927),
    [sym_continue_statement] = ACTIONS(927),
    [sym_debugger_statement] = ACTIONS(927),
    [anon_sym_assert] = ACTIONS(927),
    [anon_sym_global] = ACTIONS(927),
    [anon_sym_nonlocal] = ACTIONS(927),
    [anon_sym_yield] = ACTIONS(927),
    [anon_sym_from] = ACTIONS(927),
    [anon_sym_import] = ACTIONS(927),
    [anon_sym_DOT] = ACTIONS(927),
    [anon_sym_if] = ACTIONS(927),
    [anon_sym_for] = ACTIONS(927),
    [anon_sym_in] = ACTIONS(927),
    [anon_sym_while] = ACTIONS(927),
    [anon_sym_do] = ACTIONS(927),
    [anon_sym_try] = ACTIONS(927),
    [anon_sym_with] = ACTIONS(927),
    [anon_sym_STAR] = ACTIONS(927),
    [anon_sym_STAR_STAR] = ACTIONS(927),
    [anon_sym_class] = ACTIONS(927),
    [anon_sym_AT] = ACTIONS(927),
    [anon_sym_not] = ACTIONS(927),
    [anon_sym_and] = ACTIONS(927),
    [anon_sym_or] = ACTIONS(927),
    [anon_sym_PLUS] = ACTIONS(927),
    [anon_sym_DASH] = ACTIONS(927),
    [anon_sym_SLASH] = ACTIONS(927),
    [anon_sym_SLASH_SLASH] = ACTIONS(927),
    [anon_sym_PERCENT] = ACTIONS(927),
    [anon_sym_PIPE] = ACTIONS(927),
    [anon_sym_CARET] = ACTIONS(927),
    [anon_sym_AMP] = ACTIONS(927),
    [anon_sym_LT_LT] = ACTIONS(927),
    [anon_sym_GT_GT] = ACTIONS(927),
    [anon_sym_GT_GT_GT] = ACTIONS(927),
    [anon_sym_TILDE] = ACTIONS(927),
    [anon_sym_typeof] = ACTIONS(927),
    [anon_sym_void] = ACTIONS(927),
    [anon_sym_LT] = ACTIONS(927),
    [anon_sym_LT_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(927),
    [anon_sym_GT_EQ] = ACTIONS(927),
    [anon_sym_GT] = ACTIONS(927),
    [anon_sym_is] = ACTIONS(927),
    [anon_sym_instanceof] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
    [anon_sym_QMARK] = ACTIONS(927),
    [anon_sym_await] = ACTIONS(927),
    [anon_sym_LBRACE] = ACTIONS(927),
    [sym_this] = ACTIONS(927),
    [sym_true] = ACTIONS(927),
    [sym_false] = ACTIONS(927),
    [sym_none] = ACTIONS(927),
    [sym_string] = ACTIONS(927),
    [sym_verbatim] = ACTIONS(927),
    [sym_number] = ACTIONS(927),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(925),
    [sym_regex] = ACTIONS(925),
    [sym_fstring_start] = ACTIONS(925),
  },
  [STATE(642)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1226),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(643)] = {
    [sym_identifier] = ACTIONS(1548),
    [anon_sym_SEMI] = ACTIONS(1548),
    [anon_sym_return] = ACTIONS(1548),
    [anon_sym_async] = ACTIONS(1548),
    [anon_sym_def] = ACTIONS(1548),
    [anon_sym_COMMA] = ACTIONS(1548),
    [anon_sym_LPAREN] = ACTIONS(1548),
    [anon_sym_LBRACK] = ACTIONS(1548),
    [anon_sym_del] = ACTIONS(1548),
    [anon_sym_raise] = ACTIONS(1548),
    [sym_pass_statement] = ACTIONS(1548),
    [sym_break_statement] = ACTIONS(1548),
    [sym_continue_statement] = ACTIONS(1548),
    [sym_debugger_statement] = ACTIONS(1548),
    [anon_sym_assert] = ACTIONS(1548),
    [anon_sym_global] = ACTIONS(1548),
    [anon_sym_nonlocal] = ACTIONS(1548),
    [anon_sym_yield] = ACTIONS(1548),
    [anon_sym_from] = ACTIONS(1548),
    [anon_sym_import] = ACTIONS(1548),
    [anon_sym_DOT] = ACTIONS(1548),
    [anon_sym_if] = ACTIONS(1548),
    [anon_sym_for] = ACTIONS(1548),
    [anon_sym_in] = ACTIONS(1548),
    [anon_sym_while] = ACTIONS(1548),
    [anon_sym_do] = ACTIONS(1548),
    [anon_sym_try] = ACTIONS(1548),
    [anon_sym_with] = ACTIONS(1548),
    [anon_sym_STAR] = ACTIONS(1548),
    [anon_sym_STAR_STAR] = ACTIONS(1548),
    [anon_sym_class] = ACTIONS(1548),
    [anon_sym_AT] = ACTIONS(1548),
    [anon_sym_not] = ACTIONS(1548),
    [anon_sym_and] = ACTIONS(1548),
    [anon_sym_or] = ACTIONS(1548),
    [anon_sym_PLUS] = ACTIONS(1548),
    [anon_sym_DASH] = ACTIONS(1548),
    [anon_sym_SLASH] = ACTIONS(1548),
    [anon_sym_SLASH_SLASH] = ACTIONS(1548),
    [anon_sym_PERCENT] = ACTIONS(1548),
    [anon_sym_PIPE] = ACTIONS(1548),
    [anon_sym_CARET] = ACTIONS(1548),
    [anon_sym_AMP] = ACTIONS(1548),
    [anon_sym_LT_LT] = ACTIONS(1548),
    [anon_sym_GT_GT] = ACTIONS(1548),
    [anon_sym_GT_GT_GT] = ACTIONS(1548),
    [anon_sym_TILDE] = ACTIONS(1548),
    [anon_sym_typeof] = ACTIONS(1548),
    [anon_sym_void] = ACTIONS(1548),
    [anon_sym_LT] = ACTIONS(1548),
    [anon_sym_LT_EQ] = ACTIONS(1548),
    [anon_sym_EQ_EQ] = ACTIONS(1548),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ] = ACTIONS(1548),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1548),
    [anon_sym_GT_EQ] = ACTIONS(1548),
    [anon_sym_GT] = ACTIONS(1548),
    [anon_sym_is] = ACTIONS(1548),
    [anon_sym_instanceof] = ACTIONS(1548),
    [anon_sym_new] = ACTIONS(1548),
    [anon_sym_QMARK] = ACTIONS(1548),
    [anon_sym_await] = ACTIONS(1548),
    [anon_sym_LBRACE] = ACTIONS(1548),
    [sym_this] = ACTIONS(1548),
    [sym_true] = ACTIONS(1548),
    [sym_false] = ACTIONS(1548),
    [sym_none] = ACTIONS(1548),
    [sym_string] = ACTIONS(1548),
    [sym_verbatim] = ACTIONS(1548),
    [sym_number] = ACTIONS(1548),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1546),
    [sym_regex] = ACTIONS(1546),
    [sym_fstring_start] = ACTIONS(1546),
  },
  [STATE(644)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(645)] = {
    [sym_identifier] = ACTIONS(1554),
    [anon_sym_SEMI] = ACTIONS(1554),
    [anon_sym_return] = ACTIONS(1554),
    [anon_sym_async] = ACTIONS(1554),
    [anon_sym_def] = ACTIONS(1554),
    [anon_sym_COMMA] = ACTIONS(1554),
    [anon_sym_LPAREN] = ACTIONS(1554),
    [anon_sym_LBRACK] = ACTIONS(1554),
    [anon_sym_del] = ACTIONS(1554),
    [anon_sym_raise] = ACTIONS(1554),
    [sym_pass_statement] = ACTIONS(1554),
    [sym_break_statement] = ACTIONS(1554),
    [sym_continue_statement] = ACTIONS(1554),
    [sym_debugger_statement] = ACTIONS(1554),
    [anon_sym_assert] = ACTIONS(1554),
    [anon_sym_global] = ACTIONS(1554),
    [anon_sym_nonlocal] = ACTIONS(1554),
    [anon_sym_yield] = ACTIONS(1554),
    [anon_sym_from] = ACTIONS(1554),
    [anon_sym_import] = ACTIONS(1554),
    [anon_sym_DOT] = ACTIONS(1554),
    [anon_sym_if] = ACTIONS(1554),
    [anon_sym_for] = ACTIONS(1554),
    [anon_sym_in] = ACTIONS(1554),
    [anon_sym_while] = ACTIONS(1554),
    [anon_sym_do] = ACTIONS(1554),
    [anon_sym_try] = ACTIONS(1554),
    [anon_sym_with] = ACTIONS(1554),
    [anon_sym_STAR] = ACTIONS(1554),
    [anon_sym_STAR_STAR] = ACTIONS(1554),
    [anon_sym_class] = ACTIONS(1554),
    [anon_sym_AT] = ACTIONS(1554),
    [anon_sym_not] = ACTIONS(1554),
    [anon_sym_and] = ACTIONS(1554),
    [anon_sym_or] = ACTIONS(1554),
    [anon_sym_PLUS] = ACTIONS(1554),
    [anon_sym_DASH] = ACTIONS(1554),
    [anon_sym_SLASH] = ACTIONS(1554),
    [anon_sym_SLASH_SLASH] = ACTIONS(1554),
    [anon_sym_PERCENT] = ACTIONS(1554),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_CARET] = ACTIONS(1554),
    [anon_sym_AMP] = ACTIONS(1554),
    [anon_sym_LT_LT] = ACTIONS(1554),
    [anon_sym_GT_GT] = ACTIONS(1554),
    [anon_sym_GT_GT_GT] = ACTIONS(1554),
    [anon_sym_TILDE] = ACTIONS(1554),
    [anon_sym_typeof] = ACTIONS(1554),
    [anon_sym_void] = ACTIONS(1554),
    [anon_sym_LT] = ACTIONS(1554),
    [anon_sym_LT_EQ] = ACTIONS(1554),
    [anon_sym_EQ_EQ] = ACTIONS(1554),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1554),
    [anon_sym_BANG_EQ] = ACTIONS(1554),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1554),
    [anon_sym_GT_EQ] = ACTIONS(1554),
    [anon_sym_GT] = ACTIONS(1554),
    [anon_sym_is] = ACTIONS(1554),
    [anon_sym_instanceof] = ACTIONS(1554),
    [anon_sym_new] = ACTIONS(1554),
    [anon_sym_QMARK] = ACTIONS(1554),
    [anon_sym_await] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [sym_this] = ACTIONS(1554),
    [sym_true] = ACTIONS(1554),
    [sym_false] = ACTIONS(1554),
    [sym_none] = ACTIONS(1554),
    [sym_string] = ACTIONS(1554),
    [sym_verbatim] = ACTIONS(1554),
    [sym_number] = ACTIONS(1554),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1552),
    [sym_regex] = ACTIONS(1552),
    [sym_fstring_start] = ACTIONS(1552),
  },
  [STATE(646)] = {
    [ts_builtin_sym_end] = ACTIONS(925),
    [sym_identifier] = ACTIONS(927),
    [anon_sym_SEMI] = ACTIONS(927),
    [anon_sym_return] = ACTIONS(927),
    [anon_sym_async] = ACTIONS(927),
    [anon_sym_def] = ACTIONS(927),
    [anon_sym_COMMA] = ACTIONS(927),
    [anon_sym_LPAREN] = ACTIONS(927),
    [anon_sym_LBRACK] = ACTIONS(927),
    [anon_sym_del] = ACTIONS(927),
    [anon_sym_raise] = ACTIONS(927),
    [sym_pass_statement] = ACTIONS(927),
    [sym_break_statement] = ACTIONS(927),
    [sym_continue_statement] = ACTIONS(927),
    [sym_debugger_statement] = ACTIONS(927),
    [anon_sym_assert] = ACTIONS(927),
    [anon_sym_global] = ACTIONS(927),
    [anon_sym_nonlocal] = ACTIONS(927),
    [anon_sym_yield] = ACTIONS(927),
    [anon_sym_from] = ACTIONS(927),
    [anon_sym_import] = ACTIONS(927),
    [anon_sym_DOT] = ACTIONS(927),
    [anon_sym_if] = ACTIONS(927),
    [anon_sym_for] = ACTIONS(927),
    [anon_sym_in] = ACTIONS(927),
    [anon_sym_while] = ACTIONS(927),
    [anon_sym_do] = ACTIONS(927),
    [anon_sym_try] = ACTIONS(927),
    [anon_sym_with] = ACTIONS(927),
    [anon_sym_STAR] = ACTIONS(927),
    [anon_sym_STAR_STAR] = ACTIONS(927),
    [anon_sym_class] = ACTIONS(927),
    [anon_sym_AT] = ACTIONS(927),
    [anon_sym_not] = ACTIONS(927),
    [anon_sym_and] = ACTIONS(927),
    [anon_sym_or] = ACTIONS(927),
    [anon_sym_PLUS] = ACTIONS(927),
    [anon_sym_DASH] = ACTIONS(927),
    [anon_sym_SLASH] = ACTIONS(927),
    [anon_sym_SLASH_SLASH] = ACTIONS(927),
    [anon_sym_PERCENT] = ACTIONS(927),
    [anon_sym_PIPE] = ACTIONS(927),
    [anon_sym_CARET] = ACTIONS(927),
    [anon_sym_AMP] = ACTIONS(927),
    [anon_sym_LT_LT] = ACTIONS(927),
    [anon_sym_GT_GT] = ACTIONS(927),
    [anon_sym_GT_GT_GT] = ACTIONS(927),
    [anon_sym_TILDE] = ACTIONS(927),
    [anon_sym_typeof] = ACTIONS(927),
    [anon_sym_void] = ACTIONS(927),
    [anon_sym_LT] = ACTIONS(927),
    [anon_sym_LT_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ] = ACTIONS(927),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ] = ACTIONS(927),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(927),
    [anon_sym_GT_EQ] = ACTIONS(927),
    [anon_sym_GT] = ACTIONS(927),
    [anon_sym_is] = ACTIONS(927),
    [anon_sym_instanceof] = ACTIONS(927),
    [anon_sym_new] = ACTIONS(927),
    [anon_sym_QMARK] = ACTIONS(927),
    [anon_sym_await] = ACTIONS(927),
    [anon_sym_LBRACE] = ACTIONS(927),
    [sym_this] = ACTIONS(927),
    [sym_true] = ACTIONS(927),
    [sym_false] = ACTIONS(927),
    [sym_none] = ACTIONS(927),
    [sym_string] = ACTIONS(927),
    [sym_verbatim] = ACTIONS(927),
    [sym_number] = ACTIONS(927),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(925),
    [sym_fstring_start] = ACTIONS(925),
  },
  [STATE(647)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(647),
    [sym_identifier] = ACTIONS(1247),
    [anon_sym_SEMI] = ACTIONS(1247),
    [anon_sym_return] = ACTIONS(1247),
    [anon_sym_async] = ACTIONS(1247),
    [anon_sym_def] = ACTIONS(1247),
    [anon_sym_LPAREN] = ACTIONS(1247),
    [anon_sym_LBRACK] = ACTIONS(1247),
    [anon_sym_del] = ACTIONS(1247),
    [anon_sym_raise] = ACTIONS(1247),
    [sym_pass_statement] = ACTIONS(1247),
    [sym_break_statement] = ACTIONS(1247),
    [sym_continue_statement] = ACTIONS(1247),
    [sym_debugger_statement] = ACTIONS(1247),
    [anon_sym_assert] = ACTIONS(1247),
    [anon_sym_global] = ACTIONS(1247),
    [anon_sym_nonlocal] = ACTIONS(1247),
    [anon_sym_yield] = ACTIONS(1247),
    [anon_sym_from] = ACTIONS(1247),
    [anon_sym_import] = ACTIONS(1247),
    [anon_sym_DOT] = ACTIONS(1247),
    [anon_sym_if] = ACTIONS(1247),
    [anon_sym_for] = ACTIONS(1247),
    [anon_sym_in] = ACTIONS(1581),
    [anon_sym_while] = ACTIONS(1247),
    [anon_sym_do] = ACTIONS(1247),
    [anon_sym_try] = ACTIONS(1247),
    [anon_sym_with] = ACTIONS(1247),
    [anon_sym_STAR] = ACTIONS(1247),
    [anon_sym_STAR_STAR] = ACTIONS(1247),
    [anon_sym_class] = ACTIONS(1247),
    [anon_sym_AT] = ACTIONS(1247),
    [anon_sym_not] = ACTIONS(1584),
    [anon_sym_and] = ACTIONS(1247),
    [anon_sym_or] = ACTIONS(1247),
    [anon_sym_PLUS] = ACTIONS(1247),
    [anon_sym_DASH] = ACTIONS(1247),
    [anon_sym_SLASH] = ACTIONS(1247),
    [anon_sym_SLASH_SLASH] = ACTIONS(1247),
    [anon_sym_PERCENT] = ACTIONS(1247),
    [anon_sym_PIPE] = ACTIONS(1247),
    [anon_sym_CARET] = ACTIONS(1247),
    [anon_sym_AMP] = ACTIONS(1247),
    [anon_sym_LT_LT] = ACTIONS(1247),
    [anon_sym_GT_GT] = ACTIONS(1247),
    [anon_sym_GT_GT_GT] = ACTIONS(1247),
    [anon_sym_TILDE] = ACTIONS(1247),
    [anon_sym_typeof] = ACTIONS(1247),
    [anon_sym_void] = ACTIONS(1247),
    [anon_sym_LT] = ACTIONS(1581),
    [anon_sym_LT_EQ] = ACTIONS(1581),
    [anon_sym_EQ_EQ] = ACTIONS(1581),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1581),
    [anon_sym_BANG_EQ] = ACTIONS(1581),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1581),
    [anon_sym_GT_EQ] = ACTIONS(1581),
    [anon_sym_GT] = ACTIONS(1581),
    [anon_sym_is] = ACTIONS(1587),
    [anon_sym_instanceof] = ACTIONS(1581),
    [anon_sym_new] = ACTIONS(1247),
    [anon_sym_QMARK] = ACTIONS(1247),
    [anon_sym_await] = ACTIONS(1247),
    [anon_sym_LBRACE] = ACTIONS(1247),
    [sym_this] = ACTIONS(1247),
    [sym_true] = ACTIONS(1247),
    [sym_false] = ACTIONS(1247),
    [sym_none] = ACTIONS(1247),
    [sym_string] = ACTIONS(1247),
    [sym_verbatim] = ACTIONS(1247),
    [sym_number] = ACTIONS(1247),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1245),
    [sym_regex] = ACTIONS(1245),
    [sym_fstring_start] = ACTIONS(1245),
  },
  [STATE(648)] = {
    [ts_builtin_sym_end] = ACTIONS(1590),
    [sym_identifier] = ACTIONS(1592),
    [anon_sym_SEMI] = ACTIONS(1592),
    [anon_sym_return] = ACTIONS(1592),
    [anon_sym_async] = ACTIONS(1592),
    [anon_sym_def] = ACTIONS(1592),
    [anon_sym_COMMA] = ACTIONS(1592),
    [anon_sym_LPAREN] = ACTIONS(1592),
    [anon_sym_LBRACK] = ACTIONS(1592),
    [anon_sym_del] = ACTIONS(1592),
    [anon_sym_raise] = ACTIONS(1592),
    [sym_pass_statement] = ACTIONS(1592),
    [sym_break_statement] = ACTIONS(1592),
    [sym_continue_statement] = ACTIONS(1592),
    [sym_debugger_statement] = ACTIONS(1592),
    [anon_sym_assert] = ACTIONS(1592),
    [anon_sym_global] = ACTIONS(1592),
    [anon_sym_nonlocal] = ACTIONS(1592),
    [anon_sym_yield] = ACTIONS(1592),
    [anon_sym_from] = ACTIONS(1592),
    [anon_sym_import] = ACTIONS(1592),
    [anon_sym_DOT] = ACTIONS(1592),
    [anon_sym_if] = ACTIONS(1592),
    [anon_sym_for] = ACTIONS(1592),
    [anon_sym_in] = ACTIONS(1592),
    [anon_sym_while] = ACTIONS(1592),
    [anon_sym_do] = ACTIONS(1592),
    [anon_sym_try] = ACTIONS(1592),
    [anon_sym_with] = ACTIONS(1592),
    [anon_sym_STAR] = ACTIONS(1592),
    [anon_sym_STAR_STAR] = ACTIONS(1592),
    [anon_sym_class] = ACTIONS(1592),
    [anon_sym_AT] = ACTIONS(1592),
    [anon_sym_not] = ACTIONS(1592),
    [anon_sym_and] = ACTIONS(1592),
    [anon_sym_or] = ACTIONS(1592),
    [anon_sym_PLUS] = ACTIONS(1592),
    [anon_sym_DASH] = ACTIONS(1592),
    [anon_sym_SLASH] = ACTIONS(1592),
    [anon_sym_SLASH_SLASH] = ACTIONS(1592),
    [anon_sym_PERCENT] = ACTIONS(1592),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_CARET] = ACTIONS(1592),
    [anon_sym_AMP] = ACTIONS(1592),
    [anon_sym_LT_LT] = ACTIONS(1592),
    [anon_sym_GT_GT] = ACTIONS(1592),
    [anon_sym_GT_GT_GT] = ACTIONS(1592),
    [anon_sym_TILDE] = ACTIONS(1592),
    [anon_sym_typeof] = ACTIONS(1592),
    [anon_sym_void] = ACTIONS(1592),
    [anon_sym_LT] = ACTIONS(1592),
    [anon_sym_LT_EQ] = ACTIONS(1592),
    [anon_sym_EQ_EQ] = ACTIONS(1592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1592),
    [anon_sym_BANG_EQ] = ACTIONS(1592),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1592),
    [anon_sym_GT_EQ] = ACTIONS(1592),
    [anon_sym_GT] = ACTIONS(1592),
    [anon_sym_is] = ACTIONS(1592),
    [anon_sym_instanceof] = ACTIONS(1592),
    [anon_sym_new] = ACTIONS(1592),
    [anon_sym_QMARK] = ACTIONS(1592),
    [anon_sym_await] = ACTIONS(1592),
    [anon_sym_LBRACE] = ACTIONS(1592),
    [sym_this] = ACTIONS(1592),
    [sym_true] = ACTIONS(1592),
    [sym_false] = ACTIONS(1592),
    [sym_none] = ACTIONS(1592),
    [sym_string] = ACTIONS(1592),
    [sym_verbatim] = ACTIONS(1592),
    [sym_number] = ACTIONS(1592),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1590),
    [sym_fstring_start] = ACTIONS(1590),
  },
  [STATE(649)] = {
    [sym_identifier] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym_return] = ACTIONS(1200),
    [anon_sym_async] = ACTIONS(1200),
    [anon_sym_def] = ACTIONS(1200),
    [anon_sym_COMMA] = ACTIONS(1200),
    [anon_sym_LPAREN] = ACTIONS(1200),
    [anon_sym_LBRACK] = ACTIONS(1200),
    [anon_sym_del] = ACTIONS(1200),
    [anon_sym_raise] = ACTIONS(1200),
    [sym_pass_statement] = ACTIONS(1200),
    [sym_break_statement] = ACTIONS(1200),
    [sym_continue_statement] = ACTIONS(1200),
    [sym_debugger_statement] = ACTIONS(1200),
    [anon_sym_assert] = ACTIONS(1200),
    [anon_sym_global] = ACTIONS(1200),
    [anon_sym_nonlocal] = ACTIONS(1200),
    [anon_sym_yield] = ACTIONS(1200),
    [anon_sym_from] = ACTIONS(1200),
    [anon_sym_import] = ACTIONS(1200),
    [anon_sym_DOT] = ACTIONS(1200),
    [anon_sym_if] = ACTIONS(1200),
    [anon_sym_for] = ACTIONS(1200),
    [anon_sym_in] = ACTIONS(1200),
    [anon_sym_while] = ACTIONS(1200),
    [anon_sym_do] = ACTIONS(1200),
    [anon_sym_try] = ACTIONS(1200),
    [anon_sym_with] = ACTIONS(1200),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_STAR_STAR] = ACTIONS(1200),
    [anon_sym_class] = ACTIONS(1200),
    [anon_sym_AT] = ACTIONS(1200),
    [anon_sym_not] = ACTIONS(1200),
    [anon_sym_and] = ACTIONS(1200),
    [anon_sym_or] = ACTIONS(1200),
    [anon_sym_PLUS] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [anon_sym_SLASH] = ACTIONS(1200),
    [anon_sym_SLASH_SLASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(1200),
    [anon_sym_PIPE] = ACTIONS(1200),
    [anon_sym_CARET] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_LT_LT] = ACTIONS(1200),
    [anon_sym_GT_GT] = ACTIONS(1200),
    [anon_sym_GT_GT_GT] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_typeof] = ACTIONS(1200),
    [anon_sym_void] = ACTIONS(1200),
    [anon_sym_LT] = ACTIONS(1200),
    [anon_sym_LT_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1200),
    [anon_sym_GT_EQ] = ACTIONS(1200),
    [anon_sym_GT] = ACTIONS(1200),
    [anon_sym_is] = ACTIONS(1200),
    [anon_sym_instanceof] = ACTIONS(1200),
    [anon_sym_new] = ACTIONS(1200),
    [anon_sym_QMARK] = ACTIONS(1200),
    [anon_sym_await] = ACTIONS(1200),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [sym_this] = ACTIONS(1200),
    [sym_true] = ACTIONS(1200),
    [sym_false] = ACTIONS(1200),
    [sym_none] = ACTIONS(1200),
    [sym_string] = ACTIONS(1200),
    [sym_verbatim] = ACTIONS(1200),
    [sym_number] = ACTIONS(1200),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1198),
    [sym_regex] = ACTIONS(1198),
    [sym_fstring_start] = ACTIONS(1198),
  },
  [STATE(650)] = {
    [sym_identifier] = ACTIONS(1558),
    [anon_sym_SEMI] = ACTIONS(1558),
    [anon_sym_return] = ACTIONS(1558),
    [anon_sym_async] = ACTIONS(1558),
    [anon_sym_def] = ACTIONS(1558),
    [anon_sym_COMMA] = ACTIONS(1558),
    [anon_sym_LPAREN] = ACTIONS(1558),
    [anon_sym_LBRACK] = ACTIONS(1558),
    [anon_sym_del] = ACTIONS(1558),
    [anon_sym_raise] = ACTIONS(1558),
    [sym_pass_statement] = ACTIONS(1558),
    [sym_break_statement] = ACTIONS(1558),
    [sym_continue_statement] = ACTIONS(1558),
    [sym_debugger_statement] = ACTIONS(1558),
    [anon_sym_assert] = ACTIONS(1558),
    [anon_sym_global] = ACTIONS(1558),
    [anon_sym_nonlocal] = ACTIONS(1558),
    [anon_sym_yield] = ACTIONS(1558),
    [anon_sym_from] = ACTIONS(1558),
    [anon_sym_import] = ACTIONS(1558),
    [anon_sym_DOT] = ACTIONS(1558),
    [anon_sym_if] = ACTIONS(1558),
    [anon_sym_for] = ACTIONS(1558),
    [anon_sym_in] = ACTIONS(1558),
    [anon_sym_while] = ACTIONS(1558),
    [anon_sym_do] = ACTIONS(1558),
    [anon_sym_try] = ACTIONS(1558),
    [anon_sym_with] = ACTIONS(1558),
    [anon_sym_STAR] = ACTIONS(1558),
    [anon_sym_STAR_STAR] = ACTIONS(1558),
    [anon_sym_class] = ACTIONS(1558),
    [anon_sym_AT] = ACTIONS(1558),
    [anon_sym_not] = ACTIONS(1558),
    [anon_sym_and] = ACTIONS(1558),
    [anon_sym_or] = ACTIONS(1558),
    [anon_sym_PLUS] = ACTIONS(1558),
    [anon_sym_DASH] = ACTIONS(1558),
    [anon_sym_SLASH] = ACTIONS(1558),
    [anon_sym_SLASH_SLASH] = ACTIONS(1558),
    [anon_sym_PERCENT] = ACTIONS(1558),
    [anon_sym_PIPE] = ACTIONS(1558),
    [anon_sym_CARET] = ACTIONS(1558),
    [anon_sym_AMP] = ACTIONS(1558),
    [anon_sym_LT_LT] = ACTIONS(1558),
    [anon_sym_GT_GT] = ACTIONS(1558),
    [anon_sym_GT_GT_GT] = ACTIONS(1558),
    [anon_sym_TILDE] = ACTIONS(1558),
    [anon_sym_typeof] = ACTIONS(1558),
    [anon_sym_void] = ACTIONS(1558),
    [anon_sym_LT] = ACTIONS(1558),
    [anon_sym_LT_EQ] = ACTIONS(1558),
    [anon_sym_EQ_EQ] = ACTIONS(1558),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1558),
    [anon_sym_BANG_EQ] = ACTIONS(1558),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1558),
    [anon_sym_GT_EQ] = ACTIONS(1558),
    [anon_sym_GT] = ACTIONS(1558),
    [anon_sym_is] = ACTIONS(1558),
    [anon_sym_instanceof] = ACTIONS(1558),
    [anon_sym_new] = ACTIONS(1558),
    [anon_sym_QMARK] = ACTIONS(1558),
    [anon_sym_await] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [sym_this] = ACTIONS(1558),
    [sym_true] = ACTIONS(1558),
    [sym_false] = ACTIONS(1558),
    [sym_none] = ACTIONS(1558),
    [sym_string] = ACTIONS(1558),
    [sym_verbatim] = ACTIONS(1558),
    [sym_number] = ACTIONS(1558),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1556),
    [sym_regex] = ACTIONS(1556),
    [sym_fstring_start] = ACTIONS(1556),
  },
  [STATE(651)] = {
    [sym_identifier] = ACTIONS(1594),
    [anon_sym_SEMI] = ACTIONS(1594),
    [anon_sym_return] = ACTIONS(1594),
    [anon_sym_async] = ACTIONS(1594),
    [anon_sym_def] = ACTIONS(1594),
    [anon_sym_COMMA] = ACTIONS(1594),
    [anon_sym_LPAREN] = ACTIONS(1594),
    [anon_sym_LBRACK] = ACTIONS(1594),
    [anon_sym_del] = ACTIONS(1594),
    [anon_sym_raise] = ACTIONS(1594),
    [sym_pass_statement] = ACTIONS(1594),
    [sym_break_statement] = ACTIONS(1594),
    [sym_continue_statement] = ACTIONS(1594),
    [sym_debugger_statement] = ACTIONS(1594),
    [anon_sym_assert] = ACTIONS(1594),
    [anon_sym_global] = ACTIONS(1594),
    [anon_sym_nonlocal] = ACTIONS(1594),
    [anon_sym_yield] = ACTIONS(1594),
    [anon_sym_from] = ACTIONS(1594),
    [anon_sym_import] = ACTIONS(1594),
    [anon_sym_DOT] = ACTIONS(1594),
    [anon_sym_if] = ACTIONS(1594),
    [anon_sym_for] = ACTIONS(1594),
    [anon_sym_in] = ACTIONS(1594),
    [anon_sym_while] = ACTIONS(1594),
    [anon_sym_do] = ACTIONS(1594),
    [anon_sym_try] = ACTIONS(1594),
    [anon_sym_with] = ACTIONS(1594),
    [anon_sym_STAR] = ACTIONS(1594),
    [anon_sym_STAR_STAR] = ACTIONS(1594),
    [anon_sym_class] = ACTIONS(1594),
    [anon_sym_AT] = ACTIONS(1594),
    [anon_sym_not] = ACTIONS(1594),
    [anon_sym_and] = ACTIONS(1594),
    [anon_sym_or] = ACTIONS(1594),
    [anon_sym_PLUS] = ACTIONS(1594),
    [anon_sym_DASH] = ACTIONS(1594),
    [anon_sym_SLASH] = ACTIONS(1594),
    [anon_sym_SLASH_SLASH] = ACTIONS(1594),
    [anon_sym_PERCENT] = ACTIONS(1594),
    [anon_sym_PIPE] = ACTIONS(1594),
    [anon_sym_CARET] = ACTIONS(1594),
    [anon_sym_AMP] = ACTIONS(1594),
    [anon_sym_LT_LT] = ACTIONS(1594),
    [anon_sym_GT_GT] = ACTIONS(1594),
    [anon_sym_GT_GT_GT] = ACTIONS(1594),
    [anon_sym_TILDE] = ACTIONS(1594),
    [anon_sym_typeof] = ACTIONS(1594),
    [anon_sym_void] = ACTIONS(1594),
    [anon_sym_LT] = ACTIONS(1594),
    [anon_sym_LT_EQ] = ACTIONS(1594),
    [anon_sym_EQ_EQ] = ACTIONS(1594),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1594),
    [anon_sym_BANG_EQ] = ACTIONS(1594),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1594),
    [anon_sym_GT_EQ] = ACTIONS(1594),
    [anon_sym_GT] = ACTIONS(1594),
    [anon_sym_is] = ACTIONS(1594),
    [anon_sym_instanceof] = ACTIONS(1594),
    [anon_sym_new] = ACTIONS(1594),
    [anon_sym_QMARK] = ACTIONS(1594),
    [anon_sym_await] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(1594),
    [sym_this] = ACTIONS(1594),
    [sym_true] = ACTIONS(1594),
    [sym_false] = ACTIONS(1594),
    [sym_none] = ACTIONS(1594),
    [sym_string] = ACTIONS(1594),
    [sym_verbatim] = ACTIONS(1594),
    [sym_number] = ACTIONS(1594),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1596),
    [sym_regex] = ACTIONS(1596),
    [sym_fstring_start] = ACTIONS(1596),
  },
  [STATE(652)] = {
    [sym_argument_list] = STATE(678),
    [sym_identifier] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym_return] = ACTIONS(1200),
    [anon_sym_async] = ACTIONS(1200),
    [anon_sym_def] = ACTIONS(1200),
    [anon_sym_LPAREN] = ACTIONS(1214),
    [anon_sym_LBRACK] = ACTIONS(1216),
    [anon_sym_del] = ACTIONS(1200),
    [anon_sym_raise] = ACTIONS(1200),
    [sym_pass_statement] = ACTIONS(1200),
    [sym_break_statement] = ACTIONS(1200),
    [sym_continue_statement] = ACTIONS(1200),
    [sym_debugger_statement] = ACTIONS(1200),
    [anon_sym_assert] = ACTIONS(1200),
    [anon_sym_global] = ACTIONS(1200),
    [anon_sym_nonlocal] = ACTIONS(1200),
    [anon_sym_yield] = ACTIONS(1200),
    [anon_sym_from] = ACTIONS(1200),
    [anon_sym_import] = ACTIONS(1200),
    [anon_sym_DOT] = ACTIONS(1218),
    [anon_sym_if] = ACTIONS(1200),
    [anon_sym_for] = ACTIONS(1200),
    [anon_sym_in] = ACTIONS(1200),
    [anon_sym_while] = ACTIONS(1200),
    [anon_sym_do] = ACTIONS(1200),
    [anon_sym_try] = ACTIONS(1200),
    [anon_sym_with] = ACTIONS(1200),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_STAR_STAR] = ACTIONS(1200),
    [anon_sym_class] = ACTIONS(1200),
    [anon_sym_AT] = ACTIONS(1200),
    [anon_sym_not] = ACTIONS(1200),
    [anon_sym_and] = ACTIONS(1200),
    [anon_sym_or] = ACTIONS(1200),
    [anon_sym_PLUS] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [anon_sym_SLASH] = ACTIONS(1200),
    [anon_sym_SLASH_SLASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(1200),
    [anon_sym_PIPE] = ACTIONS(1200),
    [anon_sym_CARET] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_LT_LT] = ACTIONS(1200),
    [anon_sym_GT_GT] = ACTIONS(1200),
    [anon_sym_GT_GT_GT] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_typeof] = ACTIONS(1200),
    [anon_sym_void] = ACTIONS(1200),
    [anon_sym_LT] = ACTIONS(1200),
    [anon_sym_LT_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1200),
    [anon_sym_GT_EQ] = ACTIONS(1200),
    [anon_sym_GT] = ACTIONS(1200),
    [anon_sym_is] = ACTIONS(1200),
    [anon_sym_instanceof] = ACTIONS(1200),
    [anon_sym_new] = ACTIONS(1200),
    [anon_sym_QMARK] = ACTIONS(1598),
    [anon_sym_await] = ACTIONS(1200),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [sym_this] = ACTIONS(1200),
    [sym_true] = ACTIONS(1200),
    [sym_false] = ACTIONS(1200),
    [sym_none] = ACTIONS(1200),
    [sym_string] = ACTIONS(1200),
    [sym_verbatim] = ACTIONS(1200),
    [sym_number] = ACTIONS(1200),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1198),
    [sym_regex] = ACTIONS(1198),
    [sym_fstring_start] = ACTIONS(1198),
  },
  [STATE(653)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1260),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_return] = ACTIONS(1260),
    [anon_sym_async] = ACTIONS(1260),
    [anon_sym_def] = ACTIONS(1260),
    [anon_sym_LPAREN] = ACTIONS(1260),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [anon_sym_del] = ACTIONS(1260),
    [anon_sym_raise] = ACTIONS(1260),
    [sym_pass_statement] = ACTIONS(1260),
    [sym_break_statement] = ACTIONS(1260),
    [sym_continue_statement] = ACTIONS(1260),
    [sym_debugger_statement] = ACTIONS(1260),
    [anon_sym_assert] = ACTIONS(1260),
    [anon_sym_global] = ACTIONS(1260),
    [anon_sym_nonlocal] = ACTIONS(1260),
    [anon_sym_yield] = ACTIONS(1260),
    [anon_sym_from] = ACTIONS(1260),
    [anon_sym_import] = ACTIONS(1260),
    [anon_sym_DOT] = ACTIONS(1260),
    [anon_sym_if] = ACTIONS(1600),
    [anon_sym_for] = ACTIONS(1260),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1260),
    [anon_sym_do] = ACTIONS(1260),
    [anon_sym_try] = ACTIONS(1260),
    [anon_sym_with] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1260),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1425),
    [anon_sym_or] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1260),
    [anon_sym_typeof] = ACTIONS(1260),
    [anon_sym_void] = ACTIONS(1260),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1260),
    [anon_sym_QMARK] = ACTIONS(1260),
    [anon_sym_await] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [sym_this] = ACTIONS(1260),
    [sym_true] = ACTIONS(1260),
    [sym_false] = ACTIONS(1260),
    [sym_none] = ACTIONS(1260),
    [sym_string] = ACTIONS(1260),
    [sym_verbatim] = ACTIONS(1260),
    [sym_number] = ACTIONS(1260),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1258),
    [sym_regex] = ACTIONS(1258),
    [sym_fstring_start] = ACTIONS(1258),
  },
  [STATE(654)] = {
    [sym_identifier] = ACTIONS(1562),
    [anon_sym_SEMI] = ACTIONS(1562),
    [anon_sym_return] = ACTIONS(1562),
    [anon_sym_async] = ACTIONS(1562),
    [anon_sym_def] = ACTIONS(1562),
    [anon_sym_COMMA] = ACTIONS(1562),
    [anon_sym_LPAREN] = ACTIONS(1562),
    [anon_sym_LBRACK] = ACTIONS(1562),
    [anon_sym_del] = ACTIONS(1562),
    [anon_sym_raise] = ACTIONS(1562),
    [sym_pass_statement] = ACTIONS(1562),
    [sym_break_statement] = ACTIONS(1562),
    [sym_continue_statement] = ACTIONS(1562),
    [sym_debugger_statement] = ACTIONS(1562),
    [anon_sym_assert] = ACTIONS(1562),
    [anon_sym_global] = ACTIONS(1562),
    [anon_sym_nonlocal] = ACTIONS(1562),
    [anon_sym_yield] = ACTIONS(1562),
    [anon_sym_from] = ACTIONS(1562),
    [anon_sym_import] = ACTIONS(1562),
    [anon_sym_DOT] = ACTIONS(1562),
    [anon_sym_if] = ACTIONS(1562),
    [anon_sym_for] = ACTIONS(1562),
    [anon_sym_in] = ACTIONS(1562),
    [anon_sym_while] = ACTIONS(1562),
    [anon_sym_do] = ACTIONS(1562),
    [anon_sym_try] = ACTIONS(1562),
    [anon_sym_with] = ACTIONS(1562),
    [anon_sym_STAR] = ACTIONS(1562),
    [anon_sym_STAR_STAR] = ACTIONS(1562),
    [anon_sym_class] = ACTIONS(1562),
    [anon_sym_AT] = ACTIONS(1562),
    [anon_sym_not] = ACTIONS(1562),
    [anon_sym_and] = ACTIONS(1562),
    [anon_sym_or] = ACTIONS(1562),
    [anon_sym_PLUS] = ACTIONS(1562),
    [anon_sym_DASH] = ACTIONS(1562),
    [anon_sym_SLASH] = ACTIONS(1562),
    [anon_sym_SLASH_SLASH] = ACTIONS(1562),
    [anon_sym_PERCENT] = ACTIONS(1562),
    [anon_sym_PIPE] = ACTIONS(1562),
    [anon_sym_CARET] = ACTIONS(1562),
    [anon_sym_AMP] = ACTIONS(1562),
    [anon_sym_LT_LT] = ACTIONS(1562),
    [anon_sym_GT_GT] = ACTIONS(1562),
    [anon_sym_GT_GT_GT] = ACTIONS(1562),
    [anon_sym_TILDE] = ACTIONS(1562),
    [anon_sym_typeof] = ACTIONS(1562),
    [anon_sym_void] = ACTIONS(1562),
    [anon_sym_LT] = ACTIONS(1562),
    [anon_sym_LT_EQ] = ACTIONS(1562),
    [anon_sym_EQ_EQ] = ACTIONS(1562),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1562),
    [anon_sym_BANG_EQ] = ACTIONS(1562),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1562),
    [anon_sym_GT_EQ] = ACTIONS(1562),
    [anon_sym_GT] = ACTIONS(1562),
    [anon_sym_is] = ACTIONS(1562),
    [anon_sym_instanceof] = ACTIONS(1562),
    [anon_sym_new] = ACTIONS(1562),
    [anon_sym_QMARK] = ACTIONS(1562),
    [anon_sym_await] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [sym_this] = ACTIONS(1562),
    [sym_true] = ACTIONS(1562),
    [sym_false] = ACTIONS(1562),
    [sym_none] = ACTIONS(1562),
    [sym_string] = ACTIONS(1562),
    [sym_verbatim] = ACTIONS(1562),
    [sym_number] = ACTIONS(1562),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1560),
    [sym_regex] = ACTIONS(1560),
    [sym_fstring_start] = ACTIONS(1560),
  },
  [STATE(655)] = {
    [sym_identifier] = ACTIONS(1602),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_return] = ACTIONS(1602),
    [anon_sym_async] = ACTIONS(1602),
    [anon_sym_def] = ACTIONS(1602),
    [anon_sym_COMMA] = ACTIONS(1602),
    [anon_sym_LPAREN] = ACTIONS(1602),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [anon_sym_del] = ACTIONS(1602),
    [anon_sym_raise] = ACTIONS(1602),
    [sym_pass_statement] = ACTIONS(1602),
    [sym_break_statement] = ACTIONS(1602),
    [sym_continue_statement] = ACTIONS(1602),
    [sym_debugger_statement] = ACTIONS(1602),
    [anon_sym_assert] = ACTIONS(1602),
    [anon_sym_global] = ACTIONS(1602),
    [anon_sym_nonlocal] = ACTIONS(1602),
    [anon_sym_yield] = ACTIONS(1602),
    [anon_sym_from] = ACTIONS(1602),
    [anon_sym_import] = ACTIONS(1602),
    [anon_sym_DOT] = ACTIONS(1602),
    [anon_sym_if] = ACTIONS(1602),
    [anon_sym_for] = ACTIONS(1602),
    [anon_sym_in] = ACTIONS(1602),
    [anon_sym_while] = ACTIONS(1602),
    [anon_sym_do] = ACTIONS(1602),
    [anon_sym_try] = ACTIONS(1602),
    [anon_sym_with] = ACTIONS(1602),
    [anon_sym_STAR] = ACTIONS(1602),
    [anon_sym_STAR_STAR] = ACTIONS(1602),
    [anon_sym_class] = ACTIONS(1602),
    [anon_sym_AT] = ACTIONS(1602),
    [anon_sym_not] = ACTIONS(1602),
    [anon_sym_and] = ACTIONS(1602),
    [anon_sym_or] = ACTIONS(1602),
    [anon_sym_PLUS] = ACTIONS(1602),
    [anon_sym_DASH] = ACTIONS(1602),
    [anon_sym_SLASH] = ACTIONS(1602),
    [anon_sym_SLASH_SLASH] = ACTIONS(1602),
    [anon_sym_PERCENT] = ACTIONS(1602),
    [anon_sym_PIPE] = ACTIONS(1602),
    [anon_sym_CARET] = ACTIONS(1602),
    [anon_sym_AMP] = ACTIONS(1602),
    [anon_sym_LT_LT] = ACTIONS(1602),
    [anon_sym_GT_GT] = ACTIONS(1602),
    [anon_sym_GT_GT_GT] = ACTIONS(1602),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_typeof] = ACTIONS(1602),
    [anon_sym_void] = ACTIONS(1602),
    [anon_sym_LT] = ACTIONS(1602),
    [anon_sym_LT_EQ] = ACTIONS(1602),
    [anon_sym_EQ_EQ] = ACTIONS(1602),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1602),
    [anon_sym_BANG_EQ] = ACTIONS(1602),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1602),
    [anon_sym_GT_EQ] = ACTIONS(1602),
    [anon_sym_GT] = ACTIONS(1602),
    [anon_sym_is] = ACTIONS(1602),
    [anon_sym_instanceof] = ACTIONS(1602),
    [anon_sym_new] = ACTIONS(1602),
    [anon_sym_QMARK] = ACTIONS(1602),
    [anon_sym_await] = ACTIONS(1602),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [sym_this] = ACTIONS(1602),
    [sym_true] = ACTIONS(1602),
    [sym_false] = ACTIONS(1602),
    [sym_none] = ACTIONS(1602),
    [sym_string] = ACTIONS(1602),
    [sym_verbatim] = ACTIONS(1602),
    [sym_number] = ACTIONS(1602),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1604),
    [sym_regex] = ACTIONS(1604),
    [sym_fstring_start] = ACTIONS(1604),
  },
  [STATE(656)] = {
    [sym_identifier] = ACTIONS(1371),
    [anon_sym_SEMI] = ACTIONS(1371),
    [anon_sym_return] = ACTIONS(1371),
    [anon_sym_async] = ACTIONS(1371),
    [anon_sym_def] = ACTIONS(1371),
    [anon_sym_COMMA] = ACTIONS(1371),
    [anon_sym_LPAREN] = ACTIONS(1371),
    [anon_sym_LBRACK] = ACTIONS(1371),
    [anon_sym_del] = ACTIONS(1371),
    [anon_sym_raise] = ACTIONS(1371),
    [sym_pass_statement] = ACTIONS(1371),
    [sym_break_statement] = ACTIONS(1371),
    [sym_continue_statement] = ACTIONS(1371),
    [sym_debugger_statement] = ACTIONS(1371),
    [anon_sym_assert] = ACTIONS(1371),
    [anon_sym_global] = ACTIONS(1371),
    [anon_sym_nonlocal] = ACTIONS(1371),
    [anon_sym_yield] = ACTIONS(1371),
    [anon_sym_from] = ACTIONS(1371),
    [anon_sym_import] = ACTIONS(1371),
    [anon_sym_DOT] = ACTIONS(1371),
    [anon_sym_if] = ACTIONS(1371),
    [anon_sym_for] = ACTIONS(1371),
    [anon_sym_in] = ACTIONS(1371),
    [anon_sym_while] = ACTIONS(1371),
    [anon_sym_do] = ACTIONS(1371),
    [anon_sym_try] = ACTIONS(1371),
    [anon_sym_with] = ACTIONS(1371),
    [anon_sym_STAR] = ACTIONS(1371),
    [anon_sym_STAR_STAR] = ACTIONS(1371),
    [anon_sym_class] = ACTIONS(1371),
    [anon_sym_AT] = ACTIONS(1371),
    [anon_sym_not] = ACTIONS(1371),
    [anon_sym_and] = ACTIONS(1371),
    [anon_sym_or] = ACTIONS(1371),
    [anon_sym_PLUS] = ACTIONS(1371),
    [anon_sym_DASH] = ACTIONS(1371),
    [anon_sym_SLASH] = ACTIONS(1371),
    [anon_sym_SLASH_SLASH] = ACTIONS(1371),
    [anon_sym_PERCENT] = ACTIONS(1371),
    [anon_sym_PIPE] = ACTIONS(1371),
    [anon_sym_CARET] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(1371),
    [anon_sym_LT_LT] = ACTIONS(1371),
    [anon_sym_GT_GT] = ACTIONS(1371),
    [anon_sym_GT_GT_GT] = ACTIONS(1371),
    [anon_sym_TILDE] = ACTIONS(1371),
    [anon_sym_typeof] = ACTIONS(1371),
    [anon_sym_void] = ACTIONS(1371),
    [anon_sym_LT] = ACTIONS(1371),
    [anon_sym_LT_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1371),
    [anon_sym_GT_EQ] = ACTIONS(1371),
    [anon_sym_GT] = ACTIONS(1371),
    [anon_sym_is] = ACTIONS(1371),
    [anon_sym_instanceof] = ACTIONS(1371),
    [anon_sym_new] = ACTIONS(1371),
    [anon_sym_QMARK] = ACTIONS(1371),
    [anon_sym_await] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1371),
    [sym_this] = ACTIONS(1371),
    [sym_true] = ACTIONS(1371),
    [sym_false] = ACTIONS(1371),
    [sym_none] = ACTIONS(1371),
    [sym_string] = ACTIONS(1371),
    [sym_verbatim] = ACTIONS(1371),
    [sym_number] = ACTIONS(1371),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1369),
    [sym_regex] = ACTIONS(1369),
    [sym_fstring_start] = ACTIONS(1369),
  },
  [STATE(657)] = {
    [sym_identifier] = ACTIONS(1154),
    [anon_sym_SEMI] = ACTIONS(1154),
    [anon_sym_return] = ACTIONS(1154),
    [anon_sym_async] = ACTIONS(1154),
    [anon_sym_def] = ACTIONS(1154),
    [anon_sym_COMMA] = ACTIONS(1154),
    [anon_sym_LPAREN] = ACTIONS(1154),
    [anon_sym_LBRACK] = ACTIONS(1154),
    [anon_sym_del] = ACTIONS(1154),
    [anon_sym_raise] = ACTIONS(1154),
    [sym_pass_statement] = ACTIONS(1154),
    [sym_break_statement] = ACTIONS(1154),
    [sym_continue_statement] = ACTIONS(1154),
    [sym_debugger_statement] = ACTIONS(1154),
    [anon_sym_assert] = ACTIONS(1154),
    [anon_sym_global] = ACTIONS(1154),
    [anon_sym_nonlocal] = ACTIONS(1154),
    [anon_sym_yield] = ACTIONS(1154),
    [anon_sym_from] = ACTIONS(1154),
    [anon_sym_import] = ACTIONS(1154),
    [anon_sym_DOT] = ACTIONS(1154),
    [anon_sym_if] = ACTIONS(1154),
    [anon_sym_for] = ACTIONS(1154),
    [anon_sym_in] = ACTIONS(1154),
    [anon_sym_while] = ACTIONS(1154),
    [anon_sym_do] = ACTIONS(1154),
    [anon_sym_try] = ACTIONS(1154),
    [anon_sym_with] = ACTIONS(1154),
    [anon_sym_STAR] = ACTIONS(1154),
    [anon_sym_STAR_STAR] = ACTIONS(1154),
    [anon_sym_class] = ACTIONS(1154),
    [anon_sym_AT] = ACTIONS(1154),
    [anon_sym_not] = ACTIONS(1154),
    [anon_sym_and] = ACTIONS(1154),
    [anon_sym_or] = ACTIONS(1154),
    [anon_sym_PLUS] = ACTIONS(1154),
    [anon_sym_DASH] = ACTIONS(1154),
    [anon_sym_SLASH] = ACTIONS(1154),
    [anon_sym_SLASH_SLASH] = ACTIONS(1154),
    [anon_sym_PERCENT] = ACTIONS(1154),
    [anon_sym_PIPE] = ACTIONS(1154),
    [anon_sym_CARET] = ACTIONS(1154),
    [anon_sym_AMP] = ACTIONS(1154),
    [anon_sym_LT_LT] = ACTIONS(1154),
    [anon_sym_GT_GT] = ACTIONS(1154),
    [anon_sym_GT_GT_GT] = ACTIONS(1154),
    [anon_sym_TILDE] = ACTIONS(1154),
    [anon_sym_typeof] = ACTIONS(1154),
    [anon_sym_void] = ACTIONS(1154),
    [anon_sym_LT] = ACTIONS(1154),
    [anon_sym_LT_EQ] = ACTIONS(1154),
    [anon_sym_EQ_EQ] = ACTIONS(1154),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1154),
    [anon_sym_BANG_EQ] = ACTIONS(1154),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1154),
    [anon_sym_GT_EQ] = ACTIONS(1154),
    [anon_sym_GT] = ACTIONS(1154),
    [anon_sym_is] = ACTIONS(1154),
    [anon_sym_instanceof] = ACTIONS(1154),
    [anon_sym_new] = ACTIONS(1154),
    [anon_sym_QMARK] = ACTIONS(1154),
    [anon_sym_await] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1154),
    [sym_this] = ACTIONS(1154),
    [sym_true] = ACTIONS(1154),
    [sym_false] = ACTIONS(1154),
    [sym_none] = ACTIONS(1154),
    [sym_string] = ACTIONS(1154),
    [sym_verbatim] = ACTIONS(1154),
    [sym_number] = ACTIONS(1154),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1152),
    [sym_regex] = ACTIONS(1152),
    [sym_fstring_start] = ACTIONS(1152),
  },
  [STATE(658)] = {
    [sym_identifier] = ACTIONS(1447),
    [anon_sym_SEMI] = ACTIONS(1447),
    [anon_sym_return] = ACTIONS(1447),
    [anon_sym_async] = ACTIONS(1447),
    [anon_sym_def] = ACTIONS(1447),
    [anon_sym_COMMA] = ACTIONS(1447),
    [anon_sym_LPAREN] = ACTIONS(1447),
    [anon_sym_LBRACK] = ACTIONS(1447),
    [anon_sym_del] = ACTIONS(1447),
    [anon_sym_raise] = ACTIONS(1447),
    [sym_pass_statement] = ACTIONS(1447),
    [sym_break_statement] = ACTIONS(1447),
    [sym_continue_statement] = ACTIONS(1447),
    [sym_debugger_statement] = ACTIONS(1447),
    [anon_sym_assert] = ACTIONS(1447),
    [anon_sym_global] = ACTIONS(1447),
    [anon_sym_nonlocal] = ACTIONS(1447),
    [anon_sym_yield] = ACTIONS(1447),
    [anon_sym_from] = ACTIONS(1447),
    [anon_sym_import] = ACTIONS(1447),
    [anon_sym_DOT] = ACTIONS(1447),
    [anon_sym_if] = ACTIONS(1447),
    [anon_sym_for] = ACTIONS(1447),
    [anon_sym_in] = ACTIONS(1447),
    [anon_sym_while] = ACTIONS(1447),
    [anon_sym_do] = ACTIONS(1447),
    [anon_sym_try] = ACTIONS(1447),
    [anon_sym_with] = ACTIONS(1447),
    [anon_sym_STAR] = ACTIONS(1447),
    [anon_sym_STAR_STAR] = ACTIONS(1447),
    [anon_sym_class] = ACTIONS(1447),
    [anon_sym_AT] = ACTIONS(1447),
    [anon_sym_not] = ACTIONS(1447),
    [anon_sym_and] = ACTIONS(1447),
    [anon_sym_or] = ACTIONS(1447),
    [anon_sym_PLUS] = ACTIONS(1447),
    [anon_sym_DASH] = ACTIONS(1447),
    [anon_sym_SLASH] = ACTIONS(1447),
    [anon_sym_SLASH_SLASH] = ACTIONS(1447),
    [anon_sym_PERCENT] = ACTIONS(1447),
    [anon_sym_PIPE] = ACTIONS(1447),
    [anon_sym_CARET] = ACTIONS(1447),
    [anon_sym_AMP] = ACTIONS(1447),
    [anon_sym_LT_LT] = ACTIONS(1447),
    [anon_sym_GT_GT] = ACTIONS(1447),
    [anon_sym_GT_GT_GT] = ACTIONS(1447),
    [anon_sym_TILDE] = ACTIONS(1447),
    [anon_sym_typeof] = ACTIONS(1447),
    [anon_sym_void] = ACTIONS(1447),
    [anon_sym_LT] = ACTIONS(1447),
    [anon_sym_LT_EQ] = ACTIONS(1447),
    [anon_sym_EQ_EQ] = ACTIONS(1447),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1447),
    [anon_sym_BANG_EQ] = ACTIONS(1447),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1447),
    [anon_sym_GT_EQ] = ACTIONS(1447),
    [anon_sym_GT] = ACTIONS(1447),
    [anon_sym_is] = ACTIONS(1447),
    [anon_sym_instanceof] = ACTIONS(1447),
    [anon_sym_new] = ACTIONS(1447),
    [anon_sym_QMARK] = ACTIONS(1447),
    [anon_sym_await] = ACTIONS(1447),
    [anon_sym_LBRACE] = ACTIONS(1447),
    [sym_this] = ACTIONS(1447),
    [sym_true] = ACTIONS(1447),
    [sym_false] = ACTIONS(1447),
    [sym_none] = ACTIONS(1447),
    [sym_string] = ACTIONS(1447),
    [sym_verbatim] = ACTIONS(1447),
    [sym_number] = ACTIONS(1447),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1445),
    [sym_regex] = ACTIONS(1445),
    [sym_fstring_start] = ACTIONS(1445),
  },
  [STATE(659)] = {
    [sym_identifier] = ACTIONS(1072),
    [anon_sym_SEMI] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_async] = ACTIONS(1072),
    [anon_sym_def] = ACTIONS(1072),
    [anon_sym_COMMA] = ACTIONS(1072),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_del] = ACTIONS(1072),
    [anon_sym_raise] = ACTIONS(1072),
    [sym_pass_statement] = ACTIONS(1072),
    [sym_break_statement] = ACTIONS(1072),
    [sym_continue_statement] = ACTIONS(1072),
    [sym_debugger_statement] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_global] = ACTIONS(1072),
    [anon_sym_nonlocal] = ACTIONS(1072),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_from] = ACTIONS(1072),
    [anon_sym_import] = ACTIONS(1072),
    [anon_sym_DOT] = ACTIONS(1072),
    [anon_sym_if] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1072),
    [anon_sym_in] = ACTIONS(1072),
    [anon_sym_while] = ACTIONS(1072),
    [anon_sym_do] = ACTIONS(1072),
    [anon_sym_try] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(1072),
    [anon_sym_STAR] = ACTIONS(1072),
    [anon_sym_STAR_STAR] = ACTIONS(1072),
    [anon_sym_class] = ACTIONS(1072),
    [anon_sym_AT] = ACTIONS(1072),
    [anon_sym_not] = ACTIONS(1072),
    [anon_sym_and] = ACTIONS(1072),
    [anon_sym_or] = ACTIONS(1072),
    [anon_sym_PLUS] = ACTIONS(1072),
    [anon_sym_DASH] = ACTIONS(1072),
    [anon_sym_SLASH] = ACTIONS(1072),
    [anon_sym_SLASH_SLASH] = ACTIONS(1072),
    [anon_sym_PERCENT] = ACTIONS(1072),
    [anon_sym_PIPE] = ACTIONS(1072),
    [anon_sym_CARET] = ACTIONS(1072),
    [anon_sym_AMP] = ACTIONS(1072),
    [anon_sym_LT_LT] = ACTIONS(1072),
    [anon_sym_GT_GT] = ACTIONS(1072),
    [anon_sym_GT_GT_GT] = ACTIONS(1072),
    [anon_sym_TILDE] = ACTIONS(1072),
    [anon_sym_typeof] = ACTIONS(1072),
    [anon_sym_void] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(1072),
    [anon_sym_LT_EQ] = ACTIONS(1072),
    [anon_sym_EQ_EQ] = ACTIONS(1072),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1072),
    [anon_sym_BANG_EQ] = ACTIONS(1072),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1072),
    [anon_sym_GT_EQ] = ACTIONS(1072),
    [anon_sym_GT] = ACTIONS(1072),
    [anon_sym_is] = ACTIONS(1072),
    [anon_sym_instanceof] = ACTIONS(1072),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_QMARK] = ACTIONS(1072),
    [anon_sym_await] = ACTIONS(1072),
    [anon_sym_LBRACE] = ACTIONS(1072),
    [sym_this] = ACTIONS(1072),
    [sym_true] = ACTIONS(1072),
    [sym_false] = ACTIONS(1072),
    [sym_none] = ACTIONS(1072),
    [sym_string] = ACTIONS(1072),
    [sym_verbatim] = ACTIONS(1072),
    [sym_number] = ACTIONS(1072),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1101),
    [sym_regex] = ACTIONS(1101),
    [sym_fstring_start] = ACTIONS(1101),
  },
  [STATE(660)] = {
    [sym_identifier] = ACTIONS(1133),
    [anon_sym_SEMI] = ACTIONS(1133),
    [anon_sym_return] = ACTIONS(1133),
    [anon_sym_async] = ACTIONS(1133),
    [anon_sym_def] = ACTIONS(1133),
    [anon_sym_COMMA] = ACTIONS(1133),
    [anon_sym_LPAREN] = ACTIONS(1133),
    [anon_sym_LBRACK] = ACTIONS(1133),
    [anon_sym_del] = ACTIONS(1133),
    [anon_sym_raise] = ACTIONS(1133),
    [sym_pass_statement] = ACTIONS(1133),
    [sym_break_statement] = ACTIONS(1133),
    [sym_continue_statement] = ACTIONS(1133),
    [sym_debugger_statement] = ACTIONS(1133),
    [anon_sym_assert] = ACTIONS(1133),
    [anon_sym_global] = ACTIONS(1133),
    [anon_sym_nonlocal] = ACTIONS(1133),
    [anon_sym_yield] = ACTIONS(1133),
    [anon_sym_from] = ACTIONS(1133),
    [anon_sym_import] = ACTIONS(1133),
    [anon_sym_DOT] = ACTIONS(1133),
    [anon_sym_if] = ACTIONS(1133),
    [anon_sym_for] = ACTIONS(1133),
    [anon_sym_in] = ACTIONS(1133),
    [anon_sym_while] = ACTIONS(1133),
    [anon_sym_do] = ACTIONS(1133),
    [anon_sym_try] = ACTIONS(1133),
    [anon_sym_with] = ACTIONS(1133),
    [anon_sym_STAR] = ACTIONS(1133),
    [anon_sym_STAR_STAR] = ACTIONS(1133),
    [anon_sym_class] = ACTIONS(1133),
    [anon_sym_AT] = ACTIONS(1133),
    [anon_sym_not] = ACTIONS(1133),
    [anon_sym_and] = ACTIONS(1133),
    [anon_sym_or] = ACTIONS(1133),
    [anon_sym_PLUS] = ACTIONS(1133),
    [anon_sym_DASH] = ACTIONS(1133),
    [anon_sym_SLASH] = ACTIONS(1133),
    [anon_sym_SLASH_SLASH] = ACTIONS(1133),
    [anon_sym_PERCENT] = ACTIONS(1133),
    [anon_sym_PIPE] = ACTIONS(1133),
    [anon_sym_CARET] = ACTIONS(1133),
    [anon_sym_AMP] = ACTIONS(1133),
    [anon_sym_LT_LT] = ACTIONS(1133),
    [anon_sym_GT_GT] = ACTIONS(1133),
    [anon_sym_GT_GT_GT] = ACTIONS(1133),
    [anon_sym_TILDE] = ACTIONS(1133),
    [anon_sym_typeof] = ACTIONS(1133),
    [anon_sym_void] = ACTIONS(1133),
    [anon_sym_LT] = ACTIONS(1133),
    [anon_sym_LT_EQ] = ACTIONS(1133),
    [anon_sym_EQ_EQ] = ACTIONS(1133),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1133),
    [anon_sym_BANG_EQ] = ACTIONS(1133),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1133),
    [anon_sym_GT_EQ] = ACTIONS(1133),
    [anon_sym_GT] = ACTIONS(1133),
    [anon_sym_is] = ACTIONS(1133),
    [anon_sym_instanceof] = ACTIONS(1133),
    [anon_sym_new] = ACTIONS(1133),
    [anon_sym_QMARK] = ACTIONS(1133),
    [anon_sym_await] = ACTIONS(1133),
    [anon_sym_LBRACE] = ACTIONS(1133),
    [sym_this] = ACTIONS(1133),
    [sym_true] = ACTIONS(1133),
    [sym_false] = ACTIONS(1133),
    [sym_none] = ACTIONS(1133),
    [sym_string] = ACTIONS(1133),
    [sym_verbatim] = ACTIONS(1133),
    [sym_number] = ACTIONS(1133),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1137),
    [sym_regex] = ACTIONS(1137),
    [sym_fstring_start] = ACTIONS(1137),
  },
  [STATE(661)] = {
    [aux_sym__inline_body_repeat1] = STATE(661),
    [sym_identifier] = ACTIONS(1371),
    [anon_sym_SEMI] = ACTIONS(1606),
    [anon_sym_return] = ACTIONS(1371),
    [anon_sym_async] = ACTIONS(1371),
    [anon_sym_def] = ACTIONS(1371),
    [anon_sym_LPAREN] = ACTIONS(1371),
    [anon_sym_LBRACK] = ACTIONS(1371),
    [anon_sym_del] = ACTIONS(1371),
    [anon_sym_raise] = ACTIONS(1371),
    [sym_pass_statement] = ACTIONS(1371),
    [sym_break_statement] = ACTIONS(1371),
    [sym_continue_statement] = ACTIONS(1371),
    [sym_debugger_statement] = ACTIONS(1371),
    [anon_sym_assert] = ACTIONS(1371),
    [anon_sym_global] = ACTIONS(1371),
    [anon_sym_nonlocal] = ACTIONS(1371),
    [anon_sym_yield] = ACTIONS(1371),
    [anon_sym_from] = ACTIONS(1371),
    [anon_sym_import] = ACTIONS(1371),
    [anon_sym_DOT] = ACTIONS(1371),
    [anon_sym_if] = ACTIONS(1371),
    [anon_sym_for] = ACTIONS(1371),
    [anon_sym_in] = ACTIONS(1371),
    [anon_sym_while] = ACTIONS(1371),
    [anon_sym_do] = ACTIONS(1371),
    [anon_sym_try] = ACTIONS(1371),
    [anon_sym_with] = ACTIONS(1371),
    [anon_sym_STAR] = ACTIONS(1371),
    [anon_sym_STAR_STAR] = ACTIONS(1371),
    [anon_sym_class] = ACTIONS(1371),
    [anon_sym_AT] = ACTIONS(1371),
    [anon_sym_not] = ACTIONS(1371),
    [anon_sym_and] = ACTIONS(1371),
    [anon_sym_or] = ACTIONS(1371),
    [anon_sym_PLUS] = ACTIONS(1371),
    [anon_sym_DASH] = ACTIONS(1371),
    [anon_sym_SLASH] = ACTIONS(1371),
    [anon_sym_SLASH_SLASH] = ACTIONS(1371),
    [anon_sym_PERCENT] = ACTIONS(1371),
    [anon_sym_PIPE] = ACTIONS(1371),
    [anon_sym_CARET] = ACTIONS(1371),
    [anon_sym_AMP] = ACTIONS(1371),
    [anon_sym_LT_LT] = ACTIONS(1371),
    [anon_sym_GT_GT] = ACTIONS(1371),
    [anon_sym_GT_GT_GT] = ACTIONS(1371),
    [anon_sym_TILDE] = ACTIONS(1371),
    [anon_sym_typeof] = ACTIONS(1371),
    [anon_sym_void] = ACTIONS(1371),
    [anon_sym_LT] = ACTIONS(1371),
    [anon_sym_LT_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ] = ACTIONS(1371),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ] = ACTIONS(1371),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1371),
    [anon_sym_GT_EQ] = ACTIONS(1371),
    [anon_sym_GT] = ACTIONS(1371),
    [anon_sym_is] = ACTIONS(1371),
    [anon_sym_instanceof] = ACTIONS(1371),
    [anon_sym_new] = ACTIONS(1371),
    [anon_sym_QMARK] = ACTIONS(1371),
    [anon_sym_await] = ACTIONS(1371),
    [anon_sym_LBRACE] = ACTIONS(1371),
    [sym_this] = ACTIONS(1371),
    [sym_true] = ACTIONS(1371),
    [sym_false] = ACTIONS(1371),
    [sym_none] = ACTIONS(1371),
    [sym_string] = ACTIONS(1371),
    [sym_verbatim] = ACTIONS(1371),
    [sym_number] = ACTIONS(1371),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1369),
    [sym_regex] = ACTIONS(1369),
    [sym_fstring_start] = ACTIONS(1369),
  },
  [STATE(662)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_async] = ACTIONS(1270),
    [anon_sym_def] = ACTIONS(1270),
    [anon_sym_LPAREN] = ACTIONS(1270),
    [anon_sym_LBRACK] = ACTIONS(1270),
    [anon_sym_del] = ACTIONS(1270),
    [anon_sym_raise] = ACTIONS(1270),
    [sym_pass_statement] = ACTIONS(1270),
    [sym_break_statement] = ACTIONS(1270),
    [sym_continue_statement] = ACTIONS(1270),
    [sym_debugger_statement] = ACTIONS(1270),
    [anon_sym_assert] = ACTIONS(1270),
    [anon_sym_global] = ACTIONS(1270),
    [anon_sym_nonlocal] = ACTIONS(1270),
    [anon_sym_yield] = ACTIONS(1270),
    [anon_sym_from] = ACTIONS(1270),
    [anon_sym_import] = ACTIONS(1270),
    [anon_sym_DOT] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_do] = ACTIONS(1270),
    [anon_sym_try] = ACTIONS(1270),
    [anon_sym_with] = ACTIONS(1270),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1270),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1270),
    [anon_sym_or] = ACTIONS(1270),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1270),
    [anon_sym_typeof] = ACTIONS(1270),
    [anon_sym_void] = ACTIONS(1270),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1270),
    [anon_sym_QMARK] = ACTIONS(1270),
    [anon_sym_await] = ACTIONS(1270),
    [anon_sym_LBRACE] = ACTIONS(1270),
    [sym_this] = ACTIONS(1270),
    [sym_true] = ACTIONS(1270),
    [sym_false] = ACTIONS(1270),
    [sym_none] = ACTIONS(1270),
    [sym_string] = ACTIONS(1270),
    [sym_verbatim] = ACTIONS(1270),
    [sym_number] = ACTIONS(1270),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1268),
    [sym_regex] = ACTIONS(1268),
    [sym_fstring_start] = ACTIONS(1268),
  },
  [STATE(663)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1385),
    [anon_sym_SEMI] = ACTIONS(1385),
    [anon_sym_return] = ACTIONS(1385),
    [anon_sym_async] = ACTIONS(1385),
    [anon_sym_def] = ACTIONS(1385),
    [anon_sym_LPAREN] = ACTIONS(1385),
    [anon_sym_LBRACK] = ACTIONS(1385),
    [anon_sym_del] = ACTIONS(1385),
    [anon_sym_raise] = ACTIONS(1385),
    [sym_pass_statement] = ACTIONS(1385),
    [sym_break_statement] = ACTIONS(1385),
    [sym_continue_statement] = ACTIONS(1385),
    [sym_debugger_statement] = ACTIONS(1385),
    [anon_sym_assert] = ACTIONS(1385),
    [anon_sym_global] = ACTIONS(1385),
    [anon_sym_nonlocal] = ACTIONS(1385),
    [anon_sym_yield] = ACTIONS(1385),
    [anon_sym_from] = ACTIONS(1385),
    [anon_sym_import] = ACTIONS(1385),
    [anon_sym_DOT] = ACTIONS(1385),
    [anon_sym_if] = ACTIONS(1385),
    [anon_sym_for] = ACTIONS(1385),
    [anon_sym_in] = ACTIONS(1385),
    [anon_sym_while] = ACTIONS(1385),
    [anon_sym_do] = ACTIONS(1385),
    [anon_sym_try] = ACTIONS(1385),
    [anon_sym_with] = ACTIONS(1385),
    [anon_sym_STAR] = ACTIONS(1385),
    [anon_sym_STAR_STAR] = ACTIONS(1385),
    [anon_sym_class] = ACTIONS(1385),
    [anon_sym_AT] = ACTIONS(1385),
    [anon_sym_not] = ACTIONS(1385),
    [anon_sym_and] = ACTIONS(1385),
    [anon_sym_or] = ACTIONS(1385),
    [anon_sym_PLUS] = ACTIONS(1385),
    [anon_sym_DASH] = ACTIONS(1385),
    [anon_sym_SLASH] = ACTIONS(1385),
    [anon_sym_SLASH_SLASH] = ACTIONS(1385),
    [anon_sym_PERCENT] = ACTIONS(1385),
    [anon_sym_PIPE] = ACTIONS(1385),
    [anon_sym_CARET] = ACTIONS(1385),
    [anon_sym_AMP] = ACTIONS(1385),
    [anon_sym_LT_LT] = ACTIONS(1385),
    [anon_sym_GT_GT] = ACTIONS(1385),
    [anon_sym_GT_GT_GT] = ACTIONS(1385),
    [anon_sym_TILDE] = ACTIONS(1385),
    [anon_sym_typeof] = ACTIONS(1385),
    [anon_sym_void] = ACTIONS(1385),
    [anon_sym_LT] = ACTIONS(1385),
    [anon_sym_LT_EQ] = ACTIONS(1385),
    [anon_sym_EQ_EQ] = ACTIONS(1385),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1385),
    [anon_sym_BANG_EQ] = ACTIONS(1385),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1385),
    [anon_sym_GT_EQ] = ACTIONS(1385),
    [anon_sym_GT] = ACTIONS(1385),
    [anon_sym_is] = ACTIONS(1385),
    [anon_sym_instanceof] = ACTIONS(1385),
    [anon_sym_new] = ACTIONS(1385),
    [anon_sym_QMARK] = ACTIONS(1385),
    [anon_sym_await] = ACTIONS(1385),
    [anon_sym_LBRACE] = ACTIONS(1385),
    [sym_this] = ACTIONS(1385),
    [sym_true] = ACTIONS(1385),
    [sym_false] = ACTIONS(1385),
    [sym_none] = ACTIONS(1385),
    [sym_string] = ACTIONS(1385),
    [sym_verbatim] = ACTIONS(1385),
    [sym_number] = ACTIONS(1385),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1383),
    [sym_regex] = ACTIONS(1383),
    [sym_fstring_start] = ACTIONS(1383),
  },
  [STATE(664)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1266),
    [anon_sym_SEMI] = ACTIONS(1266),
    [anon_sym_return] = ACTIONS(1266),
    [anon_sym_async] = ACTIONS(1266),
    [anon_sym_def] = ACTIONS(1266),
    [anon_sym_LPAREN] = ACTIONS(1266),
    [anon_sym_LBRACK] = ACTIONS(1266),
    [anon_sym_del] = ACTIONS(1266),
    [anon_sym_raise] = ACTIONS(1266),
    [sym_pass_statement] = ACTIONS(1266),
    [sym_break_statement] = ACTIONS(1266),
    [sym_continue_statement] = ACTIONS(1266),
    [sym_debugger_statement] = ACTIONS(1266),
    [anon_sym_assert] = ACTIONS(1266),
    [anon_sym_global] = ACTIONS(1266),
    [anon_sym_nonlocal] = ACTIONS(1266),
    [anon_sym_yield] = ACTIONS(1266),
    [anon_sym_from] = ACTIONS(1266),
    [anon_sym_import] = ACTIONS(1266),
    [anon_sym_DOT] = ACTIONS(1266),
    [anon_sym_if] = ACTIONS(1266),
    [anon_sym_for] = ACTIONS(1266),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1266),
    [anon_sym_do] = ACTIONS(1266),
    [anon_sym_try] = ACTIONS(1266),
    [anon_sym_with] = ACTIONS(1266),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1266),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1425),
    [anon_sym_or] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1266),
    [anon_sym_typeof] = ACTIONS(1266),
    [anon_sym_void] = ACTIONS(1266),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1266),
    [anon_sym_QMARK] = ACTIONS(1266),
    [anon_sym_await] = ACTIONS(1266),
    [anon_sym_LBRACE] = ACTIONS(1266),
    [sym_this] = ACTIONS(1266),
    [sym_true] = ACTIONS(1266),
    [sym_false] = ACTIONS(1266),
    [sym_none] = ACTIONS(1266),
    [sym_string] = ACTIONS(1266),
    [sym_verbatim] = ACTIONS(1266),
    [sym_number] = ACTIONS(1266),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1264),
    [sym_regex] = ACTIONS(1264),
    [sym_fstring_start] = ACTIONS(1264),
  },
  [STATE(665)] = {
    [ts_builtin_sym_end] = ACTIONS(1198),
    [sym_identifier] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym_return] = ACTIONS(1200),
    [anon_sym_async] = ACTIONS(1200),
    [anon_sym_def] = ACTIONS(1200),
    [anon_sym_COMMA] = ACTIONS(1200),
    [anon_sym_LPAREN] = ACTIONS(1200),
    [anon_sym_LBRACK] = ACTIONS(1200),
    [anon_sym_del] = ACTIONS(1200),
    [anon_sym_raise] = ACTIONS(1200),
    [sym_pass_statement] = ACTIONS(1200),
    [sym_break_statement] = ACTIONS(1200),
    [sym_continue_statement] = ACTIONS(1200),
    [sym_debugger_statement] = ACTIONS(1200),
    [anon_sym_assert] = ACTIONS(1200),
    [anon_sym_global] = ACTIONS(1200),
    [anon_sym_nonlocal] = ACTIONS(1200),
    [anon_sym_yield] = ACTIONS(1200),
    [anon_sym_from] = ACTIONS(1200),
    [anon_sym_import] = ACTIONS(1200),
    [anon_sym_DOT] = ACTIONS(1200),
    [anon_sym_if] = ACTIONS(1200),
    [anon_sym_for] = ACTIONS(1200),
    [anon_sym_in] = ACTIONS(1200),
    [anon_sym_while] = ACTIONS(1200),
    [anon_sym_do] = ACTIONS(1200),
    [anon_sym_try] = ACTIONS(1200),
    [anon_sym_with] = ACTIONS(1200),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_STAR_STAR] = ACTIONS(1200),
    [anon_sym_class] = ACTIONS(1200),
    [anon_sym_AT] = ACTIONS(1200),
    [anon_sym_not] = ACTIONS(1200),
    [anon_sym_and] = ACTIONS(1200),
    [anon_sym_or] = ACTIONS(1200),
    [anon_sym_PLUS] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [anon_sym_SLASH] = ACTIONS(1200),
    [anon_sym_SLASH_SLASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(1200),
    [anon_sym_PIPE] = ACTIONS(1200),
    [anon_sym_CARET] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_LT_LT] = ACTIONS(1200),
    [anon_sym_GT_GT] = ACTIONS(1200),
    [anon_sym_GT_GT_GT] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_typeof] = ACTIONS(1200),
    [anon_sym_void] = ACTIONS(1200),
    [anon_sym_LT] = ACTIONS(1200),
    [anon_sym_LT_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1200),
    [anon_sym_GT_EQ] = ACTIONS(1200),
    [anon_sym_GT] = ACTIONS(1200),
    [anon_sym_is] = ACTIONS(1200),
    [anon_sym_instanceof] = ACTIONS(1200),
    [anon_sym_new] = ACTIONS(1200),
    [anon_sym_QMARK] = ACTIONS(1200),
    [anon_sym_await] = ACTIONS(1200),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [sym_this] = ACTIONS(1200),
    [sym_true] = ACTIONS(1200),
    [sym_false] = ACTIONS(1200),
    [sym_none] = ACTIONS(1200),
    [sym_string] = ACTIONS(1200),
    [sym_verbatim] = ACTIONS(1200),
    [sym_number] = ACTIONS(1200),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1198),
    [sym_fstring_start] = ACTIONS(1198),
  },
  [STATE(666)] = {
    [ts_builtin_sym_end] = ACTIONS(1609),
    [sym_identifier] = ACTIONS(1611),
    [anon_sym_SEMI] = ACTIONS(1611),
    [anon_sym_return] = ACTIONS(1611),
    [anon_sym_async] = ACTIONS(1611),
    [anon_sym_def] = ACTIONS(1611),
    [anon_sym_COMMA] = ACTIONS(1611),
    [anon_sym_LPAREN] = ACTIONS(1611),
    [anon_sym_LBRACK] = ACTIONS(1611),
    [anon_sym_del] = ACTIONS(1611),
    [anon_sym_raise] = ACTIONS(1611),
    [sym_pass_statement] = ACTIONS(1611),
    [sym_break_statement] = ACTIONS(1611),
    [sym_continue_statement] = ACTIONS(1611),
    [sym_debugger_statement] = ACTIONS(1611),
    [anon_sym_assert] = ACTIONS(1611),
    [anon_sym_global] = ACTIONS(1611),
    [anon_sym_nonlocal] = ACTIONS(1611),
    [anon_sym_yield] = ACTIONS(1611),
    [anon_sym_from] = ACTIONS(1611),
    [anon_sym_import] = ACTIONS(1611),
    [anon_sym_DOT] = ACTIONS(1611),
    [anon_sym_if] = ACTIONS(1611),
    [anon_sym_for] = ACTIONS(1611),
    [anon_sym_in] = ACTIONS(1611),
    [anon_sym_while] = ACTIONS(1611),
    [anon_sym_do] = ACTIONS(1611),
    [anon_sym_try] = ACTIONS(1611),
    [anon_sym_with] = ACTIONS(1611),
    [anon_sym_STAR] = ACTIONS(1611),
    [anon_sym_STAR_STAR] = ACTIONS(1611),
    [anon_sym_class] = ACTIONS(1611),
    [anon_sym_AT] = ACTIONS(1611),
    [anon_sym_not] = ACTIONS(1611),
    [anon_sym_and] = ACTIONS(1611),
    [anon_sym_or] = ACTIONS(1611),
    [anon_sym_PLUS] = ACTIONS(1611),
    [anon_sym_DASH] = ACTIONS(1611),
    [anon_sym_SLASH] = ACTIONS(1611),
    [anon_sym_SLASH_SLASH] = ACTIONS(1611),
    [anon_sym_PERCENT] = ACTIONS(1611),
    [anon_sym_PIPE] = ACTIONS(1611),
    [anon_sym_CARET] = ACTIONS(1611),
    [anon_sym_AMP] = ACTIONS(1611),
    [anon_sym_LT_LT] = ACTIONS(1611),
    [anon_sym_GT_GT] = ACTIONS(1611),
    [anon_sym_GT_GT_GT] = ACTIONS(1611),
    [anon_sym_TILDE] = ACTIONS(1611),
    [anon_sym_typeof] = ACTIONS(1611),
    [anon_sym_void] = ACTIONS(1611),
    [anon_sym_LT] = ACTIONS(1611),
    [anon_sym_LT_EQ] = ACTIONS(1611),
    [anon_sym_EQ_EQ] = ACTIONS(1611),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1611),
    [anon_sym_BANG_EQ] = ACTIONS(1611),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1611),
    [anon_sym_GT_EQ] = ACTIONS(1611),
    [anon_sym_GT] = ACTIONS(1611),
    [anon_sym_is] = ACTIONS(1611),
    [anon_sym_instanceof] = ACTIONS(1611),
    [anon_sym_new] = ACTIONS(1611),
    [anon_sym_QMARK] = ACTIONS(1611),
    [anon_sym_await] = ACTIONS(1611),
    [anon_sym_LBRACE] = ACTIONS(1611),
    [sym_this] = ACTIONS(1611),
    [sym_true] = ACTIONS(1611),
    [sym_false] = ACTIONS(1611),
    [sym_none] = ACTIONS(1611),
    [sym_string] = ACTIONS(1611),
    [sym_verbatim] = ACTIONS(1611),
    [sym_number] = ACTIONS(1611),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1609),
    [sym_fstring_start] = ACTIONS(1609),
  },
  [STATE(667)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1194),
    [anon_sym_SEMI] = ACTIONS(1194),
    [anon_sym_return] = ACTIONS(1194),
    [anon_sym_async] = ACTIONS(1194),
    [anon_sym_def] = ACTIONS(1194),
    [anon_sym_LPAREN] = ACTIONS(1194),
    [anon_sym_LBRACK] = ACTIONS(1194),
    [anon_sym_del] = ACTIONS(1194),
    [anon_sym_raise] = ACTIONS(1194),
    [sym_pass_statement] = ACTIONS(1194),
    [sym_break_statement] = ACTIONS(1194),
    [sym_continue_statement] = ACTIONS(1194),
    [sym_debugger_statement] = ACTIONS(1194),
    [anon_sym_assert] = ACTIONS(1194),
    [anon_sym_global] = ACTIONS(1194),
    [anon_sym_nonlocal] = ACTIONS(1194),
    [anon_sym_yield] = ACTIONS(1194),
    [anon_sym_from] = ACTIONS(1194),
    [anon_sym_import] = ACTIONS(1194),
    [anon_sym_DOT] = ACTIONS(1194),
    [anon_sym_if] = ACTIONS(1194),
    [anon_sym_for] = ACTIONS(1194),
    [anon_sym_in] = ACTIONS(1194),
    [anon_sym_while] = ACTIONS(1194),
    [anon_sym_do] = ACTIONS(1194),
    [anon_sym_try] = ACTIONS(1194),
    [anon_sym_with] = ACTIONS(1194),
    [anon_sym_STAR] = ACTIONS(1194),
    [anon_sym_STAR_STAR] = ACTIONS(1194),
    [anon_sym_class] = ACTIONS(1194),
    [anon_sym_AT] = ACTIONS(1194),
    [anon_sym_not] = ACTIONS(1194),
    [anon_sym_and] = ACTIONS(1194),
    [anon_sym_or] = ACTIONS(1194),
    [anon_sym_PLUS] = ACTIONS(1194),
    [anon_sym_DASH] = ACTIONS(1194),
    [anon_sym_SLASH] = ACTIONS(1194),
    [anon_sym_SLASH_SLASH] = ACTIONS(1194),
    [anon_sym_PERCENT] = ACTIONS(1194),
    [anon_sym_PIPE] = ACTIONS(1194),
    [anon_sym_CARET] = ACTIONS(1194),
    [anon_sym_AMP] = ACTIONS(1194),
    [anon_sym_LT_LT] = ACTIONS(1194),
    [anon_sym_GT_GT] = ACTIONS(1194),
    [anon_sym_GT_GT_GT] = ACTIONS(1194),
    [anon_sym_TILDE] = ACTIONS(1194),
    [anon_sym_typeof] = ACTIONS(1194),
    [anon_sym_void] = ACTIONS(1194),
    [anon_sym_LT] = ACTIONS(1194),
    [anon_sym_LT_EQ] = ACTIONS(1194),
    [anon_sym_EQ_EQ] = ACTIONS(1194),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1194),
    [anon_sym_BANG_EQ] = ACTIONS(1194),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1194),
    [anon_sym_GT_EQ] = ACTIONS(1194),
    [anon_sym_GT] = ACTIONS(1194),
    [anon_sym_is] = ACTIONS(1194),
    [anon_sym_instanceof] = ACTIONS(1194),
    [anon_sym_new] = ACTIONS(1194),
    [anon_sym_QMARK] = ACTIONS(1194),
    [anon_sym_await] = ACTIONS(1194),
    [anon_sym_LBRACE] = ACTIONS(1194),
    [sym_this] = ACTIONS(1194),
    [sym_true] = ACTIONS(1194),
    [sym_false] = ACTIONS(1194),
    [sym_none] = ACTIONS(1194),
    [sym_string] = ACTIONS(1194),
    [sym_verbatim] = ACTIONS(1194),
    [sym_number] = ACTIONS(1194),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1196),
    [sym_regex] = ACTIONS(1196),
    [sym_fstring_start] = ACTIONS(1196),
  },
  [STATE(668)] = {
    [ts_builtin_sym_end] = ACTIONS(1596),
    [sym_identifier] = ACTIONS(1594),
    [anon_sym_SEMI] = ACTIONS(1594),
    [anon_sym_return] = ACTIONS(1594),
    [anon_sym_async] = ACTIONS(1594),
    [anon_sym_def] = ACTIONS(1594),
    [anon_sym_COMMA] = ACTIONS(1594),
    [anon_sym_LPAREN] = ACTIONS(1594),
    [anon_sym_LBRACK] = ACTIONS(1594),
    [anon_sym_del] = ACTIONS(1594),
    [anon_sym_raise] = ACTIONS(1594),
    [sym_pass_statement] = ACTIONS(1594),
    [sym_break_statement] = ACTIONS(1594),
    [sym_continue_statement] = ACTIONS(1594),
    [sym_debugger_statement] = ACTIONS(1594),
    [anon_sym_assert] = ACTIONS(1594),
    [anon_sym_global] = ACTIONS(1594),
    [anon_sym_nonlocal] = ACTIONS(1594),
    [anon_sym_yield] = ACTIONS(1594),
    [anon_sym_from] = ACTIONS(1594),
    [anon_sym_import] = ACTIONS(1594),
    [anon_sym_DOT] = ACTIONS(1594),
    [anon_sym_if] = ACTIONS(1594),
    [anon_sym_for] = ACTIONS(1594),
    [anon_sym_in] = ACTIONS(1594),
    [anon_sym_while] = ACTIONS(1594),
    [anon_sym_do] = ACTIONS(1594),
    [anon_sym_try] = ACTIONS(1594),
    [anon_sym_with] = ACTIONS(1594),
    [anon_sym_STAR] = ACTIONS(1594),
    [anon_sym_STAR_STAR] = ACTIONS(1594),
    [anon_sym_class] = ACTIONS(1594),
    [anon_sym_AT] = ACTIONS(1594),
    [anon_sym_not] = ACTIONS(1594),
    [anon_sym_and] = ACTIONS(1594),
    [anon_sym_or] = ACTIONS(1594),
    [anon_sym_PLUS] = ACTIONS(1594),
    [anon_sym_DASH] = ACTIONS(1594),
    [anon_sym_SLASH] = ACTIONS(1594),
    [anon_sym_SLASH_SLASH] = ACTIONS(1594),
    [anon_sym_PERCENT] = ACTIONS(1594),
    [anon_sym_PIPE] = ACTIONS(1594),
    [anon_sym_CARET] = ACTIONS(1594),
    [anon_sym_AMP] = ACTIONS(1594),
    [anon_sym_LT_LT] = ACTIONS(1594),
    [anon_sym_GT_GT] = ACTIONS(1594),
    [anon_sym_GT_GT_GT] = ACTIONS(1594),
    [anon_sym_TILDE] = ACTIONS(1594),
    [anon_sym_typeof] = ACTIONS(1594),
    [anon_sym_void] = ACTIONS(1594),
    [anon_sym_LT] = ACTIONS(1594),
    [anon_sym_LT_EQ] = ACTIONS(1594),
    [anon_sym_EQ_EQ] = ACTIONS(1594),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1594),
    [anon_sym_BANG_EQ] = ACTIONS(1594),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1594),
    [anon_sym_GT_EQ] = ACTIONS(1594),
    [anon_sym_GT] = ACTIONS(1594),
    [anon_sym_is] = ACTIONS(1594),
    [anon_sym_instanceof] = ACTIONS(1594),
    [anon_sym_new] = ACTIONS(1594),
    [anon_sym_QMARK] = ACTIONS(1594),
    [anon_sym_await] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(1594),
    [sym_this] = ACTIONS(1594),
    [sym_true] = ACTIONS(1594),
    [sym_false] = ACTIONS(1594),
    [sym_none] = ACTIONS(1594),
    [sym_string] = ACTIONS(1594),
    [sym_verbatim] = ACTIONS(1594),
    [sym_number] = ACTIONS(1594),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1596),
    [sym_fstring_start] = ACTIONS(1596),
  },
  [STATE(669)] = {
    [sym_identifier] = ACTIONS(1451),
    [anon_sym_SEMI] = ACTIONS(1451),
    [anon_sym_return] = ACTIONS(1451),
    [anon_sym_async] = ACTIONS(1451),
    [anon_sym_def] = ACTIONS(1451),
    [anon_sym_COMMA] = ACTIONS(1451),
    [anon_sym_LPAREN] = ACTIONS(1451),
    [anon_sym_LBRACK] = ACTIONS(1451),
    [anon_sym_del] = ACTIONS(1451),
    [anon_sym_raise] = ACTIONS(1451),
    [sym_pass_statement] = ACTIONS(1451),
    [sym_break_statement] = ACTIONS(1451),
    [sym_continue_statement] = ACTIONS(1451),
    [sym_debugger_statement] = ACTIONS(1451),
    [anon_sym_assert] = ACTIONS(1451),
    [anon_sym_global] = ACTIONS(1451),
    [anon_sym_nonlocal] = ACTIONS(1451),
    [anon_sym_yield] = ACTIONS(1451),
    [anon_sym_from] = ACTIONS(1451),
    [anon_sym_import] = ACTIONS(1451),
    [anon_sym_DOT] = ACTIONS(1451),
    [anon_sym_if] = ACTIONS(1451),
    [anon_sym_for] = ACTIONS(1451),
    [anon_sym_in] = ACTIONS(1451),
    [anon_sym_while] = ACTIONS(1451),
    [anon_sym_do] = ACTIONS(1451),
    [anon_sym_try] = ACTIONS(1451),
    [anon_sym_with] = ACTIONS(1451),
    [anon_sym_STAR] = ACTIONS(1451),
    [anon_sym_STAR_STAR] = ACTIONS(1451),
    [anon_sym_class] = ACTIONS(1451),
    [anon_sym_AT] = ACTIONS(1451),
    [anon_sym_not] = ACTIONS(1451),
    [anon_sym_and] = ACTIONS(1451),
    [anon_sym_or] = ACTIONS(1451),
    [anon_sym_PLUS] = ACTIONS(1451),
    [anon_sym_DASH] = ACTIONS(1451),
    [anon_sym_SLASH] = ACTIONS(1451),
    [anon_sym_SLASH_SLASH] = ACTIONS(1451),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [anon_sym_PIPE] = ACTIONS(1451),
    [anon_sym_CARET] = ACTIONS(1451),
    [anon_sym_AMP] = ACTIONS(1451),
    [anon_sym_LT_LT] = ACTIONS(1451),
    [anon_sym_GT_GT] = ACTIONS(1451),
    [anon_sym_GT_GT_GT] = ACTIONS(1451),
    [anon_sym_TILDE] = ACTIONS(1451),
    [anon_sym_typeof] = ACTIONS(1451),
    [anon_sym_void] = ACTIONS(1451),
    [anon_sym_LT] = ACTIONS(1451),
    [anon_sym_LT_EQ] = ACTIONS(1451),
    [anon_sym_EQ_EQ] = ACTIONS(1451),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1451),
    [anon_sym_BANG_EQ] = ACTIONS(1451),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1451),
    [anon_sym_GT_EQ] = ACTIONS(1451),
    [anon_sym_GT] = ACTIONS(1451),
    [anon_sym_is] = ACTIONS(1451),
    [anon_sym_instanceof] = ACTIONS(1451),
    [anon_sym_new] = ACTIONS(1451),
    [anon_sym_QMARK] = ACTIONS(1451),
    [anon_sym_await] = ACTIONS(1451),
    [anon_sym_LBRACE] = ACTIONS(1451),
    [sym_this] = ACTIONS(1451),
    [sym_true] = ACTIONS(1451),
    [sym_false] = ACTIONS(1451),
    [sym_none] = ACTIONS(1451),
    [sym_string] = ACTIONS(1451),
    [sym_verbatim] = ACTIONS(1451),
    [sym_number] = ACTIONS(1451),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1449),
    [sym_regex] = ACTIONS(1449),
    [sym_fstring_start] = ACTIONS(1449),
  },
  [STATE(670)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1297),
    [anon_sym_SEMI] = ACTIONS(1297),
    [anon_sym_return] = ACTIONS(1297),
    [anon_sym_async] = ACTIONS(1297),
    [anon_sym_def] = ACTIONS(1297),
    [anon_sym_LPAREN] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1297),
    [anon_sym_del] = ACTIONS(1297),
    [anon_sym_raise] = ACTIONS(1297),
    [sym_pass_statement] = ACTIONS(1297),
    [sym_break_statement] = ACTIONS(1297),
    [sym_continue_statement] = ACTIONS(1297),
    [sym_debugger_statement] = ACTIONS(1297),
    [anon_sym_assert] = ACTIONS(1297),
    [anon_sym_global] = ACTIONS(1297),
    [anon_sym_nonlocal] = ACTIONS(1297),
    [anon_sym_yield] = ACTIONS(1297),
    [anon_sym_from] = ACTIONS(1297),
    [anon_sym_import] = ACTIONS(1297),
    [anon_sym_DOT] = ACTIONS(1297),
    [anon_sym_if] = ACTIONS(1297),
    [anon_sym_for] = ACTIONS(1297),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1297),
    [anon_sym_do] = ACTIONS(1297),
    [anon_sym_try] = ACTIONS(1297),
    [anon_sym_with] = ACTIONS(1297),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1297),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1425),
    [anon_sym_or] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1297),
    [anon_sym_typeof] = ACTIONS(1297),
    [anon_sym_void] = ACTIONS(1297),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1297),
    [anon_sym_QMARK] = ACTIONS(1297),
    [anon_sym_await] = ACTIONS(1297),
    [anon_sym_LBRACE] = ACTIONS(1297),
    [sym_this] = ACTIONS(1297),
    [sym_true] = ACTIONS(1297),
    [sym_false] = ACTIONS(1297),
    [sym_none] = ACTIONS(1297),
    [sym_string] = ACTIONS(1297),
    [sym_verbatim] = ACTIONS(1297),
    [sym_number] = ACTIONS(1297),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1295),
    [sym_regex] = ACTIONS(1295),
    [sym_fstring_start] = ACTIONS(1295),
  },
  [STATE(671)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1274),
    [anon_sym_SEMI] = ACTIONS(1274),
    [anon_sym_return] = ACTIONS(1274),
    [anon_sym_async] = ACTIONS(1274),
    [anon_sym_def] = ACTIONS(1274),
    [anon_sym_LPAREN] = ACTIONS(1274),
    [anon_sym_LBRACK] = ACTIONS(1274),
    [anon_sym_del] = ACTIONS(1274),
    [anon_sym_raise] = ACTIONS(1274),
    [sym_pass_statement] = ACTIONS(1274),
    [sym_break_statement] = ACTIONS(1274),
    [sym_continue_statement] = ACTIONS(1274),
    [sym_debugger_statement] = ACTIONS(1274),
    [anon_sym_assert] = ACTIONS(1274),
    [anon_sym_global] = ACTIONS(1274),
    [anon_sym_nonlocal] = ACTIONS(1274),
    [anon_sym_yield] = ACTIONS(1274),
    [anon_sym_from] = ACTIONS(1274),
    [anon_sym_import] = ACTIONS(1274),
    [anon_sym_DOT] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(1600),
    [anon_sym_for] = ACTIONS(1274),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1274),
    [anon_sym_do] = ACTIONS(1274),
    [anon_sym_try] = ACTIONS(1274),
    [anon_sym_with] = ACTIONS(1274),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1274),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1425),
    [anon_sym_or] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1274),
    [anon_sym_typeof] = ACTIONS(1274),
    [anon_sym_void] = ACTIONS(1274),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1274),
    [anon_sym_QMARK] = ACTIONS(1274),
    [anon_sym_await] = ACTIONS(1274),
    [anon_sym_LBRACE] = ACTIONS(1274),
    [sym_this] = ACTIONS(1274),
    [sym_true] = ACTIONS(1274),
    [sym_false] = ACTIONS(1274),
    [sym_none] = ACTIONS(1274),
    [sym_string] = ACTIONS(1274),
    [sym_verbatim] = ACTIONS(1274),
    [sym_number] = ACTIONS(1274),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1272),
    [sym_regex] = ACTIONS(1272),
    [sym_fstring_start] = ACTIONS(1272),
  },
  [STATE(672)] = {
    [sym_argument_list] = STATE(564),
    [ts_builtin_sym_end] = ACTIONS(1198),
    [sym_identifier] = ACTIONS(1200),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym_return] = ACTIONS(1200),
    [anon_sym_async] = ACTIONS(1200),
    [anon_sym_def] = ACTIONS(1200),
    [anon_sym_LPAREN] = ACTIONS(1202),
    [anon_sym_LBRACK] = ACTIONS(1204),
    [anon_sym_del] = ACTIONS(1200),
    [anon_sym_raise] = ACTIONS(1200),
    [sym_pass_statement] = ACTIONS(1200),
    [sym_break_statement] = ACTIONS(1200),
    [sym_continue_statement] = ACTIONS(1200),
    [sym_debugger_statement] = ACTIONS(1200),
    [anon_sym_assert] = ACTIONS(1200),
    [anon_sym_global] = ACTIONS(1200),
    [anon_sym_nonlocal] = ACTIONS(1200),
    [anon_sym_yield] = ACTIONS(1200),
    [anon_sym_from] = ACTIONS(1200),
    [anon_sym_import] = ACTIONS(1200),
    [anon_sym_DOT] = ACTIONS(1206),
    [anon_sym_if] = ACTIONS(1200),
    [anon_sym_for] = ACTIONS(1200),
    [anon_sym_in] = ACTIONS(1200),
    [anon_sym_while] = ACTIONS(1200),
    [anon_sym_do] = ACTIONS(1200),
    [anon_sym_try] = ACTIONS(1200),
    [anon_sym_with] = ACTIONS(1200),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_STAR_STAR] = ACTIONS(1200),
    [anon_sym_class] = ACTIONS(1200),
    [anon_sym_AT] = ACTIONS(1200),
    [anon_sym_not] = ACTIONS(1200),
    [anon_sym_and] = ACTIONS(1200),
    [anon_sym_or] = ACTIONS(1200),
    [anon_sym_PLUS] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [anon_sym_SLASH] = ACTIONS(1200),
    [anon_sym_SLASH_SLASH] = ACTIONS(1200),
    [anon_sym_PERCENT] = ACTIONS(1200),
    [anon_sym_PIPE] = ACTIONS(1200),
    [anon_sym_CARET] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_LT_LT] = ACTIONS(1200),
    [anon_sym_GT_GT] = ACTIONS(1200),
    [anon_sym_GT_GT_GT] = ACTIONS(1200),
    [anon_sym_TILDE] = ACTIONS(1200),
    [anon_sym_typeof] = ACTIONS(1200),
    [anon_sym_void] = ACTIONS(1200),
    [anon_sym_LT] = ACTIONS(1200),
    [anon_sym_LT_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ] = ACTIONS(1200),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ] = ACTIONS(1200),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1200),
    [anon_sym_GT_EQ] = ACTIONS(1200),
    [anon_sym_GT] = ACTIONS(1200),
    [anon_sym_is] = ACTIONS(1200),
    [anon_sym_instanceof] = ACTIONS(1200),
    [anon_sym_new] = ACTIONS(1200),
    [anon_sym_QMARK] = ACTIONS(1613),
    [anon_sym_await] = ACTIONS(1200),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [sym_this] = ACTIONS(1200),
    [sym_true] = ACTIONS(1200),
    [sym_false] = ACTIONS(1200),
    [sym_none] = ACTIONS(1200),
    [sym_string] = ACTIONS(1200),
    [sym_verbatim] = ACTIONS(1200),
    [sym_number] = ACTIONS(1200),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1198),
    [sym_fstring_start] = ACTIONS(1198),
  },
  [STATE(673)] = {
    [sym_identifier] = ACTIONS(1464),
    [anon_sym_SEMI] = ACTIONS(1464),
    [anon_sym_return] = ACTIONS(1464),
    [anon_sym_async] = ACTIONS(1464),
    [anon_sym_def] = ACTIONS(1464),
    [anon_sym_COMMA] = ACTIONS(1464),
    [anon_sym_LPAREN] = ACTIONS(1464),
    [anon_sym_LBRACK] = ACTIONS(1464),
    [anon_sym_del] = ACTIONS(1464),
    [anon_sym_raise] = ACTIONS(1464),
    [sym_pass_statement] = ACTIONS(1464),
    [sym_break_statement] = ACTIONS(1464),
    [sym_continue_statement] = ACTIONS(1464),
    [sym_debugger_statement] = ACTIONS(1464),
    [anon_sym_assert] = ACTIONS(1464),
    [anon_sym_global] = ACTIONS(1464),
    [anon_sym_nonlocal] = ACTIONS(1464),
    [anon_sym_yield] = ACTIONS(1464),
    [anon_sym_from] = ACTIONS(1464),
    [anon_sym_import] = ACTIONS(1464),
    [anon_sym_DOT] = ACTIONS(1464),
    [anon_sym_if] = ACTIONS(1464),
    [anon_sym_for] = ACTIONS(1464),
    [anon_sym_in] = ACTIONS(1464),
    [anon_sym_while] = ACTIONS(1464),
    [anon_sym_do] = ACTIONS(1464),
    [anon_sym_try] = ACTIONS(1464),
    [anon_sym_with] = ACTIONS(1464),
    [anon_sym_STAR] = ACTIONS(1464),
    [anon_sym_STAR_STAR] = ACTIONS(1464),
    [anon_sym_class] = ACTIONS(1464),
    [anon_sym_AT] = ACTIONS(1464),
    [anon_sym_not] = ACTIONS(1464),
    [anon_sym_and] = ACTIONS(1464),
    [anon_sym_or] = ACTIONS(1464),
    [anon_sym_PLUS] = ACTIONS(1464),
    [anon_sym_DASH] = ACTIONS(1464),
    [anon_sym_SLASH] = ACTIONS(1464),
    [anon_sym_SLASH_SLASH] = ACTIONS(1464),
    [anon_sym_PERCENT] = ACTIONS(1464),
    [anon_sym_PIPE] = ACTIONS(1464),
    [anon_sym_CARET] = ACTIONS(1464),
    [anon_sym_AMP] = ACTIONS(1464),
    [anon_sym_LT_LT] = ACTIONS(1464),
    [anon_sym_GT_GT] = ACTIONS(1464),
    [anon_sym_GT_GT_GT] = ACTIONS(1464),
    [anon_sym_TILDE] = ACTIONS(1464),
    [anon_sym_typeof] = ACTIONS(1464),
    [anon_sym_void] = ACTIONS(1464),
    [anon_sym_LT] = ACTIONS(1464),
    [anon_sym_LT_EQ] = ACTIONS(1464),
    [anon_sym_EQ_EQ] = ACTIONS(1464),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1464),
    [anon_sym_BANG_EQ] = ACTIONS(1464),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1464),
    [anon_sym_GT_EQ] = ACTIONS(1464),
    [anon_sym_GT] = ACTIONS(1464),
    [anon_sym_is] = ACTIONS(1464),
    [anon_sym_instanceof] = ACTIONS(1464),
    [anon_sym_new] = ACTIONS(1464),
    [anon_sym_QMARK] = ACTIONS(1464),
    [anon_sym_await] = ACTIONS(1464),
    [anon_sym_LBRACE] = ACTIONS(1464),
    [sym_this] = ACTIONS(1464),
    [sym_true] = ACTIONS(1464),
    [sym_false] = ACTIONS(1464),
    [sym_none] = ACTIONS(1464),
    [sym_string] = ACTIONS(1464),
    [sym_verbatim] = ACTIONS(1464),
    [sym_number] = ACTIONS(1464),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1462),
    [sym_regex] = ACTIONS(1462),
    [sym_fstring_start] = ACTIONS(1462),
  },
  [STATE(674)] = {
    [ts_builtin_sym_end] = ACTIONS(931),
    [sym_identifier] = ACTIONS(933),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym_return] = ACTIONS(933),
    [anon_sym_async] = ACTIONS(933),
    [anon_sym_def] = ACTIONS(933),
    [anon_sym_COMMA] = ACTIONS(933),
    [anon_sym_LPAREN] = ACTIONS(933),
    [anon_sym_LBRACK] = ACTIONS(933),
    [anon_sym_del] = ACTIONS(933),
    [anon_sym_raise] = ACTIONS(933),
    [sym_pass_statement] = ACTIONS(933),
    [sym_break_statement] = ACTIONS(933),
    [sym_continue_statement] = ACTIONS(933),
    [sym_debugger_statement] = ACTIONS(933),
    [anon_sym_assert] = ACTIONS(933),
    [anon_sym_global] = ACTIONS(933),
    [anon_sym_nonlocal] = ACTIONS(933),
    [anon_sym_yield] = ACTIONS(933),
    [anon_sym_from] = ACTIONS(933),
    [anon_sym_import] = ACTIONS(933),
    [anon_sym_DOT] = ACTIONS(933),
    [anon_sym_if] = ACTIONS(933),
    [anon_sym_for] = ACTIONS(933),
    [anon_sym_in] = ACTIONS(933),
    [anon_sym_while] = ACTIONS(933),
    [anon_sym_do] = ACTIONS(933),
    [anon_sym_try] = ACTIONS(933),
    [anon_sym_with] = ACTIONS(933),
    [anon_sym_STAR] = ACTIONS(933),
    [anon_sym_STAR_STAR] = ACTIONS(933),
    [anon_sym_class] = ACTIONS(933),
    [anon_sym_AT] = ACTIONS(933),
    [anon_sym_not] = ACTIONS(933),
    [anon_sym_and] = ACTIONS(933),
    [anon_sym_or] = ACTIONS(933),
    [anon_sym_PLUS] = ACTIONS(933),
    [anon_sym_DASH] = ACTIONS(933),
    [anon_sym_SLASH] = ACTIONS(933),
    [anon_sym_SLASH_SLASH] = ACTIONS(933),
    [anon_sym_PERCENT] = ACTIONS(933),
    [anon_sym_PIPE] = ACTIONS(933),
    [anon_sym_CARET] = ACTIONS(933),
    [anon_sym_AMP] = ACTIONS(933),
    [anon_sym_LT_LT] = ACTIONS(933),
    [anon_sym_GT_GT] = ACTIONS(933),
    [anon_sym_GT_GT_GT] = ACTIONS(933),
    [anon_sym_TILDE] = ACTIONS(933),
    [anon_sym_typeof] = ACTIONS(933),
    [anon_sym_void] = ACTIONS(933),
    [anon_sym_LT] = ACTIONS(933),
    [anon_sym_LT_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(933),
    [anon_sym_GT_EQ] = ACTIONS(933),
    [anon_sym_GT] = ACTIONS(933),
    [anon_sym_is] = ACTIONS(933),
    [anon_sym_instanceof] = ACTIONS(933),
    [anon_sym_new] = ACTIONS(933),
    [anon_sym_QMARK] = ACTIONS(933),
    [anon_sym_await] = ACTIONS(933),
    [anon_sym_LBRACE] = ACTIONS(933),
    [sym_this] = ACTIONS(933),
    [sym_true] = ACTIONS(933),
    [sym_false] = ACTIONS(933),
    [sym_none] = ACTIONS(933),
    [sym_string] = ACTIONS(933),
    [sym_verbatim] = ACTIONS(933),
    [sym_number] = ACTIONS(933),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(931),
    [sym_fstring_start] = ACTIONS(931),
  },
  [STATE(675)] = {
    [ts_builtin_sym_end] = ACTIONS(1604),
    [sym_identifier] = ACTIONS(1602),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_return] = ACTIONS(1602),
    [anon_sym_async] = ACTIONS(1602),
    [anon_sym_def] = ACTIONS(1602),
    [anon_sym_COMMA] = ACTIONS(1602),
    [anon_sym_LPAREN] = ACTIONS(1602),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [anon_sym_del] = ACTIONS(1602),
    [anon_sym_raise] = ACTIONS(1602),
    [sym_pass_statement] = ACTIONS(1602),
    [sym_break_statement] = ACTIONS(1602),
    [sym_continue_statement] = ACTIONS(1602),
    [sym_debugger_statement] = ACTIONS(1602),
    [anon_sym_assert] = ACTIONS(1602),
    [anon_sym_global] = ACTIONS(1602),
    [anon_sym_nonlocal] = ACTIONS(1602),
    [anon_sym_yield] = ACTIONS(1602),
    [anon_sym_from] = ACTIONS(1602),
    [anon_sym_import] = ACTIONS(1602),
    [anon_sym_DOT] = ACTIONS(1602),
    [anon_sym_if] = ACTIONS(1602),
    [anon_sym_for] = ACTIONS(1602),
    [anon_sym_in] = ACTIONS(1602),
    [anon_sym_while] = ACTIONS(1602),
    [anon_sym_do] = ACTIONS(1602),
    [anon_sym_try] = ACTIONS(1602),
    [anon_sym_with] = ACTIONS(1602),
    [anon_sym_STAR] = ACTIONS(1602),
    [anon_sym_STAR_STAR] = ACTIONS(1602),
    [anon_sym_class] = ACTIONS(1602),
    [anon_sym_AT] = ACTIONS(1602),
    [anon_sym_not] = ACTIONS(1602),
    [anon_sym_and] = ACTIONS(1602),
    [anon_sym_or] = ACTIONS(1602),
    [anon_sym_PLUS] = ACTIONS(1602),
    [anon_sym_DASH] = ACTIONS(1602),
    [anon_sym_SLASH] = ACTIONS(1602),
    [anon_sym_SLASH_SLASH] = ACTIONS(1602),
    [anon_sym_PERCENT] = ACTIONS(1602),
    [anon_sym_PIPE] = ACTIONS(1602),
    [anon_sym_CARET] = ACTIONS(1602),
    [anon_sym_AMP] = ACTIONS(1602),
    [anon_sym_LT_LT] = ACTIONS(1602),
    [anon_sym_GT_GT] = ACTIONS(1602),
    [anon_sym_GT_GT_GT] = ACTIONS(1602),
    [anon_sym_TILDE] = ACTIONS(1602),
    [anon_sym_typeof] = ACTIONS(1602),
    [anon_sym_void] = ACTIONS(1602),
    [anon_sym_LT] = ACTIONS(1602),
    [anon_sym_LT_EQ] = ACTIONS(1602),
    [anon_sym_EQ_EQ] = ACTIONS(1602),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1602),
    [anon_sym_BANG_EQ] = ACTIONS(1602),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1602),
    [anon_sym_GT_EQ] = ACTIONS(1602),
    [anon_sym_GT] = ACTIONS(1602),
    [anon_sym_is] = ACTIONS(1602),
    [anon_sym_instanceof] = ACTIONS(1602),
    [anon_sym_new] = ACTIONS(1602),
    [anon_sym_QMARK] = ACTIONS(1602),
    [anon_sym_await] = ACTIONS(1602),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [sym_this] = ACTIONS(1602),
    [sym_true] = ACTIONS(1602),
    [sym_false] = ACTIONS(1602),
    [sym_none] = ACTIONS(1602),
    [sym_string] = ACTIONS(1602),
    [sym_verbatim] = ACTIONS(1602),
    [sym_number] = ACTIONS(1602),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1604),
    [sym_fstring_start] = ACTIONS(1604),
  },
  [STATE(676)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(647),
    [sym_identifier] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1359),
    [anon_sym_return] = ACTIONS(1359),
    [anon_sym_async] = ACTIONS(1359),
    [anon_sym_def] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1359),
    [anon_sym_LBRACK] = ACTIONS(1359),
    [anon_sym_del] = ACTIONS(1359),
    [anon_sym_raise] = ACTIONS(1359),
    [sym_pass_statement] = ACTIONS(1359),
    [sym_break_statement] = ACTIONS(1359),
    [sym_continue_statement] = ACTIONS(1359),
    [sym_debugger_statement] = ACTIONS(1359),
    [anon_sym_assert] = ACTIONS(1359),
    [anon_sym_global] = ACTIONS(1359),
    [anon_sym_nonlocal] = ACTIONS(1359),
    [anon_sym_yield] = ACTIONS(1359),
    [anon_sym_from] = ACTIONS(1359),
    [anon_sym_import] = ACTIONS(1359),
    [anon_sym_DOT] = ACTIONS(1359),
    [anon_sym_if] = ACTIONS(1359),
    [anon_sym_for] = ACTIONS(1359),
    [anon_sym_in] = ACTIONS(1359),
    [anon_sym_while] = ACTIONS(1359),
    [anon_sym_do] = ACTIONS(1359),
    [anon_sym_try] = ACTIONS(1359),
    [anon_sym_with] = ACTIONS(1359),
    [anon_sym_STAR] = ACTIONS(1359),
    [anon_sym_STAR_STAR] = ACTIONS(1359),
    [anon_sym_class] = ACTIONS(1359),
    [anon_sym_AT] = ACTIONS(1359),
    [anon_sym_not] = ACTIONS(1359),
    [anon_sym_and] = ACTIONS(1359),
    [anon_sym_or] = ACTIONS(1359),
    [anon_sym_PLUS] = ACTIONS(1359),
    [anon_sym_DASH] = ACTIONS(1359),
    [anon_sym_SLASH] = ACTIONS(1359),
    [anon_sym_SLASH_SLASH] = ACTIONS(1359),
    [anon_sym_PERCENT] = ACTIONS(1359),
    [anon_sym_PIPE] = ACTIONS(1359),
    [anon_sym_CARET] = ACTIONS(1359),
    [anon_sym_AMP] = ACTIONS(1359),
    [anon_sym_LT_LT] = ACTIONS(1359),
    [anon_sym_GT_GT] = ACTIONS(1359),
    [anon_sym_GT_GT_GT] = ACTIONS(1359),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_typeof] = ACTIONS(1359),
    [anon_sym_void] = ACTIONS(1359),
    [anon_sym_LT] = ACTIONS(1359),
    [anon_sym_LT_EQ] = ACTIONS(1359),
    [anon_sym_EQ_EQ] = ACTIONS(1359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1359),
    [anon_sym_BANG_EQ] = ACTIONS(1359),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1359),
    [anon_sym_GT_EQ] = ACTIONS(1359),
    [anon_sym_GT] = ACTIONS(1359),
    [anon_sym_is] = ACTIONS(1359),
    [anon_sym_instanceof] = ACTIONS(1359),
    [anon_sym_new] = ACTIONS(1359),
    [anon_sym_QMARK] = ACTIONS(1359),
    [anon_sym_await] = ACTIONS(1359),
    [anon_sym_LBRACE] = ACTIONS(1359),
    [sym_this] = ACTIONS(1359),
    [sym_true] = ACTIONS(1359),
    [sym_false] = ACTIONS(1359),
    [sym_none] = ACTIONS(1359),
    [sym_string] = ACTIONS(1359),
    [sym_verbatim] = ACTIONS(1359),
    [sym_number] = ACTIONS(1359),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1357),
    [sym_regex] = ACTIONS(1357),
    [sym_fstring_start] = ACTIONS(1357),
  },
  [STATE(677)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(678)] = {
    [sym_identifier] = ACTIONS(1407),
    [anon_sym_SEMI] = ACTIONS(1407),
    [anon_sym_return] = ACTIONS(1407),
    [anon_sym_async] = ACTIONS(1407),
    [anon_sym_def] = ACTIONS(1407),
    [anon_sym_COMMA] = ACTIONS(1407),
    [anon_sym_LPAREN] = ACTIONS(1407),
    [anon_sym_LBRACK] = ACTIONS(1407),
    [anon_sym_del] = ACTIONS(1407),
    [anon_sym_raise] = ACTIONS(1407),
    [sym_pass_statement] = ACTIONS(1407),
    [sym_break_statement] = ACTIONS(1407),
    [sym_continue_statement] = ACTIONS(1407),
    [sym_debugger_statement] = ACTIONS(1407),
    [anon_sym_assert] = ACTIONS(1407),
    [anon_sym_global] = ACTIONS(1407),
    [anon_sym_nonlocal] = ACTIONS(1407),
    [anon_sym_yield] = ACTIONS(1407),
    [anon_sym_from] = ACTIONS(1407),
    [anon_sym_import] = ACTIONS(1407),
    [anon_sym_DOT] = ACTIONS(1407),
    [anon_sym_if] = ACTIONS(1407),
    [anon_sym_for] = ACTIONS(1407),
    [anon_sym_in] = ACTIONS(1407),
    [anon_sym_while] = ACTIONS(1407),
    [anon_sym_do] = ACTIONS(1407),
    [anon_sym_try] = ACTIONS(1407),
    [anon_sym_with] = ACTIONS(1407),
    [anon_sym_STAR] = ACTIONS(1407),
    [anon_sym_STAR_STAR] = ACTIONS(1407),
    [anon_sym_class] = ACTIONS(1407),
    [anon_sym_AT] = ACTIONS(1407),
    [anon_sym_not] = ACTIONS(1407),
    [anon_sym_and] = ACTIONS(1407),
    [anon_sym_or] = ACTIONS(1407),
    [anon_sym_PLUS] = ACTIONS(1407),
    [anon_sym_DASH] = ACTIONS(1407),
    [anon_sym_SLASH] = ACTIONS(1407),
    [anon_sym_SLASH_SLASH] = ACTIONS(1407),
    [anon_sym_PERCENT] = ACTIONS(1407),
    [anon_sym_PIPE] = ACTIONS(1407),
    [anon_sym_CARET] = ACTIONS(1407),
    [anon_sym_AMP] = ACTIONS(1407),
    [anon_sym_LT_LT] = ACTIONS(1407),
    [anon_sym_GT_GT] = ACTIONS(1407),
    [anon_sym_GT_GT_GT] = ACTIONS(1407),
    [anon_sym_TILDE] = ACTIONS(1407),
    [anon_sym_typeof] = ACTIONS(1407),
    [anon_sym_void] = ACTIONS(1407),
    [anon_sym_LT] = ACTIONS(1407),
    [anon_sym_LT_EQ] = ACTIONS(1407),
    [anon_sym_EQ_EQ] = ACTIONS(1407),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1407),
    [anon_sym_BANG_EQ] = ACTIONS(1407),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1407),
    [anon_sym_GT_EQ] = ACTIONS(1407),
    [anon_sym_GT] = ACTIONS(1407),
    [anon_sym_is] = ACTIONS(1407),
    [anon_sym_instanceof] = ACTIONS(1407),
    [anon_sym_new] = ACTIONS(1407),
    [anon_sym_QMARK] = ACTIONS(1407),
    [anon_sym_await] = ACTIONS(1407),
    [anon_sym_LBRACE] = ACTIONS(1407),
    [sym_this] = ACTIONS(1407),
    [sym_true] = ACTIONS(1407),
    [sym_false] = ACTIONS(1407),
    [sym_none] = ACTIONS(1407),
    [sym_string] = ACTIONS(1407),
    [sym_verbatim] = ACTIONS(1407),
    [sym_number] = ACTIONS(1407),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1405),
    [sym_regex] = ACTIONS(1405),
    [sym_fstring_start] = ACTIONS(1405),
  },
  [STATE(679)] = {
    [sym_identifier] = ACTIONS(1544),
    [anon_sym_SEMI] = ACTIONS(1544),
    [anon_sym_return] = ACTIONS(1544),
    [anon_sym_async] = ACTIONS(1544),
    [anon_sym_def] = ACTIONS(1544),
    [anon_sym_COMMA] = ACTIONS(1544),
    [anon_sym_LPAREN] = ACTIONS(1544),
    [anon_sym_LBRACK] = ACTIONS(1544),
    [anon_sym_del] = ACTIONS(1544),
    [anon_sym_raise] = ACTIONS(1544),
    [sym_pass_statement] = ACTIONS(1544),
    [sym_break_statement] = ACTIONS(1544),
    [sym_continue_statement] = ACTIONS(1544),
    [sym_debugger_statement] = ACTIONS(1544),
    [anon_sym_assert] = ACTIONS(1544),
    [anon_sym_global] = ACTIONS(1544),
    [anon_sym_nonlocal] = ACTIONS(1544),
    [anon_sym_yield] = ACTIONS(1544),
    [anon_sym_from] = ACTIONS(1544),
    [anon_sym_import] = ACTIONS(1544),
    [anon_sym_DOT] = ACTIONS(1544),
    [anon_sym_if] = ACTIONS(1544),
    [anon_sym_for] = ACTIONS(1544),
    [anon_sym_in] = ACTIONS(1544),
    [anon_sym_while] = ACTIONS(1544),
    [anon_sym_do] = ACTIONS(1544),
    [anon_sym_try] = ACTIONS(1544),
    [anon_sym_with] = ACTIONS(1544),
    [anon_sym_STAR] = ACTIONS(1544),
    [anon_sym_STAR_STAR] = ACTIONS(1544),
    [anon_sym_class] = ACTIONS(1544),
    [anon_sym_AT] = ACTIONS(1544),
    [anon_sym_not] = ACTIONS(1544),
    [anon_sym_and] = ACTIONS(1544),
    [anon_sym_or] = ACTIONS(1544),
    [anon_sym_PLUS] = ACTIONS(1544),
    [anon_sym_DASH] = ACTIONS(1544),
    [anon_sym_SLASH] = ACTIONS(1544),
    [anon_sym_SLASH_SLASH] = ACTIONS(1544),
    [anon_sym_PERCENT] = ACTIONS(1544),
    [anon_sym_PIPE] = ACTIONS(1544),
    [anon_sym_CARET] = ACTIONS(1544),
    [anon_sym_AMP] = ACTIONS(1544),
    [anon_sym_LT_LT] = ACTIONS(1544),
    [anon_sym_GT_GT] = ACTIONS(1544),
    [anon_sym_GT_GT_GT] = ACTIONS(1544),
    [anon_sym_TILDE] = ACTIONS(1544),
    [anon_sym_typeof] = ACTIONS(1544),
    [anon_sym_void] = ACTIONS(1544),
    [anon_sym_LT] = ACTIONS(1544),
    [anon_sym_LT_EQ] = ACTIONS(1544),
    [anon_sym_EQ_EQ] = ACTIONS(1544),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1544),
    [anon_sym_BANG_EQ] = ACTIONS(1544),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1544),
    [anon_sym_GT_EQ] = ACTIONS(1544),
    [anon_sym_GT] = ACTIONS(1544),
    [anon_sym_is] = ACTIONS(1544),
    [anon_sym_instanceof] = ACTIONS(1544),
    [anon_sym_new] = ACTIONS(1544),
    [anon_sym_QMARK] = ACTIONS(1544),
    [anon_sym_await] = ACTIONS(1544),
    [anon_sym_LBRACE] = ACTIONS(1544),
    [sym_this] = ACTIONS(1544),
    [sym_true] = ACTIONS(1544),
    [sym_false] = ACTIONS(1544),
    [sym_none] = ACTIONS(1544),
    [sym_string] = ACTIONS(1544),
    [sym_verbatim] = ACTIONS(1544),
    [sym_number] = ACTIONS(1544),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1542),
    [sym_regex] = ACTIONS(1542),
    [sym_fstring_start] = ACTIONS(1542),
  },
  [STATE(680)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1301),
    [anon_sym_SEMI] = ACTIONS(1301),
    [anon_sym_return] = ACTIONS(1301),
    [anon_sym_async] = ACTIONS(1301),
    [anon_sym_def] = ACTIONS(1301),
    [anon_sym_LPAREN] = ACTIONS(1301),
    [anon_sym_LBRACK] = ACTIONS(1301),
    [anon_sym_del] = ACTIONS(1301),
    [anon_sym_raise] = ACTIONS(1301),
    [sym_pass_statement] = ACTIONS(1301),
    [sym_break_statement] = ACTIONS(1301),
    [sym_continue_statement] = ACTIONS(1301),
    [sym_debugger_statement] = ACTIONS(1301),
    [anon_sym_assert] = ACTIONS(1301),
    [anon_sym_global] = ACTIONS(1301),
    [anon_sym_nonlocal] = ACTIONS(1301),
    [anon_sym_yield] = ACTIONS(1301),
    [anon_sym_from] = ACTIONS(1301),
    [anon_sym_import] = ACTIONS(1301),
    [anon_sym_DOT] = ACTIONS(1301),
    [anon_sym_if] = ACTIONS(1301),
    [anon_sym_for] = ACTIONS(1301),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1301),
    [anon_sym_do] = ACTIONS(1301),
    [anon_sym_try] = ACTIONS(1301),
    [anon_sym_with] = ACTIONS(1301),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1301),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1425),
    [anon_sym_or] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1301),
    [anon_sym_typeof] = ACTIONS(1301),
    [anon_sym_void] = ACTIONS(1301),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1301),
    [anon_sym_QMARK] = ACTIONS(1301),
    [anon_sym_await] = ACTIONS(1301),
    [anon_sym_LBRACE] = ACTIONS(1301),
    [sym_this] = ACTIONS(1301),
    [sym_true] = ACTIONS(1301),
    [sym_false] = ACTIONS(1301),
    [sym_none] = ACTIONS(1301),
    [sym_string] = ACTIONS(1301),
    [sym_verbatim] = ACTIONS(1301),
    [sym_number] = ACTIONS(1301),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1299),
    [sym_regex] = ACTIONS(1299),
    [sym_fstring_start] = ACTIONS(1299),
  },
  [STATE(681)] = {
    [sym_identifier] = ACTIONS(1592),
    [anon_sym_SEMI] = ACTIONS(1592),
    [anon_sym_return] = ACTIONS(1592),
    [anon_sym_async] = ACTIONS(1592),
    [anon_sym_def] = ACTIONS(1592),
    [anon_sym_COMMA] = ACTIONS(1592),
    [anon_sym_LPAREN] = ACTIONS(1592),
    [anon_sym_LBRACK] = ACTIONS(1592),
    [anon_sym_del] = ACTIONS(1592),
    [anon_sym_raise] = ACTIONS(1592),
    [sym_pass_statement] = ACTIONS(1592),
    [sym_break_statement] = ACTIONS(1592),
    [sym_continue_statement] = ACTIONS(1592),
    [sym_debugger_statement] = ACTIONS(1592),
    [anon_sym_assert] = ACTIONS(1592),
    [anon_sym_global] = ACTIONS(1592),
    [anon_sym_nonlocal] = ACTIONS(1592),
    [anon_sym_yield] = ACTIONS(1592),
    [anon_sym_from] = ACTIONS(1592),
    [anon_sym_import] = ACTIONS(1592),
    [anon_sym_DOT] = ACTIONS(1592),
    [anon_sym_if] = ACTIONS(1592),
    [anon_sym_for] = ACTIONS(1592),
    [anon_sym_in] = ACTIONS(1592),
    [anon_sym_while] = ACTIONS(1592),
    [anon_sym_do] = ACTIONS(1592),
    [anon_sym_try] = ACTIONS(1592),
    [anon_sym_with] = ACTIONS(1592),
    [anon_sym_STAR] = ACTIONS(1592),
    [anon_sym_STAR_STAR] = ACTIONS(1592),
    [anon_sym_class] = ACTIONS(1592),
    [anon_sym_AT] = ACTIONS(1592),
    [anon_sym_not] = ACTIONS(1592),
    [anon_sym_and] = ACTIONS(1592),
    [anon_sym_or] = ACTIONS(1592),
    [anon_sym_PLUS] = ACTIONS(1592),
    [anon_sym_DASH] = ACTIONS(1592),
    [anon_sym_SLASH] = ACTIONS(1592),
    [anon_sym_SLASH_SLASH] = ACTIONS(1592),
    [anon_sym_PERCENT] = ACTIONS(1592),
    [anon_sym_PIPE] = ACTIONS(1592),
    [anon_sym_CARET] = ACTIONS(1592),
    [anon_sym_AMP] = ACTIONS(1592),
    [anon_sym_LT_LT] = ACTIONS(1592),
    [anon_sym_GT_GT] = ACTIONS(1592),
    [anon_sym_GT_GT_GT] = ACTIONS(1592),
    [anon_sym_TILDE] = ACTIONS(1592),
    [anon_sym_typeof] = ACTIONS(1592),
    [anon_sym_void] = ACTIONS(1592),
    [anon_sym_LT] = ACTIONS(1592),
    [anon_sym_LT_EQ] = ACTIONS(1592),
    [anon_sym_EQ_EQ] = ACTIONS(1592),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1592),
    [anon_sym_BANG_EQ] = ACTIONS(1592),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1592),
    [anon_sym_GT_EQ] = ACTIONS(1592),
    [anon_sym_GT] = ACTIONS(1592),
    [anon_sym_is] = ACTIONS(1592),
    [anon_sym_instanceof] = ACTIONS(1592),
    [anon_sym_new] = ACTIONS(1592),
    [anon_sym_QMARK] = ACTIONS(1592),
    [anon_sym_await] = ACTIONS(1592),
    [anon_sym_LBRACE] = ACTIONS(1592),
    [sym_this] = ACTIONS(1592),
    [sym_true] = ACTIONS(1592),
    [sym_false] = ACTIONS(1592),
    [sym_none] = ACTIONS(1592),
    [sym_string] = ACTIONS(1592),
    [sym_verbatim] = ACTIONS(1592),
    [sym_number] = ACTIONS(1592),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1590),
    [sym_regex] = ACTIONS(1590),
    [sym_fstring_start] = ACTIONS(1590),
  },
  [STATE(682)] = {
    [sym_identifier] = ACTIONS(1611),
    [anon_sym_SEMI] = ACTIONS(1611),
    [anon_sym_return] = ACTIONS(1611),
    [anon_sym_async] = ACTIONS(1611),
    [anon_sym_def] = ACTIONS(1611),
    [anon_sym_COMMA] = ACTIONS(1611),
    [anon_sym_LPAREN] = ACTIONS(1611),
    [anon_sym_LBRACK] = ACTIONS(1611),
    [anon_sym_del] = ACTIONS(1611),
    [anon_sym_raise] = ACTIONS(1611),
    [sym_pass_statement] = ACTIONS(1611),
    [sym_break_statement] = ACTIONS(1611),
    [sym_continue_statement] = ACTIONS(1611),
    [sym_debugger_statement] = ACTIONS(1611),
    [anon_sym_assert] = ACTIONS(1611),
    [anon_sym_global] = ACTIONS(1611),
    [anon_sym_nonlocal] = ACTIONS(1611),
    [anon_sym_yield] = ACTIONS(1611),
    [anon_sym_from] = ACTIONS(1611),
    [anon_sym_import] = ACTIONS(1611),
    [anon_sym_DOT] = ACTIONS(1611),
    [anon_sym_if] = ACTIONS(1611),
    [anon_sym_for] = ACTIONS(1611),
    [anon_sym_in] = ACTIONS(1611),
    [anon_sym_while] = ACTIONS(1611),
    [anon_sym_do] = ACTIONS(1611),
    [anon_sym_try] = ACTIONS(1611),
    [anon_sym_with] = ACTIONS(1611),
    [anon_sym_STAR] = ACTIONS(1611),
    [anon_sym_STAR_STAR] = ACTIONS(1611),
    [anon_sym_class] = ACTIONS(1611),
    [anon_sym_AT] = ACTIONS(1611),
    [anon_sym_not] = ACTIONS(1611),
    [anon_sym_and] = ACTIONS(1611),
    [anon_sym_or] = ACTIONS(1611),
    [anon_sym_PLUS] = ACTIONS(1611),
    [anon_sym_DASH] = ACTIONS(1611),
    [anon_sym_SLASH] = ACTIONS(1611),
    [anon_sym_SLASH_SLASH] = ACTIONS(1611),
    [anon_sym_PERCENT] = ACTIONS(1611),
    [anon_sym_PIPE] = ACTIONS(1611),
    [anon_sym_CARET] = ACTIONS(1611),
    [anon_sym_AMP] = ACTIONS(1611),
    [anon_sym_LT_LT] = ACTIONS(1611),
    [anon_sym_GT_GT] = ACTIONS(1611),
    [anon_sym_GT_GT_GT] = ACTIONS(1611),
    [anon_sym_TILDE] = ACTIONS(1611),
    [anon_sym_typeof] = ACTIONS(1611),
    [anon_sym_void] = ACTIONS(1611),
    [anon_sym_LT] = ACTIONS(1611),
    [anon_sym_LT_EQ] = ACTIONS(1611),
    [anon_sym_EQ_EQ] = ACTIONS(1611),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1611),
    [anon_sym_BANG_EQ] = ACTIONS(1611),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1611),
    [anon_sym_GT_EQ] = ACTIONS(1611),
    [anon_sym_GT] = ACTIONS(1611),
    [anon_sym_is] = ACTIONS(1611),
    [anon_sym_instanceof] = ACTIONS(1611),
    [anon_sym_new] = ACTIONS(1611),
    [anon_sym_QMARK] = ACTIONS(1611),
    [anon_sym_await] = ACTIONS(1611),
    [anon_sym_LBRACE] = ACTIONS(1611),
    [sym_this] = ACTIONS(1611),
    [sym_true] = ACTIONS(1611),
    [sym_false] = ACTIONS(1611),
    [sym_none] = ACTIONS(1611),
    [sym_string] = ACTIONS(1611),
    [sym_verbatim] = ACTIONS(1611),
    [sym_number] = ACTIONS(1611),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1609),
    [sym_regex] = ACTIONS(1609),
    [sym_fstring_start] = ACTIONS(1609),
  },
  [STATE(683)] = {
    [ts_builtin_sym_end] = ACTIONS(1152),
    [sym_identifier] = ACTIONS(1154),
    [anon_sym_SEMI] = ACTIONS(1154),
    [anon_sym_return] = ACTIONS(1154),
    [anon_sym_async] = ACTIONS(1154),
    [anon_sym_def] = ACTIONS(1154),
    [anon_sym_COMMA] = ACTIONS(1154),
    [anon_sym_LPAREN] = ACTIONS(1154),
    [anon_sym_LBRACK] = ACTIONS(1154),
    [anon_sym_del] = ACTIONS(1154),
    [anon_sym_raise] = ACTIONS(1154),
    [sym_pass_statement] = ACTIONS(1154),
    [sym_break_statement] = ACTIONS(1154),
    [sym_continue_statement] = ACTIONS(1154),
    [sym_debugger_statement] = ACTIONS(1154),
    [anon_sym_assert] = ACTIONS(1154),
    [anon_sym_global] = ACTIONS(1154),
    [anon_sym_nonlocal] = ACTIONS(1154),
    [anon_sym_yield] = ACTIONS(1154),
    [anon_sym_from] = ACTIONS(1154),
    [anon_sym_import] = ACTIONS(1154),
    [anon_sym_DOT] = ACTIONS(1154),
    [anon_sym_if] = ACTIONS(1154),
    [anon_sym_for] = ACTIONS(1154),
    [anon_sym_in] = ACTIONS(1154),
    [anon_sym_while] = ACTIONS(1154),
    [anon_sym_do] = ACTIONS(1154),
    [anon_sym_try] = ACTIONS(1154),
    [anon_sym_with] = ACTIONS(1154),
    [anon_sym_STAR] = ACTIONS(1154),
    [anon_sym_STAR_STAR] = ACTIONS(1154),
    [anon_sym_class] = ACTIONS(1154),
    [anon_sym_AT] = ACTIONS(1154),
    [anon_sym_not] = ACTIONS(1154),
    [anon_sym_and] = ACTIONS(1154),
    [anon_sym_or] = ACTIONS(1154),
    [anon_sym_PLUS] = ACTIONS(1154),
    [anon_sym_DASH] = ACTIONS(1154),
    [anon_sym_SLASH] = ACTIONS(1154),
    [anon_sym_SLASH_SLASH] = ACTIONS(1154),
    [anon_sym_PERCENT] = ACTIONS(1154),
    [anon_sym_PIPE] = ACTIONS(1154),
    [anon_sym_CARET] = ACTIONS(1154),
    [anon_sym_AMP] = ACTIONS(1154),
    [anon_sym_LT_LT] = ACTIONS(1154),
    [anon_sym_GT_GT] = ACTIONS(1154),
    [anon_sym_GT_GT_GT] = ACTIONS(1154),
    [anon_sym_TILDE] = ACTIONS(1154),
    [anon_sym_typeof] = ACTIONS(1154),
    [anon_sym_void] = ACTIONS(1154),
    [anon_sym_LT] = ACTIONS(1154),
    [anon_sym_LT_EQ] = ACTIONS(1154),
    [anon_sym_EQ_EQ] = ACTIONS(1154),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1154),
    [anon_sym_BANG_EQ] = ACTIONS(1154),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1154),
    [anon_sym_GT_EQ] = ACTIONS(1154),
    [anon_sym_GT] = ACTIONS(1154),
    [anon_sym_is] = ACTIONS(1154),
    [anon_sym_instanceof] = ACTIONS(1154),
    [anon_sym_new] = ACTIONS(1154),
    [anon_sym_QMARK] = ACTIONS(1154),
    [anon_sym_await] = ACTIONS(1154),
    [anon_sym_LBRACE] = ACTIONS(1154),
    [sym_this] = ACTIONS(1154),
    [sym_true] = ACTIONS(1154),
    [sym_false] = ACTIONS(1154),
    [sym_none] = ACTIONS(1154),
    [sym_string] = ACTIONS(1154),
    [sym_verbatim] = ACTIONS(1154),
    [sym_number] = ACTIONS(1154),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1152),
    [sym_fstring_start] = ACTIONS(1152),
  },
  [STATE(684)] = {
    [sym_identifier] = ACTIONS(933),
    [anon_sym_SEMI] = ACTIONS(933),
    [anon_sym_return] = ACTIONS(933),
    [anon_sym_async] = ACTIONS(933),
    [anon_sym_def] = ACTIONS(933),
    [anon_sym_COMMA] = ACTIONS(933),
    [anon_sym_LPAREN] = ACTIONS(933),
    [anon_sym_LBRACK] = ACTIONS(933),
    [anon_sym_del] = ACTIONS(933),
    [anon_sym_raise] = ACTIONS(933),
    [sym_pass_statement] = ACTIONS(933),
    [sym_break_statement] = ACTIONS(933),
    [sym_continue_statement] = ACTIONS(933),
    [sym_debugger_statement] = ACTIONS(933),
    [anon_sym_assert] = ACTIONS(933),
    [anon_sym_global] = ACTIONS(933),
    [anon_sym_nonlocal] = ACTIONS(933),
    [anon_sym_yield] = ACTIONS(933),
    [anon_sym_from] = ACTIONS(933),
    [anon_sym_import] = ACTIONS(933),
    [anon_sym_DOT] = ACTIONS(933),
    [anon_sym_if] = ACTIONS(933),
    [anon_sym_for] = ACTIONS(933),
    [anon_sym_in] = ACTIONS(933),
    [anon_sym_while] = ACTIONS(933),
    [anon_sym_do] = ACTIONS(933),
    [anon_sym_try] = ACTIONS(933),
    [anon_sym_with] = ACTIONS(933),
    [anon_sym_STAR] = ACTIONS(933),
    [anon_sym_STAR_STAR] = ACTIONS(933),
    [anon_sym_class] = ACTIONS(933),
    [anon_sym_AT] = ACTIONS(933),
    [anon_sym_not] = ACTIONS(933),
    [anon_sym_and] = ACTIONS(933),
    [anon_sym_or] = ACTIONS(933),
    [anon_sym_PLUS] = ACTIONS(933),
    [anon_sym_DASH] = ACTIONS(933),
    [anon_sym_SLASH] = ACTIONS(933),
    [anon_sym_SLASH_SLASH] = ACTIONS(933),
    [anon_sym_PERCENT] = ACTIONS(933),
    [anon_sym_PIPE] = ACTIONS(933),
    [anon_sym_CARET] = ACTIONS(933),
    [anon_sym_AMP] = ACTIONS(933),
    [anon_sym_LT_LT] = ACTIONS(933),
    [anon_sym_GT_GT] = ACTIONS(933),
    [anon_sym_GT_GT_GT] = ACTIONS(933),
    [anon_sym_TILDE] = ACTIONS(933),
    [anon_sym_typeof] = ACTIONS(933),
    [anon_sym_void] = ACTIONS(933),
    [anon_sym_LT] = ACTIONS(933),
    [anon_sym_LT_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ] = ACTIONS(933),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ] = ACTIONS(933),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(933),
    [anon_sym_GT_EQ] = ACTIONS(933),
    [anon_sym_GT] = ACTIONS(933),
    [anon_sym_is] = ACTIONS(933),
    [anon_sym_instanceof] = ACTIONS(933),
    [anon_sym_new] = ACTIONS(933),
    [anon_sym_QMARK] = ACTIONS(933),
    [anon_sym_await] = ACTIONS(933),
    [anon_sym_LBRACE] = ACTIONS(933),
    [sym_this] = ACTIONS(933),
    [sym_true] = ACTIONS(933),
    [sym_false] = ACTIONS(933),
    [sym_none] = ACTIONS(933),
    [sym_string] = ACTIONS(933),
    [sym_verbatim] = ACTIONS(933),
    [sym_number] = ACTIONS(933),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(931),
    [sym_regex] = ACTIONS(931),
    [sym_fstring_start] = ACTIONS(931),
  },
  [STATE(685)] = {
    [sym_identifier] = ACTIONS(1571),
    [anon_sym_SEMI] = ACTIONS(1571),
    [anon_sym_return] = ACTIONS(1571),
    [anon_sym_async] = ACTIONS(1571),
    [anon_sym_def] = ACTIONS(1571),
    [anon_sym_COMMA] = ACTIONS(1571),
    [anon_sym_LPAREN] = ACTIONS(1571),
    [anon_sym_LBRACK] = ACTIONS(1571),
    [anon_sym_del] = ACTIONS(1571),
    [anon_sym_raise] = ACTIONS(1571),
    [sym_pass_statement] = ACTIONS(1571),
    [sym_break_statement] = ACTIONS(1571),
    [sym_continue_statement] = ACTIONS(1571),
    [sym_debugger_statement] = ACTIONS(1571),
    [anon_sym_assert] = ACTIONS(1571),
    [anon_sym_global] = ACTIONS(1571),
    [anon_sym_nonlocal] = ACTIONS(1571),
    [anon_sym_yield] = ACTIONS(1571),
    [anon_sym_from] = ACTIONS(1571),
    [anon_sym_import] = ACTIONS(1571),
    [anon_sym_DOT] = ACTIONS(1571),
    [anon_sym_if] = ACTIONS(1571),
    [anon_sym_for] = ACTIONS(1571),
    [anon_sym_in] = ACTIONS(1571),
    [anon_sym_while] = ACTIONS(1571),
    [anon_sym_do] = ACTIONS(1571),
    [anon_sym_try] = ACTIONS(1571),
    [anon_sym_with] = ACTIONS(1571),
    [anon_sym_STAR] = ACTIONS(1571),
    [anon_sym_STAR_STAR] = ACTIONS(1571),
    [anon_sym_class] = ACTIONS(1571),
    [anon_sym_AT] = ACTIONS(1571),
    [anon_sym_not] = ACTIONS(1571),
    [anon_sym_and] = ACTIONS(1571),
    [anon_sym_or] = ACTIONS(1571),
    [anon_sym_PLUS] = ACTIONS(1571),
    [anon_sym_DASH] = ACTIONS(1571),
    [anon_sym_SLASH] = ACTIONS(1571),
    [anon_sym_SLASH_SLASH] = ACTIONS(1571),
    [anon_sym_PERCENT] = ACTIONS(1571),
    [anon_sym_PIPE] = ACTIONS(1571),
    [anon_sym_CARET] = ACTIONS(1571),
    [anon_sym_AMP] = ACTIONS(1571),
    [anon_sym_LT_LT] = ACTIONS(1571),
    [anon_sym_GT_GT] = ACTIONS(1571),
    [anon_sym_GT_GT_GT] = ACTIONS(1571),
    [anon_sym_TILDE] = ACTIONS(1571),
    [anon_sym_typeof] = ACTIONS(1571),
    [anon_sym_void] = ACTIONS(1571),
    [anon_sym_LT] = ACTIONS(1571),
    [anon_sym_LT_EQ] = ACTIONS(1571),
    [anon_sym_EQ_EQ] = ACTIONS(1571),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1571),
    [anon_sym_BANG_EQ] = ACTIONS(1571),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1571),
    [anon_sym_GT_EQ] = ACTIONS(1571),
    [anon_sym_GT] = ACTIONS(1571),
    [anon_sym_is] = ACTIONS(1571),
    [anon_sym_instanceof] = ACTIONS(1571),
    [anon_sym_new] = ACTIONS(1571),
    [anon_sym_QMARK] = ACTIONS(1571),
    [anon_sym_await] = ACTIONS(1571),
    [anon_sym_LBRACE] = ACTIONS(1571),
    [sym_this] = ACTIONS(1571),
    [sym_true] = ACTIONS(1571),
    [sym_false] = ACTIONS(1571),
    [sym_none] = ACTIONS(1571),
    [sym_string] = ACTIONS(1571),
    [sym_verbatim] = ACTIONS(1571),
    [sym_number] = ACTIONS(1571),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1569),
    [sym_regex] = ACTIONS(1569),
    [sym_fstring_start] = ACTIONS(1569),
  },
  [STATE(686)] = {
    [sym_identifier] = ACTIONS(1468),
    [anon_sym_SEMI] = ACTIONS(1468),
    [anon_sym_return] = ACTIONS(1468),
    [anon_sym_async] = ACTIONS(1468),
    [anon_sym_def] = ACTIONS(1468),
    [anon_sym_COMMA] = ACTIONS(1468),
    [anon_sym_LPAREN] = ACTIONS(1468),
    [anon_sym_LBRACK] = ACTIONS(1468),
    [anon_sym_del] = ACTIONS(1468),
    [anon_sym_raise] = ACTIONS(1468),
    [sym_pass_statement] = ACTIONS(1468),
    [sym_break_statement] = ACTIONS(1468),
    [sym_continue_statement] = ACTIONS(1468),
    [sym_debugger_statement] = ACTIONS(1468),
    [anon_sym_assert] = ACTIONS(1468),
    [anon_sym_global] = ACTIONS(1468),
    [anon_sym_nonlocal] = ACTIONS(1468),
    [anon_sym_yield] = ACTIONS(1468),
    [anon_sym_from] = ACTIONS(1468),
    [anon_sym_import] = ACTIONS(1468),
    [anon_sym_DOT] = ACTIONS(1468),
    [anon_sym_if] = ACTIONS(1468),
    [anon_sym_for] = ACTIONS(1468),
    [anon_sym_in] = ACTIONS(1468),
    [anon_sym_while] = ACTIONS(1468),
    [anon_sym_do] = ACTIONS(1468),
    [anon_sym_try] = ACTIONS(1468),
    [anon_sym_with] = ACTIONS(1468),
    [anon_sym_STAR] = ACTIONS(1468),
    [anon_sym_STAR_STAR] = ACTIONS(1468),
    [anon_sym_class] = ACTIONS(1468),
    [anon_sym_AT] = ACTIONS(1468),
    [anon_sym_not] = ACTIONS(1468),
    [anon_sym_and] = ACTIONS(1468),
    [anon_sym_or] = ACTIONS(1468),
    [anon_sym_PLUS] = ACTIONS(1468),
    [anon_sym_DASH] = ACTIONS(1468),
    [anon_sym_SLASH] = ACTIONS(1468),
    [anon_sym_SLASH_SLASH] = ACTIONS(1468),
    [anon_sym_PERCENT] = ACTIONS(1468),
    [anon_sym_PIPE] = ACTIONS(1468),
    [anon_sym_CARET] = ACTIONS(1468),
    [anon_sym_AMP] = ACTIONS(1468),
    [anon_sym_LT_LT] = ACTIONS(1468),
    [anon_sym_GT_GT] = ACTIONS(1468),
    [anon_sym_GT_GT_GT] = ACTIONS(1468),
    [anon_sym_TILDE] = ACTIONS(1468),
    [anon_sym_typeof] = ACTIONS(1468),
    [anon_sym_void] = ACTIONS(1468),
    [anon_sym_LT] = ACTIONS(1468),
    [anon_sym_LT_EQ] = ACTIONS(1468),
    [anon_sym_EQ_EQ] = ACTIONS(1468),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1468),
    [anon_sym_BANG_EQ] = ACTIONS(1468),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1468),
    [anon_sym_GT_EQ] = ACTIONS(1468),
    [anon_sym_GT] = ACTIONS(1468),
    [anon_sym_is] = ACTIONS(1468),
    [anon_sym_instanceof] = ACTIONS(1468),
    [anon_sym_new] = ACTIONS(1468),
    [anon_sym_QMARK] = ACTIONS(1468),
    [anon_sym_await] = ACTIONS(1468),
    [anon_sym_LBRACE] = ACTIONS(1468),
    [sym_this] = ACTIONS(1468),
    [sym_true] = ACTIONS(1468),
    [sym_false] = ACTIONS(1468),
    [sym_none] = ACTIONS(1468),
    [sym_string] = ACTIONS(1468),
    [sym_verbatim] = ACTIONS(1468),
    [sym_number] = ACTIONS(1468),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1466),
    [sym_regex] = ACTIONS(1466),
    [sym_fstring_start] = ACTIONS(1466),
  },
  [STATE(687)] = {
    [ts_builtin_sym_end] = ACTIONS(937),
    [sym_identifier] = ACTIONS(939),
    [anon_sym_SEMI] = ACTIONS(939),
    [anon_sym_return] = ACTIONS(939),
    [anon_sym_async] = ACTIONS(939),
    [anon_sym_def] = ACTIONS(939),
    [anon_sym_COMMA] = ACTIONS(939),
    [anon_sym_LPAREN] = ACTIONS(939),
    [anon_sym_LBRACK] = ACTIONS(939),
    [anon_sym_del] = ACTIONS(939),
    [anon_sym_raise] = ACTIONS(939),
    [sym_pass_statement] = ACTIONS(939),
    [sym_break_statement] = ACTIONS(939),
    [sym_continue_statement] = ACTIONS(939),
    [sym_debugger_statement] = ACTIONS(939),
    [anon_sym_assert] = ACTIONS(939),
    [anon_sym_global] = ACTIONS(939),
    [anon_sym_nonlocal] = ACTIONS(939),
    [anon_sym_yield] = ACTIONS(939),
    [anon_sym_from] = ACTIONS(939),
    [anon_sym_import] = ACTIONS(939),
    [anon_sym_DOT] = ACTIONS(939),
    [anon_sym_if] = ACTIONS(939),
    [anon_sym_for] = ACTIONS(939),
    [anon_sym_in] = ACTIONS(939),
    [anon_sym_while] = ACTIONS(939),
    [anon_sym_do] = ACTIONS(939),
    [anon_sym_try] = ACTIONS(939),
    [anon_sym_with] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(939),
    [anon_sym_STAR_STAR] = ACTIONS(939),
    [anon_sym_class] = ACTIONS(939),
    [anon_sym_AT] = ACTIONS(939),
    [anon_sym_not] = ACTIONS(939),
    [anon_sym_and] = ACTIONS(939),
    [anon_sym_or] = ACTIONS(939),
    [anon_sym_PLUS] = ACTIONS(939),
    [anon_sym_DASH] = ACTIONS(939),
    [anon_sym_SLASH] = ACTIONS(939),
    [anon_sym_SLASH_SLASH] = ACTIONS(939),
    [anon_sym_PERCENT] = ACTIONS(939),
    [anon_sym_PIPE] = ACTIONS(939),
    [anon_sym_CARET] = ACTIONS(939),
    [anon_sym_AMP] = ACTIONS(939),
    [anon_sym_LT_LT] = ACTIONS(939),
    [anon_sym_GT_GT] = ACTIONS(939),
    [anon_sym_GT_GT_GT] = ACTIONS(939),
    [anon_sym_TILDE] = ACTIONS(939),
    [anon_sym_typeof] = ACTIONS(939),
    [anon_sym_void] = ACTIONS(939),
    [anon_sym_LT] = ACTIONS(939),
    [anon_sym_LT_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(939),
    [anon_sym_GT_EQ] = ACTIONS(939),
    [anon_sym_GT] = ACTIONS(939),
    [anon_sym_is] = ACTIONS(939),
    [anon_sym_instanceof] = ACTIONS(939),
    [anon_sym_new] = ACTIONS(939),
    [anon_sym_QMARK] = ACTIONS(939),
    [anon_sym_await] = ACTIONS(939),
    [anon_sym_LBRACE] = ACTIONS(939),
    [sym_this] = ACTIONS(939),
    [sym_true] = ACTIONS(939),
    [sym_false] = ACTIONS(939),
    [sym_none] = ACTIONS(939),
    [sym_string] = ACTIONS(939),
    [sym_verbatim] = ACTIONS(939),
    [sym_number] = ACTIONS(939),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(937),
    [sym_fstring_start] = ACTIONS(937),
  },
  [STATE(688)] = {
    [ts_builtin_sym_end] = ACTIONS(1615),
    [sym_identifier] = ACTIONS(1617),
    [anon_sym_SEMI] = ACTIONS(1617),
    [anon_sym_return] = ACTIONS(1617),
    [anon_sym_async] = ACTIONS(1617),
    [anon_sym_def] = ACTIONS(1617),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_LBRACK] = ACTIONS(1617),
    [anon_sym_del] = ACTIONS(1617),
    [anon_sym_raise] = ACTIONS(1617),
    [sym_pass_statement] = ACTIONS(1617),
    [sym_break_statement] = ACTIONS(1617),
    [sym_continue_statement] = ACTIONS(1617),
    [sym_debugger_statement] = ACTIONS(1617),
    [anon_sym_assert] = ACTIONS(1617),
    [anon_sym_global] = ACTIONS(1617),
    [anon_sym_nonlocal] = ACTIONS(1617),
    [anon_sym_yield] = ACTIONS(1617),
    [anon_sym_from] = ACTIONS(1617),
    [anon_sym_import] = ACTIONS(1617),
    [anon_sym_DOT] = ACTIONS(1617),
    [anon_sym_if] = ACTIONS(1617),
    [anon_sym_for] = ACTIONS(1617),
    [anon_sym_in] = ACTIONS(1617),
    [anon_sym_while] = ACTIONS(1617),
    [anon_sym_do] = ACTIONS(1617),
    [anon_sym_try] = ACTIONS(1617),
    [anon_sym_with] = ACTIONS(1617),
    [anon_sym_STAR] = ACTIONS(1617),
    [anon_sym_STAR_STAR] = ACTIONS(1617),
    [anon_sym_class] = ACTIONS(1617),
    [anon_sym_AT] = ACTIONS(1617),
    [anon_sym_not] = ACTIONS(1617),
    [anon_sym_and] = ACTIONS(1617),
    [anon_sym_or] = ACTIONS(1617),
    [anon_sym_PLUS] = ACTIONS(1617),
    [anon_sym_DASH] = ACTIONS(1617),
    [anon_sym_SLASH] = ACTIONS(1617),
    [anon_sym_SLASH_SLASH] = ACTIONS(1617),
    [anon_sym_PERCENT] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1617),
    [anon_sym_CARET] = ACTIONS(1617),
    [anon_sym_AMP] = ACTIONS(1617),
    [anon_sym_LT_LT] = ACTIONS(1617),
    [anon_sym_GT_GT] = ACTIONS(1617),
    [anon_sym_GT_GT_GT] = ACTIONS(1617),
    [anon_sym_TILDE] = ACTIONS(1617),
    [anon_sym_typeof] = ACTIONS(1617),
    [anon_sym_void] = ACTIONS(1617),
    [anon_sym_LT] = ACTIONS(1617),
    [anon_sym_LT_EQ] = ACTIONS(1617),
    [anon_sym_EQ_EQ] = ACTIONS(1617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1617),
    [anon_sym_BANG_EQ] = ACTIONS(1617),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1617),
    [anon_sym_GT_EQ] = ACTIONS(1617),
    [anon_sym_GT] = ACTIONS(1617),
    [anon_sym_is] = ACTIONS(1617),
    [anon_sym_instanceof] = ACTIONS(1617),
    [anon_sym_new] = ACTIONS(1617),
    [anon_sym_QMARK] = ACTIONS(1617),
    [anon_sym_await] = ACTIONS(1617),
    [anon_sym_LBRACE] = ACTIONS(1617),
    [sym_this] = ACTIONS(1617),
    [sym_true] = ACTIONS(1617),
    [sym_false] = ACTIONS(1617),
    [sym_none] = ACTIONS(1617),
    [sym_string] = ACTIONS(1617),
    [sym_verbatim] = ACTIONS(1617),
    [sym_number] = ACTIONS(1617),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1615),
    [sym_fstring_start] = ACTIONS(1615),
  },
  [STATE(689)] = {
    [ts_builtin_sym_end] = ACTIONS(1619),
    [sym_identifier] = ACTIONS(1621),
    [anon_sym_SEMI] = ACTIONS(1621),
    [anon_sym_return] = ACTIONS(1621),
    [anon_sym_async] = ACTIONS(1621),
    [anon_sym_def] = ACTIONS(1621),
    [anon_sym_COMMA] = ACTIONS(1621),
    [anon_sym_LPAREN] = ACTIONS(1621),
    [anon_sym_LBRACK] = ACTIONS(1621),
    [anon_sym_del] = ACTIONS(1621),
    [anon_sym_raise] = ACTIONS(1621),
    [sym_pass_statement] = ACTIONS(1621),
    [sym_break_statement] = ACTIONS(1621),
    [sym_continue_statement] = ACTIONS(1621),
    [sym_debugger_statement] = ACTIONS(1621),
    [anon_sym_assert] = ACTIONS(1621),
    [anon_sym_global] = ACTIONS(1621),
    [anon_sym_nonlocal] = ACTIONS(1621),
    [anon_sym_yield] = ACTIONS(1621),
    [anon_sym_from] = ACTIONS(1621),
    [anon_sym_import] = ACTIONS(1621),
    [anon_sym_DOT] = ACTIONS(1621),
    [anon_sym_if] = ACTIONS(1621),
    [anon_sym_for] = ACTIONS(1621),
    [anon_sym_in] = ACTIONS(1621),
    [anon_sym_while] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1621),
    [anon_sym_try] = ACTIONS(1621),
    [anon_sym_with] = ACTIONS(1621),
    [anon_sym_STAR] = ACTIONS(1621),
    [anon_sym_STAR_STAR] = ACTIONS(1621),
    [anon_sym_class] = ACTIONS(1621),
    [anon_sym_AT] = ACTIONS(1621),
    [anon_sym_not] = ACTIONS(1621),
    [anon_sym_and] = ACTIONS(1621),
    [anon_sym_or] = ACTIONS(1621),
    [anon_sym_PLUS] = ACTIONS(1621),
    [anon_sym_DASH] = ACTIONS(1621),
    [anon_sym_SLASH] = ACTIONS(1621),
    [anon_sym_SLASH_SLASH] = ACTIONS(1621),
    [anon_sym_PERCENT] = ACTIONS(1621),
    [anon_sym_PIPE] = ACTIONS(1621),
    [anon_sym_CARET] = ACTIONS(1621),
    [anon_sym_AMP] = ACTIONS(1621),
    [anon_sym_LT_LT] = ACTIONS(1621),
    [anon_sym_GT_GT] = ACTIONS(1621),
    [anon_sym_GT_GT_GT] = ACTIONS(1621),
    [anon_sym_TILDE] = ACTIONS(1621),
    [anon_sym_typeof] = ACTIONS(1621),
    [anon_sym_void] = ACTIONS(1621),
    [anon_sym_LT] = ACTIONS(1621),
    [anon_sym_LT_EQ] = ACTIONS(1621),
    [anon_sym_EQ_EQ] = ACTIONS(1621),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1621),
    [anon_sym_BANG_EQ] = ACTIONS(1621),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1621),
    [anon_sym_GT_EQ] = ACTIONS(1621),
    [anon_sym_GT] = ACTIONS(1621),
    [anon_sym_is] = ACTIONS(1621),
    [anon_sym_instanceof] = ACTIONS(1621),
    [anon_sym_new] = ACTIONS(1621),
    [anon_sym_QMARK] = ACTIONS(1621),
    [anon_sym_await] = ACTIONS(1621),
    [anon_sym_LBRACE] = ACTIONS(1621),
    [sym_this] = ACTIONS(1621),
    [sym_true] = ACTIONS(1621),
    [sym_false] = ACTIONS(1621),
    [sym_none] = ACTIONS(1621),
    [sym_string] = ACTIONS(1621),
    [sym_verbatim] = ACTIONS(1621),
    [sym_number] = ACTIONS(1621),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1619),
    [sym_fstring_start] = ACTIONS(1619),
  },
  [STATE(690)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(574),
    [ts_builtin_sym_end] = ACTIONS(1357),
    [sym_identifier] = ACTIONS(1359),
    [anon_sym_SEMI] = ACTIONS(1359),
    [anon_sym_return] = ACTIONS(1359),
    [anon_sym_async] = ACTIONS(1359),
    [anon_sym_def] = ACTIONS(1359),
    [anon_sym_LPAREN] = ACTIONS(1359),
    [anon_sym_LBRACK] = ACTIONS(1359),
    [anon_sym_del] = ACTIONS(1359),
    [anon_sym_raise] = ACTIONS(1359),
    [sym_pass_statement] = ACTIONS(1359),
    [sym_break_statement] = ACTIONS(1359),
    [sym_continue_statement] = ACTIONS(1359),
    [sym_debugger_statement] = ACTIONS(1359),
    [anon_sym_assert] = ACTIONS(1359),
    [anon_sym_global] = ACTIONS(1359),
    [anon_sym_nonlocal] = ACTIONS(1359),
    [anon_sym_yield] = ACTIONS(1359),
    [anon_sym_from] = ACTIONS(1359),
    [anon_sym_import] = ACTIONS(1359),
    [anon_sym_DOT] = ACTIONS(1359),
    [anon_sym_if] = ACTIONS(1359),
    [anon_sym_for] = ACTIONS(1359),
    [anon_sym_in] = ACTIONS(1359),
    [anon_sym_while] = ACTIONS(1359),
    [anon_sym_do] = ACTIONS(1359),
    [anon_sym_try] = ACTIONS(1359),
    [anon_sym_with] = ACTIONS(1359),
    [anon_sym_STAR] = ACTIONS(1359),
    [anon_sym_STAR_STAR] = ACTIONS(1359),
    [anon_sym_class] = ACTIONS(1359),
    [anon_sym_AT] = ACTIONS(1359),
    [anon_sym_not] = ACTIONS(1359),
    [anon_sym_and] = ACTIONS(1359),
    [anon_sym_or] = ACTIONS(1359),
    [anon_sym_PLUS] = ACTIONS(1359),
    [anon_sym_DASH] = ACTIONS(1359),
    [anon_sym_SLASH] = ACTIONS(1359),
    [anon_sym_SLASH_SLASH] = ACTIONS(1359),
    [anon_sym_PERCENT] = ACTIONS(1359),
    [anon_sym_PIPE] = ACTIONS(1359),
    [anon_sym_CARET] = ACTIONS(1359),
    [anon_sym_AMP] = ACTIONS(1359),
    [anon_sym_LT_LT] = ACTIONS(1359),
    [anon_sym_GT_GT] = ACTIONS(1359),
    [anon_sym_GT_GT_GT] = ACTIONS(1359),
    [anon_sym_TILDE] = ACTIONS(1359),
    [anon_sym_typeof] = ACTIONS(1359),
    [anon_sym_void] = ACTIONS(1359),
    [anon_sym_LT] = ACTIONS(1359),
    [anon_sym_LT_EQ] = ACTIONS(1359),
    [anon_sym_EQ_EQ] = ACTIONS(1359),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1359),
    [anon_sym_BANG_EQ] = ACTIONS(1359),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1359),
    [anon_sym_GT_EQ] = ACTIONS(1359),
    [anon_sym_GT] = ACTIONS(1359),
    [anon_sym_is] = ACTIONS(1359),
    [anon_sym_instanceof] = ACTIONS(1359),
    [anon_sym_new] = ACTIONS(1359),
    [anon_sym_QMARK] = ACTIONS(1359),
    [anon_sym_await] = ACTIONS(1359),
    [anon_sym_LBRACE] = ACTIONS(1359),
    [sym_this] = ACTIONS(1359),
    [sym_true] = ACTIONS(1359),
    [sym_false] = ACTIONS(1359),
    [sym_none] = ACTIONS(1359),
    [sym_string] = ACTIONS(1359),
    [sym_verbatim] = ACTIONS(1359),
    [sym_number] = ACTIONS(1359),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1357),
    [sym_fstring_start] = ACTIONS(1357),
  },
  [STATE(691)] = {
    [ts_builtin_sym_end] = ACTIONS(1623),
    [sym_identifier] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1626),
    [anon_sym_return] = ACTIONS(1626),
    [anon_sym_async] = ACTIONS(1626),
    [anon_sym_def] = ACTIONS(1626),
    [anon_sym_COMMA] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1626),
    [anon_sym_del] = ACTIONS(1626),
    [anon_sym_raise] = ACTIONS(1626),
    [sym_pass_statement] = ACTIONS(1626),
    [sym_break_statement] = ACTIONS(1626),
    [sym_continue_statement] = ACTIONS(1626),
    [sym_debugger_statement] = ACTIONS(1626),
    [anon_sym_assert] = ACTIONS(1626),
    [anon_sym_global] = ACTIONS(1626),
    [anon_sym_nonlocal] = ACTIONS(1626),
    [anon_sym_yield] = ACTIONS(1626),
    [anon_sym_from] = ACTIONS(1626),
    [anon_sym_import] = ACTIONS(1626),
    [anon_sym_DOT] = ACTIONS(1626),
    [anon_sym_if] = ACTIONS(1626),
    [anon_sym_for] = ACTIONS(1626),
    [anon_sym_in] = ACTIONS(1626),
    [anon_sym_while] = ACTIONS(1626),
    [anon_sym_do] = ACTIONS(1626),
    [anon_sym_try] = ACTIONS(1626),
    [anon_sym_with] = ACTIONS(1626),
    [anon_sym_STAR] = ACTIONS(1626),
    [anon_sym_STAR_STAR] = ACTIONS(1626),
    [anon_sym_class] = ACTIONS(1626),
    [anon_sym_AT] = ACTIONS(1626),
    [anon_sym_not] = ACTIONS(1626),
    [anon_sym_and] = ACTIONS(1626),
    [anon_sym_or] = ACTIONS(1626),
    [anon_sym_PLUS] = ACTIONS(1626),
    [anon_sym_DASH] = ACTIONS(1626),
    [anon_sym_SLASH] = ACTIONS(1626),
    [anon_sym_SLASH_SLASH] = ACTIONS(1626),
    [anon_sym_PERCENT] = ACTIONS(1626),
    [anon_sym_PIPE] = ACTIONS(1626),
    [anon_sym_CARET] = ACTIONS(1626),
    [anon_sym_AMP] = ACTIONS(1626),
    [anon_sym_LT_LT] = ACTIONS(1626),
    [anon_sym_GT_GT] = ACTIONS(1626),
    [anon_sym_GT_GT_GT] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1626),
    [anon_sym_typeof] = ACTIONS(1626),
    [anon_sym_void] = ACTIONS(1626),
    [anon_sym_LT] = ACTIONS(1626),
    [anon_sym_LT_EQ] = ACTIONS(1626),
    [anon_sym_EQ_EQ] = ACTIONS(1626),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1626),
    [anon_sym_BANG_EQ] = ACTIONS(1626),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1626),
    [anon_sym_GT_EQ] = ACTIONS(1626),
    [anon_sym_GT] = ACTIONS(1626),
    [anon_sym_is] = ACTIONS(1626),
    [anon_sym_instanceof] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1626),
    [anon_sym_await] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1626),
    [sym_this] = ACTIONS(1626),
    [sym_true] = ACTIONS(1626),
    [sym_false] = ACTIONS(1626),
    [sym_none] = ACTIONS(1626),
    [sym_string] = ACTIONS(1626),
    [sym_verbatim] = ACTIONS(1626),
    [sym_number] = ACTIONS(1626),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1623),
    [sym_fstring_start] = ACTIONS(1623),
  },
  [STATE(692)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1228),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1228),
    [anon_sym_SLASH_SLASH] = ACTIONS(1228),
    [anon_sym_PERCENT] = ACTIONS(1228),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(693)] = {
    [sym_identifier] = ACTIONS(1472),
    [anon_sym_SEMI] = ACTIONS(1472),
    [anon_sym_return] = ACTIONS(1472),
    [anon_sym_async] = ACTIONS(1472),
    [anon_sym_def] = ACTIONS(1472),
    [anon_sym_COMMA] = ACTIONS(1472),
    [anon_sym_LPAREN] = ACTIONS(1472),
    [anon_sym_LBRACK] = ACTIONS(1472),
    [anon_sym_del] = ACTIONS(1472),
    [anon_sym_raise] = ACTIONS(1472),
    [sym_pass_statement] = ACTIONS(1472),
    [sym_break_statement] = ACTIONS(1472),
    [sym_continue_statement] = ACTIONS(1472),
    [sym_debugger_statement] = ACTIONS(1472),
    [anon_sym_assert] = ACTIONS(1472),
    [anon_sym_global] = ACTIONS(1472),
    [anon_sym_nonlocal] = ACTIONS(1472),
    [anon_sym_yield] = ACTIONS(1472),
    [anon_sym_from] = ACTIONS(1472),
    [anon_sym_import] = ACTIONS(1472),
    [anon_sym_DOT] = ACTIONS(1472),
    [anon_sym_if] = ACTIONS(1472),
    [anon_sym_for] = ACTIONS(1472),
    [anon_sym_in] = ACTIONS(1472),
    [anon_sym_while] = ACTIONS(1472),
    [anon_sym_do] = ACTIONS(1472),
    [anon_sym_try] = ACTIONS(1472),
    [anon_sym_with] = ACTIONS(1472),
    [anon_sym_STAR] = ACTIONS(1472),
    [anon_sym_STAR_STAR] = ACTIONS(1472),
    [anon_sym_class] = ACTIONS(1472),
    [anon_sym_AT] = ACTIONS(1472),
    [anon_sym_not] = ACTIONS(1472),
    [anon_sym_and] = ACTIONS(1472),
    [anon_sym_or] = ACTIONS(1472),
    [anon_sym_PLUS] = ACTIONS(1472),
    [anon_sym_DASH] = ACTIONS(1472),
    [anon_sym_SLASH] = ACTIONS(1472),
    [anon_sym_SLASH_SLASH] = ACTIONS(1472),
    [anon_sym_PERCENT] = ACTIONS(1472),
    [anon_sym_PIPE] = ACTIONS(1472),
    [anon_sym_CARET] = ACTIONS(1472),
    [anon_sym_AMP] = ACTIONS(1472),
    [anon_sym_LT_LT] = ACTIONS(1472),
    [anon_sym_GT_GT] = ACTIONS(1472),
    [anon_sym_GT_GT_GT] = ACTIONS(1472),
    [anon_sym_TILDE] = ACTIONS(1472),
    [anon_sym_typeof] = ACTIONS(1472),
    [anon_sym_void] = ACTIONS(1472),
    [anon_sym_LT] = ACTIONS(1472),
    [anon_sym_LT_EQ] = ACTIONS(1472),
    [anon_sym_EQ_EQ] = ACTIONS(1472),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1472),
    [anon_sym_BANG_EQ] = ACTIONS(1472),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1472),
    [anon_sym_GT_EQ] = ACTIONS(1472),
    [anon_sym_GT] = ACTIONS(1472),
    [anon_sym_is] = ACTIONS(1472),
    [anon_sym_instanceof] = ACTIONS(1472),
    [anon_sym_new] = ACTIONS(1472),
    [anon_sym_QMARK] = ACTIONS(1472),
    [anon_sym_await] = ACTIONS(1472),
    [anon_sym_LBRACE] = ACTIONS(1472),
    [sym_this] = ACTIONS(1472),
    [sym_true] = ACTIONS(1472),
    [sym_false] = ACTIONS(1472),
    [sym_none] = ACTIONS(1472),
    [sym_string] = ACTIONS(1472),
    [sym_verbatim] = ACTIONS(1472),
    [sym_number] = ACTIONS(1472),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1470),
    [sym_regex] = ACTIONS(1470),
    [sym_fstring_start] = ACTIONS(1470),
  },
  [STATE(694)] = {
    [ts_builtin_sym_end] = ACTIONS(1629),
    [sym_identifier] = ACTIONS(1631),
    [anon_sym_SEMI] = ACTIONS(1631),
    [anon_sym_return] = ACTIONS(1631),
    [anon_sym_async] = ACTIONS(1631),
    [anon_sym_def] = ACTIONS(1631),
    [anon_sym_COMMA] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(1631),
    [anon_sym_LBRACK] = ACTIONS(1631),
    [anon_sym_del] = ACTIONS(1631),
    [anon_sym_raise] = ACTIONS(1631),
    [sym_pass_statement] = ACTIONS(1631),
    [sym_break_statement] = ACTIONS(1631),
    [sym_continue_statement] = ACTIONS(1631),
    [sym_debugger_statement] = ACTIONS(1631),
    [anon_sym_assert] = ACTIONS(1631),
    [anon_sym_global] = ACTIONS(1631),
    [anon_sym_nonlocal] = ACTIONS(1631),
    [anon_sym_yield] = ACTIONS(1631),
    [anon_sym_from] = ACTIONS(1631),
    [anon_sym_import] = ACTIONS(1631),
    [anon_sym_DOT] = ACTIONS(1631),
    [anon_sym_if] = ACTIONS(1631),
    [anon_sym_for] = ACTIONS(1631),
    [anon_sym_in] = ACTIONS(1631),
    [anon_sym_while] = ACTIONS(1631),
    [anon_sym_do] = ACTIONS(1631),
    [anon_sym_try] = ACTIONS(1631),
    [anon_sym_with] = ACTIONS(1631),
    [anon_sym_STAR] = ACTIONS(1631),
    [anon_sym_STAR_STAR] = ACTIONS(1631),
    [anon_sym_class] = ACTIONS(1631),
    [anon_sym_AT] = ACTIONS(1631),
    [anon_sym_not] = ACTIONS(1631),
    [anon_sym_and] = ACTIONS(1631),
    [anon_sym_or] = ACTIONS(1631),
    [anon_sym_PLUS] = ACTIONS(1631),
    [anon_sym_DASH] = ACTIONS(1631),
    [anon_sym_SLASH] = ACTIONS(1631),
    [anon_sym_SLASH_SLASH] = ACTIONS(1631),
    [anon_sym_PERCENT] = ACTIONS(1631),
    [anon_sym_PIPE] = ACTIONS(1631),
    [anon_sym_CARET] = ACTIONS(1631),
    [anon_sym_AMP] = ACTIONS(1631),
    [anon_sym_LT_LT] = ACTIONS(1631),
    [anon_sym_GT_GT] = ACTIONS(1631),
    [anon_sym_GT_GT_GT] = ACTIONS(1631),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_typeof] = ACTIONS(1631),
    [anon_sym_void] = ACTIONS(1631),
    [anon_sym_LT] = ACTIONS(1631),
    [anon_sym_LT_EQ] = ACTIONS(1631),
    [anon_sym_EQ_EQ] = ACTIONS(1631),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1631),
    [anon_sym_BANG_EQ] = ACTIONS(1631),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1631),
    [anon_sym_GT_EQ] = ACTIONS(1631),
    [anon_sym_GT] = ACTIONS(1631),
    [anon_sym_is] = ACTIONS(1631),
    [anon_sym_instanceof] = ACTIONS(1631),
    [anon_sym_new] = ACTIONS(1631),
    [anon_sym_QMARK] = ACTIONS(1631),
    [anon_sym_await] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1631),
    [sym_this] = ACTIONS(1631),
    [sym_true] = ACTIONS(1631),
    [sym_false] = ACTIONS(1631),
    [sym_none] = ACTIONS(1631),
    [sym_string] = ACTIONS(1631),
    [sym_verbatim] = ACTIONS(1631),
    [sym_number] = ACTIONS(1631),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1629),
    [sym_fstring_start] = ACTIONS(1629),
  },
  [STATE(695)] = {
    [ts_builtin_sym_end] = ACTIONS(1633),
    [sym_identifier] = ACTIONS(1635),
    [anon_sym_SEMI] = ACTIONS(1635),
    [anon_sym_return] = ACTIONS(1635),
    [anon_sym_async] = ACTIONS(1635),
    [anon_sym_def] = ACTIONS(1635),
    [anon_sym_COMMA] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(1635),
    [anon_sym_LBRACK] = ACTIONS(1635),
    [anon_sym_del] = ACTIONS(1635),
    [anon_sym_raise] = ACTIONS(1635),
    [sym_pass_statement] = ACTIONS(1635),
    [sym_break_statement] = ACTIONS(1635),
    [sym_continue_statement] = ACTIONS(1635),
    [sym_debugger_statement] = ACTIONS(1635),
    [anon_sym_assert] = ACTIONS(1635),
    [anon_sym_global] = ACTIONS(1635),
    [anon_sym_nonlocal] = ACTIONS(1635),
    [anon_sym_yield] = ACTIONS(1635),
    [anon_sym_from] = ACTIONS(1635),
    [anon_sym_import] = ACTIONS(1635),
    [anon_sym_DOT] = ACTIONS(1635),
    [anon_sym_if] = ACTIONS(1635),
    [anon_sym_for] = ACTIONS(1635),
    [anon_sym_in] = ACTIONS(1635),
    [anon_sym_while] = ACTIONS(1635),
    [anon_sym_do] = ACTIONS(1635),
    [anon_sym_try] = ACTIONS(1635),
    [anon_sym_with] = ACTIONS(1635),
    [anon_sym_STAR] = ACTIONS(1635),
    [anon_sym_STAR_STAR] = ACTIONS(1635),
    [anon_sym_class] = ACTIONS(1635),
    [anon_sym_AT] = ACTIONS(1635),
    [anon_sym_not] = ACTIONS(1635),
    [anon_sym_and] = ACTIONS(1635),
    [anon_sym_or] = ACTIONS(1635),
    [anon_sym_PLUS] = ACTIONS(1635),
    [anon_sym_DASH] = ACTIONS(1635),
    [anon_sym_SLASH] = ACTIONS(1635),
    [anon_sym_SLASH_SLASH] = ACTIONS(1635),
    [anon_sym_PERCENT] = ACTIONS(1635),
    [anon_sym_PIPE] = ACTIONS(1635),
    [anon_sym_CARET] = ACTIONS(1635),
    [anon_sym_AMP] = ACTIONS(1635),
    [anon_sym_LT_LT] = ACTIONS(1635),
    [anon_sym_GT_GT] = ACTIONS(1635),
    [anon_sym_GT_GT_GT] = ACTIONS(1635),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_typeof] = ACTIONS(1635),
    [anon_sym_void] = ACTIONS(1635),
    [anon_sym_LT] = ACTIONS(1635),
    [anon_sym_LT_EQ] = ACTIONS(1635),
    [anon_sym_EQ_EQ] = ACTIONS(1635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1635),
    [anon_sym_BANG_EQ] = ACTIONS(1635),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1635),
    [anon_sym_GT_EQ] = ACTIONS(1635),
    [anon_sym_GT] = ACTIONS(1635),
    [anon_sym_is] = ACTIONS(1635),
    [anon_sym_instanceof] = ACTIONS(1635),
    [anon_sym_new] = ACTIONS(1635),
    [anon_sym_QMARK] = ACTIONS(1635),
    [anon_sym_await] = ACTIONS(1635),
    [anon_sym_LBRACE] = ACTIONS(1635),
    [sym_this] = ACTIONS(1635),
    [sym_true] = ACTIONS(1635),
    [sym_false] = ACTIONS(1635),
    [sym_none] = ACTIONS(1635),
    [sym_string] = ACTIONS(1635),
    [sym_verbatim] = ACTIONS(1635),
    [sym_number] = ACTIONS(1635),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1633),
    [sym_fstring_start] = ACTIONS(1633),
  },
  [STATE(696)] = {
    [sym_identifier] = ACTIONS(1635),
    [anon_sym_SEMI] = ACTIONS(1635),
    [anon_sym_return] = ACTIONS(1635),
    [anon_sym_async] = ACTIONS(1635),
    [anon_sym_def] = ACTIONS(1635),
    [anon_sym_COMMA] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(1635),
    [anon_sym_LBRACK] = ACTIONS(1635),
    [anon_sym_del] = ACTIONS(1635),
    [anon_sym_raise] = ACTIONS(1635),
    [sym_pass_statement] = ACTIONS(1635),
    [sym_break_statement] = ACTIONS(1635),
    [sym_continue_statement] = ACTIONS(1635),
    [sym_debugger_statement] = ACTIONS(1635),
    [anon_sym_assert] = ACTIONS(1635),
    [anon_sym_global] = ACTIONS(1635),
    [anon_sym_nonlocal] = ACTIONS(1635),
    [anon_sym_yield] = ACTIONS(1635),
    [anon_sym_from] = ACTIONS(1635),
    [anon_sym_import] = ACTIONS(1635),
    [anon_sym_DOT] = ACTIONS(1635),
    [anon_sym_if] = ACTIONS(1635),
    [anon_sym_for] = ACTIONS(1635),
    [anon_sym_in] = ACTIONS(1635),
    [anon_sym_while] = ACTIONS(1635),
    [anon_sym_do] = ACTIONS(1635),
    [anon_sym_try] = ACTIONS(1635),
    [anon_sym_with] = ACTIONS(1635),
    [anon_sym_STAR] = ACTIONS(1635),
    [anon_sym_STAR_STAR] = ACTIONS(1635),
    [anon_sym_class] = ACTIONS(1635),
    [anon_sym_AT] = ACTIONS(1635),
    [anon_sym_not] = ACTIONS(1635),
    [anon_sym_and] = ACTIONS(1635),
    [anon_sym_or] = ACTIONS(1635),
    [anon_sym_PLUS] = ACTIONS(1635),
    [anon_sym_DASH] = ACTIONS(1635),
    [anon_sym_SLASH] = ACTIONS(1635),
    [anon_sym_SLASH_SLASH] = ACTIONS(1635),
    [anon_sym_PERCENT] = ACTIONS(1635),
    [anon_sym_PIPE] = ACTIONS(1635),
    [anon_sym_CARET] = ACTIONS(1635),
    [anon_sym_AMP] = ACTIONS(1635),
    [anon_sym_LT_LT] = ACTIONS(1635),
    [anon_sym_GT_GT] = ACTIONS(1635),
    [anon_sym_GT_GT_GT] = ACTIONS(1635),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_typeof] = ACTIONS(1635),
    [anon_sym_void] = ACTIONS(1635),
    [anon_sym_LT] = ACTIONS(1635),
    [anon_sym_LT_EQ] = ACTIONS(1635),
    [anon_sym_EQ_EQ] = ACTIONS(1635),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1635),
    [anon_sym_BANG_EQ] = ACTIONS(1635),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1635),
    [anon_sym_GT_EQ] = ACTIONS(1635),
    [anon_sym_GT] = ACTIONS(1635),
    [anon_sym_is] = ACTIONS(1635),
    [anon_sym_instanceof] = ACTIONS(1635),
    [anon_sym_new] = ACTIONS(1635),
    [anon_sym_QMARK] = ACTIONS(1635),
    [anon_sym_await] = ACTIONS(1635),
    [anon_sym_LBRACE] = ACTIONS(1635),
    [sym_this] = ACTIONS(1635),
    [sym_true] = ACTIONS(1635),
    [sym_false] = ACTIONS(1635),
    [sym_none] = ACTIONS(1635),
    [sym_string] = ACTIONS(1635),
    [sym_verbatim] = ACTIONS(1635),
    [sym_number] = ACTIONS(1635),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1633),
    [sym_regex] = ACTIONS(1633),
    [sym_fstring_start] = ACTIONS(1633),
  },
  [STATE(697)] = {
    [sym_identifier] = ACTIONS(1626),
    [anon_sym_SEMI] = ACTIONS(1626),
    [anon_sym_return] = ACTIONS(1626),
    [anon_sym_async] = ACTIONS(1626),
    [anon_sym_def] = ACTIONS(1626),
    [anon_sym_COMMA] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1626),
    [anon_sym_del] = ACTIONS(1626),
    [anon_sym_raise] = ACTIONS(1626),
    [sym_pass_statement] = ACTIONS(1626),
    [sym_break_statement] = ACTIONS(1626),
    [sym_continue_statement] = ACTIONS(1626),
    [sym_debugger_statement] = ACTIONS(1626),
    [anon_sym_assert] = ACTIONS(1626),
    [anon_sym_global] = ACTIONS(1626),
    [anon_sym_nonlocal] = ACTIONS(1626),
    [anon_sym_yield] = ACTIONS(1626),
    [anon_sym_from] = ACTIONS(1626),
    [anon_sym_import] = ACTIONS(1626),
    [anon_sym_DOT] = ACTIONS(1626),
    [anon_sym_if] = ACTIONS(1626),
    [anon_sym_for] = ACTIONS(1626),
    [anon_sym_in] = ACTIONS(1626),
    [anon_sym_while] = ACTIONS(1626),
    [anon_sym_do] = ACTIONS(1626),
    [anon_sym_try] = ACTIONS(1626),
    [anon_sym_with] = ACTIONS(1626),
    [anon_sym_STAR] = ACTIONS(1626),
    [anon_sym_STAR_STAR] = ACTIONS(1626),
    [anon_sym_class] = ACTIONS(1626),
    [anon_sym_AT] = ACTIONS(1626),
    [anon_sym_not] = ACTIONS(1626),
    [anon_sym_and] = ACTIONS(1626),
    [anon_sym_or] = ACTIONS(1626),
    [anon_sym_PLUS] = ACTIONS(1626),
    [anon_sym_DASH] = ACTIONS(1626),
    [anon_sym_SLASH] = ACTIONS(1626),
    [anon_sym_SLASH_SLASH] = ACTIONS(1626),
    [anon_sym_PERCENT] = ACTIONS(1626),
    [anon_sym_PIPE] = ACTIONS(1626),
    [anon_sym_CARET] = ACTIONS(1626),
    [anon_sym_AMP] = ACTIONS(1626),
    [anon_sym_LT_LT] = ACTIONS(1626),
    [anon_sym_GT_GT] = ACTIONS(1626),
    [anon_sym_GT_GT_GT] = ACTIONS(1626),
    [anon_sym_TILDE] = ACTIONS(1626),
    [anon_sym_typeof] = ACTIONS(1626),
    [anon_sym_void] = ACTIONS(1626),
    [anon_sym_LT] = ACTIONS(1626),
    [anon_sym_LT_EQ] = ACTIONS(1626),
    [anon_sym_EQ_EQ] = ACTIONS(1626),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1626),
    [anon_sym_BANG_EQ] = ACTIONS(1626),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1626),
    [anon_sym_GT_EQ] = ACTIONS(1626),
    [anon_sym_GT] = ACTIONS(1626),
    [anon_sym_is] = ACTIONS(1626),
    [anon_sym_instanceof] = ACTIONS(1626),
    [anon_sym_new] = ACTIONS(1626),
    [anon_sym_QMARK] = ACTIONS(1626),
    [anon_sym_await] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1626),
    [sym_this] = ACTIONS(1626),
    [sym_true] = ACTIONS(1626),
    [sym_false] = ACTIONS(1626),
    [sym_none] = ACTIONS(1626),
    [sym_string] = ACTIONS(1626),
    [sym_verbatim] = ACTIONS(1626),
    [sym_number] = ACTIONS(1626),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1623),
    [sym_regex] = ACTIONS(1623),
    [sym_fstring_start] = ACTIONS(1623),
  },
  [STATE(698)] = {
    [sym_identifier] = ACTIONS(1575),
    [anon_sym_SEMI] = ACTIONS(1575),
    [anon_sym_return] = ACTIONS(1575),
    [anon_sym_async] = ACTIONS(1575),
    [anon_sym_def] = ACTIONS(1575),
    [anon_sym_COMMA] = ACTIONS(1575),
    [anon_sym_LPAREN] = ACTIONS(1575),
    [anon_sym_LBRACK] = ACTIONS(1575),
    [anon_sym_del] = ACTIONS(1575),
    [anon_sym_raise] = ACTIONS(1575),
    [sym_pass_statement] = ACTIONS(1575),
    [sym_break_statement] = ACTIONS(1575),
    [sym_continue_statement] = ACTIONS(1575),
    [sym_debugger_statement] = ACTIONS(1575),
    [anon_sym_assert] = ACTIONS(1575),
    [anon_sym_global] = ACTIONS(1575),
    [anon_sym_nonlocal] = ACTIONS(1575),
    [anon_sym_yield] = ACTIONS(1575),
    [anon_sym_from] = ACTIONS(1575),
    [anon_sym_import] = ACTIONS(1575),
    [anon_sym_DOT] = ACTIONS(1575),
    [anon_sym_if] = ACTIONS(1575),
    [anon_sym_for] = ACTIONS(1575),
    [anon_sym_in] = ACTIONS(1575),
    [anon_sym_while] = ACTIONS(1575),
    [anon_sym_do] = ACTIONS(1575),
    [anon_sym_try] = ACTIONS(1575),
    [anon_sym_with] = ACTIONS(1575),
    [anon_sym_STAR] = ACTIONS(1575),
    [anon_sym_STAR_STAR] = ACTIONS(1575),
    [anon_sym_class] = ACTIONS(1575),
    [anon_sym_AT] = ACTIONS(1575),
    [anon_sym_not] = ACTIONS(1575),
    [anon_sym_and] = ACTIONS(1575),
    [anon_sym_or] = ACTIONS(1575),
    [anon_sym_PLUS] = ACTIONS(1575),
    [anon_sym_DASH] = ACTIONS(1575),
    [anon_sym_SLASH] = ACTIONS(1575),
    [anon_sym_SLASH_SLASH] = ACTIONS(1575),
    [anon_sym_PERCENT] = ACTIONS(1575),
    [anon_sym_PIPE] = ACTIONS(1575),
    [anon_sym_CARET] = ACTIONS(1575),
    [anon_sym_AMP] = ACTIONS(1575),
    [anon_sym_LT_LT] = ACTIONS(1575),
    [anon_sym_GT_GT] = ACTIONS(1575),
    [anon_sym_GT_GT_GT] = ACTIONS(1575),
    [anon_sym_TILDE] = ACTIONS(1575),
    [anon_sym_typeof] = ACTIONS(1575),
    [anon_sym_void] = ACTIONS(1575),
    [anon_sym_LT] = ACTIONS(1575),
    [anon_sym_LT_EQ] = ACTIONS(1575),
    [anon_sym_EQ_EQ] = ACTIONS(1575),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1575),
    [anon_sym_BANG_EQ] = ACTIONS(1575),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1575),
    [anon_sym_GT_EQ] = ACTIONS(1575),
    [anon_sym_GT] = ACTIONS(1575),
    [anon_sym_is] = ACTIONS(1575),
    [anon_sym_instanceof] = ACTIONS(1575),
    [anon_sym_new] = ACTIONS(1575),
    [anon_sym_QMARK] = ACTIONS(1575),
    [anon_sym_await] = ACTIONS(1575),
    [anon_sym_LBRACE] = ACTIONS(1575),
    [sym_this] = ACTIONS(1575),
    [sym_true] = ACTIONS(1575),
    [sym_false] = ACTIONS(1575),
    [sym_none] = ACTIONS(1575),
    [sym_string] = ACTIONS(1575),
    [sym_verbatim] = ACTIONS(1575),
    [sym_number] = ACTIONS(1575),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1573),
    [sym_regex] = ACTIONS(1573),
    [sym_fstring_start] = ACTIONS(1573),
  },
  [STATE(699)] = {
    [sym_identifier] = ACTIONS(1579),
    [anon_sym_SEMI] = ACTIONS(1579),
    [anon_sym_return] = ACTIONS(1579),
    [anon_sym_async] = ACTIONS(1579),
    [anon_sym_def] = ACTIONS(1579),
    [anon_sym_COMMA] = ACTIONS(1579),
    [anon_sym_LPAREN] = ACTIONS(1579),
    [anon_sym_LBRACK] = ACTIONS(1579),
    [anon_sym_del] = ACTIONS(1579),
    [anon_sym_raise] = ACTIONS(1579),
    [sym_pass_statement] = ACTIONS(1579),
    [sym_break_statement] = ACTIONS(1579),
    [sym_continue_statement] = ACTIONS(1579),
    [sym_debugger_statement] = ACTIONS(1579),
    [anon_sym_assert] = ACTIONS(1579),
    [anon_sym_global] = ACTIONS(1579),
    [anon_sym_nonlocal] = ACTIONS(1579),
    [anon_sym_yield] = ACTIONS(1579),
    [anon_sym_from] = ACTIONS(1579),
    [anon_sym_import] = ACTIONS(1579),
    [anon_sym_DOT] = ACTIONS(1579),
    [anon_sym_if] = ACTIONS(1579),
    [anon_sym_for] = ACTIONS(1579),
    [anon_sym_in] = ACTIONS(1579),
    [anon_sym_while] = ACTIONS(1579),
    [anon_sym_do] = ACTIONS(1579),
    [anon_sym_try] = ACTIONS(1579),
    [anon_sym_with] = ACTIONS(1579),
    [anon_sym_STAR] = ACTIONS(1579),
    [anon_sym_STAR_STAR] = ACTIONS(1579),
    [anon_sym_class] = ACTIONS(1579),
    [anon_sym_AT] = ACTIONS(1579),
    [anon_sym_not] = ACTIONS(1579),
    [anon_sym_and] = ACTIONS(1579),
    [anon_sym_or] = ACTIONS(1579),
    [anon_sym_PLUS] = ACTIONS(1579),
    [anon_sym_DASH] = ACTIONS(1579),
    [anon_sym_SLASH] = ACTIONS(1579),
    [anon_sym_SLASH_SLASH] = ACTIONS(1579),
    [anon_sym_PERCENT] = ACTIONS(1579),
    [anon_sym_PIPE] = ACTIONS(1579),
    [anon_sym_CARET] = ACTIONS(1579),
    [anon_sym_AMP] = ACTIONS(1579),
    [anon_sym_LT_LT] = ACTIONS(1579),
    [anon_sym_GT_GT] = ACTIONS(1579),
    [anon_sym_GT_GT_GT] = ACTIONS(1579),
    [anon_sym_TILDE] = ACTIONS(1579),
    [anon_sym_typeof] = ACTIONS(1579),
    [anon_sym_void] = ACTIONS(1579),
    [anon_sym_LT] = ACTIONS(1579),
    [anon_sym_LT_EQ] = ACTIONS(1579),
    [anon_sym_EQ_EQ] = ACTIONS(1579),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1579),
    [anon_sym_BANG_EQ] = ACTIONS(1579),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1579),
    [anon_sym_GT_EQ] = ACTIONS(1579),
    [anon_sym_GT] = ACTIONS(1579),
    [anon_sym_is] = ACTIONS(1579),
    [anon_sym_instanceof] = ACTIONS(1579),
    [anon_sym_new] = ACTIONS(1579),
    [anon_sym_QMARK] = ACTIONS(1579),
    [anon_sym_await] = ACTIONS(1579),
    [anon_sym_LBRACE] = ACTIONS(1579),
    [sym_this] = ACTIONS(1579),
    [sym_true] = ACTIONS(1579),
    [sym_false] = ACTIONS(1579),
    [sym_none] = ACTIONS(1579),
    [sym_string] = ACTIONS(1579),
    [sym_verbatim] = ACTIONS(1579),
    [sym_number] = ACTIONS(1579),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1577),
    [sym_regex] = ACTIONS(1577),
    [sym_fstring_start] = ACTIONS(1577),
  },
  [STATE(700)] = {
    [sym_identifier] = ACTIONS(939),
    [anon_sym_SEMI] = ACTIONS(939),
    [anon_sym_return] = ACTIONS(939),
    [anon_sym_async] = ACTIONS(939),
    [anon_sym_def] = ACTIONS(939),
    [anon_sym_COMMA] = ACTIONS(939),
    [anon_sym_LPAREN] = ACTIONS(939),
    [anon_sym_LBRACK] = ACTIONS(939),
    [anon_sym_del] = ACTIONS(939),
    [anon_sym_raise] = ACTIONS(939),
    [sym_pass_statement] = ACTIONS(939),
    [sym_break_statement] = ACTIONS(939),
    [sym_continue_statement] = ACTIONS(939),
    [sym_debugger_statement] = ACTIONS(939),
    [anon_sym_assert] = ACTIONS(939),
    [anon_sym_global] = ACTIONS(939),
    [anon_sym_nonlocal] = ACTIONS(939),
    [anon_sym_yield] = ACTIONS(939),
    [anon_sym_from] = ACTIONS(939),
    [anon_sym_import] = ACTIONS(939),
    [anon_sym_DOT] = ACTIONS(939),
    [anon_sym_if] = ACTIONS(939),
    [anon_sym_for] = ACTIONS(939),
    [anon_sym_in] = ACTIONS(939),
    [anon_sym_while] = ACTIONS(939),
    [anon_sym_do] = ACTIONS(939),
    [anon_sym_try] = ACTIONS(939),
    [anon_sym_with] = ACTIONS(939),
    [anon_sym_STAR] = ACTIONS(939),
    [anon_sym_STAR_STAR] = ACTIONS(939),
    [anon_sym_class] = ACTIONS(939),
    [anon_sym_AT] = ACTIONS(939),
    [anon_sym_not] = ACTIONS(939),
    [anon_sym_and] = ACTIONS(939),
    [anon_sym_or] = ACTIONS(939),
    [anon_sym_PLUS] = ACTIONS(939),
    [anon_sym_DASH] = ACTIONS(939),
    [anon_sym_SLASH] = ACTIONS(939),
    [anon_sym_SLASH_SLASH] = ACTIONS(939),
    [anon_sym_PERCENT] = ACTIONS(939),
    [anon_sym_PIPE] = ACTIONS(939),
    [anon_sym_CARET] = ACTIONS(939),
    [anon_sym_AMP] = ACTIONS(939),
    [anon_sym_LT_LT] = ACTIONS(939),
    [anon_sym_GT_GT] = ACTIONS(939),
    [anon_sym_GT_GT_GT] = ACTIONS(939),
    [anon_sym_TILDE] = ACTIONS(939),
    [anon_sym_typeof] = ACTIONS(939),
    [anon_sym_void] = ACTIONS(939),
    [anon_sym_LT] = ACTIONS(939),
    [anon_sym_LT_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ] = ACTIONS(939),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ] = ACTIONS(939),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(939),
    [anon_sym_GT_EQ] = ACTIONS(939),
    [anon_sym_GT] = ACTIONS(939),
    [anon_sym_is] = ACTIONS(939),
    [anon_sym_instanceof] = ACTIONS(939),
    [anon_sym_new] = ACTIONS(939),
    [anon_sym_QMARK] = ACTIONS(939),
    [anon_sym_await] = ACTIONS(939),
    [anon_sym_LBRACE] = ACTIONS(939),
    [sym_this] = ACTIONS(939),
    [sym_true] = ACTIONS(939),
    [sym_false] = ACTIONS(939),
    [sym_none] = ACTIONS(939),
    [sym_string] = ACTIONS(939),
    [sym_verbatim] = ACTIONS(939),
    [sym_number] = ACTIONS(939),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(937),
    [sym_regex] = ACTIONS(937),
    [sym_fstring_start] = ACTIONS(937),
  },
  [STATE(701)] = {
    [sym_identifier] = ACTIONS(1389),
    [anon_sym_SEMI] = ACTIONS(1389),
    [anon_sym_return] = ACTIONS(1389),
    [anon_sym_async] = ACTIONS(1389),
    [anon_sym_def] = ACTIONS(1389),
    [anon_sym_COMMA] = ACTIONS(1389),
    [anon_sym_LPAREN] = ACTIONS(1389),
    [anon_sym_LBRACK] = ACTIONS(1389),
    [anon_sym_del] = ACTIONS(1389),
    [anon_sym_raise] = ACTIONS(1389),
    [sym_pass_statement] = ACTIONS(1389),
    [sym_break_statement] = ACTIONS(1389),
    [sym_continue_statement] = ACTIONS(1389),
    [sym_debugger_statement] = ACTIONS(1389),
    [anon_sym_assert] = ACTIONS(1389),
    [anon_sym_global] = ACTIONS(1389),
    [anon_sym_nonlocal] = ACTIONS(1389),
    [anon_sym_yield] = ACTIONS(1389),
    [anon_sym_from] = ACTIONS(1389),
    [anon_sym_import] = ACTIONS(1389),
    [anon_sym_DOT] = ACTIONS(1389),
    [anon_sym_if] = ACTIONS(1389),
    [anon_sym_for] = ACTIONS(1389),
    [anon_sym_in] = ACTIONS(1389),
    [anon_sym_while] = ACTIONS(1389),
    [anon_sym_do] = ACTIONS(1389),
    [anon_sym_try] = ACTIONS(1389),
    [anon_sym_with] = ACTIONS(1389),
    [anon_sym_STAR] = ACTIONS(1389),
    [anon_sym_STAR_STAR] = ACTIONS(1389),
    [anon_sym_class] = ACTIONS(1389),
    [anon_sym_AT] = ACTIONS(1389),
    [anon_sym_not] = ACTIONS(1389),
    [anon_sym_and] = ACTIONS(1389),
    [anon_sym_or] = ACTIONS(1389),
    [anon_sym_PLUS] = ACTIONS(1389),
    [anon_sym_DASH] = ACTIONS(1389),
    [anon_sym_SLASH] = ACTIONS(1389),
    [anon_sym_SLASH_SLASH] = ACTIONS(1389),
    [anon_sym_PERCENT] = ACTIONS(1389),
    [anon_sym_PIPE] = ACTIONS(1389),
    [anon_sym_CARET] = ACTIONS(1389),
    [anon_sym_AMP] = ACTIONS(1389),
    [anon_sym_LT_LT] = ACTIONS(1389),
    [anon_sym_GT_GT] = ACTIONS(1389),
    [anon_sym_GT_GT_GT] = ACTIONS(1389),
    [anon_sym_TILDE] = ACTIONS(1389),
    [anon_sym_typeof] = ACTIONS(1389),
    [anon_sym_void] = ACTIONS(1389),
    [anon_sym_LT] = ACTIONS(1389),
    [anon_sym_LT_EQ] = ACTIONS(1389),
    [anon_sym_EQ_EQ] = ACTIONS(1389),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1389),
    [anon_sym_BANG_EQ] = ACTIONS(1389),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1389),
    [anon_sym_GT_EQ] = ACTIONS(1389),
    [anon_sym_GT] = ACTIONS(1389),
    [anon_sym_is] = ACTIONS(1389),
    [anon_sym_instanceof] = ACTIONS(1389),
    [anon_sym_new] = ACTIONS(1389),
    [anon_sym_QMARK] = ACTIONS(1389),
    [anon_sym_await] = ACTIONS(1389),
    [anon_sym_LBRACE] = ACTIONS(1389),
    [sym_this] = ACTIONS(1389),
    [sym_true] = ACTIONS(1389),
    [sym_false] = ACTIONS(1389),
    [sym_none] = ACTIONS(1389),
    [sym_string] = ACTIONS(1389),
    [sym_verbatim] = ACTIONS(1389),
    [sym_number] = ACTIONS(1389),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1387),
    [sym_regex] = ACTIONS(1387),
    [sym_fstring_start] = ACTIONS(1387),
  },
  [STATE(702)] = {
    [ts_builtin_sym_end] = ACTIONS(1101),
    [sym_identifier] = ACTIONS(1072),
    [anon_sym_SEMI] = ACTIONS(1072),
    [anon_sym_return] = ACTIONS(1072),
    [anon_sym_async] = ACTIONS(1072),
    [anon_sym_def] = ACTIONS(1072),
    [anon_sym_COMMA] = ACTIONS(1072),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_del] = ACTIONS(1072),
    [anon_sym_raise] = ACTIONS(1072),
    [sym_pass_statement] = ACTIONS(1072),
    [sym_break_statement] = ACTIONS(1072),
    [sym_continue_statement] = ACTIONS(1072),
    [sym_debugger_statement] = ACTIONS(1072),
    [anon_sym_assert] = ACTIONS(1072),
    [anon_sym_global] = ACTIONS(1072),
    [anon_sym_nonlocal] = ACTIONS(1072),
    [anon_sym_yield] = ACTIONS(1072),
    [anon_sym_from] = ACTIONS(1072),
    [anon_sym_import] = ACTIONS(1072),
    [anon_sym_DOT] = ACTIONS(1072),
    [anon_sym_if] = ACTIONS(1072),
    [anon_sym_for] = ACTIONS(1072),
    [anon_sym_in] = ACTIONS(1072),
    [anon_sym_while] = ACTIONS(1072),
    [anon_sym_do] = ACTIONS(1072),
    [anon_sym_try] = ACTIONS(1072),
    [anon_sym_with] = ACTIONS(1072),
    [anon_sym_STAR] = ACTIONS(1072),
    [anon_sym_STAR_STAR] = ACTIONS(1072),
    [anon_sym_class] = ACTIONS(1072),
    [anon_sym_AT] = ACTIONS(1072),
    [anon_sym_not] = ACTIONS(1072),
    [anon_sym_and] = ACTIONS(1072),
    [anon_sym_or] = ACTIONS(1072),
    [anon_sym_PLUS] = ACTIONS(1072),
    [anon_sym_DASH] = ACTIONS(1072),
    [anon_sym_SLASH] = ACTIONS(1072),
    [anon_sym_SLASH_SLASH] = ACTIONS(1072),
    [anon_sym_PERCENT] = ACTIONS(1072),
    [anon_sym_PIPE] = ACTIONS(1072),
    [anon_sym_CARET] = ACTIONS(1072),
    [anon_sym_AMP] = ACTIONS(1072),
    [anon_sym_LT_LT] = ACTIONS(1072),
    [anon_sym_GT_GT] = ACTIONS(1072),
    [anon_sym_GT_GT_GT] = ACTIONS(1072),
    [anon_sym_TILDE] = ACTIONS(1072),
    [anon_sym_typeof] = ACTIONS(1072),
    [anon_sym_void] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(1072),
    [anon_sym_LT_EQ] = ACTIONS(1072),
    [anon_sym_EQ_EQ] = ACTIONS(1072),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1072),
    [anon_sym_BANG_EQ] = ACTIONS(1072),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1072),
    [anon_sym_GT_EQ] = ACTIONS(1072),
    [anon_sym_GT] = ACTIONS(1072),
    [anon_sym_is] = ACTIONS(1072),
    [anon_sym_instanceof] = ACTIONS(1072),
    [anon_sym_new] = ACTIONS(1072),
    [anon_sym_QMARK] = ACTIONS(1072),
    [anon_sym_await] = ACTIONS(1072),
    [anon_sym_LBRACE] = ACTIONS(1072),
    [sym_this] = ACTIONS(1072),
    [sym_true] = ACTIONS(1072),
    [sym_false] = ACTIONS(1072),
    [sym_none] = ACTIONS(1072),
    [sym_string] = ACTIONS(1072),
    [sym_verbatim] = ACTIONS(1072),
    [sym_number] = ACTIONS(1072),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1101),
    [sym_fstring_start] = ACTIONS(1101),
  },
  [STATE(703)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(704)] = {
    [sym_identifier] = ACTIONS(1617),
    [anon_sym_SEMI] = ACTIONS(1617),
    [anon_sym_return] = ACTIONS(1617),
    [anon_sym_async] = ACTIONS(1617),
    [anon_sym_def] = ACTIONS(1617),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_LBRACK] = ACTIONS(1617),
    [anon_sym_del] = ACTIONS(1617),
    [anon_sym_raise] = ACTIONS(1617),
    [sym_pass_statement] = ACTIONS(1617),
    [sym_break_statement] = ACTIONS(1617),
    [sym_continue_statement] = ACTIONS(1617),
    [sym_debugger_statement] = ACTIONS(1617),
    [anon_sym_assert] = ACTIONS(1617),
    [anon_sym_global] = ACTIONS(1617),
    [anon_sym_nonlocal] = ACTIONS(1617),
    [anon_sym_yield] = ACTIONS(1617),
    [anon_sym_from] = ACTIONS(1617),
    [anon_sym_import] = ACTIONS(1617),
    [anon_sym_DOT] = ACTIONS(1617),
    [anon_sym_if] = ACTIONS(1617),
    [anon_sym_for] = ACTIONS(1617),
    [anon_sym_in] = ACTIONS(1617),
    [anon_sym_while] = ACTIONS(1617),
    [anon_sym_do] = ACTIONS(1617),
    [anon_sym_try] = ACTIONS(1617),
    [anon_sym_with] = ACTIONS(1617),
    [anon_sym_STAR] = ACTIONS(1617),
    [anon_sym_STAR_STAR] = ACTIONS(1617),
    [anon_sym_class] = ACTIONS(1617),
    [anon_sym_AT] = ACTIONS(1617),
    [anon_sym_not] = ACTIONS(1617),
    [anon_sym_and] = ACTIONS(1617),
    [anon_sym_or] = ACTIONS(1617),
    [anon_sym_PLUS] = ACTIONS(1617),
    [anon_sym_DASH] = ACTIONS(1617),
    [anon_sym_SLASH] = ACTIONS(1617),
    [anon_sym_SLASH_SLASH] = ACTIONS(1617),
    [anon_sym_PERCENT] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1617),
    [anon_sym_CARET] = ACTIONS(1617),
    [anon_sym_AMP] = ACTIONS(1617),
    [anon_sym_LT_LT] = ACTIONS(1617),
    [anon_sym_GT_GT] = ACTIONS(1617),
    [anon_sym_GT_GT_GT] = ACTIONS(1617),
    [anon_sym_TILDE] = ACTIONS(1617),
    [anon_sym_typeof] = ACTIONS(1617),
    [anon_sym_void] = ACTIONS(1617),
    [anon_sym_LT] = ACTIONS(1617),
    [anon_sym_LT_EQ] = ACTIONS(1617),
    [anon_sym_EQ_EQ] = ACTIONS(1617),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1617),
    [anon_sym_BANG_EQ] = ACTIONS(1617),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1617),
    [anon_sym_GT_EQ] = ACTIONS(1617),
    [anon_sym_GT] = ACTIONS(1617),
    [anon_sym_is] = ACTIONS(1617),
    [anon_sym_instanceof] = ACTIONS(1617),
    [anon_sym_new] = ACTIONS(1617),
    [anon_sym_QMARK] = ACTIONS(1617),
    [anon_sym_await] = ACTIONS(1617),
    [anon_sym_LBRACE] = ACTIONS(1617),
    [sym_this] = ACTIONS(1617),
    [sym_true] = ACTIONS(1617),
    [sym_false] = ACTIONS(1617),
    [sym_none] = ACTIONS(1617),
    [sym_string] = ACTIONS(1617),
    [sym_verbatim] = ACTIONS(1617),
    [sym_number] = ACTIONS(1617),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1615),
    [sym_regex] = ACTIONS(1615),
    [sym_fstring_start] = ACTIONS(1615),
  },
  [STATE(705)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1222),
    [sym_identifier] = ACTIONS(1224),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym_return] = ACTIONS(1224),
    [anon_sym_async] = ACTIONS(1224),
    [anon_sym_def] = ACTIONS(1224),
    [anon_sym_LPAREN] = ACTIONS(1224),
    [anon_sym_LBRACK] = ACTIONS(1224),
    [anon_sym_del] = ACTIONS(1224),
    [anon_sym_raise] = ACTIONS(1224),
    [sym_pass_statement] = ACTIONS(1224),
    [sym_break_statement] = ACTIONS(1224),
    [sym_continue_statement] = ACTIONS(1224),
    [sym_debugger_statement] = ACTIONS(1224),
    [anon_sym_assert] = ACTIONS(1224),
    [anon_sym_global] = ACTIONS(1224),
    [anon_sym_nonlocal] = ACTIONS(1224),
    [anon_sym_yield] = ACTIONS(1224),
    [anon_sym_from] = ACTIONS(1224),
    [anon_sym_import] = ACTIONS(1224),
    [anon_sym_DOT] = ACTIONS(1224),
    [anon_sym_if] = ACTIONS(1224),
    [anon_sym_for] = ACTIONS(1224),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1224),
    [anon_sym_do] = ACTIONS(1224),
    [anon_sym_try] = ACTIONS(1224),
    [anon_sym_with] = ACTIONS(1224),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1224),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1494),
    [anon_sym_or] = ACTIONS(1496),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1224),
    [anon_sym_typeof] = ACTIONS(1224),
    [anon_sym_void] = ACTIONS(1224),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1224),
    [anon_sym_QMARK] = ACTIONS(1224),
    [anon_sym_await] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [sym_this] = ACTIONS(1224),
    [sym_true] = ACTIONS(1224),
    [sym_false] = ACTIONS(1224),
    [sym_none] = ACTIONS(1224),
    [sym_string] = ACTIONS(1224),
    [sym_verbatim] = ACTIONS(1224),
    [sym_number] = ACTIONS(1224),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1222),
    [sym_fstring_start] = ACTIONS(1222),
  },
  [STATE(706)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1226),
    [sym_identifier] = ACTIONS(1228),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_return] = ACTIONS(1228),
    [anon_sym_async] = ACTIONS(1228),
    [anon_sym_def] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_del] = ACTIONS(1228),
    [anon_sym_raise] = ACTIONS(1228),
    [sym_pass_statement] = ACTIONS(1228),
    [sym_break_statement] = ACTIONS(1228),
    [sym_continue_statement] = ACTIONS(1228),
    [sym_debugger_statement] = ACTIONS(1228),
    [anon_sym_assert] = ACTIONS(1228),
    [anon_sym_global] = ACTIONS(1228),
    [anon_sym_nonlocal] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1228),
    [anon_sym_from] = ACTIONS(1228),
    [anon_sym_import] = ACTIONS(1228),
    [anon_sym_DOT] = ACTIONS(1228),
    [anon_sym_if] = ACTIONS(1228),
    [anon_sym_for] = ACTIONS(1228),
    [anon_sym_in] = ACTIONS(1228),
    [anon_sym_while] = ACTIONS(1228),
    [anon_sym_do] = ACTIONS(1228),
    [anon_sym_try] = ACTIONS(1228),
    [anon_sym_with] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1228),
    [anon_sym_AT] = ACTIONS(1228),
    [anon_sym_not] = ACTIONS(1228),
    [anon_sym_and] = ACTIONS(1228),
    [anon_sym_or] = ACTIONS(1228),
    [anon_sym_PLUS] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_SLASH] = ACTIONS(1228),
    [anon_sym_SLASH_SLASH] = ACTIONS(1228),
    [anon_sym_PERCENT] = ACTIONS(1228),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_CARET] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_LT_LT] = ACTIONS(1228),
    [anon_sym_GT_GT] = ACTIONS(1228),
    [anon_sym_GT_GT_GT] = ACTIONS(1228),
    [anon_sym_TILDE] = ACTIONS(1228),
    [anon_sym_typeof] = ACTIONS(1228),
    [anon_sym_void] = ACTIONS(1228),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_LT_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ] = ACTIONS(1228),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ] = ACTIONS(1228),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1228),
    [anon_sym_GT_EQ] = ACTIONS(1228),
    [anon_sym_GT] = ACTIONS(1228),
    [anon_sym_is] = ACTIONS(1228),
    [anon_sym_instanceof] = ACTIONS(1228),
    [anon_sym_new] = ACTIONS(1228),
    [anon_sym_QMARK] = ACTIONS(1228),
    [anon_sym_await] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [sym_this] = ACTIONS(1228),
    [sym_true] = ACTIONS(1228),
    [sym_false] = ACTIONS(1228),
    [sym_none] = ACTIONS(1228),
    [sym_string] = ACTIONS(1228),
    [sym_verbatim] = ACTIONS(1228),
    [sym_number] = ACTIONS(1228),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1226),
    [sym_fstring_start] = ACTIONS(1226),
  },
  [STATE(707)] = {
    [sym_identifier] = ACTIONS(1480),
    [anon_sym_SEMI] = ACTIONS(1480),
    [anon_sym_return] = ACTIONS(1480),
    [anon_sym_async] = ACTIONS(1480),
    [anon_sym_def] = ACTIONS(1480),
    [anon_sym_COMMA] = ACTIONS(1480),
    [anon_sym_LPAREN] = ACTIONS(1480),
    [anon_sym_LBRACK] = ACTIONS(1480),
    [anon_sym_del] = ACTIONS(1480),
    [anon_sym_raise] = ACTIONS(1480),
    [sym_pass_statement] = ACTIONS(1480),
    [sym_break_statement] = ACTIONS(1480),
    [sym_continue_statement] = ACTIONS(1480),
    [sym_debugger_statement] = ACTIONS(1480),
    [anon_sym_assert] = ACTIONS(1480),
    [anon_sym_global] = ACTIONS(1480),
    [anon_sym_nonlocal] = ACTIONS(1480),
    [anon_sym_yield] = ACTIONS(1480),
    [anon_sym_from] = ACTIONS(1480),
    [anon_sym_import] = ACTIONS(1480),
    [anon_sym_DOT] = ACTIONS(1480),
    [anon_sym_if] = ACTIONS(1480),
    [anon_sym_for] = ACTIONS(1480),
    [anon_sym_in] = ACTIONS(1480),
    [anon_sym_while] = ACTIONS(1480),
    [anon_sym_do] = ACTIONS(1480),
    [anon_sym_try] = ACTIONS(1480),
    [anon_sym_with] = ACTIONS(1480),
    [anon_sym_STAR] = ACTIONS(1480),
    [anon_sym_STAR_STAR] = ACTIONS(1480),
    [anon_sym_class] = ACTIONS(1480),
    [anon_sym_AT] = ACTIONS(1480),
    [anon_sym_not] = ACTIONS(1480),
    [anon_sym_and] = ACTIONS(1480),
    [anon_sym_or] = ACTIONS(1480),
    [anon_sym_PLUS] = ACTIONS(1480),
    [anon_sym_DASH] = ACTIONS(1480),
    [anon_sym_SLASH] = ACTIONS(1480),
    [anon_sym_SLASH_SLASH] = ACTIONS(1480),
    [anon_sym_PERCENT] = ACTIONS(1480),
    [anon_sym_PIPE] = ACTIONS(1480),
    [anon_sym_CARET] = ACTIONS(1480),
    [anon_sym_AMP] = ACTIONS(1480),
    [anon_sym_LT_LT] = ACTIONS(1480),
    [anon_sym_GT_GT] = ACTIONS(1480),
    [anon_sym_GT_GT_GT] = ACTIONS(1480),
    [anon_sym_TILDE] = ACTIONS(1480),
    [anon_sym_typeof] = ACTIONS(1480),
    [anon_sym_void] = ACTIONS(1480),
    [anon_sym_LT] = ACTIONS(1480),
    [anon_sym_LT_EQ] = ACTIONS(1480),
    [anon_sym_EQ_EQ] = ACTIONS(1480),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1480),
    [anon_sym_BANG_EQ] = ACTIONS(1480),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1480),
    [anon_sym_GT_EQ] = ACTIONS(1480),
    [anon_sym_GT] = ACTIONS(1480),
    [anon_sym_is] = ACTIONS(1480),
    [anon_sym_instanceof] = ACTIONS(1480),
    [anon_sym_new] = ACTIONS(1480),
    [anon_sym_QMARK] = ACTIONS(1480),
    [anon_sym_await] = ACTIONS(1480),
    [anon_sym_LBRACE] = ACTIONS(1480),
    [sym_this] = ACTIONS(1480),
    [sym_true] = ACTIONS(1480),
    [sym_false] = ACTIONS(1480),
    [sym_none] = ACTIONS(1480),
    [sym_string] = ACTIONS(1480),
    [sym_verbatim] = ACTIONS(1480),
    [sym_number] = ACTIONS(1480),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1478),
    [sym_regex] = ACTIONS(1478),
    [sym_fstring_start] = ACTIONS(1478),
  },
  [STATE(708)] = {
    [sym_identifier] = ACTIONS(1621),
    [anon_sym_SEMI] = ACTIONS(1621),
    [anon_sym_return] = ACTIONS(1621),
    [anon_sym_async] = ACTIONS(1621),
    [anon_sym_def] = ACTIONS(1621),
    [anon_sym_COMMA] = ACTIONS(1621),
    [anon_sym_LPAREN] = ACTIONS(1621),
    [anon_sym_LBRACK] = ACTIONS(1621),
    [anon_sym_del] = ACTIONS(1621),
    [anon_sym_raise] = ACTIONS(1621),
    [sym_pass_statement] = ACTIONS(1621),
    [sym_break_statement] = ACTIONS(1621),
    [sym_continue_statement] = ACTIONS(1621),
    [sym_debugger_statement] = ACTIONS(1621),
    [anon_sym_assert] = ACTIONS(1621),
    [anon_sym_global] = ACTIONS(1621),
    [anon_sym_nonlocal] = ACTIONS(1621),
    [anon_sym_yield] = ACTIONS(1621),
    [anon_sym_from] = ACTIONS(1621),
    [anon_sym_import] = ACTIONS(1621),
    [anon_sym_DOT] = ACTIONS(1621),
    [anon_sym_if] = ACTIONS(1621),
    [anon_sym_for] = ACTIONS(1621),
    [anon_sym_in] = ACTIONS(1621),
    [anon_sym_while] = ACTIONS(1621),
    [anon_sym_do] = ACTIONS(1621),
    [anon_sym_try] = ACTIONS(1621),
    [anon_sym_with] = ACTIONS(1621),
    [anon_sym_STAR] = ACTIONS(1621),
    [anon_sym_STAR_STAR] = ACTIONS(1621),
    [anon_sym_class] = ACTIONS(1621),
    [anon_sym_AT] = ACTIONS(1621),
    [anon_sym_not] = ACTIONS(1621),
    [anon_sym_and] = ACTIONS(1621),
    [anon_sym_or] = ACTIONS(1621),
    [anon_sym_PLUS] = ACTIONS(1621),
    [anon_sym_DASH] = ACTIONS(1621),
    [anon_sym_SLASH] = ACTIONS(1621),
    [anon_sym_SLASH_SLASH] = ACTIONS(1621),
    [anon_sym_PERCENT] = ACTIONS(1621),
    [anon_sym_PIPE] = ACTIONS(1621),
    [anon_sym_CARET] = ACTIONS(1621),
    [anon_sym_AMP] = ACTIONS(1621),
    [anon_sym_LT_LT] = ACTIONS(1621),
    [anon_sym_GT_GT] = ACTIONS(1621),
    [anon_sym_GT_GT_GT] = ACTIONS(1621),
    [anon_sym_TILDE] = ACTIONS(1621),
    [anon_sym_typeof] = ACTIONS(1621),
    [anon_sym_void] = ACTIONS(1621),
    [anon_sym_LT] = ACTIONS(1621),
    [anon_sym_LT_EQ] = ACTIONS(1621),
    [anon_sym_EQ_EQ] = ACTIONS(1621),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1621),
    [anon_sym_BANG_EQ] = ACTIONS(1621),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1621),
    [anon_sym_GT_EQ] = ACTIONS(1621),
    [anon_sym_GT] = ACTIONS(1621),
    [anon_sym_is] = ACTIONS(1621),
    [anon_sym_instanceof] = ACTIONS(1621),
    [anon_sym_new] = ACTIONS(1621),
    [anon_sym_QMARK] = ACTIONS(1621),
    [anon_sym_await] = ACTIONS(1621),
    [anon_sym_LBRACE] = ACTIONS(1621),
    [sym_this] = ACTIONS(1621),
    [sym_true] = ACTIONS(1621),
    [sym_false] = ACTIONS(1621),
    [sym_none] = ACTIONS(1621),
    [sym_string] = ACTIONS(1621),
    [sym_verbatim] = ACTIONS(1621),
    [sym_number] = ACTIONS(1621),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1619),
    [sym_regex] = ACTIONS(1619),
    [sym_fstring_start] = ACTIONS(1619),
  },
  [STATE(709)] = {
    [sym_identifier] = ACTIONS(1631),
    [anon_sym_SEMI] = ACTIONS(1631),
    [anon_sym_return] = ACTIONS(1631),
    [anon_sym_async] = ACTIONS(1631),
    [anon_sym_def] = ACTIONS(1631),
    [anon_sym_COMMA] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(1631),
    [anon_sym_LBRACK] = ACTIONS(1631),
    [anon_sym_del] = ACTIONS(1631),
    [anon_sym_raise] = ACTIONS(1631),
    [sym_pass_statement] = ACTIONS(1631),
    [sym_break_statement] = ACTIONS(1631),
    [sym_continue_statement] = ACTIONS(1631),
    [sym_debugger_statement] = ACTIONS(1631),
    [anon_sym_assert] = ACTIONS(1631),
    [anon_sym_global] = ACTIONS(1631),
    [anon_sym_nonlocal] = ACTIONS(1631),
    [anon_sym_yield] = ACTIONS(1631),
    [anon_sym_from] = ACTIONS(1631),
    [anon_sym_import] = ACTIONS(1631),
    [anon_sym_DOT] = ACTIONS(1631),
    [anon_sym_if] = ACTIONS(1631),
    [anon_sym_for] = ACTIONS(1631),
    [anon_sym_in] = ACTIONS(1631),
    [anon_sym_while] = ACTIONS(1631),
    [anon_sym_do] = ACTIONS(1631),
    [anon_sym_try] = ACTIONS(1631),
    [anon_sym_with] = ACTIONS(1631),
    [anon_sym_STAR] = ACTIONS(1631),
    [anon_sym_STAR_STAR] = ACTIONS(1631),
    [anon_sym_class] = ACTIONS(1631),
    [anon_sym_AT] = ACTIONS(1631),
    [anon_sym_not] = ACTIONS(1631),
    [anon_sym_and] = ACTIONS(1631),
    [anon_sym_or] = ACTIONS(1631),
    [anon_sym_PLUS] = ACTIONS(1631),
    [anon_sym_DASH] = ACTIONS(1631),
    [anon_sym_SLASH] = ACTIONS(1631),
    [anon_sym_SLASH_SLASH] = ACTIONS(1631),
    [anon_sym_PERCENT] = ACTIONS(1631),
    [anon_sym_PIPE] = ACTIONS(1631),
    [anon_sym_CARET] = ACTIONS(1631),
    [anon_sym_AMP] = ACTIONS(1631),
    [anon_sym_LT_LT] = ACTIONS(1631),
    [anon_sym_GT_GT] = ACTIONS(1631),
    [anon_sym_GT_GT_GT] = ACTIONS(1631),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_typeof] = ACTIONS(1631),
    [anon_sym_void] = ACTIONS(1631),
    [anon_sym_LT] = ACTIONS(1631),
    [anon_sym_LT_EQ] = ACTIONS(1631),
    [anon_sym_EQ_EQ] = ACTIONS(1631),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1631),
    [anon_sym_BANG_EQ] = ACTIONS(1631),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1631),
    [anon_sym_GT_EQ] = ACTIONS(1631),
    [anon_sym_GT] = ACTIONS(1631),
    [anon_sym_is] = ACTIONS(1631),
    [anon_sym_instanceof] = ACTIONS(1631),
    [anon_sym_new] = ACTIONS(1631),
    [anon_sym_QMARK] = ACTIONS(1631),
    [anon_sym_await] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1631),
    [sym_this] = ACTIONS(1631),
    [sym_true] = ACTIONS(1631),
    [sym_false] = ACTIONS(1631),
    [sym_none] = ACTIONS(1631),
    [sym_string] = ACTIONS(1631),
    [sym_verbatim] = ACTIONS(1631),
    [sym_number] = ACTIONS(1631),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1629),
    [sym_regex] = ACTIONS(1629),
    [sym_fstring_start] = ACTIONS(1629),
  },
  [STATE(710)] = {
    [sym_identifier] = ACTIONS(1540),
    [anon_sym_SEMI] = ACTIONS(1540),
    [anon_sym_return] = ACTIONS(1540),
    [anon_sym_async] = ACTIONS(1540),
    [anon_sym_def] = ACTIONS(1540),
    [anon_sym_COMMA] = ACTIONS(1540),
    [anon_sym_LPAREN] = ACTIONS(1540),
    [anon_sym_LBRACK] = ACTIONS(1540),
    [anon_sym_del] = ACTIONS(1540),
    [anon_sym_raise] = ACTIONS(1540),
    [sym_pass_statement] = ACTIONS(1540),
    [sym_break_statement] = ACTIONS(1540),
    [sym_continue_statement] = ACTIONS(1540),
    [sym_debugger_statement] = ACTIONS(1540),
    [anon_sym_assert] = ACTIONS(1540),
    [anon_sym_global] = ACTIONS(1540),
    [anon_sym_nonlocal] = ACTIONS(1540),
    [anon_sym_yield] = ACTIONS(1540),
    [anon_sym_from] = ACTIONS(1540),
    [anon_sym_import] = ACTIONS(1540),
    [anon_sym_DOT] = ACTIONS(1540),
    [anon_sym_if] = ACTIONS(1540),
    [anon_sym_for] = ACTIONS(1540),
    [anon_sym_in] = ACTIONS(1540),
    [anon_sym_while] = ACTIONS(1540),
    [anon_sym_do] = ACTIONS(1540),
    [anon_sym_try] = ACTIONS(1540),
    [anon_sym_with] = ACTIONS(1540),
    [anon_sym_STAR] = ACTIONS(1540),
    [anon_sym_STAR_STAR] = ACTIONS(1540),
    [anon_sym_class] = ACTIONS(1540),
    [anon_sym_AT] = ACTIONS(1540),
    [anon_sym_not] = ACTIONS(1540),
    [anon_sym_and] = ACTIONS(1540),
    [anon_sym_or] = ACTIONS(1540),
    [anon_sym_PLUS] = ACTIONS(1540),
    [anon_sym_DASH] = ACTIONS(1540),
    [anon_sym_SLASH] = ACTIONS(1540),
    [anon_sym_SLASH_SLASH] = ACTIONS(1540),
    [anon_sym_PERCENT] = ACTIONS(1540),
    [anon_sym_PIPE] = ACTIONS(1540),
    [anon_sym_CARET] = ACTIONS(1540),
    [anon_sym_AMP] = ACTIONS(1540),
    [anon_sym_LT_LT] = ACTIONS(1540),
    [anon_sym_GT_GT] = ACTIONS(1540),
    [anon_sym_GT_GT_GT] = ACTIONS(1540),
    [anon_sym_TILDE] = ACTIONS(1540),
    [anon_sym_typeof] = ACTIONS(1540),
    [anon_sym_void] = ACTIONS(1540),
    [anon_sym_LT] = ACTIONS(1540),
    [anon_sym_LT_EQ] = ACTIONS(1540),
    [anon_sym_EQ_EQ] = ACTIONS(1540),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1540),
    [anon_sym_BANG_EQ] = ACTIONS(1540),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1540),
    [anon_sym_GT_EQ] = ACTIONS(1540),
    [anon_sym_GT] = ACTIONS(1540),
    [anon_sym_is] = ACTIONS(1540),
    [anon_sym_instanceof] = ACTIONS(1540),
    [anon_sym_new] = ACTIONS(1540),
    [anon_sym_QMARK] = ACTIONS(1540),
    [anon_sym_await] = ACTIONS(1540),
    [anon_sym_LBRACE] = ACTIONS(1540),
    [sym_this] = ACTIONS(1540),
    [sym_true] = ACTIONS(1540),
    [sym_false] = ACTIONS(1540),
    [sym_none] = ACTIONS(1540),
    [sym_string] = ACTIONS(1540),
    [sym_verbatim] = ACTIONS(1540),
    [sym_number] = ACTIONS(1540),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1538),
    [sym_regex] = ACTIONS(1538),
    [sym_fstring_start] = ACTIONS(1538),
  },
  [STATE(711)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(676),
    [sym_identifier] = ACTIONS(1637),
    [anon_sym_return] = ACTIONS(1637),
    [anon_sym_async] = ACTIONS(1637),
    [anon_sym_def] = ACTIONS(1637),
    [anon_sym_LPAREN] = ACTIONS(1637),
    [anon_sym_LBRACK] = ACTIONS(1637),
    [anon_sym_del] = ACTIONS(1637),
    [anon_sym_raise] = ACTIONS(1637),
    [sym_pass_statement] = ACTIONS(1637),
    [sym_break_statement] = ACTIONS(1637),
    [sym_continue_statement] = ACTIONS(1637),
    [sym_debugger_statement] = ACTIONS(1637),
    [anon_sym_assert] = ACTIONS(1637),
    [anon_sym_global] = ACTIONS(1637),
    [anon_sym_nonlocal] = ACTIONS(1637),
    [anon_sym_yield] = ACTIONS(1637),
    [anon_sym_from] = ACTIONS(1637),
    [anon_sym_import] = ACTIONS(1637),
    [anon_sym_if] = ACTIONS(1637),
    [anon_sym_for] = ACTIONS(1637),
    [anon_sym_in] = ACTIONS(1417),
    [anon_sym_while] = ACTIONS(1637),
    [anon_sym_do] = ACTIONS(1637),
    [anon_sym_try] = ACTIONS(1637),
    [anon_sym_with] = ACTIONS(1637),
    [anon_sym_STAR] = ACTIONS(1419),
    [anon_sym_STAR_STAR] = ACTIONS(1421),
    [anon_sym_class] = ACTIONS(1637),
    [anon_sym_AT] = ACTIONS(1419),
    [anon_sym_not] = ACTIONS(1423),
    [anon_sym_and] = ACTIONS(1425),
    [anon_sym_or] = ACTIONS(1427),
    [anon_sym_PLUS] = ACTIONS(1429),
    [anon_sym_DASH] = ACTIONS(1429),
    [anon_sym_SLASH] = ACTIONS(1419),
    [anon_sym_SLASH_SLASH] = ACTIONS(1419),
    [anon_sym_PERCENT] = ACTIONS(1419),
    [anon_sym_PIPE] = ACTIONS(1431),
    [anon_sym_CARET] = ACTIONS(1433),
    [anon_sym_AMP] = ACTIONS(1435),
    [anon_sym_LT_LT] = ACTIONS(1437),
    [anon_sym_GT_GT] = ACTIONS(1437),
    [anon_sym_GT_GT_GT] = ACTIONS(1437),
    [anon_sym_TILDE] = ACTIONS(1637),
    [anon_sym_typeof] = ACTIONS(1637),
    [anon_sym_void] = ACTIONS(1637),
    [anon_sym_LT] = ACTIONS(1417),
    [anon_sym_LT_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ] = ACTIONS(1417),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ] = ACTIONS(1417),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1417),
    [anon_sym_GT_EQ] = ACTIONS(1417),
    [anon_sym_GT] = ACTIONS(1417),
    [anon_sym_is] = ACTIONS(1439),
    [anon_sym_instanceof] = ACTIONS(1417),
    [anon_sym_new] = ACTIONS(1637),
    [anon_sym_await] = ACTIONS(1637),
    [anon_sym_LBRACE] = ACTIONS(1637),
    [sym_this] = ACTIONS(1637),
    [sym_true] = ACTIONS(1637),
    [sym_false] = ACTIONS(1637),
    [sym_none] = ACTIONS(1637),
    [sym_string] = ACTIONS(1637),
    [sym_verbatim] = ACTIONS(1637),
    [sym_number] = ACTIONS(1637),
    [sym_comment] = ACTIONS(3),
    [sym__dedent] = ACTIONS(1639),
    [sym_regex] = ACTIONS(1639),
    [sym_fstring_start] = ACTIONS(1639),
  },
  [STATE(712)] = {
    [aux_sym_comparison_operator_repeat1] = STATE(690),
    [ts_builtin_sym_end] = ACTIONS(1639),
    [sym_identifier] = ACTIONS(1637),
    [anon_sym_return] = ACTIONS(1637),
    [anon_sym_async] = ACTIONS(1637),
    [anon_sym_def] = ACTIONS(1637),
    [anon_sym_LPAREN] = ACTIONS(1637),
    [anon_sym_LBRACK] = ACTIONS(1637),
    [anon_sym_del] = ACTIONS(1637),
    [anon_sym_raise] = ACTIONS(1637),
    [sym_pass_statement] = ACTIONS(1637),
    [sym_break_statement] = ACTIONS(1637),
    [sym_continue_statement] = ACTIONS(1637),
    [sym_debugger_statement] = ACTIONS(1637),
    [anon_sym_assert] = ACTIONS(1637),
    [anon_sym_global] = ACTIONS(1637),
    [anon_sym_nonlocal] = ACTIONS(1637),
    [anon_sym_yield] = ACTIONS(1637),
    [anon_sym_from] = ACTIONS(1637),
    [anon_sym_import] = ACTIONS(1637),
    [anon_sym_if] = ACTIONS(1637),
    [anon_sym_for] = ACTIONS(1637),
    [anon_sym_in] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1637),
    [anon_sym_do] = ACTIONS(1637),
    [anon_sym_try] = ACTIONS(1637),
    [anon_sym_with] = ACTIONS(1637),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_STAR_STAR] = ACTIONS(1490),
    [anon_sym_class] = ACTIONS(1637),
    [anon_sym_AT] = ACTIONS(1488),
    [anon_sym_not] = ACTIONS(1492),
    [anon_sym_and] = ACTIONS(1494),
    [anon_sym_or] = ACTIONS(1496),
    [anon_sym_PLUS] = ACTIONS(1498),
    [anon_sym_DASH] = ACTIONS(1498),
    [anon_sym_SLASH] = ACTIONS(1488),
    [anon_sym_SLASH_SLASH] = ACTIONS(1488),
    [anon_sym_PERCENT] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_CARET] = ACTIONS(1502),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_LT_LT] = ACTIONS(1506),
    [anon_sym_GT_GT] = ACTIONS(1506),
    [anon_sym_GT_GT_GT] = ACTIONS(1506),
    [anon_sym_TILDE] = ACTIONS(1637),
    [anon_sym_typeof] = ACTIONS(1637),
    [anon_sym_void] = ACTIONS(1637),
    [anon_sym_LT] = ACTIONS(1486),
    [anon_sym_LT_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ] = ACTIONS(1486),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ] = ACTIONS(1486),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1486),
    [anon_sym_GT_EQ] = ACTIONS(1486),
    [anon_sym_GT] = ACTIONS(1486),
    [anon_sym_is] = ACTIONS(1508),
    [anon_sym_instanceof] = ACTIONS(1486),
    [anon_sym_new] = ACTIONS(1637),
    [anon_sym_await] = ACTIONS(1637),
    [anon_sym_LBRACE] = ACTIONS(1637),
    [sym_this] = ACTIONS(1637),
    [sym_true] = ACTIONS(1637),
    [sym_false] = ACTIONS(1637),
    [sym_none] = ACTIONS(1637),
    [sym_string] = ACTIONS(1637),
    [sym_verbatim] = ACTIONS(1637),
    [sym_number] = ACTIONS(1637),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(1639),
    [sym_fstring_start] = ACTIONS(1639),
  },
  [STATE(713)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(3017),
    [sym__expression] = STATE(2321),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2729),
    [sym_keyword_argument] = STATE(2729),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2729),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2864),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1643),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(714)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2980),
    [sym__expression] = STATE(2313),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2702),
    [sym_keyword_argument] = STATE(2702),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2702),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2931),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1651),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(715)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2325),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2694),
    [sym_keyword_argument] = STATE(2694),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2694),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1653),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(716)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2980),
    [sym__expression] = STATE(2313),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2702),
    [sym_keyword_argument] = STATE(2702),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2702),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2931),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1655),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(717)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(3047),
    [sym__expression] = STATE(2300),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2711),
    [sym_keyword_argument] = STATE(2711),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2711),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2960),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1657),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(718)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(3047),
    [sym__expression] = STATE(2300),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2711),
    [sym_keyword_argument] = STATE(2711),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2711),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2960),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1659),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(719)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2338),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2694),
    [sym_keyword_argument] = STATE(2694),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2694),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1661),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(720)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(3017),
    [sym__expression] = STATE(2321),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2729),
    [sym_keyword_argument] = STATE(2729),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2729),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2864),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1663),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(721)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2325),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2694),
    [sym_keyword_argument] = STATE(2694),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2694),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1665),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(722)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2325),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2694),
    [sym_keyword_argument] = STATE(2694),
    [sym_list_splat_argument] = STATE(2916),
    [sym_dictionary_splat_argument] = STATE(2694),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1641),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1667),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(723)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2379),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2850),
    [sym_keyword_argument] = STATE(2850),
    [sym_list_splat_argument] = STATE(2850),
    [sym_dictionary_splat_argument] = STATE(2850),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1671),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(724)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2327),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2869),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1673),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(725)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2295),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2641),
    [sym_dictionary_splat_argument] = STATE(2639),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2641),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_pair] = STATE(2561),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_STAR_STAR] = ACTIONS(1677),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(1679),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(726)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2296),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(3037),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2737),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2737),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1683),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(727)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2318),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2694),
    [sym_keyword_argument] = STATE(2694),
    [sym_list_splat_argument] = STATE(2694),
    [sym_dictionary_splat_argument] = STATE(2694),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(728)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2327),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2869),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1687),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(729)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2379),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2850),
    [sym_keyword_argument] = STATE(2850),
    [sym_list_splat_argument] = STATE(2850),
    [sym_dictionary_splat_argument] = STATE(2850),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1689),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(730)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2275),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(2975),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2695),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2695),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1691),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(731)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2327),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2869),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1693),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(732)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2275),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(2975),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2695),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2695),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1695),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(733)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2379),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2850),
    [sym_keyword_argument] = STATE(2850),
    [sym_list_splat_argument] = STATE(2850),
    [sym_dictionary_splat_argument] = STATE(2850),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1697),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(734)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2296),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(3037),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2737),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2737),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1699),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(735)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2327),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2869),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1701),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(736)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(3017),
    [sym__expression] = STATE(2301),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2864),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2864),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1703),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(737)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2236),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(3016),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2728),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2728),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1705),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(738)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(3047),
    [sym__expression] = STATE(2306),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2960),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2960),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1707),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(739)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2226),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2698),
    [sym_dictionary_splat_argument] = STATE(2696),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2698),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_pair] = STATE(2574),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_STAR_STAR] = ACTIONS(1677),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(1709),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(740)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2335),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2711),
    [sym_keyword_argument] = STATE(2711),
    [sym_list_splat_argument] = STATE(2711),
    [sym_dictionary_splat_argument] = STATE(2711),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1711),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(741)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2327),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2869),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1713),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(742)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(3047),
    [sym__expression] = STATE(2306),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2960),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2960),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1715),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(743)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2379),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2850),
    [sym_keyword_argument] = STATE(2850),
    [sym_list_splat_argument] = STATE(2850),
    [sym_dictionary_splat_argument] = STATE(2850),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1717),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(744)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2980),
    [sym__expression] = STATE(2298),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2931),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2931),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1719),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(745)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2379),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2850),
    [sym_keyword_argument] = STATE(2850),
    [sym_list_splat_argument] = STATE(2850),
    [sym_dictionary_splat_argument] = STATE(2850),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1721),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(746)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2379),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2850),
    [sym_keyword_argument] = STATE(2850),
    [sym_list_splat_argument] = STATE(2850),
    [sym_dictionary_splat_argument] = STATE(2850),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1723),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(747)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2296),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(3037),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2737),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2737),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1725),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(748)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2980),
    [sym__expression] = STATE(2298),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2931),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2931),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1727),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(749)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2296),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(3037),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2737),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2737),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1729),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(750)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2287),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(3048),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2669),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2669),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1731),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(751)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2287),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(3048),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2669),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2669),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1733),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(752)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2310),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2729),
    [sym_keyword_argument] = STATE(2729),
    [sym_list_splat_argument] = STATE(2729),
    [sym_dictionary_splat_argument] = STATE(2729),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1735),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(753)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(2986),
    [sym__expression] = STATE(2327),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2869),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2869),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1737),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(754)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2246),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2662),
    [sym_dictionary_splat_argument] = STATE(2661),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2662),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_pair] = STATE(2567),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_STAR_STAR] = ACTIONS(1677),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(1739),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(755)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2319),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2702),
    [sym_keyword_argument] = STATE(2702),
    [sym_list_splat_argument] = STATE(2702),
    [sym_dictionary_splat_argument] = STATE(2702),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1741),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(756)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2541),
    [sym_pattern_list] = STATE(2541),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2248),
    [sym__primary_expression] = STATE(1784),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2665),
    [sym_dictionary_splat_argument] = STATE(2664),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2665),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_pair] = STATE(2568),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(533),
    [anon_sym_def] = ACTIONS(535),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_STAR_STAR] = ACTIONS(1677),
    [anon_sym_not] = ACTIONS(545),
    [anon_sym_PLUS] = ACTIONS(547),
    [anon_sym_DASH] = ACTIONS(547),
    [anon_sym_TILDE] = ACTIONS(547),
    [anon_sym_typeof] = ACTIONS(547),
    [anon_sym_void] = ACTIONS(547),
    [anon_sym_new] = ACTIONS(549),
    [anon_sym_await] = ACTIONS(551),
    [anon_sym_LBRACE] = ACTIONS(415),
    [anon_sym_RBRACE] = ACTIONS(1743),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(757)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2379),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2850),
    [sym_keyword_argument] = STATE(2850),
    [sym_list_splat_argument] = STATE(2850),
    [sym_dictionary_splat_argument] = STATE(2850),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1745),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(758)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2536),
    [sym_pattern_list] = STATE(2536),
    [sym_pattern] = STATE(2526),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__expression] = STATE(2379),
    [sym__primary_expression] = STATE(1785),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(1494),
    [sym_subscript] = STATE(1494),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym__argument] = STATE(2850),
    [sym_keyword_argument] = STATE(2850),
    [sym_list_splat_argument] = STATE(2850),
    [sym_dictionary_splat_argument] = STATE(2850),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(1669),
    [anon_sym_async] = ACTIONS(455),
    [anon_sym_def] = ACTIONS(457),
    [anon_sym_LPAREN] = ACTIONS(393),
    [anon_sym_RPAREN] = ACTIONS(1747),
    [anon_sym_LBRACK] = ACTIONS(395),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_STAR_STAR] = ACTIONS(1649),
    [anon_sym_not] = ACTIONS(467),
    [anon_sym_PLUS] = ACTIONS(469),
    [anon_sym_DASH] = ACTIONS(469),
    [anon_sym_TILDE] = ACTIONS(469),
    [anon_sym_typeof] = ACTIONS(469),
    [anon_sym_void] = ACTIONS(469),
    [anon_sym_new] = ACTIONS(471),
    [anon_sym_await] = ACTIONS(473),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(385),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(759)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2524),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3014),
    [sym_yield] = STATE(3017),
    [sym__expression] = STATE(2301),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2864),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2864),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_RPAREN] = ACTIONS(1749),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_yield] = ACTIONS(1645),
    [anon_sym_STAR] = ACTIONS(1675),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(760)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2296),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(3037),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2737),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2737),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1751),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
  [STATE(761)] = {
    [sym_assignment] = STATE(1698),
    [sym_augmented_assignment] = STATE(1698),
    [sym__assign_lhs] = STATE(2543),
    [sym_pattern_list] = STATE(2543),
    [sym_pattern] = STATE(2522),
    [sym_tuple_pattern] = STATE(2516),
    [sym_list_pattern] = STATE(2516),
    [sym__pattern_seq] = STATE(3061),
    [sym__expression] = STATE(2236),
    [sym__primary_expression] = STATE(1727),
    [sym_not_operator] = STATE(1698),
    [sym_boolean_operator] = STATE(1698),
    [sym_binary_operator] = STATE(1698),
    [sym_unary_operator] = STATE(1698),
    [sym_comparison_operator] = STATE(1698),
    [sym_conditional_expression] = STATE(1698),
    [sym_attribute] = STATE(936),
    [sym_subscript] = STATE(936),
    [sym_slice] = STATE(3016),
    [sym_call] = STATE(1682),
    [sym_slice_call] = STATE(1682),
    [sym_list_splat_argument] = STATE(2728),
    [sym_new_expression] = STATE(1682),
    [sym_existential] = STATE(1682),
    [sym_await] = STATE(1698),
    [sym_anonymous_function] = STATE(1682),
    [sym_list] = STATE(1682),
    [sym__collection_element] = STATE(2728),
    [sym_set] = STATE(1682),
    [sym_tuple] = STATE(1682),
    [sym_parenthesized_expression] = STATE(1682),
    [sym_dictionary] = STATE(1682),
    [sym_list_comprehension] = STATE(1682),
    [sym_set_comprehension] = STATE(1682),
    [sym_dictionary_comprehension] = STATE(1682),
    [sym_generator_expression] = STATE(1682),
    [sym_concatenated_string] = STATE(1682),
    [sym_f_string] = STATE(1536),
    [sym_identifier] = ACTIONS(475),
    [anon_sym_async] = ACTIONS(389),
    [anon_sym_def] = ACTIONS(391),
    [anon_sym_COLON] = ACTIONS(1681),
    [anon_sym_LPAREN] = ACTIONS(483),
    [anon_sym_LBRACK] = ACTIONS(485),
    [anon_sym_RBRACK] = ACTIONS(1753),
    [anon_sym_STAR] = ACTIONS(1647),
    [anon_sym_not] = ACTIONS(407),
    [anon_sym_PLUS] = ACTIONS(409),
    [anon_sym_DASH] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(409),
    [anon_sym_typeof] = ACTIONS(409),
    [anon_sym_void] = ACTIONS(409),
    [anon_sym_new] = ACTIONS(411),
    [anon_sym_await] = ACTIONS(413),
    [anon_sym_LBRACE] = ACTIONS(415),
    [sym_this] = ACTIONS(475),
    [sym_true] = ACTIONS(417),
    [sym_false] = ACTIONS(417),
    [sym_none] = ACTIONS(417),
    [sym_string] = ACTIONS(419),
    [sym_verbatim] = ACTIONS(417),
    [sym_number] = ACTIONS(417),
    [sym_comment] = ACTIONS(3),
    [sym_regex] = ACTIONS(421),
    [sym_fstring_start] = ACTIONS(423),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1645), 1,
      anon_sym_yield,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1755), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2301), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(3017), 1,
      sym_yield,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2864), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [123] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1645), 1,
      anon_sym_yield,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1757), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2298), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(2980), 1,
      sym_yield,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2931), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [246] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1645), 1,
      anon_sym_yield,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1759), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2327), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(2986), 1,
      sym_yield,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2869), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [369] = 29,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1645), 1,
      anon_sym_yield,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1761), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2306), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(3047), 1,
      sym_yield,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2960), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [492] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(385), 1,
      sym_this,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1649), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1669), 1,
      sym_identifier,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2379), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2850), 4,
      sym__argument,
      sym_keyword_argument,
      sym_list_splat_argument,
      sym_dictionary_splat_argument,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [613] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(661), 1,
      anon_sym_yield,
    ACTIONS(1763), 1,
      anon_sym_async,
    ACTIONS(1765), 1,
      anon_sym_def,
    STATE(1658), 1,
      sym__block_anonymous_function,
    STATE(1746), 1,
      sym_f_string,
    STATE(1881), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2245), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [731] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1725), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2334), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2737), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [851] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1751), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2334), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2737), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [971] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1683), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2334), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2737), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [1091] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1767), 1,
      anon_sym_from,
    STATE(1536), 1,
      sym_f_string,
    STATE(1697), 1,
      sym_expression_list,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2062), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(427), 3,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [1209] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1695), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2323), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2695), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [1329] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1729), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2334), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2737), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [1449] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1731), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2305), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2669), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [1569] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(661), 1,
      anon_sym_yield,
    ACTIONS(1769), 1,
      anon_sym_async,
    ACTIONS(1771), 1,
      anon_sym_def,
    STATE(1640), 1,
      sym__block_anonymous_function,
    STATE(1746), 1,
      sym_f_string,
    STATE(1881), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2245), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [1687] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1705), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2315), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2728), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [1807] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1691), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2323), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2695), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [1927] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1699), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2334), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2737), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [2047] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1733), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2305), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2669), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [2167] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1753), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2315), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2728), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [2287] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1773), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2334), 1,
      sym__expression,
    STATE(2522), 1,
      sym_pattern,
    STATE(3061), 1,
      sym__pattern_seq,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2737), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [2407] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(621), 1,
      anon_sym_yield,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1863), 1,
      sym__expression,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [2522] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(381), 1,
      sym__newline,
    ACTIONS(383), 1,
      anon_sym_SEMI,
    ACTIONS(1769), 1,
      anon_sym_async,
    ACTIONS(1771), 1,
      anon_sym_def,
    STATE(1624), 1,
      sym__block_anonymous_function,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2242), 1,
      sym_expression_list,
    STATE(2346), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [2641] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    ACTIONS(727), 1,
      anon_sym_yield,
    STATE(1746), 1,
      sym_f_string,
    STATE(1901), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2245), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [2756] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1677), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1775), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2415), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2929), 2,
      sym_dictionary_splat_argument,
      sym_pair,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [2873] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    ACTIONS(713), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1940), 1,
      sym__expression,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [2988] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(691), 1,
      anon_sym_yield,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1931), 1,
      sym__expression,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [3103] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(691), 1,
      anon_sym_yield,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1931), 1,
      sym__expression,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [3218] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    ACTIONS(713), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1940), 1,
      sym__expression,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [3333] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1777), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [3450] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(425), 1,
      sym__newline,
    ACTIONS(427), 1,
      anon_sym_SEMI,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    ACTIONS(1779), 1,
      anon_sym_from,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2241), 1,
      sym_expression_list,
    STATE(2339), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [3569] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1781), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [3686] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(173), 1,
      anon_sym_yield,
    STATE(449), 1,
      sym__expression,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(709), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [3801] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1678), 1,
      sym_expression_list,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2011), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(383), 3,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [3916] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(669), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1742), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [4031] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1783), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2368), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [4148] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1785), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2323), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2695), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [4265] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1787), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [4382] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(669), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1742), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [4497] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1789), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [4614] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1791), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [4731] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1677), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1793), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2415), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2929), 2,
      sym_dictionary_splat_argument,
      sym_pair,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [4848] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1677), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1795), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2415), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2929), 2,
      sym_dictionary_splat_argument,
      sym_pair,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [4965] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1677), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1797), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2415), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2929), 2,
      sym_dictionary_splat_argument,
      sym_pair,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [5082] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1799), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2368), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [5199] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1801), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2368), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [5316] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1803), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [5433] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1805), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [5550] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1677), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1807), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2415), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2929), 2,
      sym_dictionary_splat_argument,
      sym_pair,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [5667] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1809), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2368), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [5784] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2117), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(505), 4,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [5897] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1811), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [6014] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1813), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [6131] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1815), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2368), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [6248] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    ACTIONS(661), 1,
      anon_sym_yield,
    STATE(1746), 1,
      sym_f_string,
    STATE(1881), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2253), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [6363] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(719), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1734), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [6478] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(719), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1734), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [6593] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(579), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1710), 1,
      sym__expression,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [6708] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(465), 1,
      anon_sym_yield,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1745), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [6823] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(465), 1,
      anon_sym_yield,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1745), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [6938] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    ACTIONS(645), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1781), 1,
      sym__expression,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [7053] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(543), 1,
      anon_sym_yield,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1733), 1,
      sym__expression,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [7168] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(543), 1,
      anon_sym_yield,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1733), 1,
      sym__expression,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [7283] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    ACTIONS(645), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1781), 1,
      sym__expression,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [7398] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1725), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(505), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_for,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [7511] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1677), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1817), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2415), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2929), 2,
      sym_dictionary_splat_argument,
      sym_pair,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [7628] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1819), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2368), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [7745] = 28,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(381), 1,
      sym__newline,
    ACTIONS(383), 1,
      anon_sym_SEMI,
    ACTIONS(1763), 1,
      anon_sym_async,
    ACTIONS(1765), 1,
      anon_sym_def,
    STATE(1604), 1,
      sym__block_anonymous_function,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2242), 1,
      sym_expression_list,
    STATE(2346), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [7864] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1821), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2315), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2728), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [7981] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(165), 1,
      anon_sym_yield,
    STATE(440), 1,
      sym__expression,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(695), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [8096] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1823), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2334), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2737), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [8213] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1825), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [8330] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1827), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [8447] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    ACTIONS(727), 1,
      anon_sym_yield,
    STATE(1746), 1,
      sym_f_string,
    STATE(1901), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2253), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [8562] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1677), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1829), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2415), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2929), 2,
      sym_dictionary_splat_argument,
      sym_pair,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [8679] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1831), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2368), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [8796] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2117), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(527), 4,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [8909] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(405), 1,
      anon_sym_yield,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1707), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [9024] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1833), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [9141] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1835), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [9258] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(593), 1,
      anon_sym_yield,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1876), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [9373] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1677), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1837), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2415), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2929), 2,
      sym_dictionary_splat_argument,
      sym_pair,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [9490] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    ACTIONS(1839), 1,
      anon_sym_RBRACE,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2368), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [9607] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(593), 1,
      anon_sym_yield,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1876), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [9722] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1725), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(527), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_for,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [9835] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1841), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [9952] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(621), 1,
      anon_sym_yield,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1863), 1,
      sym__expression,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [10067] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(173), 1,
      anon_sym_yield,
    STATE(449), 1,
      sym__expression,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(696), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [10182] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1843), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [10299] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(165), 1,
      anon_sym_yield,
    STATE(440), 1,
      sym__expression,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(694), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [10414] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    ACTIONS(661), 1,
      anon_sym_yield,
    STATE(1746), 1,
      sym_f_string,
    STATE(1881), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2245), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [10529] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(493), 1,
      anon_sym_yield,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1764), 1,
      sym__expression,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [10644] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(493), 1,
      anon_sym_yield,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1764), 1,
      sym__expression,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [10759] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(139), 1,
      anon_sym_yield,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(443), 1,
      sym__expression,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(695), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [10874] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(139), 1,
      anon_sym_yield,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(443), 1,
      sym__expression,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(694), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [10989] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(93), 1,
      anon_sym_yield,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(456), 1,
      sym__expression,
    STATE(466), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(696), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [11104] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(93), 1,
      anon_sym_yield,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(456), 1,
      sym__expression,
    STATE(466), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(709), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [11219] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1845), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2305), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2669), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [11336] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(579), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1710), 1,
      sym__expression,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [11451] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    ACTIONS(1847), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [11568] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(677), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1734), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [11683] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(677), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1734), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [11798] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(705), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1742), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1688), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [11913] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(705), 1,
      anon_sym_yield,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1742), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [12028] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(405), 1,
      anon_sym_yield,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1707), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(1686), 3,
      sym__assign_rhs,
      sym_expression_list,
      sym_yield,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [12143] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1647), 1,
      anon_sym_STAR,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2352), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [12257] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(381), 1,
      sym__newline,
    ACTIONS(383), 1,
      anon_sym_SEMI,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2242), 1,
      sym_expression_list,
    STATE(2346), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [12373] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(435), 1,
      anon_sym_QMARK,
    ACTIONS(739), 1,
      anon_sym_DOT,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(765), 1,
      anon_sym_LBRACK,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(1849), 1,
      anon_sym_LPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(1685), 1,
      sym_argument_list,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1915), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [12489] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1681), 1,
      anon_sym_COLON,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2345), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(3060), 1,
      sym_expression_list,
    STATE(3078), 1,
      sym_slice,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [12605] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1856), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(505), 3,
      anon_sym_if,
      anon_sym_for,
      anon_sym_RBRACE,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [12717] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2303), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(449), 3,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [12829] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1677), 1,
      anon_sym_STAR_STAR,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2415), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2929), 2,
      sym_dictionary_splat_argument,
      sym_pair,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [12943] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1681), 1,
      anon_sym_COLON,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2341), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(3066), 1,
      sym_expression_list,
    STATE(3074), 1,
      sym_slice,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [13059] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1675), 1,
      anon_sym_STAR,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2368), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    STATE(2611), 2,
      sym_list_splat_argument,
      sym__collection_element,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [13173] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1856), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(527), 3,
      anon_sym_if,
      anon_sym_for,
      anon_sym_RBRACE,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [13285] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1681), 1,
      anon_sym_COLON,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2342), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(3023), 1,
      sym_expression_list,
    STATE(3054), 1,
      sym_slice,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [13401] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1681), 1,
      anon_sym_COLON,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2349), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(3002), 1,
      sym_slice,
    STATE(3089), 1,
      sym_expression_list,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [13517] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1681), 1,
      anon_sym_COLON,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2344), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(2993), 1,
      sym_expression_list,
    STATE(2997), 1,
      sym_slice,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [13633] = 27,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(427), 1,
      anon_sym_RPAREN,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1851), 1,
      anon_sym_from,
    STATE(1536), 1,
      sym_f_string,
    STATE(1697), 1,
      sym_expression_list,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2360), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [13749] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(525), 1,
      sym__newline,
    ACTIONS(527), 1,
      anon_sym_SEMI,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2356), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [13862] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2358), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    ACTIONS(505), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [13973] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2358), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    ACTIONS(527), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [14084] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    ACTIONS(1853), 1,
      anon_sym_COLON,
    ACTIONS(1855), 1,
      anon_sym_as,
    STATE(1536), 1,
      sym_f_string,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2350), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [14197] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    ACTIONS(1857), 1,
      anon_sym_COLON,
    ACTIONS(1859), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2381), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [14310] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    ACTIONS(1861), 1,
      anon_sym_COLON,
    ACTIONS(1863), 1,
      anon_sym_as,
    STATE(1536), 1,
      sym_f_string,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2348), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [14423] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(503), 1,
      sym__newline,
    ACTIONS(505), 1,
      anon_sym_SEMI,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2356), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [14536] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    ACTIONS(1865), 1,
      anon_sym_COLON,
    ACTIONS(1867), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2374), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [14649] = 26,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(449), 1,
      anon_sym_SEMI,
    ACTIONS(451), 1,
      sym__newline,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2369), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [14762] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1874), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [14872] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(455), 1,
      sym__expression,
    STATE(466), 1,
      sym__primary_expression,
    STATE(659), 1,
      sym_expression_list,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [14982] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1749), 1,
      sym__expression,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [15092] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2351), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(3043), 1,
      sym_expression_list,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [15202] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(2347), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(2587), 1,
      sym_expression_list,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [15312] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    ACTIONS(1867), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2488), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [15422] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(1986), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2212), 1,
      sym_expression_list,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [15532] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1753), 1,
      sym__expression,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [15642] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1756), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [15752] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(439), 1,
      sym__expression,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(659), 1,
      sym_expression_list,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [15862] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2077), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [15972] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1776), 1,
      sym__expression,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [16082] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(1886), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2212), 1,
      sym_expression_list,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [16192] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1709), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [16302] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1726), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [16412] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1783), 1,
      sym__expression,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [16522] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(451), 1,
      sym__expression,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(702), 1,
      sym_expression_list,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [16632] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1723), 1,
      sym__expression,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [16742] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2299), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(2587), 1,
      sym_expression_list,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [16852] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2364), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(2675), 1,
      sym_with_clause,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [16962] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2361), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(3005), 1,
      sym_expression_list,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [17072] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1937), 1,
      sym__expression,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [17182] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(448), 1,
      sym__expression,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(702), 1,
      sym_expression_list,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [17292] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2364), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(2764), 1,
      sym_with_clause,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [17402] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    ACTIONS(1869), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2468), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [17512] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1721), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [17622] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2212), 1,
      sym_expression_list,
    STATE(2343), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [17732] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    ACTIONS(1871), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2459), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [17842] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2364), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    STATE(2849), 1,
      sym_with_clause,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [17952] = 25,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1679), 1,
      sym_expression_list,
    STATE(1942), 1,
      sym__expression,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [18062] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2404), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [18169] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2286), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [18276] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2100), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [18383] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2108), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [18490] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2414), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [18597] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2114), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [18704] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1759), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [18811] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2367), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [18918] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2375), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [19025] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(498), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [19132] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1757), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [19239] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(652), 1,
      sym__primary_expression,
    STATE(662), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [19346] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(652), 1,
      sym__primary_expression,
    STATE(663), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [19453] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(652), 1,
      sym__primary_expression,
    STATE(667), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [19560] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2103), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [19667] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1732), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [19774] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1744), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [19881] = 6,
    ACTIONS(929), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1875), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1873), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(927), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(925), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [19952] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(652), 1,
      sym__primary_expression,
    STATE(711), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [20059] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1976), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [20166] = 6,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1882), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1880), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(933), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(931), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [20237] = 6,
    ACTIONS(941), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1887), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1885), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(939), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(937), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [20308] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2464), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [20415] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1887), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [20522] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1888), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [20629] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1889), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [20736] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1890), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [20843] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1891), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [20950] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1892), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [21057] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1894), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [21164] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1895), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [21271] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1896), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [21378] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    ACTIONS(1890), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(1887), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [21485] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1826), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [21592] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1917), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [21699] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1978), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [21806] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1979), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [21913] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1735), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22020] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1903), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22127] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1904), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22234] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1905), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22341] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1906), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22448] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1907), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22555] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1908), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22662] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1909), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22769] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1910), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22876] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1911), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [22983] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1912), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [23090] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1892), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1903), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [23197] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1921), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [23304] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1923), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [23411] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1928), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [23518] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1725), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [23625] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2122), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [23732] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1981), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [23839] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(1994), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [23946] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1747), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [24053] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1761), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [24160] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1762), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [24267] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(542), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [24374] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(555), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [24481] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(556), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [24588] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1768), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [24695] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1769), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [24802] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1877), 1,
      sym__expression,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [24909] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1828), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25016] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1748), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25123] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1736), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25230] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1925), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25337] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1926), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25444] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1751), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25551] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1724), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25658] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(547), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25765] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(557), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25872] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(461), 1,
      sym__expression,
    STATE(466), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [25979] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1770), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [26086] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1894), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1747), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [26193] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(502), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [26300] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(467), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [26407] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(468), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [26514] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(558), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [26621] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(470), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [26728] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(471), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [26835] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(472), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [26942] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(473), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [27049] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(474), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [27156] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(475), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [27263] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(476), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [27370] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(1896), 1,
      anon_sym_not,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(467), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [27477] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(503), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [27584] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(479), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [27691] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2048), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [27798] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(480), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [27905] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(1987), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28012] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(527), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28119] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(481), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28226] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(508), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28333] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(482), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28440] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(483), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28547] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(484), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28654] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(485), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28761] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(486), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28868] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(487), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [28975] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(488), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [29082] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(489), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [29189] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(490), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [29296] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(491), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [29403] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(1898), 1,
      anon_sym_not,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(482), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [29510] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(509), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [29617] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(494), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [29724] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(495), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [29831] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(496), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [29938] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1738), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [30045] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1730), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [30152] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(1998), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [30259] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(943), 19,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(945), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [30324] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(1999), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [30431] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(598), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [30538] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(672), 1,
      sym__primary_expression,
    STATE(705), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [30645] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(672), 1,
      sym__primary_expression,
    STATE(706), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [30752] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(672), 1,
      sym__primary_expression,
    STATE(692), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [30859] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(637), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [30966] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(633), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [31073] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(644), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [31180] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(672), 1,
      sym__primary_expression,
    STATE(677), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [31287] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2153), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [31394] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(622), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [31501] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(630), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [31608] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(672), 1,
      sym__primary_expression,
    STATE(703), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [31715] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    ACTIONS(1900), 1,
      anon_sym_not,
    STATE(458), 1,
      sym_f_string,
    STATE(672), 1,
      sym__primary_expression,
    STATE(705), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [31822] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(619), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [31929] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(594), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32036] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(609), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32143] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2255), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32250] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1898), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32357] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1918), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32464] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1919), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32571] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(1939), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32678] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1718), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32785] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1864), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32892] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1865), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [32999] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1866), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [33106] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1867), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [33213] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1868), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [33320] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1869), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [33427] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1870), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [33534] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1871), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [33641] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1872), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [33748] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1873), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [33855] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    ACTIONS(1902), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1864), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [33962] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(632), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [34069] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1827), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [34176] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1772), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [34283] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1773), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [34390] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1823), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [34497] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(549), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [34604] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2358), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [34711] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(947), 19,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(949), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [34776] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(947), 19,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(949), 38,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [34841] = 6,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1904), 1,
      anon_sym_COMMA,
    ACTIONS(1880), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(951), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(958), 24,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [34912] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2308), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35019] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2008), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35126] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    ACTIONS(1908), 1,
      anon_sym_not,
    STATE(1746), 1,
      sym_f_string,
    STATE(2071), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35233] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(652), 1,
      sym__primary_expression,
    STATE(680), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35340] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(569), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35447] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(573), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35554] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(580), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35661] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(581), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35768] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(587), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35875] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(596), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [35982] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(625), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [36089] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(635), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [36196] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(636), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [36303] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(642), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [36410] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    ACTIONS(1910), 1,
      anon_sym_not,
    STATE(450), 1,
      sym_f_string,
    STATE(569), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [36517] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(639), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [36624] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1934), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [36731] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2104), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [36838] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2135), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [36945] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2136), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [37052] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1780), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [37159] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2256), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [37266] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2260), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [37373] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2261), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [37480] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2262), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [37587] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2269), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [37694] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2276), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [37801] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2277), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [37908] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2278), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38015] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2279), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38122] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2284), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38229] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    ACTIONS(1912), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2256), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38336] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2291), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38443] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2292), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38550] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2163), 1,
      sym__expression,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38657] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1944), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38764] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(652), 1,
      sym__primary_expression,
    STATE(664), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38871] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1778), 1,
      sym__expression,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [38978] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(602), 1,
      sym__expression,
    STATE(652), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [39085] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(565), 1,
      anon_sym_not,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(571), 1,
      anon_sym_await,
    ACTIONS(631), 1,
      anon_sym_LPAREN,
    ACTIONS(633), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(652), 1,
      sym__primary_expression,
    STATE(671), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(555), 2,
      sym_this,
      sym_identifier,
    STATE(236), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2529), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(567), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [39192] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2057), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [39299] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1819), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [39406] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1974), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [39513] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1984), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [39620] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1985), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [39727] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2180), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [39834] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2118), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [39941] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2123), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [40048] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2125), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [40155] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2126), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [40262] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2131), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [40369] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2132), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [40476] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2137), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [40583] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2138), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [40690] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2151), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [40797] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1989), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [40904] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    ACTIONS(1914), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(2118), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41011] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1995), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41118] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1996), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41225] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2356), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41332] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2000), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41439] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2384), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41546] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2172), 1,
      sym__expression,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41653] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1852), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41760] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2110), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41867] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2139), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [41974] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2140), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [42081] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2214), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [42188] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1946), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [42295] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1947), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [42402] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1948), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [42509] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1949), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [42616] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1950), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [42723] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1951), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [42830] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1952), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [42937] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1954), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [43044] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1955), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [43151] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1956), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [43258] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    ACTIONS(1916), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1946), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [43365] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1959), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [43472] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1960), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [43579] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2353), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [43686] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1961), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [43793] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2005), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [43900] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2392), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44007] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2482), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44114] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2393), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44221] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(539), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44328] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(585), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44435] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1972), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44542] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2112), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44649] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2141), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44756] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2142), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44863] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(811), 1,
      anon_sym_not,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(817), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2007), 1,
      sym__expression,
    STATE(2157), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2533), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(813), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [44970] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2067), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [45077] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2068), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [45184] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2069), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [45291] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2070), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [45398] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2073), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [45505] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2124), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [45612] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2127), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [45719] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2128), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [45826] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2130), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [45933] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2133), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [46040] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    ACTIONS(1918), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2067), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [46147] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2148), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [46254] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2149), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [46361] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(590), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [46468] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2152), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [46575] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(565), 1,
      sym__expression,
    STATE(672), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [46682] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2371), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [46789] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1964), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [46896] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2372), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47003] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2383), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47110] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(1965), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47217] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2159), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47324] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2033), 1,
      sym__expression,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47431] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2080), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47538] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2038), 1,
      sym__expression,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47645] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2056), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47752] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2078), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47859] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2150), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [47966] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1992), 1,
      sym__expression,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [48073] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2003), 1,
      sym__expression,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [48180] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2004), 1,
      sym__expression,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [48287] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    ACTIONS(1920), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2159), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [48394] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2040), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [48501] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2041), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [48608] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2044), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [48715] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2155), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [48822] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2147), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [48929] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2389), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49036] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2382), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49143] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2376), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49250] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(825), 1,
      anon_sym_not,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(831), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2050), 1,
      sym__primary_expression,
    STATE(2156), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2546), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(827), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49357] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2051), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49464] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2411), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49571] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2466), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49678] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2181), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49785] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2182), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49892] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(851), 1,
      anon_sym_not,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(857), 1,
      anon_sym_await,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(879), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2039), 1,
      sym__primary_expression,
    STATE(2052), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(803), 2,
      sym_this,
      sym_identifier,
    STATE(1505), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2531), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(853), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [49999] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2183), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [50106] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2184), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [50213] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2185), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [50320] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2186), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [50427] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2187), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [50534] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2188), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [50641] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2191), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [50748] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2192), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [50855] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    ACTIONS(1922), 1,
      anon_sym_not,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2181), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [50962] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(517), 1,
      anon_sym_not,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(523), 1,
      anon_sym_await,
    ACTIONS(635), 1,
      anon_sym_LPAREN,
    ACTIONS(637), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(672), 1,
      sym__primary_expression,
    STATE(712), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(507), 2,
      sym_this,
      sym_identifier,
    STATE(216), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2534), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(519), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [51069] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2359), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [51176] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2476), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [51283] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1711), 1,
      sym__expression,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [51390] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2014), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [51497] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2380), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [51604] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2451), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [51711] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2373), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [51818] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2483), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [51925] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2257), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52032] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2259), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52139] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(2354), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52246] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2365), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52353] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2016), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52460] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2036), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52567] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2474), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52674] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2431), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52781] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2377), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52888] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2485), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [52995] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2378), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [53102] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2503), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [53209] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2510), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [53316] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2511), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [53423] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1800), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [53530] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1829), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [53637] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1857), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [53744] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1858), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [53851] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2312), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [53958] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(127), 1,
      anon_sym_LPAREN,
    ACTIONS(129), 1,
      anon_sym_LBRACK,
    ACTIONS(141), 1,
      anon_sym_not,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(147), 1,
      anon_sym_await,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    STATE(458), 1,
      sym_f_string,
    STATE(463), 1,
      sym__primary_expression,
    STATE(507), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(117), 2,
      sym_this,
      sym_identifier,
    STATE(234), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2535), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(143), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(151), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(665), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(646), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [54065] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2117), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [54172] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2266), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [54279] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2355), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [54386] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(81), 1,
      anon_sym_LPAREN,
    ACTIONS(83), 1,
      anon_sym_LBRACK,
    ACTIONS(95), 1,
      anon_sym_not,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(101), 1,
      anon_sym_await,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    STATE(450), 1,
      sym_f_string,
    STATE(466), 1,
      sym__primary_expression,
    STATE(511), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(71), 2,
      sym_this,
      sym_identifier,
    STATE(237), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2540), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(97), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(105), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(649), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(641), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [54493] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1817), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [54600] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1801), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [54707] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1802), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [54814] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1803), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [54921] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1804), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55028] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1805), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55135] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1806), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55242] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1807), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55349] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1808), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55456] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1809), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55563] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1810), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55670] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    ACTIONS(1924), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1801), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55777] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2120), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55884] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2121), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [55991] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1818), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [56098] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1813), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [56205] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1814), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [56312] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2489), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [56419] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1815), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [56526] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1993), 1,
      sym__expression,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [56633] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2251), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [56740] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2161), 1,
      sym__expression,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [56847] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2281), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [56954] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2196), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [57061] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2197), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [57168] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2198), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [57275] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2199), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [57382] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2200), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [57489] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2201), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [57596] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2202), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [57703] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2203), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [57810] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2204), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [57917] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2205), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58024] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    ACTIONS(1926), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2196), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58131] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2208), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58238] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2209), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58345] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2210), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58452] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2422), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58559] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2037), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58666] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2168), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58773] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2213), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58880] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1936), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [58987] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1820), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [59094] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1830), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [59201] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1831), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [59308] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2097), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [59415] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2115), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [59522] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2143), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [59629] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2144), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [59736] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2017), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [59843] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2018), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [59950] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2019), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [60057] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2021), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [60164] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2022), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [60271] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2023), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [60378] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2024), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [60485] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2025), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [60592] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2026), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [60699] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2027), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [60806] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    ACTIONS(1928), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(2017), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [60913] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2030), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61020] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2031), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61127] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2032), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61234] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2034), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61341] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(839), 1,
      anon_sym_not,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(845), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2035), 1,
      sym__expression,
    STATE(2072), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2538), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(841), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61448] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2455), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61555] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2293), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61662] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2465), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61769] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2470), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61876] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2471), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [61983] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2473), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [62090] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2010), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [62197] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2294), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [62304] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1850), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [62411] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1834), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [62518] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1835), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [62625] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1836), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [62732] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1837), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [62839] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1838), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [62946] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1839), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [63053] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1840), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [63160] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1841), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [63267] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1842), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [63374] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1843), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [63481] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    ACTIONS(1930), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1834), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [63588] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1851), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [63695] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1846), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [63802] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1847), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [63909] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1848), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64016] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1966), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64123] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2263), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64230] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2282), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64337] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(917), 1,
      anon_sym_not,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(923), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2193), 1,
      sym__primary_expression,
    STATE(2283), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2528), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(919), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64444] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(545), 1,
      anon_sym_not,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(551), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1784), 1,
      sym__primary_expression,
    STATE(1856), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2541), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(547), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64551] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1853), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64658] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(1855), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64765] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(495), 1,
      anon_sym_not,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(501), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1771), 1,
      sym__expression,
    STATE(1789), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2544), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(497), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64872] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2428), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [64979] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2357), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [65086] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2506), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [65193] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2509), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [65300] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(1969), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [65407] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(905), 1,
      anon_sym_not,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(911), 1,
      anon_sym_await,
    ACTIONS(961), 1,
      anon_sym_LPAREN,
    ACTIONS(963), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2165), 1,
      sym__primary_expression,
    STATE(2227), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(895), 2,
      sym_this,
      sym_identifier,
    STATE(1517), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2539), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(907), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [65514] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2095), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [65621] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2079), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [65728] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2160), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [65835] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2081), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [65942] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2082), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [66049] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2083), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [66156] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2084), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [66263] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2085), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [66370] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2086), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [66477] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2087), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [66584] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2088), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [66691] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    ACTIONS(1932), 1,
      anon_sym_not,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2079), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [66798] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2096), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [66905] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2091), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67012] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2092), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67119] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2093), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67226] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2129), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67333] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2098), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67440] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2101), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67547] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(693), 1,
      anon_sym_not,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(699), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2061), 1,
      sym__primary_expression,
    STATE(2102), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2547), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(695), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67654] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2500), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67761] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2071), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67868] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2462), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [67975] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2390), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [68082] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2450), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [68189] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2412), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [68296] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2418), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [68403] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2407), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [68510] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2486), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [68617] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2497), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [68724] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2396), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [68831] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2458), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [68938] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2460), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [69045] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2505), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [69152] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2480), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [69259] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2484), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [69366] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2400), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [69473] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2499), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [69580] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2501), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [69687] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2441), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [69794] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2472), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [69901] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2494), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70008] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2478), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70115] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2391), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70222] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2394), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70329] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2493), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70436] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2403), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70543] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2406), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70650] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2386), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70757] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2426), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70864] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2430), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [70971] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2432), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [71078] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2453), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [71185] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2467), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [71292] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2399), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [71399] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2475), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [71506] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2477), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [71613] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2388), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [71720] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2479), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [71827] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2481), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [71934] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2397), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [72041] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2490), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [72148] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2492), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [72255] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2419), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [72362] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2496), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [72469] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2498), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [72576] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(885), 1,
      anon_sym_not,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(891), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(2174), 1,
      sym__primary_expression,
    STATE(2448), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2542), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(887), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [72683] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2502), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [72790] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2504), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [72897] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2507), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73004] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2508), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73111] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2370), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73218] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2398), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73325] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2487), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73432] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2491), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73539] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2447), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73646] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2387), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73753] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2395), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73860] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2408), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [73967] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2413), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [74074] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2437), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [74181] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2445), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [74288] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2461), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [74395] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1924), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [74502] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(15), 1,
      anon_sym_LPAREN,
    ACTIONS(17), 1,
      anon_sym_LBRACK,
    ACTIONS(53), 1,
      anon_sym_not,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(59), 1,
      anon_sym_await,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    STATE(1746), 1,
      sym_f_string,
    STATE(2074), 1,
      sym__expression,
    STATE(2119), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(7), 2,
      sym_this,
      sym_identifier,
    STATE(1527), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2530), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(55), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(63), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(2258), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(2176), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [74609] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(467), 1,
      anon_sym_not,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(473), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1785), 1,
      sym__primary_expression,
    STATE(2366), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2536), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(469), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [74716] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(611), 1,
      anon_sym_LPAREN,
    ACTIONS(613), 1,
      anon_sym_LBRACK,
    ACTIONS(623), 1,
      anon_sym_not,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(629), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1879), 1,
      sym__primary_expression,
    STATE(2362), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(603), 2,
      sym_this,
      sym_identifier,
    STATE(1513), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2548), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(625), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [74823] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2495), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [74930] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2463), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [75037] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2469), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [75144] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(393), 1,
      anon_sym_LPAREN,
    ACTIONS(395), 1,
      anon_sym_LBRACK,
    ACTIONS(407), 1,
      anon_sym_not,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(413), 1,
      anon_sym_await,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    STATE(1536), 1,
      sym_f_string,
    STATE(1727), 1,
      sym__primary_expression,
    STATE(1953), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(385), 2,
      sym_this,
      sym_identifier,
    STATE(1494), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2543), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(409), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [75251] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2405), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [75358] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2425), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [75465] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2401), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [75572] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2402), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [75679] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2409), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [75786] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2410), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [75893] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2416), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76000] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2417), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76107] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2420), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76214] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2421), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76321] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2423), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76428] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2424), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76535] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2427), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76642] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2429), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76749] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2433), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76856] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2434), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [76963] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2435), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [77070] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2436), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [77177] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2438), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [77284] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2439), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [77391] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2440), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [77498] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2442), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [77605] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2443), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [77712] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2444), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [77819] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2446), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [77926] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2512), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [78033] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2449), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [78140] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2385), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [78247] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2452), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [78354] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2454), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [78461] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2456), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [78568] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(767), 1,
      anon_sym_not,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(773), 1,
      anon_sym_await,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1897), 1,
      sym__primary_expression,
    STATE(2457), 1,
      sym__expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2545), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(769), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [78675] = 24,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(483), 1,
      anon_sym_LPAREN,
    ACTIONS(485), 1,
      anon_sym_LBRACK,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(595), 1,
      anon_sym_not,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(601), 1,
      anon_sym_await,
    STATE(1536), 1,
      sym_f_string,
    STATE(1893), 1,
      sym__expression,
    STATE(1983), 1,
      sym__primary_expression,
    STATE(2526), 1,
      sym_pattern,
    ACTIONS(475), 2,
      sym_this,
      sym_identifier,
    STATE(936), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2532), 2,
      sym__assign_lhs,
      sym_pattern_list,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(597), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
    STATE(1698), 9,
      sym_assignment,
      sym_augmented_assignment,
      sym_not_operator,
      sym_boolean_operator,
      sym_binary_operator,
      sym_unary_operator,
      sym_comparison_operator,
      sym_conditional_expression,
      sym_await,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [78782] = 5,
    ACTIONS(929), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1873), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(927), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(925), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [78850] = 5,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1880), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(933), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(931), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [78918] = 5,
    ACTIONS(941), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1885), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(939), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(937), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [78986] = 5,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1880), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
    ACTIONS(951), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(958), 23,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [79054] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1944), 1,
      anon_sym_RPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2589), 1,
      sym__primary_expression,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(1936), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [79149] = 6,
    ACTIONS(929), 1,
      anon_sym_EQ,
    ACTIONS(1875), 1,
      anon_sym_COMMA,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1873), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(927), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(925), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [79216] = 6,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1882), 1,
      anon_sym_COMMA,
    ACTIONS(1880), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(933), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(931), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [79283] = 6,
    ACTIONS(941), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1887), 1,
      anon_sym_COMMA,
    ACTIONS(1885), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(939), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(937), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [79350] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    ACTIONS(1952), 1,
      anon_sym_RPAREN,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2589), 1,
      sym__primary_expression,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(1950), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [79445] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1948), 1,
      anon_sym_new,
    ACTIONS(1952), 1,
      anon_sym_RBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2589), 1,
      sym__primary_expression,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(1950), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [79540] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1944), 1,
      anon_sym_RBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2589), 1,
      sym__primary_expression,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(1936), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [79635] = 5,
    ACTIONS(941), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1885), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
    ACTIONS(939), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(937), 19,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [79699] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1956), 1,
      anon_sym_else,
    ACTIONS(1958), 1,
      anon_sym_except,
    ACTIONS(1960), 1,
      anon_sym_finally,
    STATE(1568), 1,
      sym_else_clause,
    STATE(1617), 1,
      sym_finally_clause,
    STATE(1518), 2,
      sym_except_clause,
      aux_sym_try_statement_repeat1,
    ACTIONS(1962), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1954), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [79771] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2589), 1,
      sym__primary_expression,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(1936), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [79863] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2589), 1,
      sym__primary_expression,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    ACTIONS(1950), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [79955] = 9,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1964), 1,
      anon_sym_else,
    ACTIONS(1966), 1,
      anon_sym_except,
    ACTIONS(1968), 1,
      anon_sym_finally,
    STATE(1578), 1,
      sym_else_clause,
    STATE(1650), 1,
      sym_finally_clause,
    STATE(1529), 2,
      sym_except_clause,
      aux_sym_try_statement_repeat1,
    ACTIONS(1962), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1954), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [80027] = 5,
    ACTIONS(929), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1873), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
    ACTIONS(927), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(925), 19,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [80091] = 5,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1880), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
    ACTIONS(933), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(931), 19,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [80155] = 6,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1970), 1,
      anon_sym_EQ,
    ACTIONS(1875), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(1873), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(925), 16,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(927), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [80219] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1976), 1,
      anon_sym_elif,
    ACTIONS(1978), 1,
      anon_sym_else,
    STATE(1525), 1,
      aux_sym_if_statement_repeat1,
    STATE(1571), 1,
      sym_elif_clause,
    STATE(1630), 1,
      sym_else_clause,
    ACTIONS(1972), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1974), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [80287] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1980), 1,
      anon_sym_elif,
    ACTIONS(1982), 1,
      anon_sym_else,
    STATE(1514), 1,
      aux_sym_if_statement_repeat1,
    STATE(1575), 1,
      sym_elif_clause,
    STATE(1603), 1,
      sym_else_clause,
    ACTIONS(1972), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1974), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [80355] = 6,
    ACTIONS(971), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1882), 1,
      anon_sym_COMMA,
    ACTIONS(1880), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(931), 17,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(933), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [80419] = 6,
    ACTIONS(965), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1887), 1,
      anon_sym_COMMA,
    ACTIONS(1885), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(937), 17,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(939), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [80483] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(947), 19,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(949), 31,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [80541] = 6,
    ACTIONS(968), 1,
      anon_sym_EQ,
    ACTIONS(1875), 1,
      anon_sym_COMMA,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1873), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(925), 17,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(927), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [80605] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1980), 1,
      anon_sym_elif,
    ACTIONS(1982), 1,
      anon_sym_else,
    STATE(1530), 1,
      aux_sym_if_statement_repeat1,
    STATE(1575), 1,
      sym_elif_clause,
    STATE(1649), 1,
      sym_else_clause,
    ACTIONS(1986), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1984), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [80673] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(947), 19,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(949), 31,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [80731] = 6,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1904), 1,
      anon_sym_COMMA,
    ACTIONS(1880), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(958), 17,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(951), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [80795] = 5,
    ACTIONS(929), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1873), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
    ACTIONS(925), 17,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(927), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [80857] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1990), 1,
      anon_sym_except,
    STATE(1518), 2,
      sym_except_clause,
      aux_sym_try_statement_repeat1,
    ACTIONS(1993), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1988), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [80919] = 5,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1880), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
    ACTIONS(931), 17,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(933), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [80981] = 5,
    ACTIONS(941), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1885), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
    ACTIONS(937), 17,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(939), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [81043] = 5,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1880), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
    ACTIONS(958), 17,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(951), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [81105] = 6,
    ACTIONS(954), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1904), 1,
      anon_sym_COMMA,
    ACTIONS(1880), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(958), 17,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(951), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [81169] = 6,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1882), 1,
      anon_sym_COMMA,
    ACTIONS(1880), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(931), 17,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(933), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [81233] = 6,
    ACTIONS(941), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1887), 1,
      anon_sym_COMMA,
    ACTIONS(1885), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(937), 17,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(939), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [81297] = 8,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1976), 1,
      anon_sym_elif,
    ACTIONS(1978), 1,
      anon_sym_else,
    STATE(1531), 1,
      aux_sym_if_statement_repeat1,
    STATE(1571), 1,
      sym_elif_clause,
    STATE(1635), 1,
      sym_else_clause,
    ACTIONS(1986), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1984), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [81365] = 6,
    ACTIONS(1875), 1,
      anon_sym_COMMA,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1970), 1,
      anon_sym_EQ,
    ACTIONS(1873), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(925), 17,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(927), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [81429] = 6,
    ACTIONS(929), 1,
      anon_sym_EQ,
    ACTIONS(1875), 1,
      anon_sym_COMMA,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1873), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(925), 17,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
    ACTIONS(927), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [81493] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(943), 19,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(945), 31,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [81551] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1995), 1,
      anon_sym_except,
    STATE(1529), 2,
      sym_except_clause,
      aux_sym_try_statement_repeat1,
    ACTIONS(1993), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1988), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [81613] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2000), 1,
      anon_sym_elif,
    STATE(1530), 1,
      aux_sym_if_statement_repeat1,
    STATE(1575), 1,
      sym_elif_clause,
    ACTIONS(2003), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1998), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [81676] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2005), 1,
      anon_sym_elif,
    STATE(1531), 1,
      aux_sym_if_statement_repeat1,
    STATE(1571), 1,
      sym_elif_clause,
    ACTIONS(2003), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1998), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [81739] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2008), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2010), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [81795] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2012), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2014), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [81851] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1192), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1190), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [81907] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2016), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2018), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [81963] = 6,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2020), 1,
      sym_string,
    STATE(1537), 2,
      sym_f_string,
      aux_sym_concatenated_string_repeat1,
    ACTIONS(927), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(925), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [82025] = 6,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2022), 1,
      sym_string,
    STATE(1556), 2,
      sym_f_string,
      aux_sym_concatenated_string_repeat1,
    ACTIONS(1070), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1068), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [82087] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2026), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2024), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82143] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2028), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2030), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82199] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2032), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2034), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82255] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2038), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2036), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82311] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2008), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2010), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82367] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2042), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2040), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82423] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2012), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2014), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82479] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2032), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2034), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82535] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2016), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2018), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82591] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2026), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2024), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82647] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2044), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82703] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2028), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2030), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82759] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2044), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82815] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2048), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2050), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82871] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2042), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2040), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82927] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2048), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2050), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [82983] = 6,
    ACTIONS(935), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(958), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(931), 11,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
    ACTIONS(1880), 14,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
    ACTIONS(933), 18,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_STAR_STAR,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
  [83045] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2038), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2036), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83101] = 6,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2052), 1,
      sym_string,
    ACTIONS(2055), 1,
      sym_fstring_start,
    STATE(1556), 2,
      sym_f_string,
      aux_sym_concatenated_string_repeat1,
    ACTIONS(1141), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1139), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [83163] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1344), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83219] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1192), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1190), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83275] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1344), 45,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83331] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2048), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2050), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83386] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2044), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83441] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1344), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83496] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2044), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83551] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1344), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83606] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1978), 1,
      anon_sym_else,
    STATE(1632), 1,
      sym_else_clause,
    ACTIONS(2058), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2060), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83665] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1192), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1190), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83720] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1982), 1,
      anon_sym_else,
    STATE(1606), 1,
      sym_else_clause,
    ACTIONS(2064), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2062), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83779] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1960), 1,
      anon_sym_finally,
    STATE(1621), 1,
      sym_finally_clause,
    ACTIONS(2068), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2066), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83838] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2048), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2050), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83893] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2016), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2018), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [83948] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2070), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2072), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84003] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2016), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2018), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84058] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1982), 1,
      anon_sym_else,
    STATE(1648), 1,
      sym_else_clause,
    ACTIONS(2058), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2060), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84117] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1192), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1190), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84172] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2070), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2072), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84227] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2028), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2030), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84282] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2076), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2074), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84337] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1968), 1,
      anon_sym_finally,
    STATE(1651), 1,
      sym_finally_clause,
    ACTIONS(2068), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2066), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84396] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2076), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2074), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84451] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1978), 1,
      anon_sym_else,
    STATE(1643), 1,
      sym_else_clause,
    ACTIONS(2064), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2062), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84510] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2028), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2030), 44,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_elif,
      anon_sym_else,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84565] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1190), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1192), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [84619] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1344), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84673] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1192), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1190), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84727] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2016), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2018), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84781] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1544), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1542), 37,
      sym_fstring_start,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
      sym_string,
  [84835] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2044), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84889] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1344), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1342), 37,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_except,
      anon_sym_finally,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [84943] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2028), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2030), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [84997] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2048), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2050), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85051] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2016), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2018), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85105] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1602), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1604), 37,
      sym_fstring_start,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
      sym_string,
  [85159] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2044), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85213] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2028), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2030), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85267] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2048), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2050), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85321] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1344), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85375] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1192), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1190), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85429] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2078), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2080), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85483] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2078), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2080), 43,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_finally,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85537] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2082), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2084), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85590] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2086), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2088), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85643] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2090), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2092), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85696] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2094), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85749] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2098), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2100), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85802] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2104), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2102), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85855] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2108), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2106), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85908] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2110), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2112), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [85961] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1212), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1210), 32,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [86022] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2120), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2122), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86075] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2048), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2050), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86128] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2124), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2126), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86181] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2110), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2112), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86234] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2028), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2030), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86287] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2128), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2130), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86340] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2120), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2122), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86393] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2134), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2132), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86446] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2068), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2066), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86499] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2090), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2092), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86552] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2028), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2030), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86605] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1344), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86658] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2138), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2136), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86711] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2134), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2132), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86764] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1192), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1190), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86817] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2098), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2100), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86870] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2124), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2126), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86923] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2086), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2088), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [86976] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2078), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2080), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87029] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1962), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1954), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87082] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2104), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2102), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87135] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2096), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2094), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87188] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2140), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2142), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87241] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2144), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2146), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87294] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2150), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2148), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87347] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1962), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(1954), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87400] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2152), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2154), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87453] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2156), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2158), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87506] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2162), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2160), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87559] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1342), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1344), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87612] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2078), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2080), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87665] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2166), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2164), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87718] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2140), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2142), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87771] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2044), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87824] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2108), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2106), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87877] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2016), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2018), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87930] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2128), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2130), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [87983] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1192), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(1190), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88036] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2016), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2018), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88089] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2144), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2146), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88142] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2152), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2154), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88195] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2068), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2066), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88248] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2138), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2136), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88301] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2048), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2050), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88354] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2150), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2148), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88407] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2156), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2158), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88460] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2046), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2044), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88513] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2162), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2160), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88566] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2082), 3,
      sym__dedent,
      sym_regex,
      sym_fstring_start,
    ACTIONS(2084), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88619] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(2166), 3,
      sym_regex,
      sym_fstring_start,
      ts_builtin_sym_end,
    ACTIONS(2164), 42,
      anon_sym_return,
      anon_sym_async,
      anon_sym_def,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_del,
      anon_sym_raise,
      sym_pass_statement,
      sym_break_statement,
      sym_continue_statement,
      sym_debugger_statement,
      anon_sym_assert,
      anon_sym_global,
      anon_sym_nonlocal,
      anon_sym_yield,
      anon_sym_from,
      anon_sym_import,
      anon_sym_if,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_with,
      anon_sym_class,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_TILDE,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_new,
      anon_sym_await,
      anon_sym_LBRACE,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_string,
      sym_verbatim,
      sym_number,
      sym_identifier,
  [88672] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1592), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1590), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [88724] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(939), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(937), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [88776] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1594), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1596), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [88828] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1540), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1538), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [88880] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1548), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1546), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [88932] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1451), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1449), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [88984] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1484), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1482), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89036] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1554), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1552), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89088] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1464), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1462), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89140] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1474), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1476), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89192] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1512), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1510), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89244] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1468), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1466), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89296] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1472), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1470), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89348] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1516), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1514), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89400] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1562), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1560), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89452] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1389), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1387), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89504] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1480), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1478), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89556] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1558), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1556), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89608] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1524), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1522), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89660] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1154), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1152), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89712] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1072), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1101), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89764] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1407), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1405), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89816] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1617), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1615), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89868] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(927), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(925), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89920] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1621), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1619), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [89972] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1626), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1623), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90024] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1397), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1395), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90076] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1631), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1629), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90128] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1528), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1526), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90180] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1635), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1633), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90232] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1575), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1573), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90284] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1579), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1577), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90336] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1411), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1409), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90388] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1415), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1413), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90440] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1611), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1609), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90492] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1393), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1391), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90544] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1443), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1441), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90596] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1532), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1530), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90648] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1133), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1137), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90700] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1200), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90752] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1536), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1534), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90804] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1371), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90856] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1401), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1399), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90908] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1447), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1445), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [90960] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(933), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(931), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [91012] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1571), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1569), 35,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [91064] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2168), 1,
      anon_sym_COMMA,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1728), 1,
      aux_sym_expression_list_repeat1,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [91147] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2201), 1,
      anon_sym_COMMA,
    STATE(1728), 1,
      aux_sym_expression_list_repeat1,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1137), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [91230] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2203), 1,
      anon_sym_COMMA,
    STATE(1728), 1,
      aux_sym_expression_list_repeat1,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [91313] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2201), 1,
      anon_sym_COMMA,
    STATE(1728), 1,
      aux_sym_expression_list_repeat1,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1152), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [91396] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2206), 1,
      anon_sym_COMMA,
    STATE(1728), 1,
      aux_sym_expression_list_repeat1,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [91479] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2209), 1,
      anon_sym_COMMA,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1862), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [91561] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [91639] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2209), 1,
      anon_sym_COMMA,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1862), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1137), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [91721] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2594), 1,
      sym__primary_expression,
    STATE(2731), 1,
      sym_pattern,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2996), 2,
      sym_pattern_list,
      sym__for_target,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [91801] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2241), 1,
      anon_sym_COMMA,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1790), 1,
      aux_sym_expression_list_repeat1,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [91883] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2274), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1862), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [91965] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2277), 1,
      anon_sym_COMMA,
    STATE(1716), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1309), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1314), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [92017] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2201), 1,
      anon_sym_COMMA,
    STATE(1728), 1,
      aux_sym_expression_list_repeat1,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [92099] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
  [92179] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2282), 1,
      anon_sym_COMMA,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1794), 1,
      aux_sym_expression_list_repeat1,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1152), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [92261] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2314), 1,
      anon_sym_SEMI,
    STATE(1750), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 30,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [92313] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2201), 1,
      anon_sym_COMMA,
    STATE(1728), 1,
      aux_sym_expression_list_repeat1,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [92395] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
  [92475] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2316), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1862), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [92557] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 23,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [92619] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1314), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [92697] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2319), 1,
      anon_sym_COMMA,
    STATE(1794), 1,
      aux_sym_expression_list_repeat1,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [92779] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2322), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 27,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [92837] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2324), 1,
      anon_sym_COMMA,
    STATE(1716), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(505), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(503), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [92889] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1758), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 31,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [92939] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 13,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [93013] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2326), 1,
      anon_sym_SEMI,
    STATE(1731), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 30,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [93065] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 31,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [93115] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2329), 1,
      anon_sym_COMMA,
    STATE(1790), 1,
      aux_sym_expression_list_repeat1,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [93197] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2282), 1,
      anon_sym_COMMA,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1794), 1,
      aux_sym_expression_list_repeat1,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [93279] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [93357] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 21,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [93423] = 6,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2332), 1,
      sym_string,
    ACTIONS(2335), 1,
      sym_fstring_start,
    STATE(1737), 2,
      sym_f_string,
      aux_sym_concatenated_string_repeat1,
    ACTIONS(1141), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1139), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [93477] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [93555] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2338), 1,
      anon_sym_COMMA,
    STATE(1790), 1,
      aux_sym_expression_list_repeat1,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1152), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [93637] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2338), 1,
      anon_sym_COMMA,
    STATE(1790), 1,
      aux_sym_expression_list_repeat1,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1137), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [93719] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2282), 1,
      anon_sym_COMMA,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1794), 1,
      aux_sym_expression_list_repeat1,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [93801] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2201), 1,
      anon_sym_COMMA,
    STATE(1728), 1,
      aux_sym_expression_list_repeat1,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [93883] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2594), 1,
      sym__primary_expression,
    STATE(2731), 1,
      sym_pattern,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2969), 2,
      sym_pattern_list,
      sym__for_target,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [93963] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 31,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [94013] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2340), 1,
      anon_sym_COMMA,
    STATE(1794), 1,
      aux_sym_expression_list_repeat1,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [94095] = 6,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2343), 1,
      sym_string,
    STATE(1763), 2,
      sym_f_string,
      aux_sym_concatenated_string_repeat1,
    ACTIONS(927), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(925), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [94149] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [94227] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [94283] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2209), 1,
      anon_sym_COMMA,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1862), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [94365] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2345), 1,
      anon_sym_SEMI,
    STATE(1731), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 30,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [94417] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [94481] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2209), 1,
      anon_sym_COMMA,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1862), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1152), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [94563] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2347), 1,
      anon_sym_COMMA,
    STATE(1790), 1,
      aux_sym_expression_list_repeat1,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [94645] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2209), 1,
      anon_sym_COMMA,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1862), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [94727] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [94805] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2282), 1,
      anon_sym_COMMA,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1794), 1,
      aux_sym_expression_list_repeat1,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [94887] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 13,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [94961] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2353), 1,
      anon_sym_not,
    ACTIONS(2359), 1,
      anon_sym_is,
    STATE(1758), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2350), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2356), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 24,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [95019] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2362), 1,
      anon_sym_COMMA,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [95099] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2282), 1,
      anon_sym_COMMA,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1794), 1,
      aux_sym_expression_list_repeat1,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1137), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [95181] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [95233] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [95285] = 6,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2364), 1,
      sym_string,
    STATE(1737), 2,
      sym_f_string,
      aux_sym_concatenated_string_repeat1,
    ACTIONS(1070), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1068), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [95339] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2366), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1862), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [95421] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    ACTIONS(2369), 1,
      sym_verbatim,
    STATE(1536), 1,
      sym_f_string,
    STATE(2594), 1,
      sym__primary_expression,
    STATE(2731), 1,
      sym_pattern,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(2998), 2,
      sym_pattern_list,
      sym__for_target,
    ACTIONS(417), 4,
      sym_true,
      sym_false,
      sym_none,
      sym_number,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [95503] = 20,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    ACTIONS(2371), 1,
      sym_verbatim,
    STATE(1536), 1,
      sym_f_string,
    STATE(2594), 1,
      sym__primary_expression,
    STATE(2731), 1,
      sym_pattern,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    STATE(3104), 2,
      sym_pattern_list,
      sym__for_target,
    ACTIONS(417), 4,
      sym_true,
      sym_false,
      sym_none,
      sym_number,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [95585] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2373), 1,
      anon_sym_COMMA,
    STATE(1794), 1,
      aux_sym_expression_list_repeat1,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [95667] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 13,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [95741] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_or,
      anon_sym_QMARK,
  [95817] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 25,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [95875] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [95924] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [95997] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [96076] = 21,
    ACTIONS(1164), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2378), 1,
      anon_sym_COMMA,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1883), 1,
      aux_sym_expression_list_repeat1,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [96159] = 21,
    ACTIONS(1154), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    ACTIONS(2411), 1,
      anon_sym_COMMA,
    STATE(1883), 1,
      aux_sym_expression_list_repeat1,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1152), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [96242] = 21,
    ACTIONS(1072), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    ACTIONS(2413), 1,
      anon_sym_COMMA,
    STATE(1883), 1,
      aux_sym_expression_list_repeat1,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [96325] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [96402] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [96479] = 21,
    ACTIONS(1133), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    ACTIONS(2411), 1,
      anon_sym_COMMA,
    STATE(1883), 1,
      aux_sym_expression_list_repeat1,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1137), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [96562] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [96639] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2338), 1,
      anon_sym_COMMA,
    STATE(1790), 1,
      aux_sym_expression_list_repeat1,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [96720] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2338), 1,
      anon_sym_COMMA,
    STATE(1790), 1,
      aux_sym_expression_list_repeat1,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [96801] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2338), 1,
      anon_sym_COMMA,
    STATE(1790), 1,
      aux_sym_expression_list_repeat1,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [96882] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2416), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 26,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_RBRACE,
  [96939] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 26,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [96996] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2420), 1,
      anon_sym_SEMI,
    STATE(1787), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 29,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [97047] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2422), 1,
      anon_sym_SEMI,
    STATE(1788), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 29,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [97098] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2424), 1,
      anon_sym_SEMI,
    STATE(1788), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 29,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [97149] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2427), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 26,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [97206] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2429), 1,
      anon_sym_COMMA,
    STATE(1859), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(505), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(503), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [97257] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2431), 1,
      anon_sym_SEMI,
    STATE(1792), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 29,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [97308] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2433), 1,
      anon_sym_SEMI,
    STATE(1799), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 29,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [97359] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2435), 1,
      anon_sym_SEMI,
    STATE(1793), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 29,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [97410] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2438), 1,
      anon_sym_COMMA,
    STATE(1833), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(505), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(503), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [97461] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2440), 1,
      anon_sym_SEMI,
    STATE(1796), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 29,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [97512] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2442), 1,
      anon_sym_SEMI,
    STATE(1793), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 29,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [97563] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    ACTIONS(1950), 1,
      anon_sym_in,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2594), 1,
      sym__primary_expression,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [97642] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2444), 1,
      anon_sym_COMMA,
    STATE(1798), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1309), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1314), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [97693] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2447), 1,
      anon_sym_SEMI,
    STATE(1799), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 29,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [97744] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2450), 1,
      anon_sym_COMMA,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [97823] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [97900] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [97951] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98002] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98075] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98150] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 26,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98205] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98270] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 21,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98333] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98394] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98451] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2455), 1,
      anon_sym_not,
    ACTIONS(2461), 1,
      anon_sym_is,
    STATE(1811), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2452), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2458), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 23,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98508] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2464), 1,
      anon_sym_if,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98587] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98664] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98737] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2464), 1,
      anon_sym_if,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98816] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98893] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [98970] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [99047] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2466), 1,
      anon_sym_COMMA,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [99126] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [99199] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2468), 1,
      anon_sym_COMMA,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1932), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1152), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [99280] = 19,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1936), 1,
      anon_sym_in,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2594), 1,
      sym__primary_expression,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [99359] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1314), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [99436] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1811), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [99485] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2468), 1,
      anon_sym_COMMA,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1932), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1137), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [99566] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2500), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [99645] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [99722] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1314), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [99799] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [99872] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [99921] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [99970] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1875), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [100019] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2502), 1,
      anon_sym_COMMA,
    STATE(1833), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1309), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1314), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [100070] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100147] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100198] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100249] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100322] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100397] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 26,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100452] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100517] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 21,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100580] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100641] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100698] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2508), 1,
      anon_sym_not,
    ACTIONS(2514), 1,
      anon_sym_is,
    STATE(1844), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2505), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2511), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 23,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100755] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100834] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100911] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [100984] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [101063] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [101140] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [101217] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [101294] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2519), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [101373] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [101446] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1844), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [101495] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [101544] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1314), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [101621] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [101670] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 30,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [101719] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2521), 1,
      anon_sym_COMMA,
    STATE(1859), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1309), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1314), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [101770] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [101849] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2524), 1,
      anon_sym_COMMA,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1932), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [101930] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2527), 1,
      anon_sym_COMMA,
    STATE(1798), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(505), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(503), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [101981] = 21,
    ACTIONS(1105), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    ACTIONS(2529), 1,
      anon_sym_COMMA,
    STATE(1883), 1,
      aux_sym_expression_list_repeat1,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [102064] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [102141] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [102192] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [102243] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 12,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [102316] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_or,
      anon_sym_QMARK,
  [102391] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 26,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [102446] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [102511] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 21,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [102574] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [102635] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [102692] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2532), 1,
      anon_sym_COMMA,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1932), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [102773] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2538), 1,
      anon_sym_not,
    ACTIONS(2544), 1,
      anon_sym_is,
    STATE(1875), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2535), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2541), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 23,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [102830] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2547), 1,
      anon_sym_COMMA,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(1932), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [102911] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [102987] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2589), 1,
      sym__primary_expression,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [103063] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2580), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [103119] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2582), 1,
      anon_sym_COMMA,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2063), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [103199] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2615), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2063), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [103279] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2618), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2063), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1152), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [103359] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2620), 1,
      anon_sym_COMMA,
    STATE(1970), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(505), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(503), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [103409] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2622), 1,
      anon_sym_SEMI,
    STATE(1885), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 27,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [103459] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2624), 1,
      anon_sym_SEMI,
    STATE(1973), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 27,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [103509] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2626), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2063), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [103589] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [103665] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [103715] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [103765] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [103837] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_or,
      anon_sym_QMARK,
  [103911] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 25,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [103965] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [104029] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [104091] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 21,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [104151] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 23,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [104207] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2629), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 25,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [104263] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [104335] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2634), 1,
      anon_sym_not,
    ACTIONS(2640), 1,
      anon_sym_is,
    STATE(1899), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2631), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2637), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [104391] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2643), 1,
      anon_sym_if,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_QMARK,
  [104469] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2618), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2063), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [104549] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1899), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [104597] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [104673] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [104723] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [104773] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [104845] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [104919] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 25,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [104973] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105037] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105099] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 21,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105159] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 23,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105215] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2645), 1,
      anon_sym_COMMA,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2158), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1152), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [105295] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2680), 1,
      anon_sym_not,
    ACTIONS(2686), 1,
      anon_sym_is,
    STATE(1914), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2677), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2683), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105351] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105429] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2691), 1,
      anon_sym_SEMI,
    STATE(1927), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 28,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105479] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105551] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [105599] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [105647] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1914), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105695] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105771] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2645), 1,
      anon_sym_COMMA,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2158), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1137), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [105851] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [105923] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2693), 1,
      anon_sym_COMMA,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [106001] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [106077] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 11,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [106149] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2695), 1,
      anon_sym_SEMI,
    STATE(1929), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 28,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [106199] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [106277] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2697), 1,
      anon_sym_SEMI,
    STATE(1929), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 28,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [106327] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [106403] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    ACTIONS(2700), 1,
      anon_sym_COMMA,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2158), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [106483] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2703), 1,
      anon_sym_COMMA,
    STATE(1980), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(505), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(503), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [106533] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2705), 1,
      anon_sym_SEMI,
    STATE(1933), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 28,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [106583] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2708), 1,
      anon_sym_COMMA,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [106661] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    ACTIONS(2710), 1,
      anon_sym_COMMA,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2158), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [106741] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1314), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [106817] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    ACTIONS(2713), 1,
      anon_sym_COMMA,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2158), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [106897] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1942), 1,
      anon_sym_LPAREN,
    ACTIONS(1946), 1,
      anon_sym_LBRACK,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2594), 1,
      sym__primary_expression,
    ACTIONS(1934), 2,
      sym_this,
      sym_identifier,
    STATE(2514), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [106973] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [107049] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2645), 1,
      anon_sym_COMMA,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2158), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1103), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [107129] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2645), 1,
      anon_sym_COMMA,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2158), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [107209] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2645), 1,
      anon_sym_COMMA,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2158), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [107289] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2618), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2063), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1137), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [107369] = 19,
    ACTIONS(1309), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1314), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [107447] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1544), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1542), 30,
      sym__newline,
      sym_fstring_start,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      sym_string,
  [107493] = 19,
    ACTIONS(1224), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [107571] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [107621] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [107671] = 17,
    ACTIONS(1243), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [107745] = 18,
    ACTIONS(1243), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_or,
      anon_sym_QMARK,
  [107821] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 7,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [107875] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [107939] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2716), 1,
      anon_sym_COMMA,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [108017] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [108079] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [108139] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 7,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [108195] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2721), 1,
      anon_sym_not,
    ACTIONS(2727), 1,
      anon_sym_is,
    STATE(1957), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 4,
      anon_sym_EQ,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2718), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2724), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 21,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [108251] = 20,
    ACTIONS(1260), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    ACTIONS(2730), 1,
      anon_sym_if,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [108331] = 19,
    ACTIONS(1266), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [108409] = 17,
    ACTIONS(1270), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [108483] = 20,
    ACTIONS(1274), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    ACTIONS(2730), 1,
      anon_sym_if,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [108563] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2732), 1,
      anon_sym_SEMI,
    STATE(1968), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 28,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [108613] = 19,
    ACTIONS(1297), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [108691] = 19,
    ACTIONS(1305), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [108769] = 19,
    ACTIONS(1301), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [108847] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2734), 1,
      anon_sym_COMMA,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [108925] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2618), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2063), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1162), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [109005] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2736), 1,
      anon_sym_SEMI,
    STATE(1933), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 28,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [109055] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2738), 1,
      anon_sym_COMMA,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [109133] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2740), 1,
      anon_sym_COMMA,
    STATE(1970), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1309), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1314), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [109183] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [109259] = 20,
    ACTIONS(1287), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    ACTIONS(2743), 1,
      anon_sym_COMMA,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [109339] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2745), 1,
      anon_sym_SEMI,
    STATE(1973), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 27,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [109389] = 17,
    ACTIONS(1270), 1,
      anon_sym_EQ,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [109463] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1957), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [109511] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [109587] = 18,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(797), 1,
      anon_sym_LPAREN,
    ACTIONS(799), 1,
      anon_sym_LBRACK,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1948), 1,
      anon_sym_new,
    STATE(1536), 1,
      sym_f_string,
    STATE(2517), 1,
      sym_pattern,
    STATE(2589), 1,
      sym__primary_expression,
    ACTIONS(757), 2,
      sym_this,
      sym_identifier,
    STATE(1489), 2,
      sym_attribute,
      sym_subscript,
    STATE(2516), 2,
      sym_tuple_pattern,
      sym_list_pattern,
    ACTIONS(417), 5,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
    STATE(1682), 15,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [109663] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [109711] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 29,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
      aux_sym_interpolation_token1,
  [109759] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2748), 1,
      anon_sym_COMMA,
    STATE(1980), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1309), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1314), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [109809] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [109885] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1602), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1604), 30,
      sym__newline,
      sym_fstring_start,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      sym_string,
  [109931] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2751), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 25,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_as,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [109987] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [110035] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 9,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [110083] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2618), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2063), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1101), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [110163] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2643), 1,
      anon_sym_if,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_as,
      anon_sym_DOT,
      anon_sym_QMARK,
  [110241] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2763), 1,
      anon_sym_and,
    ACTIONS(2765), 1,
      anon_sym_or,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [110316] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [110371] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2794), 1,
      anon_sym_not,
    ACTIONS(2800), 1,
      anon_sym_is,
    STATE(1990), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2791), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2797), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 21,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [110426] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2803), 1,
      anon_sym_if,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 7,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [110503] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [110564] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    ACTIONS(2841), 1,
      anon_sym_COMMA,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [110641] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [110704] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [110779] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [110850] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2843), 1,
      anon_sym_SEMI,
    STATE(1997), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 27,
      sym__newline,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [110899] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [110960] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [111019] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2803), 1,
      anon_sym_if,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 7,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [111096] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2846), 1,
      anon_sym_SEMI,
    STATE(1997), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 27,
      sym__newline,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [111145] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [111220] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [111279] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [111334] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [111409] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2851), 1,
      anon_sym_not,
    ACTIONS(2857), 1,
      anon_sym_is,
    STATE(2006), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2848), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2854), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 21,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [111464] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [111539] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [111594] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2860), 1,
      anon_sym_if,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(2868), 1,
      anon_sym_or,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
  [111671] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2878), 1,
      anon_sym_COMMA,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [111748] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    ACTIONS(2880), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2556), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1152), 5,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [111829] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 8,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [111904] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2882), 1,
      anon_sym_COMMA,
    STATE(2013), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1309), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1314), 27,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [111953] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [112024] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2885), 1,
      anon_sym_SEMI,
    STATE(2001), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 27,
      sym__newline,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [112073] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [112120] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2763), 1,
      anon_sym_and,
    ACTIONS(2765), 1,
      anon_sym_or,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112195] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112244] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112293] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2887), 1,
      anon_sym_LPAREN,
    ACTIONS(2889), 1,
      anon_sym_LBRACK,
    ACTIONS(2891), 1,
      anon_sym_DOT,
    STATE(2195), 1,
      sym_argument_list,
    ACTIONS(1212), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1210), 25,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [112346] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 10,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112417] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2763), 1,
      anon_sym_and,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112490] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112543] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112606] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112667] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112726] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112781] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2896), 1,
      anon_sym_not,
    ACTIONS(2902), 1,
      anon_sym_is,
    STATE(2028), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2893), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2899), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 21,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112836] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2763), 1,
      anon_sym_and,
    ACTIONS(2765), 1,
      anon_sym_or,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    ACTIONS(2905), 1,
      anon_sym_if,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112913] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2763), 1,
      anon_sym_and,
    ACTIONS(2765), 1,
      anon_sym_or,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [112988] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [113059] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2763), 1,
      anon_sym_and,
    ACTIONS(2765), 1,
      anon_sym_or,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    ACTIONS(2905), 1,
      anon_sym_if,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [113136] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [113185] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2763), 1,
      anon_sym_and,
    ACTIONS(2765), 1,
      anon_sym_or,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [113260] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2763), 1,
      anon_sym_and,
    ACTIONS(2765), 1,
      anon_sym_or,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [113335] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [113382] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2907), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [113459] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 10,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [113530] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2909), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 24,
      anon_sym_SEMI,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [113585] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(2868), 1,
      anon_sym_or,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [113660] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [113731] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2911), 1,
      anon_sym_SEMI,
    STATE(2043), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 27,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [113780] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2913), 1,
      anon_sym_SEMI,
    STATE(2047), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 27,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [113829] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2860), 1,
      anon_sym_if,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(2868), 1,
      anon_sym_or,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
  [113906] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2915), 1,
      anon_sym_SEMI,
    STATE(2046), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 27,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [113955] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2917), 1,
      anon_sym_SEMI,
    STATE(2109), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 27,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [114004] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2919), 1,
      anon_sym_SEMI,
    STATE(2047), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 27,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [114053] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 8,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [114128] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(2868), 1,
      anon_sym_or,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [114203] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2922), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 24,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [114258] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(2868), 1,
      anon_sym_or,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [114333] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(2868), 1,
      anon_sym_or,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [114408] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2924), 1,
      anon_sym_SEMI,
    STATE(2055), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 27,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [114457] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2066), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [114504] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2926), 1,
      anon_sym_SEMI,
    STATE(2060), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 27,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [114553] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_or,
      anon_sym_QMARK,
  [114626] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2928), 1,
      anon_sym_if,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [114703] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2930), 1,
      anon_sym_SEMI,
    STATE(2059), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 27,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [114752] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2932), 1,
      anon_sym_SEMI,
    STATE(2106), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 27,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [114801] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2934), 1,
      anon_sym_SEMI,
    STATE(2060), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 27,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [114850] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2937), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [114905] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    ACTIONS(2880), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2556), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1137), 5,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [114986] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2939), 1,
      anon_sym_COMMA,
    STATE(2013), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(505), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(503), 27,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [115035] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2941), 1,
      anon_sym_SEMI,
    STATE(2065), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 27,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [115084] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2943), 1,
      anon_sym_SEMI,
    STATE(2075), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 27,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [115133] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2948), 1,
      anon_sym_not,
    ACTIONS(2954), 1,
      anon_sym_is,
    STATE(2066), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2945), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2951), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 21,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [115188] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2967), 1,
      anon_sym_and,
    ACTIONS(2969), 1,
      anon_sym_or,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [115263] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [115312] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [115361] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 10,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [115432] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 8,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [115507] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(2987), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 24,
      anon_sym_SEMI,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_RBRACE,
  [115562] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2967), 1,
      anon_sym_and,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_or,
      anon_sym_QMARK,
  [115635] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [115684] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2989), 1,
      anon_sym_SEMI,
    STATE(2075), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 27,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [115733] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    ACTIONS(2880), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2556), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1162), 5,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [115814] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    ACTIONS(2880), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2556), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1101), 5,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [115895] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [115948] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [116023] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [116072] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [116121] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [116192] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_or,
      anon_sym_QMARK,
  [116265] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [116318] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [116381] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [116442] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [116501] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [116556] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2995), 1,
      anon_sym_not,
    ACTIONS(3001), 1,
      anon_sym_is,
    STATE(2089), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(2992), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2998), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 21,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [116611] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    ACTIONS(3004), 1,
      anon_sym_if,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_else,
      anon_sym_QMARK,
  [116688] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [116763] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [116834] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    ACTIONS(3004), 1,
      anon_sym_if,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_else,
      anon_sym_QMARK,
  [116911] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [116986] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [117061] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [117136] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    ACTIONS(3006), 1,
      anon_sym_COMMA,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [117213] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [117284] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2089), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [117331] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [117380] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [117427] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 28,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [117474] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1314), 8,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [117549] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [117620] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(1990), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [117667] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3008), 1,
      anon_sym_SEMI,
    STATE(2106), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 27,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [117716] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2928), 1,
      anon_sym_if,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [117793] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 10,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [117864] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3011), 1,
      anon_sym_SEMI,
    STATE(2109), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 27,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [117913] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [117984] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2145), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 28,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [118031] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [118102] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2006), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 28,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [118149] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_or,
      anon_sym_QMARK,
  [118222] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [118293] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2028), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 28,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [118340] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1314), 7,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [118417] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 8,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [118492] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2887), 1,
      anon_sym_LPAREN,
    ACTIONS(2889), 1,
      anon_sym_LBRACK,
    ACTIONS(2891), 1,
      anon_sym_DOT,
    ACTIONS(3014), 1,
      anon_sym_QMARK,
    STATE(2195), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 24,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [118547] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 8,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [118622] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [118693] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [118746] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [118795] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [118848] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [118897] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 10,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [118968] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119031] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119092] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2655), 1,
      anon_sym_not,
    ACTIONS(2657), 1,
      anon_sym_and,
    ACTIONS(2659), 1,
      anon_sym_or,
    ACTIONS(2663), 1,
      anon_sym_PIPE,
    ACTIONS(2665), 1,
      anon_sym_CARET,
    ACTIONS(2667), 1,
      anon_sym_AMP,
    ACTIONS(2671), 1,
      anon_sym_GT_GT,
    ACTIONS(2675), 1,
      anon_sym_is,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2649), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2661), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2669), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2653), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2647), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2673), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1314), 8,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [119167] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119226] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_or,
      anon_sym_QMARK,
  [119299] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 24,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119352] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 22,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119407] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3016), 1,
      anon_sym_COMMA,
    STATE(2134), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1309), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1314), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119456] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119503] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 28,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119550] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119613] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119674] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 28,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119721] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 28,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119768] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 28,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119815] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 28,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [119862] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 28,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [119909] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 28,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
      anon_sym_RBRACE,
  [119956] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3022), 1,
      anon_sym_not,
    ACTIONS(3028), 1,
      anon_sym_is,
    STATE(2145), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(3019), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3025), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 21,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [120011] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2967), 1,
      anon_sym_and,
    ACTIONS(2969), 1,
      anon_sym_or,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    ACTIONS(3031), 1,
      anon_sym_if,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
  [120088] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 8,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [120163] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2967), 1,
      anon_sym_and,
    ACTIONS(2969), 1,
      anon_sym_or,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [120238] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 10,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [120309] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [120372] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 20,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [120431] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2967), 1,
      anon_sym_and,
    ACTIONS(2969), 1,
      anon_sym_or,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    ACTIONS(3031), 1,
      anon_sym_if,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_for,
      anon_sym_QMARK,
  [120508] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(3033), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1291), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [120585] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2967), 1,
      anon_sym_and,
    ACTIONS(2969), 1,
      anon_sym_or,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [120660] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2967), 1,
      anon_sym_and,
    ACTIONS(2969), 1,
      anon_sym_or,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [120735] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2967), 1,
      anon_sym_and,
    ACTIONS(2969), 1,
      anon_sym_or,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [120810] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(3035), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 24,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [120865] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3037), 1,
      anon_sym_COMMA,
    STATE(2134), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(505), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(503), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [120914] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(2868), 1,
      anon_sym_or,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_for,
      anon_sym_QMARK,
  [120989] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2651), 1,
      anon_sym_STAR_STAR,
    STATE(2099), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 27,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121038] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121084] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3039), 1,
      anon_sym_SEMI,
    STATE(2164), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 26,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121132] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 6,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_else,
      anon_sym_QMARK,
  [121208] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3073), 1,
      anon_sym_SEMI,
    STATE(2166), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 26,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121256] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2887), 1,
      anon_sym_LPAREN,
    ACTIONS(2889), 1,
      anon_sym_LBRACK,
    ACTIONS(2891), 1,
      anon_sym_DOT,
    ACTIONS(3075), 1,
      anon_sym_QMARK,
    STATE(2195), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 23,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [121310] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3077), 1,
      anon_sym_SEMI,
    STATE(2166), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 26,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121358] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [121432] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 9,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [121502] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1575), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1573), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121546] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1540), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1538), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121590] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1579), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1577), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121634] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [121708] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2194), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 27,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121754] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(3106), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 23,
      anon_sym_SEMI,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [121808] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1393), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1391), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121852] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(927), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(925), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121896] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1554), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1552), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121940] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3108), 2,
      sym__newline,
      anon_sym_SEMI,
    ACTIONS(3113), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3111), 26,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [121986] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1397), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1395), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122030] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [122104] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [122178] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 26,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122226] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 26,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122274] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [122344] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 8,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_or,
      anon_sym_QMARK,
  [122416] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 23,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122468] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 17,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122530] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122590] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3119), 1,
      anon_sym_SEMI,
    STATE(2189), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 26,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122638] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [122712] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122770] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 21,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122824] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(3122), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 23,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [122878] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3127), 1,
      anon_sym_not,
    ACTIONS(3133), 1,
      anon_sym_is,
    STATE(2194), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(3124), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3130), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 20,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [122932] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1407), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1405), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [122976] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3146), 1,
      anon_sym_and,
    ACTIONS(3148), 1,
      anon_sym_or,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [123050] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 26,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [123098] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 26,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [123146] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [123216] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3146), 1,
      anon_sym_and,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_or,
      anon_sym_QMARK,
  [123288] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 23,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [123340] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 17,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [123402] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [123462] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [123520] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 21,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [123574] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3169), 1,
      anon_sym_not,
    ACTIONS(3175), 1,
      anon_sym_is,
    STATE(2206), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(3166), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3172), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 20,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [123628] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3146), 1,
      anon_sym_and,
    ACTIONS(3148), 1,
      anon_sym_or,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    ACTIONS(3178), 1,
      anon_sym_if,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 6,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [123704] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3146), 1,
      anon_sym_and,
    ACTIONS(3148), 1,
      anon_sym_or,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [123778] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [123848] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3146), 1,
      anon_sym_and,
    ACTIONS(3148), 1,
      anon_sym_or,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    ACTIONS(3178), 1,
      anon_sym_if,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 6,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [123924] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3146), 1,
      anon_sym_and,
    ACTIONS(3148), 1,
      anon_sym_or,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1295), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [123998] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1072), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1101), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124042] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3146), 1,
      anon_sym_and,
    ACTIONS(3148), 1,
      anon_sym_or,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [124116] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3146), 1,
      anon_sym_and,
    ACTIONS(3148), 1,
      anon_sym_or,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [124190] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1389), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1387), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124234] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1401), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1399), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124278] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1411), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1409), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124322] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    ACTIONS(3180), 1,
      anon_sym_if,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 6,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [124398] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1592), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1590), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124442] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1558), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1556), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124486] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1516), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1514), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124530] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(933), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(931), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124574] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1415), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1413), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124618] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3182), 1,
      anon_sym_SEMI,
    STATE(2225), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 26,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124666] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3184), 1,
      anon_sym_SEMI,
    STATE(2189), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 26,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124714] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2464), 1,
      anon_sym_if,
    ACTIONS(3186), 1,
      anon_sym_COLON,
    ACTIONS(3190), 1,
      anon_sym_for,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2578), 1,
      sym_for_in_clause,
    STATE(3025), 1,
      sym__comprehension_clauses,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3188), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [124798] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1299), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [124872] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1443), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1441), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124916] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1447), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1445), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [124960] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1524), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1522), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125004] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(939), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(937), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125048] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1528), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1526), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125092] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1548), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1546), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125136] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1617), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1615), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125180] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3192), 1,
      anon_sym_SEMI,
    STATE(2235), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 26,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125228] = 24,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3188), 1,
      anon_sym_COMMA,
    ACTIONS(3195), 1,
      anon_sym_COLON,
    ACTIONS(3197), 1,
      anon_sym_RBRACK,
    ACTIONS(3199), 1,
      anon_sym_for,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2570), 1,
      sym_for_in_clause,
    STATE(3001), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [125314] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3201), 1,
      anon_sym_SEMI,
    STATE(2238), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1363), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1361), 26,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125362] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3203), 1,
      anon_sym_SEMI,
    STATE(2235), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(73), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(109), 26,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125410] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1562), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1560), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125454] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1621), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1619), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125498] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1133), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1137), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125542] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1154), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1152), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125586] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1371), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1369), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125630] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1344), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1342), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125674] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1631), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1629), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125718] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2464), 1,
      anon_sym_if,
    ACTIONS(3186), 1,
      anon_sym_COLON,
    ACTIONS(3190), 1,
      anon_sym_for,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2578), 1,
      sym_for_in_clause,
    STATE(3039), 1,
      sym__comprehension_clauses,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3188), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [125802] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1451), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1449), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125846] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2464), 1,
      anon_sym_if,
    ACTIONS(3186), 1,
      anon_sym_COLON,
    ACTIONS(3190), 1,
      anon_sym_for,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2578), 1,
      sym_for_in_clause,
    STATE(3052), 1,
      sym__comprehension_clauses,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3188), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [125930] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1190), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1192), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [125974] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1512), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1510), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [126018] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [126088] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2288), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [126134] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1635), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1633), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [126178] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1532), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1530), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [126222] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    ACTIONS(3205), 1,
      anon_sym_COMMA,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1291), 5,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [126300] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1222), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [126374] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [126448] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1200), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1198), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [126492] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 9,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [126562] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 26,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [126610] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1228), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 26,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [126658] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [126728] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [126798] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2206), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1359), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1357), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [126844] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1464), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1462), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [126888] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    ACTIONS(3180), 1,
      anon_sym_if,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1272), 6,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [126964] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1468), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1466), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127008] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1611), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1609), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127052] = 17,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1241), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_or,
      anon_sym_QMARK,
  [127124] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1472), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1470), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127168] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1480), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1478), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127212] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1594), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1596), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127256] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1571), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1569), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127300] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1484), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1482), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127344] = 24,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3188), 1,
      anon_sym_COMMA,
    ACTIONS(3195), 1,
      anon_sym_COLON,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3207), 1,
      anon_sym_RBRACK,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2570), 1,
      sym_for_in_clause,
    STATE(3006), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [127430] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 23,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127482] = 12,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 17,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127544] = 11,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 18,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127604] = 10,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127662] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1626), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1623), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127706] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127752] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127798] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 27,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127844] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1228), 6,
      anon_sym_in,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1226), 21,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127898] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1536), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1534), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [127942] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1303), 7,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_QMARK,
  [128016] = 24,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3188), 1,
      anon_sym_COMMA,
    ACTIONS(3195), 1,
      anon_sym_COLON,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3209), 1,
      anon_sym_RBRACK,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2570), 1,
      sym_for_in_clause,
    STATE(3013), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [128102] = 8,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3214), 1,
      anon_sym_not,
    ACTIONS(3220), 1,
      anon_sym_is,
    STATE(2288), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1247), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
    ACTIONS(3211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3217), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1245), 20,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_QMARK,
  [128156] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1474), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1476), 28,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [128200] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1258), 6,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_else,
      anon_sym_QMARK,
  [128276] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1264), 7,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_QMARK,
  [128350] = 16,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1268), 9,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_else,
      anon_sym_and,
      anon_sym_or,
      anon_sym_QMARK,
  [128420] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1385), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1383), 27,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [128466] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1194), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1196), 27,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
      anon_sym_QMARK,
  [128512] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2464), 1,
      anon_sym_if,
    ACTIONS(3186), 1,
      anon_sym_COLON,
    ACTIONS(3190), 1,
      anon_sym_for,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2578), 1,
      sym_for_in_clause,
    STATE(3040), 1,
      sym__comprehension_clauses,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3188), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [128596] = 24,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3188), 1,
      anon_sym_COMMA,
    ACTIONS(3195), 1,
      anon_sym_COLON,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3223), 1,
      anon_sym_RBRACK,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2570), 1,
      sym_for_in_clause,
    STATE(3026), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [128682] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(681), 1,
      anon_sym_async,
    ACTIONS(683), 1,
      anon_sym_def,
    ACTIONS(697), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [128747] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3188), 1,
      anon_sym_COMMA,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3229), 1,
      anon_sym_RPAREN,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(2978), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [128830] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(3231), 1,
      anon_sym_COMMA,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2559), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3233), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_for,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [128907] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3235), 1,
      anon_sym_COMMA,
    ACTIONS(3238), 1,
      anon_sym_RPAREN,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(3004), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [128990] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3188), 1,
      anon_sym_COMMA,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3240), 1,
      anon_sym_RPAREN,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(3073), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [129073] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(57), 1,
      anon_sym_new,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(649), 1,
      anon_sym_async,
    ACTIONS(651), 1,
      anon_sym_def,
    ACTIONS(3242), 1,
      anon_sym_LPAREN,
    ACTIONS(3244), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2020), 1,
      sym__primary_expression,
    ACTIONS(63), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(2176), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [129138] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1295), 5,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [129213] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(557), 1,
      anon_sym_async,
    ACTIONS(559), 1,
      anon_sym_def,
    ACTIONS(569), 1,
      anon_sym_new,
    ACTIONS(3246), 1,
      anon_sym_LPAREN,
    ACTIONS(3248), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(538), 1,
      sym__primary_expression,
    ACTIONS(105), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(641), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [129278] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2570), 1,
      sym_for_in_clause,
    STATE(3013), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3188), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [129359] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3188), 1,
      anon_sym_COMMA,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3238), 1,
      anon_sym_RPAREN,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(2994), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [129442] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(389), 1,
      anon_sym_async,
    ACTIONS(391), 1,
      anon_sym_def,
    ACTIONS(411), 1,
      anon_sym_new,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [129507] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1303), 5,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [129582] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(61), 1,
      anon_sym_LBRACE,
    ACTIONS(65), 1,
      sym_string,
    ACTIONS(67), 1,
      sym_regex,
    ACTIONS(69), 1,
      sym_fstring_start,
    ACTIONS(897), 1,
      anon_sym_async,
    ACTIONS(899), 1,
      anon_sym_def,
    ACTIONS(909), 1,
      anon_sym_new,
    ACTIONS(3242), 1,
      anon_sym_LPAREN,
    ACTIONS(3244), 1,
      anon_sym_LBRACK,
    STATE(1746), 1,
      sym_f_string,
    STATE(2020), 1,
      sym__primary_expression,
    ACTIONS(63), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(2176), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [129647] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(3094), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3250), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [129728] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(759), 1,
      anon_sym_async,
    ACTIONS(761), 1,
      anon_sym_def,
    ACTIONS(771), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [129793] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(1299), 5,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [129868] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3229), 1,
      anon_sym_RPAREN,
    ACTIONS(3235), 1,
      anon_sym_COMMA,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(2995), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [129951] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(1938), 1,
      anon_sym_async,
    ACTIONS(1940), 1,
      anon_sym_def,
    ACTIONS(1948), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [130016] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2570), 1,
      sym_for_in_clause,
    STATE(3001), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3188), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [130097] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(583), 1,
      anon_sym_async,
    ACTIONS(585), 1,
      anon_sym_def,
    ACTIONS(599), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [130162] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(533), 1,
      anon_sym_async,
    ACTIONS(535), 1,
      anon_sym_def,
    ACTIONS(549), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [130227] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(3022), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3250), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [130308] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(3069), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3250), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [130389] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(455), 1,
      anon_sym_async,
    ACTIONS(457), 1,
      anon_sym_def,
    ACTIONS(471), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [130454] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3235), 1,
      anon_sym_COMMA,
    ACTIONS(3240), 1,
      anon_sym_RPAREN,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(3063), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [130537] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(479), 1,
      anon_sym_async,
    ACTIONS(481), 1,
      anon_sym_def,
    ACTIONS(499), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [130602] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2570), 1,
      sym_for_in_clause,
    STATE(3006), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3188), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [130683] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(881), 1,
      anon_sym_async,
    ACTIONS(883), 1,
      anon_sym_def,
    ACTIONS(889), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [130748] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3235), 1,
      anon_sym_COMMA,
    ACTIONS(3252), 1,
      anon_sym_RPAREN,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(3056), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [130831] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(913), 1,
      anon_sym_async,
    ACTIONS(915), 1,
      anon_sym_def,
    ACTIONS(921), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [130896] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3188), 1,
      anon_sym_COMMA,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3252), 1,
      anon_sym_RPAREN,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(2989), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [130979] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(805), 1,
      anon_sym_async,
    ACTIONS(807), 1,
      anon_sym_def,
    ACTIONS(815), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [131044] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(607), 1,
      anon_sym_async,
    ACTIONS(609), 1,
      anon_sym_def,
    ACTIONS(627), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [131109] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(821), 1,
      anon_sym_async,
    ACTIONS(823), 1,
      anon_sym_def,
    ACTIONS(829), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [131174] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(847), 1,
      anon_sym_async,
    ACTIONS(849), 1,
      anon_sym_def,
    ACTIONS(855), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [131239] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(415), 1,
      anon_sym_LBRACE,
    ACTIONS(419), 1,
      sym_string,
    ACTIONS(421), 1,
      sym_regex,
    ACTIONS(423), 1,
      sym_fstring_start,
    ACTIONS(835), 1,
      anon_sym_async,
    ACTIONS(837), 1,
      anon_sym_def,
    ACTIONS(843), 1,
      anon_sym_new,
    ACTIONS(3225), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_LBRACK,
    STATE(1536), 1,
      sym_f_string,
    STATE(1608), 1,
      sym__primary_expression,
    ACTIONS(417), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(1682), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [131304] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(77), 1,
      anon_sym_async,
    ACTIONS(79), 1,
      anon_sym_def,
    ACTIONS(99), 1,
      anon_sym_new,
    ACTIONS(103), 1,
      anon_sym_LBRACE,
    ACTIONS(107), 1,
      sym_string,
    ACTIONS(111), 1,
      sym_regex,
    ACTIONS(113), 1,
      sym_fstring_start,
    ACTIONS(3246), 1,
      anon_sym_LPAREN,
    ACTIONS(3248), 1,
      anon_sym_LBRACK,
    STATE(450), 1,
      sym_f_string,
    STATE(538), 1,
      sym__primary_expression,
    ACTIONS(105), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(641), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [131369] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2570), 1,
      sym_for_in_clause,
    STATE(3026), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3188), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [131450] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(2968), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3250), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [131531] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(509), 1,
      anon_sym_async,
    ACTIONS(511), 1,
      anon_sym_def,
    ACTIONS(521), 1,
      anon_sym_new,
    ACTIONS(3254), 1,
      anon_sym_LPAREN,
    ACTIONS(3256), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(464), 1,
      sym__primary_expression,
    ACTIONS(151), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(646), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [131596] = 14,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(123), 1,
      anon_sym_async,
    ACTIONS(125), 1,
      anon_sym_def,
    ACTIONS(145), 1,
      anon_sym_new,
    ACTIONS(149), 1,
      anon_sym_LBRACE,
    ACTIONS(153), 1,
      sym_string,
    ACTIONS(155), 1,
      sym_regex,
    ACTIONS(157), 1,
      sym_fstring_start,
    ACTIONS(3254), 1,
      anon_sym_LPAREN,
    ACTIONS(3256), 1,
      anon_sym_LBRACK,
    STATE(458), 1,
      sym_f_string,
    STATE(464), 1,
      sym__primary_expression,
    ACTIONS(151), 7,
      sym_this,
      sym_true,
      sym_false,
      sym_none,
      sym_verbatim,
      sym_number,
      sym_identifier,
    STATE(646), 17,
      sym_attribute,
      sym_subscript,
      sym_call,
      sym_slice_call,
      sym_new_expression,
      sym_existential,
      sym_anonymous_function,
      sym_list,
      sym_set,
      sym_tuple,
      sym_parenthesized_expression,
      sym_dictionary,
      sym_list_comprehension,
      sym_set_comprehension,
      sym_dictionary_comprehension,
      sym_generator_expression,
      sym_concatenated_string,
  [131661] = 23,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2175), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2179), 1,
      anon_sym_not,
    ACTIONS(2181), 1,
      anon_sym_and,
    ACTIONS(2183), 1,
      anon_sym_or,
    ACTIONS(2187), 1,
      anon_sym_PIPE,
    ACTIONS(2189), 1,
      anon_sym_CARET,
    ACTIONS(2191), 1,
      anon_sym_AMP,
    ACTIONS(2195), 1,
      anon_sym_GT_GT,
    ACTIONS(2199), 1,
      anon_sym_is,
    ACTIONS(2280), 1,
      anon_sym_if,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3235), 1,
      anon_sym_COMMA,
    ACTIONS(3252), 1,
      anon_sym_RPAREN,
    STATE(1729), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2566), 1,
      sym_for_in_clause,
    STATE(2987), 1,
      sym__comprehension_clauses,
    ACTIONS(2173), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2185), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2193), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2177), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2171), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2197), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [131744] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2928), 1,
      anon_sym_if,
    ACTIONS(3258), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2632), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1137), 2,
      sym__newline,
      anon_sym_SEMI,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [131822] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2928), 1,
      anon_sym_if,
    ACTIONS(3258), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2632), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1162), 2,
      sym__newline,
      anon_sym_SEMI,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [131900] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    ACTIONS(3195), 1,
      anon_sym_COLON,
    ACTIONS(3260), 1,
      anon_sym_COMMA,
    ACTIONS(3262), 1,
      anon_sym_RBRACK,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2621), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [131980] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    ACTIONS(3195), 1,
      anon_sym_COLON,
    ACTIONS(3260), 1,
      anon_sym_COMMA,
    ACTIONS(3264), 1,
      anon_sym_RBRACK,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2621), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132060] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2928), 1,
      anon_sym_if,
    ACTIONS(3258), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2632), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1101), 2,
      sym__newline,
      anon_sym_SEMI,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132138] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    ACTIONS(3195), 1,
      anon_sym_COLON,
    ACTIONS(3260), 1,
      anon_sym_COMMA,
    ACTIONS(3266), 1,
      anon_sym_RBRACK,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2621), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132218] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    ACTIONS(3195), 1,
      anon_sym_COLON,
    ACTIONS(3260), 1,
      anon_sym_COMMA,
    ACTIONS(3268), 1,
      anon_sym_RBRACK,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2621), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132298] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2928), 1,
      anon_sym_if,
    ACTIONS(3258), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2632), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1152), 2,
      sym__newline,
      anon_sym_SEMI,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132376] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(3270), 1,
      anon_sym_COMMA,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2583), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3233), 3,
      anon_sym_if,
      anon_sym_for,
      anon_sym_RBRACE,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132452] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2643), 1,
      anon_sym_if,
    ACTIONS(3272), 1,
      anon_sym_COLON,
    ACTIONS(3274), 1,
      anon_sym_COMMA,
    ACTIONS(3276), 1,
      anon_sym_as,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2630), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132532] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    ACTIONS(3195), 1,
      anon_sym_COLON,
    ACTIONS(3260), 1,
      anon_sym_COMMA,
    ACTIONS(3278), 1,
      anon_sym_RBRACK,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2621), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132612] = 22,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2643), 1,
      anon_sym_if,
    ACTIONS(3274), 1,
      anon_sym_COMMA,
    ACTIONS(3280), 1,
      anon_sym_COLON,
    ACTIONS(3282), 1,
      anon_sym_as,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2608), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132692] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    ACTIONS(2880), 1,
      anon_sym_COMMA,
    ACTIONS(3284), 1,
      anon_sym_COLON,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2556), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132769] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3188), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132842] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3286), 1,
      anon_sym_COLON,
    ACTIONS(3288), 1,
      anon_sym_RBRACE,
    ACTIONS(3290), 1,
      aux_sym_interpolation_token1,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132919] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2248), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2252), 1,
      anon_sym_not,
    ACTIONS(2254), 1,
      anon_sym_and,
    ACTIONS(2256), 1,
      anon_sym_or,
    ACTIONS(2260), 1,
      anon_sym_PIPE,
    ACTIONS(2262), 1,
      anon_sym_CARET,
    ACTIONS(2264), 1,
      anon_sym_AMP,
    ACTIONS(2268), 1,
      anon_sym_GT_GT,
    ACTIONS(2272), 1,
      anon_sym_is,
    ACTIONS(2464), 1,
      anon_sym_if,
    STATE(1824), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2246), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2258), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2266), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2250), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3292), 3,
      anon_sym_COMMA,
      anon_sym_for,
      anon_sym_RBRACE,
    ACTIONS(2244), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2270), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [132992] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2643), 1,
      anon_sym_if,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(1314), 3,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_as,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133065] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2928), 1,
      anon_sym_if,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(1314), 3,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133138] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3294), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133211] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(1314), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133284] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2589), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2593), 1,
      anon_sym_not,
    ACTIONS(2595), 1,
      anon_sym_and,
    ACTIONS(2597), 1,
      anon_sym_or,
    ACTIONS(2601), 1,
      anon_sym_PIPE,
    ACTIONS(2603), 1,
      anon_sym_CARET,
    ACTIONS(2605), 1,
      anon_sym_AMP,
    ACTIONS(2609), 1,
      anon_sym_GT_GT,
    ACTIONS(2613), 1,
      anon_sym_is,
    ACTIONS(2928), 1,
      anon_sym_if,
    ACTIONS(3296), 1,
      anon_sym_COMMA,
    STATE(2054), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1291), 2,
      sym__newline,
      anon_sym_SEMI,
    ACTIONS(2587), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2599), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2607), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2591), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2585), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2611), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133359] = 21,
    ACTIONS(1137), 1,
      anon_sym_RPAREN,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    ACTIONS(3260), 1,
      anon_sym_COMMA,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2621), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133436] = 21,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    ACTIONS(2880), 1,
      anon_sym_COMMA,
    ACTIONS(3298), 1,
      anon_sym_COLON,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    STATE(2556), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133513] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2385), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2389), 1,
      anon_sym_not,
    ACTIONS(2391), 1,
      anon_sym_and,
    ACTIONS(2393), 1,
      anon_sym_or,
    ACTIONS(2397), 1,
      anon_sym_PIPE,
    ACTIONS(2399), 1,
      anon_sym_CARET,
    ACTIONS(2401), 1,
      anon_sym_AMP,
    ACTIONS(2405), 1,
      anon_sym_GT_GT,
    ACTIONS(2409), 1,
      anon_sym_is,
    ACTIONS(2730), 1,
      anon_sym_if,
    ACTIONS(3300), 1,
      anon_sym_EQ,
    STATE(1975), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2383), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2395), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3302), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2387), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2381), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2407), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133588] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1623), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(1411), 8,
      anon_sym_in,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_GT_GT,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(1409), 21,
      anon_sym_if,
      anon_sym_STAR_STAR,
      anon_sym_AT,
      anon_sym_not,
      anon_sym_and,
      anon_sym_or,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_is,
      anon_sym_instanceof,
  [133631] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2474), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2478), 1,
      anon_sym_not,
    ACTIONS(2480), 1,
      anon_sym_and,
    ACTIONS(2482), 1,
      anon_sym_or,
    ACTIONS(2486), 1,
      anon_sym_PIPE,
    ACTIONS(2488), 1,
      anon_sym_CARET,
    ACTIONS(2490), 1,
      anon_sym_AMP,
    ACTIONS(2494), 1,
      anon_sym_GT_GT,
    ACTIONS(2498), 1,
      anon_sym_is,
    ACTIONS(2643), 1,
      anon_sym_if,
    ACTIONS(3306), 1,
      anon_sym_as,
    STATE(1902), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2472), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2484), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2492), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3304), 2,
      anon_sym_COLON,
      anon_sym_COMMA,
    ACTIONS(2476), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2470), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2496), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133706] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3308), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133778] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3310), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133850] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2829), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2835), 1,
      anon_sym_AMP,
    ACTIONS(2839), 1,
      anon_sym_GT_GT,
    ACTIONS(2864), 1,
      anon_sym_not,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(2868), 1,
      anon_sym_or,
    ACTIONS(2870), 1,
      anon_sym_PIPE,
    ACTIONS(2872), 1,
      anon_sym_CARET,
    ACTIONS(2876), 1,
      anon_sym_is,
    STATE(2113), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2827), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2833), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2837), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2831), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3312), 3,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_for,
    ACTIONS(2862), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2874), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133920] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3188), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [133992] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    ACTIONS(3180), 1,
      anon_sym_if,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1295), 2,
      sym__newline,
      anon_sym_SEMI,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134064] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3314), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134136] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3316), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134208] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3318), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134280] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3320), 2,
      anon_sym_COLON,
      anon_sym_COMMA,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134352] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2803), 1,
      anon_sym_if,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    ACTIONS(3322), 1,
      anon_sym_COLON,
    ACTIONS(3324), 1,
      anon_sym_RBRACK,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134426] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2961), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2965), 1,
      anon_sym_not,
    ACTIONS(2967), 1,
      anon_sym_and,
    ACTIONS(2969), 1,
      anon_sym_or,
    ACTIONS(2973), 1,
      anon_sym_PIPE,
    ACTIONS(2975), 1,
      anon_sym_CARET,
    ACTIONS(2977), 1,
      anon_sym_AMP,
    ACTIONS(2981), 1,
      anon_sym_GT_GT,
    ACTIONS(2985), 1,
      anon_sym_is,
    STATE(2111), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2959), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2979), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2963), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3312), 3,
      anon_sym_RPAREN,
      anon_sym_if,
      anon_sym_for,
    ACTIONS(2957), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2983), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134496] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    ACTIONS(3180), 1,
      anon_sym_if,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1299), 2,
      sym__newline,
      anon_sym_SEMI,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134568] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3292), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134640] = 18,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2757), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2761), 1,
      anon_sym_not,
    ACTIONS(2763), 1,
      anon_sym_and,
    ACTIONS(2765), 1,
      anon_sym_or,
    ACTIONS(2769), 1,
      anon_sym_PIPE,
    ACTIONS(2771), 1,
      anon_sym_CARET,
    ACTIONS(2773), 1,
      anon_sym_AMP,
    ACTIONS(2777), 1,
      anon_sym_GT_GT,
    ACTIONS(2781), 1,
      anon_sym_is,
    STATE(2116), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2755), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2767), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2775), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2759), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3312), 3,
      anon_sym_if,
      anon_sym_for,
      anon_sym_RBRACE,
    ACTIONS(2753), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2779), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134710] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3250), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134782] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3314), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134854] = 20,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2803), 1,
      anon_sym_if,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    ACTIONS(3326), 1,
      anon_sym_COLON,
    ACTIONS(3328), 1,
      anon_sym_RBRACK,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [134928] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    ACTIONS(3180), 1,
      anon_sym_if,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(1303), 2,
      sym__newline,
      anon_sym_SEMI,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135000] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2215), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2219), 1,
      anon_sym_not,
    ACTIONS(2221), 1,
      anon_sym_and,
    ACTIONS(2223), 1,
      anon_sym_or,
    ACTIONS(2227), 1,
      anon_sym_PIPE,
    ACTIONS(2229), 1,
      anon_sym_CARET,
    ACTIONS(2231), 1,
      anon_sym_AMP,
    ACTIONS(2235), 1,
      anon_sym_GT_GT,
    ACTIONS(2239), 1,
      anon_sym_is,
    ACTIONS(2517), 1,
      anon_sym_if,
    STATE(1854), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2213), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2225), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2233), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3294), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2217), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2211), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2237), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135072] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2288), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2292), 1,
      anon_sym_not,
    ACTIONS(2294), 1,
      anon_sym_and,
    ACTIONS(2296), 1,
      anon_sym_or,
    ACTIONS(2300), 1,
      anon_sym_PIPE,
    ACTIONS(2302), 1,
      anon_sym_CARET,
    ACTIONS(2304), 1,
      anon_sym_AMP,
    ACTIONS(2308), 1,
      anon_sym_GT_GT,
    ACTIONS(2312), 1,
      anon_sym_is,
    ACTIONS(2376), 1,
      anon_sym_if,
    STATE(1832), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2286), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2298), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2306), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3330), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2290), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2284), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2310), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135144] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3332), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135215] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3334), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135286] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3336), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135357] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3338), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135428] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3340), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135499] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3342), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135570] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3344), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135641] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3346), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135712] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3348), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135783] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3350), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135854] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3352), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135925] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3354), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [135996] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3356), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136067] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3358), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136138] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3360), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136209] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3362), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136280] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3364), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136351] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3366), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136422] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3368), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136493] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3370), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136564] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3372), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136635] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3374), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136706] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3376), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136777] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3378), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136848] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3380), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136919] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3382), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [136990] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3384), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137061] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3386), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137132] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3388), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137203] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3390), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137274] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3392), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137345] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3394), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137416] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3396), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137487] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3398), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137558] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3400), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137629] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3402), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137700] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3404), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137771] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3406), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137842] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3408), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137913] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3410), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [137984] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3412), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138055] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3414), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138126] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3416), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138197] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3418), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138268] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3420), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138339] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3422), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138410] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3424), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138481] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3426), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138552] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3428), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138623] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3430), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138694] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3432), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138765] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3434), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138836] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3436), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138907] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3438), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [138978] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3440), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139049] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3442), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139120] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3444), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139191] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3446), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139262] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3448), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139333] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3450), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139404] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3452), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139475] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3454), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139546] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3456), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139617] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3458), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139688] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3460), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139759] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3462), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139830] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3464), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139901] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3466), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [139972] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3468), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140043] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3470), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140114] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3472), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140185] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3474), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140256] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3476), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140327] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3478), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140398] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2803), 1,
      anon_sym_if,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    ACTIONS(3480), 1,
      anon_sym_RBRACK,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140469] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3482), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140540] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3484), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140611] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3486), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140682] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3488), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140753] = 19,
    ACTIONS(1299), 1,
      anon_sym_RPAREN,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3140), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3144), 1,
      anon_sym_not,
    ACTIONS(3146), 1,
      anon_sym_and,
    ACTIONS(3148), 1,
      anon_sym_or,
    ACTIONS(3152), 1,
      anon_sym_PIPE,
    ACTIONS(3154), 1,
      anon_sym_CARET,
    ACTIONS(3156), 1,
      anon_sym_AMP,
    ACTIONS(3160), 1,
      anon_sym_GT_GT,
    ACTIONS(3164), 1,
      anon_sym_is,
    ACTIONS(3178), 1,
      anon_sym_if,
    STATE(2264), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3138), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3150), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3158), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3142), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3136), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3162), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140824] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3490), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140895] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3492), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [140966] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3494), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141037] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2803), 1,
      anon_sym_if,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    ACTIONS(3496), 1,
      anon_sym_RBRACK,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141108] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3498), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141179] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3500), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141250] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3502), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141321] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3504), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141392] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3506), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141463] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3508), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141534] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3510), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141605] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3084), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3088), 1,
      anon_sym_not,
    ACTIONS(3092), 1,
      anon_sym_PIPE,
    ACTIONS(3094), 1,
      anon_sym_CARET,
    ACTIONS(3096), 1,
      anon_sym_AMP,
    ACTIONS(3100), 1,
      anon_sym_GT_GT,
    ACTIONS(3104), 1,
      anon_sym_is,
    ACTIONS(3115), 1,
      anon_sym_and,
    ACTIONS(3117), 1,
      anon_sym_or,
    ACTIONS(3180), 1,
      anon_sym_if,
    ACTIONS(3512), 1,
      sym__newline,
    STATE(2173), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3090), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3098), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3086), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3080), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3102), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141676] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3514), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141747] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3516), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141818] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3518), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141889] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3520), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [141960] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3522), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142031] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3524), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142102] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3526), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142173] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3528), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142244] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3530), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142315] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3532), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142386] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3534), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142457] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2785), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2803), 1,
      anon_sym_if,
    ACTIONS(2807), 1,
      anon_sym_not,
    ACTIONS(2809), 1,
      anon_sym_and,
    ACTIONS(2811), 1,
      anon_sym_or,
    ACTIONS(2813), 1,
      anon_sym_PIPE,
    ACTIONS(2815), 1,
      anon_sym_CARET,
    ACTIONS(2817), 1,
      anon_sym_AMP,
    ACTIONS(2821), 1,
      anon_sym_GT_GT,
    ACTIONS(2825), 1,
      anon_sym_is,
    ACTIONS(3324), 1,
      anon_sym_RBRACK,
    STATE(2105), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2783), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2789), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2819), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2787), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2805), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2823), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142528] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3536), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142599] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3538), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142670] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3540), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142741] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3542), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142812] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3544), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142883] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3546), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [142954] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3548), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143025] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3550), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143096] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3552), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143167] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3554), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143238] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3556), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143309] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3558), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143380] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3560), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143451] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3562), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143522] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3564), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143593] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3566), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143664] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3041), 1,
      anon_sym_if,
    ACTIONS(3047), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3051), 1,
      anon_sym_not,
    ACTIONS(3053), 1,
      anon_sym_and,
    ACTIONS(3055), 1,
      anon_sym_or,
    ACTIONS(3059), 1,
      anon_sym_PIPE,
    ACTIONS(3061), 1,
      anon_sym_CARET,
    ACTIONS(3063), 1,
      anon_sym_AMP,
    ACTIONS(3067), 1,
      anon_sym_GT_GT,
    ACTIONS(3071), 1,
      anon_sym_is,
    ACTIONS(3568), 1,
      anon_sym_else,
    STATE(2252), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(3045), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3057), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3065), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(3049), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3043), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(3069), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143735] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3570), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143806] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3572), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143877] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3574), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [143948] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3576), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [144019] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3578), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [144090] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3580), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [144161] = 19,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2554), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2558), 1,
      anon_sym_not,
    ACTIONS(2560), 1,
      anon_sym_and,
    ACTIONS(2562), 1,
      anon_sym_or,
    ACTIONS(2566), 1,
      anon_sym_PIPE,
    ACTIONS(2568), 1,
      anon_sym_CARET,
    ACTIONS(2570), 1,
      anon_sym_AMP,
    ACTIONS(2574), 1,
      anon_sym_GT_GT,
    ACTIONS(2578), 1,
      anon_sym_is,
    ACTIONS(2689), 1,
      anon_sym_if,
    ACTIONS(3582), 1,
      anon_sym_COLON,
    STATE(1920), 1,
      aux_sym_comparison_operator_repeat1,
    ACTIONS(2552), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2572), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT_GT,
    ACTIONS(2556), 3,
      anon_sym_AT,
      anon_sym_SLASH_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2550), 5,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT,
    ACTIONS(2576), 5,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [144232] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(931), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(1880), 17,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_in,
  [144261] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(925), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(1873), 17,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_in,
  [144290] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(937), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
    ACTIONS(1885), 17,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_in,
  [144319] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(1873), 18,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_in,
  [144343] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3584), 18,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_in,
  [144367] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3586), 18,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_in,
  [144391] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3588), 18,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_in,
  [144415] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3590), 1,
      anon_sym_COMMA,
    STATE(2520), 1,
      aux_sym_pattern_list_repeat1,
    ACTIONS(3584), 15,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RBRACK,
  [144442] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3593), 1,
      anon_sym_COMMA,
    STATE(2521), 1,
      aux_sym_pattern_list_repeat1,
    ACTIONS(3584), 15,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_RPAREN,
  [144469] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3598), 1,
      anon_sym_COMMA,
    ACTIONS(3600), 1,
      anon_sym_RBRACK,
    STATE(2525), 1,
      aux_sym_pattern_list_repeat1,
    ACTIONS(3596), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144498] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3604), 1,
      anon_sym_COMMA,
    ACTIONS(3606), 1,
      anon_sym_RPAREN,
    STATE(2521), 1,
      aux_sym_pattern_list_repeat1,
    ACTIONS(3602), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144527] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3600), 1,
      anon_sym_RPAREN,
    ACTIONS(3608), 1,
      anon_sym_COMMA,
    STATE(2523), 1,
      aux_sym_pattern_list_repeat1,
    ACTIONS(3596), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144556] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3606), 1,
      anon_sym_RBRACK,
    ACTIONS(3610), 1,
      anon_sym_COMMA,
    STATE(2520), 1,
      aux_sym_pattern_list_repeat1,
    ACTIONS(3602), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144585] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3612), 1,
      anon_sym_COMMA,
    STATE(2527), 1,
      aux_sym_pattern_list_repeat1,
    ACTIONS(3596), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144611] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3614), 1,
      anon_sym_COMMA,
    STATE(2520), 1,
      aux_sym_pattern_list_repeat1,
    ACTIONS(3602), 14,
      anon_sym_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144637] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3616), 1,
      anon_sym_EQ,
    ACTIONS(3618), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144659] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3620), 1,
      anon_sym_EQ,
    ACTIONS(3622), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144681] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3624), 1,
      anon_sym_EQ,
    ACTIONS(3626), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144703] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3628), 1,
      anon_sym_EQ,
    ACTIONS(3630), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144725] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3632), 1,
      anon_sym_EQ,
    ACTIONS(3634), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144747] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3636), 1,
      anon_sym_EQ,
    ACTIONS(3638), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144769] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3640), 1,
      anon_sym_EQ,
    ACTIONS(3642), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144791] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3644), 1,
      anon_sym_EQ,
    ACTIONS(3646), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144813] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3648), 1,
      anon_sym_EQ,
    ACTIONS(3650), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144835] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3652), 1,
      anon_sym_EQ,
    ACTIONS(3626), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144857] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3654), 1,
      anon_sym_EQ,
    ACTIONS(3656), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144879] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3658), 1,
      anon_sym_EQ,
    ACTIONS(3660), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144901] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3662), 1,
      anon_sym_EQ,
    ACTIONS(3664), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144923] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3666), 1,
      anon_sym_EQ,
    ACTIONS(3668), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144945] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3670), 1,
      anon_sym_EQ,
    ACTIONS(3672), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144967] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3674), 1,
      anon_sym_EQ,
    ACTIONS(3676), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [144989] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3678), 1,
      anon_sym_EQ,
    ACTIONS(3680), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145011] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3682), 1,
      anon_sym_EQ,
    ACTIONS(3684), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145033] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3686), 1,
      anon_sym_EQ,
    ACTIONS(3688), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145055] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3690), 1,
      anon_sym_EQ,
    ACTIONS(3692), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145077] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3694), 1,
      anon_sym_EQ,
    ACTIONS(3696), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145099] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3698), 1,
      anon_sym_EQ,
    ACTIONS(3626), 13,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_SLASH_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [145121] = 7,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3700), 1,
      sym_identifier,
    ACTIONS(3702), 1,
      anon_sym_RPAREN,
    ACTIONS(3704), 1,
      anon_sym_STAR,
    ACTIONS(3706), 1,
      anon_sym_STAR_STAR,
    STATE(2974), 1,
      sym__parameters,
    STATE(2660), 6,
      sym__parameter,
      sym_typed_parameter,
      sym_default_parameter,
      sym_typed_default_parameter,
      sym_list_splat_parameter,
      sym_dictionary_splat_parameter,
  [145148] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3700), 1,
      sym_identifier,
    ACTIONS(3704), 1,
      anon_sym_STAR,
    ACTIONS(3706), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3708), 1,
      anon_sym_RPAREN,
    STATE(2905), 6,
      sym__parameter,
      sym_typed_parameter,
      sym_default_parameter,
      sym_typed_default_parameter,
      sym_list_splat_parameter,
      sym_dictionary_splat_parameter,
  [145172] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3700), 1,
      sym_identifier,
    ACTIONS(3704), 1,
      anon_sym_STAR,
    ACTIONS(3706), 1,
      anon_sym_STAR_STAR,
    ACTIONS(3710), 1,
      anon_sym_RPAREN,
    STATE(2905), 6,
      sym__parameter,
      sym_typed_parameter,
      sym_default_parameter,
      sym_typed_default_parameter,
      sym_list_splat_parameter,
      sym_dictionary_splat_parameter,
  [145196] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3712), 1,
      anon_sym_async,
    ACTIONS(3714), 1,
      anon_sym_def,
    ACTIONS(3716), 1,
      anon_sym_class,
    ACTIONS(3718), 1,
      anon_sym_AT,
    STATE(2571), 2,
      sym_decorator,
      aux_sym_decorated_definition_repeat1,
    STATE(1602), 3,
      sym_function_definition,
      sym_async_function_definition,
      sym_class_definition,
  [145221] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3700), 1,
      sym_identifier,
    ACTIONS(3704), 1,
      anon_sym_STAR,
    ACTIONS(3706), 1,
      anon_sym_STAR_STAR,
    STATE(2905), 6,
      sym__parameter,
      sym_typed_parameter,
      sym_default_parameter,
      sym_typed_default_parameter,
      sym_list_splat_parameter,
      sym_dictionary_splat_parameter,
  [145242] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3718), 1,
      anon_sym_AT,
    ACTIONS(3720), 1,
      anon_sym_async,
    ACTIONS(3722), 1,
      anon_sym_def,
    ACTIONS(3724), 1,
      anon_sym_class,
    STATE(2571), 2,
      sym_decorator,
      aux_sym_decorated_definition_repeat1,
    STATE(1618), 3,
      sym_function_definition,
      sym_async_function_definition,
      sym_class_definition,
  [145267] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3726), 1,
      anon_sym_COMMA,
    STATE(2557), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(503), 6,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [145285] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3728), 1,
      anon_sym_COMMA,
    STATE(2557), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1314), 6,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [145303] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3731), 1,
      anon_sym_SEMI,
    STATE(2569), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(109), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [145319] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3733), 1,
      anon_sym_COMMA,
    STATE(1716), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(503), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_for,
  [145335] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3735), 1,
      anon_sym_RPAREN,
    ACTIONS(3737), 1,
      anon_sym_if,
    STATE(2579), 3,
      sym_for_in_clause,
      sym_if_clause,
      aux_sym__comprehension_clauses_repeat1,
  [145353] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3190), 1,
      anon_sym_for,
    ACTIONS(3739), 1,
      anon_sym_COMMA,
    ACTIONS(3741), 1,
      anon_sym_RBRACE,
    STATE(2578), 1,
      sym_for_in_clause,
    STATE(2683), 1,
      aux_sym_dictionary_repeat1,
    STATE(2992), 1,
      sym__comprehension_clauses,
  [145375] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3743), 1,
      anon_sym_SEMI,
    STATE(2558), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1361), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [145391] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(945), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_in,
      anon_sym_QMARK,
  [145403] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3747), 1,
      anon_sym_DOT,
    STATE(2564), 1,
      aux_sym_dotted_name_repeat1,
    ACTIONS(3745), 4,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_as,
  [145419] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3752), 1,
      anon_sym_DOT,
    STATE(2580), 1,
      aux_sym_dotted_name_repeat1,
    ACTIONS(3750), 4,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_as,
  [145435] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3737), 1,
      anon_sym_if,
    ACTIONS(3754), 1,
      anon_sym_RPAREN,
    STATE(2560), 3,
      sym_for_in_clause,
      sym_if_clause,
      aux_sym__comprehension_clauses_repeat1,
  [145453] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3190), 1,
      anon_sym_for,
    ACTIONS(3756), 1,
      anon_sym_COMMA,
    ACTIONS(3758), 1,
      anon_sym_RBRACE,
    STATE(2578), 1,
      sym_for_in_clause,
    STATE(2686), 1,
      aux_sym_dictionary_repeat1,
    STATE(3049), 1,
      sym__comprehension_clauses,
  [145475] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3190), 1,
      anon_sym_for,
    ACTIONS(3760), 1,
      anon_sym_COMMA,
    ACTIONS(3762), 1,
      anon_sym_RBRACE,
    STATE(2578), 1,
      sym_for_in_clause,
    STATE(2765), 1,
      aux_sym_dictionary_repeat1,
    STATE(2970), 1,
      sym__comprehension_clauses,
  [145497] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3764), 1,
      anon_sym_SEMI,
    STATE(2569), 1,
      aux_sym__inline_body_repeat1,
    ACTIONS(1369), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [145513] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3754), 1,
      anon_sym_RBRACK,
    ACTIONS(3767), 1,
      anon_sym_if,
    STATE(2577), 3,
      sym_for_in_clause,
      sym_if_clause,
      aux_sym__comprehension_clauses_repeat1,
  [145531] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3771), 1,
      anon_sym_AT,
    STATE(2571), 2,
      sym_decorator,
      aux_sym_decorated_definition_repeat1,
    ACTIONS(3769), 3,
      anon_sym_async,
      anon_sym_def,
      anon_sym_class,
  [145547] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3190), 1,
      anon_sym_for,
    ACTIONS(3735), 1,
      anon_sym_RBRACE,
    ACTIONS(3774), 1,
      anon_sym_if,
    STATE(2573), 3,
      sym_for_in_clause,
      sym_if_clause,
      aux_sym__comprehension_clauses_repeat1,
  [145565] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3776), 1,
      anon_sym_if,
    ACTIONS(3779), 1,
      anon_sym_for,
    ACTIONS(3782), 1,
      anon_sym_RBRACE,
    STATE(2573), 3,
      sym_for_in_clause,
      sym_if_clause,
      aux_sym__comprehension_clauses_repeat1,
  [145583] = 7,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3190), 1,
      anon_sym_for,
    ACTIONS(3784), 1,
      anon_sym_COMMA,
    ACTIONS(3786), 1,
      anon_sym_RBRACE,
    STATE(2578), 1,
      sym_for_in_clause,
    STATE(2706), 1,
      aux_sym_dictionary_repeat1,
    STATE(3050), 1,
      sym__comprehension_clauses,
  [145605] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(949), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_in,
      anon_sym_QMARK,
  [145617] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(949), 6,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_in,
      anon_sym_QMARK,
  [145629] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3199), 1,
      anon_sym_for,
    ACTIONS(3735), 1,
      anon_sym_RBRACK,
    ACTIONS(3767), 1,
      anon_sym_if,
    STATE(2581), 3,
      sym_for_in_clause,
      sym_if_clause,
      aux_sym__comprehension_clauses_repeat1,
  [145647] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3190), 1,
      anon_sym_for,
    ACTIONS(3754), 1,
      anon_sym_RBRACE,
    ACTIONS(3774), 1,
      anon_sym_if,
    STATE(2572), 3,
      sym_for_in_clause,
      sym_if_clause,
      aux_sym__comprehension_clauses_repeat1,
  [145665] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3782), 1,
      anon_sym_RPAREN,
    ACTIONS(3788), 1,
      anon_sym_if,
    ACTIONS(3791), 1,
      anon_sym_for,
    STATE(2579), 3,
      sym_for_in_clause,
      sym_if_clause,
      aux_sym__comprehension_clauses_repeat1,
  [145683] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3752), 1,
      anon_sym_DOT,
    STATE(2564), 1,
      aux_sym_dotted_name_repeat1,
    ACTIONS(3794), 4,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_as,
  [145699] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3782), 1,
      anon_sym_RBRACK,
    ACTIONS(3791), 1,
      anon_sym_for,
    ACTIONS(3796), 1,
      anon_sym_if,
    STATE(2581), 3,
      sym_for_in_clause,
      sym_if_clause,
      aux_sym__comprehension_clauses_repeat1,
  [145717] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3799), 1,
      anon_sym_LBRACE,
    ACTIONS(3801), 1,
      sym_fstring_content,
    ACTIONS(3803), 1,
      sym_fstring_end,
    STATE(2599), 2,
      sym_interpolation,
      aux_sym_f_string_repeat1,
  [145734] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3805), 1,
      anon_sym_COMMA,
    STATE(1859), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(503), 3,
      anon_sym_if,
      anon_sym_for,
      anon_sym_RBRACE,
  [145749] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3809), 1,
      anon_sym_COMMA,
    ACTIONS(3811), 1,
      anon_sym_as,
    STATE(2623), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(3807), 2,
      sym__newline,
      anon_sym_SEMI,
  [145766] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3799), 1,
      anon_sym_LBRACE,
    ACTIONS(3813), 1,
      sym_fstring_content,
    ACTIONS(3815), 1,
      sym_fstring_end,
    STATE(2592), 2,
      sym_interpolation,
      aux_sym_f_string_repeat1,
  [145783] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3817), 1,
      anon_sym_except,
    ACTIONS(3819), 1,
      anon_sym_finally,
    STATE(1634), 1,
      sym_finally_clause,
    STATE(1501), 2,
      sym_except_clause,
      aux_sym_try_statement_repeat1,
  [145800] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3233), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_for,
      anon_sym_RBRACE,
  [145811] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3821), 1,
      anon_sym_COLON,
    ACTIONS(925), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [145824] = 6,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(2116), 1,
      anon_sym_LBRACK,
    ACTIONS(2118), 1,
      anon_sym_DOT,
    ACTIONS(3823), 1,
      anon_sym_QMARK,
    STATE(1680), 1,
      sym_argument_list,
  [145843] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3825), 1,
      anon_sym_except,
    ACTIONS(3827), 1,
      anon_sym_finally,
    STATE(1628), 1,
      sym_finally_clause,
    STATE(1504), 2,
      sym_except_clause,
      aux_sym_try_statement_repeat1,
  [145860] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3745), 5,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_as,
      anon_sym_DOT,
  [145871] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3799), 1,
      anon_sym_LBRACE,
    ACTIONS(3829), 1,
      sym_fstring_content,
    ACTIONS(3831), 1,
      sym_fstring_end,
    STATE(2595), 2,
      sym_interpolation,
      aux_sym_f_string_repeat1,
  [145888] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3833), 1,
      anon_sym_COLON,
    ACTIONS(925), 4,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      anon_sym_QMARK,
  [145901] = 6,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(3823), 1,
      anon_sym_QMARK,
    ACTIONS(3835), 1,
      anon_sym_LBRACK,
    ACTIONS(3837), 1,
      anon_sym_DOT,
    STATE(1680), 1,
      sym_argument_list,
  [145920] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3839), 1,
      anon_sym_LBRACE,
    ACTIONS(3842), 1,
      sym_fstring_content,
    ACTIONS(3845), 1,
      sym_fstring_end,
    STATE(2595), 2,
      sym_interpolation,
      aux_sym_f_string_repeat1,
  [145937] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3799), 1,
      anon_sym_LBRACE,
    ACTIONS(3847), 1,
      sym_fstring_content,
    ACTIONS(3849), 1,
      sym_fstring_end,
    STATE(2597), 2,
      sym_interpolation,
      aux_sym_f_string_repeat1,
  [145954] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3799), 1,
      anon_sym_LBRACE,
    ACTIONS(3829), 1,
      sym_fstring_content,
    ACTIONS(3851), 1,
      sym_fstring_end,
    STATE(2595), 2,
      sym_interpolation,
      aux_sym_f_string_repeat1,
  [145971] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3799), 1,
      anon_sym_LBRACE,
    ACTIONS(3829), 1,
      sym_fstring_content,
    ACTIONS(3853), 1,
      sym_fstring_end,
    STATE(2595), 2,
      sym_interpolation,
      aux_sym_f_string_repeat1,
  [145988] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3799), 1,
      anon_sym_LBRACE,
    ACTIONS(3829), 1,
      sym_fstring_content,
    ACTIONS(3855), 1,
      sym_fstring_end,
    STATE(2595), 2,
      sym_interpolation,
      aux_sym_f_string_repeat1,
  [146005] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3799), 1,
      anon_sym_LBRACE,
    ACTIONS(3857), 1,
      sym_fstring_content,
    ACTIONS(3859), 1,
      sym_fstring_end,
    STATE(2598), 2,
      sym_interpolation,
      aux_sym_f_string_repeat1,
  [146022] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3863), 1,
      anon_sym_COMMA,
    ACTIONS(3865), 1,
      anon_sym_as,
    STATE(2624), 1,
      aux_sym__import_list_repeat1,
    ACTIONS(3861), 2,
      sym__newline,
      anon_sym_SEMI,
  [146039] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3869), 1,
      anon_sym_COMMA,
    STATE(2626), 1,
      aux_sym_global_statement_repeat1,
    ACTIONS(3867), 2,
      sym__newline,
      anon_sym_SEMI,
  [146053] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3809), 1,
      anon_sym_COMMA,
    STATE(2604), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(3807), 2,
      sym__newline,
      anon_sym_SEMI,
  [146067] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3809), 1,
      anon_sym_COMMA,
    STATE(2607), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(3871), 2,
      sym__newline,
      anon_sym_SEMI,
  [146081] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3863), 1,
      anon_sym_COMMA,
    STATE(2627), 1,
      aux_sym__import_list_repeat1,
    ACTIONS(3861), 2,
      sym__newline,
      anon_sym_SEMI,
  [146095] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3811), 1,
      anon_sym_as,
    ACTIONS(3873), 3,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [146107] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3877), 1,
      anon_sym_COMMA,
    STATE(2607), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(3875), 2,
      sym__newline,
      anon_sym_SEMI,
  [146121] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3274), 1,
      anon_sym_COMMA,
    ACTIONS(3880), 1,
      anon_sym_COLON,
    ACTIONS(3882), 1,
      anon_sym_as,
    STATE(2618), 1,
      aux_sym_expression_list_repeat1,
  [146137] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3884), 1,
      anon_sym_EQ,
    ACTIONS(3886), 1,
      anon_sym_COLON,
    ACTIONS(3888), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [146151] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3890), 1,
      sym_identifier,
    ACTIONS(3892), 1,
      anon_sym_SEMI,
    ACTIONS(3894), 1,
      sym__newline,
    STATE(2722), 1,
      sym_aliased_import_name,
  [146167] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3896), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
  [146177] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3898), 1,
      anon_sym_COMMA,
    STATE(2612), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1314), 2,
      sym__newline,
      anon_sym_SEMI,
  [146191] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3903), 1,
      anon_sym_COMMA,
    STATE(2613), 1,
      aux_sym__flag_list_repeat1,
    ACTIONS(3901), 2,
      sym__newline,
      anon_sym_SEMI,
  [146205] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3865), 1,
      anon_sym_as,
    ACTIONS(3906), 3,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [146217] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3908), 1,
      anon_sym_COMMA,
    STATE(2615), 1,
      aux_sym_list_repeat1,
    ACTIONS(3896), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [146231] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3890), 1,
      sym_identifier,
    ACTIONS(3911), 1,
      anon_sym_SEMI,
    ACTIONS(3913), 1,
      sym__newline,
    STATE(2722), 1,
      sym_aliased_import_name,
  [146247] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3915), 4,
      anon_sym_async,
      anon_sym_def,
      anon_sym_class,
      anon_sym_AT,
  [146257] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3917), 1,
      anon_sym_COMMA,
    STATE(2618), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1314), 2,
      anon_sym_COLON,
      anon_sym_as,
  [146271] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3922), 1,
      anon_sym_COMMA,
    STATE(2619), 1,
      aux_sym__import_list_repeat1,
    ACTIONS(3920), 2,
      sym__newline,
      anon_sym_SEMI,
  [146285] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3890), 1,
      sym_identifier,
    ACTIONS(3911), 1,
      anon_sym_SEMI,
    ACTIONS(3913), 1,
      sym__newline,
    STATE(2722), 1,
      sym_aliased_import_name,
  [146301] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3925), 1,
      anon_sym_COMMA,
    STATE(2622), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(503), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [146315] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3927), 1,
      anon_sym_COMMA,
    STATE(2622), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(1314), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [146329] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3809), 1,
      anon_sym_COMMA,
    STATE(2607), 1,
      aux_sym_import_statement_repeat1,
    ACTIONS(3871), 2,
      sym__newline,
      anon_sym_SEMI,
  [146343] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3932), 1,
      anon_sym_COMMA,
    STATE(2619), 1,
      aux_sym__import_list_repeat1,
    ACTIONS(3930), 2,
      sym__newline,
      anon_sym_SEMI,
  [146357] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3934), 1,
      sym_identifier,
    ACTIONS(3936), 1,
      anon_sym_LPAREN,
    STATE(2605), 1,
      sym_aliased_import_name,
    STATE(2898), 1,
      sym__import_list,
  [146373] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3869), 1,
      anon_sym_COMMA,
    STATE(2633), 1,
      aux_sym_global_statement_repeat1,
    ACTIONS(3938), 2,
      sym__newline,
      anon_sym_SEMI,
  [146387] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3940), 1,
      anon_sym_COMMA,
    STATE(2619), 1,
      aux_sym__import_list_repeat1,
    ACTIONS(3930), 2,
      sym__newline,
      anon_sym_SEMI,
  [146401] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3861), 1,
      anon_sym_RPAREN,
    ACTIONS(3942), 1,
      anon_sym_COMMA,
    ACTIONS(3944), 1,
      anon_sym_as,
    STATE(2680), 1,
      aux_sym__import_list_repeat1,
  [146417] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3869), 1,
      anon_sym_COMMA,
    STATE(2631), 1,
      aux_sym_global_statement_repeat1,
    ACTIONS(3946), 2,
      sym__newline,
      anon_sym_SEMI,
  [146431] = 5,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3274), 1,
      anon_sym_COMMA,
    ACTIONS(3948), 1,
      anon_sym_COLON,
    ACTIONS(3950), 1,
      anon_sym_as,
    STATE(2618), 1,
      aux_sym_expression_list_repeat1,
  [146447] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3869), 1,
      anon_sym_COMMA,
    STATE(2633), 1,
      aux_sym_global_statement_repeat1,
    ACTIONS(3952), 2,
      sym__newline,
      anon_sym_SEMI,
  [146461] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3954), 1,
      anon_sym_COMMA,
    STATE(2612), 1,
      aux_sym_expression_list_repeat1,
    ACTIONS(503), 2,
      sym__newline,
      anon_sym_SEMI,
  [146475] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3958), 1,
      anon_sym_COMMA,
    STATE(2633), 1,
      aux_sym_global_statement_repeat1,
    ACTIONS(3956), 2,
      sym__newline,
      anon_sym_SEMI,
  [146489] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3963), 1,
      anon_sym_COMMA,
    STATE(2635), 1,
      aux_sym__flag_list_repeat1,
    ACTIONS(3961), 2,
      sym__newline,
      anon_sym_SEMI,
  [146503] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3967), 1,
      anon_sym_COMMA,
    STATE(2613), 1,
      aux_sym__flag_list_repeat1,
    ACTIONS(3965), 2,
      sym__newline,
      anon_sym_SEMI,
  [146517] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(3969), 1,
      anon_sym_COLON,
    STATE(3129), 1,
      sym_argument_list,
  [146530] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3971), 1,
      sym_identifier,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    STATE(2855), 1,
      sym_parameters,
  [146543] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3584), 1,
      anon_sym_in,
    ACTIONS(3975), 1,
      anon_sym_COMMA,
    STATE(2638), 1,
      aux_sym_pattern_list_repeat1,
  [146556] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3739), 1,
      anon_sym_COMMA,
    ACTIONS(3741), 1,
      anon_sym_RBRACE,
    STATE(2683), 1,
      aux_sym_dictionary_repeat1,
  [146569] = 4,
    ACTIONS(982), 1,
      sym__newline,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3978), 1,
      anon_sym_SEMI,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [146582] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3980), 1,
      anon_sym_COMMA,
    ACTIONS(3982), 1,
      anon_sym_RBRACE,
    STATE(2687), 1,
      aux_sym_list_repeat1,
  [146595] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3984), 1,
      anon_sym_SEMI,
    ACTIONS(3987), 1,
      sym__newline,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [146608] = 4,
    ACTIONS(978), 1,
      sym__newline,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3989), 1,
      anon_sym_SEMI,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [146621] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3602), 1,
      anon_sym_in,
    ACTIONS(3991), 1,
      anon_sym_COMMA,
    STATE(2638), 1,
      aux_sym_pattern_list_repeat1,
  [146634] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3961), 1,
      anon_sym_RPAREN,
    ACTIONS(3993), 1,
      anon_sym_COMMA,
    STATE(2678), 1,
      aux_sym__flag_list_repeat1,
  [146647] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3794), 1,
      anon_sym_import,
    ACTIONS(3995), 1,
      anon_sym_DOT,
    STATE(2650), 1,
      aux_sym_dotted_name_repeat1,
  [146660] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3861), 1,
      anon_sym_RPAREN,
    ACTIONS(3942), 1,
      anon_sym_COMMA,
    STATE(2681), 1,
      aux_sym__import_list_repeat1,
  [146673] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3956), 3,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [146682] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3750), 1,
      anon_sym_import,
    ACTIONS(3995), 1,
      anon_sym_DOT,
    STATE(2646), 1,
      aux_sym_dotted_name_repeat1,
  [146695] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3745), 1,
      anon_sym_import,
    ACTIONS(3997), 1,
      anon_sym_DOT,
    STATE(2650), 1,
      aux_sym_dotted_name_repeat1,
  [146708] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4000), 1,
      sym_identifier,
    STATE(2647), 1,
      sym_aliased_import_name,
    STATE(3015), 1,
      sym__import_list,
  [146721] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4002), 1,
      anon_sym_COMMA,
    ACTIONS(4004), 1,
      anon_sym_RPAREN,
    STATE(2615), 1,
      aux_sym_list_repeat1,
  [146734] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4006), 1,
      anon_sym_COMMA,
    ACTIONS(4008), 1,
      anon_sym_RBRACK,
    STATE(2615), 1,
      aux_sym_list_repeat1,
  [146747] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4010), 1,
      anon_sym_COLON,
    ACTIONS(4012), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [146758] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2048), 3,
      anon_sym_DOT,
      anon_sym_except,
      anon_sym_finally,
  [146767] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3873), 3,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [146776] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4014), 1,
      anon_sym_COMMA,
    ACTIONS(4016), 1,
      anon_sym_RPAREN,
    STATE(2705), 1,
      aux_sym__parameters_repeat1,
  [146789] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4018), 3,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [146798] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2016), 3,
      anon_sym_DOT,
      anon_sym_except,
      anon_sym_finally,
  [146807] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4020), 1,
      anon_sym_COMMA,
    ACTIONS(4022), 1,
      anon_sym_RPAREN,
    STATE(2657), 1,
      aux_sym__parameters_repeat1,
  [146820] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3756), 1,
      anon_sym_COMMA,
    ACTIONS(3758), 1,
      anon_sym_RBRACE,
    STATE(2686), 1,
      aux_sym_dictionary_repeat1,
  [146833] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4024), 1,
      anon_sym_COMMA,
    ACTIONS(4026), 1,
      anon_sym_RBRACE,
    STATE(2693), 1,
      aux_sym_list_repeat1,
  [146846] = 4,
    ACTIONS(988), 1,
      sym__newline,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4028), 1,
      anon_sym_SEMI,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [146859] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3760), 1,
      anon_sym_COMMA,
    ACTIONS(3762), 1,
      anon_sym_RBRACE,
    STATE(2765), 1,
      aux_sym_dictionary_repeat1,
  [146872] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4030), 1,
      anon_sym_COMMA,
    ACTIONS(4032), 1,
      anon_sym_RBRACE,
    STATE(2738), 1,
      aux_sym_list_repeat1,
  [146885] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4034), 1,
      anon_sym_SEMI,
    ACTIONS(4036), 1,
      sym__newline,
    STATE(2668), 1,
      aux_sym__simple_statements_repeat1,
  [146898] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4038), 1,
      sym_identifier,
    STATE(2900), 1,
      sym_parameters,
  [146911] = 4,
    ACTIONS(974), 1,
      sym__newline,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4040), 1,
      anon_sym_SEMI,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [146924] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4042), 1,
      anon_sym_COMMA,
    ACTIONS(4044), 1,
      anon_sym_RBRACK,
    STATE(2742), 1,
      aux_sym_list_repeat1,
  [146937] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2046), 3,
      anon_sym_DOT,
      anon_sym_except,
      anon_sym_finally,
  [146946] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4046), 1,
      sym_identifier,
    STATE(2870), 1,
      sym_parameters,
  [146959] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4048), 1,
      anon_sym_COMMA,
    ACTIONS(4050), 1,
      anon_sym_RPAREN,
    STATE(2615), 1,
      aux_sym_list_repeat1,
  [146972] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4052), 1,
      anon_sym_COLON,
    ACTIONS(4054), 1,
      anon_sym_COMMA,
    STATE(2673), 1,
      aux_sym_with_statement_repeat1,
  [146985] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4057), 1,
      anon_sym_COMMA,
    ACTIONS(4059), 1,
      anon_sym_RPAREN,
    STATE(2615), 1,
      aux_sym_list_repeat1,
  [146998] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4061), 1,
      anon_sym_COLON,
    ACTIONS(4063), 1,
      anon_sym_COMMA,
    STATE(2685), 1,
      aux_sym_with_statement_repeat1,
  [147011] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2114), 1,
      anon_sym_LPAREN,
    ACTIONS(4065), 1,
      anon_sym_COLON,
    STATE(3088), 1,
      sym_argument_list,
  [147024] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4067), 1,
      anon_sym_COMMA,
    ACTIONS(4069), 1,
      anon_sym_RBRACK,
    STATE(2615), 1,
      aux_sym_list_repeat1,
  [147037] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3965), 1,
      anon_sym_RPAREN,
    ACTIONS(4071), 1,
      anon_sym_COMMA,
    STATE(2727), 1,
      aux_sym__flag_list_repeat1,
  [147050] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3892), 1,
      anon_sym_RPAREN,
    ACTIONS(4073), 1,
      sym_identifier,
    STATE(2888), 1,
      sym_aliased_import_name,
  [147063] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3930), 1,
      anon_sym_RPAREN,
    ACTIONS(4075), 1,
      anon_sym_COMMA,
    STATE(2735), 1,
      aux_sym__import_list_repeat1,
  [147076] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3930), 1,
      anon_sym_RPAREN,
    ACTIONS(4077), 1,
      anon_sym_COMMA,
    STATE(2735), 1,
      aux_sym__import_list_repeat1,
  [147089] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4079), 1,
      sym_identifier,
    STATE(2822), 1,
      sym_parameters,
  [147102] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4081), 1,
      anon_sym_COMMA,
    ACTIONS(4083), 1,
      anon_sym_RBRACE,
    STATE(2691), 1,
      aux_sym_dictionary_repeat1,
  [147115] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4085), 3,
      sym_fstring_content,
      sym_fstring_end,
      anon_sym_LBRACE,
  [147124] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4063), 1,
      anon_sym_COMMA,
    ACTIONS(4087), 1,
      anon_sym_COLON,
    STATE(2673), 1,
      aux_sym_with_statement_repeat1,
  [147137] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4089), 1,
      anon_sym_COMMA,
    ACTIONS(4091), 1,
      anon_sym_RBRACE,
    STATE(2691), 1,
      aux_sym_dictionary_repeat1,
  [147150] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4093), 1,
      anon_sym_COMMA,
    ACTIONS(4095), 1,
      anon_sym_RBRACE,
    STATE(2690), 1,
      aux_sym_list_repeat1,
  [147163] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4097), 1,
      anon_sym_SEMI,
    ACTIONS(4099), 1,
      sym__newline,
    STATE(2689), 1,
      aux_sym__simple_statements_repeat1,
  [147176] = 4,
    ACTIONS(1002), 1,
      sym__newline,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4101), 1,
      anon_sym_SEMI,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [147189] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3896), 1,
      anon_sym_RBRACE,
    ACTIONS(4103), 1,
      anon_sym_COMMA,
    STATE(2690), 1,
      aux_sym_list_repeat1,
  [147202] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4106), 1,
      anon_sym_COMMA,
    ACTIONS(4109), 1,
      anon_sym_RBRACE,
    STATE(2691), 1,
      aux_sym_dictionary_repeat1,
  [147215] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4111), 1,
      anon_sym_SEMI,
    ACTIONS(4113), 1,
      sym__newline,
    STATE(2701), 1,
      aux_sym__simple_statements_repeat1,
  [147228] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4115), 1,
      anon_sym_COMMA,
    ACTIONS(4117), 1,
      anon_sym_RBRACE,
    STATE(2690), 1,
      aux_sym_list_repeat1,
  [147241] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4119), 1,
      anon_sym_COMMA,
    ACTIONS(4121), 1,
      anon_sym_RPAREN,
    STATE(2749), 1,
      aux_sym_argument_list_repeat1,
  [147254] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4123), 1,
      anon_sym_COMMA,
    ACTIONS(4125), 1,
      anon_sym_RBRACK,
    STATE(2704), 1,
      aux_sym_list_repeat1,
  [147267] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3784), 1,
      anon_sym_COMMA,
    ACTIONS(3786), 1,
      anon_sym_RBRACE,
    STATE(2706), 1,
      aux_sym_dictionary_repeat1,
  [147280] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4127), 1,
      anon_sym_SEMI,
    ACTIONS(4129), 1,
      sym__newline,
    STATE(2640), 1,
      aux_sym__simple_statements_repeat1,
  [147293] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4131), 1,
      anon_sym_COMMA,
    ACTIONS(4133), 1,
      anon_sym_RBRACE,
    STATE(2709), 1,
      aux_sym_list_repeat1,
  [147306] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4135), 1,
      anon_sym_SEMI,
    ACTIONS(4137), 1,
      sym__newline,
    STATE(2643), 1,
      aux_sym__simple_statements_repeat1,
  [147319] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4139), 1,
      sym_identifier,
    STATE(2966), 1,
      sym_parameters,
  [147332] = 4,
    ACTIONS(984), 1,
      sym__newline,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4141), 1,
      anon_sym_SEMI,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [147345] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4143), 1,
      anon_sym_COMMA,
    ACTIONS(4145), 1,
      anon_sym_RPAREN,
    STATE(2714), 1,
      aux_sym_argument_list_repeat1,
  [147358] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4147), 1,
      anon_sym_COMMA,
    ACTIONS(4149), 1,
      anon_sym_RPAREN,
    STATE(2615), 1,
      aux_sym_list_repeat1,
  [147371] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4151), 1,
      anon_sym_COMMA,
    ACTIONS(4153), 1,
      anon_sym_RBRACK,
    STATE(2615), 1,
      aux_sym_list_repeat1,
  [147384] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4155), 1,
      anon_sym_COMMA,
    ACTIONS(4158), 1,
      anon_sym_RPAREN,
    STATE(2705), 1,
      aux_sym__parameters_repeat1,
  [147397] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4160), 1,
      anon_sym_COMMA,
    ACTIONS(4162), 1,
      anon_sym_RBRACE,
    STATE(2691), 1,
      aux_sym_dictionary_repeat1,
  [147410] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4164), 3,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [147419] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4063), 1,
      anon_sym_COMMA,
    ACTIONS(4166), 1,
      anon_sym_COLON,
    STATE(2673), 1,
      aux_sym_with_statement_repeat1,
  [147432] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4168), 1,
      anon_sym_COMMA,
    ACTIONS(4170), 1,
      anon_sym_RBRACE,
    STATE(2690), 1,
      aux_sym_list_repeat1,
  [147445] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4172), 1,
      sym_identifier,
    ACTIONS(4174), 1,
      anon_sym_SEMI,
    ACTIONS(4176), 1,
      sym__newline,
  [147458] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4178), 1,
      anon_sym_COMMA,
    ACTIONS(4180), 1,
      anon_sym_RPAREN,
    STATE(2723), 1,
      aux_sym_argument_list_repeat1,
  [147471] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4182), 1,
      anon_sym_SEMI,
    ACTIONS(4184), 1,
      sym__newline,
    STATE(2793), 1,
      aux_sym__simple_statements_repeat1,
  [147484] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4186), 1,
      sym_identifier,
    STATE(2861), 1,
      sym_parameters,
  [147497] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4188), 1,
      anon_sym_COMMA,
    ACTIONS(4190), 1,
      anon_sym_RPAREN,
    STATE(2739), 1,
      aux_sym_argument_list_repeat1,
  [147510] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4046), 1,
      sym_identifier,
    STATE(2861), 1,
      sym_parameters,
  [147523] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4192), 1,
      sym_identifier,
    STATE(2892), 1,
      sym_parameters,
  [147536] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4194), 1,
      sym_identifier,
    STATE(2899), 1,
      sym_parameters,
  [147549] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4079), 1,
      sym_identifier,
    STATE(2892), 1,
      sym_parameters,
  [147562] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4196), 1,
      sym_identifier,
    STATE(2920), 1,
      sym_parameters,
  [147575] = 4,
    ACTIONS(998), 1,
      sym__newline,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4198), 1,
      anon_sym_SEMI,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [147588] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4200), 1,
      anon_sym_COMMA,
    ACTIONS(4202), 1,
      anon_sym_RPAREN,
    STATE(2739), 1,
      aux_sym_argument_list_repeat1,
  [147601] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3906), 3,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [147610] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4204), 1,
      anon_sym_COMMA,
    ACTIONS(4206), 1,
      anon_sym_RPAREN,
    STATE(2739), 1,
      aux_sym_argument_list_repeat1,
  [147623] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4208), 3,
      sym__newline,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [147632] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4210), 1,
      sym_identifier,
    STATE(2959), 1,
      sym_parameters,
  [147645] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4212), 1,
      sym_identifier,
    STATE(2606), 1,
      sym_dotted_name,
    STATE(2656), 1,
      sym_aliased_import,
  [147658] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3901), 1,
      anon_sym_RPAREN,
    ACTIONS(4214), 1,
      anon_sym_COMMA,
    STATE(2727), 1,
      aux_sym__flag_list_repeat1,
  [147671] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4217), 1,
      anon_sym_COMMA,
    ACTIONS(4219), 1,
      anon_sym_RBRACK,
    STATE(2677), 1,
      aux_sym_list_repeat1,
  [147684] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4221), 1,
      anon_sym_COMMA,
    ACTIONS(4223), 1,
      anon_sym_RPAREN,
    STATE(2721), 1,
      aux_sym_argument_list_repeat1,
  [147697] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3944), 1,
      anon_sym_as,
    ACTIONS(3906), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [147708] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4225), 1,
      anon_sym_COMMA,
    ACTIONS(4227), 1,
      anon_sym_in,
    STATE(2644), 1,
      aux_sym_pattern_list_repeat1,
  [147721] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4229), 1,
      sym_identifier,
    STATE(2936), 1,
      sym_parameters,
  [147734] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4231), 1,
      sym_identifier,
    STATE(2917), 1,
      sym_parameters,
  [147747] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3911), 1,
      anon_sym_RPAREN,
    ACTIONS(4073), 1,
      sym_identifier,
    STATE(2888), 1,
      sym_aliased_import_name,
  [147760] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3920), 1,
      anon_sym_RPAREN,
    ACTIONS(4233), 1,
      anon_sym_COMMA,
    STATE(2735), 1,
      aux_sym__import_list_repeat1,
  [147773] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3911), 1,
      anon_sym_RPAREN,
    ACTIONS(4073), 1,
      sym_identifier,
    STATE(2888), 1,
      sym_aliased_import_name,
  [147786] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4236), 1,
      anon_sym_COMMA,
    ACTIONS(4238), 1,
      anon_sym_RBRACK,
    STATE(2653), 1,
      aux_sym_list_repeat1,
  [147799] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4240), 1,
      anon_sym_COMMA,
    ACTIONS(4242), 1,
      anon_sym_RBRACE,
    STATE(2690), 1,
      aux_sym_list_repeat1,
  [147812] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4244), 1,
      anon_sym_COMMA,
    ACTIONS(4247), 1,
      anon_sym_RPAREN,
    STATE(2739), 1,
      aux_sym_argument_list_repeat1,
  [147825] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4249), 1,
      sym_identifier,
    STATE(2948), 1,
      sym_parameters,
  [147838] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4251), 1,
      sym_identifier,
    STATE(2795), 1,
      sym_parameters,
  [147851] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4253), 1,
      anon_sym_COMMA,
    ACTIONS(4255), 1,
      anon_sym_RBRACK,
    STATE(2615), 1,
      aux_sym_list_repeat1,
  [147864] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4257), 1,
      sym_identifier,
    STATE(2894), 1,
      sym_parameters,
  [147877] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4259), 1,
      sym_identifier,
    STATE(2896), 1,
      sym_parameters,
  [147890] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4261), 1,
      sym_identifier,
    STATE(2930), 1,
      sym_parameters,
  [147903] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4263), 1,
      anon_sym_SEMI,
    ACTIONS(4265), 1,
      sym__newline,
    STATE(2750), 1,
      aux_sym__simple_statements_repeat1,
  [147916] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4267), 1,
      sym_identifier,
    STATE(2908), 1,
      sym_parameters,
  [147929] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4269), 1,
      sym_identifier,
    STATE(2910), 1,
      sym_parameters,
  [147942] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4271), 1,
      anon_sym_COMMA,
    ACTIONS(4273), 1,
      anon_sym_RPAREN,
    STATE(2739), 1,
      aux_sym_argument_list_repeat1,
  [147955] = 4,
    ACTIONS(1006), 1,
      sym__newline,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4275), 1,
      anon_sym_SEMI,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [147968] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4277), 1,
      sym_identifier,
    STATE(2922), 1,
      sym_parameters,
  [147981] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4279), 1,
      sym_identifier,
    STATE(2925), 1,
      sym_parameters,
  [147994] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4046), 1,
      sym_identifier,
    STATE(2966), 1,
      sym_parameters,
  [148007] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4281), 1,
      sym_identifier,
    STATE(2933), 1,
      sym_parameters,
  [148020] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4283), 1,
      sym_identifier,
    STATE(2935), 1,
      sym_parameters,
  [148033] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4285), 3,
      sym_fstring_content,
      sym_fstring_end,
      anon_sym_LBRACE,
  [148042] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4287), 1,
      sym_identifier,
    STATE(2944), 1,
      sym_parameters,
  [148055] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4289), 1,
      sym_identifier,
    STATE(2946), 1,
      sym_parameters,
  [148068] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4291), 1,
      anon_sym_SEMI,
    ACTIONS(4293), 1,
      sym__newline,
    STATE(2720), 1,
      aux_sym__simple_statements_repeat1,
  [148081] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4295), 1,
      sym_identifier,
    STATE(2953), 1,
      sym_parameters,
  [148094] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4297), 1,
      sym_identifier,
    STATE(2955), 1,
      sym_parameters,
  [148107] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4299), 1,
      sym_identifier,
    STATE(2963), 1,
      sym_parameters,
  [148120] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4301), 1,
      sym_identifier,
    STATE(2886), 1,
      sym_parameters,
  [148133] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4063), 1,
      anon_sym_COMMA,
    ACTIONS(4303), 1,
      anon_sym_COLON,
    STATE(2708), 1,
      aux_sym_with_statement_repeat1,
  [148146] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4305), 1,
      anon_sym_COMMA,
    ACTIONS(4307), 1,
      anon_sym_RBRACE,
    STATE(2691), 1,
      aux_sym_dictionary_repeat1,
  [148159] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4309), 1,
      sym_identifier,
    STATE(2800), 1,
      sym_parameters,
  [148172] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4311), 1,
      sym_identifier,
    STATE(2802), 1,
      sym_parameters,
  [148185] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4313), 3,
      sym_fstring_content,
      sym_fstring_end,
      anon_sym_LBRACE,
  [148194] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4315), 1,
      sym_identifier,
    STATE(2806), 1,
      sym_parameters,
  [148207] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4317), 1,
      sym_identifier,
    STATE(2808), 1,
      sym_parameters,
  [148220] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4319), 1,
      sym_identifier,
    STATE(2813), 1,
      sym_parameters,
  [148233] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4321), 1,
      sym_identifier,
    STATE(2815), 1,
      sym_parameters,
  [148246] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(2028), 3,
      anon_sym_DOT,
      anon_sym_except,
      anon_sym_finally,
  [148255] = 4,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4323), 1,
      anon_sym_SEMI,
    ACTIONS(4325), 1,
      sym__newline,
    STATE(2663), 1,
      aux_sym__simple_statements_repeat1,
  [148268] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4327), 1,
      sym_identifier,
    STATE(2819), 1,
      sym_parameters,
  [148281] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4329), 1,
      sym_identifier,
    STATE(2821), 1,
      sym_parameters,
  [148294] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4331), 3,
      sym_fstring_content,
      sym_fstring_end,
      anon_sym_LBRACE,
  [148303] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4079), 1,
      sym_identifier,
    STATE(2930), 1,
      sym_parameters,
  [148316] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4333), 1,
      sym_identifier,
    STATE(2825), 1,
      sym_parameters,
  [148329] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4335), 1,
      sym_identifier,
    STATE(2827), 1,
      sym_parameters,
  [148342] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4337), 1,
      sym_identifier,
    ACTIONS(4339), 1,
      anon_sym___python__,
    STATE(3008), 1,
      sym_dotted_name,
  [148355] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4212), 1,
      sym_identifier,
    STATE(2584), 1,
      sym_dotted_name,
    STATE(2603), 1,
      sym_aliased_import,
  [148368] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4341), 1,
      sym_identifier,
    STATE(2831), 1,
      sym_parameters,
  [148381] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      sym_identifier,
    STATE(2833), 1,
      sym_parameters,
  [148394] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4345), 1,
      sym_identifier,
    STATE(2811), 1,
      sym_parameters,
  [148407] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4347), 1,
      anon_sym_COLON,
    ACTIONS(4349), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [148418] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4351), 1,
      sym_identifier,
    STATE(2839), 1,
      sym_parameters,
  [148431] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4353), 1,
      sym_identifier,
    STATE(2841), 1,
      sym_parameters,
  [148444] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4355), 1,
      sym_identifier,
    ACTIONS(4357), 1,
      anon_sym_LPAREN,
    STATE(2842), 1,
      sym__flag_list,
  [148457] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4359), 1,
      sym_identifier,
    STATE(2844), 1,
      sym_parameters,
  [148470] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3973), 1,
      anon_sym_LPAREN,
    ACTIONS(4361), 1,
      sym_identifier,
    STATE(2846), 1,
      sym_parameters,
  [148483] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4172), 1,
      sym_identifier,
    ACTIONS(4363), 1,
      anon_sym_SEMI,
    ACTIONS(4365), 1,
      sym__newline,
  [148496] = 4,
    ACTIONS(994), 1,
      sym__newline,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4367), 1,
      anon_sym_SEMI,
    STATE(2642), 1,
      aux_sym__simple_statements_repeat1,
  [148509] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2866), 1,
      sym_parameters,
  [148519] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4371), 1,
      anon_sym_DASH_GT,
    ACTIONS(4373), 1,
      anon_sym_COLON,
  [148529] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2854), 1,
      sym_parameters,
  [148539] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4375), 1,
      anon_sym_DASH_GT,
    ACTIONS(4377), 1,
      anon_sym_COLON,
  [148549] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4379), 1,
      anon_sym_DASH_GT,
    ACTIONS(4381), 1,
      anon_sym_COLON,
  [148559] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2865), 1,
      sym_parameters,
  [148569] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4383), 1,
      anon_sym_DASH_GT,
    ACTIONS(4385), 1,
      anon_sym_COLON,
  [148579] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2881), 1,
      sym_parameters,
  [148589] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4387), 1,
      anon_sym_DASH_GT,
    ACTIONS(4389), 1,
      anon_sym_COLON,
  [148599] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2942), 1,
      sym_parameters,
  [148609] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4391), 2,
      anon_sym_DASH_GT,
      anon_sym_COLON,
  [148617] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2820), 1,
      sym_parameters,
  [148627] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4393), 1,
      anon_sym_DASH_GT,
    ACTIONS(4395), 1,
      anon_sym_COLON,
  [148637] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2834), 1,
      sym_parameters,
  [148647] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4397), 1,
      anon_sym_DASH_GT,
    ACTIONS(4399), 1,
      anon_sym_COLON,
  [148657] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4401), 1,
      anon_sym_DASH_GT,
    ACTIONS(4403), 1,
      anon_sym_COLON,
  [148667] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4405), 1,
      anon_sym_DASH_GT,
    ACTIONS(4407), 1,
      anon_sym_COLON,
  [148677] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4409), 1,
      anon_sym_DASH_GT,
    ACTIONS(4411), 1,
      anon_sym_COLON,
  [148687] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2885), 1,
      sym_parameters,
  [148697] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4413), 1,
      anon_sym_DASH_GT,
    ACTIONS(4415), 1,
      anon_sym_COLON,
  [148707] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2911), 1,
      sym_parameters,
  [148717] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4417), 1,
      anon_sym_DASH_GT,
    ACTIONS(4419), 1,
      anon_sym_COLON,
  [148727] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4421), 1,
      anon_sym_DASH_GT,
    ACTIONS(4423), 1,
      anon_sym_COLON,
  [148737] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4425), 1,
      anon_sym_DASH_GT,
    ACTIONS(4427), 1,
      anon_sym_COLON,
  [148747] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2858), 1,
      sym_parameters,
  [148757] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4429), 1,
      anon_sym_DASH_GT,
    ACTIONS(4431), 1,
      anon_sym_COLON,
  [148767] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4433), 1,
      anon_sym_DASH_GT,
    ACTIONS(4435), 1,
      anon_sym_COLON,
  [148777] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4437), 1,
      anon_sym_DASH_GT,
    ACTIONS(4439), 1,
      anon_sym_COLON,
  [148787] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4441), 1,
      anon_sym_DASH_GT,
    ACTIONS(4443), 1,
      anon_sym_COLON,
  [148797] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4445), 1,
      anon_sym_DASH_GT,
    ACTIONS(4447), 1,
      anon_sym_COLON,
  [148807] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2874), 1,
      sym_parameters,
  [148817] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4449), 1,
      anon_sym_DASH_GT,
    ACTIONS(4451), 1,
      anon_sym_COLON,
  [148827] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2878), 1,
      sym_parameters,
  [148837] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4453), 1,
      anon_sym_DASH_GT,
    ACTIONS(4455), 1,
      anon_sym_COLON,
  [148847] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4457), 1,
      anon_sym_DASH_GT,
    ACTIONS(4459), 1,
      anon_sym_COLON,
  [148857] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4461), 2,
      sym__newline,
      anon_sym_SEMI,
  [148865] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2904), 1,
      sym_parameters,
  [148875] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4463), 1,
      anon_sym_DASH_GT,
    ACTIONS(4465), 1,
      anon_sym_COLON,
  [148885] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2906), 1,
      sym_parameters,
  [148895] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4467), 1,
      anon_sym_DASH_GT,
    ACTIONS(4469), 1,
      anon_sym_COLON,
  [148905] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4471), 1,
      anon_sym_DASH_GT,
    ACTIONS(4473), 1,
      anon_sym_COLON,
  [148915] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4475), 1,
      sym_identifier,
    STATE(3105), 1,
      sym__flag_list,
  [148925] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4477), 1,
      anon_sym_DASH_GT,
    ACTIONS(4479), 1,
      anon_sym_COLON,
  [148935] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4481), 2,
      sym__newline,
      anon_sym_SEMI,
  [148943] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2950), 1,
      sym_parameters,
  [148953] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4483), 1,
      anon_sym_DASH_GT,
    ACTIONS(4485), 1,
      anon_sym_COLON,
  [148963] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2797), 1,
      sym_parameters,
  [148973] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4487), 1,
      anon_sym_DASH_GT,
    ACTIONS(4489), 1,
      anon_sym_COLON,
  [148983] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4491), 2,
      sym__newline,
      anon_sym_SEMI,
  [148991] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2817), 1,
      sym_parameters,
  [149001] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4493), 1,
      anon_sym_DASH_GT,
    ACTIONS(4495), 1,
      anon_sym_COLON,
  [149011] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2823), 1,
      sym_parameters,
  [149021] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4497), 1,
      anon_sym_DASH_GT,
    ACTIONS(4499), 1,
      anon_sym_COLON,
  [149031] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3745), 2,
      anon_sym_import,
      anon_sym_DOT,
  [149039] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4501), 2,
      sym__newline,
      anon_sym_SEMI,
  [149047] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4052), 2,
      anon_sym_COLON,
      anon_sym_COMMA,
  [149055] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4247), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [149063] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2856), 1,
      sym_parameters,
  [149073] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2857), 1,
      sym_parameters,
  [149083] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2876), 1,
      sym_parameters,
  [149093] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4503), 1,
      anon_sym_DASH_GT,
    ACTIONS(4505), 1,
      anon_sym_COLON,
  [149103] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4507), 1,
      anon_sym_DASH_GT,
    ACTIONS(4509), 1,
      anon_sym_COLON,
  [149113] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4511), 1,
      anon_sym_DASH_GT,
    ACTIONS(4513), 1,
      anon_sym_COLON,
  [149123] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4515), 1,
      anon_sym_DASH_GT,
    ACTIONS(4517), 1,
      anon_sym_COLON,
  [149133] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4519), 1,
      anon_sym_DASH_GT,
    ACTIONS(4521), 1,
      anon_sym_COLON,
  [149143] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2897), 1,
      sym_parameters,
  [149153] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4523), 1,
      anon_sym_DASH_GT,
    ACTIONS(4525), 1,
      anon_sym_COLON,
  [149163] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4527), 1,
      anon_sym_DASH_GT,
    ACTIONS(4529), 1,
      anon_sym_COLON,
  [149173] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2914), 1,
      sym_parameters,
  [149183] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2915), 1,
      sym_parameters,
  [149193] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4531), 1,
      anon_sym_COMMA,
    STATE(2674), 1,
      aux_sym_list_repeat1,
  [149203] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4533), 1,
      anon_sym_DASH_GT,
    ACTIONS(4535), 1,
      anon_sym_COLON,
  [149213] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4537), 1,
      anon_sym_DASH_GT,
    ACTIONS(4539), 1,
      anon_sym_COLON,
  [149223] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2940), 1,
      sym_parameters,
  [149233] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2947), 1,
      sym_parameters,
  [149243] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4541), 1,
      anon_sym_COMMA,
    STATE(2652), 1,
      aux_sym_list_repeat1,
  [149253] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4543), 1,
      anon_sym_DASH_GT,
    ACTIONS(4545), 1,
      anon_sym_COLON,
  [149263] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2961), 1,
      sym_parameters,
  [149273] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4164), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [149281] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4547), 1,
      anon_sym_DASH_GT,
    ACTIONS(4549), 1,
      anon_sym_COLON,
  [149291] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4551), 1,
      anon_sym_DASH_GT,
    ACTIONS(4553), 1,
      anon_sym_COLON,
  [149301] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(3890), 1,
      sym_identifier,
    STATE(2722), 1,
      sym_aliased_import_name,
  [149311] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4555), 1,
      anon_sym_DASH_GT,
    ACTIONS(4557), 1,
      anon_sym_COLON,
  [149321] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4174), 1,
      anon_sym_RPAREN,
    ACTIONS(4559), 1,
      sym_identifier,
  [149331] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4561), 1,
      anon_sym_DASH_GT,
    ACTIONS(4563), 1,
      anon_sym_COLON,
  [149341] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2798), 1,
      sym_parameters,
  [149351] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4363), 1,
      anon_sym_RPAREN,
    ACTIONS(4559), 1,
      sym_identifier,
  [149361] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4565), 1,
      anon_sym_DASH_GT,
    ACTIONS(4567), 1,
      anon_sym_COLON,
  [149371] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4569), 1,
      anon_sym_DASH_GT,
    ACTIONS(4571), 1,
      anon_sym_COLON,
  [149381] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4573), 2,
      anon_sym_DASH_GT,
      anon_sym_COLON,
  [149389] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3987), 2,
      sym__newline,
      anon_sym_SEMI,
  [149397] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4575), 1,
      anon_sym_DASH_GT,
    ACTIONS(4577), 1,
      anon_sym_COLON,
  [149407] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4579), 1,
      anon_sym_DASH_GT,
    ACTIONS(4581), 1,
      anon_sym_COLON,
  [149417] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4583), 1,
      anon_sym_DASH_GT,
    ACTIONS(4585), 1,
      anon_sym_COLON,
  [149427] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3906), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [149435] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4208), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [149443] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2809), 1,
      sym_parameters,
  [149453] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4587), 1,
      anon_sym_DASH_GT,
    ACTIONS(4589), 1,
      anon_sym_COLON,
  [149463] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4591), 1,
      anon_sym_DASH_GT,
    ACTIONS(4593), 1,
      anon_sym_COLON,
  [149473] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2913), 1,
      sym_parameters,
  [149483] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4595), 1,
      anon_sym_DASH_GT,
    ACTIONS(4597), 1,
      anon_sym_COLON,
  [149493] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2941), 1,
      sym_parameters,
  [149503] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4599), 1,
      anon_sym_DASH_GT,
    ACTIONS(4601), 1,
      anon_sym_COLON,
  [149513] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4603), 1,
      anon_sym_DASH_GT,
    ACTIONS(4605), 1,
      anon_sym_COLON,
  [149523] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4607), 2,
      sym__newline,
      anon_sym_SEMI,
  [149531] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4609), 1,
      anon_sym_DASH_GT,
    ACTIONS(4611), 1,
      anon_sym_COLON,
  [149541] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4613), 1,
      anon_sym_DASH_GT,
    ACTIONS(4615), 1,
      anon_sym_COLON,
  [149551] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4617), 1,
      anon_sym_DASH_GT,
    ACTIONS(4619), 1,
      anon_sym_COLON,
  [149561] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4621), 1,
      anon_sym_DASH_GT,
    ACTIONS(4623), 1,
      anon_sym_COLON,
  [149571] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2949), 1,
      sym_parameters,
  [149581] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4625), 1,
      anon_sym_DASH_GT,
    ACTIONS(4627), 1,
      anon_sym_COLON,
  [149591] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4158), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [149599] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4629), 1,
      anon_sym_DASH_GT,
    ACTIONS(4631), 1,
      anon_sym_COLON,
  [149609] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2882), 1,
      sym_parameters,
  [149619] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4633), 1,
      anon_sym_DASH_GT,
    ACTIONS(4635), 1,
      anon_sym_COLON,
  [149629] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2887), 1,
      sym_parameters,
  [149639] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4637), 1,
      anon_sym_DASH_GT,
    ACTIONS(4639), 1,
      anon_sym_COLON,
  [149649] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4641), 1,
      anon_sym_DASH_GT,
    ACTIONS(4643), 1,
      anon_sym_COLON,
  [149659] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2918), 1,
      sym_parameters,
  [149669] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4645), 1,
      anon_sym_DASH_GT,
    ACTIONS(4647), 1,
      anon_sym_COLON,
  [149679] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4649), 1,
      anon_sym_DASH_GT,
    ACTIONS(4651), 1,
      anon_sym_COLON,
  [149689] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4653), 1,
      anon_sym_DASH_GT,
    ACTIONS(4655), 1,
      anon_sym_COLON,
  [149699] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3235), 1,
      anon_sym_COMMA,
    ACTIONS(3250), 1,
      anon_sym_RPAREN,
  [149709] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4657), 1,
      anon_sym_DASH_GT,
    ACTIONS(4659), 1,
      anon_sym_COLON,
  [149719] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4661), 1,
      anon_sym_DASH_GT,
    ACTIONS(4663), 1,
      anon_sym_COLON,
  [149729] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4665), 1,
      anon_sym_DASH_GT,
    ACTIONS(4667), 1,
      anon_sym_COLON,
  [149739] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4669), 1,
      anon_sym_DASH_GT,
    ACTIONS(4671), 1,
      anon_sym_COLON,
  [149749] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2828), 1,
      sym_parameters,
  [149759] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4673), 1,
      anon_sym_DASH_GT,
    ACTIONS(4675), 1,
      anon_sym_COLON,
  [149769] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2927), 1,
      sym_parameters,
  [149779] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2873), 1,
      sym_parameters,
  [149789] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4677), 1,
      anon_sym_DASH_GT,
    ACTIONS(4679), 1,
      anon_sym_COLON,
  [149799] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2964), 1,
      sym_parameters,
  [149809] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4681), 1,
      anon_sym_DASH_GT,
    ACTIONS(4683), 1,
      anon_sym_COLON,
  [149819] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4685), 1,
      anon_sym_DASH_GT,
    ACTIONS(4687), 1,
      anon_sym_COLON,
  [149829] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4109), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [149837] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4689), 1,
      anon_sym_DASH_GT,
    ACTIONS(4691), 1,
      anon_sym_COLON,
  [149847] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4693), 1,
      anon_sym_COMMA,
    STATE(2672), 1,
      aux_sym_list_repeat1,
  [149857] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2951), 1,
      sym_parameters,
  [149867] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4695), 1,
      anon_sym_DASH_GT,
    ACTIONS(4697), 1,
      anon_sym_COLON,
  [149877] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2958), 1,
      sym_parameters,
  [149887] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4699), 1,
      anon_sym_DASH_GT,
    ACTIONS(4701), 1,
      anon_sym_COLON,
  [149897] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4703), 1,
      anon_sym_DASH_GT,
    ACTIONS(4705), 1,
      anon_sym_COLON,
  [149907] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2901), 1,
      sym_parameters,
  [149917] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4707), 1,
      anon_sym_COLON,
    ACTIONS(4709), 1,
      anon_sym_RBRACE,
  [149927] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2891), 1,
      sym_parameters,
  [149937] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4711), 1,
      anon_sym_DASH_GT,
    ACTIONS(4713), 1,
      anon_sym_COLON,
  [149947] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4715), 1,
      anon_sym_DASH_GT,
    ACTIONS(4717), 1,
      anon_sym_COLON,
  [149957] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4719), 1,
      anon_sym_DASH_GT,
    ACTIONS(4721), 1,
      anon_sym_COLON,
  [149967] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2919), 1,
      sym_parameters,
  [149977] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4723), 1,
      anon_sym_DASH_GT,
    ACTIONS(4725), 1,
      anon_sym_COLON,
  [149987] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2928), 1,
      sym_parameters,
  [149997] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4727), 1,
      anon_sym_DASH_GT,
    ACTIONS(4729), 1,
      anon_sym_COLON,
  [150007] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4731), 1,
      anon_sym_DASH_GT,
    ACTIONS(4733), 1,
      anon_sym_COLON,
  [150017] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4735), 1,
      anon_sym_DASH_GT,
    ACTIONS(4737), 1,
      anon_sym_COLON,
  [150027] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4739), 1,
      anon_sym_DASH_GT,
    ACTIONS(4741), 1,
      anon_sym_COLON,
  [150037] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4743), 1,
      anon_sym_DASH_GT,
    ACTIONS(4745), 1,
      anon_sym_COLON,
  [150047] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4747), 1,
      anon_sym_DASH_GT,
    ACTIONS(4749), 1,
      anon_sym_COLON,
  [150057] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2816), 1,
      sym_parameters,
  [150067] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4751), 1,
      anon_sym_DASH_GT,
    ACTIONS(4753), 1,
      anon_sym_COLON,
  [150077] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2810), 1,
      sym_parameters,
  [150087] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4755), 1,
      anon_sym_DASH_GT,
    ACTIONS(4757), 1,
      anon_sym_COLON,
  [150097] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2836), 1,
      sym_parameters,
  [150107] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4073), 1,
      sym_identifier,
    STATE(2888), 1,
      sym_aliased_import_name,
  [150117] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4759), 1,
      anon_sym_DASH_GT,
    ACTIONS(4761), 1,
      anon_sym_COLON,
  [150127] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4763), 1,
      anon_sym_DASH_GT,
    ACTIONS(4765), 1,
      anon_sym_COLON,
  [150137] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4767), 1,
      anon_sym_COMMA,
    STATE(2703), 1,
      aux_sym_list_repeat1,
  [150147] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4769), 1,
      anon_sym_DASH_GT,
    ACTIONS(4771), 1,
      anon_sym_COLON,
  [150157] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2902), 1,
      sym_parameters,
  [150167] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4773), 1,
      anon_sym_DASH_GT,
    ACTIONS(4775), 1,
      anon_sym_COLON,
  [150177] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4777), 1,
      anon_sym_DASH_GT,
    ACTIONS(4779), 1,
      anon_sym_COLON,
  [150187] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4369), 1,
      anon_sym_LPAREN,
    STATE(2860), 1,
      sym_parameters,
  [150197] = 3,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4781), 1,
      anon_sym_DASH_GT,
    ACTIONS(4783), 1,
      anon_sym_COLON,
  [150207] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4785), 1,
      anon_sym_def,
  [150214] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4787), 1,
      anon_sym_RPAREN,
  [150221] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4789), 1,
      anon_sym_in,
  [150228] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4791), 1,
      anon_sym_RBRACE,
  [150235] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4793), 1,
      sym_identifier,
  [150242] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4795), 1,
      anon_sym_in,
  [150249] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4797), 1,
      sym_identifier,
  [150256] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4799), 1,
      anon_sym_RPAREN,
  [150263] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3207), 1,
      anon_sym_RBRACK,
  [150270] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4801), 1,
      sym_identifier,
  [150277] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4803), 1,
      sym_identifier,
  [150284] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4805), 1,
      anon_sym_RPAREN,
  [150291] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4807), 1,
      sym_identifier,
  [150298] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3229), 1,
      anon_sym_RPAREN,
  [150305] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4809), 1,
      sym_identifier,
  [150312] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4811), 1,
      anon_sym_COLON,
  [150319] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4813), 1,
      sym_identifier,
  [150326] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4172), 1,
      sym_identifier,
  [150333] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4815), 1,
      anon_sym_in,
  [150340] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3252), 1,
      anon_sym_RPAREN,
  [150347] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4817), 1,
      anon_sym_RPAREN,
  [150354] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4819), 1,
      sym_identifier,
  [150361] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4821), 1,
      anon_sym_RPAREN,
  [150368] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4823), 1,
      anon_sym_in,
  [150375] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4825), 1,
      anon_sym_COLON,
  [150382] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4827), 1,
      anon_sym_RBRACE,
  [150389] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3266), 1,
      anon_sym_RBRACK,
  [150396] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4829), 1,
      anon_sym_RPAREN,
  [150403] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4831), 1,
      anon_sym_RPAREN,
  [150410] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4833), 1,
      anon_sym_in,
  [150417] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4835), 1,
      anon_sym_RBRACK,
  [150424] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4837), 1,
      anon_sym_in,
  [150431] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4839), 1,
      anon_sym_def,
  [150438] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4841), 1,
      anon_sym_COLON,
  [150445] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4843), 1,
      anon_sym_RBRACK,
  [150452] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4845), 1,
      anon_sym_RBRACK,
  [150459] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4847), 1,
      sym_identifier,
  [150466] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4849), 1,
      anon_sym_RPAREN,
  [150473] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3298), 1,
      anon_sym_COLON,
  [150480] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4851), 1,
      anon_sym_RBRACK,
  [150487] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4853), 1,
      sym_identifier,
  [150494] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4855), 1,
      anon_sym_import,
  [150501] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4857), 1,
      sym_identifier,
  [150508] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4859), 1,
      anon_sym_COLON,
  [150515] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4861), 1,
      anon_sym_in,
  [150522] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4863), 1,
      sym_identifier,
  [150529] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4865), 1,
      anon_sym_RBRACK,
  [150536] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4867), 1,
      anon_sym_RPAREN,
  [150543] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4869), 1,
      anon_sym_RPAREN,
  [150550] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3197), 1,
      anon_sym_RBRACK,
  [150557] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3240), 1,
      anon_sym_RPAREN,
  [150564] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4871), 1,
      sym_identifier,
  [150571] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4873), 1,
      sym_identifier,
  [150578] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4875), 1,
      sym_identifier,
  [150585] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4877), 1,
      anon_sym_import,
  [150592] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4879), 1,
      anon_sym_RPAREN,
  [150599] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3264), 1,
      anon_sym_RBRACK,
  [150606] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4881), 1,
      anon_sym_while,
  [150613] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4883), 1,
      anon_sym_RBRACE,
  [150620] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4885), 1,
      anon_sym_RBRACK,
  [150627] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4887), 1,
      sym_identifier,
  [150634] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4889), 1,
      anon_sym_def,
  [150641] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4891), 1,
      anon_sym_COLON,
  [150648] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4559), 1,
      sym_identifier,
  [150655] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4893), 1,
      sym_identifier,
  [150662] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4895), 1,
      anon_sym_COLON,
  [150669] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4897), 1,
      sym_identifier,
  [150676] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4899), 1,
      anon_sym_COLON,
  [150683] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4901), 1,
      anon_sym_in,
  [150690] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4903), 1,
      sym_identifier,
  [150697] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3223), 1,
      anon_sym_RBRACK,
  [150704] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4905), 1,
      sym_identifier,
  [150711] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4907), 1,
      anon_sym_RBRACE,
  [150718] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4909), 1,
      anon_sym_RBRACE,
  [150725] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4911), 1,
      anon_sym_in,
  [150732] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4913), 1,
      anon_sym_COLON,
  [150739] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3284), 1,
      anon_sym_COLON,
  [150746] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4915), 1,
      sym_identifier,
  [150753] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4917), 1,
      sym_identifier,
  [150760] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4919), 1,
      anon_sym_def,
  [150767] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3238), 1,
      anon_sym_RPAREN,
  [150774] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3209), 1,
      anon_sym_RBRACK,
  [150781] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4921), 1,
      anon_sym_RBRACE,
  [150788] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4923), 1,
      anon_sym_RBRACE,
  [150795] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4925), 1,
      sym_fstring_content,
  [150802] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4927), 1,
      anon_sym_RBRACE,
  [150809] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4929), 1,
      anon_sym_in,
  [150816] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4931), 1,
      anon_sym_RBRACK,
  [150823] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4933), 1,
      anon_sym_def,
  [150830] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4935), 1,
      anon_sym_RPAREN,
  [150837] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4937), 1,
      anon_sym_in,
  [150844] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4939), 1,
      anon_sym_in,
  [150851] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4941), 1,
      anon_sym_RBRACE,
  [150858] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3268), 1,
      anon_sym_RBRACK,
  [150865] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4943), 1,
      anon_sym_RBRACK,
  [150872] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4945), 1,
      anon_sym_in,
  [150879] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4947), 1,
      anon_sym_RPAREN,
  [150886] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4949), 1,
      sym_identifier,
  [150893] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4951), 1,
      sym_identifier,
  [150900] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3262), 1,
      anon_sym_RBRACK,
  [150907] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4953), 1,
      anon_sym_RBRACE,
  [150914] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4955), 1,
      sym_fstring_content,
  [150921] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4957), 1,
      anon_sym_RPAREN,
  [150928] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4959), 1,
      anon_sym_in,
  [150935] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4961), 1,
      anon_sym_in,
  [150942] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4963), 1,
      anon_sym_in,
  [150949] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4965), 1,
      anon_sym_RPAREN,
  [150956] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4967), 1,
      anon_sym_RBRACK,
  [150963] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4969), 1,
      anon_sym_COLON,
  [150970] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4971), 1,
      anon_sym_COLON,
  [150977] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4973), 1,
      sym_identifier,
  [150984] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4975), 1,
      anon_sym_RBRACK,
  [150991] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4977), 1,
      anon_sym_in,
  [150998] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4979), 1,
      anon_sym_COLON,
  [151005] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4981), 1,
      anon_sym_COLON,
  [151012] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4983), 1,
      sym_identifier,
  [151019] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4985), 1,
      anon_sym_while,
  [151026] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4987), 1,
      anon_sym_in,
  [151033] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4989), 1,
      anon_sym_COLON,
  [151040] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4991), 1,
      sym_identifier,
  [151047] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4993), 1,
      anon_sym_def,
  [151054] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4995), 1,
      anon_sym_COLON,
  [151061] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(3278), 1,
      anon_sym_RBRACK,
  [151068] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4997), 1,
      anon_sym_COLON,
  [151075] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(4999), 1,
      anon_sym_DOT,
  [151082] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5001), 1,
      anon_sym_in,
  [151089] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5003), 1,
      anon_sym_def,
  [151096] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5005), 1,
      anon_sym_RPAREN,
  [151103] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5007), 1,
      ts_builtin_sym_end,
  [151110] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5009), 1,
      anon_sym_COLON,
  [151117] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5011), 1,
      anon_sym_DOT,
  [151124] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5013), 1,
      anon_sym_in,
  [151131] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5015), 1,
      anon_sym_in,
  [151138] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5017), 1,
      sym_identifier,
  [151145] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5019), 1,
      anon_sym_def,
  [151152] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5021), 1,
      anon_sym_def,
  [151159] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5023), 1,
      anon_sym_def,
  [151166] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5025), 1,
      anon_sym_in,
  [151173] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5027), 1,
      anon_sym_RPAREN,
  [151180] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5029), 1,
      sym_identifier,
  [151187] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5031), 1,
      sym_identifier,
  [151194] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5033), 1,
      sym_identifier,
  [151201] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5035), 1,
      anon_sym_in,
  [151208] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5037), 1,
      anon_sym_def,
  [151215] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5039), 1,
      anon_sym_def,
  [151222] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5041), 1,
      anon_sym_in,
  [151229] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5043), 1,
      anon_sym_def,
  [151236] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5045), 1,
      anon_sym_def,
  [151243] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5047), 1,
      anon_sym_def,
  [151250] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5049), 1,
      anon_sym_def,
  [151257] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5051), 1,
      anon_sym_def,
  [151264] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5053), 1,
      anon_sym_def,
  [151271] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5055), 1,
      anon_sym_def,
  [151278] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5057), 1,
      anon_sym_def,
  [151285] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5059), 1,
      anon_sym_def,
  [151292] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5061), 1,
      anon_sym_def,
  [151299] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5063), 1,
      anon_sym_def,
  [151306] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5065), 1,
      anon_sym_def,
  [151313] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5067), 1,
      anon_sym_def,
  [151320] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5069), 1,
      anon_sym_def,
  [151327] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5071), 1,
      anon_sym_COLON,
  [151334] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5073), 1,
      anon_sym_def,
  [151341] = 2,
    ACTIONS(1878), 1,
      sym_comment,
    ACTIONS(5075), 1,
      anon_sym_COLON,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(762)] = 0,
  [SMALL_STATE(763)] = 123,
  [SMALL_STATE(764)] = 246,
  [SMALL_STATE(765)] = 369,
  [SMALL_STATE(766)] = 492,
  [SMALL_STATE(767)] = 613,
  [SMALL_STATE(768)] = 731,
  [SMALL_STATE(769)] = 851,
  [SMALL_STATE(770)] = 971,
  [SMALL_STATE(771)] = 1091,
  [SMALL_STATE(772)] = 1209,
  [SMALL_STATE(773)] = 1329,
  [SMALL_STATE(774)] = 1449,
  [SMALL_STATE(775)] = 1569,
  [SMALL_STATE(776)] = 1687,
  [SMALL_STATE(777)] = 1807,
  [SMALL_STATE(778)] = 1927,
  [SMALL_STATE(779)] = 2047,
  [SMALL_STATE(780)] = 2167,
  [SMALL_STATE(781)] = 2287,
  [SMALL_STATE(782)] = 2407,
  [SMALL_STATE(783)] = 2522,
  [SMALL_STATE(784)] = 2641,
  [SMALL_STATE(785)] = 2756,
  [SMALL_STATE(786)] = 2873,
  [SMALL_STATE(787)] = 2988,
  [SMALL_STATE(788)] = 3103,
  [SMALL_STATE(789)] = 3218,
  [SMALL_STATE(790)] = 3333,
  [SMALL_STATE(791)] = 3450,
  [SMALL_STATE(792)] = 3569,
  [SMALL_STATE(793)] = 3686,
  [SMALL_STATE(794)] = 3801,
  [SMALL_STATE(795)] = 3916,
  [SMALL_STATE(796)] = 4031,
  [SMALL_STATE(797)] = 4148,
  [SMALL_STATE(798)] = 4265,
  [SMALL_STATE(799)] = 4382,
  [SMALL_STATE(800)] = 4497,
  [SMALL_STATE(801)] = 4614,
  [SMALL_STATE(802)] = 4731,
  [SMALL_STATE(803)] = 4848,
  [SMALL_STATE(804)] = 4965,
  [SMALL_STATE(805)] = 5082,
  [SMALL_STATE(806)] = 5199,
  [SMALL_STATE(807)] = 5316,
  [SMALL_STATE(808)] = 5433,
  [SMALL_STATE(809)] = 5550,
  [SMALL_STATE(810)] = 5667,
  [SMALL_STATE(811)] = 5784,
  [SMALL_STATE(812)] = 5897,
  [SMALL_STATE(813)] = 6014,
  [SMALL_STATE(814)] = 6131,
  [SMALL_STATE(815)] = 6248,
  [SMALL_STATE(816)] = 6363,
  [SMALL_STATE(817)] = 6478,
  [SMALL_STATE(818)] = 6593,
  [SMALL_STATE(819)] = 6708,
  [SMALL_STATE(820)] = 6823,
  [SMALL_STATE(821)] = 6938,
  [SMALL_STATE(822)] = 7053,
  [SMALL_STATE(823)] = 7168,
  [SMALL_STATE(824)] = 7283,
  [SMALL_STATE(825)] = 7398,
  [SMALL_STATE(826)] = 7511,
  [SMALL_STATE(827)] = 7628,
  [SMALL_STATE(828)] = 7745,
  [SMALL_STATE(829)] = 7864,
  [SMALL_STATE(830)] = 7981,
  [SMALL_STATE(831)] = 8096,
  [SMALL_STATE(832)] = 8213,
  [SMALL_STATE(833)] = 8330,
  [SMALL_STATE(834)] = 8447,
  [SMALL_STATE(835)] = 8562,
  [SMALL_STATE(836)] = 8679,
  [SMALL_STATE(837)] = 8796,
  [SMALL_STATE(838)] = 8909,
  [SMALL_STATE(839)] = 9024,
  [SMALL_STATE(840)] = 9141,
  [SMALL_STATE(841)] = 9258,
  [SMALL_STATE(842)] = 9373,
  [SMALL_STATE(843)] = 9490,
  [SMALL_STATE(844)] = 9607,
  [SMALL_STATE(845)] = 9722,
  [SMALL_STATE(846)] = 9835,
  [SMALL_STATE(847)] = 9952,
  [SMALL_STATE(848)] = 10067,
  [SMALL_STATE(849)] = 10182,
  [SMALL_STATE(850)] = 10299,
  [SMALL_STATE(851)] = 10414,
  [SMALL_STATE(852)] = 10529,
  [SMALL_STATE(853)] = 10644,
  [SMALL_STATE(854)] = 10759,
  [SMALL_STATE(855)] = 10874,
  [SMALL_STATE(856)] = 10989,
  [SMALL_STATE(857)] = 11104,
  [SMALL_STATE(858)] = 11219,
  [SMALL_STATE(859)] = 11336,
  [SMALL_STATE(860)] = 11451,
  [SMALL_STATE(861)] = 11568,
  [SMALL_STATE(862)] = 11683,
  [SMALL_STATE(863)] = 11798,
  [SMALL_STATE(864)] = 11913,
  [SMALL_STATE(865)] = 12028,
  [SMALL_STATE(866)] = 12143,
  [SMALL_STATE(867)] = 12257,
  [SMALL_STATE(868)] = 12373,
  [SMALL_STATE(869)] = 12489,
  [SMALL_STATE(870)] = 12605,
  [SMALL_STATE(871)] = 12717,
  [SMALL_STATE(872)] = 12829,
  [SMALL_STATE(873)] = 12943,
  [SMALL_STATE(874)] = 13059,
  [SMALL_STATE(875)] = 13173,
  [SMALL_STATE(876)] = 13285,
  [SMALL_STATE(877)] = 13401,
  [SMALL_STATE(878)] = 13517,
  [SMALL_STATE(879)] = 13633,
  [SMALL_STATE(880)] = 13749,
  [SMALL_STATE(881)] = 13862,
  [SMALL_STATE(882)] = 13973,
  [SMALL_STATE(883)] = 14084,
  [SMALL_STATE(884)] = 14197,
  [SMALL_STATE(885)] = 14310,
  [SMALL_STATE(886)] = 14423,
  [SMALL_STATE(887)] = 14536,
  [SMALL_STATE(888)] = 14649,
  [SMALL_STATE(889)] = 14762,
  [SMALL_STATE(890)] = 14872,
  [SMALL_STATE(891)] = 14982,
  [SMALL_STATE(892)] = 15092,
  [SMALL_STATE(893)] = 15202,
  [SMALL_STATE(894)] = 15312,
  [SMALL_STATE(895)] = 15422,
  [SMALL_STATE(896)] = 15532,
  [SMALL_STATE(897)] = 15642,
  [SMALL_STATE(898)] = 15752,
  [SMALL_STATE(899)] = 15862,
  [SMALL_STATE(900)] = 15972,
  [SMALL_STATE(901)] = 16082,
  [SMALL_STATE(902)] = 16192,
  [SMALL_STATE(903)] = 16302,
  [SMALL_STATE(904)] = 16412,
  [SMALL_STATE(905)] = 16522,
  [SMALL_STATE(906)] = 16632,
  [SMALL_STATE(907)] = 16742,
  [SMALL_STATE(908)] = 16852,
  [SMALL_STATE(909)] = 16962,
  [SMALL_STATE(910)] = 17072,
  [SMALL_STATE(911)] = 17182,
  [SMALL_STATE(912)] = 17292,
  [SMALL_STATE(913)] = 17402,
  [SMALL_STATE(914)] = 17512,
  [SMALL_STATE(915)] = 17622,
  [SMALL_STATE(916)] = 17732,
  [SMALL_STATE(917)] = 17842,
  [SMALL_STATE(918)] = 17952,
  [SMALL_STATE(919)] = 18062,
  [SMALL_STATE(920)] = 18169,
  [SMALL_STATE(921)] = 18276,
  [SMALL_STATE(922)] = 18383,
  [SMALL_STATE(923)] = 18490,
  [SMALL_STATE(924)] = 18597,
  [SMALL_STATE(925)] = 18704,
  [SMALL_STATE(926)] = 18811,
  [SMALL_STATE(927)] = 18918,
  [SMALL_STATE(928)] = 19025,
  [SMALL_STATE(929)] = 19132,
  [SMALL_STATE(930)] = 19239,
  [SMALL_STATE(931)] = 19346,
  [SMALL_STATE(932)] = 19453,
  [SMALL_STATE(933)] = 19560,
  [SMALL_STATE(934)] = 19667,
  [SMALL_STATE(935)] = 19774,
  [SMALL_STATE(936)] = 19881,
  [SMALL_STATE(937)] = 19952,
  [SMALL_STATE(938)] = 20059,
  [SMALL_STATE(939)] = 20166,
  [SMALL_STATE(940)] = 20237,
  [SMALL_STATE(941)] = 20308,
  [SMALL_STATE(942)] = 20415,
  [SMALL_STATE(943)] = 20522,
  [SMALL_STATE(944)] = 20629,
  [SMALL_STATE(945)] = 20736,
  [SMALL_STATE(946)] = 20843,
  [SMALL_STATE(947)] = 20950,
  [SMALL_STATE(948)] = 21057,
  [SMALL_STATE(949)] = 21164,
  [SMALL_STATE(950)] = 21271,
  [SMALL_STATE(951)] = 21378,
  [SMALL_STATE(952)] = 21485,
  [SMALL_STATE(953)] = 21592,
  [SMALL_STATE(954)] = 21699,
  [SMALL_STATE(955)] = 21806,
  [SMALL_STATE(956)] = 21913,
  [SMALL_STATE(957)] = 22020,
  [SMALL_STATE(958)] = 22127,
  [SMALL_STATE(959)] = 22234,
  [SMALL_STATE(960)] = 22341,
  [SMALL_STATE(961)] = 22448,
  [SMALL_STATE(962)] = 22555,
  [SMALL_STATE(963)] = 22662,
  [SMALL_STATE(964)] = 22769,
  [SMALL_STATE(965)] = 22876,
  [SMALL_STATE(966)] = 22983,
  [SMALL_STATE(967)] = 23090,
  [SMALL_STATE(968)] = 23197,
  [SMALL_STATE(969)] = 23304,
  [SMALL_STATE(970)] = 23411,
  [SMALL_STATE(971)] = 23518,
  [SMALL_STATE(972)] = 23625,
  [SMALL_STATE(973)] = 23732,
  [SMALL_STATE(974)] = 23839,
  [SMALL_STATE(975)] = 23946,
  [SMALL_STATE(976)] = 24053,
  [SMALL_STATE(977)] = 24160,
  [SMALL_STATE(978)] = 24267,
  [SMALL_STATE(979)] = 24374,
  [SMALL_STATE(980)] = 24481,
  [SMALL_STATE(981)] = 24588,
  [SMALL_STATE(982)] = 24695,
  [SMALL_STATE(983)] = 24802,
  [SMALL_STATE(984)] = 24909,
  [SMALL_STATE(985)] = 25016,
  [SMALL_STATE(986)] = 25123,
  [SMALL_STATE(987)] = 25230,
  [SMALL_STATE(988)] = 25337,
  [SMALL_STATE(989)] = 25444,
  [SMALL_STATE(990)] = 25551,
  [SMALL_STATE(991)] = 25658,
  [SMALL_STATE(992)] = 25765,
  [SMALL_STATE(993)] = 25872,
  [SMALL_STATE(994)] = 25979,
  [SMALL_STATE(995)] = 26086,
  [SMALL_STATE(996)] = 26193,
  [SMALL_STATE(997)] = 26300,
  [SMALL_STATE(998)] = 26407,
  [SMALL_STATE(999)] = 26514,
  [SMALL_STATE(1000)] = 26621,
  [SMALL_STATE(1001)] = 26728,
  [SMALL_STATE(1002)] = 26835,
  [SMALL_STATE(1003)] = 26942,
  [SMALL_STATE(1004)] = 27049,
  [SMALL_STATE(1005)] = 27156,
  [SMALL_STATE(1006)] = 27263,
  [SMALL_STATE(1007)] = 27370,
  [SMALL_STATE(1008)] = 27477,
  [SMALL_STATE(1009)] = 27584,
  [SMALL_STATE(1010)] = 27691,
  [SMALL_STATE(1011)] = 27798,
  [SMALL_STATE(1012)] = 27905,
  [SMALL_STATE(1013)] = 28012,
  [SMALL_STATE(1014)] = 28119,
  [SMALL_STATE(1015)] = 28226,
  [SMALL_STATE(1016)] = 28333,
  [SMALL_STATE(1017)] = 28440,
  [SMALL_STATE(1018)] = 28547,
  [SMALL_STATE(1019)] = 28654,
  [SMALL_STATE(1020)] = 28761,
  [SMALL_STATE(1021)] = 28868,
  [SMALL_STATE(1022)] = 28975,
  [SMALL_STATE(1023)] = 29082,
  [SMALL_STATE(1024)] = 29189,
  [SMALL_STATE(1025)] = 29296,
  [SMALL_STATE(1026)] = 29403,
  [SMALL_STATE(1027)] = 29510,
  [SMALL_STATE(1028)] = 29617,
  [SMALL_STATE(1029)] = 29724,
  [SMALL_STATE(1030)] = 29831,
  [SMALL_STATE(1031)] = 29938,
  [SMALL_STATE(1032)] = 30045,
  [SMALL_STATE(1033)] = 30152,
  [SMALL_STATE(1034)] = 30259,
  [SMALL_STATE(1035)] = 30324,
  [SMALL_STATE(1036)] = 30431,
  [SMALL_STATE(1037)] = 30538,
  [SMALL_STATE(1038)] = 30645,
  [SMALL_STATE(1039)] = 30752,
  [SMALL_STATE(1040)] = 30859,
  [SMALL_STATE(1041)] = 30966,
  [SMALL_STATE(1042)] = 31073,
  [SMALL_STATE(1043)] = 31180,
  [SMALL_STATE(1044)] = 31287,
  [SMALL_STATE(1045)] = 31394,
  [SMALL_STATE(1046)] = 31501,
  [SMALL_STATE(1047)] = 31608,
  [SMALL_STATE(1048)] = 31715,
  [SMALL_STATE(1049)] = 31822,
  [SMALL_STATE(1050)] = 31929,
  [SMALL_STATE(1051)] = 32036,
  [SMALL_STATE(1052)] = 32143,
  [SMALL_STATE(1053)] = 32250,
  [SMALL_STATE(1054)] = 32357,
  [SMALL_STATE(1055)] = 32464,
  [SMALL_STATE(1056)] = 32571,
  [SMALL_STATE(1057)] = 32678,
  [SMALL_STATE(1058)] = 32785,
  [SMALL_STATE(1059)] = 32892,
  [SMALL_STATE(1060)] = 32999,
  [SMALL_STATE(1061)] = 33106,
  [SMALL_STATE(1062)] = 33213,
  [SMALL_STATE(1063)] = 33320,
  [SMALL_STATE(1064)] = 33427,
  [SMALL_STATE(1065)] = 33534,
  [SMALL_STATE(1066)] = 33641,
  [SMALL_STATE(1067)] = 33748,
  [SMALL_STATE(1068)] = 33855,
  [SMALL_STATE(1069)] = 33962,
  [SMALL_STATE(1070)] = 34069,
  [SMALL_STATE(1071)] = 34176,
  [SMALL_STATE(1072)] = 34283,
  [SMALL_STATE(1073)] = 34390,
  [SMALL_STATE(1074)] = 34497,
  [SMALL_STATE(1075)] = 34604,
  [SMALL_STATE(1076)] = 34711,
  [SMALL_STATE(1077)] = 34776,
  [SMALL_STATE(1078)] = 34841,
  [SMALL_STATE(1079)] = 34912,
  [SMALL_STATE(1080)] = 35019,
  [SMALL_STATE(1081)] = 35126,
  [SMALL_STATE(1082)] = 35233,
  [SMALL_STATE(1083)] = 35340,
  [SMALL_STATE(1084)] = 35447,
  [SMALL_STATE(1085)] = 35554,
  [SMALL_STATE(1086)] = 35661,
  [SMALL_STATE(1087)] = 35768,
  [SMALL_STATE(1088)] = 35875,
  [SMALL_STATE(1089)] = 35982,
  [SMALL_STATE(1090)] = 36089,
  [SMALL_STATE(1091)] = 36196,
  [SMALL_STATE(1092)] = 36303,
  [SMALL_STATE(1093)] = 36410,
  [SMALL_STATE(1094)] = 36517,
  [SMALL_STATE(1095)] = 36624,
  [SMALL_STATE(1096)] = 36731,
  [SMALL_STATE(1097)] = 36838,
  [SMALL_STATE(1098)] = 36945,
  [SMALL_STATE(1099)] = 37052,
  [SMALL_STATE(1100)] = 37159,
  [SMALL_STATE(1101)] = 37266,
  [SMALL_STATE(1102)] = 37373,
  [SMALL_STATE(1103)] = 37480,
  [SMALL_STATE(1104)] = 37587,
  [SMALL_STATE(1105)] = 37694,
  [SMALL_STATE(1106)] = 37801,
  [SMALL_STATE(1107)] = 37908,
  [SMALL_STATE(1108)] = 38015,
  [SMALL_STATE(1109)] = 38122,
  [SMALL_STATE(1110)] = 38229,
  [SMALL_STATE(1111)] = 38336,
  [SMALL_STATE(1112)] = 38443,
  [SMALL_STATE(1113)] = 38550,
  [SMALL_STATE(1114)] = 38657,
  [SMALL_STATE(1115)] = 38764,
  [SMALL_STATE(1116)] = 38871,
  [SMALL_STATE(1117)] = 38978,
  [SMALL_STATE(1118)] = 39085,
  [SMALL_STATE(1119)] = 39192,
  [SMALL_STATE(1120)] = 39299,
  [SMALL_STATE(1121)] = 39406,
  [SMALL_STATE(1122)] = 39513,
  [SMALL_STATE(1123)] = 39620,
  [SMALL_STATE(1124)] = 39727,
  [SMALL_STATE(1125)] = 39834,
  [SMALL_STATE(1126)] = 39941,
  [SMALL_STATE(1127)] = 40048,
  [SMALL_STATE(1128)] = 40155,
  [SMALL_STATE(1129)] = 40262,
  [SMALL_STATE(1130)] = 40369,
  [SMALL_STATE(1131)] = 40476,
  [SMALL_STATE(1132)] = 40583,
  [SMALL_STATE(1133)] = 40690,
  [SMALL_STATE(1134)] = 40797,
  [SMALL_STATE(1135)] = 40904,
  [SMALL_STATE(1136)] = 41011,
  [SMALL_STATE(1137)] = 41118,
  [SMALL_STATE(1138)] = 41225,
  [SMALL_STATE(1139)] = 41332,
  [SMALL_STATE(1140)] = 41439,
  [SMALL_STATE(1141)] = 41546,
  [SMALL_STATE(1142)] = 41653,
  [SMALL_STATE(1143)] = 41760,
  [SMALL_STATE(1144)] = 41867,
  [SMALL_STATE(1145)] = 41974,
  [SMALL_STATE(1146)] = 42081,
  [SMALL_STATE(1147)] = 42188,
  [SMALL_STATE(1148)] = 42295,
  [SMALL_STATE(1149)] = 42402,
  [SMALL_STATE(1150)] = 42509,
  [SMALL_STATE(1151)] = 42616,
  [SMALL_STATE(1152)] = 42723,
  [SMALL_STATE(1153)] = 42830,
  [SMALL_STATE(1154)] = 42937,
  [SMALL_STATE(1155)] = 43044,
  [SMALL_STATE(1156)] = 43151,
  [SMALL_STATE(1157)] = 43258,
  [SMALL_STATE(1158)] = 43365,
  [SMALL_STATE(1159)] = 43472,
  [SMALL_STATE(1160)] = 43579,
  [SMALL_STATE(1161)] = 43686,
  [SMALL_STATE(1162)] = 43793,
  [SMALL_STATE(1163)] = 43900,
  [SMALL_STATE(1164)] = 44007,
  [SMALL_STATE(1165)] = 44114,
  [SMALL_STATE(1166)] = 44221,
  [SMALL_STATE(1167)] = 44328,
  [SMALL_STATE(1168)] = 44435,
  [SMALL_STATE(1169)] = 44542,
  [SMALL_STATE(1170)] = 44649,
  [SMALL_STATE(1171)] = 44756,
  [SMALL_STATE(1172)] = 44863,
  [SMALL_STATE(1173)] = 44970,
  [SMALL_STATE(1174)] = 45077,
  [SMALL_STATE(1175)] = 45184,
  [SMALL_STATE(1176)] = 45291,
  [SMALL_STATE(1177)] = 45398,
  [SMALL_STATE(1178)] = 45505,
  [SMALL_STATE(1179)] = 45612,
  [SMALL_STATE(1180)] = 45719,
  [SMALL_STATE(1181)] = 45826,
  [SMALL_STATE(1182)] = 45933,
  [SMALL_STATE(1183)] = 46040,
  [SMALL_STATE(1184)] = 46147,
  [SMALL_STATE(1185)] = 46254,
  [SMALL_STATE(1186)] = 46361,
  [SMALL_STATE(1187)] = 46468,
  [SMALL_STATE(1188)] = 46575,
  [SMALL_STATE(1189)] = 46682,
  [SMALL_STATE(1190)] = 46789,
  [SMALL_STATE(1191)] = 46896,
  [SMALL_STATE(1192)] = 47003,
  [SMALL_STATE(1193)] = 47110,
  [SMALL_STATE(1194)] = 47217,
  [SMALL_STATE(1195)] = 47324,
  [SMALL_STATE(1196)] = 47431,
  [SMALL_STATE(1197)] = 47538,
  [SMALL_STATE(1198)] = 47645,
  [SMALL_STATE(1199)] = 47752,
  [SMALL_STATE(1200)] = 47859,
  [SMALL_STATE(1201)] = 47966,
  [SMALL_STATE(1202)] = 48073,
  [SMALL_STATE(1203)] = 48180,
  [SMALL_STATE(1204)] = 48287,
  [SMALL_STATE(1205)] = 48394,
  [SMALL_STATE(1206)] = 48501,
  [SMALL_STATE(1207)] = 48608,
  [SMALL_STATE(1208)] = 48715,
  [SMALL_STATE(1209)] = 48822,
  [SMALL_STATE(1210)] = 48929,
  [SMALL_STATE(1211)] = 49036,
  [SMALL_STATE(1212)] = 49143,
  [SMALL_STATE(1213)] = 49250,
  [SMALL_STATE(1214)] = 49357,
  [SMALL_STATE(1215)] = 49464,
  [SMALL_STATE(1216)] = 49571,
  [SMALL_STATE(1217)] = 49678,
  [SMALL_STATE(1218)] = 49785,
  [SMALL_STATE(1219)] = 49892,
  [SMALL_STATE(1220)] = 49999,
  [SMALL_STATE(1221)] = 50106,
  [SMALL_STATE(1222)] = 50213,
  [SMALL_STATE(1223)] = 50320,
  [SMALL_STATE(1224)] = 50427,
  [SMALL_STATE(1225)] = 50534,
  [SMALL_STATE(1226)] = 50641,
  [SMALL_STATE(1227)] = 50748,
  [SMALL_STATE(1228)] = 50855,
  [SMALL_STATE(1229)] = 50962,
  [SMALL_STATE(1230)] = 51069,
  [SMALL_STATE(1231)] = 51176,
  [SMALL_STATE(1232)] = 51283,
  [SMALL_STATE(1233)] = 51390,
  [SMALL_STATE(1234)] = 51497,
  [SMALL_STATE(1235)] = 51604,
  [SMALL_STATE(1236)] = 51711,
  [SMALL_STATE(1237)] = 51818,
  [SMALL_STATE(1238)] = 51925,
  [SMALL_STATE(1239)] = 52032,
  [SMALL_STATE(1240)] = 52139,
  [SMALL_STATE(1241)] = 52246,
  [SMALL_STATE(1242)] = 52353,
  [SMALL_STATE(1243)] = 52460,
  [SMALL_STATE(1244)] = 52567,
  [SMALL_STATE(1245)] = 52674,
  [SMALL_STATE(1246)] = 52781,
  [SMALL_STATE(1247)] = 52888,
  [SMALL_STATE(1248)] = 52995,
  [SMALL_STATE(1249)] = 53102,
  [SMALL_STATE(1250)] = 53209,
  [SMALL_STATE(1251)] = 53316,
  [SMALL_STATE(1252)] = 53423,
  [SMALL_STATE(1253)] = 53530,
  [SMALL_STATE(1254)] = 53637,
  [SMALL_STATE(1255)] = 53744,
  [SMALL_STATE(1256)] = 53851,
  [SMALL_STATE(1257)] = 53958,
  [SMALL_STATE(1258)] = 54065,
  [SMALL_STATE(1259)] = 54172,
  [SMALL_STATE(1260)] = 54279,
  [SMALL_STATE(1261)] = 54386,
  [SMALL_STATE(1262)] = 54493,
  [SMALL_STATE(1263)] = 54600,
  [SMALL_STATE(1264)] = 54707,
  [SMALL_STATE(1265)] = 54814,
  [SMALL_STATE(1266)] = 54921,
  [SMALL_STATE(1267)] = 55028,
  [SMALL_STATE(1268)] = 55135,
  [SMALL_STATE(1269)] = 55242,
  [SMALL_STATE(1270)] = 55349,
  [SMALL_STATE(1271)] = 55456,
  [SMALL_STATE(1272)] = 55563,
  [SMALL_STATE(1273)] = 55670,
  [SMALL_STATE(1274)] = 55777,
  [SMALL_STATE(1275)] = 55884,
  [SMALL_STATE(1276)] = 55991,
  [SMALL_STATE(1277)] = 56098,
  [SMALL_STATE(1278)] = 56205,
  [SMALL_STATE(1279)] = 56312,
  [SMALL_STATE(1280)] = 56419,
  [SMALL_STATE(1281)] = 56526,
  [SMALL_STATE(1282)] = 56633,
  [SMALL_STATE(1283)] = 56740,
  [SMALL_STATE(1284)] = 56847,
  [SMALL_STATE(1285)] = 56954,
  [SMALL_STATE(1286)] = 57061,
  [SMALL_STATE(1287)] = 57168,
  [SMALL_STATE(1288)] = 57275,
  [SMALL_STATE(1289)] = 57382,
  [SMALL_STATE(1290)] = 57489,
  [SMALL_STATE(1291)] = 57596,
  [SMALL_STATE(1292)] = 57703,
  [SMALL_STATE(1293)] = 57810,
  [SMALL_STATE(1294)] = 57917,
  [SMALL_STATE(1295)] = 58024,
  [SMALL_STATE(1296)] = 58131,
  [SMALL_STATE(1297)] = 58238,
  [SMALL_STATE(1298)] = 58345,
  [SMALL_STATE(1299)] = 58452,
  [SMALL_STATE(1300)] = 58559,
  [SMALL_STATE(1301)] = 58666,
  [SMALL_STATE(1302)] = 58773,
  [SMALL_STATE(1303)] = 58880,
  [SMALL_STATE(1304)] = 58987,
  [SMALL_STATE(1305)] = 59094,
  [SMALL_STATE(1306)] = 59201,
  [SMALL_STATE(1307)] = 59308,
  [SMALL_STATE(1308)] = 59415,
  [SMALL_STATE(1309)] = 59522,
  [SMALL_STATE(1310)] = 59629,
  [SMALL_STATE(1311)] = 59736,
  [SMALL_STATE(1312)] = 59843,
  [SMALL_STATE(1313)] = 59950,
  [SMALL_STATE(1314)] = 60057,
  [SMALL_STATE(1315)] = 60164,
  [SMALL_STATE(1316)] = 60271,
  [SMALL_STATE(1317)] = 60378,
  [SMALL_STATE(1318)] = 60485,
  [SMALL_STATE(1319)] = 60592,
  [SMALL_STATE(1320)] = 60699,
  [SMALL_STATE(1321)] = 60806,
  [SMALL_STATE(1322)] = 60913,
  [SMALL_STATE(1323)] = 61020,
  [SMALL_STATE(1324)] = 61127,
  [SMALL_STATE(1325)] = 61234,
  [SMALL_STATE(1326)] = 61341,
  [SMALL_STATE(1327)] = 61448,
  [SMALL_STATE(1328)] = 61555,
  [SMALL_STATE(1329)] = 61662,
  [SMALL_STATE(1330)] = 61769,
  [SMALL_STATE(1331)] = 61876,
  [SMALL_STATE(1332)] = 61983,
  [SMALL_STATE(1333)] = 62090,
  [SMALL_STATE(1334)] = 62197,
  [SMALL_STATE(1335)] = 62304,
  [SMALL_STATE(1336)] = 62411,
  [SMALL_STATE(1337)] = 62518,
  [SMALL_STATE(1338)] = 62625,
  [SMALL_STATE(1339)] = 62732,
  [SMALL_STATE(1340)] = 62839,
  [SMALL_STATE(1341)] = 62946,
  [SMALL_STATE(1342)] = 63053,
  [SMALL_STATE(1343)] = 63160,
  [SMALL_STATE(1344)] = 63267,
  [SMALL_STATE(1345)] = 63374,
  [SMALL_STATE(1346)] = 63481,
  [SMALL_STATE(1347)] = 63588,
  [SMALL_STATE(1348)] = 63695,
  [SMALL_STATE(1349)] = 63802,
  [SMALL_STATE(1350)] = 63909,
  [SMALL_STATE(1351)] = 64016,
  [SMALL_STATE(1352)] = 64123,
  [SMALL_STATE(1353)] = 64230,
  [SMALL_STATE(1354)] = 64337,
  [SMALL_STATE(1355)] = 64444,
  [SMALL_STATE(1356)] = 64551,
  [SMALL_STATE(1357)] = 64658,
  [SMALL_STATE(1358)] = 64765,
  [SMALL_STATE(1359)] = 64872,
  [SMALL_STATE(1360)] = 64979,
  [SMALL_STATE(1361)] = 65086,
  [SMALL_STATE(1362)] = 65193,
  [SMALL_STATE(1363)] = 65300,
  [SMALL_STATE(1364)] = 65407,
  [SMALL_STATE(1365)] = 65514,
  [SMALL_STATE(1366)] = 65621,
  [SMALL_STATE(1367)] = 65728,
  [SMALL_STATE(1368)] = 65835,
  [SMALL_STATE(1369)] = 65942,
  [SMALL_STATE(1370)] = 66049,
  [SMALL_STATE(1371)] = 66156,
  [SMALL_STATE(1372)] = 66263,
  [SMALL_STATE(1373)] = 66370,
  [SMALL_STATE(1374)] = 66477,
  [SMALL_STATE(1375)] = 66584,
  [SMALL_STATE(1376)] = 66691,
  [SMALL_STATE(1377)] = 66798,
  [SMALL_STATE(1378)] = 66905,
  [SMALL_STATE(1379)] = 67012,
  [SMALL_STATE(1380)] = 67119,
  [SMALL_STATE(1381)] = 67226,
  [SMALL_STATE(1382)] = 67333,
  [SMALL_STATE(1383)] = 67440,
  [SMALL_STATE(1384)] = 67547,
  [SMALL_STATE(1385)] = 67654,
  [SMALL_STATE(1386)] = 67761,
  [SMALL_STATE(1387)] = 67868,
  [SMALL_STATE(1388)] = 67975,
  [SMALL_STATE(1389)] = 68082,
  [SMALL_STATE(1390)] = 68189,
  [SMALL_STATE(1391)] = 68296,
  [SMALL_STATE(1392)] = 68403,
  [SMALL_STATE(1393)] = 68510,
  [SMALL_STATE(1394)] = 68617,
  [SMALL_STATE(1395)] = 68724,
  [SMALL_STATE(1396)] = 68831,
  [SMALL_STATE(1397)] = 68938,
  [SMALL_STATE(1398)] = 69045,
  [SMALL_STATE(1399)] = 69152,
  [SMALL_STATE(1400)] = 69259,
  [SMALL_STATE(1401)] = 69366,
  [SMALL_STATE(1402)] = 69473,
  [SMALL_STATE(1403)] = 69580,
  [SMALL_STATE(1404)] = 69687,
  [SMALL_STATE(1405)] = 69794,
  [SMALL_STATE(1406)] = 69901,
  [SMALL_STATE(1407)] = 70008,
  [SMALL_STATE(1408)] = 70115,
  [SMALL_STATE(1409)] = 70222,
  [SMALL_STATE(1410)] = 70329,
  [SMALL_STATE(1411)] = 70436,
  [SMALL_STATE(1412)] = 70543,
  [SMALL_STATE(1413)] = 70650,
  [SMALL_STATE(1414)] = 70757,
  [SMALL_STATE(1415)] = 70864,
  [SMALL_STATE(1416)] = 70971,
  [SMALL_STATE(1417)] = 71078,
  [SMALL_STATE(1418)] = 71185,
  [SMALL_STATE(1419)] = 71292,
  [SMALL_STATE(1420)] = 71399,
  [SMALL_STATE(1421)] = 71506,
  [SMALL_STATE(1422)] = 71613,
  [SMALL_STATE(1423)] = 71720,
  [SMALL_STATE(1424)] = 71827,
  [SMALL_STATE(1425)] = 71934,
  [SMALL_STATE(1426)] = 72041,
  [SMALL_STATE(1427)] = 72148,
  [SMALL_STATE(1428)] = 72255,
  [SMALL_STATE(1429)] = 72362,
  [SMALL_STATE(1430)] = 72469,
  [SMALL_STATE(1431)] = 72576,
  [SMALL_STATE(1432)] = 72683,
  [SMALL_STATE(1433)] = 72790,
  [SMALL_STATE(1434)] = 72897,
  [SMALL_STATE(1435)] = 73004,
  [SMALL_STATE(1436)] = 73111,
  [SMALL_STATE(1437)] = 73218,
  [SMALL_STATE(1438)] = 73325,
  [SMALL_STATE(1439)] = 73432,
  [SMALL_STATE(1440)] = 73539,
  [SMALL_STATE(1441)] = 73646,
  [SMALL_STATE(1442)] = 73753,
  [SMALL_STATE(1443)] = 73860,
  [SMALL_STATE(1444)] = 73967,
  [SMALL_STATE(1445)] = 74074,
  [SMALL_STATE(1446)] = 74181,
  [SMALL_STATE(1447)] = 74288,
  [SMALL_STATE(1448)] = 74395,
  [SMALL_STATE(1449)] = 74502,
  [SMALL_STATE(1450)] = 74609,
  [SMALL_STATE(1451)] = 74716,
  [SMALL_STATE(1452)] = 74823,
  [SMALL_STATE(1453)] = 74930,
  [SMALL_STATE(1454)] = 75037,
  [SMALL_STATE(1455)] = 75144,
  [SMALL_STATE(1456)] = 75251,
  [SMALL_STATE(1457)] = 75358,
  [SMALL_STATE(1458)] = 75465,
  [SMALL_STATE(1459)] = 75572,
  [SMALL_STATE(1460)] = 75679,
  [SMALL_STATE(1461)] = 75786,
  [SMALL_STATE(1462)] = 75893,
  [SMALL_STATE(1463)] = 76000,
  [SMALL_STATE(1464)] = 76107,
  [SMALL_STATE(1465)] = 76214,
  [SMALL_STATE(1466)] = 76321,
  [SMALL_STATE(1467)] = 76428,
  [SMALL_STATE(1468)] = 76535,
  [SMALL_STATE(1469)] = 76642,
  [SMALL_STATE(1470)] = 76749,
  [SMALL_STATE(1471)] = 76856,
  [SMALL_STATE(1472)] = 76963,
  [SMALL_STATE(1473)] = 77070,
  [SMALL_STATE(1474)] = 77177,
  [SMALL_STATE(1475)] = 77284,
  [SMALL_STATE(1476)] = 77391,
  [SMALL_STATE(1477)] = 77498,
  [SMALL_STATE(1478)] = 77605,
  [SMALL_STATE(1479)] = 77712,
  [SMALL_STATE(1480)] = 77819,
  [SMALL_STATE(1481)] = 77926,
  [SMALL_STATE(1482)] = 78033,
  [SMALL_STATE(1483)] = 78140,
  [SMALL_STATE(1484)] = 78247,
  [SMALL_STATE(1485)] = 78354,
  [SMALL_STATE(1486)] = 78461,
  [SMALL_STATE(1487)] = 78568,
  [SMALL_STATE(1488)] = 78675,
  [SMALL_STATE(1489)] = 78782,
  [SMALL_STATE(1490)] = 78850,
  [SMALL_STATE(1491)] = 78918,
  [SMALL_STATE(1492)] = 78986,
  [SMALL_STATE(1493)] = 79054,
  [SMALL_STATE(1494)] = 79149,
  [SMALL_STATE(1495)] = 79216,
  [SMALL_STATE(1496)] = 79283,
  [SMALL_STATE(1497)] = 79350,
  [SMALL_STATE(1498)] = 79445,
  [SMALL_STATE(1499)] = 79540,
  [SMALL_STATE(1500)] = 79635,
  [SMALL_STATE(1501)] = 79699,
  [SMALL_STATE(1502)] = 79771,
  [SMALL_STATE(1503)] = 79863,
  [SMALL_STATE(1504)] = 79955,
  [SMALL_STATE(1505)] = 80027,
  [SMALL_STATE(1506)] = 80091,
  [SMALL_STATE(1507)] = 80155,
  [SMALL_STATE(1508)] = 80219,
  [SMALL_STATE(1509)] = 80287,
  [SMALL_STATE(1510)] = 80355,
  [SMALL_STATE(1511)] = 80419,
  [SMALL_STATE(1512)] = 80483,
  [SMALL_STATE(1513)] = 80541,
  [SMALL_STATE(1514)] = 80605,
  [SMALL_STATE(1515)] = 80673,
  [SMALL_STATE(1516)] = 80731,
  [SMALL_STATE(1517)] = 80795,
  [SMALL_STATE(1518)] = 80857,
  [SMALL_STATE(1519)] = 80919,
  [SMALL_STATE(1520)] = 80981,
  [SMALL_STATE(1521)] = 81043,
  [SMALL_STATE(1522)] = 81105,
  [SMALL_STATE(1523)] = 81169,
  [SMALL_STATE(1524)] = 81233,
  [SMALL_STATE(1525)] = 81297,
  [SMALL_STATE(1526)] = 81365,
  [SMALL_STATE(1527)] = 81429,
  [SMALL_STATE(1528)] = 81493,
  [SMALL_STATE(1529)] = 81551,
  [SMALL_STATE(1530)] = 81613,
  [SMALL_STATE(1531)] = 81676,
  [SMALL_STATE(1532)] = 81739,
  [SMALL_STATE(1533)] = 81795,
  [SMALL_STATE(1534)] = 81851,
  [SMALL_STATE(1535)] = 81907,
  [SMALL_STATE(1536)] = 81963,
  [SMALL_STATE(1537)] = 82025,
  [SMALL_STATE(1538)] = 82087,
  [SMALL_STATE(1539)] = 82143,
  [SMALL_STATE(1540)] = 82199,
  [SMALL_STATE(1541)] = 82255,
  [SMALL_STATE(1542)] = 82311,
  [SMALL_STATE(1543)] = 82367,
  [SMALL_STATE(1544)] = 82423,
  [SMALL_STATE(1545)] = 82479,
  [SMALL_STATE(1546)] = 82535,
  [SMALL_STATE(1547)] = 82591,
  [SMALL_STATE(1548)] = 82647,
  [SMALL_STATE(1549)] = 82703,
  [SMALL_STATE(1550)] = 82759,
  [SMALL_STATE(1551)] = 82815,
  [SMALL_STATE(1552)] = 82871,
  [SMALL_STATE(1553)] = 82927,
  [SMALL_STATE(1554)] = 82983,
  [SMALL_STATE(1555)] = 83045,
  [SMALL_STATE(1556)] = 83101,
  [SMALL_STATE(1557)] = 83163,
  [SMALL_STATE(1558)] = 83219,
  [SMALL_STATE(1559)] = 83275,
  [SMALL_STATE(1560)] = 83331,
  [SMALL_STATE(1561)] = 83386,
  [SMALL_STATE(1562)] = 83441,
  [SMALL_STATE(1563)] = 83496,
  [SMALL_STATE(1564)] = 83551,
  [SMALL_STATE(1565)] = 83606,
  [SMALL_STATE(1566)] = 83665,
  [SMALL_STATE(1567)] = 83720,
  [SMALL_STATE(1568)] = 83779,
  [SMALL_STATE(1569)] = 83838,
  [SMALL_STATE(1570)] = 83893,
  [SMALL_STATE(1571)] = 83948,
  [SMALL_STATE(1572)] = 84003,
  [SMALL_STATE(1573)] = 84058,
  [SMALL_STATE(1574)] = 84117,
  [SMALL_STATE(1575)] = 84172,
  [SMALL_STATE(1576)] = 84227,
  [SMALL_STATE(1577)] = 84282,
  [SMALL_STATE(1578)] = 84337,
  [SMALL_STATE(1579)] = 84396,
  [SMALL_STATE(1580)] = 84451,
  [SMALL_STATE(1581)] = 84510,
  [SMALL_STATE(1582)] = 84565,
  [SMALL_STATE(1583)] = 84619,
  [SMALL_STATE(1584)] = 84673,
  [SMALL_STATE(1585)] = 84727,
  [SMALL_STATE(1586)] = 84781,
  [SMALL_STATE(1587)] = 84835,
  [SMALL_STATE(1588)] = 84889,
  [SMALL_STATE(1589)] = 84943,
  [SMALL_STATE(1590)] = 84997,
  [SMALL_STATE(1591)] = 85051,
  [SMALL_STATE(1592)] = 85105,
  [SMALL_STATE(1593)] = 85159,
  [SMALL_STATE(1594)] = 85213,
  [SMALL_STATE(1595)] = 85267,
  [SMALL_STATE(1596)] = 85321,
  [SMALL_STATE(1597)] = 85375,
  [SMALL_STATE(1598)] = 85429,
  [SMALL_STATE(1599)] = 85483,
  [SMALL_STATE(1600)] = 85537,
  [SMALL_STATE(1601)] = 85590,
  [SMALL_STATE(1602)] = 85643,
  [SMALL_STATE(1603)] = 85696,
  [SMALL_STATE(1604)] = 85749,
  [SMALL_STATE(1605)] = 85802,
  [SMALL_STATE(1606)] = 85855,
  [SMALL_STATE(1607)] = 85908,
  [SMALL_STATE(1608)] = 85961,
  [SMALL_STATE(1609)] = 86022,
  [SMALL_STATE(1610)] = 86075,
  [SMALL_STATE(1611)] = 86128,
  [SMALL_STATE(1612)] = 86181,
  [SMALL_STATE(1613)] = 86234,
  [SMALL_STATE(1614)] = 86287,
  [SMALL_STATE(1615)] = 86340,
  [SMALL_STATE(1616)] = 86393,
  [SMALL_STATE(1617)] = 86446,
  [SMALL_STATE(1618)] = 86499,
  [SMALL_STATE(1619)] = 86552,
  [SMALL_STATE(1620)] = 86605,
  [SMALL_STATE(1621)] = 86658,
  [SMALL_STATE(1622)] = 86711,
  [SMALL_STATE(1623)] = 86764,
  [SMALL_STATE(1624)] = 86817,
  [SMALL_STATE(1625)] = 86870,
  [SMALL_STATE(1626)] = 86923,
  [SMALL_STATE(1627)] = 86976,
  [SMALL_STATE(1628)] = 87029,
  [SMALL_STATE(1629)] = 87082,
  [SMALL_STATE(1630)] = 87135,
  [SMALL_STATE(1631)] = 87188,
  [SMALL_STATE(1632)] = 87241,
  [SMALL_STATE(1633)] = 87294,
  [SMALL_STATE(1634)] = 87347,
  [SMALL_STATE(1635)] = 87400,
  [SMALL_STATE(1636)] = 87453,
  [SMALL_STATE(1637)] = 87506,
  [SMALL_STATE(1638)] = 87559,
  [SMALL_STATE(1639)] = 87612,
  [SMALL_STATE(1640)] = 87665,
  [SMALL_STATE(1641)] = 87718,
  [SMALL_STATE(1642)] = 87771,
  [SMALL_STATE(1643)] = 87824,
  [SMALL_STATE(1644)] = 87877,
  [SMALL_STATE(1645)] = 87930,
  [SMALL_STATE(1646)] = 87983,
  [SMALL_STATE(1647)] = 88036,
  [SMALL_STATE(1648)] = 88089,
  [SMALL_STATE(1649)] = 88142,
  [SMALL_STATE(1650)] = 88195,
  [SMALL_STATE(1651)] = 88248,
  [SMALL_STATE(1652)] = 88301,
  [SMALL_STATE(1653)] = 88354,
  [SMALL_STATE(1654)] = 88407,
  [SMALL_STATE(1655)] = 88460,
  [SMALL_STATE(1656)] = 88513,
  [SMALL_STATE(1657)] = 88566,
  [SMALL_STATE(1658)] = 88619,
  [SMALL_STATE(1659)] = 88672,
  [SMALL_STATE(1660)] = 88724,
  [SMALL_STATE(1661)] = 88776,
  [SMALL_STATE(1662)] = 88828,
  [SMALL_STATE(1663)] = 88880,
  [SMALL_STATE(1664)] = 88932,
  [SMALL_STATE(1665)] = 88984,
  [SMALL_STATE(1666)] = 89036,
  [SMALL_STATE(1667)] = 89088,
  [SMALL_STATE(1668)] = 89140,
  [SMALL_STATE(1669)] = 89192,
  [SMALL_STATE(1670)] = 89244,
  [SMALL_STATE(1671)] = 89296,
  [SMALL_STATE(1672)] = 89348,
  [SMALL_STATE(1673)] = 89400,
  [SMALL_STATE(1674)] = 89452,
  [SMALL_STATE(1675)] = 89504,
  [SMALL_STATE(1676)] = 89556,
  [SMALL_STATE(1677)] = 89608,
  [SMALL_STATE(1678)] = 89660,
  [SMALL_STATE(1679)] = 89712,
  [SMALL_STATE(1680)] = 89764,
  [SMALL_STATE(1681)] = 89816,
  [SMALL_STATE(1682)] = 89868,
  [SMALL_STATE(1683)] = 89920,
  [SMALL_STATE(1684)] = 89972,
  [SMALL_STATE(1685)] = 90024,
  [SMALL_STATE(1686)] = 90076,
  [SMALL_STATE(1687)] = 90128,
  [SMALL_STATE(1688)] = 90180,
  [SMALL_STATE(1689)] = 90232,
  [SMALL_STATE(1690)] = 90284,
  [SMALL_STATE(1691)] = 90336,
  [SMALL_STATE(1692)] = 90388,
  [SMALL_STATE(1693)] = 90440,
  [SMALL_STATE(1694)] = 90492,
  [SMALL_STATE(1695)] = 90544,
  [SMALL_STATE(1696)] = 90596,
  [SMALL_STATE(1697)] = 90648,
  [SMALL_STATE(1698)] = 90700,
  [SMALL_STATE(1699)] = 90752,
  [SMALL_STATE(1700)] = 90804,
  [SMALL_STATE(1701)] = 90856,
  [SMALL_STATE(1702)] = 90908,
  [SMALL_STATE(1703)] = 90960,
  [SMALL_STATE(1704)] = 91012,
  [SMALL_STATE(1705)] = 91064,
  [SMALL_STATE(1706)] = 91147,
  [SMALL_STATE(1707)] = 91230,
  [SMALL_STATE(1708)] = 91313,
  [SMALL_STATE(1709)] = 91396,
  [SMALL_STATE(1710)] = 91479,
  [SMALL_STATE(1711)] = 91561,
  [SMALL_STATE(1712)] = 91639,
  [SMALL_STATE(1713)] = 91721,
  [SMALL_STATE(1714)] = 91801,
  [SMALL_STATE(1715)] = 91883,
  [SMALL_STATE(1716)] = 91965,
  [SMALL_STATE(1717)] = 92017,
  [SMALL_STATE(1718)] = 92099,
  [SMALL_STATE(1719)] = 92179,
  [SMALL_STATE(1720)] = 92261,
  [SMALL_STATE(1721)] = 92313,
  [SMALL_STATE(1722)] = 92395,
  [SMALL_STATE(1723)] = 92475,
  [SMALL_STATE(1724)] = 92557,
  [SMALL_STATE(1725)] = 92619,
  [SMALL_STATE(1726)] = 92697,
  [SMALL_STATE(1727)] = 92779,
  [SMALL_STATE(1728)] = 92837,
  [SMALL_STATE(1729)] = 92889,
  [SMALL_STATE(1730)] = 92939,
  [SMALL_STATE(1731)] = 93013,
  [SMALL_STATE(1732)] = 93065,
  [SMALL_STATE(1733)] = 93115,
  [SMALL_STATE(1734)] = 93197,
  [SMALL_STATE(1735)] = 93279,
  [SMALL_STATE(1736)] = 93357,
  [SMALL_STATE(1737)] = 93423,
  [SMALL_STATE(1738)] = 93477,
  [SMALL_STATE(1739)] = 93555,
  [SMALL_STATE(1740)] = 93637,
  [SMALL_STATE(1741)] = 93719,
  [SMALL_STATE(1742)] = 93801,
  [SMALL_STATE(1743)] = 93883,
  [SMALL_STATE(1744)] = 93963,
  [SMALL_STATE(1745)] = 94013,
  [SMALL_STATE(1746)] = 94095,
  [SMALL_STATE(1747)] = 94149,
  [SMALL_STATE(1748)] = 94227,
  [SMALL_STATE(1749)] = 94283,
  [SMALL_STATE(1750)] = 94365,
  [SMALL_STATE(1751)] = 94417,
  [SMALL_STATE(1752)] = 94481,
  [SMALL_STATE(1753)] = 94563,
  [SMALL_STATE(1754)] = 94645,
  [SMALL_STATE(1755)] = 94727,
  [SMALL_STATE(1756)] = 94805,
  [SMALL_STATE(1757)] = 94887,
  [SMALL_STATE(1758)] = 94961,
  [SMALL_STATE(1759)] = 95019,
  [SMALL_STATE(1760)] = 95099,
  [SMALL_STATE(1761)] = 95181,
  [SMALL_STATE(1762)] = 95233,
  [SMALL_STATE(1763)] = 95285,
  [SMALL_STATE(1764)] = 95339,
  [SMALL_STATE(1765)] = 95421,
  [SMALL_STATE(1766)] = 95503,
  [SMALL_STATE(1767)] = 95585,
  [SMALL_STATE(1768)] = 95667,
  [SMALL_STATE(1769)] = 95741,
  [SMALL_STATE(1770)] = 95817,
  [SMALL_STATE(1771)] = 95875,
  [SMALL_STATE(1772)] = 95924,
  [SMALL_STATE(1773)] = 95997,
  [SMALL_STATE(1774)] = 96076,
  [SMALL_STATE(1775)] = 96159,
  [SMALL_STATE(1776)] = 96242,
  [SMALL_STATE(1777)] = 96325,
  [SMALL_STATE(1778)] = 96402,
  [SMALL_STATE(1779)] = 96479,
  [SMALL_STATE(1780)] = 96562,
  [SMALL_STATE(1781)] = 96639,
  [SMALL_STATE(1782)] = 96720,
  [SMALL_STATE(1783)] = 96801,
  [SMALL_STATE(1784)] = 96882,
  [SMALL_STATE(1785)] = 96939,
  [SMALL_STATE(1786)] = 96996,
  [SMALL_STATE(1787)] = 97047,
  [SMALL_STATE(1788)] = 97098,
  [SMALL_STATE(1789)] = 97149,
  [SMALL_STATE(1790)] = 97206,
  [SMALL_STATE(1791)] = 97257,
  [SMALL_STATE(1792)] = 97308,
  [SMALL_STATE(1793)] = 97359,
  [SMALL_STATE(1794)] = 97410,
  [SMALL_STATE(1795)] = 97461,
  [SMALL_STATE(1796)] = 97512,
  [SMALL_STATE(1797)] = 97563,
  [SMALL_STATE(1798)] = 97642,
  [SMALL_STATE(1799)] = 97693,
  [SMALL_STATE(1800)] = 97744,
  [SMALL_STATE(1801)] = 97823,
  [SMALL_STATE(1802)] = 97900,
  [SMALL_STATE(1803)] = 97951,
  [SMALL_STATE(1804)] = 98002,
  [SMALL_STATE(1805)] = 98075,
  [SMALL_STATE(1806)] = 98150,
  [SMALL_STATE(1807)] = 98205,
  [SMALL_STATE(1808)] = 98270,
  [SMALL_STATE(1809)] = 98333,
  [SMALL_STATE(1810)] = 98394,
  [SMALL_STATE(1811)] = 98451,
  [SMALL_STATE(1812)] = 98508,
  [SMALL_STATE(1813)] = 98587,
  [SMALL_STATE(1814)] = 98664,
  [SMALL_STATE(1815)] = 98737,
  [SMALL_STATE(1816)] = 98816,
  [SMALL_STATE(1817)] = 98893,
  [SMALL_STATE(1818)] = 98970,
  [SMALL_STATE(1819)] = 99047,
  [SMALL_STATE(1820)] = 99126,
  [SMALL_STATE(1821)] = 99199,
  [SMALL_STATE(1822)] = 99280,
  [SMALL_STATE(1823)] = 99359,
  [SMALL_STATE(1824)] = 99436,
  [SMALL_STATE(1825)] = 99485,
  [SMALL_STATE(1826)] = 99566,
  [SMALL_STATE(1827)] = 99645,
  [SMALL_STATE(1828)] = 99722,
  [SMALL_STATE(1829)] = 99799,
  [SMALL_STATE(1830)] = 99872,
  [SMALL_STATE(1831)] = 99921,
  [SMALL_STATE(1832)] = 99970,
  [SMALL_STATE(1833)] = 100019,
  [SMALL_STATE(1834)] = 100070,
  [SMALL_STATE(1835)] = 100147,
  [SMALL_STATE(1836)] = 100198,
  [SMALL_STATE(1837)] = 100249,
  [SMALL_STATE(1838)] = 100322,
  [SMALL_STATE(1839)] = 100397,
  [SMALL_STATE(1840)] = 100452,
  [SMALL_STATE(1841)] = 100517,
  [SMALL_STATE(1842)] = 100580,
  [SMALL_STATE(1843)] = 100641,
  [SMALL_STATE(1844)] = 100698,
  [SMALL_STATE(1845)] = 100755,
  [SMALL_STATE(1846)] = 100834,
  [SMALL_STATE(1847)] = 100911,
  [SMALL_STATE(1848)] = 100984,
  [SMALL_STATE(1849)] = 101063,
  [SMALL_STATE(1850)] = 101140,
  [SMALL_STATE(1851)] = 101217,
  [SMALL_STATE(1852)] = 101294,
  [SMALL_STATE(1853)] = 101373,
  [SMALL_STATE(1854)] = 101446,
  [SMALL_STATE(1855)] = 101495,
  [SMALL_STATE(1856)] = 101544,
  [SMALL_STATE(1857)] = 101621,
  [SMALL_STATE(1858)] = 101670,
  [SMALL_STATE(1859)] = 101719,
  [SMALL_STATE(1860)] = 101770,
  [SMALL_STATE(1861)] = 101849,
  [SMALL_STATE(1862)] = 101930,
  [SMALL_STATE(1863)] = 101981,
  [SMALL_STATE(1864)] = 102064,
  [SMALL_STATE(1865)] = 102141,
  [SMALL_STATE(1866)] = 102192,
  [SMALL_STATE(1867)] = 102243,
  [SMALL_STATE(1868)] = 102316,
  [SMALL_STATE(1869)] = 102391,
  [SMALL_STATE(1870)] = 102446,
  [SMALL_STATE(1871)] = 102511,
  [SMALL_STATE(1872)] = 102574,
  [SMALL_STATE(1873)] = 102635,
  [SMALL_STATE(1874)] = 102692,
  [SMALL_STATE(1875)] = 102773,
  [SMALL_STATE(1876)] = 102830,
  [SMALL_STATE(1877)] = 102911,
  [SMALL_STATE(1878)] = 102987,
  [SMALL_STATE(1879)] = 103063,
  [SMALL_STATE(1880)] = 103119,
  [SMALL_STATE(1881)] = 103199,
  [SMALL_STATE(1882)] = 103279,
  [SMALL_STATE(1883)] = 103359,
  [SMALL_STATE(1884)] = 103409,
  [SMALL_STATE(1885)] = 103459,
  [SMALL_STATE(1886)] = 103509,
  [SMALL_STATE(1887)] = 103589,
  [SMALL_STATE(1888)] = 103665,
  [SMALL_STATE(1889)] = 103715,
  [SMALL_STATE(1890)] = 103765,
  [SMALL_STATE(1891)] = 103837,
  [SMALL_STATE(1892)] = 103911,
  [SMALL_STATE(1893)] = 103965,
  [SMALL_STATE(1894)] = 104029,
  [SMALL_STATE(1895)] = 104091,
  [SMALL_STATE(1896)] = 104151,
  [SMALL_STATE(1897)] = 104207,
  [SMALL_STATE(1898)] = 104263,
  [SMALL_STATE(1899)] = 104335,
  [SMALL_STATE(1900)] = 104391,
  [SMALL_STATE(1901)] = 104469,
  [SMALL_STATE(1902)] = 104549,
  [SMALL_STATE(1903)] = 104597,
  [SMALL_STATE(1904)] = 104673,
  [SMALL_STATE(1905)] = 104723,
  [SMALL_STATE(1906)] = 104773,
  [SMALL_STATE(1907)] = 104845,
  [SMALL_STATE(1908)] = 104919,
  [SMALL_STATE(1909)] = 104973,
  [SMALL_STATE(1910)] = 105037,
  [SMALL_STATE(1911)] = 105099,
  [SMALL_STATE(1912)] = 105159,
  [SMALL_STATE(1913)] = 105215,
  [SMALL_STATE(1914)] = 105295,
  [SMALL_STATE(1915)] = 105351,
  [SMALL_STATE(1916)] = 105429,
  [SMALL_STATE(1917)] = 105479,
  [SMALL_STATE(1918)] = 105551,
  [SMALL_STATE(1919)] = 105599,
  [SMALL_STATE(1920)] = 105647,
  [SMALL_STATE(1921)] = 105695,
  [SMALL_STATE(1922)] = 105771,
  [SMALL_STATE(1923)] = 105851,
  [SMALL_STATE(1924)] = 105923,
  [SMALL_STATE(1925)] = 106001,
  [SMALL_STATE(1926)] = 106077,
  [SMALL_STATE(1927)] = 106149,
  [SMALL_STATE(1928)] = 106199,
  [SMALL_STATE(1929)] = 106277,
  [SMALL_STATE(1930)] = 106327,
  [SMALL_STATE(1931)] = 106403,
  [SMALL_STATE(1932)] = 106483,
  [SMALL_STATE(1933)] = 106533,
  [SMALL_STATE(1934)] = 106583,
  [SMALL_STATE(1935)] = 106661,
  [SMALL_STATE(1936)] = 106741,
  [SMALL_STATE(1937)] = 106817,
  [SMALL_STATE(1938)] = 106897,
  [SMALL_STATE(1939)] = 106973,
  [SMALL_STATE(1940)] = 107049,
  [SMALL_STATE(1941)] = 107129,
  [SMALL_STATE(1942)] = 107209,
  [SMALL_STATE(1943)] = 107289,
  [SMALL_STATE(1944)] = 107369,
  [SMALL_STATE(1945)] = 107447,
  [SMALL_STATE(1946)] = 107493,
  [SMALL_STATE(1947)] = 107571,
  [SMALL_STATE(1948)] = 107621,
  [SMALL_STATE(1949)] = 107671,
  [SMALL_STATE(1950)] = 107745,
  [SMALL_STATE(1951)] = 107821,
  [SMALL_STATE(1952)] = 107875,
  [SMALL_STATE(1953)] = 107939,
  [SMALL_STATE(1954)] = 108017,
  [SMALL_STATE(1955)] = 108079,
  [SMALL_STATE(1956)] = 108139,
  [SMALL_STATE(1957)] = 108195,
  [SMALL_STATE(1958)] = 108251,
  [SMALL_STATE(1959)] = 108331,
  [SMALL_STATE(1960)] = 108409,
  [SMALL_STATE(1961)] = 108483,
  [SMALL_STATE(1962)] = 108563,
  [SMALL_STATE(1963)] = 108613,
  [SMALL_STATE(1964)] = 108691,
  [SMALL_STATE(1965)] = 108769,
  [SMALL_STATE(1966)] = 108847,
  [SMALL_STATE(1967)] = 108925,
  [SMALL_STATE(1968)] = 109005,
  [SMALL_STATE(1969)] = 109055,
  [SMALL_STATE(1970)] = 109133,
  [SMALL_STATE(1971)] = 109183,
  [SMALL_STATE(1972)] = 109259,
  [SMALL_STATE(1973)] = 109339,
  [SMALL_STATE(1974)] = 109389,
  [SMALL_STATE(1975)] = 109463,
  [SMALL_STATE(1976)] = 109511,
  [SMALL_STATE(1977)] = 109587,
  [SMALL_STATE(1978)] = 109663,
  [SMALL_STATE(1979)] = 109711,
  [SMALL_STATE(1980)] = 109759,
  [SMALL_STATE(1981)] = 109809,
  [SMALL_STATE(1982)] = 109885,
  [SMALL_STATE(1983)] = 109931,
  [SMALL_STATE(1984)] = 109987,
  [SMALL_STATE(1985)] = 110035,
  [SMALL_STATE(1986)] = 110083,
  [SMALL_STATE(1987)] = 110163,
  [SMALL_STATE(1988)] = 110241,
  [SMALL_STATE(1989)] = 110316,
  [SMALL_STATE(1990)] = 110371,
  [SMALL_STATE(1991)] = 110426,
  [SMALL_STATE(1992)] = 110503,
  [SMALL_STATE(1993)] = 110564,
  [SMALL_STATE(1994)] = 110641,
  [SMALL_STATE(1995)] = 110704,
  [SMALL_STATE(1996)] = 110779,
  [SMALL_STATE(1997)] = 110850,
  [SMALL_STATE(1998)] = 110899,
  [SMALL_STATE(1999)] = 110960,
  [SMALL_STATE(2000)] = 111019,
  [SMALL_STATE(2001)] = 111096,
  [SMALL_STATE(2002)] = 111145,
  [SMALL_STATE(2003)] = 111220,
  [SMALL_STATE(2004)] = 111279,
  [SMALL_STATE(2005)] = 111334,
  [SMALL_STATE(2006)] = 111409,
  [SMALL_STATE(2007)] = 111464,
  [SMALL_STATE(2008)] = 111539,
  [SMALL_STATE(2009)] = 111594,
  [SMALL_STATE(2010)] = 111671,
  [SMALL_STATE(2011)] = 111748,
  [SMALL_STATE(2012)] = 111829,
  [SMALL_STATE(2013)] = 111904,
  [SMALL_STATE(2014)] = 111953,
  [SMALL_STATE(2015)] = 112024,
  [SMALL_STATE(2016)] = 112073,
  [SMALL_STATE(2017)] = 112120,
  [SMALL_STATE(2018)] = 112195,
  [SMALL_STATE(2019)] = 112244,
  [SMALL_STATE(2020)] = 112293,
  [SMALL_STATE(2021)] = 112346,
  [SMALL_STATE(2022)] = 112417,
  [SMALL_STATE(2023)] = 112490,
  [SMALL_STATE(2024)] = 112543,
  [SMALL_STATE(2025)] = 112606,
  [SMALL_STATE(2026)] = 112667,
  [SMALL_STATE(2027)] = 112726,
  [SMALL_STATE(2028)] = 112781,
  [SMALL_STATE(2029)] = 112836,
  [SMALL_STATE(2030)] = 112913,
  [SMALL_STATE(2031)] = 112988,
  [SMALL_STATE(2032)] = 113059,
  [SMALL_STATE(2033)] = 113136,
  [SMALL_STATE(2034)] = 113185,
  [SMALL_STATE(2035)] = 113260,
  [SMALL_STATE(2036)] = 113335,
  [SMALL_STATE(2037)] = 113382,
  [SMALL_STATE(2038)] = 113459,
  [SMALL_STATE(2039)] = 113530,
  [SMALL_STATE(2040)] = 113585,
  [SMALL_STATE(2041)] = 113660,
  [SMALL_STATE(2042)] = 113731,
  [SMALL_STATE(2043)] = 113780,
  [SMALL_STATE(2044)] = 113829,
  [SMALL_STATE(2045)] = 113906,
  [SMALL_STATE(2046)] = 113955,
  [SMALL_STATE(2047)] = 114004,
  [SMALL_STATE(2048)] = 114053,
  [SMALL_STATE(2049)] = 114128,
  [SMALL_STATE(2050)] = 114203,
  [SMALL_STATE(2051)] = 114258,
  [SMALL_STATE(2052)] = 114333,
  [SMALL_STATE(2053)] = 114408,
  [SMALL_STATE(2054)] = 114457,
  [SMALL_STATE(2055)] = 114504,
  [SMALL_STATE(2056)] = 114553,
  [SMALL_STATE(2057)] = 114626,
  [SMALL_STATE(2058)] = 114703,
  [SMALL_STATE(2059)] = 114752,
  [SMALL_STATE(2060)] = 114801,
  [SMALL_STATE(2061)] = 114850,
  [SMALL_STATE(2062)] = 114905,
  [SMALL_STATE(2063)] = 114986,
  [SMALL_STATE(2064)] = 115035,
  [SMALL_STATE(2065)] = 115084,
  [SMALL_STATE(2066)] = 115133,
  [SMALL_STATE(2067)] = 115188,
  [SMALL_STATE(2068)] = 115263,
  [SMALL_STATE(2069)] = 115312,
  [SMALL_STATE(2070)] = 115361,
  [SMALL_STATE(2071)] = 115432,
  [SMALL_STATE(2072)] = 115507,
  [SMALL_STATE(2073)] = 115562,
  [SMALL_STATE(2074)] = 115635,
  [SMALL_STATE(2075)] = 115684,
  [SMALL_STATE(2076)] = 115733,
  [SMALL_STATE(2077)] = 115814,
  [SMALL_STATE(2078)] = 115895,
  [SMALL_STATE(2079)] = 115948,
  [SMALL_STATE(2080)] = 116023,
  [SMALL_STATE(2081)] = 116072,
  [SMALL_STATE(2082)] = 116121,
  [SMALL_STATE(2083)] = 116192,
  [SMALL_STATE(2084)] = 116265,
  [SMALL_STATE(2085)] = 116318,
  [SMALL_STATE(2086)] = 116381,
  [SMALL_STATE(2087)] = 116442,
  [SMALL_STATE(2088)] = 116501,
  [SMALL_STATE(2089)] = 116556,
  [SMALL_STATE(2090)] = 116611,
  [SMALL_STATE(2091)] = 116688,
  [SMALL_STATE(2092)] = 116763,
  [SMALL_STATE(2093)] = 116834,
  [SMALL_STATE(2094)] = 116911,
  [SMALL_STATE(2095)] = 116986,
  [SMALL_STATE(2096)] = 117061,
  [SMALL_STATE(2097)] = 117136,
  [SMALL_STATE(2098)] = 117213,
  [SMALL_STATE(2099)] = 117284,
  [SMALL_STATE(2100)] = 117331,
  [SMALL_STATE(2101)] = 117380,
  [SMALL_STATE(2102)] = 117427,
  [SMALL_STATE(2103)] = 117474,
  [SMALL_STATE(2104)] = 117549,
  [SMALL_STATE(2105)] = 117620,
  [SMALL_STATE(2106)] = 117667,
  [SMALL_STATE(2107)] = 117716,
  [SMALL_STATE(2108)] = 117793,
  [SMALL_STATE(2109)] = 117864,
  [SMALL_STATE(2110)] = 117913,
  [SMALL_STATE(2111)] = 117984,
  [SMALL_STATE(2112)] = 118031,
  [SMALL_STATE(2113)] = 118102,
  [SMALL_STATE(2114)] = 118149,
  [SMALL_STATE(2115)] = 118222,
  [SMALL_STATE(2116)] = 118293,
  [SMALL_STATE(2117)] = 118340,
  [SMALL_STATE(2118)] = 118417,
  [SMALL_STATE(2119)] = 118492,
  [SMALL_STATE(2120)] = 118547,
  [SMALL_STATE(2121)] = 118622,
  [SMALL_STATE(2122)] = 118693,
  [SMALL_STATE(2123)] = 118746,
  [SMALL_STATE(2124)] = 118795,
  [SMALL_STATE(2125)] = 118848,
  [SMALL_STATE(2126)] = 118897,
  [SMALL_STATE(2127)] = 118968,
  [SMALL_STATE(2128)] = 119031,
  [SMALL_STATE(2129)] = 119092,
  [SMALL_STATE(2130)] = 119167,
  [SMALL_STATE(2131)] = 119226,
  [SMALL_STATE(2132)] = 119299,
  [SMALL_STATE(2133)] = 119352,
  [SMALL_STATE(2134)] = 119407,
  [SMALL_STATE(2135)] = 119456,
  [SMALL_STATE(2136)] = 119503,
  [SMALL_STATE(2137)] = 119550,
  [SMALL_STATE(2138)] = 119613,
  [SMALL_STATE(2139)] = 119674,
  [SMALL_STATE(2140)] = 119721,
  [SMALL_STATE(2141)] = 119768,
  [SMALL_STATE(2142)] = 119815,
  [SMALL_STATE(2143)] = 119862,
  [SMALL_STATE(2144)] = 119909,
  [SMALL_STATE(2145)] = 119956,
  [SMALL_STATE(2146)] = 120011,
  [SMALL_STATE(2147)] = 120088,
  [SMALL_STATE(2148)] = 120163,
  [SMALL_STATE(2149)] = 120238,
  [SMALL_STATE(2150)] = 120309,
  [SMALL_STATE(2151)] = 120372,
  [SMALL_STATE(2152)] = 120431,
  [SMALL_STATE(2153)] = 120508,
  [SMALL_STATE(2154)] = 120585,
  [SMALL_STATE(2155)] = 120660,
  [SMALL_STATE(2156)] = 120735,
  [SMALL_STATE(2157)] = 120810,
  [SMALL_STATE(2158)] = 120865,
  [SMALL_STATE(2159)] = 120914,
  [SMALL_STATE(2160)] = 120989,
  [SMALL_STATE(2161)] = 121038,
  [SMALL_STATE(2162)] = 121084,
  [SMALL_STATE(2163)] = 121132,
  [SMALL_STATE(2164)] = 121208,
  [SMALL_STATE(2165)] = 121256,
  [SMALL_STATE(2166)] = 121310,
  [SMALL_STATE(2167)] = 121358,
  [SMALL_STATE(2168)] = 121432,
  [SMALL_STATE(2169)] = 121502,
  [SMALL_STATE(2170)] = 121546,
  [SMALL_STATE(2171)] = 121590,
  [SMALL_STATE(2172)] = 121634,
  [SMALL_STATE(2173)] = 121708,
  [SMALL_STATE(2174)] = 121754,
  [SMALL_STATE(2175)] = 121808,
  [SMALL_STATE(2176)] = 121852,
  [SMALL_STATE(2177)] = 121896,
  [SMALL_STATE(2178)] = 121940,
  [SMALL_STATE(2179)] = 121986,
  [SMALL_STATE(2180)] = 122030,
  [SMALL_STATE(2181)] = 122104,
  [SMALL_STATE(2182)] = 122178,
  [SMALL_STATE(2183)] = 122226,
  [SMALL_STATE(2184)] = 122274,
  [SMALL_STATE(2185)] = 122344,
  [SMALL_STATE(2186)] = 122416,
  [SMALL_STATE(2187)] = 122468,
  [SMALL_STATE(2188)] = 122530,
  [SMALL_STATE(2189)] = 122590,
  [SMALL_STATE(2190)] = 122638,
  [SMALL_STATE(2191)] = 122712,
  [SMALL_STATE(2192)] = 122770,
  [SMALL_STATE(2193)] = 122824,
  [SMALL_STATE(2194)] = 122878,
  [SMALL_STATE(2195)] = 122932,
  [SMALL_STATE(2196)] = 122976,
  [SMALL_STATE(2197)] = 123050,
  [SMALL_STATE(2198)] = 123098,
  [SMALL_STATE(2199)] = 123146,
  [SMALL_STATE(2200)] = 123216,
  [SMALL_STATE(2201)] = 123288,
  [SMALL_STATE(2202)] = 123340,
  [SMALL_STATE(2203)] = 123402,
  [SMALL_STATE(2204)] = 123462,
  [SMALL_STATE(2205)] = 123520,
  [SMALL_STATE(2206)] = 123574,
  [SMALL_STATE(2207)] = 123628,
  [SMALL_STATE(2208)] = 123704,
  [SMALL_STATE(2209)] = 123778,
  [SMALL_STATE(2210)] = 123848,
  [SMALL_STATE(2211)] = 123924,
  [SMALL_STATE(2212)] = 123998,
  [SMALL_STATE(2213)] = 124042,
  [SMALL_STATE(2214)] = 124116,
  [SMALL_STATE(2215)] = 124190,
  [SMALL_STATE(2216)] = 124234,
  [SMALL_STATE(2217)] = 124278,
  [SMALL_STATE(2218)] = 124322,
  [SMALL_STATE(2219)] = 124398,
  [SMALL_STATE(2220)] = 124442,
  [SMALL_STATE(2221)] = 124486,
  [SMALL_STATE(2222)] = 124530,
  [SMALL_STATE(2223)] = 124574,
  [SMALL_STATE(2224)] = 124618,
  [SMALL_STATE(2225)] = 124666,
  [SMALL_STATE(2226)] = 124714,
  [SMALL_STATE(2227)] = 124798,
  [SMALL_STATE(2228)] = 124872,
  [SMALL_STATE(2229)] = 124916,
  [SMALL_STATE(2230)] = 124960,
  [SMALL_STATE(2231)] = 125004,
  [SMALL_STATE(2232)] = 125048,
  [SMALL_STATE(2233)] = 125092,
  [SMALL_STATE(2234)] = 125136,
  [SMALL_STATE(2235)] = 125180,
  [SMALL_STATE(2236)] = 125228,
  [SMALL_STATE(2237)] = 125314,
  [SMALL_STATE(2238)] = 125362,
  [SMALL_STATE(2239)] = 125410,
  [SMALL_STATE(2240)] = 125454,
  [SMALL_STATE(2241)] = 125498,
  [SMALL_STATE(2242)] = 125542,
  [SMALL_STATE(2243)] = 125586,
  [SMALL_STATE(2244)] = 125630,
  [SMALL_STATE(2245)] = 125674,
  [SMALL_STATE(2246)] = 125718,
  [SMALL_STATE(2247)] = 125802,
  [SMALL_STATE(2248)] = 125846,
  [SMALL_STATE(2249)] = 125930,
  [SMALL_STATE(2250)] = 125974,
  [SMALL_STATE(2251)] = 126018,
  [SMALL_STATE(2252)] = 126088,
  [SMALL_STATE(2253)] = 126134,
  [SMALL_STATE(2254)] = 126178,
  [SMALL_STATE(2255)] = 126222,
  [SMALL_STATE(2256)] = 126300,
  [SMALL_STATE(2257)] = 126374,
  [SMALL_STATE(2258)] = 126448,
  [SMALL_STATE(2259)] = 126492,
  [SMALL_STATE(2260)] = 126562,
  [SMALL_STATE(2261)] = 126610,
  [SMALL_STATE(2262)] = 126658,
  [SMALL_STATE(2263)] = 126728,
  [SMALL_STATE(2264)] = 126798,
  [SMALL_STATE(2265)] = 126844,
  [SMALL_STATE(2266)] = 126888,
  [SMALL_STATE(2267)] = 126964,
  [SMALL_STATE(2268)] = 127008,
  [SMALL_STATE(2269)] = 127052,
  [SMALL_STATE(2270)] = 127124,
  [SMALL_STATE(2271)] = 127168,
  [SMALL_STATE(2272)] = 127212,
  [SMALL_STATE(2273)] = 127256,
  [SMALL_STATE(2274)] = 127300,
  [SMALL_STATE(2275)] = 127344,
  [SMALL_STATE(2276)] = 127430,
  [SMALL_STATE(2277)] = 127482,
  [SMALL_STATE(2278)] = 127544,
  [SMALL_STATE(2279)] = 127604,
  [SMALL_STATE(2280)] = 127662,
  [SMALL_STATE(2281)] = 127706,
  [SMALL_STATE(2282)] = 127752,
  [SMALL_STATE(2283)] = 127798,
  [SMALL_STATE(2284)] = 127844,
  [SMALL_STATE(2285)] = 127898,
  [SMALL_STATE(2286)] = 127942,
  [SMALL_STATE(2287)] = 128016,
  [SMALL_STATE(2288)] = 128102,
  [SMALL_STATE(2289)] = 128156,
  [SMALL_STATE(2290)] = 128200,
  [SMALL_STATE(2291)] = 128276,
  [SMALL_STATE(2292)] = 128350,
  [SMALL_STATE(2293)] = 128420,
  [SMALL_STATE(2294)] = 128466,
  [SMALL_STATE(2295)] = 128512,
  [SMALL_STATE(2296)] = 128596,
  [SMALL_STATE(2297)] = 128682,
  [SMALL_STATE(2298)] = 128747,
  [SMALL_STATE(2299)] = 128830,
  [SMALL_STATE(2300)] = 128907,
  [SMALL_STATE(2301)] = 128990,
  [SMALL_STATE(2302)] = 129073,
  [SMALL_STATE(2303)] = 129138,
  [SMALL_STATE(2304)] = 129213,
  [SMALL_STATE(2305)] = 129278,
  [SMALL_STATE(2306)] = 129359,
  [SMALL_STATE(2307)] = 129442,
  [SMALL_STATE(2308)] = 129507,
  [SMALL_STATE(2309)] = 129582,
  [SMALL_STATE(2310)] = 129647,
  [SMALL_STATE(2311)] = 129728,
  [SMALL_STATE(2312)] = 129793,
  [SMALL_STATE(2313)] = 129868,
  [SMALL_STATE(2314)] = 129951,
  [SMALL_STATE(2315)] = 130016,
  [SMALL_STATE(2316)] = 130097,
  [SMALL_STATE(2317)] = 130162,
  [SMALL_STATE(2318)] = 130227,
  [SMALL_STATE(2319)] = 130308,
  [SMALL_STATE(2320)] = 130389,
  [SMALL_STATE(2321)] = 130454,
  [SMALL_STATE(2322)] = 130537,
  [SMALL_STATE(2323)] = 130602,
  [SMALL_STATE(2324)] = 130683,
  [SMALL_STATE(2325)] = 130748,
  [SMALL_STATE(2326)] = 130831,
  [SMALL_STATE(2327)] = 130896,
  [SMALL_STATE(2328)] = 130979,
  [SMALL_STATE(2329)] = 131044,
  [SMALL_STATE(2330)] = 131109,
  [SMALL_STATE(2331)] = 131174,
  [SMALL_STATE(2332)] = 131239,
  [SMALL_STATE(2333)] = 131304,
  [SMALL_STATE(2334)] = 131369,
  [SMALL_STATE(2335)] = 131450,
  [SMALL_STATE(2336)] = 131531,
  [SMALL_STATE(2337)] = 131596,
  [SMALL_STATE(2338)] = 131661,
  [SMALL_STATE(2339)] = 131744,
  [SMALL_STATE(2340)] = 131822,
  [SMALL_STATE(2341)] = 131900,
  [SMALL_STATE(2342)] = 131980,
  [SMALL_STATE(2343)] = 132060,
  [SMALL_STATE(2344)] = 132138,
  [SMALL_STATE(2345)] = 132218,
  [SMALL_STATE(2346)] = 132298,
  [SMALL_STATE(2347)] = 132376,
  [SMALL_STATE(2348)] = 132452,
  [SMALL_STATE(2349)] = 132532,
  [SMALL_STATE(2350)] = 132612,
  [SMALL_STATE(2351)] = 132692,
  [SMALL_STATE(2352)] = 132769,
  [SMALL_STATE(2353)] = 132842,
  [SMALL_STATE(2354)] = 132919,
  [SMALL_STATE(2355)] = 132992,
  [SMALL_STATE(2356)] = 133065,
  [SMALL_STATE(2357)] = 133138,
  [SMALL_STATE(2358)] = 133211,
  [SMALL_STATE(2359)] = 133284,
  [SMALL_STATE(2360)] = 133359,
  [SMALL_STATE(2361)] = 133436,
  [SMALL_STATE(2362)] = 133513,
  [SMALL_STATE(2363)] = 133588,
  [SMALL_STATE(2364)] = 133631,
  [SMALL_STATE(2365)] = 133706,
  [SMALL_STATE(2366)] = 133778,
  [SMALL_STATE(2367)] = 133850,
  [SMALL_STATE(2368)] = 133920,
  [SMALL_STATE(2369)] = 133992,
  [SMALL_STATE(2370)] = 134064,
  [SMALL_STATE(2371)] = 134136,
  [SMALL_STATE(2372)] = 134208,
  [SMALL_STATE(2373)] = 134280,
  [SMALL_STATE(2374)] = 134352,
  [SMALL_STATE(2375)] = 134426,
  [SMALL_STATE(2376)] = 134496,
  [SMALL_STATE(2377)] = 134568,
  [SMALL_STATE(2378)] = 134640,
  [SMALL_STATE(2379)] = 134710,
  [SMALL_STATE(2380)] = 134782,
  [SMALL_STATE(2381)] = 134854,
  [SMALL_STATE(2382)] = 134928,
  [SMALL_STATE(2383)] = 135000,
  [SMALL_STATE(2384)] = 135072,
  [SMALL_STATE(2385)] = 135144,
  [SMALL_STATE(2386)] = 135215,
  [SMALL_STATE(2387)] = 135286,
  [SMALL_STATE(2388)] = 135357,
  [SMALL_STATE(2389)] = 135428,
  [SMALL_STATE(2390)] = 135499,
  [SMALL_STATE(2391)] = 135570,
  [SMALL_STATE(2392)] = 135641,
  [SMALL_STATE(2393)] = 135712,
  [SMALL_STATE(2394)] = 135783,
  [SMALL_STATE(2395)] = 135854,
  [SMALL_STATE(2396)] = 135925,
  [SMALL_STATE(2397)] = 135996,
  [SMALL_STATE(2398)] = 136067,
  [SMALL_STATE(2399)] = 136138,
  [SMALL_STATE(2400)] = 136209,
  [SMALL_STATE(2401)] = 136280,
  [SMALL_STATE(2402)] = 136351,
  [SMALL_STATE(2403)] = 136422,
  [SMALL_STATE(2404)] = 136493,
  [SMALL_STATE(2405)] = 136564,
  [SMALL_STATE(2406)] = 136635,
  [SMALL_STATE(2407)] = 136706,
  [SMALL_STATE(2408)] = 136777,
  [SMALL_STATE(2409)] = 136848,
  [SMALL_STATE(2410)] = 136919,
  [SMALL_STATE(2411)] = 136990,
  [SMALL_STATE(2412)] = 137061,
  [SMALL_STATE(2413)] = 137132,
  [SMALL_STATE(2414)] = 137203,
  [SMALL_STATE(2415)] = 137274,
  [SMALL_STATE(2416)] = 137345,
  [SMALL_STATE(2417)] = 137416,
  [SMALL_STATE(2418)] = 137487,
  [SMALL_STATE(2419)] = 137558,
  [SMALL_STATE(2420)] = 137629,
  [SMALL_STATE(2421)] = 137700,
  [SMALL_STATE(2422)] = 137771,
  [SMALL_STATE(2423)] = 137842,
  [SMALL_STATE(2424)] = 137913,
  [SMALL_STATE(2425)] = 137984,
  [SMALL_STATE(2426)] = 138055,
  [SMALL_STATE(2427)] = 138126,
  [SMALL_STATE(2428)] = 138197,
  [SMALL_STATE(2429)] = 138268,
  [SMALL_STATE(2430)] = 138339,
  [SMALL_STATE(2431)] = 138410,
  [SMALL_STATE(2432)] = 138481,
  [SMALL_STATE(2433)] = 138552,
  [SMALL_STATE(2434)] = 138623,
  [SMALL_STATE(2435)] = 138694,
  [SMALL_STATE(2436)] = 138765,
  [SMALL_STATE(2437)] = 138836,
  [SMALL_STATE(2438)] = 138907,
  [SMALL_STATE(2439)] = 138978,
  [SMALL_STATE(2440)] = 139049,
  [SMALL_STATE(2441)] = 139120,
  [SMALL_STATE(2442)] = 139191,
  [SMALL_STATE(2443)] = 139262,
  [SMALL_STATE(2444)] = 139333,
  [SMALL_STATE(2445)] = 139404,
  [SMALL_STATE(2446)] = 139475,
  [SMALL_STATE(2447)] = 139546,
  [SMALL_STATE(2448)] = 139617,
  [SMALL_STATE(2449)] = 139688,
  [SMALL_STATE(2450)] = 139759,
  [SMALL_STATE(2451)] = 139830,
  [SMALL_STATE(2452)] = 139901,
  [SMALL_STATE(2453)] = 139972,
  [SMALL_STATE(2454)] = 140043,
  [SMALL_STATE(2455)] = 140114,
  [SMALL_STATE(2456)] = 140185,
  [SMALL_STATE(2457)] = 140256,
  [SMALL_STATE(2458)] = 140327,
  [SMALL_STATE(2459)] = 140398,
  [SMALL_STATE(2460)] = 140469,
  [SMALL_STATE(2461)] = 140540,
  [SMALL_STATE(2462)] = 140611,
  [SMALL_STATE(2463)] = 140682,
  [SMALL_STATE(2464)] = 140753,
  [SMALL_STATE(2465)] = 140824,
  [SMALL_STATE(2466)] = 140895,
  [SMALL_STATE(2467)] = 140966,
  [SMALL_STATE(2468)] = 141037,
  [SMALL_STATE(2469)] = 141108,
  [SMALL_STATE(2470)] = 141179,
  [SMALL_STATE(2471)] = 141250,
  [SMALL_STATE(2472)] = 141321,
  [SMALL_STATE(2473)] = 141392,
  [SMALL_STATE(2474)] = 141463,
  [SMALL_STATE(2475)] = 141534,
  [SMALL_STATE(2476)] = 141605,
  [SMALL_STATE(2477)] = 141676,
  [SMALL_STATE(2478)] = 141747,
  [SMALL_STATE(2479)] = 141818,
  [SMALL_STATE(2480)] = 141889,
  [SMALL_STATE(2481)] = 141960,
  [SMALL_STATE(2482)] = 142031,
  [SMALL_STATE(2483)] = 142102,
  [SMALL_STATE(2484)] = 142173,
  [SMALL_STATE(2485)] = 142244,
  [SMALL_STATE(2486)] = 142315,
  [SMALL_STATE(2487)] = 142386,
  [SMALL_STATE(2488)] = 142457,
  [SMALL_STATE(2489)] = 142528,
  [SMALL_STATE(2490)] = 142599,
  [SMALL_STATE(2491)] = 142670,
  [SMALL_STATE(2492)] = 142741,
  [SMALL_STATE(2493)] = 142812,
  [SMALL_STATE(2494)] = 142883,
  [SMALL_STATE(2495)] = 142954,
  [SMALL_STATE(2496)] = 143025,
  [SMALL_STATE(2497)] = 143096,
  [SMALL_STATE(2498)] = 143167,
  [SMALL_STATE(2499)] = 143238,
  [SMALL_STATE(2500)] = 143309,
  [SMALL_STATE(2501)] = 143380,
  [SMALL_STATE(2502)] = 143451,
  [SMALL_STATE(2503)] = 143522,
  [SMALL_STATE(2504)] = 143593,
  [SMALL_STATE(2505)] = 143664,
  [SMALL_STATE(2506)] = 143735,
  [SMALL_STATE(2507)] = 143806,
  [SMALL_STATE(2508)] = 143877,
  [SMALL_STATE(2509)] = 143948,
  [SMALL_STATE(2510)] = 144019,
  [SMALL_STATE(2511)] = 144090,
  [SMALL_STATE(2512)] = 144161,
  [SMALL_STATE(2513)] = 144232,
  [SMALL_STATE(2514)] = 144261,
  [SMALL_STATE(2515)] = 144290,
  [SMALL_STATE(2516)] = 144319,
  [SMALL_STATE(2517)] = 144343,
  [SMALL_STATE(2518)] = 144367,
  [SMALL_STATE(2519)] = 144391,
  [SMALL_STATE(2520)] = 144415,
  [SMALL_STATE(2521)] = 144442,
  [SMALL_STATE(2522)] = 144469,
  [SMALL_STATE(2523)] = 144498,
  [SMALL_STATE(2524)] = 144527,
  [SMALL_STATE(2525)] = 144556,
  [SMALL_STATE(2526)] = 144585,
  [SMALL_STATE(2527)] = 144611,
  [SMALL_STATE(2528)] = 144637,
  [SMALL_STATE(2529)] = 144659,
  [SMALL_STATE(2530)] = 144681,
  [SMALL_STATE(2531)] = 144703,
  [SMALL_STATE(2532)] = 144725,
  [SMALL_STATE(2533)] = 144747,
  [SMALL_STATE(2534)] = 144769,
  [SMALL_STATE(2535)] = 144791,
  [SMALL_STATE(2536)] = 144813,
  [SMALL_STATE(2537)] = 144835,
  [SMALL_STATE(2538)] = 144857,
  [SMALL_STATE(2539)] = 144879,
  [SMALL_STATE(2540)] = 144901,
  [SMALL_STATE(2541)] = 144923,
  [SMALL_STATE(2542)] = 144945,
  [SMALL_STATE(2543)] = 144967,
  [SMALL_STATE(2544)] = 144989,
  [SMALL_STATE(2545)] = 145011,
  [SMALL_STATE(2546)] = 145033,
  [SMALL_STATE(2547)] = 145055,
  [SMALL_STATE(2548)] = 145077,
  [SMALL_STATE(2549)] = 145099,
  [SMALL_STATE(2550)] = 145121,
  [SMALL_STATE(2551)] = 145148,
  [SMALL_STATE(2552)] = 145172,
  [SMALL_STATE(2553)] = 145196,
  [SMALL_STATE(2554)] = 145221,
  [SMALL_STATE(2555)] = 145242,
  [SMALL_STATE(2556)] = 145267,
  [SMALL_STATE(2557)] = 145285,
  [SMALL_STATE(2558)] = 145303,
  [SMALL_STATE(2559)] = 145319,
  [SMALL_STATE(2560)] = 145335,
  [SMALL_STATE(2561)] = 145353,
  [SMALL_STATE(2562)] = 145375,
  [SMALL_STATE(2563)] = 145391,
  [SMALL_STATE(2564)] = 145403,
  [SMALL_STATE(2565)] = 145419,
  [SMALL_STATE(2566)] = 145435,
  [SMALL_STATE(2567)] = 145453,
  [SMALL_STATE(2568)] = 145475,
  [SMALL_STATE(2569)] = 145497,
  [SMALL_STATE(2570)] = 145513,
  [SMALL_STATE(2571)] = 145531,
  [SMALL_STATE(2572)] = 145547,
  [SMALL_STATE(2573)] = 145565,
  [SMALL_STATE(2574)] = 145583,
  [SMALL_STATE(2575)] = 145605,
  [SMALL_STATE(2576)] = 145617,
  [SMALL_STATE(2577)] = 145629,
  [SMALL_STATE(2578)] = 145647,
  [SMALL_STATE(2579)] = 145665,
  [SMALL_STATE(2580)] = 145683,
  [SMALL_STATE(2581)] = 145699,
  [SMALL_STATE(2582)] = 145717,
  [SMALL_STATE(2583)] = 145734,
  [SMALL_STATE(2584)] = 145749,
  [SMALL_STATE(2585)] = 145766,
  [SMALL_STATE(2586)] = 145783,
  [SMALL_STATE(2587)] = 145800,
  [SMALL_STATE(2588)] = 145811,
  [SMALL_STATE(2589)] = 145824,
  [SMALL_STATE(2590)] = 145843,
  [SMALL_STATE(2591)] = 145860,
  [SMALL_STATE(2592)] = 145871,
  [SMALL_STATE(2593)] = 145888,
  [SMALL_STATE(2594)] = 145901,
  [SMALL_STATE(2595)] = 145920,
  [SMALL_STATE(2596)] = 145937,
  [SMALL_STATE(2597)] = 145954,
  [SMALL_STATE(2598)] = 145971,
  [SMALL_STATE(2599)] = 145988,
  [SMALL_STATE(2600)] = 146005,
  [SMALL_STATE(2601)] = 146022,
  [SMALL_STATE(2602)] = 146039,
  [SMALL_STATE(2603)] = 146053,
  [SMALL_STATE(2604)] = 146067,
  [SMALL_STATE(2605)] = 146081,
  [SMALL_STATE(2606)] = 146095,
  [SMALL_STATE(2607)] = 146107,
  [SMALL_STATE(2608)] = 146121,
  [SMALL_STATE(2609)] = 146137,
  [SMALL_STATE(2610)] = 146151,
  [SMALL_STATE(2611)] = 146167,
  [SMALL_STATE(2612)] = 146177,
  [SMALL_STATE(2613)] = 146191,
  [SMALL_STATE(2614)] = 146205,
  [SMALL_STATE(2615)] = 146217,
  [SMALL_STATE(2616)] = 146231,
  [SMALL_STATE(2617)] = 146247,
  [SMALL_STATE(2618)] = 146257,
  [SMALL_STATE(2619)] = 146271,
  [SMALL_STATE(2620)] = 146285,
  [SMALL_STATE(2621)] = 146301,
  [SMALL_STATE(2622)] = 146315,
  [SMALL_STATE(2623)] = 146329,
  [SMALL_STATE(2624)] = 146343,
  [SMALL_STATE(2625)] = 146357,
  [SMALL_STATE(2626)] = 146373,
  [SMALL_STATE(2627)] = 146387,
  [SMALL_STATE(2628)] = 146401,
  [SMALL_STATE(2629)] = 146417,
  [SMALL_STATE(2630)] = 146431,
  [SMALL_STATE(2631)] = 146447,
  [SMALL_STATE(2632)] = 146461,
  [SMALL_STATE(2633)] = 146475,
  [SMALL_STATE(2634)] = 146489,
  [SMALL_STATE(2635)] = 146503,
  [SMALL_STATE(2636)] = 146517,
  [SMALL_STATE(2637)] = 146530,
  [SMALL_STATE(2638)] = 146543,
  [SMALL_STATE(2639)] = 146556,
  [SMALL_STATE(2640)] = 146569,
  [SMALL_STATE(2641)] = 146582,
  [SMALL_STATE(2642)] = 146595,
  [SMALL_STATE(2643)] = 146608,
  [SMALL_STATE(2644)] = 146621,
  [SMALL_STATE(2645)] = 146634,
  [SMALL_STATE(2646)] = 146647,
  [SMALL_STATE(2647)] = 146660,
  [SMALL_STATE(2648)] = 146673,
  [SMALL_STATE(2649)] = 146682,
  [SMALL_STATE(2650)] = 146695,
  [SMALL_STATE(2651)] = 146708,
  [SMALL_STATE(2652)] = 146721,
  [SMALL_STATE(2653)] = 146734,
  [SMALL_STATE(2654)] = 146747,
  [SMALL_STATE(2655)] = 146758,
  [SMALL_STATE(2656)] = 146767,
  [SMALL_STATE(2657)] = 146776,
  [SMALL_STATE(2658)] = 146789,
  [SMALL_STATE(2659)] = 146798,
  [SMALL_STATE(2660)] = 146807,
  [SMALL_STATE(2661)] = 146820,
  [SMALL_STATE(2662)] = 146833,
  [SMALL_STATE(2663)] = 146846,
  [SMALL_STATE(2664)] = 146859,
  [SMALL_STATE(2665)] = 146872,
  [SMALL_STATE(2666)] = 146885,
  [SMALL_STATE(2667)] = 146898,
  [SMALL_STATE(2668)] = 146911,
  [SMALL_STATE(2669)] = 146924,
  [SMALL_STATE(2670)] = 146937,
  [SMALL_STATE(2671)] = 146946,
  [SMALL_STATE(2672)] = 146959,
  [SMALL_STATE(2673)] = 146972,
  [SMALL_STATE(2674)] = 146985,
  [SMALL_STATE(2675)] = 146998,
  [SMALL_STATE(2676)] = 147011,
  [SMALL_STATE(2677)] = 147024,
  [SMALL_STATE(2678)] = 147037,
  [SMALL_STATE(2679)] = 147050,
  [SMALL_STATE(2680)] = 147063,
  [SMALL_STATE(2681)] = 147076,
  [SMALL_STATE(2682)] = 147089,
  [SMALL_STATE(2683)] = 147102,
  [SMALL_STATE(2684)] = 147115,
  [SMALL_STATE(2685)] = 147124,
  [SMALL_STATE(2686)] = 147137,
  [SMALL_STATE(2687)] = 147150,
  [SMALL_STATE(2688)] = 147163,
  [SMALL_STATE(2689)] = 147176,
  [SMALL_STATE(2690)] = 147189,
  [SMALL_STATE(2691)] = 147202,
  [SMALL_STATE(2692)] = 147215,
  [SMALL_STATE(2693)] = 147228,
  [SMALL_STATE(2694)] = 147241,
  [SMALL_STATE(2695)] = 147254,
  [SMALL_STATE(2696)] = 147267,
  [SMALL_STATE(2697)] = 147280,
  [SMALL_STATE(2698)] = 147293,
  [SMALL_STATE(2699)] = 147306,
  [SMALL_STATE(2700)] = 147319,
  [SMALL_STATE(2701)] = 147332,
  [SMALL_STATE(2702)] = 147345,
  [SMALL_STATE(2703)] = 147358,
  [SMALL_STATE(2704)] = 147371,
  [SMALL_STATE(2705)] = 147384,
  [SMALL_STATE(2706)] = 147397,
  [SMALL_STATE(2707)] = 147410,
  [SMALL_STATE(2708)] = 147419,
  [SMALL_STATE(2709)] = 147432,
  [SMALL_STATE(2710)] = 147445,
  [SMALL_STATE(2711)] = 147458,
  [SMALL_STATE(2712)] = 147471,
  [SMALL_STATE(2713)] = 147484,
  [SMALL_STATE(2714)] = 147497,
  [SMALL_STATE(2715)] = 147510,
  [SMALL_STATE(2716)] = 147523,
  [SMALL_STATE(2717)] = 147536,
  [SMALL_STATE(2718)] = 147549,
  [SMALL_STATE(2719)] = 147562,
  [SMALL_STATE(2720)] = 147575,
  [SMALL_STATE(2721)] = 147588,
  [SMALL_STATE(2722)] = 147601,
  [SMALL_STATE(2723)] = 147610,
  [SMALL_STATE(2724)] = 147623,
  [SMALL_STATE(2725)] = 147632,
  [SMALL_STATE(2726)] = 147645,
  [SMALL_STATE(2727)] = 147658,
  [SMALL_STATE(2728)] = 147671,
  [SMALL_STATE(2729)] = 147684,
  [SMALL_STATE(2730)] = 147697,
  [SMALL_STATE(2731)] = 147708,
  [SMALL_STATE(2732)] = 147721,
  [SMALL_STATE(2733)] = 147734,
  [SMALL_STATE(2734)] = 147747,
  [SMALL_STATE(2735)] = 147760,
  [SMALL_STATE(2736)] = 147773,
  [SMALL_STATE(2737)] = 147786,
  [SMALL_STATE(2738)] = 147799,
  [SMALL_STATE(2739)] = 147812,
  [SMALL_STATE(2740)] = 147825,
  [SMALL_STATE(2741)] = 147838,
  [SMALL_STATE(2742)] = 147851,
  [SMALL_STATE(2743)] = 147864,
  [SMALL_STATE(2744)] = 147877,
  [SMALL_STATE(2745)] = 147890,
  [SMALL_STATE(2746)] = 147903,
  [SMALL_STATE(2747)] = 147916,
  [SMALL_STATE(2748)] = 147929,
  [SMALL_STATE(2749)] = 147942,
  [SMALL_STATE(2750)] = 147955,
  [SMALL_STATE(2751)] = 147968,
  [SMALL_STATE(2752)] = 147981,
  [SMALL_STATE(2753)] = 147994,
  [SMALL_STATE(2754)] = 148007,
  [SMALL_STATE(2755)] = 148020,
  [SMALL_STATE(2756)] = 148033,
  [SMALL_STATE(2757)] = 148042,
  [SMALL_STATE(2758)] = 148055,
  [SMALL_STATE(2759)] = 148068,
  [SMALL_STATE(2760)] = 148081,
  [SMALL_STATE(2761)] = 148094,
  [SMALL_STATE(2762)] = 148107,
  [SMALL_STATE(2763)] = 148120,
  [SMALL_STATE(2764)] = 148133,
  [SMALL_STATE(2765)] = 148146,
  [SMALL_STATE(2766)] = 148159,
  [SMALL_STATE(2767)] = 148172,
  [SMALL_STATE(2768)] = 148185,
  [SMALL_STATE(2769)] = 148194,
  [SMALL_STATE(2770)] = 148207,
  [SMALL_STATE(2771)] = 148220,
  [SMALL_STATE(2772)] = 148233,
  [SMALL_STATE(2773)] = 148246,
  [SMALL_STATE(2774)] = 148255,
  [SMALL_STATE(2775)] = 148268,
  [SMALL_STATE(2776)] = 148281,
  [SMALL_STATE(2777)] = 148294,
  [SMALL_STATE(2778)] = 148303,
  [SMALL_STATE(2779)] = 148316,
  [SMALL_STATE(2780)] = 148329,
  [SMALL_STATE(2781)] = 148342,
  [SMALL_STATE(2782)] = 148355,
  [SMALL_STATE(2783)] = 148368,
  [SMALL_STATE(2784)] = 148381,
  [SMALL_STATE(2785)] = 148394,
  [SMALL_STATE(2786)] = 148407,
  [SMALL_STATE(2787)] = 148418,
  [SMALL_STATE(2788)] = 148431,
  [SMALL_STATE(2789)] = 148444,
  [SMALL_STATE(2790)] = 148457,
  [SMALL_STATE(2791)] = 148470,
  [SMALL_STATE(2792)] = 148483,
  [SMALL_STATE(2793)] = 148496,
  [SMALL_STATE(2794)] = 148509,
  [SMALL_STATE(2795)] = 148519,
  [SMALL_STATE(2796)] = 148529,
  [SMALL_STATE(2797)] = 148539,
  [SMALL_STATE(2798)] = 148549,
  [SMALL_STATE(2799)] = 148559,
  [SMALL_STATE(2800)] = 148569,
  [SMALL_STATE(2801)] = 148579,
  [SMALL_STATE(2802)] = 148589,
  [SMALL_STATE(2803)] = 148599,
  [SMALL_STATE(2804)] = 148609,
  [SMALL_STATE(2805)] = 148617,
  [SMALL_STATE(2806)] = 148627,
  [SMALL_STATE(2807)] = 148637,
  [SMALL_STATE(2808)] = 148647,
  [SMALL_STATE(2809)] = 148657,
  [SMALL_STATE(2810)] = 148667,
  [SMALL_STATE(2811)] = 148677,
  [SMALL_STATE(2812)] = 148687,
  [SMALL_STATE(2813)] = 148697,
  [SMALL_STATE(2814)] = 148707,
  [SMALL_STATE(2815)] = 148717,
  [SMALL_STATE(2816)] = 148727,
  [SMALL_STATE(2817)] = 148737,
  [SMALL_STATE(2818)] = 148747,
  [SMALL_STATE(2819)] = 148757,
  [SMALL_STATE(2820)] = 148767,
  [SMALL_STATE(2821)] = 148777,
  [SMALL_STATE(2822)] = 148787,
  [SMALL_STATE(2823)] = 148797,
  [SMALL_STATE(2824)] = 148807,
  [SMALL_STATE(2825)] = 148817,
  [SMALL_STATE(2826)] = 148827,
  [SMALL_STATE(2827)] = 148837,
  [SMALL_STATE(2828)] = 148847,
  [SMALL_STATE(2829)] = 148857,
  [SMALL_STATE(2830)] = 148865,
  [SMALL_STATE(2831)] = 148875,
  [SMALL_STATE(2832)] = 148885,
  [SMALL_STATE(2833)] = 148895,
  [SMALL_STATE(2834)] = 148905,
  [SMALL_STATE(2835)] = 148915,
  [SMALL_STATE(2836)] = 148925,
  [SMALL_STATE(2837)] = 148935,
  [SMALL_STATE(2838)] = 148943,
  [SMALL_STATE(2839)] = 148953,
  [SMALL_STATE(2840)] = 148963,
  [SMALL_STATE(2841)] = 148973,
  [SMALL_STATE(2842)] = 148983,
  [SMALL_STATE(2843)] = 148991,
  [SMALL_STATE(2844)] = 149001,
  [SMALL_STATE(2845)] = 149011,
  [SMALL_STATE(2846)] = 149021,
  [SMALL_STATE(2847)] = 149031,
  [SMALL_STATE(2848)] = 149039,
  [SMALL_STATE(2849)] = 149047,
  [SMALL_STATE(2850)] = 149055,
  [SMALL_STATE(2851)] = 149063,
  [SMALL_STATE(2852)] = 149073,
  [SMALL_STATE(2853)] = 149083,
  [SMALL_STATE(2854)] = 149093,
  [SMALL_STATE(2855)] = 149103,
  [SMALL_STATE(2856)] = 149113,
  [SMALL_STATE(2857)] = 149123,
  [SMALL_STATE(2858)] = 149133,
  [SMALL_STATE(2859)] = 149143,
  [SMALL_STATE(2860)] = 149153,
  [SMALL_STATE(2861)] = 149163,
  [SMALL_STATE(2862)] = 149173,
  [SMALL_STATE(2863)] = 149183,
  [SMALL_STATE(2864)] = 149193,
  [SMALL_STATE(2865)] = 149203,
  [SMALL_STATE(2866)] = 149213,
  [SMALL_STATE(2867)] = 149223,
  [SMALL_STATE(2868)] = 149233,
  [SMALL_STATE(2869)] = 149243,
  [SMALL_STATE(2870)] = 149253,
  [SMALL_STATE(2871)] = 149263,
  [SMALL_STATE(2872)] = 149273,
  [SMALL_STATE(2873)] = 149281,
  [SMALL_STATE(2874)] = 149291,
  [SMALL_STATE(2875)] = 149301,
  [SMALL_STATE(2876)] = 149311,
  [SMALL_STATE(2877)] = 149321,
  [SMALL_STATE(2878)] = 149331,
  [SMALL_STATE(2879)] = 149341,
  [SMALL_STATE(2880)] = 149351,
  [SMALL_STATE(2881)] = 149361,
  [SMALL_STATE(2882)] = 149371,
  [SMALL_STATE(2883)] = 149381,
  [SMALL_STATE(2884)] = 149389,
  [SMALL_STATE(2885)] = 149397,
  [SMALL_STATE(2886)] = 149407,
  [SMALL_STATE(2887)] = 149417,
  [SMALL_STATE(2888)] = 149427,
  [SMALL_STATE(2889)] = 149435,
  [SMALL_STATE(2890)] = 149443,
  [SMALL_STATE(2891)] = 149453,
  [SMALL_STATE(2892)] = 149463,
  [SMALL_STATE(2893)] = 149473,
  [SMALL_STATE(2894)] = 149483,
  [SMALL_STATE(2895)] = 149493,
  [SMALL_STATE(2896)] = 149503,
  [SMALL_STATE(2897)] = 149513,
  [SMALL_STATE(2898)] = 149523,
  [SMALL_STATE(2899)] = 149531,
  [SMALL_STATE(2900)] = 149541,
  [SMALL_STATE(2901)] = 149551,
  [SMALL_STATE(2902)] = 149561,
  [SMALL_STATE(2903)] = 149571,
  [SMALL_STATE(2904)] = 149581,
  [SMALL_STATE(2905)] = 149591,
  [SMALL_STATE(2906)] = 149599,
  [SMALL_STATE(2907)] = 149609,
  [SMALL_STATE(2908)] = 149619,
  [SMALL_STATE(2909)] = 149629,
  [SMALL_STATE(2910)] = 149639,
  [SMALL_STATE(2911)] = 149649,
  [SMALL_STATE(2912)] = 149659,
  [SMALL_STATE(2913)] = 149669,
  [SMALL_STATE(2914)] = 149679,
  [SMALL_STATE(2915)] = 149689,
  [SMALL_STATE(2916)] = 149699,
  [SMALL_STATE(2917)] = 149709,
  [SMALL_STATE(2918)] = 149719,
  [SMALL_STATE(2919)] = 149729,
  [SMALL_STATE(2920)] = 149739,
  [SMALL_STATE(2921)] = 149749,
  [SMALL_STATE(2922)] = 149759,
  [SMALL_STATE(2923)] = 149769,
  [SMALL_STATE(2924)] = 149779,
  [SMALL_STATE(2925)] = 149789,
  [SMALL_STATE(2926)] = 149799,
  [SMALL_STATE(2927)] = 149809,
  [SMALL_STATE(2928)] = 149819,
  [SMALL_STATE(2929)] = 149829,
  [SMALL_STATE(2930)] = 149837,
  [SMALL_STATE(2931)] = 149847,
  [SMALL_STATE(2932)] = 149857,
  [SMALL_STATE(2933)] = 149867,
  [SMALL_STATE(2934)] = 149877,
  [SMALL_STATE(2935)] = 149887,
  [SMALL_STATE(2936)] = 149897,
  [SMALL_STATE(2937)] = 149907,
  [SMALL_STATE(2938)] = 149917,
  [SMALL_STATE(2939)] = 149927,
  [SMALL_STATE(2940)] = 149937,
  [SMALL_STATE(2941)] = 149947,
  [SMALL_STATE(2942)] = 149957,
  [SMALL_STATE(2943)] = 149967,
  [SMALL_STATE(2944)] = 149977,
  [SMALL_STATE(2945)] = 149987,
  [SMALL_STATE(2946)] = 149997,
  [SMALL_STATE(2947)] = 150007,
  [SMALL_STATE(2948)] = 150017,
  [SMALL_STATE(2949)] = 150027,
  [SMALL_STATE(2950)] = 150037,
  [SMALL_STATE(2951)] = 150047,
  [SMALL_STATE(2952)] = 150057,
  [SMALL_STATE(2953)] = 150067,
  [SMALL_STATE(2954)] = 150077,
  [SMALL_STATE(2955)] = 150087,
  [SMALL_STATE(2956)] = 150097,
  [SMALL_STATE(2957)] = 150107,
  [SMALL_STATE(2958)] = 150117,
  [SMALL_STATE(2959)] = 150127,
  [SMALL_STATE(2960)] = 150137,
  [SMALL_STATE(2961)] = 150147,
  [SMALL_STATE(2962)] = 150157,
  [SMALL_STATE(2963)] = 150167,
  [SMALL_STATE(2964)] = 150177,
  [SMALL_STATE(2965)] = 150187,
  [SMALL_STATE(2966)] = 150197,
  [SMALL_STATE(2967)] = 150207,
  [SMALL_STATE(2968)] = 150214,
  [SMALL_STATE(2969)] = 150221,
  [SMALL_STATE(2970)] = 150228,
  [SMALL_STATE(2971)] = 150235,
  [SMALL_STATE(2972)] = 150242,
  [SMALL_STATE(2973)] = 150249,
  [SMALL_STATE(2974)] = 150256,
  [SMALL_STATE(2975)] = 150263,
  [SMALL_STATE(2976)] = 150270,
  [SMALL_STATE(2977)] = 150277,
  [SMALL_STATE(2978)] = 150284,
  [SMALL_STATE(2979)] = 150291,
  [SMALL_STATE(2980)] = 150298,
  [SMALL_STATE(2981)] = 150305,
  [SMALL_STATE(2982)] = 150312,
  [SMALL_STATE(2983)] = 150319,
  [SMALL_STATE(2984)] = 150326,
  [SMALL_STATE(2985)] = 150333,
  [SMALL_STATE(2986)] = 150340,
  [SMALL_STATE(2987)] = 150347,
  [SMALL_STATE(2988)] = 150354,
  [SMALL_STATE(2989)] = 150361,
  [SMALL_STATE(2990)] = 150368,
  [SMALL_STATE(2991)] = 150375,
  [SMALL_STATE(2992)] = 150382,
  [SMALL_STATE(2993)] = 150389,
  [SMALL_STATE(2994)] = 150396,
  [SMALL_STATE(2995)] = 150403,
  [SMALL_STATE(2996)] = 150410,
  [SMALL_STATE(2997)] = 150417,
  [SMALL_STATE(2998)] = 150424,
  [SMALL_STATE(2999)] = 150431,
  [SMALL_STATE(3000)] = 150438,
  [SMALL_STATE(3001)] = 150445,
  [SMALL_STATE(3002)] = 150452,
  [SMALL_STATE(3003)] = 150459,
  [SMALL_STATE(3004)] = 150466,
  [SMALL_STATE(3005)] = 150473,
  [SMALL_STATE(3006)] = 150480,
  [SMALL_STATE(3007)] = 150487,
  [SMALL_STATE(3008)] = 150494,
  [SMALL_STATE(3009)] = 150501,
  [SMALL_STATE(3010)] = 150508,
  [SMALL_STATE(3011)] = 150515,
  [SMALL_STATE(3012)] = 150522,
  [SMALL_STATE(3013)] = 150529,
  [SMALL_STATE(3014)] = 150536,
  [SMALL_STATE(3015)] = 150543,
  [SMALL_STATE(3016)] = 150550,
  [SMALL_STATE(3017)] = 150557,
  [SMALL_STATE(3018)] = 150564,
  [SMALL_STATE(3019)] = 150571,
  [SMALL_STATE(3020)] = 150578,
  [SMALL_STATE(3021)] = 150585,
  [SMALL_STATE(3022)] = 150592,
  [SMALL_STATE(3023)] = 150599,
  [SMALL_STATE(3024)] = 150606,
  [SMALL_STATE(3025)] = 150613,
  [SMALL_STATE(3026)] = 150620,
  [SMALL_STATE(3027)] = 150627,
  [SMALL_STATE(3028)] = 150634,
  [SMALL_STATE(3029)] = 150641,
  [SMALL_STATE(3030)] = 150648,
  [SMALL_STATE(3031)] = 150655,
  [SMALL_STATE(3032)] = 150662,
  [SMALL_STATE(3033)] = 150669,
  [SMALL_STATE(3034)] = 150676,
  [SMALL_STATE(3035)] = 150683,
  [SMALL_STATE(3036)] = 150690,
  [SMALL_STATE(3037)] = 150697,
  [SMALL_STATE(3038)] = 150704,
  [SMALL_STATE(3039)] = 150711,
  [SMALL_STATE(3040)] = 150718,
  [SMALL_STATE(3041)] = 150725,
  [SMALL_STATE(3042)] = 150732,
  [SMALL_STATE(3043)] = 150739,
  [SMALL_STATE(3044)] = 150746,
  [SMALL_STATE(3045)] = 150753,
  [SMALL_STATE(3046)] = 150760,
  [SMALL_STATE(3047)] = 150767,
  [SMALL_STATE(3048)] = 150774,
  [SMALL_STATE(3049)] = 150781,
  [SMALL_STATE(3050)] = 150788,
  [SMALL_STATE(3051)] = 150795,
  [SMALL_STATE(3052)] = 150802,
  [SMALL_STATE(3053)] = 150809,
  [SMALL_STATE(3054)] = 150816,
  [SMALL_STATE(3055)] = 150823,
  [SMALL_STATE(3056)] = 150830,
  [SMALL_STATE(3057)] = 150837,
  [SMALL_STATE(3058)] = 150844,
  [SMALL_STATE(3059)] = 150851,
  [SMALL_STATE(3060)] = 150858,
  [SMALL_STATE(3061)] = 150865,
  [SMALL_STATE(3062)] = 150872,
  [SMALL_STATE(3063)] = 150879,
  [SMALL_STATE(3064)] = 150886,
  [SMALL_STATE(3065)] = 150893,
  [SMALL_STATE(3066)] = 150900,
  [SMALL_STATE(3067)] = 150907,
  [SMALL_STATE(3068)] = 150914,
  [SMALL_STATE(3069)] = 150921,
  [SMALL_STATE(3070)] = 150928,
  [SMALL_STATE(3071)] = 150935,
  [SMALL_STATE(3072)] = 150942,
  [SMALL_STATE(3073)] = 150949,
  [SMALL_STATE(3074)] = 150956,
  [SMALL_STATE(3075)] = 150963,
  [SMALL_STATE(3076)] = 150970,
  [SMALL_STATE(3077)] = 150977,
  [SMALL_STATE(3078)] = 150984,
  [SMALL_STATE(3079)] = 150991,
  [SMALL_STATE(3080)] = 150998,
  [SMALL_STATE(3081)] = 151005,
  [SMALL_STATE(3082)] = 151012,
  [SMALL_STATE(3083)] = 151019,
  [SMALL_STATE(3084)] = 151026,
  [SMALL_STATE(3085)] = 151033,
  [SMALL_STATE(3086)] = 151040,
  [SMALL_STATE(3087)] = 151047,
  [SMALL_STATE(3088)] = 151054,
  [SMALL_STATE(3089)] = 151061,
  [SMALL_STATE(3090)] = 151068,
  [SMALL_STATE(3091)] = 151075,
  [SMALL_STATE(3092)] = 151082,
  [SMALL_STATE(3093)] = 151089,
  [SMALL_STATE(3094)] = 151096,
  [SMALL_STATE(3095)] = 151103,
  [SMALL_STATE(3096)] = 151110,
  [SMALL_STATE(3097)] = 151117,
  [SMALL_STATE(3098)] = 151124,
  [SMALL_STATE(3099)] = 151131,
  [SMALL_STATE(3100)] = 151138,
  [SMALL_STATE(3101)] = 151145,
  [SMALL_STATE(3102)] = 151152,
  [SMALL_STATE(3103)] = 151159,
  [SMALL_STATE(3104)] = 151166,
  [SMALL_STATE(3105)] = 151173,
  [SMALL_STATE(3106)] = 151180,
  [SMALL_STATE(3107)] = 151187,
  [SMALL_STATE(3108)] = 151194,
  [SMALL_STATE(3109)] = 151201,
  [SMALL_STATE(3110)] = 151208,
  [SMALL_STATE(3111)] = 151215,
  [SMALL_STATE(3112)] = 151222,
  [SMALL_STATE(3113)] = 151229,
  [SMALL_STATE(3114)] = 151236,
  [SMALL_STATE(3115)] = 151243,
  [SMALL_STATE(3116)] = 151250,
  [SMALL_STATE(3117)] = 151257,
  [SMALL_STATE(3118)] = 151264,
  [SMALL_STATE(3119)] = 151271,
  [SMALL_STATE(3120)] = 151278,
  [SMALL_STATE(3121)] = 151285,
  [SMALL_STATE(3122)] = 151292,
  [SMALL_STATE(3123)] = 151299,
  [SMALL_STATE(3124)] = 151306,
  [SMALL_STATE(3125)] = 151313,
  [SMALL_STATE(3126)] = 151320,
  [SMALL_STATE(3127)] = 151327,
  [SMALL_STATE(3128)] = 151334,
  [SMALL_STATE(3129)] = 151341,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 0, 0, 0),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1527),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(828),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3093),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2700),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(759),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(779),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(915),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(888),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2837),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1230),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3027),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3086),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(791),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2781),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2782),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1244),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1765),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1216),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3090),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3081),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(912),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3065),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1231),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1233),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1242),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2302),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1243),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(756),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2176),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1746),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2176),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2596),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(237),
  [73] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_body, 2, 0, 0),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(42),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3126),
  [79] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2790),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(748),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(780),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(898),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(49),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(656),
  [91] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1261),
  [93] = {.entry = {.count = 1, .reusable = false}}, SHIFT(46),
  [95] = {.entry = {.count = 1, .reusable = false}}, SHIFT(991),
  [97] = {.entry = {.count = 1, .reusable = false}}, SHIFT(992),
  [99] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2333),
  [101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(993),
  [103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(754),
  [105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(641),
  [107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(450),
  [109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_body, 2, 0, 0),
  [111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(641),
  [113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2585),
  [115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_body, 3, 0, 0),
  [117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(234),
  [119] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_body, 3, 0, 0),
  [121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(41),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3124),
  [125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2783),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(742),
  [129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(772),
  [131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(905),
  [133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(61),
  [135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(618),
  [137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1257),
  [139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(45),
  [141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(978),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(979),
  [145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2337),
  [147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(980),
  [149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(739),
  [151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(646),
  [153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(458),
  [155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2582),
  [159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(911),
  [161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(72),
  [163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1166),
  [165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(57),
  [167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(890),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(71),
  [171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(928),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(62),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(783),
  [177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3101),
  [179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2713),
  [181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1434),
  [183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1766),
  [185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1435),
  [187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3127),
  [189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3080),
  [191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(908),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3082),
  [195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1582),
  [197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(614),
  [199] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 1, 0, 0),
  [201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [203] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0),
  [205] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1527),
  [208] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(828),
  [211] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3093),
  [214] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(2700),
  [217] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(759),
  [220] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(779),
  [223] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(915),
  [226] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(888),
  [229] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(2837),
  [232] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1230),
  [235] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3027),
  [238] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3086),
  [241] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(791),
  [244] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(2781),
  [247] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(2782),
  [250] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1244),
  [253] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1765),
  [256] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1216),
  [259] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3090),
  [262] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3081),
  [265] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(912),
  [268] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3065),
  [271] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1231),
  [274] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1233),
  [277] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1242),
  [280] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(2302),
  [283] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1243),
  [286] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(756),
  [289] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(2176),
  [292] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1746),
  [295] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(2176),
  [298] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(2596),
  [301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [305] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(783),
  [308] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3101),
  [311] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(2713),
  [314] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1434),
  [317] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1766),
  [320] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(1435),
  [323] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3127),
  [326] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3080),
  [329] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(908),
  [332] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2, 0, 0), SHIFT_REPEAT(3082),
  [335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1562),
  [337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1638),
  [341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1646),
  [343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2244),
  [345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2249),
  [347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1559),
  [349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1534),
  [351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1566),
  [359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1620),
  [361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1623),
  [363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1597),
  [367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1557),
  [369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1583),
  [377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1584),
  [379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1574),
  [381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 1, 0, 0),
  [383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 1, 0, 0),
  [385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1494),
  [387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(96),
  [389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3046),
  [391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2667),
  [393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(728),
  [395] = {.entry = {.count = 1, .reusable = false}}, SHIFT(778),
  [397] = {.entry = {.count = 1, .reusable = false}}, SHIFT(902),
  [399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(101),
  [401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1700),
  [403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(925),
  [405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(91),
  [407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(929),
  [409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(934),
  [411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2307),
  [413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(935),
  [415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(725),
  [417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1682),
  [419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1536),
  [421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1682),
  [423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2600),
  [425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield, 1, 0, 0),
  [427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield, 1, 0, 0),
  [429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(996),
  [431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1015),
  [433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_existential, 2, 0, 9),
  [435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_existential, 2, 0, 9),
  [437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(717),
  [439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(732),
  [441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3077),
  [443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(714),
  [445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(761),
  [447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3009),
  [449] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_raise_statement, 1, 0, 0),
  [451] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_raise_statement, 1, 0, 0),
  [453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(116),
  [455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3113),
  [457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2747),
  [459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(903),
  [461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(132),
  [463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1252),
  [465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(92),
  [467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1253),
  [469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1254),
  [471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2320),
  [473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1255),
  [475] = {.entry = {.count = 1, .reusable = false}}, SHIFT(936),
  [477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(103),
  [479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3114),
  [481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2751),
  [483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(741),
  [485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(768),
  [487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(906),
  [489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(128),
  [491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1142),
  [493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(93),
  [495] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1356),
  [497] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1357),
  [499] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2322),
  [501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1358),
  [503] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_list, 2, 0, 0),
  [505] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_list, 2, 0, 0),
  [507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(216),
  [509] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3123),
  [511] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2779),
  [513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(718),
  [515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(730),
  [517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1167),
  [519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1186),
  [521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2336),
  [523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1188),
  [525] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_list, 3, 0, 0),
  [527] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_list, 3, 0, 0),
  [529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1036),
  [531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(113),
  [533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2967),
  [535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2743),
  [537] = {.entry = {.count = 1, .reusable = false}}, SHIFT(896),
  [539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(127),
  [541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1120),
  [543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1304),
  [547] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1305),
  [549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2317),
  [551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1306),
  [553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1082),
  [555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(236),
  [557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3125),
  [559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2787),
  [561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(716),
  [563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(737),
  [565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(930),
  [567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(931),
  [569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2304),
  [571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(932),
  [573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(891),
  [575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(140),
  [577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(952),
  [579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(110),
  [581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(125),
  [583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3111),
  [585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2740),
  [587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(889),
  [589] = {.entry = {.count = 1, .reusable = false}}, SHIFT(152),
  [591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1095),
  [593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(111),
  [595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1053),
  [597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1054),
  [599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2316),
  [601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1055),
  [603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1513),
  [605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(131),
  [607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3118),
  [609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2762),
  [611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(735),
  [613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(770),
  [615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(900),
  [617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(144),
  [619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1168),
  [621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(100),
  [623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1121),
  [625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1122),
  [627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2329),
  [629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1123),
  [631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(744),
  [633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(776),
  [635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(738),
  [637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(777),
  [639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(213),
  [643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1351),
  [645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(121),
  [647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(139),
  [649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3087),
  [651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2671),
  [653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(901),
  [655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(182),
  [657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2243),
  [659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1300),
  [661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(130),
  [663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(914),
  [665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(180),
  [667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1448),
  [669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(133),
  [671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(897),
  [673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(172),
  [675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1363),
  [677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(135),
  [679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(146),
  [681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3122),
  [683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2775),
  [685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(910),
  [687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(193),
  [689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1307),
  [691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(126),
  [693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1382),
  [695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1383),
  [697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2297),
  [699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1384),
  [701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(192),
  [703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1455),
  [705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(134),
  [707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(918),
  [709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(240),
  [711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1281),
  [713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(148),
  [715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(231),
  [717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1333),
  [719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(895),
  [723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(232),
  [725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1044),
  [727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(142),
  [729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(956),
  [731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1099),
  [733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1335),
  [735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(715),
  [737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(734),
  [739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2971),
  [741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1262),
  [743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2178),
  [745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(747),
  [749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1193),
  [751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1056),
  [755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(938),
  [757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1489),
  [759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3103),
  [761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2717),
  [763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(721),
  [765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(749),
  [767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(953),
  [769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(954),
  [771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2311),
  [773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(955),
  [775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(722),
  [777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(726),
  [779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1326),
  [781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1365),
  [783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1209),
  [785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1213),
  [787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1219),
  [789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1172),
  [791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(713),
  [793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(751),
  [795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3007),
  [797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(753),
  [799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(773),
  [801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1364),
  [803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1505),
  [805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3117),
  [807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2760),
  [809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(760),
  [811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1096),
  [813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1097),
  [815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2328),
  [817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1098),
  [819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1124),
  [821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3119),
  [823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2766),
  [825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1143),
  [827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1144),
  [829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2330),
  [831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1145),
  [833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1146),
  [835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3121),
  [837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2771),
  [839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1308),
  [841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1309),
  [843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2332),
  [845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1310),
  [847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3120),
  [849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2769),
  [851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1169),
  [853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1170),
  [855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2331),
  [857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1171),
  [859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(867),
  [861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(13),
  [869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(731),
  [879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(769),
  [881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3115),
  [883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2754),
  [885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1282),
  [887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1283),
  [889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2324),
  [891] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1284),
  [893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1517),
  [897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2999),
  [899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2785),
  [901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(720),
  [903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(750),
  [905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1301),
  [907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1328),
  [909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2309),
  [911] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1334),
  [913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3116),
  [915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2757),
  [917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1352),
  [919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1353),
  [921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2326),
  [923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1354),
  [925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__primary_expression, 1, 0, 0),
  [927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__primary_expression, 1, 0, 0),
  [929] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pattern, 1, 0, 0),
  [931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 2, 0, 0),
  [933] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 2, 0, 0),
  [935] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_pattern, 2, 0, 0),
  [937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 2, 0, 0),
  [939] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 2, 0, 0),
  [941] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_pattern, 2, 0, 0),
  [943] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 3, 0, 20),
  [945] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, 0, 20),
  [947] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript, 4, 0, 40),
  [949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript, 4, 0, 40),
  [951] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_argument_list, 2, 0, 0), REDUCE(sym_tuple, 2, 0, 0),
  [954] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_tuple_pattern, 2, 0, 0), REDUCE(sym_argument_list, 2, 0, 0), REDUCE(sym_tuple, 2, 0, 0),
  [958] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_argument_list, 2, 0, 0), REDUCE(sym_tuple, 2, 0, 0),
  [961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(736),
  [963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(774),
  [965] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_list_pattern, 2, 0, 0), REDUCE(sym_list, 2, 0, 0),
  [968] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_pattern, 1, 0, 0), REDUCE(sym__primary_expression, 1, 0, 0),
  [971] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_tuple_pattern, 2, 0, 0), REDUCE(sym_tuple, 2, 0, 0),
  [974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1619),
  [976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1610),
  [978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2773),
  [980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2655),
  [982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1539),
  [984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1569),
  [988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1613),
  [990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1652),
  [992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1551),
  [994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
  [996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1560),
  [998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1589),
  [1000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [1002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1549),
  [1004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [1006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1594),
  [1008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1595),
  [1010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2015),
  [1012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [1014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2042),
  [1016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1720),
  [1018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [1020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [1022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(611),
  [1024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [1026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2162),
  [1028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1916),
  [1030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1791),
  [1032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1962),
  [1034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(544),
  [1036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(550),
  [1038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(794),
  [1040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(899),
  [1042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(871),
  [1044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2562),
  [1046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1052),
  [1048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(771),
  [1050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2237),
  [1052] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1786),
  [1054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2064),
  [1056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1795),
  [1058] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1884),
  [1060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2224),
  [1062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2058),
  [1064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2053),
  [1066] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2045),
  [1068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concatenated_string, 2, 0, 0),
  [1070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_concatenated_string, 2, 0, 0),
  [1072] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_delete_statement, 2, 0, 0),
  [1074] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(58),
  [1077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1016),
  [1079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1017),
  [1081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1018),
  [1083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3053),
  [1085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1019),
  [1087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1020),
  [1089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1021),
  [1091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1022),
  [1093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1023),
  [1095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1024),
  [1097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1025),
  [1099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1026),
  [1101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_delete_statement, 2, 0, 0),
  [1103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__assign_rhs, 1, 0, 0),
  [1105] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__assign_rhs, 1, 0, 0),
  [1107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(54),
  [1109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(997),
  [1111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(998),
  [1113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [1115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3058),
  [1117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1000),
  [1119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1001),
  [1121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1002),
  [1123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1003),
  [1125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1004),
  [1127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1005),
  [1129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1006),
  [1131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1007),
  [1133] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield, 2, 0, 0),
  [1135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(58),
  [1137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield, 2, 0, 0),
  [1139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0),
  [1141] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0),
  [1143] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(442),
  [1146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2582),
  [1149] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(54),
  [1152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1156] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(446),
  [1159] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2585),
  [1162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_statement, 1, 0, 0),
  [1164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_statement, 1, 0, 0),
  [1166] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__primary_expression, 1, 0, 0), SHIFT(457),
  [1169] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__primary_expression, 1, 0, 0), SHIFT(2585),
  [1172] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(54),
  [1175] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(54),
  [1178] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(58),
  [1181] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__primary_expression, 1, 0, 0), SHIFT(438),
  [1184] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__primary_expression, 1, 0, 0), SHIFT(2582),
  [1187] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(58),
  [1190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 3, 0, 0),
  [1192] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 3, 0, 0),
  [1194] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_await, 2, 0, 0),
  [1196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_await, 2, 0, 0),
  [1198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [1200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [1202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(740),
  [1204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(876),
  [1206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2977),
  [1208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(47),
  [1210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_new_expression, 2, 0, 7),
  [1212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_new_expression, 2, 0, 7),
  [1214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(755),
  [1216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(877),
  [1218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3012),
  [1220] = {.entry = {.count = 1, .reusable = false}}, SHIFT(48),
  [1222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 18),
  [1224] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 18),
  [1226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_operator, 3, 0, 17),
  [1228] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_operator, 3, 0, 17),
  [1230] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__suite, 1, 0, 0),
  [1232] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__anon_suite, 1, 0, 0),
  [1234] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__suite, 1, 0, 0), REDUCE(sym__anon_suite, 1, 0, 0),
  [1237] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__anon_suite, 1, 0, 0),
  [1239] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__suite, 1, 0, 0),
  [1241] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean_operator, 3, 0, 17),
  [1243] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean_operator, 3, 0, 17),
  [1245] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19),
  [1247] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19),
  [1249] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(997),
  [1252] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3058),
  [1255] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1007),
  [1258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_existential, 3, 0, 21),
  [1260] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_existential, 3, 0, 21),
  [1262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1425),
  [1264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 3, 0, 39),
  [1266] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 3, 0, 39),
  [1268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_not_operator, 2, 0, 5),
  [1270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_not_operator, 2, 0, 5),
  [1272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 5, 0, 0),
  [1274] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_conditional_expression, 5, 0, 0),
  [1276] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1016),
  [1279] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3053),
  [1282] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1026),
  [1285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1431),
  [1287] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_statement, 2, 0, 0),
  [1289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1094),
  [1291] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_statement, 2, 0, 0),
  [1293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(56),
  [1295] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_raise_statement, 2, 0, 0),
  [1297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_raise_statement, 2, 0, 0),
  [1299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield, 3, 0, 12),
  [1301] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield, 3, 0, 12),
  [1303] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_statement, 4, 0, 0),
  [1305] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_statement, 4, 0, 0),
  [1307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1008),
  [1309] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0),
  [1311] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1074),
  [1314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0),
  [1316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1027),
  [1318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__block_anonymous_function, 6, 0, 60),
  [1320] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__anon_suite, 1, 0, 0), REDUCE(sym__block_anonymous_function, 6, 0, 60),
  [1323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__block_anonymous_function, 6, 0, 60),
  [1325] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__block_anonymous_function, 7, 0, 71),
  [1327] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__block_anonymous_function, 7, 0, 71),
  [1329] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__anon_suite, 1, 0, 0), REDUCE(sym__block_anonymous_function, 7, 0, 71),
  [1332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__block_anonymous_function, 4, 0, 24),
  [1334] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__anon_suite, 1, 0, 0), REDUCE(sym__block_anonymous_function, 4, 0, 24),
  [1337] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__block_anonymous_function, 4, 0, 24),
  [1339] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1013),
  [1342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 2, 0, 0),
  [1344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 2, 0, 0),
  [1346] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__block_anonymous_function, 5, 0, 44),
  [1348] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__anon_suite, 1, 0, 0), REDUCE(sym__block_anonymous_function, 5, 0, 44),
  [1351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__block_anonymous_function, 5, 0, 44),
  [1353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1049),
  [1355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(60),
  [1357] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comparison_operator, 2, 0, 8),
  [1359] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_comparison_operator, 2, 0, 8),
  [1361] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_body, 1, 0, 0),
  [1363] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_body, 1, 0, 0),
  [1365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4),
  [1367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3),
  [1369] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0),
  [1371] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0),
  [1373] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(499),
  [1376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2),
  [1378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5),
  [1380] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(505),
  [1383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_operator, 2, 0, 6),
  [1385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_operator, 2, 0, 6),
  [1387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 8, 0, 75),
  [1389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 8, 0, 75),
  [1391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 2, 0, 0),
  [1393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 2, 0, 0),
  [1395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_existential, 3, 0, 22),
  [1397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_existential, 3, 0, 22),
  [1399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 4, 0, 24),
  [1401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 4, 0, 24),
  [1403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8),
  [1405] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call, 2, 0, 10),
  [1407] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call, 2, 0, 10),
  [1409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generator_expression, 4, 0, 25),
  [1411] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generator_expression, 4, 0, 25),
  [1413] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 4, 0, 0),
  [1415] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 4, 0, 0),
  [1417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1083),
  [1419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1084),
  [1421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1085),
  [1423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3062),
  [1425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1086),
  [1427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1087),
  [1429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1088),
  [1431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1089),
  [1433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1090),
  [1435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1091),
  [1437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1092),
  [1439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1093),
  [1441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_comprehension, 4, 0, 25),
  [1443] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_comprehension, 4, 0, 25),
  [1445] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 4, 0, 0),
  [1447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 4, 0, 0),
  [1449] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary, 2, 0, 0),
  [1451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dictionary, 2, 0, 0),
  [1453] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1037),
  [1456] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3041),
  [1459] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1048),
  [1462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_comprehension, 4, 0, 25),
  [1464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_comprehension, 4, 0, 25),
  [1466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary, 4, 0, 0),
  [1468] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dictionary, 4, 0, 0),
  [1470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set, 4, 0, 0),
  [1472] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set, 4, 0, 0),
  [1474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_existential, 4, 0, 42),
  [1476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_existential, 4, 0, 42),
  [1478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary_comprehension, 4, 0, 25),
  [1480] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dictionary_comprehension, 4, 0, 25),
  [1482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 3, 0, 0),
  [1484] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 3, 0, 0),
  [1486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1037),
  [1488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1038),
  [1490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1039),
  [1492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3041),
  [1494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1040),
  [1496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1041),
  [1498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1042),
  [1500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1043),
  [1502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1045),
  [1504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1046),
  [1506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1047),
  [1508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1048),
  [1510] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 5, 0, 44),
  [1512] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 5, 0, 44),
  [1514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 5, 0, 47),
  [1516] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 5, 0, 47),
  [1518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1422),
  [1520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6),
  [1522] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 5, 0, 0),
  [1524] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 5, 0, 0),
  [1526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 5, 0, 0),
  [1528] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 5, 0, 0),
  [1530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary, 5, 0, 0),
  [1532] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dictionary, 5, 0, 0),
  [1534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set, 5, 0, 0),
  [1536] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set, 5, 0, 0),
  [1538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 4, 0, 25),
  [1540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 4, 0, 25),
  [1542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_f_string, 3, 0, 0),
  [1544] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_f_string, 3, 0, 0),
  [1546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 4, 0, 0),
  [1548] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 4, 0, 0),
  [1550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(9),
  [1552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_existential, 5, 0, 56),
  [1554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_existential, 5, 0, 56),
  [1556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 6, 0, 58),
  [1558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 6, 0, 58),
  [1560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 6, 0, 60),
  [1562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 6, 0, 60),
  [1564] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(519),
  [1567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7),
  [1569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 5, 0, 0),
  [1571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 5, 0, 0),
  [1573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 7, 0, 71),
  [1575] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 7, 0, 71),
  [1577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 7, 0, 73),
  [1579] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 7, 0, 73),
  [1581] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1083),
  [1584] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3062),
  [1587] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1093),
  [1590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [1592] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [1594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield_statement, 1, 0, 0),
  [1596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_statement, 1, 0, 0),
  [1598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(63),
  [1600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1428),
  [1602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_f_string, 2, 0, 0),
  [1604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_f_string, 2, 0, 0),
  [1606] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(536),
  [1609] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 3, 0, 0),
  [1611] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 3, 0, 0),
  [1613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(55),
  [1615] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary, 3, 0, 0),
  [1617] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dictionary, 3, 0, 0),
  [1619] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set, 3, 0, 0),
  [1621] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set, 3, 0, 0),
  [1623] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_argument_list, 4, 0, 25), REDUCE(sym_generator_expression, 4, 0, 25),
  [1626] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_argument_list, 4, 0, 25), REDUCE(sym_generator_expression, 4, 0, 25),
  [1629] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment, 3, 0, 16),
  [1631] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment, 3, 0, 16),
  [1633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_augmented_assignment, 3, 0, 17),
  [1635] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_augmented_assignment, 3, 0, 17),
  [1637] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_statement, 6, 0, 69),
  [1639] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_statement, 6, 0, 69),
  [1641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1507),
  [1643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1516),
  [1645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(879),
  [1647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1360),
  [1649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1436),
  [1651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(222),
  [1653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1078),
  [1655] = {.entry = {.count = 1, .reusable = false}}, SHIFT(228),
  [1657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(223),
  [1659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(229),
  [1661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1554),
  [1663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1521),
  [1665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1492),
  [1667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1522),
  [1669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1526),
  [1671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1663),
  [1673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2513),
  [1675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1192),
  [1677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1234),
  [1679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1664),
  [1681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(884),
  [1683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1511),
  [1685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1694),
  [1687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1495),
  [1689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2233),
  [1691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(218),
  [1693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1506),
  [1695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(239),
  [1697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1704),
  [1699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1496),
  [1701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1510),
  [1703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1519),
  [1705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(224),
  [1707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(217),
  [1709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(572),
  [1711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(560),
  [1713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(939),
  [1715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(235),
  [1717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2273),
  [1719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(226),
  [1721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(608),
  [1723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(624),
  [1725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(940),
  [1727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(238),
  [1729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1491),
  [1731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1520),
  [1733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1524),
  [1735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2175),
  [1737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1490),
  [1739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(669),
  [1741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(593),
  [1743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2247),
  [1745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(643),
  [1747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(685),
  [1749] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1523),
  [1751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1500),
  [1753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(233),
  [1755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2222),
  [1757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(684),
  [1759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1703),
  [1761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(674),
  [1763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3028),
  [1765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2753),
  [1767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1256),
  [1769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3102),
  [1771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2715),
  [1773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2515),
  [1775] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2254),
  [1777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1692),
  [1779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1212),
  [1781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1702),
  [1783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2285),
  [1785] = {.entry = {.count = 1, .reusable = false}}, SHIFT(687),
  [1787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2229),
  [1789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(568),
  [1791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(571),
  [1793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2267),
  [1795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(576),
  [1797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1670),
  [1799] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1671),
  [1801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(577),
  [1803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(599),
  [1805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(600),
  [1807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(601),
  [1809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(603),
  [1811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1677),
  [1813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1687),
  [1815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2270),
  [1817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1696),
  [1819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1699),
  [1821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [1823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1660),
  [1825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(620),
  [1827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(658),
  [1829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(686),
  [1831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(693),
  [1833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(626),
  [1835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(631),
  [1837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(634),
  [1839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(638),
  [1841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2230),
  [1843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2232),
  [1845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2231),
  [1847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2223),
  [1849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(719),
  [1851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(941),
  [1853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(165),
  [1855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2976),
  [1857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(894),
  [1859] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice, 1, 0, 0),
  [1861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(215),
  [1863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3064),
  [1865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(913),
  [1867] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice, 2, 0, 0),
  [1869] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice, 3, 0, 0),
  [1871] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice, 4, 0, 0),
  [1873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern, 1, 0, 0),
  [1875] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_pattern, 1, 0, 0), REDUCE(sym__primary_expression, 1, 0, 0),
  [1878] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [1880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 2, 0, 0),
  [1882] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_tuple_pattern, 2, 0, 0), REDUCE(sym_tuple, 2, 0, 0),
  [1885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 2, 0, 0),
  [1887] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_list_pattern, 2, 0, 0), REDUCE(sym_list, 2, 0, 0),
  [1890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(988),
  [1892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(969),
  [1894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1032),
  [1896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1011),
  [1898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1029),
  [1900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1051),
  [1902] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1071),
  [1904] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_tuple_pattern, 2, 0, 0), REDUCE(sym_argument_list, 2, 0, 0), REDUCE(sym_tuple, 2, 0, 0),
  [1908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1275),
  [1910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1117),
  [1912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1112),
  [1914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1137),
  [1916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1159),
  [1918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1185),
  [1920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1206),
  [1922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1239),
  [1924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1278),
  [1926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1297),
  [1928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1323),
  [1930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1349),
  [1932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1379),
  [1934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2514),
  [1936] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pattern_list, 3, 0, 0),
  [1938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3110),
  [1940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2732),
  [1942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(724),
  [1944] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern_seq, 3, 0, 0),
  [1946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(781),
  [1948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2314),
  [1950] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pattern_list, 2, 0, 0),
  [1952] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern_seq, 2, 0, 0),
  [1954] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 4, 0, 34),
  [1956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3034),
  [1958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(883),
  [1960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3032),
  [1962] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 4, 0, 34),
  [1964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3076),
  [1966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(885),
  [1968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3085),
  [1970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1140),
  [1972] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 32),
  [1974] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 32),
  [1976] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1215),
  [1978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2991),
  [1980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1444),
  [1982] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3096),
  [1984] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 52),
  [1986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 52),
  [1988] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0),
  [1990] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(883),
  [1993] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0),
  [1995] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_try_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(885),
  [1998] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 67),
  [2000] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 67), SHIFT_REPEAT(1444),
  [2003] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 67),
  [2005] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 2, 0, 67), SHIFT_REPEAT(1215),
  [2008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_except_clause, 5, 0, 76),
  [2010] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_except_clause, 5, 0, 76),
  [2012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_except_clause, 6, 0, 77),
  [2014] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_except_clause, 6, 0, 77),
  [2016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_statements, 2, 0, 0),
  [2018] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_statements, 2, 0, 0),
  [2020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1537),
  [2022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1556),
  [2024] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_except_clause, 3, 0, 34),
  [2026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_except_clause, 3, 0, 34),
  [2028] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_statements, 3, 0, 0),
  [2030] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_statements, 3, 0, 0),
  [2032] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_except_clause, 7, 0, 78),
  [2034] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_except_clause, 7, 0, 78),
  [2036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_except_clause, 4, 0, 35),
  [2038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_except_clause, 4, 0, 35),
  [2040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_except_clause, 5, 0, 54),
  [2042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_except_clause, 5, 0, 54),
  [2044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__suite, 1, 0, 14),
  [2046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__suite, 1, 0, 14),
  [2048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_statements, 4, 0, 0),
  [2050] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_statements, 4, 0, 0),
  [2052] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1556),
  [2055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2600),
  [2058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, 0, 68),
  [2060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, 0, 68),
  [2062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 4, 0, 33),
  [2064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 4, 0, 33),
  [2066] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 5, 0, 34),
  [2068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 5, 0, 34),
  [2070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_statement_repeat1, 1, 0, 50),
  [2072] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_statement_repeat1, 1, 0, 50),
  [2074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elif_clause, 4, 0, 32),
  [2076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elif_clause, 4, 0, 32),
  [2078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 3, 0, 34),
  [2080] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 3, 0, 34),
  [2082] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__compound_statement, 1, 0, 0),
  [2084] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__compound_statement, 1, 0, 0),
  [2086] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_js_statement, 4, 0, 33),
  [2088] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_js_statement, 4, 0, 33),
  [2090] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_decorated_definition, 2, 0, 11),
  [2092] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_decorated_definition, 2, 0, 11),
  [2094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 5, 0, 51),
  [2096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 51),
  [2098] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_tailed_statement, 2, 0, 2),
  [2100] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_tailed_statement, 2, 0, 2),
  [2102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 7, 0, 73),
  [2104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 7, 0, 73),
  [2106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 5, 0, 53),
  [2108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 5, 0, 53),
  [2110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_async_function_definition, 6, 0, 58),
  [2112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_async_function_definition, 6, 0, 58),
  [2114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(727),
  [2116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [2118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3044),
  [2120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 5, 0, 47),
  [2122] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 5, 0, 47),
  [2124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_tailed_statement, 1, 0, 1),
  [2126] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_tailed_statement, 1, 0, 1),
  [2128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 4, 0, 37),
  [2130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 4, 0, 37),
  [2132] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_finally_clause, 3, 0, 34),
  [2134] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_finally_clause, 3, 0, 34),
  [2136] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 6, 0, 34),
  [2138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 6, 0, 34),
  [2140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_with_statement, 4, 0, 35),
  [2142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_with_statement, 4, 0, 35),
  [2144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 74),
  [2146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 74),
  [2148] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_with_statement, 5, 0, 54),
  [2150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_with_statement, 5, 0, 54),
  [2152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 6, 0, 66),
  [2154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 6, 0, 66),
  [2156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_async_function_definition, 8, 0, 75),
  [2158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_async_function_definition, 8, 0, 75),
  [2160] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 5, 0, 55),
  [2162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 5, 0, 55),
  [2164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_tailed_statement, 3, 0, 15),
  [2166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_tailed_statement, 3, 0, 15),
  [2168] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(102),
  [2171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(975),
  [2173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(976),
  [2175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [2177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(976),
  [2179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3071),
  [2181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [2183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [2185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [2187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [2189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [2191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [2193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [2195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(994),
  [2197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [2199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(995),
  [2201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [2203] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(102),
  [2206] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(102),
  [2209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [2211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1336),
  [2213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1337),
  [2215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1338),
  [2217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [2219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3098),
  [2221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [2223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [2225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1341),
  [2227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1342),
  [2229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1343),
  [2231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [2233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1345),
  [2235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1345),
  [2237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1336),
  [2239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1346),
  [2241] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(129),
  [2244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1263),
  [2246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1264),
  [2248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1265),
  [2250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1264),
  [2252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3079),
  [2254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1266),
  [2256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1267),
  [2258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1268),
  [2260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1269),
  [2262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1270),
  [2264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1271),
  [2266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1272),
  [2268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1272),
  [2270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1263),
  [2272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [2274] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(122),
  [2277] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(971),
  [2280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1245),
  [2282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [2284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1058),
  [2286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1059),
  [2288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [2290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1059),
  [2292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3109),
  [2294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [2296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1062),
  [2298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [2300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [2302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1065),
  [2304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1066),
  [2306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1067),
  [2308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1067),
  [2310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1058),
  [2312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [2314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [2316] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(122),
  [2319] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(136),
  [2322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [2324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [2326] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(530),
  [2329] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(129),
  [2332] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1737),
  [2335] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_concatenated_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2596),
  [2338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [2340] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(136),
  [2343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1763),
  [2345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [2347] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(129),
  [2350] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(975),
  [2353] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3071),
  [2356] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(975),
  [2359] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(995),
  [2362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [2364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1737),
  [2366] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(122),
  [2369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2593),
  [2371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2588),
  [2373] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(136),
  [2376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(919),
  [2378] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(145),
  [2381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1147),
  [2383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1148),
  [2385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1149),
  [2387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1148),
  [2389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3035),
  [2391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [2393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1151),
  [2395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1152),
  [2397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1153),
  [2399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1154),
  [2401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1155),
  [2403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1156),
  [2405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1156),
  [2407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [2409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1157),
  [2411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [2413] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(145),
  [2416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [2418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [2420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [2422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [2424] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(514),
  [2427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(98),
  [2429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [2431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [2433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [2435] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(521),
  [2438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [2440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [2442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [2444] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(984),
  [2447] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(462),
  [2450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1116),
  [2452] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1263),
  [2455] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3079),
  [2458] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1263),
  [2461] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1273),
  [2464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [2466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1276),
  [2468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [2470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(942),
  [2472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(943),
  [2474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [2476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [2478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3070),
  [2480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(945),
  [2482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(946),
  [2484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [2486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1488),
  [2488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(948),
  [2490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [2492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(950),
  [2494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(950),
  [2496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(942),
  [2498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [2500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(983),
  [2502] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1073),
  [2505] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1336),
  [2508] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3098),
  [2511] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1336),
  [2514] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1346),
  [2517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [2519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1347),
  [2521] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1355),
  [2524] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(138),
  [2527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [2529] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(145),
  [2532] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(138),
  [2535] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1058),
  [2538] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3109),
  [2541] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1058),
  [2544] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1068),
  [2547] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(138),
  [2550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(957),
  [2552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(958),
  [2554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [2556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [2558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3099),
  [2560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(960),
  [2562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [2564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [2566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(963),
  [2568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(964),
  [2570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [2572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [2574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(966),
  [2576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [2578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [2582] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(201),
  [2585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1386),
  [2587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1449),
  [2589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(921),
  [2591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1449),
  [2593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3092),
  [2595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [2597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(924),
  [2599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [2601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(974),
  [2603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [2605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [2607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [2609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1080),
  [2611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1386),
  [2613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1081),
  [2615] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(201),
  [2618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [2622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [2624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [2626] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(201),
  [2629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [2631] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(942),
  [2634] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3070),
  [2637] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(942),
  [2640] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(951),
  [2643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1385),
  [2645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [2647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1366),
  [2649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1367),
  [2651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1368),
  [2653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1367),
  [2655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3057),
  [2657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [2659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [2661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [2663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1372),
  [2665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1373),
  [2667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1374),
  [2669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1375),
  [2671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1375),
  [2673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1366),
  [2675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1376),
  [2677] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(957),
  [2680] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3099),
  [2683] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(957),
  [2686] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(967),
  [2689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1359),
  [2691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [2693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [2695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [2697] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(533),
  [2700] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__assign_rhs, 1, 0, 0), SHIFT(185),
  [2703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [2705] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(465),
  [2708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(973),
  [2710] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression_statement, 1, 0, 0), SHIFT(185),
  [2713] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_delete_statement, 2, 0, 0), SHIFT(185),
  [2716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1214),
  [2718] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1147),
  [2721] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3035),
  [2724] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1147),
  [2727] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1157),
  [2730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1407),
  [2732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [2734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1325),
  [2736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [2738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1162),
  [2740] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1114),
  [2743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1190),
  [2745] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(522),
  [2748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1303),
  [2751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [2753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1311),
  [2755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1312),
  [2757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1313),
  [2759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [2761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3072),
  [2763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [2765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [2767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1316),
  [2769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [2771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1318),
  [2773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1319),
  [2775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1320),
  [2777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1320),
  [2779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1311),
  [2781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1321),
  [2783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1126),
  [2785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1127),
  [2787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1126),
  [2789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1130),
  [2791] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1125),
  [2794] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3011),
  [2797] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1125),
  [2800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1135),
  [2803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1404),
  [2805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1125),
  [2807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3011),
  [2809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1128),
  [2811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1129),
  [2813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1131),
  [2815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1132),
  [2817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [2819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1134),
  [2821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1134),
  [2823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1125),
  [2825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1135),
  [2827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1195),
  [2829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [2831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1195),
  [2833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1199),
  [2835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1202),
  [2837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1203),
  [2839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1203),
  [2841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1141),
  [2843] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(524),
  [2846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [2848] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1194),
  [2851] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3084),
  [2854] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1194),
  [2857] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1204),
  [2860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1413),
  [2862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1194),
  [2864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3084),
  [2866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1197),
  [2868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1198),
  [2870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1200),
  [2872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1201),
  [2874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1194),
  [2876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1204),
  [2878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [2880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [2882] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(933),
  [2885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [2887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(752),
  [2889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [2891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2973),
  [2893] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1311),
  [2896] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3072),
  [2899] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1311),
  [2902] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1321),
  [2905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1416),
  [2907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [2909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [2911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [2913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [2915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [2917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [2919] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(504),
  [2922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [2924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [2926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [2928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1299),
  [2930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [2932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [2934] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(528),
  [2937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [2939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [2941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [2943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [2945] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1386),
  [2948] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3092),
  [2951] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1386),
  [2954] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1081),
  [2957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1173),
  [2959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1174),
  [2961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1175),
  [2963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1174),
  [2965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3112),
  [2967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1176),
  [2969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1177),
  [2971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1178),
  [2973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [2975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1180),
  [2977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1181),
  [2979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1182),
  [2981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1182),
  [2983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1173),
  [2985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [2987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [2989] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(520),
  [2992] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1366),
  [2995] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3057),
  [2998] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1366),
  [3001] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1376),
  [3004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1419),
  [3006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [3008] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(525),
  [3011] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(531),
  [3014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [3016] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1381),
  [3019] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1173),
  [3022] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(3112),
  [3025] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1173),
  [3028] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1183),
  [3031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1410),
  [3033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(920),
  [3035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [3037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [3039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [3041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1398),
  [3043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1100),
  [3045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1101),
  [3047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1102),
  [3049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [3051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2990),
  [3053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [3055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [3057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [3059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [3061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [3063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [3065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [3067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1109),
  [3069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [3071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [3073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [3075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [3077] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(541),
  [3080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1217),
  [3082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1218),
  [3084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1220),
  [3086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [3088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2972),
  [3090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1223),
  [3092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1224),
  [3094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1225),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1226),
  [3098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1227),
  [3100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1227),
  [3102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1217),
  [3104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1228),
  [3106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [3108] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__simple_statement, 1, 0, 0), REDUCE(sym__inline_simple_statement, 1, 0, 0),
  [3111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__inline_simple_statement, 1, 0, 0),
  [3113] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__inline_simple_statement, 1, 0, 0),
  [3115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1221),
  [3117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1222),
  [3119] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(523),
  [3122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [3124] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1217),
  [3127] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(2972),
  [3130] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1217),
  [3133] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1228),
  [3136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1285),
  [3138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1286),
  [3140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1287),
  [3142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1286),
  [3144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2985),
  [3146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1288),
  [3148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [3150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1290),
  [3152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1291),
  [3154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1292),
  [3156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1293),
  [3158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [3160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1294),
  [3162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1285),
  [3164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [3166] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1285),
  [3169] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(2985),
  [3172] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1285),
  [3175] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1295),
  [3178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1401),
  [3180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1327),
  [3182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [3184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [3186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1240),
  [3188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__collection_element, 1, 0, 0),
  [3190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1713),
  [3192] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(512),
  [3195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [3197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(645),
  [3199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1743),
  [3201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [3203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [3205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1079),
  [3207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [3209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2177),
  [3211] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1100),
  [3214] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(2990),
  [3217] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1100),
  [3220] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_comparison_operator_repeat1, 2, 0, 19), SHIFT_REPEAT(1110),
  [3223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1666),
  [3225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(764),
  [3227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(831),
  [3229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [3231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [3233] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_in_clause, 4, 0, 61),
  [3235] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__argument, 1, 0, 0), REDUCE(sym__collection_element, 1, 0, 0),
  [3238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [3240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2219),
  [3242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(762),
  [3244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(858),
  [3246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(763),
  [3248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(829),
  [3250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__argument, 1, 0, 0),
  [3252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1659),
  [3254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(765),
  [3256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(797),
  [3258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(886),
  [3260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [3262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2575),
  [3264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [3266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [3268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1076),
  [3270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [3272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [3274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1260),
  [3276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3038),
  [3278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [3280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [3282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2979),
  [3284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [3286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3051),
  [3288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2768),
  [3290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2938),
  [3292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pair, 3, 0, 38),
  [3294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_splat_argument, 2, 0, 0),
  [3296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1211),
  [3298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [3300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1241),
  [3302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_parameter, 3, 0, 46),
  [3304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_with_clause, 1, 0, 4),
  [3306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1236),
  [3308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_default_parameter, 5, 0, 72),
  [3310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_default_parameter, 3, 0, 45),
  [3312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_clause, 2, 0, 0),
  [3314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary_splat_argument, 2, 0, 0),
  [3316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_splat_parameter, 4, 0, 59),
  [3318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary_splat_parameter, 4, 0, 59),
  [3320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_with_clause, 3, 0, 36),
  [3322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(916),
  [3324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice, 3, 0, 0),
  [3326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(913),
  [3328] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice, 2, 0, 0),
  [3330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_keyword_argument, 3, 0, 45),
  [3332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [3334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [3336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [3338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [3340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [3342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [3344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [3346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [3348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [3350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [3352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [3354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [3356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [3358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [3360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1380),
  [3362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1298),
  [3364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [3366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [3368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [3370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [3372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [3374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [3376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [3378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [3380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [3382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [3384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [3386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [3388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [3390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [3392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1246),
  [3394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [3396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [3398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [3400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [3402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [3404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [3406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [3408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [3410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [3412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [3414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [3416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [3418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [3420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [3422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [3424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1057),
  [3426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1324),
  [3428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [3430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [3432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [3434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [3436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [3438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [3440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [3442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [3444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1139),
  [3446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [3448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [3450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [3452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [3454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [3456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [3458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1030),
  [3460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [3462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1280),
  [3464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [3466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [3468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [3470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [3472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1259),
  [3474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [3476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [3478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [3480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice, 5, 0, 0),
  [3482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [3484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [3486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [3488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [3490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [3492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [3494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [3496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice, 4, 0, 0),
  [3498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [3500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [3502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [3504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [3506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [3508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [3510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [3512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2617),
  [3514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [3516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1161),
  [3518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [3520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [3522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [3524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [3526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [3528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [3530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [3532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [3534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [3536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [3538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [3540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [3542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [3544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [3546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [3548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(114),
  [3550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [3552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [3554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [3556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [3558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1012),
  [3560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [3562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [3564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [3566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [3568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1113),
  [3570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [3572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [3574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [3576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [3578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [3580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [3582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [3584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pattern_list_repeat1, 2, 0, 0),
  [3586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 3, 0, 0),
  [3588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 3, 0, 0),
  [3590] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pattern_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1977),
  [3593] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pattern_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1878),
  [3596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__assign_lhs, 1, 0, 0),
  [3598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1498),
  [3600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_seq, 1, 0, 0),
  [3602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_list, 2, 0, 0),
  [3604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1493),
  [3606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_seq, 2, 0, 0),
  [3608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1497),
  [3610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1499),
  [3612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1503),
  [3614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [3616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [3618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [3620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [3622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(848),
  [3624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(851),
  [3626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [3628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [3630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(863),
  [3632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [3634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [3636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(862),
  [3638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [3640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [3642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [3644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [3646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [3648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [3650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [3652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [3654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [3656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [3658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [3660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [3662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [3664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(856),
  [3666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(823),
  [3668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(822),
  [3670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [3672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(786),
  [3674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [3676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(838),
  [3678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [3680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [3682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [3684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(818),
  [3686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(799),
  [3688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(795),
  [3690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [3692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(787),
  [3694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [3696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [3698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(767),
  [3700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2609),
  [3702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2804),
  [3704] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3019),
  [3706] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3033),
  [3708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__parameters, 3, 0, 0),
  [3710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__parameters, 2, 0, 0),
  [3712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3055),
  [3714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3018),
  [3716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3065),
  [3718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1231),
  [3720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3128),
  [3722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3106),
  [3724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3082),
  [3726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(837),
  [3728] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1258),
  [3731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [3733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [3735] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__comprehension_clauses, 2, 0, 0),
  [3737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(927),
  [3739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [3741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1681),
  [3743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [3745] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_dotted_name_repeat1, 2, 0, 0),
  [3747] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dotted_name_repeat1, 2, 0, 0), SHIFT_REPEAT(2988),
  [3750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dotted_name, 1, 0, 0),
  [3752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2988),
  [3754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__comprehension_clauses, 1, 0, 0),
  [3756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [3758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [3760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [3762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2234),
  [3764] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__inline_body_repeat1, 2, 0, 0), SHIFT_REPEAT(553),
  [3767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(926),
  [3769] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_decorated_definition_repeat1, 2, 0, 0),
  [3771] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_decorated_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1231),
  [3774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1248),
  [3776] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__comprehension_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1248),
  [3779] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__comprehension_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1713),
  [3782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__comprehension_clauses_repeat1, 2, 0, 0),
  [3784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(803),
  [3786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [3788] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__comprehension_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(927),
  [3791] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__comprehension_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(1743),
  [3794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dotted_name, 2, 0, 0),
  [3796] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__comprehension_clauses_repeat1, 2, 0, 0), SHIFT_REPEAT(926),
  [3799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [3801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2599),
  [3803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [3805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [3807] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 2, 0, 3),
  [3809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2726),
  [3811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3036),
  [3813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2592),
  [3815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [3817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(883),
  [3819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3032),
  [3821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [3823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [3825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [3827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3085),
  [3829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2595),
  [3831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [3833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [3835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(873),
  [3837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2983),
  [3839] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_f_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1160),
  [3842] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_f_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2595),
  [3845] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_f_string_repeat1, 2, 0, 0),
  [3847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2597),
  [3849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1982),
  [3851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1945),
  [3853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1586),
  [3855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(605),
  [3857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2598),
  [3859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1592),
  [3861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__import_list, 1, 0, 28),
  [3863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2610),
  [3865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3003),
  [3867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_global_statement, 2, 0, 0),
  [3869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3020),
  [3871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 3, 0, 13),
  [3873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_import_statement_repeat1, 2, 0, 3),
  [3875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_import_statement_repeat1, 2, 0, 30),
  [3877] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_import_statement_repeat1, 2, 0, 30), SHIFT_REPEAT(2726),
  [3880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [3882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2981),
  [3884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1450),
  [3886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1451),
  [3888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__parameter, 1, 0, 0),
  [3890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2614),
  [3892] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__import_list, 2, 0, 28),
  [3894] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__import_list, 2, 0, 28),
  [3896] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0),
  [3898] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1138),
  [3901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__flag_list_repeat1, 2, 0, 64),
  [3903] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__flag_list_repeat1, 2, 0, 64), SHIFT_REPEAT(2984),
  [3906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__import_list_repeat1, 2, 0, 3),
  [3908] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(866),
  [3911] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__import_list, 3, 0, 49),
  [3913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__import_list, 3, 0, 49),
  [3915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_decorator, 3, 0, 0),
  [3917] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1260),
  [3920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__import_list_repeat1, 2, 0, 30),
  [3922] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__import_list_repeat1, 2, 0, 30), SHIFT_REPEAT(2875),
  [3925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [3927] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1075),
  [3930] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__import_list, 2, 0, 49),
  [3932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2616),
  [3934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2601),
  [3936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2651),
  [3938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_global_statement, 3, 0, 0),
  [3940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2620),
  [3942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2679),
  [3944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3108),
  [3946] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nonlocal_statement, 2, 0, 0),
  [3948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [3950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3031),
  [3952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nonlocal_statement, 3, 0, 0),
  [3954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [3956] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_global_statement_repeat1, 2, 0, 0),
  [3958] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_global_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(3020),
  [3961] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__flag_list, 1, 0, 26),
  [3963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2792),
  [3965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__flag_list, 2, 0, 48),
  [3967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2710),
  [3969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [3971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2868),
  [3973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2550),
  [3975] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pattern_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1938),
  [3978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [3980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [3982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1683),
  [3984] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__simple_statements_repeat1, 2, 0, 0), SHIFT_REPEAT(259),
  [3987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__simple_statements_repeat1, 2, 0, 0),
  [3989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [3991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1822),
  [3993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2880),
  [3995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3100),
  [3997] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dotted_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3100),
  [4000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2628),
  [4002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [4004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1692),
  [4006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [4008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1702),
  [4010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [4012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dictionary_splat_parameter, 2, 0, 3),
  [4014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2551),
  [4016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__parameters, 2, 0, 0),
  [4018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_aliased_import, 3, 0, 31),
  [4020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2552),
  [4022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__parameters, 1, 0, 0),
  [4024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(836),
  [4026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [4028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [4030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [4032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2240),
  [4034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [4036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1647),
  [4038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2890),
  [4040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [4042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [4044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2268),
  [4046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2803),
  [4048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [4050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(620),
  [4052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_with_statement_repeat1, 2, 0, 0),
  [4054] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_with_statement_repeat1, 2, 0, 0), SHIFT_REPEAT(917),
  [4057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [4059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2223),
  [4061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [4063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(917),
  [4065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [4067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [4069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [4071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2877),
  [4073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2730),
  [4075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2734),
  [4077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2736),
  [4079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2939),
  [4081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [4083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1670),
  [4085] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolation, 6, 0, 70),
  [4087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [4089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [4091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(686),
  [4093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [4095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1671),
  [4097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [4099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1546),
  [4101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [4103] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_repeat1, 2, 0, 0), SHIFT_REPEAT(874),
  [4106] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dictionary_repeat1, 2, 0, 0), SHIFT_REPEAT(872),
  [4109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_dictionary_repeat1, 2, 0, 0),
  [4111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [4113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1572),
  [4115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [4117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [4119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(723),
  [4121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1665),
  [4123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(801),
  [4125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(666),
  [4127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [4129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1535),
  [4131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(806),
  [4133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [4135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [4137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2659),
  [4139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2879),
  [4141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [4143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(757),
  [4145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(640),
  [4147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(807),
  [4149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [4151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(808),
  [4153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [4155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__parameters_repeat1, 2, 0, 0), SHIFT_REPEAT(2554),
  [4158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__parameters_repeat1, 2, 0, 0),
  [4160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(809),
  [4162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [4164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__flag_list_repeat1, 2, 0, 63),
  [4166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [4168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [4170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [4172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2707),
  [4174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__flag_list, 3, 0, 48),
  [4176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__flag_list, 3, 0, 48),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(745),
  [4180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [4182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [4184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1570),
  [4186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2859),
  [4188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [4190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [4192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2863),
  [4194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2796),
  [4196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2853),
  [4198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [4200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(743),
  [4202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2233),
  [4204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(746),
  [4206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [4208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_aliased_import_name, 3, 0, 31),
  [4210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2912),
  [4212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2565),
  [4214] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__flag_list_repeat1, 2, 0, 64), SHIFT_REPEAT(3030),
  [4217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [4219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [4221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [4223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2274),
  [4225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1797),
  [4227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_target, 1, 0, 0),
  [4229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2937),
  [4231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2956),
  [4233] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__import_list_repeat1, 2, 0, 30), SHIFT_REPEAT(2957),
  [4236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [4238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1693),
  [4240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [4242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2270),
  [4244] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(766),
  [4247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0),
  [4249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2893),
  [4251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2895),
  [4253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [4255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2229),
  [4257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2907),
  [4259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2909),
  [4261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2903),
  [4263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [4265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1591),
  [4267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2921),
  [4269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2924),
  [4271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [4273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1663),
  [4275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [4277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2932),
  [4279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2934),
  [4281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2943),
  [4283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2945),
  [4285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolation, 5, 0, 57),
  [4287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2952),
  [4289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2954),
  [4291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [4293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1585),
  [4295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2962),
  [4297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2965),
  [4299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2799),
  [4301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2801),
  [4303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [4305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [4307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2267),
  [4309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2805),
  [4311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2807),
  [4313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolation, 3, 0, 23),
  [4315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2812),
  [4317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2814),
  [4319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2818),
  [4321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2794),
  [4323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [4325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1644),
  [4327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2824),
  [4329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2826),
  [4331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolation, 4, 0, 43),
  [4333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2830),
  [4335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2832),
  [4337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2649),
  [4339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3021),
  [4341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2838),
  [4343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2840),
  [4345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2862),
  [4347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [4349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_splat_parameter, 2, 0, 3),
  [4351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2843),
  [4353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2845),
  [4355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2634),
  [4357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2835),
  [4359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2851),
  [4361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2852),
  [4363] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__flag_list, 2, 0, 26),
  [4365] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__flag_list, 2, 0, 26),
  [4367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [4369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2550),
  [4371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1457),
  [4373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [4375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1427),
  [4377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [4379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1452),
  [4381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [4383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1472),
  [4385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [4387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [4389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [4391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameters, 2, 0, 0),
  [4393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1474),
  [4395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [4397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [4399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [4401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1249),
  [4403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [4405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1403),
  [4407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [4409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1329),
  [4411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [4413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1476),
  [4415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [4417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1477),
  [4419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [4421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1402),
  [4423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [4425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1429),
  [4427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [4429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1478),
  [4431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [4433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1411),
  [4435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [4437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [4439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [4441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1250),
  [4443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [4445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1430),
  [4447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [4449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [4451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [4453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1481),
  [4455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [4457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1393),
  [4459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [4461] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_flag_statement, 6, 0, 62),
  [4463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1482),
  [4465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [4467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1483),
  [4469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [4471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1412),
  [4473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [4475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2645),
  [4477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1362),
  [4479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [4481] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_statement, 1, 0, 0),
  [4483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1484),
  [4485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [4487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [4489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [4491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_flag_statement, 4, 0, 27),
  [4493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1486),
  [4495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [4497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1487),
  [4499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [4501] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_from_statement, 6, 0, 65),
  [4503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1330),
  [4505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [4507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1331),
  [4509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [4511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1432),
  [4513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [4515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1433),
  [4517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [4519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1417),
  [4521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [4523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1406),
  [4525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [4527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1437),
  [4529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [4531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [4533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1408),
  [4535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [4537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1418),
  [4539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [4541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [4543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1247),
  [4545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [4547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [4549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [4551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [4553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [4555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1332),
  [4557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [4559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2872),
  [4561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [4563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [4565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1409),
  [4567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [4569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1390),
  [4571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [4573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameters, 3, 0, 0),
  [4575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [4577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [4579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1471),
  [4581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [4583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1391),
  [4585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [4587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1237),
  [4589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [4591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1438),
  [4593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [4595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1458),
  [4597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [4599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [4601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [4603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1439),
  [4605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [4607] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_from_statement, 4, 0, 29),
  [4609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1440),
  [4611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [4613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [4615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [4617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [4619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [4621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1405),
  [4623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [4625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [4627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [4629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [4631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [4633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1460),
  [4635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [4637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1461),
  [4639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [4641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1415),
  [4643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [4645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1387),
  [4647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [4649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1441),
  [4651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [4653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1442),
  [4655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [4657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1454),
  [4659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [4661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1251),
  [4663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [4665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1399),
  [4667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [4669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1443),
  [4671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [4673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [4675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [4677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1463),
  [4679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [4681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1235),
  [4683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [4685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1400),
  [4687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [4689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [4691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [4693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [4695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1464),
  [4697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [4699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1465),
  [4701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [4703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1453),
  [4705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [4707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3068),
  [4709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2777),
  [4711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1445),
  [4713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [4715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1388),
  [4717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [4719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1163),
  [4721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [4723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1466),
  [4725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [4727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [4729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [4731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1446),
  [4733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [4735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1456),
  [4737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [4739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1165),
  [4741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [4743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1426),
  [4745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [4747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1396),
  [4749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [4751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1468),
  [4753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [4755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1469),
  [4757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [4759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1397),
  [4761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [4763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1210),
  [4765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [4767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [4769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1447),
  [4771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [4773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1470),
  [4775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(395),
  [4777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1279),
  [4779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [4781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1164),
  [4783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [4785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2744),
  [4787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [4789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [4791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2271),
  [4793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1668),
  [4795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1238),
  [4797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1528),
  [4799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2883),
  [4801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2982),
  [4803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [4805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [4807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3000),
  [4809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3010),
  [4811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [4813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2563),
  [4815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1296),
  [4817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2363),
  [4819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2591),
  [4821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1691),
  [4823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [4825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [4827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1675),
  [4829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [4831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [4833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [4835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1515),
  [4837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(892),
  [4839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2637),
  [4841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [4843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [4845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [4847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2724),
  [4849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(691),
  [4851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [4853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2289),
  [4855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2625),
  [4857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [4859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [4861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [4863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [4865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2228),
  [4867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2519),
  [4869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2848),
  [4871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2923),
  [4873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2786),
  [4875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2648),
  [4877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2789),
  [4879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1662),
  [4881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [4883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [4885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1695),
  [4887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2602),
  [4889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2778),
  [4891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [4893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3075),
  [4895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [4897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2654),
  [4899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [4901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [4903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2658),
  [4905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3042),
  [4907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(673),
  [4909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1667),
  [4911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1050),
  [4913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [4915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [4917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2926),
  [4919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2725),
  [4921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [4923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [4925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3067),
  [4927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2265),
  [4929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [4931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [4933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3045),
  [4935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1684),
  [4937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [4939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [4941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2684),
  [4943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2518),
  [4945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1115),
  [4947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2280),
  [4949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3029),
  [4951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2636),
  [4953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2756),
  [4955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3059),
  [4957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [4959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [4961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [4963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1322),
  [4965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2217),
  [4967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2576),
  [4969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [4971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [4973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [4975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1077),
  [4977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1277),
  [4979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [4981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [4983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2676),
  [4985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1229),
  [4987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1205),
  [4989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [4991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2629),
  [4993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2682),
  [4995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [4997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [4999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3083),
  [5001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1274),
  [5003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2745),
  [5005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2170),
  [5007] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [5009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [5011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3024),
  [5013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [5015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [5017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2847),
  [5019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2716),
  [5021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2718),
  [5023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2719),
  [5025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(909),
  [5027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2829),
  [5029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2867),
  [5031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2871),
  [5033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2889),
  [5035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1070),
  [5037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2733),
  [5039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2741),
  [5041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [5043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2748),
  [5045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2752),
  [5047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2755),
  [5049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2758),
  [5051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2761),
  [5053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2763),
  [5055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2767),
  [5057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2770),
  [5059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2772),
  [5061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2776),
  [5063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2780),
  [5065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2784),
  [5067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2788),
  [5069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2791),
  [5071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [5073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3107),
  [5075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__newline = 0,
  ts_external_token__indent = 1,
  ts_external_token__dedent = 2,
  ts_external_token_regex = 3,
  ts_external_token_fstring_start = 4,
  ts_external_token_fstring_content = 5,
  ts_external_token_fstring_end = 6,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__newline] = sym__newline,
  [ts_external_token__indent] = sym__indent,
  [ts_external_token__dedent] = sym__dedent,
  [ts_external_token_regex] = sym_regex,
  [ts_external_token_fstring_start] = sym_fstring_start,
  [ts_external_token_fstring_content] = sym_fstring_content,
  [ts_external_token_fstring_end] = sym_fstring_end,
};

static const bool ts_external_scanner_states[12][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__newline] = true,
    [ts_external_token__indent] = true,
    [ts_external_token__dedent] = true,
    [ts_external_token_regex] = true,
    [ts_external_token_fstring_start] = true,
    [ts_external_token_fstring_content] = true,
    [ts_external_token_fstring_end] = true,
  },
  [2] = {
    [ts_external_token_regex] = true,
    [ts_external_token_fstring_start] = true,
  },
  [3] = {
    [ts_external_token__dedent] = true,
    [ts_external_token_regex] = true,
    [ts_external_token_fstring_start] = true,
  },
  [4] = {
    [ts_external_token__newline] = true,
    [ts_external_token_regex] = true,
    [ts_external_token_fstring_start] = true,
  },
  [5] = {
    [ts_external_token__indent] = true,
    [ts_external_token_regex] = true,
    [ts_external_token_fstring_start] = true,
  },
  [6] = {
    [ts_external_token__newline] = true,
    [ts_external_token__dedent] = true,
    [ts_external_token_regex] = true,
    [ts_external_token_fstring_start] = true,
  },
  [7] = {
    [ts_external_token__newline] = true,
  },
  [8] = {
    [ts_external_token_fstring_start] = true,
  },
  [9] = {
    [ts_external_token__newline] = true,
    [ts_external_token_fstring_start] = true,
  },
  [10] = {
    [ts_external_token_fstring_content] = true,
    [ts_external_token_fstring_end] = true,
  },
  [11] = {
    [ts_external_token_fstring_content] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_rapydscript_external_scanner_create(void);
void tree_sitter_rapydscript_external_scanner_destroy(void *);
bool tree_sitter_rapydscript_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_rapydscript_external_scanner_serialize(void *, char *);
void tree_sitter_rapydscript_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_rapydscript(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_rapydscript_external_scanner_create,
      tree_sitter_rapydscript_external_scanner_destroy,
      tree_sitter_rapydscript_external_scanner_scan,
      tree_sitter_rapydscript_external_scanner_serialize,
      tree_sitter_rapydscript_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "rapydscript",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 1,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
